(window["webpackJsonp"]=window["webpackJsonp"]||[]).push([["chunk-vendors"],{"00b4":function(t,e,n){"use strict";n("ac1f");var r=n("23e7"),i=n("c65b"),o=n("1626"),a=n("825a"),s=n("577e"),l=function(){var t=!1,e=/[ac]/;return e.exec=function(){return t=!0,/./.exec.apply(this,arguments)},!0===e.test("abc")&&t}(),u=/./.test;r({target:"RegExp",proto:!0,forced:!l},{test:function(t){var e=a(this),n=s(t),r=e.exec;if(!o(r))return i(u,e,n);var l=i(r,e,n);return null!==l&&(a(l),!0)}})},"00ee":function(t,e,n){"use strict";var r=n("b622"),i=r("toStringTag"),o={};o[i]="z",t.exports="[object z]"===String(o)},"00fd":function(t,e,n){var r=n("9e69"),i=Object.prototype,o=i.hasOwnProperty,a=i.toString,s=r?r.toStringTag:void 0;function l(t){var e=o.call(t,s),n=t[s];try{t[s]=void 0;var r=!0}catch(l){}var i=a.call(t);return r&&(e?t[s]=n:delete t[s]),i}t.exports=l},"010e":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("uz-latn",{months:"Yanvar_Fevral_Mart_Aprel_May_Iyun_Iyul_Avgust_Sentabr_Oktabr_Noyabr_Dekabr".split("_"),monthsShort:"Yan_Fev_Mar_Apr_May_Iyun_Iyul_Avg_Sen_Okt_Noy_Dek".split("_"),weekdays:"Yakshanba_Dushanba_Seshanba_Chorshanba_Payshanba_Juma_Shanba".split("_"),weekdaysShort:"Yak_Dush_Sesh_Chor_Pay_Jum_Shan".split("_"),weekdaysMin:"Ya_Du_Se_Cho_Pa_Ju_Sha".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"D MMMM YYYY, dddd HH:mm"},calendar:{sameDay:"[Bugun soat] LT [da]",nextDay:"[Ertaga] LT [da]",nextWeek:"dddd [kuni soat] LT [da]",lastDay:"[Kecha soat] LT [da]",lastWeek:"[O'tgan] dddd [kuni soat] LT [da]",sameElse:"L"},relativeTime:{future:"Yaqin %s ichida",past:"Bir necha %s oldin",s:"soniya",ss:"%d soniya",m:"bir daqiqa",mm:"%d daqiqa",h:"bir soat",hh:"%d soat",d:"bir kun",dd:"%d kun",M:"bir oy",MM:"%d oy",y:"bir yil",yy:"%d yil"},week:{dow:1,doy:7}});return e}))},"01b4":function(t,e,n){"use strict";var r=function(){this.head=null,this.tail=null};r.prototype={add:function(t){var e={item:t,next:null},n=this.tail;n?n.next=e:this.head=e,this.tail=e},get:function(){var t=this.head;if(t){var e=this.head=t.next;return null===e&&(this.tail=null),t.item}}},t.exports=r},"020f":function(t,e,n){var r=n("242e"),i=n("1304");function o(t,e){return t&&r(t,i(e))}t.exports=o},"02fb":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ml",{months:"ജനുവരി_ഫെബ്രുവരി_മാർച്ച്_ഏപ്രിൽ_മേയ്_ജൂൺ_ജൂലൈ_ഓഗസ്റ്റ്_സെപ്റ്റംബർ_ഒക്ടോബർ_നവംബർ_ഡിസംബർ".split("_"),monthsShort:"ജനു._ഫെബ്രു._മാർ._ഏപ്രി._മേയ്_ജൂൺ_ജൂലൈ._ഓഗ._സെപ്റ്റ._ഒക്ടോ._നവം._ഡിസം.".split("_"),monthsParseExact:!0,weekdays:"ഞായറാഴ്ച_തിങ്കളാഴ്ച_ചൊവ്വാഴ്ച_ബുധനാഴ്ച_വ്യാഴാഴ്ച_വെള്ളിയാഴ്ച_ശനിയാഴ്ച".split("_"),weekdaysShort:"ഞായർ_തിങ്കൾ_ചൊവ്വ_ബുധൻ_വ്യാഴം_വെള്ളി_ശനി".split("_"),weekdaysMin:"ഞാ_തി_ചൊ_ബു_വ്യാ_വെ_ശ".split("_"),longDateFormat:{LT:"A h:mm -നു",LTS:"A h:mm:ss -നു",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm -നു",LLLL:"dddd, D MMMM YYYY, A h:mm -നു"},calendar:{sameDay:"[ഇന്ന്] LT",nextDay:"[നാളെ] LT",nextWeek:"dddd, LT",lastDay:"[ഇന്നലെ] LT",lastWeek:"[കഴിഞ്ഞ] dddd, LT",sameElse:"L"},relativeTime:{future:"%s കഴിഞ്ഞ്",past:"%s മുൻപ്",s:"അൽപ നിമിഷങ്ങൾ",ss:"%d സെക്കൻഡ്",m:"ഒരു മിനിറ്റ്",mm:"%d മിനിറ്റ്",h:"ഒരു മണിക്കൂർ",hh:"%d മണിക്കൂർ",d:"ഒരു ദിവസം",dd:"%d ദിവസം",M:"ഒരു മാസം",MM:"%d മാസം",y:"ഒരു വർഷം",yy:"%d വർഷം"},meridiemParse:/രാത്രി|രാവിലെ|ഉച്ച കഴിഞ്ഞ്|വൈകുന്നേരം|രാത്രി/i,meridiemHour:function(t,e){return 12===t&&(t=0),"രാത്രി"===e&&t>=4||"ഉച്ച കഴിഞ്ഞ്"===e||"വൈകുന്നേരം"===e?t+12:t},meridiem:function(t,e,n){return t<4?"രാത്രി":t<12?"രാവിലെ":t<17?"ഉച്ച കഴിഞ്ഞ്":t<20?"വൈകുന്നേരം":"രാത്രി"}});return e}))},"0366":function(t,e,n){"use strict";var r=n("4625"),i=n("59ed"),o=n("40d5"),a=r(r.bind);t.exports=function(t,e){return i(t),void 0===e?t:o?a(t,e):function(){return t.apply(e,arguments)}}},"03dd":function(t,e,n){var r=n("eac5"),i=n("57a5"),o=Object.prototype,a=o.hasOwnProperty;function s(t){if(!r(t))return i(t);var e=[];for(var n in Object(t))a.call(t,n)&&"constructor"!=n&&e.push(n);return e}t.exports=s},"03ec":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("cv",{months:"кӑрлач_нарӑс_пуш_ака_май_ҫӗртме_утӑ_ҫурла_авӑн_юпа_чӳк_раштав".split("_"),monthsShort:"кӑр_нар_пуш_ака_май_ҫӗр_утӑ_ҫур_авн_юпа_чӳк_раш".split("_"),weekdays:"вырсарникун_тунтикун_ытларикун_юнкун_кӗҫнерникун_эрнекун_шӑматкун".split("_"),weekdaysShort:"выр_тун_ытл_юн_кӗҫ_эрн_шӑм".split("_"),weekdaysMin:"вр_тн_ыт_юн_кҫ_эр_шм".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD-MM-YYYY",LL:"YYYY [ҫулхи] MMMM [уйӑхӗн] D[-мӗшӗ]",LLL:"YYYY [ҫулхи] MMMM [уйӑхӗн] D[-мӗшӗ], HH:mm",LLLL:"dddd, YYYY [ҫулхи] MMMM [уйӑхӗн] D[-мӗшӗ], HH:mm"},calendar:{sameDay:"[Паян] LT [сехетре]",nextDay:"[Ыран] LT [сехетре]",lastDay:"[Ӗнер] LT [сехетре]",nextWeek:"[Ҫитес] dddd LT [сехетре]",lastWeek:"[Иртнӗ] dddd LT [сехетре]",sameElse:"L"},relativeTime:{future:function(t){var e=/сехет$/i.exec(t)?"рен":/ҫул$/i.exec(t)?"тан":"ран";return t+e},past:"%s каялла",s:"пӗр-ик ҫеккунт",ss:"%d ҫеккунт",m:"пӗр минут",mm:"%d минут",h:"пӗр сехет",hh:"%d сехет",d:"пӗр кун",dd:"%d кун",M:"пӗр уйӑх",MM:"%d уйӑх",y:"пӗр ҫул",yy:"%d ҫул"},dayOfMonthOrdinalParse:/\d{1,2}-мӗш/,ordinal:"%d-мӗш",week:{dow:1,doy:7}});return e}))},"0402":function(t,e,n){"use strict";var r=n("23e7"),i=n("da84"),o=n("2cf4").set,a=n("20cc"),s=i.setImmediate?a(o,!1):o;r({global:!0,bind:!0,enumerable:!0,forced:i.setImmediate!==s},{setImmediate:s})},"04d1":function(t,e,n){"use strict";var r=n("342f"),i=r.match(/firefox\/(\d+)/i);t.exports=!!i&&+i[1]},"04f8":function(t,e,n){"use strict";var r=n("2d00"),i=n("d039"),o=n("da84"),a=o.String;t.exports=!!Object.getOwnPropertySymbols&&!i((function(){var t=Symbol("symbol detection");return!a(t)||!(Object(t)instanceof Symbol)||!Symbol.sham&&r&&r<41}))},"0558":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t){return t%100===11||t%10!==1}function n(t,n,r,i){var o=t+" ";switch(r){case"s":return n||i?"nokkrar sekúndur":"nokkrum sekúndum";case"ss":return e(t)?o+(n||i?"sekúndur":"sekúndum"):o+"sekúnda";case"m":return n?"mínúta":"mínútu";case"mm":return e(t)?o+(n||i?"mínútur":"mínútum"):n?o+"mínúta":o+"mínútu";case"hh":return e(t)?o+(n||i?"klukkustundir":"klukkustundum"):o+"klukkustund";case"d":return n?"dagur":i?"dag":"degi";case"dd":return e(t)?n?o+"dagar":o+(i?"daga":"dögum"):n?o+"dagur":o+(i?"dag":"degi");case"M":return n?"mánuður":i?"mánuð":"mánuði";case"MM":return e(t)?n?o+"mánuðir":o+(i?"mánuði":"mánuðum"):n?o+"mánuður":o+(i?"mánuð":"mánuði");case"y":return n||i?"ár":"ári";case"yy":return e(t)?o+(n||i?"ár":"árum"):o+(n||i?"ár":"ári")}}var r=t.defineLocale("is",{months:"janúar_febrúar_mars_apríl_maí_júní_júlí_ágúst_september_október_nóvember_desember".split("_"),monthsShort:"jan_feb_mar_apr_maí_jún_júl_ágú_sep_okt_nóv_des".split("_"),weekdays:"sunnudagur_mánudagur_þriðjudagur_miðvikudagur_fimmtudagur_föstudagur_laugardagur".split("_"),weekdaysShort:"sun_mán_þri_mið_fim_fös_lau".split("_"),weekdaysMin:"Su_Má_Þr_Mi_Fi_Fö_La".split("_"),longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY [kl.] H:mm",LLLL:"dddd, D. MMMM YYYY [kl.] H:mm"},calendar:{sameDay:"[í dag kl.] LT",nextDay:"[á morgun kl.] LT",nextWeek:"dddd [kl.] LT",lastDay:"[í gær kl.] LT",lastWeek:"[síðasta] dddd [kl.] LT",sameElse:"L"},relativeTime:{future:"eftir %s",past:"fyrir %s síðan",s:n,ss:n,m:n,mm:n,h:"klukkustund",hh:n,d:n,dd:n,M:n,MM:n,y:n,yy:n},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return r}))},"057f":function(t,e,n){"use strict";var r=n("c6b6"),i=n("fc6a"),o=n("241c").f,a=n("f36a"),s="object"==typeof window&&window&&Object.getOwnPropertyNames?Object.getOwnPropertyNames(window):[],l=function(t){try{return o(t)}catch(e){return a(s)}};t.exports.f=function(t){return s&&"Window"===r(t)?l(t):o(i(t))}},"0621":function(t,e,n){var r=n("9e69"),i=n("d370"),o=n("6747"),a=r?r.isConcatSpreadable:void 0;function s(t){return o(t)||i(t)||!!(a&&t&&t[a])}t.exports=s},"06c5":function(t,e,n){"use strict";n.d(e,"a",(function(){return i}));n("a630"),n("fb6a"),n("b0c0"),n("d3b7"),n("ac1f"),n("00b4"),n("25f0"),n("3ca3");var r=n("6b75");function i(t,e){if(t){if("string"===typeof t)return Object(r["a"])(t,e);var n=Object.prototype.toString.call(t).slice(8,-1);return"Object"===n&&t.constructor&&(n=t.constructor.name),"Map"===n||"Set"===n?Array.from(t):"Arguments"===n||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)?Object(r["a"])(t,e):void 0}}},"06cf":function(t,e,n){"use strict";var r=n("83ab"),i=n("c65b"),o=n("d1e7"),a=n("5c6c"),s=n("fc6a"),l=n("a04b"),u=n("1a2d"),c=n("0cfb"),d=Object.getOwnPropertyDescriptor;e.f=r?d:function(t,e){if(t=s(t),e=l(e),c)try{return d(t,e)}catch(n){}if(u(t,e))return a(!i(o.f,t,e),t[e])}},"0721":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("fo",{months:"januar_februar_mars_apríl_mai_juni_juli_august_september_oktober_november_desember".split("_"),monthsShort:"jan_feb_mar_apr_mai_jun_jul_aug_sep_okt_nov_des".split("_"),weekdays:"sunnudagur_mánadagur_týsdagur_mikudagur_hósdagur_fríggjadagur_leygardagur".split("_"),weekdaysShort:"sun_mán_týs_mik_hós_frí_ley".split("_"),weekdaysMin:"su_má_tý_mi_hó_fr_le".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D. MMMM, YYYY HH:mm"},calendar:{sameDay:"[Í dag kl.] LT",nextDay:"[Í morgin kl.] LT",nextWeek:"dddd [kl.] LT",lastDay:"[Í gjár kl.] LT",lastWeek:"[síðstu] dddd [kl] LT",sameElse:"L"},relativeTime:{future:"um %s",past:"%s síðani",s:"fá sekund",ss:"%d sekundir",m:"ein minuttur",mm:"%d minuttir",h:"ein tími",hh:"%d tímar",d:"ein dagur",dd:"%d dagar",M:"ein mánaður",MM:"%d mánaðir",y:"eitt ár",yy:"%d ár"},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return e}))},"079e":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ja",{eras:[{since:"2019-05-01",offset:1,name:"令和",narrow:"㋿",abbr:"R"},{since:"1989-01-08",until:"2019-04-30",offset:1,name:"平成",narrow:"㍻",abbr:"H"},{since:"1926-12-25",until:"1989-01-07",offset:1,name:"昭和",narrow:"㍼",abbr:"S"},{since:"1912-07-30",until:"1926-12-24",offset:1,name:"大正",narrow:"㍽",abbr:"T"},{since:"1873-01-01",until:"1912-07-29",offset:6,name:"明治",narrow:"㍾",abbr:"M"},{since:"0001-01-01",until:"1873-12-31",offset:1,name:"西暦",narrow:"AD",abbr:"AD"},{since:"0000-12-31",until:-1/0,offset:1,name:"紀元前",narrow:"BC",abbr:"BC"}],eraYearOrdinalRegex:/(元|\d+)年/,eraYearOrdinalParse:function(t,e){return"元"===e[1]?1:parseInt(e[1]||t,10)},months:"1月_2月_3月_4月_5月_6月_7月_8月_9月_10月_11月_12月".split("_"),monthsShort:"1月_2月_3月_4月_5月_6月_7月_8月_9月_10月_11月_12月".split("_"),weekdays:"日曜日_月曜日_火曜日_水曜日_木曜日_金曜日_土曜日".split("_"),weekdaysShort:"日_月_火_水_木_金_土".split("_"),weekdaysMin:"日_月_火_水_木_金_土".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY/MM/DD",LL:"YYYY年M月D日",LLL:"YYYY年M月D日 HH:mm",LLLL:"YYYY年M月D日 dddd HH:mm",l:"YYYY/MM/DD",ll:"YYYY年M月D日",lll:"YYYY年M月D日 HH:mm",llll:"YYYY年M月D日(ddd) HH:mm"},meridiemParse:/午前|午後/i,isPM:function(t){return"午後"===t},meridiem:function(t,e,n){return t<12?"午前":"午後"},calendar:{sameDay:"[今日] LT",nextDay:"[明日] LT",nextWeek:function(t){return t.week()!==this.week()?"[来週]dddd LT":"dddd LT"},lastDay:"[昨日] LT",lastWeek:function(t){return this.week()!==t.week()?"[先週]dddd LT":"dddd LT"},sameElse:"L"},dayOfMonthOrdinalParse:/\d{1,2}日/,ordinal:function(t,e){switch(e){case"y":return 1===t?"元年":t+"年";case"d":case"D":case"DDD":return t+"日";default:return t}},relativeTime:{future:"%s後",past:"%s前",s:"数秒",ss:"%d秒",m:"1分",mm:"%d分",h:"1時間",hh:"%d時間",d:"1日",dd:"%d日",M:"1ヶ月",MM:"%dヶ月",y:"1年",yy:"%d年"}});return e}))},"07ac":function(t,e,n){"use strict";var r=n("23e7"),i=n("6f53").values;r({target:"Object",stat:!0},{values:function(t){return i(t)}})},"07c7":function(t,e){function n(){return!1}t.exports=n},"07fa":function(t,e,n){"use strict";var r=n("50c4");t.exports=function(t){return r(t.length)}},"083a":function(t,e,n){"use strict";var r=n("0d51"),i=TypeError;t.exports=function(t,e){if(!delete t[e])throw new i("Cannot delete property "+r(e)+" of "+r(t))}},"087d":function(t,e){function n(t,e){var n=-1,r=e.length,i=t.length;while(++n<r)t[i+n]=e[n];return t}t.exports=n},"08cc":function(t,e,n){var r=n("1a8c");function i(t){return t===t&&!r(t)}t.exports=i},"0977":function(module,__webpack_exports__,__webpack_require__){"use strict";(function(global){var version="1.3.1";function _typeof(t){return _typeof="function"===typeof Symbol&&"symbol"===typeof Symbol.iterator?function(t){return typeof t}:function(t){return t&&"function"===typeof Symbol&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},_typeof(t)}function _classCallCheck(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function _defineProperties(t,e){for(var n=0;n<e.length;n++){var r=e[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(t,r.key,r)}}function _createClass(t,e,n){return e&&_defineProperties(t.prototype,e),n&&_defineProperties(t,n),t}function _defineProperty(t,e,n){return e in t?Object.defineProperty(t,e,{value:n,enumerable:!0,configurable:!0,writable:!0}):t[e]=n,t}function _inherits(t,e){if("function"!==typeof e&&null!==e)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),e&&_setPrototypeOf(t,e)}function _getPrototypeOf(t){return _getPrototypeOf=Object.setPrototypeOf?Object.getPrototypeOf:function(t){return t.__proto__||Object.getPrototypeOf(t)},_getPrototypeOf(t)}function _setPrototypeOf(t,e){return _setPrototypeOf=Object.setPrototypeOf||function(t,e){return t.__proto__=e,t},_setPrototypeOf(t,e)}function _isNativeReflectConstruct(){if("undefined"===typeof Reflect||!Reflect.construct)return!1;if(Reflect.construct.sham)return!1;if("function"===typeof Proxy)return!0;try{return Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],(function(){}))),!0}catch(t){return!1}}function _assertThisInitialized(t){if(void 0===t)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function _possibleConstructorReturn(t,e){if(e&&("object"===typeof e||"function"===typeof e))return e;if(void 0!==e)throw new TypeError("Derived constructors may only return object or undefined");return _assertThisInitialized(t)}function _createSuper(t){var e=_isNativeReflectConstruct();return function(){var n,r=_getPrototypeOf(t);if(e){var i=_getPrototypeOf(this).constructor;n=Reflect.construct(r,arguments,i)}else n=r.apply(this,arguments);return _possibleConstructorReturn(this,n)}}function _superPropBase(t,e){while(!Object.prototype.hasOwnProperty.call(t,e))if(t=_getPrototypeOf(t),null===t)break;return t}function _get(){return _get="undefined"!==typeof Reflect&&Reflect.get?Reflect.get:function(t,e,n){var r=_superPropBase(t,e);if(r){var i=Object.getOwnPropertyDescriptor(r,e);return i.get?i.get.call(arguments.length<3?t:n):i.value}},_get.apply(this,arguments)}function _slicedToArray(t,e){return _arrayWithHoles(t)||_iterableToArrayLimit(t,e)||_unsupportedIterableToArray(t,e)||_nonIterableRest()}function _toConsumableArray(t){return _arrayWithoutHoles(t)||_iterableToArray(t)||_unsupportedIterableToArray(t)||_nonIterableSpread()}function _arrayWithoutHoles(t){if(Array.isArray(t))return _arrayLikeToArray(t)}function _arrayWithHoles(t){if(Array.isArray(t))return t}function _iterableToArray(t){if("undefined"!==typeof Symbol&&null!=t[Symbol.iterator]||null!=t["@@iterator"])return Array.from(t)}function _iterableToArrayLimit(t,e){var n=null==t?null:"undefined"!==typeof Symbol&&t[Symbol.iterator]||t["@@iterator"];if(null!=n){var r,i,o=[],a=!0,s=!1;try{for(n=n.call(t);!(a=(r=n.next()).done);a=!0)if(o.push(r.value),e&&o.length===e)break}catch(l){s=!0,i=l}finally{try{a||null==n["return"]||n["return"]()}finally{if(s)throw i}}return o}}function _unsupportedIterableToArray(t,e){if(t){if("string"===typeof t)return _arrayLikeToArray(t,e);var n=Object.prototype.toString.call(t).slice(8,-1);return"Object"===n&&t.constructor&&(n=t.constructor.name),"Map"===n||"Set"===n?Array.from(t):"Arguments"===n||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)?_arrayLikeToArray(t,e):void 0}}function _arrayLikeToArray(t,e){(null==e||e>t.length)&&(e=t.length);for(var n=0,r=new Array(e);n<e;n++)r[n]=t[n];return r}function _nonIterableSpread(){throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.")}function _nonIterableRest(){throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.")}function _createForOfIteratorHelper(t,e){var n="undefined"!==typeof Symbol&&t[Symbol.iterator]||t["@@iterator"];if(!n){if(Array.isArray(t)||(n=_unsupportedIterableToArray(t))||e&&t&&"number"===typeof t.length){n&&(t=n);var r=0,i=function(){};return{s:i,n:function(){return r>=t.length?{done:!0}:{done:!1,value:t[r++]}},e:function(t){throw t},f:i}}throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.")}var o,a=!0,s=!1;return{s:function(){n=n.call(t)},n:function(){var t=n.next();return a=t.done,t},e:function(t){s=!0,o=t},f:function(){try{a||null==n.return||n.return()}finally{if(s)throw o}}}}var commonjsGlobal="undefined"!==typeof globalThis?globalThis:"undefined"!==typeof window?window:"undefined"!==typeof global?global:"undefined"!==typeof self?self:{};function createCommonjsModule(t,e){return e={exports:{}},t(e,e.exports),e.exports}var check=function(t){return t&&t.Math==Math&&t},global_1=check("object"==typeof globalThis&&globalThis)||check("object"==typeof window&&window)||check("object"==typeof self&&self)||check("object"==typeof commonjsGlobal&&commonjsGlobal)||function(){return this}()||Function("return this")(),fails=function(t){try{return!!t()}catch(e){return!0}},descriptors=!fails((function(){return 7!=Object.defineProperty({},1,{get:function(){return 7}})[1]})),call=Function.prototype.call,functionCall=call.bind?call.bind(call):function(){return call.apply(call,arguments)},$propertyIsEnumerable={}.propertyIsEnumerable,getOwnPropertyDescriptor=Object.getOwnPropertyDescriptor,NASHORN_BUG=getOwnPropertyDescriptor&&!$propertyIsEnumerable.call({1:2},1),f=NASHORN_BUG?function(t){var e=getOwnPropertyDescriptor(this,t);return!!e&&e.enumerable}:$propertyIsEnumerable,objectPropertyIsEnumerable={f:f},createPropertyDescriptor=function(t,e){return{enumerable:!(1&t),configurable:!(2&t),writable:!(4&t),value:e}},FunctionPrototype=Function.prototype,bind=FunctionPrototype.bind,call$1=FunctionPrototype.call,callBind=bind&&bind.bind(call$1),functionUncurryThis=bind?function(t){return t&&callBind(call$1,t)}:function(t){return t&&function(){return call$1.apply(t,arguments)}},toString=functionUncurryThis({}.toString),stringSlice=functionUncurryThis("".slice),classofRaw=function(t){return stringSlice(toString(t),8,-1)},Object$1=global_1.Object,split=functionUncurryThis("".split),indexedObject=fails((function(){return!Object$1("z").propertyIsEnumerable(0)}))?function(t){return"String"==classofRaw(t)?split(t,""):Object$1(t)}:Object$1,TypeError$1=global_1.TypeError,requireObjectCoercible=function(t){if(void 0==t)throw TypeError$1("Can't call method on "+t);return t},toIndexedObject=function(t){return indexedObject(requireObjectCoercible(t))},isCallable=function(t){return"function"==typeof t},isObject=function(t){return"object"==typeof t?null!==t:isCallable(t)},aFunction=function(t){return isCallable(t)?t:void 0},getBuiltIn=function(t,e){return arguments.length<2?aFunction(global_1[t]):global_1[t]&&global_1[t][e]},objectIsPrototypeOf=functionUncurryThis({}.isPrototypeOf),engineUserAgent=getBuiltIn("navigator","userAgent")||"",process=global_1.process,Deno=global_1.Deno,versions=process&&process.versions||Deno&&Deno.version,v8=versions&&versions.v8,match,version$1;v8&&(match=v8.split("."),version$1=match[0]>0&&match[0]<4?1:+(match[0]+match[1])),!version$1&&engineUserAgent&&(match=engineUserAgent.match(/Edge\/(\d+)/),(!match||match[1]>=74)&&(match=engineUserAgent.match(/Chrome\/(\d+)/),match&&(version$1=+match[1])));var engineV8Version=version$1,nativeSymbol=!!Object.getOwnPropertySymbols&&!fails((function(){var t=Symbol();return!String(t)||!(Object(t)instanceof Symbol)||!Symbol.sham&&engineV8Version&&engineV8Version<41})),useSymbolAsUid=nativeSymbol&&!Symbol.sham&&"symbol"==typeof Symbol.iterator,Object$2=global_1.Object,isSymbol=useSymbolAsUid?function(t){return"symbol"==typeof t}:function(t){var e=getBuiltIn("Symbol");return isCallable(e)&&objectIsPrototypeOf(e.prototype,Object$2(t))},String$1=global_1.String,tryToString=function(t){try{return String$1(t)}catch(e){return"Object"}},TypeError$2=global_1.TypeError,aCallable=function(t){if(isCallable(t))return t;throw TypeError$2(tryToString(t)+" is not a function")},getMethod=function(t,e){var n=t[e];return null==n?void 0:aCallable(n)},TypeError$3=global_1.TypeError,ordinaryToPrimitive=function(t,e){var n,r;if("string"===e&&isCallable(n=t.toString)&&!isObject(r=functionCall(n,t)))return r;if(isCallable(n=t.valueOf)&&!isObject(r=functionCall(n,t)))return r;if("string"!==e&&isCallable(n=t.toString)&&!isObject(r=functionCall(n,t)))return r;throw TypeError$3("Can't convert object to primitive value")},defineProperty=Object.defineProperty,setGlobal=function(t,e){try{defineProperty(global_1,t,{value:e,configurable:!0,writable:!0})}catch(n){global_1[t]=e}return e},SHARED="__core-js_shared__",store=global_1[SHARED]||setGlobal(SHARED,{}),sharedStore=store,shared=createCommonjsModule((function(t){(t.exports=function(t,e){return sharedStore[t]||(sharedStore[t]=void 0!==e?e:{})})("versions",[]).push({version:"3.19.3",mode:"global",copyright:"© 2021 Denis Pushkarev (zloirock.ru)"})})),Object$3=global_1.Object,toObject=function(t){return Object$3(requireObjectCoercible(t))},hasOwnProperty=functionUncurryThis({}.hasOwnProperty),hasOwnProperty_1=Object.hasOwn||function(t,e){return hasOwnProperty(toObject(t),e)},id=0,postfix=Math.random(),toString$1=functionUncurryThis(1..toString),uid=function(t){return"Symbol("+(void 0===t?"":t)+")_"+toString$1(++id+postfix,36)},WellKnownSymbolsStore=shared("wks"),Symbol$1=global_1.Symbol,symbolFor=Symbol$1&&Symbol$1["for"],createWellKnownSymbol=useSymbolAsUid?Symbol$1:Symbol$1&&Symbol$1.withoutSetter||uid,wellKnownSymbol=function(t){if(!hasOwnProperty_1(WellKnownSymbolsStore,t)||!nativeSymbol&&"string"!=typeof WellKnownSymbolsStore[t]){var e="Symbol."+t;nativeSymbol&&hasOwnProperty_1(Symbol$1,t)?WellKnownSymbolsStore[t]=Symbol$1[t]:WellKnownSymbolsStore[t]=useSymbolAsUid&&symbolFor?symbolFor(e):createWellKnownSymbol(e)}return WellKnownSymbolsStore[t]},TypeError$4=global_1.TypeError,TO_PRIMITIVE=wellKnownSymbol("toPrimitive"),toPrimitive=function(t,e){if(!isObject(t)||isSymbol(t))return t;var n,r=getMethod(t,TO_PRIMITIVE);if(r){if(void 0===e&&(e="default"),n=functionCall(r,t,e),!isObject(n)||isSymbol(n))return n;throw TypeError$4("Can't convert object to primitive value")}return void 0===e&&(e="number"),ordinaryToPrimitive(t,e)},toPropertyKey=function(t){var e=toPrimitive(t,"string");return isSymbol(e)?e:e+""},document=global_1.document,EXISTS=isObject(document)&&isObject(document.createElement),documentCreateElement=function(t){return EXISTS?document.createElement(t):{}},ie8DomDefine=!descriptors&&!fails((function(){return 7!=Object.defineProperty(documentCreateElement("div"),"a",{get:function(){return 7}}).a})),$getOwnPropertyDescriptor=Object.getOwnPropertyDescriptor,f$1=descriptors?$getOwnPropertyDescriptor:function(t,e){if(t=toIndexedObject(t),e=toPropertyKey(e),ie8DomDefine)try{return $getOwnPropertyDescriptor(t,e)}catch(n){}if(hasOwnProperty_1(t,e))return createPropertyDescriptor(!functionCall(objectPropertyIsEnumerable.f,t,e),t[e])},objectGetOwnPropertyDescriptor={f:f$1},String$2=global_1.String,TypeError$5=global_1.TypeError,anObject=function(t){if(isObject(t))return t;throw TypeError$5(String$2(t)+" is not an object")},TypeError$6=global_1.TypeError,$defineProperty=Object.defineProperty,f$2=descriptors?$defineProperty:function(t,e,n){if(anObject(t),e=toPropertyKey(e),anObject(n),ie8DomDefine)try{return $defineProperty(t,e,n)}catch(r){}if("get"in n||"set"in n)throw TypeError$6("Accessors not supported");return"value"in n&&(t[e]=n.value),t},objectDefineProperty={f:f$2},createNonEnumerableProperty=descriptors?function(t,e,n){return objectDefineProperty.f(t,e,createPropertyDescriptor(1,n))}:function(t,e,n){return t[e]=n,t},functionToString=functionUncurryThis(Function.toString);isCallable(sharedStore.inspectSource)||(sharedStore.inspectSource=function(t){return functionToString(t)});var inspectSource=sharedStore.inspectSource,WeakMap=global_1.WeakMap,nativeWeakMap=isCallable(WeakMap)&&/native code/.test(inspectSource(WeakMap)),keys=shared("keys"),sharedKey=function(t){return keys[t]||(keys[t]=uid(t))},hiddenKeys={},OBJECT_ALREADY_INITIALIZED="Object already initialized",TypeError$7=global_1.TypeError,WeakMap$1=global_1.WeakMap,set,get,has,enforce=function(t){return has(t)?get(t):set(t,{})},getterFor=function(t){return function(e){var n;if(!isObject(e)||(n=get(e)).type!==t)throw TypeError$7("Incompatible receiver, "+t+" required");return n}};if(nativeWeakMap||sharedStore.state){var store$1=sharedStore.state||(sharedStore.state=new WeakMap$1),wmget=functionUncurryThis(store$1.get),wmhas=functionUncurryThis(store$1.has),wmset=functionUncurryThis(store$1.set);set=function(t,e){if(wmhas(store$1,t))throw new TypeError$7(OBJECT_ALREADY_INITIALIZED);return e.facade=t,wmset(store$1,t,e),e},get=function(t){return wmget(store$1,t)||{}},has=function(t){return wmhas(store$1,t)}}else{var STATE=sharedKey("state");hiddenKeys[STATE]=!0,set=function(t,e){if(hasOwnProperty_1(t,STATE))throw new TypeError$7(OBJECT_ALREADY_INITIALIZED);return e.facade=t,createNonEnumerableProperty(t,STATE,e),e},get=function(t){return hasOwnProperty_1(t,STATE)?t[STATE]:{}},has=function(t){return hasOwnProperty_1(t,STATE)}}var internalState={set:set,get:get,has:has,enforce:enforce,getterFor:getterFor},FunctionPrototype$1=Function.prototype,getDescriptor=descriptors&&Object.getOwnPropertyDescriptor,EXISTS$1=hasOwnProperty_1(FunctionPrototype$1,"name"),PROPER=EXISTS$1&&"something"===function(){}.name,CONFIGURABLE=EXISTS$1&&(!descriptors||descriptors&&getDescriptor(FunctionPrototype$1,"name").configurable),functionName={EXISTS:EXISTS$1,PROPER:PROPER,CONFIGURABLE:CONFIGURABLE},redefine=createCommonjsModule((function(t){var e=functionName.CONFIGURABLE,n=internalState.get,r=internalState.enforce,i=String(String).split("String");(t.exports=function(t,n,o,a){var s,l=!!a&&!!a.unsafe,u=!!a&&!!a.enumerable,c=!!a&&!!a.noTargetGet,d=a&&void 0!==a.name?a.name:n;isCallable(o)&&("Symbol("===String(d).slice(0,7)&&(d="["+String(d).replace(/^Symbol\(([^)]*)\)/,"$1")+"]"),(!hasOwnProperty_1(o,"name")||e&&o.name!==d)&&createNonEnumerableProperty(o,"name",d),s=r(o),s.source||(s.source=i.join("string"==typeof d?d:""))),t!==global_1?(l?!c&&t[n]&&(u=!0):delete t[n],u?t[n]=o:createNonEnumerableProperty(t,n,o)):u?t[n]=o:setGlobal(n,o)})(Function.prototype,"toString",(function(){return isCallable(this)&&n(this).source||inspectSource(this)}))})),ceil=Math.ceil,floor=Math.floor,toIntegerOrInfinity=function(t){var e=+t;return e!==e||0===e?0:(e>0?floor:ceil)(e)},max=Math.max,min=Math.min,toAbsoluteIndex=function(t,e){var n=toIntegerOrInfinity(t);return n<0?max(n+e,0):min(n,e)},min$1=Math.min,toLength=function(t){return t>0?min$1(toIntegerOrInfinity(t),9007199254740991):0},lengthOfArrayLike=function(t){return toLength(t.length)},createMethod=function(t){return function(e,n,r){var i,o=toIndexedObject(e),a=lengthOfArrayLike(o),s=toAbsoluteIndex(r,a);if(t&&n!=n){while(a>s)if(i=o[s++],i!=i)return!0}else for(;a>s;s++)if((t||s in o)&&o[s]===n)return t||s||0;return!t&&-1}},arrayIncludes={includes:createMethod(!0),indexOf:createMethod(!1)},indexOf=arrayIncludes.indexOf,push=functionUncurryThis([].push),objectKeysInternal=function(t,e){var n,r=toIndexedObject(t),i=0,o=[];for(n in r)!hasOwnProperty_1(hiddenKeys,n)&&hasOwnProperty_1(r,n)&&push(o,n);while(e.length>i)hasOwnProperty_1(r,n=e[i++])&&(~indexOf(o,n)||push(o,n));return o},enumBugKeys=["constructor","hasOwnProperty","isPrototypeOf","propertyIsEnumerable","toLocaleString","toString","valueOf"],hiddenKeys$1=enumBugKeys.concat("length","prototype"),f$3=Object.getOwnPropertyNames||function(t){return objectKeysInternal(t,hiddenKeys$1)},objectGetOwnPropertyNames={f:f$3},f$4=Object.getOwnPropertySymbols,objectGetOwnPropertySymbols={f:f$4},concat=functionUncurryThis([].concat),ownKeys=getBuiltIn("Reflect","ownKeys")||function(t){var e=objectGetOwnPropertyNames.f(anObject(t)),n=objectGetOwnPropertySymbols.f;return n?concat(e,n(t)):e},copyConstructorProperties=function(t,e){for(var n=ownKeys(e),r=objectDefineProperty.f,i=objectGetOwnPropertyDescriptor.f,o=0;o<n.length;o++){var a=n[o];hasOwnProperty_1(t,a)||r(t,a,i(e,a))}},replacement=/#|\.prototype\./,isForced=function(t,e){var n=data[normalize(t)];return n==POLYFILL||n!=NATIVE&&(isCallable(e)?fails(e):!!e)},normalize=isForced.normalize=function(t){return String(t).replace(replacement,".").toLowerCase()},data=isForced.data={},NATIVE=isForced.NATIVE="N",POLYFILL=isForced.POLYFILL="P",isForced_1=isForced,getOwnPropertyDescriptor$1=objectGetOwnPropertyDescriptor.f,_export=function(t,e){var n,r,i,o,a,s,l=t.target,u=t.global,c=t.stat;if(r=u?global_1:c?global_1[l]||setGlobal(l,{}):(global_1[l]||{}).prototype,r)for(i in e){if(a=e[i],t.noTargetGet?(s=getOwnPropertyDescriptor$1(r,i),o=s&&s.value):o=r[i],n=isForced_1(u?i:l+(c?".":"#")+i,t.forced),!n&&void 0!==o){if(typeof a==typeof o)continue;copyConstructorProperties(a,o)}(t.sham||o&&o.sham)&&createNonEnumerableProperty(a,"sham",!0),redefine(r,i,a,t)}},FunctionPrototype$2=Function.prototype,apply=FunctionPrototype$2.apply,bind$1=FunctionPrototype$2.bind,call$2=FunctionPrototype$2.call,functionApply="object"==typeof Reflect&&Reflect.apply||(bind$1?call$2.bind(apply):function(){return call$2.apply(apply,arguments)}),bind$2=functionUncurryThis(functionUncurryThis.bind),functionBindContext=function(t,e){return aCallable(t),void 0===e?t:bind$2?bind$2(t,e):function(){return t.apply(e,arguments)}},html=getBuiltIn("document","documentElement"),arraySlice=functionUncurryThis([].slice),engineIsIos=/(?:ipad|iphone|ipod).*applewebkit/i.test(engineUserAgent),engineIsNode="process"==classofRaw(global_1.process),set$1=global_1.setImmediate,clear=global_1.clearImmediate,process$1=global_1.process,Dispatch=global_1.Dispatch,Function$1=global_1.Function,MessageChannel=global_1.MessageChannel,String$3=global_1.String,counter=0,queue={},ONREADYSTATECHANGE="onreadystatechange",location,defer,channel,port;try{location=global_1.location}catch(error){}var run=function(t){if(hasOwnProperty_1(queue,t)){var e=queue[t];delete queue[t],e()}},runner=function(t){return function(){run(t)}},listener=function(t){run(t.data)},post=function(t){global_1.postMessage(String$3(t),location.protocol+"//"+location.host)};set$1&&clear||(set$1=function(t){var e=arraySlice(arguments,1);return queue[++counter]=function(){functionApply(isCallable(t)?t:Function$1(t),void 0,e)},defer(counter),counter},clear=function(t){delete queue[t]},engineIsNode?defer=function(t){process$1.nextTick(runner(t))}:Dispatch&&Dispatch.now?defer=function(t){Dispatch.now(runner(t))}:MessageChannel&&!engineIsIos?(channel=new MessageChannel,port=channel.port2,channel.port1.onmessage=listener,defer=functionBindContext(port.postMessage,port)):global_1.addEventListener&&isCallable(global_1.postMessage)&&!global_1.importScripts&&location&&"file:"!==location.protocol&&!fails(post)?(defer=post,global_1.addEventListener("message",listener,!1)):defer=ONREADYSTATECHANGE in documentCreateElement("script")?function(t){html.appendChild(documentCreateElement("script"))[ONREADYSTATECHANGE]=function(){html.removeChild(this),run(t)}}:function(t){setTimeout(runner(t),0)});var task={set:set$1,clear:clear},FORCED=!global_1.setImmediate||!global_1.clearImmediate;_export({global:!0,bind:!0,enumerable:!0,forced:FORCED},{setImmediate:task.set,clearImmediate:task.clear});var path=global_1,setImmediate$1=path.setImmediate,eventemitter3=createCommonjsModule((function(t){var e=Object.prototype.hasOwnProperty,n="~";function r(){}function i(t,e,n){this.fn=t,this.context=e,this.once=n||!1}function o(t,e,r,o,a){if("function"!==typeof r)throw new TypeError("The listener must be a function");var s=new i(r,o||t,a),l=n?n+e:e;return t._events[l]?t._events[l].fn?t._events[l]=[t._events[l],s]:t._events[l].push(s):(t._events[l]=s,t._eventsCount++),t}function a(t,e){0===--t._eventsCount?t._events=new r:delete t._events[e]}function s(){this._events=new r,this._eventsCount=0}Object.create&&(r.prototype=Object.create(null),(new r).__proto__||(n=!1)),s.prototype.eventNames=function(){var t,r,i=[];if(0===this._eventsCount)return i;for(r in t=this._events)e.call(t,r)&&i.push(n?r.slice(1):r);return Object.getOwnPropertySymbols?i.concat(Object.getOwnPropertySymbols(t)):i},s.prototype.listeners=function(t){var e=n?n+t:t,r=this._events[e];if(!r)return[];if(r.fn)return[r.fn];for(var i=0,o=r.length,a=new Array(o);i<o;i++)a[i]=r[i].fn;return a},s.prototype.listenerCount=function(t){var e=n?n+t:t,r=this._events[e];return r?r.fn?1:r.length:0},s.prototype.emit=function(t,e,r,i,o,a){var s=n?n+t:t;if(!this._events[s])return!1;var l,u,c=this._events[s],d=arguments.length;if(c.fn){switch(c.once&&this.removeListener(t,c.fn,void 0,!0),d){case 1:return c.fn.call(c.context),!0;case 2:return c.fn.call(c.context,e),!0;case 3:return c.fn.call(c.context,e,r),!0;case 4:return c.fn.call(c.context,e,r,i),!0;case 5:return c.fn.call(c.context,e,r,i,o),!0;case 6:return c.fn.call(c.context,e,r,i,o,a),!0}for(u=1,l=new Array(d-1);u<d;u++)l[u-1]=arguments[u];c.fn.apply(c.context,l)}else{var h,f=c.length;for(u=0;u<f;u++)switch(c[u].once&&this.removeListener(t,c[u].fn,void 0,!0),d){case 1:c[u].fn.call(c[u].context);break;case 2:c[u].fn.call(c[u].context,e);break;case 3:c[u].fn.call(c[u].context,e,r);break;case 4:c[u].fn.call(c[u].context,e,r,i);break;default:if(!l)for(h=1,l=new Array(d-1);h<d;h++)l[h-1]=arguments[h];c[u].fn.apply(c[u].context,l)}}return!0},s.prototype.on=function(t,e,n){return o(this,t,e,n,!1)},s.prototype.once=function(t,e,n){return o(this,t,e,n,!0)},s.prototype.removeListener=function(t,e,r,i){var o=n?n+t:t;if(!this._events[o])return this;if(!e)return a(this,o),this;var s=this._events[o];if(s.fn)s.fn!==e||i&&!s.once||r&&s.context!==r||a(this,o);else{for(var l=0,u=[],c=s.length;l<c;l++)(s[l].fn!==e||i&&!s[l].once||r&&s[l].context!==r)&&u.push(s[l]);u.length?this._events[o]=1===u.length?u[0]:u:a(this,o)}return this},s.prototype.removeAllListeners=function(t){var e;return t?(e=n?n+t:t,this._events[e]&&a(this,e)):(this._events=new r,this._eventsCount=0),this},s.prototype.off=s.prototype.removeListener,s.prototype.addListener=s.prototype.on,s.prefixed=n,s.EventEmitter=s,t.exports=s})),IsEmptyObject=function(t){return t&&t.constructor===Object&&0===Object.keys(t).length},RandomStr=function(){for(var t=arguments.length>0&&void 0!==arguments[0]?arguments[0]:8,e="0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ".split(""),n="",r=0;r<t;r++)n+=e[62*Math.random()|0];return n};function _genList(t){for(var e=[],n=t.split("|"),r=0;r<n.length;r++){var i=n[r].trim();(0!==r||i)&&((r!==n.length-1||i)&&e.push(i))}return e}function _genItem(t,e){for(var n={},r=0;r<t.length;r++)n[t[r]]=e[r];return n}var ParseSettlementContent=function(){var t=arguments.length>0&&void 0!==arguments[0]?arguments[0]:"";if(""===t)return t;for(var e=t.split("\n"),n="",r={positionClosed:{title:"平仓明细 Position Closed",colNames:[]},transactionRecords:{title:"成交记录 Transaction Record",colNames:[]}},i={account:{},positionClosed:[],transactionRecords:[]},o=0;o<e.length;o++){var a=e[o].trim();if(a.indexOf("资金状况")>-1)n="account",o++;else if(a.indexOf("平仓明细")>-1||a.indexOf("成交记录")>-1){n=a.indexOf("平仓明细")>-1?"positionClosed":"transactionRecords";while(o++){var s=e[o].trim();if(""===s.replace(/-/g,"")){if(0===r[n].colNames.length)continue;break}r[n].colNames=_genList(s)}}else if("account"===n){if(0===a.length||""===a.replace("-","")){n="";continue}for(var l=a.match(/([A-Z][a-zA-Z\.\/\(\)\s]+)[:：]+/g),u=a.match(/(-?[\d]+\.\d\d)/g),c=0;c<l.length;c++)i.account[l[c].split(/[:：]/)[0]]=u[c]}else if("positionClosed"===n||"transactionRecords"===n){if(0===a.length||""===a.replace(/-/g,"")){n="";continue}var d=r[n].colNames,h=_genList(a),f=_genItem(d,h);if(d.length!==h.length&&"transactionRecords"===n)for(var p=d.indexOf("Lots"),m=d.indexOf("Fee"),g=1;g<h.length;g++)if(/^\d+$/.test(h[g])){f.Lots=h[g],f.Fee=h[g+(m-p)];break}i[n].push(f)}}return i},TqWebsocket=function(_EventEmitter){_inherits(TqWebsocket,_EventEmitter);var _super=_createSuper(TqWebsocket);function TqWebsocket(t){var e,n=arguments.length>1&&void 0!==arguments[1]?arguments[1]:{};return _classCallCheck(this,TqWebsocket),e=_super.call(this),e.urlList=t instanceof Array?t:[t],e.ws=null,e.queue=[],e.reconnect=!0,e.reconnectTask=null,e.reconnectInterval=n.reconnectInterval?n.reconnectInterval:3e3,e.reconnectMaxTimes=n.reconnectMaxTimes?n.reconnectMaxTimes:2,e.WebSocket=n.WebSocket?n.WebSocket:WebSocket,e.reconnectTimes=0,e.reconnectUrlIndex=0,e.STATUS={CONNECTING:0,OPEN:1,CLOSING:2,CLOSED:3},e.__init(!1),e}return _createClass(TqWebsocket,[{key:"send",value:function(t){var e=JSON.stringify(t);this.isReady()?this.ws.send(e):this.queue.push(e)}},{key:"isReady",value:function(){return this.ws.readyState===this.WebSocket.OPEN}},{key:"__init",value:function __init(){var isReconnection=!(arguments.length>0&&void 0!==arguments[0])||arguments[0];this.ws=new this.WebSocket(this.urlList[this.reconnectUrlIndex]),isReconnection&&this.reconnectUrlIndex===this.urlList.length-1&&(this.reconnectTimes+=1);var _this=this;this.ws.onmessage=function(message){var data=eval("("+message.data+")");_this.emit("message",data),setImmediate((function(){_this.ws.send('{"aid":"peek_message"}')}))},this.ws.onclose=function(t){_this.emit("close"),_this.queue=[],_this.reconnect&&(_this.reconnectMaxTimes<=_this.reconnectTimes?(clearTimeout(_this.reconnectTask),_this.emit("death",{msg:"超过重连次数"+_this.reconnectMaxTimes})):_this.reconnectTask=setTimeout((function(){3===_this.ws.readyState&&(_this.reconnectUrlIndex=_this.reconnectUrlIndex+1<_this.urlList.length?_this.reconnectUrlIndex+1:0,_this.__init(!0),_this.emit("reconnect",{msg:"发起重连第 "+_this.reconnectTimes+" 次"}))}),_this.reconnectInterval))},this.ws.onerror=function(t){_this.emit("error",t),_this.ws.close()},this.ws.onopen=function(){_this.emit("open",{msg:"发起重连第 "+_this.reconnectTimes+" 次, 成功"}),this.reconnectTask&&clearTimeout(_this.reconnectTask);while(_this.queue.length>0){if(1!==_this.ws.readyState)break;_this.ws.send(_this.queue.shift())}}}},{key:"close",value:function(){this.ws.onclose=function(){},this.ws.close()}}]),TqWebsocket}(eventemitter3),TqTradeWebsocket=function(t){_inherits(n,t);var e=_createSuper(n);function n(t,r){var i,o=arguments.length>2&&void 0!==arguments[2]?arguments[2]:{};return _classCallCheck(this,n),i=e.call(this,t,o),i.dm=r,i.req_login=null,i.init(),i}return _createClass(n,[{key:"init",value:function(){var t=this;this.on("message",(function(e){if("rtn_data"===e.aid){for(var n=t._separateNotifies(e.data),r=0;r<n.length;r++)t.emit("notify",n[r]);t.dm.mergeData(e.data)}else if("rtn_brokers"===e.aid)t.emit("rtn_brokers",e.brokers);else if("qry_settlement_info"===e.aid){var i=ParseSettlementContent(e.settlement_info);t.dm.mergeData({trade:_defineProperty({},e.user_name,{his_settlements:_defineProperty({},e.trading_day,i)})}),Tqsdk.store&&Tqsdk.store.setContent(e.user_name,e.trading_day,e.settlement_info)}})),this.on("reconnect",(function(){t.req_login&&t.send(t.req_login)}))}},{key:"_separateNotifies",value:function(t){for(var e=[],n=0;n<t.length;n++)if(t[n].notify){var r=t.splice(n--,1)[0].notify;for(var i in r)e.push(r[i])}return e}},{key:"send",value:function(t){"req_login"===t.aid&&(this.req_login=t),_get(_getPrototypeOf(n.prototype),"send",this).call(this,t)}}]),n}(TqWebsocket),TqQuoteWebsocket=function(t){_inherits(n,t);var e=_createSuper(n);function n(t,r){var i,o=arguments.length>2&&void 0!==arguments[2]?arguments[2]:{};return _classCallCheck(this,n),i=e.call(this,t,o),i.dm=r,i.subscribe_quote=null,i.charts={},i.init(),i}return _createClass(n,[{key:"init",value:function(){var t=this;this.on("message",(function(e){"rtn_data"===e.aid&&t.dm.mergeData(e.data)})),this.on("reconnect",(function(e){for(var n in t.subscribe_quote&&t.send(t.subscribe_quote),t.charts)t.charts[n].view_width>0&&t.send(t.charts[n])}))}},{key:"send",value:function(t){"subscribe_quote"===t.aid?null!==this.subscribe_quote&&JSON.stringify(t.ins_list)===JSON.stringify(this.subscribe_quote.ins_list)||(this.subscribe_quote=t,_get(_getPrototypeOf(n.prototype),"send",this).call(this,t)):"set_chart"===t.aid&&(0===t.view_width?this.charts[t.chart_id]&&delete this.charts[t.chart_id]:this.charts[t.chart_id]=t,_get(_getPrototypeOf(n.prototype),"send",this).call(this,t))}}]),n}(TqWebsocket),TqRecvOnlyWebsocket=function(t){_inherits(n,t);var e=_createSuper(n);function n(t,r){var i,o=arguments.length>2&&void 0!==arguments[2]?arguments[2]:{};return _classCallCheck(this,n),i=e.call(this,t,o),i.dm=r,i.init(),i}return _createClass(n,[{key:"init",value:function(){var t=this;this.on("message",(function(e){"rtn_data"===e.aid&&t.dm.mergeData(e.data)})),this.on("reconnect",(function(t){}))}}]),n}(TqWebsocket),Quote=function(){function t(){_classCallCheck(this,t),this.instrument_id="",this.datetime="",this._last_price="-",this.ask_price1="-",this.ask_volume1="-",this.bid_price1="-",this.bid_volume1="-",this.highest="-",this.lowest="-",this.open="-",this.close="-",this.average="-",this.volume="-",this.amount="-",this.open_interest="-",this.lower_limit="-",this.upper_limit="-",this.settlement="-",this.change="-",this.change_percent="-",this.strike_price=NaN,this.pre_open_interest="-",this.pre_close="-",this.pre_volume="-",this._pre_settlement="-",this.margin="-",this.commission="-"}return _createClass(t,[{key:"last_price",get:function(){return this._last_price},set:function(t){this._last_price=t,this.setChange()}},{key:"pre_settlement",get:function(){return this._pre_settlement},set:function(t){this._pre_settlement=t,this.setChange()}},{key:"setChange",value:function(){Number.isFinite(this._last_price)&&Number.isFinite(this._pre_settlement)&&0!==this._pre_settlement&&(this.change=this._last_price-this._pre_settlement,this.change_percent=this.change/this._pre_settlement*100)}}]),t}(),Chart=function t(){var e=arguments.length>0&&void 0!==arguments[0]?arguments[0]:{};_classCallCheck(this,t),this.left_id=-1,this.right_id=-1,this.more_data=!0,this.state=e},DataManager=function(t){_inherits(n,t);var e=_createSuper(n);function n(){var t,r=arguments.length>0&&void 0!==arguments[0]?arguments[0]:{};return _classCallCheck(this,n),t=e.call(this),t._epoch=0,t._data=r,t._diffs=[],t}return _createClass(n,[{key:"mergeData",value:function(t){var e=!(arguments.length>1&&void 0!==arguments[1])||arguments[1],n=!(arguments.length>2&&void 0!==arguments[2])||arguments[2],r=Array.isArray(t)?t:[t];e&&(this._epoch+=1,this._diffs=r);var i,o=_createForOfIteratorHelper(r);try{for(o.s();!(i=o.n()).done;){var a=i.value;null===a||IsEmptyObject(a)||this._mergeObject(this._data,a,this._epoch,n)}}catch(s){o.e(s)}finally{o.f()}e&&this._data._epoch===this._epoch&&this.emit("data",null)}},{key:"_mergeObject",value:function(t,e){var n=arguments.length>2&&void 0!==arguments[2]?arguments[2]:0,r=!(arguments.length>3&&void 0!==arguments[3])||arguments[3];for(var i in e){var o=e[i],a=_typeof(o);if(["string","boolean","number"].includes(a))t[i]="NaN"===o?NaN:o;else if(null===o&&r)delete t[i];else if(Array.isArray(o))t[i]=o,o._epoch||Object.defineProperty(o,"_epoch",{configurable:!1,enumerable:!1,writable:!0}),o._epoch=n;else if("object"===a)if(t[i]=t[i]||("data"===i?[]:{}),"quotes"===i)for(var s in o){var l=o[s];null!==l?(t[i][s]||(t[i][s]=new Quote),this._mergeObject(t[i][s],l,n,r)):r&&s&&delete t[i][s]}else this._mergeObject(t[i],o,n,r)}t._root||Object.defineProperty(t,"_root",{value:this._data,configurable:!1,enumerable:!1,writable:!1}),t._epoch||Object.defineProperty(t,"_epoch",{configurable:!1,enumerable:!1,writable:!0}),t._epoch=n}},{key:"isChanging",value:function(t){if(Array.isArray(t)){for(var e=this._data,n=0;n<t.length;n++){if(e=e[t[n]],void 0===e)return!1;if(e._epoch&&e._epoch===this._epoch)return!0}return!1}return!(!t||!t._epoch)&&t._epoch===this._epoch}},{key:"setDefault",value:function(t){for(var e=arguments.length>1&&void 0!==arguments[1]?arguments[1]:{},n=arguments.length>2&&void 0!==arguments[2]?arguments[2]:this._data,r=n,i=0;i<t.length;i++){if("string"!==typeof t[i]&&"number"!==typeof t[i])break;var o=t[i];if(o in r||(r[o]=i===t.length-1?e:{}),i===t.length-1)return r[o];r=r[o]}return r}},{key:"getByPath",value:function(t){for(var e=arguments.length>1&&void 0!==arguments[1]?arguments[1]:this._data,n=e,r=0;r<t.length;r++)if(n=n[t[r]],void 0===n||null===n)return null;return n}}]),n}(eventemitter3),Tqsdk=function(t){_inherits(n,t);var e=_createSuper(n);function n(){var t,r=arguments.length>0&&void 0!==arguments[0]?arguments[0]:{},i=r.symbolsServerUrl,o=void 0===i?"https://openmd.shinnytech.com/t/md/symbols/latest.json":i,a=r.wsQuoteUrl,s=void 0===a?"wss://openmd.shinnytech.com/t/md/front/mobile":a,l=r.requestMethod,u=void 0===l?null:l,c=r.clientSystemInfo,d=void 0===c?"":c,h=r.clientAppId,f=void 0===h?"":h,p=r.autoInit,m=void 0===p||p,g=r.accessToken,y=void 0===g?"eyJhbGciOiJSUzI1NiIsInR5cCIgOiAiSldUIiwia2lkIiA6ICJobi1MZ3ZwbWlFTTJHZHAtRmlScjV5MUF5MnZrQmpLSFFyQVlnQ0UwR1JjIn0.eyJqdGkiOiJlZGFmNDllMy03MjVhLTQwNDMtOGY3OS0yMGU4NWRhZTVlZGEiLCJleHAiOjE2NzM2ODI5MzgsIm5iZiI6MCwiaWF0IjoxNjQyMTQ2OTM4LCJpc3MiOiJodHRwOi8vYXV0aC5zaGlubnl0ZWNoLmNvbS9hdXRoL3JlYWxtcy9zaGlubnl0ZWNoIiwic3ViIjoiNGJiMmEwZGMtMjZkMS00M2Y1LTlkNjctOTlkYzVkZGE4Y2ZlIiwidHlwIjoiQmVhcmVyIiwiYXpwIjoic2hpbm55X3dlYiIsImF1dGhfdGltZSI6MCwic2Vzc2lvbl9zdGF0ZSI6IjRkYmM1NzU2LWY0OTMtNDRiMi05MDA1LWNlMGU2NzhiOTRhYiIsImFjciI6IjEiLCJzY29wZSI6ImF0dHJpYnV0ZXMiLCJncmFudHMiOnsiZmVhdHVyZXMiOlsiIl0sImFjY291bnRzIjpbIioiXX19.F1DHbKSbFt4lEBcjI3Qt1qFuPXUFAOxtjWDk53dY2JIrpk4scQwqYfhU5I34kiaMBrsFNj9bnmnXHbJhCa2ktcP82yb1p2rRH3goaoFUYEAqX0SBCR_58a3b4h165CqF3K1TsUqJYWbngo25MqElgq37StaSzJGsTKPF2GzWMghFsBPwDL8fVhJnTq3vzwPcxI1EARmyGImc8OHPHSPFTQchb3J5YVG6K_me1mM9PqNTiWracEB9R_ID6PfDVbzU-mzTmtFZjw3tLFqVC9nWtPcoCHZmfSfiRkdrKsq0QLh-NLg6CTIVB4QQ8T5BXH58micbhgAEW6ly1GFGgrVkug":g,v=r.data,_=void 0===v?{klines:{},quotes:{},charts:{},ticks:{},trade:{}}:v,b=arguments.length>1&&void 0!==arguments[1]?arguments[1]:{};_classCallCheck(this,n),t=e.call(this),t._insUrl=o,t._mdUrl=s,t.clientSystemInfo=d,t.clientAppId=f,t.wsOption=b,t.requestMethod=u||function(t){return fetch(t,{headers:{Accept:"application/json; charset=utf-8"}}).then((function(t){return t.json()}))},t._accessToken=y,t._prefix="TQJS_";var x=_assertThisInitialized(t);return t.dm=new DataManager(_),t.dm.on("data",(function(){x.emit("rtn_data",null)})),t.brokers_list=null,t.brokers=null,t.trade_accounts={},t.subscribeQuotesSet=new Set,t.quotesWs=null,t.quotesInfo={},m&&(t.initMdWebsocket(),t.initTdWebsocket()),t}return _createClass(n,[{key:"initMdWebsocket",value:function(){if(!this.quotesWs){var t=this;this.requestMethod(this._insUrl).then((function(e){for(var n in e)"FUTURE_OPTION"===e[n]["class"]&&(e[n]["class"]="OPTION");t.quotesInfo=e,t.emit("ready"),t.emit("rtn_data",null)}))["catch"]((function(e){return t.emit("error",e),e})),this.quotesWs=new TqQuoteWebsocket(this._mdUrl,this.dm,this.wsOption)}}},{key:"initTdWebsocket",value:function(){if(!this.brokers){var t=this;this.requestMethod("https://files.shinnytech.com/broker-list.json").then((function(e){t.brokers_list=e,t.brokers=Object.keys(e).filter((function(t){return!t.endsWith(" ")})).sort(),t.emit("rtn_brokers",t.brokers)}))["catch"]((function(e){return t.emit("error",e),e}))}}},{key:"addWebSocket",value:function(){var t=arguments.length>0&&void 0!==arguments[0]?arguments[0]:"";return t?new TqRecvOnlyWebsocket(t,this.dm,this.wsOption):null}},{key:"get",value:function(){var t=arguments.length>0&&void 0!==arguments[0]?arguments[0]:{},e=t.name,n=void 0===e?"users":e,r=t.bid,i=void 0===r?"":r,o=t.user_id,a=void 0===o?"":o,s=t.currency,l=void 0===s?"CNY":s,u=t.symbol,c=void 0===u?"":u,d=t.order_id,h=void 0===d?"":d,f=t.trade_id,p=void 0===f?"":f,m=t.trading_day,g=void 0===m?"":m,y=t.chart_id,v=void 0===y?"":y,_=t.input,b=void 0===_?"":_,x=t.duration,w=void 0===x?0:x;if("users"===n)return Object.keys(this.trade_accounts);if(a){var M=this._getAccountInfoByPaths({bid:i,user_id:a},[]);if(null===M&&(M=this.dm.getByPath("trade",a)),"user"===n)return M;if(["session","accounts","positions","orders","trades","his_settlements"].indexOf(n)>-1)return M&&M[n]?M[n]:null;if(M&&M[n+"s"]){var S="account"===n?l:"position"===n?c:"order"===n?h:"trade"===n?p:"his_settlement"===n?g:"";return M[n+"s"][S]}return null}return"quotes"===n?b?this.get_quotes_by_input(b):[]:"quote"===n?this.getQuote(c):"klines"===n?this.getKlines(c,w):"ticks"===n?this.getTicks(c):"charts"===n?this.dm.getByPath(["charts"]):"chart"===n?this.dm.getByPath(["charts",v]):void 0}},{key:"getByPath",value:function(t){var e=arguments.length>1&&void 0!==arguments[1]?arguments[1]:this.dm;return e.getByPath(t)}},{key:"getQuotesByInput",value:function(t){var e=arguments.length>1&&void 0!==arguments[1]?arguments[1]:{};if("string"!==typeof t)return[];var n={input:t.toLowerCase(),symbol:"boolean"!==typeof e.symbol||e.symbol,pinyin:"boolean"!==typeof e.pinyin||e.pinyin,include_expired:"boolean"===typeof e.include_expired&&e.include_expired,FUTURE:"boolean"!==typeof e.future||e.future,FUTURE_INDEX:"boolean"===typeof e.future_index&&e.future_index,FUTURE_CONT:"boolean"===typeof e.future_cont&&e.future_cont,OPTION:"boolean"===typeof e.option&&e.option,COMBINE:"boolean"===typeof e.combine&&e.combine},r=[];for(var i in this.quotesInfo)this._filterSymbol(n,this.quotesInfo[i])&&r.push(i);return r}},{key:"_filterSymbol",value:function(t,e){if(t[e["class"]]&&(t.include_expired||!t.include_expired&&!e.expired)){if(t.symbol)if(e.underlying_product){var n=e.underlying_product.toLowerCase().split("."),r=_slicedToArray(n,2),i=r[0],o=r[1];if(i===t.input||o===t.input)return!0}else{if(e.product_id&&e.product_id.toLowerCase()===t.input)return!0;if(t.input.length>2&&e.instrument_id.toLowerCase().indexOf(t.input)>-1)return!0}if(t.pinyin){var a,s=e.py.split(","),l=_createForOfIteratorHelper(s);try{for(l.s();!(a=l.n()).done;){var u=a.value;if(u.indexOf(t.input)>-1)return!0}}catch(c){l.e(c)}finally{l.f()}}}return!1}},{key:"getQuote",value:function(t){if(""===t)return{};if(!this.quotesInfo[t])return{};var e=this.dm.setDefault(["quotes",t],new Quote);if(!e["class"]){var n=e.last_price?e.last_price:this.quotesInfo[t].last_price;Object.assign(e,this.quotesInfo[t],{last_price:n})}return this.subscribeQuote(t),e}},{key:"setChart",value:function(t){var e={};t.trading_day_start||t.trading_day_count?(e.trading_day_start=t.trading_day_start?t.trading_day_start:0,e.trading_day_count=t.trading_day_count?t.trading_day_count:864e11):(e.view_width=t.view_width?t.view_width:500,t.left_kline_id?e.left_kline_id=t.left_kline_id:t.focus_datetime&&(e.focus_datetime=t.focus_datetime,e.focus_position=t.focus_position?t.focus_position:0));var n=Object.assign({aid:"set_chart",chart_id:t.chart_id?t.chart_id:this._prefix+"kline_chart",ins_list:t.ins_list?t.ins_list.join(","):t.symbol,duration:t.duration},e);return this.quotesWs.send(n),this.dm.setDefault(["charts",n.chart_id],new Chart(n))}},{key:"getChart",value:function(t){return""===t?null:this.dm.getByPath(["charts",t])}},{key:"getKlines",value:function(t,e){if(""===t)return null;var n=this.dm.getByPath(["klines",t,e]);return n&&n.data&&-1!==n.last_id||(this.dm.mergeData({klines:_defineProperty({},t,_defineProperty({},e,{trading_day_end_id:-1,trading_day_start_id:-1,last_id:-1,data:{}}))},!1,!1),n=this.dm.getByPath(["klines",t,e])),n}},{key:"getTicks",value:function(t){if(""===t)return null;var e=this.dm.getByPath(["ticks",t]);return e&&e.data||this.dm.mergeData({ticks:_defineProperty({},t,{last_id:-1,data:{}})},!1,!1),this.dm.getByPath(["ticks",t])}},{key:"isChanging",value:function(t){return t&&t._epoch&&t._root&&t._root._epoch?t._epoch===t._root._epoch:!!Array.isArray(t)&&this.dm.isChanging(t)}},{key:"subscribeQuote",value:function(){var t,e=arguments.length>0&&void 0!==arguments[0]?arguments[0]:[],n=this.subscribeQuotesSet.size;for(var r in this.trade_accounts){var i=this.getPositions(this.trade_accounts[r]);if(i)for(var o in i)this.subscribeQuotesSet.add(o)}e="string"===typeof e?e.split(","):e;var a,s=_createForOfIteratorHelper(e);try{for(s.s();!(a=s.n()).done;){var l=a.value;this.subscribeQuotesSet.add(l)}}catch(u){s.e(u)}finally{s.f()}n!==this.subscribeQuotesSet.size&&this.quotesWs.send({aid:"subscribe_quote",ins_list:(t=[]).concat.apply(t,_toConsumableArray(this.subscribeQuotesSet)).join(",")})}},{key:"addAccount",value:function(t){if(this.brokers){if(t.bid&&t.user_id&&t.password){if(-1===this.brokers.indexOf(t.bid))return null;var e=this._getAccountKey(t);if(e&&!this.trade_accounts[e]){var n=new DataManager({trade:_defineProperty({},t.user_id,{accounts:{CNY:{}},trades:{},positions:{},orders:{},his_settlements:{}})}),r=[this.brokers_list[t.bid].url+"?access_token="+this._accessToken,this.brokers_list[t.bid].url],i=new TqTradeWebsocket(r,n,this.wsOption),o=this;n.on("data",(function(){o.emit("rtn_data",null)})),i.on("notify",(function(e){o.emit("notify",Object.assign(e,{bid:t.bid,user_id:t.user_id}))})),this.trade_accounts[e]={bid:t.bid,user_id:t.user_id,password:t.password,ws:i,dm:n}}return this.trade_accounts[e]}return null}}},{key:"removeAccount",value:function(t){var e=this._getAccountKey(t);this.trade_accounts[e]&&(this.trade_accounts[e].ws.close(),delete this.trade_accounts[e])}},{key:"login",value:function(t){var e=this._getAccountRef(t);IsEmptyObject(e)&&(e=this.addAccount(t));var n={aid:"req_login"};this.clientAppId&&(n.client_app_id=this.clientAppId,n.client_system_info=this.clientSystemInfo),e&&e.ws&&(n.bid=t.bid,n.user_name=t.user_id,n.password=t.password||e.password,e.ws.send(n))}},{key:"isLogined",value:function(t){var e=this._getAccountRef(t);if(e&&e.dm){var n=e.dm.getByPath(["trade",t.user_id,"session"]),r=e.dm.getByPath(["trade",e.user_id,"trade_more_data"]);if(n&&n.trading_day&&!1===r)return!0}return!1}},{key:"refreshAccount",value:function(t){var e=this._getAccountRef(t);e&&e.ws&&(e.ws.send({aid:"qry_account_info"}),e.ws.send({aid:"qry_account_register"}))}},{key:"refreshAccounts",value:function(){for(var t in this.trade_accounts)this.trade_accounts[t].ws.send({aid:"qry_account_info"}),this.trade_accounts[t].ws.send({aid:"qry_account_register"})}},{key:"getAllAccounts",value:function(t){var e=[];return Object.values(this.trade_accounts).forEach((function(t){e.push({bid:t.bid,user_id:t.user_id,password:t.password})})),e}},{key:"getAccount",value:function(t){return this._getAccountInfoByPaths(t,["accounts","CNY"])}},{key:"_getAccountKey",value:function(t){if(t&&t.user_id){if(t.bid)return t.bid+","+t.user_id;var e="";for(var n in this.trade_accounts)if(t.user_id===this.trade_accounts[n].user_id){if(e)return"";e=this.trade_accounts[n].bid}return e+","+t.user_id}return""}},{key:"_getAccountRef",value:function(t){var e=this._getAccountKey(t);return this.trade_accounts[e]||{}}},{key:"insertOrder",value:function(t){if(!this.isLogined(t))return null;var e=this._getAccountRef(t),n=this._prefix+RandomStr(8),r={user_id:t.user_id,order_id:n,exchange_id:t.exchange_id,instrument_id:t.instrument_id,direction:t.direction,offset:t.offset,price_type:t.price_type?t.price_type:"LIMIT",limit_price:Number(t.limit_price),volume_condition:"ANY",time_condition:"ANY"===t.price_type?"IOC":"GFD"},i=Object.assign({aid:"insert_order",volume:t.volume},r);e.ws.send(i);var o=Object.assign({volume_orign:t.volume,status:"ALIVE",volume_left:t.volume},r);return e.dm.mergeData({trade:_defineProperty({},t.user_id,{orders:_defineProperty({},n,o)})},!1,!1),e.dm.getByPath(["trade",t.user_id,"orders",n])}},{key:"autoInsertOrder",value:function(t){if(!this.is_logined(t))return null;var e={bid:t.bid,user_id:t.user_id,exchange_id:t.exchange_id,instrument_id:t.instrument_id,direction:t.direction,price_type:t.price_type?t.price_type:"LIMIT",volume_condition:"ANY",time_condition:"ANY"===t.price_type?"IOC":"GFD",limit_price:Number(t.limit_price)};if("SHFE"!==t.exchange_id&&"INE"!==t.exchange_id||"CLOSE"!==t.offset)return[this.insert_order(Object.assign({offset:t.offset,volume:t.volume},e))];var n=this.getPosition({bid:t.bid,user_id:t.user_id,symbol:t.exchange_id+"."+t.instrument_id}),r=0;"BUY"===t.direction&&n.volume_short_today>0?r=Math.min(n.volume_short_today,t.volume):"SELL"===t.direction&&n.volume_long_today>0&&(r=Math.min(n.volume_long_today,t.volume));var i=[];if(r>0){var o=this.insert_order(Object.assign({offset:"CLOSETODAY",volume:r},e));i.push(o)}if(t.volume-r>0){var a=this.insert_order(Object.assign({offset:"CLOSE",volume:t.volume-r},e));i.push(a)}return i}},{key:"cancelOrder",value:function(t){var e=this._getAccountRef(t);e&&e.ws&&e.ws.send({aid:"cancel_order",user_id:t.user_id,order_id:t.order_id})}},{key:"getPosition",value:function(t){return this._getAccountInfoByPaths(t,["positions",t.symbol])}},{key:"getPositions",value:function(t){return this._getAccountInfoByPaths(t,["positions"])}},{key:"getOrder",value:function(t){return this._getAccountInfoByPaths(t,["orders",t.order_id])}},{key:"getOrders",value:function(t){return this._getAccountInfoByPaths(t,["orders"])}},{key:"getOrdersBySymbol",value:function(t){var e=this._getAccountInfoByPaths(t,["orders"]),n=t.symbol.split("."),r=_slicedToArray(n,2),i=r[0],o=r[1],a={};for(var s in e)e[s].exchange_id===i&&e[s].instrument_id===o&&(a[s]=e[s]);return a}},{key:"getTrade",value:function(t){return this._getAccountInfoByPaths(t,["trades",t.trade_id])}},{key:"getTrades",value:function(t){return this._getAccountInfoByPaths(t,["trades"])}},{key:"getTradesByOrder",value:function(t){var e=this._getAccountInfoByPaths(t,["trades"]),n={};for(var r in e)e[r].order_id===t.order_id&&(n[r]=e[r]);return n}},{key:"getTradesBySymbol",value:function(t){var e=this._getAccountInfoByPaths(t,["trades"]),n=t.symbol.split("."),r=_slicedToArray(n,2),i=r[0],o=r[1],a={};for(var s in e)e[s].exchange_id===i&&e[s].instrument_id===o&&(a[s]=e[s]);return a}},{key:"getHisSettlements",value:function(t){return this._getAccountInfoByPaths(t,["his_settlements"])}},{key:"getHisSettlement",value:function(t){return this._getAccountInfoByPaths(t,["his_settlements",t.trading_day])}},{key:"_getAccountInfoByPaths",value:function(t){var e=arguments.length>1&&void 0!==arguments[1]?arguments[1]:[],n=this._getAccountRef(t);return n&&n.dm?n.dm.getByPath(["trade",n.user_id].concat(e)):null}},{key:"confirmSettlement",value:function(t){var e=this._getAccountRef(t);e&&e.ws&&!e._hasConfirmed&&(e._hasConfirmed=!0,e.ws.send({aid:"confirm_settlement"}))}},{key:"transfer",value:function(t){var e=this._getAccountRef(t);e&&e.ws&&e.ws.send({aid:"req_transfer",bank_id:t.bank_id,bank_password:t.bank_password,future_account:t.future_account,future_password:t.future_password,currency:t.currency||"CNY",amount:t.amount})}},{key:"hisSettlement",value:function(t){if(!n.store)return null;var e=this._getAccountRef(t);if(e&&e.dm){var r=e.dm.getByPath(["trade",t.user_id,"his_settlements",t.trading_day]);if(r)return;n.store.getContent(t.user_id,t.trading_day).then((function(n){if(null===n)e.ws.send({aid:"qry_settlement_info",trading_day:Number(t.trading_day)});else{var r=ParseSettlementContent(n);e.dm.mergeData({trade:_defineProperty({},t.user_id,{his_settlements:_defineProperty({},t.trading_day,r)})},!0,!1)}}))["catch"]((function(t){}))}}}]),n}(eventemitter3);Tqsdk.prototype.subscribe_quote=Tqsdk.prototype.subscribeQuote,Tqsdk.prototype.his_settlement=Tqsdk.prototype.hisSettlement,Tqsdk.prototype.confirm_settlement=Tqsdk.prototype.confirmSettlement,Tqsdk.prototype.add_account=Tqsdk.prototype.addAccount,Tqsdk.prototype.remove_account=Tqsdk.prototype.removeAccount,Tqsdk.prototype.get_by_path=Tqsdk.prototype.getByPath,Tqsdk.prototype.get_quotes_by_input=Tqsdk.prototype.getQuotesByInput,Tqsdk.prototype.get_quote=Tqsdk.prototype.getQuote,Tqsdk.prototype.set_chart=Tqsdk.prototype.setChart,Tqsdk.prototype.get_user=Tqsdk.prototype.getAccount,Tqsdk.prototype.is_logined=Tqsdk.prototype.isLogined,Tqsdk.prototype.is_changed=Tqsdk.prototype.isChanging,Tqsdk.prototype.insert_order=Tqsdk.prototype.insertOrder,Tqsdk.prototype.auto_insert_order=Tqsdk.prototype.autoInsertOrder,Tqsdk.prototype.cancel_order=Tqsdk.prototype.cancelOrder,Tqsdk.TqWebsocket=TqWebsocket,Tqsdk.DataManager=DataManager,Tqsdk.version=version,__webpack_exports__["a"]=Tqsdk}).call(this,__webpack_require__("c8ba"))},"099a":function(t,e,n){"use strict";var r=function(){var t=this,e=t._self._c;return e("i",{class:t.classes,style:t.styles,on:{click:t.handleClick}})},i=[],o=n("ade3"),a=(n("99af"),n("a9e3"),"ivu-icon"),s={name:"Icon",props:{type:{type:String,default:""},size:[Number,String],color:String,custom:{type:String,default:""}},computed:{classes:function(){return["".concat(a),Object(o["a"])(Object(o["a"])({},"".concat(a,"-").concat(this.type),""!==this.type),"".concat(this.custom),""!==this.custom)]},styles:function(){var t={};return this.size&&(t["font-size"]="".concat(this.size,"px")),this.color&&(t.color=this.color),t}},methods:{handleClick:function(t){this.$emit("click",t)}}},l=s,u=n("2877"),c=Object(u["a"])(l,r,i,!1,null,null,null);e["a"]=c.exports},"0a3c":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="ene._feb._mar._abr._may._jun._jul._ago._sep._oct._nov._dic.".split("_"),n="ene_feb_mar_abr_may_jun_jul_ago_sep_oct_nov_dic".split("_"),r=[/^ene/i,/^feb/i,/^mar/i,/^abr/i,/^may/i,/^jun/i,/^jul/i,/^ago/i,/^sep/i,/^oct/i,/^nov/i,/^dic/i],i=/^(enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre|ene\.?|feb\.?|mar\.?|abr\.?|may\.?|jun\.?|jul\.?|ago\.?|sep\.?|oct\.?|nov\.?|dic\.?)/i,o=t.defineLocale("es-do",{months:"enero_febrero_marzo_abril_mayo_junio_julio_agosto_septiembre_octubre_noviembre_diciembre".split("_"),monthsShort:function(t,r){return t?/-MMM-/.test(r)?n[t.month()]:e[t.month()]:e},monthsRegex:i,monthsShortRegex:i,monthsStrictRegex:/^(enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre)/i,monthsShortStrictRegex:/^(ene\.?|feb\.?|mar\.?|abr\.?|may\.?|jun\.?|jul\.?|ago\.?|sep\.?|oct\.?|nov\.?|dic\.?)/i,monthsParse:r,longMonthsParse:r,shortMonthsParse:r,weekdays:"domingo_lunes_martes_miércoles_jueves_viernes_sábado".split("_"),weekdaysShort:"dom._lun._mar._mié._jue._vie._sáb.".split("_"),weekdaysMin:"do_lu_ma_mi_ju_vi_sá".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"h:mm A",LTS:"h:mm:ss A",L:"DD/MM/YYYY",LL:"D [de] MMMM [de] YYYY",LLL:"D [de] MMMM [de] YYYY h:mm A",LLLL:"dddd, D [de] MMMM [de] YYYY h:mm A"},calendar:{sameDay:function(){return"[hoy a la"+(1!==this.hours()?"s":"")+"] LT"},nextDay:function(){return"[mañana a la"+(1!==this.hours()?"s":"")+"] LT"},nextWeek:function(){return"dddd [a la"+(1!==this.hours()?"s":"")+"] LT"},lastDay:function(){return"[ayer a la"+(1!==this.hours()?"s":"")+"] LT"},lastWeek:function(){return"[el] dddd [pasado a la"+(1!==this.hours()?"s":"")+"] LT"},sameElse:"L"},relativeTime:{future:"en %s",past:"hace %s",s:"unos segundos",ss:"%d segundos",m:"un minuto",mm:"%d minutos",h:"una hora",hh:"%d horas",d:"un día",dd:"%d días",w:"una semana",ww:"%d semanas",M:"un mes",MM:"%d meses",y:"un año",yy:"%d años"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:1,doy:4}});return o}))},"0a84":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ar-ma",{months:"يناير_فبراير_مارس_أبريل_ماي_يونيو_يوليوز_غشت_شتنبر_أكتوبر_نونبر_دجنبر".split("_"),monthsShort:"يناير_فبراير_مارس_أبريل_ماي_يونيو_يوليوز_غشت_شتنبر_أكتوبر_نونبر_دجنبر".split("_"),weekdays:"الأحد_الإثنين_الثلاثاء_الأربعاء_الخميس_الجمعة_السبت".split("_"),weekdaysShort:"احد_اثنين_ثلاثاء_اربعاء_خميس_جمعة_سبت".split("_"),weekdaysMin:"ح_ن_ث_ر_خ_ج_س".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[اليوم على الساعة] LT",nextDay:"[غدا على الساعة] LT",nextWeek:"dddd [على الساعة] LT",lastDay:"[أمس على الساعة] LT",lastWeek:"dddd [على الساعة] LT",sameElse:"L"},relativeTime:{future:"في %s",past:"منذ %s",s:"ثوان",ss:"%d ثانية",m:"دقيقة",mm:"%d دقائق",h:"ساعة",hh:"%d ساعات",d:"يوم",dd:"%d أيام",M:"شهر",MM:"%d أشهر",y:"سنة",yy:"%d سنوات"},week:{dow:1,doy:4}});return e}))},"0b07":function(t,e,n){var r=n("34ac"),i=n("3698");function o(t,e){var n=i(t,e);return r(n)?n:void 0}t.exports=o},"0b42":function(t,e,n){"use strict";var r=n("e8b5"),i=n("68ee"),o=n("861d"),a=n("b622"),s=a("species"),l=Array;t.exports=function(t){var e;return r(t)&&(e=t.constructor,i(e)&&(e===l||r(e.prototype))?e=void 0:o(e)&&(e=e[s],null===e&&(e=void 0))),void 0===e?l:e}},"0b43":function(t,e,n){"use strict";var r=n("04f8");t.exports=r&&!!Symbol["for"]&&!!Symbol.keyFor},"0b49":function(t,e,n){t.exports=n("f542")},"0caa":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,r){var i={s:["thoddea sekondamni","thodde sekond"],ss:[t+" sekondamni",t+" sekond"],m:["eka mintan","ek minut"],mm:[t+" mintamni",t+" mintam"],h:["eka voran","ek vor"],hh:[t+" voramni",t+" voram"],d:["eka disan","ek dis"],dd:[t+" disamni",t+" dis"],M:["eka mhoinean","ek mhoino"],MM:[t+" mhoineamni",t+" mhoine"],y:["eka vorsan","ek voros"],yy:[t+" vorsamni",t+" vorsam"]};return r?i[n][0]:i[n][1]}var n=t.defineLocale("gom-latn",{months:{standalone:"Janer_Febrer_Mars_Abril_Mai_Jun_Julai_Agost_Setembr_Otubr_Novembr_Dezembr".split("_"),format:"Janerachea_Febrerachea_Marsachea_Abrilachea_Maiachea_Junachea_Julaiachea_Agostachea_Setembrachea_Otubrachea_Novembrachea_Dezembrachea".split("_"),isFormat:/MMMM(\s)+D[oD]?/},monthsShort:"Jan._Feb._Mars_Abr._Mai_Jun_Jul._Ago._Set._Otu._Nov._Dez.".split("_"),monthsParseExact:!0,weekdays:"Aitar_Somar_Mongllar_Budhvar_Birestar_Sukrar_Son'var".split("_"),weekdaysShort:"Ait._Som._Mon._Bud._Bre._Suk._Son.".split("_"),weekdaysMin:"Ai_Sm_Mo_Bu_Br_Su_Sn".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"A h:mm [vazta]",LTS:"A h:mm:ss [vazta]",L:"DD-MM-YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY A h:mm [vazta]",LLLL:"dddd, MMMM Do, YYYY, A h:mm [vazta]",llll:"ddd, D MMM YYYY, A h:mm [vazta]"},calendar:{sameDay:"[Aiz] LT",nextDay:"[Faleam] LT",nextWeek:"[Fuddlo] dddd[,] LT",lastDay:"[Kal] LT",lastWeek:"[Fattlo] dddd[,] LT",sameElse:"L"},relativeTime:{future:"%s",past:"%s adim",s:e,ss:e,m:e,mm:e,h:e,hh:e,d:e,dd:e,M:e,MM:e,y:e,yy:e},dayOfMonthOrdinalParse:/\d{1,2}(er)/,ordinal:function(t,e){switch(e){case"D":return t+"er";default:case"M":case"Q":case"DDD":case"d":case"w":case"W":return t}},week:{dow:0,doy:3},meridiemParse:/rati|sokallim|donparam|sanje/,meridiemHour:function(t,e){return 12===t&&(t=0),"rati"===e?t<4?t:t+12:"sokallim"===e?t:"donparam"===e?t>12?t:t+12:"sanje"===e?t+12:void 0},meridiem:function(t,e,n){return t<4?"rati":t<12?"sokallim":t<16?"donparam":t<20?"sanje":"rati"}});return n}))},"0cfb":function(t,e,n){"use strict";var r=n("83ab"),i=n("d039"),o=n("cc12");t.exports=!r&&!i((function(){return 7!==Object.defineProperty(o("div"),"a",{get:function(){return 7}}).a}))},"0d24":function(t,e,n){(function(t){var r=n("2b3e"),i=n("07c7"),o=e&&!e.nodeType&&e,a=o&&"object"==typeof t&&t&&!t.nodeType&&t,s=a&&a.exports===o,l=s?r.Buffer:void 0,u=l?l.isBuffer:void 0,c=u||i;t.exports=c}).call(this,n("62e4")(t))},"0d26":function(t,e,n){"use strict";var r=n("e330"),i=Error,o=r("".replace),a=function(t){return String(new i(t).stack)}("zxcasd"),s=/\n\s*at [^:]*:[^\n]*/,l=s.test(a);t.exports=function(t,e){if(l&&"string"==typeof t&&!i.prepareStackTrace)while(e--)t=o(t,s,"");return t}},"0d51":function(t,e,n){"use strict";var r=String;t.exports=function(t){try{return r(t)}catch(e){return"Object"}}},"0e49":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("fr-ch",{months:"janvier_février_mars_avril_mai_juin_juillet_août_septembre_octobre_novembre_décembre".split("_"),monthsShort:"janv._févr._mars_avr._mai_juin_juil._août_sept._oct._nov._déc.".split("_"),monthsParseExact:!0,weekdays:"dimanche_lundi_mardi_mercredi_jeudi_vendredi_samedi".split("_"),weekdaysShort:"dim._lun._mar._mer._jeu._ven._sam.".split("_"),weekdaysMin:"di_lu_ma_me_je_ve_sa".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[Aujourd’hui à] LT",nextDay:"[Demain à] LT",nextWeek:"dddd [à] LT",lastDay:"[Hier à] LT",lastWeek:"dddd [dernier à] LT",sameElse:"L"},relativeTime:{future:"dans %s",past:"il y a %s",s:"quelques secondes",ss:"%d secondes",m:"une minute",mm:"%d minutes",h:"une heure",hh:"%d heures",d:"un jour",dd:"%d jours",M:"un mois",MM:"%d mois",y:"un an",yy:"%d ans"},dayOfMonthOrdinalParse:/\d{1,2}(er|e)/,ordinal:function(t,e){switch(e){default:case"M":case"Q":case"D":case"DDD":case"d":return t+(1===t?"er":"e");case"w":case"W":return t+(1===t?"re":"e")}},week:{dow:1,doy:4}});return e}))},"0e6b":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("en-au",{months:"January_February_March_April_May_June_July_August_September_October_November_December".split("_"),monthsShort:"Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec".split("_"),weekdays:"Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday".split("_"),weekdaysShort:"Sun_Mon_Tue_Wed_Thu_Fri_Sat".split("_"),weekdaysMin:"Su_Mo_Tu_We_Th_Fr_Sa".split("_"),longDateFormat:{LT:"h:mm A",LTS:"h:mm:ss A",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY h:mm A",LLLL:"dddd, D MMMM YYYY h:mm A"},calendar:{sameDay:"[Today at] LT",nextDay:"[Tomorrow at] LT",nextWeek:"dddd [at] LT",lastDay:"[Yesterday at] LT",lastWeek:"[Last] dddd [at] LT",sameElse:"L"},relativeTime:{future:"in %s",past:"%s ago",s:"a few seconds",ss:"%d seconds",m:"a minute",mm:"%d minutes",h:"an hour",hh:"%d hours",d:"a day",dd:"%d days",M:"a month",MM:"%d months",y:"a year",yy:"%d years"},dayOfMonthOrdinalParse:/\d{1,2}(st|nd|rd|th)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n},week:{dow:0,doy:4}});return e}))},"0e81":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"'inci",5:"'inci",8:"'inci",70:"'inci",80:"'inci",2:"'nci",7:"'nci",20:"'nci",50:"'nci",3:"'üncü",4:"'üncü",100:"'üncü",6:"'ncı",9:"'uncu",10:"'uncu",30:"'uncu",60:"'ıncı",90:"'ıncı"},n=t.defineLocale("tr",{months:"Ocak_Şubat_Mart_Nisan_Mayıs_Haziran_Temmuz_Ağustos_Eylül_Ekim_Kasım_Aralık".split("_"),monthsShort:"Oca_Şub_Mar_Nis_May_Haz_Tem_Ağu_Eyl_Eki_Kas_Ara".split("_"),weekdays:"Pazar_Pazartesi_Salı_Çarşamba_Perşembe_Cuma_Cumartesi".split("_"),weekdaysShort:"Paz_Pzt_Sal_Çar_Per_Cum_Cmt".split("_"),weekdaysMin:"Pz_Pt_Sa_Ça_Pe_Cu_Ct".split("_"),meridiem:function(t,e,n){return t<12?n?"öö":"ÖÖ":n?"ös":"ÖS"},meridiemParse:/öö|ÖÖ|ös|ÖS/,isPM:function(t){return"ös"===t||"ÖS"===t},longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[bugün saat] LT",nextDay:"[yarın saat] LT",nextWeek:"[gelecek] dddd [saat] LT",lastDay:"[dün] LT",lastWeek:"[geçen] dddd [saat] LT",sameElse:"L"},relativeTime:{future:"%s sonra",past:"%s önce",s:"birkaç saniye",ss:"%d saniye",m:"bir dakika",mm:"%d dakika",h:"bir saat",hh:"%d saat",d:"bir gün",dd:"%d gün",w:"bir hafta",ww:"%d hafta",M:"bir ay",MM:"%d ay",y:"bir yıl",yy:"%d yıl"},ordinal:function(t,n){switch(n){case"d":case"D":case"Do":case"DD":return t;default:if(0===t)return t+"'ıncı";var r=t%10,i=t%100-r,o=t>=100?100:null;return t+(e[r]||e[i]||e[o])}},week:{dow:1,doy:7}});return n}))},"0f0f":function(t,e,n){var r=n("8eeb"),i=n("9934");function o(t,e){return t&&r(e,i(e),t)}t.exports=o},"0f14":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("da",{months:"januar_februar_marts_april_maj_juni_juli_august_september_oktober_november_december".split("_"),monthsShort:"jan_feb_mar_apr_maj_jun_jul_aug_sep_okt_nov_dec".split("_"),weekdays:"søndag_mandag_tirsdag_onsdag_torsdag_fredag_lørdag".split("_"),weekdaysShort:"søn_man_tir_ons_tor_fre_lør".split("_"),weekdaysMin:"sø_ma_ti_on_to_fr_lø".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY HH:mm",LLLL:"dddd [d.] D. MMMM YYYY [kl.] HH:mm"},calendar:{sameDay:"[i dag kl.] LT",nextDay:"[i morgen kl.] LT",nextWeek:"på dddd [kl.] LT",lastDay:"[i går kl.] LT",lastWeek:"[i] dddd[s kl.] LT",sameElse:"L"},relativeTime:{future:"om %s",past:"%s siden",s:"få sekunder",ss:"%d sekunder",m:"et minut",mm:"%d minutter",h:"en time",hh:"%d timer",d:"en dag",dd:"%d dage",M:"en måned",MM:"%d måneder",y:"et år",yy:"%d år"},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return e}))},"0f38":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("tl-ph",{months:"Enero_Pebrero_Marso_Abril_Mayo_Hunyo_Hulyo_Agosto_Setyembre_Oktubre_Nobyembre_Disyembre".split("_"),monthsShort:"Ene_Peb_Mar_Abr_May_Hun_Hul_Ago_Set_Okt_Nob_Dis".split("_"),weekdays:"Linggo_Lunes_Martes_Miyerkules_Huwebes_Biyernes_Sabado".split("_"),weekdaysShort:"Lin_Lun_Mar_Miy_Huw_Biy_Sab".split("_"),weekdaysMin:"Li_Lu_Ma_Mi_Hu_Bi_Sab".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"MM/D/YYYY",LL:"MMMM D, YYYY",LLL:"MMMM D, YYYY HH:mm",LLLL:"dddd, MMMM DD, YYYY HH:mm"},calendar:{sameDay:"LT [ngayong araw]",nextDay:"[Bukas ng] LT",nextWeek:"LT [sa susunod na] dddd",lastDay:"LT [kahapon]",lastWeek:"LT [noong nakaraang] dddd",sameElse:"L"},relativeTime:{future:"sa loob ng %s",past:"%s ang nakalipas",s:"ilang segundo",ss:"%d segundo",m:"isang minuto",mm:"%d minuto",h:"isang oras",hh:"%d oras",d:"isang araw",dd:"%d araw",M:"isang buwan",MM:"%d buwan",y:"isang taon",yy:"%d taon"},dayOfMonthOrdinalParse:/\d{1,2}/,ordinal:function(t){return t},week:{dow:1,doy:4}});return e}))},"0f5c":function(t,e,n){var r=n("159a");function i(t,e,n){return null==t?t:r(t,e,n)}t.exports=i},"0ff2":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("eu",{months:"urtarrila_otsaila_martxoa_apirila_maiatza_ekaina_uztaila_abuztua_iraila_urria_azaroa_abendua".split("_"),monthsShort:"urt._ots._mar._api._mai._eka._uzt._abu._ira._urr._aza._abe.".split("_"),monthsParseExact:!0,weekdays:"igandea_astelehena_asteartea_asteazkena_osteguna_ostirala_larunbata".split("_"),weekdaysShort:"ig._al._ar._az._og._ol._lr.".split("_"),weekdaysMin:"ig_al_ar_az_og_ol_lr".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY-MM-DD",LL:"YYYY[ko] MMMM[ren] D[a]",LLL:"YYYY[ko] MMMM[ren] D[a] HH:mm",LLLL:"dddd, YYYY[ko] MMMM[ren] D[a] HH:mm",l:"YYYY-M-D",ll:"YYYY[ko] MMM D[a]",lll:"YYYY[ko] MMM D[a] HH:mm",llll:"ddd, YYYY[ko] MMM D[a] HH:mm"},calendar:{sameDay:"[gaur] LT[etan]",nextDay:"[bihar] LT[etan]",nextWeek:"dddd LT[etan]",lastDay:"[atzo] LT[etan]",lastWeek:"[aurreko] dddd LT[etan]",sameElse:"L"},relativeTime:{future:"%s barru",past:"duela %s",s:"segundo batzuk",ss:"%d segundo",m:"minutu bat",mm:"%d minutu",h:"ordu bat",hh:"%d ordu",d:"egun bat",dd:"%d egun",M:"hilabete bat",MM:"%d hilabete",y:"urte bat",yy:"%d urte"},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:7}});return e}))},"100e":function(t,e,n){var r=n("cd9d"),i=n("2286"),o=n("c1c9");function a(t,e){return o(i(t,e,r),t+"")}t.exports=a},1041:function(t,e,n){var r=n("8eeb"),i=n("a029");function o(t,e){return r(t,i(t),e)}t.exports=o},"107c":function(t,e,n){"use strict";var r=n("d039"),i=n("da84"),o=i.RegExp;t.exports=r((function(){var t=o("(?<a>b)","g");return"b"!==t.exec("b").groups.a||"bc"!=="b".replace(t,"$<a>c")}))},"10e8":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("th",{months:"มกราคม_กุมภาพันธ์_มีนาคม_เมษายน_พฤษภาคม_มิถุนายน_กรกฎาคม_สิงหาคม_กันยายน_ตุลาคม_พฤศจิกายน_ธันวาคม".split("_"),monthsShort:"ม.ค._ก.พ._มี.ค._เม.ย._พ.ค._มิ.ย._ก.ค._ส.ค._ก.ย._ต.ค._พ.ย._ธ.ค.".split("_"),monthsParseExact:!0,weekdays:"อาทิตย์_จันทร์_อังคาร_พุธ_พฤหัสบดี_ศุกร์_เสาร์".split("_"),weekdaysShort:"อาทิตย์_จันทร์_อังคาร_พุธ_พฤหัส_ศุกร์_เสาร์".split("_"),weekdaysMin:"อา._จ._อ._พ._พฤ._ศ._ส.".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY เวลา H:mm",LLLL:"วันddddที่ D MMMM YYYY เวลา H:mm"},meridiemParse:/ก่อนเที่ยง|หลังเที่ยง/,isPM:function(t){return"หลังเที่ยง"===t},meridiem:function(t,e,n){return t<12?"ก่อนเที่ยง":"หลังเที่ยง"},calendar:{sameDay:"[วันนี้ เวลา] LT",nextDay:"[พรุ่งนี้ เวลา] LT",nextWeek:"dddd[หน้า เวลา] LT",lastDay:"[เมื่อวานนี้ เวลา] LT",lastWeek:"[วัน]dddd[ที่แล้ว เวลา] LT",sameElse:"L"},relativeTime:{future:"อีก %s",past:"%sที่แล้ว",s:"ไม่กี่วินาที",ss:"%d วินาที",m:"1 นาที",mm:"%d นาที",h:"1 ชั่วโมง",hh:"%d ชั่วโมง",d:"1 วัน",dd:"%d วัน",w:"1 สัปดาห์",ww:"%d สัปดาห์",M:"1 เดือน",MM:"%d เดือน",y:"1 ปี",yy:"%d ปี"}});return e}))},1148:function(t,e,n){"use strict";var r=n("5926"),i=n("577e"),o=n("1d80"),a=RangeError;t.exports=function(t){var e=i(o(this)),n="",s=r(t);if(s<0||s===1/0)throw new a("Wrong number of repetitions");for(;s>0;(s>>>=1)&&(e+=e))1&s&&(n+=e);return n}},1290:function(t,e){function n(t){var e=typeof t;return"string"==e||"number"==e||"symbol"==e||"boolean"==e?"__proto__"!==t:null===t}t.exports=n},1304:function(t,e,n){var r=n("cd9d");function i(t){return"function"==typeof t?t:r}t.exports=i},"130f":function(t,e,n){"use strict";n("e20c"),n("0402")},1310:function(t,e){function n(t){return null!=t&&"object"==typeof t}t.exports=n},1368:function(t,e,n){var r=n("da03"),i=function(){var t=/[^.]+$/.exec(r&&r.keys&&r.keys.IE_PROTO||"");return t?"Symbol(src)_1."+t:""}();function o(t){return!!i&&i in t}t.exports=o},"13d2":function(t,e,n){"use strict";var r=n("e330"),i=n("d039"),o=n("1626"),a=n("1a2d"),s=n("83ab"),l=n("5e77").CONFIGURABLE,u=n("8925"),c=n("69f3"),d=c.enforce,h=c.get,f=String,p=Object.defineProperty,m=r("".slice),g=r("".replace),y=r([].join),v=s&&!i((function(){return 8!==p((function(){}),"length",{value:8}).length})),_=String(String).split("String"),b=t.exports=function(t,e,n){"Symbol("===m(f(e),0,7)&&(e="["+g(f(e),/^Symbol\(([^)]*)\).*$/,"$1")+"]"),n&&n.getter&&(e="get "+e),n&&n.setter&&(e="set "+e),(!a(t,"name")||l&&t.name!==e)&&(s?p(t,"name",{value:e,configurable:!0}):t.name=e),v&&n&&a(n,"arity")&&t.length!==n.arity&&p(t,"length",{value:n.arity});try{n&&a(n,"constructor")&&n.constructor?s&&p(t,"prototype",{writable:!1}):t.prototype&&(t.prototype=void 0)}catch(i){}var r=d(t);return a(r,"source")||(r.source=y(_,"string"==typeof e?e:"")),t};Function.prototype.toString=b((function(){return o(this)&&h(this).source||u(this)}),"toString")},"13e9":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={words:{ss:["секунда","секунде","секунди"],m:["један минут","једног минута"],mm:["минут","минута","минута"],h:["један сат","једног сата"],hh:["сат","сата","сати"],d:["један дан","једног дана"],dd:["дан","дана","дана"],M:["један месец","једног месеца"],MM:["месец","месеца","месеци"],y:["једну годину","једне године"],yy:["годину","године","година"]},correctGrammaticalCase:function(t,e){return t%10>=1&&t%10<=4&&(t%100<10||t%100>=20)?t%10===1?e[0]:e[1]:e[2]},translate:function(t,n,r,i){var o,a=e.words[r];return 1===r.length?"y"===r&&n?"једна година":i||n?a[0]:a[1]:(o=e.correctGrammaticalCase(t,a),"yy"===r&&n&&"годину"===o?t+" година":t+" "+o)}},n=t.defineLocale("sr-cyrl",{months:"јануар_фебруар_март_април_мај_јун_јул_август_септембар_октобар_новембар_децембар".split("_"),monthsShort:"јан._феб._мар._апр._мај_јун_јул_авг._сеп._окт._нов._дец.".split("_"),monthsParseExact:!0,weekdays:"недеља_понедељак_уторак_среда_четвртак_петак_субота".split("_"),weekdaysShort:"нед._пон._уто._сре._чет._пет._суб.".split("_"),weekdaysMin:"не_по_ут_ср_че_пе_су".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"D. M. YYYY.",LL:"D. MMMM YYYY.",LLL:"D. MMMM YYYY. H:mm",LLLL:"dddd, D. MMMM YYYY. H:mm"},calendar:{sameDay:"[данас у] LT",nextDay:"[сутра у] LT",nextWeek:function(){switch(this.day()){case 0:return"[у] [недељу] [у] LT";case 3:return"[у] [среду] [у] LT";case 6:return"[у] [суботу] [у] LT";case 1:case 2:case 4:case 5:return"[у] dddd [у] LT"}},lastDay:"[јуче у] LT",lastWeek:function(){var t=["[прошле] [недеље] [у] LT","[прошлог] [понедељка] [у] LT","[прошлог] [уторка] [у] LT","[прошле] [среде] [у] LT","[прошлог] [четвртка] [у] LT","[прошлог] [петка] [у] LT","[прошле] [суботе] [у] LT"];return t[this.day()]},sameElse:"L"},relativeTime:{future:"за %s",past:"пре %s",s:"неколико секунди",ss:e.translate,m:e.translate,mm:e.translate,h:e.translate,hh:e.translate,d:e.translate,dd:e.translate,M:e.translate,MM:e.translate,y:e.translate,yy:e.translate},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:7}});return n}))},"14c3":function(t,e,n){"use strict";var r=n("c65b"),i=n("825a"),o=n("1626"),a=n("c6b6"),s=n("9263"),l=TypeError;t.exports=function(t,e){var n=t.exec;if(o(n)){var u=r(n,t,e);return null!==u&&i(u),u}if("RegExp"===a(t))return r(s,t,e);throw new l("RegExp#exec called on incompatible receiver")}},"14d9":function(t,e,n){"use strict";var r=n("23e7"),i=n("7b0b"),o=n("07fa"),a=n("3a34"),s=n("3511"),l=n("d039"),u=l((function(){return 4294967297!==[].push.call({length:4294967296},1)})),c=function(){try{Object.defineProperty([],"length",{writable:!1}).push()}catch(t){return t instanceof TypeError}},d=u||!c();r({target:"Array",proto:!0,arity:1,forced:d},{push:function(t){var e=i(this),n=o(e),r=arguments.length;s(n+r);for(var l=0;l<r;l++)e[n]=arguments[l],n++;return a(e,n),n}})},"14e5":function(t,e,n){"use strict";var r=n("23e7"),i=n("c65b"),o=n("59ed"),a=n("f069"),s=n("e667"),l=n("2266"),u=n("5eed");r({target:"Promise",stat:!0,forced:u},{all:function(t){var e=this,n=a.f(e),r=n.resolve,u=n.reject,c=s((function(){var n=o(e.resolve),a=[],s=0,c=1;l(t,(function(t){var o=s++,l=!1;c++,i(n,e,t).then((function(t){l||(l=!0,a[o]=t,--c||r(a))}),u)})),--c||r(a)}));return c.error&&u(c.value),n.promise}})},"157a":function(t,e,n){"use strict";var r=n("da84"),i=n("83ab"),o=Object.getOwnPropertyDescriptor;t.exports=function(t){if(!i)return r[t];var e=o(r,t);return e&&e.value}},"159a":function(t,e,n){var r=n("32b3"),i=n("e2e4"),o=n("c098"),a=n("1a8c"),s=n("f4d6");function l(t,e,n,l){if(!a(t))return t;e=i(e,t);var u=-1,c=e.length,d=c-1,h=t;while(null!=h&&++u<c){var f=s(e[u]),p=n;if("__proto__"===f||"constructor"===f||"prototype"===f)return t;if(u!=d){var m=h[f];p=l?l(m,f,h):void 0,void 0===p&&(p=a(m)?m:o(e[u+1])?[]:{})}r(h,f,p),h=h[f]}return t}t.exports=l},"159b":function(t,e,n){"use strict";var r=n("da84"),i=n("fdbc"),o=n("785a"),a=n("17c2"),s=n("9112"),l=function(t){if(t&&t.forEach!==a)try{s(t,"forEach",a)}catch(e){t.forEach=a}};for(var u in i)i[u]&&l(r[u]&&r[u].prototype);l(o)},"15f3":function(t,e,n){var r=n("89d9"),i=n("8604");function o(t,e){return r(t,e,(function(e,n){return i(t,n)}))}t.exports=o},1626:function(t,e,n){"use strict";var r="object"==typeof document&&document.all;t.exports="undefined"==typeof r&&void 0!==r?function(t){return"function"==typeof t||t===r}:function(t){return"function"==typeof t}},"167b":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("oc-lnc",{months:{standalone:"genièr_febrièr_març_abril_mai_junh_julhet_agost_setembre_octòbre_novembre_decembre".split("_"),format:"de genièr_de febrièr_de març_d'abril_de mai_de junh_de julhet_d'agost_de setembre_d'octòbre_de novembre_de decembre".split("_"),isFormat:/D[oD]?(\s)+MMMM/},monthsShort:"gen._febr._març_abr._mai_junh_julh._ago._set._oct._nov._dec.".split("_"),monthsParseExact:!0,weekdays:"dimenge_diluns_dimars_dimècres_dijòus_divendres_dissabte".split("_"),weekdaysShort:"dg._dl._dm._dc._dj._dv._ds.".split("_"),weekdaysMin:"dg_dl_dm_dc_dj_dv_ds".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM [de] YYYY",ll:"D MMM YYYY",LLL:"D MMMM [de] YYYY [a] H:mm",lll:"D MMM YYYY, H:mm",LLLL:"dddd D MMMM [de] YYYY [a] H:mm",llll:"ddd D MMM YYYY, H:mm"},calendar:{sameDay:"[uèi a] LT",nextDay:"[deman a] LT",nextWeek:"dddd [a] LT",lastDay:"[ièr a] LT",lastWeek:"dddd [passat a] LT",sameElse:"L"},relativeTime:{future:"d'aquí %s",past:"fa %s",s:"unas segondas",ss:"%d segondas",m:"una minuta",mm:"%d minutas",h:"una ora",hh:"%d oras",d:"un jorn",dd:"%d jorns",M:"un mes",MM:"%d meses",y:"un an",yy:"%d ans"},dayOfMonthOrdinalParse:/\d{1,2}(r|n|t|è|a)/,ordinal:function(t,e){var n=1===t?"r":2===t?"n":3===t?"r":4===t?"t":"è";return"w"!==e&&"W"!==e||(n="a"),t+n},week:{dow:1,doy:4}});return e}))},1693:function(t,e,n){var r=n("afb9"),i=n("9934");function o(t){return null==t?[]:r(t,i(t))}t.exports=o},1766:function(t,e,n){"use strict";e["a"]={name:"RenderCell",functional:!0,props:{render:Function},render:(t,e)=>e.props.render(t)}},1787:function(t,e,n){"use strict";var r=n("861d");t.exports=function(t){return r(t)||null===t}},"17c2":function(t,e,n){"use strict";var r=n("b727").forEach,i=n("a640"),o=i("forEach");t.exports=o?[].forEach:function(t){return r(this,t,arguments.length>1?arguments[1]:void 0)}},1838:function(t,e,n){var r=n("c05f"),i=n("9b02"),o=n("8604"),a=n("f608"),s=n("08cc"),l=n("20ec"),u=n("f4d6"),c=1,d=2;function h(t,e){return a(t)&&s(e)?l(u(t),e):function(n){var a=i(n,t);return void 0===a&&a===e?o(n,t):r(e,a,c|d)}}t.exports=h},"18d2":function(t,e,n){"use strict";var r=n("18e9");t.exports=function(t){t=t||{};var e=t.reporter,n=t.batchProcessor,i=t.stateHandler.getState;if(!e)throw new Error("Missing required dependency: reporter.");function o(t,e){function n(){e(t)}if(r.isIE(8))i(t).object={proxy:n},t.attachEvent("onresize",n);else{var o=l(t);if(!o)throw new Error("Element is not detectable by this strategy.");o.contentDocument.defaultView.addEventListener("resize",n)}}function a(e){var n=t.important?" !important; ":"; ";return(e.join(n)+n).trim()}function s(t,o,s){s||(s=o,o=t,t=null),t=t||{};t.debug;function l(o,s){var l=a(["display: block","position: absolute","top: 0","left: 0","width: 100%","height: 100%","border: none","padding: 0","margin: 0","opacity: 0","z-index: -1000","pointer-events: none"]),u=!1,c=window.getComputedStyle(o),d=o.offsetWidth,h=o.offsetHeight;function f(){function n(){if("static"===c.position){o.style.setProperty("position","relative",t.important?"important":"");var n=function(e,n,r,i){function o(t){return t.replace(/[^-\d\.]/g,"")}var a=r[i];"auto"!==a&&"0"!==o(a)&&(e.warn("An element that is positioned static has style."+i+"="+a+" which is ignored due to the static positioning. The element will need to be positioned relative, so the style."+i+" will be set to 0. Element: ",n),n.style.setProperty(i,"0",t.important?"important":""))};n(e,o,c,"top"),n(e,o,c,"right"),n(e,o,c,"bottom"),n(e,o,c,"left")}}function a(){function t(e,n){if(!e.contentDocument){var r=i(e);return r.checkForObjectDocumentTimeoutId&&window.clearTimeout(r.checkForObjectDocumentTimeoutId),void(r.checkForObjectDocumentTimeoutId=setTimeout((function(){r.checkForObjectDocumentTimeoutId=0,t(e,n)}),100))}n(e.contentDocument)}u||n();var e=this;t(e,(function(t){s(o)}))}""!==c.position&&(n(c),u=!0);var d=document.createElement("object");d.style.cssText=l,d.tabIndex=-1,d.type="text/html",d.setAttribute("aria-hidden","true"),d.onload=a,r.isIE()||(d.data="about:blank"),i(o)&&(o.appendChild(d),i(o).object=d,r.isIE()&&(d.data="about:blank"))}i(o).startSize={width:d,height:h},n?n.add(f):f()}r.isIE(8)?s(o):l(o,s)}function l(t){return i(t).object}function u(t){if(i(t)){var e=l(t);e&&(r.isIE(8)?t.detachEvent("onresize",e.proxy):t.removeChild(e),i(t).checkForObjectDocumentTimeoutId&&window.clearTimeout(i(t).checkForObjectDocumentTimeoutId),delete i(t).object)}}return{makeDetectable:s,addListener:o,uninstall:u}}},"18d8":function(t,e,n){var r=n("234d"),i=/[^.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(?:\.|\[\])(?:\.|\[\]|$))/g,o=/\\(\\)?/g,a=r((function(t){var e=[];return 46===t.charCodeAt(0)&&e.push(""),t.replace(i,(function(t,n,r,i){e.push(r?i.replace(o,"$1"):n||t)})),e}));t.exports=a},"18e9":function(t,e,n){"use strict";var r=t.exports={};r.isIE=function(t){function e(){var t=navigator.userAgent.toLowerCase();return-1!==t.indexOf("msie")||-1!==t.indexOf("trident")||-1!==t.indexOf(" edge/")}if(!e())return!1;if(!t)return!0;var n=function(){var t,e=3,n=document.createElement("div"),r=n.getElementsByTagName("i");do{n.innerHTML="\x3c!--[if gt IE "+ ++e+"]><i></i><![endif]--\x3e"}while(r[0]);return e>4?e:t}();return t===n},r.isLegacyOpera=function(){return!!window.opera}},"18f6":function(t,e,n){var r=n("30e0"),i=n("1304"),o=n("9934");function a(t,e){return null==t?t:r(t,i(e),o)}t.exports=a},"193b":function(t,e,n){var r=n("159a");function i(t,e,n,i){return i="function"==typeof i?i:void 0,null==t?t:r(t,e,n,i)}t.exports=i},"19aa":function(t,e,n){"use strict";var r=n("3a9b"),i=TypeError;t.exports=function(t,e){if(r(e,t))return t;throw new i("Incorrect invocation")}},"1a2d":function(t,e,n){"use strict";var r=n("e330"),i=n("7b0b"),o=r({}.hasOwnProperty);t.exports=Object.hasOwn||function(t,e){return o(i(t),e)}},"1a2d0":function(t,e,n){var r=n("42a2"),i=n("1310"),o="[object Map]";function a(t){return i(t)&&r(t)==o}t.exports=a},"1a8c":function(t,e){function n(t){var e=typeof t;return null!=t&&("object"==e||"function"==e)}t.exports=n},"1b45":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("mt",{months:"Jannar_Frar_Marzu_April_Mejju_Ġunju_Lulju_Awwissu_Settembru_Ottubru_Novembru_Diċembru".split("_"),monthsShort:"Jan_Fra_Mar_Apr_Mej_Ġun_Lul_Aww_Set_Ott_Nov_Diċ".split("_"),weekdays:"Il-Ħadd_It-Tnejn_It-Tlieta_L-Erbgħa_Il-Ħamis_Il-Ġimgħa_Is-Sibt".split("_"),weekdaysShort:"Ħad_Tne_Tli_Erb_Ħam_Ġim_Sib".split("_"),weekdaysMin:"Ħa_Tn_Tl_Er_Ħa_Ġi_Si".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Illum fil-]LT",nextDay:"[Għada fil-]LT",nextWeek:"dddd [fil-]LT",lastDay:"[Il-bieraħ fil-]LT",lastWeek:"dddd [li għadda] [fil-]LT",sameElse:"L"},relativeTime:{future:"f’ %s",past:"%s ilu",s:"ftit sekondi",ss:"%d sekondi",m:"minuta",mm:"%d minuti",h:"siegħa",hh:"%d siegħat",d:"ġurnata",dd:"%d ġranet",M:"xahar",MM:"%d xhur",y:"sena",yy:"%d sni"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:1,doy:4}});return e}))},"1bac":function(t,e,n){var r=n("7d1f"),i=n("a029"),o=n("9934");function a(t){return r(t,o,i)}t.exports=a},"1be4":function(t,e,n){"use strict";var r=n("d066");t.exports=r("document","documentElement")},"1c3c":function(t,e,n){var r=n("9e69"),i=n("2474"),o=n("9638"),a=n("a2be"),s=n("edfa"),l=n("ac41"),u=1,c=2,d="[object Boolean]",h="[object Date]",f="[object Error]",p="[object Map]",m="[object Number]",g="[object RegExp]",y="[object Set]",v="[object String]",_="[object Symbol]",b="[object ArrayBuffer]",x="[object DataView]",w=r?r.prototype:void 0,M=w?w.valueOf:void 0;function S(t,e,n,r,w,S,k){switch(n){case x:if(t.byteLength!=e.byteLength||t.byteOffset!=e.byteOffset)return!1;t=t.buffer,e=e.buffer;case b:return!(t.byteLength!=e.byteLength||!S(new i(t),new i(e)));case d:case h:case m:return o(+t,+e);case f:return t.name==e.name&&t.message==e.message;case g:case v:return t==e+"";case p:var L=s;case y:var T=r&u;if(L||(L=l),t.size!=e.size&&!T)return!1;var D=k.get(t);if(D)return D==e;r|=c,k.set(t,e);var O=a(L(t),L(e),r,w,S,k);return k["delete"](t),O;case _:if(M)return M.call(t)==M.call(e)}return!1}t.exports=S},"1c7e":function(t,e,n){"use strict";var r=n("b622"),i=r("iterator"),o=!1;try{var a=0,s={next:function(){return{done:!!a++}},return:function(){o=!0}};s[i]=function(){return this},Array.from(s,(function(){throw 2}))}catch(l){}t.exports=function(t,e){try{if(!e&&!o)return!1}catch(l){return!1}var n=!1;try{var r={};r[i]=function(){return{next:function(){return{done:n=!0}}}},t(r)}catch(l){}return n}},"1cdc":function(t,e,n){"use strict";var r=n("342f");t.exports=/(?:ipad|iphone|ipod).*applewebkit/i.test(r)},"1cec":function(t,e,n){var r=n("0b07"),i=n("2b3e"),o=r(i,"Promise");t.exports=o},"1cfd":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"1",2:"2",3:"3",4:"4",5:"5",6:"6",7:"7",8:"8",9:"9",0:"0"},n=function(t){return 0===t?0:1===t?1:2===t?2:t%100>=3&&t%100<=10?3:t%100>=11?4:5},r={s:["أقل من ثانية","ثانية واحدة",["ثانيتان","ثانيتين"],"%d ثوان","%d ثانية","%d ثانية"],m:["أقل من دقيقة","دقيقة واحدة",["دقيقتان","دقيقتين"],"%d دقائق","%d دقيقة","%d دقيقة"],h:["أقل من ساعة","ساعة واحدة",["ساعتان","ساعتين"],"%d ساعات","%d ساعة","%d ساعة"],d:["أقل من يوم","يوم واحد",["يومان","يومين"],"%d أيام","%d يومًا","%d يوم"],M:["أقل من شهر","شهر واحد",["شهران","شهرين"],"%d أشهر","%d شهرا","%d شهر"],y:["أقل من عام","عام واحد",["عامان","عامين"],"%d أعوام","%d عامًا","%d عام"]},i=function(t){return function(e,i,o,a){var s=n(e),l=r[t][n(e)];return 2===s&&(l=l[i?0:1]),l.replace(/%d/i,e)}},o=["يناير","فبراير","مارس","أبريل","مايو","يونيو","يوليو","أغسطس","سبتمبر","أكتوبر","نوفمبر","ديسمبر"],a=t.defineLocale("ar-ly",{months:o,monthsShort:o,weekdays:"الأحد_الإثنين_الثلاثاء_الأربعاء_الخميس_الجمعة_السبت".split("_"),weekdaysShort:"أحد_إثنين_ثلاثاء_أربعاء_خميس_جمعة_سبت".split("_"),weekdaysMin:"ح_ن_ث_ر_خ_ج_س".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"D/‏M/‏YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},meridiemParse:/ص|م/,isPM:function(t){return"م"===t},meridiem:function(t,e,n){return t<12?"ص":"م"},calendar:{sameDay:"[اليوم عند الساعة] LT",nextDay:"[غدًا عند الساعة] LT",nextWeek:"dddd [عند الساعة] LT",lastDay:"[أمس عند الساعة] LT",lastWeek:"dddd [عند الساعة] LT",sameElse:"L"},relativeTime:{future:"بعد %s",past:"منذ %s",s:i("s"),ss:i("s"),m:i("m"),mm:i("m"),h:i("h"),hh:i("h"),d:i("d"),dd:i("d"),M:i("M"),MM:i("M"),y:i("y"),yy:i("y")},preparse:function(t){return t.replace(/،/g,",")},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]})).replace(/,/g,"،")},week:{dow:6,doy:12}});return a}))},"1d80":function(t,e,n){"use strict";var r=n("7234"),i=TypeError;t.exports=function(t){if(r(t))throw new i("Can't call method on "+t);return t}},"1dde":function(t,e,n){"use strict";var r=n("d039"),i=n("b622"),o=n("2d00"),a=i("species");t.exports=function(t){return o>=51||!r((function(){var e=[],n=e.constructor={};return n[a]=function(){return{foo:1}},1!==e[t](Boolean).foo}))}},"1efc":function(t,e){function n(t){var e=this.has(t)&&delete this.__data__[t];return this.size-=e?1:0,e}t.exports=n},"1fb5":function(t,e,n){"use strict";e.byteLength=c,e.toByteArray=h,e.fromByteArray=m;for(var r=[],i=[],o="undefined"!==typeof Uint8Array?Uint8Array:Array,a="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/",s=0,l=a.length;s<l;++s)r[s]=a[s],i[a.charCodeAt(s)]=s;function u(t){var e=t.length;if(e%4>0)throw new Error("Invalid string. Length must be a multiple of 4");var n=t.indexOf("=");-1===n&&(n=e);var r=n===e?0:4-n%4;return[n,r]}function c(t){var e=u(t),n=e[0],r=e[1];return 3*(n+r)/4-r}function d(t,e,n){return 3*(e+n)/4-n}function h(t){var e,n,r=u(t),a=r[0],s=r[1],l=new o(d(t,a,s)),c=0,h=s>0?a-4:a;for(n=0;n<h;n+=4)e=i[t.charCodeAt(n)]<<18|i[t.charCodeAt(n+1)]<<12|i[t.charCodeAt(n+2)]<<6|i[t.charCodeAt(n+3)],l[c++]=e>>16&255,l[c++]=e>>8&255,l[c++]=255&e;return 2===s&&(e=i[t.charCodeAt(n)]<<2|i[t.charCodeAt(n+1)]>>4,l[c++]=255&e),1===s&&(e=i[t.charCodeAt(n)]<<10|i[t.charCodeAt(n+1)]<<4|i[t.charCodeAt(n+2)]>>2,l[c++]=e>>8&255,l[c++]=255&e),l}function f(t){return r[t>>18&63]+r[t>>12&63]+r[t>>6&63]+r[63&t]}function p(t,e,n){for(var r,i=[],o=e;o<n;o+=3)r=(t[o]<<16&16711680)+(t[o+1]<<8&65280)+(255&t[o+2]),i.push(f(r));return i.join("")}function m(t){for(var e,n=t.length,i=n%3,o=[],a=16383,s=0,l=n-i;s<l;s+=a)o.push(p(t,s,s+a>l?l:s+a));return 1===i?(e=t[n-1],o.push(r[e>>2]+r[e<<4&63]+"==")):2===i&&(e=(t[n-2]<<8)+t[n-1],o.push(r[e>>10]+r[e>>4&63]+r[e<<2&63]+"=")),o.join("")}i["-".charCodeAt(0)]=62,i["_".charCodeAt(0)]=63},"1fc1":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e){var n=t.split("_");return e%10===1&&e%100!==11?n[0]:e%10>=2&&e%10<=4&&(e%100<10||e%100>=20)?n[1]:n[2]}function n(t,n,r){var i={ss:n?"секунда_секунды_секунд":"секунду_секунды_секунд",mm:n?"хвіліна_хвіліны_хвілін":"хвіліну_хвіліны_хвілін",hh:n?"гадзіна_гадзіны_гадзін":"гадзіну_гадзіны_гадзін",dd:"дзень_дні_дзён",MM:"месяц_месяцы_месяцаў",yy:"год_гады_гадоў"};return"m"===r?n?"хвіліна":"хвіліну":"h"===r?n?"гадзіна":"гадзіну":t+" "+e(i[r],+t)}var r=t.defineLocale("be",{months:{format:"студзеня_лютага_сакавіка_красавіка_траўня_чэрвеня_ліпеня_жніўня_верасня_кастрычніка_лістапада_снежня".split("_"),standalone:"студзень_люты_сакавік_красавік_травень_чэрвень_ліпень_жнівень_верасень_кастрычнік_лістапад_снежань".split("_")},monthsShort:"студ_лют_сак_крас_трав_чэрв_ліп_жнів_вер_каст_ліст_снеж".split("_"),weekdays:{format:"нядзелю_панядзелак_аўторак_сераду_чацвер_пятніцу_суботу".split("_"),standalone:"нядзеля_панядзелак_аўторак_серада_чацвер_пятніца_субота".split("_"),isFormat:/\[ ?[Ууў] ?(?:мінулую|наступную)? ?\] ?dddd/},weekdaysShort:"нд_пн_ат_ср_чц_пт_сб".split("_"),weekdaysMin:"нд_пн_ат_ср_чц_пт_сб".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY г.",LLL:"D MMMM YYYY г., HH:mm",LLLL:"dddd, D MMMM YYYY г., HH:mm"},calendar:{sameDay:"[Сёння ў] LT",nextDay:"[Заўтра ў] LT",lastDay:"[Учора ў] LT",nextWeek:function(){return"[У] dddd [ў] LT"},lastWeek:function(){switch(this.day()){case 0:case 3:case 5:case 6:return"[У мінулую] dddd [ў] LT";case 1:case 2:case 4:return"[У мінулы] dddd [ў] LT"}},sameElse:"L"},relativeTime:{future:"праз %s",past:"%s таму",s:"некалькі секунд",m:n,mm:n,h:n,hh:n,d:"дзень",dd:n,M:"месяц",MM:n,y:"год",yy:n},meridiemParse:/ночы|раніцы|дня|вечара/,isPM:function(t){return/^(дня|вечара)$/.test(t)},meridiem:function(t,e,n){return t<4?"ночы":t<12?"раніцы":t<17?"дня":"вечара"},dayOfMonthOrdinalParse:/\d{1,2}-(і|ы|га)/,ordinal:function(t,e){switch(e){case"M":case"d":case"DDD":case"w":case"W":return t%10!==2&&t%10!==3||t%100===12||t%100===13?t+"-ы":t+"-і";case"D":return t+"-га";default:return t}},week:{dow:1,doy:7}});return r}))},"1fc8":function(t,e,n){var r=n("4245");function i(t,e){var n=r(this,t),i=n.size;return n.set(t,e),this.size+=n.size==i?0:1,this}t.exports=i},"201b":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ka",{months:"იანვარი_თებერვალი_მარტი_აპრილი_მაისი_ივნისი_ივლისი_აგვისტო_სექტემბერი_ოქტომბერი_ნოემბერი_დეკემბერი".split("_"),monthsShort:"იან_თებ_მარ_აპრ_მაი_ივნ_ივლ_აგვ_სექ_ოქტ_ნოე_დეკ".split("_"),weekdays:{standalone:"კვირა_ორშაბათი_სამშაბათი_ოთხშაბათი_ხუთშაბათი_პარასკევი_შაბათი".split("_"),format:"კვირას_ორშაბათს_სამშაბათს_ოთხშაბათს_ხუთშაბათს_პარასკევს_შაბათს".split("_"),isFormat:/(წინა|შემდეგ)/},weekdaysShort:"კვი_ორშ_სამ_ოთხ_ხუთ_პარ_შაბ".split("_"),weekdaysMin:"კვ_ორ_სა_ოთ_ხუ_პა_შა".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[დღეს] LT[-ზე]",nextDay:"[ხვალ] LT[-ზე]",lastDay:"[გუშინ] LT[-ზე]",nextWeek:"[შემდეგ] dddd LT[-ზე]",lastWeek:"[წინა] dddd LT-ზე",sameElse:"L"},relativeTime:{future:function(t){return t.replace(/(წამ|წუთ|საათ|წელ|დღ|თვ)(ი|ე)/,(function(t,e,n){return"ი"===n?e+"ში":e+n+"ში"}))},past:function(t){return/(წამი|წუთი|საათი|დღე|თვე)/.test(t)?t.replace(/(ი|ე)$/,"ის წინ"):/წელი/.test(t)?t.replace(/წელი$/,"წლის წინ"):t},s:"რამდენიმე წამი",ss:"%d წამი",m:"წუთი",mm:"%d წუთი",h:"საათი",hh:"%d საათი",d:"დღე",dd:"%d დღე",M:"თვე",MM:"%d თვე",y:"წელი",yy:"%d წელი"},dayOfMonthOrdinalParse:/0|1-ლი|მე-\d{1,2}|\d{1,2}-ე/,ordinal:function(t){return 0===t?t:1===t?t+"-ლი":t<20||t<=100&&t%20===0||t%100===0?"მე-"+t:t+"-ე"},week:{dow:1,doy:7}});return e}))},"20cc":function(t,e,n){"use strict";var r=n("da84"),i=n("2ba4"),o=n("1626"),a=n("c6a7"),s=n("342f"),l=n("f36a"),u=n("d6d6"),c=r.Function,d=/MSIE .\./.test(s)||a&&function(){var t=r.Bun.version.split(".");return t.length<3||"0"===t[0]&&(t[1]<3||"3"===t[1]&&"0"===t[2])}();t.exports=function(t,e){var n=e?2:1;return d?function(r,a){var s=u(arguments.length,1)>n,d=o(r)?r:c(r),h=s?l(arguments,n):[],f=s?function(){i(d,this,h)}:d;return e?t(f,a):t(f)}:t}},"20ec":function(t,e){function n(t,e){return function(n){return null!=n&&(n[t]===e&&(void 0!==e||t in Object(n)))}}t.exports=n},2266:function(t,e,n){"use strict";var r=n("0366"),i=n("c65b"),o=n("825a"),a=n("0d51"),s=n("e95a"),l=n("07fa"),u=n("3a9b"),c=n("9a1f"),d=n("35a1"),h=n("2a62"),f=TypeError,p=function(t,e){this.stopped=t,this.result=e},m=p.prototype;t.exports=function(t,e,n){var g,y,v,_,b,x,w,M=n&&n.that,S=!(!n||!n.AS_ENTRIES),k=!(!n||!n.IS_RECORD),L=!(!n||!n.IS_ITERATOR),T=!(!n||!n.INTERRUPTED),D=r(e,M),O=function(t){return g&&h(g,"normal",t),new p(!0,t)},I=function(t){return S?(o(t),T?D(t[0],t[1],O):D(t[0],t[1])):T?D(t,O):D(t)};if(k)g=t.iterator;else if(L)g=t;else{if(y=d(t),!y)throw new f(a(t)+" is not iterable");if(s(y)){for(v=0,_=l(t);_>v;v++)if(b=I(t[v]),b&&u(m,b))return b;return new p(!1)}g=c(t,y)}x=k?t.next:g.next;while(!(w=i(x,g)).done){try{b=I(w.value)}catch(C){h(g,"throw",C)}if("object"==typeof b&&b&&u(m,b))return b}return new p(!1)}},2286:function(t,e,n){var r=n("85e3"),i=Math.max;function o(t,e,n){return e=i(void 0===e?t.length-1:e,0),function(){var o=arguments,a=-1,s=i(o.length-e,0),l=Array(s);while(++a<s)l[a]=o[e+a];a=-1;var u=Array(e+1);while(++a<e)u[a]=o[a];return u[e]=n(l),r(t,this,u)}}t.exports=o},"22d1":function(t,e,n){"use strict";var r=function(){function t(){this.firefox=!1,this.ie=!1,this.edge=!1,this.newEdge=!1,this.weChat=!1}return t}(),i=function(){function t(){this.browser=new r,this.node=!1,this.wxa=!1,this.worker=!1,this.svgSupported=!1,this.touchEventsSupported=!1,this.pointerEventsSupported=!1,this.domSupported=!1,this.transformSupported=!1,this.transform3dSupported=!1,this.hasGlobalWindow="undefined"!==typeof window}return t}(),o=new i;function a(t,e){var n=e.browser,r=t.match(/Firefox\/([\d.]+)/),i=t.match(/MSIE\s([\d.]+)/)||t.match(/Trident\/.+?rv:(([\d.]+))/),o=t.match(/Edge?\/([\d.]+)/),a=/micromessenger/i.test(t);r&&(n.firefox=!0,n.version=r[1]),i&&(n.ie=!0,n.version=i[1]),o&&(n.edge=!0,n.version=o[1],n.newEdge=+o[1].split(".")[0]>18),a&&(n.weChat=!0),e.svgSupported="undefined"!==typeof SVGRect,e.touchEventsSupported="ontouchstart"in window&&!n.ie&&!n.edge,e.pointerEventsSupported="onpointerdown"in window&&(n.edge||n.ie&&+n.version>=11),e.domSupported="undefined"!==typeof document;var s=document.documentElement.style;e.transform3dSupported=(n.ie&&"transition"in s||n.edge||"WebKitCSSMatrix"in window&&"m11"in new WebKitCSSMatrix||"MozPerspective"in s)&&!("OTransition"in s),e.transformSupported=e.transform3dSupported||n.ie&&+n.version>=9}"object"===typeof wx&&"function"===typeof wx.getSystemInfoSync?(o.wxa=!0,o.touchEventsSupported=!0):"undefined"===typeof document&&"undefined"!==typeof self?o.worker=!0:"undefined"===typeof navigator||0===navigator.userAgent.indexOf("Node.js")?(o.node=!0,o.svgSupported=!0):a(navigator.userAgent,o),e["a"]=o},"22f8":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ko",{months:"1월_2월_3월_4월_5월_6월_7월_8월_9월_10월_11월_12월".split("_"),monthsShort:"1월_2월_3월_4월_5월_6월_7월_8월_9월_10월_11월_12월".split("_"),weekdays:"일요일_월요일_화요일_수요일_목요일_금요일_토요일".split("_"),weekdaysShort:"일_월_화_수_목_금_토".split("_"),weekdaysMin:"일_월_화_수_목_금_토".split("_"),longDateFormat:{LT:"A h:mm",LTS:"A h:mm:ss",L:"YYYY.MM.DD.",LL:"YYYY년 MMMM D일",LLL:"YYYY년 MMMM D일 A h:mm",LLLL:"YYYY년 MMMM D일 dddd A h:mm",l:"YYYY.MM.DD.",ll:"YYYY년 MMMM D일",lll:"YYYY년 MMMM D일 A h:mm",llll:"YYYY년 MMMM D일 dddd A h:mm"},calendar:{sameDay:"오늘 LT",nextDay:"내일 LT",nextWeek:"dddd LT",lastDay:"어제 LT",lastWeek:"지난주 dddd LT",sameElse:"L"},relativeTime:{future:"%s 후",past:"%s 전",s:"몇 초",ss:"%d초",m:"1분",mm:"%d분",h:"한 시간",hh:"%d시간",d:"하루",dd:"%d일",M:"한 달",MM:"%d달",y:"일 년",yy:"%d년"},dayOfMonthOrdinalParse:/\d{1,2}(일|월|주)/,ordinal:function(t,e){switch(e){case"d":case"D":case"DDD":return t+"일";case"M":return t+"월";case"w":case"W":return t+"주";default:return t}},meridiemParse:/오전|오후/,isPM:function(t){return"오후"===t},meridiem:function(t,e,n){return t<12?"오전":"오후"}});return e}))},"22fe":function(t,e,n){},"234d":function(t,e,n){var r=n("e380"),i=500;function o(t){var e=r(t,(function(t){return n.size===i&&n.clear(),t})),n=e.cache;return e}t.exports=o},"23cb":function(t,e,n){"use strict";var r=n("5926"),i=Math.max,o=Math.min;t.exports=function(t,e){var n=r(t);return n<0?i(n+e,0):o(n,e)}},"23e2":function(t,e,n){var r=n("242e");function i(t,e,n,i){return r(t,(function(t,r,o){e(i,n(t),r,o)})),i}t.exports=i},"23e7":function(t,e,n){"use strict";var r=n("da84"),i=n("06cf").f,o=n("9112"),a=n("cb2d"),s=n("6374"),l=n("e893"),u=n("94ca");t.exports=function(t,e){var n,c,d,h,f,p,m=t.target,g=t.global,y=t.stat;if(c=g?r:y?r[m]||s(m,{}):r[m]&&r[m].prototype,c)for(d in e){if(f=e[d],t.dontCallGetSet?(p=i(c,d),h=p&&p.value):h=c[d],n=u(g?d:m+(y?".":"#")+d,t.forced),!n&&void 0!==h){if(typeof f==typeof h)continue;l(f,h)}(t.sham||h&&h.sham)&&o(f,"sham",!0),a(c,d,f,t)}}},2411:function(t,e,n){var r=n("f909"),i=n("2ec1"),o=i((function(t,e,n,i){r(t,e,n,i)}));t.exports=o},"241c":function(t,e,n){"use strict";var r=n("ca84"),i=n("7839"),o=i.concat("length","prototype");e.f=Object.getOwnPropertyNames||function(t){return r(t,o)}},2421:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"١",2:"٢",3:"٣",4:"٤",5:"٥",6:"٦",7:"٧",8:"٨",9:"٩",0:"٠"},n={"١":"1","٢":"2","٣":"3","٤":"4","٥":"5","٦":"6","٧":"7","٨":"8","٩":"9","٠":"0"},r=["کانونی دووەم","شوبات","ئازار","نیسان","ئایار","حوزەیران","تەمموز","ئاب","ئەیلوول","تشرینی یەكەم","تشرینی دووەم","كانونی یەکەم"],i=t.defineLocale("ku",{months:r,monthsShort:r,weekdays:"یه‌كشه‌ممه‌_دووشه‌ممه‌_سێشه‌ممه‌_چوارشه‌ممه‌_پێنجشه‌ممه‌_هه‌ینی_شه‌ممه‌".split("_"),weekdaysShort:"یه‌كشه‌م_دووشه‌م_سێشه‌م_چوارشه‌م_پێنجشه‌م_هه‌ینی_شه‌ممه‌".split("_"),weekdaysMin:"ی_د_س_چ_پ_ه_ش".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},meridiemParse:/ئێواره‌|به‌یانی/,isPM:function(t){return/ئێواره‌/.test(t)},meridiem:function(t,e,n){return t<12?"به‌یانی":"ئێواره‌"},calendar:{sameDay:"[ئه‌مرۆ كاتژمێر] LT",nextDay:"[به‌یانی كاتژمێر] LT",nextWeek:"dddd [كاتژمێر] LT",lastDay:"[دوێنێ كاتژمێر] LT",lastWeek:"dddd [كاتژمێر] LT",sameElse:"L"},relativeTime:{future:"له‌ %s",past:"%s",s:"چه‌ند چركه‌یه‌ك",ss:"چركه‌ %d",m:"یه‌ك خوله‌ك",mm:"%d خوله‌ك",h:"یه‌ك كاتژمێر",hh:"%d كاتژمێر",d:"یه‌ك ڕۆژ",dd:"%d ڕۆژ",M:"یه‌ك مانگ",MM:"%d مانگ",y:"یه‌ك ساڵ",yy:"%d ساڵ"},preparse:function(t){return t.replace(/[١٢٣٤٥٦٧٨٩٠]/g,(function(t){return n[t]})).replace(/،/g,",")},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]})).replace(/,/g,"،")},week:{dow:6,doy:12}});return i}))},"242e":function(t,e,n){var r=n("72af"),i=n("ec69");function o(t,e){return t&&r(t,e,i)}t.exports=o},2430:function(t,e,n){"use strict";n("b0c0");var r=function(){var t=this,e=t._self._c;return e("div",{class:t.classes,attrs:{name:t.name}},[t._t("default")],2)},i=[],o=n("ade3"),a=(n("99af"),n("a9e3"),n("d3b7"),n("159b"),n("3a27")),s=n("2507"),l="ivu-radio-group",u=0,c=Date.now(),d=function(){return"ivuRadioGroup_".concat(c,"_").concat(u++)},h={name:"RadioGroup",mixins:[s["a"]],props:{value:{type:[String,Number],default:""},size:{validator:function(t){return Object(a["j"])(t,["small","large","default"])},default:function(){return this.$IVIEW&&""!==this.$IVIEW.size?this.$IVIEW.size:"default"}},type:{validator:function(t){return Object(a["j"])(t,["button"])}},vertical:{type:Boolean,default:!1},name:{type:String,default:d},buttonStyle:{validator:function(t){return Object(a["j"])(t,["default","solid"])},default:"default"}},data:function(){return{currentValue:this.value,children:[]}},computed:{classes:function(){return["".concat(l),Object(o["a"])(Object(o["a"])(Object(o["a"])(Object(o["a"])(Object(o["a"])({},"".concat(l,"-").concat(this.size),!!this.size),"ivu-radio-".concat(this.size),!!this.size),"".concat(l,"-").concat(this.type),!!this.type),"".concat(l,"-button-").concat(this.buttonStyle),"button"===this.type&&"default"!==this.buttonStyle),"".concat(l,"-vertical"),this.vertical)]}},mounted:function(){this.updateValue()},methods:{updateValue:function(){var t=this;this.children=Object(a["g"])(this,"Radio"),this.children&&this.children.forEach((function(e){e.currentValue=t.currentValue===e.label,e.group=!0}))},change:function(t){this.currentValue=t.value,this.updateValue(),this.$emit("input",t.value),this.$emit("on-change",t.value),this.dispatch("FormItem","on-form-change",t.value)}},watch:{value:function(){var t=this;this.currentValue!==this.value&&(this.currentValue=this.value,this.$nextTick((function(){t.updateValue()})))}}},f=h,p=n("2877"),m=Object(p["a"])(f,r,i,!1,null,null,null);e["a"]=m.exports},"245a":function(t,e,n){"use strict";var r=function(){var t=this,e=t._self._c;return e("li",{class:t.classes,on:{click:t.handleClick}},[t._t("default")],2)},i=[],o=n("ade3"),a=(n("b0c0"),n("a9e3"),n("3a27")),s="ivu-dropdown-item",l={name:"DropdownItem",props:{name:{type:[String,Number]},disabled:{type:Boolean,default:!1},selected:{type:Boolean,default:!1},divided:{type:Boolean,default:!1}},computed:{classes:function(){return["".concat(s),Object(o["a"])(Object(o["a"])(Object(o["a"])({},"".concat(s,"-disabled"),this.disabled),"".concat(s,"-selected"),this.selected),"".concat(s,"-divided"),this.divided)]}},methods:{handleClick:function(){if(!this.disabled){var t=Object(a["f"])(this,"Dropdown"),e=this.$parent&&"Dropdown"===this.$parent.$options.name;e?this.$parent.$emit("on-haschild-click"):t&&"Dropdown"===t.$options.name&&t.$emit("on-hover-click"),t.$emit("on-click",this.name)}}}},u=l,c=n("2877"),d=Object(c["a"])(u,r,i,!1,null,null,null);e["a"]=d.exports},2474:function(t,e,n){var r=n("2b3e"),i=r.Uint8Array;t.exports=i},2478:function(t,e,n){var r=n("4245");function i(t){return r(this,t).get(t)}t.exports=i},2507:function(t,e,n){"use strict";function r(t,e,n){this.$children.forEach(i=>{const o=i.$options.name;o===t?i.$emit.apply(i,[e].concat(n)):r.apply(i,[t,e].concat([n]))})}e["a"]={methods:{dispatch(t,e,n){let r=this.$parent||this.$root,i=r.$options.name;while(r&&(!i||i!==t))r=r.$parent,r&&(i=r.$options.name);r&&r.$emit.apply(r,[e].concat(n))},broadcast(t,e,n){r.call(this,t,e,n)}}}},2524:function(t,e,n){var r=n("6044"),i="__lodash_hash_undefined__";function o(t,e){var n=this.__data__;return this.size+=this.has(t)?0:1,n[t]=r&&void 0===e?i:e,this}t.exports=o},2532:function(t,e,n){"use strict";var r=n("23e7"),i=n("e330"),o=n("5a34"),a=n("1d80"),s=n("577e"),l=n("ab13"),u=i("".indexOf);r({target:"String",proto:!0,forced:!l("includes")},{includes:function(t){return!!~u(s(a(this)),s(o(t)),arguments.length>1?arguments[1]:void 0)}})},"253c":function(t,e,n){var r=n("3729"),i=n("1310"),o="[object Arguments]";function a(t){return i(t)&&r(t)==o}t.exports=a},2554:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,r){switch(n){case"m":return e?"jedna minuta":r?"jednu minutu":"jedne minute"}}function n(t,e,n){var r=t+" ";switch(n){case"ss":return r+=1===t?"sekunda":2===t||3===t||4===t?"sekunde":"sekundi",r;case"mm":return r+=1===t?"minuta":2===t||3===t||4===t?"minute":"minuta",r;case"h":return"jedan sat";case"hh":return r+=1===t?"sat":2===t||3===t||4===t?"sata":"sati",r;case"dd":return r+=1===t?"dan":"dana",r;case"MM":return r+=1===t?"mjesec":2===t||3===t||4===t?"mjeseca":"mjeseci",r;case"yy":return r+=1===t?"godina":2===t||3===t||4===t?"godine":"godina",r}}var r=t.defineLocale("bs",{months:"januar_februar_mart_april_maj_juni_juli_august_septembar_oktobar_novembar_decembar".split("_"),monthsShort:"jan._feb._mar._apr._maj._jun._jul._aug._sep._okt._nov._dec.".split("_"),monthsParseExact:!0,weekdays:"nedjelja_ponedjeljak_utorak_srijeda_četvrtak_petak_subota".split("_"),weekdaysShort:"ned._pon._uto._sri._čet._pet._sub.".split("_"),weekdaysMin:"ne_po_ut_sr_če_pe_su".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY H:mm",LLLL:"dddd, D. MMMM YYYY H:mm"},calendar:{sameDay:"[danas u] LT",nextDay:"[sutra u] LT",nextWeek:function(){switch(this.day()){case 0:return"[u] [nedjelju] [u] LT";case 3:return"[u] [srijedu] [u] LT";case 6:return"[u] [subotu] [u] LT";case 1:case 2:case 4:case 5:return"[u] dddd [u] LT"}},lastDay:"[jučer u] LT",lastWeek:function(){switch(this.day()){case 0:case 3:return"[prošlu] dddd [u] LT";case 6:return"[prošle] [subote] [u] LT";case 1:case 2:case 4:case 5:return"[prošli] dddd [u] LT"}},sameElse:"L"},relativeTime:{future:"za %s",past:"prije %s",s:"par sekundi",ss:n,m:e,mm:n,h:n,hh:n,d:"dan",dd:n,M:"mjesec",MM:n,y:"godinu",yy:n},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:7}});return r}))},2593:function(t,e,n){var r=n("15f3"),i=n("c6cf"),o=i((function(t,e){return null==t?{}:r(t,e)}));t.exports=o},"25eb":function(t,e,n){"use strict";var r=n("23e7"),i=n("c20d");r({target:"Number",stat:!0,forced:Number.parseInt!==i},{parseInt:i})},"25f0":function(t,e,n){"use strict";var r=n("5e77").PROPER,i=n("cb2d"),o=n("825a"),a=n("577e"),s=n("d039"),l=n("90d8"),u="toString",c=RegExp.prototype,d=c[u],h=s((function(){return"/a/b"!==d.call({source:"a",flags:"b"})})),f=r&&d.name!==u;(h||f)&&i(c,u,(function(){var t=o(this),e=a(t.source),n=a(l(t));return"/"+e+"/"+n}),{unsafe:!0})},2626:function(t,e,n){"use strict";var r=n("d066"),i=n("edd0"),o=n("b622"),a=n("83ab"),s=o("species");t.exports=function(t){var e=r(t);a&&e&&!e[s]&&i(e,s,{configurable:!0,get:function(){return this}})}},"266a":function(t,e,n){var r=n("7948");function i(t,e){return r(e,(function(e){return t[e]}))}t.exports=i},"26e8":function(t,e){function n(t,e){return null!=t&&e in Object(t)}t.exports=n},"26f9":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={ss:"sekundė_sekundžių_sekundes",m:"minutė_minutės_minutę",mm:"minutės_minučių_minutes",h:"valanda_valandos_valandą",hh:"valandos_valandų_valandas",d:"diena_dienos_dieną",dd:"dienos_dienų_dienas",M:"mėnuo_mėnesio_mėnesį",MM:"mėnesiai_mėnesių_mėnesius",y:"metai_metų_metus",yy:"metai_metų_metus"};function n(t,e,n,r){return e?"kelios sekundės":r?"kelių sekundžių":"kelias sekundes"}function r(t,e,n,r){return e?o(n)[0]:r?o(n)[1]:o(n)[2]}function i(t){return t%10===0||t>10&&t<20}function o(t){return e[t].split("_")}function a(t,e,n,a){var s=t+" ";return 1===t?s+r(t,e,n[0],a):e?s+(i(t)?o(n)[1]:o(n)[0]):a?s+o(n)[1]:s+(i(t)?o(n)[1]:o(n)[2])}var s=t.defineLocale("lt",{months:{format:"sausio_vasario_kovo_balandžio_gegužės_birželio_liepos_rugpjūčio_rugsėjo_spalio_lapkričio_gruodžio".split("_"),standalone:"sausis_vasaris_kovas_balandis_gegužė_birželis_liepa_rugpjūtis_rugsėjis_spalis_lapkritis_gruodis".split("_"),isFormat:/D[oD]?(\[[^\[\]]*\]|\s)+MMMM?|MMMM?(\[[^\[\]]*\]|\s)+D[oD]?/},monthsShort:"sau_vas_kov_bal_geg_bir_lie_rgp_rgs_spa_lap_grd".split("_"),weekdays:{format:"sekmadienį_pirmadienį_antradienį_trečiadienį_ketvirtadienį_penktadienį_šeštadienį".split("_"),standalone:"sekmadienis_pirmadienis_antradienis_trečiadienis_ketvirtadienis_penktadienis_šeštadienis".split("_"),isFormat:/dddd HH:mm/},weekdaysShort:"Sek_Pir_Ant_Tre_Ket_Pen_Šeš".split("_"),weekdaysMin:"S_P_A_T_K_Pn_Š".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY-MM-DD",LL:"YYYY [m.] MMMM D [d.]",LLL:"YYYY [m.] MMMM D [d.], HH:mm [val.]",LLLL:"YYYY [m.] MMMM D [d.], dddd, HH:mm [val.]",l:"YYYY-MM-DD",ll:"YYYY [m.] MMMM D [d.]",lll:"YYYY [m.] MMMM D [d.], HH:mm [val.]",llll:"YYYY [m.] MMMM D [d.], ddd, HH:mm [val.]"},calendar:{sameDay:"[Šiandien] LT",nextDay:"[Rytoj] LT",nextWeek:"dddd LT",lastDay:"[Vakar] LT",lastWeek:"[Praėjusį] dddd LT",sameElse:"L"},relativeTime:{future:"po %s",past:"prieš %s",s:n,ss:a,m:r,mm:a,h:r,hh:a,d:r,dd:a,M:r,MM:a,y:r,yy:a},dayOfMonthOrdinalParse:/\d{1,2}-oji/,ordinal:function(t){return t+"-oji"},week:{dow:1,doy:4}});return s}))},"27f3":function(t,e,n){var r=n("72f0"),i=n("43ad"),o=n("cd9d"),a=Object.prototype,s=a.toString,l=i((function(t,e,n){null!=e&&"function"!=typeof e.toString&&(e=s.call(e)),t[e]=n}),r(o));t.exports=l},2877:function(t,e,n){"use strict";function r(t,e,n,r,i,o,a,s){var l,u="function"===typeof t?t.options:t;if(e&&(u.render=e,u.staticRenderFns=n,u._compiled=!0),r&&(u.functional=!0),o&&(u._scopeId="data-v-"+o),a?(l=function(t){t=t||this.$vnode&&this.$vnode.ssrContext||this.parent&&this.parent.$vnode&&this.parent.$vnode.ssrContext,t||"undefined"===typeof __VUE_SSR_CONTEXT__||(t=__VUE_SSR_CONTEXT__),i&&i.call(this,t),t&&t._registeredComponents&&t._registeredComponents.add(a)},u._ssrRegister=l):i&&(l=s?function(){i.call(this,(u.functional?this.parent:this).$root.$options.shadowRoot)}:i),l)if(u.functional){u._injectStyles=l;var c=u.render;u.render=function(t,e){return l.call(e),c(t,e)}}else{var d=u.beforeCreate;u.beforeCreate=d?[].concat(d,l):[l]}return{exports:t,options:u}}n.d(e,"a",(function(){return r}))},"28c9":function(t,e){function n(){this.__data__=[],this.size=0}t.exports=n},2921:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("vi",{months:"tháng 1_tháng 2_tháng 3_tháng 4_tháng 5_tháng 6_tháng 7_tháng 8_tháng 9_tháng 10_tháng 11_tháng 12".split("_"),monthsShort:"Thg 01_Thg 02_Thg 03_Thg 04_Thg 05_Thg 06_Thg 07_Thg 08_Thg 09_Thg 10_Thg 11_Thg 12".split("_"),monthsParseExact:!0,weekdays:"chủ nhật_thứ hai_thứ ba_thứ tư_thứ năm_thứ sáu_thứ bảy".split("_"),weekdaysShort:"CN_T2_T3_T4_T5_T6_T7".split("_"),weekdaysMin:"CN_T2_T3_T4_T5_T6_T7".split("_"),weekdaysParseExact:!0,meridiemParse:/sa|ch/i,isPM:function(t){return/^ch$/i.test(t)},meridiem:function(t,e,n){return t<12?n?"sa":"SA":n?"ch":"CH"},longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM [năm] YYYY",LLL:"D MMMM [năm] YYYY HH:mm",LLLL:"dddd, D MMMM [năm] YYYY HH:mm",l:"DD/M/YYYY",ll:"D MMM YYYY",lll:"D MMM YYYY HH:mm",llll:"ddd, D MMM YYYY HH:mm"},calendar:{sameDay:"[Hôm nay lúc] LT",nextDay:"[Ngày mai lúc] LT",nextWeek:"dddd [tuần tới lúc] LT",lastDay:"[Hôm qua lúc] LT",lastWeek:"dddd [tuần trước lúc] LT",sameElse:"L"},relativeTime:{future:"%s tới",past:"%s trước",s:"vài giây",ss:"%d giây",m:"một phút",mm:"%d phút",h:"một giờ",hh:"%d giờ",d:"một ngày",dd:"%d ngày",w:"một tuần",ww:"%d tuần",M:"một tháng",MM:"%d tháng",y:"một năm",yy:"%d năm"},dayOfMonthOrdinalParse:/\d{1,2}/,ordinal:function(t){return t},week:{dow:1,doy:4}});return e}))},"293c":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={words:{ss:["sekund","sekunda","sekundi"],m:["jedan minut","jednog minuta"],mm:["minut","minuta","minuta"],h:["jedan sat","jednog sata"],hh:["sat","sata","sati"],dd:["dan","dana","dana"],MM:["mjesec","mjeseca","mjeseci"],yy:["godina","godine","godina"]},correctGrammaticalCase:function(t,e){return 1===t?e[0]:t>=2&&t<=4?e[1]:e[2]},translate:function(t,n,r){var i=e.words[r];return 1===r.length?n?i[0]:i[1]:t+" "+e.correctGrammaticalCase(t,i)}},n=t.defineLocale("me",{months:"januar_februar_mart_april_maj_jun_jul_avgust_septembar_oktobar_novembar_decembar".split("_"),monthsShort:"jan._feb._mar._apr._maj_jun_jul_avg._sep._okt._nov._dec.".split("_"),monthsParseExact:!0,weekdays:"nedjelja_ponedjeljak_utorak_srijeda_četvrtak_petak_subota".split("_"),weekdaysShort:"ned._pon._uto._sri._čet._pet._sub.".split("_"),weekdaysMin:"ne_po_ut_sr_če_pe_su".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY H:mm",LLLL:"dddd, D. MMMM YYYY H:mm"},calendar:{sameDay:"[danas u] LT",nextDay:"[sjutra u] LT",nextWeek:function(){switch(this.day()){case 0:return"[u] [nedjelju] [u] LT";case 3:return"[u] [srijedu] [u] LT";case 6:return"[u] [subotu] [u] LT";case 1:case 2:case 4:case 5:return"[u] dddd [u] LT"}},lastDay:"[juče u] LT",lastWeek:function(){var t=["[prošle] [nedjelje] [u] LT","[prošlog] [ponedjeljka] [u] LT","[prošlog] [utorka] [u] LT","[prošle] [srijede] [u] LT","[prošlog] [četvrtka] [u] LT","[prošlog] [petka] [u] LT","[prošle] [subote] [u] LT"];return t[this.day()]},sameElse:"L"},relativeTime:{future:"za %s",past:"prije %s",s:"nekoliko sekundi",ss:e.translate,m:e.translate,mm:e.translate,h:e.translate,hh:e.translate,d:"dan",dd:e.translate,M:"mjesec",MM:e.translate,y:"godinu",yy:e.translate},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:7}});return n}))},"29f3":function(t,e){var n=Object.prototype,r=n.toString;function i(t){return r.call(t)}t.exports=i},"2a62":function(t,e,n){"use strict";var r=n("c65b"),i=n("825a"),o=n("dc4a");t.exports=function(t,e,n){var a,s;i(t);try{if(a=o(t,"return"),!a){if("throw"===e)throw n;return n}a=r(a,t)}catch(l){s=!0,a=l}if("throw"===e)throw n;if(s)throw a;return i(a),n}},"2b0e":function(t,e,n){"use strict";(function(t){n.d(e,"a",(function(){return Kr}));
/*!
 * Vue.js v2.7.16
 * (c) 2014-2023 Evan You
 * Released under the MIT License.
 */
var r=Object.freeze({}),i=Array.isArray;function o(t){return void 0===t||null===t}function a(t){return void 0!==t&&null!==t}function s(t){return!0===t}function l(t){return!1===t}function u(t){return"string"===typeof t||"number"===typeof t||"symbol"===typeof t||"boolean"===typeof t}function c(t){return"function"===typeof t}function d(t){return null!==t&&"object"===typeof t}var h=Object.prototype.toString;function f(t){return"[object Object]"===h.call(t)}function p(t){return"[object RegExp]"===h.call(t)}function m(t){var e=parseFloat(String(t));return e>=0&&Math.floor(e)===e&&isFinite(t)}function g(t){return a(t)&&"function"===typeof t.then&&"function"===typeof t.catch}function y(t){return null==t?"":Array.isArray(t)||f(t)&&t.toString===h?JSON.stringify(t,v,2):String(t)}function v(t,e){return e&&e.__v_isRef?e.value:e}function _(t){var e=parseFloat(t);return isNaN(e)?t:e}function b(t,e){for(var n=Object.create(null),r=t.split(","),i=0;i<r.length;i++)n[r[i]]=!0;return e?function(t){return n[t.toLowerCase()]}:function(t){return n[t]}}b("slot,component",!0);var x=b("key,ref,slot,slot-scope,is");function w(t,e){var n=t.length;if(n){if(e===t[n-1])return void(t.length=n-1);var r=t.indexOf(e);if(r>-1)return t.splice(r,1)}}var M=Object.prototype.hasOwnProperty;function S(t,e){return M.call(t,e)}function k(t){var e=Object.create(null);return function(n){var r=e[n];return r||(e[n]=t(n))}}var L=/-(\w)/g,T=k((function(t){return t.replace(L,(function(t,e){return e?e.toUpperCase():""}))})),D=k((function(t){return t.charAt(0).toUpperCase()+t.slice(1)})),O=/\B([A-Z])/g,I=k((function(t){return t.replace(O,"-$1").toLowerCase()}));function C(t,e){function n(n){var r=arguments.length;return r?r>1?t.apply(e,arguments):t.call(e,n):t.call(e)}return n._length=t.length,n}function A(t,e){return t.bind(e)}var Y=Function.prototype.bind?A:C;function j(t,e){e=e||0;var n=t.length-e,r=new Array(n);while(n--)r[n]=t[n+e];return r}function P(t,e){for(var n in e)t[n]=e[n];return t}function E(t){for(var e={},n=0;n<t.length;n++)t[n]&&P(e,t[n]);return e}function N(t,e,n){}var R=function(t,e,n){return!1},H=function(t){return t};function z(t,e){if(t===e)return!0;var n=d(t),r=d(e);if(!n||!r)return!n&&!r&&String(t)===String(e);try{var i=Array.isArray(t),o=Array.isArray(e);if(i&&o)return t.length===e.length&&t.every((function(t,n){return z(t,e[n])}));if(t instanceof Date&&e instanceof Date)return t.getTime()===e.getTime();if(i||o)return!1;var a=Object.keys(t),s=Object.keys(e);return a.length===s.length&&a.every((function(n){return z(t[n],e[n])}))}catch(l){return!1}}function F(t,e){for(var n=0;n<t.length;n++)if(z(t[n],e))return n;return-1}function B(t){var e=!1;return function(){e||(e=!0,t.apply(this,arguments))}}function W(t,e){return t===e?0===t&&1/t!==1/e:t===t||e===e}var V="data-server-rendered",G=["component","directive","filter"],$=["beforeCreate","created","beforeMount","mounted","beforeUpdate","updated","beforeDestroy","destroyed","activated","deactivated","errorCaptured","serverPrefetch","renderTracked","renderTriggered"],U={optionMergeStrategies:Object.create(null),silent:!1,productionTip:!1,devtools:!1,performance:!1,errorHandler:null,warnHandler:null,ignoredElements:[],keyCodes:Object.create(null),isReservedTag:R,isReservedAttr:R,isUnknownElement:R,getTagNamespace:N,parsePlatformTagName:H,mustUseProp:R,async:!0,_lifecycleHooks:$},q=/a-zA-Z\u00B7\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u037D\u037F-\u1FFF\u200C-\u200D\u203F-\u2040\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD/;function X(t){var e=(t+"").charCodeAt(0);return 36===e||95===e}function Z(t,e,n,r){Object.defineProperty(t,e,{value:n,enumerable:!!r,writable:!0,configurable:!0})}var J=new RegExp("[^".concat(q.source,".$_\\d]"));function K(t){if(!J.test(t)){var e=t.split(".");return function(t){for(var n=0;n<e.length;n++){if(!t)return;t=t[e[n]]}return t}}}var Q="__proto__"in{},tt="undefined"!==typeof window,et=tt&&window.navigator.userAgent.toLowerCase(),nt=et&&/msie|trident/.test(et),rt=et&&et.indexOf("msie 9.0")>0,it=et&&et.indexOf("edge/")>0;et&&et.indexOf("android");var ot=et&&/iphone|ipad|ipod|ios/.test(et);et&&/chrome\/\d+/.test(et),et&&/phantomjs/.test(et);var at,st=et&&et.match(/firefox\/(\d+)/),lt={}.watch,ut=!1;if(tt)try{var ct={};Object.defineProperty(ct,"passive",{get:function(){ut=!0}}),window.addEventListener("test-passive",null,ct)}catch(Qa){}var dt=function(){return void 0===at&&(at=!tt&&"undefined"!==typeof t&&(t["process"]&&"server"===t["process"].env.VUE_ENV)),at},ht=tt&&window.__VUE_DEVTOOLS_GLOBAL_HOOK__;function ft(t){return"function"===typeof t&&/native code/.test(t.toString())}var pt,mt="undefined"!==typeof Symbol&&ft(Symbol)&&"undefined"!==typeof Reflect&&ft(Reflect.ownKeys);pt="undefined"!==typeof Set&&ft(Set)?Set:function(){function t(){this.set=Object.create(null)}return t.prototype.has=function(t){return!0===this.set[t]},t.prototype.add=function(t){this.set[t]=!0},t.prototype.clear=function(){this.set=Object.create(null)},t}();var gt=null;function yt(t){void 0===t&&(t=null),t||gt&&gt._scope.off(),gt=t,t&&t._scope.on()}var vt=function(){function t(t,e,n,r,i,o,a,s){this.tag=t,this.data=e,this.children=n,this.text=r,this.elm=i,this.ns=void 0,this.context=o,this.fnContext=void 0,this.fnOptions=void 0,this.fnScopeId=void 0,this.key=e&&e.key,this.componentOptions=a,this.componentInstance=void 0,this.parent=void 0,this.raw=!1,this.isStatic=!1,this.isRootInsert=!0,this.isComment=!1,this.isCloned=!1,this.isOnce=!1,this.asyncFactory=s,this.asyncMeta=void 0,this.isAsyncPlaceholder=!1}return Object.defineProperty(t.prototype,"child",{get:function(){return this.componentInstance},enumerable:!1,configurable:!0}),t}(),_t=function(t){void 0===t&&(t="");var e=new vt;return e.text=t,e.isComment=!0,e};function bt(t){return new vt(void 0,void 0,void 0,String(t))}function xt(t){var e=new vt(t.tag,t.data,t.children&&t.children.slice(),t.text,t.elm,t.context,t.componentOptions,t.asyncFactory);return e.ns=t.ns,e.isStatic=t.isStatic,e.key=t.key,e.isComment=t.isComment,e.fnContext=t.fnContext,e.fnOptions=t.fnOptions,e.fnScopeId=t.fnScopeId,e.asyncMeta=t.asyncMeta,e.isCloned=!0,e}"function"===typeof SuppressedError&&SuppressedError;var wt=0,Mt=[],St=function(){for(var t=0;t<Mt.length;t++){var e=Mt[t];e.subs=e.subs.filter((function(t){return t})),e._pending=!1}Mt.length=0},kt=function(){function t(){this._pending=!1,this.id=wt++,this.subs=[]}return t.prototype.addSub=function(t){this.subs.push(t)},t.prototype.removeSub=function(t){this.subs[this.subs.indexOf(t)]=null,this._pending||(this._pending=!0,Mt.push(this))},t.prototype.depend=function(e){t.target&&t.target.addDep(this)},t.prototype.notify=function(t){var e=this.subs.filter((function(t){return t}));for(var n=0,r=e.length;n<r;n++){var i=e[n];0,i.update()}},t}();kt.target=null;var Lt=[];function Tt(t){Lt.push(t),kt.target=t}function Dt(){Lt.pop(),kt.target=Lt[Lt.length-1]}var Ot=Array.prototype,It=Object.create(Ot),Ct=["push","pop","shift","unshift","splice","sort","reverse"];Ct.forEach((function(t){var e=Ot[t];Z(It,t,(function(){for(var n=[],r=0;r<arguments.length;r++)n[r]=arguments[r];var i,o=e.apply(this,n),a=this.__ob__;switch(t){case"push":case"unshift":i=n;break;case"splice":i=n.slice(2);break}return i&&a.observeArray(i),a.dep.notify(),o}))}));var At=Object.getOwnPropertyNames(It),Yt={},jt=!0;function Pt(t){jt=t}var Et={notify:N,depend:N,addSub:N,removeSub:N},Nt=function(){function t(t,e,n){if(void 0===e&&(e=!1),void 0===n&&(n=!1),this.value=t,this.shallow=e,this.mock=n,this.dep=n?Et:new kt,this.vmCount=0,Z(t,"__ob__",this),i(t)){if(!n)if(Q)t.__proto__=It;else for(var r=0,o=At.length;r<o;r++){var a=At[r];Z(t,a,It[a])}e||this.observeArray(t)}else{var s=Object.keys(t);for(r=0;r<s.length;r++){a=s[r];Ht(t,a,Yt,void 0,e,n)}}}return t.prototype.observeArray=function(t){for(var e=0,n=t.length;e<n;e++)Rt(t[e],!1,this.mock)},t}();function Rt(t,e,n){return t&&S(t,"__ob__")&&t.__ob__ instanceof Nt?t.__ob__:!jt||!n&&dt()||!i(t)&&!f(t)||!Object.isExtensible(t)||t.__v_skip||$t(t)||t instanceof vt?void 0:new Nt(t,e,n)}function Ht(t,e,n,r,o,a,s){void 0===s&&(s=!1);var l=new kt,u=Object.getOwnPropertyDescriptor(t,e);if(!u||!1!==u.configurable){var c=u&&u.get,d=u&&u.set;c&&!d||n!==Yt&&2!==arguments.length||(n=t[e]);var h=o?n&&n.__ob__:Rt(n,!1,a);return Object.defineProperty(t,e,{enumerable:!0,configurable:!0,get:function(){var e=c?c.call(t):n;return kt.target&&(l.depend(),h&&(h.dep.depend(),i(e)&&Bt(e))),$t(e)&&!o?e.value:e},set:function(e){var r=c?c.call(t):n;if(W(r,e)){if(d)d.call(t,e);else{if(c)return;if(!o&&$t(r)&&!$t(e))return void(r.value=e);n=e}h=o?e&&e.__ob__:Rt(e,!1,a),l.notify()}}}),l}}function zt(t,e,n){if(!Gt(t)){var r=t.__ob__;return i(t)&&m(e)?(t.length=Math.max(t.length,e),t.splice(e,1,n),r&&!r.shallow&&r.mock&&Rt(n,!1,!0),n):e in t&&!(e in Object.prototype)?(t[e]=n,n):t._isVue||r&&r.vmCount?n:r?(Ht(r.value,e,n,void 0,r.shallow,r.mock),r.dep.notify(),n):(t[e]=n,n)}}function Ft(t,e){if(i(t)&&m(e))t.splice(e,1);else{var n=t.__ob__;t._isVue||n&&n.vmCount||Gt(t)||S(t,e)&&(delete t[e],n&&n.dep.notify())}}function Bt(t){for(var e=void 0,n=0,r=t.length;n<r;n++)e=t[n],e&&e.__ob__&&e.__ob__.dep.depend(),i(e)&&Bt(e)}function Wt(t){return Vt(t,!0),Z(t,"__v_isShallow",!0),t}function Vt(t,e){if(!Gt(t)){Rt(t,e,dt());0}}function Gt(t){return!(!t||!t.__v_isReadonly)}function $t(t){return!(!t||!0!==t.__v_isRef)}function Ut(t,e,n){Object.defineProperty(t,n,{enumerable:!0,configurable:!0,get:function(){var t=e[n];if($t(t))return t.value;var r=t&&t.__ob__;return r&&r.dep.depend(),t},set:function(t){var r=e[n];$t(r)&&!$t(t)?r.value=t:e[n]=t}})}var qt="watcher";"".concat(qt," callback"),"".concat(qt," getter"),"".concat(qt," cleanup");var Xt;var Zt=function(){function t(t){void 0===t&&(t=!1),this.detached=t,this.active=!0,this.effects=[],this.cleanups=[],this.parent=Xt,!t&&Xt&&(this.index=(Xt.scopes||(Xt.scopes=[])).push(this)-1)}return t.prototype.run=function(t){if(this.active){var e=Xt;try{return Xt=this,t()}finally{Xt=e}}else 0},t.prototype.on=function(){Xt=this},t.prototype.off=function(){Xt=this.parent},t.prototype.stop=function(t){if(this.active){var e=void 0,n=void 0;for(e=0,n=this.effects.length;e<n;e++)this.effects[e].teardown();for(e=0,n=this.cleanups.length;e<n;e++)this.cleanups[e]();if(this.scopes)for(e=0,n=this.scopes.length;e<n;e++)this.scopes[e].stop(!0);if(!this.detached&&this.parent&&!t){var r=this.parent.scopes.pop();r&&r!==this&&(this.parent.scopes[this.index]=r,r.index=this.index)}this.parent=void 0,this.active=!1}},t}();function Jt(t,e){void 0===e&&(e=Xt),e&&e.active&&e.effects.push(t)}function Kt(){return Xt}function Qt(t){var e=t._provided,n=t.$parent&&t.$parent._provided;return n===e?t._provided=Object.create(n):e}var te=k((function(t){var e="&"===t.charAt(0);t=e?t.slice(1):t;var n="~"===t.charAt(0);t=n?t.slice(1):t;var r="!"===t.charAt(0);return t=r?t.slice(1):t,{name:t,once:n,capture:r,passive:e}}));function ee(t,e){function n(){var t=n.fns;if(!i(t))return Je(t,null,arguments,e,"v-on handler");for(var r=t.slice(),o=0;o<r.length;o++)Je(r[o],null,arguments,e,"v-on handler")}return n.fns=t,n}function ne(t,e,n,r,i,a){var l,u,c,d;for(l in t)u=t[l],c=e[l],d=te(l),o(u)||(o(c)?(o(u.fns)&&(u=t[l]=ee(u,a)),s(d.once)&&(u=t[l]=i(d.name,u,d.capture)),n(d.name,u,d.capture,d.passive,d.params)):u!==c&&(c.fns=u,t[l]=c));for(l in e)o(t[l])&&(d=te(l),r(d.name,e[l],d.capture))}function re(t,e,n){var r;t instanceof vt&&(t=t.data.hook||(t.data.hook={}));var i=t[e];function l(){n.apply(this,arguments),w(r.fns,l)}o(i)?r=ee([l]):a(i.fns)&&s(i.merged)?(r=i,r.fns.push(l)):r=ee([i,l]),r.merged=!0,t[e]=r}function ie(t,e,n){var r=e.options.props;if(!o(r)){var i={},s=t.attrs,l=t.props;if(a(s)||a(l))for(var u in r){var c=I(u);oe(i,l,u,c,!0)||oe(i,s,u,c,!1)}return i}}function oe(t,e,n,r,i){if(a(e)){if(S(e,n))return t[n]=e[n],i||delete e[n],!0;if(S(e,r))return t[n]=e[r],i||delete e[r],!0}return!1}function ae(t){for(var e=0;e<t.length;e++)if(i(t[e]))return Array.prototype.concat.apply([],t);return t}function se(t){return u(t)?[bt(t)]:i(t)?ue(t):void 0}function le(t){return a(t)&&a(t.text)&&l(t.isComment)}function ue(t,e){var n,r,l,c,d=[];for(n=0;n<t.length;n++)r=t[n],o(r)||"boolean"===typeof r||(l=d.length-1,c=d[l],i(r)?r.length>0&&(r=ue(r,"".concat(e||"","_").concat(n)),le(r[0])&&le(c)&&(d[l]=bt(c.text+r[0].text),r.shift()),d.push.apply(d,r)):u(r)?le(c)?d[l]=bt(c.text+r):""!==r&&d.push(bt(r)):le(r)&&le(c)?d[l]=bt(c.text+r.text):(s(t._isVList)&&a(r.tag)&&o(r.key)&&a(e)&&(r.key="__vlist".concat(e,"_").concat(n,"__")),d.push(r)));return d}function ce(t,e){var n,r,o,s,l=null;if(i(t)||"string"===typeof t)for(l=new Array(t.length),n=0,r=t.length;n<r;n++)l[n]=e(t[n],n);else if("number"===typeof t)for(l=new Array(t),n=0;n<t;n++)l[n]=e(n+1,n);else if(d(t))if(mt&&t[Symbol.iterator]){l=[];var u=t[Symbol.iterator](),c=u.next();while(!c.done)l.push(e(c.value,l.length)),c=u.next()}else for(o=Object.keys(t),l=new Array(o.length),n=0,r=o.length;n<r;n++)s=o[n],l[n]=e(t[s],s,n);return a(l)||(l=[]),l._isVList=!0,l}function de(t,e,n,r){var i,o=this.$scopedSlots[t];o?(n=n||{},r&&(n=P(P({},r),n)),i=o(n)||(c(e)?e():e)):i=this.$slots[t]||(c(e)?e():e);var a=n&&n.slot;return a?this.$createElement("template",{slot:a},i):i}function he(t){return kr(this.$options,"filters",t,!0)||H}function fe(t,e){return i(t)?-1===t.indexOf(e):t!==e}function pe(t,e,n,r,i){var o=U.keyCodes[e]||n;return i&&r&&!U.keyCodes[e]?fe(i,r):o?fe(o,t):r?I(r)!==e:void 0===t}function me(t,e,n,r,o){if(n)if(d(n)){i(n)&&(n=E(n));var a=void 0,s=function(i){if("class"===i||"style"===i||x(i))a=t;else{var s=t.attrs&&t.attrs.type;a=r||U.mustUseProp(e,s,i)?t.domProps||(t.domProps={}):t.attrs||(t.attrs={})}var l=T(i),u=I(i);if(!(l in a)&&!(u in a)&&(a[i]=n[i],o)){var c=t.on||(t.on={});c["update:".concat(i)]=function(t){n[i]=t}}};for(var l in n)s(l)}else;return t}function ge(t,e){var n=this._staticTrees||(this._staticTrees=[]),r=n[t];return r&&!e||(r=n[t]=this.$options.staticRenderFns[t].call(this._renderProxy,this._c,this),ve(r,"__static__".concat(t),!1)),r}function ye(t,e,n){return ve(t,"__once__".concat(e).concat(n?"_".concat(n):""),!0),t}function ve(t,e,n){if(i(t))for(var r=0;r<t.length;r++)t[r]&&"string"!==typeof t[r]&&_e(t[r],"".concat(e,"_").concat(r),n);else _e(t,e,n)}function _e(t,e,n){t.isStatic=!0,t.key=e,t.isOnce=n}function be(t,e){if(e)if(f(e)){var n=t.on=t.on?P({},t.on):{};for(var r in e){var i=n[r],o=e[r];n[r]=i?[].concat(i,o):o}}else;return t}function xe(t,e,n,r){e=e||{$stable:!n};for(var o=0;o<t.length;o++){var a=t[o];i(a)?xe(a,e,n):a&&(a.proxy&&(a.fn.proxy=!0),e[a.key]=a.fn)}return r&&(e.$key=r),e}function we(t,e){for(var n=0;n<e.length;n+=2){var r=e[n];"string"===typeof r&&r&&(t[e[n]]=e[n+1])}return t}function Me(t,e){return"string"===typeof t?e+t:t}function Se(t){t._o=ye,t._n=_,t._s=y,t._l=ce,t._t=de,t._q=z,t._i=F,t._m=ge,t._f=he,t._k=pe,t._b=me,t._v=bt,t._e=_t,t._u=xe,t._g=be,t._d=we,t._p=Me}function ke(t,e){if(!t||!t.length)return{};for(var n={},r=0,i=t.length;r<i;r++){var o=t[r],a=o.data;if(a&&a.attrs&&a.attrs.slot&&delete a.attrs.slot,o.context!==e&&o.fnContext!==e||!a||null==a.slot)(n.default||(n.default=[])).push(o);else{var s=a.slot,l=n[s]||(n[s]=[]);"template"===o.tag?l.push.apply(l,o.children||[]):l.push(o)}}for(var u in n)n[u].every(Le)&&delete n[u];return n}function Le(t){return t.isComment&&!t.asyncFactory||" "===t.text}function Te(t){return t.isComment&&t.asyncFactory}function De(t,e,n,i){var o,a=Object.keys(n).length>0,s=e?!!e.$stable:!a,l=e&&e.$key;if(e){if(e._normalized)return e._normalized;if(s&&i&&i!==r&&l===i.$key&&!a&&!i.$hasNormal)return i;for(var u in o={},e)e[u]&&"$"!==u[0]&&(o[u]=Oe(t,n,u,e[u]))}else o={};for(var c in n)c in o||(o[c]=Ie(n,c));return e&&Object.isExtensible(e)&&(e._normalized=o),Z(o,"$stable",s),Z(o,"$key",l),Z(o,"$hasNormal",a),o}function Oe(t,e,n,r){var o=function(){var e=gt;yt(t);var n=arguments.length?r.apply(null,arguments):r({});n=n&&"object"===typeof n&&!i(n)?[n]:se(n);var o=n&&n[0];return yt(e),n&&(!o||1===n.length&&o.isComment&&!Te(o))?void 0:n};return r.proxy&&Object.defineProperty(e,n,{get:o,enumerable:!0,configurable:!0}),o}function Ie(t,e){return function(){return t[e]}}function Ce(t){var e=t.$options,n=e.setup;if(n){var r=t._setupContext=Ae(t);yt(t),Tt();var i=Je(n,null,[t._props||Wt({}),r],t,"setup");if(Dt(),yt(),c(i))e.render=i;else if(d(i))if(t._setupState=i,i.__sfc){var o=t._setupProxy={};for(var a in i)"__sfc"!==a&&Ut(o,i,a)}else for(var a in i)X(a)||Ut(t,i,a);else 0}}function Ae(t){return{get attrs(){if(!t._attrsProxy){var e=t._attrsProxy={};Z(e,"_v_attr_proxy",!0),Ye(e,t.$attrs,r,t,"$attrs")}return t._attrsProxy},get listeners(){if(!t._listenersProxy){var e=t._listenersProxy={};Ye(e,t.$listeners,r,t,"$listeners")}return t._listenersProxy},get slots(){return Pe(t)},emit:Y(t.$emit,t),expose:function(e){e&&Object.keys(e).forEach((function(n){return Ut(t,e,n)}))}}}function Ye(t,e,n,r,i){var o=!1;for(var a in e)a in t?e[a]!==n[a]&&(o=!0):(o=!0,je(t,a,r,i));for(var a in t)a in e||(o=!0,delete t[a]);return o}function je(t,e,n,r){Object.defineProperty(t,e,{enumerable:!0,configurable:!0,get:function(){return n[r][e]}})}function Pe(t){return t._slotsProxy||Ee(t._slotsProxy={},t.$scopedSlots),t._slotsProxy}function Ee(t,e){for(var n in e)t[n]=e[n];for(var n in t)n in e||delete t[n]}function Ne(t){t._vnode=null,t._staticTrees=null;var e=t.$options,n=t.$vnode=e._parentVnode,i=n&&n.context;t.$slots=ke(e._renderChildren,i),t.$scopedSlots=n?De(t.$parent,n.data.scopedSlots,t.$slots):r,t._c=function(e,n,r,i){return $e(t,e,n,r,i,!1)},t.$createElement=function(e,n,r,i){return $e(t,e,n,r,i,!0)};var o=n&&n.data;Ht(t,"$attrs",o&&o.attrs||r,null,!0),Ht(t,"$listeners",e._parentListeners||r,null,!0)}var Re=null;function He(t){Se(t.prototype),t.prototype.$nextTick=function(t){return cn(t,this)},t.prototype._render=function(){var t=this,e=t.$options,n=e.render,r=e._parentVnode;r&&t._isMounted&&(t.$scopedSlots=De(t.$parent,r.data.scopedSlots,t.$slots,t.$scopedSlots),t._slotsProxy&&Ee(t._slotsProxy,t.$scopedSlots)),t.$vnode=r;var o,a=gt,s=Re;try{yt(t),Re=t,o=n.call(t._renderProxy,t.$createElement)}catch(Qa){Ze(Qa,t,"render"),o=t._vnode}finally{Re=s,yt(a)}return i(o)&&1===o.length&&(o=o[0]),o instanceof vt||(o=_t()),o.parent=r,o}}function ze(t,e){return(t.__esModule||mt&&"Module"===t[Symbol.toStringTag])&&(t=t.default),d(t)?e.extend(t):t}function Fe(t,e,n,r,i){var o=_t();return o.asyncFactory=t,o.asyncMeta={data:e,context:n,children:r,tag:i},o}function Be(t,e){if(s(t.error)&&a(t.errorComp))return t.errorComp;if(a(t.resolved))return t.resolved;var n=Re;if(n&&a(t.owners)&&-1===t.owners.indexOf(n)&&t.owners.push(n),s(t.loading)&&a(t.loadingComp))return t.loadingComp;if(n&&!a(t.owners)){var r=t.owners=[n],i=!0,l=null,u=null;n.$on("hook:destroyed",(function(){return w(r,n)}));var c=function(t){for(var e=0,n=r.length;e<n;e++)r[e].$forceUpdate();t&&(r.length=0,null!==l&&(clearTimeout(l),l=null),null!==u&&(clearTimeout(u),u=null))},h=B((function(n){t.resolved=ze(n,e),i?r.length=0:c(!0)})),f=B((function(e){a(t.errorComp)&&(t.error=!0,c(!0))})),p=t(h,f);return d(p)&&(g(p)?o(t.resolved)&&p.then(h,f):g(p.component)&&(p.component.then(h,f),a(p.error)&&(t.errorComp=ze(p.error,e)),a(p.loading)&&(t.loadingComp=ze(p.loading,e),0===p.delay?t.loading=!0:l=setTimeout((function(){l=null,o(t.resolved)&&o(t.error)&&(t.loading=!0,c(!1))}),p.delay||200)),a(p.timeout)&&(u=setTimeout((function(){u=null,o(t.resolved)&&f(null)}),p.timeout)))),i=!1,t.loading?t.loadingComp:t.resolved}}function We(t){if(i(t))for(var e=0;e<t.length;e++){var n=t[e];if(a(n)&&(a(n.componentOptions)||Te(n)))return n}}var Ve=1,Ge=2;function $e(t,e,n,r,o,a){return(i(n)||u(n))&&(o=r,r=n,n=void 0),s(a)&&(o=Ge),Ue(t,e,n,r,o)}function Ue(t,e,n,r,o){if(a(n)&&a(n.__ob__))return _t();if(a(n)&&a(n.is)&&(e=n.is),!e)return _t();var s,l;if(i(r)&&c(r[0])&&(n=n||{},n.scopedSlots={default:r[0]},r.length=0),o===Ge?r=se(r):o===Ve&&(r=ae(r)),"string"===typeof e){var u=void 0;l=t.$vnode&&t.$vnode.ns||U.getTagNamespace(e),s=U.isReservedTag(e)?new vt(U.parsePlatformTagName(e),n,r,void 0,void 0,t):n&&n.pre||!a(u=kr(t.$options,"components",e))?new vt(e,n,r,void 0,void 0,t):lr(u,n,t,r,e)}else s=lr(e,n,t,r);return i(s)?s:a(s)?(a(l)&&qe(s,l),a(n)&&Xe(n),s):_t()}function qe(t,e,n){if(t.ns=e,"foreignObject"===t.tag&&(e=void 0,n=!0),a(t.children))for(var r=0,i=t.children.length;r<i;r++){var l=t.children[r];a(l.tag)&&(o(l.ns)||s(n)&&"svg"!==l.tag)&&qe(l,e,n)}}function Xe(t){d(t.style)&&mn(t.style),d(t.class)&&mn(t.class)}function Ze(t,e,n){Tt();try{if(e){var r=e;while(r=r.$parent){var i=r.$options.errorCaptured;if(i)for(var o=0;o<i.length;o++)try{var a=!1===i[o].call(r,t,e,n);if(a)return}catch(Qa){Ke(Qa,r,"errorCaptured hook")}}}Ke(t,e,n)}finally{Dt()}}function Je(t,e,n,r,i){var o;try{o=n?t.apply(e,n):t.call(e),o&&!o._isVue&&g(o)&&!o._handled&&(o.catch((function(t){return Ze(t,r,i+" (Promise/async)")})),o._handled=!0)}catch(Qa){Ze(Qa,r,i)}return o}function Ke(t,e,n){if(U.errorHandler)try{return U.errorHandler.call(null,t,e,n)}catch(Qa){Qa!==t&&Qe(Qa,null,"config.errorHandler")}Qe(t,e,n)}function Qe(t,e,n){if(!tt||"undefined"===typeof console)throw t}var tn,en=!1,nn=[],rn=!1;function on(){rn=!1;var t=nn.slice(0);nn.length=0;for(var e=0;e<t.length;e++)t[e]()}if("undefined"!==typeof Promise&&ft(Promise)){var an=Promise.resolve();tn=function(){an.then(on),ot&&setTimeout(N)},en=!0}else if(nt||"undefined"===typeof MutationObserver||!ft(MutationObserver)&&"[object MutationObserverConstructor]"!==MutationObserver.toString())tn="undefined"!==typeof setImmediate&&ft(setImmediate)?function(){setImmediate(on)}:function(){setTimeout(on,0)};else{var sn=1,ln=new MutationObserver(on),un=document.createTextNode(String(sn));ln.observe(un,{characterData:!0}),tn=function(){sn=(sn+1)%2,un.data=String(sn)},en=!0}function cn(t,e){var n;if(nn.push((function(){if(t)try{t.call(e)}catch(Qa){Ze(Qa,e,"nextTick")}else n&&n(e)})),rn||(rn=!0,tn()),!t&&"undefined"!==typeof Promise)return new Promise((function(t){n=t}))}function dn(t){return function(e,n){if(void 0===n&&(n=gt),n)return hn(n,t,e)}}function hn(t,e,n){var r=t.$options;r[e]=yr(r[e],n)}dn("beforeMount"),dn("mounted"),dn("beforeUpdate"),dn("updated"),dn("beforeDestroy"),dn("destroyed"),dn("activated"),dn("deactivated"),dn("serverPrefetch"),dn("renderTracked"),dn("renderTriggered"),dn("errorCaptured");var fn="2.7.16";var pn=new pt;function mn(t){return gn(t,pn),pn.clear(),t}function gn(t,e){var n,r,o=i(t);if(!(!o&&!d(t)||t.__v_skip||Object.isFrozen(t)||t instanceof vt)){if(t.__ob__){var a=t.__ob__.dep.id;if(e.has(a))return;e.add(a)}if(o){n=t.length;while(n--)gn(t[n],e)}else if($t(t))gn(t.value,e);else{r=Object.keys(t),n=r.length;while(n--)gn(t[r[n]],e)}}}var yn,vn=0,_n=function(){function t(t,e,n,r,i){Jt(this,Xt&&!Xt._vm?Xt:t?t._scope:void 0),(this.vm=t)&&i&&(t._watcher=this),r?(this.deep=!!r.deep,this.user=!!r.user,this.lazy=!!r.lazy,this.sync=!!r.sync,this.before=r.before):this.deep=this.user=this.lazy=this.sync=!1,this.cb=n,this.id=++vn,this.active=!0,this.post=!1,this.dirty=this.lazy,this.deps=[],this.newDeps=[],this.depIds=new pt,this.newDepIds=new pt,this.expression="",c(e)?this.getter=e:(this.getter=K(e),this.getter||(this.getter=N)),this.value=this.lazy?void 0:this.get()}return t.prototype.get=function(){var t;Tt(this);var e=this.vm;try{t=this.getter.call(e,e)}catch(Qa){if(!this.user)throw Qa;Ze(Qa,e,'getter for watcher "'.concat(this.expression,'"'))}finally{this.deep&&mn(t),Dt(),this.cleanupDeps()}return t},t.prototype.addDep=function(t){var e=t.id;this.newDepIds.has(e)||(this.newDepIds.add(e),this.newDeps.push(t),this.depIds.has(e)||t.addSub(this))},t.prototype.cleanupDeps=function(){var t=this.deps.length;while(t--){var e=this.deps[t];this.newDepIds.has(e.id)||e.removeSub(this)}var n=this.depIds;this.depIds=this.newDepIds,this.newDepIds=n,this.newDepIds.clear(),n=this.deps,this.deps=this.newDeps,this.newDeps=n,this.newDeps.length=0},t.prototype.update=function(){this.lazy?this.dirty=!0:this.sync?this.run():Jn(this)},t.prototype.run=function(){if(this.active){var t=this.get();if(t!==this.value||d(t)||this.deep){var e=this.value;if(this.value=t,this.user){var n='callback for watcher "'.concat(this.expression,'"');Je(this.cb,this.vm,[t,e],this.vm,n)}else this.cb.call(this.vm,t,e)}}},t.prototype.evaluate=function(){this.value=this.get(),this.dirty=!1},t.prototype.depend=function(){var t=this.deps.length;while(t--)this.deps[t].depend()},t.prototype.teardown=function(){if(this.vm&&!this.vm._isBeingDestroyed&&w(this.vm._scope.effects,this),this.active){var t=this.deps.length;while(t--)this.deps[t].removeSub(this);this.active=!1,this.onStop&&this.onStop()}},t}();function bn(t){t._events=Object.create(null),t._hasHookEvent=!1;var e=t.$options._parentListeners;e&&Sn(t,e)}function xn(t,e){yn.$on(t,e)}function wn(t,e){yn.$off(t,e)}function Mn(t,e){var n=yn;return function r(){var i=e.apply(null,arguments);null!==i&&n.$off(t,r)}}function Sn(t,e,n){yn=t,ne(e,n||{},xn,wn,Mn,t),yn=void 0}function kn(t){var e=/^hook:/;t.prototype.$on=function(t,n){var r=this;if(i(t))for(var o=0,a=t.length;o<a;o++)r.$on(t[o],n);else(r._events[t]||(r._events[t]=[])).push(n),e.test(t)&&(r._hasHookEvent=!0);return r},t.prototype.$once=function(t,e){var n=this;function r(){n.$off(t,r),e.apply(n,arguments)}return r.fn=e,n.$on(t,r),n},t.prototype.$off=function(t,e){var n=this;if(!arguments.length)return n._events=Object.create(null),n;if(i(t)){for(var r=0,o=t.length;r<o;r++)n.$off(t[r],e);return n}var a,s=n._events[t];if(!s)return n;if(!e)return n._events[t]=null,n;var l=s.length;while(l--)if(a=s[l],a===e||a.fn===e){s.splice(l,1);break}return n},t.prototype.$emit=function(t){var e=this,n=e._events[t];if(n){n=n.length>1?j(n):n;for(var r=j(arguments,1),i='event handler for "'.concat(t,'"'),o=0,a=n.length;o<a;o++)Je(n[o],e,r,e,i)}return e}}var Ln=null;function Tn(t){var e=Ln;return Ln=t,function(){Ln=e}}function Dn(t){var e=t.$options,n=e.parent;if(n&&!e.abstract){while(n.$options.abstract&&n.$parent)n=n.$parent;n.$children.push(t)}t.$parent=n,t.$root=n?n.$root:t,t.$children=[],t.$refs={},t._provided=n?n._provided:Object.create(null),t._watcher=null,t._inactive=null,t._directInactive=!1,t._isMounted=!1,t._isDestroyed=!1,t._isBeingDestroyed=!1}function On(t){t.prototype._update=function(t,e){var n=this,r=n.$el,i=n._vnode,o=Tn(n);n._vnode=t,n.$el=i?n.__patch__(i,t):n.__patch__(n.$el,t,e,!1),o(),r&&(r.__vue__=null),n.$el&&(n.$el.__vue__=n);var a=n;while(a&&a.$vnode&&a.$parent&&a.$vnode===a.$parent._vnode)a.$parent.$el=a.$el,a=a.$parent},t.prototype.$forceUpdate=function(){var t=this;t._watcher&&t._watcher.update()},t.prototype.$destroy=function(){var t=this;if(!t._isBeingDestroyed){Pn(t,"beforeDestroy"),t._isBeingDestroyed=!0;var e=t.$parent;!e||e._isBeingDestroyed||t.$options.abstract||w(e.$children,t),t._scope.stop(),t._data.__ob__&&t._data.__ob__.vmCount--,t._isDestroyed=!0,t.__patch__(t._vnode,null),Pn(t,"destroyed"),t.$off(),t.$el&&(t.$el.__vue__=null),t.$vnode&&(t.$vnode.parent=null)}}}function In(t,e,n){var r;t.$el=e,t.$options.render||(t.$options.render=_t),Pn(t,"beforeMount"),r=function(){t._update(t._render(),n)};var i={before:function(){t._isMounted&&!t._isDestroyed&&Pn(t,"beforeUpdate")}};new _n(t,r,N,i,!0),n=!1;var o=t._preWatchers;if(o)for(var a=0;a<o.length;a++)o[a].run();return null==t.$vnode&&(t._isMounted=!0,Pn(t,"mounted")),t}function Cn(t,e,n,i,o){var a=i.data.scopedSlots,s=t.$scopedSlots,l=!!(a&&!a.$stable||s!==r&&!s.$stable||a&&t.$scopedSlots.$key!==a.$key||!a&&t.$scopedSlots.$key),u=!!(o||t.$options._renderChildren||l),c=t.$vnode;t.$options._parentVnode=i,t.$vnode=i,t._vnode&&(t._vnode.parent=i),t.$options._renderChildren=o;var d=i.data.attrs||r;t._attrsProxy&&Ye(t._attrsProxy,d,c.data&&c.data.attrs||r,t,"$attrs")&&(u=!0),t.$attrs=d,n=n||r;var h=t.$options._parentListeners;if(t._listenersProxy&&Ye(t._listenersProxy,n,h||r,t,"$listeners"),t.$listeners=t.$options._parentListeners=n,Sn(t,n,h),e&&t.$options.props){Pt(!1);for(var f=t._props,p=t.$options._propKeys||[],m=0;m<p.length;m++){var g=p[m],y=t.$options.props;f[g]=Lr(g,y,e,t)}Pt(!0),t.$options.propsData=e}u&&(t.$slots=ke(o,i.context),t.$forceUpdate())}function An(t){while(t&&(t=t.$parent))if(t._inactive)return!0;return!1}function Yn(t,e){if(e){if(t._directInactive=!1,An(t))return}else if(t._directInactive)return;if(t._inactive||null===t._inactive){t._inactive=!1;for(var n=0;n<t.$children.length;n++)Yn(t.$children[n]);Pn(t,"activated")}}function jn(t,e){if((!e||(t._directInactive=!0,!An(t)))&&!t._inactive){t._inactive=!0;for(var n=0;n<t.$children.length;n++)jn(t.$children[n]);Pn(t,"deactivated")}}function Pn(t,e,n,r){void 0===r&&(r=!0),Tt();var i=gt,o=Kt();r&&yt(t);var a=t.$options[e],s="".concat(e," hook");if(a)for(var l=0,u=a.length;l<u;l++)Je(a[l],t,n||null,t,s);t._hasHookEvent&&t.$emit("hook:"+e),r&&(yt(i),o&&o.on()),Dt()}var En=[],Nn=[],Rn={},Hn=!1,zn=!1,Fn=0;function Bn(){Fn=En.length=Nn.length=0,Rn={},Hn=zn=!1}var Wn=0,Vn=Date.now;if(tt&&!nt){var Gn=window.performance;Gn&&"function"===typeof Gn.now&&Vn()>document.createEvent("Event").timeStamp&&(Vn=function(){return Gn.now()})}var $n=function(t,e){if(t.post){if(!e.post)return 1}else if(e.post)return-1;return t.id-e.id};function Un(){var t,e;for(Wn=Vn(),zn=!0,En.sort($n),Fn=0;Fn<En.length;Fn++)t=En[Fn],t.before&&t.before(),e=t.id,Rn[e]=null,t.run();var n=Nn.slice(),r=En.slice();Bn(),Zn(n),qn(r),St(),ht&&U.devtools&&ht.emit("flush")}function qn(t){var e=t.length;while(e--){var n=t[e],r=n.vm;r&&r._watcher===n&&r._isMounted&&!r._isDestroyed&&Pn(r,"updated")}}function Xn(t){t._inactive=!1,Nn.push(t)}function Zn(t){for(var e=0;e<t.length;e++)t[e]._inactive=!0,Yn(t[e],!0)}function Jn(t){var e=t.id;if(null==Rn[e]&&(t!==kt.target||!t.noRecurse)){if(Rn[e]=!0,zn){var n=En.length-1;while(n>Fn&&En[n].id>t.id)n--;En.splice(n+1,0,t)}else En.push(t);Hn||(Hn=!0,cn(Un))}}function Kn(t){var e=t.$options.provide;if(e){var n=c(e)?e.call(t):e;if(!d(n))return;for(var r=Qt(t),i=mt?Reflect.ownKeys(n):Object.keys(n),o=0;o<i.length;o++){var a=i[o];Object.defineProperty(r,a,Object.getOwnPropertyDescriptor(n,a))}}}function Qn(t){var e=tr(t.$options.inject,t);e&&(Pt(!1),Object.keys(e).forEach((function(n){Ht(t,n,e[n])})),Pt(!0))}function tr(t,e){if(t){for(var n=Object.create(null),r=mt?Reflect.ownKeys(t):Object.keys(t),i=0;i<r.length;i++){var o=r[i];if("__ob__"!==o){var a=t[o].from;if(a in e._provided)n[o]=e._provided[a];else if("default"in t[o]){var s=t[o].default;n[o]=c(s)?s.call(e):s}else 0}}return n}}function er(t,e,n,o,a){var l,u=this,c=a.options;S(o,"_uid")?(l=Object.create(o),l._original=o):(l=o,o=o._original);var d=s(c._compiled),h=!d;this.data=t,this.props=e,this.children=n,this.parent=o,this.listeners=t.on||r,this.injections=tr(c.inject,o),this.slots=function(){return u.$slots||De(o,t.scopedSlots,u.$slots=ke(n,o)),u.$slots},Object.defineProperty(this,"scopedSlots",{enumerable:!0,get:function(){return De(o,t.scopedSlots,this.slots())}}),d&&(this.$options=c,this.$slots=this.slots(),this.$scopedSlots=De(o,t.scopedSlots,this.$slots)),c._scopeId?this._c=function(t,e,n,r){var a=$e(l,t,e,n,r,h);return a&&!i(a)&&(a.fnScopeId=c._scopeId,a.fnContext=o),a}:this._c=function(t,e,n,r){return $e(l,t,e,n,r,h)}}function nr(t,e,n,o,s){var l=t.options,u={},c=l.props;if(a(c))for(var d in c)u[d]=Lr(d,c,e||r);else a(n.attrs)&&ir(u,n.attrs),a(n.props)&&ir(u,n.props);var h=new er(n,u,s,o,t),f=l.render.call(null,h._c,h);if(f instanceof vt)return rr(f,n,h.parent,l,h);if(i(f)){for(var p=se(f)||[],m=new Array(p.length),g=0;g<p.length;g++)m[g]=rr(p[g],n,h.parent,l,h);return m}}function rr(t,e,n,r,i){var o=xt(t);return o.fnContext=n,o.fnOptions=r,e.slot&&((o.data||(o.data={})).slot=e.slot),o}function ir(t,e){for(var n in e)t[T(n)]=e[n]}function or(t){return t.name||t.__name||t._componentTag}Se(er.prototype);var ar={init:function(t,e){if(t.componentInstance&&!t.componentInstance._isDestroyed&&t.data.keepAlive){var n=t;ar.prepatch(n,n)}else{var r=t.componentInstance=ur(t,Ln);r.$mount(e?t.elm:void 0,e)}},prepatch:function(t,e){var n=e.componentOptions,r=e.componentInstance=t.componentInstance;Cn(r,n.propsData,n.listeners,e,n.children)},insert:function(t){var e=t.context,n=t.componentInstance;n._isMounted||(n._isMounted=!0,Pn(n,"mounted")),t.data.keepAlive&&(e._isMounted?Xn(n):Yn(n,!0))},destroy:function(t){var e=t.componentInstance;e._isDestroyed||(t.data.keepAlive?jn(e,!0):e.$destroy())}},sr=Object.keys(ar);function lr(t,e,n,r,i){if(!o(t)){var l=n.$options._base;if(d(t)&&(t=l.extend(t)),"function"===typeof t){var u;if(o(t.cid)&&(u=t,t=Be(u,l),void 0===t))return Fe(u,e,n,r,i);e=e||{},Zr(t),a(e.model)&&hr(t.options,e);var c=ie(e,t,i);if(s(t.options.functional))return nr(t,c,e,n,r);var h=e.on;if(e.on=e.nativeOn,s(t.options.abstract)){var f=e.slot;e={},f&&(e.slot=f)}cr(e);var p=or(t.options)||i,m=new vt("vue-component-".concat(t.cid).concat(p?"-".concat(p):""),e,void 0,void 0,void 0,n,{Ctor:t,propsData:c,listeners:h,tag:i,children:r},u);return m}}}function ur(t,e){var n={_isComponent:!0,_parentVnode:t,parent:e},r=t.data.inlineTemplate;return a(r)&&(n.render=r.render,n.staticRenderFns=r.staticRenderFns),new t.componentOptions.Ctor(n)}function cr(t){for(var e=t.hook||(t.hook={}),n=0;n<sr.length;n++){var r=sr[n],i=e[r],o=ar[r];i===o||i&&i._merged||(e[r]=i?dr(o,i):o)}}function dr(t,e){var n=function(n,r){t(n,r),e(n,r)};return n._merged=!0,n}function hr(t,e){var n=t.model&&t.model.prop||"value",r=t.model&&t.model.event||"input";(e.attrs||(e.attrs={}))[n]=e.model.value;var o=e.on||(e.on={}),s=o[r],l=e.model.callback;a(s)?(i(s)?-1===s.indexOf(l):s!==l)&&(o[r]=[l].concat(s)):o[r]=l}var fr=N,pr=U.optionMergeStrategies;function mr(t,e,n){if(void 0===n&&(n=!0),!e)return t;for(var r,i,o,a=mt?Reflect.ownKeys(e):Object.keys(e),s=0;s<a.length;s++)r=a[s],"__ob__"!==r&&(i=t[r],o=e[r],n&&S(t,r)?i!==o&&f(i)&&f(o)&&mr(i,o):zt(t,r,o));return t}function gr(t,e,n){return n?function(){var r=c(e)?e.call(n,n):e,i=c(t)?t.call(n,n):t;return r?mr(r,i):i}:e?t?function(){return mr(c(e)?e.call(this,this):e,c(t)?t.call(this,this):t)}:e:t}function yr(t,e){var n=e?t?t.concat(e):i(e)?e:[e]:t;return n?vr(n):n}function vr(t){for(var e=[],n=0;n<t.length;n++)-1===e.indexOf(t[n])&&e.push(t[n]);return e}function _r(t,e,n,r){var i=Object.create(t||null);return e?P(i,e):i}pr.data=function(t,e,n){return n?gr(t,e,n):e&&"function"!==typeof e?t:gr(t,e)},$.forEach((function(t){pr[t]=yr})),G.forEach((function(t){pr[t+"s"]=_r})),pr.watch=function(t,e,n,r){if(t===lt&&(t=void 0),e===lt&&(e=void 0),!e)return Object.create(t||null);if(!t)return e;var o={};for(var a in P(o,t),e){var s=o[a],l=e[a];s&&!i(s)&&(s=[s]),o[a]=s?s.concat(l):i(l)?l:[l]}return o},pr.props=pr.methods=pr.inject=pr.computed=function(t,e,n,r){if(!t)return e;var i=Object.create(null);return P(i,t),e&&P(i,e),i},pr.provide=function(t,e){return t?function(){var n=Object.create(null);return mr(n,c(t)?t.call(this):t),e&&mr(n,c(e)?e.call(this):e,!1),n}:e};var br=function(t,e){return void 0===e?t:e};function xr(t,e){var n=t.props;if(n){var r,o,a,s={};if(i(n)){r=n.length;while(r--)o=n[r],"string"===typeof o&&(a=T(o),s[a]={type:null})}else if(f(n))for(var l in n)o=n[l],a=T(l),s[a]=f(o)?o:{type:o};else 0;t.props=s}}function wr(t,e){var n=t.inject;if(n){var r=t.inject={};if(i(n))for(var o=0;o<n.length;o++)r[n[o]]={from:n[o]};else if(f(n))for(var a in n){var s=n[a];r[a]=f(s)?P({from:a},s):{from:s}}else 0}}function Mr(t){var e=t.directives;if(e)for(var n in e){var r=e[n];c(r)&&(e[n]={bind:r,update:r})}}function Sr(t,e,n){if(c(e)&&(e=e.options),xr(e,n),wr(e,n),Mr(e),!e._base&&(e.extends&&(t=Sr(t,e.extends,n)),e.mixins))for(var r=0,i=e.mixins.length;r<i;r++)t=Sr(t,e.mixins[r],n);var o,a={};for(o in t)s(o);for(o in e)S(t,o)||s(o);function s(r){var i=pr[r]||br;a[r]=i(t[r],e[r],n,r)}return a}function kr(t,e,n,r){if("string"===typeof n){var i=t[e];if(S(i,n))return i[n];var o=T(n);if(S(i,o))return i[o];var a=D(o);if(S(i,a))return i[a];var s=i[n]||i[o]||i[a];return s}}function Lr(t,e,n,r){var i=e[t],o=!S(n,t),a=n[t],s=Cr(Boolean,i.type);if(s>-1)if(o&&!S(i,"default"))a=!1;else if(""===a||a===I(t)){var l=Cr(String,i.type);(l<0||s<l)&&(a=!0)}if(void 0===a){a=Tr(r,i,t);var u=jt;Pt(!0),Rt(a),Pt(u)}return a}function Tr(t,e,n){if(S(e,"default")){var r=e.default;return t&&t.$options.propsData&&void 0===t.$options.propsData[n]&&void 0!==t._props[n]?t._props[n]:c(r)&&"Function"!==Or(e.type)?r.call(t):r}}var Dr=/^\s*function (\w+)/;function Or(t){var e=t&&t.toString().match(Dr);return e?e[1]:""}function Ir(t,e){return Or(t)===Or(e)}function Cr(t,e){if(!i(e))return Ir(e,t)?0:-1;for(var n=0,r=e.length;n<r;n++)if(Ir(e[n],t))return n;return-1}var Ar={enumerable:!0,configurable:!0,get:N,set:N};function Yr(t,e,n){Ar.get=function(){return this[e][n]},Ar.set=function(t){this[e][n]=t},Object.defineProperty(t,n,Ar)}function jr(t){var e=t.$options;if(e.props&&Pr(t,e.props),Ce(t),e.methods&&Wr(t,e.methods),e.data)Er(t);else{var n=Rt(t._data={});n&&n.vmCount++}e.computed&&Hr(t,e.computed),e.watch&&e.watch!==lt&&Vr(t,e.watch)}function Pr(t,e){var n=t.$options.propsData||{},r=t._props=Wt({}),i=t.$options._propKeys=[],o=!t.$parent;o||Pt(!1);var a=function(o){i.push(o);var a=Lr(o,e,n,t);Ht(r,o,a,void 0,!0),o in t||Yr(t,"_props",o)};for(var s in e)a(s);Pt(!0)}function Er(t){var e=t.$options.data;e=t._data=c(e)?Nr(e,t):e||{},f(e)||(e={});var n=Object.keys(e),r=t.$options.props,i=(t.$options.methods,n.length);while(i--){var o=n[i];0,r&&S(r,o)||X(o)||Yr(t,"_data",o)}var a=Rt(e);a&&a.vmCount++}function Nr(t,e){Tt();try{return t.call(e,e)}catch(Qa){return Ze(Qa,e,"data()"),{}}finally{Dt()}}var Rr={lazy:!0};function Hr(t,e){var n=t._computedWatchers=Object.create(null),r=dt();for(var i in e){var o=e[i],a=c(o)?o:o.get;0,r||(n[i]=new _n(t,a||N,N,Rr)),i in t||zr(t,i,o)}}function zr(t,e,n){var r=!dt();c(n)?(Ar.get=r?Fr(e):Br(n),Ar.set=N):(Ar.get=n.get?r&&!1!==n.cache?Fr(e):Br(n.get):N,Ar.set=n.set||N),Object.defineProperty(t,e,Ar)}function Fr(t){return function(){var e=this._computedWatchers&&this._computedWatchers[t];if(e)return e.dirty&&e.evaluate(),kt.target&&e.depend(),e.value}}function Br(t){return function(){return t.call(this,this)}}function Wr(t,e){t.$options.props;for(var n in e)t[n]="function"!==typeof e[n]?N:Y(e[n],t)}function Vr(t,e){for(var n in e){var r=e[n];if(i(r))for(var o=0;o<r.length;o++)Gr(t,n,r[o]);else Gr(t,n,r)}}function Gr(t,e,n,r){return f(n)&&(r=n,n=n.handler),"string"===typeof n&&(n=t[n]),t.$watch(e,n,r)}function $r(t){var e={get:function(){return this._data}},n={get:function(){return this._props}};Object.defineProperty(t.prototype,"$data",e),Object.defineProperty(t.prototype,"$props",n),t.prototype.$set=zt,t.prototype.$delete=Ft,t.prototype.$watch=function(t,e,n){var r=this;if(f(e))return Gr(r,t,e,n);n=n||{},n.user=!0;var i=new _n(r,t,e,n);if(n.immediate){var o='callback for immediate watcher "'.concat(i.expression,'"');Tt(),Je(e,r,[i.value],r,o),Dt()}return function(){i.teardown()}}}var Ur=0;function qr(t){t.prototype._init=function(t){var e=this;e._uid=Ur++,e._isVue=!0,e.__v_skip=!0,e._scope=new Zt(!0),e._scope.parent=void 0,e._scope._vm=!0,t&&t._isComponent?Xr(e,t):e.$options=Sr(Zr(e.constructor),t||{},e),e._renderProxy=e,e._self=e,Dn(e),bn(e),Ne(e),Pn(e,"beforeCreate",void 0,!1),Qn(e),jr(e),Kn(e),Pn(e,"created"),e.$options.el&&e.$mount(e.$options.el)}}function Xr(t,e){var n=t.$options=Object.create(t.constructor.options),r=e._parentVnode;n.parent=e.parent,n._parentVnode=r;var i=r.componentOptions;n.propsData=i.propsData,n._parentListeners=i.listeners,n._renderChildren=i.children,n._componentTag=i.tag,e.render&&(n.render=e.render,n.staticRenderFns=e.staticRenderFns)}function Zr(t){var e=t.options;if(t.super){var n=Zr(t.super),r=t.superOptions;if(n!==r){t.superOptions=n;var i=Jr(t);i&&P(t.extendOptions,i),e=t.options=Sr(n,t.extendOptions),e.name&&(e.components[e.name]=t)}}return e}function Jr(t){var e,n=t.options,r=t.sealedOptions;for(var i in n)n[i]!==r[i]&&(e||(e={}),e[i]=n[i]);return e}function Kr(t){this._init(t)}function Qr(t){t.use=function(t){var e=this._installedPlugins||(this._installedPlugins=[]);if(e.indexOf(t)>-1)return this;var n=j(arguments,1);return n.unshift(this),c(t.install)?t.install.apply(t,n):c(t)&&t.apply(null,n),e.push(t),this}}function ti(t){t.mixin=function(t){return this.options=Sr(this.options,t),this}}function ei(t){t.cid=0;var e=1;t.extend=function(t){t=t||{};var n=this,r=n.cid,i=t._Ctor||(t._Ctor={});if(i[r])return i[r];var o=or(t)||or(n.options);var a=function(t){this._init(t)};return a.prototype=Object.create(n.prototype),a.prototype.constructor=a,a.cid=e++,a.options=Sr(n.options,t),a["super"]=n,a.options.props&&ni(a),a.options.computed&&ri(a),a.extend=n.extend,a.mixin=n.mixin,a.use=n.use,G.forEach((function(t){a[t]=n[t]})),o&&(a.options.components[o]=a),a.superOptions=n.options,a.extendOptions=t,a.sealedOptions=P({},a.options),i[r]=a,a}}function ni(t){var e=t.options.props;for(var n in e)Yr(t.prototype,"_props",n)}function ri(t){var e=t.options.computed;for(var n in e)zr(t.prototype,n,e[n])}function ii(t){G.forEach((function(e){t[e]=function(t,n){return n?("component"===e&&f(n)&&(n.name=n.name||t,n=this.options._base.extend(n)),"directive"===e&&c(n)&&(n={bind:n,update:n}),this.options[e+"s"][t]=n,n):this.options[e+"s"][t]}}))}function oi(t){return t&&(or(t.Ctor.options)||t.tag)}function ai(t,e){return i(t)?t.indexOf(e)>-1:"string"===typeof t?t.split(",").indexOf(e)>-1:!!p(t)&&t.test(e)}function si(t,e){var n=t.cache,r=t.keys,i=t._vnode,o=t.$vnode;for(var a in n){var s=n[a];if(s){var l=s.name;l&&!e(l)&&li(n,a,r,i)}}o.componentOptions.children=void 0}function li(t,e,n,r){var i=t[e];!i||r&&i.tag===r.tag||i.componentInstance.$destroy(),t[e]=null,w(n,e)}qr(Kr),$r(Kr),kn(Kr),On(Kr),He(Kr);var ui=[String,RegExp,Array],ci={name:"keep-alive",abstract:!0,props:{include:ui,exclude:ui,max:[String,Number]},methods:{cacheVNode:function(){var t=this,e=t.cache,n=t.keys,r=t.vnodeToCache,i=t.keyToCache;if(r){var o=r.tag,a=r.componentInstance,s=r.componentOptions;e[i]={name:oi(s),tag:o,componentInstance:a},n.push(i),this.max&&n.length>parseInt(this.max)&&li(e,n[0],n,this._vnode),this.vnodeToCache=null}}},created:function(){this.cache=Object.create(null),this.keys=[]},destroyed:function(){for(var t in this.cache)li(this.cache,t,this.keys)},mounted:function(){var t=this;this.cacheVNode(),this.$watch("include",(function(e){si(t,(function(t){return ai(e,t)}))})),this.$watch("exclude",(function(e){si(t,(function(t){return!ai(e,t)}))}))},updated:function(){this.cacheVNode()},render:function(){var t=this.$slots.default,e=We(t),n=e&&e.componentOptions;if(n){var r=oi(n),i=this,o=i.include,a=i.exclude;if(o&&(!r||!ai(o,r))||a&&r&&ai(a,r))return e;var s=this,l=s.cache,u=s.keys,c=null==e.key?n.Ctor.cid+(n.tag?"::".concat(n.tag):""):e.key;l[c]?(e.componentInstance=l[c].componentInstance,w(u,c),u.push(c)):(this.vnodeToCache=e,this.keyToCache=c),e.data.keepAlive=!0}return e||t&&t[0]}},di={KeepAlive:ci};function hi(t){var e={get:function(){return U}};Object.defineProperty(t,"config",e),t.util={warn:fr,extend:P,mergeOptions:Sr,defineReactive:Ht},t.set=zt,t.delete=Ft,t.nextTick=cn,t.observable=function(t){return Rt(t),t},t.options=Object.create(null),G.forEach((function(e){t.options[e+"s"]=Object.create(null)})),t.options._base=t,P(t.options.components,di),Qr(t),ti(t),ei(t),ii(t)}hi(Kr),Object.defineProperty(Kr.prototype,"$isServer",{get:dt}),Object.defineProperty(Kr.prototype,"$ssrContext",{get:function(){return this.$vnode&&this.$vnode.ssrContext}}),Object.defineProperty(Kr,"FunctionalRenderContext",{value:er}),Kr.version=fn;var fi=b("style,class"),pi=b("input,textarea,option,select,progress"),mi=function(t,e,n){return"value"===n&&pi(t)&&"button"!==e||"selected"===n&&"option"===t||"checked"===n&&"input"===t||"muted"===n&&"video"===t},gi=b("contenteditable,draggable,spellcheck"),yi=b("events,caret,typing,plaintext-only"),vi=function(t,e){return Mi(e)||"false"===e?"false":"contenteditable"===t&&yi(e)?e:"true"},_i=b("allowfullscreen,async,autofocus,autoplay,checked,compact,controls,declare,default,defaultchecked,defaultmuted,defaultselected,defer,disabled,enabled,formnovalidate,hidden,indeterminate,inert,ismap,itemscope,loop,multiple,muted,nohref,noresize,noshade,novalidate,nowrap,open,pauseonexit,readonly,required,reversed,scoped,seamless,selected,sortable,truespeed,typemustmatch,visible"),bi="http://www.w3.org/1999/xlink",xi=function(t){return":"===t.charAt(5)&&"xlink"===t.slice(0,5)},wi=function(t){return xi(t)?t.slice(6,t.length):""},Mi=function(t){return null==t||!1===t};function Si(t){var e=t.data,n=t,r=t;while(a(r.componentInstance))r=r.componentInstance._vnode,r&&r.data&&(e=ki(r.data,e));while(a(n=n.parent))n&&n.data&&(e=ki(e,n.data));return Li(e.staticClass,e.class)}function ki(t,e){return{staticClass:Ti(t.staticClass,e.staticClass),class:a(t.class)?[t.class,e.class]:e.class}}function Li(t,e){return a(t)||a(e)?Ti(t,Di(e)):""}function Ti(t,e){return t?e?t+" "+e:t:e||""}function Di(t){return Array.isArray(t)?Oi(t):d(t)?Ii(t):"string"===typeof t?t:""}function Oi(t){for(var e,n="",r=0,i=t.length;r<i;r++)a(e=Di(t[r]))&&""!==e&&(n&&(n+=" "),n+=e);return n}function Ii(t){var e="";for(var n in t)t[n]&&(e&&(e+=" "),e+=n);return e}var Ci={svg:"http://www.w3.org/2000/svg",math:"http://www.w3.org/1998/Math/MathML"},Ai=b("html,body,base,head,link,meta,style,title,address,article,aside,footer,header,h1,h2,h3,h4,h5,h6,hgroup,nav,section,div,dd,dl,dt,figcaption,figure,picture,hr,img,li,main,ol,p,pre,ul,a,b,abbr,bdi,bdo,br,cite,code,data,dfn,em,i,kbd,mark,q,rp,rt,rtc,ruby,s,samp,small,span,strong,sub,sup,time,u,var,wbr,area,audio,map,track,video,embed,object,param,source,canvas,script,noscript,del,ins,caption,col,colgroup,table,thead,tbody,td,th,tr,button,datalist,fieldset,form,input,label,legend,meter,optgroup,option,output,progress,select,textarea,details,dialog,menu,menuitem,summary,content,element,shadow,template,blockquote,iframe,tfoot"),Yi=b("svg,animate,circle,clippath,cursor,defs,desc,ellipse,filter,font-face,foreignobject,g,glyph,image,line,marker,mask,missing-glyph,path,pattern,polygon,polyline,rect,switch,symbol,text,textpath,tspan,use,view",!0),ji=function(t){return Ai(t)||Yi(t)};function Pi(t){return Yi(t)?"svg":"math"===t?"math":void 0}var Ei=Object.create(null);function Ni(t){if(!tt)return!0;if(ji(t))return!1;if(t=t.toLowerCase(),null!=Ei[t])return Ei[t];var e=document.createElement(t);return t.indexOf("-")>-1?Ei[t]=e.constructor===window.HTMLUnknownElement||e.constructor===window.HTMLElement:Ei[t]=/HTMLUnknownElement/.test(e.toString())}var Ri=b("text,number,password,search,email,tel,url");function Hi(t){if("string"===typeof t){var e=document.querySelector(t);return e||document.createElement("div")}return t}function zi(t,e){var n=document.createElement(t);return"select"!==t||e.data&&e.data.attrs&&void 0!==e.data.attrs.multiple&&n.setAttribute("multiple","multiple"),n}function Fi(t,e){return document.createElementNS(Ci[t],e)}function Bi(t){return document.createTextNode(t)}function Wi(t){return document.createComment(t)}function Vi(t,e,n){t.insertBefore(e,n)}function Gi(t,e){t.removeChild(e)}function $i(t,e){t.appendChild(e)}function Ui(t){return t.parentNode}function qi(t){return t.nextSibling}function Xi(t){return t.tagName}function Zi(t,e){t.textContent=e}function Ji(t,e){t.setAttribute(e,"")}var Ki=Object.freeze({__proto__:null,createElement:zi,createElementNS:Fi,createTextNode:Bi,createComment:Wi,insertBefore:Vi,removeChild:Gi,appendChild:$i,parentNode:Ui,nextSibling:qi,tagName:Xi,setTextContent:Zi,setStyleScope:Ji}),Qi={create:function(t,e){to(e)},update:function(t,e){t.data.ref!==e.data.ref&&(to(t,!0),to(e))},destroy:function(t){to(t,!0)}};function to(t,e){var n=t.data.ref;if(a(n)){var r=t.context,o=t.componentInstance||t.elm,s=e?null:o,l=e?void 0:o;if(c(n))Je(n,r,[s],r,"template ref function");else{var u=t.data.refInFor,d="string"===typeof n||"number"===typeof n,h=$t(n),f=r.$refs;if(d||h)if(u){var p=d?f[n]:n.value;e?i(p)&&w(p,o):i(p)?p.includes(o)||p.push(o):d?(f[n]=[o],eo(r,n,f[n])):n.value=[o]}else if(d){if(e&&f[n]!==o)return;f[n]=l,eo(r,n,s)}else if(h){if(e&&n.value!==o)return;n.value=s}else 0}}}function eo(t,e,n){var r=t._setupState;r&&S(r,e)&&($t(r[e])?r[e].value=n:r[e]=n)}var no=new vt("",{},[]),ro=["create","activate","update","remove","destroy"];function io(t,e){return t.key===e.key&&t.asyncFactory===e.asyncFactory&&(t.tag===e.tag&&t.isComment===e.isComment&&a(t.data)===a(e.data)&&oo(t,e)||s(t.isAsyncPlaceholder)&&o(e.asyncFactory.error))}function oo(t,e){if("input"!==t.tag)return!0;var n,r=a(n=t.data)&&a(n=n.attrs)&&n.type,i=a(n=e.data)&&a(n=n.attrs)&&n.type;return r===i||Ri(r)&&Ri(i)}function ao(t,e,n){var r,i,o={};for(r=e;r<=n;++r)i=t[r].key,a(i)&&(o[i]=r);return o}function so(t){var e,n,r={},l=t.modules,c=t.nodeOps;for(e=0;e<ro.length;++e)for(r[ro[e]]=[],n=0;n<l.length;++n)a(l[n][ro[e]])&&r[ro[e]].push(l[n][ro[e]]);function d(t){return new vt(c.tagName(t).toLowerCase(),{},[],void 0,t)}function h(t,e){function n(){0===--n.listeners&&f(t)}return n.listeners=e,n}function f(t){var e=c.parentNode(t);a(e)&&c.removeChild(e,t)}function p(t,e,n,r,i,o,l){if(a(t.elm)&&a(o)&&(t=o[l]=xt(t)),t.isRootInsert=!i,!m(t,e,n,r)){var u=t.data,d=t.children,h=t.tag;a(h)?(t.elm=t.ns?c.createElementNS(t.ns,h):c.createElement(h,t),M(t),_(t,d,e),a(u)&&w(t,e),v(n,t.elm,r)):s(t.isComment)?(t.elm=c.createComment(t.text),v(n,t.elm,r)):(t.elm=c.createTextNode(t.text),v(n,t.elm,r))}}function m(t,e,n,r){var i=t.data;if(a(i)){var o=a(t.componentInstance)&&i.keepAlive;if(a(i=i.hook)&&a(i=i.init)&&i(t,!1),a(t.componentInstance))return g(t,e),v(n,t.elm,r),s(o)&&y(t,e,n,r),!0}}function g(t,e){a(t.data.pendingInsert)&&(e.push.apply(e,t.data.pendingInsert),t.data.pendingInsert=null),t.elm=t.componentInstance.$el,x(t)?(w(t,e),M(t)):(to(t),e.push(t))}function y(t,e,n,i){var o,s=t;while(s.componentInstance)if(s=s.componentInstance._vnode,a(o=s.data)&&a(o=o.transition)){for(o=0;o<r.activate.length;++o)r.activate[o](no,s);e.push(s);break}v(n,t.elm,i)}function v(t,e,n){a(t)&&(a(n)?c.parentNode(n)===t&&c.insertBefore(t,e,n):c.appendChild(t,e))}function _(t,e,n){if(i(e)){0;for(var r=0;r<e.length;++r)p(e[r],n,t.elm,null,!0,e,r)}else u(t.text)&&c.appendChild(t.elm,c.createTextNode(String(t.text)))}function x(t){while(t.componentInstance)t=t.componentInstance._vnode;return a(t.tag)}function w(t,n){for(var i=0;i<r.create.length;++i)r.create[i](no,t);e=t.data.hook,a(e)&&(a(e.create)&&e.create(no,t),a(e.insert)&&n.push(t))}function M(t){var e;if(a(e=t.fnScopeId))c.setStyleScope(t.elm,e);else{var n=t;while(n)a(e=n.context)&&a(e=e.$options._scopeId)&&c.setStyleScope(t.elm,e),n=n.parent}a(e=Ln)&&e!==t.context&&e!==t.fnContext&&a(e=e.$options._scopeId)&&c.setStyleScope(t.elm,e)}function S(t,e,n,r,i,o){for(;r<=i;++r)p(n[r],o,t,e,!1,n,r)}function k(t){var e,n,i=t.data;if(a(i))for(a(e=i.hook)&&a(e=e.destroy)&&e(t),e=0;e<r.destroy.length;++e)r.destroy[e](t);if(a(e=t.children))for(n=0;n<t.children.length;++n)k(t.children[n])}function L(t,e,n){for(;e<=n;++e){var r=t[e];a(r)&&(a(r.tag)?(T(r),k(r)):f(r.elm))}}function T(t,e){if(a(e)||a(t.data)){var n,i=r.remove.length+1;for(a(e)?e.listeners+=i:e=h(t.elm,i),a(n=t.componentInstance)&&a(n=n._vnode)&&a(n.data)&&T(n,e),n=0;n<r.remove.length;++n)r.remove[n](t,e);a(n=t.data.hook)&&a(n=n.remove)?n(t,e):e()}else f(t.elm)}function D(t,e,n,r,i){var s,l,u,d,h=0,f=0,m=e.length-1,g=e[0],y=e[m],v=n.length-1,_=n[0],b=n[v],x=!i;while(h<=m&&f<=v)o(g)?g=e[++h]:o(y)?y=e[--m]:io(g,_)?(I(g,_,r,n,f),g=e[++h],_=n[++f]):io(y,b)?(I(y,b,r,n,v),y=e[--m],b=n[--v]):io(g,b)?(I(g,b,r,n,v),x&&c.insertBefore(t,g.elm,c.nextSibling(y.elm)),g=e[++h],b=n[--v]):io(y,_)?(I(y,_,r,n,f),x&&c.insertBefore(t,y.elm,g.elm),y=e[--m],_=n[++f]):(o(s)&&(s=ao(e,h,m)),l=a(_.key)?s[_.key]:O(_,e,h,m),o(l)?p(_,r,t,g.elm,!1,n,f):(u=e[l],io(u,_)?(I(u,_,r,n,f),e[l]=void 0,x&&c.insertBefore(t,u.elm,g.elm)):p(_,r,t,g.elm,!1,n,f)),_=n[++f]);h>m?(d=o(n[v+1])?null:n[v+1].elm,S(t,d,n,f,v,r)):f>v&&L(e,h,m)}function O(t,e,n,r){for(var i=n;i<r;i++){var o=e[i];if(a(o)&&io(t,o))return i}}function I(t,e,n,i,l,u){if(t!==e){a(e.elm)&&a(i)&&(e=i[l]=xt(e));var d=e.elm=t.elm;if(s(t.isAsyncPlaceholder))a(e.asyncFactory.resolved)?Y(t.elm,e,n):e.isAsyncPlaceholder=!0;else if(s(e.isStatic)&&s(t.isStatic)&&e.key===t.key&&(s(e.isCloned)||s(e.isOnce)))e.componentInstance=t.componentInstance;else{var h,f=e.data;a(f)&&a(h=f.hook)&&a(h=h.prepatch)&&h(t,e);var p=t.children,m=e.children;if(a(f)&&x(e)){for(h=0;h<r.update.length;++h)r.update[h](t,e);a(h=f.hook)&&a(h=h.update)&&h(t,e)}o(e.text)?a(p)&&a(m)?p!==m&&D(d,p,m,n,u):a(m)?(a(t.text)&&c.setTextContent(d,""),S(d,null,m,0,m.length-1,n)):a(p)?L(p,0,p.length-1):a(t.text)&&c.setTextContent(d,""):t.text!==e.text&&c.setTextContent(d,e.text),a(f)&&a(h=f.hook)&&a(h=h.postpatch)&&h(t,e)}}}function C(t,e,n){if(s(n)&&a(t.parent))t.parent.data.pendingInsert=e;else for(var r=0;r<e.length;++r)e[r].data.hook.insert(e[r])}var A=b("attrs,class,staticClass,staticStyle,key");function Y(t,e,n,r){var i,o=e.tag,l=e.data,u=e.children;if(r=r||l&&l.pre,e.elm=t,s(e.isComment)&&a(e.asyncFactory))return e.isAsyncPlaceholder=!0,!0;if(a(l)&&(a(i=l.hook)&&a(i=i.init)&&i(e,!0),a(i=e.componentInstance)))return g(e,n),!0;if(a(o)){if(a(u))if(t.hasChildNodes())if(a(i=l)&&a(i=i.domProps)&&a(i=i.innerHTML)){if(i!==t.innerHTML)return!1}else{for(var c=!0,d=t.firstChild,h=0;h<u.length;h++){if(!d||!Y(d,u[h],n,r)){c=!1;break}d=d.nextSibling}if(!c||d)return!1}else _(e,u,n);if(a(l)){var f=!1;for(var p in l)if(!A(p)){f=!0,w(e,n);break}!f&&l["class"]&&mn(l["class"])}}else t.data!==e.text&&(t.data=e.text);return!0}return function(t,e,n,i){if(!o(e)){var l=!1,u=[];if(o(t))l=!0,p(e,u);else{var h=a(t.nodeType);if(!h&&io(t,e))I(t,e,u,null,null,i);else{if(h){if(1===t.nodeType&&t.hasAttribute(V)&&(t.removeAttribute(V),n=!0),s(n)&&Y(t,e,u))return C(e,u,!0),t;t=d(t)}var f=t.elm,m=c.parentNode(f);if(p(e,u,f._leaveCb?null:m,c.nextSibling(f)),a(e.parent)){var g=e.parent,y=x(e);while(g){for(var v=0;v<r.destroy.length;++v)r.destroy[v](g);if(g.elm=e.elm,y){for(var _=0;_<r.create.length;++_)r.create[_](no,g);var b=g.data.hook.insert;if(b.merged)for(var w=b.fns.slice(1),M=0;M<w.length;M++)w[M]()}else to(g);g=g.parent}}a(m)?L([t],0,0):a(t.tag)&&k(t)}}return C(e,u,l),e.elm}a(t)&&k(t)}}var lo={create:uo,update:uo,destroy:function(t){uo(t,no)}};function uo(t,e){(t.data.directives||e.data.directives)&&co(t,e)}function co(t,e){var n,r,i,o=t===no,a=e===no,s=fo(t.data.directives,t.context),l=fo(e.data.directives,e.context),u=[],c=[];for(n in l)r=s[n],i=l[n],r?(i.oldValue=r.value,i.oldArg=r.arg,mo(i,"update",e,t),i.def&&i.def.componentUpdated&&c.push(i)):(mo(i,"bind",e,t),i.def&&i.def.inserted&&u.push(i));if(u.length){var d=function(){for(var n=0;n<u.length;n++)mo(u[n],"inserted",e,t)};o?re(e,"insert",d):d()}if(c.length&&re(e,"postpatch",(function(){for(var n=0;n<c.length;n++)mo(c[n],"componentUpdated",e,t)})),!o)for(n in s)l[n]||mo(s[n],"unbind",t,t,a)}var ho=Object.create(null);function fo(t,e){var n,r,i=Object.create(null);if(!t)return i;for(n=0;n<t.length;n++){if(r=t[n],r.modifiers||(r.modifiers=ho),i[po(r)]=r,e._setupState&&e._setupState.__sfc){var o=r.def||kr(e,"_setupState","v-"+r.name);r.def="function"===typeof o?{bind:o,update:o}:o}r.def=r.def||kr(e.$options,"directives",r.name,!0)}return i}function po(t){return t.rawName||"".concat(t.name,".").concat(Object.keys(t.modifiers||{}).join("."))}function mo(t,e,n,r,i){var o=t.def&&t.def[e];if(o)try{o(n.elm,t,n,r,i)}catch(Qa){Ze(Qa,n.context,"directive ".concat(t.name," ").concat(e," hook"))}}var go=[Qi,lo];function yo(t,e){var n=e.componentOptions;if((!a(n)||!1!==n.Ctor.options.inheritAttrs)&&(!o(t.data.attrs)||!o(e.data.attrs))){var r,i,l,u=e.elm,c=t.data.attrs||{},d=e.data.attrs||{};for(r in(a(d.__ob__)||s(d._v_attr_proxy))&&(d=e.data.attrs=P({},d)),d)i=d[r],l=c[r],l!==i&&vo(u,r,i,e.data.pre);for(r in(nt||it)&&d.value!==c.value&&vo(u,"value",d.value),c)o(d[r])&&(xi(r)?u.removeAttributeNS(bi,wi(r)):gi(r)||u.removeAttribute(r))}}function vo(t,e,n,r){r||t.tagName.indexOf("-")>-1?_o(t,e,n):_i(e)?Mi(n)?t.removeAttribute(e):(n="allowfullscreen"===e&&"EMBED"===t.tagName?"true":e,t.setAttribute(e,n)):gi(e)?t.setAttribute(e,vi(e,n)):xi(e)?Mi(n)?t.removeAttributeNS(bi,wi(e)):t.setAttributeNS(bi,e,n):_o(t,e,n)}function _o(t,e,n){if(Mi(n))t.removeAttribute(e);else{if(nt&&!rt&&"TEXTAREA"===t.tagName&&"placeholder"===e&&""!==n&&!t.__ieph){var r=function(e){e.stopImmediatePropagation(),t.removeEventListener("input",r)};t.addEventListener("input",r),t.__ieph=!0}t.setAttribute(e,n)}}var bo={create:yo,update:yo};function xo(t,e){var n=e.elm,r=e.data,i=t.data;if(!(o(r.staticClass)&&o(r.class)&&(o(i)||o(i.staticClass)&&o(i.class)))){var s=Si(e),l=n._transitionClasses;a(l)&&(s=Ti(s,Di(l))),s!==n._prevClass&&(n.setAttribute("class",s),n._prevClass=s)}}var wo,Mo={create:xo,update:xo},So="__r",ko="__c";function Lo(t){if(a(t[So])){var e=nt?"change":"input";t[e]=[].concat(t[So],t[e]||[]),delete t[So]}a(t[ko])&&(t.change=[].concat(t[ko],t.change||[]),delete t[ko])}function To(t,e,n){var r=wo;return function i(){var o=e.apply(null,arguments);null!==o&&Io(t,i,n,r)}}var Do=en&&!(st&&Number(st[1])<=53);function Oo(t,e,n,r){if(Do){var i=Wn,o=e;e=o._wrapper=function(t){if(t.target===t.currentTarget||t.timeStamp>=i||t.timeStamp<=0||t.target.ownerDocument!==document)return o.apply(this,arguments)}}wo.addEventListener(t,e,ut?{capture:n,passive:r}:n)}function Io(t,e,n,r){(r||wo).removeEventListener(t,e._wrapper||e,n)}function Co(t,e){if(!o(t.data.on)||!o(e.data.on)){var n=e.data.on||{},r=t.data.on||{};wo=e.elm||t.elm,Lo(n),ne(n,r,Oo,Io,To,e.context),wo=void 0}}var Ao,Yo={create:Co,update:Co,destroy:function(t){return Co(t,no)}};function jo(t,e){if(!o(t.data.domProps)||!o(e.data.domProps)){var n,r,i=e.elm,l=t.data.domProps||{},u=e.data.domProps||{};for(n in(a(u.__ob__)||s(u._v_attr_proxy))&&(u=e.data.domProps=P({},u)),l)n in u||(i[n]="");for(n in u){if(r=u[n],"textContent"===n||"innerHTML"===n){if(e.children&&(e.children.length=0),r===l[n])continue;1===i.childNodes.length&&i.removeChild(i.childNodes[0])}if("value"===n&&"PROGRESS"!==i.tagName){i._value=r;var c=o(r)?"":String(r);Po(i,c)&&(i.value=c)}else if("innerHTML"===n&&Yi(i.tagName)&&o(i.innerHTML)){Ao=Ao||document.createElement("div"),Ao.innerHTML="<svg>".concat(r,"</svg>");var d=Ao.firstChild;while(i.firstChild)i.removeChild(i.firstChild);while(d.firstChild)i.appendChild(d.firstChild)}else if(r!==l[n])try{i[n]=r}catch(Qa){}}}}function Po(t,e){return!t.composing&&("OPTION"===t.tagName||Eo(t,e)||No(t,e))}function Eo(t,e){var n=!0;try{n=document.activeElement!==t}catch(Qa){}return n&&t.value!==e}function No(t,e){var n=t.value,r=t._vModifiers;if(a(r)){if(r.number)return _(n)!==_(e);if(r.trim)return n.trim()!==e.trim()}return n!==e}var Ro={create:jo,update:jo},Ho=k((function(t){var e={},n=/;(?![^(]*\))/g,r=/:(.+)/;return t.split(n).forEach((function(t){if(t){var n=t.split(r);n.length>1&&(e[n[0].trim()]=n[1].trim())}})),e}));function zo(t){var e=Fo(t.style);return t.staticStyle?P(t.staticStyle,e):e}function Fo(t){return Array.isArray(t)?E(t):"string"===typeof t?Ho(t):t}function Bo(t,e){var n,r={};if(e){var i=t;while(i.componentInstance)i=i.componentInstance._vnode,i&&i.data&&(n=zo(i.data))&&P(r,n)}(n=zo(t.data))&&P(r,n);var o=t;while(o=o.parent)o.data&&(n=zo(o.data))&&P(r,n);return r}var Wo,Vo=/^--/,Go=/\s*!important$/,$o=function(t,e,n){if(Vo.test(e))t.style.setProperty(e,n);else if(Go.test(n))t.style.setProperty(I(e),n.replace(Go,""),"important");else{var r=qo(e);if(Array.isArray(n))for(var i=0,o=n.length;i<o;i++)t.style[r]=n[i];else t.style[r]=n}},Uo=["Webkit","Moz","ms"],qo=k((function(t){if(Wo=Wo||document.createElement("div").style,t=T(t),"filter"!==t&&t in Wo)return t;for(var e=t.charAt(0).toUpperCase()+t.slice(1),n=0;n<Uo.length;n++){var r=Uo[n]+e;if(r in Wo)return r}}));function Xo(t,e){var n=e.data,r=t.data;if(!(o(n.staticStyle)&&o(n.style)&&o(r.staticStyle)&&o(r.style))){var i,s,l=e.elm,u=r.staticStyle,c=r.normalizedStyle||r.style||{},d=u||c,h=Fo(e.data.style)||{};e.data.normalizedStyle=a(h.__ob__)?P({},h):h;var f=Bo(e,!0);for(s in d)o(f[s])&&$o(l,s,"");for(s in f)i=f[s],$o(l,s,null==i?"":i)}}var Zo={create:Xo,update:Xo},Jo=/\s+/;function Ko(t,e){if(e&&(e=e.trim()))if(t.classList)e.indexOf(" ")>-1?e.split(Jo).forEach((function(e){return t.classList.add(e)})):t.classList.add(e);else{var n=" ".concat(t.getAttribute("class")||""," ");n.indexOf(" "+e+" ")<0&&t.setAttribute("class",(n+e).trim())}}function Qo(t,e){if(e&&(e=e.trim()))if(t.classList)e.indexOf(" ")>-1?e.split(Jo).forEach((function(e){return t.classList.remove(e)})):t.classList.remove(e),t.classList.length||t.removeAttribute("class");else{var n=" ".concat(t.getAttribute("class")||""," "),r=" "+e+" ";while(n.indexOf(r)>=0)n=n.replace(r," ");n=n.trim(),n?t.setAttribute("class",n):t.removeAttribute("class")}}function ta(t){if(t){if("object"===typeof t){var e={};return!1!==t.css&&P(e,ea(t.name||"v")),P(e,t),e}return"string"===typeof t?ea(t):void 0}}var ea=k((function(t){return{enterClass:"".concat(t,"-enter"),enterToClass:"".concat(t,"-enter-to"),enterActiveClass:"".concat(t,"-enter-active"),leaveClass:"".concat(t,"-leave"),leaveToClass:"".concat(t,"-leave-to"),leaveActiveClass:"".concat(t,"-leave-active")}})),na=tt&&!rt,ra="transition",ia="animation",oa="transition",aa="transitionend",sa="animation",la="animationend";na&&(void 0===window.ontransitionend&&void 0!==window.onwebkittransitionend&&(oa="WebkitTransition",aa="webkitTransitionEnd"),void 0===window.onanimationend&&void 0!==window.onwebkitanimationend&&(sa="WebkitAnimation",la="webkitAnimationEnd"));var ua=tt?window.requestAnimationFrame?window.requestAnimationFrame.bind(window):setTimeout:function(t){return t()};function ca(t){ua((function(){ua(t)}))}function da(t,e){var n=t._transitionClasses||(t._transitionClasses=[]);n.indexOf(e)<0&&(n.push(e),Ko(t,e))}function ha(t,e){t._transitionClasses&&w(t._transitionClasses,e),Qo(t,e)}function fa(t,e,n){var r=ma(t,e),i=r.type,o=r.timeout,a=r.propCount;if(!i)return n();var s=i===ra?aa:la,l=0,u=function(){t.removeEventListener(s,c),n()},c=function(e){e.target===t&&++l>=a&&u()};setTimeout((function(){l<a&&u()}),o+1),t.addEventListener(s,c)}var pa=/\b(transform|all)(,|$)/;function ma(t,e){var n,r=window.getComputedStyle(t),i=(r[oa+"Delay"]||"").split(", "),o=(r[oa+"Duration"]||"").split(", "),a=ga(i,o),s=(r[sa+"Delay"]||"").split(", "),l=(r[sa+"Duration"]||"").split(", "),u=ga(s,l),c=0,d=0;e===ra?a>0&&(n=ra,c=a,d=o.length):e===ia?u>0&&(n=ia,c=u,d=l.length):(c=Math.max(a,u),n=c>0?a>u?ra:ia:null,d=n?n===ra?o.length:l.length:0);var h=n===ra&&pa.test(r[oa+"Property"]);return{type:n,timeout:c,propCount:d,hasTransform:h}}function ga(t,e){while(t.length<e.length)t=t.concat(t);return Math.max.apply(null,e.map((function(e,n){return ya(e)+ya(t[n])})))}function ya(t){return 1e3*Number(t.slice(0,-1).replace(",","."))}function va(t,e){var n=t.elm;a(n._leaveCb)&&(n._leaveCb.cancelled=!0,n._leaveCb());var r=ta(t.data.transition);if(!o(r)&&!a(n._enterCb)&&1===n.nodeType){var i=r.css,s=r.type,l=r.enterClass,u=r.enterToClass,h=r.enterActiveClass,f=r.appearClass,p=r.appearToClass,m=r.appearActiveClass,g=r.beforeEnter,y=r.enter,v=r.afterEnter,b=r.enterCancelled,x=r.beforeAppear,w=r.appear,M=r.afterAppear,S=r.appearCancelled,k=r.duration,L=Ln,T=Ln.$vnode;while(T&&T.parent)L=T.context,T=T.parent;var D=!L._isMounted||!t.isRootInsert;if(!D||w||""===w){var O=D&&f?f:l,I=D&&m?m:h,C=D&&p?p:u,A=D&&x||g,Y=D&&c(w)?w:y,j=D&&M||v,P=D&&S||b,E=_(d(k)?k.enter:k);0;var N=!1!==i&&!rt,R=xa(Y),H=n._enterCb=B((function(){N&&(ha(n,C),ha(n,I)),H.cancelled?(N&&ha(n,O),P&&P(n)):j&&j(n),n._enterCb=null}));t.data.show||re(t,"insert",(function(){var e=n.parentNode,r=e&&e._pending&&e._pending[t.key];r&&r.tag===t.tag&&r.elm._leaveCb&&r.elm._leaveCb(),Y&&Y(n,H)})),A&&A(n),N&&(da(n,O),da(n,I),ca((function(){ha(n,O),H.cancelled||(da(n,C),R||(ba(E)?setTimeout(H,E):fa(n,s,H)))}))),t.data.show&&(e&&e(),Y&&Y(n,H)),N||R||H()}}}function _a(t,e){var n=t.elm;a(n._enterCb)&&(n._enterCb.cancelled=!0,n._enterCb());var r=ta(t.data.transition);if(o(r)||1!==n.nodeType)return e();if(!a(n._leaveCb)){var i=r.css,s=r.type,l=r.leaveClass,u=r.leaveToClass,c=r.leaveActiveClass,h=r.beforeLeave,f=r.leave,p=r.afterLeave,m=r.leaveCancelled,g=r.delayLeave,y=r.duration,v=!1!==i&&!rt,b=xa(f),x=_(d(y)?y.leave:y);0;var w=n._leaveCb=B((function(){n.parentNode&&n.parentNode._pending&&(n.parentNode._pending[t.key]=null),v&&(ha(n,u),ha(n,c)),w.cancelled?(v&&ha(n,l),m&&m(n)):(e(),p&&p(n)),n._leaveCb=null}));g?g(M):M()}function M(){w.cancelled||(!t.data.show&&n.parentNode&&((n.parentNode._pending||(n.parentNode._pending={}))[t.key]=t),h&&h(n),v&&(da(n,l),da(n,c),ca((function(){ha(n,l),w.cancelled||(da(n,u),b||(ba(x)?setTimeout(w,x):fa(n,s,w)))}))),f&&f(n,w),v||b||w())}}function ba(t){return"number"===typeof t&&!isNaN(t)}function xa(t){if(o(t))return!1;var e=t.fns;return a(e)?xa(Array.isArray(e)?e[0]:e):(t._length||t.length)>1}function wa(t,e){!0!==e.data.show&&va(e)}var Ma=tt?{create:wa,activate:wa,remove:function(t,e){!0!==t.data.show?_a(t,e):e()}}:{},Sa=[bo,Mo,Yo,Ro,Zo,Ma],ka=Sa.concat(go),La=so({nodeOps:Ki,modules:ka});rt&&document.addEventListener("selectionchange",(function(){var t=document.activeElement;t&&t.vmodel&&ja(t,"input")}));var Ta={inserted:function(t,e,n,r){"select"===n.tag?(r.elm&&!r.elm._vOptions?re(n,"postpatch",(function(){Ta.componentUpdated(t,e,n)})):Da(t,e,n.context),t._vOptions=[].map.call(t.options,Ca)):("textarea"===n.tag||Ri(t.type))&&(t._vModifiers=e.modifiers,e.modifiers.lazy||(t.addEventListener("compositionstart",Aa),t.addEventListener("compositionend",Ya),t.addEventListener("change",Ya),rt&&(t.vmodel=!0)))},componentUpdated:function(t,e,n){if("select"===n.tag){Da(t,e,n.context);var r=t._vOptions,i=t._vOptions=[].map.call(t.options,Ca);if(i.some((function(t,e){return!z(t,r[e])}))){var o=t.multiple?e.value.some((function(t){return Ia(t,i)})):e.value!==e.oldValue&&Ia(e.value,i);o&&ja(t,"change")}}}};function Da(t,e,n){Oa(t,e,n),(nt||it)&&setTimeout((function(){Oa(t,e,n)}),0)}function Oa(t,e,n){var r=e.value,i=t.multiple;if(!i||Array.isArray(r)){for(var o,a,s=0,l=t.options.length;s<l;s++)if(a=t.options[s],i)o=F(r,Ca(a))>-1,a.selected!==o&&(a.selected=o);else if(z(Ca(a),r))return void(t.selectedIndex!==s&&(t.selectedIndex=s));i||(t.selectedIndex=-1)}}function Ia(t,e){return e.every((function(e){return!z(e,t)}))}function Ca(t){return"_value"in t?t._value:t.value}function Aa(t){t.target.composing=!0}function Ya(t){t.target.composing&&(t.target.composing=!1,ja(t.target,"input"))}function ja(t,e){var n=document.createEvent("HTMLEvents");n.initEvent(e,!0,!0),t.dispatchEvent(n)}function Pa(t){return!t.componentInstance||t.data&&t.data.transition?t:Pa(t.componentInstance._vnode)}var Ea={bind:function(t,e,n){var r=e.value;n=Pa(n);var i=n.data&&n.data.transition,o=t.__vOriginalDisplay="none"===t.style.display?"":t.style.display;r&&i?(n.data.show=!0,va(n,(function(){t.style.display=o}))):t.style.display=r?o:"none"},update:function(t,e,n){var r=e.value,i=e.oldValue;if(!r!==!i){n=Pa(n);var o=n.data&&n.data.transition;o?(n.data.show=!0,r?va(n,(function(){t.style.display=t.__vOriginalDisplay})):_a(n,(function(){t.style.display="none"}))):t.style.display=r?t.__vOriginalDisplay:"none"}},unbind:function(t,e,n,r,i){i||(t.style.display=t.__vOriginalDisplay)}},Na={model:Ta,show:Ea},Ra={name:String,appear:Boolean,css:Boolean,mode:String,type:String,enterClass:String,leaveClass:String,enterToClass:String,leaveToClass:String,enterActiveClass:String,leaveActiveClass:String,appearClass:String,appearActiveClass:String,appearToClass:String,duration:[Number,String,Object]};function Ha(t){var e=t&&t.componentOptions;return e&&e.Ctor.options.abstract?Ha(We(e.children)):t}function za(t){var e={},n=t.$options;for(var r in n.propsData)e[r]=t[r];var i=n._parentListeners;for(var r in i)e[T(r)]=i[r];return e}function Fa(t,e){if(/\d-keep-alive$/.test(e.tag))return t("keep-alive",{props:e.componentOptions.propsData})}function Ba(t){while(t=t.parent)if(t.data.transition)return!0}function Wa(t,e){return e.key===t.key&&e.tag===t.tag}var Va=function(t){return t.tag||Te(t)},Ga=function(t){return"show"===t.name},$a={name:"transition",props:Ra,abstract:!0,render:function(t){var e=this,n=this.$slots.default;if(n&&(n=n.filter(Va),n.length)){0;var r=this.mode;0;var i=n[0];if(Ba(this.$vnode))return i;var o=Ha(i);if(!o)return i;if(this._leaving)return Fa(t,i);var a="__transition-".concat(this._uid,"-");o.key=null==o.key?o.isComment?a+"comment":a+o.tag:u(o.key)?0===String(o.key).indexOf(a)?o.key:a+o.key:o.key;var s=(o.data||(o.data={})).transition=za(this),l=this._vnode,c=Ha(l);if(o.data.directives&&o.data.directives.some(Ga)&&(o.data.show=!0),c&&c.data&&!Wa(o,c)&&!Te(c)&&(!c.componentInstance||!c.componentInstance._vnode.isComment)){var d=c.data.transition=P({},s);if("out-in"===r)return this._leaving=!0,re(d,"afterLeave",(function(){e._leaving=!1,e.$forceUpdate()})),Fa(t,i);if("in-out"===r){if(Te(o))return l;var h,f=function(){h()};re(s,"afterEnter",f),re(s,"enterCancelled",f),re(d,"delayLeave",(function(t){h=t}))}}return i}}},Ua=P({tag:String,moveClass:String},Ra);delete Ua.mode;var qa={props:Ua,beforeMount:function(){var t=this,e=this._update;this._update=function(n,r){var i=Tn(t);t.__patch__(t._vnode,t.kept,!1,!0),t._vnode=t.kept,i(),e.call(t,n,r)}},render:function(t){for(var e=this.tag||this.$vnode.data.tag||"span",n=Object.create(null),r=this.prevChildren=this.children,i=this.$slots.default||[],o=this.children=[],a=za(this),s=0;s<i.length;s++){var l=i[s];if(l.tag)if(null!=l.key&&0!==String(l.key).indexOf("__vlist"))o.push(l),n[l.key]=l,(l.data||(l.data={})).transition=a;else;}if(r){var u=[],c=[];for(s=0;s<r.length;s++){l=r[s];l.data.transition=a,l.data.pos=l.elm.getBoundingClientRect(),n[l.key]?u.push(l):c.push(l)}this.kept=t(e,null,u),this.removed=c}return t(e,null,o)},updated:function(){var t=this.prevChildren,e=this.moveClass||(this.name||"v")+"-move";t.length&&this.hasMove(t[0].elm,e)&&(t.forEach(Xa),t.forEach(Za),t.forEach(Ja),this._reflow=document.body.offsetHeight,t.forEach((function(t){if(t.data.moved){var n=t.elm,r=n.style;da(n,e),r.transform=r.WebkitTransform=r.transitionDuration="",n.addEventListener(aa,n._moveCb=function t(r){r&&r.target!==n||r&&!/transform$/.test(r.propertyName)||(n.removeEventListener(aa,t),n._moveCb=null,ha(n,e))})}})))},methods:{hasMove:function(t,e){if(!na)return!1;if(this._hasMove)return this._hasMove;var n=t.cloneNode();t._transitionClasses&&t._transitionClasses.forEach((function(t){Qo(n,t)})),Ko(n,e),n.style.display="none",this.$el.appendChild(n);var r=ma(n);return this.$el.removeChild(n),this._hasMove=r.hasTransform}}};function Xa(t){t.elm._moveCb&&t.elm._moveCb(),t.elm._enterCb&&t.elm._enterCb()}function Za(t){t.data.newPos=t.elm.getBoundingClientRect()}function Ja(t){var e=t.data.pos,n=t.data.newPos,r=e.left-n.left,i=e.top-n.top;if(r||i){t.data.moved=!0;var o=t.elm.style;o.transform=o.WebkitTransform="translate(".concat(r,"px,").concat(i,"px)"),o.transitionDuration="0s"}}var Ka={Transition:$a,TransitionGroup:qa};Kr.config.mustUseProp=mi,Kr.config.isReservedTag=ji,Kr.config.isReservedAttr=fi,Kr.config.getTagNamespace=Pi,Kr.config.isUnknownElement=Ni,P(Kr.options.directives,Na),P(Kr.options.components,Ka),Kr.prototype.__patch__=tt?La:N,Kr.prototype.$mount=function(t,e){return t=t&&tt?Hi(t):void 0,In(this,t,e)},tt&&setTimeout((function(){U.devtools&&ht&&ht.emit("init",Kr)}),0)}).call(this,n("c8ba"))},"2b10":function(t,e){function n(t,e,n){var r=-1,i=t.length;e<0&&(e=-e>i?0:i+e),n=n>i?i:n,n<0&&(n+=i),i=e>n?0:n-e>>>0,e>>>=0;var o=Array(i);while(++r<i)o[r]=t[r+e];return o}t.exports=n},"2b3e":function(t,e,n){var r=n("585a"),i="object"==typeof self&&self&&self.Object===Object&&self,o=r||i||Function("return this")();t.exports=o},"2ba4":function(t,e,n){"use strict";var r=n("40d5"),i=Function.prototype,o=i.apply,a=i.call;t.exports="object"==typeof Reflect&&Reflect.apply||(r?a.bind(o):function(){return a.apply(o,arguments)})},"2bfb":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("af",{months:"Januarie_Februarie_Maart_April_Mei_Junie_Julie_Augustus_September_Oktober_November_Desember".split("_"),monthsShort:"Jan_Feb_Mrt_Apr_Mei_Jun_Jul_Aug_Sep_Okt_Nov_Des".split("_"),weekdays:"Sondag_Maandag_Dinsdag_Woensdag_Donderdag_Vrydag_Saterdag".split("_"),weekdaysShort:"Son_Maa_Din_Woe_Don_Vry_Sat".split("_"),weekdaysMin:"So_Ma_Di_Wo_Do_Vr_Sa".split("_"),meridiemParse:/vm|nm/i,isPM:function(t){return/^nm$/i.test(t)},meridiem:function(t,e,n){return t<12?n?"vm":"VM":n?"nm":"NM"},longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Vandag om] LT",nextDay:"[Môre om] LT",nextWeek:"dddd [om] LT",lastDay:"[Gister om] LT",lastWeek:"[Laas] dddd [om] LT",sameElse:"L"},relativeTime:{future:"oor %s",past:"%s gelede",s:"'n paar sekondes",ss:"%d sekondes",m:"'n minuut",mm:"%d minute",h:"'n uur",hh:"%d ure",d:"'n dag",dd:"%d dae",M:"'n maand",MM:"%d maande",y:"'n jaar",yy:"%d jaar"},dayOfMonthOrdinalParse:/\d{1,2}(ste|de)/,ordinal:function(t){return t+(1===t||8===t||t>=20?"ste":"de")},week:{dow:1,doy:4}});return e}))},"2c3e":function(t,e,n){"use strict";var r=n("83ab"),i=n("9f7f").MISSED_STICKY,o=n("c6b6"),a=n("edd0"),s=n("69f3").get,l=RegExp.prototype,u=TypeError;r&&i&&a(l,"sticky",{configurable:!0,get:function(){if(this!==l){if("RegExp"===o(this))return!!s(this).sticky;throw new u("Incompatible receiver, RegExp required")}}})},"2cef":function(t,e,n){"use strict";t.exports=function(){var t=1;function e(){return t++}return{generate:e}}},"2cf4":function(t,e,n){"use strict";var r,i,o,a,s=n("da84"),l=n("2ba4"),u=n("0366"),c=n("1626"),d=n("1a2d"),h=n("d039"),f=n("1be4"),p=n("f36a"),m=n("cc12"),g=n("d6d6"),y=n("1cdc"),v=n("605d"),_=s.setImmediate,b=s.clearImmediate,x=s.process,w=s.Dispatch,M=s.Function,S=s.MessageChannel,k=s.String,L=0,T={},D="onreadystatechange";h((function(){r=s.location}));var O=function(t){if(d(T,t)){var e=T[t];delete T[t],e()}},I=function(t){return function(){O(t)}},C=function(t){O(t.data)},A=function(t){s.postMessage(k(t),r.protocol+"//"+r.host)};_&&b||(_=function(t){g(arguments.length,1);var e=c(t)?t:M(t),n=p(arguments,1);return T[++L]=function(){l(e,void 0,n)},i(L),L},b=function(t){delete T[t]},v?i=function(t){x.nextTick(I(t))}:w&&w.now?i=function(t){w.now(I(t))}:S&&!y?(o=new S,a=o.port2,o.port1.onmessage=C,i=u(a.postMessage,a)):s.addEventListener&&c(s.postMessage)&&!s.importScripts&&r&&"file:"!==r.protocol&&!h(A)?(i=A,s.addEventListener("message",C,!1)):i=D in m("script")?function(t){f.appendChild(m("script"))[D]=function(){f.removeChild(this),O(t)}}:function(t){setTimeout(I(t),0)}),t.exports={set:_,clear:b}},"2d00":function(t,e,n){"use strict";var r,i,o=n("da84"),a=n("342f"),s=o.process,l=o.Deno,u=s&&s.versions||l&&l.version,c=u&&u.v8;c&&(r=c.split("."),i=r[0]>0&&r[0]<4?1:+(r[0]+r[1])),!i&&a&&(r=a.match(/Edge\/(\d+)/),(!r||r[1]>=74)&&(r=a.match(/Chrome\/(\d+)/),r&&(i=+r[1]))),t.exports=i},"2d48":function(t,e,n){"use strict";var r=function(){var t=this,e=t._self._c;return e("label",{class:t.wrapClasses},[e("span",{class:t.radioClasses},[e("span",{class:t.innerClasses}),e("input",{class:t.inputClasses,attrs:{type:"radio",disabled:t.itemDisabled,name:t.groupName},domProps:{checked:t.currentValue},on:{change:t.change,focus:t.onFocus,blur:t.onBlur}})]),t._t("default",(function(){return[t._v(t._s(t.label))]}))],2)},i=[],o=n("ade3"),a=(n("99af"),n("b0c0"),n("a9e3"),n("3a27")),s=n("2507"),l=n("9f81"),u="ivu-radio",c={name:"Radio",mixins:[s["a"],l["a"]],props:{value:{type:[String,Number,Boolean],default:!1},trueValue:{type:[String,Number,Boolean],default:!0},falseValue:{type:[String,Number,Boolean],default:!1},label:{type:[String,Number]},disabled:{type:Boolean,default:!1},size:{validator:function(t){return Object(a["j"])(t,["small","large","default"])},default:function(){return this.$IVIEW&&""!==this.$IVIEW.size?this.$IVIEW.size:"default"}},name:{type:String},border:{type:Boolean,default:!1}},data:function(){return{currentValue:this.value,group:!1,groupName:this.name,parent:Object(a["f"])(this,"RadioGroup"),focusWrapper:!1,focusInner:!1}},computed:{wrapClasses:function(){return["".concat(u,"-wrapper"),Object(o["a"])(Object(o["a"])(Object(o["a"])(Object(o["a"])(Object(o["a"])(Object(o["a"])({},"".concat(u,"-group-item"),this.group),"".concat(u,"-wrapper-checked"),this.currentValue),"".concat(u,"-wrapper-disabled"),this.itemDisabled),"".concat(u,"-").concat(this.size),!!this.size),"".concat(u,"-focus"),this.focusWrapper),"".concat(u,"-border"),this.border)]},radioClasses:function(){return["".concat(u),Object(o["a"])(Object(o["a"])({},"".concat(u,"-checked"),this.currentValue),"".concat(u,"-disabled"),this.itemDisabled)]},innerClasses:function(){return["".concat(u,"-inner"),Object(o["a"])({},"".concat(u,"-focus"),this.focusInner)]},inputClasses:function(){return"".concat(u,"-input")}},mounted:function(){this.parent&&(this.group=!0,this.name&&this.name!==this.parent.name?console.warn:this.groupName=this.parent.name),this.group?this.parent.updateValue():this.updateValue()},methods:{change:function(t){if(this.itemDisabled)return!1;var e=t.target.checked;this.currentValue=e;var n=e?this.trueValue:this.falseValue;this.$emit("input",n),this.group?void 0!==this.label&&this.parent.change({value:this.label,checked:this.value}):(this.$emit("on-change",n),this.dispatch("FormItem","on-form-change",n))},updateValue:function(){this.currentValue=this.value===this.trueValue},onBlur:function(){this.focusWrapper=!1,this.focusInner=!1},onFocus:function(){this.group&&"button"===this.parent.type?this.focusWrapper=!0:this.focusInner=!0}},watch:{value:function(t){if(t!==this.trueValue&&t!==this.falseValue)throw"Value should be trueValue or falseValue.";this.updateValue()}}},d=c,h=n("2877"),f=Object(h["a"])(d,r,i,!1,null,null,null),p=f.exports,m=n("2430");p.Group=m["a"];e["a"]=p},"2d7c":function(t,e){function n(t,e){var n=-1,r=null==t?0:t.length,i=0,o=[];while(++n<r){var a=t[n];e(a,n,t)&&(o[i++]=a)}return o}t.exports=n},"2dcb":function(t,e,n){var r=n("91e9"),i=r(Object.getPrototypeOf,Object);t.exports=i},"2e8c":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("uz",{months:"январ_феврал_март_апрел_май_июн_июл_август_сентябр_октябр_ноябр_декабр".split("_"),monthsShort:"янв_фев_мар_апр_май_июн_июл_авг_сен_окт_ноя_дек".split("_"),weekdays:"Якшанба_Душанба_Сешанба_Чоршанба_Пайшанба_Жума_Шанба".split("_"),weekdaysShort:"Якш_Душ_Сеш_Чор_Пай_Жум_Шан".split("_"),weekdaysMin:"Як_Ду_Се_Чо_Па_Жу_Ша".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"D MMMM YYYY, dddd HH:mm"},calendar:{sameDay:"[Бугун соат] LT [да]",nextDay:"[Эртага] LT [да]",nextWeek:"dddd [куни соат] LT [да]",lastDay:"[Кеча соат] LT [да]",lastWeek:"[Утган] dddd [куни соат] LT [да]",sameElse:"L"},relativeTime:{future:"Якин %s ичида",past:"Бир неча %s олдин",s:"фурсат",ss:"%d фурсат",m:"бир дакика",mm:"%d дакика",h:"бир соат",hh:"%d соат",d:"бир кун",dd:"%d кун",M:"бир ой",MM:"%d ой",y:"бир йил",yy:"%d йил"},week:{dow:1,doy:7}});return e}))},"2ec1":function(t,e,n){var r=n("100e"),i=n("9aff");function o(t){return r((function(e,n){var r=-1,o=n.length,a=o>1?n[o-1]:void 0,s=o>2?n[2]:void 0;a=t.length>3&&"function"==typeof a?(o--,a):void 0,s&&i(n[0],n[1],s)&&(a=o<3?void 0:a,o=1),e=Object(e);while(++r<o){var l=n[r];l&&t(e,l,r,a)}return e}))}t.exports=o},"2f62":function(t,e,n){"use strict";(function(t){
/*!
 * vuex v3.6.2
 * (c) 2021 Evan You
 * @license MIT
 */
function n(t){var e=Number(t.version.split(".")[0]);if(e>=2)t.mixin({beforeCreate:r});else{var n=t.prototype._init;t.prototype._init=function(t){void 0===t&&(t={}),t.init=t.init?[r].concat(t.init):r,n.call(this,t)}}function r(){var t=this.$options;t.store?this.$store="function"===typeof t.store?t.store():t.store:t.parent&&t.parent.$store&&(this.$store=t.parent.$store)}}var r="undefined"!==typeof window?window:"undefined"!==typeof t?t:{},i=r.__VUE_DEVTOOLS_GLOBAL_HOOK__;function o(t){i&&(t._devtoolHook=i,i.emit("vuex:init",t),i.on("vuex:travel-to-state",(function(e){t.replaceState(e)})),t.subscribe((function(t,e){i.emit("vuex:mutation",t,e)}),{prepend:!0}),t.subscribeAction((function(t,e){i.emit("vuex:action",t,e)}),{prepend:!0}))}function a(t,e){return t.filter(e)[0]}function s(t,e){if(void 0===e&&(e=[]),null===t||"object"!==typeof t)return t;var n=a(e,(function(e){return e.original===t}));if(n)return n.copy;var r=Array.isArray(t)?[]:{};return e.push({original:t,copy:r}),Object.keys(t).forEach((function(n){r[n]=s(t[n],e)})),r}function l(t,e){Object.keys(t).forEach((function(n){return e(t[n],n)}))}function u(t){return null!==t&&"object"===typeof t}function c(t){return t&&"function"===typeof t.then}function d(t,e){return function(){return t(e)}}var h=function(t,e){this.runtime=e,this._children=Object.create(null),this._rawModule=t;var n=t.state;this.state=("function"===typeof n?n():n)||{}},f={namespaced:{configurable:!0}};f.namespaced.get=function(){return!!this._rawModule.namespaced},h.prototype.addChild=function(t,e){this._children[t]=e},h.prototype.removeChild=function(t){delete this._children[t]},h.prototype.getChild=function(t){return this._children[t]},h.prototype.hasChild=function(t){return t in this._children},h.prototype.update=function(t){this._rawModule.namespaced=t.namespaced,t.actions&&(this._rawModule.actions=t.actions),t.mutations&&(this._rawModule.mutations=t.mutations),t.getters&&(this._rawModule.getters=t.getters)},h.prototype.forEachChild=function(t){l(this._children,t)},h.prototype.forEachGetter=function(t){this._rawModule.getters&&l(this._rawModule.getters,t)},h.prototype.forEachAction=function(t){this._rawModule.actions&&l(this._rawModule.actions,t)},h.prototype.forEachMutation=function(t){this._rawModule.mutations&&l(this._rawModule.mutations,t)},Object.defineProperties(h.prototype,f);var p=function(t){this.register([],t,!1)};function m(t,e,n){if(e.update(n),n.modules)for(var r in n.modules){if(!e.getChild(r))return void 0;m(t.concat(r),e.getChild(r),n.modules[r])}}p.prototype.get=function(t){return t.reduce((function(t,e){return t.getChild(e)}),this.root)},p.prototype.getNamespace=function(t){var e=this.root;return t.reduce((function(t,n){return e=e.getChild(n),t+(e.namespaced?n+"/":"")}),"")},p.prototype.update=function(t){m([],this.root,t)},p.prototype.register=function(t,e,n){var r=this;void 0===n&&(n=!0);var i=new h(e,n);if(0===t.length)this.root=i;else{var o=this.get(t.slice(0,-1));o.addChild(t[t.length-1],i)}e.modules&&l(e.modules,(function(e,i){r.register(t.concat(i),e,n)}))},p.prototype.unregister=function(t){var e=this.get(t.slice(0,-1)),n=t[t.length-1],r=e.getChild(n);r&&r.runtime&&e.removeChild(n)},p.prototype.isRegistered=function(t){var e=this.get(t.slice(0,-1)),n=t[t.length-1];return!!e&&e.hasChild(n)};var g;var y=function(t){var e=this;void 0===t&&(t={}),!g&&"undefined"!==typeof window&&window.Vue&&C(window.Vue);var n=t.plugins;void 0===n&&(n=[]);var r=t.strict;void 0===r&&(r=!1),this._committing=!1,this._actions=Object.create(null),this._actionSubscribers=[],this._mutations=Object.create(null),this._wrappedGetters=Object.create(null),this._modules=new p(t),this._modulesNamespaceMap=Object.create(null),this._subscribers=[],this._watcherVM=new g,this._makeLocalGettersCache=Object.create(null);var i=this,a=this,s=a.dispatch,l=a.commit;this.dispatch=function(t,e){return s.call(i,t,e)},this.commit=function(t,e,n){return l.call(i,t,e,n)},this.strict=r;var u=this._modules.root.state;w(this,u,[],this._modules.root),x(this,u),n.forEach((function(t){return t(e)}));var c=void 0!==t.devtools?t.devtools:g.config.devtools;c&&o(this)},v={state:{configurable:!0}};function _(t,e,n){return e.indexOf(t)<0&&(n&&n.prepend?e.unshift(t):e.push(t)),function(){var n=e.indexOf(t);n>-1&&e.splice(n,1)}}function b(t,e){t._actions=Object.create(null),t._mutations=Object.create(null),t._wrappedGetters=Object.create(null),t._modulesNamespaceMap=Object.create(null);var n=t.state;w(t,n,[],t._modules.root,!0),x(t,n,e)}function x(t,e,n){var r=t._vm;t.getters={},t._makeLocalGettersCache=Object.create(null);var i=t._wrappedGetters,o={};l(i,(function(e,n){o[n]=d(e,t),Object.defineProperty(t.getters,n,{get:function(){return t._vm[n]},enumerable:!0})}));var a=g.config.silent;g.config.silent=!0,t._vm=new g({data:{$$state:e},computed:o}),g.config.silent=a,t.strict&&D(t),r&&(n&&t._withCommit((function(){r._data.$$state=null})),g.nextTick((function(){return r.$destroy()})))}function w(t,e,n,r,i){var o=!n.length,a=t._modules.getNamespace(n);if(r.namespaced&&(t._modulesNamespaceMap[a],t._modulesNamespaceMap[a]=r),!o&&!i){var s=O(e,n.slice(0,-1)),l=n[n.length-1];t._withCommit((function(){g.set(s,l,r.state)}))}var u=r.context=M(t,a,n);r.forEachMutation((function(e,n){var r=a+n;k(t,r,e,u)})),r.forEachAction((function(e,n){var r=e.root?n:a+n,i=e.handler||e;L(t,r,i,u)})),r.forEachGetter((function(e,n){var r=a+n;T(t,r,e,u)})),r.forEachChild((function(r,o){w(t,e,n.concat(o),r,i)}))}function M(t,e,n){var r=""===e,i={dispatch:r?t.dispatch:function(n,r,i){var o=I(n,r,i),a=o.payload,s=o.options,l=o.type;return s&&s.root||(l=e+l),t.dispatch(l,a)},commit:r?t.commit:function(n,r,i){var o=I(n,r,i),a=o.payload,s=o.options,l=o.type;s&&s.root||(l=e+l),t.commit(l,a,s)}};return Object.defineProperties(i,{getters:{get:r?function(){return t.getters}:function(){return S(t,e)}},state:{get:function(){return O(t.state,n)}}}),i}function S(t,e){if(!t._makeLocalGettersCache[e]){var n={},r=e.length;Object.keys(t.getters).forEach((function(i){if(i.slice(0,r)===e){var o=i.slice(r);Object.defineProperty(n,o,{get:function(){return t.getters[i]},enumerable:!0})}})),t._makeLocalGettersCache[e]=n}return t._makeLocalGettersCache[e]}function k(t,e,n,r){var i=t._mutations[e]||(t._mutations[e]=[]);i.push((function(e){n.call(t,r.state,e)}))}function L(t,e,n,r){var i=t._actions[e]||(t._actions[e]=[]);i.push((function(e){var i=n.call(t,{dispatch:r.dispatch,commit:r.commit,getters:r.getters,state:r.state,rootGetters:t.getters,rootState:t.state},e);return c(i)||(i=Promise.resolve(i)),t._devtoolHook?i.catch((function(e){throw t._devtoolHook.emit("vuex:error",e),e})):i}))}function T(t,e,n,r){t._wrappedGetters[e]||(t._wrappedGetters[e]=function(t){return n(r.state,r.getters,t.state,t.getters)})}function D(t){t._vm.$watch((function(){return this._data.$$state}),(function(){0}),{deep:!0,sync:!0})}function O(t,e){return e.reduce((function(t,e){return t[e]}),t)}function I(t,e,n){return u(t)&&t.type&&(n=e,e=t,t=t.type),{type:t,payload:e,options:n}}function C(t){g&&t===g||(g=t,n(g))}v.state.get=function(){return this._vm._data.$$state},v.state.set=function(t){0},y.prototype.commit=function(t,e,n){var r=this,i=I(t,e,n),o=i.type,a=i.payload,s=(i.options,{type:o,payload:a}),l=this._mutations[o];l&&(this._withCommit((function(){l.forEach((function(t){t(a)}))})),this._subscribers.slice().forEach((function(t){return t(s,r.state)})))},y.prototype.dispatch=function(t,e){var n=this,r=I(t,e),i=r.type,o=r.payload,a={type:i,payload:o},s=this._actions[i];if(s){try{this._actionSubscribers.slice().filter((function(t){return t.before})).forEach((function(t){return t.before(a,n.state)}))}catch(u){0}var l=s.length>1?Promise.all(s.map((function(t){return t(o)}))):s[0](o);return new Promise((function(t,e){l.then((function(e){try{n._actionSubscribers.filter((function(t){return t.after})).forEach((function(t){return t.after(a,n.state)}))}catch(u){0}t(e)}),(function(t){try{n._actionSubscribers.filter((function(t){return t.error})).forEach((function(e){return e.error(a,n.state,t)}))}catch(u){0}e(t)}))}))}},y.prototype.subscribe=function(t,e){return _(t,this._subscribers,e)},y.prototype.subscribeAction=function(t,e){var n="function"===typeof t?{before:t}:t;return _(n,this._actionSubscribers,e)},y.prototype.watch=function(t,e,n){var r=this;return this._watcherVM.$watch((function(){return t(r.state,r.getters)}),e,n)},y.prototype.replaceState=function(t){var e=this;this._withCommit((function(){e._vm._data.$$state=t}))},y.prototype.registerModule=function(t,e,n){void 0===n&&(n={}),"string"===typeof t&&(t=[t]),this._modules.register(t,e),w(this,this.state,t,this._modules.get(t),n.preserveState),x(this,this.state)},y.prototype.unregisterModule=function(t){var e=this;"string"===typeof t&&(t=[t]),this._modules.unregister(t),this._withCommit((function(){var n=O(e.state,t.slice(0,-1));g.delete(n,t[t.length-1])})),b(this)},y.prototype.hasModule=function(t){return"string"===typeof t&&(t=[t]),this._modules.isRegistered(t)},y.prototype.hotUpdate=function(t){this._modules.update(t),b(this,!0)},y.prototype._withCommit=function(t){var e=this._committing;this._committing=!0,t(),this._committing=e},Object.defineProperties(y.prototype,v);var A=H((function(t,e){var n={};return N(e).forEach((function(e){var r=e.key,i=e.val;n[r]=function(){var e=this.$store.state,n=this.$store.getters;if(t){var r=z(this.$store,"mapState",t);if(!r)return;e=r.context.state,n=r.context.getters}return"function"===typeof i?i.call(this,e,n):e[i]},n[r].vuex=!0})),n})),Y=H((function(t,e){var n={};return N(e).forEach((function(e){var r=e.key,i=e.val;n[r]=function(){var e=[],n=arguments.length;while(n--)e[n]=arguments[n];var r=this.$store.commit;if(t){var o=z(this.$store,"mapMutations",t);if(!o)return;r=o.context.commit}return"function"===typeof i?i.apply(this,[r].concat(e)):r.apply(this.$store,[i].concat(e))}})),n})),j=H((function(t,e){var n={};return N(e).forEach((function(e){var r=e.key,i=e.val;i=t+i,n[r]=function(){if(!t||z(this.$store,"mapGetters",t))return this.$store.getters[i]},n[r].vuex=!0})),n})),P=H((function(t,e){var n={};return N(e).forEach((function(e){var r=e.key,i=e.val;n[r]=function(){var e=[],n=arguments.length;while(n--)e[n]=arguments[n];var r=this.$store.dispatch;if(t){var o=z(this.$store,"mapActions",t);if(!o)return;r=o.context.dispatch}return"function"===typeof i?i.apply(this,[r].concat(e)):r.apply(this.$store,[i].concat(e))}})),n})),E=function(t){return{mapState:A.bind(null,t),mapGetters:j.bind(null,t),mapMutations:Y.bind(null,t),mapActions:P.bind(null,t)}};function N(t){return R(t)?Array.isArray(t)?t.map((function(t){return{key:t,val:t}})):Object.keys(t).map((function(e){return{key:e,val:t[e]}})):[]}function R(t){return Array.isArray(t)||u(t)}function H(t){return function(e,n){return"string"!==typeof e?(n=e,e=""):"/"!==e.charAt(e.length-1)&&(e+="/"),t(e,n)}}function z(t,e,n){var r=t._modulesNamespaceMap[n];return r}function F(t){void 0===t&&(t={});var e=t.collapsed;void 0===e&&(e=!0);var n=t.filter;void 0===n&&(n=function(t,e,n){return!0});var r=t.transformer;void 0===r&&(r=function(t){return t});var i=t.mutationTransformer;void 0===i&&(i=function(t){return t});var o=t.actionFilter;void 0===o&&(o=function(t,e){return!0});var a=t.actionTransformer;void 0===a&&(a=function(t){return t});var l=t.logMutations;void 0===l&&(l=!0);var u=t.logActions;void 0===u&&(u=!0);var c=t.logger;return void 0===c&&(c=console),function(t){var d=s(t.state);"undefined"!==typeof c&&(l&&t.subscribe((function(t,o){var a=s(o);if(n(t,d,a)){var l=V(),u=i(t),h="mutation "+t.type+l;B(c,h,e),c.log("%c prev state","color: #9E9E9E; font-weight: bold",r(d)),c.log("%c mutation","color: #03A9F4; font-weight: bold",u),c.log("%c next state","color: #4CAF50; font-weight: bold",r(a)),W(c)}d=a})),u&&t.subscribeAction((function(t,n){if(o(t,n)){var r=V(),i=a(t),s="action "+t.type+r;B(c,s,e),c.log("%c action","color: #03A9F4; font-weight: bold",i),W(c)}})))}}function B(t,e,n){var r=n?t.groupCollapsed:t.group;try{r.call(t,e)}catch(i){t.log(e)}}function W(t){try{t.groupEnd()}catch(e){t.log("—— log end ——")}}function V(){var t=new Date;return" @ "+$(t.getHours(),2)+":"+$(t.getMinutes(),2)+":"+$(t.getSeconds(),2)+"."+$(t.getMilliseconds(),3)}function G(t,e){return new Array(e+1).join(t)}function $(t,e){return G("0",e-t.toString().length)+t}var U={Store:y,install:C,version:"3.6.2",mapState:A,mapMutations:Y,mapGetters:j,mapActions:P,createNamespacedHelpers:E,createLogger:F};e["a"]=U}).call(this,n("c8ba"))},"2fcc":function(t,e){function n(t){var e=this.__data__,n=e["delete"](t);return this.size=e.size,n}t.exports=n},"30c9":function(t,e,n){var r=n("9520"),i=n("b218");function o(t){return null!=t&&i(t.length)&&!r(t)}t.exports=o},"30e0":function(t,e,n){var r=n("99cd"),i=r(!0);t.exports=i},"313e":function(t,e,n){"use strict";n.d(e,"a",(function(){return mb}));var r={};n.r(r),n.d(r,"updateProps",(function(){return Wl})),n.d(r,"initProps",(function(){return Vl})),n.d(r,"removeElement",(function(){return $l})),n.d(r,"removeElementWithFadeOut",(function(){return ql})),n.d(r,"isElementRemoved",(function(){return Gl})),n.d(r,"extendShape",(function(){return ug})),n.d(r,"extendPath",(function(){return dg})),n.d(r,"registerShape",(function(){return hg})),n.d(r,"getShapeClass",(function(){return fg})),n.d(r,"makePath",(function(){return pg})),n.d(r,"makeImage",(function(){return mg})),n.d(r,"mergePath",(function(){return yg})),n.d(r,"resizePath",(function(){return vg})),n.d(r,"subPixelOptimizeLine",(function(){return _g})),n.d(r,"subPixelOptimizeRect",(function(){return bg})),n.d(r,"subPixelOptimize",(function(){return xg})),n.d(r,"getTransform",(function(){return wg})),n.d(r,"applyTransform",(function(){return Mg})),n.d(r,"transformDirection",(function(){return Sg})),n.d(r,"groupTransition",(function(){return Tg})),n.d(r,"clipPointsByRect",(function(){return Dg})),n.d(r,"clipRectByRect",(function(){return Og})),n.d(r,"createIcon",(function(){return Ig})),n.d(r,"linePolygonIntersect",(function(){return Cg})),n.d(r,"lineLineIntersect",(function(){return Ag})),n.d(r,"setTooltipConfig",(function(){return Pg})),n.d(r,"traverseElements",(function(){return Ng})),n.d(r,"Group",(function(){return wr})),n.d(r,"Image",(function(){return Ka})),n.d(r,"Text",(function(){return ks})),n.d(r,"Circle",(function(){return Kp})),n.d(r,"Ellipse",(function(){return em})),n.d(r,"Sector",(function(){return _m})),n.d(r,"Ring",(function(){return wm})),n.d(r,"Polygon",(function(){return Tm})),n.d(r,"Polyline",(function(){return Im})),n.d(r,"Rect",(function(){return ss})),n.d(r,"Line",(function(){return jm})),n.d(r,"BezierCurve",(function(){return Hm})),n.d(r,"Arc",(function(){return Bm})),n.d(r,"IncrementalDisplayable",(function(){return og})),n.d(r,"CompoundPath",(function(){return Vm})),n.d(r,"LinearGradient",(function(){return qm})),n.d(r,"RadialGradient",(function(){return Zm})),n.d(r,"BoundingRect",(function(){return Tt})),n.d(r,"OrientedBoundingRect",(function(){return ng})),n.d(r,"Point",(function(){return yt})),n.d(r,"Path",(function(){return Va}));
/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
var i=function(t,e){return i=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)Object.prototype.hasOwnProperty.call(e,n)&&(t[n]=e[n])},i(t,e)};function o(t,e){if("function"!==typeof e&&null!==e)throw new TypeError("Class extends value "+String(e)+" is not a constructor or null");function n(){this.constructor=t}i(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}Object.create;Object.create;var a=n("22d1"),s=n("6d8b");function l(t,e){return null==t&&(t=0),null==e&&(e=0),[t,e]}function u(t,e){return t[0]=e[0],t[1]=e[1],t}function c(t){return[t[0],t[1]]}function d(t,e,n){return t[0]=e,t[1]=n,t}function h(t,e,n){return t[0]=e[0]+n[0],t[1]=e[1]+n[1],t}function f(t,e,n,r){return t[0]=e[0]+n[0]*r,t[1]=e[1]+n[1]*r,t}function p(t,e,n){return t[0]=e[0]-n[0],t[1]=e[1]-n[1],t}function m(t){return Math.sqrt(g(t))}function g(t){return t[0]*t[0]+t[1]*t[1]}function y(t,e,n){return t[0]=e[0]*n,t[1]=e[1]*n,t}function v(t,e){var n=m(e);return 0===n?(t[0]=0,t[1]=0):(t[0]=e[0]/n,t[1]=e[1]/n),t}function _(t,e){return Math.sqrt((t[0]-e[0])*(t[0]-e[0])+(t[1]-e[1])*(t[1]-e[1]))}var b=_;function x(t,e){return(t[0]-e[0])*(t[0]-e[0])+(t[1]-e[1])*(t[1]-e[1])}var w=x;function M(t,e,n,r){return t[0]=e[0]+r*(n[0]-e[0]),t[1]=e[1]+r*(n[1]-e[1]),t}function S(t,e,n){var r=e[0],i=e[1];return t[0]=n[0]*r+n[2]*i+n[4],t[1]=n[1]*r+n[3]*i+n[5],t}function k(t,e,n){return t[0]=Math.min(e[0],n[0]),t[1]=Math.min(e[1],n[1]),t}function L(t,e,n){return t[0]=Math.max(e[0],n[0]),t[1]=Math.max(e[1],n[1]),t}var T=function(){function t(t,e){this.target=t,this.topTarget=e&&e.topTarget}return t}(),D=function(){function t(t){this.handler=t,t.on("mousedown",this._dragStart,this),t.on("mousemove",this._drag,this),t.on("mouseup",this._dragEnd,this)}return t.prototype._dragStart=function(t){var e=t.target;while(e&&!e.draggable)e=e.parent||e.__hostTarget;e&&(this._draggingTarget=e,e.dragging=!0,this._x=t.offsetX,this._y=t.offsetY,this.handler.dispatchToElement(new T(e,t),"dragstart",t.event))},t.prototype._drag=function(t){var e=this._draggingTarget;if(e){var n=t.offsetX,r=t.offsetY,i=n-this._x,o=r-this._y;this._x=n,this._y=r,e.drift(i,o,t),this.handler.dispatchToElement(new T(e,t),"drag",t.event);var a=this.handler.findHover(n,r,e).target,s=this._dropTarget;this._dropTarget=a,e!==a&&(s&&a!==s&&this.handler.dispatchToElement(new T(s,t),"dragleave",t.event),a&&a!==s&&this.handler.dispatchToElement(new T(a,t),"dragenter",t.event))}},t.prototype._dragEnd=function(t){var e=this._draggingTarget;e&&(e.dragging=!1),this.handler.dispatchToElement(new T(e,t),"dragend",t.event),this._dropTarget&&this.handler.dispatchToElement(new T(this._dropTarget,t),"drop",t.event),this._draggingTarget=null,this._dropTarget=null},t}(),O=D,I=function(){function t(t){t&&(this._$eventProcessor=t)}return t.prototype.on=function(t,e,n,r){this._$handlers||(this._$handlers={});var i=this._$handlers;if("function"===typeof e&&(r=n,n=e,e=null),!n||!t)return this;var o=this._$eventProcessor;null!=e&&o&&o.normalizeQuery&&(e=o.normalizeQuery(e)),i[t]||(i[t]=[]);for(var a=0;a<i[t].length;a++)if(i[t][a].h===n)return this;var s={h:n,query:e,ctx:r||this,callAtLast:n.zrEventfulCallAtLast},l=i[t].length-1,u=i[t][l];return u&&u.callAtLast?i[t].splice(l,0,s):i[t].push(s),this},t.prototype.isSilent=function(t){var e=this._$handlers;return!e||!e[t]||!e[t].length},t.prototype.off=function(t,e){var n=this._$handlers;if(!n)return this;if(!t)return this._$handlers={},this;if(e){if(n[t]){for(var r=[],i=0,o=n[t].length;i<o;i++)n[t][i].h!==e&&r.push(n[t][i]);n[t]=r}n[t]&&0===n[t].length&&delete n[t]}else delete n[t];return this},t.prototype.trigger=function(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];if(!this._$handlers)return this;var r=this._$handlers[t],i=this._$eventProcessor;if(r)for(var o=e.length,a=r.length,s=0;s<a;s++){var l=r[s];if(!i||!i.filter||null==l.query||i.filter(t,l.query))switch(o){case 0:l.h.call(l.ctx);break;case 1:l.h.call(l.ctx,e[0]);break;case 2:l.h.call(l.ctx,e[0],e[1]);break;default:l.h.apply(l.ctx,e);break}}return i&&i.afterTrigger&&i.afterTrigger(t),this},t.prototype.triggerWithContext=function(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];if(!this._$handlers)return this;var r=this._$handlers[t],i=this._$eventProcessor;if(r)for(var o=e.length,a=e[o-1],s=r.length,l=0;l<s;l++){var u=r[l];if(!i||!i.filter||null==u.query||i.filter(t,u.query))switch(o){case 0:u.h.call(a);break;case 1:u.h.call(a,e[0]);break;case 2:u.h.call(a,e[0],e[1]);break;default:u.h.apply(a,e.slice(1,o-1));break}}return i&&i.afterTrigger&&i.afterTrigger(t),this},t}(),C=I,A=Math.log(2);function Y(t,e,n,r,i,o){var a=r+"-"+i,s=t.length;if(o.hasOwnProperty(a))return o[a];if(1===e){var l=Math.round(Math.log((1<<s)-1&~i)/A);return t[n][l]}var u=r|1<<n,c=n+1;while(r&1<<c)c++;for(var d=0,h=0,f=0;h<s;h++){var p=1<<h;p&i||(d+=(f%2?-1:1)*t[n][h]*Y(t,e-1,c,u,i|p,o),f++)}return o[a]=d,d}function j(t,e){var n=[[t[0],t[1],1,0,0,0,-e[0]*t[0],-e[0]*t[1]],[0,0,0,t[0],t[1],1,-e[1]*t[0],-e[1]*t[1]],[t[2],t[3],1,0,0,0,-e[2]*t[2],-e[2]*t[3]],[0,0,0,t[2],t[3],1,-e[3]*t[2],-e[3]*t[3]],[t[4],t[5],1,0,0,0,-e[4]*t[4],-e[4]*t[5]],[0,0,0,t[4],t[5],1,-e[5]*t[4],-e[5]*t[5]],[t[6],t[7],1,0,0,0,-e[6]*t[6],-e[6]*t[7]],[0,0,0,t[6],t[7],1,-e[7]*t[6],-e[7]*t[7]]],r={},i=Y(n,8,0,0,0,r);if(0!==i){for(var o=[],a=0;a<8;a++)for(var s=0;s<8;s++)null==o[s]&&(o[s]=0),o[s]+=((a+s)%2?-1:1)*Y(n,7,0===a?1:0,1<<a,1<<s,r)/i*e[a];return function(t,e,n){var r=e*o[6]+n*o[7]+1;t[0]=(e*o[0]+n*o[1]+o[2])/r,t[1]=(e*o[3]+n*o[4]+o[5])/r}}}var P="___zrEVENTSAVED",E=[];function N(t,e,n,r,i){return R(E,e,r,i,!0)&&R(t,n,E[0],E[1])}function R(t,e,n,r,i){if(e.getBoundingClientRect&&a["a"].domSupported&&!F(e)){var o=e[P]||(e[P]={}),s=H(e,o),l=z(s,o,i);if(l)return l(t,n,r),!0}return!1}function H(t,e){var n=e.markers;if(n)return n;n=e.markers=[];for(var r=["left","right"],i=["top","bottom"],o=0;o<4;o++){var a=document.createElement("div"),s=a.style,l=o%2,u=(o>>1)%2;s.cssText=["position: absolute","visibility: hidden","padding: 0","margin: 0","border-width: 0","user-select: none","width:0","height:0",r[l]+":0",i[u]+":0",r[1-l]+":auto",i[1-u]+":auto",""].join("!important;"),t.appendChild(a),n.push(a)}return n}function z(t,e,n){for(var r=n?"invTrans":"trans",i=e[r],o=e.srcCoords,a=[],s=[],l=!0,u=0;u<4;u++){var c=t[u].getBoundingClientRect(),d=2*u,h=c.left,f=c.top;a.push(h,f),l=l&&o&&h===o[d]&&f===o[d+1],s.push(t[u].offsetLeft,t[u].offsetTop)}return l&&i?i:(e.srcCoords=a,e[r]=n?j(s,a):j(a,s))}function F(t){return"CANVAS"===t.nodeName.toUpperCase()}var B=/([&<>"'])/g,W={"&":"&amp;","<":"&lt;",">":"&gt;",'"':"&quot;","'":"&#39;"};function V(t){return null==t?"":(t+"").replace(B,(function(t,e){return W[e]}))}var G=/^(?:mouse|pointer|contextmenu|drag|drop)|click/,$=[],U=a["a"].browser.firefox&&+a["a"].browser.version.split(".")[0]<39;function q(t,e,n,r){return n=n||{},r?X(t,e,n):U&&null!=e.layerX&&e.layerX!==e.offsetX?(n.zrX=e.layerX,n.zrY=e.layerY):null!=e.offsetX?(n.zrX=e.offsetX,n.zrY=e.offsetY):X(t,e,n),n}function X(t,e,n){if(a["a"].domSupported&&t.getBoundingClientRect){var r=e.clientX,i=e.clientY;if(F(t)){var o=t.getBoundingClientRect();return n.zrX=r-o.left,void(n.zrY=i-o.top)}if(R($,t,r,i))return n.zrX=$[0],void(n.zrY=$[1])}n.zrX=n.zrY=0}function Z(t){return t||window.event}function J(t,e,n){if(e=Z(e),null!=e.zrX)return e;var r=e.type,i=r&&r.indexOf("touch")>=0;if(i){var o="touchend"!==r?e.targetTouches[0]:e.changedTouches[0];o&&q(t,o,e,n)}else{q(t,e,e,n);var a=K(e);e.zrDelta=a?a/120:-(e.detail||0)/3}var s=e.button;return null==e.which&&void 0!==s&&G.test(e.type)&&(e.which=1&s?1:2&s?3:4&s?2:0),e}function K(t){var e=t.wheelDelta;if(e)return e;var n=t.deltaX,r=t.deltaY;if(null==n||null==r)return e;var i=0!==r?Math.abs(r):Math.abs(n),o=r>0?-1:r<0?1:n>0?-1:1;return 3*i*o}function Q(t,e,n,r){t.addEventListener(e,n,r)}function tt(t,e,n,r){t.removeEventListener(e,n,r)}var et=function(t){t.preventDefault(),t.stopPropagation(),t.cancelBubble=!0};function nt(t){return 2===t.which||3===t.which}var rt=function(){function t(){this._track=[]}return t.prototype.recognize=function(t,e,n){return this._doTrack(t,e,n),this._recognize(t)},t.prototype.clear=function(){return this._track.length=0,this},t.prototype._doTrack=function(t,e,n){var r=t.touches;if(r){for(var i={points:[],touches:[],target:e,event:t},o=0,a=r.length;o<a;o++){var s=r[o],l=q(n,s,{});i.points.push([l.zrX,l.zrY]),i.touches.push(s)}this._track.push(i)}},t.prototype._recognize=function(t){for(var e in at)if(at.hasOwnProperty(e)){var n=at[e](this._track,t);if(n)return n}},t}();function it(t){var e=t[1][0]-t[0][0],n=t[1][1]-t[0][1];return Math.sqrt(e*e+n*n)}function ot(t){return[(t[0][0]+t[1][0])/2,(t[0][1]+t[1][1])/2]}var at={pinch:function(t,e){var n=t.length;if(n){var r=(t[n-1]||{}).points,i=(t[n-2]||{}).points||r;if(i&&i.length>1&&r&&r.length>1){var o=it(r)/it(i);!isFinite(o)&&(o=1),e.pinchScale=o;var a=ot(r);return e.pinchX=a[0],e.pinchY=a[1],{type:"pinch",target:t[0].target,event:e}}}}};function st(){return[1,0,0,1,0,0]}function lt(t){return t[0]=1,t[1]=0,t[2]=0,t[3]=1,t[4]=0,t[5]=0,t}function ut(t,e){return t[0]=e[0],t[1]=e[1],t[2]=e[2],t[3]=e[3],t[4]=e[4],t[5]=e[5],t}function ct(t,e,n){var r=e[0]*n[0]+e[2]*n[1],i=e[1]*n[0]+e[3]*n[1],o=e[0]*n[2]+e[2]*n[3],a=e[1]*n[2]+e[3]*n[3],s=e[0]*n[4]+e[2]*n[5]+e[4],l=e[1]*n[4]+e[3]*n[5]+e[5];return t[0]=r,t[1]=i,t[2]=o,t[3]=a,t[4]=s,t[5]=l,t}function dt(t,e,n){return t[0]=e[0],t[1]=e[1],t[2]=e[2],t[3]=e[3],t[4]=e[4]+n[0],t[5]=e[5]+n[1],t}function ht(t,e,n,r){void 0===r&&(r=[0,0]);var i=e[0],o=e[2],a=e[4],s=e[1],l=e[3],u=e[5],c=Math.sin(n),d=Math.cos(n);return t[0]=i*d+s*c,t[1]=-i*c+s*d,t[2]=o*d+l*c,t[3]=-o*c+d*l,t[4]=d*(a-r[0])+c*(u-r[1])+r[0],t[5]=d*(u-r[1])-c*(a-r[0])+r[1],t}function ft(t,e,n){var r=n[0],i=n[1];return t[0]=e[0]*r,t[1]=e[1]*i,t[2]=e[2]*r,t[3]=e[3]*i,t[4]=e[4]*r,t[5]=e[5]*i,t}function pt(t,e){var n=e[0],r=e[2],i=e[4],o=e[1],a=e[3],s=e[5],l=n*a-o*r;return l?(l=1/l,t[0]=a*l,t[1]=-o*l,t[2]=-r*l,t[3]=n*l,t[4]=(r*s-a*i)*l,t[5]=(o*i-n*s)*l,t):null}function mt(t){var e=st();return ut(e,t),e}var gt=function(){function t(t,e){this.x=t||0,this.y=e||0}return t.prototype.copy=function(t){return this.x=t.x,this.y=t.y,this},t.prototype.clone=function(){return new t(this.x,this.y)},t.prototype.set=function(t,e){return this.x=t,this.y=e,this},t.prototype.equal=function(t){return t.x===this.x&&t.y===this.y},t.prototype.add=function(t){return this.x+=t.x,this.y+=t.y,this},t.prototype.scale=function(t){this.x*=t,this.y*=t},t.prototype.scaleAndAdd=function(t,e){this.x+=t.x*e,this.y+=t.y*e},t.prototype.sub=function(t){return this.x-=t.x,this.y-=t.y,this},t.prototype.dot=function(t){return this.x*t.x+this.y*t.y},t.prototype.len=function(){return Math.sqrt(this.x*this.x+this.y*this.y)},t.prototype.lenSquare=function(){return this.x*this.x+this.y*this.y},t.prototype.normalize=function(){var t=this.len();return this.x/=t,this.y/=t,this},t.prototype.distance=function(t){var e=this.x-t.x,n=this.y-t.y;return Math.sqrt(e*e+n*n)},t.prototype.distanceSquare=function(t){var e=this.x-t.x,n=this.y-t.y;return e*e+n*n},t.prototype.negate=function(){return this.x=-this.x,this.y=-this.y,this},t.prototype.transform=function(t){if(t){var e=this.x,n=this.y;return this.x=t[0]*e+t[2]*n+t[4],this.y=t[1]*e+t[3]*n+t[5],this}},t.prototype.toArray=function(t){return t[0]=this.x,t[1]=this.y,t},t.prototype.fromArray=function(t){this.x=t[0],this.y=t[1]},t.set=function(t,e,n){t.x=e,t.y=n},t.copy=function(t,e){t.x=e.x,t.y=e.y},t.len=function(t){return Math.sqrt(t.x*t.x+t.y*t.y)},t.lenSquare=function(t){return t.x*t.x+t.y*t.y},t.dot=function(t,e){return t.x*e.x+t.y*e.y},t.add=function(t,e,n){t.x=e.x+n.x,t.y=e.y+n.y},t.sub=function(t,e,n){t.x=e.x-n.x,t.y=e.y-n.y},t.scale=function(t,e,n){t.x=e.x*n,t.y=e.y*n},t.scaleAndAdd=function(t,e,n,r){t.x=e.x+n.x*r,t.y=e.y+n.y*r},t.lerp=function(t,e,n,r){var i=1-r;t.x=i*e.x+r*n.x,t.y=i*e.y+r*n.y},t}(),yt=gt,vt=Math.min,_t=Math.max,bt=new yt,xt=new yt,wt=new yt,Mt=new yt,St=new yt,kt=new yt,Lt=function(){function t(t,e,n,r){n<0&&(t+=n,n=-n),r<0&&(e+=r,r=-r),this.x=t,this.y=e,this.width=n,this.height=r}return t.prototype.union=function(t){var e=vt(t.x,this.x),n=vt(t.y,this.y);isFinite(this.x)&&isFinite(this.width)?this.width=_t(t.x+t.width,this.x+this.width)-e:this.width=t.width,isFinite(this.y)&&isFinite(this.height)?this.height=_t(t.y+t.height,this.y+this.height)-n:this.height=t.height,this.x=e,this.y=n},t.prototype.applyTransform=function(e){t.applyTransform(this,this,e)},t.prototype.calculateTransform=function(t){var e=this,n=t.width/e.width,r=t.height/e.height,i=st();return dt(i,i,[-e.x,-e.y]),ft(i,i,[n,r]),dt(i,i,[t.x,t.y]),i},t.prototype.intersect=function(e,n){if(!e)return!1;e instanceof t||(e=t.create(e));var r=this,i=r.x,o=r.x+r.width,a=r.y,s=r.y+r.height,l=e.x,u=e.x+e.width,c=e.y,d=e.y+e.height,h=!(o<l||u<i||s<c||d<a);if(n){var f=1/0,p=0,m=Math.abs(o-l),g=Math.abs(u-i),y=Math.abs(s-c),v=Math.abs(d-a),_=Math.min(m,g),b=Math.min(y,v);o<l||u<i?_>p&&(p=_,m<g?yt.set(kt,-m,0):yt.set(kt,g,0)):_<f&&(f=_,m<g?yt.set(St,m,0):yt.set(St,-g,0)),s<c||d<a?b>p&&(p=b,y<v?yt.set(kt,0,-y):yt.set(kt,0,v)):_<f&&(f=_,y<v?yt.set(St,0,y):yt.set(St,0,-v))}return n&&yt.copy(n,h?St:kt),h},t.prototype.contain=function(t,e){var n=this;return t>=n.x&&t<=n.x+n.width&&e>=n.y&&e<=n.y+n.height},t.prototype.clone=function(){return new t(this.x,this.y,this.width,this.height)},t.prototype.copy=function(e){t.copy(this,e)},t.prototype.plain=function(){return{x:this.x,y:this.y,width:this.width,height:this.height}},t.prototype.isFinite=function(){return isFinite(this.x)&&isFinite(this.y)&&isFinite(this.width)&&isFinite(this.height)},t.prototype.isZero=function(){return 0===this.width||0===this.height},t.create=function(e){return new t(e.x,e.y,e.width,e.height)},t.copy=function(t,e){t.x=e.x,t.y=e.y,t.width=e.width,t.height=e.height},t.applyTransform=function(e,n,r){if(r){if(r[1]<1e-5&&r[1]>-1e-5&&r[2]<1e-5&&r[2]>-1e-5){var i=r[0],o=r[3],a=r[4],s=r[5];return e.x=n.x*i+a,e.y=n.y*o+s,e.width=n.width*i,e.height=n.height*o,e.width<0&&(e.x+=e.width,e.width=-e.width),void(e.height<0&&(e.y+=e.height,e.height=-e.height))}bt.x=wt.x=n.x,bt.y=Mt.y=n.y,xt.x=Mt.x=n.x+n.width,xt.y=wt.y=n.y+n.height,bt.transform(r),Mt.transform(r),xt.transform(r),wt.transform(r),e.x=vt(bt.x,xt.x,wt.x,Mt.x),e.y=vt(bt.y,xt.y,wt.y,Mt.y);var l=_t(bt.x,xt.x,wt.x,Mt.x),u=_t(bt.y,xt.y,wt.y,Mt.y);e.width=l-e.x,e.height=u-e.y}else e!==n&&t.copy(e,n)},t}(),Tt=Lt,Dt="silent";function Ot(t,e,n){return{type:t,event:n,target:e.target,topTarget:e.topTarget,cancelBubble:!1,offsetX:n.zrX,offsetY:n.zrY,gestureEvent:n.gestureEvent,pinchX:n.pinchX,pinchY:n.pinchY,pinchScale:n.pinchScale,wheelDelta:n.zrDelta,zrByTouch:n.zrByTouch,which:n.which,stop:It}}function It(){et(this.event)}var Ct=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.handler=null,e}return o(e,t),e.prototype.dispose=function(){},e.prototype.setCursor=function(){},e}(C),At=function(){function t(t,e){this.x=t,this.y=e}return t}(),Yt=["click","dblclick","mousewheel","mouseout","mouseup","mousedown","mousemove","contextmenu"],jt=new Tt(0,0,0,0),Pt=function(t){function e(e,n,r,i,o){var a=t.call(this)||this;return a._hovered=new At(0,0),a.storage=e,a.painter=n,a.painterRoot=i,a._pointerSize=o,r=r||new Ct,a.proxy=null,a.setHandlerProxy(r),a._draggingMgr=new O(a),a}return o(e,t),e.prototype.setHandlerProxy=function(t){this.proxy&&this.proxy.dispose(),t&&(s["k"](Yt,(function(e){t.on&&t.on(e,this[e],this)}),this),t.handler=this),this.proxy=t},e.prototype.mousemove=function(t){var e=t.zrX,n=t.zrY,r=Rt(this,e,n),i=this._hovered,o=i.target;o&&!o.__zr&&(i=this.findHover(i.x,i.y),o=i.target);var a=this._hovered=r?new At(e,n):this.findHover(e,n),s=a.target,l=this.proxy;l.setCursor&&l.setCursor(s?s.cursor:"default"),o&&s!==o&&this.dispatchToElement(i,"mouseout",t),this.dispatchToElement(a,"mousemove",t),s&&s!==o&&this.dispatchToElement(a,"mouseover",t)},e.prototype.mouseout=function(t){var e=t.zrEventControl;"only_globalout"!==e&&this.dispatchToElement(this._hovered,"mouseout",t),"no_globalout"!==e&&this.trigger("globalout",{type:"globalout",event:t})},e.prototype.resize=function(){this._hovered=new At(0,0)},e.prototype.dispatch=function(t,e){var n=this[t];n&&n.call(this,e)},e.prototype.dispose=function(){this.proxy.dispose(),this.storage=null,this.proxy=null,this.painter=null},e.prototype.setCursorStyle=function(t){var e=this.proxy;e.setCursor&&e.setCursor(t)},e.prototype.dispatchToElement=function(t,e,n){t=t||{};var r=t.target;if(!r||!r.silent){var i="on"+e,o=Ot(e,t,n);while(r)if(r[i]&&(o.cancelBubble=!!r[i].call(r,o)),r.trigger(e,o),r=r.__hostTarget?r.__hostTarget:r.parent,o.cancelBubble)break;o.cancelBubble||(this.trigger(e,o),this.painter&&this.painter.eachOtherLayer&&this.painter.eachOtherLayer((function(t){"function"===typeof t[i]&&t[i].call(t,o),t.trigger&&t.trigger(e,o)})))}},e.prototype.findHover=function(t,e,n){var r=this.storage.getDisplayList(),i=new At(t,e);if(Nt(r,i,t,e,n),this._pointerSize&&!i.target){for(var o=[],a=this._pointerSize,s=a/2,l=new Tt(t-s,e-s,a,a),u=r.length-1;u>=0;u--){var c=r[u];c===n||c.ignore||c.ignoreCoarsePointer||c.parent&&c.parent.ignoreCoarsePointer||(jt.copy(c.getBoundingRect()),c.transform&&jt.applyTransform(c.transform),jt.intersect(l)&&o.push(c))}if(o.length)for(var d=4,h=Math.PI/12,f=2*Math.PI,p=0;p<s;p+=d)for(var m=0;m<f;m+=h){var g=t+p*Math.cos(m),y=e+p*Math.sin(m);if(Nt(o,i,g,y,n),i.target)return i}}return i},e.prototype.processGesture=function(t,e){this._gestureMgr||(this._gestureMgr=new rt);var n=this._gestureMgr;"start"===e&&n.clear();var r=n.recognize(t,this.findHover(t.zrX,t.zrY,null).target,this.proxy.dom);if("end"===e&&n.clear(),r){var i=r.type;t.gestureEvent=i;var o=new At;o.target=r.target,this.dispatchToElement(o,i,r.event)}},e}(C);function Et(t,e,n){if(t[t.rectHover?"rectContain":"contain"](e,n)){var r=t,i=void 0,o=!1;while(r){if(r.ignoreClip&&(o=!0),!o){var a=r.getClipPath();if(a&&!a.contain(e,n))return!1}r.silent&&(i=!0);var s=r.__hostTarget;r=s||r.parent}return!i||Dt}return!1}function Nt(t,e,n,r,i){for(var o=t.length-1;o>=0;o--){var a=t[o],s=void 0;if(a!==i&&!a.ignore&&(s=Et(a,n,r))&&(!e.topTarget&&(e.topTarget=a),s!==Dt)){e.target=a;break}}}function Rt(t,e,n){var r=t.painter;return e<0||e>r.getWidth()||n<0||n>r.getHeight()}s["k"](["click","mousedown","mouseup","mousewheel","dblclick","contextmenu"],(function(t){Pt.prototype[t]=function(e){var n,r,i=e.zrX,o=e.zrY,a=Rt(this,i,o);if("mouseup"===t&&a||(n=this.findHover(i,o),r=n.target),"mousedown"===t)this._downEl=r,this._downPoint=[e.zrX,e.zrY],this._upEl=r;else if("mouseup"===t)this._upEl=r;else if("click"===t){if(this._downEl!==this._upEl||!this._downPoint||b(this._downPoint,[e.zrX,e.zrY])>4)return;this._downPoint=null}this.dispatchToElement(n,t,e)}}));var Ht=Pt,zt=32,Ft=7;function Bt(t){var e=0;while(t>=zt)e|=1&t,t>>=1;return t+e}function Wt(t,e,n,r){var i=e+1;if(i===n)return 1;if(r(t[i++],t[e])<0){while(i<n&&r(t[i],t[i-1])<0)i++;Vt(t,e,i)}else while(i<n&&r(t[i],t[i-1])>=0)i++;return i-e}function Vt(t,e,n){n--;while(e<n){var r=t[e];t[e++]=t[n],t[n--]=r}}function Gt(t,e,n,r,i){for(r===e&&r++;r<n;r++){var o,a=t[r],s=e,l=r;while(s<l)o=s+l>>>1,i(a,t[o])<0?l=o:s=o+1;var u=r-s;switch(u){case 3:t[s+3]=t[s+2];case 2:t[s+2]=t[s+1];case 1:t[s+1]=t[s];break;default:while(u>0)t[s+u]=t[s+u-1],u--}t[s]=a}}function $t(t,e,n,r,i,o){var a=0,s=0,l=1;if(o(t,e[n+i])>0){s=r-i;while(l<s&&o(t,e[n+i+l])>0)a=l,l=1+(l<<1),l<=0&&(l=s);l>s&&(l=s),a+=i,l+=i}else{s=i+1;while(l<s&&o(t,e[n+i-l])<=0)a=l,l=1+(l<<1),l<=0&&(l=s);l>s&&(l=s);var u=a;a=i-l,l=i-u}a++;while(a<l){var c=a+(l-a>>>1);o(t,e[n+c])>0?a=c+1:l=c}return l}function Ut(t,e,n,r,i,o){var a=0,s=0,l=1;if(o(t,e[n+i])<0){s=i+1;while(l<s&&o(t,e[n+i-l])<0)a=l,l=1+(l<<1),l<=0&&(l=s);l>s&&(l=s);var u=a;a=i-l,l=i-u}else{s=r-i;while(l<s&&o(t,e[n+i+l])>=0)a=l,l=1+(l<<1),l<=0&&(l=s);l>s&&(l=s),a+=i,l+=i}a++;while(a<l){var c=a+(l-a>>>1);o(t,e[n+c])<0?l=c:a=c+1}return l}function qt(t,e){var n,r,i=Ft,o=0,a=[];function s(t,e){n[o]=t,r[o]=e,o+=1}function l(){while(o>1){var t=o-2;if(t>=1&&r[t-1]<=r[t]+r[t+1]||t>=2&&r[t-2]<=r[t]+r[t-1])r[t-1]<r[t+1]&&t--;else if(r[t]>r[t+1])break;c(t)}}function u(){while(o>1){var t=o-2;t>0&&r[t-1]<r[t+1]&&t--,c(t)}}function c(i){var a=n[i],s=r[i],l=n[i+1],u=r[i+1];r[i]=s+u,i===o-3&&(n[i+1]=n[i+2],r[i+1]=r[i+2]),o--;var c=Ut(t[l],t,a,s,0,e);a+=c,s-=c,0!==s&&(u=$t(t[a+s-1],t,l,u,u-1,e),0!==u&&(s<=u?d(a,s,l,u):h(a,s,l,u)))}function d(n,r,o,s){var l=0;for(l=0;l<r;l++)a[l]=t[n+l];var u=0,c=o,d=n;if(t[d++]=t[c++],0!==--s)if(1!==r){var h,f,p,m=i;while(1){h=0,f=0,p=!1;do{if(e(t[c],a[u])<0){if(t[d++]=t[c++],f++,h=0,0===--s){p=!0;break}}else if(t[d++]=a[u++],h++,f=0,1===--r){p=!0;break}}while((h|f)<m);if(p)break;do{if(h=Ut(t[c],a,u,r,0,e),0!==h){for(l=0;l<h;l++)t[d+l]=a[u+l];if(d+=h,u+=h,r-=h,r<=1){p=!0;break}}if(t[d++]=t[c++],0===--s){p=!0;break}if(f=$t(a[u],t,c,s,0,e),0!==f){for(l=0;l<f;l++)t[d+l]=t[c+l];if(d+=f,c+=f,s-=f,0===s){p=!0;break}}if(t[d++]=a[u++],1===--r){p=!0;break}m--}while(h>=Ft||f>=Ft);if(p)break;m<0&&(m=0),m+=2}if(i=m,i<1&&(i=1),1===r){for(l=0;l<s;l++)t[d+l]=t[c+l];t[d+s]=a[u]}else{if(0===r)throw new Error;for(l=0;l<r;l++)t[d+l]=a[u+l]}}else{for(l=0;l<s;l++)t[d+l]=t[c+l];t[d+s]=a[u]}else for(l=0;l<r;l++)t[d+l]=a[u+l]}function h(n,r,o,s){var l=0;for(l=0;l<s;l++)a[l]=t[o+l];var u=n+r-1,c=s-1,d=o+s-1,h=0,f=0;if(t[d--]=t[u--],0!==--r)if(1!==s){var p=i;while(1){var m=0,g=0,y=!1;do{if(e(a[c],t[u])<0){if(t[d--]=t[u--],m++,g=0,0===--r){y=!0;break}}else if(t[d--]=a[c--],g++,m=0,1===--s){y=!0;break}}while((m|g)<p);if(y)break;do{if(m=r-Ut(a[c],t,n,r,r-1,e),0!==m){for(d-=m,u-=m,r-=m,f=d+1,h=u+1,l=m-1;l>=0;l--)t[f+l]=t[h+l];if(0===r){y=!0;break}}if(t[d--]=a[c--],1===--s){y=!0;break}if(g=s-$t(t[u],a,0,s,s-1,e),0!==g){for(d-=g,c-=g,s-=g,f=d+1,h=c+1,l=0;l<g;l++)t[f+l]=a[h+l];if(s<=1){y=!0;break}}if(t[d--]=t[u--],0===--r){y=!0;break}p--}while(m>=Ft||g>=Ft);if(y)break;p<0&&(p=0),p+=2}if(i=p,i<1&&(i=1),1===s){for(d-=r,u-=r,f=d+1,h=u+1,l=r-1;l>=0;l--)t[f+l]=t[h+l];t[d]=a[c]}else{if(0===s)throw new Error;for(h=d-(s-1),l=0;l<s;l++)t[h+l]=a[l]}}else{for(d-=r,u-=r,f=d+1,h=u+1,l=r-1;l>=0;l--)t[f+l]=t[h+l];t[d]=a[c]}else for(h=d-(s-1),l=0;l<s;l++)t[h+l]=a[l]}return n=[],r=[],{mergeRuns:l,forceMergeRuns:u,pushRun:s}}function Xt(t,e,n,r){n||(n=0),r||(r=t.length);var i=r-n;if(!(i<2)){var o=0;if(i<zt)return o=Wt(t,n,r,e),void Gt(t,n,r,n+o,e);var a=qt(t,e),s=Bt(i);do{if(o=Wt(t,n,r,e),o<s){var l=i;l>s&&(l=s),Gt(t,n,n+l,n+o,e),o=l}a.pushRun(n,o),a.mergeRuns(),i-=o,n+=o}while(0!==i);a.forceMergeRuns()}}var Zt=1,Jt=2,Kt=4,Qt=!1;function te(){Qt||(Qt=!0)}function ee(t,e){return t.zlevel===e.zlevel?t.z===e.z?t.z2-e.z2:t.z-e.z:t.zlevel-e.zlevel}var ne,re=function(){function t(){this._roots=[],this._displayList=[],this._displayListLen=0,this.displayableSortFunc=ee}return t.prototype.traverse=function(t,e){for(var n=0;n<this._roots.length;n++)this._roots[n].traverse(t,e)},t.prototype.getDisplayList=function(t,e){e=e||!1;var n=this._displayList;return!t&&n.length||this.updateDisplayList(e),n},t.prototype.updateDisplayList=function(t){this._displayListLen=0;for(var e=this._roots,n=this._displayList,r=0,i=e.length;r<i;r++)this._updateAndAddDisplayable(e[r],null,t);n.length=this._displayListLen,Xt(n,ee)},t.prototype._updateAndAddDisplayable=function(t,e,n){if(!t.ignore||n){t.beforeUpdate(),t.update(),t.afterUpdate();var r=t.getClipPath();if(t.ignoreClip)e=null;else if(r){e=e?e.slice():[];var i=r,o=t;while(i)i.parent=o,i.updateTransform(),e.push(i),o=i,i=i.getClipPath()}if(t.childrenRef){for(var a=t.childrenRef(),s=0;s<a.length;s++){var l=a[s];t.__dirty&&(l.__dirty|=Zt),this._updateAndAddDisplayable(l,e,n)}t.__dirty=0}else{var u=t;e&&e.length?u.__clipPaths=e:u.__clipPaths&&u.__clipPaths.length>0&&(u.__clipPaths=[]),isNaN(u.z)&&(te(),u.z=0),isNaN(u.z2)&&(te(),u.z2=0),isNaN(u.zlevel)&&(te(),u.zlevel=0),this._displayList[this._displayListLen++]=u}var c=t.getDecalElement&&t.getDecalElement();c&&this._updateAndAddDisplayable(c,e,n);var d=t.getTextGuideLine();d&&this._updateAndAddDisplayable(d,e,n);var h=t.getTextContent();h&&this._updateAndAddDisplayable(h,e,n)}},t.prototype.addRoot=function(t){t.__zr&&t.__zr.storage===this||this._roots.push(t)},t.prototype.delRoot=function(t){if(t instanceof Array)for(var e=0,n=t.length;e<n;e++)this.delRoot(t[e]);else{var r=s["r"](this._roots,t);r>=0&&this._roots.splice(r,1)}},t.prototype.delAllRoots=function(){this._roots=[],this._displayList=[],this._displayListLen=0},t.prototype.getRoots=function(){return this._roots},t.prototype.dispose=function(){this._displayList=null,this._roots=null},t}(),ie=re;ne=a["a"].hasGlobalWindow&&(window.requestAnimationFrame&&window.requestAnimationFrame.bind(window)||window.msRequestAnimationFrame&&window.msRequestAnimationFrame.bind(window)||window.mozRequestAnimationFrame||window.webkitRequestAnimationFrame)||function(t){return setTimeout(t,16)};var oe=ne,ae={linear:function(t){return t},quadraticIn:function(t){return t*t},quadraticOut:function(t){return t*(2-t)},quadraticInOut:function(t){return(t*=2)<1?.5*t*t:-.5*(--t*(t-2)-1)},cubicIn:function(t){return t*t*t},cubicOut:function(t){return--t*t*t+1},cubicInOut:function(t){return(t*=2)<1?.5*t*t*t:.5*((t-=2)*t*t+2)},quarticIn:function(t){return t*t*t*t},quarticOut:function(t){return 1- --t*t*t*t},quarticInOut:function(t){return(t*=2)<1?.5*t*t*t*t:-.5*((t-=2)*t*t*t-2)},quinticIn:function(t){return t*t*t*t*t},quinticOut:function(t){return--t*t*t*t*t+1},quinticInOut:function(t){return(t*=2)<1?.5*t*t*t*t*t:.5*((t-=2)*t*t*t*t+2)},sinusoidalIn:function(t){return 1-Math.cos(t*Math.PI/2)},sinusoidalOut:function(t){return Math.sin(t*Math.PI/2)},sinusoidalInOut:function(t){return.5*(1-Math.cos(Math.PI*t))},exponentialIn:function(t){return 0===t?0:Math.pow(1024,t-1)},exponentialOut:function(t){return 1===t?1:1-Math.pow(2,-10*t)},exponentialInOut:function(t){return 0===t?0:1===t?1:(t*=2)<1?.5*Math.pow(1024,t-1):.5*(2-Math.pow(2,-10*(t-1)))},circularIn:function(t){return 1-Math.sqrt(1-t*t)},circularOut:function(t){return Math.sqrt(1- --t*t)},circularInOut:function(t){return(t*=2)<1?-.5*(Math.sqrt(1-t*t)-1):.5*(Math.sqrt(1-(t-=2)*t)+1)},elasticIn:function(t){var e,n=.1,r=.4;return 0===t?0:1===t?1:(!n||n<1?(n=1,e=r/4):e=r*Math.asin(1/n)/(2*Math.PI),-n*Math.pow(2,10*(t-=1))*Math.sin((t-e)*(2*Math.PI)/r))},elasticOut:function(t){var e,n=.1,r=.4;return 0===t?0:1===t?1:(!n||n<1?(n=1,e=r/4):e=r*Math.asin(1/n)/(2*Math.PI),n*Math.pow(2,-10*t)*Math.sin((t-e)*(2*Math.PI)/r)+1)},elasticInOut:function(t){var e,n=.1,r=.4;return 0===t?0:1===t?1:(!n||n<1?(n=1,e=r/4):e=r*Math.asin(1/n)/(2*Math.PI),(t*=2)<1?n*Math.pow(2,10*(t-=1))*Math.sin((t-e)*(2*Math.PI)/r)*-.5:n*Math.pow(2,-10*(t-=1))*Math.sin((t-e)*(2*Math.PI)/r)*.5+1)},backIn:function(t){var e=1.70158;return t*t*((e+1)*t-e)},backOut:function(t){var e=1.70158;return--t*t*((e+1)*t+e)+1},backInOut:function(t){var e=2.5949095;return(t*=2)<1?t*t*((e+1)*t-e)*.5:.5*((t-=2)*t*((e+1)*t+e)+2)},bounceIn:function(t){return 1-ae.bounceOut(1-t)},bounceOut:function(t){return t<1/2.75?7.5625*t*t:t<2/2.75?7.5625*(t-=1.5/2.75)*t+.75:t<2.5/2.75?7.5625*(t-=2.25/2.75)*t+.9375:7.5625*(t-=2.625/2.75)*t+.984375},bounceInOut:function(t){return t<.5?.5*ae.bounceIn(2*t):.5*ae.bounceOut(2*t-1)+.5}},se=ae,le=Math.pow,ue=Math.sqrt,ce=1e-8,de=1e-4,he=ue(3),fe=1/3,pe=l(),me=l(),ge=l();function ye(t){return t>-ce&&t<ce}function ve(t){return t>ce||t<-ce}function _e(t,e,n,r,i){var o=1-i;return o*o*(o*t+3*i*e)+i*i*(i*r+3*o*n)}function be(t,e,n,r,i){var o=1-i;return 3*(((e-t)*o+2*(n-e)*i)*o+(r-n)*i*i)}function xe(t,e,n,r,i,o){var a=r+3*(e-n)-t,s=3*(n-2*e+t),l=3*(e-t),u=t-i,c=s*s-3*a*l,d=s*l-9*a*u,h=l*l-3*s*u,f=0;if(ye(c)&&ye(d))if(ye(s))o[0]=0;else{var p=-l/s;p>=0&&p<=1&&(o[f++]=p)}else{var m=d*d-4*c*h;if(ye(m)){var g=d/c,y=(p=-s/a+g,-g/2);p>=0&&p<=1&&(o[f++]=p),y>=0&&y<=1&&(o[f++]=y)}else if(m>0){var v=ue(m),_=c*s+1.5*a*(-d+v),b=c*s+1.5*a*(-d-v);_=_<0?-le(-_,fe):le(_,fe),b=b<0?-le(-b,fe):le(b,fe);p=(-s-(_+b))/(3*a);p>=0&&p<=1&&(o[f++]=p)}else{var x=(2*c*s-3*a*d)/(2*ue(c*c*c)),w=Math.acos(x)/3,M=ue(c),S=Math.cos(w),k=(p=(-s-2*M*S)/(3*a),y=(-s+M*(S+he*Math.sin(w)))/(3*a),(-s+M*(S-he*Math.sin(w)))/(3*a));p>=0&&p<=1&&(o[f++]=p),y>=0&&y<=1&&(o[f++]=y),k>=0&&k<=1&&(o[f++]=k)}}return f}function we(t,e,n,r,i){var o=6*n-12*e+6*t,a=9*e+3*r-3*t-9*n,s=3*e-3*t,l=0;if(ye(a)){if(ve(o)){var u=-s/o;u>=0&&u<=1&&(i[l++]=u)}}else{var c=o*o-4*a*s;if(ye(c))i[0]=-o/(2*a);else if(c>0){var d=ue(c),h=(u=(-o+d)/(2*a),(-o-d)/(2*a));u>=0&&u<=1&&(i[l++]=u),h>=0&&h<=1&&(i[l++]=h)}}return l}function Me(t,e,n,r,i,o){var a=(e-t)*i+t,s=(n-e)*i+e,l=(r-n)*i+n,u=(s-a)*i+a,c=(l-s)*i+s,d=(c-u)*i+u;o[0]=t,o[1]=a,o[2]=u,o[3]=d,o[4]=d,o[5]=c,o[6]=l,o[7]=r}function Se(t,e,n,r,i,o,a,s,l,u,c){var d,h,f,p,m,g=.005,y=1/0;pe[0]=l,pe[1]=u;for(var v=0;v<1;v+=.05)me[0]=_e(t,n,i,a,v),me[1]=_e(e,r,o,s,v),p=w(pe,me),p<y&&(d=v,y=p);y=1/0;for(var _=0;_<32;_++){if(g<de)break;h=d-g,f=d+g,me[0]=_e(t,n,i,a,h),me[1]=_e(e,r,o,s,h),p=w(me,pe),h>=0&&p<y?(d=h,y=p):(ge[0]=_e(t,n,i,a,f),ge[1]=_e(e,r,o,s,f),m=w(ge,pe),f<=1&&m<y?(d=f,y=m):g*=.5)}return c&&(c[0]=_e(t,n,i,a,d),c[1]=_e(e,r,o,s,d)),ue(y)}function ke(t,e,n,r,i,o,a,s,l){for(var u=t,c=e,d=0,h=1/l,f=1;f<=l;f++){var p=f*h,m=_e(t,n,i,a,p),g=_e(e,r,o,s,p),y=m-u,v=g-c;d+=Math.sqrt(y*y+v*v),u=m,c=g}return d}function Le(t,e,n,r){var i=1-r;return i*(i*t+2*r*e)+r*r*n}function Te(t,e,n,r){return 2*((1-r)*(e-t)+r*(n-e))}function De(t,e,n,r,i){var o=t-2*e+n,a=2*(e-t),s=t-r,l=0;if(ye(o)){if(ve(a)){var u=-s/a;u>=0&&u<=1&&(i[l++]=u)}}else{var c=a*a-4*o*s;if(ye(c)){u=-a/(2*o);u>=0&&u<=1&&(i[l++]=u)}else if(c>0){var d=ue(c),h=(u=(-a+d)/(2*o),(-a-d)/(2*o));u>=0&&u<=1&&(i[l++]=u),h>=0&&h<=1&&(i[l++]=h)}}return l}function Oe(t,e,n){var r=t+n-2*e;return 0===r?.5:(t-e)/r}function Ie(t,e,n,r,i){var o=(e-t)*r+t,a=(n-e)*r+e,s=(a-o)*r+o;i[0]=t,i[1]=o,i[2]=s,i[3]=s,i[4]=a,i[5]=n}function Ce(t,e,n,r,i,o,a,s,l){var u,c=.005,d=1/0;pe[0]=a,pe[1]=s;for(var h=0;h<1;h+=.05){me[0]=Le(t,n,i,h),me[1]=Le(e,r,o,h);var f=w(pe,me);f<d&&(u=h,d=f)}d=1/0;for(var p=0;p<32;p++){if(c<de)break;var m=u-c,g=u+c;me[0]=Le(t,n,i,m),me[1]=Le(e,r,o,m);f=w(me,pe);if(m>=0&&f<d)u=m,d=f;else{ge[0]=Le(t,n,i,g),ge[1]=Le(e,r,o,g);var y=w(ge,pe);g<=1&&y<d?(u=g,d=y):c*=.5}}return l&&(l[0]=Le(t,n,i,u),l[1]=Le(e,r,o,u)),ue(d)}function Ae(t,e,n,r,i,o,a){for(var s=t,l=e,u=0,c=1/a,d=1;d<=a;d++){var h=d*c,f=Le(t,n,i,h),p=Le(e,r,o,h),m=f-s,g=p-l;u+=Math.sqrt(m*m+g*g),s=f,l=p}return u}var Ye=/cubic-bezier\(([0-9,\.e ]+)\)/;function je(t){var e=t&&Ye.exec(t);if(e){var n=e[1].split(","),r=+Object(s["T"])(n[0]),i=+Object(s["T"])(n[1]),o=+Object(s["T"])(n[2]),a=+Object(s["T"])(n[3]);if(isNaN(r+i+o+a))return;var l=[];return function(t){return t<=0?0:t>=1?1:xe(0,r,o,1,t,l)&&_e(0,i,a,1,l[0])}}}var Pe=function(){function t(t){this._inited=!1,this._startTime=0,this._pausedTime=0,this._paused=!1,this._life=t.life||1e3,this._delay=t.delay||0,this.loop=t.loop||!1,this.onframe=t.onframe||s["L"],this.ondestroy=t.ondestroy||s["L"],this.onrestart=t.onrestart||s["L"],t.easing&&this.setEasing(t.easing)}return t.prototype.step=function(t,e){if(this._inited||(this._startTime=t+this._delay,this._inited=!0),!this._paused){var n=this._life,r=t-this._startTime-this._pausedTime,i=r/n;i<0&&(i=0),i=Math.min(i,1);var o=this.easingFunc,a=o?o(i):i;if(this.onframe(a),1===i){if(!this.loop)return!0;var s=r%n;this._startTime=t-s,this._pausedTime=0,this.onrestart()}return!1}this._pausedTime+=e},t.prototype.pause=function(){this._paused=!0},t.prototype.resume=function(){this._paused=!1},t.prototype.setEasing=function(t){this.easing=t,this.easingFunc=Object(s["w"])(t)?t:se[t]||je(t)},t}(),Ee=Pe,Ne=n("41ef"),Re=n("7a29"),He=Array.prototype.slice;function ze(t,e,n){return(e-t)*n+t}function Fe(t,e,n,r){for(var i=e.length,o=0;o<i;o++)t[o]=ze(e[o],n[o],r);return t}function Be(t,e,n,r){for(var i=e.length,o=i&&e[0].length,a=0;a<i;a++){t[a]||(t[a]=[]);for(var s=0;s<o;s++)t[a][s]=ze(e[a][s],n[a][s],r)}return t}function We(t,e,n,r){for(var i=e.length,o=0;o<i;o++)t[o]=e[o]+n[o]*r;return t}function Ve(t,e,n,r){for(var i=e.length,o=i&&e[0].length,a=0;a<i;a++){t[a]||(t[a]=[]);for(var s=0;s<o;s++)t[a][s]=e[a][s]+n[a][s]*r}return t}function Ge(t,e){for(var n=t.length,r=e.length,i=n>r?e:t,o=Math.min(n,r),a=i[o-1]||{color:[0,0,0,0],offset:0},s=o;s<Math.max(n,r);s++)i.push({offset:a.offset,color:a.color.slice()})}function $e(t,e,n){var r=t,i=e;if(r.push&&i.push){var o=r.length,a=i.length;if(o!==a){var s=o>a;if(s)r.length=a;else for(var l=o;l<a;l++)r.push(1===n?i[l]:He.call(i[l]))}var u=r[0]&&r[0].length;for(l=0;l<r.length;l++)if(1===n)isNaN(r[l])&&(r[l]=i[l]);else for(var c=0;c<u;c++)isNaN(r[l][c])&&(r[l][c]=i[l][c])}}function Ue(t){if(Object(s["u"])(t)){var e=t.length;if(Object(s["u"])(t[0])){for(var n=[],r=0;r<e;r++)n.push(He.call(t[r]));return n}return He.call(t)}return t}function qe(t){return t[0]=Math.floor(t[0])||0,t[1]=Math.floor(t[1])||0,t[2]=Math.floor(t[2])||0,t[3]=null==t[3]?1:t[3],"rgba("+t.join(",")+")"}function Xe(t){return Object(s["u"])(t&&t[0])?2:1}var Ze=0,Je=1,Ke=2,Qe=3,tn=4,en=5,nn=6;function rn(t){return t===tn||t===en}function on(t){return t===Je||t===Ke}var an=[0,0,0,0],sn=function(){function t(t){this.keyframes=[],this.discrete=!1,this._invalid=!1,this._needsSort=!1,this._lastFr=0,this._lastFrP=0,this.propName=t}return t.prototype.isFinished=function(){return this._finished},t.prototype.setFinished=function(){this._finished=!0,this._additiveTrack&&this._additiveTrack.setFinished()},t.prototype.needsAnimate=function(){return this.keyframes.length>=1},t.prototype.getAdditiveTrack=function(){return this._additiveTrack},t.prototype.addKeyframe=function(t,e,n){this._needsSort=!0;var r=this.keyframes,i=r.length,o=!1,a=nn,l=e;if(Object(s["u"])(e)){var u=Xe(e);a=u,(1===u&&!Object(s["z"])(e[0])||2===u&&!Object(s["z"])(e[0][0]))&&(o=!0)}else if(Object(s["z"])(e)&&!Object(s["l"])(e))a=Ze;else if(Object(s["C"])(e))if(isNaN(+e)){var c=Ne["h"](e);c&&(l=c,a=Qe)}else a=Ze;else if(Object(s["x"])(e)){var d=Object(s["m"])({},l);d.colorStops=Object(s["H"])(e.colorStops,(function(t){return{offset:t.offset,color:Ne["h"](t.color)}})),Object(Re["m"])(e)?a=tn:Object(Re["o"])(e)&&(a=en),l=d}0===i?this.valType=a:a===this.valType&&a!==nn||(o=!0),this.discrete=this.discrete||o;var h={time:t,value:l,rawValue:e,percent:0};return n&&(h.easing=n,h.easingFunc=Object(s["w"])(n)?n:se[n]||je(n)),r.push(h),h},t.prototype.prepare=function(t,e){var n=this.keyframes;this._needsSort&&n.sort((function(t,e){return t.time-e.time}));for(var r=this.valType,i=n.length,o=n[i-1],a=this.discrete,s=on(r),l=rn(r),u=0;u<i;u++){var c=n[u],d=c.value,h=o.value;c.percent=c.time/t,a||(s&&u!==i-1?$e(d,h,r):l&&Ge(d.colorStops,h.colorStops))}if(!a&&r!==en&&e&&this.needsAnimate()&&e.needsAnimate()&&r===e.valType&&!e._finished){this._additiveTrack=e;var f=n[0].value;for(u=0;u<i;u++)r===Ze?n[u].additiveValue=n[u].value-f:r===Qe?n[u].additiveValue=We([],n[u].value,f,-1):on(r)&&(n[u].additiveValue=r===Je?We([],n[u].value,f,-1):Ve([],n[u].value,f,-1))}},t.prototype.step=function(t,e){if(!this._finished){this._additiveTrack&&this._additiveTrack._finished&&(this._additiveTrack=null);var n,r,i,o=null!=this._additiveTrack,a=o?"additiveValue":"value",l=this.valType,u=this.keyframes,c=u.length,d=this.propName,h=l===Qe,f=this._lastFr,p=Math.min;if(1===c)r=i=u[0];else{if(e<0)n=0;else if(e<this._lastFrP){var m=p(f+1,c-1);for(n=m;n>=0;n--)if(u[n].percent<=e)break;n=p(n,c-2)}else{for(n=f;n<c;n++)if(u[n].percent>e)break;n=p(n-1,c-2)}i=u[n+1],r=u[n]}if(r&&i){this._lastFr=n,this._lastFrP=e;var g=i.percent-r.percent,y=0===g?1:p((e-r.percent)/g,1);i.easingFunc&&(y=i.easingFunc(y));var v=o?this._additiveValue:h?an:t[d];if(!on(l)&&!h||v||(v=this._additiveValue=[]),this.discrete)t[d]=y<1?r.rawValue:i.rawValue;else if(on(l))l===Je?Fe(v,r[a],i[a],y):Be(v,r[a],i[a],y);else if(rn(l)){var _=r[a],b=i[a],x=l===tn;t[d]={type:x?"linear":"radial",x:ze(_.x,b.x,y),y:ze(_.y,b.y,y),colorStops:Object(s["H"])(_.colorStops,(function(t,e){var n=b.colorStops[e];return{offset:ze(t.offset,n.offset,y),color:qe(Fe([],t.color,n.color,y))}})),global:b.global},x?(t[d].x2=ze(_.x2,b.x2,y),t[d].y2=ze(_.y2,b.y2,y)):t[d].r=ze(_.r,b.r,y)}else if(h)Fe(v,r[a],i[a],y),o||(t[d]=qe(v));else{var w=ze(r[a],i[a],y);o?this._additiveValue=w:t[d]=w}o&&this._addToTarget(t)}}},t.prototype._addToTarget=function(t){var e=this.valType,n=this.propName,r=this._additiveValue;e===Ze?t[n]=t[n]+r:e===Qe?(Ne["h"](t[n],an),We(an,an,r,1),t[n]=qe(an)):e===Je?We(t[n],t[n],r,1):e===Ke&&Ve(t[n],t[n],r,1)},t}(),ln=function(){function t(t,e,n,r){this._tracks={},this._trackKeys=[],this._maxTime=0,this._started=0,this._clip=null,this._target=t,this._loop=e,e&&r?Object(s["G"])("Can' use additive animation on looped animation."):(this._additiveAnimators=r,this._allowDiscrete=n)}return t.prototype.getMaxTime=function(){return this._maxTime},t.prototype.getDelay=function(){return this._delay},t.prototype.getLoop=function(){return this._loop},t.prototype.getTarget=function(){return this._target},t.prototype.changeTarget=function(t){this._target=t},t.prototype.when=function(t,e,n){return this.whenWithKeys(t,e,Object(s["F"])(e),n)},t.prototype.whenWithKeys=function(t,e,n,r){for(var i=this._tracks,o=0;o<n.length;o++){var a=n[o],s=i[a];if(!s){s=i[a]=new sn(a);var l=void 0,u=this._getAdditiveTrack(a);if(u){var c=u.keyframes,d=c[c.length-1];l=d&&d.value,u.valType===Qe&&l&&(l=qe(l))}else l=this._target[a];if(null==l)continue;t>0&&s.addKeyframe(0,Ue(l),r),this._trackKeys.push(a)}s.addKeyframe(t,Ue(e[a]),r)}return this._maxTime=Math.max(this._maxTime,t),this},t.prototype.pause=function(){this._clip.pause(),this._paused=!0},t.prototype.resume=function(){this._clip.resume(),this._paused=!1},t.prototype.isPaused=function(){return!!this._paused},t.prototype.duration=function(t){return this._maxTime=t,this._force=!0,this},t.prototype._doneCallback=function(){this._setTracksFinished(),this._clip=null;var t=this._doneCbs;if(t)for(var e=t.length,n=0;n<e;n++)t[n].call(this)},t.prototype._abortedCallback=function(){this._setTracksFinished();var t=this.animation,e=this._abortedCbs;if(t&&t.removeClip(this._clip),this._clip=null,e)for(var n=0;n<e.length;n++)e[n].call(this)},t.prototype._setTracksFinished=function(){for(var t=this._tracks,e=this._trackKeys,n=0;n<e.length;n++)t[e[n]].setFinished()},t.prototype._getAdditiveTrack=function(t){var e,n=this._additiveAnimators;if(n)for(var r=0;r<n.length;r++){var i=n[r].getTrack(t);i&&(e=i)}return e},t.prototype.start=function(t){if(!(this._started>0)){this._started=1;for(var e=this,n=[],r=this._maxTime||0,i=0;i<this._trackKeys.length;i++){var o=this._trackKeys[i],a=this._tracks[o],s=this._getAdditiveTrack(o),l=a.keyframes,u=l.length;if(a.prepare(r,s),a.needsAnimate())if(!this._allowDiscrete&&a.discrete){var c=l[u-1];c&&(e._target[a.propName]=c.rawValue),a.setFinished()}else n.push(a)}if(n.length||this._force){var d=new Ee({life:r,loop:this._loop,delay:this._delay||0,onframe:function(t){e._started=2;var r=e._additiveAnimators;if(r){for(var i=!1,o=0;o<r.length;o++)if(r[o]._clip){i=!0;break}i||(e._additiveAnimators=null)}for(o=0;o<n.length;o++)n[o].step(e._target,t);var a=e._onframeCbs;if(a)for(o=0;o<a.length;o++)a[o](e._target,t)},ondestroy:function(){e._doneCallback()}});this._clip=d,this.animation&&this.animation.addClip(d),t&&d.setEasing(t)}else this._doneCallback();return this}},t.prototype.stop=function(t){if(this._clip){var e=this._clip;t&&e.onframe(1),this._abortedCallback()}},t.prototype.delay=function(t){return this._delay=t,this},t.prototype.during=function(t){return t&&(this._onframeCbs||(this._onframeCbs=[]),this._onframeCbs.push(t)),this},t.prototype.done=function(t){return t&&(this._doneCbs||(this._doneCbs=[]),this._doneCbs.push(t)),this},t.prototype.aborted=function(t){return t&&(this._abortedCbs||(this._abortedCbs=[]),this._abortedCbs.push(t)),this},t.prototype.getClip=function(){return this._clip},t.prototype.getTrack=function(t){return this._tracks[t]},t.prototype.getTracks=function(){var t=this;return Object(s["H"])(this._trackKeys,(function(e){return t._tracks[e]}))},t.prototype.stopTracks=function(t,e){if(!t.length||!this._clip)return!0;for(var n=this._tracks,r=this._trackKeys,i=0;i<t.length;i++){var o=n[t[i]];o&&!o.isFinished()&&(e?o.step(this._target,1):1===this._started&&o.step(this._target,0),o.setFinished())}var a=!0;for(i=0;i<r.length;i++)if(!n[r[i]].isFinished()){a=!1;break}return a&&this._abortedCallback(),a},t.prototype.saveTo=function(t,e,n){if(t){e=e||this._trackKeys;for(var r=0;r<e.length;r++){var i=e[r],o=this._tracks[i];if(o&&!o.isFinished()){var a=o.keyframes,s=a[n?0:a.length-1];s&&(t[i]=Ue(s.rawValue))}}}},t.prototype.__changeFinalValue=function(t,e){e=e||Object(s["F"])(t);for(var n=0;n<e.length;n++){var r=e[n],i=this._tracks[r];if(i){var o=i.keyframes;if(o.length>1){var a=o.pop();i.addKeyframe(a.time,t[r]),i.prepare(this._maxTime,i.getAdditiveTrack())}}}},t}(),un=ln;function cn(){return(new Date).getTime()}var dn=function(t){function e(e){var n=t.call(this)||this;return n._running=!1,n._time=0,n._pausedTime=0,n._pauseStart=0,n._paused=!1,e=e||{},n.stage=e.stage||{},n}return o(e,t),e.prototype.addClip=function(t){t.animation&&this.removeClip(t),this._head?(this._tail.next=t,t.prev=this._tail,t.next=null,this._tail=t):this._head=this._tail=t,t.animation=this},e.prototype.addAnimator=function(t){t.animation=this;var e=t.getClip();e&&this.addClip(e)},e.prototype.removeClip=function(t){if(t.animation){var e=t.prev,n=t.next;e?e.next=n:this._head=n,n?n.prev=e:this._tail=e,t.next=t.prev=t.animation=null}},e.prototype.removeAnimator=function(t){var e=t.getClip();e&&this.removeClip(e),t.animation=null},e.prototype.update=function(t){var e=cn()-this._pausedTime,n=e-this._time,r=this._head;while(r){var i=r.next,o=r.step(e,n);o?(r.ondestroy(),this.removeClip(r),r=i):r=i}this._time=e,t||(this.trigger("frame",n),this.stage.update&&this.stage.update())},e.prototype._startLoop=function(){var t=this;function e(){t._running&&(oe(e),!t._paused&&t.update())}this._running=!0,oe(e)},e.prototype.start=function(){this._running||(this._time=cn(),this._pausedTime=0,this._startLoop())},e.prototype.stop=function(){this._running=!1},e.prototype.pause=function(){this._paused||(this._pauseStart=cn(),this._paused=!0)},e.prototype.resume=function(){this._paused&&(this._pausedTime+=cn()-this._pauseStart,this._paused=!1)},e.prototype.clear=function(){var t=this._head;while(t){var e=t.next;t.prev=t.next=t.animation=null,t=e}this._head=this._tail=null},e.prototype.isFinished=function(){return null==this._head},e.prototype.animate=function(t,e){e=e||{},this.start();var n=new un(t,e.loop);return this.addAnimator(n),n},e}(C),hn=dn,fn=300,pn=a["a"].domSupported,mn=function(){var t=["click","dblclick","mousewheel","wheel","mouseout","mouseup","mousedown","mousemove","contextmenu"],e=["touchstart","touchend","touchmove"],n={pointerdown:1,pointerup:1,pointermove:1,pointerout:1},r=s["H"](t,(function(t){var e=t.replace("mouse","pointer");return n.hasOwnProperty(e)?e:t}));return{mouse:t,touch:e,pointer:r}}(),gn={mouse:["mousemove","mouseup"],pointer:["pointermove","pointerup"]},yn=!1;function vn(t){var e=t.pointerType;return"pen"===e||"touch"===e}function _n(t){t.touching=!0,null!=t.touchTimer&&(clearTimeout(t.touchTimer),t.touchTimer=null),t.touchTimer=setTimeout((function(){t.touching=!1,t.touchTimer=null}),700)}function bn(t){t&&(t.zrByTouch=!0)}function xn(t,e){return J(t.dom,new Mn(t,e),!0)}function wn(t,e){var n=e,r=!1;while(n&&9!==n.nodeType&&!(r=n.domBelongToZr||n!==e&&n===t.painterRoot))n=n.parentNode;return r}var Mn=function(){function t(t,e){this.stopPropagation=s["L"],this.stopImmediatePropagation=s["L"],this.preventDefault=s["L"],this.type=e.type,this.target=this.currentTarget=t.dom,this.pointerType=e.pointerType,this.clientX=e.clientX,this.clientY=e.clientY}return t}(),Sn={mousedown:function(t){t=J(this.dom,t),this.__mayPointerCapture=[t.zrX,t.zrY],this.trigger("mousedown",t)},mousemove:function(t){t=J(this.dom,t);var e=this.__mayPointerCapture;!e||t.zrX===e[0]&&t.zrY===e[1]||this.__togglePointerCapture(!0),this.trigger("mousemove",t)},mouseup:function(t){t=J(this.dom,t),this.__togglePointerCapture(!1),this.trigger("mouseup",t)},mouseout:function(t){t=J(this.dom,t);var e=t.toElement||t.relatedTarget;wn(this,e)||(this.__pointerCapturing&&(t.zrEventControl="no_globalout"),this.trigger("mouseout",t))},wheel:function(t){yn=!0,t=J(this.dom,t),this.trigger("mousewheel",t)},mousewheel:function(t){yn||(t=J(this.dom,t),this.trigger("mousewheel",t))},touchstart:function(t){t=J(this.dom,t),bn(t),this.__lastTouchMoment=new Date,this.handler.processGesture(t,"start"),Sn.mousemove.call(this,t),Sn.mousedown.call(this,t)},touchmove:function(t){t=J(this.dom,t),bn(t),this.handler.processGesture(t,"change"),Sn.mousemove.call(this,t)},touchend:function(t){t=J(this.dom,t),bn(t),this.handler.processGesture(t,"end"),Sn.mouseup.call(this,t),+new Date-+this.__lastTouchMoment<fn&&Sn.click.call(this,t)},pointerdown:function(t){Sn.mousedown.call(this,t)},pointermove:function(t){vn(t)||Sn.mousemove.call(this,t)},pointerup:function(t){Sn.mouseup.call(this,t)},pointerout:function(t){vn(t)||Sn.mouseout.call(this,t)}};s["k"](["click","dblclick","contextmenu"],(function(t){Sn[t]=function(e){e=J(this.dom,e),this.trigger(t,e)}}));var kn={pointermove:function(t){vn(t)||kn.mousemove.call(this,t)},pointerup:function(t){kn.mouseup.call(this,t)},mousemove:function(t){this.trigger("mousemove",t)},mouseup:function(t){var e=this.__pointerCapturing;this.__togglePointerCapture(!1),this.trigger("mouseup",t),e&&(t.zrEventControl="only_globalout",this.trigger("mouseout",t))}};function Ln(t,e){var n=e.domHandlers;a["a"].pointerEventsSupported?s["k"](mn.pointer,(function(r){Dn(e,r,(function(e){n[r].call(t,e)}))})):(a["a"].touchEventsSupported&&s["k"](mn.touch,(function(r){Dn(e,r,(function(i){n[r].call(t,i),_n(e)}))})),s["k"](mn.mouse,(function(r){Dn(e,r,(function(i){i=Z(i),e.touching||n[r].call(t,i)}))})))}function Tn(t,e){function n(n){function r(r){r=Z(r),wn(t,r.target)||(r=xn(t,r),e.domHandlers[n].call(t,r))}Dn(e,n,r,{capture:!0})}a["a"].pointerEventsSupported?s["k"](gn.pointer,n):a["a"].touchEventsSupported||s["k"](gn.mouse,n)}function Dn(t,e,n,r){t.mounted[e]=n,t.listenerOpts[e]=r,Q(t.domTarget,e,n,r)}function On(t){var e=t.mounted;for(var n in e)e.hasOwnProperty(n)&&tt(t.domTarget,n,e[n],t.listenerOpts[n]);t.mounted={}}var In=function(){function t(t,e){this.mounted={},this.listenerOpts={},this.touching=!1,this.domTarget=t,this.domHandlers=e}return t}(),Cn=function(t){function e(e,n){var r=t.call(this)||this;return r.__pointerCapturing=!1,r.dom=e,r.painterRoot=n,r._localHandlerScope=new In(e,Sn),pn&&(r._globalHandlerScope=new In(document,kn)),Ln(r,r._localHandlerScope),r}return o(e,t),e.prototype.dispose=function(){On(this._localHandlerScope),pn&&On(this._globalHandlerScope)},e.prototype.setCursor=function(t){this.dom.style&&(this.dom.style.cursor=t||"default")},e.prototype.__togglePointerCapture=function(t){if(this.__mayPointerCapture=null,pn&&+this.__pointerCapturing^+t){this.__pointerCapturing=t;var e=this._globalHandlerScope;t?Tn(this,e):On(e)}},e}(C),An=Cn,Yn=1;a["a"].hasGlobalWindow&&(Yn=Math.max(window.devicePixelRatio||window.screen&&window.screen.deviceXDPI/window.screen.logicalXDPI||1,1));var jn=Yn,Pn=.4,En="#333",Nn="#ccc",Rn="#eee",Hn=lt,zn=5e-5;function Fn(t){return t>zn||t<-zn}var Bn=[],Wn=[],Vn=st(),Gn=Math.abs,$n=function(){function t(){}return t.prototype.getLocalTransform=function(e){return t.getLocalTransform(this,e)},t.prototype.setPosition=function(t){this.x=t[0],this.y=t[1]},t.prototype.setScale=function(t){this.scaleX=t[0],this.scaleY=t[1]},t.prototype.setSkew=function(t){this.skewX=t[0],this.skewY=t[1]},t.prototype.setOrigin=function(t){this.originX=t[0],this.originY=t[1]},t.prototype.needLocalTransform=function(){return Fn(this.rotation)||Fn(this.x)||Fn(this.y)||Fn(this.scaleX-1)||Fn(this.scaleY-1)||Fn(this.skewX)||Fn(this.skewY)},t.prototype.updateTransform=function(){var t=this.parent&&this.parent.transform,e=this.needLocalTransform(),n=this.transform;e||t?(n=n||st(),e?this.getLocalTransform(n):Hn(n),t&&(e?ct(n,t,n):ut(n,t)),this.transform=n,this._resolveGlobalScaleRatio(n)):n&&(Hn(n),this.invTransform=null)},t.prototype._resolveGlobalScaleRatio=function(t){var e=this.globalScaleRatio;if(null!=e&&1!==e){this.getGlobalScale(Bn);var n=Bn[0]<0?-1:1,r=Bn[1]<0?-1:1,i=((Bn[0]-n)*e+n)/Bn[0]||0,o=((Bn[1]-r)*e+r)/Bn[1]||0;t[0]*=i,t[1]*=i,t[2]*=o,t[3]*=o}this.invTransform=this.invTransform||st(),pt(this.invTransform,t)},t.prototype.getComputedTransform=function(){var t=this,e=[];while(t)e.push(t),t=t.parent;while(t=e.pop())t.updateTransform();return this.transform},t.prototype.setLocalTransform=function(t){if(t){var e=t[0]*t[0]+t[1]*t[1],n=t[2]*t[2]+t[3]*t[3],r=Math.atan2(t[1],t[0]),i=Math.PI/2+r-Math.atan2(t[3],t[2]);n=Math.sqrt(n)*Math.cos(i),e=Math.sqrt(e),this.skewX=i,this.skewY=0,this.rotation=-r,this.x=+t[4],this.y=+t[5],this.scaleX=e,this.scaleY=n,this.originX=0,this.originY=0}},t.prototype.decomposeTransform=function(){if(this.transform){var t=this.parent,e=this.transform;t&&t.transform&&(t.invTransform=t.invTransform||st(),ct(Wn,t.invTransform,e),e=Wn);var n=this.originX,r=this.originY;(n||r)&&(Vn[4]=n,Vn[5]=r,ct(Wn,e,Vn),Wn[4]-=n,Wn[5]-=r,e=Wn),this.setLocalTransform(e)}},t.prototype.getGlobalScale=function(t){var e=this.transform;return t=t||[],e?(t[0]=Math.sqrt(e[0]*e[0]+e[1]*e[1]),t[1]=Math.sqrt(e[2]*e[2]+e[3]*e[3]),e[0]<0&&(t[0]=-t[0]),e[3]<0&&(t[1]=-t[1]),t):(t[0]=1,t[1]=1,t)},t.prototype.transformCoordToLocal=function(t,e){var n=[t,e],r=this.invTransform;return r&&S(n,n,r),n},t.prototype.transformCoordToGlobal=function(t,e){var n=[t,e],r=this.transform;return r&&S(n,n,r),n},t.prototype.getLineScale=function(){var t=this.transform;return t&&Gn(t[0]-1)>1e-10&&Gn(t[3]-1)>1e-10?Math.sqrt(Gn(t[0]*t[3]-t[2]*t[1])):1},t.prototype.copyTransform=function(t){qn(this,t)},t.getLocalTransform=function(t,e){e=e||[];var n=t.originX||0,r=t.originY||0,i=t.scaleX,o=t.scaleY,a=t.anchorX,s=t.anchorY,l=t.rotation||0,u=t.x,c=t.y,d=t.skewX?Math.tan(t.skewX):0,h=t.skewY?Math.tan(-t.skewY):0;if(n||r||a||s){var f=n+a,p=r+s;e[4]=-f*i-d*p*o,e[5]=-p*o-h*f*i}else e[4]=e[5]=0;return e[0]=i,e[3]=o,e[1]=h*i,e[2]=d*o,l&&ht(e,e,l),e[4]+=n+u,e[5]+=r+c,e},t.initDefaultProps=function(){var e=t.prototype;e.scaleX=e.scaleY=e.globalScaleRatio=1,e.x=e.y=e.originX=e.originY=e.skewX=e.skewY=e.rotation=e.anchorX=e.anchorY=0}(),t}(),Un=["x","y","originX","originY","anchorX","anchorY","rotation","scaleX","scaleY","skewX","skewY"];function qn(t,e){for(var n=0;n<Un.length;n++){var r=Un[n];t[r]=e[r]}}var Xn=$n,Zn=n("d51b"),Jn=n("726e"),Kn={};function Qn(t,e){e=e||Jn["a"];var n=Kn[e];n||(n=Kn[e]=new Zn["a"](500));var r=n.get(t);return null==r&&(r=Jn["d"].measureText(t,e).width,n.put(t,r)),r}function tr(t,e,n,r){var i=Qn(t,e),o=ir(e),a=nr(0,i,n),s=rr(0,o,r),l=new Tt(a,s,i,o);return l}function er(t,e,n,r){var i=((t||"")+"").split("\n"),o=i.length;if(1===o)return tr(i[0],e,n,r);for(var a=new Tt(0,0,0,0),s=0;s<i.length;s++){var l=tr(i[s],e,n,r);0===s?a.copy(l):a.union(l)}return a}function nr(t,e,n){return"right"===n?t-=e:"center"===n&&(t-=e/2),t}function rr(t,e,n){return"middle"===n?t-=e/2:"bottom"===n&&(t-=e),t}function ir(t){return Qn("国",t)}function or(t,e){return"string"===typeof t?t.lastIndexOf("%")>=0?parseFloat(t)/100*e:parseFloat(t):t}function ar(t,e,n){var r=e.position||"inside",i=null!=e.distance?e.distance:5,o=n.height,a=n.width,s=o/2,l=n.x,u=n.y,c="left",d="top";if(r instanceof Array)l+=or(r[0],n.width),u+=or(r[1],n.height),c=null,d=null;else switch(r){case"left":l-=i,u+=s,c="right",d="middle";break;case"right":l+=i+a,u+=s,d="middle";break;case"top":l+=a/2,u-=i,c="center",d="bottom";break;case"bottom":l+=a/2,u+=o+i,c="center";break;case"inside":l+=a/2,u+=s,c="center",d="middle";break;case"insideLeft":l+=i,u+=s,d="middle";break;case"insideRight":l+=a-i,u+=s,c="right",d="middle";break;case"insideTop":l+=a/2,u+=i,c="center";break;case"insideBottom":l+=a/2,u+=o-i,c="center",d="bottom";break;case"insideTopLeft":l+=i,u+=i;break;case"insideTopRight":l+=a-i,u+=i,c="right";break;case"insideBottomLeft":l+=i,u+=o-i,d="bottom";break;case"insideBottomRight":l+=a-i,u+=o-i,c="right",d="bottom";break}return t=t||{},t.x=l,t.y=u,t.align=c,t.verticalAlign=d,t}var sr="__zr_normal__",lr=Un.concat(["ignore"]),ur=Object(s["N"])(Un,(function(t,e){return t[e]=!0,t}),{ignore:!1}),cr={},dr=new Tt(0,0,0,0),hr=function(){function t(t){this.id=Object(s["p"])(),this.animators=[],this.currentStates=[],this.states={},this._init(t)}return t.prototype._init=function(t){this.attr(t)},t.prototype.drift=function(t,e,n){switch(this.draggable){case"horizontal":e=0;break;case"vertical":t=0;break}var r=this.transform;r||(r=this.transform=[1,0,0,1,0,0]),r[4]+=t,r[5]+=e,this.decomposeTransform(),this.markRedraw()},t.prototype.beforeUpdate=function(){},t.prototype.afterUpdate=function(){},t.prototype.update=function(){this.updateTransform(),this.__dirty&&this.updateInnerText()},t.prototype.updateInnerText=function(t){var e=this._textContent;if(e&&(!e.ignore||t)){this.textConfig||(this.textConfig={});var n=this.textConfig,r=n.local,i=e.innerTransformable,o=void 0,a=void 0,s=!1;i.parent=r?this:null;var l=!1;if(i.copyTransform(e),null!=n.position){var u=dr;n.layoutRect?u.copy(n.layoutRect):u.copy(this.getBoundingRect()),r||u.applyTransform(this.transform),this.calculateTextPosition?this.calculateTextPosition(cr,n,u):ar(cr,n,u),i.x=cr.x,i.y=cr.y,o=cr.align,a=cr.verticalAlign;var c=n.origin;if(c&&null!=n.rotation){var d=void 0,h=void 0;"center"===c?(d=.5*u.width,h=.5*u.height):(d=or(c[0],u.width),h=or(c[1],u.height)),l=!0,i.originX=-i.x+d+(r?0:u.x),i.originY=-i.y+h+(r?0:u.y)}}null!=n.rotation&&(i.rotation=n.rotation);var f=n.offset;f&&(i.x+=f[0],i.y+=f[1],l||(i.originX=-f[0],i.originY=-f[1]));var p=null==n.inside?"string"===typeof n.position&&n.position.indexOf("inside")>=0:n.inside,m=this._innerTextDefaultStyle||(this._innerTextDefaultStyle={}),g=void 0,y=void 0,v=void 0;p&&this.canBeInsideText()?(g=n.insideFill,y=n.insideStroke,null!=g&&"auto"!==g||(g=this.getInsideTextFill()),null!=y&&"auto"!==y||(y=this.getInsideTextStroke(g),v=!0)):(g=n.outsideFill,y=n.outsideStroke,null!=g&&"auto"!==g||(g=this.getOutsideFill()),null!=y&&"auto"!==y||(y=this.getOutsideStroke(g),v=!0)),g=g||"#000",g===m.fill&&y===m.stroke&&v===m.autoStroke&&o===m.align&&a===m.verticalAlign||(s=!0,m.fill=g,m.stroke=y,m.autoStroke=v,m.align=o,m.verticalAlign=a,e.setDefaultTextStyle(m)),e.__dirty|=Zt,s&&e.dirtyStyle(!0)}},t.prototype.canBeInsideText=function(){return!0},t.prototype.getInsideTextFill=function(){return"#fff"},t.prototype.getInsideTextStroke=function(t){return"#000"},t.prototype.getOutsideFill=function(){return this.__zr&&this.__zr.isDarkMode()?Nn:En},t.prototype.getOutsideStroke=function(t){var e=this.__zr&&this.__zr.getBackgroundColor(),n="string"===typeof e&&Object(Ne["h"])(e);n||(n=[255,255,255,1]);for(var r=n[3],i=this.__zr.isDarkMode(),o=0;o<3;o++)n[o]=n[o]*r+(i?0:255)*(1-r);return n[3]=1,Object(Ne["i"])(n,"rgba")},t.prototype.traverse=function(t,e){},t.prototype.attrKV=function(t,e){"textConfig"===t?this.setTextConfig(e):"textContent"===t?this.setTextContent(e):"clipPath"===t?this.setClipPath(e):"extra"===t?(this.extra=this.extra||{},Object(s["m"])(this.extra,e)):this[t]=e},t.prototype.hide=function(){this.ignore=!0,this.markRedraw()},t.prototype.show=function(){this.ignore=!1,this.markRedraw()},t.prototype.attr=function(t,e){if("string"===typeof t)this.attrKV(t,e);else if(Object(s["A"])(t))for(var n=t,r=Object(s["F"])(n),i=0;i<r.length;i++){var o=r[i];this.attrKV(o,t[o])}return this.markRedraw(),this},t.prototype.saveCurrentToNormalState=function(t){this._innerSaveToNormal(t);for(var e=this._normalState,n=0;n<this.animators.length;n++){var r=this.animators[n],i=r.__fromStateTransition;if(!(r.getLoop()||i&&i!==sr)){var o=r.targetName,a=o?e[o]:e;r.saveTo(a)}}},t.prototype._innerSaveToNormal=function(t){var e=this._normalState;e||(e=this._normalState={}),t.textConfig&&!e.textConfig&&(e.textConfig=this.textConfig),this._savePrimaryToNormal(t,e,lr)},t.prototype._savePrimaryToNormal=function(t,e,n){for(var r=0;r<n.length;r++){var i=n[r];null==t[i]||i in e||(e[i]=this[i])}},t.prototype.hasState=function(){return this.currentStates.length>0},t.prototype.getState=function(t){return this.states[t]},t.prototype.ensureState=function(t){var e=this.states;return e[t]||(e[t]={}),e[t]},t.prototype.clearStates=function(t){this.useState(sr,!1,t)},t.prototype.useState=function(t,e,n,r){var i=t===sr,o=this.hasState();if(o||!i){var a=this.currentStates,l=this.stateTransition;if(!(Object(s["r"])(a,t)>=0)||!e&&1!==a.length){var u;if(this.stateProxy&&!i&&(u=this.stateProxy(t)),u||(u=this.states&&this.states[t]),u||i){i||this.saveCurrentToNormalState(u);var c=!!(u&&u.hoverLayer||r);c&&this._toggleHoverLayerFlag(!0),this._applyStateObj(t,u,this._normalState,e,!n&&!this.__inHover&&l&&l.duration>0,l);var d=this._textContent,h=this._textGuide;return d&&d.useState(t,e,n,c),h&&h.useState(t,e,n,c),i?(this.currentStates=[],this._normalState={}):e?this.currentStates.push(t):this.currentStates=[t],this._updateAnimationTargets(),this.markRedraw(),!c&&this.__inHover&&(this._toggleHoverLayerFlag(!1),this.__dirty&=~Zt),u}Object(s["G"])("State "+t+" not exists.")}}},t.prototype.useStates=function(t,e,n){if(t.length){var r=[],i=this.currentStates,o=t.length,a=o===i.length;if(a)for(var s=0;s<o;s++)if(t[s]!==i[s]){a=!1;break}if(a)return;for(s=0;s<o;s++){var l=t[s],u=void 0;this.stateProxy&&(u=this.stateProxy(l,t)),u||(u=this.states[l]),u&&r.push(u)}var c=r[o-1],d=!!(c&&c.hoverLayer||n);d&&this._toggleHoverLayerFlag(!0);var h=this._mergeStates(r),f=this.stateTransition;this.saveCurrentToNormalState(h),this._applyStateObj(t.join(","),h,this._normalState,!1,!e&&!this.__inHover&&f&&f.duration>0,f);var p=this._textContent,m=this._textGuide;p&&p.useStates(t,e,d),m&&m.useStates(t,e,d),this._updateAnimationTargets(),this.currentStates=t.slice(),this.markRedraw(),!d&&this.__inHover&&(this._toggleHoverLayerFlag(!1),this.__dirty&=~Zt)}else this.clearStates()},t.prototype.isSilent=function(){var t=this.silent,e=this.parent;while(!t&&e){if(e.silent){t=!0;break}e=e.parent}return t},t.prototype._updateAnimationTargets=function(){for(var t=0;t<this.animators.length;t++){var e=this.animators[t];e.targetName&&e.changeTarget(this[e.targetName])}},t.prototype.removeState=function(t){var e=Object(s["r"])(this.currentStates,t);if(e>=0){var n=this.currentStates.slice();n.splice(e,1),this.useStates(n)}},t.prototype.replaceState=function(t,e,n){var r=this.currentStates.slice(),i=Object(s["r"])(r,t),o=Object(s["r"])(r,e)>=0;i>=0?o?r.splice(i,1):r[i]=e:n&&!o&&r.push(e),this.useStates(r)},t.prototype.toggleState=function(t,e){e?this.useState(t,!0):this.removeState(t)},t.prototype._mergeStates=function(t){for(var e,n={},r=0;r<t.length;r++){var i=t[r];Object(s["m"])(n,i),i.textConfig&&(e=e||{},Object(s["m"])(e,i.textConfig))}return e&&(n.textConfig=e),n},t.prototype._applyStateObj=function(t,e,n,r,i,o){var a=!(e&&r);e&&e.textConfig?(this.textConfig=Object(s["m"])({},r?this.textConfig:n.textConfig),Object(s["m"])(this.textConfig,e.textConfig)):a&&n.textConfig&&(this.textConfig=n.textConfig);for(var l={},u=!1,c=0;c<lr.length;c++){var d=lr[c],h=i&&ur[d];e&&null!=e[d]?h?(u=!0,l[d]=e[d]):this[d]=e[d]:a&&null!=n[d]&&(h?(u=!0,l[d]=n[d]):this[d]=n[d])}if(!i)for(c=0;c<this.animators.length;c++){var f=this.animators[c],p=f.targetName;f.getLoop()||f.__changeFinalValue(p?(e||n)[p]:e||n)}u&&this._transitionState(t,l,o)},t.prototype._attachComponent=function(t){if((!t.__zr||t.__hostTarget)&&t!==this){var e=this.__zr;e&&t.addSelfToZr(e),t.__zr=e,t.__hostTarget=this}},t.prototype._detachComponent=function(t){t.__zr&&t.removeSelfFromZr(t.__zr),t.__zr=null,t.__hostTarget=null},t.prototype.getClipPath=function(){return this._clipPath},t.prototype.setClipPath=function(t){this._clipPath&&this._clipPath!==t&&this.removeClipPath(),this._attachComponent(t),this._clipPath=t,this.markRedraw()},t.prototype.removeClipPath=function(){var t=this._clipPath;t&&(this._detachComponent(t),this._clipPath=null,this.markRedraw())},t.prototype.getTextContent=function(){return this._textContent},t.prototype.setTextContent=function(t){var e=this._textContent;e!==t&&(e&&e!==t&&this.removeTextContent(),t.innerTransformable=new Xn,this._attachComponent(t),this._textContent=t,this.markRedraw())},t.prototype.setTextConfig=function(t){this.textConfig||(this.textConfig={}),Object(s["m"])(this.textConfig,t),this.markRedraw()},t.prototype.removeTextConfig=function(){this.textConfig=null,this.markRedraw()},t.prototype.removeTextContent=function(){var t=this._textContent;t&&(t.innerTransformable=null,this._detachComponent(t),this._textContent=null,this._innerTextDefaultStyle=null,this.markRedraw())},t.prototype.getTextGuideLine=function(){return this._textGuide},t.prototype.setTextGuideLine=function(t){this._textGuide&&this._textGuide!==t&&this.removeTextGuideLine(),this._attachComponent(t),this._textGuide=t,this.markRedraw()},t.prototype.removeTextGuideLine=function(){var t=this._textGuide;t&&(this._detachComponent(t),this._textGuide=null,this.markRedraw())},t.prototype.markRedraw=function(){this.__dirty|=Zt;var t=this.__zr;t&&(this.__inHover?t.refreshHover():t.refresh()),this.__hostTarget&&this.__hostTarget.markRedraw()},t.prototype.dirty=function(){this.markRedraw()},t.prototype._toggleHoverLayerFlag=function(t){this.__inHover=t;var e=this._textContent,n=this._textGuide;e&&(e.__inHover=t),n&&(n.__inHover=t)},t.prototype.addSelfToZr=function(t){if(this.__zr!==t){this.__zr=t;var e=this.animators;if(e)for(var n=0;n<e.length;n++)t.animation.addAnimator(e[n]);this._clipPath&&this._clipPath.addSelfToZr(t),this._textContent&&this._textContent.addSelfToZr(t),this._textGuide&&this._textGuide.addSelfToZr(t)}},t.prototype.removeSelfFromZr=function(t){if(this.__zr){this.__zr=null;var e=this.animators;if(e)for(var n=0;n<e.length;n++)t.animation.removeAnimator(e[n]);this._clipPath&&this._clipPath.removeSelfFromZr(t),this._textContent&&this._textContent.removeSelfFromZr(t),this._textGuide&&this._textGuide.removeSelfFromZr(t)}},t.prototype.animate=function(t,e,n){var r=t?this[t]:this;var i=new un(r,e,n);return t&&(i.targetName=t),this.addAnimator(i,t),i},t.prototype.addAnimator=function(t,e){var n=this.__zr,r=this;t.during((function(){r.updateDuringAnimation(e)})).done((function(){var e=r.animators,n=Object(s["r"])(e,t);n>=0&&e.splice(n,1)})),this.animators.push(t),n&&n.animation.addAnimator(t),n&&n.wakeUp()},t.prototype.updateDuringAnimation=function(t){this.markRedraw()},t.prototype.stopAnimation=function(t,e){for(var n=this.animators,r=n.length,i=[],o=0;o<r;o++){var a=n[o];t&&t!==a.scope?i.push(a):a.stop(e)}return this.animators=i,this},t.prototype.animateTo=function(t,e,n){fr(this,t,e,n)},t.prototype.animateFrom=function(t,e,n){fr(this,t,e,n,!0)},t.prototype._transitionState=function(t,e,n,r){for(var i=fr(this,e,n,r),o=0;o<i.length;o++)i[o].__fromStateTransition=t},t.prototype.getBoundingRect=function(){return null},t.prototype.getPaintRect=function(){return null},t.initDefaultProps=function(){var e=t.prototype;e.type="element",e.name="",e.ignore=e.silent=e.isGroup=e.draggable=e.dragging=e.ignoreClip=e.__inHover=!1,e.__dirty=Zt;function n(t,n,r,i){function o(t,e){Object.defineProperty(e,0,{get:function(){return t[r]},set:function(e){t[r]=e}}),Object.defineProperty(e,1,{get:function(){return t[i]},set:function(e){t[i]=e}})}Object.defineProperty(e,t,{get:function(){if(!this[n]){var t=this[n]=[];o(this,t)}return this[n]},set:function(t){this[r]=t[0],this[i]=t[1],this[n]=t,o(this,t)}})}Object.defineProperty&&(n("position","_legacyPos","x","y"),n("scale","_legacyScale","scaleX","scaleY"),n("origin","_legacyOrigin","originX","originY"))}(),t}();function fr(t,e,n,r,i){n=n||{};var o=[];_r(t,"",t,e,n,r,o,i);var a=o.length,s=!1,l=n.done,u=n.aborted,c=function(){s=!0,a--,a<=0&&(s?l&&l():u&&u())},d=function(){a--,a<=0&&(s?l&&l():u&&u())};a||l&&l(),o.length>0&&n.during&&o[0].during((function(t,e){n.during(e)}));for(var h=0;h<o.length;h++){var f=o[h];c&&f.done(c),d&&f.aborted(d),n.force&&f.duration(n.duration),f.start(n.easing)}return o}function pr(t,e,n){for(var r=0;r<n;r++)t[r]=e[r]}function mr(t){return Object(s["u"])(t[0])}function gr(t,e,n){if(Object(s["u"])(e[n]))if(Object(s["u"])(t[n])||(t[n]=[]),Object(s["E"])(e[n])){var r=e[n].length;t[n].length!==r&&(t[n]=new e[n].constructor(r),pr(t[n],e[n],r))}else{var i=e[n],o=t[n],a=i.length;if(mr(i))for(var l=i[0].length,u=0;u<a;u++)o[u]?pr(o[u],i[u],l):o[u]=Array.prototype.slice.call(i[u]);else pr(o,i,a);o.length=i.length}else t[n]=e[n]}function yr(t,e){return t===e||Object(s["u"])(t)&&Object(s["u"])(e)&&vr(t,e)}function vr(t,e){var n=t.length;if(n!==e.length)return!1;for(var r=0;r<n;r++)if(t[r]!==e[r])return!1;return!0}function _r(t,e,n,r,i,o,a,l){for(var u=Object(s["F"])(r),c=i.duration,d=i.delay,h=i.additive,f=i.setToFinal,p=!Object(s["A"])(o),m=t.animators,g=[],y=0;y<u.length;y++){var v=u[y],_=r[v];if(null!=_&&null!=n[v]&&(p||o[v]))if(!Object(s["A"])(_)||Object(s["u"])(_)||Object(s["x"])(_))g.push(v);else{if(e){l||(n[v]=_,t.updateDuringAnimation(e));continue}_r(t,v,n[v],_,i,o&&o[v],a,l)}else l||(n[v]=_,t.updateDuringAnimation(e),g.push(v))}var b=g.length;if(!h&&b)for(var x=0;x<m.length;x++){var w=m[x];if(w.targetName===e){var M=w.stopTracks(g);if(M){var S=Object(s["r"])(m,w);m.splice(S,1)}}}if(i.force||(g=Object(s["n"])(g,(function(t){return!yr(r[t],n[t])})),b=g.length),b>0||i.force&&!a.length){var k=void 0,L=void 0,T=void 0;if(l){L={},f&&(k={});for(x=0;x<b;x++){v=g[x];L[v]=n[v],f?k[v]=r[v]:n[v]=r[v]}}else if(f){T={};for(x=0;x<b;x++){v=g[x];T[v]=Ue(n[v]),gr(n,r,v)}}w=new un(n,!1,!1,h?Object(s["n"])(m,(function(t){return t.targetName===e})):null);w.targetName=e,i.scope&&(w.scope=i.scope),f&&k&&w.whenWithKeys(0,k,g),T&&w.whenWithKeys(0,T,g),w.whenWithKeys(null==c?500:c,l?L:r,g).delay(d||0),t.addAnimator(w,e),a.push(w)}}Object(s["K"])(hr,C),Object(s["K"])(hr,Xn);var br=hr,xr=function(t){function e(e){var n=t.call(this)||this;return n.isGroup=!0,n._children=[],n.attr(e),n}return o(e,t),e.prototype.childrenRef=function(){return this._children},e.prototype.children=function(){return this._children.slice()},e.prototype.childAt=function(t){return this._children[t]},e.prototype.childOfName=function(t){for(var e=this._children,n=0;n<e.length;n++)if(e[n].name===t)return e[n]},e.prototype.childCount=function(){return this._children.length},e.prototype.add=function(t){return t&&t!==this&&t.parent!==this&&(this._children.push(t),this._doAdd(t)),this},e.prototype.addBefore=function(t,e){if(t&&t!==this&&t.parent!==this&&e&&e.parent===this){var n=this._children,r=n.indexOf(e);r>=0&&(n.splice(r,0,t),this._doAdd(t))}return this},e.prototype.replace=function(t,e){var n=s["r"](this._children,t);return n>=0&&this.replaceAt(e,n),this},e.prototype.replaceAt=function(t,e){var n=this._children,r=n[e];if(t&&t!==this&&t.parent!==this&&t!==r){n[e]=t,r.parent=null;var i=this.__zr;i&&r.removeSelfFromZr(i),this._doAdd(t)}return this},e.prototype._doAdd=function(t){t.parent&&t.parent.remove(t),t.parent=this;var e=this.__zr;e&&e!==t.__zr&&t.addSelfToZr(e),e&&e.refresh()},e.prototype.remove=function(t){var e=this.__zr,n=this._children,r=s["r"](n,t);return r<0||(n.splice(r,1),t.parent=null,e&&t.removeSelfFromZr(e),e&&e.refresh()),this},e.prototype.removeAll=function(){for(var t=this._children,e=this.__zr,n=0;n<t.length;n++){var r=t[n];e&&r.removeSelfFromZr(e),r.parent=null}return t.length=0,this},e.prototype.eachChild=function(t,e){for(var n=this._children,r=0;r<n.length;r++){var i=n[r];t.call(e,i,r)}return this},e.prototype.traverse=function(t,e){for(var n=0;n<this._children.length;n++){var r=this._children[n],i=t.call(e,r);r.isGroup&&!i&&r.traverse(t,e)}return this},e.prototype.addSelfToZr=function(e){t.prototype.addSelfToZr.call(this,e);for(var n=0;n<this._children.length;n++){var r=this._children[n];r.addSelfToZr(e)}},e.prototype.removeSelfFromZr=function(e){t.prototype.removeSelfFromZr.call(this,e);for(var n=0;n<this._children.length;n++){var r=this._children[n];r.removeSelfFromZr(e)}},e.prototype.getBoundingRect=function(t){for(var e=new Tt(0,0,0,0),n=t||this._children,r=[],i=null,o=0;o<n.length;o++){var a=n[o];if(!a.ignore&&!a.invisible){var s=a.getBoundingRect(),l=a.getLocalTransform(r);l?(Tt.applyTransform(e,s,l),i=i||e.clone(),i.union(e)):(i=i||s.clone(),i.union(s))}}return i||e},e}(br);xr.prototype.type="group";var wr=xr,Mr={},Sr={};
/*!
* ZRender, a high performance 2d drawing library.
*
* Copyright (c) 2013, Baidu Inc.
* All rights reserved.
*
* LICENSE
* https://github.com/ecomfe/zrender/blob/master/LICENSE.txt
*/function kr(t){delete Sr[t]}function Lr(t){if(!t)return!1;if("string"===typeof t)return Object(Ne["e"])(t,1)<Pn;if(t.colorStops){for(var e=t.colorStops,n=0,r=e.length,i=0;i<r;i++)n+=Object(Ne["e"])(e[i].color,1);return n/=r,n<Pn}return!1}var Tr,Dr=function(){function t(t,e,n){var r=this;this._sleepAfterStill=10,this._stillFrameAccum=0,this._needsRefresh=!0,this._needsRefreshHover=!0,this._darkMode=!1,n=n||{},this.dom=e,this.id=t;var i=new ie,o=n.renderer||"canvas";Mr[o]||(o=s["F"](Mr)[0]),n.useDirtyRect=null!=n.useDirtyRect&&n.useDirtyRect;var l=new Mr[o](e,i,n,t),u=n.ssr||l.ssrOnly;this.storage=i,this.painter=l;var c,d=a["a"].node||a["a"].worker||u?null:new An(l.getViewportRoot(),l.root),h=n.useCoarsePointer,f=null==h||"auto"===h?a["a"].touchEventsSupported:!!h,p=44;f&&(c=s["P"](n.pointerSize,p)),this.handler=new Ht(i,l,d,l.root,c),this.animation=new hn({stage:{update:u?null:function(){return r._flush(!0)}}}),u||this.animation.start()}return t.prototype.add=function(t){!this._disposed&&t&&(this.storage.addRoot(t),t.addSelfToZr(this),this.refresh())},t.prototype.remove=function(t){!this._disposed&&t&&(this.storage.delRoot(t),t.removeSelfFromZr(this),this.refresh())},t.prototype.configLayer=function(t,e){this._disposed||(this.painter.configLayer&&this.painter.configLayer(t,e),this.refresh())},t.prototype.setBackgroundColor=function(t){this._disposed||(this.painter.setBackgroundColor&&this.painter.setBackgroundColor(t),this.refresh(),this._backgroundColor=t,this._darkMode=Lr(t))},t.prototype.getBackgroundColor=function(){return this._backgroundColor},t.prototype.setDarkMode=function(t){this._darkMode=t},t.prototype.isDarkMode=function(){return this._darkMode},t.prototype.refreshImmediately=function(t){this._disposed||(t||this.animation.update(!0),this._needsRefresh=!1,this.painter.refresh(),this._needsRefresh=!1)},t.prototype.refresh=function(){this._disposed||(this._needsRefresh=!0,this.animation.start())},t.prototype.flush=function(){this._disposed||this._flush(!1)},t.prototype._flush=function(t){var e,n=cn();this._needsRefresh&&(e=!0,this.refreshImmediately(t)),this._needsRefreshHover&&(e=!0,this.refreshHoverImmediately());var r=cn();e?(this._stillFrameAccum=0,this.trigger("rendered",{elapsedTime:r-n})):this._sleepAfterStill>0&&(this._stillFrameAccum++,this._stillFrameAccum>this._sleepAfterStill&&this.animation.stop())},t.prototype.setSleepAfterStill=function(t){this._sleepAfterStill=t},t.prototype.wakeUp=function(){this._disposed||(this.animation.start(),this._stillFrameAccum=0)},t.prototype.refreshHover=function(){this._needsRefreshHover=!0},t.prototype.refreshHoverImmediately=function(){this._disposed||(this._needsRefreshHover=!1,this.painter.refreshHover&&"canvas"===this.painter.getType()&&this.painter.refreshHover())},t.prototype.resize=function(t){this._disposed||(t=t||{},this.painter.resize(t.width,t.height),this.handler.resize())},t.prototype.clearAnimation=function(){this._disposed||this.animation.clear()},t.prototype.getWidth=function(){if(!this._disposed)return this.painter.getWidth()},t.prototype.getHeight=function(){if(!this._disposed)return this.painter.getHeight()},t.prototype.setCursorStyle=function(t){this._disposed||this.handler.setCursorStyle(t)},t.prototype.findHover=function(t,e){if(!this._disposed)return this.handler.findHover(t,e)},t.prototype.on=function(t,e,n){return this._disposed||this.handler.on(t,e,n),this},t.prototype.off=function(t,e){this._disposed||this.handler.off(t,e)},t.prototype.trigger=function(t,e){this._disposed||this.handler.trigger(t,e)},t.prototype.clear=function(){if(!this._disposed){for(var t=this.storage.getRoots(),e=0;e<t.length;e++)t[e]instanceof wr&&t[e].removeSelfFromZr(this);this.storage.delAllRoots(),this.painter.clear()}},t.prototype.dispose=function(){this._disposed||(this.animation.stop(),this.clear(),this.storage.dispose(),this.painter.dispose(),this.handler.dispose(),this.animation=this.storage=this.painter=this.handler=null,this._disposed=!0,kr(this.id))},t}();function Or(t,e){var n=new Dr(s["p"](),t,e);return Sr[n.id]=n,n}function Ir(t,e){Mr[t]=e}function Cr(t){if("function"===typeof Tr)return Tr(t)}function Ar(t){Tr=t}var Yr=1e-4,jr=20;function Pr(t){return t.replace(/^\s+|\s+$/g,"")}function Er(t,e,n,r){var i=e[0],o=e[1],a=n[0],s=n[1],l=o-i,u=s-a;if(0===l)return 0===u?a:(a+s)/2;if(r)if(l>0){if(t<=i)return a;if(t>=o)return s}else{if(t>=i)return a;if(t<=o)return s}else{if(t===i)return a;if(t===o)return s}return(t-i)/l*u+a}function Nr(t,e){switch(t){case"center":case"middle":t="50%";break;case"left":case"top":t="0%";break;case"right":case"bottom":t="100%";break}return s["C"](t)?Pr(t).match(/%$/)?parseFloat(t)/100*e:parseFloat(t):null==t?NaN:+t}function Rr(t,e,n){return null==e&&(e=10),e=Math.min(Math.max(0,e),jr),t=(+t).toFixed(e),n?t:+t}function Hr(t){return t.sort((function(t,e){return t-e})),t}function zr(t){if(t=+t,isNaN(t))return 0;if(t>1e-14)for(var e=1,n=0;n<15;n++,e*=10)if(Math.round(t*e)/e===t)return n;return Fr(t)}function Fr(t){var e=t.toString().toLowerCase(),n=e.indexOf("e"),r=n>0?+e.slice(n+1):0,i=n>0?n:e.length,o=e.indexOf("."),a=o<0?0:i-1-o;return Math.max(0,a-r)}function Br(t,e){var n=Math.log,r=Math.LN10,i=Math.floor(n(t[1]-t[0])/r),o=Math.round(n(Math.abs(e[1]-e[0]))/r),a=Math.min(Math.max(-i+o,0),20);return isFinite(a)?a:20}function Wr(t,e){var n=s["N"](t,(function(t,e){return t+(isNaN(e)?0:e)}),0);if(0===n)return[];var r=Math.pow(10,e),i=s["H"](t,(function(t){return(isNaN(t)?0:t)/n*r*100})),o=100*r,a=s["H"](i,(function(t){return Math.floor(t)})),l=s["N"](a,(function(t,e){return t+e}),0),u=s["H"](i,(function(t,e){return t-a[e]}));while(l<o){for(var c=Number.NEGATIVE_INFINITY,d=null,h=0,f=u.length;h<f;++h)u[h]>c&&(c=u[h],d=h);++a[d],u[d]=0,++l}return s["H"](a,(function(t){return t/r}))}function Vr(t,e){var n=Math.max(zr(t),zr(e)),r=t+e;return n>jr?r:Rr(r,n)}var Gr=9007199254740991;function $r(t){var e=2*Math.PI;return(t%e+e)%e}function Ur(t){return t>-Yr&&t<Yr}var qr=/^(?:(\d{4})(?:[-\/](\d{1,2})(?:[-\/](\d{1,2})(?:[T ](\d{1,2})(?::(\d{1,2})(?::(\d{1,2})(?:[.,](\d+))?)?)?(Z|[\+\-]\d\d:?\d\d)?)?)?)?)?$/;function Xr(t){if(t instanceof Date)return t;if(s["C"](t)){var e=qr.exec(t);if(!e)return new Date(NaN);if(e[8]){var n=+e[4]||0;return"Z"!==e[8].toUpperCase()&&(n-=+e[8].slice(0,3)),new Date(Date.UTC(+e[1],+(e[2]||1)-1,+e[3]||1,n,+(e[5]||0),+e[6]||0,e[7]?+e[7].substring(0,3):0))}return new Date(+e[1],+(e[2]||1)-1,+e[3]||1,+e[4]||0,+(e[5]||0),+e[6]||0,e[7]?+e[7].substring(0,3):0)}return null==t?new Date(NaN):new Date(Math.round(t))}function Zr(t){return Math.pow(10,Jr(t))}function Jr(t){if(0===t)return 0;var e=Math.floor(Math.log(t)/Math.LN10);return t/Math.pow(10,e)>=10&&e++,e}function Kr(t,e){var n,r=Jr(t),i=Math.pow(10,r),o=t/i;return n=e?o<1.5?1:o<2.5?2:o<4?3:o<7?5:10:o<1?1:o<2?2:o<3?3:o<5?5:10,t=n*i,r>=-20?+t.toFixed(r<0?-r:0):t}function Qr(t,e){var n=(t.length-1)*e+1,r=Math.floor(n),i=+t[r-1],o=n-r;return o?i+o*(t[r]-i):i}function ti(t){t.sort((function(t,e){return s(t,e,0)?-1:1}));for(var e=-1/0,n=1,r=0;r<t.length;){for(var i=t[r].interval,o=t[r].close,a=0;a<2;a++)i[a]<=e&&(i[a]=e,o[a]=a?1:1-n),e=i[a],n=o[a];i[0]===i[1]&&o[0]*o[1]!==1?t.splice(r,1):r++}return t;function s(t,e,n){return t.interval[n]<e.interval[n]||t.interval[n]===e.interval[n]&&(t.close[n]-e.close[n]===(n?-1:1)||!n&&s(t,e,1))}}function ei(t){var e=parseFloat(t);return e==t&&(0!==e||!s["C"](t)||t.indexOf("x")<=0)?e:NaN}function ni(t){return!isNaN(ei(t))}function ri(){return Math.round(9*Math.random())}function ii(t,e){return 0===e?t:ii(e,t%e)}function oi(t,e){return null==t?e:null==e?t:t*e/ii(t,e)}function ai(t,e,n){return(e-t)*n+t}var si="series\0",li="\0_ec_\0";function ui(t){return t instanceof Array?t:null==t?[]:[t]}function ci(t,e,n){if(t){t[e]=t[e]||{},t.emphasis=t.emphasis||{},t.emphasis[e]=t.emphasis[e]||{};for(var r=0,i=n.length;r<i;r++){var o=n[r];!t.emphasis[e].hasOwnProperty(o)&&t[e].hasOwnProperty(o)&&(t.emphasis[e][o]=t[e][o])}}}var di=["fontStyle","fontWeight","fontSize","fontFamily","rich","tag","color","textBorderColor","textBorderWidth","width","height","lineHeight","align","verticalAlign","baseline","shadowColor","shadowBlur","shadowOffsetX","shadowOffsetY","textShadowColor","textShadowBlur","textShadowOffsetX","textShadowOffsetY","backgroundColor","borderColor","borderWidth","borderRadius","padding"];function hi(t){return!Object(s["A"])(t)||Object(s["t"])(t)||t instanceof Date?t:t.value}function fi(t){return Object(s["A"])(t)&&!(t instanceof Array)}function pi(t,e,n){var r="normalMerge"===n,i="replaceMerge"===n,o="replaceAll"===n;t=t||[],e=(e||[]).slice();var a=Object(s["f"])();Object(s["k"])(e,(function(t,n){Object(s["A"])(t)||(e[n]=null)}));var l=mi(t,a,n);return(r||i)&&gi(l,t,a,e),r&&yi(l,e),r||i?vi(l,e,i):o&&_i(l,e),bi(l),l}function mi(t,e,n){var r=[];if("replaceAll"===n)return r;for(var i=0;i<t.length;i++){var o=t[i];o&&null!=o.id&&e.set(o.id,i),r.push({existing:"replaceMerge"===n||ki(o)?null:o,newOption:null,keyInfo:null,brandNew:null})}return r}function gi(t,e,n,r){Object(s["k"])(r,(function(i,o){if(i&&null!=i.id){var a=wi(i.id),l=n.get(a);if(null!=l){var u=t[l];Object(s["b"])(!u.newOption,'Duplicated option on id "'+a+'".'),u.newOption=i,u.existing=e[l],r[o]=null}}}))}function yi(t,e){Object(s["k"])(e,(function(n,r){if(n&&null!=n.name)for(var i=0;i<t.length;i++){var o=t[i].existing;if(!t[i].newOption&&o&&(null==o.id||null==n.id)&&!ki(n)&&!ki(o)&&xi("name",o,n))return t[i].newOption=n,void(e[r]=null)}}))}function vi(t,e,n){Object(s["k"])(e,(function(e){if(e){var r,i=0;while((r=t[i])&&(r.newOption||ki(r.existing)||r.existing&&null!=e.id&&!xi("id",e,r.existing)))i++;r?(r.newOption=e,r.brandNew=n):t.push({newOption:e,brandNew:n,existing:null,keyInfo:null}),i++}}))}function _i(t,e){Object(s["k"])(e,(function(e){t.push({newOption:e,brandNew:!0,existing:null,keyInfo:null})}))}function bi(t){var e=Object(s["f"])();Object(s["k"])(t,(function(t){var n=t.existing;n&&e.set(n.id,t)})),Object(s["k"])(t,(function(t){var n=t.newOption;Object(s["b"])(!n||null==n.id||!e.get(n.id)||e.get(n.id)===t,"id duplicates: "+(n&&n.id)),n&&null!=n.id&&e.set(n.id,t),!t.keyInfo&&(t.keyInfo={})})),Object(s["k"])(t,(function(t,n){var r=t.existing,i=t.newOption,o=t.keyInfo;if(Object(s["A"])(i)){if(o.name=null!=i.name?wi(i.name):r?r.name:si+n,r)o.id=wi(r.id);else if(null!=i.id)o.id=wi(i.id);else{var a=0;do{o.id="\0"+o.name+"\0"+a++}while(e.get(o.id))}e.set(o.id,t)}}))}function xi(t,e,n){var r=Mi(e[t],null),i=Mi(n[t],null);return null!=r&&null!=i&&r===i}function wi(t){return Mi(t,"")}function Mi(t,e){return null==t?e:Object(s["C"])(t)?t:Object(s["z"])(t)||Object(s["D"])(t)?t+"":e}function Si(t){var e=t.name;return!(!e||!e.indexOf(si))}function ki(t){return t&&null!=t.id&&0===wi(t.id).indexOf(li)}function Li(t){return li+t}function Ti(t,e,n){Object(s["k"])(t,(function(t){var r=t.newOption;Object(s["A"])(r)&&(t.keyInfo.mainType=e,t.keyInfo.subType=Di(e,r,t.existing,n))}))}function Di(t,e,n,r){var i=e.type?e.type:n?n.subType:r.determineSubType(t,e);return i}function Oi(t,e){var n={},r={};return i(t||[],n),i(e||[],r,n),[o(n),o(r)];function i(t,e,n){for(var r=0,i=t.length;r<i;r++){var o=Mi(t[r].seriesId,null);if(null==o)return;for(var a=ui(t[r].dataIndex),s=n&&n[o],l=0,u=a.length;l<u;l++){var c=a[l];s&&s[c]?s[c]=null:(e[o]||(e[o]={}))[c]=1}}}function o(t,e){var n=[];for(var r in t)if(t.hasOwnProperty(r)&&null!=t[r])if(e)n.push(+r);else{var i=o(t[r],!0);i.length&&n.push({seriesId:r,dataIndex:i})}return n}}function Ii(t,e){return null!=e.dataIndexInside?e.dataIndexInside:null!=e.dataIndex?Object(s["t"])(e.dataIndex)?Object(s["H"])(e.dataIndex,(function(e){return t.indexOfRawIndex(e)})):t.indexOfRawIndex(e.dataIndex):null!=e.name?Object(s["t"])(e.name)?Object(s["H"])(e.name,(function(e){return t.indexOfName(e)})):t.indexOfName(e.name):void 0}function Ci(){var t="__ec_inner_"+Ai++;return function(e){return e[t]||(e[t]={})}}var Ai=ri();function Yi(t,e,n){var r=ji(e,n),i=r.mainTypeSpecified,o=r.queryOptionMap,a=r.others,s=a,l=n?n.defaultMainType:null;return!i&&l&&o.set(l,{}),o.each((function(e,r){var i=Ni(t,r,e,{useDefault:l===r,enableAll:!n||null==n.enableAll||n.enableAll,enableNone:!n||null==n.enableNone||n.enableNone});s[r+"Models"]=i.models,s[r+"Model"]=i.models[0]})),s}function ji(t,e){var n;if(Object(s["C"])(t)){var r={};r[t+"Index"]=0,n=r}else n=t;var i=Object(s["f"])(),o={},a=!1;return Object(s["k"])(n,(function(t,n){if("dataIndex"!==n&&"dataIndexInside"!==n){var r=n.match(/^(\w+)(Index|Id|Name)$/)||[],l=r[1],u=(r[2]||"").toLowerCase();if(l&&u&&!(e&&e.includeMainTypes&&Object(s["r"])(e.includeMainTypes,l)<0)){a=a||!!l;var c=i.get(l)||i.set(l,{});c[u]=t}}else o[n]=t})),{mainTypeSpecified:a,queryOptionMap:i,others:o}}var Pi={useDefault:!0,enableAll:!1,enableNone:!1},Ei={useDefault:!1,enableAll:!0,enableNone:!0};function Ni(t,e,n,r){r=r||Pi;var i=n.index,o=n.id,a=n.name,l={models:null,specified:null!=i||null!=o||null!=a};if(!l.specified){var u=void 0;return l.models=r.useDefault&&(u=t.getComponent(e))?[u]:[],l}return"none"===i||!1===i?(Object(s["b"])(r.enableNone,'`"none"` or `false` is not a valid value on index option.'),l.models=[],l):("all"===i&&(Object(s["b"])(r.enableAll,'`"all"` is not a valid value on index option.'),i=o=a=null),l.models=t.queryComponents({mainType:e,index:i,id:o,name:a}),l)}function Ri(t,e,n){t.setAttribute?t.setAttribute(e,n):t[e]=n}function Hi(t,e){return t.getAttribute?t.getAttribute(e):t[e]}function zi(t){return"auto"===t?a["a"].domSupported?"html":"richText":t||"html"}function Fi(t,e){var n=Object(s["f"])(),r=[];return Object(s["k"])(t,(function(t){var i=e(t);(n.get(i)||(r.push(i),n.set(i,[]))).push(t)})),{keys:r,buckets:n}}function Bi(t,e,n,r,i){var o=null==e||"auto"===e;if(null==r)return r;if(Object(s["z"])(r)){var a=ai(n||0,r,i);return Rr(a,o?Math.max(zr(n||0),zr(r)):e)}if(Object(s["C"])(r))return i<1?n:r;for(var l=[],u=n,c=r,d=Math.max(u?u.length:0,c.length),h=0;h<d;++h){var f=t.getDimensionInfo(h);if(f&&"ordinal"===f.type)l[h]=(i<1&&u?u:c)[h];else{var p=u&&u[h]?u[h]:0,m=c[h];a=ai(p,m,i);l[h]=Rr(a,o?Math.max(zr(p),zr(m)):e)}}return l}var Wi=".",Vi="___EC__COMPONENT__CONTAINER___",Gi="___EC__EXTENDED_CLASS___";function $i(t){var e={main:"",sub:""};if(t){var n=t.split(Wi);e.main=n[0]||"",e.sub=n[1]||""}return e}function Ui(t){s["b"](/^[a-zA-Z0-9_]+([.][a-zA-Z0-9_]+)?$/.test(t),'componentType "'+t+'" illegal')}function qi(t){return!(!t||!t[Gi])}function Xi(t,e){t.$constructor=t,t.extend=function(t){var e,n=this;return Zi(n)?e=function(t){function e(){return t.apply(this,arguments)||this}return o(e,t),e}(n):(e=function(){(t.$constructor||n).apply(this,arguments)},s["s"](e,this)),s["m"](e.prototype,t),e[Gi]=!0,e.extend=this.extend,e.superCall=to,e.superApply=eo,e.superClass=n,e}}function Zi(t){return s["w"](t)&&/^class\s/.test(Function.prototype.toString.call(t))}function Ji(t,e){t.extend=e.extend}var Ki=Math.round(10*Math.random());function Qi(t){var e=["__\0is_clz",Ki++].join("_");t.prototype[e]=!0,t.isInstance=function(t){return!(!t||!t[e])}}function to(t,e){for(var n=[],r=2;r<arguments.length;r++)n[r-2]=arguments[r];return this.superClass.prototype[e].apply(t,n)}function eo(t,e,n){return this.superClass.prototype[e].apply(t,n)}function no(t){var e={};function n(t){var n=e[t.main];return n&&n[Vi]||(n=e[t.main]={},n[Vi]=!0),n}t.registerClass=function(t){var r=t.type||t.prototype.type;if(r){Ui(r),t.prototype.type=r;var i=$i(r);if(i.sub){if(i.sub!==Vi){var o=n(i);o[i.sub]=t}}else e[i.main]=t}return t},t.getClass=function(t,n,r){var i=e[t];if(i&&i[Vi]&&(i=n?i[n]:null),r&&!i)throw new Error(n?"Component "+t+"."+(n||"")+" is used but not imported.":t+".type should be specified.");return i},t.getClassesByMainType=function(t){var n=$i(t),r=[],i=e[n.main];return i&&i[Vi]?s["k"](i,(function(t,e){e!==Vi&&r.push(t)})):r.push(i),r},t.hasClass=function(t){var n=$i(t);return!!e[n.main]},t.getAllClassMainTypes=function(){var t=[];return s["k"](e,(function(e,n){t.push(n)})),t},t.hasSubTypes=function(t){var n=$i(t),r=e[n.main];return r&&r[Vi]}}function ro(t,e){for(var n=0;n<t.length;n++)t[n][1]||(t[n][1]=t[n][0]);return e=e||!1,function(n,r,i){for(var o={},a=0;a<t.length;a++){var l=t[a][1];if(!(r&&s["r"](r,l)>=0||i&&s["r"](i,l)<0)){var u=n.getShallow(l,e);null!=u&&(o[t[a][0]]=u)}}return o}}var io=[["fill","color"],["shadowBlur"],["shadowOffsetX"],["shadowOffsetY"],["opacity"],["shadowColor"]],oo=ro(io),ao=function(){function t(){}return t.prototype.getAreaStyle=function(t,e){return oo(this,t,e)},t}(),so=new Zn["a"](50);function lo(t){if("string"===typeof t){var e=so.get(t);return e&&e.image}return t}function uo(t,e,n,r,i){if(t){if("string"===typeof t){if(e&&e.__zrImageSrc===t||!n)return e;var o=so.get(t),a={hostEl:n,cb:r,cbPayload:i};return o?(e=o.image,!ho(e)&&o.pending.push(a)):(e=Jn["d"].loadImage(t,co,co),e.__zrImageSrc=t,so.put(t,e.__cachedImgObj={image:e,pending:[a]})),e}return t}return e}function co(){var t=this.__cachedImgObj;this.onload=this.onerror=this.__cachedImgObj=null;for(var e=0;e<t.pending.length;e++){var n=t.pending[e],r=n.cb;r&&r(this,n.cbPayload),n.hostEl.dirty()}t.pending.length=0}function ho(t){return t&&t.width&&t.height}var fo=/\{([a-zA-Z0-9_]+)\|([^}]*)\}/g;function po(t,e,n,r,i){if(!e)return"";var o=(t+"").split("\n");i=mo(e,n,r,i);for(var a=0,s=o.length;a<s;a++)o[a]=go(o[a],i);return o.join("\n")}function mo(t,e,n,r){r=r||{};var i=Object(s["m"])({},r);i.font=e,n=Object(s["P"])(n,"..."),i.maxIterations=Object(s["P"])(r.maxIterations,2);var o=i.minChar=Object(s["P"])(r.minChar,0);i.cnCharWidth=Qn("国",e);var a=i.ascCharWidth=Qn("a",e);i.placeholder=Object(s["P"])(r.placeholder,"");for(var l=t=Math.max(0,t-1),u=0;u<o&&l>=a;u++)l-=a;var c=Qn(n,e);return c>l&&(n="",c=0),l=t-c,i.ellipsis=n,i.ellipsisWidth=c,i.contentWidth=l,i.containerWidth=t,i}function go(t,e){var n=e.containerWidth,r=e.font,i=e.contentWidth;if(!n)return"";var o=Qn(t,r);if(o<=n)return t;for(var a=0;;a++){if(o<=i||a>=e.maxIterations){t+=e.ellipsis;break}var s=0===a?yo(t,i,e.ascCharWidth,e.cnCharWidth):o>0?Math.floor(t.length*i/o):0;t=t.substr(0,s),o=Qn(t,r)}return""===t&&(t=e.placeholder),t}function yo(t,e,n,r){for(var i=0,o=0,a=t.length;o<a&&i<e;o++){var s=t.charCodeAt(o);i+=0<=s&&s<=127?n:r}return o}function vo(t,e){null!=t&&(t+="");var n,r=e.overflow,i=e.padding,o=e.font,a="truncate"===r,l=ir(o),u=Object(s["P"])(e.lineHeight,l),c=!!e.backgroundColor,d="truncate"===e.lineOverflow,h=e.width;n=null==h||"break"!==r&&"breakAll"!==r?t?t.split("\n"):[]:t?To(t,e.font,h,"breakAll"===r,0).lines:[];var f=n.length*u,p=Object(s["P"])(e.height,f);if(f>p&&d){var m=Math.floor(p/u);n=n.slice(0,m)}if(t&&a&&null!=h)for(var g=mo(h,o,e.ellipsis,{minChar:e.truncateMinChar,placeholder:e.placeholder}),y=0;y<n.length;y++)n[y]=go(n[y],g);var v=p,_=0;for(y=0;y<n.length;y++)_=Math.max(Qn(n[y],o),_);null==h&&(h=_);var b=_;return i&&(v+=i[0]+i[2],b+=i[1]+i[3],h+=i[1]+i[3]),c&&(b=h),{lines:n,height:p,outerWidth:b,outerHeight:v,lineHeight:u,calculatedLineHeight:l,contentWidth:_,contentHeight:f,width:h}}var _o=function(){function t(){}return t}(),bo=function(){function t(t){this.tokens=[],t&&(this.tokens=t)}return t}(),xo=function(){function t(){this.width=0,this.height=0,this.contentWidth=0,this.contentHeight=0,this.outerWidth=0,this.outerHeight=0,this.lines=[]}return t}();function wo(t,e){var n=new xo;if(null!=t&&(t+=""),!t)return n;var r,i=e.width,o=e.height,a=e.overflow,l="break"!==a&&"breakAll"!==a||null==i?null:{width:i,accumWidth:0,breakAll:"breakAll"===a},u=fo.lastIndex=0;while(null!=(r=fo.exec(t))){var c=r.index;c>u&&Mo(n,t.substring(u,c),e,l),Mo(n,r[2],e,l,r[1]),u=fo.lastIndex}u<t.length&&Mo(n,t.substring(u,t.length),e,l);var d=[],h=0,f=0,p=e.padding,m="truncate"===a,g="truncate"===e.lineOverflow;function y(t,e,n){t.width=e,t.lineHeight=n,h+=n,f=Math.max(f,e)}t:for(var v=0;v<n.lines.length;v++){for(var _=n.lines[v],b=0,x=0,w=0;w<_.tokens.length;w++){var M=_.tokens[w],S=M.styleName&&e.rich[M.styleName]||{},k=M.textPadding=S.padding,L=k?k[1]+k[3]:0,T=M.font=S.font||e.font;M.contentHeight=ir(T);var D=Object(s["P"])(S.height,M.contentHeight);if(M.innerHeight=D,k&&(D+=k[0]+k[2]),M.height=D,M.lineHeight=Object(s["Q"])(S.lineHeight,e.lineHeight,D),M.align=S&&S.align||e.align,M.verticalAlign=S&&S.verticalAlign||"middle",g&&null!=o&&h+M.lineHeight>o){w>0?(_.tokens=_.tokens.slice(0,w),y(_,x,b),n.lines=n.lines.slice(0,v+1)):n.lines=n.lines.slice(0,v);break t}var O=S.width,I=null==O||"auto"===O;if("string"===typeof O&&"%"===O.charAt(O.length-1))M.percentWidth=O,d.push(M),M.contentWidth=Qn(M.text,T);else{if(I){var C=S.backgroundColor,A=C&&C.image;A&&(A=lo(A),ho(A)&&(M.width=Math.max(M.width,A.width*D/A.height)))}var Y=m&&null!=i?i-x:null;null!=Y&&Y<M.width?!I||Y<L?(M.text="",M.width=M.contentWidth=0):(M.text=po(M.text,Y-L,T,e.ellipsis,{minChar:e.truncateMinChar}),M.width=M.contentWidth=Qn(M.text,T)):M.contentWidth=Qn(M.text,T)}M.width+=L,x+=M.width,S&&(b=Math.max(b,M.lineHeight))}y(_,x,b)}n.outerWidth=n.width=Object(s["P"])(i,f),n.outerHeight=n.height=Object(s["P"])(o,h),n.contentHeight=h,n.contentWidth=f,p&&(n.outerWidth+=p[1]+p[3],n.outerHeight+=p[0]+p[2]);for(v=0;v<d.length;v++){M=d[v];var j=M.percentWidth;M.width=parseInt(j,10)/100*n.width}return n}function Mo(t,e,n,r,i){var o,a,s=""===e,l=i&&n.rich[i]||{},u=t.lines,c=l.font||n.font,d=!1;if(r){var h=l.padding,f=h?h[1]+h[3]:0;if(null!=l.width&&"auto"!==l.width){var p=or(l.width,r.width)+f;u.length>0&&p+r.accumWidth>r.width&&(o=e.split("\n"),d=!0),r.accumWidth=p}else{var m=To(e,c,r.width,r.breakAll,r.accumWidth);r.accumWidth=m.accumWidth+f,a=m.linesWidths,o=m.lines}}else o=e.split("\n");for(var g=0;g<o.length;g++){var y=o[g],v=new _o;if(v.styleName=i,v.text=y,v.isLineHolder=!y&&!s,"number"===typeof l.width?v.width=l.width:v.width=a?a[g]:Qn(y,c),g||d)u.push(new bo([v]));else{var _=(u[u.length-1]||(u[0]=new bo)).tokens,b=_.length;1===b&&_[0].isLineHolder?_[0]=v:(y||!b||s)&&_.push(v)}}}function So(t){var e=t.charCodeAt(0);return e>=32&&e<=591||e>=880&&e<=4351||e>=4608&&e<=5119||e>=7680&&e<=8303}var ko=Object(s["N"])(",&?/;] ".split(""),(function(t,e){return t[e]=!0,t}),{});function Lo(t){return!So(t)||!!ko[t]}function To(t,e,n,r,i){for(var o=[],a=[],s="",l="",u=0,c=0,d=0;d<t.length;d++){var h=t.charAt(d);if("\n"!==h){var f=Qn(h,e),p=!r&&!Lo(h);(o.length?c+f>n:i+c+f>n)?c?(s||l)&&(p?(s||(s=l,l="",u=0,c=u),o.push(s),a.push(c-u),l+=h,u+=f,s="",c=u):(l&&(s+=l,l="",u=0),o.push(s),a.push(c),s=h,c=f)):p?(o.push(l),a.push(u),l=h,u=f):(o.push(h),a.push(f)):(c+=f,p?(l+=h,u+=f):(l&&(s+=l,l="",u=0),s+=h))}else l&&(s+=l,c+=u),o.push(s),a.push(c),s="",l="",u=0,c=0}return o.length||s||(s=t,l="",u=0),l&&(s+=l),s&&(o.push(s),a.push(c)),1===o.length&&(c+=i),{accumWidth:c,lines:o,linesWidths:a}}var Do="__zr_style_"+Math.round(10*Math.random()),Oo={shadowBlur:0,shadowOffsetX:0,shadowOffsetY:0,shadowColor:"#000",opacity:1,blend:"source-over"},Io={style:{shadowBlur:!0,shadowOffsetX:!0,shadowOffsetY:!0,shadowColor:!0,opacity:!0}};Oo[Do]=!0;var Co=["z","z2","invisible"],Ao=["invisible"],Yo=function(t){function e(e){return t.call(this,e)||this}return o(e,t),e.prototype._init=function(e){for(var n=Object(s["F"])(e),r=0;r<n.length;r++){var i=n[r];"style"===i?this.useStyle(e[i]):t.prototype.attrKV.call(this,i,e[i])}this.style||this.useStyle({})},e.prototype.beforeBrush=function(){},e.prototype.afterBrush=function(){},e.prototype.innerBeforeBrush=function(){},e.prototype.innerAfterBrush=function(){},e.prototype.shouldBePainted=function(t,e,n,r){var i=this.transform;if(this.ignore||this.invisible||0===this.style.opacity||this.culling&&Eo(this,t,e)||i&&!i[0]&&!i[3])return!1;if(n&&this.__clipPaths)for(var o=0;o<this.__clipPaths.length;++o)if(this.__clipPaths[o].isZeroArea())return!1;if(r&&this.parent){var a=this.parent;while(a){if(a.ignore)return!1;a=a.parent}}return!0},e.prototype.contain=function(t,e){return this.rectContain(t,e)},e.prototype.traverse=function(t,e){t.call(e,this)},e.prototype.rectContain=function(t,e){var n=this.transformCoordToLocal(t,e),r=this.getBoundingRect();return r.contain(n[0],n[1])},e.prototype.getPaintRect=function(){var t=this._paintRect;if(!this._paintRect||this.__dirty){var e=this.transform,n=this.getBoundingRect(),r=this.style,i=r.shadowBlur||0,o=r.shadowOffsetX||0,a=r.shadowOffsetY||0;t=this._paintRect||(this._paintRect=new Tt(0,0,0,0)),e?Tt.applyTransform(t,n,e):t.copy(n),(i||o||a)&&(t.width+=2*i+Math.abs(o),t.height+=2*i+Math.abs(a),t.x=Math.min(t.x,t.x+o-i),t.y=Math.min(t.y,t.y+a-i));var s=this.dirtyRectTolerance;t.isZero()||(t.x=Math.floor(t.x-s),t.y=Math.floor(t.y-s),t.width=Math.ceil(t.width+1+2*s),t.height=Math.ceil(t.height+1+2*s))}return t},e.prototype.setPrevPaintRect=function(t){t?(this._prevPaintRect=this._prevPaintRect||new Tt(0,0,0,0),this._prevPaintRect.copy(t)):this._prevPaintRect=null},e.prototype.getPrevPaintRect=function(){return this._prevPaintRect},e.prototype.animateStyle=function(t){return this.animate("style",t)},e.prototype.updateDuringAnimation=function(t){"style"===t?this.dirtyStyle():this.markRedraw()},e.prototype.attrKV=function(e,n){"style"!==e?t.prototype.attrKV.call(this,e,n):this.style?this.setStyle(n):this.useStyle(n)},e.prototype.setStyle=function(t,e){return"string"===typeof t?this.style[t]=e:Object(s["m"])(this.style,t),this.dirtyStyle(),this},e.prototype.dirtyStyle=function(t){t||this.markRedraw(),this.__dirty|=Jt,this._rect&&(this._rect=null)},e.prototype.dirty=function(){this.dirtyStyle()},e.prototype.styleChanged=function(){return!!(this.__dirty&Jt)},e.prototype.styleUpdated=function(){this.__dirty&=~Jt},e.prototype.createStyle=function(t){return Object(s["g"])(Oo,t)},e.prototype.useStyle=function(t){t[Do]||(t=this.createStyle(t)),this.__inHover?this.__hoverStyle=t:this.style=t,this.dirtyStyle()},e.prototype.isStyleObject=function(t){return t[Do]},e.prototype._innerSaveToNormal=function(e){t.prototype._innerSaveToNormal.call(this,e);var n=this._normalState;e.style&&!n.style&&(n.style=this._mergeStyle(this.createStyle(),this.style)),this._savePrimaryToNormal(e,n,Co)},e.prototype._applyStateObj=function(e,n,r,i,o,a){t.prototype._applyStateObj.call(this,e,n,r,i,o,a);var l,u=!(n&&i);if(n&&n.style?o?i?l=n.style:(l=this._mergeStyle(this.createStyle(),r.style),this._mergeStyle(l,n.style)):(l=this._mergeStyle(this.createStyle(),i?this.style:r.style),this._mergeStyle(l,n.style)):u&&(l=r.style),l)if(o){var c=this.style;if(this.style=this.createStyle(u?{}:c),u)for(var d=Object(s["F"])(c),h=0;h<d.length;h++){var f=d[h];f in l&&(l[f]=l[f],this.style[f]=c[f])}var p=Object(s["F"])(l);for(h=0;h<p.length;h++){f=p[h];this.style[f]=this.style[f]}this._transitionState(e,{style:l},a,this.getAnimationStyleProps())}else this.useStyle(l);var m=this.__inHover?Ao:Co;for(h=0;h<m.length;h++){f=m[h];n&&null!=n[f]?this[f]=n[f]:u&&null!=r[f]&&(this[f]=r[f])}},e.prototype._mergeStates=function(e){for(var n,r=t.prototype._mergeStates.call(this,e),i=0;i<e.length;i++){var o=e[i];o.style&&(n=n||{},this._mergeStyle(n,o.style))}return n&&(r.style=n),r},e.prototype._mergeStyle=function(t,e){return Object(s["m"])(t,e),t},e.prototype.getAnimationStyleProps=function(){return Io},e.initDefaultProps=function(){var t=e.prototype;t.type="displayable",t.invisible=!1,t.z=0,t.z2=0,t.zlevel=0,t.culling=!1,t.cursor="pointer",t.rectHover=!1,t.incremental=!1,t._rect=null,t.dirtyRectTolerance=0,t.__dirty=Zt|Jt}(),e}(br),jo=new Tt(0,0,0,0),Po=new Tt(0,0,0,0);function Eo(t,e,n){return jo.copy(t.getBoundingRect()),t.transform&&jo.applyTransform(t.transform),Po.width=e,Po.height=n,!jo.intersect(Po)}var No=Yo,Ro=Math.min,Ho=Math.max,zo=Math.sin,Fo=Math.cos,Bo=2*Math.PI,Wo=l(),Vo=l(),Go=l();function $o(t,e,n){if(0!==t.length){for(var r=t[0],i=r[0],o=r[0],a=r[1],s=r[1],l=1;l<t.length;l++)r=t[l],i=Ro(i,r[0]),o=Ho(o,r[0]),a=Ro(a,r[1]),s=Ho(s,r[1]);e[0]=i,e[1]=a,n[0]=o,n[1]=s}}function Uo(t,e,n,r,i,o){i[0]=Ro(t,n),i[1]=Ro(e,r),o[0]=Ho(t,n),o[1]=Ho(e,r)}var qo=[],Xo=[];function Zo(t,e,n,r,i,o,a,s,l,u){var c=we,d=_e,h=c(t,n,i,a,qo);l[0]=1/0,l[1]=1/0,u[0]=-1/0,u[1]=-1/0;for(var f=0;f<h;f++){var p=d(t,n,i,a,qo[f]);l[0]=Ro(p,l[0]),u[0]=Ho(p,u[0])}h=c(e,r,o,s,Xo);for(f=0;f<h;f++){var m=d(e,r,o,s,Xo[f]);l[1]=Ro(m,l[1]),u[1]=Ho(m,u[1])}l[0]=Ro(t,l[0]),u[0]=Ho(t,u[0]),l[0]=Ro(a,l[0]),u[0]=Ho(a,u[0]),l[1]=Ro(e,l[1]),u[1]=Ho(e,u[1]),l[1]=Ro(s,l[1]),u[1]=Ho(s,u[1])}function Jo(t,e,n,r,i,o,a,s){var l=Oe,u=Le,c=Ho(Ro(l(t,n,i),1),0),d=Ho(Ro(l(e,r,o),1),0),h=u(t,n,i,c),f=u(e,r,o,d);a[0]=Ro(t,i,h),a[1]=Ro(e,o,f),s[0]=Ho(t,i,h),s[1]=Ho(e,o,f)}function Ko(t,e,n,r,i,o,a,s,l){var u=k,c=L,d=Math.abs(i-o);if(d%Bo<1e-4&&d>1e-4)return s[0]=t-n,s[1]=e-r,l[0]=t+n,void(l[1]=e+r);if(Wo[0]=Fo(i)*n+t,Wo[1]=zo(i)*r+e,Vo[0]=Fo(o)*n+t,Vo[1]=zo(o)*r+e,u(s,Wo,Vo),c(l,Wo,Vo),i%=Bo,i<0&&(i+=Bo),o%=Bo,o<0&&(o+=Bo),i>o&&!a?o+=Bo:i<o&&a&&(i+=Bo),a){var h=o;o=i,i=h}for(var f=0;f<o;f+=Math.PI/2)f>i&&(Go[0]=Fo(f)*n+t,Go[1]=zo(f)*r+e,u(s,Go,s),c(l,Go,l))}var Qo={M:1,L:2,C:3,Q:4,A:5,Z:6,R:7},ta=[],ea=[],na=[],ra=[],ia=[],oa=[],aa=Math.min,sa=Math.max,la=Math.cos,ua=Math.sin,ca=Math.abs,da=Math.PI,ha=2*da,fa="undefined"!==typeof Float32Array,pa=[];function ma(t){var e=Math.round(t/da*1e8)/1e8;return e%2*da}function ga(t,e){var n=ma(t[0]);n<0&&(n+=ha);var r=n-t[0],i=t[1];i+=r,!e&&i-n>=ha?i=n+ha:e&&n-i>=ha?i=n-ha:!e&&n>i?i=n+(ha-ma(n-i)):e&&n<i&&(i=n-(ha-ma(i-n))),t[0]=n,t[1]=i}var ya=function(){function t(t){this.dpr=1,this._xi=0,this._yi=0,this._x0=0,this._y0=0,this._len=0,t&&(this._saveData=!1),this._saveData&&(this.data=[])}return t.prototype.increaseVersion=function(){this._version++},t.prototype.getVersion=function(){return this._version},t.prototype.setScale=function(t,e,n){n=n||0,n>0&&(this._ux=ca(n/jn/t)||0,this._uy=ca(n/jn/e)||0)},t.prototype.setDPR=function(t){this.dpr=t},t.prototype.setContext=function(t){this._ctx=t},t.prototype.getContext=function(){return this._ctx},t.prototype.beginPath=function(){return this._ctx&&this._ctx.beginPath(),this.reset(),this},t.prototype.reset=function(){this._saveData&&(this._len=0),this._pathSegLen&&(this._pathSegLen=null,this._pathLen=0),this._version++},t.prototype.moveTo=function(t,e){return this._drawPendingPt(),this.addData(Qo.M,t,e),this._ctx&&this._ctx.moveTo(t,e),this._x0=t,this._y0=e,this._xi=t,this._yi=e,this},t.prototype.lineTo=function(t,e){var n=ca(t-this._xi),r=ca(e-this._yi),i=n>this._ux||r>this._uy;if(this.addData(Qo.L,t,e),this._ctx&&i&&this._ctx.lineTo(t,e),i)this._xi=t,this._yi=e,this._pendingPtDist=0;else{var o=n*n+r*r;o>this._pendingPtDist&&(this._pendingPtX=t,this._pendingPtY=e,this._pendingPtDist=o)}return this},t.prototype.bezierCurveTo=function(t,e,n,r,i,o){return this._drawPendingPt(),this.addData(Qo.C,t,e,n,r,i,o),this._ctx&&this._ctx.bezierCurveTo(t,e,n,r,i,o),this._xi=i,this._yi=o,this},t.prototype.quadraticCurveTo=function(t,e,n,r){return this._drawPendingPt(),this.addData(Qo.Q,t,e,n,r),this._ctx&&this._ctx.quadraticCurveTo(t,e,n,r),this._xi=n,this._yi=r,this},t.prototype.arc=function(t,e,n,r,i,o){this._drawPendingPt(),pa[0]=r,pa[1]=i,ga(pa,o),r=pa[0],i=pa[1];var a=i-r;return this.addData(Qo.A,t,e,n,n,r,a,0,o?0:1),this._ctx&&this._ctx.arc(t,e,n,r,i,o),this._xi=la(i)*n+t,this._yi=ua(i)*n+e,this},t.prototype.arcTo=function(t,e,n,r,i){return this._drawPendingPt(),this._ctx&&this._ctx.arcTo(t,e,n,r,i),this},t.prototype.rect=function(t,e,n,r){return this._drawPendingPt(),this._ctx&&this._ctx.rect(t,e,n,r),this.addData(Qo.R,t,e,n,r),this},t.prototype.closePath=function(){this._drawPendingPt(),this.addData(Qo.Z);var t=this._ctx,e=this._x0,n=this._y0;return t&&t.closePath(),this._xi=e,this._yi=n,this},t.prototype.fill=function(t){t&&t.fill(),this.toStatic()},t.prototype.stroke=function(t){t&&t.stroke(),this.toStatic()},t.prototype.len=function(){return this._len},t.prototype.setData=function(t){var e=t.length;this.data&&this.data.length===e||!fa||(this.data=new Float32Array(e));for(var n=0;n<e;n++)this.data[n]=t[n];this._len=e},t.prototype.appendPath=function(t){t instanceof Array||(t=[t]);for(var e=t.length,n=0,r=this._len,i=0;i<e;i++)n+=t[i].len();fa&&this.data instanceof Float32Array&&(this.data=new Float32Array(r+n));for(i=0;i<e;i++)for(var o=t[i].data,a=0;a<o.length;a++)this.data[r++]=o[a];this._len=r},t.prototype.addData=function(t,e,n,r,i,o,a,s,l){if(this._saveData){var u=this.data;this._len+arguments.length>u.length&&(this._expandData(),u=this.data);for(var c=0;c<arguments.length;c++)u[this._len++]=arguments[c]}},t.prototype._drawPendingPt=function(){this._pendingPtDist>0&&(this._ctx&&this._ctx.lineTo(this._pendingPtX,this._pendingPtY),this._pendingPtDist=0)},t.prototype._expandData=function(){if(!(this.data instanceof Array)){for(var t=[],e=0;e<this._len;e++)t[e]=this.data[e];this.data=t}},t.prototype.toStatic=function(){if(this._saveData){this._drawPendingPt();var t=this.data;t instanceof Array&&(t.length=this._len,fa&&this._len>11&&(this.data=new Float32Array(t)))}},t.prototype.getBoundingRect=function(){na[0]=na[1]=ia[0]=ia[1]=Number.MAX_VALUE,ra[0]=ra[1]=oa[0]=oa[1]=-Number.MAX_VALUE;var t,e=this.data,n=0,r=0,i=0,o=0;for(t=0;t<this._len;){var a=e[t++],s=1===t;switch(s&&(n=e[t],r=e[t+1],i=n,o=r),a){case Qo.M:n=i=e[t++],r=o=e[t++],ia[0]=i,ia[1]=o,oa[0]=i,oa[1]=o;break;case Qo.L:Uo(n,r,e[t],e[t+1],ia,oa),n=e[t++],r=e[t++];break;case Qo.C:Zo(n,r,e[t++],e[t++],e[t++],e[t++],e[t],e[t+1],ia,oa),n=e[t++],r=e[t++];break;case Qo.Q:Jo(n,r,e[t++],e[t++],e[t],e[t+1],ia,oa),n=e[t++],r=e[t++];break;case Qo.A:var l=e[t++],u=e[t++],c=e[t++],d=e[t++],h=e[t++],f=e[t++]+h;t+=1;var p=!e[t++];s&&(i=la(h)*c+l,o=ua(h)*d+u),Ko(l,u,c,d,h,f,p,ia,oa),n=la(f)*c+l,r=ua(f)*d+u;break;case Qo.R:i=n=e[t++],o=r=e[t++];var m=e[t++],g=e[t++];Uo(i,o,i+m,o+g,ia,oa);break;case Qo.Z:n=i,r=o;break}k(na,na,ia),L(ra,ra,oa)}return 0===t&&(na[0]=na[1]=ra[0]=ra[1]=0),new Tt(na[0],na[1],ra[0]-na[0],ra[1]-na[1])},t.prototype._calculateLength=function(){var t=this.data,e=this._len,n=this._ux,r=this._uy,i=0,o=0,a=0,s=0;this._pathSegLen||(this._pathSegLen=[]);for(var l=this._pathSegLen,u=0,c=0,d=0;d<e;){var h=t[d++],f=1===d;f&&(i=t[d],o=t[d+1],a=i,s=o);var p=-1;switch(h){case Qo.M:i=a=t[d++],o=s=t[d++];break;case Qo.L:var m=t[d++],g=t[d++],y=m-i,v=g-o;(ca(y)>n||ca(v)>r||d===e-1)&&(p=Math.sqrt(y*y+v*v),i=m,o=g);break;case Qo.C:var _=t[d++],b=t[d++],x=(m=t[d++],g=t[d++],t[d++]),w=t[d++];p=ke(i,o,_,b,m,g,x,w,10),i=x,o=w;break;case Qo.Q:_=t[d++],b=t[d++],m=t[d++],g=t[d++];p=Ae(i,o,_,b,m,g,10),i=m,o=g;break;case Qo.A:var M=t[d++],S=t[d++],k=t[d++],L=t[d++],T=t[d++],D=t[d++],O=D+T;d+=1,f&&(a=la(T)*k+M,s=ua(T)*L+S),p=sa(k,L)*aa(ha,Math.abs(D)),i=la(O)*k+M,o=ua(O)*L+S;break;case Qo.R:a=i=t[d++],s=o=t[d++];var I=t[d++],C=t[d++];p=2*I+2*C;break;case Qo.Z:y=a-i,v=s-o;p=Math.sqrt(y*y+v*v),i=a,o=s;break}p>=0&&(l[c++]=p,u+=p)}return this._pathLen=u,u},t.prototype.rebuildPath=function(t,e){var n,r,i,o,a,s,l,u,c,d,h,f=this.data,p=this._ux,m=this._uy,g=this._len,y=e<1,v=0,_=0,b=0;if(!y||(this._pathSegLen||this._calculateLength(),l=this._pathSegLen,u=this._pathLen,c=e*u,c))t:for(var x=0;x<g;){var w=f[x++],M=1===x;switch(M&&(i=f[x],o=f[x+1],n=i,r=o),w!==Qo.L&&b>0&&(t.lineTo(d,h),b=0),w){case Qo.M:n=i=f[x++],r=o=f[x++],t.moveTo(i,o);break;case Qo.L:a=f[x++],s=f[x++];var S=ca(a-i),k=ca(s-o);if(S>p||k>m){if(y){var L=l[_++];if(v+L>c){var T=(c-v)/L;t.lineTo(i*(1-T)+a*T,o*(1-T)+s*T);break t}v+=L}t.lineTo(a,s),i=a,o=s,b=0}else{var D=S*S+k*k;D>b&&(d=a,h=s,b=D)}break;case Qo.C:var O=f[x++],I=f[x++],C=f[x++],A=f[x++],Y=f[x++],j=f[x++];if(y){L=l[_++];if(v+L>c){T=(c-v)/L;Me(i,O,C,Y,T,ta),Me(o,I,A,j,T,ea),t.bezierCurveTo(ta[1],ea[1],ta[2],ea[2],ta[3],ea[3]);break t}v+=L}t.bezierCurveTo(O,I,C,A,Y,j),i=Y,o=j;break;case Qo.Q:O=f[x++],I=f[x++],C=f[x++],A=f[x++];if(y){L=l[_++];if(v+L>c){T=(c-v)/L;Ie(i,O,C,T,ta),Ie(o,I,A,T,ea),t.quadraticCurveTo(ta[1],ea[1],ta[2],ea[2]);break t}v+=L}t.quadraticCurveTo(O,I,C,A),i=C,o=A;break;case Qo.A:var P=f[x++],E=f[x++],N=f[x++],R=f[x++],H=f[x++],z=f[x++],F=f[x++],B=!f[x++],W=N>R?N:R,V=ca(N-R)>.001,G=H+z,$=!1;if(y){L=l[_++];v+L>c&&(G=H+z*(c-v)/L,$=!0),v+=L}if(V&&t.ellipse?t.ellipse(P,E,N,R,F,H,G,B):t.arc(P,E,W,H,G,B),$)break t;M&&(n=la(H)*N+P,r=ua(H)*R+E),i=la(G)*N+P,o=ua(G)*R+E;break;case Qo.R:n=i=f[x],r=o=f[x+1],a=f[x++],s=f[x++];var U=f[x++],q=f[x++];if(y){L=l[_++];if(v+L>c){var X=c-v;t.moveTo(a,s),t.lineTo(a+aa(X,U),s),X-=U,X>0&&t.lineTo(a+U,s+aa(X,q)),X-=q,X>0&&t.lineTo(a+sa(U-X,0),s+q),X-=U,X>0&&t.lineTo(a,s+sa(q-X,0));break t}v+=L}t.rect(a,s,U,q);break;case Qo.Z:if(y){L=l[_++];if(v+L>c){T=(c-v)/L;t.lineTo(i*(1-T)+n*T,o*(1-T)+r*T);break t}v+=L}t.closePath(),i=n,o=r}}},t.prototype.clone=function(){var e=new t,n=this.data;return e.data=n.slice?n.slice():Array.prototype.slice.call(n),e._len=this._len,e},t.CMD=Qo,t.initDefaultProps=function(){var e=t.prototype;e._saveData=!0,e._ux=0,e._uy=0,e._pendingPtDist=0,e._version=0}(),t}(),va=ya;function _a(t,e,n,r,i,o,a){if(0===i)return!1;var s=i,l=0,u=t;if(a>e+s&&a>r+s||a<e-s&&a<r-s||o>t+s&&o>n+s||o<t-s&&o<n-s)return!1;if(t===n)return Math.abs(o-t)<=s/2;l=(e-r)/(t-n),u=(t*r-n*e)/(t-n);var c=l*o-a+u,d=c*c/(l*l+1);return d<=s/2*s/2}function ba(t,e,n,r,i,o,a,s,l,u,c){if(0===l)return!1;var d=l;if(c>e+d&&c>r+d&&c>o+d&&c>s+d||c<e-d&&c<r-d&&c<o-d&&c<s-d||u>t+d&&u>n+d&&u>i+d&&u>a+d||u<t-d&&u<n-d&&u<i-d&&u<a-d)return!1;var h=Se(t,e,n,r,i,o,a,s,u,c,null);return h<=d/2}function xa(t,e,n,r,i,o,a,s,l){if(0===a)return!1;var u=a;if(l>e+u&&l>r+u&&l>o+u||l<e-u&&l<r-u&&l<o-u||s>t+u&&s>n+u&&s>i+u||s<t-u&&s<n-u&&s<i-u)return!1;var c=Ce(t,e,n,r,i,o,s,l,null);return c<=u/2}var wa=2*Math.PI;function Ma(t){return t%=wa,t<0&&(t+=wa),t}var Sa=2*Math.PI;function ka(t,e,n,r,i,o,a,s,l){if(0===a)return!1;var u=a;s-=t,l-=e;var c=Math.sqrt(s*s+l*l);if(c-u>n||c+u<n)return!1;if(Math.abs(r-i)%Sa<1e-4)return!0;if(o){var d=r;r=Ma(i),i=Ma(d)}else r=Ma(r),i=Ma(i);r>i&&(i+=Sa);var h=Math.atan2(l,s);return h<0&&(h+=Sa),h>=r&&h<=i||h+Sa>=r&&h+Sa<=i}function La(t,e,n,r,i,o){if(o>e&&o>r||o<e&&o<r)return 0;if(r===e)return 0;var a=(o-e)/(r-e),s=r<e?1:-1;1!==a&&0!==a||(s=r<e?.5:-.5);var l=a*(n-t)+t;return l===i?1/0:l>i?s:0}var Ta=va.CMD,Da=2*Math.PI,Oa=1e-4;function Ia(t,e){return Math.abs(t-e)<Oa}var Ca=[-1,-1,-1],Aa=[-1,-1];function Ya(){var t=Aa[0];Aa[0]=Aa[1],Aa[1]=t}function ja(t,e,n,r,i,o,a,s,l,u){if(u>e&&u>r&&u>o&&u>s||u<e&&u<r&&u<o&&u<s)return 0;var c=xe(e,r,o,s,u,Ca);if(0===c)return 0;for(var d=0,h=-1,f=void 0,p=void 0,m=0;m<c;m++){var g=Ca[m],y=0===g||1===g?.5:1,v=_e(t,n,i,a,g);v<l||(h<0&&(h=we(e,r,o,s,Aa),Aa[1]<Aa[0]&&h>1&&Ya(),f=_e(e,r,o,s,Aa[0]),h>1&&(p=_e(e,r,o,s,Aa[1]))),2===h?g<Aa[0]?d+=f<e?y:-y:g<Aa[1]?d+=p<f?y:-y:d+=s<p?y:-y:g<Aa[0]?d+=f<e?y:-y:d+=s<f?y:-y)}return d}function Pa(t,e,n,r,i,o,a,s){if(s>e&&s>r&&s>o||s<e&&s<r&&s<o)return 0;var l=De(e,r,o,s,Ca);if(0===l)return 0;var u=Oe(e,r,o);if(u>=0&&u<=1){for(var c=0,d=Le(e,r,o,u),h=0;h<l;h++){var f=0===Ca[h]||1===Ca[h]?.5:1,p=Le(t,n,i,Ca[h]);p<a||(Ca[h]<u?c+=d<e?f:-f:c+=o<d?f:-f)}return c}f=0===Ca[0]||1===Ca[0]?.5:1,p=Le(t,n,i,Ca[0]);return p<a?0:o<e?f:-f}function Ea(t,e,n,r,i,o,a,s){if(s-=e,s>n||s<-n)return 0;var l=Math.sqrt(n*n-s*s);Ca[0]=-l,Ca[1]=l;var u=Math.abs(r-i);if(u<1e-4)return 0;if(u>=Da-1e-4){r=0,i=Da;var c=o?1:-1;return a>=Ca[0]+t&&a<=Ca[1]+t?c:0}if(r>i){var d=r;r=i,i=d}r<0&&(r+=Da,i+=Da);for(var h=0,f=0;f<2;f++){var p=Ca[f];if(p+t>a){var m=Math.atan2(s,p);c=o?1:-1;m<0&&(m=Da+m),(m>=r&&m<=i||m+Da>=r&&m+Da<=i)&&(m>Math.PI/2&&m<1.5*Math.PI&&(c=-c),h+=c)}}return h}function Na(t,e,n,r,i){for(var o,a,s=t.data,l=t.len(),u=0,c=0,d=0,h=0,f=0,p=0;p<l;){var m=s[p++],g=1===p;switch(m===Ta.M&&p>1&&(n||(u+=La(c,d,h,f,r,i))),g&&(c=s[p],d=s[p+1],h=c,f=d),m){case Ta.M:h=s[p++],f=s[p++],c=h,d=f;break;case Ta.L:if(n){if(_a(c,d,s[p],s[p+1],e,r,i))return!0}else u+=La(c,d,s[p],s[p+1],r,i)||0;c=s[p++],d=s[p++];break;case Ta.C:if(n){if(ba(c,d,s[p++],s[p++],s[p++],s[p++],s[p],s[p+1],e,r,i))return!0}else u+=ja(c,d,s[p++],s[p++],s[p++],s[p++],s[p],s[p+1],r,i)||0;c=s[p++],d=s[p++];break;case Ta.Q:if(n){if(xa(c,d,s[p++],s[p++],s[p],s[p+1],e,r,i))return!0}else u+=Pa(c,d,s[p++],s[p++],s[p],s[p+1],r,i)||0;c=s[p++],d=s[p++];break;case Ta.A:var y=s[p++],v=s[p++],_=s[p++],b=s[p++],x=s[p++],w=s[p++];p+=1;var M=!!(1-s[p++]);o=Math.cos(x)*_+y,a=Math.sin(x)*b+v,g?(h=o,f=a):u+=La(c,d,o,a,r,i);var S=(r-y)*b/_+y;if(n){if(ka(y,v,b,x,x+w,M,e,S,i))return!0}else u+=Ea(y,v,b,x,x+w,M,S,i);c=Math.cos(x+w)*_+y,d=Math.sin(x+w)*b+v;break;case Ta.R:h=c=s[p++],f=d=s[p++];var k=s[p++],L=s[p++];if(o=h+k,a=f+L,n){if(_a(h,f,o,f,e,r,i)||_a(o,f,o,a,e,r,i)||_a(o,a,h,a,e,r,i)||_a(h,a,h,f,e,r,i))return!0}else u+=La(o,f,o,a,r,i),u+=La(h,a,h,f,r,i);break;case Ta.Z:if(n){if(_a(c,d,h,f,e,r,i))return!0}else u+=La(c,d,h,f,r,i);c=h,d=f;break}}return n||Ia(d,f)||(u+=La(c,d,h,f,r,i)||0),0!==u}function Ra(t,e,n){return Na(t,0,!1,e,n)}function Ha(t,e,n,r){return Na(t,e,!0,n,r)}var za=Object(s["i"])({fill:"#000",stroke:null,strokePercent:1,fillOpacity:1,strokeOpacity:1,lineDashOffset:0,lineWidth:1,lineCap:"butt",miterLimit:10,strokeNoScale:!1,strokeFirst:!1},Oo),Fa={style:Object(s["i"])({fill:!0,stroke:!0,strokePercent:!0,fillOpacity:!0,strokeOpacity:!0,lineDashOffset:!0,lineWidth:!0,miterLimit:!0},Io.style)},Ba=Un.concat(["invisible","culling","z","z2","zlevel","parent"]),Wa=function(t){function e(e){return t.call(this,e)||this}return o(e,t),e.prototype.update=function(){var n=this;t.prototype.update.call(this);var r=this.style;if(r.decal){var i=this._decalEl=this._decalEl||new e;i.buildPath===e.prototype.buildPath&&(i.buildPath=function(t){n.buildPath(t,n.shape)}),i.silent=!0;var o=i.style;for(var a in r)o[a]!==r[a]&&(o[a]=r[a]);o.fill=r.fill?r.decal:null,o.decal=null,o.shadowColor=null,r.strokeFirst&&(o.stroke=null);for(var s=0;s<Ba.length;++s)i[Ba[s]]=this[Ba[s]];i.__dirty|=Zt}else this._decalEl&&(this._decalEl=null)},e.prototype.getDecalElement=function(){return this._decalEl},e.prototype._init=function(e){var n=Object(s["F"])(e);this.shape=this.getDefaultShape();var r=this.getDefaultStyle();r&&this.useStyle(r);for(var i=0;i<n.length;i++){var o=n[i],a=e[o];"style"===o?this.style?Object(s["m"])(this.style,a):this.useStyle(a):"shape"===o?Object(s["m"])(this.shape,a):t.prototype.attrKV.call(this,o,a)}this.style||this.useStyle({})},e.prototype.getDefaultStyle=function(){return null},e.prototype.getDefaultShape=function(){return{}},e.prototype.canBeInsideText=function(){return this.hasFill()},e.prototype.getInsideTextFill=function(){var t=this.style.fill;if("none"!==t){if(Object(s["C"])(t)){var e=Object(Ne["e"])(t,0);return e>.5?En:e>.2?Rn:Nn}if(t)return Nn}return En},e.prototype.getInsideTextStroke=function(t){var e=this.style.fill;if(Object(s["C"])(e)){var n=this.__zr,r=!(!n||!n.isDarkMode()),i=Object(Ne["e"])(t,0)<Pn;if(r===i)return e}},e.prototype.buildPath=function(t,e,n){},e.prototype.pathUpdated=function(){this.__dirty&=~Kt},e.prototype.getUpdatedPathProxy=function(t){return!this.path&&this.createPathProxy(),this.path.beginPath(),this.buildPath(this.path,this.shape,t),this.path},e.prototype.createPathProxy=function(){this.path=new va(!1)},e.prototype.hasStroke=function(){var t=this.style,e=t.stroke;return!(null==e||"none"===e||!(t.lineWidth>0))},e.prototype.hasFill=function(){var t=this.style,e=t.fill;return null!=e&&"none"!==e},e.prototype.getBoundingRect=function(){var t=this._rect,e=this.style,n=!t;if(n){var r=!1;this.path||(r=!0,this.createPathProxy());var i=this.path;(r||this.__dirty&Kt)&&(i.beginPath(),this.buildPath(i,this.shape,!1),this.pathUpdated()),t=i.getBoundingRect()}if(this._rect=t,this.hasStroke()&&this.path&&this.path.len()>0){var o=this._rectStroke||(this._rectStroke=t.clone());if(this.__dirty||n){o.copy(t);var a=e.strokeNoScale?this.getLineScale():1,s=e.lineWidth;if(!this.hasFill()){var l=this.strokeContainThreshold;s=Math.max(s,null==l?4:l)}a>1e-10&&(o.width+=s/a,o.height+=s/a,o.x-=s/a/2,o.y-=s/a/2)}return o}return t},e.prototype.contain=function(t,e){var n=this.transformCoordToLocal(t,e),r=this.getBoundingRect(),i=this.style;if(t=n[0],e=n[1],r.contain(t,e)){var o=this.path;if(this.hasStroke()){var a=i.lineWidth,s=i.strokeNoScale?this.getLineScale():1;if(s>1e-10&&(this.hasFill()||(a=Math.max(a,this.strokeContainThreshold)),Ha(o,a/s,t,e)))return!0}if(this.hasFill())return Ra(o,t,e)}return!1},e.prototype.dirtyShape=function(){this.__dirty|=Kt,this._rect&&(this._rect=null),this._decalEl&&this._decalEl.dirtyShape(),this.markRedraw()},e.prototype.dirty=function(){this.dirtyStyle(),this.dirtyShape()},e.prototype.animateShape=function(t){return this.animate("shape",t)},e.prototype.updateDuringAnimation=function(t){"style"===t?this.dirtyStyle():"shape"===t?this.dirtyShape():this.markRedraw()},e.prototype.attrKV=function(e,n){"shape"===e?this.setShape(n):t.prototype.attrKV.call(this,e,n)},e.prototype.setShape=function(t,e){var n=this.shape;return n||(n=this.shape={}),"string"===typeof t?n[t]=e:Object(s["m"])(n,t),this.dirtyShape(),this},e.prototype.shapeChanged=function(){return!!(this.__dirty&Kt)},e.prototype.createStyle=function(t){return Object(s["g"])(za,t)},e.prototype._innerSaveToNormal=function(e){t.prototype._innerSaveToNormal.call(this,e);var n=this._normalState;e.shape&&!n.shape&&(n.shape=Object(s["m"])({},this.shape))},e.prototype._applyStateObj=function(e,n,r,i,o,a){t.prototype._applyStateObj.call(this,e,n,r,i,o,a);var l,u=!(n&&i);if(n&&n.shape?o?i?l=n.shape:(l=Object(s["m"])({},r.shape),Object(s["m"])(l,n.shape)):(l=Object(s["m"])({},i?this.shape:r.shape),Object(s["m"])(l,n.shape)):u&&(l=r.shape),l)if(o){this.shape=Object(s["m"])({},this.shape);for(var c={},d=Object(s["F"])(l),h=0;h<d.length;h++){var f=d[h];"object"===typeof l[f]?this.shape[f]=l[f]:c[f]=l[f]}this._transitionState(e,{shape:c},a)}else this.shape=l,this.dirtyShape()},e.prototype._mergeStates=function(e){for(var n,r=t.prototype._mergeStates.call(this,e),i=0;i<e.length;i++){var o=e[i];o.shape&&(n=n||{},this._mergeStyle(n,o.shape))}return n&&(r.shape=n),r},e.prototype.getAnimationStyleProps=function(){return Fa},e.prototype.isZeroArea=function(){return!1},e.extend=function(t){var n=function(e){function n(n){var r=e.call(this,n)||this;return t.init&&t.init.call(r,n),r}return o(n,e),n.prototype.getDefaultStyle=function(){return Object(s["d"])(t.style)},n.prototype.getDefaultShape=function(){return Object(s["d"])(t.shape)},n}(e);for(var r in t)"function"===typeof t[r]&&(n.prototype[r]=t[r]);return n},e.initDefaultProps=function(){var t=e.prototype;t.type="path",t.strokeContainThreshold=5,t.segmentIgnoreThreshold=0,t.subPixelOptimize=!1,t.autoBatch=!1,t.__dirty=Zt|Jt|Kt}(),e}(No),Va=Wa,Ga=Object(s["i"])({strokeFirst:!0,font:Jn["a"],x:0,y:0,textAlign:"left",textBaseline:"top",miterLimit:2},za),$a=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.hasStroke=function(){var t=this.style,e=t.stroke;return null!=e&&"none"!==e&&t.lineWidth>0},e.prototype.hasFill=function(){var t=this.style,e=t.fill;return null!=e&&"none"!==e},e.prototype.createStyle=function(t){return Object(s["g"])(Ga,t)},e.prototype.setBoundingRect=function(t){this._rect=t},e.prototype.getBoundingRect=function(){var t=this.style;if(!this._rect){var e=t.text;null!=e?e+="":e="";var n=er(e,t.font,t.textAlign,t.textBaseline);if(n.x+=t.x||0,n.y+=t.y||0,this.hasStroke()){var r=t.lineWidth;n.x-=r/2,n.y-=r/2,n.width+=r,n.height+=r}this._rect=n}return this._rect},e.initDefaultProps=function(){var t=e.prototype;t.dirtyRectTolerance=10}(),e}(No);$a.prototype.type="tspan";var Ua=$a,qa=Object(s["i"])({x:0,y:0},Oo),Xa={style:Object(s["i"])({x:!0,y:!0,width:!0,height:!0,sx:!0,sy:!0,sWidth:!0,sHeight:!0},Io.style)};function Za(t){return!!(t&&"string"!==typeof t&&t.width&&t.height)}var Ja=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.createStyle=function(t){return Object(s["g"])(qa,t)},e.prototype._getSize=function(t){var e=this.style,n=e[t];if(null!=n)return n;var r=Za(e.image)?e.image:this.__image;if(!r)return 0;var i="width"===t?"height":"width",o=e[i];return null==o?r[t]:r[t]/r[i]*o},e.prototype.getWidth=function(){return this._getSize("width")},e.prototype.getHeight=function(){return this._getSize("height")},e.prototype.getAnimationStyleProps=function(){return Xa},e.prototype.getBoundingRect=function(){var t=this.style;return this._rect||(this._rect=new Tt(t.x||0,t.y||0,this.getWidth(),this.getHeight())),this._rect},e}(No);Ja.prototype.type="image";var Ka=Ja;function Qa(t,e){var n,r,i,o,a,s=e.x,l=e.y,u=e.width,c=e.height,d=e.r;u<0&&(s+=u,u=-u),c<0&&(l+=c,c=-c),"number"===typeof d?n=r=i=o=d:d instanceof Array?1===d.length?n=r=i=o=d[0]:2===d.length?(n=i=d[0],r=o=d[1]):3===d.length?(n=d[0],r=o=d[1],i=d[2]):(n=d[0],r=d[1],i=d[2],o=d[3]):n=r=i=o=0,n+r>u&&(a=n+r,n*=u/a,r*=u/a),i+o>u&&(a=i+o,i*=u/a,o*=u/a),r+i>c&&(a=r+i,r*=c/a,i*=c/a),n+o>c&&(a=n+o,n*=c/a,o*=c/a),t.moveTo(s+n,l),t.lineTo(s+u-r,l),0!==r&&t.arc(s+u-r,l+r,r,-Math.PI/2,0),t.lineTo(s+u,l+c-i),0!==i&&t.arc(s+u-i,l+c-i,i,0,Math.PI/2),t.lineTo(s+o,l+c),0!==o&&t.arc(s+o,l+c-o,o,Math.PI/2,Math.PI),t.lineTo(s,l+n),0!==n&&t.arc(s+n,l+n,n,Math.PI,1.5*Math.PI)}var ts=Math.round;function es(t,e,n){if(e){var r=e.x1,i=e.x2,o=e.y1,a=e.y2;t.x1=r,t.x2=i,t.y1=o,t.y2=a;var s=n&&n.lineWidth;return s?(ts(2*r)===ts(2*i)&&(t.x1=t.x2=rs(r,s,!0)),ts(2*o)===ts(2*a)&&(t.y1=t.y2=rs(o,s,!0)),t):t}}function ns(t,e,n){if(e){var r=e.x,i=e.y,o=e.width,a=e.height;t.x=r,t.y=i,t.width=o,t.height=a;var s=n&&n.lineWidth;return s?(t.x=rs(r,s,!0),t.y=rs(i,s,!0),t.width=Math.max(rs(r+o,s,!1)-t.x,0===o?0:1),t.height=Math.max(rs(i+a,s,!1)-t.y,0===a?0:1),t):t}}function rs(t,e,n){if(!e)return t;var r=ts(2*t);return(r+ts(e))%2===0?r/2:(r+(n?1:-1))/2}var is=function(){function t(){this.x=0,this.y=0,this.width=0,this.height=0}return t}(),os={},as=function(t){function e(e){return t.call(this,e)||this}return o(e,t),e.prototype.getDefaultShape=function(){return new is},e.prototype.buildPath=function(t,e){var n,r,i,o;if(this.subPixelOptimize){var a=ns(os,e,this.style);n=a.x,r=a.y,i=a.width,o=a.height,a.r=e.r,e=a}else n=e.x,r=e.y,i=e.width,o=e.height;e.r?Qa(t,e):t.rect(n,r,i,o)},e.prototype.isZeroArea=function(){return!this.shape.width||!this.shape.height},e}(Va);as.prototype.type="rect";var ss=as,ls={fill:"#000"},us=2,cs={style:Object(s["i"])({fill:!0,stroke:!0,fillOpacity:!0,strokeOpacity:!0,lineWidth:!0,fontSize:!0,lineHeight:!0,width:!0,height:!0,textShadowColor:!0,textShadowBlur:!0,textShadowOffsetX:!0,textShadowOffsetY:!0,backgroundColor:!0,padding:!0,borderColor:!0,borderWidth:!0,borderRadius:!0},Io.style)},ds=function(t){function e(e){var n=t.call(this)||this;return n.type="text",n._children=[],n._defaultStyle=ls,n.attr(e),n}return o(e,t),e.prototype.childrenRef=function(){return this._children},e.prototype.update=function(){t.prototype.update.call(this),this.styleChanged()&&this._updateSubTexts();for(var e=0;e<this._children.length;e++){var n=this._children[e];n.zlevel=this.zlevel,n.z=this.z,n.z2=this.z2,n.culling=this.culling,n.cursor=this.cursor,n.invisible=this.invisible}},e.prototype.updateTransform=function(){var e=this.innerTransformable;e?(e.updateTransform(),e.transform&&(this.transform=e.transform)):t.prototype.updateTransform.call(this)},e.prototype.getLocalTransform=function(e){var n=this.innerTransformable;return n?n.getLocalTransform(e):t.prototype.getLocalTransform.call(this,e)},e.prototype.getComputedTransform=function(){return this.__hostTarget&&(this.__hostTarget.getComputedTransform(),this.__hostTarget.updateInnerText(!0)),t.prototype.getComputedTransform.call(this)},e.prototype._updateSubTexts=function(){this._childCursor=0,vs(this.style),this.style.rich?this._updateRichTexts():this._updatePlainTexts(),this._children.length=this._childCursor,this.styleUpdated()},e.prototype.addSelfToZr=function(e){t.prototype.addSelfToZr.call(this,e);for(var n=0;n<this._children.length;n++)this._children[n].__zr=e},e.prototype.removeSelfFromZr=function(e){t.prototype.removeSelfFromZr.call(this,e);for(var n=0;n<this._children.length;n++)this._children[n].__zr=null},e.prototype.getBoundingRect=function(){if(this.styleChanged()&&this._updateSubTexts(),!this._rect){for(var t=new Tt(0,0,0,0),e=this._children,n=[],r=null,i=0;i<e.length;i++){var o=e[i],a=o.getBoundingRect(),s=o.getLocalTransform(n);s?(t.copy(a),t.applyTransform(s),r=r||t.clone(),r.union(t)):(r=r||a.clone(),r.union(a))}this._rect=r||t}return this._rect},e.prototype.setDefaultTextStyle=function(t){this._defaultStyle=t||ls},e.prototype.setTextContent=function(t){0},e.prototype._mergeStyle=function(t,e){if(!e)return t;var n=e.rich,r=t.rich||n&&{};return Object(s["m"])(t,e),n&&r?(this._mergeRich(r,n),t.rich=r):r&&(t.rich=r),t},e.prototype._mergeRich=function(t,e){for(var n=Object(s["F"])(e),r=0;r<n.length;r++){var i=n[r];t[i]=t[i]||{},Object(s["m"])(t[i],e[i])}},e.prototype.getAnimationStyleProps=function(){return cs},e.prototype._getOrCreateChild=function(t){var e=this._children[this._childCursor];return e&&e instanceof t||(e=new t),this._children[this._childCursor++]=e,e.__zr=this.__zr,e.parent=this,e},e.prototype._updatePlainTexts=function(){var t=this.style,e=t.font||Jn["a"],n=t.padding,r=Ms(t),i=vo(r,t),o=Ss(t),a=!!t.backgroundColor,s=i.outerHeight,l=i.outerWidth,u=i.contentWidth,c=i.lines,d=i.lineHeight,h=this._defaultStyle,f=t.x||0,p=t.y||0,m=t.align||h.align||"left",g=t.verticalAlign||h.verticalAlign||"top",y=f,v=rr(p,i.contentHeight,g);if(o||n){var _=nr(f,l,m),b=rr(p,s,g);o&&this._renderBackground(t,t,_,b,l,s)}v+=d/2,n&&(y=ws(f,m,n),"top"===g?v+=n[0]:"bottom"===g&&(v-=n[2]));for(var x=0,w=!1,M=(xs("fill"in t?t.fill:(w=!0,h.fill))),S=(bs("stroke"in t?t.stroke:a||h.autoStroke&&!w?null:(x=us,h.stroke))),k=t.textShadowBlur>0,L=null!=t.width&&("truncate"===t.overflow||"break"===t.overflow||"breakAll"===t.overflow),T=i.calculatedLineHeight,D=0;D<c.length;D++){var O=this._getOrCreateChild(Ua),I=O.createStyle();O.useStyle(I),I.text=c[D],I.x=y,I.y=v,m&&(I.textAlign=m),I.textBaseline="middle",I.opacity=t.opacity,I.strokeFirst=!0,k&&(I.shadowBlur=t.textShadowBlur||0,I.shadowColor=t.textShadowColor||"transparent",I.shadowOffsetX=t.textShadowOffsetX||0,I.shadowOffsetY=t.textShadowOffsetY||0),I.stroke=S,I.fill=M,S&&(I.lineWidth=t.lineWidth||x,I.lineDash=t.lineDash,I.lineDashOffset=t.lineDashOffset||0),I.font=e,gs(I,t),v+=d,L&&O.setBoundingRect(new Tt(nr(I.x,t.width,I.textAlign),rr(I.y,T,I.textBaseline),u,T))}},e.prototype._updateRichTexts=function(){var t=this.style,e=Ms(t),n=wo(e,t),r=n.width,i=n.outerWidth,o=n.outerHeight,a=t.padding,s=t.x||0,l=t.y||0,u=this._defaultStyle,c=t.align||u.align,d=t.verticalAlign||u.verticalAlign,h=nr(s,i,c),f=rr(l,o,d),p=h,m=f;a&&(p+=a[3],m+=a[0]);var g=p+r;Ss(t)&&this._renderBackground(t,t,h,f,i,o);for(var y=!!t.backgroundColor,v=0;v<n.lines.length;v++){var _=n.lines[v],b=_.tokens,x=b.length,w=_.lineHeight,M=_.width,S=0,k=p,L=g,T=x-1,D=void 0;while(S<x&&(D=b[S],!D.align||"left"===D.align))this._placeToken(D,t,w,m,k,"left",y),M-=D.width,k+=D.width,S++;while(T>=0&&(D=b[T],"right"===D.align))this._placeToken(D,t,w,m,L,"right",y),M-=D.width,L-=D.width,T--;k+=(r-(k-p)-(g-L)-M)/2;while(S<=T)D=b[S],this._placeToken(D,t,w,m,k+D.width/2,"center",y),k+=D.width,S++;m+=w}},e.prototype._placeToken=function(t,e,n,r,i,o,a){var l=e.rich[t.styleName]||{};l.text=t.text;var u=t.verticalAlign,c=r+n/2;"top"===u?c=r+t.height/2:"bottom"===u&&(c=r+n-t.height/2);var d=!t.isLineHolder&&Ss(l);d&&this._renderBackground(l,e,"right"===o?i-t.width:"center"===o?i-t.width/2:i,c-t.height/2,t.width,t.height);var h=!!l.backgroundColor,f=t.textPadding;f&&(i=ws(i,o,f),c-=t.height/2-f[0]-t.innerHeight/2);var p=this._getOrCreateChild(Ua),m=p.createStyle();p.useStyle(m);var g=this._defaultStyle,y=!1,v=0,_=xs("fill"in l?l.fill:"fill"in e?e.fill:(y=!0,g.fill)),b=bs("stroke"in l?l.stroke:"stroke"in e?e.stroke:h||a||g.autoStroke&&!y?null:(v=us,g.stroke)),x=l.textShadowBlur>0||e.textShadowBlur>0;m.text=t.text,m.x=i,m.y=c,x&&(m.shadowBlur=l.textShadowBlur||e.textShadowBlur||0,m.shadowColor=l.textShadowColor||e.textShadowColor||"transparent",m.shadowOffsetX=l.textShadowOffsetX||e.textShadowOffsetX||0,m.shadowOffsetY=l.textShadowOffsetY||e.textShadowOffsetY||0),m.textAlign=o,m.textBaseline="middle",m.font=t.font||Jn["a"],m.opacity=Object(s["Q"])(l.opacity,e.opacity,1),gs(m,l),b&&(m.lineWidth=Object(s["Q"])(l.lineWidth,e.lineWidth,v),m.lineDash=Object(s["P"])(l.lineDash,e.lineDash),m.lineDashOffset=e.lineDashOffset||0,m.stroke=b),_&&(m.fill=_);var w=t.contentWidth,M=t.contentHeight;p.setBoundingRect(new Tt(nr(m.x,w,m.textAlign),rr(m.y,M,m.textBaseline),w,M))},e.prototype._renderBackground=function(t,e,n,r,i,o){var a,l,u=t.backgroundColor,c=t.borderWidth,d=t.borderColor,h=u&&u.image,f=u&&!h,p=t.borderRadius,m=this;if(f||t.lineHeight||c&&d){a=this._getOrCreateChild(ss),a.useStyle(a.createStyle()),a.style.fill=null;var g=a.shape;g.x=n,g.y=r,g.width=i,g.height=o,g.r=p,a.dirtyShape()}if(f){var y=a.style;y.fill=u||null,y.fillOpacity=Object(s["P"])(t.fillOpacity,1)}else if(h){l=this._getOrCreateChild(Ka),l.onload=function(){m.dirtyStyle()};var v=l.style;v.image=u.image,v.x=n,v.y=r,v.width=i,v.height=o}if(c&&d){y=a.style;y.lineWidth=c,y.stroke=d,y.strokeOpacity=Object(s["P"])(t.strokeOpacity,1),y.lineDash=t.borderDash,y.lineDashOffset=t.borderDashOffset||0,a.strokeContainThreshold=0,a.hasFill()&&a.hasStroke()&&(y.strokeFirst=!0,y.lineWidth*=2)}var _=(a||l).style;_.shadowBlur=t.shadowBlur||0,_.shadowColor=t.shadowColor||"transparent",_.shadowOffsetX=t.shadowOffsetX||0,_.shadowOffsetY=t.shadowOffsetY||0,_.opacity=Object(s["Q"])(t.opacity,e.opacity,1)},e.makeFont=function(t){var e="";return ys(t)&&(e=[t.fontStyle,t.fontWeight,ms(t.fontSize),t.fontFamily||"sans-serif"].join(" ")),e&&Object(s["T"])(e)||t.textFont||t.font},e}(No),hs={left:!0,right:1,center:1},fs={top:1,bottom:1,middle:1},ps=["fontStyle","fontWeight","fontSize","fontFamily"];function ms(t){return"string"!==typeof t||-1===t.indexOf("px")&&-1===t.indexOf("rem")&&-1===t.indexOf("em")?isNaN(+t)?Jn["c"]+"px":t+"px":t}function gs(t,e){for(var n=0;n<ps.length;n++){var r=ps[n],i=e[r];null!=i&&(t[r]=i)}}function ys(t){return null!=t.fontSize||t.fontFamily||t.fontWeight}function vs(t){return _s(t),Object(s["k"])(t.rich,_s),t}function _s(t){if(t){t.font=ds.makeFont(t);var e=t.align;"middle"===e&&(e="center"),t.align=null==e||hs[e]?e:"left";var n=t.verticalAlign;"center"===n&&(n="middle"),t.verticalAlign=null==n||fs[n]?n:"top";var r=t.padding;r&&(t.padding=Object(s["M"])(t.padding))}}function bs(t,e){return null==t||e<=0||"transparent"===t||"none"===t?null:t.image||t.colorStops?"#000":t}function xs(t){return null==t||"none"===t?null:t.image||t.colorStops?"#000":t}function ws(t,e,n){return"right"===e?t-n[1]:"center"===e?t+n[3]/2-n[1]/2:t+n[3]}function Ms(t){var e=t.text;return null!=e&&(e+=""),e}function Ss(t){return!!(t.backgroundColor||t.lineHeight||t.borderWidth&&t.borderColor)}var ks=ds,Ls=Ci(),Ts=function(t,e,n,r){if(r){var i=Ls(r);i.dataIndex=n,i.dataType=e,i.seriesIndex=t,i.ssrType="chart","group"===r.type&&r.traverse((function(r){var i=Ls(r);i.seriesIndex=t,i.dataIndex=n,i.dataType=e,i.ssrType="chart"}))}},Ds=1,Os={},Is=Ci(),Cs=Ci(),As=0,Ys=1,js=2,Ps=["emphasis","blur","select"],Es=["normal","emphasis","blur","select"],Ns=10,Rs=9,Hs="highlight",zs="downplay",Fs="select",Bs="unselect",Ws="toggleSelect";function Vs(t){return null!=t&&"none"!==t}function Gs(t,e,n){t.onHoverStateChange&&(t.hoverState||0)!==n&&t.onHoverStateChange(e),t.hoverState=n}function $s(t){Gs(t,"emphasis",js)}function Us(t){t.hoverState===js&&Gs(t,"normal",As)}function qs(t){Gs(t,"blur",Ys)}function Xs(t){t.hoverState===Ys&&Gs(t,"normal",As)}function Zs(t){t.selected=!0}function Js(t){t.selected=!1}function Ks(t,e,n){e(t,n)}function Qs(t,e,n){Ks(t,e,n),t.isGroup&&t.traverse((function(t){Ks(t,e,n)}))}function tl(t,e){switch(e){case"emphasis":t.hoverState=js;break;case"normal":t.hoverState=As;break;case"blur":t.hoverState=Ys;break;case"select":t.selected=!0}}function el(t,e,n,r){for(var i=t.style,o={},a=0;a<e.length;a++){var s=e[a],l=i[s];o[s]=null==l?r&&r[s]:l}for(a=0;a<t.animators.length;a++){var u=t.animators[a];u.__fromStateTransition&&u.__fromStateTransition.indexOf(n)<0&&"style"===u.targetName&&u.saveTo(o,e)}return o}function nl(t,e,n,r){var i=n&&Object(s["r"])(n,"select")>=0,o=!1;if(t instanceof Va){var a=Is(t),l=i&&a.selectFill||a.normalFill,u=i&&a.selectStroke||a.normalStroke;if(Vs(l)||Vs(u)){r=r||{};var c=r.style||{};"inherit"===c.fill?(o=!0,r=Object(s["m"])({},r),c=Object(s["m"])({},c),c.fill=l):!Vs(c.fill)&&Vs(l)?(o=!0,r=Object(s["m"])({},r),c=Object(s["m"])({},c),c.fill=Object(Ne["d"])(l)):!Vs(c.stroke)&&Vs(u)&&(o||(r=Object(s["m"])({},r),c=Object(s["m"])({},c)),c.stroke=Object(Ne["d"])(u)),r.style=c}}if(r&&null==r.z2){o||(r=Object(s["m"])({},r));var d=t.z2EmphasisLift;r.z2=t.z2+(null!=d?d:Ns)}return r}function rl(t,e,n){if(n&&null==n.z2){n=Object(s["m"])({},n);var r=t.z2SelectLift;n.z2=t.z2+(null!=r?r:Rs)}return n}function il(t,e,n){var r=Object(s["r"])(t.currentStates,e)>=0,i=t.style.opacity,o=r?null:el(t,["opacity"],e,{opacity:1});n=n||{};var a=n.style||{};return null==a.opacity&&(n=Object(s["m"])({},n),a=Object(s["m"])({opacity:r?i:.1*o.opacity},a),n.style=a),n}function ol(t,e){var n=this.states[t];if(this.style){if("emphasis"===t)return nl(this,t,e,n);if("blur"===t)return il(this,t,n);if("select"===t)return rl(this,t,n)}return n}function al(t){t.stateProxy=ol;var e=t.getTextContent(),n=t.getTextGuideLine();e&&(e.stateProxy=ol),n&&(n.stateProxy=ol)}function sl(t,e){!ml(t,e)&&!t.__highByOuter&&Qs(t,$s)}function ll(t,e){!ml(t,e)&&!t.__highByOuter&&Qs(t,Us)}function ul(t,e){t.__highByOuter|=1<<(e||0),Qs(t,$s)}function cl(t,e){!(t.__highByOuter&=~(1<<(e||0)))&&Qs(t,Us)}function dl(t){Qs(t,qs)}function hl(t){Qs(t,Xs)}function fl(t){Qs(t,Zs)}function pl(t){Qs(t,Js)}function ml(t,e){return t.__highDownSilentOnTouch&&e.zrByTouch}function gl(t){var e=t.getModel(),n=[],r=[];e.eachComponent((function(e,i){var o=Cs(i),a="series"===e,s=a?t.getViewOfSeriesModel(i):t.getViewOfComponentModel(i);!a&&r.push(s),o.isBlured&&(s.group.traverse((function(t){Xs(t)})),a&&n.push(i)),o.isBlured=!1})),Object(s["k"])(r,(function(t){t&&t.toggleBlurSeries&&t.toggleBlurSeries(n,!1,e)}))}function yl(t,e,n,r){var i=r.getModel();function o(t,e){for(var n=0;n<e.length;n++){var r=t.getItemGraphicEl(e[n]);r&&hl(r)}}if(n=n||"coordinateSystem",null!=t&&e&&"none"!==e){var a=i.getSeriesByIndex(t),l=a.coordinateSystem;l&&l.master&&(l=l.master);var u=[];i.eachSeries((function(t){var i=a===t,c=t.coordinateSystem;c&&c.master&&(c=c.master);var d=c&&l?c===l:i;if(!("series"===n&&!i||"coordinateSystem"===n&&!d||"series"===e&&i)){var h=r.getViewOfSeriesModel(t);if(h.group.traverse((function(t){t.__highByOuter&&i&&"self"===e||qs(t)})),Object(s["u"])(e))o(t.getData(),e);else if(Object(s["A"])(e))for(var f=Object(s["F"])(e),p=0;p<f.length;p++)o(t.getData(f[p]),e[f[p]]);u.push(t),Cs(t).isBlured=!0}})),i.eachComponent((function(t,e){if("series"!==t){var n=r.getViewOfComponentModel(e);n&&n.toggleBlurSeries&&n.toggleBlurSeries(u,!0,i)}}))}}function vl(t,e,n){if(null!=t&&null!=e){var r=n.getModel().getComponent(t,e);if(r){Cs(r).isBlured=!0;var i=n.getViewOfComponentModel(r);i&&i.focusBlurEnabled&&i.group.traverse((function(t){qs(t)}))}}}function _l(t,e,n){var r=t.seriesIndex,i=t.getData(e.dataType);if(i){var o=Ii(i,e);o=(Object(s["t"])(o)?o[0]:o)||0;var a=i.getItemGraphicEl(o);if(!a){var l=i.count(),u=0;while(!a&&u<l)a=i.getItemGraphicEl(u++)}if(a){var c=Ls(a);yl(r,c.focus,c.blurScope,n)}else{var d=t.get(["emphasis","focus"]),h=t.get(["emphasis","blurScope"]);null!=d&&yl(r,d,h,n)}}}function bl(t,e,n,r){var i={focusSelf:!1,dispatchers:null};if(null==t||"series"===t||null==e||null==n)return i;var o=r.getModel().getComponent(t,e);if(!o)return i;var a=r.getViewOfComponentModel(o);if(!a||!a.findHighDownDispatchers)return i;for(var s,l=a.findHighDownDispatchers(n),u=0;u<l.length;u++)if("self"===Ls(l[u]).focus){s=!0;break}return{focusSelf:s,dispatchers:l}}function xl(t,e,n){var r=Ls(t),i=bl(r.componentMainType,r.componentIndex,r.componentHighDownName,n),o=i.dispatchers,a=i.focusSelf;o?(a&&vl(r.componentMainType,r.componentIndex,n),Object(s["k"])(o,(function(t){return sl(t,e)}))):(yl(r.seriesIndex,r.focus,r.blurScope,n),"self"===r.focus&&vl(r.componentMainType,r.componentIndex,n),sl(t,e))}function wl(t,e,n){gl(n);var r=Ls(t),i=bl(r.componentMainType,r.componentIndex,r.componentHighDownName,n).dispatchers;i?Object(s["k"])(i,(function(t){return ll(t,e)})):ll(t,e)}function Ml(t,e,n){if(Nl(e)){var r=e.dataType,i=t.getData(r),o=Ii(i,e);Object(s["t"])(o)||(o=[o]),t[e.type===Ws?"toggleSelect":e.type===Fs?"select":"unselect"](o,r)}}function Sl(t){var e=t.getAllData();Object(s["k"])(e,(function(e){var n=e.data,r=e.type;n.eachItemGraphicEl((function(e,n){t.isSelected(n,r)?fl(e):pl(e)}))}))}function kl(t){var e=[];return t.eachSeries((function(t){var n=t.getAllData();Object(s["k"])(n,(function(n){n.data;var r=n.type,i=t.getSelectedDataIndices();if(i.length>0){var o={dataIndex:i,seriesIndex:t.seriesIndex};null!=r&&(o.dataType=r),e.push(o)}}))})),e}function Ll(t,e,n){Yl(t,!0),Qs(t,al),Ol(t,e,n)}function Tl(t){Yl(t,!1)}function Dl(t,e,n,r){r?Tl(t):Ll(t,e,n)}function Ol(t,e,n){var r=Ls(t);null!=e?(r.focus=e,r.blurScope=n):r.focus&&(r.focus=null)}var Il=["emphasis","blur","select"],Cl={itemStyle:"getItemStyle",lineStyle:"getLineStyle",areaStyle:"getAreaStyle"};function Al(t,e,n,r){n=n||"itemStyle";for(var i=0;i<Il.length;i++){var o=Il[i],a=e.getModel([o,n]),s=t.ensureState(o);s.style=r?r(a):a[Cl[n]]()}}function Yl(t,e){var n=!1===e,r=t;t.highDownSilentOnTouch&&(r.__highDownSilentOnTouch=t.highDownSilentOnTouch),n&&!r.__highDownDispatcher||(r.__highByOuter=r.__highByOuter||0,r.__highDownDispatcher=!n)}function jl(t){return!(!t||!t.__highDownDispatcher)}function Pl(t,e,n){var r=Ls(t);r.componentMainType=e.mainType,r.componentIndex=e.componentIndex,r.componentHighDownName=n}function El(t){var e=Os[t];return null==e&&Ds<=32&&(e=Os[t]=Ds++),e}function Nl(t){var e=t.type;return e===Fs||e===Bs||e===Ws}function Rl(t){var e=t.type;return e===Hs||e===zs}function Hl(t){var e=Is(t);e.normalFill=t.style.fill,e.normalStroke=t.style.stroke;var n=t.states.select||{};e.selectFill=n.style&&n.style.fill||null,e.selectStroke=n.style&&n.style.stroke||null}var zl=Ci();function Fl(t,e,n,r,i){var o;if(e&&e.ecModel){var a=e.ecModel.getUpdatePayload();o=a&&a.animation}var l=e&&e.isAnimationEnabled(),u="update"===t;if(l){var c=void 0,d=void 0,h=void 0;r?(c=Object(s["P"])(r.duration,200),d=Object(s["P"])(r.easing,"cubicOut"),h=0):(c=e.getShallow(u?"animationDurationUpdate":"animationDuration"),d=e.getShallow(u?"animationEasingUpdate":"animationEasing"),h=e.getShallow(u?"animationDelayUpdate":"animationDelay")),o&&(null!=o.duration&&(c=o.duration),null!=o.easing&&(d=o.easing),null!=o.delay&&(h=o.delay)),Object(s["w"])(h)&&(h=h(n,i)),Object(s["w"])(c)&&(c=c(n));var f={duration:c||0,delay:h,easing:d};return f}return null}function Bl(t,e,n,r,i,o,a){var l,u=!1;Object(s["w"])(i)?(a=o,o=i,i=null):Object(s["A"])(i)&&(o=i.cb,a=i.during,u=i.isFrom,l=i.removeOpt,i=i.dataIndex);var c="leave"===t;c||e.stopAnimation("leave");var d=Fl(t,r,i,c?l||{}:null,r&&r.getAnimationDelayParams?r.getAnimationDelayParams(e,i):null);if(d&&d.duration>0){var h=d.duration,f=d.delay,p=d.easing,m={duration:h,delay:f||0,easing:p,done:o,force:!!o||!!a,setToFinal:!c,scope:t,during:a};u?e.animateFrom(n,m):e.animateTo(n,m)}else e.stopAnimation(),!u&&e.attr(n),a&&a(1),o&&o()}function Wl(t,e,n,r,i,o){Bl("update",t,e,n,r,i,o)}function Vl(t,e,n,r,i,o){Bl("enter",t,e,n,r,i,o)}function Gl(t){if(!t.__zr)return!0;for(var e=0;e<t.animators.length;e++){var n=t.animators[e];if("leave"===n.scope)return!0}return!1}function $l(t,e,n,r,i,o){Gl(t)||Bl("leave",t,e,n,r,i,o)}function Ul(t,e,n,r){t.removeTextContent(),t.removeTextGuideLine(),$l(t,{style:{opacity:0}},e,n,r)}function ql(t,e,n){function r(){t.parent&&t.parent.remove(t)}t.isGroup?t.traverse((function(t){t.isGroup||Ul(t,e,n,r)})):Ul(t,e,n,r)}function Xl(t){zl(t).oldStyle=t.style}function Zl(t){return zl(t).oldStyle}var Jl={};function Kl(t,e){for(var n=0;n<Ps.length;n++){var r=Ps[n],i=e[r],o=t.ensureState(r);o.style=o.style||{},o.style.text=i}var a=t.currentStates.slice();t.clearStates(!0),t.setStyle({text:e.normal}),t.useStates(a,!0)}function Ql(t,e,n){var r,i=t.labelFetcher,o=t.labelDataIndex,a=t.labelDimIndex,l=e.normal;i&&(r=i.getFormattedLabel(o,"normal",null,a,l&&l.get("formatter"),null!=n?{interpolatedValue:n}:null)),null==r&&(r=Object(s["w"])(t.defaultText)?t.defaultText(o,t,n):t.defaultText);for(var u={normal:r},c=0;c<Ps.length;c++){var d=Ps[c],h=e[d];u[d]=Object(s["P"])(i?i.getFormattedLabel(o,d,null,a,h&&h.get("formatter")):null,r)}return u}function tu(t,e,n,r){n=n||Jl;for(var i=t instanceof ks,o=!1,a=0;a<Es.length;a++){var l=e[Es[a]];if(l&&l.getShallow("show")){o=!0;break}}var u=i?t:t.getTextContent();if(o){i||(u||(u=new ks,t.setTextContent(u)),t.stateProxy&&(u.stateProxy=t.stateProxy));var c=Ql(n,e),d=e.normal,h=!!d.getShallow("show"),f=nu(d,r&&r.normal,n,!1,!i);f.text=c.normal,i||t.setTextConfig(ru(d,n,!1));for(a=0;a<Ps.length;a++){var p=Ps[a];l=e[p];if(l){var m=u.ensureState(p),g=!!Object(s["P"])(l.getShallow("show"),h);if(g!==h&&(m.ignore=!g),m.style=nu(l,r&&r[p],n,!0,!i),m.style.text=c[p],!i){var y=t.ensureState(p);y.textConfig=ru(l,n,!0)}}}u.silent=!!d.getShallow("silent"),null!=u.style.x&&(f.x=u.style.x),null!=u.style.y&&(f.y=u.style.y),u.ignore=!h,u.useStyle(f),u.dirty(),n.enableTextSetter&&(du(u).setLabelText=function(t){var r=Ql(n,e,t);Kl(u,r)})}else u&&(u.ignore=!0);t.dirty()}function eu(t,e){e=e||"label";for(var n={normal:t.getModel(e)},r=0;r<Ps.length;r++){var i=Ps[r];n[i]=t.getModel([i,e])}return n}function nu(t,e,n,r,i){var o={};return iu(o,t,n,r,i),e&&Object(s["m"])(o,e),o}function ru(t,e,n){e=e||{};var r,i={},o=t.getShallow("rotate"),a=Object(s["P"])(t.getShallow("distance"),n?null:5),l=t.getShallow("offset");return r=t.getShallow("position")||(n?null:"inside"),"outside"===r&&(r=e.defaultOutsidePosition||"top"),null!=r&&(i.position=r),null!=l&&(i.offset=l),null!=o&&(o*=Math.PI/180,i.rotation=o),null!=a&&(i.distance=a),i.outsideFill="inherit"===t.get("color")?e.inheritColor||null:"auto",i}function iu(t,e,n,r,i){n=n||Jl;var o,a=e.ecModel,s=a&&a.option.textStyle,l=ou(e);if(l)for(var u in o={},l)if(l.hasOwnProperty(u)){var c=e.getModel(["rich",u]);uu(o[u]={},c,s,n,r,i,!1,!0)}o&&(t.rich=o);var d=e.get("overflow");d&&(t.overflow=d);var h=e.get("minMargin");null!=h&&(t.margin=h),uu(t,e,s,n,r,i,!0,!1)}function ou(t){var e;while(t&&t!==t.ecModel){var n=(t.option||Jl).rich;if(n){e=e||{};for(var r=Object(s["F"])(n),i=0;i<r.length;i++){var o=r[i];e[o]=1}}t=t.parentModel}return e}var au=["fontStyle","fontWeight","fontSize","fontFamily","textShadowColor","textShadowBlur","textShadowOffsetX","textShadowOffsetY"],su=["align","lineHeight","width","height","tag","verticalAlign","ellipsis"],lu=["padding","borderWidth","borderRadius","borderDashOffset","backgroundColor","borderColor","shadowColor","shadowBlur","shadowOffsetX","shadowOffsetY"];function uu(t,e,n,r,i,o,a,l){n=!i&&n||Jl;var u=r&&r.inheritColor,c=e.getShallow("color"),d=e.getShallow("textBorderColor"),h=Object(s["P"])(e.getShallow("opacity"),n.opacity);"inherit"!==c&&"auto"!==c||(c=u||null),"inherit"!==d&&"auto"!==d||(d=u||null),o||(c=c||n.color,d=d||n.textBorderColor),null!=c&&(t.fill=c),null!=d&&(t.stroke=d);var f=Object(s["P"])(e.getShallow("textBorderWidth"),n.textBorderWidth);null!=f&&(t.lineWidth=f);var p=Object(s["P"])(e.getShallow("textBorderType"),n.textBorderType);null!=p&&(t.lineDash=p);var m=Object(s["P"])(e.getShallow("textBorderDashOffset"),n.textBorderDashOffset);null!=m&&(t.lineDashOffset=m),i||null!=h||l||(h=r&&r.defaultOpacity),null!=h&&(t.opacity=h),i||o||null==t.fill&&r.inheritColor&&(t.fill=r.inheritColor);for(var g=0;g<au.length;g++){var y=au[g],v=Object(s["P"])(e.getShallow(y),n[y]);null!=v&&(t[y]=v)}for(g=0;g<su.length;g++){y=su[g],v=e.getShallow(y);null!=v&&(t[y]=v)}if(null==t.verticalAlign){var _=e.getShallow("baseline");null!=_&&(t.verticalAlign=_)}if(!a||!r.disableBox){for(g=0;g<lu.length;g++){y=lu[g],v=e.getShallow(y);null!=v&&(t[y]=v)}var b=e.getShallow("borderType");null!=b&&(t.borderDash=b),"auto"!==t.backgroundColor&&"inherit"!==t.backgroundColor||!u||(t.backgroundColor=u),"auto"!==t.borderColor&&"inherit"!==t.borderColor||!u||(t.borderColor=u)}}function cu(t,e){var n=e&&e.getModel("textStyle");return Object(s["T"])([t.fontStyle||n&&n.getShallow("fontStyle")||"",t.fontWeight||n&&n.getShallow("fontWeight")||"",(t.fontSize||n&&n.getShallow("fontSize")||12)+"px",t.fontFamily||n&&n.getShallow("fontFamily")||"sans-serif"].join(" "))}var du=Ci();function hu(t,e,n,r){if(t){var i=du(t);i.prevValue=i.value,i.value=n;var o=e.normal;i.valueAnimation=o.get("valueAnimation"),i.valueAnimation&&(i.precision=o.get("precision"),i.defaultInterpolatedText=r,i.statesModels=e)}}function fu(t,e,n,r,i){var o=du(t);if(o.valueAnimation&&o.prevValue!==o.value){var a=o.defaultInterpolatedText,l=Object(s["P"])(o.interpolatedValue,o.prevValue),u=o.value;t.percent=0,(null==o.prevValue?Vl:Wl)(t,{percent:1},r,e,null,c)}function c(r){var s=Bi(n,o.precision,l,u,r);o.interpolatedValue=1===r?null:s;var c=Ql({labelDataIndex:e,labelFetcher:i,defaultText:a?a(s):s+""},o.statesModels,s);Kl(t,c)}}var pu=["textStyle","color"],mu=["fontStyle","fontWeight","fontSize","fontFamily","padding","lineHeight","rich","width","height","overflow"],gu=new ks,yu=function(){function t(){}return t.prototype.getTextColor=function(t){var e=this.ecModel;return this.getShallow("color")||(!t&&e?e.get(pu):null)},t.prototype.getFont=function(){return cu({fontStyle:this.getShallow("fontStyle"),fontWeight:this.getShallow("fontWeight"),fontSize:this.getShallow("fontSize"),fontFamily:this.getShallow("fontFamily")},this.ecModel)},t.prototype.getTextRect=function(t){for(var e={text:t,verticalAlign:this.getShallow("verticalAlign")||this.getShallow("baseline")},n=0;n<mu.length;n++)e[mu[n]]=this.getShallow(mu[n]);return gu.useStyle(e),gu.update(),gu.getBoundingRect()},t}(),vu=yu,_u=[["lineWidth","width"],["stroke","color"],["opacity"],["shadowBlur"],["shadowOffsetX"],["shadowOffsetY"],["shadowColor"],["lineDash","type"],["lineDashOffset","dashOffset"],["lineCap","cap"],["lineJoin","join"],["miterLimit"]],bu=ro(_u),xu=function(){function t(){}return t.prototype.getLineStyle=function(t){return bu(this,t)},t}(),wu=[["fill","color"],["stroke","borderColor"],["lineWidth","borderWidth"],["opacity"],["shadowBlur"],["shadowOffsetX"],["shadowOffsetY"],["shadowColor"],["lineDash","borderType"],["lineDashOffset","borderDashOffset"],["lineCap","borderCap"],["lineJoin","borderJoin"],["miterLimit","borderMiterLimit"]],Mu=ro(wu),Su=function(){function t(){}return t.prototype.getItemStyle=function(t,e){return Mu(this,t,e)},t}(),ku=function(){function t(t,e,n){this.parentModel=e,this.ecModel=n,this.option=t}return t.prototype.init=function(t,e,n){for(var r=[],i=3;i<arguments.length;i++)r[i-3]=arguments[i]},t.prototype.mergeOption=function(t,e){Object(s["I"])(this.option,t,!0)},t.prototype.get=function(t,e){return null==t?this.option:this._doGet(this.parsePath(t),!e&&this.parentModel)},t.prototype.getShallow=function(t,e){var n=this.option,r=null==n?n:n[t];if(null==r&&!e){var i=this.parentModel;i&&(r=i.getShallow(t))}return r},t.prototype.getModel=function(e,n){var r=null!=e,i=r?this.parsePath(e):null,o=r?this._doGet(i):this.option;return n=n||this.parentModel&&this.parentModel.getModel(this.resolveParentPath(i)),new t(o,n,this.ecModel)},t.prototype.isEmpty=function(){return null==this.option},t.prototype.restoreData=function(){},t.prototype.clone=function(){var t=this.constructor;return new t(Object(s["d"])(this.option))},t.prototype.parsePath=function(t){return"string"===typeof t?t.split("."):t},t.prototype.resolveParentPath=function(t){return t},t.prototype.isAnimationEnabled=function(){if(!a["a"].node&&this.option){if(null!=this.option.animation)return!!this.option.animation;if(this.parentModel)return this.parentModel.isAnimationEnabled()}},t.prototype._doGet=function(t,e){var n=this.option;if(!t)return n;for(var r=0;r<t.length;r++)if(t[r]&&(n=n&&"object"===typeof n?n[t[r]]:null,null==n))break;return null==n&&e&&(n=e._doGet(this.resolveParentPath(t),e.parentModel)),n},t}();Xi(ku),Qi(ku),Object(s["K"])(ku,xu),Object(s["K"])(ku,Su),Object(s["K"])(ku,ao),Object(s["K"])(ku,vu);var Lu=ku,Tu=Math.round(10*Math.random());function Du(t){return[t||"",Tu++].join("_")}function Ou(t){var e={};t.registerSubTypeDefaulter=function(t,n){var r=$i(t);e[r.main]=n},t.determineSubType=function(n,r){var i=r.type;if(!i){var o=$i(n).main;t.hasSubTypes(n)&&e[o]&&(i=e[o](r))}return i}}function Iu(t,e){function n(t){var n={},o=[];return s["k"](t,(function(a){var l=r(n,a),u=l.originalDeps=e(a),c=i(u,t);l.entryCount=c.length,0===l.entryCount&&o.push(a),s["k"](c,(function(t){s["r"](l.predecessor,t)<0&&l.predecessor.push(t);var e=r(n,t);s["r"](e.successor,t)<0&&e.successor.push(a)}))})),{graph:n,noEntryList:o}}function r(t,e){return t[e]||(t[e]={predecessor:[],successor:[]}),t[e]}function i(t,e){var n=[];return s["k"](t,(function(t){s["r"](e,t)>=0&&n.push(t)})),n}t.topologicalTravel=function(t,e,r,i){if(t.length){var o=n(e),a=o.graph,l=o.noEntryList,u={};s["k"](t,(function(t){u[t]=!0}));while(l.length){var c=l.pop(),d=a[c],h=!!u[c];h&&(r.call(i,c,d.originalDeps.slice()),delete u[c]),s["k"](d.successor,h?p:f)}s["k"](u,(function(){var t="";throw new Error(t)}))}function f(t){a[t].entryCount--,0===a[t].entryCount&&l.push(t)}function p(t){u[t]=!0,f(t)}}}function Cu(t,e){return s["I"](s["I"]({},t,!0),e,!0)}var Au={time:{month:["January","February","March","April","May","June","July","August","September","October","November","December"],monthAbbr:["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"],dayOfWeek:["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"],dayOfWeekAbbr:["Sun","Mon","Tue","Wed","Thu","Fri","Sat"]},legend:{selector:{all:"All",inverse:"Inv"}},toolbox:{brush:{title:{rect:"Box Select",polygon:"Lasso Select",lineX:"Horizontally Select",lineY:"Vertically Select",keep:"Keep Selections",clear:"Clear Selections"}},dataView:{title:"Data View",lang:["Data View","Close","Refresh"]},dataZoom:{title:{zoom:"Zoom",back:"Zoom Reset"}},magicType:{title:{line:"Switch to Line Chart",bar:"Switch to Bar Chart",stack:"Stack",tiled:"Tile"}},restore:{title:"Restore"},saveAsImage:{title:"Save as Image",lang:["Right Click to Save Image"]}},series:{typeNames:{pie:"Pie chart",bar:"Bar chart",line:"Line chart",scatter:"Scatter plot",effectScatter:"Ripple scatter plot",radar:"Radar chart",tree:"Tree",treemap:"Treemap",boxplot:"Boxplot",candlestick:"Candlestick",k:"K line chart",heatmap:"Heat map",map:"Map",parallel:"Parallel coordinate map",lines:"Line graph",graph:"Relationship graph",sankey:"Sankey diagram",funnel:"Funnel chart",gauge:"Gauge",pictorialBar:"Pictorial bar",themeRiver:"Theme River Map",sunburst:"Sunburst",custom:"Custom chart",chart:"Chart"}},aria:{general:{withTitle:'This is a chart about "{title}"',withoutTitle:"This is a chart"},series:{single:{prefix:"",withName:" with type {seriesType} named {seriesName}.",withoutName:" with type {seriesType}."},multiple:{prefix:". It consists of {seriesCount} series count.",withName:" The {seriesId} series is a {seriesType} representing {seriesName}.",withoutName:" The {seriesId} series is a {seriesType}.",separator:{middle:"",end:""}}},data:{allData:"The data is as follows: ",partialData:"The first {displayCnt} items are: ",withName:"the data for {name} is {value}",withoutName:"{value}",separator:{middle:", ",end:". "}}}},Yu={time:{month:["一月","二月","三月","四月","五月","六月","七月","八月","九月","十月","十一月","十二月"],monthAbbr:["1月","2月","3月","4月","5月","6月","7月","8月","9月","10月","11月","12月"],dayOfWeek:["星期日","星期一","星期二","星期三","星期四","星期五","星期六"],dayOfWeekAbbr:["日","一","二","三","四","五","六"]},legend:{selector:{all:"全选",inverse:"反选"}},toolbox:{brush:{title:{rect:"矩形选择",polygon:"圈选",lineX:"横向选择",lineY:"纵向选择",keep:"保持选择",clear:"清除选择"}},dataView:{title:"数据视图",lang:["数据视图","关闭","刷新"]},dataZoom:{title:{zoom:"区域缩放",back:"区域缩放还原"}},magicType:{title:{line:"切换为折线图",bar:"切换为柱状图",stack:"切换为堆叠",tiled:"切换为平铺"}},restore:{title:"还原"},saveAsImage:{title:"保存为图片",lang:["右键另存为图片"]}},series:{typeNames:{pie:"饼图",bar:"柱状图",line:"折线图",scatter:"散点图",effectScatter:"涟漪散点图",radar:"雷达图",tree:"树图",treemap:"矩形树图",boxplot:"箱型图",candlestick:"K线图",k:"K线图",heatmap:"热力图",map:"地图",parallel:"平行坐标图",lines:"线图",graph:"关系图",sankey:"桑基图",funnel:"漏斗图",gauge:"仪表盘图",pictorialBar:"象形柱图",themeRiver:"主题河流图",sunburst:"旭日图",custom:"自定义图表",chart:"图表"}},aria:{general:{withTitle:"这是一个关于“{title}”的图表。",withoutTitle:"这是一个图表，"},series:{single:{prefix:"",withName:"图表类型是{seriesType}，表示{seriesName}。",withoutName:"图表类型是{seriesType}。"},multiple:{prefix:"它由{seriesCount}个图表系列组成。",withName:"第{seriesId}个系列是一个表示{seriesName}的{seriesType}，",withoutName:"第{seriesId}个系列是一个{seriesType}，",separator:{middle:"；",end:"。"}}},data:{allData:"其数据是——",partialData:"其中，前{displayCnt}项是——",withName:"{name}的数据是{value}",withoutName:"{value}",separator:{middle:"，",end:""}}}},ju="ZH",Pu="EN",Eu=Pu,Nu={},Ru={},Hu=a["a"].domSupported?function(){var t=(document.documentElement.lang||navigator.language||navigator.browserLanguage||Eu).toUpperCase();return t.indexOf(ju)>-1?ju:Eu}():Eu;function zu(t,e){t=t.toUpperCase(),Ru[t]=new Lu(e),Nu[t]=e}function Fu(t){if(Object(s["C"])(t)){var e=Nu[t.toUpperCase()]||{};return t===ju||t===Pu?Object(s["d"])(e):Object(s["I"])(Object(s["d"])(e),Object(s["d"])(Nu[Eu]),!1)}return Object(s["I"])(Object(s["d"])(t),Object(s["d"])(Nu[Eu]),!1)}function Bu(t){return Ru[t]}function Wu(){return Ru[Eu]}zu(Pu,Au),zu(ju,Yu);var Vu=1e3,Gu=60*Vu,$u=60*Gu,Uu=24*$u,qu=365*Uu,Xu={year:"{yyyy}",month:"{MMM}",day:"{d}",hour:"{HH}:{mm}",minute:"{HH}:{mm}",second:"{HH}:{mm}:{ss}",millisecond:"{HH}:{mm}:{ss} {SSS}",none:"{yyyy}-{MM}-{dd} {HH}:{mm}:{ss} {SSS}"},Zu="{yyyy}-{MM}-{dd}",Ju={year:"{yyyy}",month:"{yyyy}-{MM}",day:Zu,hour:Zu+" "+Xu.hour,minute:Zu+" "+Xu.minute,second:Zu+" "+Xu.second,millisecond:Xu.none},Ku=["year","month","day","hour","minute","second","millisecond"],Qu=["year","half-year","quarter","month","week","half-week","day","half-day","quarter-day","hour","minute","second","millisecond"];function tc(t,e){return t+="","0000".substr(0,e-t.length)+t}function ec(t){switch(t){case"half-year":case"quarter":return"month";case"week":case"half-week":return"day";case"half-day":case"quarter-day":return"hour";default:return t}}function nc(t){return t===ec(t)}function rc(t){switch(t){case"year":case"month":return"day";case"millisecond":return"millisecond";default:return"second"}}function ic(t,e,n,r){var i=Xr(t),o=i[lc(n)](),a=i[uc(n)]()+1,s=Math.floor((a-1)/3)+1,l=i[cc(n)](),u=i["get"+(n?"UTC":"")+"Day"](),c=i[dc(n)](),d=(c-1)%12+1,h=i[hc(n)](),f=i[fc(n)](),p=i[pc(n)](),m=r instanceof Lu?r:Bu(r||Hu)||Wu(),g=m.getModel("time"),y=g.get("month"),v=g.get("monthAbbr"),_=g.get("dayOfWeek"),b=g.get("dayOfWeekAbbr");return(e||"").replace(/{yyyy}/g,o+"").replace(/{yy}/g,tc(o%100+"",2)).replace(/{Q}/g,s+"").replace(/{MMMM}/g,y[a-1]).replace(/{MMM}/g,v[a-1]).replace(/{MM}/g,tc(a,2)).replace(/{M}/g,a+"").replace(/{dd}/g,tc(l,2)).replace(/{d}/g,l+"").replace(/{eeee}/g,_[u]).replace(/{ee}/g,b[u]).replace(/{e}/g,u+"").replace(/{HH}/g,tc(c,2)).replace(/{H}/g,c+"").replace(/{hh}/g,tc(d+"",2)).replace(/{h}/g,d+"").replace(/{mm}/g,tc(h,2)).replace(/{m}/g,h+"").replace(/{ss}/g,tc(f,2)).replace(/{s}/g,f+"").replace(/{SSS}/g,tc(p,3)).replace(/{S}/g,p+"")}function oc(t,e,n,r,i){var o=null;if(s["C"](n))o=n;else if(s["w"](n))o=n(t.value,e,{level:t.level});else{var a=s["m"]({},Xu);if(t.level>0)for(var l=0;l<Ku.length;++l)a[Ku[l]]="{primary|"+a[Ku[l]]+"}";var u=n?!1===n.inherit?n:s["i"](n,a):a,c=ac(t.value,i);if(u[c])o=u[c];else if(u.inherit){var d=Qu.indexOf(c);for(l=d-1;l>=0;--l)if(u[c]){o=u[c];break}o=o||a.none}if(s["t"](o)){var h=null==t.level?0:t.level>=0?t.level:o.length+t.level;h=Math.min(h,o.length-1),o=o[h]}}return ic(new Date(t.value),o,i,r)}function ac(t,e){var n=Xr(t),r=n[uc(e)]()+1,i=n[cc(e)](),o=n[dc(e)](),a=n[hc(e)](),s=n[fc(e)](),l=n[pc(e)](),u=0===l,c=u&&0===s,d=c&&0===a,h=d&&0===o,f=h&&1===i,p=f&&1===r;return p?"year":f?"month":h?"day":d?"hour":c?"minute":u?"second":"millisecond"}function sc(t,e,n){var r=s["z"](t)?Xr(t):t;switch(e=e||ac(t,n),e){case"year":return r[lc(n)]();case"half-year":return r[uc(n)]()>=6?1:0;case"quarter":return Math.floor((r[uc(n)]()+1)/4);case"month":return r[uc(n)]();case"day":return r[cc(n)]();case"half-day":return r[dc(n)]()/24;case"hour":return r[dc(n)]();case"minute":return r[hc(n)]();case"second":return r[fc(n)]();case"millisecond":return r[pc(n)]()}}function lc(t){return t?"getUTCFullYear":"getFullYear"}function uc(t){return t?"getUTCMonth":"getMonth"}function cc(t){return t?"getUTCDate":"getDate"}function dc(t){return t?"getUTCHours":"getHours"}function hc(t){return t?"getUTCMinutes":"getMinutes"}function fc(t){return t?"getUTCSeconds":"getSeconds"}function pc(t){return t?"getUTCMilliseconds":"getMilliseconds"}function mc(t){return t?"setUTCFullYear":"setFullYear"}function gc(t){return t?"setUTCMonth":"setMonth"}function yc(t){return t?"setUTCDate":"setDate"}function vc(t){return t?"setUTCHours":"setHours"}function _c(t){return t?"setUTCMinutes":"setMinutes"}function bc(t){return t?"setUTCSeconds":"setSeconds"}function xc(t){return t?"setUTCMilliseconds":"setMilliseconds"}function wc(t){if(!ni(t))return s["C"](t)?t:"-";var e=(t+"").split(".");return e[0].replace(/(\d{1,3})(?=(?:\d{3})+(?!\d))/g,"$1,")+(e.length>1?"."+e[1]:"")}function Mc(t,e){return t=(t||"").toLowerCase().replace(/-(.)/g,(function(t,e){return e.toUpperCase()})),e&&t&&(t=t.charAt(0).toUpperCase()+t.slice(1)),t}var Sc=s["M"];function kc(t,e,n){var r="{yyyy}-{MM}-{dd} {HH}:{mm}:{ss}";function i(t){return t&&s["T"](t)?t:"-"}function o(t){return!(null==t||isNaN(t)||!isFinite(t))}var a="time"===e,l=t instanceof Date;if(a||l){var u=a?Xr(t):t;if(!isNaN(+u))return ic(u,r,n);if(l)return"-"}if("ordinal"===e)return s["D"](t)?i(t):s["z"](t)&&o(t)?t+"":"-";var c=ei(t);return o(c)?wc(c):s["D"](t)?i(t):"boolean"===typeof t?t+"":"-"}var Lc=["a","b","c","d","e","f","g"],Tc=function(t,e){return"{"+t+(null==e?"":e)+"}"};function Dc(t,e,n){s["t"](e)||(e=[e]);var r=e.length;if(!r)return"";for(var i=e[0].$vars||[],o=0;o<i.length;o++){var a=Lc[o];t=t.replace(Tc(a),Tc(a,0))}for(var l=0;l<r;l++)for(var u=0;u<i.length;u++){var c=e[l][i[u]];t=t.replace(Tc(Lc[u],l),n?V(c):c)}return t}function Oc(t,e,n){return s["k"](e,(function(e,r){t=t.replace("{"+r+"}",n?V(e):e)})),t}function Ic(t,e){var n=s["C"](t)?{color:t,extraCssText:e}:t||{},r=n.color,i=n.type;e=n.extraCssText;var o=n.renderMode||"html";if(!r)return"";if("html"===o)return"subItem"===i?'<span style="display:inline-block;vertical-align:middle;margin-right:8px;margin-left:3px;border-radius:4px;width:4px;height:4px;background-color:'+V(r)+";"+(e||"")+'"></span>':'<span style="display:inline-block;margin-right:4px;border-radius:10px;width:10px;height:10px;background-color:'+V(r)+";"+(e||"")+'"></span>';var a=n.markerId||"markerX";return{renderMode:o,content:"{"+a+"|}  ",style:"subItem"===i?{width:4,height:4,borderRadius:2,backgroundColor:r}:{width:10,height:10,borderRadius:5,backgroundColor:r}}}function Cc(t,e){return e=e||"transparent",s["C"](t)?t:s["A"](t)&&t.colorStops&&(t.colorStops[0]||{}).color||e}function Ac(t,e){if("_blank"===e||"blank"===e){var n=window.open();n.opener=null,n.location.href=t}else window.open(t,e)}var Yc=s["k"],jc=["left","right","top","bottom","width","height"],Pc=[["width","left","right"],["height","top","bottom"]];function Ec(t,e,n,r,i){var o=0,a=0;null==r&&(r=1/0),null==i&&(i=1/0);var s=0;e.eachChild((function(l,u){var c,d,h=l.getBoundingRect(),f=e.childAt(u+1),p=f&&f.getBoundingRect();if("horizontal"===t){var m=h.width+(p?-p.x+h.x:0);c=o+m,c>r||l.newline?(o=0,c=m,a+=s+n,s=h.height):s=Math.max(s,h.height)}else{var g=h.height+(p?-p.y+h.y:0);d=a+g,d>i||l.newline?(o+=s+n,a=0,d=g,s=h.width):s=Math.max(s,h.width)}l.newline||(l.x=o,l.y=a,l.markRedraw(),"horizontal"===t?o=c+n:a=d+n)}))}var Nc=Ec;s["h"](Ec,"vertical"),s["h"](Ec,"horizontal");function Rc(t,e,n){var r=e.width,i=e.height,o=Nr(t.left,r),a=Nr(t.top,i),s=Nr(t.right,r),l=Nr(t.bottom,i);return(isNaN(o)||isNaN(parseFloat(t.left)))&&(o=0),(isNaN(s)||isNaN(parseFloat(t.right)))&&(s=r),(isNaN(a)||isNaN(parseFloat(t.top)))&&(a=0),(isNaN(l)||isNaN(parseFloat(t.bottom)))&&(l=i),n=Sc(n||0),{width:Math.max(s-o-n[1]-n[3],0),height:Math.max(l-a-n[0]-n[2],0)}}function Hc(t,e,n){n=Sc(n||0);var r=e.width,i=e.height,o=Nr(t.left,r),a=Nr(t.top,i),s=Nr(t.right,r),l=Nr(t.bottom,i),u=Nr(t.width,r),c=Nr(t.height,i),d=n[2]+n[0],h=n[1]+n[3],f=t.aspect;switch(isNaN(u)&&(u=r-s-h-o),isNaN(c)&&(c=i-l-d-a),null!=f&&(isNaN(u)&&isNaN(c)&&(f>r/i?u=.8*r:c=.8*i),isNaN(u)&&(u=f*c),isNaN(c)&&(c=u/f)),isNaN(o)&&(o=r-s-u-h),isNaN(a)&&(a=i-l-c-d),t.left||t.right){case"center":o=r/2-u/2-n[3];break;case"right":o=r-u-h;break}switch(t.top||t.bottom){case"middle":case"center":a=i/2-c/2-n[0];break;case"bottom":a=i-c-d;break}o=o||0,a=a||0,isNaN(u)&&(u=r-h-o-(s||0)),isNaN(c)&&(c=i-d-a-(l||0));var p=new Tt(o+n[3],a+n[0],u,c);return p.margin=n,p}function zc(t,e,n,r,i,o){var a,l=!i||!i.hv||i.hv[0],u=!i||!i.hv||i.hv[1],c=i&&i.boundingMode||"all";if(o=o||t,o.x=t.x,o.y=t.y,!l&&!u)return!1;if("raw"===c)a="group"===t.type?new Tt(0,0,+e.width||0,+e.height||0):t.getBoundingRect();else if(a=t.getBoundingRect(),t.needLocalTransform()){var d=t.getLocalTransform();a=a.clone(),a.applyTransform(d)}var h=Hc(s["i"]({width:a.width,height:a.height},e),n,r),f=l?h.x-a.x:0,p=u?h.y-a.y:0;return"raw"===c?(o.x=f,o.y=p):(o.x+=f,o.y+=p),o===t&&t.markRedraw(),!0}function Fc(t,e){return null!=t[Pc[e][0]]||null!=t[Pc[e][1]]&&null!=t[Pc[e][2]]}function Bc(t){var e=t.layoutMode||t.constructor.layoutMode;return s["A"](e)?e:e?{type:e}:null}function Wc(t,e,n){var r=n&&n.ignoreSize;!s["t"](r)&&(r=[r,r]);var i=a(Pc[0],0),o=a(Pc[1],1);function a(n,i){var o={},a=0,s={},c=0,d=2;if(Yc(n,(function(e){s[e]=t[e]})),Yc(n,(function(t){l(e,t)&&(o[t]=s[t]=e[t]),u(o,t)&&a++,u(s,t)&&c++})),r[i])return u(e,n[1])?s[n[2]]=null:u(e,n[2])&&(s[n[1]]=null),s;if(c!==d&&a){if(a>=d)return o;for(var h=0;h<n.length;h++){var f=n[h];if(!l(o,f)&&l(t,f)){o[f]=t[f];break}}return o}return s}function l(t,e){return t.hasOwnProperty(e)}function u(t,e){return null!=t[e]&&"auto"!==t[e]}function c(t,e,n){Yc(t,(function(t){e[t]=n[t]}))}c(Pc[0],t,i),c(Pc[1],t,o)}function Vc(t){return Gc({},t)}function Gc(t,e){return e&&t&&Yc(jc,(function(n){e.hasOwnProperty(n)&&(t[n]=e[n])})),t}var $c=Ci(),Uc=function(t){function e(e,n,r){var i=t.call(this,e,n,r)||this;return i.uid=Du("ec_cpt_model"),i}return o(e,t),e.prototype.init=function(t,e,n){this.mergeDefaultAndTheme(t,n)},e.prototype.mergeDefaultAndTheme=function(t,e){var n=Bc(this),r=n?Vc(t):{},i=e.getTheme();s["I"](t,i.get(this.mainType)),s["I"](t,this.getDefaultOption()),n&&Wc(t,r,n)},e.prototype.mergeOption=function(t,e){s["I"](this.option,t,!0);var n=Bc(this);n&&Wc(this.option,t,n)},e.prototype.optionUpdated=function(t,e){},e.prototype.getDefaultOption=function(){var t=this.constructor;if(!qi(t))return t.defaultOption;var e=$c(this);if(!e.defaultOption){var n=[],r=t;while(r){var i=r.prototype.defaultOption;i&&n.push(i),r=r.superClass}for(var o={},a=n.length-1;a>=0;a--)o=s["I"](o,n[a],!0);e.defaultOption=o}return e.defaultOption},e.prototype.getReferringComponents=function(t,e){var n=t+"Index",r=t+"Id";return Ni(this.ecModel,t,{index:this.get(n,!0),id:this.get(r,!0)},e)},e.prototype.getBoxLayoutParams=function(){var t=this;return{left:t.get("left"),top:t.get("top"),right:t.get("right"),bottom:t.get("bottom"),width:t.get("width"),height:t.get("height")}},e.prototype.getZLevelKey=function(){return""},e.prototype.setZLevel=function(t){this.option.zlevel=t},e.protoInitialize=function(){var t=e.prototype;t.type="component",t.id="",t.name="",t.mainType="",t.subType="",t.componentIndex=0}(),e}(Lu);function qc(t){var e=[];return s["k"](Uc.getClassesByMainType(t),(function(t){e=e.concat(t.dependencies||t.prototype.dependencies||[])})),e=s["H"](e,(function(t){return $i(t).main})),"dataset"!==t&&s["r"](e,"dataset")<=0&&e.unshift("dataset"),e}Ji(Uc,Lu),no(Uc),Ou(Uc),Iu(Uc,qc);var Xc=Uc,Zc="";"undefined"!==typeof navigator&&(Zc=navigator.platform||"");var Jc="rgba(0, 0, 0, 0.2)",Kc={darkMode:"auto",colorBy:"series",color:["#5470c6","#91cc75","#fac858","#ee6666","#73c0de","#3ba272","#fc8452","#9a60b4","#ea7ccc"],gradientColor:["#f6efa6","#d88273","#bf444c"],aria:{decal:{decals:[{color:Jc,dashArrayX:[1,0],dashArrayY:[2,5],symbolSize:1,rotation:Math.PI/6},{color:Jc,symbol:"circle",dashArrayX:[[8,8],[0,8,8,0]],dashArrayY:[6,0],symbolSize:.8},{color:Jc,dashArrayX:[1,0],dashArrayY:[4,3],rotation:-Math.PI/4},{color:Jc,dashArrayX:[[6,6],[0,6,6,0]],dashArrayY:[6,0]},{color:Jc,dashArrayX:[[1,0],[1,6]],dashArrayY:[1,0,6,0],rotation:Math.PI/4},{color:Jc,symbol:"triangle",dashArrayX:[[9,9],[0,9,9,0]],dashArrayY:[7,2],symbolSize:.75}]}},textStyle:{fontFamily:Zc.match(/^Win/)?"Microsoft YaHei":"sans-serif",fontSize:12,fontStyle:"normal",fontWeight:"normal"},blendMode:null,stateAnimation:{duration:300,easing:"cubicOut"},animation:"auto",animationDuration:1e3,animationDurationUpdate:500,animationEasing:"cubicInOut",animationEasingUpdate:"cubicInOut",animationThreshold:2e3,progressiveThreshold:3e3,progressive:400,hoverLayerThreshold:3e3,useUTC:!1},Qc=Object(s["f"])(["tooltip","label","itemName","itemId","itemGroupId","itemChildGroupId","seriesName"]),td="original",ed="arrayRows",nd="objectRows",rd="keyedColumns",id="typedArray",od="unknown",ad="column",sd="row",ld={Must:1,Might:2,Not:3},ud=Ci();function cd(t){ud(t).datasetMap=Object(s["f"])()}function dd(t,e,n){var r={},i=fd(e);if(!i||!t)return r;var o,a,l=[],u=[],c=e.ecModel,d=ud(c).datasetMap,h=i.uid+"_"+n.seriesLayoutBy;t=t.slice(),Object(s["k"])(t,(function(e,n){var i=Object(s["A"])(e)?e:t[n]={name:e};"ordinal"===i.type&&null==o&&(o=n,a=m(i)),r[i.name]=[]}));var f=d.get(h)||d.set(h,{categoryWayDim:a,valueWayDim:0});function p(t,e,n){for(var r=0;r<n;r++)t.push(e+r)}function m(t){var e=t.dimsDef;return e?e.length:1}return Object(s["k"])(t,(function(t,e){var n=t.name,i=m(t);if(null==o){var a=f.valueWayDim;p(r[n],a,i),p(u,a,i),f.valueWayDim+=i}else if(o===e)p(r[n],0,i),p(l,0,i);else{a=f.categoryWayDim;p(r[n],a,i),p(u,a,i),f.categoryWayDim+=i}})),l.length&&(r.itemName=l),u.length&&(r.seriesName=u),r}function hd(t,e,n){var r={},i=fd(t);if(!i)return r;var o,a=e.sourceFormat,l=e.dimensionsDefine;a!==nd&&a!==rd||Object(s["k"])(l,(function(t,e){"name"===(Object(s["A"])(t)?t.name:t)&&(o=e)}));var u=function(){for(var t={},r={},i=[],s=0,u=Math.min(5,n);s<u;s++){var c=gd(e.data,a,e.seriesLayoutBy,l,e.startIndex,s);i.push(c);var d=c===ld.Not;if(d&&null==t.v&&s!==o&&(t.v=s),(null==t.n||t.n===t.v||!d&&i[t.n]===ld.Not)&&(t.n=s),h(t)&&i[t.n]!==ld.Not)return t;d||(c===ld.Might&&null==r.v&&s!==o&&(r.v=s),null!=r.n&&r.n!==r.v||(r.n=s))}function h(t){return null!=t.v&&null!=t.n}return h(t)?t:h(r)?r:null}();if(u){r.value=[u.v];var c=null!=o?o:u.n;r.itemName=[c],r.seriesName=[c]}return r}function fd(t){var e=t.get("data",!0);if(!e)return Ni(t.ecModel,"dataset",{index:t.get("datasetIndex",!0),id:t.get("datasetId",!0)},Pi).models[0]}function pd(t){return t.get("transform",!0)||t.get("fromTransformResult",!0)?Ni(t.ecModel,"dataset",{index:t.get("fromDatasetIndex",!0),id:t.get("fromDatasetId",!0)},Pi).models:[]}function md(t,e){return gd(t.data,t.sourceFormat,t.seriesLayoutBy,t.dimensionsDefine,t.startIndex,e)}function gd(t,e,n,r,i,o){var a,l,u,c=5;if(Object(s["E"])(t))return ld.Not;if(r){var d=r[o];Object(s["A"])(d)?(l=d.name,u=d.type):Object(s["C"])(d)&&(l=d)}if(null!=u)return"ordinal"===u?ld.Must:ld.Not;if(e===ed){var h=t;if(n===sd){for(var f=h[o],p=0;p<(f||[]).length&&p<c;p++)if(null!=(a=x(f[i+p])))return a}else for(p=0;p<h.length&&p<c;p++){var m=h[i+p];if(m&&null!=(a=x(m[o])))return a}}else if(e===nd){var g=t;if(!l)return ld.Not;for(p=0;p<g.length&&p<c;p++){var y=g[p];if(y&&null!=(a=x(y[l])))return a}}else if(e===rd){var v=t;if(!l)return ld.Not;f=v[l];if(!f||Object(s["E"])(f))return ld.Not;for(p=0;p<f.length&&p<c;p++)if(null!=(a=x(f[p])))return a}else if(e===td){var _=t;for(p=0;p<_.length&&p<c;p++){y=_[p];var b=hi(y);if(!Object(s["t"])(b))return ld.Not;if(null!=(a=x(b[o])))return a}}function x(t){var e=Object(s["C"])(t);return null!=t&&isFinite(t)&&""!==t?e?ld.Might:ld.Not:e&&"-"!==t?ld.Must:void 0}return ld.Not}var yd=Object(s["f"])();function vd(t,e){Object(s["b"])(null==yd.get(t)&&e),yd.set(t,e)}function _d(t,e,n){var r=yd.get(e);if(!r)return n;var i=r(t);return i?n.concat(i):n}var bd=Ci(),xd=Ci(),wd=function(){function t(){}return t.prototype.getColorFromPalette=function(t,e,n){var r=ui(this.get("color",!0)),i=this.get("colorLayer",!0);return kd(this,bd,r,i,t,e,n)},t.prototype.clearColorPalette=function(){Ld(this,bd)},t}();function Md(t,e,n,r){var i=ui(t.get(["aria","decal","decals"]));return kd(t,xd,i,null,e,n,r)}function Sd(t,e){for(var n=t.length,r=0;r<n;r++)if(t[r].length>e)return t[r];return t[n-1]}function kd(t,e,n,r,i,o,a){o=o||t;var s=e(o),l=s.paletteIdx||0,u=s.paletteNameMap=s.paletteNameMap||{};if(u.hasOwnProperty(i))return u[i];var c=null!=a&&r?Sd(r,a):n;if(c=c||n,c&&c.length){var d=c[l];return i&&(u[i]=d),s.paletteIdx=(l+1)%c.length,d}}function Ld(t,e){e(t).paletteIdx=0,e(t).paletteNameMap={}}var Td,Dd,Od;"undefined"!==typeof console&&console.warn&&console.log;function Id(t){0}function Cd(t){throw new Error(t)}var Ad="\0_ec_inner",Yd=1;var jd=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.init=function(t,e,n,r,i,o){r=r||{},this.option=null,this._theme=new Lu(r),this._locale=new Lu(i),this._optionManager=o},e.prototype.setOption=function(t,e,n){var r=Hd(e);this._optionManager.setOption(t,n,r),this._resetOption(null,r)},e.prototype.resetOption=function(t,e){return this._resetOption(t,Hd(e))},e.prototype._resetOption=function(t,e){var n=!1,r=this._optionManager;if(!t||"recreate"===t){var i=r.mountOption("recreate"===t);0,this.option&&"recreate"!==t?(this.restoreData(),this._mergeOption(i,e)):Od(this,i),n=!0}if("timeline"!==t&&"media"!==t||this.restoreData(),!t||"recreate"===t||"timeline"===t){var o=r.getTimelineOption(this);o&&(n=!0,this._mergeOption(o,e))}if(!t||"recreate"===t||"media"===t){var a=r.getMediaOption(this);a.length&&Object(s["k"])(a,(function(t){n=!0,this._mergeOption(t,e)}),this)}return n},e.prototype.mergeOption=function(t){this._mergeOption(t,null)},e.prototype._mergeOption=function(t,e){var n=this.option,r=this._componentsMap,i=this._componentsCount,o=[],a=Object(s["f"])(),l=e&&e.replaceMergeMainTypeMap;function u(e){var o=_d(this,e,ui(t[e])),a=r.get(e),u=a?l&&l.get(e)?"replaceMerge":"normalMerge":"replaceAll",c=pi(a,o,u);Ti(c,e,Xc),n[e]=null,r.set(e,null),i.set(e,0);var d,h=[],f=[],p=0;Object(s["k"])(c,(function(t,n){var r=t.existing,i=t.newOption;if(i){var o="series"===e,a=Xc.getClass(e,t.keyInfo.subType,!o);if(!a)return;if("tooltip"===e){if(d)return void 0;d=!0}if(r&&r.constructor===a)r.name=t.keyInfo.name,r.mergeOption(i,this),r.optionUpdated(i,!1);else{var l=Object(s["m"])({componentIndex:n},t.keyInfo);r=new a(i,this,this,l),Object(s["m"])(r,l),t.brandNew&&(r.__requireNewView=!0),r.init(i,this,this),r.optionUpdated(null,!0)}}else r&&(r.mergeOption({},this),r.optionUpdated({},!1));r?(h.push(r.option),f.push(r),p++):(h.push(void 0),f.push(void 0))}),this),n[e]=h,r.set(e,f),i.set(e,p),"series"===e&&Td(this)}cd(this),Object(s["k"])(t,(function(t,e){null!=t&&(Xc.hasClass(e)?e&&(o.push(e),a.set(e,!0)):n[e]=null==n[e]?Object(s["d"])(t):Object(s["I"])(n[e],t,!0))})),l&&l.each((function(t,e){Xc.hasClass(e)&&!a.get(e)&&(o.push(e),a.set(e,!0))})),Xc.topologicalTravel(o,Xc.getAllClassMainTypes(),u,this),this._seriesIndices||Td(this)},e.prototype.getOption=function(){var t=Object(s["d"])(this.option);return Object(s["k"])(t,(function(e,n){if(Xc.hasClass(n)){for(var r=ui(e),i=r.length,o=!1,a=i-1;a>=0;a--)r[a]&&!ki(r[a])?o=!0:(r[a]=null,!o&&i--);r.length=i,t[n]=r}})),delete t[Ad],t},e.prototype.getTheme=function(){return this._theme},e.prototype.getLocaleModel=function(){return this._locale},e.prototype.setUpdatePayload=function(t){this._payload=t},e.prototype.getUpdatePayload=function(){return this._payload},e.prototype.getComponent=function(t,e){var n=this._componentsMap.get(t);if(n){var r=n[e||0];if(r)return r;if(null==e)for(var i=0;i<n.length;i++)if(n[i])return n[i]}},e.prototype.queryComponents=function(t){var e=t.mainType;if(!e)return[];var n,r=t.index,i=t.id,o=t.name,a=this._componentsMap.get(e);return a&&a.length?(null!=r?(n=[],Object(s["k"])(ui(r),(function(t){a[t]&&n.push(a[t])}))):n=null!=i?Nd("id",i,a):null!=o?Nd("name",o,a):Object(s["n"])(a,(function(t){return!!t})),Rd(n,t)):[]},e.prototype.findComponents=function(t){var e=t.query,n=t.mainType,r=o(e),i=r?this.queryComponents(r):Object(s["n"])(this._componentsMap.get(n),(function(t){return!!t}));return a(Rd(i,t));function o(t){var e=n+"Index",r=n+"Id",i=n+"Name";return!t||null==t[e]&&null==t[r]&&null==t[i]?null:{mainType:n,index:t[e],id:t[r],name:t[i]}}function a(e){return t.filter?Object(s["n"])(e,t.filter):e}},e.prototype.eachComponent=function(t,e,n){var r=this._componentsMap;if(Object(s["w"])(t)){var i=e,o=t;r.each((function(t,e){for(var n=0;t&&n<t.length;n++){var r=t[n];r&&o.call(i,e,r,r.componentIndex)}}))}else for(var a=Object(s["C"])(t)?r.get(t):Object(s["A"])(t)?this.findComponents(t):null,l=0;a&&l<a.length;l++){var u=a[l];u&&e.call(n,u,u.componentIndex)}},e.prototype.getSeriesByName=function(t){var e=Mi(t,null);return Object(s["n"])(this._componentsMap.get("series"),(function(t){return!!t&&null!=e&&t.name===e}))},e.prototype.getSeriesByIndex=function(t){return this._componentsMap.get("series")[t]},e.prototype.getSeriesByType=function(t){return Object(s["n"])(this._componentsMap.get("series"),(function(e){return!!e&&e.subType===t}))},e.prototype.getSeries=function(){return Object(s["n"])(this._componentsMap.get("series"),(function(t){return!!t}))},e.prototype.getSeriesCount=function(){return this._componentsCount.get("series")},e.prototype.eachSeries=function(t,e){Dd(this),Object(s["k"])(this._seriesIndices,(function(n){var r=this._componentsMap.get("series")[n];t.call(e,r,n)}),this)},e.prototype.eachRawSeries=function(t,e){Object(s["k"])(this._componentsMap.get("series"),(function(n){n&&t.call(e,n,n.componentIndex)}))},e.prototype.eachSeriesByType=function(t,e,n){Dd(this),Object(s["k"])(this._seriesIndices,(function(r){var i=this._componentsMap.get("series")[r];i.subType===t&&e.call(n,i,r)}),this)},e.prototype.eachRawSeriesByType=function(t,e,n){return Object(s["k"])(this.getSeriesByType(t),e,n)},e.prototype.isSeriesFiltered=function(t){return Dd(this),null==this._seriesIndicesMap.get(t.componentIndex)},e.prototype.getCurrentSeriesIndices=function(){return(this._seriesIndices||[]).slice()},e.prototype.filterSeries=function(t,e){Dd(this);var n=[];Object(s["k"])(this._seriesIndices,(function(r){var i=this._componentsMap.get("series")[r];t.call(e,i,r)&&n.push(r)}),this),this._seriesIndices=n,this._seriesIndicesMap=Object(s["f"])(n)},e.prototype.restoreData=function(t){Td(this);var e=this._componentsMap,n=[];e.each((function(t,e){Xc.hasClass(e)&&n.push(e)})),Xc.topologicalTravel(n,Xc.getAllClassMainTypes(),(function(n){Object(s["k"])(e.get(n),(function(e){!e||"series"===n&&Pd(e,t)||e.restoreData()}))}))},e.internalField=function(){Td=function(t){var e=t._seriesIndices=[];Object(s["k"])(t._componentsMap.get("series"),(function(t){t&&e.push(t.componentIndex)})),t._seriesIndicesMap=Object(s["f"])(e)},Dd=function(t){0},Od=function(t,e){t.option={},t.option[Ad]=Yd,t._componentsMap=Object(s["f"])({series:[]}),t._componentsCount=Object(s["f"])();var n=e.aria;Object(s["A"])(n)&&null==n.enabled&&(n.enabled=!0),Ed(e,t._theme.option),Object(s["I"])(e,Kc,!1),t._mergeOption(e,null)}}(),e}(Lu);function Pd(t,e){if(e){var n=e.seriesIndex,r=e.seriesId,i=e.seriesName;return null!=n&&t.componentIndex!==n||null!=r&&t.id!==r||null!=i&&t.name!==i}}function Ed(t,e){var n=t.color&&!t.colorLayer;Object(s["k"])(e,(function(e,r){"colorLayer"===r&&n||Xc.hasClass(r)||("object"===typeof e?t[r]=t[r]?Object(s["I"])(t[r],e,!1):Object(s["d"])(e):null==t[r]&&(t[r]=e))}))}function Nd(t,e,n){if(Object(s["t"])(e)){var r=Object(s["f"])();return Object(s["k"])(e,(function(t){if(null!=t){var e=Mi(t,null);null!=e&&r.set(t,!0)}})),Object(s["n"])(n,(function(e){return e&&r.get(e[t])}))}var i=Mi(e,null);return Object(s["n"])(n,(function(e){return e&&null!=i&&e[t]===i}))}function Rd(t,e){return e.hasOwnProperty("subType")?Object(s["n"])(t,(function(t){return t&&t.subType===e.subType})):t}function Hd(t){var e=Object(s["f"])();return t&&Object(s["k"])(ui(t.replaceMerge),(function(t){e.set(t,!0)})),{replaceMergeMainTypeMap:e}}Object(s["K"])(jd,wd);var zd=jd,Fd=["getDom","getZr","getWidth","getHeight","getDevicePixelRatio","dispatchAction","isSSR","isDisposed","on","off","getDataURL","getConnectedDataURL","getOption","getId","updateLabelLayout"],Bd=function(){function t(t){s["k"](Fd,(function(e){this[e]=s["c"](t[e],t)}),this)}return t}(),Wd=Bd,Vd={},Gd=function(){function t(){this._coordinateSystems=[]}return t.prototype.create=function(t,e){var n=[];s["k"](Vd,(function(r,i){var o=r.create(t,e);n=n.concat(o||[])})),this._coordinateSystems=n},t.prototype.update=function(t,e){s["k"](this._coordinateSystems,(function(n){n.update&&n.update(t,e)}))},t.prototype.getCoordinateSystems=function(){return this._coordinateSystems.slice()},t.register=function(t,e){Vd[t]=e},t.get=function(t){return Vd[t]},t}(),$d=Gd,Ud=/^(min|max)?(.+)$/,qd=function(){function t(t){this._timelineOptions=[],this._mediaList=[],this._currentMediaIndices=[],this._api=t}return t.prototype.setOption=function(t,e,n){t&&(Object(s["k"])(ui(t.series),(function(t){t&&t.data&&Object(s["E"])(t.data)&&Object(s["R"])(t.data)})),Object(s["k"])(ui(t.dataset),(function(t){t&&t.source&&Object(s["E"])(t.source)&&Object(s["R"])(t.source)}))),t=Object(s["d"])(t);var r=this._optionBackup,i=Xd(t,e,!r);this._newBaseOption=i.baseOption,r?(i.timelineOptions.length&&(r.timelineOptions=i.timelineOptions),i.mediaList.length&&(r.mediaList=i.mediaList),i.mediaDefault&&(r.mediaDefault=i.mediaDefault)):this._optionBackup=i},t.prototype.mountOption=function(t){var e=this._optionBackup;return this._timelineOptions=e.timelineOptions,this._mediaList=e.mediaList,this._mediaDefault=e.mediaDefault,this._currentMediaIndices=[],Object(s["d"])(t?e.baseOption:this._newBaseOption)},t.prototype.getTimelineOption=function(t){var e,n=this._timelineOptions;if(n.length){var r=t.getComponent("timeline");r&&(e=Object(s["d"])(n[r.getCurrentIndex()]))}return e},t.prototype.getMediaOption=function(t){var e=this._api.getWidth(),n=this._api.getHeight(),r=this._mediaList,i=this._mediaDefault,o=[],a=[];if(!r.length&&!i)return a;for(var l=0,u=r.length;l<u;l++)Zd(r[l].query,e,n)&&o.push(l);return!o.length&&i&&(o=[-1]),o.length&&!Kd(o,this._currentMediaIndices)&&(a=Object(s["H"])(o,(function(t){return Object(s["d"])(-1===t?i.option:r[t].option)}))),this._currentMediaIndices=o,a},t}();function Xd(t,e,n){var r,i,o=[],a=t.baseOption,l=t.timeline,u=t.options,c=t.media,d=!!t.media,h=!!(u||l||a&&a.timeline);function f(t){Object(s["k"])(e,(function(e){e(t,n)}))}return a?(i=a,i.timeline||(i.timeline=l)):((h||d)&&(t.options=t.media=null),i=t),d&&Object(s["t"])(c)&&Object(s["k"])(c,(function(t){t&&t.option&&(t.query?o.push(t):r||(r=t))})),f(i),Object(s["k"])(u,(function(t){return f(t)})),Object(s["k"])(o,(function(t){return f(t.option)})),{baseOption:i,timelineOptions:u||[],mediaDefault:r,mediaList:o}}function Zd(t,e,n){var r={width:e,height:n,aspectratio:e/n},i=!0;return Object(s["k"])(t,(function(t,e){var n=e.match(Ud);if(n&&n[1]&&n[2]){var o=n[1],a=n[2].toLowerCase();Jd(r[a],t,o)||(i=!1)}})),i}function Jd(t,e,n){return"min"===n?t>=e:"max"===n?t<=e:t===e}function Kd(t,e){return t.join(",")===e.join(",")}var Qd=qd,th=s["k"],eh=s["A"],nh=["areaStyle","lineStyle","nodeStyle","linkStyle","chordStyle","label","labelLine"];function rh(t){var e=t&&t.itemStyle;if(e)for(var n=0,r=nh.length;n<r;n++){var i=nh[n],o=e.normal,a=e.emphasis;o&&o[i]&&(t[i]=t[i]||{},t[i].normal?s["I"](t[i].normal,o[i]):t[i].normal=o[i],o[i]=null),a&&a[i]&&(t[i]=t[i]||{},t[i].emphasis?s["I"](t[i].emphasis,a[i]):t[i].emphasis=a[i],a[i]=null)}}function ih(t,e,n){if(t&&t[e]&&(t[e].normal||t[e].emphasis)){var r=t[e].normal,i=t[e].emphasis;r&&(n?(t[e].normal=t[e].emphasis=null,s["i"](t[e],r)):t[e]=r),i&&(t.emphasis=t.emphasis||{},t.emphasis[e]=i,i.focus&&(t.emphasis.focus=i.focus),i.blurScope&&(t.emphasis.blurScope=i.blurScope))}}function oh(t){ih(t,"itemStyle"),ih(t,"lineStyle"),ih(t,"areaStyle"),ih(t,"label"),ih(t,"labelLine"),ih(t,"upperLabel"),ih(t,"edgeLabel")}function ah(t,e){var n=eh(t)&&t[e],r=eh(n)&&n.textStyle;if(r){0;for(var i=0,o=di.length;i<o;i++){var a=di[i];r.hasOwnProperty(a)&&(n[a]=r[a])}}}function sh(t){t&&(oh(t),ah(t,"label"),t.emphasis&&ah(t.emphasis,"label"))}function lh(t){if(eh(t)){rh(t),oh(t),ah(t,"label"),ah(t,"upperLabel"),ah(t,"edgeLabel"),t.emphasis&&(ah(t.emphasis,"label"),ah(t.emphasis,"upperLabel"),ah(t.emphasis,"edgeLabel"));var e=t.markPoint;e&&(rh(e),sh(e));var n=t.markLine;n&&(rh(n),sh(n));var r=t.markArea;r&&sh(r);var i=t.data;if("graph"===t.type){i=i||t.nodes;var o=t.links||t.edges;if(o&&!s["E"](o))for(var a=0;a<o.length;a++)sh(o[a]);s["k"](t.categories,(function(t){oh(t)}))}if(i&&!s["E"](i))for(a=0;a<i.length;a++)sh(i[a]);if(e=t.markPoint,e&&e.data){var l=e.data;for(a=0;a<l.length;a++)sh(l[a])}if(n=t.markLine,n&&n.data){var u=n.data;for(a=0;a<u.length;a++)s["t"](u[a])?(sh(u[a][0]),sh(u[a][1])):sh(u[a])}"gauge"===t.type?(ah(t,"axisLabel"),ah(t,"title"),ah(t,"detail")):"treemap"===t.type?(ih(t.breadcrumb,"itemStyle"),s["k"](t.levels,(function(t){oh(t)}))):"tree"===t.type&&oh(t.leaves)}}function uh(t){return s["t"](t)?t:t?[t]:[]}function ch(t){return(s["t"](t)?t[0]:t)||{}}function dh(t,e){th(uh(t.series),(function(t){eh(t)&&lh(t)}));var n=["xAxis","yAxis","radiusAxis","angleAxis","singleAxis","parallelAxis","radar"];e&&n.push("valueAxis","categoryAxis","logAxis","timeAxis"),th(n,(function(e){th(uh(t[e]),(function(t){t&&(ah(t,"axisLabel"),ah(t.axisPointer,"label"))}))})),th(uh(t.parallel),(function(t){var e=t&&t.parallelAxisDefault;ah(e,"axisLabel"),ah(e&&e.axisPointer,"label")})),th(uh(t.calendar),(function(t){ih(t,"itemStyle"),ah(t,"dayLabel"),ah(t,"monthLabel"),ah(t,"yearLabel")})),th(uh(t.radar),(function(t){ah(t,"name"),t.name&&null==t.axisName&&(t.axisName=t.name,delete t.name),null!=t.nameGap&&null==t.axisNameGap&&(t.axisNameGap=t.nameGap,delete t.nameGap)})),th(uh(t.geo),(function(t){eh(t)&&(sh(t),th(uh(t.regions),(function(t){sh(t)})))})),th(uh(t.timeline),(function(t){sh(t),ih(t,"label"),ih(t,"itemStyle"),ih(t,"controlStyle",!0);var e=t.data;s["t"](e)&&s["k"](e,(function(t){s["A"](t)&&(ih(t,"label"),ih(t,"itemStyle"))}))})),th(uh(t.toolbox),(function(t){ih(t,"iconStyle"),th(t.feature,(function(t){ih(t,"iconStyle")}))})),ah(ch(t.axisPointer),"label"),ah(ch(t.tooltip).axisPointer,"label")}function hh(t,e){for(var n=e.split(","),r=t,i=0;i<n.length;i++)if(r=r&&r[n[i]],null==r)break;return r}function fh(t,e,n,r){for(var i,o=e.split(","),a=t,s=0;s<o.length-1;s++)i=o[s],null==a[i]&&(a[i]={}),a=a[i];(r||null==a[o[s]])&&(a[o[s]]=n)}function ph(t){t&&Object(s["k"])(mh,(function(e){e[0]in t&&!(e[1]in t)&&(t[e[1]]=t[e[0]])}))}var mh=[["x","left"],["y","top"],["x2","right"],["y2","bottom"]],gh=["grid","geo","parallel","legend","toolbox","title","visualMap","dataZoom","timeline"],yh=[["borderRadius","barBorderRadius"],["borderColor","barBorderColor"],["borderWidth","barBorderWidth"]];function vh(t){var e=t&&t.itemStyle;if(e)for(var n=0;n<yh.length;n++){var r=yh[n][1],i=yh[n][0];null!=e[r]&&(e[i]=e[r])}}function _h(t){t&&"edge"===t.alignTo&&null!=t.margin&&null==t.edgeDistance&&(t.edgeDistance=t.margin)}function bh(t){t&&t.downplay&&!t.blur&&(t.blur=t.downplay)}function xh(t){t&&null!=t.focusNodeAdjacency&&(t.emphasis=t.emphasis||{},null==t.emphasis.focus&&(t.emphasis.focus="adjacency"))}function wh(t,e){if(t)for(var n=0;n<t.length;n++)e(t[n]),t[n]&&wh(t[n].children,e)}function Mh(t,e){dh(t,e),t.series=ui(t.series),Object(s["k"])(t.series,(function(t){if(Object(s["A"])(t)){var e=t.type;if("line"===e)null!=t.clipOverflow&&(t.clip=t.clipOverflow);else if("pie"===e||"gauge"===e){null!=t.clockWise&&(t.clockwise=t.clockWise),_h(t.label);var n=t.data;if(n&&!Object(s["E"])(n))for(var r=0;r<n.length;r++)_h(n[r]);null!=t.hoverOffset&&(t.emphasis=t.emphasis||{},(t.emphasis.scaleSize=null)&&(t.emphasis.scaleSize=t.hoverOffset))}else if("gauge"===e){var i=hh(t,"pointer.color");null!=i&&fh(t,"itemStyle.color",i)}else if("bar"===e){vh(t),vh(t.backgroundStyle),vh(t.emphasis);n=t.data;if(n&&!Object(s["E"])(n))for(r=0;r<n.length;r++)"object"===typeof n[r]&&(vh(n[r]),vh(n[r]&&n[r].emphasis))}else if("sunburst"===e){var o=t.highlightPolicy;o&&(t.emphasis=t.emphasis||{},t.emphasis.focus||(t.emphasis.focus=o)),bh(t),wh(t.data,bh)}else"graph"===e||"sankey"===e?xh(t):"map"===e&&(t.mapType&&!t.map&&(t.map=t.mapType),t.mapLocation&&Object(s["i"])(t,t.mapLocation));null!=t.hoverAnimation&&(t.emphasis=t.emphasis||{},t.emphasis&&null==t.emphasis.scale&&(t.emphasis.scale=t.hoverAnimation)),ph(t)}})),t.dataRange&&(t.visualMap=t.dataRange),Object(s["k"])(gh,(function(e){var n=t[e];n&&(Object(s["t"])(n)||(n=[n]),Object(s["k"])(n,(function(t){ph(t)})))}))}function Sh(t){var e=Object(s["f"])();t.eachSeries((function(t){var n=t.get("stack");if(n){var r=e.get(n)||e.set(n,[]),i=t.getData(),o={stackResultDimension:i.getCalculationInfo("stackResultDimension"),stackedOverDimension:i.getCalculationInfo("stackedOverDimension"),stackedDimension:i.getCalculationInfo("stackedDimension"),stackedByDimension:i.getCalculationInfo("stackedByDimension"),isStackedByIndex:i.getCalculationInfo("isStackedByIndex"),data:i,seriesModel:t};if(!o.stackedDimension||!o.isStackedByIndex&&!o.stackedByDimension)return;r.length&&i.setCalculationInfo("stackedOnSeries",r[r.length-1].seriesModel),r.push(o)}})),e.each(kh)}function kh(t){Object(s["k"])(t,(function(e,n){var r=[],i=[NaN,NaN],o=[e.stackResultDimension,e.stackedOverDimension],a=e.data,s=e.isStackedByIndex,l=e.seriesModel.get("stackStrategy")||"samesign";a.modify(o,(function(o,u,c){var d,h,f=a.get(e.stackedDimension,c);if(isNaN(f))return i;s?h=a.getRawIndex(c):d=a.get(e.stackedByDimension,c);for(var p=NaN,m=n-1;m>=0;m--){var g=t[m];if(s||(h=g.data.rawIndexOf(g.stackedByDimension,d)),h>=0){var y=g.data.getByRawIndex(g.stackResultDimension,h);if("all"===l||"positive"===l&&y>0||"negative"===l&&y<0||"samesign"===l&&f>=0&&y>0||"samesign"===l&&f<=0&&y<0){f=Vr(f,y),p=y;break}}}return r[0]=f,r[1]=p,r}))}))}var Lh,Th,Dh,Oh,Ih,Ch=function(){function t(t){this.data=t.data||(t.sourceFormat===rd?{}:[]),this.sourceFormat=t.sourceFormat||od,this.seriesLayoutBy=t.seriesLayoutBy||ad,this.startIndex=t.startIndex||0,this.dimensionsDetectedCount=t.dimensionsDetectedCount,this.metaRawOption=t.metaRawOption;var e=this.dimensionsDefine=t.dimensionsDefine;if(e)for(var n=0;n<e.length;n++){var r=e[n];null==r.type&&md(this,n)===ld.Must&&(r.type="ordinal")}}return t}();function Ah(t){return t instanceof Ch}function Yh(t,e,n){n=n||Eh(t);var r=e.seriesLayoutBy,i=Nh(t,n,r,e.sourceHeader,e.dimensions),o=new Ch({data:t,sourceFormat:n,seriesLayoutBy:r,dimensionsDefine:i.dimensionsDefine,startIndex:i.startIndex,dimensionsDetectedCount:i.dimensionsDetectedCount,metaRawOption:Object(s["d"])(e)});return o}function jh(t){return new Ch({data:t,sourceFormat:Object(s["E"])(t)?id:td})}function Ph(t){return new Ch({data:t.data,sourceFormat:t.sourceFormat,seriesLayoutBy:t.seriesLayoutBy,dimensionsDefine:Object(s["d"])(t.dimensionsDefine),startIndex:t.startIndex,dimensionsDetectedCount:t.dimensionsDetectedCount})}function Eh(t){var e=od;if(Object(s["E"])(t))e=id;else if(Object(s["t"])(t)){0===t.length&&(e=ed);for(var n=0,r=t.length;n<r;n++){var i=t[n];if(null!=i){if(Object(s["t"])(i)||Object(s["E"])(i)){e=ed;break}if(Object(s["A"])(i)){e=nd;break}}}}else if(Object(s["A"])(t))for(var o in t)if(Object(s["q"])(t,o)&&Object(s["u"])(t[o])){e=rd;break}return e}function Nh(t,e,n,r,i){var o,a;if(!t)return{dimensionsDefine:Hh(i),startIndex:a,dimensionsDetectedCount:o};if(e===ed){var l=t;"auto"===r||null==r?zh((function(t){null!=t&&"-"!==t&&(Object(s["C"])(t)?null==a&&(a=1):a=0)}),n,l,10):a=Object(s["z"])(r)?r:r?1:0,i||1!==a||(i=[],zh((function(t,e){i[e]=null!=t?t+"":""}),n,l,1/0)),o=i?i.length:n===sd?l.length:l[0]?l[0].length:null}else if(e===nd)i||(i=Rh(t));else if(e===rd)i||(i=[],Object(s["k"])(t,(function(t,e){i.push(e)})));else if(e===td){var u=hi(t[0]);o=Object(s["t"])(u)&&u.length||1}return{startIndex:a,dimensionsDefine:Hh(i),dimensionsDetectedCount:o}}function Rh(t){var e,n=0;while(n<t.length&&!(e=t[n++]));if(e)return Object(s["F"])(e)}function Hh(t){if(t){var e=Object(s["f"])();return Object(s["H"])(t,(function(t,n){t=Object(s["A"])(t)?t:{name:t};var r={name:t.name,displayName:t.displayName,type:t.type};if(null==r.name)return r;r.name+="",null==r.displayName&&(r.displayName=r.name);var i=e.get(r.name);return i?r.name+="-"+i.count++:e.set(r.name,{count:1}),r}))}}function zh(t,e,n,r){if(e===sd)for(var i=0;i<n.length&&i<r;i++)t(n[i]?n[i][0]:null,i);else{var o=n[0]||[];for(i=0;i<o.length&&i<r;i++)t(o[i],i)}}function Fh(t){var e=t.sourceFormat;return e===nd||e===rd}var Bh=function(){function t(t,e){var n=Ah(t)?t:jh(t);this._source=n;var r=this._data=n.data;n.sourceFormat===id&&(this._offset=0,this._dimSize=e,this._data=r),Ih(this,r,n)}return t.prototype.getSource=function(){return this._source},t.prototype.count=function(){return 0},t.prototype.getItem=function(t,e){},t.prototype.appendData=function(t){},t.prototype.clean=function(){},t.protoInitialize=function(){var e=t.prototype;e.pure=!1,e.persistent=!0}(),t.internalField=function(){var t;Ih=function(t,i,o){var a=o.sourceFormat,l=o.seriesLayoutBy,u=o.startIndex,c=o.dimensionsDefine,d=Oh[Kh(a,l)];if(Object(s["m"])(t,d),a===id)t.getItem=e,t.count=r,t.fillStorage=n;else{var h=Gh(a,l);t.getItem=Object(s["c"])(h,null,i,u,c);var f=qh(a,l);t.count=Object(s["c"])(f,null,i,u,c)}};var e=function(t,e){t-=this._offset,e=e||[];for(var n=this._data,r=this._dimSize,i=r*t,o=0;o<r;o++)e[o]=n[i+o];return e},n=function(t,e,n,r){for(var i=this._data,o=this._dimSize,a=0;a<o;a++){for(var s=r[a],l=null==s[0]?1/0:s[0],u=null==s[1]?-1/0:s[1],c=e-t,d=n[a],h=0;h<c;h++){var f=i[h*o+a];d[t+h]=f,f<l&&(l=f),f>u&&(u=f)}s[0]=l,s[1]=u}},r=function(){return this._data?this._data.length/this._dimSize:0};function i(t){for(var e=0;e<t.length;e++)this._data.push(t[e])}t={},t[ed+"_"+ad]={pure:!0,appendData:i},t[ed+"_"+sd]={pure:!0,appendData:function(){throw new Error('Do not support appendData when set seriesLayoutBy: "row".')}},t[nd]={pure:!0,appendData:i},t[rd]={pure:!0,appendData:function(t){var e=this._data;Object(s["k"])(t,(function(t,n){for(var r=e[n]||(e[n]=[]),i=0;i<(t||[]).length;i++)r.push(t[i])}))}},t[td]={appendData:i},t[id]={persistent:!1,pure:!0,appendData:function(t){this._data=t},clean:function(){this._offset+=this.count(),this._data=null}},Oh=t}(),t}(),Wh=function(t,e,n,r){return t[r]},Vh=(Lh={},Lh[ed+"_"+ad]=function(t,e,n,r){return t[r+e]},Lh[ed+"_"+sd]=function(t,e,n,r,i){r+=e;for(var o=i||[],a=t,s=0;s<a.length;s++){var l=a[s];o[s]=l?l[r]:null}return o},Lh[nd]=Wh,Lh[rd]=function(t,e,n,r,i){for(var o=i||[],a=0;a<n.length;a++){var s=n[a].name;0;var l=t[s];o[a]=l?l[r]:null}return o},Lh[td]=Wh,Lh);function Gh(t,e){var n=Vh[Kh(t,e)];return n}var $h=function(t,e,n){return t.length},Uh=(Th={},Th[ed+"_"+ad]=function(t,e,n){return Math.max(0,t.length-e)},Th[ed+"_"+sd]=function(t,e,n){var r=t[0];return r?Math.max(0,r.length-e):0},Th[nd]=$h,Th[rd]=function(t,e,n){var r=n[0].name;var i=t[r];return i?i.length:0},Th[td]=$h,Th);function qh(t,e){var n=Uh[Kh(t,e)];return n}var Xh=function(t,e,n){return t[e]},Zh=(Dh={},Dh[ed]=Xh,Dh[nd]=function(t,e,n){return t[n]},Dh[rd]=Xh,Dh[td]=function(t,e,n){var r=hi(t);return r instanceof Array?r[e]:r},Dh[id]=Xh,Dh);function Jh(t){var e=Zh[t];return e}function Kh(t,e){return t===ed?t+"_"+e:t}function Qh(t,e,n){if(t){var r=t.getRawDataItem(e);if(null!=r){var i=t.getStore(),o=i.getSource().sourceFormat;if(null!=n){var a=t.getDimensionIndex(n),s=i.getDimensionProperty(a);return Jh(o)(r,a,s)}var l=r;return o===td&&(l=hi(r)),l}}}var tf=/\{@(.+?)\}/g,ef=function(){function t(){}return t.prototype.getDataParams=function(t,e){var n=this.getData(e),r=this.getRawValue(t,e),i=n.getRawIndex(t),o=n.getName(t),a=n.getRawDataItem(t),s=n.getItemVisual(t,"style"),l=s&&s[n.getItemVisual(t,"drawType")||"fill"],u=s&&s.stroke,c=this.mainType,d="series"===c,h=n.userOutput&&n.userOutput.get();return{componentType:c,componentSubType:this.subType,componentIndex:this.componentIndex,seriesType:d?this.subType:null,seriesIndex:this.seriesIndex,seriesId:d?this.id:null,seriesName:d?this.name:null,name:o,dataIndex:i,data:a,dataType:e,value:r,color:l,borderColor:u,dimensionNames:h?h.fullDimensions:null,encode:h?h.encode:null,$vars:["seriesName","name","value"]}},t.prototype.getFormattedLabel=function(t,e,n,r,i,o){e=e||"normal";var a=this.getData(n),l=this.getDataParams(t,n);if(o&&(l.value=o.interpolatedValue),null!=r&&s["t"](l.value)&&(l.value=l.value[r]),!i){var u=a.getItemModel(t);i=u.get("normal"===e?["label","formatter"]:[e,"label","formatter"])}if(s["w"](i))return l.status=e,l.dimensionIndex=r,i(l);if(s["C"](i)){var c=Dc(i,l);return c.replace(tf,(function(e,n){var r=n.length,i=n;"["===i.charAt(0)&&"]"===i.charAt(r-1)&&(i=+i.slice(1,r-1));var l=Qh(a,t,i);if(o&&s["t"](o.interpolatedValue)){var u=a.getDimensionIndex(i);u>=0&&(l=o.interpolatedValue[u])}return null!=l?l+"":""}))}},t.prototype.getRawValue=function(t,e){return Qh(this.getData(e),t)},t.prototype.formatTooltip=function(t,e,n){},t}();function nf(t){var e,n;return s["A"](t)?t.type&&(n=t):e=t,{text:e,frag:n}}function rf(t){return new of(t)}var of=function(){function t(t){t=t||{},this._reset=t.reset,this._plan=t.plan,this._count=t.count,this._onDirty=t.onDirty,this._dirty=!0}return t.prototype.perform=function(t){var e,n=this._upstream,r=t&&t.skip;if(this._dirty&&n){var i=this.context;i.data=i.outputData=n.context.outputData}this.__pipeline&&(this.__pipeline.currentTask=this),this._plan&&!r&&(e=this._plan(this.context));var o,a=d(this._modBy),l=this._modDataCount||0,u=d(t&&t.modBy),c=t&&t.modDataCount||0;function d(t){return!(t>=1)&&(t=1),t}a===u&&l===c||(e="reset"),(this._dirty||"reset"===e)&&(this._dirty=!1,o=this._doReset(r)),this._modBy=u,this._modDataCount=c;var h=t&&t.step;if(this._dueEnd=n?n._outputDueEnd:this._count?this._count(this.context):1/0,this._progress){var f=this._dueIndex,p=Math.min(null!=h?this._dueIndex+h:1/0,this._dueEnd);if(!r&&(o||f<p)){var m=this._progress;if(Object(s["t"])(m))for(var g=0;g<m.length;g++)this._doProgress(m[g],f,p,u,c);else this._doProgress(m,f,p,u,c)}this._dueIndex=p;var y=null!=this._settedOutputEnd?this._settedOutputEnd:p;0,this._outputDueEnd=y}else this._dueIndex=this._outputDueEnd=null!=this._settedOutputEnd?this._settedOutputEnd:this._dueEnd;return this.unfinished()},t.prototype.dirty=function(){this._dirty=!0,this._onDirty&&this._onDirty(this.context)},t.prototype._doProgress=function(t,e,n,r,i){af.reset(e,n,r,i),this._callingProgress=t,this._callingProgress({start:e,end:n,count:n-e,next:af.next},this.context)},t.prototype._doReset=function(t){var e,n;this._dueIndex=this._outputDueEnd=this._dueEnd=0,this._settedOutputEnd=null,!t&&this._reset&&(e=this._reset(this.context),e&&e.progress&&(n=e.forceFirstProgress,e=e.progress),Object(s["t"])(e)&&!e.length&&(e=null)),this._progress=e,this._modBy=this._modDataCount=null;var r=this._downstream;return r&&r.dirty(),n},t.prototype.unfinished=function(){return this._progress&&this._dueIndex<this._dueEnd},t.prototype.pipe=function(t){(this._downstream!==t||this._dirty)&&(this._downstream=t,t._upstream=this,t.dirty())},t.prototype.dispose=function(){this._disposed||(this._upstream&&(this._upstream._downstream=null),this._downstream&&(this._downstream._upstream=null),this._dirty=!1,this._disposed=!0)},t.prototype.getUpstream=function(){return this._upstream},t.prototype.getDownstream=function(){return this._downstream},t.prototype.setOutputEnd=function(t){this._outputDueEnd=this._settedOutputEnd=t},t}(),af=function(){var t,e,n,r,i,o={reset:function(l,u,c,d){e=l,t=u,n=c,r=d,i=Math.ceil(r/n),o.next=n>1&&r>0?s:a}};return o;function a(){return e<t?e++:null}function s(){var o=e%i*n+Math.ceil(e/i),a=e>=t?null:o<r?o:e;return e++,a}}();function sf(t,e){var n=e&&e.type;return"ordinal"===n?t:("time"!==n||Object(s["z"])(t)||null==t||"-"===t||(t=+Xr(t)),null==t||""===t?NaN:+t)}var lf=Object(s["f"])({number:function(t){return parseFloat(t)},time:function(t){return+Xr(t)},trim:function(t){return Object(s["C"])(t)?Object(s["T"])(t):t}});function uf(t){return lf.get(t)}var cf={lt:function(t,e){return t<e},lte:function(t,e){return t<=e},gt:function(t,e){return t>e},gte:function(t,e){return t>=e}},df=function(){function t(t,e){if(!Object(s["z"])(e)){var n="";0,Cd(n)}this._opFn=cf[t],this._rvalFloat=ei(e)}return t.prototype.evaluate=function(t){return Object(s["z"])(t)?this._opFn(t,this._rvalFloat):this._opFn(ei(t),this._rvalFloat)},t}(),hf=function(){function t(t,e){var n="desc"===t;this._resultLT=n?1:-1,null==e&&(e=n?"min":"max"),this._incomparable="min"===e?-1/0:1/0}return t.prototype.evaluate=function(t,e){var n=Object(s["z"])(t)?t:ei(t),r=Object(s["z"])(e)?e:ei(e),i=isNaN(n),o=isNaN(r);if(i&&(n=this._incomparable),o&&(r=this._incomparable),i&&o){var a=Object(s["C"])(t),l=Object(s["C"])(e);a&&(n=l?t:0),l&&(r=a?e:0)}return n<r?this._resultLT:n>r?-this._resultLT:0},t}(),ff=function(){function t(t,e){this._rval=e,this._isEQ=t,this._rvalTypeof=typeof e,this._rvalFloat=ei(e)}return t.prototype.evaluate=function(t){var e=t===this._rval;if(!e){var n=typeof t;n===this._rvalTypeof||"number"!==n&&"number"!==this._rvalTypeof||(e=ei(t)===this._rvalFloat)}return this._isEQ?e:!e},t}();function pf(t,e){return"eq"===t||"ne"===t?new ff("eq"===t,e):Object(s["q"])(cf,t)?new df(t,e):null}var mf=function(){function t(){}return t.prototype.getRawData=function(){throw new Error("not supported")},t.prototype.getRawDataItem=function(t){throw new Error("not supported")},t.prototype.cloneRawData=function(){},t.prototype.getDimensionInfo=function(t){},t.prototype.cloneAllDimensionInfo=function(){},t.prototype.count=function(){},t.prototype.retrieveValue=function(t,e){},t.prototype.retrieveValueFromItem=function(t,e){},t.prototype.convertValue=function(t,e){return sf(t,e)},t}();function gf(t,e){var n=new mf,r=t.data,i=n.sourceFormat=t.sourceFormat,o=t.startIndex,a="";t.seriesLayoutBy!==ad&&Cd(a);var l=[],u={},c=t.dimensionsDefine;if(c)Object(s["k"])(c,(function(t,e){var n=t.name,r={index:e,name:n,displayName:t.displayName};if(l.push(r),null!=n){var i="";Object(s["q"])(u,n)&&Cd(i),u[n]=r}}));else for(var d=0;d<t.dimensionsDetectedCount;d++)l.push({index:d});var h=Gh(i,ad);e.__isBuiltIn&&(n.getRawDataItem=function(t){return h(r,o,l,t)},n.getRawData=Object(s["c"])(yf,null,t)),n.cloneRawData=Object(s["c"])(vf,null,t);var f=qh(i,ad);n.count=Object(s["c"])(f,null,r,o,l);var p=Jh(i);n.retrieveValue=function(t,e){var n=h(r,o,l,t);return m(n,e)};var m=n.retrieveValueFromItem=function(t,e){if(null!=t){var n=l[e];return n?p(t,e,n.name):void 0}};return n.getDimensionInfo=Object(s["c"])(_f,null,l,u),n.cloneAllDimensionInfo=Object(s["c"])(bf,null,l),n}function yf(t){var e=t.sourceFormat;if(!kf(e)){var n="";0,Cd(n)}return t.data}function vf(t){var e=t.sourceFormat,n=t.data;if(!kf(e)){var r="";0,Cd(r)}if(e===ed){for(var i=[],o=0,a=n.length;o<a;o++)i.push(n[o].slice());return i}if(e===nd){for(i=[],o=0,a=n.length;o<a;o++)i.push(Object(s["m"])({},n[o]));return i}}function _f(t,e,n){if(null!=n)return Object(s["z"])(n)||!isNaN(n)&&!Object(s["q"])(e,n)?t[n]:Object(s["q"])(e,n)?e[n]:void 0}function bf(t){return Object(s["d"])(t)}var xf=Object(s["f"])();function wf(t){t=Object(s["d"])(t);var e=t.type,n="";e||Cd(n);var r=e.split(":");2!==r.length&&Cd(n);var i=!1;"echarts"===r[0]&&(e=r[1],i=!0),t.__isBuiltIn=i,xf.set(e,t)}function Mf(t,e,n){var r=ui(t),i=r.length,o="";i||Cd(o);for(var a=0,s=i;a<s;a++){var l=r[a];e=Sf(l,e,n,1===i?null:a),a!==s-1&&(e.length=Math.max(e.length,1))}return e}function Sf(t,e,n,r){var i="";e.length||Cd(i),Object(s["A"])(t)||Cd(i);var o=t.type,a=xf.get(o);a||Cd(i);var l=Object(s["H"])(e,(function(t){return gf(t,a)})),u=ui(a.transform({upstream:l[0],upstreamList:l,config:Object(s["d"])(t.config)}));return Object(s["H"])(u,(function(t,n){var r="";Object(s["A"])(t)||Cd(r),t.data||Cd(r);var i,o=Eh(t.data);kf(o)||Cd(r);var a=e[0];if(a&&0===n&&!t.dimensions){var l=a.startIndex;l&&(t.data=a.data.slice(0,l).concat(t.data)),i={seriesLayoutBy:ad,sourceHeader:l,dimensions:a.metaRawOption.dimensions}}else i={seriesLayoutBy:ad,sourceHeader:0,dimensions:t.dimensions};return Yh(t.data,i,null)}))}function kf(t){return t===ed||t===nd}var Lf,Tf="undefined",Df=typeof Uint32Array===Tf?Array:Uint32Array,Of=typeof Uint16Array===Tf?Array:Uint16Array,If=typeof Int32Array===Tf?Array:Int32Array,Cf=typeof Float64Array===Tf?Array:Float64Array,Af={float:Cf,int:If,ordinal:Array,number:Array,time:Cf};function Yf(t){return t>65535?Df:Of}function jf(){return[1/0,-1/0]}function Pf(t){var e=t.constructor;return e===Array?t.slice():new e(t)}function Ef(t,e,n,r,i){var o=Af[n||"float"];if(i){var a=t[e],s=a&&a.length;if(s!==r){for(var l=new o(r),u=0;u<s;u++)l[u]=a[u];t[e]=l}}else t[e]=new o(r)}var Nf=function(){function t(){this._chunks=[],this._rawExtent=[],this._extent=[],this._count=0,this._rawCount=0,this._calcDimNameToIdx=Object(s["f"])()}return t.prototype.initData=function(t,e,n){this._provider=t,this._chunks=[],this._indices=null,this.getRawIndex=this._getRawIdxIdentity;var r=t.getSource(),i=this.defaultDimValueGetter=Lf[r.sourceFormat];this._dimValueGetter=n||i,this._rawExtent=[];Fh(r);this._dimensions=Object(s["H"])(e,(function(t){return{type:t.type,property:t.property}})),this._initDataFromProvider(0,t.count())},t.prototype.getProvider=function(){return this._provider},t.prototype.getSource=function(){return this._provider.getSource()},t.prototype.ensureCalculationDimension=function(t,e){var n=this._calcDimNameToIdx,r=this._dimensions,i=n.get(t);if(null!=i){if(r[i].type===e)return i}else i=r.length;return r[i]={type:e},n.set(t,i),this._chunks[i]=new Af[e||"float"](this._rawCount),this._rawExtent[i]=jf(),i},t.prototype.collectOrdinalMeta=function(t,e){var n=this._chunks[t],r=this._dimensions[t],i=this._rawExtent,o=r.ordinalOffset||0,a=n.length;0===o&&(i[t]=jf());for(var s=i[t],l=o;l<a;l++){var u=n[l]=e.parseAndCollect(n[l]);isNaN(u)||(s[0]=Math.min(u,s[0]),s[1]=Math.max(u,s[1]))}r.ordinalMeta=e,r.ordinalOffset=a,r.type="ordinal"},t.prototype.getOrdinalMeta=function(t){var e=this._dimensions[t],n=e.ordinalMeta;return n},t.prototype.getDimensionProperty=function(t){var e=this._dimensions[t];return e&&e.property},t.prototype.appendData=function(t){var e=this._provider,n=this.count();e.appendData(t);var r=e.count();return e.persistent||(r+=n),n<r&&this._initDataFromProvider(n,r,!0),[n,r]},t.prototype.appendValues=function(t,e){for(var n=this._chunks,r=this._dimensions,i=r.length,o=this._rawExtent,a=this.count(),s=a+Math.max(t.length,e||0),l=0;l<i;l++){var u=r[l];Ef(n,l,u.type,s,!0)}for(var c=[],d=a;d<s;d++)for(var h=d-a,f=0;f<i;f++){u=r[f];var p=Lf.arrayRows.call(this,t[h]||c,u.property,h,f);n[f][d]=p;var m=o[f];p<m[0]&&(m[0]=p),p>m[1]&&(m[1]=p)}return this._rawCount=this._count=s,{start:a,end:s}},t.prototype._initDataFromProvider=function(t,e,n){for(var r=this._provider,i=this._chunks,o=this._dimensions,a=o.length,l=this._rawExtent,u=Object(s["H"])(o,(function(t){return t.property})),c=0;c<a;c++){var d=o[c];l[c]||(l[c]=jf()),Ef(i,c,d.type,e,n)}if(r.fillStorage)r.fillStorage(t,e,i,l);else for(var h=[],f=t;f<e;f++){h=r.getItem(f,h);for(var p=0;p<a;p++){var m=i[p],g=this._dimValueGetter(h,u[p],f,p);m[f]=g;var y=l[p];g<y[0]&&(y[0]=g),g>y[1]&&(y[1]=g)}}!r.persistent&&r.clean&&r.clean(),this._rawCount=this._count=e,this._extent=[]},t.prototype.count=function(){return this._count},t.prototype.get=function(t,e){if(!(e>=0&&e<this._count))return NaN;var n=this._chunks[t];return n?n[this.getRawIndex(e)]:NaN},t.prototype.getValues=function(t,e){var n=[],r=[];if(null==e){e=t,t=[];for(var i=0;i<this._dimensions.length;i++)r.push(i)}else r=t;i=0;for(var o=r.length;i<o;i++)n.push(this.get(r[i],e));return n},t.prototype.getByRawIndex=function(t,e){if(!(e>=0&&e<this._rawCount))return NaN;var n=this._chunks[t];return n?n[e]:NaN},t.prototype.getSum=function(t){var e=this._chunks[t],n=0;if(e)for(var r=0,i=this.count();r<i;r++){var o=this.get(t,r);isNaN(o)||(n+=o)}return n},t.prototype.getMedian=function(t){var e=[];this.each([t],(function(t){isNaN(t)||e.push(t)}));var n=e.sort((function(t,e){return t-e})),r=this.count();return 0===r?0:r%2===1?n[(r-1)/2]:(n[r/2]+n[r/2-1])/2},t.prototype.indexOfRawIndex=function(t){if(t>=this._rawCount||t<0)return-1;if(!this._indices)return t;var e=this._indices,n=e[t];if(null!=n&&n<this._count&&n===t)return t;var r=0,i=this._count-1;while(r<=i){var o=(r+i)/2|0;if(e[o]<t)r=o+1;else{if(!(e[o]>t))return o;i=o-1}}return-1},t.prototype.indicesOfNearest=function(t,e,n){var r=this._chunks,i=r[t],o=[];if(!i)return o;null==n&&(n=1/0);for(var a=1/0,s=-1,l=0,u=0,c=this.count();u<c;u++){var d=this.getRawIndex(u),h=e-i[d],f=Math.abs(h);f<=n&&((f<a||f===a&&h>=0&&s<0)&&(a=f,s=h,l=0),h===s&&(o[l++]=u))}return o.length=l,o},t.prototype.getIndices=function(){var t,e=this._indices;if(e){var n=e.constructor,r=this._count;if(n===Array){t=new n(r);for(var i=0;i<r;i++)t[i]=e[i]}else t=new n(e.buffer,0,r)}else{n=Yf(this._rawCount);t=new n(this.count());for(i=0;i<t.length;i++)t[i]=i}return t},t.prototype.filter=function(t,e){if(!this._count)return this;for(var n=this.clone(),r=n.count(),i=Yf(n._rawCount),o=new i(r),a=[],s=t.length,l=0,u=t[0],c=n._chunks,d=0;d<r;d++){var h=void 0,f=n.getRawIndex(d);if(0===s)h=e(d);else if(1===s){var p=c[u][f];h=e(p,d)}else{for(var m=0;m<s;m++)a[m]=c[t[m]][f];a[m]=d,h=e.apply(null,a)}h&&(o[l++]=f)}return l<r&&(n._indices=o),n._count=l,n._extent=[],n._updateGetRawIdx(),n},t.prototype.selectRange=function(t){var e=this.clone(),n=e._count;if(!n)return this;var r=Object(s["F"])(t),i=r.length;if(!i)return this;var o=e.count(),a=Yf(e._rawCount),l=new a(o),u=0,c=r[0],d=t[c][0],h=t[c][1],f=e._chunks,p=!1;if(!e._indices){var m=0;if(1===i){for(var g=f[r[0]],y=0;y<n;y++){var v=g[y];(v>=d&&v<=h||isNaN(v))&&(l[u++]=m),m++}p=!0}else if(2===i){g=f[r[0]];var _=f[r[1]],b=t[r[1]][0],x=t[r[1]][1];for(y=0;y<n;y++){v=g[y];var w=_[y];(v>=d&&v<=h||isNaN(v))&&(w>=b&&w<=x||isNaN(w))&&(l[u++]=m),m++}p=!0}}if(!p)if(1===i)for(y=0;y<o;y++){var M=e.getRawIndex(y);v=f[r[0]][M];(v>=d&&v<=h||isNaN(v))&&(l[u++]=M)}else for(y=0;y<o;y++){for(var S=!0,k=(M=e.getRawIndex(y),0);k<i;k++){var L=r[k];v=f[L][M];(v<t[L][0]||v>t[L][1])&&(S=!1)}S&&(l[u++]=e.getRawIndex(y))}return u<o&&(e._indices=l),e._count=u,e._extent=[],e._updateGetRawIdx(),e},t.prototype.map=function(t,e){var n=this.clone(t);return this._updateDims(n,t,e),n},t.prototype.modify=function(t,e){this._updateDims(this,t,e)},t.prototype._updateDims=function(t,e,n){for(var r=t._chunks,i=[],o=e.length,a=t.count(),s=[],l=t._rawExtent,u=0;u<e.length;u++)l[e[u]]=jf();for(var c=0;c<a;c++){for(var d=t.getRawIndex(c),h=0;h<o;h++)s[h]=r[e[h]][d];s[o]=c;var f=n&&n.apply(null,s);if(null!=f){"object"!==typeof f&&(i[0]=f,f=i);for(u=0;u<f.length;u++){var p=e[u],m=f[u],g=l[p],y=r[p];y&&(y[d]=m),m<g[0]&&(g[0]=m),m>g[1]&&(g[1]=m)}}}},t.prototype.lttbDownSample=function(t,e){var n,r,i,o=this.clone([t],!0),a=o._chunks,s=a[t],l=this.count(),u=0,c=Math.floor(1/e),d=this.getRawIndex(0),h=new(Yf(this._rawCount))(Math.min(2*(Math.ceil(l/c)+2),l));h[u++]=d;for(var f=1;f<l-1;f+=c){for(var p=Math.min(f+c,l-1),m=Math.min(f+2*c,l),g=(m+p)/2,y=0,v=p;v<m;v++){var _=this.getRawIndex(v),b=s[_];isNaN(b)||(y+=b)}y/=m-p;var x=f,w=Math.min(f+c,l),M=f-1,S=s[d];n=-1,i=x;var k=-1,L=0;for(v=x;v<w;v++){_=this.getRawIndex(v),b=s[_];isNaN(b)?(L++,k<0&&(k=_)):(r=Math.abs((M-g)*(b-S)-(M-v)*(y-S)),r>n&&(n=r,i=_))}L>0&&L<w-x&&(h[u++]=Math.min(k,i),i=Math.max(k,i)),h[u++]=i,d=i}return h[u++]=this.getRawIndex(l-1),o._count=u,o._indices=h,o.getRawIndex=this._getRawIdx,o},t.prototype.downSample=function(t,e,n,r){for(var i=this.clone([t],!0),o=i._chunks,a=[],s=Math.floor(1/e),l=o[t],u=this.count(),c=i._rawExtent[t]=jf(),d=new(Yf(this._rawCount))(Math.ceil(u/s)),h=0,f=0;f<u;f+=s){s>u-f&&(s=u-f,a.length=s);for(var p=0;p<s;p++){var m=this.getRawIndex(f+p);a[p]=l[m]}var g=n(a),y=this.getRawIndex(Math.min(f+r(a,g)||0,u-1));l[y]=g,g<c[0]&&(c[0]=g),g>c[1]&&(c[1]=g),d[h++]=y}return i._count=h,i._indices=d,i._updateGetRawIdx(),i},t.prototype.each=function(t,e){if(this._count)for(var n=t.length,r=this._chunks,i=0,o=this.count();i<o;i++){var a=this.getRawIndex(i);switch(n){case 0:e(i);break;case 1:e(r[t[0]][a],i);break;case 2:e(r[t[0]][a],r[t[1]][a],i);break;default:for(var s=0,l=[];s<n;s++)l[s]=r[t[s]][a];l[s]=i,e.apply(null,l)}}},t.prototype.getDataExtent=function(t){var e=this._chunks[t],n=jf();if(!e)return n;var r,i=this.count(),o=!this._indices;if(o)return this._rawExtent[t].slice();if(r=this._extent[t],r)return r.slice();r=n;for(var a=r[0],s=r[1],l=0;l<i;l++){var u=this.getRawIndex(l),c=e[u];c<a&&(a=c),c>s&&(s=c)}return r=[a,s],this._extent[t]=r,r},t.prototype.getRawDataItem=function(t){var e=this.getRawIndex(t);if(this._provider.persistent)return this._provider.getItem(e);for(var n=[],r=this._chunks,i=0;i<r.length;i++)n.push(r[i][e]);return n},t.prototype.clone=function(e,n){var r=new t,i=this._chunks,o=e&&Object(s["N"])(e,(function(t,e){return t[e]=!0,t}),{});if(o)for(var a=0;a<i.length;a++)r._chunks[a]=o[a]?Pf(i[a]):i[a];else r._chunks=i;return this._copyCommonProps(r),n||(r._indices=this._cloneIndices()),r._updateGetRawIdx(),r},t.prototype._copyCommonProps=function(t){t._count=this._count,t._rawCount=this._rawCount,t._provider=this._provider,t._dimensions=this._dimensions,t._extent=Object(s["d"])(this._extent),t._rawExtent=Object(s["d"])(this._rawExtent)},t.prototype._cloneIndices=function(){if(this._indices){var t=this._indices.constructor,e=void 0;if(t===Array){var n=this._indices.length;e=new t(n);for(var r=0;r<n;r++)e[r]=this._indices[r]}else e=new t(this._indices);return e}return null},t.prototype._getRawIdxIdentity=function(t){return t},t.prototype._getRawIdx=function(t){return t<this._count&&t>=0?this._indices[t]:-1},t.prototype._updateGetRawIdx=function(){this.getRawIndex=this._indices?this._getRawIdx:this._getRawIdxIdentity},t.internalField=function(){function t(t,e,n,r){return sf(t[r],this._dimensions[r])}Lf={arrayRows:t,objectRows:function(t,e,n,r){return sf(t[e],this._dimensions[r])},keyedColumns:t,original:function(t,e,n,r){var i=t&&(null==t.value?t:t.value);return sf(i instanceof Array?i[r]:i,this._dimensions[r])},typedArray:function(t,e,n,r){return t[r]}}}(),t}(),Rf=Nf,Hf=function(){function t(t){this._sourceList=[],this._storeList=[],this._upstreamSignList=[],this._versionSignBase=0,this._dirty=!0,this._sourceHost=t}return t.prototype.dirty=function(){this._setLocalSource([],[]),this._storeList=[],this._dirty=!0},t.prototype._setLocalSource=function(t,e){this._sourceList=t,this._upstreamSignList=e,this._versionSignBase++,this._versionSignBase>9e10&&(this._versionSignBase=0)},t.prototype._getVersionSign=function(){return this._sourceHost.uid+"_"+this._versionSignBase},t.prototype.prepareSource=function(){this._isDirty()&&(this._createSource(),this._dirty=!1)},t.prototype._createSource=function(){this._setLocalSource([],[]);var t,e,n=this._sourceHost,r=this._getUpstreamSourceManagers(),i=!!r.length;if(Ff(n)){var o=n,a=void 0,l=void 0,u=void 0;if(i){var c=r[0];c.prepareSource(),u=c.getSource(),a=u.data,l=u.sourceFormat,e=[c._getVersionSign()]}else a=o.get("data",!0),l=Object(s["E"])(a)?id:td,e=[];var d=this._getSourceMetaRawOption()||{},h=u&&u.metaRawOption||{},f=Object(s["P"])(d.seriesLayoutBy,h.seriesLayoutBy)||null,p=Object(s["P"])(d.sourceHeader,h.sourceHeader),m=Object(s["P"])(d.dimensions,h.dimensions),g=f!==h.seriesLayoutBy||!!p!==!!h.sourceHeader||m;t=g?[Yh(a,{seriesLayoutBy:f,sourceHeader:p,dimensions:m},l)]:[]}else{var y=n;if(i){var v=this._applyTransform(r);t=v.sourceList,e=v.upstreamSignList}else{var _=y.get("source",!0);t=[Yh(_,this._getSourceMetaRawOption(),null)],e=[]}}this._setLocalSource(t,e)},t.prototype._applyTransform=function(t){var e,n=this._sourceHost,r=n.get("transform",!0),i=n.get("fromTransformResult",!0);if(null!=i){var o="";1!==t.length&&Bf(o)}var a=[],l=[];return Object(s["k"])(t,(function(t){t.prepareSource();var e=t.getSource(i||0),n="";null==i||e||Bf(n),a.push(e),l.push(t._getVersionSign())})),r?e=Mf(r,a,{datasetIndex:n.componentIndex}):null!=i&&(e=[Ph(a[0])]),{sourceList:e,upstreamSignList:l}},t.prototype._isDirty=function(){if(this._dirty)return!0;for(var t=this._getUpstreamSourceManagers(),e=0;e<t.length;e++){var n=t[e];if(n._isDirty()||this._upstreamSignList[e]!==n._getVersionSign())return!0}},t.prototype.getSource=function(t){t=t||0;var e=this._sourceList[t];if(!e){var n=this._getUpstreamSourceManagers();return n[0]&&n[0].getSource(t)}return e},t.prototype.getSharedDataStore=function(t){var e=t.makeStoreSchema();return this._innerGetDataStore(e.dimensions,t.source,e.hash)},t.prototype._innerGetDataStore=function(t,e,n){var r=0,i=this._storeList,o=i[r];o||(o=i[r]={});var a=o[n];if(!a){var s=this._getUpstreamSourceManagers()[0];Ff(this._sourceHost)&&s?a=s._innerGetDataStore(t,e,n):(a=new Rf,a.initData(new Bh(e,t.length),t)),o[n]=a}return a},t.prototype._getUpstreamSourceManagers=function(){var t=this._sourceHost;if(Ff(t)){var e=fd(t);return e?[e.getSourceManager()]:[]}return Object(s["H"])(pd(t),(function(t){return t.getSourceManager()}))},t.prototype._getSourceMetaRawOption=function(){var t,e,n,r=this._sourceHost;if(Ff(r))t=r.get("seriesLayoutBy",!0),e=r.get("sourceHeader",!0),n=r.get("dimensions",!0);else if(!this._getUpstreamSourceManagers().length){var i=r;t=i.get("seriesLayoutBy",!0),e=i.get("sourceHeader",!0),n=i.get("dimensions",!0)}return{seriesLayoutBy:t,sourceHeader:e,dimensions:n}},t}();function zf(t){var e=t.option.transform;e&&Object(s["R"])(t.option.transform)}function Ff(t){return"series"===t.mainType}function Bf(t){throw new Error(t)}var Wf="line-height:1";function Vf(t,e){var n=t.color||"#6e7079",r=t.fontSize||12,i=t.fontWeight||"400",o=t.color||"#464646",a=t.fontSize||14,s=t.fontWeight||"900";return"html"===e?{nameStyle:"font-size:"+V(r+"")+"px;color:"+V(n)+";font-weight:"+V(i+""),valueStyle:"font-size:"+V(a+"")+"px;color:"+V(o)+";font-weight:"+V(s+"")}:{nameStyle:{fontSize:r,fill:n,fontWeight:i},valueStyle:{fontSize:a,fill:o,fontWeight:s}}}var Gf=[0,10,20,30],$f=["","\n","\n\n","\n\n\n"];function Uf(t,e){return e.type=t,e}function qf(t){return"section"===t.type}function Xf(t){return qf(t)?Jf:Kf}function Zf(t){if(qf(t)){var e=0,n=t.blocks.length,r=n>1||n>0&&!t.noHeader;return Object(s["k"])(t.blocks,(function(t){var n=Zf(t);n>=e&&(e=n+ +(r&&(!n||qf(t)&&!t.noHeader)))})),e}return 0}function Jf(t,e,n,r){var i=e.noHeader,o=tp(Zf(e)),a=[],l=e.blocks||[];Object(s["b"])(!l||Object(s["t"])(l)),l=l||[];var u=t.orderMode;if(e.sortBlocks&&u){l=l.slice();var c={valueAsc:"asc",valueDesc:"desc"};if(Object(s["q"])(c,u)){var d=new hf(c[u],null);l.sort((function(t,e){return d.evaluate(t.sortParam,e.sortParam)}))}else"seriesDesc"===u&&l.reverse()}Object(s["k"])(l,(function(n,i){var l=e.valueFormatter,u=Xf(n)(l?Object(s["m"])(Object(s["m"])({},t),{valueFormatter:l}):t,n,i>0?o.html:0,r);null!=u&&a.push(u)}));var h="richText"===t.renderMode?a.join(o.richText):ep(a.join(""),i?n:o.html);if(i)return h;var f=kc(e.header,"ordinal",t.useUTC),p=Vf(r,t.renderMode).nameStyle;return"richText"===t.renderMode?ip(t,f,p)+o.richText+h:ep('<div style="'+p+";"+Wf+';">'+V(f)+"</div>"+h,n)}function Kf(t,e,n,r){var i=t.renderMode,o=e.noName,a=e.noValue,l=!e.markerType,u=e.name,c=t.useUTC,d=e.valueFormatter||t.valueFormatter||function(t){return t=Object(s["t"])(t)?t:[t],Object(s["H"])(t,(function(t,e){return kc(t,Object(s["t"])(p)?p[e]:p,c)}))};if(!o||!a){var h=l?"":t.markupStyleCreator.makeTooltipMarker(e.markerType,e.markerColor||"#333",i),f=o?"":kc(u,"ordinal",c),p=e.valueType,m=a?[]:d(e.value,e.dataIndex),g=!l||!o,y=!l&&o,v=Vf(r,i),_=v.nameStyle,b=v.valueStyle;return"richText"===i?(l?"":h)+(o?"":ip(t,f,_))+(a?"":op(t,m,g,y,b)):ep((l?"":h)+(o?"":np(f,!l,_))+(a?"":rp(m,g,y,b)),n)}}function Qf(t,e,n,r,i,o){if(t){var a=Xf(t),s={useUTC:i,renderMode:n,orderMode:r,markupStyleCreator:e,valueFormatter:t.valueFormatter};return a(s,t,0,o)}}function tp(t){return{html:Gf[t],richText:$f[t]}}function ep(t,e){var n='<div style="clear:both"></div>',r="margin: "+e+"px 0 0";return'<div style="'+r+";"+Wf+';">'+t+n+"</div>"}function np(t,e,n){var r=e?"margin-left:2px":"";return'<span style="'+n+";"+r+'">'+V(t)+"</span>"}function rp(t,e,n,r){var i=n?"10px":"20px",o=e?"float:right;margin-left:"+i:"";return t=Object(s["t"])(t)?t:[t],'<span style="'+o+";"+r+'">'+Object(s["H"])(t,(function(t){return V(t)})).join("&nbsp;&nbsp;")+"</span>"}function ip(t,e,n){return t.markupStyleCreator.wrapRichTextStyle(e,n)}function op(t,e,n,r,i){var o=[i],a=r?10:20;return n&&o.push({padding:[0,0,0,a],align:"right"}),t.markupStyleCreator.wrapRichTextStyle(Object(s["t"])(e)?e.join("  "):e,o)}function ap(t,e){var n=t.getData().getItemVisual(e,"style"),r=n[t.visualDrawType];return Cc(r)}function sp(t,e){var n=t.get("padding");return null!=n?n:"richText"===e?[8,10]:10}var lp=function(){function t(){this.richTextStyles={},this._nextStyleNameId=ri()}return t.prototype._generateStyleName=function(){return"__EC_aUTo_"+this._nextStyleNameId++},t.prototype.makeTooltipMarker=function(t,e,n){var r="richText"===n?this._generateStyleName():null,i=Ic({color:e,type:t,renderMode:n,markerId:r});return Object(s["C"])(i)?i:(this.richTextStyles[r]=i.style,i.content)},t.prototype.wrapRichTextStyle=function(t,e){var n={};Object(s["t"])(e)?Object(s["k"])(e,(function(t){return Object(s["m"])(n,t)})):Object(s["m"])(n,e);var r=this._generateStyleName();return this.richTextStyles[r]=n,"{"+r+"|"+t+"}"},t}();function up(t){var e,n,r,i,o=t.series,a=t.dataIndex,l=t.multipleSeries,u=o.getData(),c=u.mapDimensionsAll("defaultedTooltip"),d=c.length,h=o.getRawValue(a),f=Object(s["t"])(h),p=ap(o,a);if(d>1||f&&!d){var m=cp(h,o,a,c,p);e=m.inlineValues,n=m.inlineValueTypes,r=m.blocks,i=m.inlineValues[0]}else if(d){var g=u.getDimensionInfo(c[0]);i=e=Qh(u,a,c[0]),n=g.type}else i=e=f?h[0]:h;var y=Si(o),v=y&&o.name||"",_=u.getName(a),b=l?v:_;return Uf("section",{header:v,noHeader:l||!y,sortParam:i,blocks:[Uf("nameValue",{markerType:"item",markerColor:p,name:b,noName:!Object(s["T"])(b),value:e,valueType:n,dataIndex:a})].concat(r||[])})}function cp(t,e,n,r,i){var o=e.getData(),a=Object(s["N"])(t,(function(t,e,n){var r=o.getDimensionInfo(n);return t||r&&!1!==r.tooltip&&null!=r.displayName}),!1),l=[],u=[],c=[];function d(t,e){var n=o.getDimensionInfo(e);n&&!1!==n.otherDims.tooltip&&(a?c.push(Uf("nameValue",{markerType:"subItem",markerColor:i,name:n.displayName,value:t,valueType:n.type})):(l.push(t),u.push(n.type)))}return r.length?Object(s["k"])(r,(function(t){d(Qh(o,n,t),t)})):Object(s["k"])(t,d),{inlineValues:l,inlineValueTypes:u,blocks:c}}var dp=Ci();function hp(t,e){return t.getName(e)||t.getId(e)}var fp="__universalTransitionEnabled",pp=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e._selectedDataIndicesMap={},e}return o(e,t),e.prototype.init=function(t,e,n){this.seriesIndex=this.componentIndex,this.dataTask=rf({count:yp,reset:vp}),this.dataTask.context={model:this},this.mergeDefaultAndTheme(t,n);var r=dp(this).sourceManager=new Hf(this);r.prepareSource();var i=this.getInitialData(t,n);bp(i,this),this.dataTask.context.data=i,dp(this).dataBeforeProcessed=i,mp(this),this._initSelectedMapFromData(i)},e.prototype.mergeDefaultAndTheme=function(t,e){var n=Bc(this),r=n?Vc(t):{},i=this.subType;Xc.hasClass(i)&&(i+="Series"),s["I"](t,e.getTheme().get(this.subType)),s["I"](t,this.getDefaultOption()),ci(t,"label",["show"]),this.fillDataTextStyle(t.data),n&&Wc(t,r,n)},e.prototype.mergeOption=function(t,e){t=s["I"](this.option,t,!0),this.fillDataTextStyle(t.data);var n=Bc(this);n&&Wc(this.option,t,n);var r=dp(this).sourceManager;r.dirty(),r.prepareSource();var i=this.getInitialData(t,e);bp(i,this),this.dataTask.dirty(),this.dataTask.context.data=i,dp(this).dataBeforeProcessed=i,mp(this),this._initSelectedMapFromData(i)},e.prototype.fillDataTextStyle=function(t){if(t&&!s["E"](t))for(var e=["show"],n=0;n<t.length;n++)t[n]&&t[n].label&&ci(t[n],"label",e)},e.prototype.getInitialData=function(t,e){},e.prototype.appendData=function(t){var e=this.getRawData();e.appendData(t.data)},e.prototype.getData=function(t){var e=wp(this);if(e){var n=e.context.data;return null==t?n:n.getLinkedData(t)}return dp(this).data},e.prototype.getAllData=function(){var t=this.getData();return t&&t.getLinkedDataAll?t.getLinkedDataAll():[{data:t}]},e.prototype.setData=function(t){var e=wp(this);if(e){var n=e.context;n.outputData=t,e!==this.dataTask&&(n.data=t)}dp(this).data=t},e.prototype.getEncode=function(){var t=this.get("encode",!0);if(t)return s["f"](t)},e.prototype.getSourceManager=function(){return dp(this).sourceManager},e.prototype.getSource=function(){return this.getSourceManager().getSource()},e.prototype.getRawData=function(){return dp(this).dataBeforeProcessed},e.prototype.getColorBy=function(){var t=this.get("colorBy");return t||"series"},e.prototype.isColorBySeries=function(){return"series"===this.getColorBy()},e.prototype.getBaseAxis=function(){var t=this.coordinateSystem;return t&&t.getBaseAxis&&t.getBaseAxis()},e.prototype.formatTooltip=function(t,e,n){return up({series:this,dataIndex:t,multipleSeries:e})},e.prototype.isAnimationEnabled=function(){var t=this.ecModel;if(a["a"].node&&(!t||!t.ssr))return!1;var e=this.getShallow("animation");return e&&this.getData().count()>this.getShallow("animationThreshold")&&(e=!1),!!e},e.prototype.restoreData=function(){this.dataTask.dirty()},e.prototype.getColorFromPalette=function(t,e,n){var r=this.ecModel,i=wd.prototype.getColorFromPalette.call(this,t,e,n);return i||(i=r.getColorFromPalette(t,e,n)),i},e.prototype.coordDimToDataDim=function(t){return this.getRawData().mapDimensionsAll(t)},e.prototype.getProgressive=function(){return this.get("progressive")},e.prototype.getProgressiveThreshold=function(){return this.get("progressiveThreshold")},e.prototype.select=function(t,e){this._innerSelect(this.getData(e),t)},e.prototype.unselect=function(t,e){var n=this.option.selectedMap;if(n){var r=this.option.selectedMode,i=this.getData(e);if("series"===r||"all"===n)return this.option.selectedMap={},void(this._selectedDataIndicesMap={});for(var o=0;o<t.length;o++){var a=t[o],s=hp(i,a);n[s]=!1,this._selectedDataIndicesMap[s]=-1}}},e.prototype.toggleSelect=function(t,e){for(var n=[],r=0;r<t.length;r++)n[0]=t[r],this.isSelected(t[r],e)?this.unselect(n,e):this.select(n,e)},e.prototype.getSelectedDataIndices=function(){if("all"===this.option.selectedMap)return[].slice.call(this.getData().getIndices());for(var t=this._selectedDataIndicesMap,e=s["F"](t),n=[],r=0;r<e.length;r++){var i=t[e[r]];i>=0&&n.push(i)}return n},e.prototype.isSelected=function(t,e){var n=this.option.selectedMap;if(!n)return!1;var r=this.getData(e);return("all"===n||n[hp(r,t)])&&!r.getItemModel(t).get(["select","disabled"])},e.prototype.isUniversalTransitionEnabled=function(){if(this[fp])return!0;var t=this.option.universalTransition;return!!t&&(!0===t||t&&t.enabled)},e.prototype._innerSelect=function(t,e){var n,r,i=this.option,o=i.selectedMode,a=e.length;if(o&&a)if("series"===o)i.selectedMap="all";else if("multiple"===o){s["A"](i.selectedMap)||(i.selectedMap={});for(var l=i.selectedMap,u=0;u<a;u++){var c=e[u],d=hp(t,c);l[d]=!0,this._selectedDataIndicesMap[d]=t.getRawIndex(c)}}else if("single"===o||!0===o){var h=e[a-1];d=hp(t,h);i.selectedMap=(n={},n[d]=!0,n),this._selectedDataIndicesMap=(r={},r[d]=t.getRawIndex(h),r)}},e.prototype._initSelectedMapFromData=function(t){if(!this.option.selectedMap){var e=[];t.hasItemOption&&t.each((function(n){var r=t.getRawDataItem(n);r&&r.selected&&e.push(n)})),e.length>0&&this._innerSelect(t,e)}},e.registerClass=function(t){return Xc.registerClass(t)},e.protoInitialize=function(){var t=e.prototype;t.type="series.__base__",t.seriesIndex=0,t.ignoreStyleOnData=!1,t.hasSymbolVisual=!1,t.defaultSymbol="circle",t.visualStyleAccessPath="itemStyle",t.visualDrawType="fill"}(),e}(Xc);function mp(t){var e=t.name;Si(t)||(t.name=gp(t)||e)}function gp(t){var e=t.getRawData(),n=e.mapDimensionsAll("seriesName"),r=[];return s["k"](n,(function(t){var n=e.getDimensionInfo(t);n.displayName&&r.push(n.displayName)})),r.join(" ")}function yp(t){return t.model.getRawData().count()}function vp(t){var e=t.model;return e.setData(e.getRawData().cloneShallow()),_p}function _p(t,e){e.outputData&&t.end>e.outputData.count()&&e.model.getRawData().cloneShallow(e.outputData)}function bp(t,e){s["k"](s["e"](t.CHANGABLE_METHODS,t.DOWNSAMPLE_METHODS),(function(n){t.wrapMethod(n,s["h"](xp,e))}))}function xp(t,e){var n=wp(t);return n&&n.setOutputEnd((e||this).count()),e}function wp(t){var e=(t.ecModel||{}).scheduler,n=e&&e.getPipeline(t.uid);if(n){var r=n.currentTask;if(r){var i=r.agentStubMap;i&&(r=i.get(t.uid))}return r}}s["K"](pp,ef),s["K"](pp,wd),Ji(pp,Xc);var Mp=pp,Sp=function(){function t(){this.group=new wr,this.uid=Du("viewComponent")}return t.prototype.init=function(t,e){},t.prototype.render=function(t,e,n,r){},t.prototype.dispose=function(t,e){},t.prototype.updateView=function(t,e,n,r){},t.prototype.updateLayout=function(t,e,n,r){},t.prototype.updateVisual=function(t,e,n,r){},t.prototype.toggleBlurSeries=function(t,e,n){},t.prototype.eachRendered=function(t){var e=this.group;e&&e.traverse(t)},t}();Xi(Sp),no(Sp);var kp=Sp;function Lp(){var t=Ci();return function(e){var n=t(e),r=e.pipelineContext,i=!!n.large,o=!!n.progressiveRender,a=n.large=!(!r||!r.large),s=n.progressiveRender=!(!r||!r.progressiveRender);return!(i===a&&o===s)&&"reset"}}var Tp=va.CMD,Dp=[[],[],[]],Op=Math.sqrt,Ip=Math.atan2;function Cp(t,e){if(e){var n,r,i,o,a,s,l=t.data,u=t.len(),c=Tp.M,d=Tp.C,h=Tp.L,f=Tp.R,p=Tp.A,m=Tp.Q;for(i=0,o=0;i<u;){switch(n=l[i++],o=i,r=0,n){case c:r=1;break;case h:r=1;break;case d:r=3;break;case m:r=2;break;case p:var g=e[4],y=e[5],v=Op(e[0]*e[0]+e[1]*e[1]),_=Op(e[2]*e[2]+e[3]*e[3]),b=Ip(-e[1]/_,e[0]/v);l[i]*=v,l[i++]+=g,l[i]*=_,l[i++]+=y,l[i++]*=v,l[i++]*=_,l[i++]+=b,l[i++]+=b,i+=2,o=i;break;case f:s[0]=l[i++],s[1]=l[i++],S(s,s,e),l[o++]=s[0],l[o++]=s[1],s[0]+=l[i++],s[1]+=l[i++],S(s,s,e),l[o++]=s[0],l[o++]=s[1]}for(a=0;a<r;a++){var x=Dp[a];x[0]=l[i++],x[1]=l[i++],S(x,x,e),l[o++]=x[0],l[o++]=x[1]}}t.increaseVersion()}}var Ap=Math.sqrt,Yp=Math.sin,jp=Math.cos,Pp=Math.PI;function Ep(t){return Math.sqrt(t[0]*t[0]+t[1]*t[1])}function Np(t,e){return(t[0]*e[0]+t[1]*e[1])/(Ep(t)*Ep(e))}function Rp(t,e){return(t[0]*e[1]<t[1]*e[0]?-1:1)*Math.acos(Np(t,e))}function Hp(t,e,n,r,i,o,a,s,l,u,c){var d=l*(Pp/180),h=jp(d)*(t-n)/2+Yp(d)*(e-r)/2,f=-1*Yp(d)*(t-n)/2+jp(d)*(e-r)/2,p=h*h/(a*a)+f*f/(s*s);p>1&&(a*=Ap(p),s*=Ap(p));var m=(i===o?-1:1)*Ap((a*a*(s*s)-a*a*(f*f)-s*s*(h*h))/(a*a*(f*f)+s*s*(h*h)))||0,g=m*a*f/s,y=m*-s*h/a,v=(t+n)/2+jp(d)*g-Yp(d)*y,_=(e+r)/2+Yp(d)*g+jp(d)*y,b=Rp([1,0],[(h-g)/a,(f-y)/s]),x=[(h-g)/a,(f-y)/s],w=[(-1*h-g)/a,(-1*f-y)/s],M=Rp(x,w);if(Np(x,w)<=-1&&(M=Pp),Np(x,w)>=1&&(M=0),M<0){var S=Math.round(M/Pp*1e6)/1e6;M=2*Pp+S%2*Pp}c.addData(u,v,_,a,s,b,M,d,o)}var zp=/([mlvhzcqtsa])([^mlvhzcqtsa]*)/gi,Fp=/-?([0-9]*\.)?[0-9]+([eE]-?[0-9]+)?/g;function Bp(t){var e=new va;if(!t)return e;var n,r=0,i=0,o=r,a=i,s=va.CMD,l=t.match(zp);if(!l)return e;for(var u=0;u<l.length;u++){for(var c=l[u],d=c.charAt(0),h=void 0,f=c.match(Fp)||[],p=f.length,m=0;m<p;m++)f[m]=parseFloat(f[m]);var g=0;while(g<p){var y=void 0,v=void 0,_=void 0,b=void 0,x=void 0,w=void 0,M=void 0,S=r,k=i,L=void 0,T=void 0;switch(d){case"l":r+=f[g++],i+=f[g++],h=s.L,e.addData(h,r,i);break;case"L":r=f[g++],i=f[g++],h=s.L,e.addData(h,r,i);break;case"m":r+=f[g++],i+=f[g++],h=s.M,e.addData(h,r,i),o=r,a=i,d="l";break;case"M":r=f[g++],i=f[g++],h=s.M,e.addData(h,r,i),o=r,a=i,d="L";break;case"h":r+=f[g++],h=s.L,e.addData(h,r,i);break;case"H":r=f[g++],h=s.L,e.addData(h,r,i);break;case"v":i+=f[g++],h=s.L,e.addData(h,r,i);break;case"V":i=f[g++],h=s.L,e.addData(h,r,i);break;case"C":h=s.C,e.addData(h,f[g++],f[g++],f[g++],f[g++],f[g++],f[g++]),r=f[g-2],i=f[g-1];break;case"c":h=s.C,e.addData(h,f[g++]+r,f[g++]+i,f[g++]+r,f[g++]+i,f[g++]+r,f[g++]+i),r+=f[g-2],i+=f[g-1];break;case"S":y=r,v=i,L=e.len(),T=e.data,n===s.C&&(y+=r-T[L-4],v+=i-T[L-3]),h=s.C,S=f[g++],k=f[g++],r=f[g++],i=f[g++],e.addData(h,y,v,S,k,r,i);break;case"s":y=r,v=i,L=e.len(),T=e.data,n===s.C&&(y+=r-T[L-4],v+=i-T[L-3]),h=s.C,S=r+f[g++],k=i+f[g++],r+=f[g++],i+=f[g++],e.addData(h,y,v,S,k,r,i);break;case"Q":S=f[g++],k=f[g++],r=f[g++],i=f[g++],h=s.Q,e.addData(h,S,k,r,i);break;case"q":S=f[g++]+r,k=f[g++]+i,r+=f[g++],i+=f[g++],h=s.Q,e.addData(h,S,k,r,i);break;case"T":y=r,v=i,L=e.len(),T=e.data,n===s.Q&&(y+=r-T[L-4],v+=i-T[L-3]),r=f[g++],i=f[g++],h=s.Q,e.addData(h,y,v,r,i);break;case"t":y=r,v=i,L=e.len(),T=e.data,n===s.Q&&(y+=r-T[L-4],v+=i-T[L-3]),r+=f[g++],i+=f[g++],h=s.Q,e.addData(h,y,v,r,i);break;case"A":_=f[g++],b=f[g++],x=f[g++],w=f[g++],M=f[g++],S=r,k=i,r=f[g++],i=f[g++],h=s.A,Hp(S,k,r,i,w,M,_,b,x,h,e);break;case"a":_=f[g++],b=f[g++],x=f[g++],w=f[g++],M=f[g++],S=r,k=i,r+=f[g++],i+=f[g++],h=s.A,Hp(S,k,r,i,w,M,_,b,x,h,e);break}}"z"!==d&&"Z"!==d||(h=s.Z,e.addData(h),r=o,i=a),n=h}return e.toStatic(),e}var Wp=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.applyTransform=function(t){},e}(Va);function Vp(t){return null!=t.setData}function Gp(t,e){var n=Bp(t),r=Object(s["m"])({},e);return r.buildPath=function(t){if(Vp(t)){t.setData(n.data);var e=t.getContext();e&&t.rebuildPath(e,1)}else{e=t;n.rebuildPath(e,1)}},r.applyTransform=function(t){Cp(n,t),this.dirtyShape()},r}function $p(t,e){return new Wp(Gp(t,e))}function Up(t,e){var n=Gp(t,e),r=function(t){function e(e){var r=t.call(this,e)||this;return r.applyTransform=n.applyTransform,r.buildPath=n.buildPath,r}return o(e,t),e}(Wp);return r}function qp(t,e){for(var n=[],r=t.length,i=0;i<r;i++){var o=t[i];n.push(o.getUpdatedPathProxy(!0))}var a=new Va(e);return a.createPathProxy(),a.buildPath=function(t){if(Vp(t)){t.appendPath(n);var e=t.getContext();e&&t.rebuildPath(e,1)}},a}function Xp(t,e){e=e||{};var n=new Va;return t.shape&&n.setShape(t.shape),n.setStyle(t.style),e.bakeTransform?Cp(n.path,t.getComputedTransform()):e.toLocal?n.setLocalTransform(t.getComputedTransform()):n.copyTransform(t),n.buildPath=t.buildPath,n.applyTransform=n.applyTransform,n.z=t.z,n.z2=t.z2,n.zlevel=t.zlevel,n}var Zp=function(){function t(){this.cx=0,this.cy=0,this.r=0}return t}(),Jp=function(t){function e(e){return t.call(this,e)||this}return o(e,t),e.prototype.getDefaultShape=function(){return new Zp},e.prototype.buildPath=function(t,e){t.moveTo(e.cx+e.r,e.cy),t.arc(e.cx,e.cy,e.r,0,2*Math.PI)},e}(Va);Jp.prototype.type="circle";var Kp=Jp,Qp=function(){function t(){this.cx=0,this.cy=0,this.rx=0,this.ry=0}return t}(),tm=function(t){function e(e){return t.call(this,e)||this}return o(e,t),e.prototype.getDefaultShape=function(){return new Qp},e.prototype.buildPath=function(t,e){var n=.5522848,r=e.cx,i=e.cy,o=e.rx,a=e.ry,s=o*n,l=a*n;t.moveTo(r-o,i),t.bezierCurveTo(r-o,i-l,r-s,i-a,r,i-a),t.bezierCurveTo(r+s,i-a,r+o,i-l,r+o,i),t.bezierCurveTo(r+o,i+l,r+s,i+a,r,i+a),t.bezierCurveTo(r-s,i+a,r-o,i+l,r-o,i),t.closePath()},e}(Va);tm.prototype.type="ellipse";var em=tm,nm=Math.PI,rm=2*nm,im=Math.sin,om=Math.cos,am=Math.acos,sm=Math.atan2,lm=Math.abs,um=Math.sqrt,cm=Math.max,dm=Math.min,hm=1e-4;function fm(t,e,n,r,i,o,a,s){var l=n-t,u=r-e,c=a-i,d=s-o,h=d*l-c*u;if(!(h*h<hm))return h=(c*(e-o)-d*(t-i))/h,[t+h*l,e+h*u]}function pm(t,e,n,r,i,o,a){var s=t-n,l=e-r,u=(a?o:-o)/um(s*s+l*l),c=u*l,d=-u*s,h=t+c,f=e+d,p=n+c,m=r+d,g=(h+p)/2,y=(f+m)/2,v=p-h,_=m-f,b=v*v+_*_,x=i-o,w=h*m-p*f,M=(_<0?-1:1)*um(cm(0,x*x*b-w*w)),S=(w*_-v*M)/b,k=(-w*v-_*M)/b,L=(w*_+v*M)/b,T=(-w*v+_*M)/b,D=S-g,O=k-y,I=L-g,C=T-y;return D*D+O*O>I*I+C*C&&(S=L,k=T),{cx:S,cy:k,x0:-c,y0:-d,x1:S*(i/x-1),y1:k*(i/x-1)}}function mm(t){var e;if(Object(s["t"])(t)){var n=t.length;if(!n)return t;e=1===n?[t[0],t[0],0,0]:2===n?[t[0],t[0],t[1],t[1]]:3===n?t.concat(t[2]):t}else e=[t,t,t,t];return e}function gm(t,e){var n,r=cm(e.r,0),i=cm(e.r0||0,0),o=r>0,a=i>0;if(o||a){if(o||(r=i,i=0),i>r){var s=r;r=i,i=s}var l=e.startAngle,u=e.endAngle;if(!isNaN(l)&&!isNaN(u)){var c=e.cx,d=e.cy,h=!!e.clockwise,f=lm(u-l),p=f>rm&&f%rm;if(p>hm&&(f=p),r>hm)if(f>rm-hm)t.moveTo(c+r*om(l),d+r*im(l)),t.arc(c,d,r,l,u,!h),i>hm&&(t.moveTo(c+i*om(u),d+i*im(u)),t.arc(c,d,i,u,l,h));else{var m=void 0,g=void 0,y=void 0,v=void 0,_=void 0,b=void 0,x=void 0,w=void 0,M=void 0,S=void 0,k=void 0,L=void 0,T=void 0,D=void 0,O=void 0,I=void 0,C=r*om(l),A=r*im(l),Y=i*om(u),j=i*im(u),P=f>hm;if(P){var E=e.cornerRadius;E&&(n=mm(E),m=n[0],g=n[1],y=n[2],v=n[3]);var N=lm(r-i)/2;if(_=dm(N,y),b=dm(N,v),x=dm(N,m),w=dm(N,g),k=M=cm(_,b),L=S=cm(x,w),(M>hm||S>hm)&&(T=r*om(u),D=r*im(u),O=i*om(l),I=i*im(l),f<nm)){var R=fm(C,A,O,I,T,D,Y,j);if(R){var H=C-R[0],z=A-R[1],F=T-R[0],B=D-R[1],W=1/im(am((H*F+z*B)/(um(H*H+z*z)*um(F*F+B*B)))/2),V=um(R[0]*R[0]+R[1]*R[1]);k=dm(M,(r-V)/(W+1)),L=dm(S,(i-V)/(W-1))}}}if(P)if(k>hm){var G=dm(y,k),$=dm(v,k),U=pm(O,I,C,A,r,G,h),q=pm(T,D,Y,j,r,$,h);t.moveTo(c+U.cx+U.x0,d+U.cy+U.y0),k<M&&G===$?t.arc(c+U.cx,d+U.cy,k,sm(U.y0,U.x0),sm(q.y0,q.x0),!h):(G>0&&t.arc(c+U.cx,d+U.cy,G,sm(U.y0,U.x0),sm(U.y1,U.x1),!h),t.arc(c,d,r,sm(U.cy+U.y1,U.cx+U.x1),sm(q.cy+q.y1,q.cx+q.x1),!h),$>0&&t.arc(c+q.cx,d+q.cy,$,sm(q.y1,q.x1),sm(q.y0,q.x0),!h))}else t.moveTo(c+C,d+A),t.arc(c,d,r,l,u,!h);else t.moveTo(c+C,d+A);if(i>hm&&P)if(L>hm){G=dm(m,L),$=dm(g,L),U=pm(Y,j,T,D,i,-$,h),q=pm(C,A,O,I,i,-G,h);t.lineTo(c+U.cx+U.x0,d+U.cy+U.y0),L<S&&G===$?t.arc(c+U.cx,d+U.cy,L,sm(U.y0,U.x0),sm(q.y0,q.x0),!h):($>0&&t.arc(c+U.cx,d+U.cy,$,sm(U.y0,U.x0),sm(U.y1,U.x1),!h),t.arc(c,d,i,sm(U.cy+U.y1,U.cx+U.x1),sm(q.cy+q.y1,q.cx+q.x1),h),G>0&&t.arc(c+q.cx,d+q.cy,G,sm(q.y1,q.x1),sm(q.y0,q.x0),!h))}else t.lineTo(c+Y,d+j),t.arc(c,d,i,u,l,h);else t.lineTo(c+Y,d+j)}else t.moveTo(c,d);t.closePath()}}}var ym=function(){function t(){this.cx=0,this.cy=0,this.r0=0,this.r=0,this.startAngle=0,this.endAngle=2*Math.PI,this.clockwise=!0,this.cornerRadius=0}return t}(),vm=function(t){function e(e){return t.call(this,e)||this}return o(e,t),e.prototype.getDefaultShape=function(){return new ym},e.prototype.buildPath=function(t,e){gm(t,e)},e.prototype.isZeroArea=function(){return this.shape.startAngle===this.shape.endAngle||this.shape.r===this.shape.r0},e}(Va);vm.prototype.type="sector";var _m=vm,bm=function(){function t(){this.cx=0,this.cy=0,this.r=0,this.r0=0}return t}(),xm=function(t){function e(e){return t.call(this,e)||this}return o(e,t),e.prototype.getDefaultShape=function(){return new bm},e.prototype.buildPath=function(t,e){var n=e.cx,r=e.cy,i=2*Math.PI;t.moveTo(n+e.r,r),t.arc(n,r,e.r,0,i,!1),t.moveTo(n+e.r0,r),t.arc(n,r,e.r0,0,i,!0)},e}(Va);xm.prototype.type="ring";var wm=xm;function Mm(t,e,n,r){var i,o,a,s,l=[],u=[],d=[],f=[];if(r){a=[1/0,1/0],s=[-1/0,-1/0];for(var m=0,g=t.length;m<g;m++)k(a,a,t[m]),L(s,s,t[m]);k(a,a,r[0]),L(s,s,r[1])}for(m=0,g=t.length;m<g;m++){var v=t[m];if(n)i=t[m?m-1:g-1],o=t[(m+1)%g];else{if(0===m||m===g-1){l.push(c(t[m]));continue}i=t[m-1],o=t[m+1]}p(u,o,i),y(u,u,e);var b=_(v,i),x=_(v,o),w=b+x;0!==w&&(b/=w,x/=w),y(d,u,-b),y(f,u,x);var M=h([],v,d),S=h([],v,f);r&&(L(M,M,a),k(M,M,s),L(S,S,a),k(S,S,s)),l.push(M),l.push(S)}return n&&l.push(l.shift()),l}function Sm(t,e,n){var r=e.smooth,i=e.points;if(i&&i.length>=2){if(r){var o=Mm(i,r,n,e.smoothConstraint);t.moveTo(i[0][0],i[0][1]);for(var a=i.length,s=0;s<(n?a:a-1);s++){var l=o[2*s],u=o[2*s+1],c=i[(s+1)%a];t.bezierCurveTo(l[0],l[1],u[0],u[1],c[0],c[1])}}else{t.moveTo(i[0][0],i[0][1]);s=1;for(var d=i.length;s<d;s++)t.lineTo(i[s][0],i[s][1])}n&&t.closePath()}}var km=function(){function t(){this.points=null,this.smooth=0,this.smoothConstraint=null}return t}(),Lm=function(t){function e(e){return t.call(this,e)||this}return o(e,t),e.prototype.getDefaultShape=function(){return new km},e.prototype.buildPath=function(t,e){Sm(t,e,!0)},e}(Va);Lm.prototype.type="polygon";var Tm=Lm,Dm=function(){function t(){this.points=null,this.percent=1,this.smooth=0,this.smoothConstraint=null}return t}(),Om=function(t){function e(e){return t.call(this,e)||this}return o(e,t),e.prototype.getDefaultStyle=function(){return{stroke:"#000",fill:null}},e.prototype.getDefaultShape=function(){return new Dm},e.prototype.buildPath=function(t,e){Sm(t,e,!1)},e}(Va);Om.prototype.type="polyline";var Im=Om,Cm={},Am=function(){function t(){this.x1=0,this.y1=0,this.x2=0,this.y2=0,this.percent=1}return t}(),Ym=function(t){function e(e){return t.call(this,e)||this}return o(e,t),e.prototype.getDefaultStyle=function(){return{stroke:"#000",fill:null}},e.prototype.getDefaultShape=function(){return new Am},e.prototype.buildPath=function(t,e){var n,r,i,o;if(this.subPixelOptimize){var a=es(Cm,e,this.style);n=a.x1,r=a.y1,i=a.x2,o=a.y2}else n=e.x1,r=e.y1,i=e.x2,o=e.y2;var s=e.percent;0!==s&&(t.moveTo(n,r),s<1&&(i=n*(1-s)+i*s,o=r*(1-s)+o*s),t.lineTo(i,o))},e.prototype.pointAt=function(t){var e=this.shape;return[e.x1*(1-t)+e.x2*t,e.y1*(1-t)+e.y2*t]},e}(Va);Ym.prototype.type="line";var jm=Ym,Pm=[],Em=function(){function t(){this.x1=0,this.y1=0,this.x2=0,this.y2=0,this.cpx1=0,this.cpy1=0,this.percent=1}return t}();function Nm(t,e,n){var r=t.cpx2,i=t.cpy2;return null!=r||null!=i?[(n?be:_e)(t.x1,t.cpx1,t.cpx2,t.x2,e),(n?be:_e)(t.y1,t.cpy1,t.cpy2,t.y2,e)]:[(n?Te:Le)(t.x1,t.cpx1,t.x2,e),(n?Te:Le)(t.y1,t.cpy1,t.y2,e)]}var Rm=function(t){function e(e){return t.call(this,e)||this}return o(e,t),e.prototype.getDefaultStyle=function(){return{stroke:"#000",fill:null}},e.prototype.getDefaultShape=function(){return new Em},e.prototype.buildPath=function(t,e){var n=e.x1,r=e.y1,i=e.x2,o=e.y2,a=e.cpx1,s=e.cpy1,l=e.cpx2,u=e.cpy2,c=e.percent;0!==c&&(t.moveTo(n,r),null==l||null==u?(c<1&&(Ie(n,a,i,c,Pm),a=Pm[1],i=Pm[2],Ie(r,s,o,c,Pm),s=Pm[1],o=Pm[2]),t.quadraticCurveTo(a,s,i,o)):(c<1&&(Me(n,a,l,i,c,Pm),a=Pm[1],l=Pm[2],i=Pm[3],Me(r,s,u,o,c,Pm),s=Pm[1],u=Pm[2],o=Pm[3]),t.bezierCurveTo(a,s,l,u,i,o)))},e.prototype.pointAt=function(t){return Nm(this.shape,t,!1)},e.prototype.tangentAt=function(t){var e=Nm(this.shape,t,!0);return v(e,e)},e}(Va);Rm.prototype.type="bezier-curve";var Hm=Rm,zm=function(){function t(){this.cx=0,this.cy=0,this.r=0,this.startAngle=0,this.endAngle=2*Math.PI,this.clockwise=!0}return t}(),Fm=function(t){function e(e){return t.call(this,e)||this}return o(e,t),e.prototype.getDefaultStyle=function(){return{stroke:"#000",fill:null}},e.prototype.getDefaultShape=function(){return new zm},e.prototype.buildPath=function(t,e){var n=e.cx,r=e.cy,i=Math.max(e.r,0),o=e.startAngle,a=e.endAngle,s=e.clockwise,l=Math.cos(o),u=Math.sin(o);t.moveTo(l*i+n,u*i+r),t.arc(n,r,i,o,a,!s)},e}(Va);Fm.prototype.type="arc";var Bm=Fm,Wm=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.type="compound",e}return o(e,t),e.prototype._updatePathDirty=function(){for(var t=this.shape.paths,e=this.shapeChanged(),n=0;n<t.length;n++)e=e||t[n].shapeChanged();e&&this.dirtyShape()},e.prototype.beforeBrush=function(){this._updatePathDirty();for(var t=this.shape.paths||[],e=this.getGlobalScale(),n=0;n<t.length;n++)t[n].path||t[n].createPathProxy(),t[n].path.setScale(e[0],e[1],t[n].segmentIgnoreThreshold)},e.prototype.buildPath=function(t,e){for(var n=e.paths||[],r=0;r<n.length;r++)n[r].buildPath(t,n[r].shape,!0)},e.prototype.afterBrush=function(){for(var t=this.shape.paths||[],e=0;e<t.length;e++)t[e].pathUpdated()},e.prototype.getBoundingRect=function(){return this._updatePathDirty.call(this),Va.prototype.getBoundingRect.call(this)},e}(Va),Vm=Wm,Gm=function(){function t(t){this.colorStops=t||[]}return t.prototype.addColorStop=function(t,e){this.colorStops.push({offset:t,color:e})},t}(),$m=Gm,Um=function(t){function e(e,n,r,i,o,a){var s=t.call(this,o)||this;return s.x=null==e?0:e,s.y=null==n?0:n,s.x2=null==r?1:r,s.y2=null==i?0:i,s.type="linear",s.global=a||!1,s}return o(e,t),e}($m),qm=Um,Xm=function(t){function e(e,n,r,i,o){var a=t.call(this,i)||this;return a.x=null==e?.5:e,a.y=null==n?.5:n,a.r=null==r?.5:r,a.type="radial",a.global=o||!1,a}return o(e,t),e}($m),Zm=Xm,Jm=[0,0],Km=[0,0],Qm=new yt,tg=new yt,eg=function(){function t(t,e){this._corners=[],this._axes=[],this._origin=[0,0];for(var n=0;n<4;n++)this._corners[n]=new yt;for(n=0;n<2;n++)this._axes[n]=new yt;t&&this.fromBoundingRect(t,e)}return t.prototype.fromBoundingRect=function(t,e){var n=this._corners,r=this._axes,i=t.x,o=t.y,a=i+t.width,s=o+t.height;if(n[0].set(i,o),n[1].set(a,o),n[2].set(a,s),n[3].set(i,s),e)for(var l=0;l<4;l++)n[l].transform(e);yt.sub(r[0],n[1],n[0]),yt.sub(r[1],n[3],n[0]),r[0].normalize(),r[1].normalize();for(l=0;l<2;l++)this._origin[l]=r[l].dot(n[0])},t.prototype.intersect=function(t,e){var n=!0,r=!e;return Qm.set(1/0,1/0),tg.set(0,0),!this._intersectCheckOneSide(this,t,Qm,tg,r,1)&&(n=!1,r)||!this._intersectCheckOneSide(t,this,Qm,tg,r,-1)&&(n=!1,r)||r||yt.copy(e,n?Qm:tg),n},t.prototype._intersectCheckOneSide=function(t,e,n,r,i,o){for(var a=!0,s=0;s<2;s++){var l=this._axes[s];if(this._getProjMinMaxOnAxis(s,t._corners,Jm),this._getProjMinMaxOnAxis(s,e._corners,Km),Jm[1]<Km[0]||Jm[0]>Km[1]){if(a=!1,i)return a;var u=Math.abs(Km[0]-Jm[1]),c=Math.abs(Jm[0]-Km[1]);Math.min(u,c)>r.len()&&(u<c?yt.scale(r,l,-u*o):yt.scale(r,l,c*o))}else if(n){u=Math.abs(Km[0]-Jm[1]),c=Math.abs(Jm[0]-Km[1]);Math.min(u,c)<n.len()&&(u<c?yt.scale(n,l,u*o):yt.scale(n,l,-c*o))}}return a},t.prototype._getProjMinMaxOnAxis=function(t,e,n){for(var r=this._axes[t],i=this._origin,o=e[0].dot(r)+i[t],a=o,s=o,l=1;l<e.length;l++){var u=e[l].dot(r)+i[t];a=Math.min(u,a),s=Math.max(u,s)}n[0]=a,n[1]=s},t}(),ng=eg,rg=[],ig=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.notClear=!0,e.incremental=!0,e._displayables=[],e._temporaryDisplayables=[],e._cursor=0,e}return o(e,t),e.prototype.traverse=function(t,e){t.call(e,this)},e.prototype.useStyle=function(){this.style={}},e.prototype.getCursor=function(){return this._cursor},e.prototype.innerAfterBrush=function(){this._cursor=this._displayables.length},e.prototype.clearDisplaybles=function(){this._displayables=[],this._temporaryDisplayables=[],this._cursor=0,this.markRedraw(),this.notClear=!1},e.prototype.clearTemporalDisplayables=function(){this._temporaryDisplayables=[]},e.prototype.addDisplayable=function(t,e){e?this._temporaryDisplayables.push(t):this._displayables.push(t),this.markRedraw()},e.prototype.addDisplayables=function(t,e){e=e||!1;for(var n=0;n<t.length;n++)this.addDisplayable(t[n],e)},e.prototype.getDisplayables=function(){return this._displayables},e.prototype.getTemporalDisplayables=function(){return this._temporaryDisplayables},e.prototype.eachPendingDisplayable=function(t){for(var e=this._cursor;e<this._displayables.length;e++)t&&t(this._displayables[e]);for(e=0;e<this._temporaryDisplayables.length;e++)t&&t(this._temporaryDisplayables[e])},e.prototype.update=function(){this.updateTransform();for(var t=this._cursor;t<this._displayables.length;t++){var e=this._displayables[t];e.parent=this,e.update(),e.parent=null}for(t=0;t<this._temporaryDisplayables.length;t++){e=this._temporaryDisplayables[t];e.parent=this,e.update(),e.parent=null}},e.prototype.getBoundingRect=function(){if(!this._rect){for(var t=new Tt(1/0,1/0,-1/0,-1/0),e=0;e<this._displayables.length;e++){var n=this._displayables[e],r=n.getBoundingRect().clone();n.needLocalTransform()&&r.applyTransform(n.getLocalTransform(rg)),t.union(r)}this._rect=t}return this._rect},e.prototype.contain=function(t,e){var n=this.transformCoordToLocal(t,e),r=this.getBoundingRect();if(r.contain(n[0],n[1]))for(var i=0;i<this._displayables.length;i++){var o=this._displayables[i];if(o.contain(t,e))return!0}return!1},e}(No),og=ig,ag=Math.max,sg=Math.min,lg={};function ug(t){return Va.extend(t)}var cg=Up;function dg(t,e){return cg(t,e)}function hg(t,e){lg[t]=e}function fg(t){if(lg.hasOwnProperty(t))return lg[t]}function pg(t,e,n,r){var i=$p(t,e);return n&&("center"===r&&(n=gg(n,i.getBoundingRect())),vg(i,n)),i}function mg(t,e,n){var r=new Ka({style:{image:t,x:e.x,y:e.y,width:e.width,height:e.height},onload:function(t){if("center"===n){var i={width:t.width,height:t.height};r.setStyle(gg(e,i))}}});return r}function gg(t,e){var n,r=e.width/e.height,i=t.height*r;i<=t.width?n=t.height:(i=t.width,n=i/r);var o=t.x+t.width/2,a=t.y+t.height/2;return{x:o-i/2,y:a-n/2,width:i,height:n}}var yg=qp;function vg(t,e){if(t.applyTransform){var n=t.getBoundingRect(),r=n.calculateTransform(e);t.applyTransform(r)}}function _g(t,e){return es(t,t,{lineWidth:e}),t}function bg(t){return ns(t.shape,t.shape,t.style),t}var xg=rs;function wg(t,e){var n=lt([]);while(t&&t!==e)ct(n,t.getLocalTransform(),n),t=t.parent;return n}function Mg(t,e,n){return e&&!Object(s["u"])(e)&&(e=Xn.getLocalTransform(e)),n&&(e=pt([],e)),S([],t,e)}function Sg(t,e,n){var r=0===e[4]||0===e[5]||0===e[0]?1:Math.abs(2*e[4]/e[0]),i=0===e[4]||0===e[5]||0===e[2]?1:Math.abs(2*e[4]/e[2]),o=["left"===t?-r:"right"===t?r:0,"top"===t?-i:"bottom"===t?i:0];return o=Mg(o,e,n),Math.abs(o[0])>Math.abs(o[1])?o[0]>0?"right":"left":o[1]>0?"bottom":"top"}function kg(t){return!t.isGroup}function Lg(t){return null!=t.shape}function Tg(t,e,n){if(t&&e){var r=i(t);e.traverse((function(t){if(kg(t)&&t.anid){var e=r[t.anid];if(e){var i=o(t);t.attr(o(e)),Wl(t,i,n,Ls(t).dataIndex)}}}))}function i(t){var e={};return t.traverse((function(t){kg(t)&&t.anid&&(e[t.anid]=t)})),e}function o(t){var e={x:t.x,y:t.y,rotation:t.rotation};return Lg(t)&&(e.shape=Object(s["m"])({},t.shape)),e}}function Dg(t,e){return Object(s["H"])(t,(function(t){var n=t[0];n=ag(n,e.x),n=sg(n,e.x+e.width);var r=t[1];return r=ag(r,e.y),r=sg(r,e.y+e.height),[n,r]}))}function Og(t,e){var n=ag(t.x,e.x),r=sg(t.x+t.width,e.x+e.width),i=ag(t.y,e.y),o=sg(t.y+t.height,e.y+e.height);if(r>=n&&o>=i)return{x:n,y:i,width:r-n,height:o-i}}function Ig(t,e,n){var r=Object(s["m"])({rectHover:!0},e),i=r.style={strokeNoScale:!0};if(n=n||{x:-1,y:-1,width:2,height:2},t)return 0===t.indexOf("image://")?(i.image=t.slice(8),Object(s["i"])(i,n),new Ka(r)):pg(t.replace("path://",""),r,n,"center")}function Cg(t,e,n,r,i){for(var o=0,a=i[i.length-1];o<i.length;o++){var s=i[o];if(Ag(t,e,n,r,s[0],s[1],a[0],a[1]))return!0;a=s}}function Ag(t,e,n,r,i,o,a,s){var l=n-t,u=r-e,c=a-i,d=s-o,h=Yg(c,d,l,u);if(jg(h))return!1;var f=t-i,p=e-o,m=Yg(f,p,l,u)/h;if(m<0||m>1)return!1;var g=Yg(f,p,c,d)/h;return!(g<0||g>1)}function Yg(t,e,n,r){return t*r-n*e}function jg(t){return t<=1e-6&&t>=-1e-6}function Pg(t){var e=t.itemTooltipOption,n=t.componentModel,r=t.itemName,i=Object(s["C"])(e)?{formatter:e}:e,o=n.mainType,a=n.componentIndex,l={componentType:o,name:r,$vars:["name"]};l[o+"Index"]=a;var u=t.formatterParamsExtra;u&&Object(s["k"])(Object(s["F"])(u),(function(t){Object(s["q"])(l,t)||(l[t]=u[t],l.$vars.push(t))}));var c=Ls(t.el);c.componentMainType=o,c.componentIndex=a,c.tooltipConfig={name:r,option:Object(s["i"])({content:r,formatterParams:l},i)}}function Eg(t,e){var n;t.isGroup&&(n=e(t)),n||t.traverse(e)}function Ng(t,e){if(t)if(Object(s["t"])(t))for(var n=0;n<t.length;n++)Eg(t[n],e);else Eg(t,e)}hg("circle",Kp),hg("ellipse",em),hg("sector",_m),hg("ring",wm),hg("polygon",Tm),hg("polyline",Im),hg("rect",ss),hg("line",jm),hg("bezierCurve",Hm),hg("arc",Bm);var Rg=Ci(),Hg=Lp(),zg=function(){function t(){this.group=new wr,this.uid=Du("viewChart"),this.renderTask=rf({plan:Wg,reset:Vg}),this.renderTask.context={view:this}}return t.prototype.init=function(t,e){},t.prototype.render=function(t,e,n,r){0},t.prototype.highlight=function(t,e,n,r){var i=t.getData(r&&r.dataType);i&&Bg(i,r,"emphasis")},t.prototype.downplay=function(t,e,n,r){var i=t.getData(r&&r.dataType);i&&Bg(i,r,"normal")},t.prototype.remove=function(t,e){this.group.removeAll()},t.prototype.dispose=function(t,e){},t.prototype.updateView=function(t,e,n,r){this.render(t,e,n,r)},t.prototype.updateLayout=function(t,e,n,r){this.render(t,e,n,r)},t.prototype.updateVisual=function(t,e,n,r){this.render(t,e,n,r)},t.prototype.eachRendered=function(t){Ng(this.group,t)},t.markUpdateMethod=function(t,e){Rg(t).updateMethod=e},t.protoInitialize=function(){var e=t.prototype;e.type="chart"}(),t}();function Fg(t,e,n){t&&jl(t)&&("emphasis"===e?ul:cl)(t,n)}function Bg(t,e,n){var r=Ii(t,e),i=e&&null!=e.highlightKey?El(e.highlightKey):null;null!=r?Object(s["k"])(ui(r),(function(e){Fg(t.getItemGraphicEl(e),n,i)})):t.eachItemGraphicEl((function(t){Fg(t,n,i)}))}function Wg(t){return Hg(t.model)}function Vg(t){var e=t.model,n=t.ecModel,r=t.api,i=t.payload,o=e.pipelineContext.progressiveRender,a=t.view,s=i&&Rg(i).updateMethod,l=o?"incrementalPrepareRender":s&&a[s]?s:"render";return"render"!==l&&a[l](e,n,r,i),Gg[l]}Xi(zg,["dispose"]),no(zg);var Gg={incrementalPrepareRender:{progress:function(t,e){e.view.incrementalRender(t,e.model,e.ecModel,e.api,e.payload)}},render:{forceFirstProgress:!0,progress:function(t,e){e.view.render(e.model,e.ecModel,e.api,e.payload)}}},$g=zg,Ug="\0__throttleOriginMethod",qg="\0__throttleRate",Xg="\0__throttleType";function Zg(t,e,n){var r,i,o,a,s,l=0,u=0,c=null;function d(){u=(new Date).getTime(),c=null,t.apply(o,a||[])}e=e||0;var h=function(){for(var t=[],h=0;h<arguments.length;h++)t[h]=arguments[h];r=(new Date).getTime(),o=this,a=t;var f=s||e,p=s||n;s=null,i=r-(p?l:u)-f,clearTimeout(c),p?c=setTimeout(d,f):i>=0?d():c=setTimeout(d,-i),l=r};return h.clear=function(){c&&(clearTimeout(c),c=null)},h.debounceNextCall=function(t){s=t},h}function Jg(t,e,n,r){var i=t[e];if(i){var o=i[Ug]||i,a=i[Xg],s=i[qg];if(s!==n||a!==r){if(null==n||!r)return t[e]=o;i=t[e]=Zg(o,n,"debounce"===r),i[Ug]=o,i[Xg]=r,i[qg]=n}return i}}function Kg(t,e){var n=t[e];n&&n[Ug]&&(n.clear&&n.clear(),t[e]=n[Ug])}var Qg=Ci(),ty={itemStyle:ro(wu,!0),lineStyle:ro(_u,!0)},ey={lineStyle:"stroke",itemStyle:"fill"};function ny(t,e){var n=t.visualStyleMapper||ty[e];return n||ty.itemStyle}function ry(t,e){var n=t.visualDrawType||ey[e];return n||"fill"}var iy={createOnAllSeries:!0,performRawSeries:!0,reset:function(t,e){var n=t.getData(),r=t.visualStyleAccessPath||"itemStyle",i=t.getModel(r),o=ny(t,r),a=o(i),l=i.getShallow("decal");l&&(n.setVisual("decal",l),l.dirty=!0);var u=ry(t,r),c=a[u],d=Object(s["w"])(c)?c:null,h="auto"===a.fill||"auto"===a.stroke;if(!a[u]||d||h){var f=t.getColorFromPalette(t.name,null,e.getSeriesCount());a[u]||(a[u]=f,n.setVisual("colorFromPalette",!0)),a.fill="auto"===a.fill||Object(s["w"])(a.fill)?f:a.fill,a.stroke="auto"===a.stroke||Object(s["w"])(a.stroke)?f:a.stroke}if(n.setVisual("style",a),n.setVisual("drawType",u),!e.isSeriesFiltered(t)&&d)return n.setVisual("colorFromPalette",!1),{dataEach:function(e,n){var r=t.getDataParams(n),i=Object(s["m"])({},a);i[u]=d(r),e.setItemVisual(n,"style",i)}}}},oy=new Lu,ay={createOnAllSeries:!0,performRawSeries:!0,reset:function(t,e){if(!t.ignoreStyleOnData&&!e.isSeriesFiltered(t)){var n=t.getData(),r=t.visualStyleAccessPath||"itemStyle",i=ny(t,r),o=n.getVisual("drawType");return{dataEach:n.hasItemOption?function(t,e){var n=t.getRawDataItem(e);if(n&&n[r]){oy.option=n[r];var a=i(oy),l=t.ensureUniqueItemVisual(e,"style");Object(s["m"])(l,a),oy.option.decal&&(t.setItemVisual(e,"decal",oy.option.decal),oy.option.decal.dirty=!0),o in a&&t.setItemVisual(e,"colorFromPalette",!1)}}:null}}}},sy={performRawSeries:!0,overallReset:function(t){var e=Object(s["f"])();t.eachSeries((function(t){var n=t.getColorBy();if(!t.isColorBySeries()){var r=t.type+"-"+n,i=e.get(r);i||(i={},e.set(r,i)),Qg(t).scope=i}})),t.eachSeries((function(e){if(!e.isColorBySeries()&&!t.isSeriesFiltered(e)){var n=e.getRawData(),r={},i=e.getData(),o=Qg(e).scope,a=e.visualStyleAccessPath||"itemStyle",s=ry(e,a);i.each((function(t){var e=i.getRawIndex(t);r[e]=t})),n.each((function(t){var a=r[t],l=i.getItemVisual(a,"colorFromPalette");if(l){var u=i.ensureUniqueItemVisual(a,"style"),c=n.getName(t)||t+"",d=n.count();u[s]=e.getColorFromPalette(c,o,d)}}))}}))}},ly=Math.PI;function uy(t,e){e=e||{},s["i"](e,{text:"loading",textColor:"#000",fontSize:12,fontWeight:"normal",fontStyle:"normal",fontFamily:"sans-serif",maskColor:"rgba(255, 255, 255, 0.8)",showSpinner:!0,color:"#5470c6",spinnerRadius:10,lineWidth:5,zlevel:0});var n=new wr,r=new ss({style:{fill:e.maskColor},zlevel:e.zlevel,z:1e4});n.add(r);var i,o=new ks({style:{text:e.text,fill:e.textColor,fontSize:e.fontSize,fontWeight:e.fontWeight,fontStyle:e.fontStyle,fontFamily:e.fontFamily},zlevel:e.zlevel,z:10001}),a=new ss({style:{fill:"none"},textContent:o,textConfig:{position:"right",distance:10},zlevel:e.zlevel,z:10001});return n.add(a),e.showSpinner&&(i=new Bm({shape:{startAngle:-ly/2,endAngle:-ly/2+.1,r:e.spinnerRadius},style:{stroke:e.color,lineCap:"round",lineWidth:e.lineWidth},zlevel:e.zlevel,z:10001}),i.animateShape(!0).when(1e3,{endAngle:3*ly/2}).start("circularInOut"),i.animateShape(!0).when(1e3,{startAngle:3*ly/2}).delay(300).start("circularInOut"),n.add(i)),n.resize=function(){var n=o.getBoundingRect().width,s=e.showSpinner?e.spinnerRadius:0,l=(t.getWidth()-2*s-(e.showSpinner&&n?10:0)-n)/2-(e.showSpinner&&n?0:5+n/2)+(e.showSpinner?0:n/2)+(n?0:s),u=t.getHeight()/2;e.showSpinner&&i.setShape({cx:l,cy:u}),a.setShape({x:l-s,y:u-s,width:2*s,height:2*s}),r.setShape({x:0,y:0,width:t.getWidth(),height:t.getHeight()})},n.resize(),n}var cy=function(){function t(t,e,n,r){this._stageTaskMap=Object(s["f"])(),this.ecInstance=t,this.api=e,n=this._dataProcessorHandlers=n.slice(),r=this._visualHandlers=r.slice(),this._allHandlers=n.concat(r)}return t.prototype.restoreData=function(t,e){t.restoreData(e),this._stageTaskMap.each((function(t){var e=t.overallTask;e&&e.dirty()}))},t.prototype.getPerformArgs=function(t,e){if(t.__pipeline){var n=this._pipelineMap.get(t.__pipeline.id),r=n.context,i=!e&&n.progressiveEnabled&&(!r||r.progressiveRender)&&t.__idxInPipeline>n.blockIndex,o=i?n.step:null,a=r&&r.modDataCount,s=null!=a?Math.ceil(a/o):null;return{step:o,modBy:s,modDataCount:a}}},t.prototype.getPipeline=function(t){return this._pipelineMap.get(t)},t.prototype.updateStreamModes=function(t,e){var n=this._pipelineMap.get(t.uid),r=t.getData(),i=r.count(),o=n.progressiveEnabled&&e.incrementalPrepareRender&&i>=n.threshold,a=t.get("large")&&i>=t.get("largeThreshold"),s="mod"===t.get("progressiveChunkMode")?i:null;t.pipelineContext=n.context={progressiveRender:o,modDataCount:s,large:a}},t.prototype.restorePipelines=function(t){var e=this,n=e._pipelineMap=Object(s["f"])();t.eachSeries((function(t){var r=t.getProgressive(),i=t.uid;n.set(i,{id:i,head:null,tail:null,threshold:t.getProgressiveThreshold(),progressiveEnabled:r&&!(t.preventIncremental&&t.preventIncremental()),blockIndex:-1,step:Math.round(r||700),count:0}),e._pipe(t,t.dataTask)}))},t.prototype.prepareStageTasks=function(){var t=this._stageTaskMap,e=this.api.getModel(),n=this.api;Object(s["k"])(this._allHandlers,(function(r){var i=t.get(r.uid)||t.set(r.uid,{}),o="";Object(s["b"])(!(r.reset&&r.overallReset),o),r.reset&&this._createSeriesStageTask(r,i,e,n),r.overallReset&&this._createOverallStageTask(r,i,e,n)}),this)},t.prototype.prepareView=function(t,e,n,r){var i=t.renderTask,o=i.context;o.model=e,o.ecModel=n,o.api=r,i.__block=!t.incrementalPrepareRender,this._pipe(e,i)},t.prototype.performDataProcessorTasks=function(t,e){this._performStageTasks(this._dataProcessorHandlers,t,e,{block:!0})},t.prototype.performVisualTasks=function(t,e,n){this._performStageTasks(this._visualHandlers,t,e,n)},t.prototype._performStageTasks=function(t,e,n,r){r=r||{};var i=!1,o=this;function a(t,e){return t.setDirty&&(!t.dirtyMap||t.dirtyMap.get(e.__pipeline.id))}Object(s["k"])(t,(function(t,s){if(!r.visualType||r.visualType===t.visualType){var l=o._stageTaskMap.get(t.uid),u=l.seriesTaskMap,c=l.overallTask;if(c){var d,h=c.agentStubMap;h.each((function(t){a(r,t)&&(t.dirty(),d=!0)})),d&&c.dirty(),o.updatePayload(c,n);var f=o.getPerformArgs(c,r.block);h.each((function(t){t.perform(f)})),c.perform(f)&&(i=!0)}else u&&u.each((function(s,l){a(r,s)&&s.dirty();var u=o.getPerformArgs(s,r.block);u.skip=!t.performRawSeries&&e.isSeriesFiltered(s.context.model),o.updatePayload(s,n),s.perform(u)&&(i=!0)}))}})),this.unfinished=i||this.unfinished},t.prototype.performSeriesTasks=function(t){var e;t.eachSeries((function(t){e=t.dataTask.perform()||e})),this.unfinished=e||this.unfinished},t.prototype.plan=function(){this._pipelineMap.each((function(t){var e=t.tail;do{if(e.__block){t.blockIndex=e.__idxInPipeline;break}e=e.getUpstream()}while(e)}))},t.prototype.updatePayload=function(t,e){"remain"!==e&&(t.context.payload=e)},t.prototype._createSeriesStageTask=function(t,e,n,r){var i=this,o=e.seriesTaskMap,a=e.seriesTaskMap=Object(s["f"])(),l=t.seriesType,u=t.getTargetSeries;function c(e){var s=e.uid,l=a.set(s,o&&o.get(s)||rf({plan:my,reset:gy,count:_y}));l.context={model:e,ecModel:n,api:r,useClearVisual:t.isVisual&&!t.isLayout,plan:t.plan,reset:t.reset,scheduler:i},i._pipe(e,l)}t.createOnAllSeries?n.eachRawSeries(c):l?n.eachRawSeriesByType(l,c):u&&u(n,r).each(c)},t.prototype._createOverallStageTask=function(t,e,n,r){var i=this,o=e.overallTask=e.overallTask||rf({reset:dy});o.context={ecModel:n,api:r,overallReset:t.overallReset,scheduler:i};var a=o.agentStubMap,l=o.agentStubMap=Object(s["f"])(),u=t.seriesType,c=t.getTargetSeries,d=!0,h=!1,f="";function p(t){var e=t.uid,n=l.set(e,a&&a.get(e)||(h=!0,rf({reset:hy,onDirty:py})));n.context={model:t,overallProgress:d},n.agent=o,n.__block=d,i._pipe(t,n)}Object(s["b"])(!t.createOnAllSeries,f),u?n.eachRawSeriesByType(u,p):c?c(n,r).each(p):(d=!1,Object(s["k"])(n.getSeries(),p)),h&&o.dirty()},t.prototype._pipe=function(t,e){var n=t.uid,r=this._pipelineMap.get(n);!r.head&&(r.head=e),r.tail&&r.tail.pipe(e),r.tail=e,e.__idxInPipeline=r.count++,e.__pipeline=r},t.wrapStageHandler=function(t,e){return Object(s["w"])(t)&&(t={overallReset:t,seriesType:by(t)}),t.uid=Du("stageHandler"),e&&(t.visualType=e),t},t}();function dy(t){t.overallReset(t.ecModel,t.api,t.payload)}function hy(t){return t.overallProgress&&fy}function fy(){this.agent.dirty(),this.getDownstream().dirty()}function py(){this.agent&&this.agent.dirty()}function my(t){return t.plan?t.plan(t.model,t.ecModel,t.api,t.payload):null}function gy(t){t.useClearVisual&&t.data.clearAllVisual();var e=t.resetDefines=ui(t.reset(t.model,t.ecModel,t.api,t.payload));return e.length>1?Object(s["H"])(e,(function(t,e){return vy(e)})):yy}var yy=vy(0);function vy(t){return function(e,n){var r=n.data,i=n.resetDefines[t];if(i&&i.dataEach)for(var o=e.start;o<e.end;o++)i.dataEach(r,o);else i&&i.progress&&i.progress(e,r)}}function _y(t){return t.data.count()}function by(t){xy=null;try{t(wy,My)}catch(e){}return xy}var xy,wy={},My={};function Sy(t,e){for(var n in e.prototype)t[n]=s["L"]}Sy(wy,zd),Sy(My,Wd),wy.eachSeriesByType=wy.eachRawSeriesByType=function(t){xy=t},wy.eachComponent=function(t){"series"===t.mainType&&t.subType&&(xy=t.subType)};var ky=cy,Ly=["#37A2DA","#32C5E9","#67E0E3","#9FE6B8","#FFDB5C","#ff9f7f","#fb7293","#E062AE","#E690D1","#e7bcf3","#9d96f5","#8378EA","#96BFFF"],Ty={color:Ly,colorLayer:[["#37A2DA","#ffd85c","#fd7b5f"],["#37A2DA","#67E0E3","#FFDB5C","#ff9f7f","#E062AE","#9d96f5"],["#37A2DA","#32C5E9","#9FE6B8","#FFDB5C","#ff9f7f","#fb7293","#e7bcf3","#8378EA","#96BFFF"],Ly]},Dy="#B9B8CE",Oy="#100C2A",Iy=function(){return{axisLine:{lineStyle:{color:Dy}},splitLine:{lineStyle:{color:"#484753"}},splitArea:{areaStyle:{color:["rgba(255,255,255,0.02)","rgba(255,255,255,0.05)"]}},minorSplitLine:{lineStyle:{color:"#20203B"}}}},Cy=["#4992ff","#7cffb2","#fddd60","#ff6e76","#58d9f9","#05c091","#ff8a45","#8d48e3","#dd79ff"],Ay={darkMode:!0,color:Cy,backgroundColor:Oy,axisPointer:{lineStyle:{color:"#817f91"},crossStyle:{color:"#817f91"},label:{color:"#fff"}},legend:{textStyle:{color:Dy}},textStyle:{color:Dy},title:{textStyle:{color:"#EEF1FA"},subtextStyle:{color:"#B9B8CE"}},toolbox:{iconStyle:{borderColor:Dy}},dataZoom:{borderColor:"#71708A",textStyle:{color:Dy},brushStyle:{color:"rgba(135,163,206,0.3)"},handleStyle:{color:"#353450",borderColor:"#C5CBE3"},moveHandleStyle:{color:"#B0B6C3",opacity:.3},fillerColor:"rgba(135,163,206,0.2)",emphasis:{handleStyle:{borderColor:"#91B7F2",color:"#4D587D"},moveHandleStyle:{color:"#636D9A",opacity:.7}},dataBackground:{lineStyle:{color:"#71708A",width:1},areaStyle:{color:"#71708A"}},selectedDataBackground:{lineStyle:{color:"#87A3CE"},areaStyle:{color:"#87A3CE"}}},visualMap:{textStyle:{color:Dy}},timeline:{lineStyle:{color:Dy},label:{color:Dy},controlStyle:{color:Dy,borderColor:Dy}},calendar:{itemStyle:{color:Oy},dayLabel:{color:Dy},monthLabel:{color:Dy},yearLabel:{color:Dy}},timeAxis:Iy(),logAxis:Iy(),valueAxis:Iy(),categoryAxis:Iy(),line:{symbol:"circle"},graph:{color:Cy},gauge:{title:{color:Dy},axisLine:{lineStyle:{color:[[1,"rgba(207,212,219,0.2)"]]}},axisLabel:{color:Dy},detail:{color:"#EEF1FA"}},candlestick:{itemStyle:{color:"#f64e56",color0:"#54ea92",borderColor:"#f64e56",borderColor0:"#54ea92"}}};Ay.categoryAxis.splitLine.show=!1;var Yy=Ay,jy=function(){function t(){}return t.prototype.normalizeQuery=function(t){var e={},n={},r={};if(s["C"](t)){var i=$i(t);e.mainType=i.main||null,e.subType=i.sub||null}else{var o=["Index","Name","Id"],a={name:1,dataIndex:1,dataType:1};s["k"](t,(function(t,i){for(var s=!1,l=0;l<o.length;l++){var u=o[l],c=i.lastIndexOf(u);if(c>0&&c===i.length-u.length){var d=i.slice(0,c);"data"!==d&&(e.mainType=d,e[u.toLowerCase()]=t,s=!0)}}a.hasOwnProperty(i)&&(n[i]=t,s=!0),s||(r[i]=t)}))}return{cptQuery:e,dataQuery:n,otherQuery:r}},t.prototype.filter=function(t,e){var n=this.eventInfo;if(!n)return!0;var r=n.targetEl,i=n.packedEvent,o=n.model,a=n.view;if(!o||!a)return!0;var s=e.cptQuery,l=e.dataQuery;return u(s,o,"mainType")&&u(s,o,"subType")&&u(s,o,"index","componentIndex")&&u(s,o,"name")&&u(s,o,"id")&&u(l,i,"name")&&u(l,i,"dataIndex")&&u(l,i,"dataType")&&(!a.filterForExposedEvent||a.filterForExposedEvent(t,e.otherQuery,r,i));function u(t,e,n,r){return null==t[n]||e[r||n]===t[n]}},t.prototype.afterTrigger=function(){this.eventInfo=null},t}(),Py=["symbol","symbolSize","symbolRotate","symbolOffset"],Ey=Py.concat(["symbolKeepAspect"]),Ny={createOnAllSeries:!0,performRawSeries:!0,reset:function(t,e){var n=t.getData();if(t.legendIcon&&n.setVisual("legendIcon",t.legendIcon),t.hasSymbolVisual){for(var r={},i={},o=!1,a=0;a<Py.length;a++){var l=Py[a],u=t.get(l);Object(s["w"])(u)?(o=!0,i[l]=u):r[l]=u}if(r.symbol=r.symbol||t.defaultSymbol,n.setVisual(Object(s["m"])({legendIcon:t.legendIcon||r.symbol,symbolKeepAspect:t.get("symbolKeepAspect")},r)),!e.isSeriesFiltered(t)){var c=Object(s["F"])(i);return{dataEach:o?d:null}}}function d(e,n){for(var r=t.getRawValue(n),o=t.getDataParams(n),a=0;a<c.length;a++){var s=c[a];e.setItemVisual(n,s,i[s](r,o))}}}},Ry={createOnAllSeries:!0,performRawSeries:!0,reset:function(t,e){if(t.hasSymbolVisual&&!e.isSeriesFiltered(t)){var n=t.getData();return{dataEach:n.hasItemOption?r:null}}function r(t,e){for(var n=t.getItemModel(e),r=0;r<Ey.length;r++){var i=Ey[r],o=n.getShallow(i,!0);null!=o&&t.setItemVisual(e,i,o)}}}};function Hy(t,e,n){switch(n){case"color":var r=t.getItemVisual(e,"style");return r[t.getVisual("drawType")];case"opacity":return t.getItemVisual(e,"style").opacity;case"symbol":case"symbolSize":case"liftZ":return t.getItemVisual(e,n);default:0}}function zy(t,e){switch(e){case"color":var n=t.getVisual("style");return n[t.getVisual("drawType")];case"opacity":return t.getVisual("style").opacity;case"symbol":case"symbolSize":case"liftZ":return t.getVisual(e);default:0}}function Fy(t,e,n,r){switch(n){case"color":var i=t.ensureUniqueItemVisual(e,"style");i[t.getVisual("drawType")]=r,t.setItemVisual(e,"colorFromPalette",!1);break;case"opacity":t.ensureUniqueItemVisual(e,"style").opacity=r;break;case"symbol":case"symbolSize":case"liftZ":t.setItemVisual(e,n,r);break;default:0}}function By(t,e){function n(e,n){var r=[];return e.eachComponent({mainType:"series",subType:t,query:n},(function(t){r.push(t.seriesIndex)})),r}Object(s["k"])([[t+"ToggleSelect","toggleSelect"],[t+"Select","select"],[t+"UnSelect","unselect"]],(function(t){e(t[0],(function(e,r,i){e=Object(s["m"])({},e),i.dispatchAction(Object(s["m"])(e,{type:t[1],seriesIndex:n(r,e)}))}))}))}function Wy(t,e,n,r,i){var o=t+e;n.isSilent(o)||r.eachComponent({mainType:"series",subType:"pie"},(function(t){for(var e=t.seriesIndex,r=t.option.selectedMap,a=i.selected,l=0;l<a.length;l++)if(a[l].seriesIndex===e){var u=t.getData(),c=Ii(u,i.fromActionPayload);n.trigger(o,{type:o,seriesId:t.id,name:Object(s["t"])(c)?u.getName(c[0]):u.getName(c),selected:Object(s["C"])(r)?r:Object(s["m"])({},r)})}}))}function Vy(t,e,n){t.on("selectchanged",(function(t){var r=n.getModel();t.isFromClick?(Wy("map","selectchanged",e,r,t),Wy("pie","selectchanged",e,r,t)):"select"===t.fromAction?(Wy("map","selected",e,r,t),Wy("pie","selected",e,r,t)):"unselect"===t.fromAction&&(Wy("map","unselected",e,r,t),Wy("pie","unselected",e,r,t))}))}function Gy(t,e,n){var r;while(t){if(e(t)&&(r=t,n))break;t=t.__hostTarget||t.parent}return r}var $y=Math.round(9*Math.random()),Uy="function"===typeof Object.defineProperty,qy=function(){function t(){this._id="__ec_inner_"+$y++}return t.prototype.get=function(t){return this._guard(t)[this._id]},t.prototype.set=function(t,e){var n=this._guard(t);return Uy?Object.defineProperty(n,this._id,{value:e,enumerable:!1,configurable:!0}):n[this._id]=e,this},t.prototype["delete"]=function(t){return!!this.has(t)&&(delete this._guard(t)[this._id],!0)},t.prototype.has=function(t){return!!this._guard(t)[this._id]},t.prototype._guard=function(t){if(t!==Object(t))throw TypeError("Value of WeakMap is not a non-null object.");return t},t}(),Xy=qy,Zy=Va.extend({type:"triangle",shape:{cx:0,cy:0,width:0,height:0},buildPath:function(t,e){var n=e.cx,r=e.cy,i=e.width/2,o=e.height/2;t.moveTo(n,r-o),t.lineTo(n+i,r+o),t.lineTo(n-i,r+o),t.closePath()}}),Jy=Va.extend({type:"diamond",shape:{cx:0,cy:0,width:0,height:0},buildPath:function(t,e){var n=e.cx,r=e.cy,i=e.width/2,o=e.height/2;t.moveTo(n,r-o),t.lineTo(n+i,r),t.lineTo(n,r+o),t.lineTo(n-i,r),t.closePath()}}),Ky=Va.extend({type:"pin",shape:{x:0,y:0,width:0,height:0},buildPath:function(t,e){var n=e.x,r=e.y,i=e.width/5*3,o=Math.max(i,e.height),a=i/2,s=a*a/(o-a),l=r-o+a+s,u=Math.asin(s/a),c=Math.cos(u)*a,d=Math.sin(u),h=Math.cos(u),f=.6*a,p=.7*a;t.moveTo(n-c,l+s),t.arc(n,l,a,Math.PI-u,2*Math.PI+u),t.bezierCurveTo(n+c-d*f,l+s+h*f,n,r-p,n,r),t.bezierCurveTo(n,r-p,n-c+d*f,l+s+h*f,n-c,l+s),t.closePath()}}),Qy=Va.extend({type:"arrow",shape:{x:0,y:0,width:0,height:0},buildPath:function(t,e){var n=e.height,r=e.width,i=e.x,o=e.y,a=r/3*2;t.moveTo(i,o),t.lineTo(i+a,o+n),t.lineTo(i,o+n/4*3),t.lineTo(i-a,o+n),t.lineTo(i,o),t.closePath()}}),tv={line:jm,rect:ss,roundRect:ss,square:ss,circle:Kp,diamond:Jy,pin:Ky,arrow:Qy,triangle:Zy},ev={line:function(t,e,n,r,i){i.x1=t,i.y1=e+r/2,i.x2=t+n,i.y2=e+r/2},rect:function(t,e,n,r,i){i.x=t,i.y=e,i.width=n,i.height=r},roundRect:function(t,e,n,r,i){i.x=t,i.y=e,i.width=n,i.height=r,i.r=Math.min(n,r)/4},square:function(t,e,n,r,i){var o=Math.min(n,r);i.x=t,i.y=e,i.width=o,i.height=o},circle:function(t,e,n,r,i){i.cx=t+n/2,i.cy=e+r/2,i.r=Math.min(n,r)/2},diamond:function(t,e,n,r,i){i.cx=t+n/2,i.cy=e+r/2,i.width=n,i.height=r},pin:function(t,e,n,r,i){i.x=t+n/2,i.y=e+r/2,i.width=n,i.height=r},arrow:function(t,e,n,r,i){i.x=t+n/2,i.y=e+r/2,i.width=n,i.height=r},triangle:function(t,e,n,r,i){i.cx=t+n/2,i.cy=e+r/2,i.width=n,i.height=r}},nv={};Object(s["k"])(tv,(function(t,e){nv[e]=new t}));var rv=Va.extend({type:"symbol",shape:{symbolType:"",x:0,y:0,width:0,height:0},calculateTextPosition:function(t,e,n){var r=ar(t,e,n),i=this.shape;return i&&"pin"===i.symbolType&&"inside"===e.position&&(r.y=n.y+.4*n.height),r},buildPath:function(t,e,n){var r=e.symbolType;if("none"!==r){var i=nv[r];i||(r="rect",i=nv[r]),ev[r](e.x,e.y,e.width,e.height,i.shape),i.buildPath(t,i.shape,n)}}});function iv(t,e){if("image"!==this.type){var n=this.style;this.__isEmptyBrush?(n.stroke=t,n.fill=e||"#fff",n.lineWidth=2):"line"===this.shape.symbolType?n.stroke=t:n.fill=t,this.markRedraw()}}function ov(t,e,n,r,i,o,a){var s,l=0===t.indexOf("empty");return l&&(t=t.substr(5,1).toLowerCase()+t.substr(6)),s=0===t.indexOf("image://")?mg(t.slice(8),new Tt(e,n,r,i),a?"center":"cover"):0===t.indexOf("path://")?pg(t.slice(7),{},new Tt(e,n,r,i),a?"center":"cover"):new rv({shape:{symbolType:t,x:e,y:n,width:r,height:i}}),s.__isEmptyBrush=l,s.setColor=iv,o&&s.setColor(o),s}function av(t){return Object(s["t"])(t)||(t=[+t,+t]),[t[0]||0,t[1]||0]}function sv(t,e){if(null!=t)return Object(s["t"])(t)||(t=[t,t]),[Nr(t[0],e[0])||0,Nr(Object(s["P"])(t[1],t[0]),e[1])||0]}function lv(t){return isFinite(t)}function uv(t,e,n){var r=null==e.x?0:e.x,i=null==e.x2?1:e.x2,o=null==e.y?0:e.y,a=null==e.y2?0:e.y2;e.global||(r=r*n.width+n.x,i=i*n.width+n.x,o=o*n.height+n.y,a=a*n.height+n.y),r=lv(r)?r:0,i=lv(i)?i:1,o=lv(o)?o:0,a=lv(a)?a:0;var s=t.createLinearGradient(r,o,i,a);return s}function cv(t,e,n){var r=n.width,i=n.height,o=Math.min(r,i),a=null==e.x?.5:e.x,s=null==e.y?.5:e.y,l=null==e.r?.5:e.r;e.global||(a=a*r+n.x,s=s*i+n.y,l*=o),a=lv(a)?a:.5,s=lv(s)?s:.5,l=l>=0&&lv(l)?l:.5;var u=t.createRadialGradient(a,s,0,a,s,l);return u}function dv(t,e,n){for(var r="radial"===e.type?cv(t,e,n):uv(t,e,n),i=e.colorStops,o=0;o<i.length;o++)r.addColorStop(i[o].offset,i[o].color);return r}function hv(t,e){if(t===e||!t&&!e)return!1;if(!t||!e||t.length!==e.length)return!0;for(var n=0;n<t.length;n++)if(t[n]!==e[n])return!0;return!1}function fv(t){return parseInt(t,10)}function pv(t,e,n){var r=["width","height"][e],i=["clientWidth","clientHeight"][e],o=["paddingLeft","paddingTop"][e],a=["paddingRight","paddingBottom"][e];if(null!=n[r]&&"auto"!==n[r])return parseFloat(n[r]);var s=document.defaultView.getComputedStyle(t);return(t[i]||fv(s[r])||fv(t.style[r]))-(fv(s[o])||0)-(fv(s[a])||0)|0}function mv(t,e){return t&&"solid"!==t&&e>0?"dashed"===t?[4*e,2*e]:"dotted"===t?[e]:Object(s["z"])(t)?[t]:Object(s["t"])(t)?t:null:null}function gv(t){var e=t.style,n=e.lineDash&&e.lineWidth>0&&mv(e.lineDash,e.lineWidth),r=e.lineDashOffset;if(n){var i=e.strokeNoScale&&t.getLineScale?t.getLineScale():1;i&&1!==i&&(n=Object(s["H"])(n,(function(t){return t/i})),r/=i)}return[n,r]}var yv=new va(!0);function vv(t){var e=t.stroke;return!(null==e||"none"===e||!(t.lineWidth>0))}function _v(t){return"string"===typeof t&&"none"!==t}function bv(t){var e=t.fill;return null!=e&&"none"!==e}function xv(t,e){if(null!=e.fillOpacity&&1!==e.fillOpacity){var n=t.globalAlpha;t.globalAlpha=e.fillOpacity*e.opacity,t.fill(),t.globalAlpha=n}else t.fill()}function wv(t,e){if(null!=e.strokeOpacity&&1!==e.strokeOpacity){var n=t.globalAlpha;t.globalAlpha=e.strokeOpacity*e.opacity,t.stroke(),t.globalAlpha=n}else t.stroke()}function Mv(t,e,n){var r=uo(e.image,e.__image,n);if(ho(r)){var i=t.createPattern(r,e.repeat||"repeat");if("function"===typeof DOMMatrix&&i&&i.setTransform){var o=new DOMMatrix;o.translateSelf(e.x||0,e.y||0),o.rotateSelf(0,0,(e.rotation||0)*s["a"]),o.scaleSelf(e.scaleX||1,e.scaleY||1),i.setTransform(o)}return i}}function Sv(t,e,n,r){var i,o=vv(n),a=bv(n),s=n.strokePercent,l=s<1,u=!e.path;e.silent&&!l||!u||e.createPathProxy();var c=e.path||yv,d=e.__dirty;if(!r){var h=n.fill,f=n.stroke,p=a&&!!h.colorStops,m=o&&!!f.colorStops,g=a&&!!h.image,y=o&&!!f.image,v=void 0,_=void 0,b=void 0,x=void 0,w=void 0;(p||m)&&(w=e.getBoundingRect()),p&&(v=d?dv(t,h,w):e.__canvasFillGradient,e.__canvasFillGradient=v),m&&(_=d?dv(t,f,w):e.__canvasStrokeGradient,e.__canvasStrokeGradient=_),g&&(b=d||!e.__canvasFillPattern?Mv(t,h,e):e.__canvasFillPattern,e.__canvasFillPattern=b),y&&(x=d||!e.__canvasStrokePattern?Mv(t,f,e):e.__canvasStrokePattern,e.__canvasStrokePattern=b),p?t.fillStyle=v:g&&(b?t.fillStyle=b:a=!1),m?t.strokeStyle=_:y&&(x?t.strokeStyle=x:o=!1)}var M,S,k=e.getGlobalScale();c.setScale(k[0],k[1],e.segmentIgnoreThreshold),t.setLineDash&&n.lineDash&&(i=gv(e),M=i[0],S=i[1]);var L=!0;(u||d&Kt)&&(c.setDPR(t.dpr),l?c.setContext(null):(c.setContext(t),L=!1),c.reset(),e.buildPath(c,e.shape,r),c.toStatic(),e.pathUpdated()),L&&c.rebuildPath(t,l?s:1),M&&(t.setLineDash(M),t.lineDashOffset=S),r||(n.strokeFirst?(o&&wv(t,n),a&&xv(t,n)):(a&&xv(t,n),o&&wv(t,n))),M&&t.setLineDash([])}function kv(t,e,n){var r=e.__image=uo(n.image,e.__image,e,e.onload);if(r&&ho(r)){var i=n.x||0,o=n.y||0,a=e.getWidth(),s=e.getHeight(),l=r.width/r.height;if(null==a&&null!=s?a=s*l:null==s&&null!=a?s=a/l:null==a&&null==s&&(a=r.width,s=r.height),n.sWidth&&n.sHeight){var u=n.sx||0,c=n.sy||0;t.drawImage(r,u,c,n.sWidth,n.sHeight,i,o,a,s)}else if(n.sx&&n.sy){u=n.sx,c=n.sy;var d=a-u,h=s-c;t.drawImage(r,u,c,d,h,i,o,a,s)}else t.drawImage(r,i,o,a,s)}}function Lv(t,e,n){var r,i=n.text;if(null!=i&&(i+=""),i){t.font=n.font||Jn["a"],t.textAlign=n.textAlign,t.textBaseline=n.textBaseline;var o=void 0,a=void 0;t.setLineDash&&n.lineDash&&(r=gv(e),o=r[0],a=r[1]),o&&(t.setLineDash(o),t.lineDashOffset=a),n.strokeFirst?(vv(n)&&t.strokeText(i,n.x,n.y),bv(n)&&t.fillText(i,n.x,n.y)):(bv(n)&&t.fillText(i,n.x,n.y),vv(n)&&t.strokeText(i,n.x,n.y)),o&&t.setLineDash([])}}var Tv=["shadowBlur","shadowOffsetX","shadowOffsetY"],Dv=[["lineCap","butt"],["lineJoin","miter"],["miterLimit",10]];function Ov(t,e,n,r,i){var o=!1;if(!r&&(n=n||{},e===n))return!1;if(r||e.opacity!==n.opacity){zv(t,i),o=!0;var a=Math.max(Math.min(e.opacity,1),0);t.globalAlpha=isNaN(a)?Oo.opacity:a}(r||e.blend!==n.blend)&&(o||(zv(t,i),o=!0),t.globalCompositeOperation=e.blend||Oo.blend);for(var s=0;s<Tv.length;s++){var l=Tv[s];(r||e[l]!==n[l])&&(o||(zv(t,i),o=!0),t[l]=t.dpr*(e[l]||0))}return(r||e.shadowColor!==n.shadowColor)&&(o||(zv(t,i),o=!0),t.shadowColor=e.shadowColor||Oo.shadowColor),o}function Iv(t,e,n,r,i){var o=Fv(e,i.inHover),a=r?null:n&&Fv(n,i.inHover)||{};if(o===a)return!1;var s=Ov(t,o,a,r,i);if((r||o.fill!==a.fill)&&(s||(zv(t,i),s=!0),_v(o.fill)&&(t.fillStyle=o.fill)),(r||o.stroke!==a.stroke)&&(s||(zv(t,i),s=!0),_v(o.stroke)&&(t.strokeStyle=o.stroke)),(r||o.opacity!==a.opacity)&&(s||(zv(t,i),s=!0),t.globalAlpha=null==o.opacity?1:o.opacity),e.hasStroke()){var l=o.lineWidth,u=l/(o.strokeNoScale&&e.getLineScale?e.getLineScale():1);t.lineWidth!==u&&(s||(zv(t,i),s=!0),t.lineWidth=u)}for(var c=0;c<Dv.length;c++){var d=Dv[c],h=d[0];(r||o[h]!==a[h])&&(s||(zv(t,i),s=!0),t[h]=o[h]||d[1])}return s}function Cv(t,e,n,r,i){return Ov(t,Fv(e,i.inHover),n&&Fv(n,i.inHover),r,i)}function Av(t,e){var n=e.transform,r=t.dpr||1;n?t.setTransform(r*n[0],r*n[1],r*n[2],r*n[3],r*n[4],r*n[5]):t.setTransform(r,0,0,r,0,0)}function Yv(t,e,n){for(var r=!1,i=0;i<t.length;i++){var o=t[i];r=r||o.isZeroArea(),Av(e,o),e.beginPath(),o.buildPath(e,o.shape),e.clip()}n.allClipped=r}function jv(t,e){return t&&e?t[0]!==e[0]||t[1]!==e[1]||t[2]!==e[2]||t[3]!==e[3]||t[4]!==e[4]||t[5]!==e[5]:!(!t&&!e)}var Pv=1,Ev=2,Nv=3,Rv=4;function Hv(t){var e=bv(t),n=vv(t);return!(t.lineDash||!(+e^+n)||e&&"string"!==typeof t.fill||n&&"string"!==typeof t.stroke||t.strokePercent<1||t.strokeOpacity<1||t.fillOpacity<1)}function zv(t,e){e.batchFill&&t.fill(),e.batchStroke&&t.stroke(),e.batchFill="",e.batchStroke=""}function Fv(t,e){return e&&t.__hoverStyle||t.style}function Bv(t,e){Wv(t,e,{inHover:!1,viewWidth:0,viewHeight:0},!0)}function Wv(t,e,n,r){var i=e.transform;if(!e.shouldBePainted(n.viewWidth,n.viewHeight,!1,!1))return e.__dirty&=~Zt,void(e.__isRendered=!1);var o=e.__clipPaths,a=n.prevElClipPaths,s=!1,l=!1;if(a&&!hv(o,a)||(a&&a.length&&(zv(t,n),t.restore(),l=s=!0,n.prevElClipPaths=null,n.allClipped=!1,n.prevEl=null),o&&o.length&&(zv(t,n),t.save(),Yv(o,t,n),s=!0),n.prevElClipPaths=o),n.allClipped)e.__isRendered=!1;else{e.beforeBrush&&e.beforeBrush(),e.innerBeforeBrush();var u=n.prevEl;u||(l=s=!0);var c=e instanceof Va&&e.autoBatch&&Hv(e.style);s||jv(i,u.transform)?(zv(t,n),Av(t,e)):c||zv(t,n);var d=Fv(e,n.inHover);e instanceof Va?(n.lastDrawType!==Pv&&(l=!0,n.lastDrawType=Pv),Iv(t,e,u,l,n),c&&(n.batchFill||n.batchStroke)||t.beginPath(),Sv(t,e,d,c),c&&(n.batchFill=d.fill||"",n.batchStroke=d.stroke||"")):e instanceof Ua?(n.lastDrawType!==Nv&&(l=!0,n.lastDrawType=Nv),Iv(t,e,u,l,n),Lv(t,e,d)):e instanceof Ka?(n.lastDrawType!==Ev&&(l=!0,n.lastDrawType=Ev),Cv(t,e,u,l,n),kv(t,e,d)):e.getTemporalDisplayables&&(n.lastDrawType!==Rv&&(l=!0,n.lastDrawType=Rv),Vv(t,e,n)),c&&r&&zv(t,n),e.innerAfterBrush(),e.afterBrush&&e.afterBrush(),n.prevEl=e,e.__dirty=0,e.__isRendered=!0}}function Vv(t,e,n){var r=e.getDisplayables(),i=e.getTemporalDisplayables();t.save();var o,a,s={prevElClipPaths:null,prevEl:null,allClipped:!1,viewWidth:n.viewWidth,viewHeight:n.viewHeight,inHover:n.inHover};for(o=e.getCursor(),a=r.length;o<a;o++){var l=r[o];l.beforeBrush&&l.beforeBrush(),l.innerBeforeBrush(),Wv(t,l,s,o===a-1),l.innerAfterBrush(),l.afterBrush&&l.afterBrush(),s.prevEl=l}for(var u=0,c=i.length;u<c;u++){l=i[u];l.beforeBrush&&l.beforeBrush(),l.innerBeforeBrush(),Wv(t,l,s,u===c-1),l.innerAfterBrush(),l.afterBrush&&l.afterBrush(),s.prevEl=l}e.clearTemporalDisplayables(),e.notClear=!0,t.restore()}var Gv=new Xy,$v=new Zn["a"](100),Uv=["symbol","symbolSize","symbolKeepAspect","color","backgroundColor","dashArrayX","dashArrayY","maxTileWidth","maxTileHeight"];function qv(t,e){if("none"===t)return null;var n=e.getDevicePixelRatio(),r=e.getZr(),i="svg"===r.painter.type;t.dirty&&Gv["delete"](t);var o=Gv.get(t);if(o)return o;var a=Object(s["i"])(t,{symbol:"rect",symbolSize:1,symbolKeepAspect:!0,color:"rgba(0, 0, 0, 0.2)",backgroundColor:null,dashArrayX:5,dashArrayY:5,rotation:0,maxTileWidth:512,maxTileHeight:512});"none"===a.backgroundColor&&(a.backgroundColor=null);var l={repeat:"repeat"};return u(l),l.rotation=a.rotation,l.scaleX=l.scaleY=i?1:1/n,Gv.set(t,l),t.dirty=!1,l;function u(t){for(var e,o=[n],l=!0,u=0;u<Uv.length;++u){var c=a[Uv[u]];if(null!=c&&!Object(s["t"])(c)&&!Object(s["C"])(c)&&!Object(s["z"])(c)&&"boolean"!==typeof c){l=!1;break}o.push(c)}if(l){e=o.join(",")+(i?"-svg":"");var d=$v.get(e);d&&(i?t.svgElement=d:t.image=d)}var h,f=Zv(a.dashArrayX),p=Jv(a.dashArrayY),m=Xv(a.symbol),g=Kv(f),y=Qv(p),v=!i&&Jn["d"].createCanvas(),_=i&&{tag:"g",attrs:{},key:"dcl",children:[]},b=x();function x(){for(var t=1,e=0,n=g.length;e<n;++e)t=oi(t,g[e]);var r=1;for(e=0,n=m.length;e<n;++e)r=oi(r,m[e].length);t*=r;var i=y*g.length*m.length;return{width:Math.max(1,Math.min(t,a.maxTileWidth)),height:Math.max(1,Math.min(i,a.maxTileHeight))}}function w(){h&&(h.clearRect(0,0,v.width,v.height),a.backgroundColor&&(h.fillStyle=a.backgroundColor,h.fillRect(0,0,v.width,v.height)));for(var t=0,e=0;e<p.length;++e)t+=p[e];if(!(t<=0)){var o=-y,s=0,l=0,u=0;while(o<b.height){if(s%2===0){var c=l/2%m.length,d=0,g=0,x=0;while(d<2*b.width){var w=0;for(e=0;e<f[u].length;++e)w+=f[u][e];if(w<=0)break;if(g%2===0){var M=.5*(1-a.symbolSize),S=d+f[u][g]*M,k=o+p[s]*M,L=f[u][g]*a.symbolSize,T=p[s]*a.symbolSize,D=x/2%m[c].length;O(S,k,L,T,m[c][D])}d+=f[u][g],++x,++g,g===f[u].length&&(g=0)}++u,u===f.length&&(u=0)}o+=p[s],++l,++s,s===p.length&&(s=0)}}function O(t,e,o,s,l){var u=i?1:n,c=ov(l,t*u,e*u,o*u,s*u,a.color,a.symbolKeepAspect);if(i){var d=r.painter.renderOneToVNode(c);d&&_.children.push(d)}else Bv(h,c)}}v&&(v.width=b.width*n,v.height=b.height*n,h=v.getContext("2d")),w(),l&&$v.put(e,v||_),t.image=v,t.svgElement=_,t.svgWidth=b.width,t.svgHeight=b.height}}function Xv(t){if(!t||0===t.length)return[["rect"]];if(Object(s["C"])(t))return[[t]];for(var e=!0,n=0;n<t.length;++n)if(!Object(s["C"])(t[n])){e=!1;break}if(e)return Xv([t]);var r=[];for(n=0;n<t.length;++n)Object(s["C"])(t[n])?r.push([t[n]]):r.push(t[n]);return r}function Zv(t){if(!t||0===t.length)return[[0,0]];if(Object(s["z"])(t)){var e=Math.ceil(t);return[[e,e]]}for(var n=!0,r=0;r<t.length;++r)if(!Object(s["z"])(t[r])){n=!1;break}if(n)return Zv([t]);var i=[];for(r=0;r<t.length;++r)if(Object(s["z"])(t[r])){e=Math.ceil(t[r]);i.push([e,e])}else{e=Object(s["H"])(t[r],(function(t){return Math.ceil(t)}));e.length%2===1?i.push(e.concat(e)):i.push(e)}return i}function Jv(t){if(!t||"object"===typeof t&&0===t.length)return[0,0];if(Object(s["z"])(t)){var e=Math.ceil(t);return[e,e]}var n=Object(s["H"])(t,(function(t){return Math.ceil(t)}));return t.length%2?n.concat(n):n}function Kv(t){return Object(s["H"])(t,(function(t){return Qv(t)}))}function Qv(t){for(var e=0,n=0;n<t.length;++n)e+=t[n];return t.length%2===1?2*e:e}function t_(t,e){t.eachRawSeries((function(n){if(!t.isSeriesFiltered(n)){var r=n.getData();r.hasItemVisual()&&r.each((function(t){var n=r.getItemVisual(t,"decal");if(n){var i=r.ensureUniqueItemVisual(t,"style");i.decal=qv(n,e)}}));var i=r.getVisual("decal");if(i){var o=r.getVisual("style");o.decal=qv(i,e)}}}))}var e_=new C,n_=e_,r_={};function i_(t,e){r_[t]=e}function o_(t){return r_[t]}var a_=1,s_=800,l_=900,u_=1e3,c_=2e3,d_=5e3,h_=1e3,f_=1100,p_=2e3,m_=3e3,g_=4e3,y_=4500,v_=4600,__=5e3,b_=6e3,x_=7e3,w_={PROCESSOR:{FILTER:u_,SERIES_FILTER:s_,STATISTIC:d_},VISUAL:{LAYOUT:h_,PROGRESSIVE_LAYOUT:f_,GLOBAL:p_,CHART:m_,POST_CHART_LAYOUT:v_,COMPONENT:g_,BRUSH:__,CHART_ITEM:y_,ARIA:b_,DECAL:x_}},M_="__flagInMainProcess",S_="__pendingUpdate",k_="__needsUpdateStatus",L_=/^[a-zA-Z0-9_]+$/,T_="__connectUpdateStatus",D_=0,O_=1,I_=2;function C_(t){return function(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];if(!this.isDisposed())return Y_(this,t,e);rb(this.id)}}function A_(t){return function(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];return Y_(this,t,e)}}function Y_(t,e,n){return n[0]=n[0]&&n[0].toLowerCase(),C.prototype[e].apply(t,n)}var j_,P_,E_,N_,R_,H_,z_,F_,B_,W_,V_,G_,$_,U_,q_,X_,Z_,J_,K_=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e}(C),Q_=K_.prototype;Q_.on=A_("on"),Q_.off=A_("off");var tb=function(t){function e(e,n,r){var i=t.call(this,new jy)||this;i._chartsViews=[],i._chartsMap={},i._componentsViews=[],i._componentsMap={},i._pendingActions=[],r=r||{},Object(s["C"])(n)&&(n=ub[n]),i._dom=e;var o="canvas",a="auto",l=!1;r.ssr&&Ar((function(t){var e=Ls(t),n=e.dataIndex;if(null!=n){var r=Object(s["f"])();return r.set("series_index",e.seriesIndex),r.set("data_index",n),e.ssrType&&r.set("ssr_type",e.ssrType),r}}));var u=i._zr=Or(e,{renderer:r.renderer||o,devicePixelRatio:r.devicePixelRatio,width:r.width,height:r.height,ssr:r.ssr,useDirtyRect:Object(s["P"])(r.useDirtyRect,l),useCoarsePointer:Object(s["P"])(r.useCoarsePointer,a),pointerSize:r.pointerSize});i._ssr=r.ssr,i._throttledZrFlush=Zg(Object(s["c"])(u.flush,u),17),n=Object(s["d"])(n),n&&Mh(n,!0),i._theme=n,i._locale=Fu(r.locale||Hu),i._coordSysMgr=new $d;var c=i._api=q_(i);function d(t,e){return t.__prio-e.__prio}return Xt(lb,d),Xt(ab,d),i._scheduler=new ky(i,c,ab,lb),i._messageCenter=new K_,i._initEvents(),i.resize=Object(s["c"])(i.resize,i),u.animation.on("frame",i._onframe,i),W_(u,i),V_(u,i),Object(s["R"])(i),i}return o(e,t),e.prototype._onframe=function(){if(!this._disposed){J_(this);var t=this._scheduler;if(this[S_]){var e=this[S_].silent;this[M_]=!0;try{j_(this),N_.update.call(this,null,this[S_].updateParams)}catch(a){throw this[M_]=!1,this[S_]=null,a}this._zr.flush(),this[M_]=!1,this[S_]=null,F_.call(this,e),B_.call(this,e)}else if(t.unfinished){var n=a_,r=this._model,i=this._api;t.unfinished=!1;do{var o=+new Date;t.performSeriesTasks(r),t.performDataProcessorTasks(r),H_(this,r),t.performVisualTasks(r),U_(this,this._model,i,"remain",{}),n-=+new Date-o}while(n>0&&t.unfinished);t.unfinished||this._zr.flush()}}},e.prototype.getDom=function(){return this._dom},e.prototype.getId=function(){return this.id},e.prototype.getZr=function(){return this._zr},e.prototype.isSSR=function(){return this._ssr},e.prototype.setOption=function(t,e,n){if(!this[M_])if(this._disposed)rb(this.id);else{var r,i,o;if(Object(s["A"])(e)&&(n=e.lazyUpdate,r=e.silent,i=e.replaceMerge,o=e.transition,e=e.notMerge),this[M_]=!0,!this._model||e){var a=new Qd(this._api),l=this._theme,u=this._model=new zd;u.scheduler=this._scheduler,u.ssr=this._ssr,u.init(null,null,null,l,this._locale,a)}this._model.setOption(t,{replaceMerge:i},sb);var c={seriesTransition:o,optionChanged:!0};if(n)this[S_]={silent:r,updateParams:c},this[M_]=!1,this.getZr().wakeUp();else{try{j_(this),N_.update.call(this,null,c)}catch(d){throw this[S_]=null,this[M_]=!1,d}this._ssr||this._zr.flush(),this[S_]=null,this[M_]=!1,F_.call(this,r),B_.call(this,r)}}},e.prototype.setTheme=function(){Id("ECharts#setTheme() is DEPRECATED in ECharts 3.0")},e.prototype.getModel=function(){return this._model},e.prototype.getOption=function(){return this._model&&this._model.getOption()},e.prototype.getWidth=function(){return this._zr.getWidth()},e.prototype.getHeight=function(){return this._zr.getHeight()},e.prototype.getDevicePixelRatio=function(){return this._zr.painter.dpr||a["a"].hasGlobalWindow&&window.devicePixelRatio||1},e.prototype.getRenderedCanvas=function(t){return this.renderToCanvas(t)},e.prototype.renderToCanvas=function(t){t=t||{};var e=this._zr.painter;return e.getRenderedCanvas({backgroundColor:t.backgroundColor||this._model.get("backgroundColor"),pixelRatio:t.pixelRatio||this.getDevicePixelRatio()})},e.prototype.renderToSVGString=function(t){t=t||{};var e=this._zr.painter;return e.renderToString({useViewBox:t.useViewBox})},e.prototype.getSvgDataURL=function(){if(a["a"].svgSupported){var t=this._zr,e=t.storage.getDisplayList();return Object(s["k"])(e,(function(t){t.stopAnimation(null,!0)})),t.painter.toDataURL()}},e.prototype.getDataURL=function(t){if(!this._disposed){t=t||{};var e=t.excludeComponents,n=this._model,r=[],i=this;Object(s["k"])(e,(function(t){n.eachComponent({mainType:t},(function(t){var e=i._componentsMap[t.__viewId];e.group.ignore||(r.push(e),e.group.ignore=!0)}))}));var o="svg"===this._zr.painter.getType()?this.getSvgDataURL():this.renderToCanvas(t).toDataURL("image/"+(t&&t.type||"png"));return Object(s["k"])(r,(function(t){t.group.ignore=!1})),o}rb(this.id)},e.prototype.getConnectedDataURL=function(t){if(!this._disposed){var e="svg"===t.type,n=this.group,r=Math.min,i=Math.max,o=1/0;if(hb[n]){var a=o,l=o,u=-o,c=-o,d=[],h=t&&t.pixelRatio||this.getDevicePixelRatio();Object(s["k"])(db,(function(o,h){if(o.group===n){var f=e?o.getZr().painter.getSvgDom().innerHTML:o.renderToCanvas(Object(s["d"])(t)),p=o.getDom().getBoundingClientRect();a=r(p.left,a),l=r(p.top,l),u=i(p.right,u),c=i(p.bottom,c),d.push({dom:f,left:p.left,top:p.top})}})),a*=h,l*=h,u*=h,c*=h;var f=u-a,p=c-l,m=Jn["d"].createCanvas(),g=Or(m,{renderer:e?"svg":"canvas"});if(g.resize({width:f,height:p}),e){var y="";return Object(s["k"])(d,(function(t){var e=t.left-a,n=t.top-l;y+='<g transform="translate('+e+","+n+')">'+t.dom+"</g>"})),g.painter.getSvgRoot().innerHTML=y,t.connectedBackgroundColor&&g.painter.setBackgroundColor(t.connectedBackgroundColor),g.refreshImmediately(),g.painter.toDataURL()}return t.connectedBackgroundColor&&g.add(new ss({shape:{x:0,y:0,width:f,height:p},style:{fill:t.connectedBackgroundColor}})),Object(s["k"])(d,(function(t){var e=new Ka({style:{x:t.left*h-a,y:t.top*h-l,image:t.dom}});g.add(e)})),g.refreshImmediately(),m.toDataURL("image/"+(t&&t.type||"png"))}return this.getDataURL(t)}rb(this.id)},e.prototype.convertToPixel=function(t,e){return R_(this,"convertToPixel",t,e)},e.prototype.convertFromPixel=function(t,e){return R_(this,"convertFromPixel",t,e)},e.prototype.containPixel=function(t,e){if(!this._disposed){var n,r=this._model,i=Yi(r,t);return Object(s["k"])(i,(function(t,r){r.indexOf("Models")>=0&&Object(s["k"])(t,(function(t){var i=t.coordinateSystem;if(i&&i.containPoint)n=n||!!i.containPoint(e);else if("seriesModels"===r){var o=this._chartsMap[t.__viewId];o&&o.containPoint&&(n=n||o.containPoint(e,t))}else 0}),this)}),this),!!n}rb(this.id)},e.prototype.getVisual=function(t,e){var n=this._model,r=Yi(n,t,{defaultMainType:"series"}),i=r.seriesModel;var o=i.getData(),a=r.hasOwnProperty("dataIndexInside")?r.dataIndexInside:r.hasOwnProperty("dataIndex")?o.indexOfRawIndex(r.dataIndex):null;return null!=a?Hy(o,a,e):zy(o,e)},e.prototype.getViewOfComponentModel=function(t){return this._componentsMap[t.__viewId]},e.prototype.getViewOfSeriesModel=function(t){return this._chartsMap[t.__viewId]},e.prototype._initEvents=function(){var t=this;Object(s["k"])(nb,(function(e){var n=function(n){var r,i=t.getModel(),o=n.target,a="globalout"===e;if(a?r={}:o&&Gy(o,(function(t){var e=Ls(t);if(e&&null!=e.dataIndex){var n=e.dataModel||i.getSeriesByIndex(e.seriesIndex);return r=n&&n.getDataParams(e.dataIndex,e.dataType,o)||{},!0}if(e.eventData)return r=Object(s["m"])({},e.eventData),!0}),!0),r){var l=r.componentType,u=r.componentIndex;"markLine"!==l&&"markPoint"!==l&&"markArea"!==l||(l="series",u=r.seriesIndex);var c=l&&null!=u&&i.getComponent(l,u),d=c&&t["series"===c.mainType?"_chartsMap":"_componentsMap"][c.__viewId];0,r.event=n,r.type=e,t._$eventProcessor.eventInfo={targetEl:o,packedEvent:r,model:c,view:d},t.trigger(e,r)}};n.zrEventfulCallAtLast=!0,t._zr.on(e,n,t)})),Object(s["k"])(ob,(function(e,n){t._messageCenter.on(n,(function(t){this.trigger(n,t)}),t)})),Object(s["k"])(["selectchanged"],(function(e){t._messageCenter.on(e,(function(t){this.trigger(e,t)}),t)})),Vy(this._messageCenter,this,this._api)},e.prototype.isDisposed=function(){return this._disposed},e.prototype.clear=function(){this._disposed?rb(this.id):this.setOption({series:[]},!0)},e.prototype.dispose=function(){if(this._disposed)rb(this.id);else{this._disposed=!0;var t=this.getDom();t&&Ri(this.getDom(),pb,"");var e=this,n=e._api,r=e._model;Object(s["k"])(e._componentsViews,(function(t){t.dispose(r,n)})),Object(s["k"])(e._chartsViews,(function(t){t.dispose(r,n)})),e._zr.dispose(),e._dom=e._model=e._chartsMap=e._componentsMap=e._chartsViews=e._componentsViews=e._scheduler=e._api=e._zr=e._throttledZrFlush=e._theme=e._coordSysMgr=e._messageCenter=null,delete db[e.id]}},e.prototype.resize=function(t){if(!this[M_])if(this._disposed)rb(this.id);else{this._zr.resize(t);var e=this._model;if(this._loadingFX&&this._loadingFX.resize(),e){var n=e.resetOption("media"),r=t&&t.silent;this[S_]&&(null==r&&(r=this[S_].silent),n=!0,this[S_]=null),this[M_]=!0;try{n&&j_(this),N_.update.call(this,{type:"resize",animation:Object(s["m"])({duration:0},t&&t.animation)})}catch(i){throw this[M_]=!1,i}this[M_]=!1,F_.call(this,r),B_.call(this,r)}}},e.prototype.showLoading=function(t,e){if(this._disposed)rb(this.id);else if(Object(s["A"])(t)&&(e=t,t=""),t=t||"default",this.hideLoading(),cb[t]){var n=cb[t](this._api,e),r=this._zr;this._loadingFX=n,r.add(n)}},e.prototype.hideLoading=function(){this._disposed?rb(this.id):(this._loadingFX&&this._zr.remove(this._loadingFX),this._loadingFX=null)},e.prototype.makeActionFromEvent=function(t){var e=Object(s["m"])({},t);return e.type=ob[t.type],e},e.prototype.dispatchAction=function(t,e){if(this._disposed)rb(this.id);else if(Object(s["A"])(e)||(e={silent:!!e}),ib[t.type]&&this._model)if(this[M_])this._pendingActions.push(t);else{var n=e.silent;z_.call(this,t,n);var r=e.flush;r?this._zr.flush():!1!==r&&a["a"].browser.weChat&&this._throttledZrFlush(),F_.call(this,n),B_.call(this,n)}},e.prototype.updateLabelLayout=function(){n_.trigger("series:layoutlabels",this._model,this._api,{updatedSeries:[]})},e.prototype.appendData=function(t){if(this._disposed)rb(this.id);else{var e=t.seriesIndex,n=this.getModel(),r=n.getSeriesByIndex(e);0,r.appendData(t),this._scheduler.unfinished=!0,this.getZr().wakeUp()}},e.internalField=function(){function t(t){t.clearColorPalette(),t.eachSeries((function(t){t.clearColorPalette()}))}function e(t){var e=[],n=[],r=!1;if(t.eachComponent((function(t,i){var o=i.get("zlevel")||0,a=i.get("z")||0,s=i.getZLevelKey();r=r||!!s,("series"===t?n:e).push({zlevel:o,z:a,idx:i.componentIndex,type:t,key:s})})),r){var i,o,a=e.concat(n);Xt(a,(function(t,e){return t.zlevel===e.zlevel?t.z-e.z:t.zlevel-e.zlevel})),Object(s["k"])(a,(function(e){var n=t.getComponent(e.type,e.idx),r=e.zlevel,a=e.key;null!=i&&(r=Math.max(i,r)),a?(r===i&&a!==o&&r++,o=a):o&&(r===i&&r++,o=""),i=r,n.setZLevel(r)}))}}function n(t){for(var e=[],n=t.currentStates,r=0;r<n.length;r++){var i=n[r];"emphasis"!==i&&"blur"!==i&&"select"!==i&&e.push(i)}t.selected&&t.states.select&&e.push("select"),t.hoverState===js&&t.states.emphasis?e.push("emphasis"):t.hoverState===Ys&&t.states.blur&&e.push("blur"),t.useStates(e)}function r(t,e){var n=t._zr,r=n.storage,i=0;r.traverse((function(t){t.isGroup||i++})),i>e.get("hoverLayerThreshold")&&!a["a"].node&&!a["a"].worker&&e.eachSeries((function(e){if(!e.preventUsingHoverLayer){var n=t._chartsMap[e.__viewId];n.__alive&&n.eachRendered((function(t){t.states.emphasis&&(t.states.emphasis.hoverLayer=!0)}))}}))}function i(t,e){var n=t.get("blendMode")||null;e.eachRendered((function(t){t.isGroup||(t.style.blend=n)}))}function l(t,e){if(!t.preventAutoZ){var n=t.get("z")||0,r=t.get("zlevel")||0;e.eachRendered((function(t){return u(t,n,r,-1/0),!0}))}}function u(t,e,n,r){var i=t.getTextContent(),o=t.getTextGuideLine(),a=t.isGroup;if(a)for(var s=t.childrenRef(),l=0;l<s.length;l++)r=Math.max(u(s[l],e,n,r),r);else t.z=e,t.zlevel=n,r=Math.max(t.z2,r);if(i&&(i.z=e,i.zlevel=n,isFinite(r)&&(i.z2=r+2)),o){var c=t.textGuideLineConfig;o.z=e,o.zlevel=n,isFinite(r)&&(o.z2=r+(c&&c.showAbove?1:-1))}return r}function c(t,e){e.eachRendered((function(t){if(!Gl(t)){var e=t.getTextContent(),n=t.getTextGuideLine();t.stateTransition&&(t.stateTransition=null),e&&e.stateTransition&&(e.stateTransition=null),n&&n.stateTransition&&(n.stateTransition=null),t.hasState()?(t.prevStates=t.currentStates,t.clearStates()):t.prevStates&&(t.prevStates=null)}}))}function d(t,e){var r=t.getModel("stateAnimation"),i=t.isAnimationEnabled(),o=r.get("duration"),a=o>0?{duration:o,delay:r.get("delay"),easing:r.get("easing")}:null;e.eachRendered((function(t){if(t.states&&t.states.emphasis){if(Gl(t))return;if(t instanceof Va&&Hl(t),t.__dirty){var e=t.prevStates;e&&t.useStates(e)}if(i){t.stateTransition=a;var r=t.getTextContent(),o=t.getTextGuideLine();r&&(r.stateTransition=a),o&&(o.stateTransition=a)}t.__dirty&&n(t)}}))}j_=function(t){var e=t._scheduler;e.restorePipelines(t._model),e.prepareStageTasks(),P_(t,!0),P_(t,!1),e.plan()},P_=function(t,e){for(var n=t._model,r=t._scheduler,i=e?t._componentsViews:t._chartsViews,o=e?t._componentsMap:t._chartsMap,a=t._zr,s=t._api,l=0;l<i.length;l++)i[l].__alive=!1;function u(t){var l=t.__requireNewView;t.__requireNewView=!1;var u="_ec_"+t.id+"_"+t.type,c=!l&&o[u];if(!c){var d=$i(t.type),h=e?kp.getClass(d.main,d.sub):$g.getClass(d.sub);0,c=new h,c.init(n,s),o[u]=c,i.push(c),a.add(c.group)}t.__viewId=c.__id=u,c.__alive=!0,c.__model=t,c.group.__ecComponentInfo={mainType:t.mainType,index:t.componentIndex},!e&&r.prepareView(c,t,n,s)}e?n.eachComponent((function(t,e){"series"!==t&&u(e)})):n.eachSeries(u);for(l=0;l<i.length;){var c=i[l];c.__alive?l++:(!e&&c.renderTask.dispose(),a.remove(c.group),c.dispose(n,s),i.splice(l,1),o[c.__id]===c&&delete o[c.__id],c.__id=c.group.__ecComponentInfo=null)}},E_=function(t,e,n,r,i){var o=t._model;if(o.setUpdatePayload(n),r){var a={};a[r+"Id"]=n[r+"Id"],a[r+"Index"]=n[r+"Index"],a[r+"Name"]=n[r+"Name"];var l={mainType:r,query:a};i&&(l.subType=i);var u,c=n.excludeSeriesId;null!=c&&(u=Object(s["f"])(),Object(s["k"])(ui(c),(function(t){var e=Mi(t,null);null!=e&&u.set(e,!0)}))),o&&o.eachComponent(l,(function(e){var r=u&&null!=u.get(e.id);if(!r)if(Rl(n))if(e instanceof Mp)n.type!==Hs||n.notBlur||e.get(["emphasis","disabled"])||_l(e,n,t._api);else{var i=bl(e.mainType,e.componentIndex,n.name,t._api),o=i.focusSelf,a=i.dispatchers;n.type===Hs&&o&&!n.notBlur&&vl(e.mainType,e.componentIndex,t._api),a&&Object(s["k"])(a,(function(t){n.type===Hs?ul(t):cl(t)}))}else Nl(n)&&e instanceof Mp&&(Ml(e,n,t._api),Sl(e),Z_(t))}),t),o&&o.eachComponent(l,(function(e){var n=u&&null!=u.get(e.id);n||d(t["series"===r?"_chartsMap":"_componentsMap"][e.__viewId])}),t)}else Object(s["k"])([].concat(t._componentsViews).concat(t._chartsViews),d);function d(r){r&&r.__alive&&r[e]&&r[e](r.__model,o,t._api,n)}},N_={prepareAndUpdate:function(t){j_(this),N_.update.call(this,t,{optionChanged:null!=t.newOption})},update:function(e,n){var r=this._model,i=this._api,o=this._zr,a=this._coordSysMgr,s=this._scheduler;if(r){r.setUpdatePayload(e),s.restoreData(r,e),s.performSeriesTasks(r),a.create(r,i),s.performDataProcessorTasks(r,e),H_(this,r),a.update(r,i),t(r),s.performVisualTasks(r,e),G_(this,r,i,e,n);var l=r.get("backgroundColor")||"transparent",u=r.get("darkMode");o.setBackgroundColor(l),null!=u&&"auto"!==u&&o.setDarkMode(u),n_.trigger("afterupdate",r,i)}},updateTransform:function(e){var n=this,r=this._model,i=this._api;if(r){r.setUpdatePayload(e);var o=[];r.eachComponent((function(t,a){if("series"!==t){var s=n.getViewOfComponentModel(a);if(s&&s.__alive)if(s.updateTransform){var l=s.updateTransform(a,r,i,e);l&&l.update&&o.push(s)}else o.push(s)}}));var a=Object(s["f"])();r.eachSeries((function(t){var o=n._chartsMap[t.__viewId];if(o.updateTransform){var s=o.updateTransform(t,r,i,e);s&&s.update&&a.set(t.uid,1)}else a.set(t.uid,1)})),t(r),this._scheduler.performVisualTasks(r,e,{setDirty:!0,dirtyMap:a}),U_(this,r,i,e,{},a),n_.trigger("afterupdate",r,i)}},updateView:function(e){var n=this._model;n&&(n.setUpdatePayload(e),$g.markUpdateMethod(e,"updateView"),t(n),this._scheduler.performVisualTasks(n,e,{setDirty:!0}),G_(this,n,this._api,e,{}),n_.trigger("afterupdate",n,this._api))},updateVisual:function(e){var n=this,r=this._model;r&&(r.setUpdatePayload(e),r.eachSeries((function(t){t.getData().clearAllVisual()})),$g.markUpdateMethod(e,"updateVisual"),t(r),this._scheduler.performVisualTasks(r,e,{visualType:"visual",setDirty:!0}),r.eachComponent((function(t,i){if("series"!==t){var o=n.getViewOfComponentModel(i);o&&o.__alive&&o.updateVisual(i,r,n._api,e)}})),r.eachSeries((function(t){var i=n._chartsMap[t.__viewId];i.updateVisual(t,r,n._api,e)})),n_.trigger("afterupdate",r,this._api))},updateLayout:function(t){N_.update.call(this,t)}},R_=function(t,e,n,r){if(t._disposed)rb(t.id);else{for(var i,o=t._model,a=t._coordSysMgr.getCoordinateSystems(),s=Yi(o,n),l=0;l<a.length;l++){var u=a[l];if(u[e]&&null!=(i=u[e](o,s,r)))return i}0}},H_=function(t,e){var n=t._chartsMap,r=t._scheduler;e.eachSeries((function(t){r.updateStreamModes(t,n[t.__viewId])}))},z_=function(t,e){var n=this,r=this.getModel(),i=t.type,o=t.escapeConnect,a=ib[i],l=a.actionInfo,u=(l.update||"update").split(":"),c=u.pop(),d=null!=u[0]&&$i(u[0]);this[M_]=!0;var h=[t],f=!1;t.batch&&(f=!0,h=Object(s["H"])(t.batch,(function(e){return e=Object(s["i"])(Object(s["m"])({},e),t),e.batch=null,e})));var p,m=[],g=Nl(t),y=Rl(t);if(y&&gl(this._api),Object(s["k"])(h,(function(e){if(p=a.action(e,n._model,n._api),p=p||Object(s["m"])({},e),p.type=l.event||p.type,m.push(p),y){var r=ji(t),i=r.queryOptionMap,o=r.mainTypeSpecified,u=o?i.keys()[0]:"series";E_(n,c,e,u),Z_(n)}else g?(E_(n,c,e,"series"),Z_(n)):d&&E_(n,c,e,d.main,d.sub)})),"none"!==c&&!y&&!g&&!d)try{this[S_]?(j_(this),N_.update.call(this,t),this[S_]=null):N_[c].call(this,t)}catch(b){throw this[M_]=!1,b}if(p=f?{type:l.event||i,escapeConnect:o,batch:m}:m[0],this[M_]=!1,!e){var v=this._messageCenter;if(v.trigger(p.type,p),g){var _={type:"selectchanged",escapeConnect:o,selected:kl(r),isFromClick:t.isFromClick||!1,fromAction:t.type,fromActionPayload:t};v.trigger(_.type,_)}}},F_=function(t){var e=this._pendingActions;while(e.length){var n=e.shift();z_.call(this,n,t)}},B_=function(t){!t&&this.trigger("updated")},W_=function(t,e){t.on("rendered",(function(n){e.trigger("rendered",n),!t.animation.isFinished()||e[S_]||e._scheduler.unfinished||e._pendingActions.length||e.trigger("finished")}))},V_=function(t,e){t.on("mouseover",(function(t){var n=t.target,r=Gy(n,jl);r&&(xl(r,t,e._api),Z_(e))})).on("mouseout",(function(t){var n=t.target,r=Gy(n,jl);r&&(wl(r,t,e._api),Z_(e))})).on("click",(function(t){var n=t.target,r=Gy(n,(function(t){return null!=Ls(t).dataIndex}),!0);if(r){var i=r.selected?"unselect":"select",o=Ls(r);e._api.dispatchAction({type:i,dataType:o.dataType,dataIndexInside:o.dataIndex,seriesIndex:o.seriesIndex,isFromClick:!0})}}))},G_=function(t,n,r,i,o){e(n),$_(t,n,r,i,o),Object(s["k"])(t._chartsViews,(function(t){t.__alive=!1})),U_(t,n,r,i,o),Object(s["k"])(t._chartsViews,(function(t){t.__alive||t.remove(n,r)}))},$_=function(t,e,n,r,i,o){Object(s["k"])(o||t._componentsViews,(function(t){var i=t.__model;c(i,t),t.render(i,e,n,r),l(i,t),d(i,t)}))},U_=function(t,e,n,o,a,u){var h=t._scheduler;a=Object(s["m"])(a||{},{updatedSeries:e.getSeries()}),n_.trigger("series:beforeupdate",e,n,a);var f=!1;e.eachSeries((function(e){var n=t._chartsMap[e.__viewId];n.__alive=!0;var r=n.renderTask;h.updatePayload(r,o),c(e,n),u&&u.get(e.uid)&&r.dirty(),r.perform(h.getPerformArgs(r))&&(f=!0),n.group.silent=!!e.get("silent"),i(e,n),Sl(e)})),h.unfinished=f||h.unfinished,n_.trigger("series:layoutlabels",e,n,a),n_.trigger("series:transition",e,n,a),e.eachSeries((function(e){var n=t._chartsMap[e.__viewId];l(e,n),d(e,n)})),r(t,e),n_.trigger("series:afterupdate",e,n,a)},Z_=function(t){t[k_]=!0,t.getZr().wakeUp()},J_=function(t){t[k_]&&(t.getZr().storage.traverse((function(t){Gl(t)||n(t)})),t[k_]=!1)},q_=function(t){return new(function(e){function n(){return null!==e&&e.apply(this,arguments)||this}return o(n,e),n.prototype.getCoordinateSystems=function(){return t._coordSysMgr.getCoordinateSystems()},n.prototype.getComponentByElement=function(e){while(e){var n=e.__ecComponentInfo;if(null!=n)return t._model.getComponent(n.mainType,n.index);e=e.parent}},n.prototype.enterEmphasis=function(e,n){ul(e,n),Z_(t)},n.prototype.leaveEmphasis=function(e,n){cl(e,n),Z_(t)},n.prototype.enterBlur=function(e){dl(e),Z_(t)},n.prototype.leaveBlur=function(e){hl(e),Z_(t)},n.prototype.enterSelect=function(e){fl(e),Z_(t)},n.prototype.leaveSelect=function(e){pl(e),Z_(t)},n.prototype.getModel=function(){return t.getModel()},n.prototype.getViewOfComponentModel=function(e){return t.getViewOfComponentModel(e)},n.prototype.getViewOfSeriesModel=function(e){return t.getViewOfSeriesModel(e)},n}(Wd))(t)},X_=function(t){function e(t,e){for(var n=0;n<t.length;n++){var r=t[n];r[T_]=e}}Object(s["k"])(ob,(function(n,r){t._messageCenter.on(r,(function(n){if(hb[t.group]&&t[T_]!==D_){if(n&&n.escapeConnect)return;var r=t.makeActionFromEvent(n),i=[];Object(s["k"])(db,(function(e){e!==t&&e.group===t.group&&i.push(e)})),e(i,D_),Object(s["k"])(i,(function(t){t[T_]!==O_&&t.dispatchAction(r)})),e(i,I_)}}))}))}}(),e}(C),eb=tb.prototype;eb.on=C_("on"),eb.off=C_("off"),eb.one=function(t,e,n){var r=this;function i(){for(var n=[],o=0;o<arguments.length;o++)n[o]=arguments[o];e&&e.apply&&e.apply(this,n),r.off(t,i)}Id("ECharts#one is deprecated."),this.on.call(this,t,i,n)};var nb=["click","dblclick","mouseover","mouseout","mousemove","mousedown","mouseup","globalout","contextmenu"];function rb(t){0}var ib={},ob={},ab=[],sb=[],lb=[],ub={},cb={},db={},hb={},fb=+new Date-0,pb=(new Date,"_echarts_instance_");function mb(t,e,n){var r=!(n&&n.ssr);if(r){0;var i=gb(t);if(i)return i;0}var o=new tb(t,e,n);return o.id="ec_"+fb++,db[o.id]=o,r&&Ri(t,pb,o.id),X_(o),n_.trigger("afterinit",o),o}function gb(t){return db[Hi(t,pb)]}function yb(t,e){ub[t]=e}function vb(t){Object(s["r"])(sb,t)<0&&sb.push(t)}function _b(t,e){Db(ab,t,e,c_)}function bb(t){wb("afterinit",t)}function xb(t){wb("afterupdate",t)}function wb(t,e){n_.on(t,e)}function Mb(t,e,n){Object(s["w"])(e)&&(n=e,e="");var r=Object(s["A"])(t)?t.type:[t,t={event:e}][0];t.event=(t.event||r).toLowerCase(),e=t.event,ob[e]||(Object(s["b"])(L_.test(r)&&L_.test(e)),ib[r]||(ib[r]={action:n,actionInfo:t}),ob[e]=r)}function Sb(t,e){$d.register(t,e)}function kb(t,e){Db(lb,t,e,h_,"layout")}function Lb(t,e){Db(lb,t,e,m_,"visual")}var Tb=[];function Db(t,e,n,r,i){if((Object(s["w"])(e)||Object(s["A"])(e))&&(n=e,e=r),!(Object(s["r"])(Tb,n)>=0)){Tb.push(n);var o=ky.wrapStageHandler(n,i);o.__prio=e,o.__raw=n,t.push(o)}}function Ob(t,e){cb[t]=e}function Ib(t,e,n){var r=o_("registerMap");r&&r(t,e,n)}var Cb=wf;Lb(p_,iy),Lb(y_,ay),Lb(y_,sy),Lb(p_,Ny),Lb(y_,Ry),Lb(x_,t_),vb(Mh),_b(l_,Sh),Ob("default",uy),Mb({type:Hs,event:Hs,update:Hs},s["L"]),Mb({type:zs,event:zs,update:zs},s["L"]),Mb({type:Fs,event:Fs,update:Fs},s["L"]),Mb({type:Bs,event:Bs,update:Bs},s["L"]),Mb({type:Ws,event:Ws,update:Ws},s["L"]),yb("light",Ty),yb("dark",Yy);var Ab=[],Yb={registerPreprocessor:vb,registerProcessor:_b,registerPostInit:bb,registerPostUpdate:xb,registerUpdateLifecycle:wb,registerAction:Mb,registerCoordinateSystem:Sb,registerLayout:kb,registerVisual:Lb,registerTransform:Cb,registerLoading:Ob,registerMap:Ib,registerImpl:i_,PRIORITY:w_,ComponentModel:Xc,ComponentView:kp,SeriesModel:Mp,ChartView:$g,registerComponentModel:function(t){Xc.registerClass(t)},registerComponentView:function(t){kp.registerClass(t)},registerSeriesModel:function(t){Mp.registerClass(t)},registerChartView:function(t){$g.registerClass(t)},registerSubTypeDefaulter:function(t,e){Xc.registerSubTypeDefaulter(t,e)},registerPainter:function(t,e){Ir(t,e)}};function jb(t){Object(s["t"])(t)?Object(s["k"])(t,(function(t){jb(t)})):Object(s["r"])(Ab,t)>=0||(Ab.push(t),Object(s["w"])(t)&&(t={install:t}),t.install(Yb))}var Pb=2*Math.PI,Eb=va.CMD,Nb=["top","right","bottom","left"];function Rb(t,e,n,r,i){var o=n.width,a=n.height;switch(t){case"top":r.set(n.x+o/2,n.y-e),i.set(0,-1);break;case"bottom":r.set(n.x+o/2,n.y+a+e),i.set(0,1);break;case"left":r.set(n.x-e,n.y+a/2),i.set(-1,0);break;case"right":r.set(n.x+o+e,n.y+a/2),i.set(1,0);break}}function Hb(t,e,n,r,i,o,a,s,l){a-=t,s-=e;var u=Math.sqrt(a*a+s*s);a/=u,s/=u;var c=a*n+t,d=s*n+e;if(Math.abs(r-i)%Pb<1e-4)return l[0]=c,l[1]=d,u-n;if(o){var h=r;r=Ma(i),i=Ma(h)}else r=Ma(r),i=Ma(i);r>i&&(i+=Pb);var f=Math.atan2(s,a);if(f<0&&(f+=Pb),f>=r&&f<=i||f+Pb>=r&&f+Pb<=i)return l[0]=c,l[1]=d,u-n;var p=n*Math.cos(r)+t,m=n*Math.sin(r)+e,g=n*Math.cos(i)+t,y=n*Math.sin(i)+e,v=(p-a)*(p-a)+(m-s)*(m-s),_=(g-a)*(g-a)+(y-s)*(y-s);return v<_?(l[0]=p,l[1]=m,Math.sqrt(v)):(l[0]=g,l[1]=y,Math.sqrt(_))}function zb(t,e,n,r,i,o,a,s){var l=i-t,u=o-e,c=n-t,d=r-e,h=Math.sqrt(c*c+d*d);c/=h,d/=h;var f=l*c+u*d,p=f/h;s&&(p=Math.min(Math.max(p,0),1)),p*=h;var m=a[0]=t+p*c,g=a[1]=e+p*d;return Math.sqrt((m-i)*(m-i)+(g-o)*(g-o))}function Fb(t,e,n,r,i,o,a){n<0&&(t+=n,n=-n),r<0&&(e+=r,r=-r);var s=t+n,l=e+r,u=a[0]=Math.min(Math.max(i,t),s),c=a[1]=Math.min(Math.max(o,e),l);return Math.sqrt((u-i)*(u-i)+(c-o)*(c-o))}var Bb=[];function Wb(t,e,n){var r=Fb(e.x,e.y,e.width,e.height,t.x,t.y,Bb);return n.set(Bb[0],Bb[1]),r}function Vb(t,e,n){for(var r,i,o=0,a=0,s=0,l=0,u=1/0,c=e.data,d=t.x,h=t.y,f=0;f<c.length;){var p=c[f++];1===f&&(o=c[f],a=c[f+1],s=o,l=a);var m=u;switch(p){case Eb.M:s=c[f++],l=c[f++],o=s,a=l;break;case Eb.L:m=zb(o,a,c[f],c[f+1],d,h,Bb,!0),o=c[f++],a=c[f++];break;case Eb.C:m=Se(o,a,c[f++],c[f++],c[f++],c[f++],c[f],c[f+1],d,h,Bb),o=c[f++],a=c[f++];break;case Eb.Q:m=Ce(o,a,c[f++],c[f++],c[f],c[f+1],d,h,Bb),o=c[f++],a=c[f++];break;case Eb.A:var g=c[f++],y=c[f++],v=c[f++],_=c[f++],b=c[f++],x=c[f++];f+=1;var w=!!(1-c[f++]);r=Math.cos(b)*v+g,i=Math.sin(b)*_+y,f<=1&&(s=r,l=i);var M=(d-g)*_/v+g;m=Hb(g,y,_,b,b+x,w,M,h,Bb),o=Math.cos(b+x)*v+g,a=Math.sin(b+x)*_+y;break;case Eb.R:s=o=c[f++],l=a=c[f++];var S=c[f++],k=c[f++];m=Fb(s,l,S,k,d,h,Bb);break;case Eb.Z:m=zb(o,a,s,l,d,h,Bb,!0),o=s,a=l;break}m<u&&(u=m,n.set(Bb[0],Bb[1]))}return u}var Gb=new yt,$b=new yt,Ub=new yt,qb=new yt,Xb=new yt;function Zb(t,e){if(t){var n=t.getTextGuideLine(),r=t.getTextContent();if(r&&n){var i=t.textGuideLineConfig||{},o=[[0,0],[0,0],[0,0]],a=i.candidates||Nb,s=r.getBoundingRect().clone();s.applyTransform(r.getComputedTransform());var l=1/0,u=i.anchor,c=t.getComputedTransform(),d=c&&pt([],c),h=e.get("length2")||0;u&&Ub.copy(u);for(var f=0;f<a.length;f++){var p=a[f];Rb(p,0,s,Gb,qb),yt.scaleAndAdd($b,Gb,qb,h),$b.transform(d);var m=t.getBoundingRect(),g=u?u.distance($b):t instanceof Va?Vb($b,t.path,Ub):Wb($b,m,Ub);g<l&&(l=g,$b.transform(c),Ub.transform(c),Ub.toArray(o[0]),$b.toArray(o[1]),Gb.toArray(o[2]))}Qb(o,e.get("minTurnAngle")),n.setShape({points:o})}}}var Jb=[],Kb=new yt;function Qb(t,e){if(e<=180&&e>0){e=e/180*Math.PI,Gb.fromArray(t[0]),$b.fromArray(t[1]),Ub.fromArray(t[2]),yt.sub(qb,Gb,$b),yt.sub(Xb,Ub,$b);var n=qb.len(),r=Xb.len();if(!(n<.001||r<.001)){qb.scale(1/n),Xb.scale(1/r);var i=qb.dot(Xb),o=Math.cos(e);if(o<i){var a=zb($b.x,$b.y,Ub.x,Ub.y,Gb.x,Gb.y,Jb,!1);Kb.fromArray(Jb),Kb.scaleAndAdd(Xb,a/Math.tan(Math.PI-e));var s=Ub.x!==$b.x?(Kb.x-$b.x)/(Ub.x-$b.x):(Kb.y-$b.y)/(Ub.y-$b.y);if(isNaN(s))return;s<0?yt.copy(Kb,$b):s>1&&yt.copy(Kb,Ub),Kb.toArray(t[1])}}}}function tx(t,e,n){if(n<=180&&n>0){n=n/180*Math.PI,Gb.fromArray(t[0]),$b.fromArray(t[1]),Ub.fromArray(t[2]),yt.sub(qb,$b,Gb),yt.sub(Xb,Ub,$b);var r=qb.len(),i=Xb.len();if(!(r<.001||i<.001)){qb.scale(1/r),Xb.scale(1/i);var o=qb.dot(e),a=Math.cos(n);if(o<a){var s=zb($b.x,$b.y,Ub.x,Ub.y,Gb.x,Gb.y,Jb,!1);Kb.fromArray(Jb);var l=Math.PI/2,u=Math.acos(Xb.dot(e)),c=l+u-n;if(c>=l)yt.copy(Kb,Ub);else{Kb.scaleAndAdd(Xb,s/Math.tan(Math.PI/2-c));var d=Ub.x!==$b.x?(Kb.x-$b.x)/(Ub.x-$b.x):(Kb.y-$b.y)/(Ub.y-$b.y);if(isNaN(d))return;d<0?yt.copy(Kb,$b):d>1&&yt.copy(Kb,Ub)}Kb.toArray(t[1])}}}}function ex(t,e,n,r){var i="normal"===n,o=i?t:t.ensureState(n);o.ignore=e;var a=r.get("smooth");a&&!0===a&&(a=.3),o.shape=o.shape||{},a>0&&(o.shape.smooth=a);var s=r.getModel("lineStyle").getLineStyle();i?t.useStyle(s):o.style=s}function nx(t,e){var n=e.smooth,r=e.points;if(r)if(t.moveTo(r[0][0],r[0][1]),n>0&&r.length>=3){var i=b(r[0],r[1]),o=b(r[1],r[2]);if(!i||!o)return t.lineTo(r[1][0],r[1][1]),void t.lineTo(r[2][0],r[2][1]);var a=Math.min(i,o)*n,s=M([],r[1],r[0],a/i),l=M([],r[1],r[2],a/o),u=M([],s,l,.5);t.bezierCurveTo(s[0],s[1],s[0],s[1],u[0],u[1]),t.bezierCurveTo(l[0],l[1],l[0],l[1],r[2][0],r[2][1])}else for(var c=1;c<r.length;c++)t.lineTo(r[c][0],r[c][1])}function rx(t,e,n){var r=t.getTextGuideLine(),i=t.getTextContent();if(i){for(var o=e.normal,a=o.get("show"),l=i.ignore,u=0;u<Es.length;u++){var c=Es[u],d=e[c],h="normal"===c;if(d){var f=d.get("show"),p=h?l:Object(s["P"])(i.states[c]&&i.states[c].ignore,l);if(p||!Object(s["P"])(f,a)){var m=h?r:r&&r.states[c];m&&(m.ignore=!0),r&&ex(r,!0,c,d);continue}r||(r=new Im,t.setTextGuideLine(r),h||!l&&a||ex(r,!0,"normal",e.normal),t.stateProxy&&(r.stateProxy=t.stateProxy)),ex(r,!1,c,d)}}if(r){Object(s["i"])(r.style,n),r.style.fill=null;var g=o.get("showAbove"),y=t.textGuideLineConfig=t.textGuideLineConfig||{};y.showAbove=g||!1,r.buildPath=nx}}else r&&t.removeTextGuideLine()}function ix(t,e){e=e||"labelLine";for(var n={normal:t.getModel(e)},r=0;r<Ps.length;r++){var i=Ps[r];n[i]=t.getModel([i,e])}return n}function ox(t){for(var e=[],n=0;n<t.length;n++){var r=t[n];if(!r.defaultAttr.ignore){var i=r.label,o=i.getComputedTransform(),a=i.getBoundingRect(),s=!o||o[1]<1e-5&&o[2]<1e-5,l=i.style.margin||0,u=a.clone();u.applyTransform(o),u.x-=l/2,u.y-=l/2,u.width+=l,u.height+=l;var c=s?new ng(a,o):null;e.push({label:i,labelLine:r.labelLine,rect:u,localRect:a,obb:c,priority:r.priority,defaultAttr:r.defaultAttr,layoutOption:r.computedLayoutOption,axisAligned:s,transform:o})}}return e}function ax(t,e,n,r,i,o){var a=t.length;if(!(a<2)){t.sort((function(t,n){return t.rect[e]-n.rect[e]}));for(var s,l=0,u=!1,c=[],d=0,h=0;h<a;h++){var f=t[h],p=f.rect;s=p[e]-l,s<0&&(p[e]-=s,f.label[e]-=s,u=!0);var m=Math.max(-s,0);c.push(m),d+=m,l=p[e]+p[n]}d>0&&o&&w(-d/a,0,a);var g,y,v=t[0],_=t[a-1];return b(),g<0&&M(-g,.8),y<0&&M(y,.8),b(),x(g,y,1),x(y,g,-1),b(),g<0&&S(-g),y<0&&S(y),u}function b(){g=v.rect[e]-r,y=i-_.rect[e]-_.rect[n]}function x(t,e,n){if(t<0){var r=Math.min(e,-t);if(r>0){w(r*n,0,a);var i=r+t;i<0&&M(-i*n,1)}else M(-t*n,1)}}function w(n,r,i){0!==n&&(u=!0);for(var o=r;o<i;o++){var a=t[o],s=a.rect;s[e]+=n,a.label[e]+=n}}function M(r,i){for(var o=[],s=0,l=1;l<a;l++){var u=t[l-1].rect,c=Math.max(t[l].rect[e]-u[e]-u[n],0);o.push(c),s+=c}if(s){var d=Math.min(Math.abs(r)/s,i);if(r>0)for(l=0;l<a-1;l++){var h=o[l]*d;w(h,0,l+1)}else for(l=a-1;l>0;l--){h=o[l-1]*d;w(-h,l,a)}}}function S(t){var e=t<0?-1:1;t=Math.abs(t);for(var n=Math.ceil(t/(a-1)),r=0;r<a-1;r++)if(e>0?w(n,0,r+1):w(-n,a-r-1,a),t-=n,t<=0)return}}function sx(t,e,n,r){return ax(t,"x","width",e,n,r)}function lx(t,e,n,r){return ax(t,"y","height",e,n,r)}function ux(t){var e=[];t.sort((function(t,e){return e.priority-t.priority}));var n=new Tt(0,0,0,0);function r(t){if(!t.ignore){var e=t.ensureState("emphasis");null==e.ignore&&(e.ignore=!1)}t.ignore=!0}for(var i=0;i<t.length;i++){var o=t[i],a=o.axisAligned,s=o.localRect,l=o.transform,u=o.label,c=o.labelLine;n.copy(o.rect),n.width-=.1,n.height-=.1,n.x+=.05,n.y+=.05;for(var d=o.obb,h=!1,f=0;f<e.length;f++){var p=e[f];if(n.intersect(p.rect)){if(a&&p.axisAligned){h=!0;break}if(p.obb||(p.obb=new ng(p.localRect,p.transform)),d||(d=new ng(s,l)),d.intersect(p.obb)){h=!0;break}}}h?(r(u),c&&r(c)):(u.attr("ignore",o.defaultAttr.ignore),c&&c.attr("ignore",o.defaultAttr.labelGuideIgnore),e.push(o))}}function cx(t){if(t){for(var e=[],n=0;n<t.length;n++)e.push(t[n].slice());return e}}function dx(t,e){var n=t.label,r=e&&e.getTextGuideLine();return{dataIndex:t.dataIndex,dataType:t.dataType,seriesIndex:t.seriesModel.seriesIndex,text:t.label.style.text,rect:t.hostRect,labelRect:t.rect,align:n.style.align,verticalAlign:n.style.verticalAlign,labelLinePoints:cx(r&&r.shape.points)}}var hx=["align","verticalAlign","width","height","fontSize"],fx=new Xn,px=Ci(),mx=Ci();function gx(t,e,n){for(var r=0;r<n.length;r++){var i=n[r];null!=e[i]&&(t[i]=e[i])}}var yx=["x","y","rotation"],vx=function(){function t(){this._labelList=[],this._chartViewList=[]}return t.prototype.clearLabels=function(){this._labelList=[],this._chartViewList=[]},t.prototype._addLabel=function(t,e,n,r,i){var o=r.style,a=r.__hostTarget,s=a.textConfig||{},l=r.getComputedTransform(),u=r.getBoundingRect().plain();Tt.applyTransform(u,u,l),l?fx.setLocalTransform(l):(fx.x=fx.y=fx.rotation=fx.originX=fx.originY=0,fx.scaleX=fx.scaleY=1),fx.rotation=Ma(fx.rotation);var c,d=r.__hostTarget;if(d){c=d.getBoundingRect().plain();var h=d.getComputedTransform();Tt.applyTransform(c,c,h)}var f=c&&d.getTextGuideLine();this._labelList.push({label:r,labelLine:f,seriesModel:n,dataIndex:t,dataType:e,layoutOption:i,computedLayoutOption:null,rect:u,hostRect:c,priority:c?c.width*c.height:0,defaultAttr:{ignore:r.ignore,labelGuideIgnore:f&&f.ignore,x:fx.x,y:fx.y,scaleX:fx.scaleX,scaleY:fx.scaleY,rotation:fx.rotation,style:{x:o.x,y:o.y,align:o.align,verticalAlign:o.verticalAlign,width:o.width,height:o.height,fontSize:o.fontSize},cursor:r.cursor,attachedPos:s.position,attachedRot:s.rotation}})},t.prototype.addLabelsOfSeries=function(t){var e=this;this._chartViewList.push(t);var n=t.__model,r=n.get("labelLayout");(Object(s["w"])(r)||Object(s["F"])(r).length)&&t.group.traverse((function(t){if(t.ignore)return!0;var i=t.getTextContent(),o=Ls(t);i&&!i.disableLabelLayout&&e._addLabel(o.dataIndex,o.dataType,n,i,r)}))},t.prototype.updateLayoutConfig=function(t){var e=t.getWidth(),n=t.getHeight();function r(t,e){return function(){Zb(t,e)}}for(var i=0;i<this._labelList.length;i++){var o=this._labelList[i],a=o.label,l=a.__hostTarget,u=o.defaultAttr,c=void 0;c=Object(s["w"])(o.layoutOption)?o.layoutOption(dx(o,l)):o.layoutOption,c=c||{},o.computedLayoutOption=c;var d=Math.PI/180;l&&l.setTextConfig({local:!1,position:null!=c.x||null!=c.y?null:u.attachedPos,rotation:null!=c.rotate?c.rotate*d:u.attachedRot,offset:[c.dx||0,c.dy||0]});var h=!1;if(null!=c.x?(a.x=Nr(c.x,e),a.setStyle("x",0),h=!0):(a.x=u.x,a.setStyle("x",u.style.x)),null!=c.y?(a.y=Nr(c.y,n),a.setStyle("y",0),h=!0):(a.y=u.y,a.setStyle("y",u.style.y)),c.labelLinePoints){var f=l.getTextGuideLine();f&&(f.setShape({points:c.labelLinePoints}),h=!1)}var p=px(a);p.needsUpdateLabelLine=h,a.rotation=null!=c.rotate?c.rotate*d:u.rotation,a.scaleX=u.scaleX,a.scaleY=u.scaleY;for(var m=0;m<hx.length;m++){var g=hx[m];a.setStyle(g,null!=c[g]?c[g]:u.style[g])}if(c.draggable){if(a.draggable=!0,a.cursor="move",l){var y=o.seriesModel;if(null!=o.dataIndex){var v=o.seriesModel.getData(o.dataType);y=v.getItemModel(o.dataIndex)}a.on("drag",r(l,y.getModel("labelLine")))}}else a.off("drag"),a.cursor=u.cursor}},t.prototype.layout=function(t){var e=t.getWidth(),n=t.getHeight(),r=ox(this._labelList),i=Object(s["n"])(r,(function(t){return"shiftX"===t.layoutOption.moveOverlap})),o=Object(s["n"])(r,(function(t){return"shiftY"===t.layoutOption.moveOverlap}));sx(i,0,e),lx(o,0,n);var a=Object(s["n"])(r,(function(t){return t.layoutOption.hideOverlap}));ux(a)},t.prototype.processLabelsOverall=function(){var t=this;Object(s["k"])(this._chartViewList,(function(e){var n=e.__model,r=e.ignoreLabelLineUpdate,i=n.isAnimationEnabled();e.group.traverse((function(e){if(e.ignore&&!e.forceLabelAnimation)return!0;var o=!r,a=e.getTextContent();!o&&a&&(o=px(a).needsUpdateLabelLine),o&&t._updateLabelLine(e,n),i&&t._animateLabels(e,n)}))}))},t.prototype._updateLabelLine=function(t,e){var n=t.getTextContent(),r=Ls(t),i=r.dataIndex;if(n&&null!=i){var o=e.getData(r.dataType),a=o.getItemModel(i),s={},l=o.getItemVisual(i,"style");if(l){var u=o.getVisual("drawType");s.stroke=l[u]}var c=a.getModel("labelLine");rx(t,ix(a),s),Zb(t,c)}},t.prototype._animateLabels=function(t,e){var n=t.getTextContent(),r=t.getTextGuideLine();if(n&&(t.forceLabelAnimation||!n.ignore&&!n.invisible&&!t.disableLabelAnimation&&!Gl(t))){var i=px(n),o=i.oldLayout,a=Ls(t),l=a.dataIndex,u={x:n.x,y:n.y,rotation:n.rotation},c=e.getData(a.dataType);if(o){n.attr(o);var d=t.prevStates;d&&(Object(s["r"])(d,"select")>=0&&n.attr(i.oldLayoutSelect),Object(s["r"])(d,"emphasis")>=0&&n.attr(i.oldLayoutEmphasis)),Wl(n,u,e,l)}else if(n.attr(u),!du(n).valueAnimation){var h=Object(s["P"])(n.style.opacity,1);n.style.opacity=0,Vl(n,{style:{opacity:h}},e,l)}if(i.oldLayout=u,n.states.select){var f=i.oldLayoutSelect={};gx(f,u,yx),gx(f,n.states.select,yx)}if(n.states.emphasis){var p=i.oldLayoutEmphasis={};gx(p,u,yx),gx(p,n.states.emphasis,yx)}fu(n,l,c,e,e)}if(r&&!r.ignore&&!r.invisible){i=mx(r),o=i.oldLayout;var m={points:r.shape.points};o?(r.attr({shape:o}),Wl(r,{shape:m},e)):(r.setShape(m),r.style.strokePercent=0,Vl(r,{style:{strokePercent:1}},e)),i.oldLayout=m}},t}(),_x=vx,bx=Ci();function xx(t){t.registerUpdateLifecycle("series:beforeupdate",(function(t,e,n){var r=bx(e).labelManager;r||(r=bx(e).labelManager=new _x),r.clearLabels()})),t.registerUpdateLifecycle("series:layoutlabels",(function(t,e,n){var r=bx(e).labelManager;n.updatedSeries.forEach((function(t){r.addLabelsOfSeries(e.getViewOfSeriesModel(t))})),r.updateLayoutConfig(e),r.layout(e),r.processLabelsOverall()}))}function wx(t,e,n){var r=Jn["d"].createCanvas(),i=e.getWidth(),o=e.getHeight(),a=r.style;return a&&(a.position="absolute",a.left="0",a.top="0",a.width=i+"px",a.height=o+"px",r.setAttribute("data-zr-dom-id",t)),r.width=i*n,r.height=o*n,r}jb(xx);var Mx=function(t){function e(e,n,r){var i,o=t.call(this)||this;o.motionBlur=!1,o.lastFrameAlpha=.7,o.dpr=1,o.virtual=!1,o.config={},o.incremental=!1,o.zlevel=0,o.maxRepaintRectCount=5,o.__dirty=!0,o.__firstTimePaint=!0,o.__used=!1,o.__drawIndex=0,o.__startIndex=0,o.__endIndex=0,o.__prevStartIndex=null,o.__prevEndIndex=null,r=r||jn,"string"===typeof e?i=wx(e,n,r):s["A"](e)&&(i=e,e=i.id),o.id=e,o.dom=i;var a=i.style;return a&&(s["j"](i),i.onselectstart=function(){return!1},a.padding="0",a.margin="0",a.borderWidth="0"),o.painter=n,o.dpr=r,o}return o(e,t),e.prototype.getElementCount=function(){return this.__endIndex-this.__startIndex},e.prototype.afterBrush=function(){this.__prevStartIndex=this.__startIndex,this.__prevEndIndex=this.__endIndex},e.prototype.initContext=function(){this.ctx=this.dom.getContext("2d"),this.ctx.dpr=this.dpr},e.prototype.setUnpainted=function(){this.__firstTimePaint=!0},e.prototype.createBackBuffer=function(){var t=this.dpr;this.domBack=wx("back-"+this.id,this.painter,t),this.ctxBack=this.domBack.getContext("2d"),1!==t&&this.ctxBack.scale(t,t)},e.prototype.createRepaintRects=function(t,e,n,r){if(this.__firstTimePaint)return this.__firstTimePaint=!1,null;var i,o=[],a=this.maxRepaintRectCount,s=!1,l=new Tt(0,0,0,0);function u(t){if(t.isFinite()&&!t.isZero())if(0===o.length){var e=new Tt(0,0,0,0);e.copy(t),o.push(e)}else{for(var n=!1,r=1/0,i=0,u=0;u<o.length;++u){var c=o[u];if(c.intersect(t)){var d=new Tt(0,0,0,0);d.copy(c),d.union(t),o[u]=d,n=!0;break}if(s){l.copy(t),l.union(c);var h=t.width*t.height,f=c.width*c.height,p=l.width*l.height,m=p-h-f;m<r&&(r=m,i=u)}}if(s&&(o[i].union(t),n=!0),!n){e=new Tt(0,0,0,0);e.copy(t),o.push(e)}s||(s=o.length>=a)}}for(var c=this.__startIndex;c<this.__endIndex;++c){var d=t[c];if(d){var h=d.shouldBePainted(n,r,!0,!0),f=d.__isRendered&&(d.__dirty&Zt||!h)?d.getPrevPaintRect():null;f&&u(f);var p=h&&(d.__dirty&Zt||!d.__isRendered)?d.getPaintRect():null;p&&u(p)}}for(c=this.__prevStartIndex;c<this.__prevEndIndex;++c){d=e[c],h=d&&d.shouldBePainted(n,r,!0,!0);if(d&&(!h||!d.__zr)&&d.__isRendered){f=d.getPrevPaintRect();f&&u(f)}}do{i=!1;for(c=0;c<o.length;)if(o[c].isZero())o.splice(c,1);else{for(var m=c+1;m<o.length;)o[c].intersect(o[m])?(i=!0,o[c].union(o[m]),o.splice(m,1)):m++;c++}}while(i);return this._paintRects=o,o},e.prototype.debugGetPaintRects=function(){return(this._paintRects||[]).slice()},e.prototype.resize=function(t,e){var n=this.dpr,r=this.dom,i=r.style,o=this.domBack;i&&(i.width=t+"px",i.height=e+"px"),r.width=t*n,r.height=e*n,o&&(o.width=t*n,o.height=e*n,1!==n&&this.ctxBack.scale(n,n))},e.prototype.clear=function(t,e,n){var r=this.dom,i=this.ctx,o=r.width,a=r.height;e=e||this.clearColor;var l=this.motionBlur&&!t,u=this.lastFrameAlpha,c=this.dpr,d=this;l&&(this.domBack||this.createBackBuffer(),this.ctxBack.globalCompositeOperation="copy",this.ctxBack.drawImage(r,0,0,o/c,a/c));var h=this.domBack;function f(t,n,r,o){if(i.clearRect(t,n,r,o),e&&"transparent"!==e){var a=void 0;if(s["x"](e)){var f=e.global||e.__width===r&&e.__height===o;a=f&&e.__canvasGradient||dv(i,e,{x:0,y:0,width:r,height:o}),e.__canvasGradient=a,e.__width=r,e.__height=o}else s["y"](e)&&(e.scaleX=e.scaleX||c,e.scaleY=e.scaleY||c,a=Mv(i,e,{dirty:function(){d.setUnpainted(),d.painter.refresh()}}));i.save(),i.fillStyle=a||e,i.fillRect(t,n,r,o),i.restore()}l&&(i.save(),i.globalAlpha=u,i.drawImage(h,t,n,r,o),i.restore())}!n||l?f(0,0,o,a):n.length&&s["k"](n,(function(t){f(t.x*c,t.y*c,t.width*c,t.height*c)}))},e}(C),Sx=Mx,kx=1e5,Lx=314159,Tx=.01,Dx=.001;function Ox(t){return!!t&&(!!t.__builtin__||"function"===typeof t.resize&&"function"===typeof t.refresh)}function Ix(t,e){var n=document.createElement("div");return n.style.cssText=["position:relative","width:"+t+"px","height:"+e+"px","padding:0","margin:0","border-width:0"].join(";")+";",n}var Cx=function(){function t(t,e,n,r){this.type="canvas",this._zlevelList=[],this._prevDisplayList=[],this._layers={},this._layerConfig={},this._needsManuallyCompositing=!1,this.type="canvas";var i=!t.nodeName||"CANVAS"===t.nodeName.toUpperCase();this._opts=n=s["m"]({},n||{}),this.dpr=n.devicePixelRatio||jn,this._singleCanvas=i,this.root=t;var o=t.style;o&&(s["j"](t),t.innerHTML=""),this.storage=e;var a=this._zlevelList;this._prevDisplayList=[];var l=this._layers;if(i){var u=t,c=u.width,d=u.height;null!=n.width&&(c=n.width),null!=n.height&&(d=n.height),this.dpr=n.devicePixelRatio||1,u.width=c*this.dpr,u.height=d*this.dpr,this._width=c,this._height=d;var h=new Sx(u,this,this.dpr);h.__builtin__=!0,h.initContext(),l[Lx]=h,h.zlevel=Lx,a.push(Lx),this._domRoot=t}else{this._width=pv(t,0,n),this._height=pv(t,1,n);var f=this._domRoot=Ix(this._width,this._height);t.appendChild(f)}}return t.prototype.getType=function(){return"canvas"},t.prototype.isSingleCanvas=function(){return this._singleCanvas},t.prototype.getViewportRoot=function(){return this._domRoot},t.prototype.getViewportRootOffset=function(){var t=this.getViewportRoot();if(t)return{offsetLeft:t.offsetLeft||0,offsetTop:t.offsetTop||0}},t.prototype.refresh=function(t){var e=this.storage.getDisplayList(!0),n=this._prevDisplayList,r=this._zlevelList;this._redrawId=Math.random(),this._paintList(e,n,t,this._redrawId);for(var i=0;i<r.length;i++){var o=r[i],a=this._layers[o];if(!a.__builtin__&&a.refresh){var s=0===i?this._backgroundColor:null;a.refresh(s)}}return this._opts.useDirtyRect&&(this._prevDisplayList=e.slice()),this},t.prototype.refreshHover=function(){this._paintHoverList(this.storage.getDisplayList(!1))},t.prototype._paintHoverList=function(t){var e=t.length,n=this._hoverlayer;if(n&&n.clear(),e){for(var r,i={inHover:!0,viewWidth:this._width,viewHeight:this._height},o=0;o<e;o++){var a=t[o];a.__inHover&&(n||(n=this._hoverlayer=this.getLayer(kx)),r||(r=n.ctx,r.save()),Wv(r,a,i,o===e-1))}r&&r.restore()}},t.prototype.getHoverLayer=function(){return this.getLayer(kx)},t.prototype.paintOne=function(t,e){Bv(t,e)},t.prototype._paintList=function(t,e,n,r){if(this._redrawId===r){n=n||!1,this._updateLayerStatus(t);var i=this._doPaintList(t,e,n),o=i.finished,a=i.needsRefreshHover;if(this._needsManuallyCompositing&&this._compositeManually(),a&&this._paintHoverList(t),o)this.eachLayer((function(t){t.afterBrush&&t.afterBrush()}));else{var s=this;oe((function(){s._paintList(t,e,n,r)}))}}},t.prototype._compositeManually=function(){var t=this.getLayer(Lx).ctx,e=this._domRoot.width,n=this._domRoot.height;t.clearRect(0,0,e,n),this.eachBuiltinLayer((function(r){r.virtual&&t.drawImage(r.dom,0,0,e,n)}))},t.prototype._doPaintList=function(t,e,n){for(var r=this,i=[],o=this._opts.useDirtyRect,l=0;l<this._zlevelList.length;l++){var u=this._zlevelList[l],c=this._layers[u];c.__builtin__&&c!==this._hoverlayer&&(c.__dirty||n)&&i.push(c)}for(var d=!0,h=!1,f=function(a){var s,l=i[a],u=l.ctx,c=o&&l.createRepaintRects(t,e,p._width,p._height),f=n?l.__startIndex:l.__drawIndex,m=!n&&l.incremental&&Date.now,g=m&&Date.now(),y=l.zlevel===p._zlevelList[0]?p._backgroundColor:null;if(l.__startIndex===l.__endIndex)l.clear(!1,y,c);else if(f===l.__startIndex){var v=t[f];v.incremental&&v.notClear&&!n||l.clear(!1,y,c)}-1===f&&(f=l.__startIndex);var _=function(e){var n={inHover:!1,allClipped:!1,prevEl:null,viewWidth:r._width,viewHeight:r._height};for(s=f;s<l.__endIndex;s++){var i=t[s];if(i.__inHover&&(h=!0),r._doPaintEl(i,l,o,e,n,s===l.__endIndex-1),m){var a=Date.now()-g;if(a>15)break}}n.prevElClipPaths&&u.restore()};if(c)if(0===c.length)s=l.__endIndex;else for(var b=p.dpr,x=0;x<c.length;++x){var w=c[x];u.save(),u.beginPath(),u.rect(w.x*b,w.y*b,w.width*b,w.height*b),u.clip(),_(w),u.restore()}else u.save(),_(),u.restore();l.__drawIndex=s,l.__drawIndex<l.__endIndex&&(d=!1)},p=this,m=0;m<i.length;m++)f(m);return a["a"].wxa&&s["k"](this._layers,(function(t){t&&t.ctx&&t.ctx.draw&&t.ctx.draw()})),{finished:d,needsRefreshHover:h}},t.prototype._doPaintEl=function(t,e,n,r,i,o){var a=e.ctx;if(n){var s=t.getPaintRect();(!r||s&&s.intersect(r))&&(Wv(a,t,i,o),t.setPrevPaintRect(s))}else Wv(a,t,i,o)},t.prototype.getLayer=function(t,e){this._singleCanvas&&!this._needsManuallyCompositing&&(t=Lx);var n=this._layers[t];return n||(n=new Sx("zr_"+t,this,this.dpr),n.zlevel=t,n.__builtin__=!0,this._layerConfig[t]?s["I"](n,this._layerConfig[t],!0):this._layerConfig[t-Tx]&&s["I"](n,this._layerConfig[t-Tx],!0),e&&(n.virtual=e),this.insertLayer(t,n),n.initContext()),n},t.prototype.insertLayer=function(t,e){var n=this._layers,r=this._zlevelList,i=r.length,o=this._domRoot,a=null,s=-1;if(!n[t]&&Ox(e)){if(i>0&&t>r[0]){for(s=0;s<i-1;s++)if(r[s]<t&&r[s+1]>t)break;a=n[r[s]]}if(r.splice(s+1,0,t),n[t]=e,!e.virtual)if(a){var l=a.dom;l.nextSibling?o.insertBefore(e.dom,l.nextSibling):o.appendChild(e.dom)}else o.firstChild?o.insertBefore(e.dom,o.firstChild):o.appendChild(e.dom);e.painter||(e.painter=this)}},t.prototype.eachLayer=function(t,e){for(var n=this._zlevelList,r=0;r<n.length;r++){var i=n[r];t.call(e,this._layers[i],i)}},t.prototype.eachBuiltinLayer=function(t,e){for(var n=this._zlevelList,r=0;r<n.length;r++){var i=n[r],o=this._layers[i];o.__builtin__&&t.call(e,o,i)}},t.prototype.eachOtherLayer=function(t,e){for(var n=this._zlevelList,r=0;r<n.length;r++){var i=n[r],o=this._layers[i];o.__builtin__||t.call(e,o,i)}},t.prototype.getLayers=function(){return this._layers},t.prototype._updateLayerStatus=function(t){function e(t){a&&(a.__endIndex!==t&&(a.__dirty=!0),a.__endIndex=t)}if(this.eachBuiltinLayer((function(t,e){t.__dirty=t.__used=!1})),this._singleCanvas)for(var n=1;n<t.length;n++){var r=t[n];if(r.zlevel!==t[n-1].zlevel||r.incremental){this._needsManuallyCompositing=!0;break}}var i,o,a=null,l=0;for(o=0;o<t.length;o++){r=t[o];var u=r.zlevel,c=void 0;i!==u&&(i=u,l=0),r.incremental?(c=this.getLayer(u+Dx,this._needsManuallyCompositing),c.incremental=!0,l=1):c=this.getLayer(u+(l>0?Tx:0),this._needsManuallyCompositing),c.__builtin__||s["G"]("ZLevel "+u+" has been used by unkown layer "+c.id),c!==a&&(c.__used=!0,c.__startIndex!==o&&(c.__dirty=!0),c.__startIndex=o,c.incremental?c.__drawIndex=-1:c.__drawIndex=o,e(o),a=c),r.__dirty&Zt&&!r.__inHover&&(c.__dirty=!0,c.incremental&&c.__drawIndex<0&&(c.__drawIndex=o))}e(o),this.eachBuiltinLayer((function(t,e){!t.__used&&t.getElementCount()>0&&(t.__dirty=!0,t.__startIndex=t.__endIndex=t.__drawIndex=0),t.__dirty&&t.__drawIndex<0&&(t.__drawIndex=t.__startIndex)}))},t.prototype.clear=function(){return this.eachBuiltinLayer(this._clearLayer),this},t.prototype._clearLayer=function(t){t.clear()},t.prototype.setBackgroundColor=function(t){this._backgroundColor=t,s["k"](this._layers,(function(t){t.setUnpainted()}))},t.prototype.configLayer=function(t,e){if(e){var n=this._layerConfig;n[t]?s["I"](n[t],e,!0):n[t]=e;for(var r=0;r<this._zlevelList.length;r++){var i=this._zlevelList[r];if(i===t||i===t+Tx){var o=this._layers[i];s["I"](o,n[t],!0)}}}},t.prototype.delLayer=function(t){var e=this._layers,n=this._zlevelList,r=e[t];r&&(r.dom.parentNode.removeChild(r.dom),delete e[t],n.splice(s["r"](n,t),1))},t.prototype.resize=function(t,e){if(this._domRoot.style){var n=this._domRoot;n.style.display="none";var r=this._opts,i=this.root;if(null!=t&&(r.width=t),null!=e&&(r.height=e),t=pv(i,0,r),e=pv(i,1,r),n.style.display="",this._width!==t||e!==this._height){for(var o in n.style.width=t+"px",n.style.height=e+"px",this._layers)this._layers.hasOwnProperty(o)&&this._layers[o].resize(t,e);this.refresh(!0)}this._width=t,this._height=e}else{if(null==t||null==e)return;this._width=t,this._height=e,this.getLayer(Lx).resize(t,e)}return this},t.prototype.clearLayer=function(t){var e=this._layers[t];e&&e.clear()},t.prototype.dispose=function(){this.root.innerHTML="",this.root=this.storage=this._domRoot=this._layers=null},t.prototype.getRenderedCanvas=function(t){if(t=t||{},this._singleCanvas&&!this._compositeManually)return this._layers[Lx].dom;var e=new Sx("image",this,t.pixelRatio||this.dpr);e.initContext(),e.clear(!1,t.backgroundColor||this._backgroundColor);var n=e.ctx;if(t.pixelRatio<=this.dpr){this.refresh();var r=e.dom.width,i=e.dom.height;this.eachLayer((function(t){t.__builtin__?n.drawImage(t.dom,0,0,r,i):t.renderToCanvas&&(n.save(),t.renderToCanvas(n),n.restore())}))}else for(var o={inHover:!1,viewWidth:this._width,viewHeight:this._height},a=this.storage.getDisplayList(!0),s=0,l=a.length;s<l;s++){var u=a[s];Wv(n,u,o,s===l-1)}return e.dom},t.prototype.getWidth=function(){return this._width},t.prototype.getHeight=function(){return this._height},t}(),Ax=Cx;function Yx(t){t.registerPainter("canvas",Ax)}var jx=Math.sin,Px=Math.cos,Ex=Math.PI,Nx=2*Math.PI,Rx=180/Ex,Hx=function(){function t(){}return t.prototype.reset=function(t){this._start=!0,this._d=[],this._str="",this._p=Math.pow(10,t||4)},t.prototype.moveTo=function(t,e){this._add("M",t,e)},t.prototype.lineTo=function(t,e){this._add("L",t,e)},t.prototype.bezierCurveTo=function(t,e,n,r,i,o){this._add("C",t,e,n,r,i,o)},t.prototype.quadraticCurveTo=function(t,e,n,r){this._add("Q",t,e,n,r)},t.prototype.arc=function(t,e,n,r,i,o){this.ellipse(t,e,n,n,0,r,i,o)},t.prototype.ellipse=function(t,e,n,r,i,o,a,s){var l=a-o,u=!s,c=Math.abs(l),d=Object(Re["j"])(c-Nx)||(u?l>=Nx:-l>=Nx),h=l>0?l%Nx:l%Nx+Nx,f=!1;f=!!d||!Object(Re["j"])(c)&&h>=Ex===!!u;var p=t+n*Px(o),m=e+r*jx(o);this._start&&this._add("M",p,m);var g=Math.round(i*Rx);if(d){var y=1/this._p,v=(u?1:-1)*(Nx-y);this._add("A",n,r,g,1,+u,t+n*Px(o+v),e+r*jx(o+v)),y>.01&&this._add("A",n,r,g,0,+u,p,m)}else{var _=t+n*Px(a),b=e+r*jx(a);this._add("A",n,r,g,+f,+u,_,b)}},t.prototype.rect=function(t,e,n,r){this._add("M",t,e),this._add("l",n,0),this._add("l",0,r),this._add("l",-n,0),this._add("Z")},t.prototype.closePath=function(){this._d.length>0&&this._add("Z")},t.prototype._add=function(t,e,n,r,i,o,a,s,l){for(var u=[],c=this._p,d=1;d<arguments.length;d++){var h=arguments[d];if(isNaN(h))return void(this._invalid=!0);u.push(Math.round(h*c)/c)}this._d.push(t+u.join(" ")),this._start="Z"===t},t.prototype.generateStr=function(){this._str=this._invalid?"":this._d.join(""),this._d=[]},t.prototype.getStr=function(){return this._str},t}(),zx=Hx,Fx="none",Bx=Math.round;function Wx(t){var e=t.fill;return null!=e&&e!==Fx}function Vx(t){var e=t.stroke;return null!=e&&e!==Fx}var Gx=["lineCap","miterLimit","lineJoin"],$x=Object(s["H"])(Gx,(function(t){return"stroke-"+t.toLowerCase()}));function Ux(t,e,n,r){var i=null==e.opacity?1:e.opacity;if(n instanceof Ka)t("opacity",i);else{if(Wx(e)){var o=Object(Re["p"])(e.fill);t("fill",o.color);var a=null!=e.fillOpacity?e.fillOpacity*o.opacity*i:o.opacity*i;(r||a<1)&&t("fill-opacity",a)}else t("fill",Fx);if(Vx(e)){var s=Object(Re["p"])(e.stroke);t("stroke",s.color);var l=e.strokeNoScale?n.getLineScale():1,u=l?(e.lineWidth||0)/l:0,c=null!=e.strokeOpacity?e.strokeOpacity*s.opacity*i:s.opacity*i,d=e.strokeFirst;if((r||1!==u)&&t("stroke-width",u),(r||d)&&t("paint-order",d?"stroke":"fill"),(r||c<1)&&t("stroke-opacity",c),e.lineDash){var h=gv(n),f=h[0],p=h[1];f&&(p=Bx(p||0),t("stroke-dasharray",f.join(",")),(p||r)&&t("stroke-dashoffset",p))}else r&&t("stroke-dasharray",Fx);for(var m=0;m<Gx.length;m++){var g=Gx[m];if(r||e[g]!==za[g]){var y=e[g]||za[g];y&&t($x[m],y)}}}else r&&t("stroke",Fx)}}var qx="http://www.w3.org/2000/svg",Xx="http://www.w3.org/1999/xlink",Zx="http://www.w3.org/2000/xmlns/",Jx="http://www.w3.org/XML/1998/namespace",Kx="ecmeta_";function Qx(t){return document.createElementNS(qx,t)}function tw(t,e,n,r,i){return{tag:t,attrs:n||{},children:r,text:i,key:e}}function ew(t,e){var n=[];if(e)for(var r in e){var i=e[r],o=r;!1!==i&&(!0!==i&&null!=i&&(o+='="'+i+'"'),n.push(o))}return"<"+t+" "+n.join(" ")+">"}function nw(t){return"</"+t+">"}function rw(t,e){e=e||{};var n=e.newline?"\n":"";function r(t){var e=t.children,i=t.tag,o=t.attrs,a=t.text;return ew(i,o)+("style"!==i?V(a):a||"")+(e?""+n+Object(s["H"])(e,(function(t){return r(t)})).join(n)+n:"")+nw(i)}return r(t)}function iw(t,e,n){n=n||{};var r=n.newline?"\n":"",i=" {"+r,o=r+"}",a=Object(s["H"])(Object(s["F"])(t),(function(e){return e+i+Object(s["H"])(Object(s["F"])(t[e]),(function(n){return n+":"+t[e][n]+";"})).join(r)+o})).join(r),l=Object(s["H"])(Object(s["F"])(e),(function(t){return"@keyframes "+t+i+Object(s["H"])(Object(s["F"])(e[t]),(function(n){return n+i+Object(s["H"])(Object(s["F"])(e[t][n]),(function(r){var i=e[t][n][r];return"d"===r&&(i='path("'+i+'")'),r+":"+i+";"})).join(r)+o})).join(r)+o})).join(r);return a||l?["<![CDATA[",a,l,"]]>"].join(r):""}function ow(t){return{zrId:t,shadowCache:{},patternCache:{},gradientCache:{},clipPathCache:{},defs:{},cssNodes:{},cssAnims:{},cssStyleCache:{},cssAnimIdx:0,shadowIdx:0,gradientIdx:0,patternIdx:0,clipPathIdx:0}}function aw(t,e,n,r){return tw("svg","root",{width:t,height:e,xmlns:qx,"xmlns:xlink":Xx,version:"1.1",baseProfile:"full",viewBox:!!r&&"0 0 "+t+" "+e},n)}var sw=0;function lw(){return sw++}var uw={cubicIn:"0.32,0,0.67,0",cubicOut:"0.33,1,0.68,1",cubicInOut:"0.65,0,0.35,1",quadraticIn:"0.11,0,0.5,0",quadraticOut:"0.5,1,0.89,1",quadraticInOut:"0.45,0,0.55,1",quarticIn:"0.5,0,0.75,0",quarticOut:"0.25,1,0.5,1",quarticInOut:"0.76,0,0.24,1",quinticIn:"0.64,0,0.78,0",quinticOut:"0.22,1,0.36,1",quinticInOut:"0.83,0,0.17,1",sinusoidalIn:"0.12,0,0.39,0",sinusoidalOut:"0.61,1,0.88,1",sinusoidalInOut:"0.37,0,0.63,1",exponentialIn:"0.7,0,0.84,0",exponentialOut:"0.16,1,0.3,1",exponentialInOut:"0.87,0,0.13,1",circularIn:"0.55,0,1,0.45",circularOut:"0,0.55,0.45,1",circularInOut:"0.85,0,0.15,1"},cw="transform-origin";function dw(t,e,n){var r=Object(s["m"])({},t.shape);Object(s["m"])(r,e),t.buildPath(n,r);var i=new zx;return i.reset(Object(Re["f"])(t)),n.rebuildPath(i,1),i.generateStr(),i.getStr()}function hw(t,e){var n=e.originX,r=e.originY;(n||r)&&(t[cw]=n+"px "+r+"px")}var fw={fill:"fill",opacity:"opacity",lineWidth:"stroke-width",lineDashOffset:"stroke-dashoffset"};function pw(t,e){var n=e.zrId+"-ani-"+e.cssAnimIdx++;return e.cssAnims[n]=t,n}function mw(t,e,n){var r,i,o=t.shape.paths,a={};if(Object(s["k"])(o,(function(t){var e=ow(n.zrId);e.animation=!0,yw(t,{},e,!0);var o=e.cssAnims,l=e.cssNodes,u=Object(s["F"])(o),c=u.length;if(c){i=u[c-1];var d=o[i];for(var h in d){var f=d[h];a[h]=a[h]||{d:""},a[h].d+=f.d||""}for(var p in l){var m=l[p].animation;m.indexOf(i)>=0&&(r=m)}}})),r){e.d=!1;var l=pw(a,n);return r.replace(i,l)}}function gw(t){return Object(s["C"])(t)?uw[t]?"cubic-bezier("+uw[t]+")":je(t)?t:"":""}function yw(t,e,n,r){var i=t.animators,o=i.length,a=[];if(t instanceof Vm){var l=mw(t,e,n);if(l)a.push(l);else if(!o)return}else if(!o)return;for(var u={},c=0;c<o;c++){var d=i[c],h=[d.getMaxTime()/1e3+"s"],f=gw(d.getClip().easing),p=d.getDelay();f?h.push(f):h.push("linear"),p&&h.push(p/1e3+"s"),d.getLoop()&&h.push("infinite");var m=h.join(" ");u[m]=u[m]||[m,[]],u[m][1].push(d)}function g(i){var o,a=i[1],l=a.length,u={},c={},d={},h="animation-timing-function";function f(t,e,n){for(var r=t.getTracks(),i=t.getMaxTime(),o=0;o<r.length;o++){var a=r[o];if(a.needsAnimate()){var l=a.keyframes,u=a.propName;if(n&&(u=n(u)),u)for(var c=0;c<l.length;c++){var d=l[c],f=Math.round(d.time/i*100)+"%",p=gw(d.easing),m=d.rawValue;(Object(s["C"])(m)||Object(s["z"])(m))&&(e[f]=e[f]||{},e[f][u]=d.rawValue,p&&(e[f][h]=p))}}}}for(var p=0;p<l;p++){var m=a[p],g=m.targetName;g?"shape"===g&&f(m,c):!r&&f(m,u)}for(var y in u){var v={};qn(v,t),Object(s["m"])(v,u[y]);var _=Object(Re["g"])(v),b=u[y][h];d[y]=_?{transform:_}:{},hw(d[y],v),b&&(d[y][h]=b)}var x=!0;for(var y in c){d[y]=d[y]||{};var w=!o;b=c[y][h];w&&(o=new va);var M=o.len();o.reset(),d[y].d=dw(t,c[y],o);var S=o.len();if(!w&&M!==S){x=!1;break}b&&(d[y][h]=b)}if(!x)for(var y in d)delete d[y].d;if(!r)for(p=0;p<l;p++){m=a[p],g=m.targetName;"style"===g&&f(m,d,(function(t){return fw[t]}))}var k,L=Object(s["F"])(d),T=!0;for(p=1;p<L.length;p++){var D=L[p-1],O=L[p];if(d[D][cw]!==d[O][cw]){T=!1;break}k=d[D][cw]}if(T&&k){for(var y in d)d[y][cw]&&delete d[y][cw];e[cw]=k}if(Object(s["n"])(L,(function(t){return Object(s["F"])(d[t]).length>0})).length){var I=pw(d,n);return I+" "+i[0]+" both"}}for(var y in u){l=g(u[y]);l&&a.push(l)}if(a.length){var v=n.zrId+"-cls-"+lw();n.cssNodes["."+v]={animation:a.join(",")},e["class"]=v}}function vw(t,e,n){if(!t.ignore)if(t.isSilent()){var r={"pointer-events":"none"};_w(r,e,n,!0)}else{var i=t.states.emphasis&&t.states.emphasis.style?t.states.emphasis.style:{},o=i.fill;if(!o){var a=t.style&&t.style.fill,s=t.states.select&&t.states.select.style&&t.states.select.style.fill,l=t.currentStates.indexOf("select")>=0&&s||a;l&&(o=Object(Ne["d"])(l))}var u=i.lineWidth;if(u){var c=!i.strokeNoScale&&t.transform?t.transform[0]:1;u/=c}r={cursor:"pointer"};o&&(r.fill=o),i.stroke&&(r.stroke=i.stroke),u&&(r["stroke-width"]=u),_w(r,e,n,!0)}}function _w(t,e,n,r){var i=JSON.stringify(t),o=n.cssStyleCache[i];o||(o=n.zrId+"-cls-"+lw(),n.cssStyleCache[i]=o,n.cssNodes["."+o+(r?":hover":"")]=t),e["class"]=e["class"]?e["class"]+" "+o:o}var bw=Math.round;function xw(t){return t&&Object(s["C"])(t.src)}function ww(t){return t&&Object(s["w"])(t.toDataURL)}function Mw(t,e,n,r){Ux((function(i,o){var a="fill"===i||"stroke"===i;a&&Object(Re["k"])(o)?Rw(e,t,i,r):a&&Object(Re["n"])(o)?Hw(n,t,i,r):t[i]=a&&"none"===o?"transparent":o}),e,n,!1),Nw(n,t,r)}function Sw(t,e){var n=Cr(e);n&&(n.each((function(e,n){null!=e&&(t[(Kx+n).toLowerCase()]=e+"")})),e.isSilent()&&(t[Kx+"silent"]="true"))}function kw(t){return Object(Re["j"])(t[0]-1)&&Object(Re["j"])(t[1])&&Object(Re["j"])(t[2])&&Object(Re["j"])(t[3]-1)}function Lw(t){return Object(Re["j"])(t[4])&&Object(Re["j"])(t[5])}function Tw(t,e,n){if(e&&(!Lw(e)||!kw(e))){var r=n?10:1e4;t.transform=kw(e)?"translate("+bw(e[4]*r)/r+" "+bw(e[5]*r)/r+")":Object(Re["e"])(e)}}function Dw(t,e,n){for(var r=t.points,i=[],o=0;o<r.length;o++)i.push(bw(r[o][0]*n)/n),i.push(bw(r[o][1]*n)/n);e.points=i.join(" ")}function Ow(t){return!t.smooth}function Iw(t){var e=Object(s["H"])(t,(function(t){return"string"===typeof t?[t,t]:t}));return function(t,n,r){for(var i=0;i<e.length;i++){var o=e[i],a=t[o[0]];null!=a&&(n[o[1]]=bw(a*r)/r)}}}var Cw={circle:[Iw(["cx","cy","r"])],polyline:[Dw,Ow],polygon:[Dw,Ow]};function Aw(t){for(var e=t.animators,n=0;n<e.length;n++)if("shape"===e[n].targetName)return!0;return!1}function Yw(t,e){var n=t.style,r=t.shape,i=Cw[t.type],o={},a=e.animation,s="path",l=t.style.strokePercent,u=e.compress&&Object(Re["f"])(t)||4;if(!i||e.willUpdate||i[1]&&!i[1](r)||a&&Aw(t)||l<1){var c=!t.path||t.shapeChanged();t.path||t.createPathProxy();var d=t.path;c&&(d.beginPath(),t.buildPath(d,t.shape),t.pathUpdated());var h=d.getVersion(),f=t,p=f.__svgPathBuilder;f.__svgPathVersion===h&&p&&l===f.__svgPathStrokePercent||(p||(p=f.__svgPathBuilder=new zx),p.reset(u),d.rebuildPath(p,l),p.generateStr(),f.__svgPathVersion=h,f.__svgPathStrokePercent=l),o.d=p.getStr()}else{s=t.type;var m=Math.pow(10,u);i[0](r,o,m)}return Tw(o,t.transform),Mw(o,n,t,e),Sw(o,t),e.animation&&yw(t,o,e),e.emphasis&&vw(t,o,e),tw(s,t.id+"",o)}function jw(t,e){var n=t.style,r=n.image;if(r&&!Object(s["C"])(r)&&(xw(r)?r=r.src:ww(r)&&(r=r.toDataURL())),r){var i=n.x||0,o=n.y||0,a=n.width,l=n.height,u={href:r,width:a,height:l};return i&&(u.x=i),o&&(u.y=o),Tw(u,t.transform),Mw(u,n,t,e),Sw(u,t),e.animation&&yw(t,u,e),tw("image",t.id+"",u)}}function Pw(t,e){var n=t.style,r=n.text;if(null!=r&&(r+=""),r&&!isNaN(n.x)&&!isNaN(n.y)){var i=n.font||Jn["a"],o=n.x||0,a=Object(Re["b"])(n.y||0,ir(i),n.textBaseline),s=Re["a"][n.textAlign]||n.textAlign,l={"dominant-baseline":"central","text-anchor":s};if(ys(n)){var u="",c=n.fontStyle,d=ms(n.fontSize);if(!parseFloat(d))return;var h=n.fontFamily||Jn["b"],f=n.fontWeight;u+="font-size:"+d+";font-family:"+h+";",c&&"normal"!==c&&(u+="font-style:"+c+";"),f&&"normal"!==f&&(u+="font-weight:"+f+";"),l.style=u}else l.style="font: "+i;return r.match(/\s/)&&(l["xml:space"]="preserve"),o&&(l.x=o),a&&(l.y=a),Tw(l,t.transform),Mw(l,n,t,e),Sw(l,t),e.animation&&yw(t,l,e),tw("text",t.id+"",l,void 0,r)}}function Ew(t,e){return t instanceof Va?Yw(t,e):t instanceof Ka?jw(t,e):t instanceof Ua?Pw(t,e):void 0}function Nw(t,e,n){var r=t.style;if(Object(Re["i"])(r)){var i=Object(Re["h"])(t),o=n.shadowCache,a=o[i];if(!a){var s=t.getGlobalScale(),l=s[0],u=s[1];if(!l||!u)return;var c=r.shadowOffsetX||0,d=r.shadowOffsetY||0,h=r.shadowBlur,f=Object(Re["p"])(r.shadowColor),p=f.opacity,m=f.color,g=h/2/l,y=h/2/u,v=g+" "+y;a=n.zrId+"-s"+n.shadowIdx++,n.defs[a]=tw("filter",a,{id:a,x:"-100%",y:"-100%",width:"300%",height:"300%"},[tw("feDropShadow","",{dx:c/l,dy:d/u,stdDeviation:v,"flood-color":m,"flood-opacity":p})]),o[i]=a}e.filter=Object(Re["d"])(a)}}function Rw(t,e,n,r){var i,o=t[n],a={gradientUnits:o.global?"userSpaceOnUse":"objectBoundingBox"};if(Object(Re["m"])(o))i="linearGradient",a.x1=o.x,a.y1=o.y,a.x2=o.x2,a.y2=o.y2;else{if(!Object(Re["o"])(o))return void 0;i="radialGradient",a.cx=Object(s["P"])(o.x,.5),a.cy=Object(s["P"])(o.y,.5),a.r=Object(s["P"])(o.r,.5)}for(var l=o.colorStops,u=[],c=0,d=l.length;c<d;++c){var h=100*Object(Re["q"])(l[c].offset)+"%",f=l[c].color,p=Object(Re["p"])(f),m=p.color,g=p.opacity,y={offset:h};y["stop-color"]=m,g<1&&(y["stop-opacity"]=g),u.push(tw("stop",c+"",y))}var v=tw(i,"",a,u),_=rw(v),b=r.gradientCache,x=b[_];x||(x=r.zrId+"-g"+r.gradientIdx++,b[_]=x,a.id=x,r.defs[x]=tw(i,x,a,u)),e[n]=Object(Re["d"])(x)}function Hw(t,e,n,r){var i,o=t.style[n],a=t.getBoundingRect(),l={},u=o.repeat,c="no-repeat"===u,d="repeat-x"===u,h="repeat-y"===u;if(Object(Re["l"])(o)){var f=o.imageWidth,p=o.imageHeight,m=void 0,g=o.image;if(Object(s["C"])(g)?m=g:xw(g)?m=g.src:ww(g)&&(m=g.toDataURL()),"undefined"===typeof Image){var y="Image width/height must been given explictly in svg-ssr renderer.";Object(s["b"])(f,y),Object(s["b"])(p,y)}else if(null==f||null==p){var v=function(t,e){if(t){var n=t.elm,r=f||e.width,i=p||e.height;"pattern"===t.tag&&(d?(i=1,r/=a.width):h&&(r=1,i/=a.height)),t.attrs.width=r,t.attrs.height=i,n&&(n.setAttribute("width",r),n.setAttribute("height",i))}},_=uo(m,null,t,(function(t){c||v(M,t),v(i,t)}));_&&_.width&&_.height&&(f=f||_.width,p=p||_.height)}i=tw("image","img",{href:m,width:f,height:p}),l.width=f,l.height=p}else o.svgElement&&(i=Object(s["d"])(o.svgElement),l.width=o.svgWidth,l.height=o.svgHeight);if(i){var b,x;c?b=x=1:d?(x=1,b=l.width/a.width):h?(b=1,x=l.height/a.height):l.patternUnits="userSpaceOnUse",null==b||isNaN(b)||(l.width=b),null==x||isNaN(x)||(l.height=x);var w=Object(Re["g"])(o);w&&(l.patternTransform=w);var M=tw("pattern","",l,[i]),S=rw(M),k=r.patternCache,L=k[S];L||(L=r.zrId+"-p"+r.patternIdx++,k[S]=L,l.id=L,M=r.defs[L]=tw("pattern",L,l,[i])),e[n]=Object(Re["d"])(L)}}function zw(t,e,n){var r=n.clipPathCache,i=n.defs,o=r[t.id];if(!o){o=n.zrId+"-c"+n.clipPathIdx++;var a={id:o};r[t.id]=o,i[o]=tw("clipPath",o,a,[Yw(t,n)])}e["clip-path"]=Object(Re["d"])(o)}function Fw(t){return document.createTextNode(t)}function Bw(t,e,n){t.insertBefore(e,n)}function Ww(t,e){t.removeChild(e)}function Vw(t,e){t.appendChild(e)}function Gw(t){return t.parentNode}function $w(t){return t.nextSibling}function Uw(t,e){t.textContent=e}var qw=58,Xw=120,Zw=tw("","");function Jw(t){return void 0===t}function Kw(t){return void 0!==t}function Qw(t,e,n){for(var r={},i=e;i<=n;++i){var o=t[i].key;void 0!==o&&(r[o]=i)}return r}function tM(t,e){var n=t.key===e.key,r=t.tag===e.tag;return r&&n}function eM(t){var e,n=t.children,r=t.tag;if(Kw(r)){var i=t.elm=Qx(r);if(iM(Zw,t),Object(s["t"])(n))for(e=0;e<n.length;++e){var o=n[e];null!=o&&Vw(i,eM(o))}else Kw(t.text)&&!Object(s["A"])(t.text)&&Vw(i,Fw(t.text))}else t.elm=Fw(t.text);return t.elm}function nM(t,e,n,r,i){for(;r<=i;++r){var o=n[r];null!=o&&Bw(t,eM(o),e)}}function rM(t,e,n,r){for(;n<=r;++n){var i=e[n];if(null!=i)if(Kw(i.tag)){var o=Gw(i.elm);Ww(o,i.elm)}else Ww(t,i.elm)}}function iM(t,e){var n,r=e.elm,i=t&&t.attrs||{},o=e.attrs||{};if(i!==o){for(n in o){var a=o[n],s=i[n];s!==a&&(!0===a?r.setAttribute(n,""):!1===a?r.removeAttribute(n):"style"===n?r.style.cssText=a:n.charCodeAt(0)!==Xw?r.setAttribute(n,a):"xmlns:xlink"===n||"xmlns"===n?r.setAttributeNS(Zx,n,a):n.charCodeAt(3)===qw?r.setAttributeNS(Jx,n,a):n.charCodeAt(5)===qw?r.setAttributeNS(Xx,n,a):r.setAttribute(n,a))}for(n in i)n in o||r.removeAttribute(n)}}function oM(t,e,n){var r,i,o,a,s=0,l=0,u=e.length-1,c=e[0],d=e[u],h=n.length-1,f=n[0],p=n[h];while(s<=u&&l<=h)null==c?c=e[++s]:null==d?d=e[--u]:null==f?f=n[++l]:null==p?p=n[--h]:tM(c,f)?(aM(c,f),c=e[++s],f=n[++l]):tM(d,p)?(aM(d,p),d=e[--u],p=n[--h]):tM(c,p)?(aM(c,p),Bw(t,c.elm,$w(d.elm)),c=e[++s],p=n[--h]):tM(d,f)?(aM(d,f),Bw(t,d.elm,c.elm),d=e[--u],f=n[++l]):(Jw(r)&&(r=Qw(e,s,u)),i=r[f.key],Jw(i)?Bw(t,eM(f),c.elm):(o=e[i],o.tag!==f.tag?Bw(t,eM(f),c.elm):(aM(o,f),e[i]=void 0,Bw(t,o.elm,c.elm))),f=n[++l]);(s<=u||l<=h)&&(s>u?(a=null==n[h+1]?null:n[h+1].elm,nM(t,a,n,l,h)):rM(t,e,s,u))}function aM(t,e){var n=e.elm=t.elm,r=t.children,i=e.children;t!==e&&(iM(t,e),Jw(e.text)?Kw(r)&&Kw(i)?r!==i&&oM(n,r,i):Kw(i)?(Kw(t.text)&&Uw(n,""),nM(n,null,i,0,i.length-1)):Kw(r)?rM(n,r,0,r.length-1):Kw(t.text)&&Uw(n,""):t.text!==e.text&&(Kw(r)&&rM(n,r,0,r.length-1),Uw(n,e.text)))}function sM(t,e){if(tM(t,e))aM(t,e);else{var n=t.elm,r=Gw(n);eM(e),null!==r&&(Bw(r,e.elm,$w(n)),rM(r,[t],0,0))}return e}var lM=0,uM=function(){function t(t,e,n){if(this.type="svg",this.refreshHover=cM("refreshHover"),this.configLayer=cM("configLayer"),this.storage=e,this._opts=n=Object(s["m"])({},n),this.root=t,this._id="zr"+lM++,this._oldVNode=aw(n.width,n.height),t&&!n.ssr){var r=this._viewport=document.createElement("div");r.style.cssText="position:relative;overflow:hidden";var i=this._svgDom=this._oldVNode.elm=Qx("svg");iM(null,this._oldVNode),r.appendChild(i),t.appendChild(r)}this.resize(n.width,n.height)}return t.prototype.getType=function(){return this.type},t.prototype.getViewportRoot=function(){return this._viewport},t.prototype.getViewportRootOffset=function(){var t=this.getViewportRoot();if(t)return{offsetLeft:t.offsetLeft||0,offsetTop:t.offsetTop||0}},t.prototype.getSvgDom=function(){return this._svgDom},t.prototype.refresh=function(){if(this.root){var t=this.renderToVNode({willUpdate:!0});t.attrs.style="position:absolute;left:0;top:0;user-select:none",sM(this._oldVNode,t),this._oldVNode=t}},t.prototype.renderOneToVNode=function(t){return Ew(t,ow(this._id))},t.prototype.renderToVNode=function(t){t=t||{};var e=this.storage.getDisplayList(!0),n=this._width,r=this._height,i=ow(this._id);i.animation=t.animation,i.willUpdate=t.willUpdate,i.compress=t.compress,i.emphasis=t.emphasis;var o=[],a=this._bgVNode=dM(n,r,this._backgroundColor,i);a&&o.push(a);var l=t.compress?null:this._mainVNode=tw("g","main",{},[]);this._paintList(e,i,l?l.children:o),l&&o.push(l);var u=Object(s["H"])(Object(s["F"])(i.defs),(function(t){return i.defs[t]}));if(u.length&&o.push(tw("defs","defs",{},u)),t.animation){var c=iw(i.cssNodes,i.cssAnims,{newline:!0});if(c){var d=tw("style","stl",{},[],c);o.push(d)}}return aw(n,r,o,t.useViewBox)},t.prototype.renderToString=function(t){return t=t||{},rw(this.renderToVNode({animation:Object(s["P"])(t.cssAnimation,!0),emphasis:Object(s["P"])(t.cssEmphasis,!0),willUpdate:!1,compress:!0,useViewBox:Object(s["P"])(t.useViewBox,!0)}),{newline:!0})},t.prototype.setBackgroundColor=function(t){this._backgroundColor=t},t.prototype.getSvgRoot=function(){return this._mainVNode&&this._mainVNode.elm},t.prototype._paintList=function(t,e,n){for(var r,i,o=t.length,a=[],s=0,l=0,u=0;u<o;u++){var c=t[u];if(!c.invisible){var d=c.__clipPaths,h=d&&d.length||0,f=i&&i.length||0,p=void 0;for(p=Math.max(h-1,f-1);p>=0;p--)if(d&&i&&d[p]===i[p])break;for(var m=f-1;m>p;m--)s--,r=a[s-1];for(var g=p+1;g<h;g++){var y={};zw(d[g],y,e);var v=tw("g","clip-g-"+l++,y,[]);(r?r.children:n).push(v),a[s++]=v,r=v}i=d;var _=Ew(c,e);_&&(r?r.children:n).push(_)}}},t.prototype.resize=function(t,e){var n=this._opts,r=this.root,i=this._viewport;if(null!=t&&(n.width=t),null!=e&&(n.height=e),r&&i&&(i.style.display="none",t=pv(r,0,n),e=pv(r,1,n),i.style.display=""),this._width!==t||this._height!==e){if(this._width=t,this._height=e,i){var o=i.style;o.width=t+"px",o.height=e+"px"}if(Object(Re["n"])(this._backgroundColor))this.refresh();else{var a=this._svgDom;a&&(a.setAttribute("width",t),a.setAttribute("height",e));var s=this._bgVNode&&this._bgVNode.elm;s&&(s.setAttribute("width",t),s.setAttribute("height",e))}}},t.prototype.getWidth=function(){return this._width},t.prototype.getHeight=function(){return this._height},t.prototype.dispose=function(){this.root&&(this.root.innerHTML=""),this._svgDom=this._viewport=this.storage=this._oldVNode=this._bgVNode=this._mainVNode=null},t.prototype.clear=function(){this._svgDom&&(this._svgDom.innerHTML=null),this._oldVNode=null},t.prototype.toDataURL=function(t){var e=this.renderToString(),n="data:image/svg+xml;";return t?(e=Object(Re["c"])(e),e&&n+"base64,"+e):n+"charset=UTF-8,"+encodeURIComponent(e)},t}();function cM(t){return function(){0}}function dM(t,e,n,r){var i;if(n&&"none"!==n)if(i=tw("rect","bg",{width:t,height:e,x:"0",y:"0"}),Object(Re["k"])(n))Rw({fill:n},i.attrs,"fill",r);else if(Object(Re["n"])(n))Hw({style:{fill:n},dirty:s["L"],getBoundingRect:function(){return{width:t,height:e}}},i.attrs,"fill",r);else{var o=Object(Re["p"])(n),a=o.color,l=o.opacity;i.attrs.fill=a,l<1&&(i.attrs["fill-opacity"]=l)}return i}var hM=uM;function fM(t){t.registerPainter("svg",hM)}function pM(t){return null==t?0:t.length||1}function mM(t){return t}var gM=function(){function t(t,e,n,r,i,o){this._old=t,this._new=e,this._oldKeyGetter=n||mM,this._newKeyGetter=r||mM,this.context=i,this._diffModeMultiple="multiple"===o}return t.prototype.add=function(t){return this._add=t,this},t.prototype.update=function(t){return this._update=t,this},t.prototype.updateManyToOne=function(t){return this._updateManyToOne=t,this},t.prototype.updateOneToMany=function(t){return this._updateOneToMany=t,this},t.prototype.updateManyToMany=function(t){return this._updateManyToMany=t,this},t.prototype.remove=function(t){return this._remove=t,this},t.prototype.execute=function(){this[this._diffModeMultiple?"_executeMultiple":"_executeOneToOne"]()},t.prototype._executeOneToOne=function(){var t=this._old,e=this._new,n={},r=new Array(t.length),i=new Array(e.length);this._initIndexMap(t,null,r,"_oldKeyGetter"),this._initIndexMap(e,n,i,"_newKeyGetter");for(var o=0;o<t.length;o++){var a=r[o],s=n[a],l=pM(s);if(l>1){var u=s.shift();1===s.length&&(n[a]=s[0]),this._update&&this._update(u,o)}else 1===l?(n[a]=null,this._update&&this._update(s,o)):this._remove&&this._remove(o)}this._performRestAdd(i,n)},t.prototype._executeMultiple=function(){var t=this._old,e=this._new,n={},r={},i=[],o=[];this._initIndexMap(t,n,i,"_oldKeyGetter"),this._initIndexMap(e,r,o,"_newKeyGetter");for(var a=0;a<i.length;a++){var s=i[a],l=n[s],u=r[s],c=pM(l),d=pM(u);if(c>1&&1===d)this._updateManyToOne&&this._updateManyToOne(u,l),r[s]=null;else if(1===c&&d>1)this._updateOneToMany&&this._updateOneToMany(u,l),r[s]=null;else if(1===c&&1===d)this._update&&this._update(u,l),r[s]=null;else if(c>1&&d>1)this._updateManyToMany&&this._updateManyToMany(u,l),r[s]=null;else if(c>1)for(var h=0;h<c;h++)this._remove&&this._remove(l[h]);else this._remove&&this._remove(l)}this._performRestAdd(o,r)},t.prototype._performRestAdd=function(t,e){for(var n=0;n<t.length;n++){var r=t[n],i=e[r],o=pM(i);if(o>1)for(var a=0;a<o;a++)this._add&&this._add(i[a]);else 1===o&&this._add&&this._add(i);e[r]=null}},t.prototype._initIndexMap=function(t,e,n,r){for(var i=this._diffModeMultiple,o=0;o<t.length;o++){var a="_ec_"+this[r](t[o],o);if(i||(n[o]=a),e){var s=e[a],l=pM(s);0===l?(e[a]=o,i&&n.push(a)):1===l?e[a]=[s,o]:s.push(o)}}},t}(),yM=gM,vM=function(){function t(t,e){this._encode=t,this._schema=e}return t.prototype.get=function(){return{fullDimensions:this._getFullDimensionNames(),encode:this._encode}},t.prototype._getFullDimensionNames=function(){return this._cachedDimNames||(this._cachedDimNames=this._schema?this._schema.makeOutputDimensionNames():[]),this._cachedDimNames},t}();function _M(t,e){var n={},r=n.encode={},i=Object(s["f"])(),o=[],a=[],l={};Object(s["k"])(t.dimensions,(function(e){var n=t.getDimensionInfo(e),s=n.coordDim;if(s){0;var u=n.coordDimIndex;bM(r,s)[u]=e,n.isExtraCoord||(i.set(s,1),wM(n.type)&&(o[0]=e),bM(l,s)[u]=t.getDimensionIndex(n.name)),n.defaultTooltip&&a.push(e)}Qc.each((function(t,e){var i=bM(r,e),o=n.otherDims[e];null!=o&&!1!==o&&(i[o]=n.name)}))}));var u=[],c={};i.each((function(t,e){var n=r[e];c[e]=n[0],u=u.concat(n)})),n.dataDimsOnCoord=u,n.dataDimIndicesOnCoord=Object(s["H"])(u,(function(e){return t.getDimensionInfo(e).storeDimIndex})),n.encodeFirstDimNotExtra=c;var d=r.label;d&&d.length&&(o=d.slice());var h=r.tooltip;return h&&h.length?a=h.slice():a.length||(a=o.slice()),r.defaultedLabel=o,r.defaultedTooltip=a,n.userOutput=new vM(l,e),n}function bM(t,e){return t.hasOwnProperty(e)||(t[e]=[]),t[e]}function xM(t){return"category"===t?"ordinal":"time"===t?"time":"float"}function wM(t){return!("ordinal"===t||"time"===t)}var MM=function(){function t(t){this.otherDims={},null!=t&&s["m"](this,t)}return t}(),SM=MM,kM=Ci(),LM={float:"f",int:"i",ordinal:"o",number:"n",time:"t"},TM=function(){function t(t){this.dimensions=t.dimensions,this._dimOmitted=t.dimensionOmitted,this.source=t.source,this._fullDimCount=t.fullDimensionCount,this._updateDimOmitted(t.dimensionOmitted)}return t.prototype.isDimensionOmitted=function(){return this._dimOmitted},t.prototype._updateDimOmitted=function(t){this._dimOmitted=t,t&&(this._dimNameMap||(this._dimNameMap=IM(this.source)))},t.prototype.getSourceDimensionIndex=function(t){return Object(s["P"])(this._dimNameMap.get(t),-1)},t.prototype.getSourceDimension=function(t){var e=this.source.dimensionsDefine;if(e)return e[t]},t.prototype.makeStoreSchema=function(){for(var t=this._fullDimCount,e=Fh(this.source),n=!CM(t),r="",i=[],o=0,a=0;o<t;o++){var s=void 0,l=void 0,u=void 0,c=this.dimensions[a];if(c&&c.storeDimIndex===o)s=e?c.name:null,l=c.type,u=c.ordinalMeta,a++;else{var d=this.getSourceDimension(o);d&&(s=e?d.name:null,l=d.type)}i.push({property:s,type:l,ordinalMeta:u}),!e||null==s||c&&c.isCalculationCoord||(r+=n?s.replace(/\`/g,"`1").replace(/\$/g,"`2"):s),r+="$",r+=LM[l]||"f",u&&(r+=u.uid),r+="$"}var h=this.source,f=[h.seriesLayoutBy,h.startIndex,r].join("$$");return{dimensions:i,hash:f}},t.prototype.makeOutputDimensionNames=function(){for(var t=[],e=0,n=0;e<this._fullDimCount;e++){var r=void 0,i=this.dimensions[n];if(i&&i.storeDimIndex===e)i.isCalculationCoord||(r=i.name),n++;else{var o=this.getSourceDimension(e);o&&(r=o.name)}t.push(r)}return t},t.prototype.appendCalculationDimension=function(t){this.dimensions.push(t),t.isCalculationCoord=!0,this._fullDimCount++,this._updateDimOmitted(!0)},t}();function DM(t){return t instanceof TM}function OM(t){for(var e=Object(s["f"])(),n=0;n<(t||[]).length;n++){var r=t[n],i=Object(s["A"])(r)?r.name:r;null!=i&&null==e.get(i)&&e.set(i,n)}return e}function IM(t){var e=kM(t);return e.dimNameMap||(e.dimNameMap=OM(t.dimensionsDefine))}function CM(t){return t>30}var AM,YM,jM,PM,EM,NM,RM,HM=s["A"],zM=s["H"],FM="undefined"===typeof Int32Array?Array:Int32Array,BM="e\0\0",WM=-1,VM=["hasItemOption","_nameList","_idList","_invertedIndicesMap","_dimSummary","userOutput","_rawData","_dimValueGetter","_nameDimIdx","_idDimIdx","_nameRepeatCount"],GM=["_approximateExtent"],$M=function(){function t(t,e){var n;this.type="list",this._dimOmitted=!1,this._nameList=[],this._idList=[],this._visual={},this._layout={},this._itemVisuals=[],this._itemLayouts=[],this._graphicEls=[],this._approximateExtent={},this._calculationInfo={},this.hasItemOption=!1,this.TRANSFERABLE_METHODS=["cloneShallow","downSample","lttbDownSample","map"],this.CHANGABLE_METHODS=["filterSelf","selectRange"],this.DOWNSAMPLE_METHODS=["downSample","lttbDownSample"];var r=!1;DM(t)?(n=t.dimensions,this._dimOmitted=t.isDimensionOmitted(),this._schema=t):(r=!0,n=t),n=n||["x","y"];for(var i={},o=[],a={},l=!1,u={},c=0;c<n.length;c++){var d=n[c],h=s["C"](d)?new SM({name:d}):d instanceof SM?d:new SM(d),f=h.name;h.type=h.type||"float",h.coordDim||(h.coordDim=f,h.coordDimIndex=0);var p=h.otherDims=h.otherDims||{};o.push(f),i[f]=h,null!=u[f]&&(l=!0),h.createInvertedIndices&&(a[f]=[]),0===p.itemName&&(this._nameDimIdx=c),0===p.itemId&&(this._idDimIdx=c),r&&(h.storeDimIndex=c)}if(this.dimensions=o,this._dimInfos=i,this._initGetDimensionInfo(l),this.hostModel=e,this._invertedIndicesMap=a,this._dimOmitted){var m=this._dimIdxToName=s["f"]();s["k"](o,(function(t){m.set(i[t].storeDimIndex,t)}))}}return t.prototype.getDimension=function(t){var e=this._recognizeDimIndex(t);if(null==e)return t;if(e=t,!this._dimOmitted)return this.dimensions[e];var n=this._dimIdxToName.get(e);if(null!=n)return n;var r=this._schema.getSourceDimension(e);return r?r.name:void 0},t.prototype.getDimensionIndex=function(t){var e=this._recognizeDimIndex(t);if(null!=e)return e;if(null==t)return-1;var n=this._getDimInfo(t);return n?n.storeDimIndex:this._dimOmitted?this._schema.getSourceDimensionIndex(t):-1},t.prototype._recognizeDimIndex=function(t){if(s["z"](t)||null!=t&&!isNaN(t)&&!this._getDimInfo(t)&&(!this._dimOmitted||this._schema.getSourceDimensionIndex(t)<0))return+t},t.prototype._getStoreDimIndex=function(t){var e=this.getDimensionIndex(t);return e},t.prototype.getDimensionInfo=function(t){return this._getDimInfo(this.getDimension(t))},t.prototype._initGetDimensionInfo=function(t){var e=this._dimInfos;this._getDimInfo=t?function(t){return e.hasOwnProperty(t)?e[t]:void 0}:function(t){return e[t]}},t.prototype.getDimensionsOnCoord=function(){return this._dimSummary.dataDimsOnCoord.slice()},t.prototype.mapDimension=function(t,e){var n=this._dimSummary;if(null==e)return n.encodeFirstDimNotExtra[t];var r=n.encode[t];return r?r[e]:null},t.prototype.mapDimensionsAll=function(t){var e=this._dimSummary,n=e.encode[t];return(n||[]).slice()},t.prototype.getStore=function(){return this._store},t.prototype.initData=function(t,e,n){var r,i=this;if(t instanceof Rf&&(r=t),!r){var o=this.dimensions,a=Ah(t)||s["u"](t)?new Bh(t,o.length):t;r=new Rf;var l=zM(o,(function(t){return{type:i._dimInfos[t].type,property:t}}));r.initData(a,l,n)}this._store=r,this._nameList=(e||[]).slice(),this._idList=[],this._nameRepeatCount={},this._doInit(0,r.count()),this._dimSummary=_M(this,this._schema),this.userOutput=this._dimSummary.userOutput},t.prototype.appendData=function(t){var e=this._store.appendData(t);this._doInit(e[0],e[1])},t.prototype.appendValues=function(t,e){var n=this._store.appendValues(t,e.length),r=n.start,i=n.end,o=this._shouldMakeIdFromName();if(this._updateOrdinalMeta(),e)for(var a=r;a<i;a++){var s=a-r;this._nameList[a]=e[s],o&&RM(this,a)}},t.prototype._updateOrdinalMeta=function(){for(var t=this._store,e=this.dimensions,n=0;n<e.length;n++){var r=this._dimInfos[e[n]];r.ordinalMeta&&t.collectOrdinalMeta(r.storeDimIndex,r.ordinalMeta)}},t.prototype._shouldMakeIdFromName=function(){var t=this._store.getProvider();return null==this._idDimIdx&&t.getSource().sourceFormat!==id&&!t.fillStorage},t.prototype._doInit=function(t,e){if(!(t>=e)){var n=this._store,r=n.getProvider();this._updateOrdinalMeta();var i=this._nameList,o=this._idList,a=r.getSource().sourceFormat,s=a===td;if(s&&!r.pure)for(var l=[],u=t;u<e;u++){var c=r.getItem(u,l);if(!this.hasItemOption&&fi(c)&&(this.hasItemOption=!0),c){var d=c.name;null==i[u]&&null!=d&&(i[u]=Mi(d,null));var h=c.id;null==o[u]&&null!=h&&(o[u]=Mi(h,null))}}if(this._shouldMakeIdFromName())for(u=t;u<e;u++)RM(this,u);AM(this)}},t.prototype.getApproximateExtent=function(t){return this._approximateExtent[t]||this._store.getDataExtent(this._getStoreDimIndex(t))},t.prototype.setApproximateExtent=function(t,e){e=this.getDimension(e),this._approximateExtent[e]=t.slice()},t.prototype.getCalculationInfo=function(t){return this._calculationInfo[t]},t.prototype.setCalculationInfo=function(t,e){HM(t)?s["m"](this._calculationInfo,t):this._calculationInfo[t]=e},t.prototype.getName=function(t){var e=this.getRawIndex(t),n=this._nameList[e];return null==n&&null!=this._nameDimIdx&&(n=jM(this,this._nameDimIdx,e)),null==n&&(n=""),n},t.prototype._getCategory=function(t,e){var n=this._store.get(t,e),r=this._store.getOrdinalMeta(t);return r?r.categories[n]:n},t.prototype.getId=function(t){return YM(this,this.getRawIndex(t))},t.prototype.count=function(){return this._store.count()},t.prototype.get=function(t,e){var n=this._store,r=this._dimInfos[t];if(r)return n.get(r.storeDimIndex,e)},t.prototype.getByRawIndex=function(t,e){var n=this._store,r=this._dimInfos[t];if(r)return n.getByRawIndex(r.storeDimIndex,e)},t.prototype.getIndices=function(){return this._store.getIndices()},t.prototype.getDataExtent=function(t){return this._store.getDataExtent(this._getStoreDimIndex(t))},t.prototype.getSum=function(t){return this._store.getSum(this._getStoreDimIndex(t))},t.prototype.getMedian=function(t){return this._store.getMedian(this._getStoreDimIndex(t))},t.prototype.getValues=function(t,e){var n=this,r=this._store;return s["t"](t)?r.getValues(zM(t,(function(t){return n._getStoreDimIndex(t)})),e):r.getValues(t)},t.prototype.hasValue=function(t){for(var e=this._dimSummary.dataDimIndicesOnCoord,n=0,r=e.length;n<r;n++)if(isNaN(this._store.get(e[n],t)))return!1;return!0},t.prototype.indexOfName=function(t){for(var e=0,n=this._store.count();e<n;e++)if(this.getName(e)===t)return e;return-1},t.prototype.getRawIndex=function(t){return this._store.getRawIndex(t)},t.prototype.indexOfRawIndex=function(t){return this._store.indexOfRawIndex(t)},t.prototype.rawIndexOf=function(t,e){var n=t&&this._invertedIndicesMap[t];var r=n[e];return null==r||isNaN(r)?WM:r},t.prototype.indicesOfNearest=function(t,e,n){return this._store.indicesOfNearest(this._getStoreDimIndex(t),e,n)},t.prototype.each=function(t,e,n){s["w"](t)&&(n=e,e=t,t=[]);var r=n||this,i=zM(PM(t),this._getStoreDimIndex,this);this._store.each(i,r?s["c"](e,r):e)},t.prototype.filterSelf=function(t,e,n){s["w"](t)&&(n=e,e=t,t=[]);var r=n||this,i=zM(PM(t),this._getStoreDimIndex,this);return this._store=this._store.filter(i,r?s["c"](e,r):e),this},t.prototype.selectRange=function(t){var e=this,n={},r=s["F"](t),i=[];return s["k"](r,(function(r){var o=e._getStoreDimIndex(r);n[o]=t[r],i.push(o)})),this._store=this._store.selectRange(n),this},t.prototype.mapArray=function(t,e,n){s["w"](t)&&(n=e,e=t,t=[]),n=n||this;var r=[];return this.each(t,(function(){r.push(e&&e.apply(this,arguments))}),n),r},t.prototype.map=function(t,e,n,r){var i=n||r||this,o=zM(PM(t),this._getStoreDimIndex,this),a=NM(this);return a._store=this._store.map(o,i?s["c"](e,i):e),a},t.prototype.modify=function(t,e,n,r){var i=n||r||this;var o=zM(PM(t),this._getStoreDimIndex,this);this._store.modify(o,i?s["c"](e,i):e)},t.prototype.downSample=function(t,e,n,r){var i=NM(this);return i._store=this._store.downSample(this._getStoreDimIndex(t),e,n,r),i},t.prototype.lttbDownSample=function(t,e){var n=NM(this);return n._store=this._store.lttbDownSample(this._getStoreDimIndex(t),e),n},t.prototype.getRawDataItem=function(t){return this._store.getRawDataItem(t)},t.prototype.getItemModel=function(t){var e=this.hostModel,n=this.getRawDataItem(t);return new Lu(n,e,e&&e.ecModel)},t.prototype.diff=function(t){var e=this;return new yM(t?t.getStore().getIndices():[],this.getStore().getIndices(),(function(e){return YM(t,e)}),(function(t){return YM(e,t)}))},t.prototype.getVisual=function(t){var e=this._visual;return e&&e[t]},t.prototype.setVisual=function(t,e){this._visual=this._visual||{},HM(t)?s["m"](this._visual,t):this._visual[t]=e},t.prototype.getItemVisual=function(t,e){var n=this._itemVisuals[t],r=n&&n[e];return null==r?this.getVisual(e):r},t.prototype.hasItemVisual=function(){return this._itemVisuals.length>0},t.prototype.ensureUniqueItemVisual=function(t,e){var n=this._itemVisuals,r=n[t];r||(r=n[t]={});var i=r[e];return null==i&&(i=this.getVisual(e),s["t"](i)?i=i.slice():HM(i)&&(i=s["m"]({},i)),r[e]=i),i},t.prototype.setItemVisual=function(t,e,n){var r=this._itemVisuals[t]||{};this._itemVisuals[t]=r,HM(e)?s["m"](r,e):r[e]=n},t.prototype.clearAllVisual=function(){this._visual={},this._itemVisuals=[]},t.prototype.setLayout=function(t,e){HM(t)?s["m"](this._layout,t):this._layout[t]=e},t.prototype.getLayout=function(t){return this._layout[t]},t.prototype.getItemLayout=function(t){return this._itemLayouts[t]},t.prototype.setItemLayout=function(t,e,n){this._itemLayouts[t]=n?s["m"](this._itemLayouts[t]||{},e):e},t.prototype.clearItemLayouts=function(){this._itemLayouts.length=0},t.prototype.setItemGraphicEl=function(t,e){var n=this.hostModel&&this.hostModel.seriesIndex;Ts(n,this.dataType,t,e),this._graphicEls[t]=e},t.prototype.getItemGraphicEl=function(t){return this._graphicEls[t]},t.prototype.eachItemGraphicEl=function(t,e){s["k"](this._graphicEls,(function(n,r){n&&t&&t.call(e,n,r)}))},t.prototype.cloneShallow=function(e){return e||(e=new t(this._schema?this._schema:zM(this.dimensions,this._getDimInfo,this),this.hostModel)),EM(e,this),e._store=this._store,e},t.prototype.wrapMethod=function(t,e){var n=this[t];s["w"](n)&&(this.__wrappedMethods=this.__wrappedMethods||[],this.__wrappedMethods.push(t),this[t]=function(){var t=n.apply(this,arguments);return e.apply(this,[t].concat(s["S"](arguments)))})},t.internalField=function(){AM=function(t){var e=t._invertedIndicesMap;s["k"](e,(function(n,r){var i=t._dimInfos[r],o=i.ordinalMeta,a=t._store;if(o){n=e[r]=new FM(o.categories.length);for(var s=0;s<n.length;s++)n[s]=WM;for(s=0;s<a.count();s++)n[a.get(i.storeDimIndex,s)]=s}}))},jM=function(t,e,n){return Mi(t._getCategory(e,n),null)},YM=function(t,e){var n=t._idList[e];return null==n&&null!=t._idDimIdx&&(n=jM(t,t._idDimIdx,e)),null==n&&(n=BM+e),n},PM=function(t){return s["t"](t)||(t=null!=t?[t]:[]),t},NM=function(e){var n=new t(e._schema?e._schema:zM(e.dimensions,e._getDimInfo,e),e.hostModel);return EM(n,e),n},EM=function(t,e){s["k"](VM.concat(e.__wrappedMethods||[]),(function(n){e.hasOwnProperty(n)&&(t[n]=e[n])})),t.__wrappedMethods=e.__wrappedMethods,s["k"](GM,(function(n){t[n]=s["d"](e[n])})),t._calculationInfo=s["m"]({},e._calculationInfo)},RM=function(t,e){var n=t._nameList,r=t._idList,i=t._nameDimIdx,o=t._idDimIdx,a=n[e],s=r[e];if(null==a&&null!=i&&(n[e]=a=jM(t,i,e)),null==s&&null!=o&&(r[e]=s=jM(t,o,e)),null==s&&null!=a){var l=t._nameRepeatCount,u=l[a]=(l[a]||0)+1;s=a,u>1&&(s+="__ec__"+u),r[e]=s}}}(),t}(),UM=$M;function qM(t,e){Ah(t)||(t=jh(t)),e=e||{};var n=e.coordDimensions||[],r=e.dimensionsDefine||t.dimensionsDefine||[],i=Object(s["f"])(),o=[],a=ZM(t,n,r,e.dimensionsCount),l=e.canOmitUnusedDimensions&&CM(a),u=r===t.dimensionsDefine,c=u?IM(t):OM(r),d=e.encodeDefine;!d&&e.encodeDefaulter&&(d=e.encodeDefaulter(t,a));for(var h=Object(s["f"])(d),f=new If(a),p=0;p<f.length;p++)f[p]=-1;function m(t){var e=f[t];if(e<0){var n=r[t],i=Object(s["A"])(n)?n:{name:n},a=new SM,l=i.name;null!=l&&null!=c.get(l)&&(a.name=a.displayName=l),null!=i.type&&(a.type=i.type),null!=i.displayName&&(a.displayName=i.displayName);var u=o.length;return f[t]=u,a.storeDimIndex=t,o.push(a),a}return o[e]}if(!l)for(p=0;p<a;p++)m(p);h.each((function(t,e){var n=ui(t).slice();if(1===n.length&&!Object(s["C"])(n[0])&&n[0]<0)h.set(e,!1);else{var r=h.set(e,[]);Object(s["k"])(n,(function(t,n){var i=Object(s["C"])(t)?c.get(t):t;null!=i&&i<a&&(r[n]=i,y(m(i),e,n))}))}}));var g=0;function y(t,e,n){null!=Qc.get(e)?t.otherDims[e]=n:(t.coordDim=e,t.coordDimIndex=n,i.set(e,!0))}Object(s["k"])(n,(function(t){var e,n,r,i;if(Object(s["C"])(t))e=t,i={};else{i=t,e=i.name;var o=i.ordinalMeta;i.ordinalMeta=null,i=Object(s["m"])({},i),i.ordinalMeta=o,n=i.dimsDef,r=i.otherDims,i.name=i.coordDim=i.coordDimIndex=i.dimsDef=i.otherDims=null}var l=h.get(e);if(!1!==l){if(l=ui(l),!l.length)for(var c=0;c<(n&&n.length||1);c++){while(g<a&&null!=m(g).coordDim)g++;g<a&&l.push(g++)}Object(s["k"])(l,(function(t,o){var a=m(t);if(u&&null!=i.type&&(a.type=i.type),y(Object(s["i"])(a,i),e,o),null==a.name&&n){var l=n[o];!Object(s["A"])(l)&&(l={name:l}),a.name=a.displayName=l.name,a.defaultTooltip=l.defaultTooltip}r&&Object(s["i"])(a.otherDims,r)}))}}));var v=e.generateCoord,_=e.generateCoordCount,b=null!=_;_=v?_||1:0;var x=v||"value";function w(t){null==t.name&&(t.name=t.coordDim)}if(l)Object(s["k"])(o,(function(t){w(t)})),o.sort((function(t,e){return t.storeDimIndex-e.storeDimIndex}));else for(var M=0;M<a;M++){var S=m(M),k=S.coordDim;null==k&&(S.coordDim=JM(x,i,b),S.coordDimIndex=0,(!v||_<=0)&&(S.isExtraCoord=!0),_--),w(S),null!=S.type||md(t,M)!==ld.Must&&(!S.isExtraCoord||null==S.otherDims.itemName&&null==S.otherDims.seriesName)||(S.type="ordinal")}return XM(o),new TM({source:t,dimensions:o,fullDimensionCount:a,dimensionOmitted:l})}function XM(t){for(var e=Object(s["f"])(),n=0;n<t.length;n++){var r=t[n],i=r.name,o=e.get(i)||0;o>0&&(r.name=i+(o-1)),o++,e.set(i,o)}}function ZM(t,e,n,r){var i=Math.max(t.dimensionsDetectedCount||1,e.length,n.length,r||0);return Object(s["k"])(e,(function(t){var e;Object(s["A"])(t)&&(e=t.dimsDef)&&(i=Math.max(i,e.length))})),i}function JM(t,e,n){if(n||e.hasKey(t)){var r=0;while(e.hasKey(t+r))r++;t+=r}return e.set(t,!0),t}var KM=function(){function t(t){this.coordSysDims=[],this.axisMap=Object(s["f"])(),this.categoryAxisMap=Object(s["f"])(),this.coordSysName=t}return t}();function QM(t){var e=t.get("coordinateSystem"),n=new KM(e),r=tS[e];if(r)return r(t,n,n.axisMap,n.categoryAxisMap),n}var tS={cartesian2d:function(t,e,n,r){var i=t.getReferringComponents("xAxis",Pi).models[0],o=t.getReferringComponents("yAxis",Pi).models[0];e.coordSysDims=["x","y"],n.set("x",i),n.set("y",o),eS(i)&&(r.set("x",i),e.firstCategoryDimIndex=0),eS(o)&&(r.set("y",o),null==e.firstCategoryDimIndex&&(e.firstCategoryDimIndex=1))},singleAxis:function(t,e,n,r){var i=t.getReferringComponents("singleAxis",Pi).models[0];e.coordSysDims=["single"],n.set("single",i),eS(i)&&(r.set("single",i),e.firstCategoryDimIndex=0)},polar:function(t,e,n,r){var i=t.getReferringComponents("polar",Pi).models[0],o=i.findAxisModel("radiusAxis"),a=i.findAxisModel("angleAxis");e.coordSysDims=["radius","angle"],n.set("radius",o),n.set("angle",a),eS(o)&&(r.set("radius",o),e.firstCategoryDimIndex=0),eS(a)&&(r.set("angle",a),null==e.firstCategoryDimIndex&&(e.firstCategoryDimIndex=1))},geo:function(t,e,n,r){e.coordSysDims=["lng","lat"]},parallel:function(t,e,n,r){var i=t.ecModel,o=i.getComponent("parallel",t.get("parallelIndex")),a=e.coordSysDims=o.dimensions.slice();Object(s["k"])(o.parallelAxisIndex,(function(t,o){var s=i.getComponent("parallelAxis",t),l=a[o];n.set(l,s),eS(s)&&(r.set(l,s),null==e.firstCategoryDimIndex&&(e.firstCategoryDimIndex=o))}))}};function eS(t){return"category"===t.get("type")}function nS(t,e,n){n=n||{};var r,i,o,a=n.byIndex,l=n.stackedCoordDimension;rS(e)?r=e:(i=e.schema,r=i.dimensions,o=e.store);var u,c,d,h,f=!(!t||!t.get("stack"));if(Object(s["k"])(r,(function(t,e){Object(s["C"])(t)&&(r[e]=t={name:t}),f&&!t.isExtraCoord&&(a||u||!t.ordinalMeta||(u=t),c||"ordinal"===t.type||"time"===t.type||l&&l!==t.coordDim||(c=t))})),!c||a||u||(a=!0),c){d="__\0ecstackresult_"+t.id,h="__\0ecstackedover_"+t.id,u&&(u.createInvertedIndices=!0);var p=c.coordDim,m=c.type,g=0;Object(s["k"])(r,(function(t){t.coordDim===p&&g++}));var y={name:d,coordDim:p,coordDimIndex:g,type:m,isExtraCoord:!0,isCalculationCoord:!0,storeDimIndex:r.length},v={name:h,coordDim:h,coordDimIndex:g+1,type:m,isExtraCoord:!0,isCalculationCoord:!0,storeDimIndex:r.length+1};i?(o&&(y.storeDimIndex=o.ensureCalculationDimension(h,m),v.storeDimIndex=o.ensureCalculationDimension(d,m)),i.appendCalculationDimension(y),i.appendCalculationDimension(v)):(r.push(y),r.push(v))}return{stackedDimension:c&&c.name,stackedByDimension:u&&u.name,isStackedByIndex:a,stackedOverDimension:h,stackResultDimension:d}}function rS(t){return!DM(t.schema)}function iS(t,e){return!!e&&e===t.getCalculationInfo("stackedDimension")}function oS(t,e){return iS(t,e)?t.getCalculationInfo("stackResultDimension"):e}function aS(t,e){var n,r=t.get("coordinateSystem"),i=$d.get(r);return e&&e.coordSysDims&&(n=s["H"](e.coordSysDims,(function(t){var n={name:t},r=e.axisMap.get(t);if(r){var i=r.get("type");n.type=xM(i)}return n}))),n||(n=i&&(i.getDimensionsInfo?i.getDimensionsInfo():i.dimensions.slice())||["x","y"]),n}function sS(t,e,n){var r,i;return n&&s["k"](t,(function(t,o){var a=t.coordDim,s=n.categoryAxisMap.get(a);s&&(null==r&&(r=o),t.ordinalMeta=s.getOrdinalMeta(),e&&(t.createInvertedIndices=!0)),null!=t.otherDims.itemName&&(i=!0)})),i||null==r||(t[r].otherDims.itemName=0),r}function lS(t,e,n){n=n||{};var r,i=e.getSourceManager(),o=!1;t?(o=!0,r=jh(t)):(r=i.getSource(),o=r.sourceFormat===td);var a=QM(e),l=aS(e,a),u=n.useEncodeDefaulter,c=s["w"](u)?u:u?s["h"](dd,l,e):null,d={coordDimensions:l,generateCoord:n.generateCoord,encodeDefine:e.getEncode(),encodeDefaulter:c,canOmitUnusedDimensions:!o},h=qM(r,d),f=sS(h.dimensions,n.createInvertedIndices,a),p=o?null:i.getSharedDataStore(h),m=nS(e,{schema:h,store:p}),g=new UM(h,e);g.setCalculationInfo(m);var y=null!=f&&uS(r)?function(t,e,n,r){return r===f?n:this.defaultDimValueGetter(t,e,n,r)}:null;return g.hasItemOption=!1,g.initData(o?r:p,null,y),g}function uS(t){if(t.sourceFormat===td){var e=cS(t.data||[]);return!s["t"](hi(e))}}function cS(t){var e=0;while(e<t.length&&null==t[e])e++;return t[e]}var dS=lS,hS=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.hasSymbolVisual=!0,n}return o(e,t),e.prototype.getInitialData=function(t){return dS(null,this,{useEncodeDefaulter:!0})},e.prototype.getLegendIcon=function(t){var e=new wr,n=ov("line",0,t.itemHeight/2,t.itemWidth,0,t.lineStyle.stroke,!1);e.add(n),n.setStyle(t.lineStyle);var r=this.getData().getVisual("symbol"),i=this.getData().getVisual("symbolRotate"),o="none"===r?"circle":r,a=.8*t.itemHeight,s=ov(o,(t.itemWidth-a)/2,(t.itemHeight-a)/2,a,a,t.itemStyle.fill);e.add(s),s.setStyle(t.itemStyle);var l="inherit"===t.iconRotate?i:t.iconRotate||0;return s.rotation=l*Math.PI/180,s.setOrigin([t.itemWidth/2,t.itemHeight/2]),o.indexOf("empty")>-1&&(s.style.stroke=s.style.fill,s.style.fill="#fff",s.style.lineWidth=2),e},e.type="series.line",e.dependencies=["grid","polar"],e.defaultOption={z:3,coordinateSystem:"cartesian2d",legendHoverLink:!0,clip:!0,label:{position:"top"},endLabel:{show:!1,valueAnimation:!0,distance:8},lineStyle:{width:2,type:"solid"},emphasis:{scale:!0},step:!1,smooth:!1,smoothMonotone:null,symbol:"emptyCircle",symbolSize:4,symbolRotate:null,showSymbol:!0,showAllSymbol:"auto",connectNulls:!1,sampling:"none",animationEasing:"linear",progressive:0,hoverLayerThreshold:1/0,universalTransition:{divideShape:"clone"},triggerLineEvent:!1},e}(Mp),fS=hS;function pS(t,e){var n=t.mapDimensionsAll("defaultedLabel"),r=n.length;if(1===r){var i=Qh(t,e,n[0]);return null!=i?i+"":null}if(r){for(var o=[],a=0;a<n.length;a++)o.push(Qh(t,e,n[a]));return o.join(" ")}}function mS(t,e){var n=t.mapDimensionsAll("defaultedLabel");if(!Object(s["t"])(e))return e+"";for(var r=[],i=0;i<n.length;i++){var o=t.getDimensionIndex(n[i]);o>=0&&r.push(e[o])}return r.join(" ")}var gS=function(t){function e(e,n,r,i){var o=t.call(this)||this;return o.updateData(e,n,r,i),o}return o(e,t),e.prototype._createSymbol=function(t,e,n,r,i){this.removeAll();var o=ov(t,-1,-1,2,2,null,i);o.attr({z2:100,culling:!0,scaleX:r[0]/2,scaleY:r[1]/2}),o.drift=yS,this._symbolType=t,this.add(o)},e.prototype.stopSymbolAnimation=function(t){this.childAt(0).stopAnimation(null,t)},e.prototype.getSymbolType=function(){return this._symbolType},e.prototype.getSymbolPath=function(){return this.childAt(0)},e.prototype.highlight=function(){ul(this.childAt(0))},e.prototype.downplay=function(){cl(this.childAt(0))},e.prototype.setZ=function(t,e){var n=this.childAt(0);n.zlevel=t,n.z=e},e.prototype.setDraggable=function(t,e){var n=this.childAt(0);n.draggable=t,n.cursor=!e&&t?"move":n.cursor},e.prototype.updateData=function(t,n,r,i){this.silent=!1;var o=t.getItemVisual(n,"symbol")||"circle",a=t.hostModel,s=e.getSymbolSize(t,n),l=o!==this._symbolType,u=i&&i.disableAnimation;if(l){var c=t.getItemVisual(n,"symbolKeepAspect");this._createSymbol(o,t,n,s,c)}else{var d=this.childAt(0);d.silent=!1;var h={scaleX:s[0]/2,scaleY:s[1]/2};u?d.attr(h):Wl(d,h,a,n),Xl(d)}if(this._updateCommon(t,n,s,r,i),l){d=this.childAt(0);if(!u){h={scaleX:this._sizeX,scaleY:this._sizeY,style:{opacity:d.style.opacity}};d.scaleX=d.scaleY=0,d.style.opacity=0,Vl(d,h,a,n)}}u&&this.childAt(0).stopAnimation("leave")},e.prototype._updateCommon=function(t,e,n,r,i){var o,a,l,u,c,d,h,f,p,m=this.childAt(0),g=t.hostModel;if(r&&(o=r.emphasisItemStyle,a=r.blurItemStyle,l=r.selectItemStyle,u=r.focus,c=r.blurScope,h=r.labelStatesModels,f=r.hoverScale,p=r.cursorStyle,d=r.emphasisDisabled),!r||t.hasItemOption){var y=r&&r.itemModel?r.itemModel:t.getItemModel(e),v=y.getModel("emphasis");o=v.getModel("itemStyle").getItemStyle(),l=y.getModel(["select","itemStyle"]).getItemStyle(),a=y.getModel(["blur","itemStyle"]).getItemStyle(),u=v.get("focus"),c=v.get("blurScope"),d=v.get("disabled"),h=eu(y),f=v.getShallow("scale"),p=y.getShallow("cursor")}var _=t.getItemVisual(e,"symbolRotate");m.attr("rotation",(_||0)*Math.PI/180||0);var b=sv(t.getItemVisual(e,"symbolOffset"),n);b&&(m.x=b[0],m.y=b[1]),p&&m.attr("cursor",p);var x=t.getItemVisual(e,"style"),w=x.fill;if(m instanceof Ka){var M=m.style;m.useStyle(Object(s["m"])({image:M.image,x:M.x,y:M.y,width:M.width,height:M.height},x))}else m.__isEmptyBrush?m.useStyle(Object(s["m"])({},x)):m.useStyle(x),m.style.decal=null,m.setColor(w,i&&i.symbolInnerColor),m.style.strokeNoScale=!0;var S=t.getItemVisual(e,"liftZ"),k=this._z2;null!=S?null==k&&(this._z2=m.z2,m.z2+=S):null!=k&&(m.z2=k,this._z2=null);var L=i&&i.useNameLabel;function T(e){return L?t.getName(e):pS(t,e)}tu(m,h,{labelFetcher:g,labelDataIndex:e,defaultText:T,inheritColor:w,defaultOpacity:x.opacity}),this._sizeX=n[0]/2,this._sizeY=n[1]/2;var D=m.ensureState("emphasis");D.style=o,m.ensureState("select").style=l,m.ensureState("blur").style=a;var O=null==f||!0===f?Math.max(1.1,3/this._sizeY):isFinite(f)&&f>0?+f:1;D.scaleX=this._sizeX*O,D.scaleY=this._sizeY*O,this.setSymbolScale(1),Dl(this,u,c,d)},e.prototype.setSymbolScale=function(t){this.scaleX=this.scaleY=t},e.prototype.fadeOut=function(t,e,n){var r=this.childAt(0),i=Ls(this).dataIndex,o=n&&n.animation;if(this.silent=r.silent=!0,n&&n.fadeLabel){var a=r.getTextContent();a&&$l(a,{style:{opacity:0}},e,{dataIndex:i,removeOpt:o,cb:function(){r.removeTextContent()}})}else r.removeTextContent();$l(r,{style:{opacity:0},scaleX:0,scaleY:0},e,{dataIndex:i,cb:t,removeOpt:o})},e.getSymbolSize=function(t,e){return av(t.getItemVisual(e,"symbolSize"))},e}(wr);function yS(t,e){this.parent.drift(t,e)}var vS=gS;function _S(t,e,n,r){return e&&!isNaN(e[0])&&!isNaN(e[1])&&!(r.isIgnore&&r.isIgnore(n))&&!(r.clipShape&&!r.clipShape.contain(e[0],e[1]))&&"none"!==t.getItemVisual(n,"symbol")}function bS(t){return null==t||Object(s["A"])(t)||(t={isIgnore:t}),t||{}}function xS(t){var e=t.hostModel,n=e.getModel("emphasis");return{emphasisItemStyle:n.getModel("itemStyle").getItemStyle(),blurItemStyle:e.getModel(["blur","itemStyle"]).getItemStyle(),selectItemStyle:e.getModel(["select","itemStyle"]).getItemStyle(),focus:n.get("focus"),blurScope:n.get("blurScope"),emphasisDisabled:n.get("disabled"),hoverScale:n.get("scale"),labelStatesModels:eu(e),cursorStyle:e.get("cursor")}}var wS=function(){function t(t){this.group=new wr,this._SymbolCtor=t||vS}return t.prototype.updateData=function(t,e){this._progressiveEls=null,e=bS(e);var n=this.group,r=t.hostModel,i=this._data,o=this._SymbolCtor,a=e.disableAnimation,s=xS(t),l={disableAnimation:a},u=e.getSymbolPoint||function(e){return t.getItemLayout(e)};i||n.removeAll(),t.diff(i).add((function(r){var i=u(r);if(_S(t,i,r,e)){var a=new o(t,r,s,l);a.setPosition(i),t.setItemGraphicEl(r,a),n.add(a)}})).update((function(c,d){var h=i.getItemGraphicEl(d),f=u(c);if(_S(t,f,c,e)){var p=t.getItemVisual(c,"symbol")||"circle",m=h&&h.getSymbolType&&h.getSymbolType();if(!h||m&&m!==p)n.remove(h),h=new o(t,c,s,l),h.setPosition(f);else{h.updateData(t,c,s,l);var g={x:f[0],y:f[1]};a?h.attr(g):Wl(h,g,r)}n.add(h),t.setItemGraphicEl(c,h)}else n.remove(h)})).remove((function(t){var e=i.getItemGraphicEl(t);e&&e.fadeOut((function(){n.remove(e)}),r)})).execute(),this._getSymbolPoint=u,this._data=t},t.prototype.updateLayout=function(){var t=this,e=this._data;e&&e.eachItemGraphicEl((function(e,n){var r=t._getSymbolPoint(n);e.setPosition(r),e.markRedraw()}))},t.prototype.incrementalPrepareUpdate=function(t){this._seriesScope=xS(t),this._data=null,this.group.removeAll()},t.prototype.incrementalUpdate=function(t,e,n){function r(t){t.isGroup||(t.incremental=!0,t.ensureState("emphasis").hoverLayer=!0)}this._progressiveEls=[],n=bS(n);for(var i=t.start;i<t.end;i++){var o=e.getItemLayout(i);if(_S(e,o,i,n)){var a=new this._SymbolCtor(e,i,this._seriesScope);a.traverse(r),a.setPosition(o),this.group.add(a),e.setItemGraphicEl(i,a),this._progressiveEls.push(a)}}},t.prototype.eachRendered=function(t){Ng(this._progressiveEls||this.group,t)},t.prototype.remove=function(t){var e=this.group,n=this._data;n&&t?n.eachItemGraphicEl((function(t){t.fadeOut((function(){e.remove(t)}),n.hostModel)})):e.removeAll()},t}(),MS=wS;function SS(t,e,n){var r=t.getBaseAxis(),i=t.getOtherAxis(r),o=kS(i,n),a=r.dim,l=i.dim,u=e.mapDimension(l),c=e.mapDimension(a),d="x"===l||"radius"===l?1:0,h=Object(s["H"])(t.dimensions,(function(t){return e.mapDimension(t)})),f=!1,p=e.getCalculationInfo("stackResultDimension");return iS(e,h[0])&&(f=!0,h[0]=p),iS(e,h[1])&&(f=!0,h[1]=p),{dataDimsForPoint:h,valueStart:o,valueAxisDim:l,baseAxisDim:a,stacked:!!f,valueDim:u,baseDim:c,baseDataOffset:d,stackedOverDimension:e.getCalculationInfo("stackedOverDimension")}}function kS(t,e){var n=0,r=t.scale.getExtent();return"start"===e?n=r[0]:"end"===e?n=r[1]:Object(s["z"])(e)&&!isNaN(e)?n=e:r[0]>0?n=r[0]:r[1]<0&&(n=r[1]),n}function LS(t,e,n,r){var i=NaN;t.stacked&&(i=n.get(n.getCalculationInfo("stackedOverDimension"),r)),isNaN(i)&&(i=t.valueStart);var o=t.baseDataOffset,a=[];return a[o]=n.get(t.baseDim,r),a[1-o]=i,e.dataToPoint(a)}var TS="undefined"!==typeof Float32Array,DS=TS?Float32Array:Array;function OS(t){return Object(s["t"])(t)?TS?new Float32Array(t):t:new DS(t)}function IS(t,e){var n=[];return e.diff(t).add((function(t){n.push({cmd:"+",idx:t})})).update((function(t,e){n.push({cmd:"=",idx:e,idx1:t})})).remove((function(t){n.push({cmd:"-",idx:t})})).execute(),n}function CS(t,e,n,r,i,o,a,s){for(var l=IS(t,e),u=[],c=[],d=[],h=[],f=[],p=[],m=[],g=SS(i,e,a),y=t.getLayout("points")||[],v=e.getLayout("points")||[],_=0;_<l.length;_++){var b=l[_],x=!0,w=void 0,M=void 0;switch(b.cmd){case"=":w=2*b.idx,M=2*b.idx1;var S=y[w],k=y[w+1],L=v[M],T=v[M+1];(isNaN(S)||isNaN(k))&&(S=L,k=T),u.push(S,k),c.push(L,T),d.push(n[w],n[w+1]),h.push(r[M],r[M+1]),m.push(e.getRawIndex(b.idx1));break;case"+":var D=b.idx,O=g.dataDimsForPoint,I=i.dataToPoint([e.get(O[0],D),e.get(O[1],D)]);M=2*D,u.push(I[0],I[1]),c.push(v[M],v[M+1]);var C=LS(g,i,e,D);d.push(C[0],C[1]),h.push(r[M],r[M+1]),m.push(e.getRawIndex(D));break;case"-":x=!1}x&&(f.push(b),p.push(p.length))}p.sort((function(t,e){return m[t]-m[e]}));var A=u.length,Y=OS(A),j=OS(A),P=OS(A),E=OS(A),N=[];for(_=0;_<p.length;_++){var R=p[_],H=2*_,z=2*R;Y[H]=u[z],Y[H+1]=u[z+1],j[H]=c[z],j[H+1]=c[z+1],P[H]=d[z],P[H+1]=d[z+1],E[H]=h[z],E[H+1]=h[z+1],N[_]=f[R]}return{current:Y,next:j,stackedOnCurrent:P,stackedOnNext:E,status:N}}var AS=Math.min,YS=Math.max;function jS(t,e){return isNaN(t)||isNaN(e)}function PS(t,e,n,r,i,o,a,s,l){for(var u,c,d,h,f,p,m=n,g=0;g<r;g++){var y=e[2*m],v=e[2*m+1];if(m>=i||m<0)break;if(jS(y,v)){if(l){m+=o;continue}break}if(m===n)t[o>0?"moveTo":"lineTo"](y,v),d=y,h=v;else{var _=y-u,b=v-c;if(_*_+b*b<.5){m+=o;continue}if(a>0){var x=m+o,w=e[2*x],M=e[2*x+1];while(w===y&&M===v&&g<r)g++,x+=o,m+=o,w=e[2*x],M=e[2*x+1],y=e[2*m],v=e[2*m+1],_=y-u,b=v-c;var S=g+1;if(l)while(jS(w,M)&&S<r)S++,x+=o,w=e[2*x],M=e[2*x+1];var k=.5,L=0,T=0,D=void 0,O=void 0;if(S>=r||jS(w,M))f=y,p=v;else{L=w-u,T=M-c;var I=y-u,C=w-y,A=v-c,Y=M-v,j=void 0,P=void 0;if("x"===s){j=Math.abs(I),P=Math.abs(C);var E=L>0?1:-1;f=y-E*j*a,p=v,D=y+E*P*a,O=v}else if("y"===s){j=Math.abs(A),P=Math.abs(Y);var N=T>0?1:-1;f=y,p=v-N*j*a,D=y,O=v+N*P*a}else j=Math.sqrt(I*I+A*A),P=Math.sqrt(C*C+Y*Y),k=P/(P+j),f=y-L*a*(1-k),p=v-T*a*(1-k),D=y+L*a*k,O=v+T*a*k,D=AS(D,YS(w,y)),O=AS(O,YS(M,v)),D=YS(D,AS(w,y)),O=YS(O,AS(M,v)),L=D-y,T=O-v,f=y-L*j/P,p=v-T*j/P,f=AS(f,YS(u,y)),p=AS(p,YS(c,v)),f=YS(f,AS(u,y)),p=YS(p,AS(c,v)),L=y-f,T=v-p,D=y+L*P/j,O=v+T*P/j}t.bezierCurveTo(d,h,f,p,y,v),d=D,h=O}else t.lineTo(y,v)}u=y,c=v,m+=o}return g}var ES=function(){function t(){this.smooth=0,this.smoothConstraint=!0}return t}(),NS=function(t){function e(e){var n=t.call(this,e)||this;return n.type="ec-polyline",n}return o(e,t),e.prototype.getDefaultStyle=function(){return{stroke:"#000",fill:null}},e.prototype.getDefaultShape=function(){return new ES},e.prototype.buildPath=function(t,e){var n=e.points,r=0,i=n.length/2;if(e.connectNulls){for(;i>0;i--)if(!jS(n[2*i-2],n[2*i-1]))break;for(;r<i;r++)if(!jS(n[2*r],n[2*r+1]))break}while(r<i)r+=PS(t,n,r,i,i,1,e.smooth,e.smoothMonotone,e.connectNulls)+1},e.prototype.getPointOn=function(t,e){this.path||(this.createPathProxy(),this.buildPath(this.path,this.shape));for(var n,r,i=this.path,o=i.data,a=va.CMD,s="x"===e,l=[],u=0;u<o.length;){var c=o[u++],d=void 0,h=void 0,f=void 0,p=void 0,m=void 0,g=void 0,y=void 0;switch(c){case a.M:n=o[u++],r=o[u++];break;case a.L:if(d=o[u++],h=o[u++],y=s?(t-n)/(d-n):(t-r)/(h-r),y<=1&&y>=0){var v=s?(h-r)*y+r:(d-n)*y+n;return s?[t,v]:[v,t]}n=d,r=h;break;case a.C:d=o[u++],h=o[u++],f=o[u++],p=o[u++],m=o[u++],g=o[u++];var _=s?xe(n,d,f,m,t,l):xe(r,h,p,g,t,l);if(_>0)for(var b=0;b<_;b++){var x=l[b];if(x<=1&&x>=0){v=s?_e(r,h,p,g,x):_e(n,d,f,m,x);return s?[t,v]:[v,t]}}n=m,r=g;break}}},e}(Va),RS=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e}(ES),HS=function(t){function e(e){var n=t.call(this,e)||this;return n.type="ec-polygon",n}return o(e,t),e.prototype.getDefaultShape=function(){return new RS},e.prototype.buildPath=function(t,e){var n=e.points,r=e.stackedOnPoints,i=0,o=n.length/2,a=e.smoothMonotone;if(e.connectNulls){for(;o>0;o--)if(!jS(n[2*o-2],n[2*o-1]))break;for(;i<o;i++)if(!jS(n[2*i],n[2*i+1]))break}while(i<o){var s=PS(t,n,i,o,o,1,e.smooth,a,e.connectNulls);PS(t,r,i+s-1,s,o,-1,e.stackedOnSmooth,a,e.connectNulls),i+=s+1,t.closePath()}},e}(Va);function zS(t,e,n,r,i){var o=t.getArea(),a=o.x,l=o.y,u=o.width,c=o.height,d=n.get(["lineStyle","width"])||2;a-=d/2,l-=d/2,u+=d,c+=d,u=Math.ceil(u),a!==Math.floor(a)&&(a=Math.floor(a),u++);var h=new ss({shape:{x:a,y:l,width:u,height:c}});if(e){var f=t.getBaseAxis(),p=f.isHorizontal(),m=f.inverse;p?(m&&(h.shape.x+=u),h.shape.width=0):(m||(h.shape.y+=c),h.shape.height=0);var g=Object(s["w"])(i)?function(t){i(t,h)}:null;Vl(h,{shape:{width:u,height:c,x:a,y:l}},n,null,r,g)}return h}function FS(t,e,n){var r=t.getArea(),i=Rr(r.r0,1),o=Rr(r.r,1),a=new _m({shape:{cx:Rr(t.cx,1),cy:Rr(t.cy,1),r0:i,r:o,startAngle:r.startAngle,endAngle:r.endAngle,clockwise:r.clockwise}});if(e){var s="angle"===t.getBaseAxis().dim;s?a.shape.endAngle=r.startAngle:a.shape.r=i,Vl(a,{shape:{endAngle:r.endAngle,r:o}},n)}return a}function BS(t,e,n,r,i){return t?"polar"===t.type?FS(t,e,n):"cartesian2d"===t.type?zS(t,e,n,r,i):null:null}function WS(t,e){return t.type===e}function VS(t,e){if(t.length===e.length){for(var n=0;n<t.length;n++)if(t[n]!==e[n])return;return!0}}function GS(t){for(var e=1/0,n=1/0,r=-1/0,i=-1/0,o=0;o<t.length;){var a=t[o++],s=t[o++];isNaN(a)||(e=Math.min(a,e),r=Math.max(a,r)),isNaN(s)||(n=Math.min(s,n),i=Math.max(s,i))}return[[e,n],[r,i]]}function $S(t,e){var n=GS(t),r=n[0],i=n[1],o=GS(e),a=o[0],s=o[1];return Math.max(Math.abs(r[0]-a[0]),Math.abs(r[1]-a[1]),Math.abs(i[0]-s[0]),Math.abs(i[1]-s[1]))}function US(t){return s["z"](t)?t:t?.5:0}function qS(t,e,n){if(!n.valueDim)return[];for(var r=e.count(),i=OS(2*r),o=0;o<r;o++){var a=LS(n,t,e,o);i[2*o]=a[0],i[2*o+1]=a[1]}return i}function XS(t,e,n,r){var i=e.getBaseAxis(),o="x"===i.dim||"radius"===i.dim?0:1,a=[],s=0,l=[],u=[],c=[],d=[];if(r){for(s=0;s<t.length;s+=2)isNaN(t[s])||isNaN(t[s+1])||d.push(t[s],t[s+1]);t=d}for(s=0;s<t.length-2;s+=2)switch(c[0]=t[s+2],c[1]=t[s+3],u[0]=t[s],u[1]=t[s+1],a.push(u[0],u[1]),n){case"end":l[o]=c[o],l[1-o]=u[1-o],a.push(l[0],l[1]);break;case"middle":var h=(u[o]+c[o])/2,f=[];l[o]=f[o]=h,l[1-o]=u[1-o],f[1-o]=c[1-o],a.push(l[0],l[1]),a.push(f[0],f[1]);break;default:l[o]=u[o],l[1-o]=c[1-o],a.push(l[0],l[1])}return a.push(t[s++],t[s++]),a}function ZS(t,e){var n,r,i=[],o=t.length;function a(t,e,n){var r=t.coord,i=(n-r)/(e.coord-r),o=Object(Ne["b"])(i,[t.color,e.color]);return{coord:n,color:o}}for(var s=0;s<o;s++){var l=t[s],u=l.coord;if(u<0)n=l;else{if(u>e){r?i.push(a(r,l,e)):n&&i.push(a(n,l,0),a(n,l,e));break}n&&(i.push(a(n,l,0)),n=null),i.push(l),r=l}}return i}function JS(t,e,n){var r=t.getVisual("visualMeta");if(r&&r.length&&t.count()&&"cartesian2d"===e.type){for(var i,o,a=r.length-1;a>=0;a--){var l=t.getDimensionInfo(r[a].dimension);if(i=l&&l.coordDim,"x"===i||"y"===i){o=r[a];break}}if(o){var u=e.getAxis(i),c=s["H"](o.stops,(function(t){return{coord:u.toGlobalCoord(u.dataToCoord(t.value)),color:t.color}})),d=c.length,h=o.outerColors.slice();d&&c[0].coord>c[d-1].coord&&(c.reverse(),h.reverse());var f=ZS(c,"x"===i?n.getWidth():n.getHeight()),p=f.length;if(!p&&d)return c[0].coord<0?h[1]?h[1]:c[d-1].color:h[0]?h[0]:c[0].color;var m=10,g=f[0].coord-m,y=f[p-1].coord+m,v=y-g;if(v<.001)return"transparent";s["k"](f,(function(t){t.offset=(t.coord-g)/v})),f.push({offset:p?f[p-1].offset:.5,color:h[1]||"transparent"}),f.unshift({offset:p?f[0].offset:.5,color:h[0]||"transparent"});var _=new qm(0,0,0,0,f,!0);return _[i]=g,_[i+"2"]=y,_}}}function KS(t,e,n){var r=t.get("showAllSymbol"),i="auto"===r;if(!r||i){var o=n.getAxesByScale("ordinal")[0];if(o&&(!i||!QS(o,e))){var a=e.mapDimension(o.dim),l={};return s["k"](o.getViewLabels(),(function(t){var e=o.scale.getRawOrdinalNumber(t.tickValue);l[e]=1})),function(t){return!l.hasOwnProperty(e.get(a,t))}}}}function QS(t,e){var n=t.getExtent(),r=Math.abs(n[1]-n[0])/t.scale.count();isNaN(r)&&(r=0);for(var i=e.count(),o=Math.max(1,Math.round(i/5)),a=0;a<i;a+=o)if(1.5*vS.getSymbolSize(e,a)[t.isHorizontal()?1:0]>r)return!1;return!0}function tk(t,e){return isNaN(t)||isNaN(e)}function ek(t){for(var e=t.length/2;e>0;e--)if(!tk(t[2*e-2],t[2*e-1]))break;return e-1}function nk(t,e){return[t[2*e],t[2*e+1]]}function rk(t,e,n){for(var r,i,o=t.length/2,a="x"===n?0:1,s=0,l=-1,u=0;u<o;u++)if(i=t[2*u+a],!isNaN(i)&&!isNaN(t[2*u+1-a]))if(0!==u){if(r<=e&&i>=e||r>=e&&i<=e){l=u;break}s=u,r=i}else r=i;return{range:[s,l],t:(e-r)/(i-r)}}function ik(t){if(t.get(["endLabel","show"]))return!0;for(var e=0;e<Ps.length;e++)if(t.get([Ps[e],"endLabel","show"]))return!0;return!1}function ok(t,e,n,r){if(WS(e,"cartesian2d")){var i=r.getModel("endLabel"),o=i.get("valueAnimation"),a=r.getData(),s={lastFrameIndex:0},l=ik(r)?function(n,r){t._endLabelOnDuring(n,r,a,s,o,i,e)}:null,u=e.getBaseAxis().isHorizontal(),c=zS(e,n,r,(function(){var e=t._endLabel;e&&n&&null!=s.originalX&&e.attr({x:s.originalX,y:s.originalY})}),l);if(!r.get("clip",!0)){var d=c.shape,h=Math.max(d.width,d.height);u?(d.y-=h,d.height+=2*h):(d.x-=h,d.width+=2*h)}return l&&l(1,c),c}return FS(e,n,r)}function ak(t,e){var n=e.getBaseAxis(),r=n.isHorizontal(),i=n.inverse,o=r?i?"right":"left":"center",a=r?"middle":i?"top":"bottom";return{normal:{align:t.get("align")||o,verticalAlign:t.get("verticalAlign")||a}}}var sk=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.init=function(){var t=new wr,e=new MS;this.group.add(e.group),this._symbolDraw=e,this._lineGroup=t},e.prototype.render=function(t,e,n){var r=this,i=t.coordinateSystem,o=this.group,a=t.getData(),l=t.getModel("lineStyle"),u=t.getModel("areaStyle"),c=a.getLayout("points")||[],d="polar"===i.type,h=this._coordSys,f=this._symbolDraw,p=this._polyline,m=this._polygon,g=this._lineGroup,y=!e.ssr&&t.get("animation"),v=!u.isEmpty(),_=u.get("origin"),b=SS(i,a,_),x=v&&qS(i,a,b),w=t.get("showSymbol"),M=t.get("connectNulls"),S=w&&!d&&KS(t,a,i),k=this._data;k&&k.eachItemGraphicEl((function(t,e){t.__temp&&(o.remove(t),k.setItemGraphicEl(e,null))})),w||f.remove(),o.add(g);var L,T=!d&&t.get("step");i&&i.getArea&&t.get("clip",!0)&&(L=i.getArea(),null!=L.width?(L.x-=.1,L.y-=.1,L.width+=.2,L.height+=.2):L.r0&&(L.r0-=.5,L.r+=.5)),this._clipShapeForSymbol=L;var D=JS(a,i,n)||a.getVisual("style")[a.getVisual("drawType")];if(p&&h.type===i.type&&T===this._step){v&&!m?m=this._newPolygon(c,x):m&&!v&&(g.remove(m),m=this._polygon=null),d||this._initOrUpdateEndLabel(t,i,Cc(D));var O=g.getClipPath();if(O){var I=ok(this,i,!1,t);Vl(O,{shape:I.shape},t)}else g.setClipPath(ok(this,i,!0,t));w&&f.updateData(a,{isIgnore:S,clipShape:L,disableAnimation:!0,getSymbolPoint:function(t){return[c[2*t],c[2*t+1]]}}),VS(this._stackedOnPoints,x)&&VS(this._points,c)||(y?this._doUpdateAnimation(a,x,i,n,T,_,M):(T&&(c=XS(c,i,T,M),x&&(x=XS(x,i,T,M))),p.setShape({points:c}),m&&m.setShape({points:c,stackedOnPoints:x})))}else w&&f.updateData(a,{isIgnore:S,clipShape:L,disableAnimation:!0,getSymbolPoint:function(t){return[c[2*t],c[2*t+1]]}}),y&&this._initSymbolLabelAnimation(a,i,L),T&&(c=XS(c,i,T,M),x&&(x=XS(x,i,T,M))),p=this._newPolyline(c),v?m=this._newPolygon(c,x):m&&(g.remove(m),m=this._polygon=null),d||this._initOrUpdateEndLabel(t,i,Cc(D)),g.setClipPath(ok(this,i,!0,t));var C=t.getModel("emphasis"),A=C.get("focus"),Y=C.get("blurScope"),j=C.get("disabled");if(p.useStyle(s["i"](l.getLineStyle(),{fill:"none",stroke:D,lineJoin:"bevel"})),Al(p,t,"lineStyle"),p.style.lineWidth>0&&"bolder"===t.get(["emphasis","lineStyle","width"])){var P=p.getState("emphasis").style;P.lineWidth=+p.style.lineWidth+1}Ls(p).seriesIndex=t.seriesIndex,Dl(p,A,Y,j);var E=US(t.get("smooth")),N=t.get("smoothMonotone");if(p.setShape({smooth:E,smoothMonotone:N,connectNulls:M}),m){var R=a.getCalculationInfo("stackedOnSeries"),H=0;m.useStyle(s["i"](u.getAreaStyle(),{fill:D,opacity:.7,lineJoin:"bevel",decal:a.getVisual("style").decal})),R&&(H=US(R.get("smooth"))),m.setShape({smooth:E,stackedOnSmooth:H,smoothMonotone:N,connectNulls:M}),Al(m,t,"areaStyle"),Ls(m).seriesIndex=t.seriesIndex,Dl(m,A,Y,j)}var z=function(t){r._changePolyState(t)};a.eachItemGraphicEl((function(t){t&&(t.onHoverStateChange=z)})),this._polyline.onHoverStateChange=z,this._data=a,this._coordSys=i,this._stackedOnPoints=x,this._points=c,this._step=T,this._valueOrigin=_,t.get("triggerLineEvent")&&(this.packEventData(t,p),m&&this.packEventData(t,m))},e.prototype.packEventData=function(t,e){Ls(e).eventData={componentType:"series",componentSubType:"line",componentIndex:t.componentIndex,seriesIndex:t.seriesIndex,seriesName:t.name,seriesType:"line"}},e.prototype.highlight=function(t,e,n,r){var i=t.getData(),o=Ii(i,r);if(this._changePolyState("emphasis"),!(o instanceof Array)&&null!=o&&o>=0){var a=i.getLayout("points"),s=i.getItemGraphicEl(o);if(!s){var l=a[2*o],u=a[2*o+1];if(isNaN(l)||isNaN(u))return;if(this._clipShapeForSymbol&&!this._clipShapeForSymbol.contain(l,u))return;var c=t.get("zlevel")||0,d=t.get("z")||0;s=new vS(i,o),s.x=l,s.y=u,s.setZ(c,d);var h=s.getSymbolPath().getTextContent();h&&(h.zlevel=c,h.z=d,h.z2=this._polyline.z2+1),s.__temp=!0,i.setItemGraphicEl(o,s),s.stopSymbolAnimation(!0),this.group.add(s)}s.highlight()}else $g.prototype.highlight.call(this,t,e,n,r)},e.prototype.downplay=function(t,e,n,r){var i=t.getData(),o=Ii(i,r);if(this._changePolyState("normal"),null!=o&&o>=0){var a=i.getItemGraphicEl(o);a&&(a.__temp?(i.setItemGraphicEl(o,null),this.group.remove(a)):a.downplay())}else $g.prototype.downplay.call(this,t,e,n,r)},e.prototype._changePolyState=function(t){var e=this._polygon;tl(this._polyline,t),e&&tl(e,t)},e.prototype._newPolyline=function(t){var e=this._polyline;return e&&this._lineGroup.remove(e),e=new NS({shape:{points:t},segmentIgnoreThreshold:2,z2:10}),this._lineGroup.add(e),this._polyline=e,e},e.prototype._newPolygon=function(t,e){var n=this._polygon;return n&&this._lineGroup.remove(n),n=new HS({shape:{points:t,stackedOnPoints:e},segmentIgnoreThreshold:2}),this._lineGroup.add(n),this._polygon=n,n},e.prototype._initSymbolLabelAnimation=function(t,e,n){var r,i,o=e.getBaseAxis(),a=o.inverse;"cartesian2d"===e.type?(r=o.isHorizontal(),i=!1):"polar"===e.type&&(r="angle"===o.dim,i=!0);var l=t.hostModel,u=l.get("animationDuration");s["w"](u)&&(u=u(null));var c=l.get("animationDelay")||0,d=s["w"](c)?c(null):c;t.eachItemGraphicEl((function(t,o){var l=t;if(l){var h=[t.x,t.y],f=void 0,p=void 0,m=void 0;if(n)if(i){var g=n,y=e.pointToCoord(h);r?(f=g.startAngle,p=g.endAngle,m=-y[1]/180*Math.PI):(f=g.r0,p=g.r,m=y[0])}else{var v=n;r?(f=v.x,p=v.x+v.width,m=t.x):(f=v.y+v.height,p=v.y,m=t.y)}var _=p===f?0:(m-f)/(p-f);a&&(_=1-_);var b=s["w"](c)?c(o):u*_+d,x=l.getSymbolPath(),w=x.getTextContent();l.attr({scaleX:0,scaleY:0}),l.animateTo({scaleX:1,scaleY:1},{duration:200,setToFinal:!0,delay:b}),w&&w.animateFrom({style:{opacity:0}},{duration:300,delay:b}),x.disableLabelAnimation=!0}}))},e.prototype._initOrUpdateEndLabel=function(t,e,n){var r=t.getModel("endLabel");if(ik(t)){var i=t.getData(),o=this._polyline,a=i.getLayout("points");if(!a)return o.removeTextContent(),void(this._endLabel=null);var s=this._endLabel;s||(s=this._endLabel=new ks({z2:200}),s.ignoreClip=!0,o.setTextContent(this._endLabel),o.disableLabelAnimation=!0);var l=ek(a);l>=0&&(tu(o,eu(t,"endLabel"),{inheritColor:n,labelFetcher:t,labelDataIndex:l,defaultText:function(t,e,n){return null!=n?mS(i,n):pS(i,t)},enableTextSetter:!0},ak(r,e)),o.textConfig.position=null)}else this._endLabel&&(this._polyline.removeTextContent(),this._endLabel=null)},e.prototype._endLabelOnDuring=function(t,e,n,r,i,o,a){var s=this._endLabel,l=this._polyline;if(s){t<1&&null==r.originalX&&(r.originalX=s.x,r.originalY=s.y);var u=n.getLayout("points"),c=n.hostModel,d=c.get("connectNulls"),h=o.get("precision"),f=o.get("distance")||0,p=a.getBaseAxis(),m=p.isHorizontal(),g=p.inverse,y=e.shape,v=g?m?y.x:y.y+y.height:m?y.x+y.width:y.y,_=(m?f:0)*(g?-1:1),b=(m?0:-f)*(g?-1:1),x=m?"x":"y",w=rk(u,v,x),M=w.range,S=M[1]-M[0],k=void 0;if(S>=1){if(S>1&&!d){var L=nk(u,M[0]);s.attr({x:L[0]+_,y:L[1]+b}),i&&(k=c.getRawValue(M[0]))}else{L=l.getPointOn(v,x);L&&s.attr({x:L[0]+_,y:L[1]+b});var T=c.getRawValue(M[0]),D=c.getRawValue(M[1]);i&&(k=Bi(n,h,T,D,w.t))}r.lastFrameIndex=M[0]}else{var O=1===t||r.lastFrameIndex>0?M[0]:0;L=nk(u,O);i&&(k=c.getRawValue(O)),s.attr({x:L[0]+_,y:L[1]+b})}if(i){var I=du(s);"function"===typeof I.setLabelText&&I.setLabelText(k)}}},e.prototype._doUpdateAnimation=function(t,e,n,r,i,o,a){var s=this._polyline,l=this._polygon,u=t.hostModel,c=CS(this._data,t,this._stackedOnPoints,e,this._coordSys,n,this._valueOrigin,o),d=c.current,h=c.stackedOnCurrent,f=c.next,p=c.stackedOnNext;if(i&&(d=XS(c.current,n,i,a),h=XS(c.stackedOnCurrent,n,i,a),f=XS(c.next,n,i,a),p=XS(c.stackedOnNext,n,i,a)),$S(d,f)>3e3||l&&$S(h,p)>3e3)return s.stopAnimation(),s.setShape({points:f}),void(l&&(l.stopAnimation(),l.setShape({points:f,stackedOnPoints:p})));s.shape.__points=c.current,s.shape.points=d;var m={shape:{points:f}};c.current!==d&&(m.shape.__points=c.next),s.stopAnimation(),Wl(s,m,u),l&&(l.setShape({points:d,stackedOnPoints:h}),l.stopAnimation(),Wl(l,{shape:{stackedOnPoints:p}},u),s.shape.points!==l.shape.points&&(l.shape.points=s.shape.points));for(var g=[],y=c.status,v=0;v<y.length;v++){var _=y[v].cmd;if("="===_){var b=t.getItemGraphicEl(y[v].idx1);b&&g.push({el:b,ptIdx:v})}}s.animators&&s.animators.length&&s.animators[0].during((function(){l&&l.dirtyShape();for(var t=s.shape.__points,e=0;e<g.length;e++){var n=g[e].el,r=2*g[e].ptIdx;n.x=t[r],n.y=t[r+1],n.markRedraw()}}))},e.prototype.remove=function(t){var e=this.group,n=this._data;this._lineGroup.removeAll(),this._symbolDraw.remove(!0),n&&n.eachItemGraphicEl((function(t,r){t.__temp&&(e.remove(t),n.setItemGraphicEl(r,null))})),this._polyline=this._polygon=this._coordSys=this._points=this._stackedOnPoints=this._endLabel=this._data=null},e.type="line",e}($g),lk=sk;function uk(t,e){return{seriesType:t,plan:Lp(),reset:function(t){var n=t.getData(),r=t.coordinateSystem,i=t.pipelineContext,o=e||i.large;if(r){var a=Object(s["H"])(r.dimensions,(function(t){return n.mapDimension(t)})).slice(0,2),l=a.length,u=n.getCalculationInfo("stackResultDimension");iS(n,a[0])&&(a[0]=u),iS(n,a[1])&&(a[1]=u);var c=n.getStore(),d=n.getDimensionIndex(a[0]),h=n.getDimensionIndex(a[1]);return l&&{progress:function(t,e){for(var n=t.end-t.start,i=o&&OS(n*l),a=[],s=[],u=t.start,f=0;u<t.end;u++){var p=void 0;if(1===l){var m=c.get(d,u);p=r.dataToPoint(m,null,s)}else a[0]=c.get(d,u),a[1]=c.get(h,u),p=r.dataToPoint(a,null,s);o?(i[f++]=p[0],i[f++]=p[1]):e.setItemLayout(u,p.slice())}o&&e.setLayout("points",i)}}}}}}var ck={average:function(t){for(var e=0,n=0,r=0;r<t.length;r++)isNaN(t[r])||(e+=t[r],n++);return 0===n?NaN:e/n},sum:function(t){for(var e=0,n=0;n<t.length;n++)e+=t[n]||0;return e},max:function(t){for(var e=-1/0,n=0;n<t.length;n++)t[n]>e&&(e=t[n]);return isFinite(e)?e:NaN},min:function(t){for(var e=1/0,n=0;n<t.length;n++)t[n]<e&&(e=t[n]);return isFinite(e)?e:NaN},minmax:function(t){for(var e=-1/0,n=-1/0,r=0;r<t.length;r++){var i=t[r],o=Math.abs(i);o>e&&(e=o,n=i)}return isFinite(n)?n:NaN},nearest:function(t){return t[0]}},dk=function(t){return Math.round(t.length/2)};function hk(t){return{seriesType:t,reset:function(t,e,n){var r=t.getData(),i=t.get("sampling"),o=t.coordinateSystem,a=r.count();if(a>10&&"cartesian2d"===o.type&&i){var l=o.getBaseAxis(),u=o.getOtherAxis(l),c=l.getExtent(),d=n.getDevicePixelRatio(),h=Math.abs(c[1]-c[0])*(d||1),f=Math.round(a/h);if(isFinite(f)&&f>1){"lttb"===i&&t.setData(r.lttbDownSample(r.mapDimension(u.dim),1/f));var p=void 0;Object(s["C"])(i)?p=ck[i]:Object(s["w"])(i)&&(p=i),p&&t.setData(r.downSample(r.mapDimension(u.dim),1/f,p,dk))}}}}}function fk(t){t.registerChartView(lk),t.registerSeriesModel(fS),t.registerLayout(uk("line",!0)),t.registerVisual({seriesType:"line",reset:function(t){var e=t.getData(),n=t.getModel("lineStyle").getLineStyle();n&&!n.stroke&&(n.stroke=e.getVisual("style").fill),e.setVisual("legendLineStyle",n)}}),t.registerProcessor(t.PRIORITY.PROCESSOR.STATISTIC,hk("line"))}var pk="__ec_stack_";function mk(t){return t.get("stack")||pk+t.seriesIndex}function gk(t){return t.dim+t.index}function yk(t){var e=[],n=t.axis,r="axis0";if("category"===n.type){for(var i=n.getBandWidth(),o=0;o<t.count;o++)e.push(Object(s["i"])({bandWidth:i,axisKey:r,stackId:pk+o},t));var a=xk(e),l=[];for(o=0;o<t.count;o++){var u=a[r][pk+o];u.offsetCenter=u.offset+u.width/2,l.push(u)}return l}}function vk(t,e){var n=[];return e.eachSeriesByType(t,(function(t){kk(t)&&n.push(t)})),n}function _k(t){var e={};Object(s["k"])(t,(function(t){var n=t.coordinateSystem,r=n.getBaseAxis();if("time"===r.type||"value"===r.type)for(var i=t.getData(),o=r.dim+"_"+r.index,a=i.getDimensionIndex(i.mapDimension(r.dim)),s=i.getStore(),l=0,u=s.count();l<u;++l){var c=s.get(a,l);e[o]?e[o].push(c):e[o]=[c]}}));var n={};for(var r in e)if(e.hasOwnProperty(r)){var i=e[r];if(i){i.sort((function(t,e){return t-e}));for(var o=null,a=1;a<i.length;++a){var l=i[a]-i[a-1];l>0&&(o=null===o?l:Math.min(o,l))}n[r]=o}}return n}function bk(t){var e=_k(t),n=[];return Object(s["k"])(t,(function(t){var r,i=t.coordinateSystem,o=i.getBaseAxis(),a=o.getExtent();if("category"===o.type)r=o.getBandWidth();else if("value"===o.type||"time"===o.type){var s=o.dim+"_"+o.index,l=e[s],u=Math.abs(a[1]-a[0]),c=o.scale.getExtent(),d=Math.abs(c[1]-c[0]);r=l?u/d*l:u}else{var h=t.getData();r=Math.abs(a[1]-a[0])/h.count()}var f=Nr(t.get("barWidth"),r),p=Nr(t.get("barMaxWidth"),r),m=Nr(t.get("barMinWidth")||(Lk(t)?.5:1),r),g=t.get("barGap"),y=t.get("barCategoryGap");n.push({bandWidth:r,barWidth:f,barMaxWidth:p,barMinWidth:m,barGap:g,barCategoryGap:y,axisKey:gk(o),stackId:mk(t)})})),xk(n)}function xk(t){var e={};Object(s["k"])(t,(function(t,n){var r=t.axisKey,i=t.bandWidth,o=e[r]||{bandWidth:i,remainedWidth:i,autoWidthCount:0,categoryGap:null,gap:"20%",stacks:{}},a=o.stacks;e[r]=o;var s=t.stackId;a[s]||o.autoWidthCount++,a[s]=a[s]||{width:0,maxWidth:0};var l=t.barWidth;l&&!a[s].width&&(a[s].width=l,l=Math.min(o.remainedWidth,l),o.remainedWidth-=l);var u=t.barMaxWidth;u&&(a[s].maxWidth=u);var c=t.barMinWidth;c&&(a[s].minWidth=c);var d=t.barGap;null!=d&&(o.gap=d);var h=t.barCategoryGap;null!=h&&(o.categoryGap=h)}));var n={};return Object(s["k"])(e,(function(t,e){n[e]={};var r=t.stacks,i=t.bandWidth,o=t.categoryGap;if(null==o){var a=Object(s["F"])(r).length;o=Math.max(35-4*a,15)+"%"}var l=Nr(o,i),u=Nr(t.gap,1),c=t.remainedWidth,d=t.autoWidthCount,h=(c-l)/(d+(d-1)*u);h=Math.max(h,0),Object(s["k"])(r,(function(t){var e=t.maxWidth,n=t.minWidth;if(t.width){r=t.width;e&&(r=Math.min(r,e)),n&&(r=Math.max(r,n)),t.width=r,c-=r+u*r,d--}else{var r=h;e&&e<r&&(r=Math.min(e,c)),n&&n>r&&(r=n),r!==h&&(t.width=r,c-=r+u*r,d--)}})),h=(c-l)/(d+(d-1)*u),h=Math.max(h,0);var f,p=0;Object(s["k"])(r,(function(t,e){t.width||(t.width=h),f=t,p+=t.width*(1+u)})),f&&(p-=f.width*u);var m=-p/2;Object(s["k"])(r,(function(t,r){n[e][r]=n[e][r]||{bandWidth:i,offset:m,width:t.width},m+=t.width*(1+u)}))})),n}function wk(t,e,n){if(t&&e){var r=t[gk(e)];return null!=r&&null!=n?r[mk(n)]:r}}function Mk(t,e){var n=vk(t,e),r=bk(n);Object(s["k"])(n,(function(t){var e=t.getData(),n=t.coordinateSystem,i=n.getBaseAxis(),o=mk(t),a=r[gk(i)][o],s=a.offset,l=a.width;e.setLayout({bandWidth:a.bandWidth,offset:s,size:l})}))}function Sk(t){return{seriesType:t,plan:Lp(),reset:function(t){if(kk(t)){var e=t.getData(),n=t.coordinateSystem,r=n.getBaseAxis(),i=n.getOtherAxis(r),o=e.getDimensionIndex(e.mapDimension(i.dim)),a=e.getDimensionIndex(e.mapDimension(r.dim)),s=t.get("showBackground",!0),l=e.mapDimension(i.dim),u=e.getCalculationInfo("stackResultDimension"),c=iS(e,l)&&!!e.getCalculationInfo("stackedOnSeries"),d=i.isHorizontal(),h=Tk(r,i),f=Lk(t),p=t.get("barMinHeight")||0,m=u&&e.getDimensionIndex(u),g=e.getLayout("size"),y=e.getLayout("offset");return{progress:function(t,e){var r,i=t.count,l=f&&OS(3*i),u=f&&s&&OS(3*i),v=f&&OS(i),_=n.master.getRect(),b=d?_.width:_.height,x=e.getStore(),w=0;while(null!=(r=t.next())){var M=x.get(c?m:o,r),S=x.get(a,r),k=h,L=void 0;c&&(L=+M-x.get(o,r));var T=void 0,D=void 0,O=void 0,I=void 0;if(d){var C=n.dataToPoint([M,S]);if(c){var A=n.dataToPoint([L,S]);k=A[0]}T=k,D=C[1]+y,O=C[0]-k,I=g,Math.abs(O)<p&&(O=(O<0?-1:1)*p)}else{C=n.dataToPoint([S,M]);if(c){A=n.dataToPoint([S,L]);k=A[1]}T=C[0]+y,D=k,O=g,I=C[1]-k,Math.abs(I)<p&&(I=(I<=0?-1:1)*p)}f?(l[w]=T,l[w+1]=D,l[w+2]=d?O:I,u&&(u[w]=d?_.x:T,u[w+1]=d?D:_.y,u[w+2]=b),v[r]=r):e.setItemLayout(r,{x:T,y:D,width:O,height:I}),w+=3}f&&e.setLayout({largePoints:l,largeDataIndices:v,largeBackgroundPoints:u,valueAxisHorizontal:d})}}}}}}function kk(t){return t.coordinateSystem&&"cartesian2d"===t.coordinateSystem.type}function Lk(t){return t.pipelineContext&&t.pipelineContext.large}function Tk(t,e){return e.toGlobalCoord(e.dataToCoord("log"===e.type?1:0))}var Dk=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.getInitialData=function(t,e){return dS(null,this,{useEncodeDefaulter:!0})},e.prototype.getMarkerPosition=function(t,e,n){var r=this.coordinateSystem;if(r&&r.clampData){var i=r.clampData(t),o=r.dataToPoint(i);if(n)Object(s["k"])(r.getAxes(),(function(t,n){if("category"===t.type&&null!=e){var r=t.getTicksCoords(),a=t.getTickModel().get("alignWithLabel"),s=i[n],l="x1"===e[n]||"y1"===e[n];if(l&&!a&&(s+=1),r.length<2)return;if(2===r.length)return void(o[n]=t.toGlobalCoord(t.getExtent()[l?1:0]));for(var u=void 0,c=void 0,d=1,h=0;h<r.length;h++){var f=r[h].coord,p=h===r.length-1?r[h-1].tickValue+d:r[h].tickValue;if(p===s){c=f;break}if(p<s)u=f;else if(null!=u&&p>s){c=(f+u)/2;break}1===h&&(d=p-r[0].tickValue)}null==c&&(u?u&&(c=r[r.length-1].coord):c=r[0].coord),o[n]=t.toGlobalCoord(c)}}));else{var a=this.getData(),l=a.getLayout("offset"),u=a.getLayout("size"),c=r.getBaseAxis().isHorizontal()?0:1;o[c]+=l+u/2}return o}return[NaN,NaN]},e.type="series.__base_bar__",e.defaultOption={z:2,coordinateSystem:"cartesian2d",legendHoverLink:!0,barMinHeight:0,barMinAngle:0,large:!1,largeThreshold:400,progressive:3e3,progressiveChunkMode:"mod"},e}(Mp);Mp.registerClass(Dk);var Ok=Dk,Ik=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.getInitialData=function(){return dS(null,this,{useEncodeDefaulter:!0,createInvertedIndices:!!this.get("realtimeSort",!0)||null})},e.prototype.getProgressive=function(){return!!this.get("large")&&this.get("progressive")},e.prototype.getProgressiveThreshold=function(){var t=this.get("progressiveThreshold"),e=this.get("largeThreshold");return e>t&&(t=e),t},e.prototype.brushSelector=function(t,e,n){return n.rect(e.getItemLayout(t))},e.type="series.bar",e.dependencies=["grid","polar"],e.defaultOption=Cu(Ok.defaultOption,{clip:!0,roundCap:!1,showBackground:!1,backgroundStyle:{color:"rgba(180, 180, 180, 0.2)",borderColor:null,borderWidth:0,borderType:"solid",borderRadius:0,shadowBlur:0,shadowColor:null,shadowOffsetX:0,shadowOffsetY:0,opacity:1},select:{itemStyle:{borderColor:"#212121"}},realtimeSort:!1}),e}(Ok),Ck=Ik,Ak=function(){function t(){this.cx=0,this.cy=0,this.r0=0,this.r=0,this.startAngle=0,this.endAngle=2*Math.PI,this.clockwise=!0}return t}(),Yk=function(t){function e(e){var n=t.call(this,e)||this;return n.type="sausage",n}return o(e,t),e.prototype.getDefaultShape=function(){return new Ak},e.prototype.buildPath=function(t,e){var n=e.cx,r=e.cy,i=Math.max(e.r0||0,0),o=Math.max(e.r,0),a=.5*(o-i),s=i+a,l=e.startAngle,u=e.endAngle,c=e.clockwise,d=2*Math.PI,h=c?u-l<d:l-u<d;h||(l=u-(c?d:-d));var f=Math.cos(l),p=Math.sin(l),m=Math.cos(u),g=Math.sin(u);h?(t.moveTo(f*i+n,p*i+r),t.arc(f*s+n,p*s+r,a,-Math.PI+l,l,!c)):t.moveTo(f*o+n,p*o+r),t.arc(n,r,o,l,u,!c),t.arc(m*s+n,g*s+r,a,u-2*Math.PI,u-Math.PI,!c),0!==i&&t.arc(n,r,i,u,l,c)},e}(Va),jk=Yk;function Pk(t,e){e=e||{};var n=e.isRoundCap;return function(e,r,i){var o=r.position;if(!o||o instanceof Array)return ar(e,r,i);var a=t(o),s=null!=r.distance?r.distance:5,l=this.shape,u=l.cx,c=l.cy,d=l.r,h=l.r0,f=(d+h)/2,p=l.startAngle,m=l.endAngle,g=(p+m)/2,y=n?Math.abs(d-h)/2:0,v=Math.cos,_=Math.sin,b=u+d*v(p),x=c+d*_(p),w="left",M="top";switch(a){case"startArc":b=u+(h-s)*v(g),x=c+(h-s)*_(g),w="center",M="top";break;case"insideStartArc":b=u+(h+s)*v(g),x=c+(h+s)*_(g),w="center",M="bottom";break;case"startAngle":b=u+f*v(p)+Nk(p,s+y,!1),x=c+f*_(p)+Rk(p,s+y,!1),w="right",M="middle";break;case"insideStartAngle":b=u+f*v(p)+Nk(p,-s+y,!1),x=c+f*_(p)+Rk(p,-s+y,!1),w="left",M="middle";break;case"middle":b=u+f*v(g),x=c+f*_(g),w="center",M="middle";break;case"endArc":b=u+(d+s)*v(g),x=c+(d+s)*_(g),w="center",M="bottom";break;case"insideEndArc":b=u+(d-s)*v(g),x=c+(d-s)*_(g),w="center",M="top";break;case"endAngle":b=u+f*v(m)+Nk(m,s+y,!0),x=c+f*_(m)+Rk(m,s+y,!0),w="left",M="middle";break;case"insideEndAngle":b=u+f*v(m)+Nk(m,-s+y,!0),x=c+f*_(m)+Rk(m,-s+y,!0),w="right",M="middle";break;default:return ar(e,r,i)}return e=e||{},e.x=b,e.y=x,e.align=w,e.verticalAlign=M,e}}function Ek(t,e,n,r){if(Object(s["z"])(r))t.setTextConfig({rotation:r});else if(Object(s["t"])(e))t.setTextConfig({rotation:0});else{var i,o=t.shape,a=o.clockwise?o.startAngle:o.endAngle,l=o.clockwise?o.endAngle:o.startAngle,u=(a+l)/2,c=n(e);switch(c){case"startArc":case"insideStartArc":case"middle":case"insideEndArc":case"endArc":i=u;break;case"startAngle":case"insideStartAngle":i=a;break;case"endAngle":case"insideEndAngle":i=l;break;default:return void t.setTextConfig({rotation:0})}var d=1.5*Math.PI-i;"middle"===c&&d>Math.PI/2&&d<1.5*Math.PI&&(d-=Math.PI),t.setTextConfig({rotation:d})}}function Nk(t,e,n){return e*Math.sin(t)*(n?-1:1)}function Rk(t,e,n){return e*Math.cos(t)*(n?1:-1)}function Hk(t,e,n){var r=t.get("borderRadius");if(null==r)return n?{cornerRadius:0}:null;Object(s["t"])(r)||(r=[r,r,r,r]);var i=Math.abs(e.r||0-e.r0||0);return{cornerRadius:Object(s["H"])(r,(function(t){return or(t,i)}))}}var zk=Math.max,Fk=Math.min;function Bk(t,e){var n=t.getArea&&t.getArea();if(WS(t,"cartesian2d")){var r=t.getBaseAxis();if("category"!==r.type||!r.onBand){var i=e.getLayout("bandWidth");r.isHorizontal()?(n.x-=i,n.width+=2*i):(n.y-=i,n.height+=2*i)}}return n}var Wk=function(t){function e(){var n=t.call(this)||this;return n.type=e.type,n._isFirstFrame=!0,n}return o(e,t),e.prototype.render=function(t,e,n,r){this._model=t,this._removeOnRenderedListener(n),this._updateDrawMode(t);var i=t.get("coordinateSystem");("cartesian2d"===i||"polar"===i)&&(this._progressiveEls=null,this._isLargeDraw?this._renderLarge(t,e,n):this._renderNormal(t,e,n,r))},e.prototype.incrementalPrepareRender=function(t){this._clear(),this._updateDrawMode(t),this._updateLargeClip(t)},e.prototype.incrementalRender=function(t,e){this._progressiveEls=[],this._incrementalRenderLarge(t,e)},e.prototype.eachRendered=function(t){Ng(this._progressiveEls||this.group,t)},e.prototype._updateDrawMode=function(t){var e=t.pipelineContext.large;null!=this._isLargeDraw&&e===this._isLargeDraw||(this._isLargeDraw=e,this._clear())},e.prototype._renderNormal=function(t,e,n,r){var i,o=this.group,a=t.getData(),s=this._data,l=t.coordinateSystem,u=l.getBaseAxis();"cartesian2d"===l.type?i=u.isHorizontal():"polar"===l.type&&(i="angle"===u.dim);var c=t.isAnimationEnabled()?t:null,d=$k(t,l);d&&this._enableRealtimeSort(d,a,n);var h=t.get("clip",!0)||d,f=Bk(l,a);o.removeClipPath();var p=t.get("roundCap",!0),m=t.get("showBackground",!0),g=t.getModel("backgroundStyle"),y=g.get("borderRadius")||0,v=[],_=this._backgroundEls,b=r&&r.isInitSort,x=r&&"changeAxisOrder"===r.type;function w(t){var e=Kk[l.type](a,t),n=uL(l,i,e);return n.useStyle(g.getItemStyle()),"cartesian2d"===l.type?n.setShape("r",y):n.setShape("cornerRadius",y),v[t]=n,n}a.diff(s).add((function(e){var n=a.getItemModel(e),r=Kk[l.type](a,e,n);if(m&&w(e),a.hasValue(e)&&Jk[l.type](r)){var s=!1;h&&(s=Vk[l.type](f,r));var g=Gk[l.type](t,a,e,r,i,c,u.model,!1,p);d&&(g.forceLabelAnimation=!0),eL(g,a,e,n,r,t,i,"polar"===l.type),b?g.attr({shape:r}):d?Uk(d,c,g,r,e,i,!1,!1):Vl(g,{shape:r},t,e),a.setItemGraphicEl(e,g),o.add(g),g.ignore=s}})).update((function(e,n){var r=a.getItemModel(e),M=Kk[l.type](a,e,r);if(m){var S=void 0;0===_.length?S=w(n):(S=_[n],S.useStyle(g.getItemStyle()),"cartesian2d"===l.type?S.setShape("r",y):S.setShape("cornerRadius",y),v[e]=S);var k=Kk[l.type](a,e),L=lL(i,k,l);Wl(S,{shape:L},c,e)}var T=s.getItemGraphicEl(n);if(a.hasValue(e)&&Jk[l.type](M)){var D=!1;if(h&&(D=Vk[l.type](f,M),D&&o.remove(T)),T?Xl(T):T=Gk[l.type](t,a,e,M,i,c,u.model,!!T,p),d&&(T.forceLabelAnimation=!0),x){var O=T.getTextContent();if(O){var I=du(O);null!=I.prevValue&&(I.prevValue=I.value)}}else eL(T,a,e,r,M,t,i,"polar"===l.type);b?T.attr({shape:M}):d?Uk(d,c,T,M,e,i,!0,x):Wl(T,{shape:M},t,e,null),a.setItemGraphicEl(e,T),T.ignore=D,o.add(T)}else o.remove(T)})).remove((function(e){var n=s.getItemGraphicEl(e);n&&ql(n,t,e)})).execute();var M=this._backgroundGroup||(this._backgroundGroup=new wr);M.removeAll();for(var S=0;S<v.length;++S)M.add(v[S]);o.add(M),this._backgroundEls=v,this._data=a},e.prototype._renderLarge=function(t,e,n){this._clear(),oL(t,this.group),this._updateLargeClip(t)},e.prototype._incrementalRenderLarge=function(t,e){this._removeBackground(),oL(e,this.group,this._progressiveEls,!0)},e.prototype._updateLargeClip=function(t){var e=t.get("clip",!0)&&BS(t.coordinateSystem,!1,t),n=this.group;e?n.setClipPath(e):n.removeClipPath()},e.prototype._enableRealtimeSort=function(t,e,n){var r=this;if(e.count()){var i=t.baseAxis;if(this._isFirstFrame)this._dispatchInitSort(e,t,n),this._isFirstFrame=!1;else{var o=function(t){var n=e.getItemGraphicEl(t),r=n&&n.shape;return r&&Math.abs(i.isHorizontal()?r.height:r.width)||0};this._onRendered=function(){r._updateSortWithinSameData(e,o,i,n)},n.getZr().on("rendered",this._onRendered)}}},e.prototype._dataSort=function(t,e,n){var r=[];return t.each(t.mapDimension(e.dim),(function(t,e){var i=n(e);i=null==i?NaN:i,r.push({dataIndex:e,mappedValue:i,ordinalNumber:t})})),r.sort((function(t,e){return e.mappedValue-t.mappedValue})),{ordinalNumbers:Object(s["H"])(r,(function(t){return t.ordinalNumber}))}},e.prototype._isOrderChangedWithinSameData=function(t,e,n){for(var r=n.scale,i=t.mapDimension(n.dim),o=Number.MAX_VALUE,a=0,s=r.getOrdinalMeta().categories.length;a<s;++a){var l=t.rawIndexOf(i,r.getRawOrdinalNumber(a)),u=l<0?Number.MIN_VALUE:e(t.indexOfRawIndex(l));if(u>o)return!0;o=u}return!1},e.prototype._isOrderDifferentInView=function(t,e){for(var n=e.scale,r=n.getExtent(),i=Math.max(0,r[0]),o=Math.min(r[1],n.getOrdinalMeta().categories.length-1);i<=o;++i)if(t.ordinalNumbers[i]!==n.getRawOrdinalNumber(i))return!0},e.prototype._updateSortWithinSameData=function(t,e,n,r){if(this._isOrderChangedWithinSameData(t,e,n)){var i=this._dataSort(t,n,e);this._isOrderDifferentInView(i,n)&&(this._removeOnRenderedListener(r),r.dispatchAction({type:"changeAxisOrder",componentType:n.dim+"Axis",axisId:n.index,sortInfo:i}))}},e.prototype._dispatchInitSort=function(t,e,n){var r=e.baseAxis,i=this._dataSort(t,r,(function(n){return t.get(t.mapDimension(e.otherAxis.dim),n)}));n.dispatchAction({type:"changeAxisOrder",componentType:r.dim+"Axis",isInitSort:!0,axisId:r.index,sortInfo:i})},e.prototype.remove=function(t,e){this._clear(this._model),this._removeOnRenderedListener(e)},e.prototype.dispose=function(t,e){this._removeOnRenderedListener(e)},e.prototype._removeOnRenderedListener=function(t){this._onRendered&&(t.getZr().off("rendered",this._onRendered),this._onRendered=null)},e.prototype._clear=function(t){var e=this.group,n=this._data;t&&t.isAnimationEnabled()&&n&&!this._isLargeDraw?(this._removeBackground(),this._backgroundEls=[],n.eachItemGraphicEl((function(e){ql(e,t,Ls(e).dataIndex)}))):e.removeAll(),this._data=null,this._isFirstFrame=!0},e.prototype._removeBackground=function(){this.group.remove(this._backgroundGroup),this._backgroundGroup=null},e.type="bar",e}($g),Vk={cartesian2d:function(t,e){var n=e.width<0?-1:1,r=e.height<0?-1:1;n<0&&(e.x+=e.width,e.width=-e.width),r<0&&(e.y+=e.height,e.height=-e.height);var i=t.x+t.width,o=t.y+t.height,a=zk(e.x,t.x),s=Fk(e.x+e.width,i),l=zk(e.y,t.y),u=Fk(e.y+e.height,o),c=s<a,d=u<l;return e.x=c&&a>i?s:a,e.y=d&&l>o?u:l,e.width=c?0:s-a,e.height=d?0:u-l,n<0&&(e.x+=e.width,e.width=-e.width),r<0&&(e.y+=e.height,e.height=-e.height),c||d},polar:function(t,e){var n=e.r0<=e.r?1:-1;if(n<0){var r=e.r;e.r=e.r0,e.r0=r}var i=Fk(e.r,t.r),o=zk(e.r0,t.r0);e.r=i,e.r0=o;var a=i-o<0;if(n<0){r=e.r;e.r=e.r0,e.r0=r}return a}},Gk={cartesian2d:function(t,e,n,r,i,o,a,l,u){var c=new ss({shape:Object(s["m"])({},r),z2:1});if(c.__dataIndex=n,c.name="item",o){var d=c.shape,h=i?"height":"width";d[h]=0}return c},polar:function(t,e,n,r,i,o,a,s,l){var u=!i&&l?jk:_m,c=new u({shape:r,z2:1});c.name="item";var d=tL(i);if(c.calculateTextPosition=Pk(d,{isRoundCap:u===jk}),o){var h=c.shape,f=i?"r":"endAngle",p={};h[f]=i?r.r0:r.startAngle,p[f]=r[f],(s?Wl:Vl)(c,{shape:p},o)}return c}};function $k(t,e){var n=t.get("realtimeSort",!0),r=e.getBaseAxis();if(n&&"category"===r.type&&"cartesian2d"===e.type)return{baseAxis:r,otherAxis:e.getOtherAxis(r)}}function Uk(t,e,n,r,i,o,a,s){var l,u;o?(u={x:r.x,width:r.width},l={y:r.y,height:r.height}):(u={y:r.y,height:r.height},l={x:r.x,width:r.width}),s||(a?Wl:Vl)(n,{shape:l},e,i,null);var c=e?t.baseAxis.model:null;(a?Wl:Vl)(n,{shape:u},c,i)}function qk(t,e){for(var n=0;n<e.length;n++)if(!isFinite(t[e[n]]))return!0;return!1}var Xk=["x","y","width","height"],Zk=["cx","cy","r","startAngle","endAngle"],Jk={cartesian2d:function(t){return!qk(t,Xk)},polar:function(t){return!qk(t,Zk)}},Kk={cartesian2d:function(t,e,n){var r=t.getItemLayout(e),i=n?nL(n,r):0,o=r.width>0?1:-1,a=r.height>0?1:-1;return{x:r.x+o*i/2,y:r.y+a*i/2,width:r.width-o*i,height:r.height-a*i}},polar:function(t,e,n){var r=t.getItemLayout(e);return{cx:r.cx,cy:r.cy,r0:r.r0,r:r.r,startAngle:r.startAngle,endAngle:r.endAngle,clockwise:r.clockwise}}};function Qk(t){return null!=t.startAngle&&null!=t.endAngle&&t.startAngle===t.endAngle}function tL(t){return function(t){var e=t?"Arc":"Angle";return function(t){switch(t){case"start":case"insideStart":case"end":case"insideEnd":return t+e;default:return t}}}(t)}function eL(t,e,n,r,i,o,a,l){var u=e.getItemVisual(n,"style");if(l){if(!o.get("roundCap")){var c=t.shape,d=Hk(r.getModel("itemStyle"),c,!0);Object(s["m"])(c,d),t.setShape(c)}}else{var h=r.get(["itemStyle","borderRadius"])||0;t.setShape("r",h)}t.useStyle(u);var f=r.getShallow("cursor");f&&t.attr("cursor",f);var p=l?a?i.r>=i.r0?"endArc":"startArc":i.endAngle>=i.startAngle?"endAngle":"startAngle":a?i.height>=0?"bottom":"top":i.width>=0?"right":"left",m=eu(r);tu(t,m,{labelFetcher:o,labelDataIndex:n,defaultText:pS(o.getData(),n),inheritColor:u.fill,defaultOpacity:u.opacity,defaultOutsidePosition:p});var g=t.getTextContent();if(l&&g){var y=r.get(["label","position"]);t.textConfig.inside="middle"===y||null,Ek(t,"outside"===y?p:y,tL(a),r.get(["label","rotate"]))}hu(g,m,o.getRawValue(n),(function(t){return mS(e,t)}));var v=r.getModel(["emphasis"]);Dl(t,v.get("focus"),v.get("blurScope"),v.get("disabled")),Al(t,r),Qk(i)&&(t.style.fill="none",t.style.stroke="none",Object(s["k"])(t.states,(function(t){t.style&&(t.style.fill=t.style.stroke="none")})))}function nL(t,e){var n=t.get(["itemStyle","borderColor"]);if(!n||"none"===n)return 0;var r=t.get(["itemStyle","borderWidth"])||0,i=isNaN(e.width)?Number.MAX_VALUE:Math.abs(e.width),o=isNaN(e.height)?Number.MAX_VALUE:Math.abs(e.height);return Math.min(r,i,o)}var rL=function(){function t(){}return t}(),iL=function(t){function e(e){var n=t.call(this,e)||this;return n.type="largeBar",n}return o(e,t),e.prototype.getDefaultShape=function(){return new rL},e.prototype.buildPath=function(t,e){for(var n=e.points,r=this.baseDimIdx,i=1-this.baseDimIdx,o=[],a=[],s=this.barWidth,l=0;l<n.length;l+=3)a[r]=s,a[i]=n[l+2],o[r]=n[l+r],o[i]=n[l+i],t.rect(o[0],o[1],a[0],a[1])},e}(Va);function oL(t,e,n,r){var i=t.getData(),o=i.getLayout("valueAxisHorizontal")?1:0,a=i.getLayout("largeDataIndices"),s=i.getLayout("size"),l=t.getModel("backgroundStyle"),u=i.getLayout("largeBackgroundPoints");if(u){var c=new iL({shape:{points:u},incremental:!!r,silent:!0,z2:0});c.baseDimIdx=o,c.largeDataIndices=a,c.barWidth=s,c.useStyle(l.getItemStyle()),e.add(c),n&&n.push(c)}var d=new iL({shape:{points:i.getLayout("largePoints")},incremental:!!r,ignoreCoarsePointer:!0,z2:1});d.baseDimIdx=o,d.largeDataIndices=a,d.barWidth=s,e.add(d),d.useStyle(i.getVisual("style")),Ls(d).seriesIndex=t.seriesIndex,t.get("silent")||(d.on("mousedown",aL),d.on("mousemove",aL)),n&&n.push(d)}var aL=Zg((function(t){var e=this,n=sL(e,t.offsetX,t.offsetY);Ls(e).dataIndex=n>=0?n:null}),30,!1);function sL(t,e,n){for(var r=t.baseDimIdx,i=1-r,o=t.shape.points,a=t.largeDataIndices,s=[],l=[],u=t.barWidth,c=0,d=o.length/3;c<d;c++){var h=3*c;if(l[r]=u,l[i]=o[h+2],s[r]=o[h+r],s[i]=o[h+i],l[i]<0&&(s[i]+=l[i],l[i]=-l[i]),e>=s[0]&&e<=s[0]+l[0]&&n>=s[1]&&n<=s[1]+l[1])return a[c]}return-1}function lL(t,e,n){if(WS(n,"cartesian2d")){var r=e,i=n.getArea();return{x:t?r.x:i.x,y:t?i.y:r.y,width:t?r.width:i.width,height:t?i.height:r.height}}i=n.getArea();var o=e;return{cx:i.cx,cy:i.cy,r0:t?i.r0:o.r0,r:t?i.r:o.r,startAngle:t?o.startAngle:0,endAngle:t?o.endAngle:2*Math.PI}}function uL(t,e,n){var r="polar"===t.type?_m:ss;return new r({shape:lL(e,n,t),silent:!0,z2:0})}var cL=Wk;function dL(t){t.registerChartView(cL),t.registerSeriesModel(Ck),t.registerLayout(t.PRIORITY.VISUAL.LAYOUT,s["h"](Mk,"bar")),t.registerLayout(t.PRIORITY.VISUAL.PROGRESSIVE_LAYOUT,Sk("bar")),t.registerProcessor(t.PRIORITY.PROCESSOR.STATISTIC,hk("bar")),t.registerAction({type:"changeAxisOrder",event:"changeAxisOrder",update:"update"},(function(t,e){var n=t.componentType||"series";e.eachComponent({mainType:n,query:t},(function(e){t.sortInfo&&e.axis.setCategorySortInfo(t.sortInfo)}))}))}var hL=2*Math.PI,fL=Math.PI/180;function pL(t,e){return Hc(t.getBoxLayoutParams(),{width:e.getWidth(),height:e.getHeight()})}function mL(t,e){var n=pL(t,e),r=t.get("center"),i=t.get("radius");s["t"](i)||(i=[0,i]);var o,a,l=Nr(n.width,e.getWidth()),u=Nr(n.height,e.getHeight()),c=Math.min(l,u),d=Nr(i[0],c/2),h=Nr(i[1],c/2),f=t.coordinateSystem;if(f){var p=f.dataToPoint(r);o=p[0]||0,a=p[1]||0}else s["t"](r)||(r=[r,r]),o=Nr(r[0],l)+n.x,a=Nr(r[1],u)+n.y;return{cx:o,cy:a,r0:d,r:h}}function gL(t,e,n){e.eachSeriesByType(t,(function(t){var e=t.getData(),r=e.mapDimension("value"),i=pL(t,n),o=mL(t,n),a=o.cx,s=o.cy,l=o.r,u=o.r0,c=-t.get("startAngle")*fL,d=t.get("endAngle"),h=t.get("padAngle")*fL;d="auto"===d?c-hL:-d*fL;var f=t.get("minAngle")*fL,p=f+h,m=0;e.each(r,(function(t){!isNaN(t)&&m++}));var g=e.getSum(r),y=Math.PI/(g||m)*2,v=t.get("clockwise"),_=t.get("roseType"),b=t.get("stillShowZeroSum"),x=e.getDataExtent(r);x[0]=0;var w=v?1:-1,M=[c,d],S=w*h/2;ga(M,!v),c=M[0],d=M[1];var k=Math.abs(d-c),L=k,T=0,D=c;if(e.setLayout({viewRect:i,r:l}),e.each(r,(function(t,n){var r;if(isNaN(t))e.setItemLayout(n,{angle:NaN,startAngle:NaN,endAngle:NaN,clockwise:v,cx:a,cy:s,r0:u,r:_?NaN:l});else{r="area"!==_?0===g&&b?y:t*y:k/m,r<p?(r=p,L-=p):T+=t;var i=D+w*r,o=0,c=0;h>r?(o=D+w*r/2,c=o):(o=D+S,c=i-S),e.setItemLayout(n,{angle:r,startAngle:o,endAngle:c,clockwise:v,cx:a,cy:s,r0:u,r:_?Er(t,x,[u,l]):l}),D=i}})),L<hL&&m)if(L<=.001){var O=k/m;e.each(r,(function(t,n){if(!isNaN(t)){var r=e.getItemLayout(n);r.angle=O;var i=0,o=0;O<h?(i=c+w*(n+.5)*O,o=i):(i=c+w*n*O+S,o=c+w*(n+1)*O-S),r.startAngle=i,r.endAngle=o}}))}else y=L/T,D=c,e.each(r,(function(t,n){if(!isNaN(t)){var r=e.getItemLayout(n),i=r.angle===p?p:t*y,o=0,a=0;i<h?(o=D+w*i/2,a=o):(o=D+S,a=D+w*i-S),r.startAngle=o,r.endAngle=a,D+=w*i}}))}))}function yL(t){return{seriesType:t,reset:function(t,e){var n=e.findComponents({mainType:"legend"});if(n&&n.length){var r=t.getData();r.filterSelf((function(t){for(var e=r.getName(t),i=0;i<n.length;i++)if(!n[i].isSelected(e))return!1;return!0}))}}}}var vL=Math.PI/180;function _L(t,e,n,r,i,o,a,s,l,u){if(!(t.length<2)){for(var c=t.length,d=0;d<c;d++)if("outer"===t[d].position&&"labelLine"===t[d].labelAlignTo){var h=t[d].label.x-u;t[d].linePoints[1][0]+=h,t[d].label.x=u}lx(t,l,l+a)&&p(t)}function f(t){for(var o=t.rB,a=o*o,s=0;s<t.list.length;s++){var l=t.list[s],u=Math.abs(l.label.y-n),c=r+l.len,d=c*c,h=Math.sqrt((1-Math.abs(u*u/a))*d),f=e+(h+l.len2)*i,p=f-l.label.x,m=l.targetTextWidth-p*i;xL(l,m,!0),l.label.x=f}}function p(t){for(var o={list:[],maxY:0},a={list:[],maxY:0},s=0;s<t.length;s++)if("none"===t[s].labelAlignTo){var l=t[s],u=l.label.y>n?a:o,c=Math.abs(l.label.y-n);if(c>=u.maxY){var d=l.label.x-e-l.len2*i,h=r+l.len,p=Math.abs(d)<h?Math.sqrt(c*c/(1-d*d/h/h)):h;u.rB=p,u.maxY=c}u.list.push(l)}f(o),f(a)}}function bL(t,e,n,r,i,o,a,s){for(var l=[],u=[],c=Number.MAX_VALUE,d=-Number.MAX_VALUE,h=0;h<t.length;h++){var f=t[h].label;wL(t[h])||(f.x<e?(c=Math.min(c,f.x),l.push(t[h])):(d=Math.max(d,f.x),u.push(t[h])))}for(h=0;h<t.length;h++){var p=t[h];if(!wL(p)&&p.linePoints){if(null!=p.labelStyleWidth)continue;f=p.label;var m=p.linePoints,g=void 0;g="edge"===p.labelAlignTo?f.x<e?m[2][0]-p.labelDistance-a-p.edgeDistance:a+i-p.edgeDistance-m[2][0]-p.labelDistance:"labelLine"===p.labelAlignTo?f.x<e?c-a-p.bleedMargin:a+i-d-p.bleedMargin:f.x<e?f.x-a-p.bleedMargin:a+i-f.x-p.bleedMargin,p.targetTextWidth=g,xL(p,g)}}_L(u,e,n,r,1,i,o,a,s,d),_L(l,e,n,r,-1,i,o,a,s,c);for(h=0;h<t.length;h++){p=t[h];if(!wL(p)&&p.linePoints){f=p.label,m=p.linePoints;var y="edge"===p.labelAlignTo,v=f.style.padding,_=v?v[1]+v[3]:0,b=f.style.backgroundColor?0:_,x=p.rect.width+b,w=m[1][0]-m[2][0];y?f.x<e?m[2][0]=a+p.edgeDistance+x+p.labelDistance:m[2][0]=a+i-p.edgeDistance-x-p.labelDistance:(f.x<e?m[2][0]=f.x+p.labelDistance:m[2][0]=f.x-p.labelDistance,m[1][0]=m[2][0]+w),m[1][1]=m[2][1]=f.y}}}function xL(t,e,n){if(void 0===n&&(n=!1),null==t.labelStyleWidth){var r=t.label,i=r.style,o=t.rect,a=i.backgroundColor,s=i.padding,l=s?s[1]+s[3]:0,u=i.overflow,c=o.width+(a?0:l);if(e<c||n){var d=o.height;if(u&&u.match("break")){r.setStyle("backgroundColor",null),r.setStyle("width",e-l);var h=r.getBoundingRect();r.setStyle("width",Math.ceil(h.width)),r.setStyle("backgroundColor",a)}else{var f=e-l,p=e<c?f:n?f>t.unconstrainedWidth?null:f:null;r.setStyle("width",p)}var m=r.getBoundingRect();o.width=m.width;var g=(r.style.margin||0)+2.1;o.height=m.height+g,o.y-=(o.height-d)/2}}}function wL(t){return"center"===t.position}function ML(t){var e,n,r=t.getData(),i=[],o=!1,a=(t.get("minShowLabelAngle")||0)*vL,l=r.getLayout("viewRect"),u=r.getLayout("r"),c=l.width,d=l.x,h=l.y,f=l.height;function p(t){t.ignore=!0}function m(t){if(!t.ignore)return!0;for(var e in t.states)if(!1===t.states[e].ignore)return!0;return!1}r.each((function(t){var l=r.getItemGraphicEl(t),h=l.shape,f=l.getTextContent(),g=l.getTextGuideLine(),y=r.getItemModel(t),v=y.getModel("label"),_=v.get("position")||y.get(["emphasis","label","position"]),b=v.get("distanceToLabelLine"),x=v.get("alignTo"),w=Nr(v.get("edgeDistance"),c),M=v.get("bleedMargin"),S=y.getModel("labelLine"),k=S.get("length");k=Nr(k,c);var L=S.get("length2");if(L=Nr(L,c),Math.abs(h.endAngle-h.startAngle)<a)return Object(s["k"])(f.states,p),f.ignore=!0,void(g&&(Object(s["k"])(g.states,p),g.ignore=!0));if(m(f)){var T,D,O,I,C=(h.startAngle+h.endAngle)/2,A=Math.cos(C),Y=Math.sin(C);e=h.cx,n=h.cy;var j="inside"===_||"inner"===_;if("center"===_)T=h.cx,D=h.cy,I="center";else{var P=(j?(h.r+h.r0)/2*A:h.r*A)+e,E=(j?(h.r+h.r0)/2*Y:h.r*Y)+n;if(T=P+3*A,D=E+3*Y,!j){var N=P+A*(k+u-h.r),R=E+Y*(k+u-h.r),H=N+(A<0?-1:1)*L,z=R;T="edge"===x?A<0?d+w:d+c-w:H+(A<0?-b:b),D=z,O=[[P,E],[N,R],[H,z]]}I=j?"center":"edge"===x?A>0?"right":"left":A>0?"left":"right"}var F=Math.PI,B=0,W=v.get("rotate");if(Object(s["z"])(W))B=W*(F/180);else if("center"===_)B=0;else if("radial"===W||!0===W){var V=A<0?-C+F:-C;B=V}else if("tangential"===W&&"outside"!==_&&"outer"!==_){var G=Math.atan2(A,Y);G<0&&(G=2*F+G);var $=Y>0;$&&(G=F+G),B=G-F}if(o=!!B,f.x=T,f.y=D,f.rotation=B,f.setStyle({verticalAlign:"middle"}),j){f.setStyle({align:I});var U=f.states.select;U&&(U.x+=f.x,U.y+=f.y)}else{var q=f.getBoundingRect().clone();q.applyTransform(f.getComputedTransform());var X=(f.style.margin||0)+2.1;q.y-=X/2,q.height+=X,i.push({label:f,labelLine:g,position:_,len:k,len2:L,minTurnAngle:S.get("minTurnAngle"),maxSurfaceAngle:S.get("maxSurfaceAngle"),surfaceNormal:new yt(A,Y),linePoints:O,textAlign:I,labelDistance:b,labelAlignTo:x,edgeDistance:w,bleedMargin:M,rect:q,unconstrainedWidth:q.width,labelStyleWidth:f.style.width})}l.setTextConfig({inside:j})}})),!o&&t.get("avoidLabelOverlap")&&bL(i,e,n,u,c,f,d,h);for(var g=0;g<i.length;g++){var y=i[g],v=y.label,_=y.labelLine,b=isNaN(v.x)||isNaN(v.y);if(v){v.setStyle({align:y.textAlign}),b&&(Object(s["k"])(v.states,p),v.ignore=!0);var x=v.states.select;x&&(x.x+=v.x,x.y+=v.y)}if(_){var w=y.linePoints;b||!w?(Object(s["k"])(_.states,p),_.ignore=!0):(Qb(w,y.minTurnAngle),tx(w,y.surfaceNormal,y.maxSurfaceAngle),_.setShape({points:w}),v.__hostTarget.textGuideLineConfig={anchor:new yt(w[0][0],w[0][1])})}}}var SL=function(t){function e(e,n,r){var i=t.call(this)||this;i.z2=2;var o=new ks;return i.setTextContent(o),i.updateData(e,n,r,!0),i}return o(e,t),e.prototype.updateData=function(t,e,n,r){var i=this,o=t.hostModel,a=t.getItemModel(e),l=a.getModel("emphasis"),u=t.getItemLayout(e),c=Object(s["m"])(Hk(a.getModel("itemStyle"),u,!0),u);if(isNaN(c.startAngle))i.setShape(c);else{if(r){i.setShape(c);var d=o.getShallow("animationType");o.ecModel.ssr?(Vl(i,{scaleX:0,scaleY:0},o,{dataIndex:e,isFrom:!0}),i.originX=c.cx,i.originY=c.cy):"scale"===d?(i.shape.r=u.r0,Vl(i,{shape:{r:u.r}},o,e)):null!=n?(i.setShape({startAngle:n,endAngle:n}),Vl(i,{shape:{startAngle:u.startAngle,endAngle:u.endAngle}},o,e)):(i.shape.endAngle=u.startAngle,Wl(i,{shape:{endAngle:u.endAngle}},o,e))}else Xl(i),Wl(i,{shape:c},o,e);i.useStyle(t.getItemVisual(e,"style")),Al(i,a);var h=(u.startAngle+u.endAngle)/2,f=o.get("selectedOffset"),p=Math.cos(h)*f,m=Math.sin(h)*f,g=a.getShallow("cursor");g&&i.attr("cursor",g),this._updateLabel(o,t,e),i.ensureState("emphasis").shape=Object(s["m"])({r:u.r+(l.get("scale")&&l.get("scaleSize")||0)},Hk(l.getModel("itemStyle"),u)),Object(s["m"])(i.ensureState("select"),{x:p,y:m,shape:Hk(a.getModel(["select","itemStyle"]),u)}),Object(s["m"])(i.ensureState("blur"),{shape:Hk(a.getModel(["blur","itemStyle"]),u)});var y=i.getTextGuideLine(),v=i.getTextContent();y&&Object(s["m"])(y.ensureState("select"),{x:p,y:m}),Object(s["m"])(v.ensureState("select"),{x:p,y:m}),Dl(this,l.get("focus"),l.get("blurScope"),l.get("disabled"))}},e.prototype._updateLabel=function(t,e,n){var r=this,i=e.getItemModel(n),o=i.getModel("labelLine"),a=e.getItemVisual(n,"style"),l=a&&a.fill,u=a&&a.opacity;tu(r,eu(i),{labelFetcher:e.hostModel,labelDataIndex:n,inheritColor:l,defaultOpacity:u,defaultText:t.getFormattedLabel(n,"normal")||e.getName(n)});var c=r.getTextContent();r.setTextConfig({position:null,rotation:null}),c.attr({z2:10});var d=t.get(["label","position"]);if("outside"!==d&&"outer"!==d)r.removeTextGuideLine();else{var h=this.getTextGuideLine();h||(h=new Im,this.setTextGuideLine(h)),rx(this,ix(i),{stroke:l,opacity:Object(s["Q"])(o.get(["lineStyle","opacity"]),u,1)})}},e}(_m),kL=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.ignoreLabelLineUpdate=!0,e}return o(e,t),e.prototype.render=function(t,e,n,r){var i,o=t.getData(),a=this._data,s=this.group;if(!a&&o.count()>0){for(var l=o.getItemLayout(0),u=1;isNaN(l&&l.startAngle)&&u<o.count();++u)l=o.getItemLayout(u);l&&(i=l.startAngle)}if(this._emptyCircleSector&&s.remove(this._emptyCircleSector),0===o.count()&&t.get("showEmptyCircle")){var c=new _m({shape:mL(t,n)});c.useStyle(t.getModel("emptyCircleStyle").getItemStyle()),this._emptyCircleSector=c,s.add(c)}o.diff(a).add((function(t){var e=new SL(o,t,i);o.setItemGraphicEl(t,e),s.add(e)})).update((function(t,e){var n=a.getItemGraphicEl(e);n.updateData(o,t,i),n.off("click"),s.add(n),o.setItemGraphicEl(t,n)})).remove((function(e){var n=a.getItemGraphicEl(e);ql(n,t,e)})).execute(),ML(t),"expansion"!==t.get("animationTypeUpdate")&&(this._data=o)},e.prototype.dispose=function(){},e.prototype.containPoint=function(t,e){var n=e.getData(),r=n.getItemLayout(0);if(r){var i=t[0]-r.cx,o=t[1]-r.cy,a=Math.sqrt(i*i+o*o);return a<=r.r&&a>=r.r0}},e.type="pie",e}($g),LL=kL;function TL(t,e,n){e=Object(s["t"])(e)&&{coordDimensions:e}||Object(s["m"])({encodeDefine:t.getEncode()},e);var r=t.getSource(),i=qM(r,e).dimensions,o=new UM(i,t);return o.initData(r,n),o}var DL=function(){function t(t,e){this._getDataWithEncodedVisual=t,this._getRawData=e}return t.prototype.getAllNames=function(){var t=this._getRawData();return t.mapArray(t.getName)},t.prototype.containName=function(t){var e=this._getRawData();return e.indexOfName(t)>=0},t.prototype.indexOfName=function(t){var e=this._getDataWithEncodedVisual();return e.indexOfName(t)},t.prototype.getItemVisual=function(t,e){var n=this._getDataWithEncodedVisual();return n.getItemVisual(t,e)},t}(),OL=DL,IL=Ci(),CL=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.init=function(e){t.prototype.init.apply(this,arguments),this.legendVisualProvider=new OL(s["c"](this.getData,this),s["c"](this.getRawData,this)),this._defaultLabelLine(e)},e.prototype.mergeOption=function(){t.prototype.mergeOption.apply(this,arguments)},e.prototype.getInitialData=function(){return TL(this,{coordDimensions:["value"],encodeDefaulter:s["h"](hd,this)})},e.prototype.getDataParams=function(e){var n=this.getData(),r=IL(n),i=r.seats;if(!i){var o=[];n.each(n.mapDimension("value"),(function(t){o.push(t)})),i=r.seats=Wr(o,n.hostModel.get("percentPrecision"))}var a=t.prototype.getDataParams.call(this,e);return a.percent=i[e]||0,a.$vars.push("percent"),a},e.prototype._defaultLabelLine=function(t){ci(t,"labelLine",["show"]);var e=t.labelLine,n=t.emphasis.labelLine;e.show=e.show&&t.label.show,n.show=n.show&&t.emphasis.label.show},e.type="series.pie",e.defaultOption={z:2,legendHoverLink:!0,colorBy:"data",center:["50%","50%"],radius:[0,"75%"],clockwise:!0,startAngle:90,endAngle:"auto",padAngle:0,minAngle:0,minShowLabelAngle:0,selectedOffset:10,percentPrecision:2,stillShowZeroSum:!0,left:0,top:0,right:0,bottom:0,width:null,height:null,label:{rotate:0,show:!0,overflow:"truncate",position:"outer",alignTo:"none",edgeDistance:"25%",bleedMargin:10,distanceToLabelLine:5},labelLine:{show:!0,length:15,length2:15,smooth:!1,minTurnAngle:90,maxSurfaceAngle:90,lineStyle:{width:1,type:"solid"}},itemStyle:{borderWidth:1,borderJoin:"round"},showEmptyCircle:!0,emptyCircleStyle:{color:"lightgray",opacity:1},labelLayout:{hideOverlap:!0},emphasis:{scale:!0,scaleSize:5},avoidLabelOverlap:!0,animationType:"expansion",animationDuration:1e3,animationTypeUpdate:"transition",animationEasingUpdate:"cubicInOut",animationDurationUpdate:500,animationEasing:"cubicInOut"},e}(Mp),AL=CL;function YL(t){return{seriesType:t,reset:function(t,e){var n=t.getData();n.filterSelf((function(t){var e=n.mapDimension("value"),r=n.get(e,t);return!(Object(s["z"])(r)&&!isNaN(r)&&r<0)}))}}}function jL(t){t.registerChartView(LL),t.registerSeriesModel(AL),By("pie",t.registerAction),t.registerLayout(Object(s["h"])(gL,"pie")),t.registerProcessor(yL("pie")),t.registerProcessor(YL("pie"))}var PL=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.hasSymbolVisual=!0,n}return o(e,t),e.prototype.getInitialData=function(t,e){return dS(null,this,{useEncodeDefaulter:!0})},e.prototype.getProgressive=function(){var t=this.option.progressive;return null==t?this.option.large?5e3:this.get("progressive"):t},e.prototype.getProgressiveThreshold=function(){var t=this.option.progressiveThreshold;return null==t?this.option.large?1e4:this.get("progressiveThreshold"):t},e.prototype.brushSelector=function(t,e,n){return n.point(e.getItemLayout(t))},e.prototype.getZLevelKey=function(){return this.getData().count()>this.getProgressiveThreshold()?this.id:""},e.type="series.scatter",e.dependencies=["grid","polar","geo","singleAxis","calendar"],e.defaultOption={coordinateSystem:"cartesian2d",z:2,legendHoverLink:!0,symbolSize:10,large:!1,largeThreshold:2e3,itemStyle:{opacity:.8},emphasis:{scale:!0},clip:!0,select:{itemStyle:{borderColor:"#212121"}},universalTransition:{divideShape:"clone"}},e}(Mp),EL=PL,NL=4,RL=function(){function t(){}return t}(),HL=function(t){function e(e){var n=t.call(this,e)||this;return n._off=0,n.hoverDataIdx=-1,n}return o(e,t),e.prototype.getDefaultShape=function(){return new RL},e.prototype.reset=function(){this.notClear=!1,this._off=0},e.prototype.buildPath=function(t,e){var n,r=e.points,i=e.size,o=this.symbolProxy,a=o.shape,s=t.getContext?t.getContext():t,l=s&&i[0]<NL,u=this.softClipShape;if(l)this._ctx=s;else{for(this._ctx=null,n=this._off;n<r.length;){var c=r[n++],d=r[n++];isNaN(c)||isNaN(d)||(u&&!u.contain(c,d)||(a.x=c-i[0]/2,a.y=d-i[1]/2,a.width=i[0],a.height=i[1],o.buildPath(t,a,!0)))}this.incremental&&(this._off=n,this.notClear=!0)}},e.prototype.afterBrush=function(){var t,e=this.shape,n=e.points,r=e.size,i=this._ctx,o=this.softClipShape;if(i){for(t=this._off;t<n.length;){var a=n[t++],s=n[t++];isNaN(a)||isNaN(s)||(o&&!o.contain(a,s)||i.fillRect(a-r[0]/2,s-r[1]/2,r[0],r[1]))}this.incremental&&(this._off=t,this.notClear=!0)}},e.prototype.findDataIndex=function(t,e){for(var n=this.shape,r=n.points,i=n.size,o=Math.max(i[0],4),a=Math.max(i[1],4),s=r.length/2-1;s>=0;s--){var l=2*s,u=r[l]-o/2,c=r[l+1]-a/2;if(t>=u&&e>=c&&t<=u+o&&e<=c+a)return s}return-1},e.prototype.contain=function(t,e){var n=this.transformCoordToLocal(t,e),r=this.getBoundingRect();if(t=n[0],e=n[1],r.contain(t,e)){var i=this.hoverDataIdx=this.findDataIndex(t,e);return i>=0}return this.hoverDataIdx=-1,!1},e.prototype.getBoundingRect=function(){var t=this._rect;if(!t){for(var e=this.shape,n=e.points,r=e.size,i=r[0],o=r[1],a=1/0,s=1/0,l=-1/0,u=-1/0,c=0;c<n.length;){var d=n[c++],h=n[c++];a=Math.min(d,a),l=Math.max(d,l),s=Math.min(h,s),u=Math.max(h,u)}t=this._rect=new Tt(a-i/2,s-o/2,l-a+i,u-s+o)}return t},e}(Va),zL=function(){function t(){this.group=new wr}return t.prototype.updateData=function(t,e){this._clear();var n=this._create();n.setShape({points:t.getLayout("points")}),this._setCommon(n,t,e)},t.prototype.updateLayout=function(t){var e=t.getLayout("points");this.group.eachChild((function(t){if(null!=t.startIndex){var n=2*(t.endIndex-t.startIndex),r=4*t.startIndex*2;e=new Float32Array(e.buffer,r,n)}t.setShape("points",e),t.reset()}))},t.prototype.incrementalPrepareUpdate=function(t){this._clear()},t.prototype.incrementalUpdate=function(t,e,n){var r=this._newAdded[0],i=e.getLayout("points"),o=r&&r.shape.points;if(o&&o.length<2e4){var a=o.length,s=new Float32Array(a+i.length);s.set(o),s.set(i,a),r.endIndex=t.end,r.setShape({points:s})}else{this._newAdded=[];var l=this._create();l.startIndex=t.start,l.endIndex=t.end,l.incremental=!0,l.setShape({points:i}),this._setCommon(l,e,n)}},t.prototype.eachRendered=function(t){this._newAdded[0]&&t(this._newAdded[0])},t.prototype._create=function(){var t=new HL({cursor:"default"});return t.ignoreCoarsePointer=!0,this.group.add(t),this._newAdded.push(t),t},t.prototype._setCommon=function(t,e,n){var r=e.hostModel;n=n||{};var i=e.getVisual("symbolSize");t.setShape("size",i instanceof Array?i:[i,i]),t.softClipShape=n.clipShape||null,t.symbolProxy=ov(e.getVisual("symbol"),0,0,0,0),t.setColor=t.symbolProxy.setColor;var o=t.shape.size[0]<NL;t.useStyle(r.getModel("itemStyle").getItemStyle(o?["color","shadowBlur","shadowColor"]:["color"]));var a=e.getVisual("style"),s=a&&a.fill;s&&t.setColor(s);var l=Ls(t);l.seriesIndex=r.seriesIndex,t.on("mousemove",(function(e){l.dataIndex=null;var n=t.hoverDataIdx;n>=0&&(l.dataIndex=n+(t.startIndex||0))}))},t.prototype.remove=function(){this._clear()},t.prototype._clear=function(){this._newAdded=[],this.group.removeAll()},t}(),FL=zL,BL=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){var r=t.getData(),i=this._updateSymbolDraw(r,t);i.updateData(r,{clipShape:this._getClipShape(t)}),this._finished=!0},e.prototype.incrementalPrepareRender=function(t,e,n){var r=t.getData(),i=this._updateSymbolDraw(r,t);i.incrementalPrepareUpdate(r),this._finished=!1},e.prototype.incrementalRender=function(t,e,n){this._symbolDraw.incrementalUpdate(t,e.getData(),{clipShape:this._getClipShape(e)}),this._finished=t.end===e.getData().count()},e.prototype.updateTransform=function(t,e,n){var r=t.getData();if(this.group.dirty(),!this._finished||r.count()>1e4)return{update:!0};var i=uk("").reset(t,e,n);i.progress&&i.progress({start:0,end:r.count(),count:r.count()},r),this._symbolDraw.updateLayout(r)},e.prototype.eachRendered=function(t){this._symbolDraw&&this._symbolDraw.eachRendered(t)},e.prototype._getClipShape=function(t){if(t.get("clip",!0)){var e=t.coordinateSystem;return e&&e.getArea&&e.getArea(.1)}},e.prototype._updateSymbolDraw=function(t,e){var n=this._symbolDraw,r=e.pipelineContext,i=r.large;return n&&i===this._isLargeDraw||(n&&n.remove(),n=this._symbolDraw=i?new FL:new MS,this._isLargeDraw=i,this.group.removeAll()),this.group.add(n.group),n},e.prototype.remove=function(t,e){this._symbolDraw&&this._symbolDraw.remove(!0),this._symbolDraw=null},e.prototype.dispose=function(){},e.type="scatter",e}($g),WL=BL,VL=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.type="grid",e.dependencies=["xAxis","yAxis"],e.layoutMode="box",e.defaultOption={show:!1,z:0,left:"10%",top:60,right:"10%",bottom:70,containLabel:!1,backgroundColor:"rgba(0,0,0,0)",borderWidth:1,borderColor:"#ccc"},e}(Xc),GL=VL,$L=function(){function t(){}return t.prototype.getNeedCrossZero=function(){var t=this.option;return!t.scale},t.prototype.getCoordSysModel=function(){},t}(),UL=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.getCoordSysModel=function(){return this.getReferringComponents("grid",Pi).models[0]},e.type="cartesian2dAxis",e}(Xc);s["K"](UL,$L);var qL={show:!0,z:0,inverse:!1,name:"",nameLocation:"end",nameRotate:null,nameTruncate:{maxWidth:null,ellipsis:"...",placeholder:"."},nameTextStyle:{},nameGap:15,silent:!1,triggerEvent:!1,tooltip:{show:!1},axisPointer:{},axisLine:{show:!0,onZero:!0,onZeroAxisIndex:null,lineStyle:{color:"#6E7079",width:1,type:"solid"},symbol:["none","none"],symbolSize:[10,15]},axisTick:{show:!0,inside:!1,length:5,lineStyle:{width:1}},axisLabel:{show:!0,inside:!1,rotate:0,showMinLabel:null,showMaxLabel:null,margin:8,fontSize:12},splitLine:{show:!0,lineStyle:{color:["#E0E6F1"],width:1,type:"solid"}},splitArea:{show:!1,areaStyle:{color:["rgba(250,250,250,0.2)","rgba(210,219,238,0.2)"]}}},XL=s["I"]({boundaryGap:!0,deduplication:null,splitLine:{show:!1},axisTick:{alignWithLabel:!1,interval:"auto"},axisLabel:{interval:"auto"}},qL),ZL=s["I"]({boundaryGap:[0,0],axisLine:{show:"auto"},axisTick:{show:"auto"},splitNumber:5,minorTick:{show:!1,splitNumber:5,length:3,lineStyle:{}},minorSplitLine:{show:!1,lineStyle:{color:"#F4F7FD",width:1}}},qL),JL=s["I"]({splitNumber:6,axisLabel:{showMinLabel:!1,showMaxLabel:!1,rich:{primary:{fontWeight:"bold"}}},splitLine:{show:!1}},ZL),KL=s["i"]({logBase:10},ZL),QL={category:XL,value:ZL,time:JL,log:KL},tT=0,eT=function(){function t(t){this.categories=t.categories||[],this._needCollect=t.needCollect,this._deduplication=t.deduplication,this.uid=++tT}return t.createByAxisModel=function(e){var n=e.option,r=n.data,i=r&&Object(s["H"])(r,nT);return new t({categories:i,needCollect:!i,deduplication:!1!==n.dedplication})},t.prototype.getOrdinal=function(t){return this._getOrCreateMap().get(t)},t.prototype.parseAndCollect=function(t){var e,n=this._needCollect;if(!Object(s["C"])(t)&&!n)return t;if(n&&!this._deduplication)return e=this.categories.length,this.categories[e]=t,e;var r=this._getOrCreateMap();return e=r.get(t),null==e&&(n?(e=this.categories.length,this.categories[e]=t,r.set(t,e)):e=NaN),e},t.prototype._getOrCreateMap=function(){return this._map||(this._map=Object(s["f"])(this.categories))},t}();function nT(t){return Object(s["A"])(t)&&null!=t.value?t.value:t+""}var rT=eT,iT={value:1,category:1,time:1,log:1};function oT(t,e,n,r){Object(s["k"])(iT,(function(i,a){var l=Object(s["I"])(Object(s["I"])({},QL[a],!0),r,!0),u=function(t){function n(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e+"Axis."+a,n}return o(n,t),n.prototype.mergeDefaultAndTheme=function(t,e){var n=Bc(this),r=n?Vc(t):{},i=e.getTheme();Object(s["I"])(t,i.get(a+"Axis")),Object(s["I"])(t,this.getDefaultOption()),t.type=aT(t),n&&Wc(t,r,n)},n.prototype.optionUpdated=function(){var t=this.option;"category"===t.type&&(this.__ordinalMeta=rT.createByAxisModel(this))},n.prototype.getCategories=function(t){var e=this.option;if("category"===e.type)return t?e.data:this.__ordinalMeta.categories},n.prototype.getOrdinalMeta=function(){return this.__ordinalMeta},n.type=e+"Axis."+a,n.defaultOption=l,n}(n);t.registerComponentModel(u)})),t.registerSubTypeDefaulter(e+"Axis",aT)}function aT(t){return t.type||(t.data?"category":"value")}var sT=function(){function t(t){this._setting=t||{},this._extent=[1/0,-1/0]}return t.prototype.getSetting=function(t){return this._setting[t]},t.prototype.unionExtent=function(t){var e=this._extent;t[0]<e[0]&&(e[0]=t[0]),t[1]>e[1]&&(e[1]=t[1])},t.prototype.unionExtentFromData=function(t,e){this.unionExtent(t.getApproximateExtent(e))},t.prototype.getExtent=function(){return this._extent.slice()},t.prototype.setExtent=function(t,e){var n=this._extent;isNaN(t)||(n[0]=t),isNaN(e)||(n[1]=e)},t.prototype.isInExtentRange=function(t){return this._extent[0]<=t&&this._extent[1]>=t},t.prototype.isBlank=function(){return this._isBlank},t.prototype.setBlank=function(t){this._isBlank=t},t}();no(sT);var lT=sT;function uT(t){return"interval"===t.type||"log"===t.type}function cT(t,e,n,r){var i={},o=t[1]-t[0],a=i.interval=Kr(o/e,!0);null!=n&&a<n&&(a=i.interval=n),null!=r&&a>r&&(a=i.interval=r);var s=i.intervalPrecision=hT(a),l=i.niceTickExtent=[Rr(Math.ceil(t[0]/a)*a,s),Rr(Math.floor(t[1]/a)*a,s)];return pT(l,t),i}function dT(t){var e=Math.pow(10,Jr(t)),n=t/e;return n?2===n?n=3:3===n?n=5:n*=2:n=1,Rr(n*e)}function hT(t){return zr(t)+2}function fT(t,e,n){t[e]=Math.max(Math.min(t[e],n[1]),n[0])}function pT(t,e){!isFinite(t[0])&&(t[0]=e[0]),!isFinite(t[1])&&(t[1]=e[1]),fT(t,0,e),fT(t,1,e),t[0]>t[1]&&(t[0]=t[1])}function mT(t,e){return t>=e[0]&&t<=e[1]}function gT(t,e){return e[1]===e[0]?.5:(t-e[0])/(e[1]-e[0])}function yT(t,e){return t*(e[1]-e[0])+e[0]}var vT=function(t){function e(e){var n=t.call(this,e)||this;n.type="ordinal";var r=n.getSetting("ordinalMeta");return r||(r=new rT({})),Object(s["t"])(r)&&(r=new rT({categories:Object(s["H"])(r,(function(t){return Object(s["A"])(t)?t.value:t}))})),n._ordinalMeta=r,n._extent=n.getSetting("extent")||[0,r.categories.length-1],n}return o(e,t),e.prototype.parse=function(t){return null==t?NaN:Object(s["C"])(t)?this._ordinalMeta.getOrdinal(t):Math.round(t)},e.prototype.contain=function(t){return t=this.parse(t),mT(t,this._extent)&&null!=this._ordinalMeta.categories[t]},e.prototype.normalize=function(t){return t=this._getTickNumber(this.parse(t)),gT(t,this._extent)},e.prototype.scale=function(t){return t=Math.round(yT(t,this._extent)),this.getRawOrdinalNumber(t)},e.prototype.getTicks=function(){var t=[],e=this._extent,n=e[0];while(n<=e[1])t.push({value:n}),n++;return t},e.prototype.getMinorTicks=function(t){},e.prototype.setSortInfo=function(t){if(null!=t){for(var e=t.ordinalNumbers,n=this._ordinalNumbersByTick=[],r=this._ticksByOrdinalNumber=[],i=0,o=this._ordinalMeta.categories.length,a=Math.min(o,e.length);i<a;++i){var s=e[i];n[i]=s,r[s]=i}for(var l=0;i<o;++i){while(null!=r[l])l++;n.push(l),r[l]=i}}else this._ordinalNumbersByTick=this._ticksByOrdinalNumber=null},e.prototype._getTickNumber=function(t){var e=this._ticksByOrdinalNumber;return e&&t>=0&&t<e.length?e[t]:t},e.prototype.getRawOrdinalNumber=function(t){var e=this._ordinalNumbersByTick;return e&&t>=0&&t<e.length?e[t]:t},e.prototype.getLabel=function(t){if(!this.isBlank()){var e=this.getRawOrdinalNumber(t.value),n=this._ordinalMeta.categories[e];return null==n?"":n+""}},e.prototype.count=function(){return this._extent[1]-this._extent[0]+1},e.prototype.unionExtentFromData=function(t,e){this.unionExtent(t.getApproximateExtent(e))},e.prototype.isInExtentRange=function(t){return t=this._getTickNumber(t),this._extent[0]<=t&&this._extent[1]>=t},e.prototype.getOrdinalMeta=function(){return this._ordinalMeta},e.prototype.calcNiceTicks=function(){},e.prototype.calcNiceExtent=function(){},e.type="ordinal",e}(lT);lT.registerClass(vT);var _T=vT,bT=Rr,xT=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.type="interval",e._interval=0,e._intervalPrecision=2,e}return o(e,t),e.prototype.parse=function(t){return t},e.prototype.contain=function(t){return mT(t,this._extent)},e.prototype.normalize=function(t){return gT(t,this._extent)},e.prototype.scale=function(t){return yT(t,this._extent)},e.prototype.setExtent=function(t,e){var n=this._extent;isNaN(t)||(n[0]=parseFloat(t)),isNaN(e)||(n[1]=parseFloat(e))},e.prototype.unionExtent=function(t){var e=this._extent;t[0]<e[0]&&(e[0]=t[0]),t[1]>e[1]&&(e[1]=t[1]),this.setExtent(e[0],e[1])},e.prototype.getInterval=function(){return this._interval},e.prototype.setInterval=function(t){this._interval=t,this._niceExtent=this._extent.slice(),this._intervalPrecision=hT(t)},e.prototype.getTicks=function(t){var e=this._interval,n=this._extent,r=this._niceExtent,i=this._intervalPrecision,o=[];if(!e)return o;var a=1e4;n[0]<r[0]&&(t?o.push({value:bT(r[0]-e,i)}):o.push({value:n[0]}));var s=r[0];while(s<=r[1]){if(o.push({value:s}),s=bT(s+e,i),s===o[o.length-1].value)break;if(o.length>a)return[]}var l=o.length?o[o.length-1].value:r[1];return n[1]>l&&(t?o.push({value:bT(l+e,i)}):o.push({value:n[1]})),o},e.prototype.getMinorTicks=function(t){for(var e=this.getTicks(!0),n=[],r=this.getExtent(),i=1;i<e.length;i++){var o=e[i],a=e[i-1],s=0,l=[],u=o.value-a.value,c=u/t;while(s<t-1){var d=bT(a.value+(s+1)*c);d>r[0]&&d<r[1]&&l.push(d),s++}n.push(l)}return n},e.prototype.getLabel=function(t,e){if(null==t)return"";var n=e&&e.precision;null==n?n=zr(t.value)||0:"auto"===n&&(n=this._intervalPrecision);var r=bT(t.value,n,!0);return wc(r)},e.prototype.calcNiceTicks=function(t,e,n){t=t||5;var r=this._extent,i=r[1]-r[0];if(isFinite(i)){i<0&&(i=-i,r.reverse());var o=cT(r,t,e,n);this._intervalPrecision=o.intervalPrecision,this._interval=o.interval,this._niceExtent=o.niceTickExtent}},e.prototype.calcNiceExtent=function(t){var e=this._extent;if(e[0]===e[1])if(0!==e[0]){var n=Math.abs(e[0]);t.fixMax||(e[1]+=n/2),e[0]-=n/2}else e[1]=1;var r=e[1]-e[0];isFinite(r)||(e[0]=0,e[1]=1),this.calcNiceTicks(t.splitNumber,t.minInterval,t.maxInterval);var i=this._interval;t.fixMin||(e[0]=bT(Math.floor(e[0]/i)*i)),t.fixMax||(e[1]=bT(Math.ceil(e[1]/i)*i))},e.prototype.setNiceExtent=function(t,e){this._niceExtent=[t,e]},e.type="interval",e}(lT);lT.registerClass(xT);var wT=xT,MT=function(t,e,n,r){while(n<r){var i=n+r>>>1;t[i][1]<e?n=i+1:r=i}return n},ST=function(t){function e(e){var n=t.call(this,e)||this;return n.type="time",n}return o(e,t),e.prototype.getLabel=function(t){var e=this.getSetting("useUTC");return ic(t.value,Ju[rc(ec(this._minLevelUnit))]||Ju.second,e,this.getSetting("locale"))},e.prototype.getFormattedLabel=function(t,e,n){var r=this.getSetting("useUTC"),i=this.getSetting("locale");return oc(t,e,n,i,r)},e.prototype.getTicks=function(){var t=this._interval,e=this._extent,n=[];if(!t)return n;n.push({value:e[0],level:0});var r=this.getSetting("useUTC"),i=YT(this._minLevelUnit,this._approxInterval,r,e);return n=n.concat(i),n.push({value:e[1],level:0}),n},e.prototype.calcNiceExtent=function(t){var e=this._extent;if(e[0]===e[1]&&(e[0]-=Uu,e[1]+=Uu),e[1]===-1/0&&e[0]===1/0){var n=new Date;e[1]=+new Date(n.getFullYear(),n.getMonth(),n.getDate()),e[0]=e[1]-Uu}this.calcNiceTicks(t.splitNumber,t.minInterval,t.maxInterval)},e.prototype.calcNiceTicks=function(t,e,n){t=t||10;var r=this._extent,i=r[1]-r[0];this._approxInterval=i/t,null!=e&&this._approxInterval<e&&(this._approxInterval=e),null!=n&&this._approxInterval>n&&(this._approxInterval=n);var o=kT.length,a=Math.min(MT(kT,this._approxInterval,0,o),o-1);this._interval=kT[a][1],this._minLevelUnit=kT[Math.max(a-1,0)][0]},e.prototype.parse=function(t){return Object(s["z"])(t)?t:+Xr(t)},e.prototype.contain=function(t){return mT(this.parse(t),this._extent)},e.prototype.normalize=function(t){return gT(this.parse(t),this._extent)},e.prototype.scale=function(t){return yT(t,this._extent)},e.type="time",e}(wT),kT=[["second",Vu],["minute",Gu],["hour",$u],["quarter-day",6*$u],["half-day",12*$u],["day",1.2*Uu],["half-week",3.5*Uu],["week",7*Uu],["month",31*Uu],["quarter",95*Uu],["half-year",qu/2],["year",qu]];function LT(t,e,n,r){var i=Xr(e),o=Xr(n),a=function(t){return sc(i,t,r)===sc(o,t,r)},s=function(){return a("year")},l=function(){return s()&&a("month")},u=function(){return l()&&a("day")},c=function(){return u()&&a("hour")},d=function(){return c()&&a("minute")},h=function(){return d()&&a("second")},f=function(){return h()&&a("millisecond")};switch(t){case"year":return s();case"month":return l();case"day":return u();case"hour":return c();case"minute":return d();case"second":return h();case"millisecond":return f()}}function TT(t,e){return t/=Uu,t>16?16:t>7.5?7:t>3.5?4:t>1.5?2:1}function DT(t){var e=30*Uu;return t/=e,t>6?6:t>3?3:t>2?2:1}function OT(t){return t/=$u,t>12?12:t>6?6:t>3.5?4:t>2?2:1}function IT(t,e){return t/=e?Gu:Vu,t>30?30:t>20?20:t>15?15:t>10?10:t>5?5:t>2?2:1}function CT(t){return Kr(t,!0)}function AT(t,e,n){var r=new Date(t);switch(ec(e)){case"year":case"month":r[gc(n)](0);case"day":r[yc(n)](1);case"hour":r[vc(n)](0);case"minute":r[_c(n)](0);case"second":r[bc(n)](0),r[xc(n)](0)}return r.getTime()}function YT(t,e,n,r){var i=1e4,o=Qu,a=0;function l(t,e,n,i,o,a,s){var l=new Date(e),u=e,c=l[i]();while(u<n&&u<=r[1])s.push({value:u}),c+=t,l[o](c),u=l.getTime();s.push({value:u,notAdd:!0})}function u(t,i,o){var a=[],s=!i.length;if(!LT(ec(t),r[0],r[1],n)){s&&(i=[{value:AT(new Date(r[0]),t,n)},{value:r[1]}]);for(var u=0;u<i.length-1;u++){var c=i[u].value,d=i[u+1].value;if(c!==d){var h=void 0,f=void 0,p=void 0,m=!1;switch(t){case"year":h=Math.max(1,Math.round(e/Uu/365)),f=lc(n),p=mc(n);break;case"half-year":case"quarter":case"month":h=DT(e),f=uc(n),p=gc(n);break;case"week":case"half-week":case"day":h=TT(e,31),f=cc(n),p=yc(n),m=!0;break;case"half-day":case"quarter-day":case"hour":h=OT(e),f=dc(n),p=vc(n);break;case"minute":h=IT(e,!0),f=hc(n),p=_c(n);break;case"second":h=IT(e,!1),f=fc(n),p=bc(n);break;case"millisecond":h=CT(e),f=pc(n),p=xc(n);break}l(h,c,d,f,p,m,a),"year"===t&&o.length>1&&0===u&&o.unshift({value:o[0].value-h})}}for(u=0;u<a.length;u++)o.push(a[u]);return a}}for(var c=[],d=[],h=0,f=0,p=0;p<o.length&&a++<i;++p){var m=ec(o[p]);if(nc(o[p])){u(o[p],c[c.length-1]||[],d);var g=o[p+1]?ec(o[p+1]):null;if(m!==g){if(d.length){f=h,d.sort((function(t,e){return t.value-e.value}));for(var y=[],v=0;v<d.length;++v){var _=d[v].value;0!==v&&d[v-1].value===_||(y.push(d[v]),_>=r[0]&&_<=r[1]&&h++)}var b=(r[1]-r[0])/e;if(h>1.5*b&&f>b/1.5)break;if(c.push(y),h>b||t===o[p])break}d=[]}}}var x=Object(s["n"])(Object(s["H"])(c,(function(t){return Object(s["n"])(t,(function(t){return t.value>=r[0]&&t.value<=r[1]&&!t.notAdd}))})),(function(t){return t.length>0})),w=[],M=x.length-1;for(p=0;p<x.length;++p)for(var S=x[p],k=0;k<S.length;++k)w.push({value:S[k].value,level:M-p});w.sort((function(t,e){return t.value-e.value}));var L=[];for(p=0;p<w.length;++p)0!==p&&w[p].value===w[p-1].value||L.push(w[p]);return L}lT.registerClass(ST);var jT=ST,PT=lT.prototype,ET=wT.prototype,NT=Rr,RT=Math.floor,HT=Math.ceil,zT=Math.pow,FT=Math.log,BT=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.type="log",e.base=10,e._originalScale=new wT,e._interval=0,e}return o(e,t),e.prototype.getTicks=function(t){var e=this._originalScale,n=this._extent,r=e.getExtent(),i=ET.getTicks.call(this,t);return s["H"](i,(function(t){var e=t.value,i=Rr(zT(this.base,e));return i=e===n[0]&&this._fixMin?VT(i,r[0]):i,i=e===n[1]&&this._fixMax?VT(i,r[1]):i,{value:i}}),this)},e.prototype.setExtent=function(t,e){var n=FT(this.base);t=FT(Math.max(0,t))/n,e=FT(Math.max(0,e))/n,ET.setExtent.call(this,t,e)},e.prototype.getExtent=function(){var t=this.base,e=PT.getExtent.call(this);e[0]=zT(t,e[0]),e[1]=zT(t,e[1]);var n=this._originalScale,r=n.getExtent();return this._fixMin&&(e[0]=VT(e[0],r[0])),this._fixMax&&(e[1]=VT(e[1],r[1])),e},e.prototype.unionExtent=function(t){this._originalScale.unionExtent(t);var e=this.base;t[0]=FT(t[0])/FT(e),t[1]=FT(t[1])/FT(e),PT.unionExtent.call(this,t)},e.prototype.unionExtentFromData=function(t,e){this.unionExtent(t.getApproximateExtent(e))},e.prototype.calcNiceTicks=function(t){t=t||10;var e=this._extent,n=e[1]-e[0];if(!(n===1/0||n<=0)){var r=Zr(n),i=t/n*r;i<=.5&&(r*=10);while(!isNaN(r)&&Math.abs(r)<1&&Math.abs(r)>0)r*=10;var o=[Rr(HT(e[0]/r)*r),Rr(RT(e[1]/r)*r)];this._interval=r,this._niceExtent=o}},e.prototype.calcNiceExtent=function(t){ET.calcNiceExtent.call(this,t),this._fixMin=t.fixMin,this._fixMax=t.fixMax},e.prototype.parse=function(t){return t},e.prototype.contain=function(t){return t=FT(t)/FT(this.base),mT(t,this._extent)},e.prototype.normalize=function(t){return t=FT(t)/FT(this.base),gT(t,this._extent)},e.prototype.scale=function(t){return t=yT(t,this._extent),zT(this.base,t)},e.type="log",e}(lT),WT=BT.prototype;function VT(t,e){return NT(t,zr(e))}WT.getMinorTicks=ET.getMinorTicks,WT.getLabel=ET.getLabel,lT.registerClass(BT);var GT=BT,$T=function(){function t(t,e,n){this._prepareParams(t,e,n)}return t.prototype._prepareParams=function(t,e,n){n[1]<n[0]&&(n=[NaN,NaN]),this._dataMin=n[0],this._dataMax=n[1];var r=this._isOrdinal="ordinal"===t.type;this._needCrossZero="interval"===t.type&&e.getNeedCrossZero&&e.getNeedCrossZero();var i=this._modelMinRaw=e.get("min",!0);Object(s["w"])(i)?this._modelMinNum=ZT(t,i({min:n[0],max:n[1]})):"dataMin"!==i&&(this._modelMinNum=ZT(t,i));var o=this._modelMaxRaw=e.get("max",!0);if(Object(s["w"])(o)?this._modelMaxNum=ZT(t,o({min:n[0],max:n[1]})):"dataMax"!==o&&(this._modelMaxNum=ZT(t,o)),r)this._axisDataLen=e.getCategories().length;else{var a=e.get("boundaryGap"),l=Object(s["t"])(a)?a:[a||0,a||0];"boolean"===typeof l[0]||"boolean"===typeof l[1]?this._boundaryGapInner=[0,0]:this._boundaryGapInner=[or(l[0],1),or(l[1],1)]}},t.prototype.calculate=function(){var t=this._isOrdinal,e=this._dataMin,n=this._dataMax,r=this._axisDataLen,i=this._boundaryGapInner,o=t?null:n-e||Math.abs(e),a="dataMin"===this._modelMinRaw?e:this._modelMinNum,l="dataMax"===this._modelMaxRaw?n:this._modelMaxNum,u=null!=a,c=null!=l;null==a&&(a=t?r?0:NaN:e-i[0]*o),null==l&&(l=t?r?r-1:NaN:n+i[1]*o),(null==a||!isFinite(a))&&(a=NaN),(null==l||!isFinite(l))&&(l=NaN);var d=Object(s["l"])(a)||Object(s["l"])(l)||t&&!r;this._needCrossZero&&(a>0&&l>0&&!u&&(a=0),a<0&&l<0&&!c&&(l=0));var h=this._determinedMin,f=this._determinedMax;return null!=h&&(a=h,u=!0),null!=f&&(l=f,c=!0),{min:a,max:l,minFixed:u,maxFixed:c,isBlank:d}},t.prototype.modifyDataMinMax=function(t,e){this[qT[t]]=e},t.prototype.setDeterminedMinMax=function(t,e){var n=UT[t];this[n]=e},t.prototype.freeze=function(){this.frozen=!0},t}(),UT={min:"_determinedMin",max:"_determinedMax"},qT={min:"_dataMin",max:"_dataMax"};function XT(t,e,n){var r=t.rawExtentInfo;return r||(r=new $T(t,e,n),t.rawExtentInfo=r,r)}function ZT(t,e){return null==e?null:Object(s["l"])(e)?NaN:t.parse(e)}function JT(t,e){var n=t.type,r=XT(t,e,t.getExtent()).calculate();t.setBlank(r.isBlank);var i=r.min,o=r.max,a=e.ecModel;if(a&&"time"===n){var l=vk("bar",a),u=!1;if(s["k"](l,(function(t){u=u||t.getBaseAxis()===e.axis})),u){var c=bk(l),d=KT(i,o,e,c);i=d.min,o=d.max}}return{extent:[i,o],fixMin:r.minFixed,fixMax:r.maxFixed}}function KT(t,e,n,r){var i=n.axis.getExtent(),o=i[1]-i[0],a=wk(r,n.axis);if(void 0===a)return{min:t,max:e};var l=1/0;s["k"](a,(function(t){l=Math.min(t.offset,l)}));var u=-1/0;s["k"](a,(function(t){u=Math.max(t.offset+t.width,u)})),l=Math.abs(l),u=Math.abs(u);var c=l+u,d=e-t,h=1-(l+u)/o,f=d/h-d;return e+=f*(u/c),t-=f*(l/c),{min:t,max:e}}function QT(t,e){var n=e,r=JT(t,n),i=r.extent,o=n.get("splitNumber");t instanceof GT&&(t.base=n.get("logBase"));var a=t.type,s=n.get("interval"),l="interval"===a||"time"===a;t.setExtent(i[0],i[1]),t.calcNiceExtent({splitNumber:o,fixMin:r.fixMin,fixMax:r.fixMax,minInterval:l?n.get("minInterval"):null,maxInterval:l?n.get("maxInterval"):null}),null!=s&&t.setInterval&&t.setInterval(s)}function tD(t,e){if(e=e||t.get("type"),e)switch(e){case"category":return new _T({ordinalMeta:t.getOrdinalMeta?t.getOrdinalMeta():t.getCategories(),extent:[1/0,-1/0]});case"time":return new jT({locale:t.ecModel.getLocaleModel(),useUTC:t.ecModel.get("useUTC")});default:return new(lT.getClass(e)||wT)}}function eD(t){var e=t.scale.getExtent(),n=e[0],r=e[1];return!(n>0&&r>0||n<0&&r<0)}function nD(t){var e=t.getLabelModel().get("formatter"),n="category"===t.type?t.scale.getExtent()[0]:null;return"time"===t.scale.type?function(e){return function(n,r){return t.scale.getFormattedLabel(n,r,e)}}(e):s["C"](e)?function(e){return function(n){var r=t.scale.getLabel(n),i=e.replace("{value}",null!=r?r:"");return i}}(e):s["w"](e)?function(e){return function(r,i){return null!=n&&(i=r.value-n),e(rD(t,r),i,null!=r.level?{level:r.level}:null)}}(e):function(e){return t.scale.getLabel(e)}}function rD(t,e){return"category"===t.type?t.scale.getLabel(e):e.value}function iD(t){var e=t.model,n=t.scale;if(e.get(["axisLabel","show"])&&!n.isBlank()){var r,i,o=n.getExtent();n instanceof _T?i=n.count():(r=n.getTicks(),i=r.length);var a,s=t.getLabelModel(),l=nD(t),u=1;i>40&&(u=Math.ceil(i/40));for(var c=0;c<i;c+=u){var d=r?r[c]:{value:o[0]+c},h=l(d,c),f=s.getTextRect(h),p=oD(f,s.get("rotate")||0);a?a.union(p):a=p}return a}}function oD(t,e){var n=e*Math.PI/180,r=t.width,i=t.height,o=r*Math.abs(Math.cos(n))+Math.abs(i*Math.sin(n)),a=r*Math.abs(Math.sin(n))+Math.abs(i*Math.cos(n)),s=new Tt(t.x,t.y,o,a);return s}function aD(t){var e=t.get("interval");return null==e?"auto":e}function sD(t){return"category"===t.type&&0===aD(t.getLabelModel())}function lD(t,e){var n={};return s["k"](t.mapDimensionsAll(e),(function(e){n[oS(t,e)]=!0})),s["F"](n)}function uD(t,e,n){e&&s["k"](lD(e,n),(function(n){var r=e.getApproximateExtent(n);r[0]<t[0]&&(t[0]=r[0]),r[1]>t[1]&&(t[1]=r[1])}))}var cD=function(){function t(t){this.type="cartesian",this._dimList=[],this._axes={},this.name=t||""}return t.prototype.getAxis=function(t){return this._axes[t]},t.prototype.getAxes=function(){return s["H"](this._dimList,(function(t){return this._axes[t]}),this)},t.prototype.getAxesByScale=function(t){return t=t.toLowerCase(),s["n"](this.getAxes(),(function(e){return e.scale.type===t}))},t.prototype.addAxis=function(t){var e=t.dim;this._axes[e]=t,this._dimList.push(e)},t}(),dD=cD,hD=["x","y"];function fD(t){return"interval"===t.type||"time"===t.type}var pD=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.type="cartesian2d",e.dimensions=hD,e}return o(e,t),e.prototype.calcAffineTransform=function(){this._transform=this._invTransform=null;var t=this.getAxis("x").scale,e=this.getAxis("y").scale;if(fD(t)&&fD(e)){var n=t.getExtent(),r=e.getExtent(),i=this.dataToPoint([n[0],r[0]]),o=this.dataToPoint([n[1],r[1]]),a=n[1]-n[0],s=r[1]-r[0];if(a&&s){var l=(o[0]-i[0])/a,u=(o[1]-i[1])/s,c=i[0]-n[0]*l,d=i[1]-r[0]*u,h=this._transform=[l,0,0,u,c,d];this._invTransform=pt([],h)}}},e.prototype.getBaseAxis=function(){return this.getAxesByScale("ordinal")[0]||this.getAxesByScale("time")[0]||this.getAxis("x")},e.prototype.containPoint=function(t){var e=this.getAxis("x"),n=this.getAxis("y");return e.contain(e.toLocalCoord(t[0]))&&n.contain(n.toLocalCoord(t[1]))},e.prototype.containData=function(t){return this.getAxis("x").containData(t[0])&&this.getAxis("y").containData(t[1])},e.prototype.containZone=function(t,e){var n=this.dataToPoint(t),r=this.dataToPoint(e),i=this.getArea(),o=new Tt(n[0],n[1],r[0]-n[0],r[1]-n[1]);return i.intersect(o)},e.prototype.dataToPoint=function(t,e,n){n=n||[];var r=t[0],i=t[1];if(this._transform&&null!=r&&isFinite(r)&&null!=i&&isFinite(i))return S(n,t,this._transform);var o=this.getAxis("x"),a=this.getAxis("y");return n[0]=o.toGlobalCoord(o.dataToCoord(r,e)),n[1]=a.toGlobalCoord(a.dataToCoord(i,e)),n},e.prototype.clampData=function(t,e){var n=this.getAxis("x").scale,r=this.getAxis("y").scale,i=n.getExtent(),o=r.getExtent(),a=n.parse(t[0]),s=r.parse(t[1]);return e=e||[],e[0]=Math.min(Math.max(Math.min(i[0],i[1]),a),Math.max(i[0],i[1])),e[1]=Math.min(Math.max(Math.min(o[0],o[1]),s),Math.max(o[0],o[1])),e},e.prototype.pointToData=function(t,e){var n=[];if(this._invTransform)return S(n,t,this._invTransform);var r=this.getAxis("x"),i=this.getAxis("y");return n[0]=r.coordToData(r.toLocalCoord(t[0]),e),n[1]=i.coordToData(i.toLocalCoord(t[1]),e),n},e.prototype.getOtherAxis=function(t){return this.getAxis("x"===t.dim?"y":"x")},e.prototype.getArea=function(t){t=t||0;var e=this.getAxis("x").getGlobalExtent(),n=this.getAxis("y").getGlobalExtent(),r=Math.min(e[0],e[1])-t,i=Math.min(n[0],n[1])-t,o=Math.max(e[0],e[1])-r+t,a=Math.max(n[0],n[1])-i+t;return new Tt(r,i,o,a)},e}(dD),mD=pD,gD=Ci();function yD(t){return"category"===t.type?_D(t):wD(t)}function vD(t,e){return"category"===t.type?xD(t,e):{ticks:s["H"](t.scale.getTicks(),(function(t){return t.value}))}}function _D(t){var e=t.getLabelModel(),n=bD(t,e);return!e.get("show")||t.scale.isBlank()?{labels:[],labelCategoryInterval:n.labelCategoryInterval}:n}function bD(t,e){var n,r,i=MD(t,"labels"),o=aD(e),a=SD(i,o);return a||(s["w"](o)?n=ID(t,o):(r="auto"===o?LD(t):o,n=OD(t,r)),kD(i,o,{labels:n,labelCategoryInterval:r}))}function xD(t,e){var n,r,i=MD(t,"ticks"),o=aD(e),a=SD(i,o);if(a)return a;if(e.get("show")&&!t.scale.isBlank()||(n=[]),s["w"](o))n=ID(t,o,!0);else if("auto"===o){var l=bD(t,t.getLabelModel());r=l.labelCategoryInterval,n=s["H"](l.labels,(function(t){return t.tickValue}))}else r=o,n=OD(t,r,!0);return kD(i,o,{ticks:n,tickCategoryInterval:r})}function wD(t){var e=t.scale.getTicks(),n=nD(t);return{labels:s["H"](e,(function(e,r){return{level:e.level,formattedLabel:n(e,r),rawLabel:t.scale.getLabel(e),tickValue:e.value}}))}}function MD(t,e){return gD(t)[e]||(gD(t)[e]=[])}function SD(t,e){for(var n=0;n<t.length;n++)if(t[n].key===e)return t[n].value}function kD(t,e,n){return t.push({key:e,value:n}),n}function LD(t){var e=gD(t).autoInterval;return null!=e?e:gD(t).autoInterval=t.calculateCategoryInterval()}function TD(t){var e=DD(t),n=nD(t),r=(e.axisRotate-e.labelRotate)/180*Math.PI,i=t.scale,o=i.getExtent(),a=i.count();if(o[1]-o[0]<1)return 0;var s=1;a>40&&(s=Math.max(1,Math.floor(a/40)));for(var l=o[0],u=t.dataToCoord(l+1)-t.dataToCoord(l),c=Math.abs(u*Math.cos(r)),d=Math.abs(u*Math.sin(r)),h=0,f=0;l<=o[1];l+=s){var p=0,m=0,g=er(n({value:l}),e.font,"center","top");p=1.3*g.width,m=1.3*g.height,h=Math.max(h,p,7),f=Math.max(f,m,7)}var y=h/c,v=f/d;isNaN(y)&&(y=1/0),isNaN(v)&&(v=1/0);var _=Math.max(0,Math.floor(Math.min(y,v))),b=gD(t.model),x=t.getExtent(),w=b.lastAutoInterval,M=b.lastTickCount;return null!=w&&null!=M&&Math.abs(w-_)<=1&&Math.abs(M-a)<=1&&w>_&&b.axisExtent0===x[0]&&b.axisExtent1===x[1]?_=w:(b.lastTickCount=a,b.lastAutoInterval=_,b.axisExtent0=x[0],b.axisExtent1=x[1]),_}function DD(t){var e=t.getLabelModel();return{axisRotate:t.getRotate?t.getRotate():t.isHorizontal&&!t.isHorizontal()?90:0,labelRotate:e.get("rotate")||0,font:e.getFont()}}function OD(t,e,n){var r=nD(t),i=t.scale,o=i.getExtent(),a=t.getLabelModel(),s=[],l=Math.max((e||0)+1,1),u=o[0],c=i.count();0!==u&&l>1&&c/l>2&&(u=Math.round(Math.ceil(u/l)*l));var d=sD(t),h=a.get("showMinLabel")||d,f=a.get("showMaxLabel")||d;h&&u!==o[0]&&m(o[0]);for(var p=u;p<=o[1];p+=l)m(p);function m(t){var e={value:t};s.push(n?t:{formattedLabel:r(e),rawLabel:i.getLabel(e),tickValue:t})}return f&&p-l!==o[1]&&m(o[1]),s}function ID(t,e,n){var r=t.scale,i=nD(t),o=[];return s["k"](r.getTicks(),(function(t){var a=r.getLabel(t),s=t.value;e(t.value,a)&&o.push(n?s:{formattedLabel:i(t),rawLabel:a,tickValue:s})})),o}var CD=[0,1],AD=function(){function t(t,e,n){this.onBand=!1,this.inverse=!1,this.dim=t,this.scale=e,this._extent=n||[0,0]}return t.prototype.contain=function(t){var e=this._extent,n=Math.min(e[0],e[1]),r=Math.max(e[0],e[1]);return t>=n&&t<=r},t.prototype.containData=function(t){return this.scale.contain(t)},t.prototype.getExtent=function(){return this._extent.slice()},t.prototype.getPixelPrecision=function(t){return Br(t||this.scale.getExtent(),this._extent)},t.prototype.setExtent=function(t,e){var n=this._extent;n[0]=t,n[1]=e},t.prototype.dataToCoord=function(t,e){var n=this._extent,r=this.scale;return t=r.normalize(t),this.onBand&&"ordinal"===r.type&&(n=n.slice(),YD(n,r.count())),Er(t,CD,n,e)},t.prototype.coordToData=function(t,e){var n=this._extent,r=this.scale;this.onBand&&"ordinal"===r.type&&(n=n.slice(),YD(n,r.count()));var i=Er(t,n,CD,e);return this.scale.scale(i)},t.prototype.pointToData=function(t,e){},t.prototype.getTicksCoords=function(t){t=t||{};var e=t.tickModel||this.getTickModel(),n=vD(this,e),r=n.ticks,i=Object(s["H"])(r,(function(t){return{coord:this.dataToCoord("ordinal"===this.scale.type?this.scale.getRawOrdinalNumber(t):t),tickValue:t}}),this),o=e.get("alignWithLabel");return jD(this,i,o,t.clamp),i},t.prototype.getMinorTicksCoords=function(){if("ordinal"===this.scale.type)return[];var t=this.model.getModel("minorTick"),e=t.get("splitNumber");e>0&&e<100||(e=5);var n=this.scale.getMinorTicks(e),r=Object(s["H"])(n,(function(t){return Object(s["H"])(t,(function(t){return{coord:this.dataToCoord(t),tickValue:t}}),this)}),this);return r},t.prototype.getViewLabels=function(){return yD(this).labels},t.prototype.getLabelModel=function(){return this.model.getModel("axisLabel")},t.prototype.getTickModel=function(){return this.model.getModel("axisTick")},t.prototype.getBandWidth=function(){var t=this._extent,e=this.scale.getExtent(),n=e[1]-e[0]+(this.onBand?1:0);0===n&&(n=1);var r=Math.abs(t[1]-t[0]);return Math.abs(r)/n},t.prototype.calculateCategoryInterval=function(){return TD(this)},t}();function YD(t,e){var n=t[1]-t[0],r=e,i=n/r/2;t[0]+=i,t[1]-=i}function jD(t,e,n,r){var i=e.length;if(t.onBand&&!n&&i){var o,a,l=t.getExtent();if(1===i)e[0].coord=l[0],o=e[1]={coord:l[1]};else{var u=e[i-1].tickValue-e[0].tickValue,c=(e[i-1].coord-e[0].coord)/u;Object(s["k"])(e,(function(t){t.coord-=c/2}));var d=t.scale.getExtent();a=1+d[1]-e[i-1].tickValue,o={coord:e[i-1].coord+c*a},e.push(o)}var h=l[0]>l[1];f(e[0].coord,l[0])&&(r?e[0].coord=l[0]:e.shift()),r&&f(l[0],e[0].coord)&&e.unshift({coord:l[0]}),f(l[1],o.coord)&&(r?o.coord=l[1]:e.pop()),r&&f(o.coord,l[1])&&e.push({coord:l[1]})}function f(t,e){return t=Rr(t),e=Rr(e),h?t>e:t<e}}var PD=AD,ED=function(t){function e(e,n,r,i,o){var a=t.call(this,e,n,r)||this;return a.index=0,a.type=i||"value",a.position=o||"bottom",a}return o(e,t),e.prototype.isHorizontal=function(){var t=this.position;return"top"===t||"bottom"===t},e.prototype.getGlobalExtent=function(t){var e=this.getExtent();return e[0]=this.toGlobalCoord(e[0]),e[1]=this.toGlobalCoord(e[1]),t&&e[0]>e[1]&&e.reverse(),e},e.prototype.pointToData=function(t,e){return this.coordToData(this.toLocalCoord(t["x"===this.dim?0:1]),e)},e.prototype.setCategorySortInfo=function(t){if("category"!==this.type)return!1;this.model.option.categorySortInfo=t,this.scale.setSortInfo(t)},e}(PD),ND=ED;function RD(t,e,n){n=n||{};var r=t.coordinateSystem,i=e.axis,o={},a=i.getAxesOnZeroOf()[0],l=i.position,u=a?"onZero":l,c=i.dim,d=r.getRect(),h=[d.x,d.x+d.width,d.y,d.y+d.height],f={left:0,right:1,top:0,bottom:1,onZero:2},p=e.get("offset")||0,m="x"===c?[h[2]-p,h[3]+p]:[h[0]-p,h[1]+p];if(a){var g=a.toGlobalCoord(a.dataToCoord(0));m[f.onZero]=Math.max(Math.min(g,m[1]),m[0])}o.position=["y"===c?m[f[u]]:h[0],"x"===c?m[f[u]]:h[3]],o.rotation=Math.PI/2*("x"===c?0:1);var y={top:-1,bottom:1,left:-1,right:1};o.labelDirection=o.tickDirection=o.nameDirection=y[l],o.labelOffset=a?m[f[l]]-m[f.onZero]:0,e.get(["axisTick","inside"])&&(o.tickDirection=-o.tickDirection),s["O"](n.labelInside,e.get(["axisLabel","inside"]))&&(o.labelDirection=-o.labelDirection);var v=e.get(["axisLabel","rotate"]);return o.labelRotate="top"===u?-v:v,o.z2=1,o}function HD(t){return"cartesian2d"===t.get("coordinateSystem")}function zD(t){var e={xAxisModel:null,yAxisModel:null};return s["k"](e,(function(n,r){var i=r.replace(/Model$/,""),o=t.getReferringComponents(i,Pi).models[0];e[r]=o})),e}var FD=Math.log;function BD(t,e,n){var r=wT.prototype,i=r.getTicks.call(n),o=r.getTicks.call(n,!0),a=i.length-1,s=r.getInterval.call(n),l=JT(t,e),u=l.extent,c=l.fixMin,d=l.fixMax;if("log"===t.type){var h=FD(t.base);u=[FD(u[0])/h,FD(u[1])/h]}t.setExtent(u[0],u[1]),t.calcNiceExtent({splitNumber:a,fixMin:c,fixMax:d});var f=r.getExtent.call(t);c&&(u[0]=f[0]),d&&(u[1]=f[1]);var p=r.getInterval.call(t),m=u[0],g=u[1];if(c&&d)p=(g-m)/a;else if(c){g=u[0]+p*a;while(g<u[1]&&isFinite(g)&&isFinite(u[1]))p=dT(p),g=u[0]+p*a}else if(d){m=u[1]-p*a;while(m>u[0]&&isFinite(m)&&isFinite(u[0]))p=dT(p),m=u[1]-p*a}else{var y=t.getTicks().length-1;y>a&&(p=dT(p));var v=p*a;g=Math.ceil(u[1]/p)*p,m=Rr(g-v),m<0&&u[0]>=0?(m=0,g=Rr(v)):g>0&&u[1]<=0&&(g=0,m=-Rr(v))}var _=(i[0].value-o[0].value)/s,b=(i[a].value-o[a].value)/s;r.setExtent.call(t,m+p*_,g+p*b),r.setInterval.call(t,p),(_||b)&&r.setNiceExtent.call(t,m+p,g-p)}var WD=function(){function t(t,e,n){this.type="grid",this._coordsMap={},this._coordsList=[],this._axesMap={},this._axesList=[],this.axisPointerEnabled=!0,this.dimensions=hD,this._initCartesian(t,e,n),this.model=t}return t.prototype.getRect=function(){return this._rect},t.prototype.update=function(t,e){var n=this._axesMap;function r(t){var e,n=Object(s["F"])(t),r=n.length;if(r){for(var i=[],o=r-1;o>=0;o--){var a=+n[o],l=t[a],u=l.model,c=l.scale;uT(c)&&u.get("alignTicks")&&null==u.get("interval")?i.push(l):(QT(c,u),uT(c)&&(e=l))}i.length&&(e||(e=i.pop(),QT(e.scale,e.model)),Object(s["k"])(i,(function(t){BD(t.scale,t.model,e.scale)})))}}this._updateScale(t,this.model),r(n.x),r(n.y);var i={};Object(s["k"])(n.x,(function(t){GD(n,"y",t,i)})),Object(s["k"])(n.y,(function(t){GD(n,"x",t,i)})),this.resize(this.model,e)},t.prototype.resize=function(t,e,n){var r=t.getBoxLayoutParams(),i=!n&&t.get("containLabel"),o=Hc(r,{width:e.getWidth(),height:e.getHeight()});this._rect=o;var a=this._axesList;function l(){Object(s["k"])(a,(function(t){var e=t.isHorizontal(),n=e?[0,o.width]:[0,o.height],r=t.inverse?1:0;t.setExtent(n[r],n[1-r]),UD(t,e?o.x:o.y)}))}l(),i&&(Object(s["k"])(a,(function(t){if(!t.model.get(["axisLabel","inside"])){var e=iD(t);if(e){var n=t.isHorizontal()?"height":"width",r=t.model.get(["axisLabel","margin"]);o[n]-=e[n]+r,"top"===t.position?o.y+=e.height+r:"left"===t.position&&(o.x+=e.width+r)}}})),l()),Object(s["k"])(this._coordsList,(function(t){t.calcAffineTransform()}))},t.prototype.getAxis=function(t,e){var n=this._axesMap[t];if(null!=n)return n[e||0]},t.prototype.getAxes=function(){return this._axesList.slice()},t.prototype.getCartesian=function(t,e){if(null!=t&&null!=e){var n="x"+t+"y"+e;return this._coordsMap[n]}Object(s["A"])(t)&&(e=t.yAxisIndex,t=t.xAxisIndex);for(var r=0,i=this._coordsList;r<i.length;r++)if(i[r].getAxis("x").index===t||i[r].getAxis("y").index===e)return i[r]},t.prototype.getCartesians=function(){return this._coordsList.slice()},t.prototype.convertToPixel=function(t,e,n){var r=this._findConvertTarget(e);return r.cartesian?r.cartesian.dataToPoint(n):r.axis?r.axis.toGlobalCoord(r.axis.dataToCoord(n)):null},t.prototype.convertFromPixel=function(t,e,n){var r=this._findConvertTarget(e);return r.cartesian?r.cartesian.pointToData(n):r.axis?r.axis.coordToData(r.axis.toLocalCoord(n)):null},t.prototype._findConvertTarget=function(t){var e,n,r=t.seriesModel,i=t.xAxisModel||r&&r.getReferringComponents("xAxis",Pi).models[0],o=t.yAxisModel||r&&r.getReferringComponents("yAxis",Pi).models[0],a=t.gridModel,l=this._coordsList;if(r)e=r.coordinateSystem,Object(s["r"])(l,e)<0&&(e=null);else if(i&&o)e=this.getCartesian(i.componentIndex,o.componentIndex);else if(i)n=this.getAxis("x",i.componentIndex);else if(o)n=this.getAxis("y",o.componentIndex);else if(a){var u=a.coordinateSystem;u===this&&(e=this._coordsList[0])}return{cartesian:e,axis:n}},t.prototype.containPoint=function(t){var e=this._coordsList[0];if(e)return e.containPoint(t)},t.prototype._initCartesian=function(t,e,n){var r=this,i=this,o={left:!1,right:!1,top:!1,bottom:!1},a={x:{},y:{}},l={x:0,y:0};if(e.eachComponent("xAxis",u("x"),this),e.eachComponent("yAxis",u("y"),this),!l.x||!l.y)return this._axesMap={},void(this._axesList=[]);function u(e){return function(n,r){if(VD(n,t)){var s=n.get("position");"x"===e?"top"!==s&&"bottom"!==s&&(s=o.bottom?"top":"bottom"):"left"!==s&&"right"!==s&&(s=o.left?"right":"left"),o[s]=!0;var u=new ND(e,tD(n),[0,0],n.get("type"),s),c="category"===u.type;u.onBand=c&&n.get("boundaryGap"),u.inverse=n.get("inverse"),n.axis=u,u.model=n,u.grid=i,u.index=r,i._axesList.push(u),a[e][r]=u,l[e]++}}}this._axesMap=a,Object(s["k"])(a.x,(function(e,n){Object(s["k"])(a.y,(function(i,o){var a="x"+n+"y"+o,s=new mD(a);s.master=r,s.model=t,r._coordsMap[a]=s,r._coordsList.push(s),s.addAxis(e),s.addAxis(i)}))}))},t.prototype._updateScale=function(t,e){function n(t,e){Object(s["k"])(lD(t,e.dim),(function(n){e.scale.unionExtentFromData(t,n)}))}Object(s["k"])(this._axesList,(function(t){if(t.scale.setExtent(1/0,-1/0),"category"===t.type){var e=t.model.get("categorySortInfo");t.scale.setSortInfo(e)}})),t.eachSeries((function(t){if(HD(t)){var r=zD(t),i=r.xAxisModel,o=r.yAxisModel;if(!VD(i,e)||!VD(o,e))return;var a=this.getCartesian(i.componentIndex,o.componentIndex),s=t.getData(),l=a.getAxis("x"),u=a.getAxis("y");n(s,l),n(s,u)}}),this)},t.prototype.getTooltipAxes=function(t){var e=[],n=[];return Object(s["k"])(this.getCartesians(),(function(r){var i=null!=t&&"auto"!==t?r.getAxis(t):r.getBaseAxis(),o=r.getOtherAxis(i);Object(s["r"])(e,i)<0&&e.push(i),Object(s["r"])(n,o)<0&&n.push(o)})),{baseAxes:e,otherAxes:n}},t.create=function(e,n){var r=[];return e.eachComponent("grid",(function(i,o){var a=new t(i,e,n);a.name="grid_"+o,a.resize(i,n,!0),i.coordinateSystem=a,r.push(a)})),e.eachSeries((function(t){if(HD(t)){var e=zD(t),n=e.xAxisModel,r=e.yAxisModel,i=n.getCoordSysModel();0;var o=i.coordinateSystem;t.coordinateSystem=o.getCartesian(n.componentIndex,r.componentIndex)}})),r},t.dimensions=hD,t}();function VD(t,e){return t.getCoordSysModel()===e}function GD(t,e,n,r){n.getAxesOnZeroOf=function(){return i?[i]:[]};var i,o=t[e],a=n.model,s=a.get(["axisLine","onZero"]),l=a.get(["axisLine","onZeroAxisIndex"]);if(s){if(null!=l)$D(o[l])&&(i=o[l]);else for(var u in o)if(o.hasOwnProperty(u)&&$D(o[u])&&!r[c(o[u])]){i=o[u];break}i&&(r[c(i)]=!0)}function c(t){return t.dim+"_"+t.index}}function $D(t){return t&&"category"!==t.type&&"time"!==t.type&&eD(t)}function UD(t,e){var n=t.getExtent(),r=n[0]+n[1];t.toGlobalCoord="x"===t.dim?function(t){return t+e}:function(t){return r-t+e},t.toLocalCoord="x"===t.dim?function(t){return t-e}:function(t){return r-t+e}}var qD=WD,XD=Math.PI,ZD=function(){function t(t,e){this.group=new wr,this.opt=e,this.axisModel=t,Object(s["i"])(e,{labelOffset:0,nameDirection:1,tickDirection:1,labelDirection:1,silent:!0,handleAutoShown:function(){return!0}});var n=new wr({x:e.position[0],y:e.position[1],rotation:e.rotation});n.updateTransform(),this._transformGroup=n}return t.prototype.hasBuilder=function(t){return!!JD[t]},t.prototype.add=function(t){JD[t](this.opt,this.axisModel,this.group,this._transformGroup)},t.prototype.getGroup=function(){return this.group},t.innerTextLayout=function(t,e,n){var r,i,o=$r(e-t);return Ur(o)?(i=n>0?"top":"bottom",r="center"):Ur(o-XD)?(i=n>0?"bottom":"top",r="center"):(i="middle",r=o>0&&o<XD?n>0?"right":"left":n>0?"left":"right"),{rotation:o,textAlign:r,textVerticalAlign:i}},t.makeAxisEventDataBase=function(t){var e={componentType:t.mainType,componentIndex:t.componentIndex};return e[t.mainType+"Index"]=t.componentIndex,e},t.isLabelSilent=function(t){var e=t.get("tooltip");return t.get("silent")||!(t.get("triggerEvent")||e&&e.show)},t}(),JD={axisLine:function(t,e,n,r){var i=e.get(["axisLine","show"]);if("auto"===i&&t.handleAutoShown&&(i=t.handleAutoShown("axisLine")),i){var o=e.axis.getExtent(),a=r.transform,l=[o[0],0],u=[o[1],0],c=l[0]>u[0];a&&(S(l,l,a),S(u,u,a));var d=Object(s["m"])({lineCap:"round"},e.getModel(["axisLine","lineStyle"]).getLineStyle()),h=new jm({shape:{x1:l[0],y1:l[1],x2:u[0],y2:u[1]},style:d,strokeContainThreshold:t.strokeContainThreshold||5,silent:!0,z2:1});_g(h.shape,h.style.lineWidth),h.anid="line",n.add(h);var f=e.get(["axisLine","symbol"]);if(null!=f){var p=e.get(["axisLine","symbolSize"]);Object(s["C"])(f)&&(f=[f,f]),(Object(s["C"])(p)||Object(s["z"])(p))&&(p=[p,p]);var m=sv(e.get(["axisLine","symbolOffset"])||0,p),g=p[0],y=p[1];Object(s["k"])([{rotate:t.rotation+Math.PI/2,offset:m[0],r:0},{rotate:t.rotation-Math.PI/2,offset:m[1],r:Math.sqrt((l[0]-u[0])*(l[0]-u[0])+(l[1]-u[1])*(l[1]-u[1]))}],(function(e,r){if("none"!==f[r]&&null!=f[r]){var i=ov(f[r],-g/2,-y/2,g,y,d.stroke,!0),o=e.r+e.offset,a=c?u:l;i.attr({rotation:e.rotate,x:a[0]+o*Math.cos(t.rotation),y:a[1]-o*Math.sin(t.rotation),silent:!0,z2:11}),n.add(i)}}))}}},axisTickLabel:function(t,e,n,r){var i=iO(n,r,e,t),o=aO(n,r,e,t);if(QD(e,o,i),oO(n,r,e,t.tickDirection),e.get(["axisLabel","hideOverlap"])){var a=ox(Object(s["H"])(o,(function(t){return{label:t,priority:t.z2,defaultAttr:{ignore:t.ignore}}})));ux(a)}},axisName:function(t,e,n,r){var i=Object(s["O"])(t.axisName,e.get("name"));if(i){var o,a,l=e.get("nameLocation"),u=t.nameDirection,c=e.getModel("nameTextStyle"),d=e.get("nameGap")||0,h=e.axis.getExtent(),f=h[0]>h[1]?-1:1,p=["start"===l?h[0]-f*d:"end"===l?h[1]+f*d:(h[0]+h[1])/2,nO(l)?t.labelOffset+u*d:0],m=e.get("nameRotate");null!=m&&(m=m*XD/180),nO(l)?o=ZD.innerTextLayout(t.rotation,null!=m?m:t.rotation,u):(o=KD(t.rotation,l,m||0,h),a=t.axisNameAvailableWidth,null!=a&&(a=Math.abs(a/Math.sin(o.rotation)),!isFinite(a)&&(a=null)));var g=c.getFont(),y=e.get("nameTruncate",!0)||{},v=y.ellipsis,_=Object(s["O"])(t.nameTruncateMaxWidth,y.maxWidth,a),b=new ks({x:p[0],y:p[1],rotation:o.rotation,silent:ZD.isLabelSilent(e),style:nu(c,{text:i,font:g,overflow:"truncate",width:_,ellipsis:v,fill:c.getTextColor()||e.get(["axisLine","lineStyle","color"]),align:c.get("align")||o.textAlign,verticalAlign:c.get("verticalAlign")||o.textVerticalAlign}),z2:1});if(Pg({el:b,componentModel:e,itemName:i}),b.__fullText=i,b.anid="name",e.get("triggerEvent")){var x=ZD.makeAxisEventDataBase(e);x.targetType="axisName",x.name=i,Ls(b).eventData=x}r.add(b),b.updateTransform(),n.add(b),b.decomposeTransform()}}};function KD(t,e,n,r){var i,o,a=$r(n-t),s=r[0]>r[1],l="start"===e&&!s||"start"!==e&&s;return Ur(a-XD/2)?(o=l?"bottom":"top",i="center"):Ur(a-1.5*XD)?(o=l?"top":"bottom",i="center"):(o="middle",i=a<1.5*XD&&a>XD/2?l?"left":"right":l?"right":"left"),{rotation:a,textAlign:i,textVerticalAlign:o}}function QD(t,e,n){if(!sD(t.axis)){var r=t.get(["axisLabel","showMinLabel"]),i=t.get(["axisLabel","showMaxLabel"]);e=e||[],n=n||[];var o=e[0],a=e[1],s=e[e.length-1],l=e[e.length-2],u=n[0],c=n[1],d=n[n.length-1],h=n[n.length-2];!1===r?(tO(o),tO(u)):eO(o,a)&&(r?(tO(a),tO(c)):(tO(o),tO(u))),!1===i?(tO(s),tO(d)):eO(l,s)&&(i?(tO(l),tO(h)):(tO(s),tO(d)))}}function tO(t){t&&(t.ignore=!0)}function eO(t,e){var n=t&&t.getBoundingRect().clone(),r=e&&e.getBoundingRect().clone();if(n&&r){var i=lt([]);return ht(i,i,-t.rotation),n.applyTransform(ct([],i,t.getLocalTransform())),r.applyTransform(ct([],i,e.getLocalTransform())),n.intersect(r)}}function nO(t){return"middle"===t||"center"===t}function rO(t,e,n,r,i){for(var o=[],a=[],s=[],l=0;l<t.length;l++){var u=t[l].coord;a[0]=u,a[1]=0,s[0]=u,s[1]=n,e&&(S(a,a,e),S(s,s,e));var c=new jm({shape:{x1:a[0],y1:a[1],x2:s[0],y2:s[1]},style:r,z2:2,autoBatch:!0,silent:!0});_g(c.shape,c.style.lineWidth),c.anid=i+"_"+t[l].tickValue,o.push(c)}return o}function iO(t,e,n,r){var i=n.axis,o=n.getModel("axisTick"),a=o.get("show");if("auto"===a&&r.handleAutoShown&&(a=r.handleAutoShown("axisTick")),a&&!i.scale.isBlank()){for(var l=o.getModel("lineStyle"),u=r.tickDirection*o.get("length"),c=i.getTicksCoords(),d=rO(c,e.transform,u,Object(s["i"])(l.getLineStyle(),{stroke:n.get(["axisLine","lineStyle","color"])}),"ticks"),h=0;h<d.length;h++)t.add(d[h]);return d}}function oO(t,e,n,r){var i=n.axis,o=n.getModel("minorTick");if(o.get("show")&&!i.scale.isBlank()){var a=i.getMinorTicksCoords();if(a.length)for(var l=o.getModel("lineStyle"),u=r*o.get("length"),c=Object(s["i"])(l.getLineStyle(),Object(s["i"])(n.getModel("axisTick").getLineStyle(),{stroke:n.get(["axisLine","lineStyle","color"])})),d=0;d<a.length;d++)for(var h=rO(a[d],e.transform,u,c,"minorticks_"+d),f=0;f<h.length;f++)t.add(h[f])}}function aO(t,e,n,r){var i=n.axis,o=Object(s["O"])(r.axisLabelShow,n.get(["axisLabel","show"]));if(o&&!i.scale.isBlank()){var a=n.getModel("axisLabel"),l=a.get("margin"),u=i.getViewLabels(),c=(Object(s["O"])(r.labelRotate,a.get("rotate"))||0)*XD/180,d=ZD.innerTextLayout(r.rotation,c,r.labelDirection),h=n.getCategories&&n.getCategories(!0),f=[],p=ZD.isLabelSilent(n),m=n.get("triggerEvent");return Object(s["k"])(u,(function(o,c){var g="ordinal"===i.scale.type?i.scale.getRawOrdinalNumber(o.tickValue):o.tickValue,y=o.formattedLabel,v=o.rawLabel,_=a;if(h&&h[g]){var b=h[g];Object(s["A"])(b)&&b.textStyle&&(_=new Lu(b.textStyle,a,n.ecModel))}var x=_.getTextColor()||n.get(["axisLine","lineStyle","color"]),w=i.dataToCoord(g),M=_.getShallow("align",!0)||d.textAlign,S=Object(s["P"])(_.getShallow("alignMinLabel",!0),M),k=Object(s["P"])(_.getShallow("alignMaxLabel",!0),M),L=_.getShallow("verticalAlign",!0)||_.getShallow("baseline",!0)||d.textVerticalAlign,T=Object(s["P"])(_.getShallow("verticalAlignMinLabel",!0),L),D=Object(s["P"])(_.getShallow("verticalAlignMaxLabel",!0),L),O=new ks({x:w,y:r.labelOffset+r.labelDirection*l,rotation:d.rotation,silent:p,z2:10+(o.level||0),style:nu(_,{text:y,align:0===c?S:c===u.length-1?k:M,verticalAlign:0===c?T:c===u.length-1?D:L,fill:Object(s["w"])(x)?x("category"===i.type?v:"value"===i.type?g+"":g,c):x})});if(O.anid="label_"+g,m){var I=ZD.makeAxisEventDataBase(n);I.targetType="axisLabel",I.value=v,I.tickIndex=c,"category"===i.type&&(I.dataIndex=g),Ls(O).eventData=I}e.add(O),O.updateTransform(),f.push(O),t.add(O),O.decomposeTransform()})),f}}var sO=ZD;function lO(t,e){var n={axesInfo:{},seriesInvolved:!1,coordSysAxesInfo:{},coordSysMap:{}};return uO(n,t,e),n.seriesInvolved&&dO(n,t),n}function uO(t,e,n){var r=e.getComponent("tooltip"),i=e.getComponent("axisPointer"),o=i.get("link",!0)||[],a=[];Object(s["k"])(n.getCoordinateSystems(),(function(n){if(n.axisPointerEnabled){var l=vO(n.model),u=t.coordSysAxesInfo[l]={};t.coordSysMap[l]=n;var c=n.model,d=c.getModel("tooltip",r);if(Object(s["k"])(n.getAxes(),Object(s["h"])(m,!1,null)),n.getTooltipAxes&&r&&d.get("show")){var h="axis"===d.get("trigger"),f="cross"===d.get(["axisPointer","type"]),p=n.getTooltipAxes(d.get(["axisPointer","axis"]));(h||f)&&Object(s["k"])(p.baseAxes,Object(s["h"])(m,!f||"cross",h)),f&&Object(s["k"])(p.otherAxes,Object(s["h"])(m,"cross",!1))}}function m(r,s,l){var c=l.model.getModel("axisPointer",i),h=c.get("show");if(h&&("auto"!==h||r||yO(c))){null==s&&(s=c.get("triggerTooltip")),c=r?cO(l,d,i,e,r,s):c;var f=c.get("snap"),p=c.get("triggerEmphasis"),m=vO(l.model),g=s||f||"category"===l.type,y=t.axesInfo[m]={key:m,axis:l,coordSys:n,axisPointerModel:c,triggerTooltip:s,triggerEmphasis:p,involveSeries:g,snap:f,useHandle:yO(c),seriesModels:[],linkGroup:null};u[m]=y,t.seriesInvolved=t.seriesInvolved||g;var v=hO(o,l);if(null!=v){var _=a[v]||(a[v]={axesInfo:{}});_.axesInfo[m]=y,_.mapper=o[v].mapper,y.linkGroup=_}}}}))}function cO(t,e,n,r,i,o){var a=e.getModel("axisPointer"),l=["type","snap","lineStyle","shadowStyle","label","animation","animationDurationUpdate","animationEasingUpdate","z"],u={};Object(s["k"])(l,(function(t){u[t]=Object(s["d"])(a.get(t))})),u.snap="category"!==t.type&&!!o,"cross"===a.get("type")&&(u.type="line");var c=u.label||(u.label={});if(null==c.show&&(c.show=!1),"cross"===i){var d=a.get(["label","show"]);if(c.show=null==d||d,!o){var h=u.lineStyle=a.get("crossStyle");h&&Object(s["i"])(c,h.textStyle)}}return t.model.getModel("axisPointer",new Lu(u,n,r))}function dO(t,e){e.eachSeries((function(e){var n=e.coordinateSystem,r=e.get(["tooltip","trigger"],!0),i=e.get(["tooltip","show"],!0);n&&"none"!==r&&!1!==r&&"item"!==r&&!1!==i&&!1!==e.get(["axisPointer","show"],!0)&&Object(s["k"])(t.coordSysAxesInfo[vO(n.model)],(function(t){var r=t.axis;n.getAxis(r.dim)===r&&(t.seriesModels.push(e),null==t.seriesDataCount&&(t.seriesDataCount=0),t.seriesDataCount+=e.getData().count())}))}))}function hO(t,e){for(var n=e.model,r=e.dim,i=0;i<t.length;i++){var o=t[i]||{};if(fO(o[r+"AxisId"],n.id)||fO(o[r+"AxisIndex"],n.componentIndex)||fO(o[r+"AxisName"],n.name))return i}}function fO(t,e){return"all"===t||Object(s["t"])(t)&&Object(s["r"])(t,e)>=0||t===e}function pO(t){var e=mO(t);if(e){var n=e.axisPointerModel,r=e.axis.scale,i=n.option,o=n.get("status"),a=n.get("value");null!=a&&(a=r.parse(a));var s=yO(n);null==o&&(i.status=s?"show":"hide");var l=r.getExtent().slice();l[0]>l[1]&&l.reverse(),(null==a||a>l[1])&&(a=l[1]),a<l[0]&&(a=l[0]),i.value=a,s&&(i.status=e.axis.scale.isBlank()?"hide":"show")}}function mO(t){var e=(t.ecModel.getComponent("axisPointer")||{}).coordSysAxesInfo;return e&&e.axesInfo[vO(t)]}function gO(t){var e=mO(t);return e&&e.axisPointerModel}function yO(t){return!!t.get(["handle","show"])}function vO(t){return t.type+"||"+t.id}var _O={},bO=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(e,n,r,i){this.axisPointerClass&&pO(e),t.prototype.render.apply(this,arguments),this._doUpdateAxisPointerClass(e,r,!0)},e.prototype.updateAxisPointer=function(t,e,n,r){this._doUpdateAxisPointerClass(t,n,!1)},e.prototype.remove=function(t,e){var n=this._axisPointer;n&&n.remove(e)},e.prototype.dispose=function(e,n){this._disposeAxisPointer(n),t.prototype.dispose.apply(this,arguments)},e.prototype._doUpdateAxisPointerClass=function(t,n,r){var i=e.getAxisPointerClass(this.axisPointerClass);if(i){var o=gO(t);o?(this._axisPointer||(this._axisPointer=new i)).render(t,o,n,r):this._disposeAxisPointer(n)}},e.prototype._disposeAxisPointer=function(t){this._axisPointer&&this._axisPointer.dispose(t),this._axisPointer=null},e.registerAxisPointerClass=function(t,e){_O[t]=e},e.getAxisPointerClass=function(t){return t&&_O[t]},e.type="axis",e}(kp),xO=bO,wO=Ci();function MO(t,e,n,r){var i=n.axis;if(!i.scale.isBlank()){var o=n.getModel("splitArea"),a=o.getModel("areaStyle"),l=a.get("color"),u=r.coordinateSystem.getRect(),c=i.getTicksCoords({tickModel:o,clamp:!0});if(c.length){var d=l.length,h=wO(t).splitAreaColors,f=s["f"](),p=0;if(h)for(var m=0;m<c.length;m++){var g=h.get(c[m].tickValue);if(null!=g){p=(g+(d-1)*m)%d;break}}var y=i.toGlobalCoord(c[0].coord),v=a.getAreaStyle();l=s["t"](l)?l:[l];for(m=1;m<c.length;m++){var _=i.toGlobalCoord(c[m].coord),b=void 0,x=void 0,w=void 0,M=void 0;i.isHorizontal()?(b=y,x=u.y,w=_-b,M=u.height,y=b+w):(b=u.x,x=y,w=u.width,M=_-x,y=x+M);var S=c[m-1].tickValue;null!=S&&f.set(S,p),e.add(new ss({anid:null!=S?"area_"+S:null,shape:{x:b,y:x,width:w,height:M},style:s["i"]({fill:l[p]},v),autoBatch:!0,silent:!0})),p=(p+1)%d}wO(t).splitAreaColors=f}}}function SO(t){wO(t).splitAreaColors=null}var kO=["axisLine","axisTickLabel","axisName"],LO=["splitArea","splitLine","minorSplitLine"],TO=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.axisPointerClass="CartesianAxisPointer",n}return o(e,t),e.prototype.render=function(e,n,r,i){this.group.removeAll();var o=this._axisGroup;if(this._axisGroup=new wr,this.group.add(this._axisGroup),e.get("show")){var a=e.getCoordSysModel(),l=RD(a,e),u=new sO(e,s["m"]({handleAutoShown:function(t){for(var n=a.coordinateSystem.getCartesians(),r=0;r<n.length;r++)if(uT(n[r].getOtherAxis(e.axis).scale))return!0;return!1}},l));s["k"](kO,u.add,u),this._axisGroup.add(u.getGroup()),s["k"](LO,(function(t){e.get([t,"show"])&&DO[t](this,this._axisGroup,e,a)}),this);var c=i&&"changeAxisOrder"===i.type&&i.isInitSort;c||Tg(o,this._axisGroup,e),t.prototype.render.call(this,e,n,r,i)}},e.prototype.remove=function(){SO(this)},e.type="cartesianAxis",e}(xO),DO={splitLine:function(t,e,n,r){var i=n.axis;if(!i.scale.isBlank()){var o=n.getModel("splitLine"),a=o.getModel("lineStyle"),l=a.get("color");l=s["t"](l)?l:[l];for(var u=r.coordinateSystem.getRect(),c=i.isHorizontal(),d=0,h=i.getTicksCoords({tickModel:o}),f=[],p=[],m=a.getLineStyle(),g=0;g<h.length;g++){var y=i.toGlobalCoord(h[g].coord);c?(f[0]=y,f[1]=u.y,p[0]=y,p[1]=u.y+u.height):(f[0]=u.x,f[1]=y,p[0]=u.x+u.width,p[1]=y);var v=d++%l.length,_=h[g].tickValue,b=new jm({anid:null!=_?"line_"+h[g].tickValue:null,autoBatch:!0,shape:{x1:f[0],y1:f[1],x2:p[0],y2:p[1]},style:s["i"]({stroke:l[v]},m),silent:!0});_g(b.shape,m.lineWidth),e.add(b)}}},minorSplitLine:function(t,e,n,r){var i=n.axis,o=n.getModel("minorSplitLine"),a=o.getModel("lineStyle"),s=r.coordinateSystem.getRect(),l=i.isHorizontal(),u=i.getMinorTicksCoords();if(u.length)for(var c=[],d=[],h=a.getLineStyle(),f=0;f<u.length;f++)for(var p=0;p<u[f].length;p++){var m=i.toGlobalCoord(u[f][p].coord);l?(c[0]=m,c[1]=s.y,d[0]=m,d[1]=s.y+s.height):(c[0]=s.x,c[1]=m,d[0]=s.x+s.width,d[1]=m);var g=new jm({anid:"minor_line_"+u[f][p].tickValue,autoBatch:!0,shape:{x1:c[0],y1:c[1],x2:d[0],y2:d[1]},style:h,silent:!0});_g(g.shape,h.lineWidth),e.add(g)}},splitArea:function(t,e,n,r){MO(t,e,n,r)}},OO=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="xAxis",e}(TO),IO=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.type=OO.type,e}return o(e,t),e.type="yAxis",e}(TO),CO=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.type="grid",e}return o(e,t),e.prototype.render=function(t,e){this.group.removeAll(),t.get("show")&&this.group.add(new ss({shape:t.coordinateSystem.getRect(),style:Object(s["i"])({fill:t.get("backgroundColor")},t.getItemStyle()),silent:!0,z2:-1}))},e.type="grid",e}(kp),AO={offset:0};function YO(t){t.registerComponentView(CO),t.registerComponentModel(GL),t.registerCoordinateSystem("cartesian2d",qD),oT(t,"x",UL,AO),oT(t,"y",UL,AO),t.registerComponentView(OO),t.registerComponentView(IO),t.registerPreprocessor((function(t){t.xAxis&&t.yAxis&&!t.grid&&(t.grid={})}))}function jO(t){jb(YO),t.registerSeriesModel(EL),t.registerChartView(WL),t.registerLayout(uk("scatter"))}function PO(t){t.eachSeriesByType("radar",(function(t){var e=t.getData(),n=[],r=t.coordinateSystem;if(r){var i=r.getIndicatorAxes();s["k"](i,(function(t,o){e.each(e.mapDimension(i[o].dim),(function(t,e){n[e]=n[e]||[];var i=r.dataToPoint(t,o);n[e][o]=EO(i)?i:NO(r)}))})),e.each((function(t){var i=s["o"](n[t],(function(t){return EO(t)}))||NO(r);n[t].push(i.slice()),e.setItemLayout(t,n[t])}))}}))}function EO(t){return!isNaN(t[0])&&!isNaN(t[1])}function NO(t){return[t.cx,t.cy]}function RO(t){var e=t.polar;if(e){s["t"](e)||(e=[e]);var n=[];s["k"](e,(function(e,r){e.indicator?(e.type&&!e.shape&&(e.shape=e.type),t.radar=t.radar||[],s["t"](t.radar)||(t.radar=[t.radar]),t.radar.push(e)):n.push(e)})),t.polar=n}s["k"](t.series,(function(t){t&&"radar"===t.type&&t.polarIndex&&(t.radarIndex=t.polarIndex)}))}var HO=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){var i=t.coordinateSystem,o=this.group,a=t.getData(),l=this._data;function u(t,e){var n=t.getItemVisual(e,"symbol")||"circle";if("none"!==n){var r=av(t.getItemVisual(e,"symbolSize")),i=ov(n,-1,-1,2,2),o=t.getItemVisual(e,"symbolRotate")||0;return i.attr({style:{strokeNoScale:!0},z2:100,scaleX:r[0]/2,scaleY:r[1]/2,rotation:o*Math.PI/180||0}),i}}function c(e,n,i,o,a,s){i.removeAll();for(var l=0;l<n.length-1;l++){var c=u(o,a);c&&(c.__dimIdx=l,e[l]?(c.setPosition(e[l]),r[s?"initProps":"updateProps"](c,{x:n[l][0],y:n[l][1]},t,a)):c.setPosition(n[l]),i.add(c))}}function d(t){return s["H"](t,(function(t){return[i.cx,i.cy]}))}a.diff(l).add((function(e){var n=a.getItemLayout(e);if(n){var r=new Tm,i=new Im,o={shape:{points:n}};r.shape.points=d(n),i.shape.points=d(n),Vl(r,o,t,e),Vl(i,o,t,e);var s=new wr,l=new wr;s.add(i),s.add(r),s.add(l),c(i.shape.points,n,l,a,e,!0),a.setItemGraphicEl(e,s)}})).update((function(e,n){var r=l.getItemGraphicEl(n),i=r.childAt(0),o=r.childAt(1),s=r.childAt(2),u={shape:{points:a.getItemLayout(e)}};u.shape.points&&(c(i.shape.points,u.shape.points,s,a,e,!1),Xl(o),Xl(i),Wl(i,u,t),Wl(o,u,t),a.setItemGraphicEl(e,r))})).remove((function(t){o.remove(l.getItemGraphicEl(t))})).execute(),a.eachItemGraphicEl((function(t,e){var n=a.getItemModel(e),r=t.childAt(0),i=t.childAt(1),l=t.childAt(2),u=a.getItemVisual(e,"style"),c=u.fill;o.add(t),r.useStyle(s["i"](n.getModel("lineStyle").getLineStyle(),{fill:"none",stroke:c})),Al(r,n,"lineStyle"),Al(i,n,"areaStyle");var d=n.getModel("areaStyle"),h=d.isEmpty()&&d.parentModel.isEmpty();i.ignore=h,s["k"](["emphasis","select","blur"],(function(t){var e=n.getModel([t,"areaStyle"]),r=e.isEmpty()&&e.parentModel.isEmpty();i.ensureState(t).ignore=r&&h})),i.useStyle(s["i"](d.getAreaStyle(),{fill:c,opacity:.7,decal:u.decal}));var f=n.getModel("emphasis"),p=f.getModel("itemStyle").getItemStyle();l.eachChild((function(t){if(t instanceof Ka){var r=t.style;t.useStyle(s["m"]({image:r.image,x:r.x,y:r.y,width:r.width,height:r.height},u))}else t.useStyle(u),t.setColor(c),t.style.strokeNoScale=!0;var i=t.ensureState("emphasis");i.style=s["d"](p);var o=a.getStore().get(a.getDimensionIndex(t.__dimIdx),e);(null==o||isNaN(o))&&(o=""),tu(t,eu(n),{labelFetcher:a.hostModel,labelDataIndex:e,labelDimIndex:t.__dimIdx,defaultText:o,inheritColor:c,defaultOpacity:u.opacity})})),Dl(t,f.get("focus"),f.get("blurScope"),f.get("disabled"))})),this._data=a},e.prototype.remove=function(){this.group.removeAll(),this._data=null},e.type="radar",e}($g),zO=HO,FO=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.hasSymbolVisual=!0,n}return o(e,t),e.prototype.init=function(e){t.prototype.init.apply(this,arguments),this.legendVisualProvider=new OL(s["c"](this.getData,this),s["c"](this.getRawData,this))},e.prototype.getInitialData=function(t,e){return TL(this,{generateCoord:"indicator_",generateCoordCount:1/0})},e.prototype.formatTooltip=function(t,e,n){var r=this.getData(),i=this.coordinateSystem,o=i.getIndicatorAxes(),a=this.getData().getName(t),l=""===a?this.name:a,u=ap(this,t);return Uf("section",{header:l,sortBlocks:!0,blocks:s["H"](o,(function(e){var n=r.get(r.mapDimension(e.dim),t);return Uf("nameValue",{markerType:"subItem",markerColor:u,name:e.name,value:n,sortParam:n})}))})},e.prototype.getTooltipPosition=function(t){if(null!=t)for(var e=this.getData(),n=this.coordinateSystem,r=e.getValues(s["H"](n.dimensions,(function(t){return e.mapDimension(t)})),t),i=0,o=r.length;i<o;i++)if(!isNaN(r[i])){var a=n.getIndicatorAxes();return n.coordToPoint(a[i].dataToCoord(r[i]),i)}},e.type="series.radar",e.dependencies=["radar"],e.defaultOption={z:2,colorBy:"data",coordinateSystem:"radar",legendHoverLink:!0,radarIndex:0,lineStyle:{width:2,type:"solid",join:"round"},label:{position:"top"},symbolSize:8},e}(Mp),BO=FO,WO=QL.value;function VO(t,e){return s["i"]({show:e},t)}var GO=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.optionUpdated=function(){var t=this.get("boundaryGap"),e=this.get("splitNumber"),n=this.get("scale"),r=this.get("axisLine"),i=this.get("axisTick"),o=this.get("axisLabel"),a=this.get("axisName"),l=this.get(["axisName","show"]),u=this.get(["axisName","formatter"]),c=this.get("axisNameGap"),d=this.get("triggerEvent"),h=s["H"](this.get("indicator")||[],(function(h){null!=h.max&&h.max>0&&!h.min?h.min=0:null!=h.min&&h.min<0&&!h.max&&(h.max=0);var f=a;null!=h.color&&(f=s["i"]({color:h.color},a));var p=s["I"](s["d"](h),{boundaryGap:t,splitNumber:e,scale:n,axisLine:r,axisTick:i,axisLabel:o,name:h.text,showName:l,nameLocation:"end",nameGap:c,nameTextStyle:f,triggerEvent:d},!1);if(s["C"](u)){var m=p.name;p.name=u.replace("{value}",null!=m?m:"")}else s["w"](u)&&(p.name=u(p.name,p));var g=new Lu(p,null,this.ecModel);return s["K"](g,$L.prototype),g.mainType="radar",g.componentIndex=this.componentIndex,g}),this);this._indicatorModels=h},e.prototype.getIndicatorModels=function(){return this._indicatorModels},e.type="radar",e.defaultOption={z:0,center:["50%","50%"],radius:"75%",startAngle:90,axisName:{show:!0},boundaryGap:[0,0],splitNumber:5,axisNameGap:15,scale:!1,shape:"polygon",axisLine:s["I"]({lineStyle:{color:"#bbb"}},WO.axisLine),axisLabel:VO(WO.axisLabel,!1),axisTick:VO(WO.axisTick,!1),splitLine:VO(WO.splitLine,!0),splitArea:VO(WO.splitArea,!0),indicator:[]},e}(Xc),$O=GO,UO=["axisLine","axisTickLabel","axisName"],qO=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){var r=this.group;r.removeAll(),this._buildAxes(t),this._buildSplitLineAndArea(t)},e.prototype._buildAxes=function(t){var e=t.coordinateSystem,n=e.getIndicatorAxes(),r=s["H"](n,(function(t){var n=t.model.get("showName")?t.name:"",r=new sO(t.model,{axisName:n,position:[e.cx,e.cy],rotation:t.angle,labelDirection:-1,tickDirection:-1,nameDirection:1});return r}));s["k"](r,(function(t){s["k"](UO,t.add,t),this.group.add(t.getGroup())}),this)},e.prototype._buildSplitLineAndArea=function(t){var e=t.coordinateSystem,n=e.getIndicatorAxes();if(n.length){var r=t.get("shape"),i=t.getModel("splitLine"),o=t.getModel("splitArea"),a=i.getModel("lineStyle"),l=o.getModel("areaStyle"),u=i.get("show"),c=o.get("show"),d=a.get("color"),h=l.get("color"),f=s["t"](d)?d:[d],p=s["t"](h)?h:[h],m=[],g=[];if("circle"===r)for(var y=n[0].getTicksCoords(),v=e.cx,_=e.cy,b=0;b<y.length;b++){if(u){var x=O(m,f,b);m[x].push(new Kp({shape:{cx:v,cy:_,r:y[b].coord}}))}if(c&&b<y.length-1){x=O(g,p,b);g[x].push(new wm({shape:{cx:v,cy:_,r0:y[b].coord,r:y[b+1].coord}}))}}else{var w,M=s["H"](n,(function(t,n){var r=t.getTicksCoords();return w=null==w?r.length-1:Math.min(r.length-1,w),s["H"](r,(function(t){return e.coordToPoint(t.coord,n)}))})),S=[];for(b=0;b<=w;b++){for(var k=[],L=0;L<n.length;L++)k.push(M[L][b]);if(k[0]&&k.push(k[0].slice()),u){x=O(m,f,b);m[x].push(new Im({shape:{points:k}}))}if(c&&S){x=O(g,p,b-1);g[x].push(new Tm({shape:{points:k.concat(S)}}))}S=k.slice().reverse()}}var T=a.getLineStyle(),D=l.getAreaStyle();s["k"](g,(function(t,e){this.group.add(yg(t,{style:s["i"]({stroke:"none",fill:p[e%p.length]},D),silent:!0}))}),this),s["k"](m,(function(t,e){this.group.add(yg(t,{style:s["i"]({fill:"none",stroke:f[e%f.length]},T),silent:!0}))}),this)}function O(t,e,n){var r=n%e.length;return t[r]=t[r]||[],r}},e.type="radar",e}(kp),XO=qO,ZO=function(t){function e(e,n,r){var i=t.call(this,e,n,r)||this;return i.type="value",i.angle=0,i.name="",i}return o(e,t),e}(PD),JO=ZO,KO=function(){function t(t,e,n){this.dimensions=[],this._model=t,this._indicatorAxes=Object(s["H"])(t.getIndicatorModels(),(function(t,e){var n="indicator_"+e,r=new JO(n,new wT);return r.name=t.get("name"),r.model=t,t.axis=r,this.dimensions.push(n),r}),this),this.resize(t,n)}return t.prototype.getIndicatorAxes=function(){return this._indicatorAxes},t.prototype.dataToPoint=function(t,e){var n=this._indicatorAxes[e];return this.coordToPoint(n.dataToCoord(t),e)},t.prototype.coordToPoint=function(t,e){var n=this._indicatorAxes[e],r=n.angle,i=this.cx+t*Math.cos(r),o=this.cy-t*Math.sin(r);return[i,o]},t.prototype.pointToData=function(t){var e=t[0]-this.cx,n=t[1]-this.cy,r=Math.sqrt(e*e+n*n);e/=r,n/=r;for(var i,o=Math.atan2(-n,e),a=1/0,s=-1,l=0;l<this._indicatorAxes.length;l++){var u=this._indicatorAxes[l],c=Math.abs(o-u.angle);c<a&&(i=u,s=l,a=c)}return[s,+(i&&i.coordToData(r))]},t.prototype.resize=function(t,e){var n=t.get("center"),r=e.getWidth(),i=e.getHeight(),o=Math.min(r,i)/2;this.cx=Nr(n[0],r),this.cy=Nr(n[1],i),this.startAngle=t.get("startAngle")*Math.PI/180;var a=t.get("radius");(Object(s["C"])(a)||Object(s["z"])(a))&&(a=[0,a]),this.r0=Nr(a[0],o),this.r=Nr(a[1],o),Object(s["k"])(this._indicatorAxes,(function(t,e){t.setExtent(this.r0,this.r);var n=this.startAngle+e*Math.PI*2/this._indicatorAxes.length;n=Math.atan2(Math.sin(n),Math.cos(n)),t.angle=n}),this)},t.prototype.update=function(t,e){var n=this._indicatorAxes,r=this._model;Object(s["k"])(n,(function(t){t.scale.setExtent(1/0,-1/0)})),t.eachSeriesByType("radar",(function(e,i){if("radar"===e.get("coordinateSystem")&&t.getComponent("radar",e.get("radarIndex"))===r){var o=e.getData();Object(s["k"])(n,(function(t){t.scale.unionExtentFromData(o,o.mapDimension(t.dim))}))}}),this);var i=r.get("splitNumber"),o=new wT;o.setExtent(0,i),o.setInterval(1),Object(s["k"])(n,(function(t,e){BD(t.scale,t.model,o)}))},t.prototype.convertToPixel=function(t,e,n){return null},t.prototype.convertFromPixel=function(t,e,n){return null},t.prototype.containPoint=function(t){return!1},t.create=function(e,n){var r=[];return e.eachComponent("radar",(function(i){var o=new t(i,e,n);r.push(o),i.coordinateSystem=o})),e.eachSeriesByType("radar",(function(t){"radar"===t.get("coordinateSystem")&&(t.coordinateSystem=r[t.get("radarIndex")||0])})),r},t.dimensions=[],t}(),QO=KO;function tI(t){t.registerCoordinateSystem("radar",QO),t.registerComponentModel($O),t.registerComponentView(XO),t.registerVisual({seriesType:"radar",reset:function(t){var e=t.getData();e.each((function(t){e.setItemVisual(t,"legendIcon","roundRect")})),e.setVisual("legendIcon","roundRect")}})}function eI(t){jb(tI),t.registerChartView(zO),t.registerSeriesModel(BO),t.registerLayout(PO),t.registerProcessor(yL("radar")),t.registerPreprocessor(RO)}var nI="\0_ec_interaction_mutex";function rI(t,e,n){var r=aI(t);r[e]=n}function iI(t,e,n){var r=aI(t),i=r[e];i===n&&(r[e]=null)}function oI(t,e){return!!aI(t)[e]}function aI(t){return t[nI]||(t[nI]={})}Mb({type:"takeGlobalCursor",event:"globalCursorTaken",update:"update"},s["L"]);var sI=function(t){function e(e){var n=t.call(this)||this;n._zr=e;var r=Object(s["c"])(n._mousedownHandler,n),i=Object(s["c"])(n._mousemoveHandler,n),o=Object(s["c"])(n._mouseupHandler,n),a=Object(s["c"])(n._mousewheelHandler,n),l=Object(s["c"])(n._pinchHandler,n);return n.enable=function(t,n){this.disable(),this._opt=Object(s["i"])(Object(s["d"])(n)||{},{zoomOnMouseWheel:!0,moveOnMouseMove:!0,moveOnMouseWheel:!1,preventDefaultMouseMove:!0}),null==t&&(t=!0),!0!==t&&"move"!==t&&"pan"!==t||(e.on("mousedown",r),e.on("mousemove",i),e.on("mouseup",o)),!0!==t&&"scale"!==t&&"zoom"!==t||(e.on("mousewheel",a),e.on("pinch",l))},n.disable=function(){e.off("mousedown",r),e.off("mousemove",i),e.off("mouseup",o),e.off("mousewheel",a),e.off("pinch",l)},n}return o(e,t),e.prototype.isDragging=function(){return this._dragging},e.prototype.isPinching=function(){return this._pinching},e.prototype.setPointerChecker=function(t){this.pointerChecker=t},e.prototype.dispose=function(){this.disable()},e.prototype._mousedownHandler=function(t){if(!nt(t)){var e=t.target;while(e){if(e.draggable)return;e=e.__hostTarget||e.parent}var n=t.offsetX,r=t.offsetY;this.pointerChecker&&this.pointerChecker(t,n,r)&&(this._x=n,this._y=r,this._dragging=!0)}},e.prototype._mousemoveHandler=function(t){if(this._dragging&&cI("moveOnMouseMove",t,this._opt)&&"pinch"!==t.gestureEvent&&!oI(this._zr,"globalPan")){var e=t.offsetX,n=t.offsetY,r=this._x,i=this._y,o=e-r,a=n-i;this._x=e,this._y=n,this._opt.preventDefaultMouseMove&&et(t.event),uI(this,"pan","moveOnMouseMove",t,{dx:o,dy:a,oldX:r,oldY:i,newX:e,newY:n,isAvailableBehavior:null})}},e.prototype._mouseupHandler=function(t){nt(t)||(this._dragging=!1)},e.prototype._mousewheelHandler=function(t){var e=cI("zoomOnMouseWheel",t,this._opt),n=cI("moveOnMouseWheel",t,this._opt),r=t.wheelDelta,i=Math.abs(r),o=t.offsetX,a=t.offsetY;if(0!==r&&(e||n)){if(e){var s=i>3?1.4:i>1?1.2:1.1,l=r>0?s:1/s;lI(this,"zoom","zoomOnMouseWheel",t,{scale:l,originX:o,originY:a,isAvailableBehavior:null})}if(n){var u=Math.abs(r),c=(r>0?1:-1)*(u>3?.4:u>1?.15:.05);lI(this,"scrollMove","moveOnMouseWheel",t,{scrollDelta:c,originX:o,originY:a,isAvailableBehavior:null})}}},e.prototype._pinchHandler=function(t){if(!oI(this._zr,"globalPan")){var e=t.pinchScale>1?1.1:1/1.1;lI(this,"zoom",null,t,{scale:e,originX:t.pinchX,originY:t.pinchY,isAvailableBehavior:null})}},e}(C);function lI(t,e,n,r,i){t.pointerChecker&&t.pointerChecker(r,i.originX,i.originY)&&(et(r.event),uI(t,e,n,r,i))}function uI(t,e,n,r,i){i.isAvailableBehavior=Object(s["c"])(cI,null,n,r),t.trigger(e,i)}function cI(t,e,n){var r=n[t];return!t||r&&(!Object(s["C"])(r)||e.event[r+"Key"])}var dI=sI;function hI(t,e,n){var r=t.target;r.x+=e,r.y+=n,r.dirty()}function fI(t,e,n,r){var i=t.target,o=t.zoomLimit,a=t.zoom=t.zoom||1;if(a*=e,o){var s=o.min||0,l=o.max||1/0;a=Math.max(Math.min(l,a),s)}var u=a/t.zoom;t.zoom=a,i.x-=(n-i.x)*(u-1),i.y-=(r-i.y)*(u-1),i.scaleX*=u,i.scaleY*=u,i.dirty()}var pI,mI={axisPointer:1,tooltip:1,brush:1};function gI(t,e,n){var r=e.getComponentByElement(t.topTarget),i=r&&r.coordinateSystem;return r&&r!==n&&!mI.hasOwnProperty(r.mainType)&&i&&i.model!==n}function yI(t){if(Object(s["C"])(t)){var e=new DOMParser;t=e.parseFromString(t,"text/xml")}var n=t;9===n.nodeType&&(n=n.firstChild);while("svg"!==n.nodeName.toLowerCase()||1!==n.nodeType)n=n.nextSibling;return n}var vI={fill:"fill",stroke:"stroke","stroke-width":"lineWidth",opacity:"opacity","fill-opacity":"fillOpacity","stroke-opacity":"strokeOpacity","stroke-dasharray":"lineDash","stroke-dashoffset":"lineDashOffset","stroke-linecap":"lineCap","stroke-linejoin":"lineJoin","stroke-miterlimit":"miterLimit","font-family":"fontFamily","font-size":"fontSize","font-style":"fontStyle","font-weight":"fontWeight","text-anchor":"textAlign",visibility:"visibility",display:"display"},_I=Object(s["F"])(vI),bI={"alignment-baseline":"textBaseline","stop-color":"stopColor"},xI=Object(s["F"])(bI),wI=function(){function t(){this._defs={},this._root=null}return t.prototype.parse=function(t,e){e=e||{};var n=yI(t);this._defsUsePending=[];var r=new wr;this._root=r;var i=[],o=n.getAttribute("viewBox")||"",a=parseFloat(n.getAttribute("width")||e.width),s=parseFloat(n.getAttribute("height")||e.height);isNaN(a)&&(a=null),isNaN(s)&&(s=null),DI(n,r,null,!0,!1);var l,u,c=n.firstChild;while(c)this._parseNode(c,r,i,null,!1,!1),c=c.nextSibling;if(AI(this._defs,this._defsUsePending),this._defsUsePending=[],o){var d=jI(o);d.length>=4&&(l={x:parseFloat(d[0]||0),y:parseFloat(d[1]||0),width:parseFloat(d[2]),height:parseFloat(d[3])})}if(l&&null!=a&&null!=s&&(u=FI(l,{x:0,y:0,width:a,height:s}),!e.ignoreViewBox)){var h=r;r=new wr,r.add(h),h.scaleX=h.scaleY=u.scale,h.x=u.x,h.y=u.y}return e.ignoreRootClip||null==a||null==s||r.setClipPath(new ss({shape:{x:0,y:0,width:a,height:s}})),{root:r,width:a,height:s,viewBoxRect:l,viewBoxTransform:u,named:i}},t.prototype._parseNode=function(t,e,n,r,i,o){var a,l=t.nodeName.toLowerCase(),u=r;if("defs"===l&&(i=!0),"text"===l&&(o=!0),"defs"===l||"switch"===l)a=e;else{if(!i){var c=pI[l];if(c&&Object(s["q"])(pI,l)){a=c.call(this,t,e);var d=t.getAttribute("name");if(d){var h={name:d,namedFrom:null,svgNodeTagLower:l,el:a};n.push(h),"g"===l&&(u=h)}else r&&n.push({name:r.name,namedFrom:r,svgNodeTagLower:l,el:a});e.add(a)}}var f=MI[l];if(f&&Object(s["q"])(MI,l)){var p=f.call(this,t),m=t.getAttribute("id");m&&(this._defs[m]=p)}}if(a&&a.isGroup){var g=t.firstChild;while(g)1===g.nodeType?this._parseNode(g,a,n,u,i,o):3===g.nodeType&&o&&this._parseText(g,a),g=g.nextSibling}},t.prototype._parseText=function(t,e){var n=new Ua({style:{text:t.textContent},silent:!0,x:this._textX||0,y:this._textY||0});LI(e,n),DI(t,n,this._defsUsePending,!1,!1),OI(n,e);var r=n.style,i=r.fontSize;i&&i<9&&(r.fontSize=9,n.scaleX*=i/9,n.scaleY*=i/9);var o=(r.fontSize||r.fontFamily)&&[r.fontStyle,r.fontWeight,(r.fontSize||12)+"px",r.fontFamily||"sans-serif"].join(" ");r.font=o;var a=n.getBoundingRect();return this._textX+=a.width,e.add(n),n},t.internalField=function(){pI={g:function(t,e){var n=new wr;return LI(e,n),DI(t,n,this._defsUsePending,!1,!1),n},rect:function(t,e){var n=new ss;return LI(e,n),DI(t,n,this._defsUsePending,!1,!1),n.setShape({x:parseFloat(t.getAttribute("x")||"0"),y:parseFloat(t.getAttribute("y")||"0"),width:parseFloat(t.getAttribute("width")||"0"),height:parseFloat(t.getAttribute("height")||"0")}),n.silent=!0,n},circle:function(t,e){var n=new Kp;return LI(e,n),DI(t,n,this._defsUsePending,!1,!1),n.setShape({cx:parseFloat(t.getAttribute("cx")||"0"),cy:parseFloat(t.getAttribute("cy")||"0"),r:parseFloat(t.getAttribute("r")||"0")}),n.silent=!0,n},line:function(t,e){var n=new jm;return LI(e,n),DI(t,n,this._defsUsePending,!1,!1),n.setShape({x1:parseFloat(t.getAttribute("x1")||"0"),y1:parseFloat(t.getAttribute("y1")||"0"),x2:parseFloat(t.getAttribute("x2")||"0"),y2:parseFloat(t.getAttribute("y2")||"0")}),n.silent=!0,n},ellipse:function(t,e){var n=new em;return LI(e,n),DI(t,n,this._defsUsePending,!1,!1),n.setShape({cx:parseFloat(t.getAttribute("cx")||"0"),cy:parseFloat(t.getAttribute("cy")||"0"),rx:parseFloat(t.getAttribute("rx")||"0"),ry:parseFloat(t.getAttribute("ry")||"0")}),n.silent=!0,n},polygon:function(t,e){var n,r=t.getAttribute("points");r&&(n=TI(r));var i=new Tm({shape:{points:n||[]},silent:!0});return LI(e,i),DI(t,i,this._defsUsePending,!1,!1),i},polyline:function(t,e){var n,r=t.getAttribute("points");r&&(n=TI(r));var i=new Im({shape:{points:n||[]},silent:!0});return LI(e,i),DI(t,i,this._defsUsePending,!1,!1),i},image:function(t,e){var n=new Ka;return LI(e,n),DI(t,n,this._defsUsePending,!1,!1),n.setStyle({image:t.getAttribute("xlink:href")||t.getAttribute("href"),x:+t.getAttribute("x"),y:+t.getAttribute("y"),width:+t.getAttribute("width"),height:+t.getAttribute("height")}),n.silent=!0,n},text:function(t,e){var n=t.getAttribute("x")||"0",r=t.getAttribute("y")||"0",i=t.getAttribute("dx")||"0",o=t.getAttribute("dy")||"0";this._textX=parseFloat(n)+parseFloat(i),this._textY=parseFloat(r)+parseFloat(o);var a=new wr;return LI(e,a),DI(t,a,this._defsUsePending,!1,!0),a},tspan:function(t,e){var n=t.getAttribute("x"),r=t.getAttribute("y");null!=n&&(this._textX=parseFloat(n)),null!=r&&(this._textY=parseFloat(r));var i=t.getAttribute("dx")||"0",o=t.getAttribute("dy")||"0",a=new wr;return LI(e,a),DI(t,a,this._defsUsePending,!1,!0),this._textX+=parseFloat(i),this._textY+=parseFloat(o),a},path:function(t,e){var n=t.getAttribute("d")||"",r=$p(n);return LI(e,r),DI(t,r,this._defsUsePending,!1,!1),r.silent=!0,r}}}(),t}(),MI={lineargradient:function(t){var e=parseInt(t.getAttribute("x1")||"0",10),n=parseInt(t.getAttribute("y1")||"0",10),r=parseInt(t.getAttribute("x2")||"10",10),i=parseInt(t.getAttribute("y2")||"0",10),o=new qm(e,n,r,i);return SI(t,o),kI(t,o),o},radialgradient:function(t){var e=parseInt(t.getAttribute("cx")||"0",10),n=parseInt(t.getAttribute("cy")||"0",10),r=parseInt(t.getAttribute("r")||"0",10),i=new Zm(e,n,r);return SI(t,i),kI(t,i),i}};function SI(t,e){var n=t.getAttribute("gradientUnits");"userSpaceOnUse"===n&&(e.global=!0)}function kI(t,e){var n=t.firstChild;while(n){if(1===n.nodeType&&"stop"===n.nodeName.toLocaleLowerCase()){var r=n.getAttribute("offset"),i=void 0;i=r&&r.indexOf("%")>0?parseInt(r,10)/100:r?parseFloat(r):0;var o={};HI(n,o,o);var a=o.stopColor||n.getAttribute("stop-color")||"#000000";e.colorStops.push({offset:i,color:a})}n=n.nextSibling}}function LI(t,e){t&&t.__inheritedStyle&&(e.__inheritedStyle||(e.__inheritedStyle={}),Object(s["i"])(e.__inheritedStyle,t.__inheritedStyle))}function TI(t){for(var e=jI(t),n=[],r=0;r<e.length;r+=2){var i=parseFloat(e[r]),o=parseFloat(e[r+1]);n.push([i,o])}return n}function DI(t,e,n,r,i){var o=e,a=o.__inheritedStyle=o.__inheritedStyle||{},l={};1===t.nodeType&&(NI(t,e),HI(t,a,l),r||zI(t,a,l)),o.style=o.style||{},null!=a.fill&&(o.style.fill=CI(o,"fill",a.fill,n)),null!=a.stroke&&(o.style.stroke=CI(o,"stroke",a.stroke,n)),Object(s["k"])(["lineWidth","opacity","fillOpacity","strokeOpacity","miterLimit","fontSize"],(function(t){null!=a[t]&&(o.style[t]=parseFloat(a[t]))})),Object(s["k"])(["lineDashOffset","lineCap","lineJoin","fontWeight","fontFamily","fontStyle","textAlign"],(function(t){null!=a[t]&&(o.style[t]=a[t])})),i&&(o.__selfStyle=l),a.lineDash&&(o.style.lineDash=Object(s["H"])(jI(a.lineDash),(function(t){return parseFloat(t)}))),"hidden"!==a.visibility&&"collapse"!==a.visibility||(o.invisible=!0),"none"===a.display&&(o.ignore=!0)}function OI(t,e){var n=e.__selfStyle;if(n){var r=n.textBaseline,i=r;r&&"auto"!==r?"baseline"===r?i="alphabetic":"before-edge"===r||"text-before-edge"===r?i="top":"after-edge"===r||"text-after-edge"===r?i="bottom":"central"!==r&&"mathematical"!==r||(i="middle"):i="alphabetic",t.style.textBaseline=i}var o=e.__inheritedStyle;if(o){var a=o.textAlign,s=a;a&&("middle"===a&&(s="center"),t.style.textAlign=s)}}var II=/^url\(\s*#(.*?)\)/;function CI(t,e,n,r){var i=n&&n.match(II);if(!i)return"none"===n&&(n=null),n;var o=Object(s["T"])(i[1]);r.push([t,e,o])}function AI(t,e){for(var n=0;n<e.length;n++){var r=e[n];r[0].style[r[1]]=t[r[2]]}}var YI=/-?([0-9]*\.)?[0-9]+([eE]-?[0-9]+)?/g;function jI(t){return t.match(YI)||[]}var PI=/(translate|scale|rotate|skewX|skewY|matrix)\(([\-\s0-9\.eE,]*)\)/g,EI=Math.PI/180;function NI(t,e){var n=t.getAttribute("transform");if(n){n=n.replace(/,/g," ");var r=[],i=null;n.replace(PI,(function(t,e,n){return r.push(e,n),""}));for(var o=r.length-1;o>0;o-=2){var a=r[o],s=r[o-1],l=jI(a);switch(i=i||st(),s){case"translate":dt(i,i,[parseFloat(l[0]),parseFloat(l[1]||"0")]);break;case"scale":ft(i,i,[parseFloat(l[0]),parseFloat(l[1]||l[0])]);break;case"rotate":ht(i,i,-parseFloat(l[0])*EI,[parseFloat(l[1]||"0"),parseFloat(l[2]||"0")]);break;case"skewX":var u=Math.tan(parseFloat(l[0])*EI);ct(i,[1,0,u,1,0,0],i);break;case"skewY":var c=Math.tan(parseFloat(l[0])*EI);ct(i,[1,c,0,1,0,0],i);break;case"matrix":i[0]=parseFloat(l[0]),i[1]=parseFloat(l[1]),i[2]=parseFloat(l[2]),i[3]=parseFloat(l[3]),i[4]=parseFloat(l[4]),i[5]=parseFloat(l[5]);break}}e.setLocalTransform(i)}}var RI=/([^\s:;]+)\s*:\s*([^:;]+)/g;function HI(t,e,n){var r=t.getAttribute("style");if(r){var i;RI.lastIndex=0;while(null!=(i=RI.exec(r))){var o=i[1],a=Object(s["q"])(vI,o)?vI[o]:null;a&&(e[a]=i[2]);var l=Object(s["q"])(bI,o)?bI[o]:null;l&&(n[l]=i[2])}}}function zI(t,e,n){for(var r=0;r<_I.length;r++){var i=_I[r],o=t.getAttribute(i);null!=o&&(e[vI[i]]=o)}for(r=0;r<xI.length;r++){i=xI[r],o=t.getAttribute(i);null!=o&&(n[bI[i]]=o)}}function FI(t,e){var n=e.width/t.width,r=e.height/t.height,i=Math.min(n,r);return{scale:i,x:-(t.x+t.width/2)*i+(e.x+e.width/2),y:-(t.y+t.height/2)*i+(e.y+e.height/2)}}function BI(t,e){var n=new wI;return n.parse(t,e)}var WI=1e-8;function VI(t,e){return Math.abs(t-e)<WI}function GI(t,e,n){var r=0,i=t[0];if(!i)return!1;for(var o=1;o<t.length;o++){var a=t[o];r+=La(i[0],i[1],a[0],a[1],e,n),i=a}var s=t[0];return VI(i[0],s[0])&&VI(i[1],s[1])||(r+=La(i[0],i[1],s[0],s[1],e,n)),0!==r}var $I=[];function UI(t,e){for(var n=0;n<t.length;n++)S(t[n],t[n],e)}function qI(t,e,n,r){for(var i=0;i<t.length;i++){var o=t[i];r&&(o=r.project(o)),o&&isFinite(o[0])&&isFinite(o[1])&&(k(e,e,o),L(n,n,o))}}function XI(t){for(var e=0,n=0,r=0,i=t.length,o=t[i-1][0],a=t[i-1][1],s=0;s<i;s++){var l=t[s][0],u=t[s][1],c=o*u-l*a;e+=c,n+=(o+l)*c,r+=(a+u)*c,o=l,a=u}return e?[n/e/3,r/e/3,e]:[t[0][0]||0,t[0][1]||0]}var ZI=function(){function t(t){this.name=t}return t.prototype.setCenter=function(t){this._center=t},t.prototype.getCenter=function(){var t=this._center;return t||(t=this._center=this.calcCenter()),t},t}(),JI=function(){function t(t,e){this.type="polygon",this.exterior=t,this.interiors=e}return t}(),KI=function(){function t(t){this.type="linestring",this.points=t}return t}(),QI=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.type="geoJSON",i.geometries=n,i._center=r&&[r[0],r[1]],i}return o(e,t),e.prototype.calcCenter=function(){for(var t,e=this.geometries,n=0,r=0;r<e.length;r++){var i=e[r],o=i.exterior,a=o&&o.length;a>n&&(t=i,n=a)}if(t)return XI(t.exterior);var s=this.getBoundingRect();return[s.x+s.width/2,s.y+s.height/2]},e.prototype.getBoundingRect=function(t){var e=this._rect;if(e&&!t)return e;var n=[1/0,1/0],r=[-1/0,-1/0],i=this.geometries;return Object(s["k"])(i,(function(e){"polygon"===e.type?qI(e.exterior,n,r,t):Object(s["k"])(e.points,(function(e){qI(e,n,r,t)}))})),isFinite(n[0])&&isFinite(n[1])&&isFinite(r[0])&&isFinite(r[1])||(n[0]=n[1]=r[0]=r[1]=0),e=new Tt(n[0],n[1],r[0]-n[0],r[1]-n[1]),t||(this._rect=e),e},e.prototype.contain=function(t){var e=this.getBoundingRect(),n=this.geometries;if(!e.contain(t[0],t[1]))return!1;t:for(var r=0,i=n.length;r<i;r++){var o=n[r];if("polygon"===o.type){var a=o.exterior,s=o.interiors;if(GI(a,t[0],t[1])){for(var l=0;l<(s?s.length:0);l++)if(GI(s[l],t[0],t[1]))continue t;return!0}}}return!1},e.prototype.transformTo=function(t,e,n,r){var i=this.getBoundingRect(),o=i.width/i.height;n?r||(r=n/o):n=o*r;for(var a=new Tt(t,e,n,r),l=i.calculateTransform(a),u=this.geometries,c=0;c<u.length;c++){var d=u[c];"polygon"===d.type?(UI(d.exterior,l),Object(s["k"])(d.interiors,(function(t){UI(t,l)}))):Object(s["k"])(d.points,(function(t){UI(t,l)}))}i=this._rect,i.copy(a),this._center=[i.x+i.width/2,i.y+i.height/2]},e.prototype.cloneShallow=function(t){null==t&&(t=this.name);var n=new e(t,this.geometries,this._center);return n._rect=this._rect,n.transformTo=null,n},e}(ZI),tC=function(t){function e(e,n){var r=t.call(this,e)||this;return r.type="geoSVG",r._elOnlyForCalculate=n,r}return o(e,t),e.prototype.calcCenter=function(){var t=this._elOnlyForCalculate,e=t.getBoundingRect(),n=[e.x+e.width/2,e.y+e.height/2],r=lt($I),i=t;while(i&&!i.isGeoSVGGraphicRoot)ct(r,i.getLocalTransform(),r),i=i.parent;return pt(r,r),S(n,n,r),n},e}(ZI),eC=Object(s["f"])(["rect","circle","line","ellipse","polygon","polyline","path","text","tspan","g"]),nC=function(){function t(t,e){this.type="geoSVG",this._usedGraphicMap=Object(s["f"])(),this._freedGraphics=[],this._mapName=t,this._parsedXML=yI(e)}return t.prototype.load=function(){var t=this._firstGraphic;if(!t){t=this._firstGraphic=this._buildGraphic(this._parsedXML),this._freedGraphics.push(t),this._boundingRect=this._firstGraphic.boundingRect.clone();var e=iC(t.named),n=e.regions,r=e.regionsMap;this._regions=n,this._regionsMap=r}return{boundingRect:this._boundingRect,regions:this._regions,regionsMap:this._regionsMap}},t.prototype._buildGraphic=function(t){var e,n;try{e=t&&BI(t,{ignoreViewBox:!0,ignoreRootClip:!0})||{},n=e.root,Object(s["b"])(null!=n)}catch(g){throw new Error("Invalid svg format\n"+g.message)}var r=new wr;r.add(n),r.isGeoSVGGraphicRoot=!0;var i=e.width,o=e.height,a=e.viewBoxRect,l=this._boundingRect;if(!l){var u=void 0,c=void 0,d=void 0,h=void 0;if(null!=i?(u=0,d=i):a&&(u=a.x,d=a.width),null!=o?(c=0,h=o):a&&(c=a.y,h=a.height),null==u||null==c){var f=n.getBoundingRect();null==u&&(u=f.x,d=f.width),null==c&&(c=f.y,h=f.height)}l=this._boundingRect=new Tt(u,c,d,h)}if(a){var p=FI(a,l);n.scaleX=n.scaleY=p.scale,n.x=p.x,n.y=p.y}r.setClipPath(new ss({shape:l.plain()}));var m=[];return Object(s["k"])(e.named,(function(t){null!=eC.get(t.svgNodeTagLower)&&(m.push(t),rC(t.el))})),{root:r,boundingRect:l,named:m}},t.prototype.useGraphic=function(t){var e=this._usedGraphicMap,n=e.get(t);return n||(n=this._freedGraphics.pop()||this._buildGraphic(this._parsedXML),e.set(t,n),n)},t.prototype.freeGraphic=function(t){var e=this._usedGraphicMap,n=e.get(t);n&&(e.removeKey(t),this._freedGraphics.push(n))},t}();function rC(t){t.silent=!1,t.isGroup&&t.traverse((function(t){t.silent=!1}))}function iC(t){var e=[],n=Object(s["f"])();return Object(s["k"])(t,(function(t){if(null==t.namedFrom){var r=new tC(t.name,t.el);e.push(r),n.set(t.name,r)}})),{regions:e,regionsMap:n}}function oC(t){if(!t.UTF8Encoding)return t;var e=t,n=e.UTF8Scale;null==n&&(n=1024);var r=e.features;return s["k"](r,(function(t){var e=t.geometry,r=e.encodeOffsets,i=e.coordinates;if(r)switch(e.type){case"LineString":e.coordinates=sC(i,r,n);break;case"Polygon":aC(i,r,n);break;case"MultiLineString":aC(i,r,n);break;case"MultiPolygon":s["k"](i,(function(t,e){return aC(t,r[e],n)}))}})),e.UTF8Encoding=!1,e}function aC(t,e,n){for(var r=0;r<t.length;r++)t[r]=sC(t[r],e[r],n)}function sC(t,e,n){for(var r=[],i=e[0],o=e[1],a=0;a<t.length;a+=2){var s=t.charCodeAt(a)-64,l=t.charCodeAt(a+1)-64;s=s>>1^-(1&s),l=l>>1^-(1&l),s+=i,l+=o,i=s,o=l,r.push([s/n,l/n])}return r}function lC(t,e){return t=oC(t),s["H"](s["n"](t.features,(function(t){return t.geometry&&t.properties&&t.geometry.coordinates.length>0})),(function(t){var n=t.properties,r=t.geometry,i=[];switch(r.type){case"Polygon":var o=r.coordinates;i.push(new JI(o[0],o.slice(1)));break;case"MultiPolygon":s["k"](r.coordinates,(function(t){t[0]&&i.push(new JI(t[0],t.slice(1)))}));break;case"LineString":i.push(new KI([r.coordinates]));break;case"MultiLineString":i.push(new KI(r.coordinates))}var a=new QI(n[e||"name"],i,n.cp);return a.properties=n,a}))}for(var uC=[126,25],cC="南海诸岛",dC=[[[0,3.5],[7,11.2],[15,11.9],[30,7],[42,.7],[52,.7],[56,7.7],[59,.7],[64,.7],[64,0],[5,0],[0,3.5]],[[13,16.1],[19,14.7],[16,21.7],[11,23.1],[13,16.1]],[[12,32.2],[14,38.5],[15,38.5],[13,32.2],[12,32.2]],[[16,47.6],[12,53.2],[13,53.2],[18,47.6],[16,47.6]],[[6,64.4],[8,70],[9,70],[8,64.4],[6,64.4]],[[23,82.6],[29,79.8],[30,79.8],[25,82.6],[23,82.6]],[[37,70.7],[43,62.3],[44,62.3],[39,70.7],[37,70.7]],[[48,51.1],[51,45.5],[53,45.5],[50,51.1],[48,51.1]],[[51,35],[51,28.7],[53,28.7],[53,35],[51,35]],[[52,22.4],[55,17.5],[56,17.5],[53,22.4],[52,22.4]],[[58,12.6],[62,7],[63,7],[60,12.6],[58,12.6]],[[0,3.5],[0,93.1],[64,93.1],[64,0],[63,0],[63,92.4],[1,92.4],[1,3.5],[0,3.5]]],hC=0;hC<dC.length;hC++)for(var fC=0;fC<dC[hC].length;fC++)dC[hC][fC][0]/=10.5,dC[hC][fC][1]/=-14,dC[hC][fC][0]+=uC[0],dC[hC][fC][1]+=uC[1];function pC(t,e){if("china"===t){for(var n=0;n<e.length;n++)if(e[n].name===cC)return;e.push(new QI(cC,s["H"](dC,(function(t){return{type:"polygon",exterior:t}})),uC))}}var mC={"南海诸岛":[32,80],"广东":[0,-10],"香港":[10,5],"澳门":[-10,10],"天津":[5,5]};function gC(t,e){if("china"===t){var n=mC[e.name];if(n){var r=e.getCenter();r[0]+=n[0]/10.5,r[1]+=-n[1]/14,e.setCenter(r)}}}var yC=[[[123.45165252685547,25.73527164402261],[123.49731445312499,25.73527164402261],[123.49731445312499,25.750734064600884],[123.45165252685547,25.750734064600884],[123.45165252685547,25.73527164402261]]];function vC(t,e){"china"===t&&"台湾"===e.name&&e.geometries.push({type:"polygon",exterior:yC[0]})}var _C="name",bC=function(){function t(t,e,n){this.type="geoJSON",this._parsedMap=Object(s["f"])(),this._mapName=t,this._specialAreas=n,this._geoJSON=wC(e)}return t.prototype.load=function(t,e){e=e||_C;var n=this._parsedMap.get(e);if(!n){var r=this._parseToRegions(e);n=this._parsedMap.set(e,{regions:r,boundingRect:xC(r)})}var i=Object(s["f"])(),o=[];return Object(s["k"])(n.regions,(function(e){var n=e.name;t&&Object(s["q"])(t,n)&&(e=e.cloneShallow(n=t[n])),o.push(e),i.set(n,e)})),{regions:o,boundingRect:n.boundingRect||new Tt(0,0,0,0),regionsMap:i}},t.prototype._parseToRegions=function(t){var e,n=this._mapName,r=this._geoJSON;try{e=r?lC(r,t):[]}catch(i){throw new Error("Invalid geoJson format\n"+i.message)}return pC(n,e),Object(s["k"])(e,(function(t){var e=t.name;gC(n,t),vC(n,t);var r=this._specialAreas&&this._specialAreas[e];r&&t.transformTo(r.left,r.top,r.width,r.height)}),this),e},t.prototype.getMapForUser=function(){return{geoJson:this._geoJSON,geoJSON:this._geoJSON,specialAreas:this._specialAreas}},t}();function xC(t){for(var e,n=0;n<t.length;n++){var r=t[n].getBoundingRect();e=e||r.clone(),e.union(r)}return e}function wC(t){return Object(s["C"])(t)?"undefined"!==typeof JSON&&JSON.parse?JSON.parse(t):new Function("return ("+t+");")():t}var MC=Object(s["f"])(),SC={registerMap:function(t,e,n){if(e.svg){var r=new nC(t,e.svg);MC.set(t,r)}else{var i=e.geoJson||e.geoJSON;i&&!e.features?n=e.specialAreas:i=e;r=new bC(t,i,n);MC.set(t,r)}},getGeoResource:function(t){return MC.get(t)},getMapForUser:function(t){var e=MC.get(t);return e&&"geoJSON"===e.type&&e.getMapForUser()},load:function(t,e,n){var r=MC.get(t);if(r)return r.load(e,n)}},kC=["rect","circle","line","ellipse","polygon","polyline","path"],LC=s["f"](kC),TC=s["f"](kC.concat(["g"])),DC=s["f"](kC.concat(["g"])),OC=Ci();function IC(t){var e=t.getItemStyle(),n=t.get("areaColor");return null!=n&&(e.fill=n),e}function CC(t){var e=t.style;e&&(e.stroke=e.stroke||e.fill,e.fill=null)}var AC=function(){function t(t){var e=new wr;this.uid=Du("ec_map_draw"),this._controller=new dI(t.getZr()),this._controllerHost={target:e},this.group=e,e.add(this._regionsGroup=new wr),e.add(this._svgGroup=new wr)}return t.prototype.draw=function(t,e,n,r,i){var o="geo"===t.mainType,a=t.getData&&t.getData();o&&e.eachComponent({mainType:"series",subType:"map"},(function(e){a||e.getHostGeoModel()!==t||(a=e.getData())}));var s=t.coordinateSystem,l=this._regionsGroup,u=this.group,c=s.getTransformInfo(),d=c.raw,h=c.roam,f=!l.childAt(0)||i;f?(u.x=h.x,u.y=h.y,u.scaleX=h.scaleX,u.scaleY=h.scaleY,u.dirty()):Wl(u,h,t);var p=a&&a.getVisual("visualMeta")&&a.getVisual("visualMeta").length>0,m={api:n,geo:s,mapOrGeoModel:t,data:a,isVisualEncodedByVisualMap:p,isGeo:o,transformInfoRaw:d};"geoJSON"===s.resourceType?this._buildGeoJSON(m):"geoSVG"===s.resourceType&&this._buildSVG(m),this._updateController(t,e,n),this._updateMapSelectHandler(t,l,n,r)},t.prototype._buildGeoJSON=function(t){var e=this._regionsGroupByName=s["f"](),n=s["f"](),r=this._regionsGroup,i=t.transformInfoRaw,o=t.mapOrGeoModel,a=t.data,l=t.geo.projection,u=l&&l.stream;function c(t,e){return e&&(t=e(t)),t&&[t[0]*i.scaleX+i.x,t[1]*i.scaleY+i.y]}function d(t){for(var e=[],n=!u&&l&&l.project,r=0;r<t.length;++r){var i=c(t[r],n);i&&e.push(i)}return e}function h(t){return{shape:{points:d(t)}}}r.removeAll(),s["k"](t.geo.regions,(function(i){var d=i.name,f=e.get(d),p=n.get(d)||{},m=p.dataIdx,g=p.regionModel;f||(f=e.set(d,new wr),r.add(f),m=a?a.indexOfName(d):null,g=t.isGeo?o.getRegionModel(d):a?a.getItemModel(m):null,n.set(d,{dataIdx:m,regionModel:g}));var y=[],v=[];s["k"](i.geometries,(function(t){if("polygon"===t.type){var e=[t.exterior].concat(t.interiors||[]);u&&(e=RC(e,u)),s["k"](e,(function(t){y.push(new Tm(h(t)))}))}else{var n=t.points;u&&(n=RC(n,u,!0)),s["k"](n,(function(t){v.push(new Im(h(t)))}))}}));var _=c(i.getCenter(),l&&l.project);function b(e,n){if(e.length){var r=new Vm({culling:!0,segmentIgnoreThreshold:1,shape:{paths:e}});f.add(r),YC(t,r,m,g),jC(t,r,d,g,o,m,_),n&&(CC(r),s["k"](r.states,CC))}}b(y),b(v,!0)})),e.each((function(e,r){var i=n.get(r),a=i.dataIdx,s=i.regionModel;PC(t,e,r,s,o,a),EC(t,e,r,s,o),NC(t,e,r,s,o)}),this)},t.prototype._buildSVG=function(t){var e=t.geo.map,n=t.transformInfoRaw;this._svgGroup.x=n.x,this._svgGroup.y=n.y,this._svgGroup.scaleX=n.scaleX,this._svgGroup.scaleY=n.scaleY,this._svgResourceChanged(e)&&(this._freeSVG(),this._useSVG(e));var r=this._svgDispatcherMap=s["f"](),i=!1;s["k"](this._svgGraphicRecord.named,(function(e){var n=e.name,o=t.mapOrGeoModel,a=t.data,s=e.svgNodeTagLower,l=e.el,u=a?a.indexOfName(n):null,c=o.getRegionModel(n);if(null!=LC.get(s)&&l instanceof No&&YC(t,l,u,c),l instanceof No&&(l.culling=!0),l.z2EmphasisLift=0,!e.namedFrom&&(null!=DC.get(s)&&jC(t,l,n,c,o,u,null),PC(t,l,n,c,o,u),EC(t,l,n,c,o),null!=TC.get(s))){var d=NC(t,l,n,c,o);"self"===d&&(i=!0);var h=r.get(n)||r.set(n,[]);h.push(l)}}),this),this._enableBlurEntireSVG(i,t)},t.prototype._enableBlurEntireSVG=function(t,e){if(t&&e.isGeo){var n=e.mapOrGeoModel.getModel(["blur","itemStyle"]).getItemStyle(),r=n.opacity;this._svgGraphicRecord.root.traverse((function(t){if(!t.isGroup){al(t);var e=t.ensureState("blur").style||{};null==e.opacity&&null!=r&&(e.opacity=r),t.ensureState("emphasis")}}))}},t.prototype.remove=function(){this._regionsGroup.removeAll(),this._regionsGroupByName=null,this._svgGroup.removeAll(),this._freeSVG(),this._controller.dispose(),this._controllerHost=null},t.prototype.findHighDownDispatchers=function(t,e){if(null==t)return[];var n=e.coordinateSystem;if("geoJSON"===n.resourceType){var r=this._regionsGroupByName;if(r){var i=r.get(t);return i?[i]:[]}}else if("geoSVG"===n.resourceType)return this._svgDispatcherMap&&this._svgDispatcherMap.get(t)||[]},t.prototype._svgResourceChanged=function(t){return this._svgMapName!==t},t.prototype._useSVG=function(t){var e=SC.getGeoResource(t);if(e&&"geoSVG"===e.type){var n=e.useGraphic(this.uid);this._svgGroup.add(n.root),this._svgGraphicRecord=n,this._svgMapName=t}},t.prototype._freeSVG=function(){var t=this._svgMapName;if(null!=t){var e=SC.getGeoResource(t);e&&"geoSVG"===e.type&&e.freeGraphic(this.uid),this._svgGraphicRecord=null,this._svgDispatcherMap=null,this._svgGroup.removeAll(),this._svgMapName=null}},t.prototype._updateController=function(t,e,n){var r=t.coordinateSystem,i=this._controller,o=this._controllerHost;o.zoomLimit=t.get("scaleLimit"),o.zoom=r.getZoom(),i.enable(t.get("roam")||!1);var a=t.mainType;function l(){var e={type:"geoRoam",componentType:a};return e[a+"Id"]=t.id,e}i.off("pan").on("pan",(function(t){this._mouseDownFlag=!1,hI(o,t.dx,t.dy),n.dispatchAction(s["m"](l(),{dx:t.dx,dy:t.dy,animation:{duration:0}}))}),this),i.off("zoom").on("zoom",(function(t){this._mouseDownFlag=!1,fI(o,t.scale,t.originX,t.originY),n.dispatchAction(s["m"](l(),{zoom:t.scale,originX:t.originX,originY:t.originY,animation:{duration:0}}))}),this),i.setPointerChecker((function(e,i,o){return r.containPoint([i,o])&&!gI(e,n,t)}))},t.prototype.resetForLabelLayout=function(){this.group.traverse((function(t){var e=t.getTextContent();e&&(e.ignore=OC(e).ignore)}))},t.prototype._updateMapSelectHandler=function(t,e,n,r){var i=this;e.off("mousedown"),e.off("click"),t.get("selectedMode")&&(e.on("mousedown",(function(){i._mouseDownFlag=!0})),e.on("click",(function(t){i._mouseDownFlag&&(i._mouseDownFlag=!1)})))},t}();function YC(t,e,n,r){var i=r.getModel("itemStyle"),o=r.getModel(["emphasis","itemStyle"]),a=r.getModel(["blur","itemStyle"]),s=r.getModel(["select","itemStyle"]),l=IC(i),u=IC(o),c=IC(s),d=IC(a),h=t.data;if(h){var f=h.getItemVisual(n,"style"),p=h.getItemVisual(n,"decal");t.isVisualEncodedByVisualMap&&f.fill&&(l.fill=f.fill),p&&(l.decal=qv(p,t.api))}e.setStyle(l),e.style.strokeNoScale=!0,e.ensureState("emphasis").style=u,e.ensureState("select").style=c,e.ensureState("blur").style=d,al(e)}function jC(t,e,n,r,i,o,a){var s=t.data,l=t.isGeo,u=s&&isNaN(s.get(s.mapDimension("value"),o)),c=s&&s.getItemLayout(o);if(l||u||c&&c.showLabel){var d=l?n:o,h=void 0;(!s||o>=0)&&(h=i);var f=a?{normal:{align:"center",verticalAlign:"middle"}}:null;tu(e,eu(r),{labelFetcher:h,labelDataIndex:d,defaultText:n},f);var p=e.getTextContent();if(p&&(OC(p).ignore=p.ignore,e.textConfig&&a)){var m=e.getBoundingRect().clone();e.textConfig.layoutRect=m,e.textConfig.position=[(a[0]-m.x)/m.width*100+"%",(a[1]-m.y)/m.height*100+"%"]}e.disableLabelAnimation=!0}else e.removeTextContent(),e.removeTextConfig(),e.disableLabelAnimation=null}function PC(t,e,n,r,i,o){t.data?t.data.setItemGraphicEl(o,e):Ls(e).eventData={componentType:"geo",componentIndex:i.componentIndex,geoIndex:i.componentIndex,name:n,region:r&&r.option||{}}}function EC(t,e,n,r,i){t.data||Pg({el:e,componentModel:i,itemName:n,itemTooltipOption:r.get("tooltip")})}function NC(t,e,n,r,i){e.highDownSilentOnTouch=!!i.get("selectedMode");var o=r.getModel("emphasis"),a=o.get("focus");return Dl(e,a,o.get("blurScope"),o.get("disabled")),t.isGeo&&Pl(e,i,n),a}function RC(t,e,n){var r,i=[];function o(){r=[]}function a(){r.length&&(i.push(r),r=[])}var l=e({polygonStart:o,polygonEnd:a,lineStart:o,lineEnd:a,point:function(t,e){isFinite(t)&&isFinite(e)&&r.push([t,e])},sphere:function(){}});return!n&&l.polygonStart(),s["k"](t,(function(t){l.lineStart();for(var e=0;e<t.length;e++)l.point(t[e][0],t[e][1]);l.lineEnd()})),!n&&l.polygonEnd(),i}var HC=AC,zC=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n,r){if(!r||"mapToggleSelect"!==r.type||r.from!==this.uid){var i=this.group;if(i.removeAll(),!t.getHostGeoModel()){if(this._mapDraw&&r&&"geoRoam"===r.type&&this._mapDraw.resetForLabelLayout(),r&&"geoRoam"===r.type&&"series"===r.componentType&&r.seriesId===t.id){o=this._mapDraw;o&&i.add(o.group)}else if(t.needsDrawMap){var o=this._mapDraw||new HC(n);i.add(o.group),o.draw(t,e,n,this,r),this._mapDraw=o}else this._mapDraw&&this._mapDraw.remove(),this._mapDraw=null;t.get("showLegendSymbol")&&e.getComponent("legend")&&this._renderSymbols(t,e,n)}}},e.prototype.remove=function(){this._mapDraw&&this._mapDraw.remove(),this._mapDraw=null,this.group.removeAll()},e.prototype.dispose=function(){this._mapDraw&&this._mapDraw.remove(),this._mapDraw=null},e.prototype._renderSymbols=function(t,e,n){var r=t.originalData,i=this.group;r.each(r.mapDimension("value"),(function(e,n){if(!isNaN(e)){var o=r.getItemLayout(n);if(o&&o.point){var a=o.point,s=o.offset,l=new Kp({style:{fill:t.getData().getVisual("style").fill},shape:{cx:a[0]+9*s,cy:a[1],r:3},silent:!0,z2:8+(s?0:Ns+1)});if(!s){var u=t.mainSeries.getData(),c=r.getName(n),d=u.indexOfName(c),h=r.getItemModel(n),f=h.getModel("label"),p=u.getItemGraphicEl(d);tu(l,eu(h),{labelFetcher:{getFormattedLabel:function(e,n){return t.getFormattedLabel(d,n)}},defaultText:c}),l.disableLabelAnimation=!0,f.get("position")||l.setTextConfig({position:"bottom"}),p.onHoverStateChange=function(t){tl(l,t)}}i.add(l)}}}))},e.type="map",e}($g),FC=zC,BC=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.needsDrawMap=!1,n.seriesGroup=[],n.getTooltipPosition=function(t){if(null!=t){var e=this.getData().getName(t),n=this.coordinateSystem,r=n.getRegion(e);return r&&n.dataToPoint(r.getCenter())}},n}return o(e,t),e.prototype.getInitialData=function(t){for(var e=TL(this,{coordDimensions:["value"],encodeDefaulter:s["h"](hd,this)}),n=s["f"](),r=[],i=0,o=e.count();i<o;i++){var a=e.getName(i);n.set(a,!0)}var l=SC.load(this.getMapType(),this.option.nameMap,this.option.nameProperty);return s["k"](l.regions,(function(t){var e=t.name;n.get(e)||r.push(e)})),e.appendValues([],r),e},e.prototype.getHostGeoModel=function(){var t=this.option.geoIndex;return null!=t?this.ecModel.getComponent("geo",t):null},e.prototype.getMapType=function(){return(this.getHostGeoModel()||this).option.map},e.prototype.getRawValue=function(t){var e=this.getData();return e.get(e.mapDimension("value"),t)},e.prototype.getRegionModel=function(t){var e=this.getData();return e.getItemModel(e.indexOfName(t))},e.prototype.formatTooltip=function(t,e,n){for(var r=this.getData(),i=this.getRawValue(t),o=r.getName(t),a=this.seriesGroup,s=[],l=0;l<a.length;l++){var u=a[l].originalData.indexOfName(o),c=r.mapDimension("value");isNaN(a[l].originalData.get(c,u))||s.push(a[l].name)}return Uf("section",{header:s.join(", "),noHeader:!s.length,blocks:[Uf("nameValue",{name:o,value:i})]})},e.prototype.setZoom=function(t){this.option.zoom=t},e.prototype.setCenter=function(t){this.option.center=t},e.prototype.getLegendIcon=function(t){var e=t.icon||"roundRect",n=ov(e,0,0,t.itemWidth,t.itemHeight,t.itemStyle.fill);return n.setStyle(t.itemStyle),n.style.stroke="none",e.indexOf("empty")>-1&&(n.style.stroke=n.style.fill,n.style.fill="#fff",n.style.lineWidth=2),n},e.type="series.map",e.dependencies=["geo"],e.layoutMode="box",e.defaultOption={z:2,coordinateSystem:"geo",map:"",left:"center",top:"center",aspectScale:null,showLegendSymbol:!0,boundingCoords:null,center:null,zoom:1,scaleLimit:null,selectedMode:!0,label:{show:!1,color:"#000"},itemStyle:{borderWidth:.5,borderColor:"#444",areaColor:"#eee"},emphasis:{label:{show:!0,color:"rgb(100,0,0)"},itemStyle:{areaColor:"rgba(255,215,0,0.8)"}},select:{label:{show:!0,color:"rgb(100,0,0)"},itemStyle:{color:"rgba(255,215,0,0.8)"}},nameProperty:"name"},e}(Mp),WC=BC;function VC(t,e){var n={};return s["k"](t,(function(t){t.each(t.mapDimension("value"),(function(e,r){var i="ec-"+t.getName(r);n[i]=n[i]||[],isNaN(e)||n[i].push(e)}))})),t[0].map(t[0].mapDimension("value"),(function(r,i){for(var o,a="ec-"+t[0].getName(i),s=0,l=1/0,u=-1/0,c=n[a].length,d=0;d<c;d++)l=Math.min(l,n[a][d]),u=Math.max(u,n[a][d]),s+=n[a][d];return o="min"===e?l:"max"===e?u:"average"===e?s/c:s,0===c?NaN:o}))}function GC(t){var e={};t.eachSeriesByType("map",(function(t){var n=t.getHostGeoModel(),r=n?"o"+n.id:"i"+t.getMapType();(e[r]=e[r]||[]).push(t)})),s["k"](e,(function(t,e){for(var n=VC(s["H"](t,(function(t){return t.getData()})),t[0].get("mapValueCalculation")),r=0;r<t.length;r++)t[r].originalData=t[r].getData();for(r=0;r<t.length;r++)t[r].seriesGroup=t,t[r].needsDrawMap=0===r&&!t[r].getHostGeoModel(),t[r].setData(n.cloneShallow()),t[r].mainSeries=t[0]}))}function $C(t){var e={};t.eachSeriesByType("map",(function(n){var r=n.getMapType();if(!n.getHostGeoModel()&&!e[r]){var i={};s["k"](n.seriesGroup,(function(e){var n=e.coordinateSystem,r=e.originalData;e.get("showLegendSymbol")&&t.getComponent("legend")&&r.each(r.mapDimension("value"),(function(t,e){var o=r.getName(e),a=n.getRegion(o);if(a&&!isNaN(t)){var s=i[o]||0,l=n.dataToPoint(a.getCenter());i[o]=s+1,r.setItemLayout(e,{point:l,offset:s})}}))}));var o=n.getData();o.each((function(t){var e=o.getName(t),n=o.getItemLayout(t)||{};n.showLabel=!i[e],o.setItemLayout(t,n)})),e[r]=!0}}))}var UC=S,qC=function(t){function e(e){var n=t.call(this)||this;return n.type="view",n.dimensions=["x","y"],n._roamTransformable=new Xn,n._rawTransformable=new Xn,n.name=e,n}return o(e,t),e.prototype.setBoundingRect=function(t,e,n,r){return this._rect=new Tt(t,e,n,r),this._rect},e.prototype.getBoundingRect=function(){return this._rect},e.prototype.setViewRect=function(t,e,n,r){this._transformTo(t,e,n,r),this._viewRect=new Tt(t,e,n,r)},e.prototype._transformTo=function(t,e,n,r){var i=this.getBoundingRect(),o=this._rawTransformable;o.transform=i.calculateTransform(new Tt(t,e,n,r));var a=o.parent;o.parent=null,o.decomposeTransform(),o.parent=a,this._updateTransform()},e.prototype.setCenter=function(t,e){t&&(this._center=[Nr(t[0],e.getWidth()),Nr(t[1],e.getHeight())],this._updateCenterAndZoom())},e.prototype.setZoom=function(t){t=t||1;var e=this.zoomLimit;e&&(null!=e.max&&(t=Math.min(e.max,t)),null!=e.min&&(t=Math.max(e.min,t))),this._zoom=t,this._updateCenterAndZoom()},e.prototype.getDefaultCenter=function(){var t=this.getBoundingRect(),e=t.x+t.width/2,n=t.y+t.height/2;return[e,n]},e.prototype.getCenter=function(){return this._center||this.getDefaultCenter()},e.prototype.getZoom=function(){return this._zoom||1},e.prototype.getRoamTransform=function(){return this._roamTransformable.getLocalTransform()},e.prototype._updateCenterAndZoom=function(){var t=this._rawTransformable.getLocalTransform(),e=this._roamTransformable,n=this.getDefaultCenter(),r=this.getCenter(),i=this.getZoom();r=S([],r,t),n=S([],n,t),e.originX=r[0],e.originY=r[1],e.x=n[0]-r[0],e.y=n[1]-r[1],e.scaleX=e.scaleY=i,this._updateTransform()},e.prototype._updateTransform=function(){var t=this._roamTransformable,e=this._rawTransformable;e.parent=t,t.updateTransform(),e.updateTransform(),ut(this.transform||(this.transform=[]),e.transform||st()),this._rawTransform=e.getLocalTransform(),this.invTransform=this.invTransform||[],pt(this.invTransform,this.transform),this.decomposeTransform()},e.prototype.getTransformInfo=function(){var t=this._rawTransformable,e=this._roamTransformable,n=new Xn;return n.transform=e.transform,n.decomposeTransform(),{roam:{x:n.x,y:n.y,scaleX:n.scaleX,scaleY:n.scaleY},raw:{x:t.x,y:t.y,scaleX:t.scaleX,scaleY:t.scaleY}}},e.prototype.getViewRect=function(){return this._viewRect},e.prototype.getViewRectAfterRoam=function(){var t=this.getBoundingRect().clone();return t.applyTransform(this.transform),t},e.prototype.dataToPoint=function(t,e,n){var r=e?this._rawTransform:this.transform;return n=n||[],r?UC(n,t,r):u(n,t)},e.prototype.pointToData=function(t){var e=this.invTransform;return e?UC([],t,e):[t[0],t[1]]},e.prototype.convertToPixel=function(t,e,n){var r=XC(e);return r===this?r.dataToPoint(n):null},e.prototype.convertFromPixel=function(t,e,n){var r=XC(e);return r===this?r.pointToData(n):null},e.prototype.containPoint=function(t){return this.getViewRectAfterRoam().contain(t[0],t[1])},e.dimensions=["x","y"],e}(Xn);function XC(t){var e=t.seriesModel;return e?e.coordinateSystem:null}var ZC=qC,JC={geoJSON:{aspectScale:.75,invertLongitute:!0},geoSVG:{aspectScale:1,invertLongitute:!1}},KC=["lng","lat"],QC=function(t){function e(e,n,r){var i=t.call(this,e)||this;i.dimensions=KC,i.type="geo",i._nameCoordMap=s["f"](),i.map=n;var o,a=r.projection,l=SC.load(n,r.nameMap,r.nameProperty),u=SC.getGeoResource(n),c=(i.resourceType=u?u.type:null,i.regions=l.regions),d=JC[u.type];if(i._regionsMap=l.regionsMap,i.regions=l.regions,i.projection=a,a)for(var h=0;h<c.length;h++){var f=c[h].getBoundingRect(a);o=o||f.clone(),o.union(f)}else o=l.boundingRect;return i.setBoundingRect(o.x,o.y,o.width,o.height),i.aspectScale=a?1:s["P"](r.aspectScale,d.aspectScale),i._invertLongitute=!a&&d.invertLongitute,i}return o(e,t),e.prototype._transformTo=function(t,e,n,r){var i=this.getBoundingRect(),o=this._invertLongitute;i=i.clone(),o&&(i.y=-i.y-i.height);var a=this._rawTransformable;a.transform=i.calculateTransform(new Tt(t,e,n,r));var s=a.parent;a.parent=null,a.decomposeTransform(),a.parent=s,o&&(a.scaleY=-a.scaleY),this._updateTransform()},e.prototype.getRegion=function(t){return this._regionsMap.get(t)},e.prototype.getRegionByCoord=function(t){for(var e=this.regions,n=0;n<e.length;n++){var r=e[n];if("geoJSON"===r.type&&r.contain(t))return e[n]}},e.prototype.addGeoCoord=function(t,e){this._nameCoordMap.set(t,e)},e.prototype.getGeoCoord=function(t){var e=this._regionsMap.get(t);return this._nameCoordMap.get(t)||e&&e.getCenter()},e.prototype.dataToPoint=function(t,e,n){if(s["C"](t)&&(t=this.getGeoCoord(t)),t){var r=this.projection;return r&&(t=r.project(t)),t&&this.projectedToPoint(t,e,n)}},e.prototype.pointToData=function(t){var e=this.projection;return e&&(t=e.unproject(t)),t&&this.pointToProjected(t)},e.prototype.pointToProjected=function(e){return t.prototype.pointToData.call(this,e)},e.prototype.projectedToPoint=function(e,n,r){return t.prototype.dataToPoint.call(this,e,n,r)},e.prototype.convertToPixel=function(t,e,n){var r=tA(e);return r===this?r.dataToPoint(n):null},e.prototype.convertFromPixel=function(t,e,n){var r=tA(e);return r===this?r.pointToData(n):null},e}(ZC);function tA(t){var e=t.geoModel,n=t.seriesModel;return e?e.coordinateSystem:n?n.coordinateSystem||(n.getReferringComponents("geo",Pi).models[0]||{}).coordinateSystem:null}s["K"](QC,ZC);var eA=QC;function nA(t,e){var n=t.get("boundingCoords");if(null!=n){var r=n[0],i=n[1];if(isFinite(r[0])&&isFinite(r[1])&&isFinite(i[0])&&isFinite(i[1])){var o=this.projection;if(o){var a=r[0],s=r[1],l=i[0],u=i[1];r=[1/0,1/0],i=[-1/0,-1/0];var c=function(t,e,n,a){for(var s=n-t,l=a-e,u=0;u<=100;u++){var c=u/100,d=o.project([t+s*c,e+l*c]);k(r,r,d),L(i,i,d)}};c(a,s,l,s),c(l,s,l,u),c(l,u,a,u),c(a,u,l,s)}this.setBoundingRect(r[0],r[1],i[0]-r[0],i[1]-r[1])}else 0}var d,h,f,p=this.getBoundingRect(),m=t.get("layoutCenter"),g=t.get("layoutSize"),y=e.getWidth(),v=e.getHeight(),_=p.width/p.height*this.aspectScale,b=!1;if(m&&g&&(d=[Nr(m[0],y),Nr(m[1],v)],h=Nr(g,Math.min(y,v)),isNaN(d[0])||isNaN(d[1])||isNaN(h)||(b=!0)),b)f={},_>1?(f.width=h,f.height=h/_):(f.height=h,f.width=h*_),f.y=d[1]-f.height/2,f.x=d[0]-f.width/2;else{var x=t.getBoxLayoutParams();x.aspect=_,f=Hc(x,{width:y,height:v})}this.setViewRect(f.x,f.y,f.width,f.height),this.setCenter(t.get("center"),e),this.setZoom(t.get("zoom"))}function rA(t,e){s["k"](e.get("geoCoord"),(function(e,n){t.addGeoCoord(n,e)}))}var iA=function(){function t(){this.dimensions=KC}return t.prototype.create=function(t,e){var n=[];function r(t){return{nameProperty:t.get("nameProperty"),aspectScale:t.get("aspectScale"),projection:t.get("projection")}}t.eachComponent("geo",(function(t,i){var o=t.get("map"),a=new eA(o+i,o,s["m"]({nameMap:t.get("nameMap")},r(t)));a.zoomLimit=t.get("scaleLimit"),n.push(a),t.coordinateSystem=a,a.model=t,a.resize=nA,a.resize(t,e)})),t.eachSeries((function(t){var e=t.get("coordinateSystem");if("geo"===e){var r=t.get("geoIndex")||0;t.coordinateSystem=n[r]}}));var i={};return t.eachSeriesByType("map",(function(t){if(!t.getHostGeoModel()){var e=t.getMapType();i[e]=i[e]||[],i[e].push(t)}})),s["k"](i,(function(t,i){var o=s["H"](t,(function(t){return t.get("nameMap")})),a=new eA(i,i,s["m"]({nameMap:s["J"](o)},r(t[0])));a.zoomLimit=s["O"].apply(null,s["H"](t,(function(t){return t.get("scaleLimit")}))),n.push(a),a.resize=nA,a.resize(t[0],e),s["k"](t,(function(t){t.coordinateSystem=a,rA(a,t)}))})),n},t.prototype.getFilledRegions=function(t,e,n,r){for(var i=(t||[]).slice(),o=s["f"](),a=0;a<i.length;a++)o.set(i[a].name,i[a]);var l=SC.load(e,n,r);return s["k"](l.regions,(function(t){var e=t.name;!o.get(e)&&i.push({name:e})})),i},t}(),oA=new iA,aA=oA,sA=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(t,e,n){var r=SC.getGeoResource(t.map);if(r&&"geoJSON"===r.type){var i=t.itemStyle=t.itemStyle||{};"color"in i||(i.color="#eee")}this.mergeDefaultAndTheme(t,n),ci(t,"label",["show"])},e.prototype.optionUpdated=function(){var t=this,e=this.option;e.regions=aA.getFilledRegions(e.regions,e.map,e.nameMap,e.nameProperty);var n={};this._optionModelMap=s["N"](e.regions||[],(function(e,r){var i=r.name;return i&&(e.set(i,new Lu(r,t,t.ecModel)),r.selected&&(n[i]=!0)),e}),s["f"]()),e.selectedMap||(e.selectedMap=n)},e.prototype.getRegionModel=function(t){return this._optionModelMap.get(t)||new Lu(null,this,this.ecModel)},e.prototype.getFormattedLabel=function(t,e){var n=this.getRegionModel(t),r="normal"===e?n.get(["label","formatter"]):n.get(["emphasis","label","formatter"]),i={name:t};return s["w"](r)?(i.status=e,r(i)):s["C"](r)?r.replace("{a}",null!=t?t:""):void 0},e.prototype.setZoom=function(t){this.option.zoom=t},e.prototype.setCenter=function(t){this.option.center=t},e.prototype.select=function(t){var e=this.option,n=e.selectedMode;if(n){"multiple"!==n&&(e.selectedMap=null);var r=e.selectedMap||(e.selectedMap={});r[t]=!0}},e.prototype.unSelect=function(t){var e=this.option.selectedMap;e&&(e[t]=!1)},e.prototype.toggleSelected=function(t){this[this.isSelected(t)?"unSelect":"select"](t)},e.prototype.isSelected=function(t){var e=this.option.selectedMap;return!(!e||!e[t])},e.type="geo",e.layoutMode="box",e.defaultOption={z:0,show:!0,left:"center",top:"center",aspectScale:null,silent:!1,map:"",boundingCoords:null,center:null,zoom:1,scaleLimit:null,label:{show:!1,color:"#000"},itemStyle:{borderWidth:.5,borderColor:"#444"},emphasis:{label:{show:!0,color:"rgb(100,0,0)"},itemStyle:{color:"rgba(255,215,0,0.8)"}},select:{label:{show:!0,color:"rgb(100,0,0)"},itemStyle:{color:"rgba(255,215,0,0.8)"}},regions:[]},e}(Xc),lA=sA;function uA(t,e){return t.pointToProjected?t.pointToProjected(e):t.pointToData(e)}function cA(t,e,n,r){var i=t.getZoom(),o=t.getCenter(),a=e.zoom,s=t.projectedToPoint?t.projectedToPoint(o):t.dataToPoint(o);if(null!=e.dx&&null!=e.dy&&(s[0]-=e.dx,s[1]-=e.dy,t.setCenter(uA(t,s),r)),null!=a){if(n){var l=n.min||0,u=n.max||1/0;a=Math.max(Math.min(i*a,u),l)/i}t.scaleX*=a,t.scaleY*=a;var c=(e.originX-t.x)*(a-1),d=(e.originY-t.y)*(a-1);t.x-=c,t.y-=d,t.updateTransform(),t.setCenter(uA(t,s),r),t.setZoom(a*i)}return{center:t.getCenter(),zoom:t.getZoom()}}var dA=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.focusBlurEnabled=!0,n}return o(e,t),e.prototype.init=function(t,e){this._api=e},e.prototype.render=function(t,e,n,r){if(this._model=t,!t.get("show"))return this._mapDraw&&this._mapDraw.remove(),void(this._mapDraw=null);this._mapDraw||(this._mapDraw=new HC(n));var i=this._mapDraw;i.draw(t,e,n,this,r),i.group.on("click",this._handleRegionClick,this),i.group.silent=t.get("silent"),this.group.add(i.group),this.updateSelectStatus(t,e,n)},e.prototype._handleRegionClick=function(t){var e;Gy(t.target,(function(t){return null!=(e=Ls(t).eventData)}),!0),e&&this._api.dispatchAction({type:"geoToggleSelect",geoId:this._model.id,name:e.name})},e.prototype.updateSelectStatus=function(t,e,n){var r=this;this._mapDraw.group.traverse((function(t){var e=Ls(t).eventData;if(e)return r._model.isSelected(e.name)?n.enterSelect(t):n.leaveSelect(t),!0}))},e.prototype.findHighDownDispatchers=function(t){return this._mapDraw&&this._mapDraw.findHighDownDispatchers(t,this._model)},e.prototype.dispose=function(){this._mapDraw&&this._mapDraw.remove()},e.type="geo",e}(kp),hA=dA;function fA(t,e,n){SC.registerMap(t,e,n)}function pA(t){function e(e,n){n.update="geo:updateSelectStatus",t.registerAction(n,(function(t,n){var r={},i=[];return n.eachComponent({mainType:"geo",query:t},(function(n){n[e](t.name);var o=n.coordinateSystem;Object(s["k"])(o.regions,(function(t){r[t.name]=n.isSelected(t.name)||!1}));var a=[];Object(s["k"])(r,(function(t,e){r[e]&&a.push(e)})),i.push({geoIndex:n.componentIndex,name:a})})),{selected:r,allSelected:i,name:t.name}}))}t.registerCoordinateSystem("geo",aA),t.registerComponentModel(lA),t.registerComponentView(hA),t.registerImpl("registerMap",fA),t.registerImpl("getMap",(function(t){return SC.getMapForUser(t)})),e("toggleSelected",{type:"geoToggleSelect",event:"geoselectchanged"}),e("select",{type:"geoSelect",event:"geoselected"}),e("unSelect",{type:"geoUnSelect",event:"geounselected"}),t.registerAction({type:"geoRoam",event:"geoRoam",update:"updateTransform"},(function(t,e,n){var r=t.componentType||"series";e.eachComponent({mainType:r,query:t},(function(e){var i=e.coordinateSystem;if("geo"===i.type){var o=cA(i,t,e.get("scaleLimit"),n);e.setCenter&&e.setCenter(o.center),e.setZoom&&e.setZoom(o.zoom),"series"===r&&Object(s["k"])(e.seriesGroup,(function(t){t.setCenter(o.center),t.setZoom(o.zoom)}))}}))}))}function mA(t){jb(pA),t.registerChartView(FC),t.registerSeriesModel(WC),t.registerLayout($C),t.registerProcessor(t.PRIORITY.PROCESSOR.STATISTIC,GC),By("map",t.registerAction)}function gA(t){var e=t;e.hierNode={defaultAncestor:null,ancestor:e,prelim:0,modifier:0,change:0,shift:0,i:0,thread:null};var n,r,i=[e];while(n=i.pop())if(r=n.children,n.isExpand&&r.length)for(var o=r.length,a=o-1;a>=0;a--){var s=r[a];s.hierNode={defaultAncestor:null,ancestor:s,prelim:0,modifier:0,change:0,shift:0,i:a,thread:null},i.push(s)}}function yA(t,e){var n=t.isExpand?t.children:[],r=t.parentNode.children,i=t.hierNode.i?r[t.hierNode.i-1]:null;if(n.length){wA(t);var o=(n[0].hierNode.prelim+n[n.length-1].hierNode.prelim)/2;i?(t.hierNode.prelim=i.hierNode.prelim+e(t,i),t.hierNode.modifier=t.hierNode.prelim-o):t.hierNode.prelim=o}else i&&(t.hierNode.prelim=i.hierNode.prelim+e(t,i));t.parentNode.hierNode.defaultAncestor=MA(t,i,t.parentNode.hierNode.defaultAncestor||r[0],e)}function vA(t){var e=t.hierNode.prelim+t.parentNode.hierNode.modifier;t.setLayout({x:e},!0),t.hierNode.modifier+=t.parentNode.hierNode.modifier}function _A(t){return arguments.length?t:DA}function bA(t,e){return t-=Math.PI/2,{x:e*Math.cos(t),y:e*Math.sin(t)}}function xA(t,e){return Hc(t.getBoxLayoutParams(),{width:e.getWidth(),height:e.getHeight()})}function wA(t){var e=t.children,n=e.length,r=0,i=0;while(--n>=0){var o=e[n];o.hierNode.prelim+=r,o.hierNode.modifier+=r,i+=o.hierNode.change,r+=o.hierNode.shift+i}}function MA(t,e,n,r){if(e){var i=t,o=t,a=o.parentNode.children[0],s=e,l=i.hierNode.modifier,u=o.hierNode.modifier,c=a.hierNode.modifier,d=s.hierNode.modifier;while(s=SA(s),o=kA(o),s&&o){i=SA(i),a=kA(a),i.hierNode.ancestor=t;var h=s.hierNode.prelim+d-o.hierNode.prelim-u+r(s,o);h>0&&(TA(LA(s,t,n),t,h),u+=h,l+=h),d+=s.hierNode.modifier,u+=o.hierNode.modifier,l+=i.hierNode.modifier,c+=a.hierNode.modifier}s&&!SA(i)&&(i.hierNode.thread=s,i.hierNode.modifier+=d-l),o&&!kA(a)&&(a.hierNode.thread=o,a.hierNode.modifier+=u-c,n=t)}return n}function SA(t){var e=t.children;return e.length&&t.isExpand?e[e.length-1]:t.hierNode.thread}function kA(t){var e=t.children;return e.length&&t.isExpand?e[0]:t.hierNode.thread}function LA(t,e,n){return t.hierNode.ancestor.parentNode===e.parentNode?t.hierNode.ancestor:n}function TA(t,e,n){var r=n/(e.hierNode.i-t.hierNode.i);e.hierNode.change-=r,e.hierNode.shift+=n,e.hierNode.modifier+=n,e.hierNode.prelim+=n,t.hierNode.change+=r}function DA(t,e){return t.parentNode===e.parentNode?1:2}var OA=function(){function t(){this.parentPoint=[],this.childPoints=[]}return t}(),IA=function(t){function e(e){return t.call(this,e)||this}return o(e,t),e.prototype.getDefaultStyle=function(){return{stroke:"#000",fill:null}},e.prototype.getDefaultShape=function(){return new OA},e.prototype.buildPath=function(t,e){var n=e.childPoints,r=n.length,i=e.parentPoint,o=n[0],a=n[r-1];if(1===r)return t.moveTo(i[0],i[1]),void t.lineTo(o[0],o[1]);var s=e.orient,l="TB"===s||"BT"===s?0:1,u=1-l,c=Nr(e.forkPosition,1),d=[];d[l]=i[l],d[u]=i[u]+(a[u]-i[u])*c,t.moveTo(i[0],i[1]),t.lineTo(d[0],d[1]),t.moveTo(o[0],o[1]),d[l]=o[l],t.lineTo(d[0],d[1]),d[l]=a[l],t.lineTo(d[0],d[1]),t.lineTo(a[0],a[1]);for(var h=1;h<r-1;h++){var f=n[h];t.moveTo(f[0],f[1]),d[l]=f[l],t.lineTo(d[0],d[1])}},e}(Va),CA=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n._mainGroup=new wr,n}return o(e,t),e.prototype.init=function(t,e){this._controller=new dI(e.getZr()),this._controllerHost={target:this.group},this.group.add(this._mainGroup)},e.prototype.render=function(t,e,n){var r=t.getData(),i=t.layoutInfo,o=this._mainGroup,a=t.get("layout");"radial"===a?(o.x=i.x+i.width/2,o.y=i.y+i.height/2):(o.x=i.x,o.y=i.y),this._updateViewCoordSys(t,n),this._updateController(t,e,n);var s=this._data;r.diff(s).add((function(e){AA(r,e)&&YA(r,e,null,o,t)})).update((function(e,n){var i=s.getItemGraphicEl(n);AA(r,e)?YA(r,e,i,o,t):i&&NA(s,n,i,o,t)})).remove((function(e){var n=s.getItemGraphicEl(e);n&&NA(s,e,n,o,t)})).execute(),this._nodeScaleRatio=t.get("nodeScaleRatio"),this._updateNodeAndLinkScale(t),!0===t.get("expandAndCollapse")&&r.eachItemGraphicEl((function(e,r){e.off("click").on("click",(function(){n.dispatchAction({type:"treeExpandAndCollapse",seriesId:t.id,dataIndex:r})}))})),this._data=r},e.prototype._updateViewCoordSys=function(t,e){var n=t.getData(),r=[];n.each((function(t){var e=n.getItemLayout(t);!e||isNaN(e.x)||isNaN(e.y)||r.push([+e.x,+e.y])}));var i=[],o=[];$o(r,i,o);var a=this._min,s=this._max;o[0]-i[0]===0&&(i[0]=a?a[0]:i[0]-1,o[0]=s?s[0]:o[0]+1),o[1]-i[1]===0&&(i[1]=a?a[1]:i[1]-1,o[1]=s?s[1]:o[1]+1);var l=t.coordinateSystem=new ZC;l.zoomLimit=t.get("scaleLimit"),l.setBoundingRect(i[0],i[1],o[0]-i[0],o[1]-i[1]),l.setCenter(t.get("center"),e),l.setZoom(t.get("zoom")),this.group.attr({x:l.x,y:l.y,scaleX:l.scaleX,scaleY:l.scaleY}),this._min=i,this._max=o},e.prototype._updateController=function(t,e,n){var r=this,i=this._controller,o=this._controllerHost,a=this.group;i.setPointerChecker((function(e,r,i){var o=a.getBoundingRect();return o.applyTransform(a.transform),o.contain(r,i)&&!gI(e,n,t)})),i.enable(t.get("roam")),o.zoomLimit=t.get("scaleLimit"),o.zoom=t.coordinateSystem.getZoom(),i.off("pan").off("zoom").on("pan",(function(e){hI(o,e.dx,e.dy),n.dispatchAction({seriesId:t.id,type:"treeRoam",dx:e.dx,dy:e.dy})})).on("zoom",(function(e){fI(o,e.scale,e.originX,e.originY),n.dispatchAction({seriesId:t.id,type:"treeRoam",zoom:e.scale,originX:e.originX,originY:e.originY}),r._updateNodeAndLinkScale(t),n.updateLabelLayout()}))},e.prototype._updateNodeAndLinkScale=function(t){var e=t.getData(),n=this._getNodeGlobalScale(t);e.eachItemGraphicEl((function(t,e){t.setSymbolScale(n)}))},e.prototype._getNodeGlobalScale=function(t){var e=t.coordinateSystem;if("view"!==e.type)return 1;var n=this._nodeScaleRatio,r=e.scaleX||1,i=e.getZoom(),o=(i-1)*n+1;return o/r},e.prototype.dispose=function(){this._controller&&this._controller.dispose(),this._controllerHost=null},e.prototype.remove=function(){this._mainGroup.removeAll(),this._data=null},e.type="tree",e}($g);function AA(t,e){var n=t.getItemLayout(e);return n&&!isNaN(n.x)&&!isNaN(n.y)}function YA(t,e,n,r,i){var o=!n,a=t.tree.getNodeByDataIndex(e),l=a.getModel(),u=a.getVisual("style").fill,c=!1===a.isExpand&&0!==a.children.length?u:"#fff",d=t.tree.root,h=a.parentNode===d?a:a.parentNode||a,f=t.getItemGraphicEl(h.dataIndex),p=h.getLayout(),m=f?{x:f.__oldX,y:f.__oldY,rawX:f.__radialOldRawX,rawY:f.__radialOldRawY}:p,g=a.getLayout();o?(n=new vS(t,e,null,{symbolInnerColor:c,useNameLabel:!0}),n.x=m.x,n.y=m.y):n.updateData(t,e,null,{symbolInnerColor:c,useNameLabel:!0}),n.__radialOldRawX=n.__radialRawX,n.__radialOldRawY=n.__radialRawY,n.__radialRawX=g.rawX,n.__radialRawY=g.rawY,r.add(n),t.setItemGraphicEl(e,n),n.__oldX=n.x,n.__oldY=n.y,Wl(n,{x:g.x,y:g.y},i);var y=n.getSymbolPath();if("radial"===i.get("layout")){var v=d.children[0],_=v.getLayout(),b=v.children.length,x=void 0,w=void 0;if(g.x===_.x&&!0===a.isExpand&&v.children.length){var M={x:(v.children[0].getLayout().x+v.children[b-1].getLayout().x)/2,y:(v.children[0].getLayout().y+v.children[b-1].getLayout().y)/2};x=Math.atan2(M.y-_.y,M.x-_.x),x<0&&(x=2*Math.PI+x),w=M.x<_.x,w&&(x-=Math.PI)}else x=Math.atan2(g.y-_.y,g.x-_.x),x<0&&(x=2*Math.PI+x),0===a.children.length||0!==a.children.length&&!1===a.isExpand?(w=g.x<_.x,w&&(x-=Math.PI)):(w=g.x>_.x,w||(x-=Math.PI));var S=w?"left":"right",k=l.getModel("label"),L=k.get("rotate"),T=L*(Math.PI/180),D=y.getTextContent();D&&(y.setTextConfig({position:k.get("position")||S,rotation:null==L?-x:T,origin:"center"}),D.setStyle("verticalAlign","middle"))}var O=l.get(["emphasis","focus"]),I="relative"===O?s["e"](a.getAncestorsIndices(),a.getDescendantIndices()):"ancestor"===O?a.getAncestorsIndices():"descendant"===O?a.getDescendantIndices():null;I&&(Ls(n).focus=I),jA(i,a,d,n,m,p,g,r),n.__edge&&(n.onHoverStateChange=function(e){if("blur"!==e){var r=a.parentNode&&t.getItemGraphicEl(a.parentNode.dataIndex);r&&r.hoverState===Ys||tl(n.__edge,e)}})}function jA(t,e,n,r,i,o,a,l){var u=e.getModel(),c=t.get("edgeShape"),d=t.get("layout"),h=t.getOrient(),f=t.get(["lineStyle","curveness"]),p=t.get("edgeForkPosition"),m=u.getModel("lineStyle").getLineStyle(),g=r.__edge;if("curve"===c)e.parentNode&&e.parentNode!==n&&(g||(g=r.__edge=new Hm({shape:RA(d,h,f,i,i)})),Wl(g,{shape:RA(d,h,f,o,a)},t));else if("polyline"===c)if("orthogonal"===d){if(e!==n&&e.children&&0!==e.children.length&&!0===e.isExpand){for(var y=e.children,v=[],_=0;_<y.length;_++){var b=y[_].getLayout();v.push([b.x,b.y])}g||(g=r.__edge=new IA({shape:{parentPoint:[a.x,a.y],childPoints:[[a.x,a.y]],orient:h,forkPosition:p}})),Wl(g,{shape:{parentPoint:[a.x,a.y],childPoints:v}},t)}}else 0;g&&("polyline"!==c||e.isExpand)&&(g.useStyle(s["i"]({strokeNoScale:!0,fill:null},m)),Al(g,u,"lineStyle"),al(g),l.add(g))}function PA(t,e,n,r,i){var o=e.tree.root,a=EA(o,t),s=a.source,l=a.sourceLayout,u=e.getItemGraphicEl(t.dataIndex);if(u){var c=e.getItemGraphicEl(s.dataIndex),d=c.__edge,h=u.__edge||(!1===s.isExpand||1===s.children.length?d:void 0),f=r.get("edgeShape"),p=r.get("layout"),m=r.get("orient"),g=r.get(["lineStyle","curveness"]);h&&("curve"===f?$l(h,{shape:RA(p,m,g,l,l),style:{opacity:0}},r,{cb:function(){n.remove(h)},removeOpt:i}):"polyline"===f&&"orthogonal"===r.get("layout")&&$l(h,{shape:{parentPoint:[l.x,l.y],childPoints:[[l.x,l.y]]},style:{opacity:0}},r,{cb:function(){n.remove(h)},removeOpt:i}))}}function EA(t,e){var n,r=e.parentNode===t?e:e.parentNode||e;while(n=r.getLayout(),null==n)r=r.parentNode===t?r:r.parentNode||r;return{source:r,sourceLayout:n}}function NA(t,e,n,r,i){var o=t.tree.getNodeByDataIndex(e),a=t.tree.root,s=EA(a,o).sourceLayout,l={duration:i.get("animationDurationUpdate"),easing:i.get("animationEasingUpdate")};$l(n,{x:s.x+1,y:s.y+1},i,{cb:function(){r.remove(n),t.setItemGraphicEl(e,null)},removeOpt:l}),n.fadeOut(null,t.hostModel,{fadeLabel:!0,animation:l}),o.children.forEach((function(e){PA(e,t,r,i,l)})),PA(o,t,r,i,l)}function RA(t,e,n,r,i){var o,a,s,l,u,c,d,h;if("radial"===t){u=r.rawX,d=r.rawY,c=i.rawX,h=i.rawY;var f=bA(u,d),p=bA(u,d+(h-d)*n),m=bA(c,h+(d-h)*n),g=bA(c,h);return{x1:f.x||0,y1:f.y||0,x2:g.x||0,y2:g.y||0,cpx1:p.x||0,cpy1:p.y||0,cpx2:m.x||0,cpy2:m.y||0}}return u=r.x,d=r.y,c=i.x,h=i.y,"LR"!==e&&"RL"!==e||(o=u+(c-u)*n,a=d,s=c+(u-c)*n,l=h),"TB"!==e&&"BT"!==e||(o=u,a=d+(h-d)*n,s=c,l=h+(d-h)*n),{x1:u,y1:d,x2:c,y2:h,cpx1:o,cpy1:a,cpx2:s,cpy2:l}}var HA=CA,zA=Ci();function FA(t){var e=t.mainData,n=t.datas;n||(n={main:e},t.datasAttr={main:"data"}),t.datas=t.mainData=null,qA(e,n,t),Object(s["k"])(n,(function(n){Object(s["k"])(e.TRANSFERABLE_METHODS,(function(e){n.wrapMethod(e,Object(s["h"])(BA,t))}))})),e.wrapMethod("cloneShallow",Object(s["h"])(VA,t)),Object(s["k"])(e.CHANGABLE_METHODS,(function(n){e.wrapMethod(n,Object(s["h"])(WA,t))})),Object(s["b"])(n[e.dataType]===e)}function BA(t,e){if(UA(this)){var n=Object(s["m"])({},zA(this).datas);n[this.dataType]=e,qA(e,n,t)}else XA(e,this.dataType,zA(this).mainData,t);return e}function WA(t,e){return t.struct&&t.struct.update(),e}function VA(t,e){return Object(s["k"])(zA(e).datas,(function(n,r){n!==e&&XA(n.cloneShallow(),r,e,t)})),e}function GA(t){var e=zA(this).mainData;return null==t||null==e?e:zA(e).datas[t]}function $A(){var t=zA(this).mainData;return null==t?[{data:t}]:Object(s["H"])(Object(s["F"])(zA(t).datas),(function(e){return{type:e,data:zA(t).datas[e]}}))}function UA(t){return zA(t).mainData===t}function qA(t,e,n){zA(t).datas={},Object(s["k"])(e,(function(e,r){XA(e,r,t,n)}))}function XA(t,e,n,r){zA(n).datas[e]=t,zA(t).mainData=n,t.dataType=e,r.struct&&(t[r.structAttr]=r.struct,r.struct[r.datasAttr[e]]=t),t.getLinkedData=GA,t.getLinkedDataAll=$A}var ZA=FA,JA=function(){function t(t,e){this.depth=0,this.height=0,this.dataIndex=-1,this.children=[],this.viewChildren=[],this.isExpand=!1,this.name=t||"",this.hostTree=e}return t.prototype.isRemoved=function(){return this.dataIndex<0},t.prototype.eachNode=function(t,e,n){s["w"](t)&&(n=e,e=t,t=null),t=t||{},s["C"](t)&&(t={order:t});var r,i=t.order||"preorder",o=this[t.attr||"children"];"preorder"===i&&(r=e.call(n,this));for(var a=0;!r&&a<o.length;a++)o[a].eachNode(t,e,n);"postorder"===i&&e.call(n,this)},t.prototype.updateDepthAndHeight=function(t){var e=0;this.depth=t;for(var n=0;n<this.children.length;n++){var r=this.children[n];r.updateDepthAndHeight(t+1),r.height>e&&(e=r.height)}this.height=e+1},t.prototype.getNodeById=function(t){if(this.getId()===t)return this;for(var e=0,n=this.children,r=n.length;e<r;e++){var i=n[e].getNodeById(t);if(i)return i}},t.prototype.contains=function(t){if(t===this)return!0;for(var e=0,n=this.children,r=n.length;e<r;e++){var i=n[e].contains(t);if(i)return i}},t.prototype.getAncestors=function(t){var e=[],n=t?this:this.parentNode;while(n)e.push(n),n=n.parentNode;return e.reverse(),e},t.prototype.getAncestorsIndices=function(){var t=[],e=this;while(e)t.push(e.dataIndex),e=e.parentNode;return t.reverse(),t},t.prototype.getDescendantIndices=function(){var t=[];return this.eachNode((function(e){t.push(e.dataIndex)})),t},t.prototype.getValue=function(t){var e=this.hostTree.data;return e.getStore().get(e.getDimensionIndex(t||"value"),this.dataIndex)},t.prototype.setLayout=function(t,e){this.dataIndex>=0&&this.hostTree.data.setItemLayout(this.dataIndex,t,e)},t.prototype.getLayout=function(){return this.hostTree.data.getItemLayout(this.dataIndex)},t.prototype.getModel=function(t){if(!(this.dataIndex<0)){var e=this.hostTree,n=e.data.getItemModel(this.dataIndex);return n.getModel(t)}},t.prototype.getLevelModel=function(){return(this.hostTree.levelModels||[])[this.depth]},t.prototype.setVisual=function(t,e){this.dataIndex>=0&&this.hostTree.data.setItemVisual(this.dataIndex,t,e)},t.prototype.getVisual=function(t){return this.hostTree.data.getItemVisual(this.dataIndex,t)},t.prototype.getRawIndex=function(){return this.hostTree.data.getRawIndex(this.dataIndex)},t.prototype.getId=function(){return this.hostTree.data.getId(this.dataIndex)},t.prototype.getChildIndex=function(){if(this.parentNode){for(var t=this.parentNode.children,e=0;e<t.length;++e)if(t[e]===this)return e;return-1}return-1},t.prototype.isAncestorOf=function(t){var e=t.parentNode;while(e){if(e===this)return!0;e=e.parentNode}return!1},t.prototype.isDescendantOf=function(t){return t!==this&&t.isAncestorOf(this)},t}(),KA=function(){function t(t){this.type="tree",this._nodes=[],this.hostModel=t}return t.prototype.eachNode=function(t,e,n){this.root.eachNode(t,e,n)},t.prototype.getNodeByDataIndex=function(t){var e=this.data.getRawIndex(t);return this._nodes[e]},t.prototype.getNodeById=function(t){return this.root.getNodeById(t)},t.prototype.update=function(){for(var t=this.data,e=this._nodes,n=0,r=e.length;n<r;n++)e[n].dataIndex=-1;for(n=0,r=t.count();n<r;n++)e[t.getRawIndex(n)].dataIndex=n},t.prototype.clearLayouts=function(){this.data.clearItemLayouts()},t.createTree=function(e,n,r){var i=new t(n),o=[],a=1;function l(t,e){var n=t.value;a=Math.max(a,s["t"](n)?n.length:1),o.push(t);var r=new JA(Mi(t.name,""),i);e?QA(r,e):i.root=r,i._nodes.push(r);var u=t.children;if(u)for(var c=0;c<u.length;c++)l(u[c],r)}l(e),i.root.updateDepthAndHeight(0);var u=qM(o,{coordDimensions:["value"],dimensionsCount:a}).dimensions,c=new UM(u,n);return c.initData(o),r&&r(c),ZA({mainData:c,struct:i,structAttr:"tree"}),i.update(),i},t}();function QA(t,e){var n=e.children;t.parentNode!==e&&(n.push(t),t.parentNode=e)}var tY=KA;function eY(t,e,n){if(t&&s["r"](e,t.type)>=0){var r=n.getData().tree.root,i=t.targetNode;if(s["C"](i)&&(i=r.getNodeById(i)),i&&r.contains(i))return{node:i};var o=t.targetNodeId;if(null!=o&&(i=r.getNodeById(o)))return{node:i}}}function nY(t){var e=[];while(t)t=t.parentNode,t&&e.push(t);return e.reverse()}function rY(t,e){var n=nY(t);return s["r"](n,e)>=0}function iY(t,e){var n=[];while(t){var r=t.dataIndex;n.push({name:t.name,dataIndex:r,value:e.getRawValue(r)}),t=t.parentNode}return n.reverse(),n}var oY=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.hasSymbolVisual=!0,e.ignoreStyleOnData=!0,e}return o(e,t),e.prototype.getInitialData=function(t){var e={name:t.name,children:t.data},n=t.leaves||{},r=new Lu(n,this,this.ecModel),i=tY.createTree(e,this,o);function o(t){t.wrapMethod("getItemModel",(function(t,e){var n=i.getNodeByDataIndex(e);return n&&n.children.length&&n.isExpand||(t.parentModel=r),t}))}var a=0;i.eachNode("preorder",(function(t){t.depth>a&&(a=t.depth)}));var s=t.expandAndCollapse,l=s&&t.initialTreeDepth>=0?t.initialTreeDepth:a;return i.root.eachNode("preorder",(function(t){var e=t.hostTree.data.getRawDataItem(t.dataIndex);t.isExpand=e&&null!=e.collapsed?!e.collapsed:t.depth<=l})),i.data},e.prototype.getOrient=function(){var t=this.get("orient");return"horizontal"===t?t="LR":"vertical"===t&&(t="TB"),t},e.prototype.setZoom=function(t){this.option.zoom=t},e.prototype.setCenter=function(t){this.option.center=t},e.prototype.formatTooltip=function(t,e,n){var r=this.getData().tree,i=r.root.children[0],o=r.getNodeByDataIndex(t),a=o.getValue(),s=o.name;while(o&&o!==i)s=o.parentNode.name+"."+s,o=o.parentNode;return Uf("nameValue",{name:s,value:a,noValue:isNaN(a)||null==a})},e.prototype.getDataParams=function(e){var n=t.prototype.getDataParams.apply(this,arguments),r=this.getData().tree.getNodeByDataIndex(e);return n.treeAncestors=iY(r,this),n.collapsed=!r.isExpand,n},e.type="series.tree",e.layoutMode="box",e.defaultOption={z:2,coordinateSystem:"view",left:"12%",top:"12%",right:"12%",bottom:"12%",layout:"orthogonal",edgeShape:"curve",edgeForkPosition:"50%",roam:!1,nodeScaleRatio:.4,center:null,zoom:1,orient:"LR",symbol:"emptyCircle",symbolSize:7,expandAndCollapse:!0,initialTreeDepth:2,lineStyle:{color:"#ccc",width:1.5,curveness:.5},itemStyle:{color:"lightsteelblue",borderWidth:1.5},label:{show:!0},animationEasing:"linear",animationDuration:700,animationDurationUpdate:500},e}(Mp),aY=oY;function sY(t,e,n){var r,i=[t],o=[];while(r=i.pop())if(o.push(r),r.isExpand){var a=r.children;if(a.length)for(var s=0;s<a.length;s++)i.push(a[s])}while(r=o.pop())e(r,n)}function lY(t,e){var n,r=[t];while(n=r.pop())if(e(n),n.isExpand){var i=n.children;if(i.length)for(var o=i.length-1;o>=0;o--)r.push(i[o])}}function uY(t,e){t.eachSeriesByType("tree",(function(t){cY(t,e)}))}function cY(t,e){var n=xA(t,e);t.layoutInfo=n;var r=t.get("layout"),i=0,o=0,a=null;"radial"===r?(i=2*Math.PI,o=Math.min(n.height,n.width)/2,a=_A((function(t,e){return(t.parentNode===e.parentNode?1:2)/t.depth}))):(i=n.width,o=n.height,a=_A());var s=t.getData().tree.root,l=s.children[0];if(l){gA(s),sY(l,yA,a),s.hierNode.modifier=-l.hierNode.prelim,lY(l,vA);var u=l,c=l,d=l;lY(l,(function(t){var e=t.getLayout().x;e<u.getLayout().x&&(u=t),e>c.getLayout().x&&(c=t),t.depth>d.depth&&(d=t)}));var h=u===c?1:a(u,c)/2,f=h-u.getLayout().x,p=0,m=0,g=0,y=0;if("radial"===r)p=i/(c.getLayout().x+h+f),m=o/(d.depth-1||1),lY(l,(function(t){g=(t.getLayout().x+f)*p,y=(t.depth-1)*m;var e=bA(g,y);t.setLayout({x:e.x,y:e.y,rawX:g,rawY:y},!0)}));else{var v=t.getOrient();"RL"===v||"LR"===v?(m=o/(c.getLayout().x+h+f),p=i/(d.depth-1||1),lY(l,(function(t){y=(t.getLayout().x+f)*m,g="LR"===v?(t.depth-1)*p:i-(t.depth-1)*p,t.setLayout({x:g,y:y},!0)}))):"TB"!==v&&"BT"!==v||(p=i/(c.getLayout().x+h+f),m=o/(d.depth-1||1),lY(l,(function(t){g=(t.getLayout().x+f)*p,y="TB"===v?(t.depth-1)*m:o-(t.depth-1)*m,t.setLayout({x:g,y:y},!0)})))}}}function dY(t){t.eachSeriesByType("tree",(function(t){var e=t.getData(),n=e.tree;n.eachNode((function(t){var n=t.getModel(),r=n.getModel("itemStyle").getItemStyle(),i=e.ensureUniqueItemVisual(t.dataIndex,"style");Object(s["m"])(i,r)}))}))}function hY(t){t.registerAction({type:"treeExpandAndCollapse",event:"treeExpandAndCollapse",update:"update"},(function(t,e){e.eachComponent({mainType:"series",subType:"tree",query:t},(function(e){var n=t.dataIndex,r=e.getData().tree,i=r.getNodeByDataIndex(n);i.isExpand=!i.isExpand}))})),t.registerAction({type:"treeRoam",event:"treeRoam",update:"none"},(function(t,e,n){e.eachComponent({mainType:"series",subType:"tree",query:t},(function(e){var r=e.coordinateSystem,i=cA(r,t,void 0,n);e.setCenter&&e.setCenter(i.center),e.setZoom&&e.setZoom(i.zoom)}))}))}function fY(t){t.registerChartView(HA),t.registerSeriesModel(aY),t.registerLayout(uY),t.registerVisual(dY),hY(t)}var pY=["treemapZoomToNode","treemapRender","treemapMove"];function mY(t){for(var e=0;e<pY.length;e++)t.registerAction({type:pY[e],update:"updateView"},s["L"]);t.registerAction({type:"treemapRootToNode",update:"updateView"},(function(t,e){function n(e,n){var r=["treemapZoomToNode","treemapRootToNode"],i=eY(t,r,e);if(i){var o=e.getViewRoot();o&&(t.direction=rY(o,i.node)?"rollUp":"drillDown"),e.resetViewRoot(i.node)}}e.eachComponent({mainType:"series",subType:"treemap",query:t},n)}))}function gY(t){var e=t.getData(),n=e.tree,r={};n.eachNode((function(e){var n=e;while(n&&n.depth>1)n=n.parentNode;var i=Md(t.ecModel,n.name||n.dataIndex+"",r);e.setVisual("decal",i)}))}var yY=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.preventUsingHoverLayer=!0,n}return o(e,t),e.prototype.getInitialData=function(t,e){var n={name:t.name,children:t.data};vY(n);var r=t.levels||[],i=this.designatedVisualItemStyle={},o=new Lu({itemStyle:i},this,e);r=t.levels=_Y(r,e);var a=s["H"](r||[],(function(t){return new Lu(t,o,e)}),this),l=tY.createTree(n,this,u);function u(t){t.wrapMethod("getItemModel",(function(t,e){var n=l.getNodeByDataIndex(e),r=n?a[n.depth]:null;return t.parentModel=r||o,t}))}return l.data},e.prototype.optionUpdated=function(){this.resetViewRoot()},e.prototype.formatTooltip=function(t,e,n){var r=this.getData(),i=this.getRawValue(t),o=r.getName(t);return Uf("nameValue",{name:o,value:i})},e.prototype.getDataParams=function(e){var n=t.prototype.getDataParams.apply(this,arguments),r=this.getData().tree.getNodeByDataIndex(e);return n.treeAncestors=iY(r,this),n.treePathInfo=n.treeAncestors,n},e.prototype.setLayoutInfo=function(t){this.layoutInfo=this.layoutInfo||{},s["m"](this.layoutInfo,t)},e.prototype.mapIdToIndex=function(t){var e=this._idIndexMap;e||(e=this._idIndexMap=s["f"](),this._idIndexMapCount=0);var n=e.get(t);return null==n&&e.set(t,n=this._idIndexMapCount++),n},e.prototype.getViewRoot=function(){return this._viewRoot},e.prototype.resetViewRoot=function(t){t?this._viewRoot=t:t=this._viewRoot;var e=this.getRawData().tree.root;t&&(t===e||e.contains(t))||(this._viewRoot=e)},e.prototype.enableAriaDecal=function(){gY(this)},e.type="series.treemap",e.layoutMode="box",e.defaultOption={progressive:0,left:"center",top:"middle",width:"80%",height:"80%",sort:!0,clipWindow:"origin",squareRatio:.5*(1+Math.sqrt(5)),leafDepth:null,drillDownIcon:"▶",zoomToNodeRatio:.1024,roam:!0,nodeClick:"zoomToNode",animation:!0,animationDurationUpdate:900,animationEasing:"quinticInOut",breadcrumb:{show:!0,height:22,left:"center",top:"bottom",emptyItemWidth:25,itemStyle:{color:"rgba(0,0,0,0.7)",textStyle:{color:"#fff"}},emphasis:{itemStyle:{color:"rgba(0,0,0,0.9)"}}},label:{show:!0,distance:0,padding:5,position:"inside",color:"#fff",overflow:"truncate"},upperLabel:{show:!1,position:[0,"50%"],height:20,overflow:"truncate",verticalAlign:"middle"},itemStyle:{color:null,colorAlpha:null,colorSaturation:null,borderWidth:0,gapWidth:0,borderColor:"#fff",borderColorSaturation:null},emphasis:{upperLabel:{show:!0,position:[0,"50%"],overflow:"truncate",verticalAlign:"middle"}},visualDimension:0,visualMin:null,visualMax:null,color:[],colorAlpha:null,colorSaturation:null,colorMappingBy:"index",visibleMin:10,childrenVisibleMin:null,levels:[]},e}(Mp);function vY(t){var e=0;s["k"](t.children,(function(t){vY(t);var n=t.value;s["t"](n)&&(n=n[0]),e+=n}));var n=t.value;s["t"](n)&&(n=n[0]),(null==n||isNaN(n))&&(n=e),n<0&&(n=0),s["t"](t.value)?t.value[0]=n:t.value=n}function _Y(t,e){var n=ui(e.get("color")),r=ui(e.get(["aria","decal","decals"]));if(n){var i,o;t=t||[],s["k"](t,(function(t){var e=new Lu(t),n=e.get("color"),r=e.get("decal");(e.get(["itemStyle","color"])||n&&"none"!==n)&&(i=!0),(e.get(["itemStyle","decal"])||r&&"none"!==r)&&(o=!0)}));var a=t[0]||(t[0]={});return i||(a.color=n.slice()),!o&&r&&(a.decal=r.slice()),t}}var bY=yY,xY=8,wY=8,MY=5,SY=function(){function t(t){this.group=new wr,t.add(this.group)}return t.prototype.render=function(t,e,n,r){var i=t.getModel("breadcrumb"),o=this.group;if(o.removeAll(),i.get("show")&&n){var a=i.getModel("itemStyle"),s=i.getModel("emphasis"),l=a.getModel("textStyle"),u=s.getModel(["itemStyle","textStyle"]),c={pos:{left:i.get("left"),right:i.get("right"),top:i.get("top"),bottom:i.get("bottom")},box:{width:e.getWidth(),height:e.getHeight()},emptyItemWidth:i.get("emptyItemWidth"),totalWidth:0,renderList:[]};this._prepare(n,c,l),this._renderContent(t,c,a,s,l,u,r),zc(o,c.pos,c.box)}},t.prototype._prepare=function(t,e,n){for(var r=t;r;r=r.parentNode){var i=Mi(r.getModel().get("name"),""),o=n.getTextRect(i),a=Math.max(o.width+2*xY,e.emptyItemWidth);e.totalWidth+=a+wY,e.renderList.push({node:r,text:i,width:a})}},t.prototype._renderContent=function(t,e,n,r,i,o,a){for(var l=0,u=e.emptyItemWidth,c=t.get(["breadcrumb","height"]),d=Rc(e.pos,e.box),h=e.totalWidth,f=e.renderList,p=r.getModel("itemStyle").getItemStyle(),m=f.length-1;m>=0;m--){var g=f[m],y=g.node,v=g.width,_=g.text;h>d.width&&(h-=v-u,v=u,_=null);var b=new Tm({shape:{points:kY(l,0,v,c,m===f.length-1,0===m)},style:Object(s["i"])(n.getItemStyle(),{lineJoin:"bevel"}),textContent:new ks({style:nu(i,{text:_})}),textConfig:{position:"inside"},z2:1e4*Ns,onclick:Object(s["h"])(a,y)});b.disableLabelAnimation=!0,b.getTextContent().ensureState("emphasis").style=nu(o,{text:_}),b.ensureState("emphasis").style=p,Dl(b,r.get("focus"),r.get("blurScope"),r.get("disabled")),this.group.add(b),LY(b,t,y),l+=v+wY}},t.prototype.remove=function(){this.group.removeAll()},t}();function kY(t,e,n,r,i,o){var a=[[i?t:t-MY,e],[t+n,e],[t+n,e+r],[i?t:t-MY,e+r]];return!o&&a.splice(2,0,[t+n+MY,e+r/2]),!i&&a.push([t,e+r/2]),a}function LY(t,e,n){Ls(t).eventData={componentType:"series",componentSubType:"treemap",componentIndex:e.componentIndex,seriesIndex:e.seriesIndex,seriesName:e.name,seriesType:"treemap",selfType:"breadcrumb",nodeData:{dataIndex:n&&n.dataIndex,name:n&&n.name},treePathInfo:n&&iY(n,e)}}var TY=SY,DY=function(){function t(){this._storage=[],this._elExistsMap={}}return t.prototype.add=function(t,e,n,r,i){return!this._elExistsMap[t.id]&&(this._elExistsMap[t.id]=!0,this._storage.push({el:t,target:e,duration:n,delay:r,easing:i}),!0)},t.prototype.finished=function(t){return this._finishedCallback=t,this},t.prototype.start=function(){for(var t=this,e=this._storage.length,n=function(){e--,e<=0&&(t._storage.length=0,t._elExistsMap={},t._finishedCallback&&t._finishedCallback())},r=0,i=this._storage.length;r<i;r++){var o=this._storage[r];o.el.animateTo(o.target,{duration:o.duration,delay:o.delay,easing:o.easing,setToFinal:!0,done:n,aborted:n})}return this},t}();function OY(){return new DY}var IY=wr,CY=ss,AY=3,YY="label",jY="upperLabel",PY=10*Ns,EY=2*Ns,NY=3*Ns,RY=ro([["fill","color"],["stroke","strokeColor"],["lineWidth","strokeWidth"],["shadowBlur"],["shadowOffsetX"],["shadowOffsetY"],["shadowColor"]]),HY=function(t){var e=RY(t);return e.stroke=e.fill=e.lineWidth=null,e},zY=Ci(),FY=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n._state="ready",n._storage=BY(),n}return o(e,t),e.prototype.render=function(t,e,n,r){var i=e.findComponents({mainType:"series",subType:"treemap",query:r});if(!(Object(s["r"])(i,t)<0)){this.seriesModel=t,this.api=n,this.ecModel=e;var o=["treemapZoomToNode","treemapRootToNode"],a=eY(r,o,t),l=r&&r.type,u=t.layoutInfo,c=!this._oldTree,d=this._storage,h="treemapRootToNode"===l&&a&&d?{rootNodeGroup:d.nodeGroup[a.node.getRawIndex()],direction:r.direction}:null,f=this._giveContainerGroup(u),p=t.get("animation"),m=this._doRender(f,t,h);!p||c||l&&"treemapZoomToNode"!==l&&"treemapRootToNode"!==l?m.renderFinally():this._doAnimation(f,m,t,h),this._resetController(n),this._renderBreadcrumb(t,n,a)}},e.prototype._giveContainerGroup=function(t){var e=this._containerGroup;return e||(e=this._containerGroup=new IY,this._initEvents(e),this.group.add(e)),e.x=t.x,e.y=t.y,e},e.prototype._doRender=function(t,e,n){var r=e.getData().tree,i=this._oldTree,o=BY(),a=BY(),l=this._storage,u=[];function c(t,r,i,s){return WY(e,a,l,n,o,u,t,r,i,s)}h(r.root?[r.root]:[],i&&i.root?[i.root]:[],t,r===i||!i,0);var d=f(l);return this._oldTree=r,this._storage=a,{lastsForAnimation:o,willDeleteEls:d,renderFinally:p};function h(t,e,n,r,i){function o(t){return t.getId()}function a(o,a){var s=null!=o?t[o]:null,l=null!=a?e[a]:null,u=c(s,l,n,i);u&&h(s&&s.viewChildren||[],l&&l.viewChildren||[],u,r,i+1)}r?(e=t,Object(s["k"])(t,(function(t,e){!t.isRemoved()&&a(e,e)}))):new yM(e,t,o,o).add(a).update(a).remove(Object(s["h"])(a,null)).execute()}function f(t){var e=BY();return t&&Object(s["k"])(t,(function(t,n){var r=e[n];Object(s["k"])(t,(function(t){t&&(r.push(t),zY(t).willDelete=!0)}))})),e}function p(){Object(s["k"])(d,(function(t){Object(s["k"])(t,(function(t){t.parent&&t.parent.remove(t)}))})),Object(s["k"])(u,(function(t){t.invisible=!0,t.dirty()}))}},e.prototype._doAnimation=function(t,e,n,r){var i=n.get("animationDurationUpdate"),o=n.get("animationEasing"),a=(Object(s["w"])(i)?0:i)||0,l=(Object(s["w"])(o)?null:o)||"cubicOut",u=OY();Object(s["k"])(e.willDeleteEls,(function(t,e){Object(s["k"])(t,(function(t,n){if(!t.invisible){var i,o=t.parent,s=zY(o);if(r&&"drillDown"===r.direction)i=o===r.rootNodeGroup?{shape:{x:0,y:0,width:s.nodeWidth,height:s.nodeHeight},style:{opacity:0}}:{style:{opacity:0}};else{var c=0,d=0;s.willDelete||(c=s.nodeWidth/2,d=s.nodeHeight/2),i="nodeGroup"===e?{x:c,y:d,style:{opacity:0}}:{shape:{x:c,y:d,width:0,height:0},style:{opacity:0}}}i&&u.add(t,i,a,0,l)}}))})),Object(s["k"])(this._storage,(function(t,n){Object(s["k"])(t,(function(t,r){var i=e.lastsForAnimation[n][r],o={};i&&(t instanceof wr?null!=i.oldX&&(o.x=t.x,o.y=t.y,t.x=i.oldX,t.y=i.oldY):(i.oldShape&&(o.shape=Object(s["m"])({},t.shape),t.setShape(i.oldShape)),i.fadein?(t.setStyle("opacity",0),o.style={opacity:1}):1!==t.style.opacity&&(o.style={opacity:1})),u.add(t,o,a,0,l))}))}),this),this._state="animating",u.finished(Object(s["c"])((function(){this._state="ready",e.renderFinally()}),this)).start()},e.prototype._resetController=function(t){var e=this._controller;e||(e=this._controller=new dI(t.getZr()),e.enable(this.seriesModel.get("roam")),e.on("pan",Object(s["c"])(this._onPan,this)),e.on("zoom",Object(s["c"])(this._onZoom,this)));var n=new Tt(0,0,t.getWidth(),t.getHeight());e.setPointerChecker((function(t,e,r){return n.contain(e,r)}))},e.prototype._clearController=function(){var t=this._controller;t&&(t.dispose(),t=null)},e.prototype._onPan=function(t){if("animating"!==this._state&&(Math.abs(t.dx)>AY||Math.abs(t.dy)>AY)){var e=this.seriesModel.getData().tree.root;if(!e)return;var n=e.getLayout();if(!n)return;this.api.dispatchAction({type:"treemapMove",from:this.uid,seriesId:this.seriesModel.id,rootRect:{x:n.x+t.dx,y:n.y+t.dy,width:n.width,height:n.height}})}},e.prototype._onZoom=function(t){var e=t.originX,n=t.originY;if("animating"!==this._state){var r=this.seriesModel.getData().tree.root;if(!r)return;var i=r.getLayout();if(!i)return;var o=new Tt(i.x,i.y,i.width,i.height),a=this.seriesModel.layoutInfo;e-=a.x,n-=a.y;var s=st();dt(s,s,[-e,-n]),ft(s,s,[t.scale,t.scale]),dt(s,s,[e,n]),o.applyTransform(s),this.api.dispatchAction({type:"treemapRender",from:this.uid,seriesId:this.seriesModel.id,rootRect:{x:o.x,y:o.y,width:o.width,height:o.height}})}},e.prototype._initEvents=function(t){var e=this;t.on("click",(function(t){if("ready"===e._state){var n=e.seriesModel.get("nodeClick",!0);if(n){var r=e.findTarget(t.offsetX,t.offsetY);if(r){var i=r.node;if(i.getLayout().isLeafRoot)e._rootToNode(r);else if("zoomToNode"===n)e._zoomToNode(r);else if("link"===n){var o=i.hostTree.data.getItemModel(i.dataIndex),a=o.get("link",!0),s=o.get("target",!0)||"blank";a&&Ac(a,s)}}}}}),this)},e.prototype._renderBreadcrumb=function(t,e,n){var r=this;n||(n=null!=t.get("leafDepth",!0)?{node:t.getViewRoot()}:this.findTarget(e.getWidth()/2,e.getHeight()/2),n||(n={node:t.getData().tree.root})),(this._breadcrumb||(this._breadcrumb=new TY(this.group))).render(t,e,n.node,(function(e){"animating"!==r._state&&(rY(t.getViewRoot(),e)?r._rootToNode({node:e}):r._zoomToNode({node:e}))}))},e.prototype.remove=function(){this._clearController(),this._containerGroup&&this._containerGroup.removeAll(),this._storage=BY(),this._state="ready",this._breadcrumb&&this._breadcrumb.remove()},e.prototype.dispose=function(){this._clearController()},e.prototype._zoomToNode=function(t){this.api.dispatchAction({type:"treemapZoomToNode",from:this.uid,seriesId:this.seriesModel.id,targetNode:t.node})},e.prototype._rootToNode=function(t){this.api.dispatchAction({type:"treemapRootToNode",from:this.uid,seriesId:this.seriesModel.id,targetNode:t.node})},e.prototype.findTarget=function(t,e){var n,r=this.seriesModel.getViewRoot();return r.eachNode({attr:"viewChildren",order:"preorder"},(function(r){var i=this._storage.background[r.getRawIndex()];if(i){var o=i.transformCoordToLocal(t,e),a=i.shape;if(!(a.x<=o[0]&&o[0]<=a.x+a.width&&a.y<=o[1]&&o[1]<=a.y+a.height))return!1;n={node:r,offsetX:o[0],offsetY:o[1]}}}),this),n},e.type="treemap",e}($g);function BY(){return{nodeGroup:[],background:[],content:[]}}function WY(t,e,n,r,i,o,a,l,u,c){if(a){var d=a.getLayout(),h=t.getData(),f=a.getModel();if(h.setItemGraphicEl(a.dataIndex,null),d&&d.isInView){var p=d.width,m=d.height,g=d.borderWidth,y=d.invisible,v=a.getRawIndex(),_=l&&l.getRawIndex(),b=a.viewChildren,x=d.upperHeight,w=b&&b.length,M=f.getModel("itemStyle"),S=f.getModel(["emphasis","itemStyle"]),k=f.getModel(["blur","itemStyle"]),L=f.getModel(["select","itemStyle"]),T=M.get("borderRadius")||0,D=F("nodeGroup",IY);if(D){if(u.add(D),D.x=d.x||0,D.y=d.y||0,D.markRedraw(),zY(D).nodeWidth=p,zY(D).nodeHeight=m,d.isAboveViewRoot)return D;var O=F("background",CY,c,EY);O&&E(D,O,w&&d.upperLabelHeight);var I=f.getModel("emphasis"),C=I.get("focus"),A=I.get("blurScope"),Y=I.get("disabled"),j="ancestor"===C?a.getAncestorsIndices():"descendant"===C?a.getDescendantIndices():C;if(w)jl(D)&&Yl(D,!1),O&&(Yl(O,!Y),h.setItemGraphicEl(a.dataIndex,O),Ol(O,j,A));else{var P=F("content",CY,c,NY);P&&N(D,P),O.disableMorphing=!0,O&&jl(O)&&Yl(O,!1),Yl(D,!Y),h.setItemGraphicEl(a.dataIndex,D),Ol(D,j,A)}return D}}}function E(e,n,r){var i=Ls(n);if(i.dataIndex=a.dataIndex,i.seriesIndex=t.seriesIndex,n.setShape({x:0,y:0,width:p,height:m,r:T}),y)R(n);else{n.invisible=!1;var o=a.getVisual("style"),s=o.stroke,l=HY(M);l.fill=s;var u=RY(S);u.fill=S.get("borderColor");var c=RY(k);c.fill=k.get("borderColor");var d=RY(L);if(d.fill=L.get("borderColor"),r){var h=p-2*g;H(n,s,o.opacity,{x:g,y:0,width:h,height:x})}else n.removeTextContent();n.setStyle(l),n.ensureState("emphasis").style=u,n.ensureState("blur").style=c,n.ensureState("select").style=d,al(n)}e.add(n)}function N(e,n){var r=Ls(n);r.dataIndex=a.dataIndex,r.seriesIndex=t.seriesIndex;var i=Math.max(p-2*g,0),o=Math.max(m-2*g,0);if(n.culling=!0,n.setShape({x:g,y:g,width:i,height:o,r:T}),y)R(n);else{n.invisible=!1;var s=a.getVisual("style"),l=s.fill,u=HY(M);u.fill=l,u.decal=s.decal;var c=RY(S),d=RY(k),h=RY(L);H(n,l,s.opacity,null),n.setStyle(u),n.ensureState("emphasis").style=c,n.ensureState("blur").style=d,n.ensureState("select").style=h,al(n)}e.add(n)}function R(t){!t.invisible&&o.push(t)}function H(e,n,r,i){var o=f.getModel(i?jY:YY),l=Mi(f.get("name"),null),u=o.getShallow("show");tu(e,eu(f,i?jY:YY),{defaultText:u?l:null,inheritColor:n,defaultOpacity:r,labelFetcher:t,labelDataIndex:a.dataIndex});var c=e.getTextContent();if(c){var h=c.style,p=Object(s["M"])(h.padding||0);i&&(e.setTextConfig({layoutRect:i}),c.disableLabelLayout=!0),c.beforeUpdate=function(){var t=Math.max((i?i.width:e.shape.width)-p[1]-p[3],0),n=Math.max((i?i.height:e.shape.height)-p[0]-p[2],0);h.width===t&&h.height===n||c.setStyle({width:t,height:n})},h.truncateMinChar=2,h.lineOverflow="truncate",z(h,i,d);var m=c.getState("emphasis");z(m?m.style:null,i,d)}}function z(e,n,r){var i=e?e.text:null;if(!n&&r.isLeafRoot&&null!=i){var o=t.get("drillDownIcon",!0);e.text=o?o+" "+i:i}}function F(t,r,o,a){var s=null!=_&&n[t][_],l=i[t];return s?(n[t][_]=null,B(l,s)):y||(s=new r,s instanceof No&&(s.z2=VY(o,a)),W(l,s)),e[t][v]=s}function B(t,e){var n=t[v]={};e instanceof IY?(n.oldX=e.x,n.oldY=e.y):n.oldShape=Object(s["m"])({},e.shape)}function W(t,e){var n=t[v]={},o=a.parentNode,s=e instanceof wr;if(o&&(!r||"drillDown"===r.direction)){var l=0,u=0,c=i.background[o.getRawIndex()];!r&&c&&c.oldShape&&(l=c.oldShape.width,u=c.oldShape.height),s?(n.oldX=0,n.oldY=u):n.oldShape={x:l,y:u,width:0,height:0}}n.fadein=!s}}function VY(t,e){return t*PY+e}var GY=FY,$Y=s["k"],UY=s["A"],qY=-1,XY=function(){function t(e){var n=e.mappingMethod,r=e.type,i=this.option=s["d"](e);this.type=r,this.mappingMethod=n,this._normalizeData=sj[n];var o=t.visualHandlers[r];this.applyVisual=o.applyVisual,this.getColorMapper=o.getColorMapper,this._normalizedToVisual=o._normalizedToVisual[n],"piecewise"===n?(KY(i),ZY(i)):"category"===n?i.categories?JY(i):KY(i,!0):(s["b"]("linear"!==n||i.dataExtent),KY(i))}return t.prototype.mapValueToVisual=function(t){var e=this._normalizeData(t);return this._normalizedToVisual(e,t)},t.prototype.getNormalizer=function(){return s["c"](this._normalizeData,this)},t.listVisualTypes=function(){return s["F"](t.visualHandlers)},t.isValidType=function(e){return t.visualHandlers.hasOwnProperty(e)},t.eachVisual=function(t,e,n){s["A"](t)?s["k"](t,e,n):e.call(n,t)},t.mapVisual=function(e,n,r){var i,o=s["t"](e)?[]:s["A"](e)?{}:(i=!0,null);return t.eachVisual(e,(function(t,e){var a=n.call(r,t,e);i?o=a:o[e]=a})),o},t.retrieveVisuals=function(e){var n,r={};return e&&$Y(t.visualHandlers,(function(t,i){e.hasOwnProperty(i)&&(r[i]=e[i],n=!0)})),n?r:null},t.prepareVisualTypes=function(t){if(s["t"](t))t=t.slice();else{if(!UY(t))return[];var e=[];$Y(t,(function(t,n){e.push(n)})),t=e}return t.sort((function(t,e){return"color"===e&&"color"!==t&&0===t.indexOf("color")?1:-1})),t},t.dependsOn=function(t,e){return"color"===e?!(!t||0!==t.indexOf(e)):t===e},t.findPieceIndex=function(t,e,n){for(var r,i=1/0,o=0,a=e.length;o<a;o++){var l=e[o].value;if(null!=l){if(l===t||s["C"](l)&&l===t+"")return o;n&&h(l,o)}}for(o=0,a=e.length;o<a;o++){var u=e[o],c=u.interval,d=u.close;if(c){if(c[0]===-1/0){if(lj(d[1],t,c[1]))return o}else if(c[1]===1/0){if(lj(d[0],c[0],t))return o}else if(lj(d[0],c[0],t)&&lj(d[1],t,c[1]))return o;n&&h(c[0],o),n&&h(c[1],o)}}if(n)return t===1/0?e.length-1:t===-1/0?0:r;function h(e,n){var o=Math.abs(e-t);o<i&&(i=o,r=n)}},t.visualHandlers={color:{applyVisual:ej("color"),getColorMapper:function(){var t=this.option;return s["c"]("category"===t.mappingMethod?function(t,e){return!e&&(t=this._normalizeData(t)),nj.call(this,t)}:function(e,n,r){var i=!!r;return!n&&(e=this._normalizeData(e)),r=Ne["a"](e,t.parsedVisual,r),i?r:Ne["i"](r,"rgba")},this)},_normalizedToVisual:{linear:function(t){return Ne["i"](Ne["a"](t,this.option.parsedVisual),"rgba")},category:nj,piecewise:function(t,e){var n=oj.call(this,e);return null==n&&(n=Ne["i"](Ne["a"](t,this.option.parsedVisual),"rgba")),n},fixed:rj}},colorHue:QY((function(t,e){return Ne["g"](t,e)})),colorSaturation:QY((function(t,e){return Ne["g"](t,null,e)})),colorLightness:QY((function(t,e){return Ne["g"](t,null,null,e)})),colorAlpha:QY((function(t,e){return Ne["f"](t,e)})),decal:{applyVisual:ej("decal"),_normalizedToVisual:{linear:null,category:nj,piecewise:null,fixed:null}},opacity:{applyVisual:ej("opacity"),_normalizedToVisual:ij([0,1])},liftZ:{applyVisual:ej("liftZ"),_normalizedToVisual:{linear:rj,category:rj,piecewise:rj,fixed:rj}},symbol:{applyVisual:function(t,e,n){var r=this.mapValueToVisual(t);n("symbol",r)},_normalizedToVisual:{linear:tj,category:nj,piecewise:function(t,e){var n=oj.call(this,e);return null==n&&(n=tj.call(this,t)),n},fixed:rj}},symbolSize:{applyVisual:ej("symbolSize"),_normalizedToVisual:ij([0,1])}},t}();function ZY(t){var e=t.pieceList;t.hasSpecialVisual=!1,s["k"](e,(function(e,n){e.originIndex=n,null!=e.visual&&(t.hasSpecialVisual=!0)}))}function JY(t){var e=t.categories,n=t.categoryMap={},r=t.visual;if($Y(e,(function(t,e){n[t]=e})),!s["t"](r)){var i=[];s["A"](r)?$Y(r,(function(t,e){var r=n[e];i[null!=r?r:qY]=t})):i[qY]=r,r=aj(t,i)}for(var o=e.length-1;o>=0;o--)null==r[o]&&(delete n[e[o]],e.pop())}function KY(t,e){var n=t.visual,r=[];s["A"](n)?$Y(n,(function(t){r.push(t)})):null!=n&&r.push(n);var i={color:1,symbol:1};e||1!==r.length||i.hasOwnProperty(t.type)||(r[1]=r[0]),aj(t,r)}function QY(t){return{applyVisual:function(e,n,r){var i=this.mapValueToVisual(e);r("color",t(n("color"),i))},_normalizedToVisual:ij([0,1])}}function tj(t){var e=this.option.visual;return e[Math.round(Er(t,[0,1],[0,e.length-1],!0))]||{}}function ej(t){return function(e,n,r){r(t,this.mapValueToVisual(e))}}function nj(t){var e=this.option.visual;return e[this.option.loop&&t!==qY?t%e.length:t]}function rj(){return this.option.visual[0]}function ij(t){return{linear:function(e){return Er(e,t,this.option.visual,!0)},category:nj,piecewise:function(e,n){var r=oj.call(this,n);return null==r&&(r=Er(e,t,this.option.visual,!0)),r},fixed:rj}}function oj(t){var e=this.option,n=e.pieceList;if(e.hasSpecialVisual){var r=XY.findPieceIndex(t,n),i=n[r];if(i&&i.visual)return i.visual[this.type]}}function aj(t,e){return t.visual=e,"color"===t.type&&(t.parsedVisual=s["H"](e,(function(t){var e=Ne["h"](t);return e||[0,0,0,1]}))),e}var sj={linear:function(t){return Er(t,this.option.dataExtent,[0,1],!0)},piecewise:function(t){var e=this.option.pieceList,n=XY.findPieceIndex(t,e,!0);if(null!=n)return Er(n,[0,e.length-1],[0,1],!0)},category:function(t){var e=this.option.categories?this.option.categoryMap[t]:t;return null==e?qY:e},fixed:s["L"]};function lj(t,e,n){return t?e<=n:e<n}var uj=XY,cj="itemStyle",dj=Ci(),hj={seriesType:"treemap",reset:function(t){var e=t.getData().tree,n=e.root;n.isRemoved()||fj(n,{},t.getViewRoot().getAncestors(),t)}};function fj(t,e,n,r){var i=t.getModel(),o=t.getLayout(),a=t.hostTree.data;if(o&&!o.invisible&&o.isInView){var l,u=i.getModel(cj),c=pj(u,e,r),d=a.ensureUniqueItemVisual(t.dataIndex,"style"),h=u.get("borderColor"),f=u.get("borderColorSaturation");null!=f&&(l=mj(c),h=gj(f,l)),d.stroke=h;var p=t.viewChildren;if(p&&p.length){var m=vj(t,i,o,u,c,p);Object(s["k"])(p,(function(t,e){if(t.depth>=n.length||t===n[t.depth]){var o=bj(i,c,t,e,m,r);fj(t,o,n,r)}}))}else l=mj(c),d.fill=l}}function pj(t,e,n){var r=Object(s["m"])({},e),i=n.designatedVisualItemStyle;return Object(s["k"])(["color","colorAlpha","colorSaturation"],(function(n){i[n]=e[n];var o=t.get(n);i[n]=null,null!=o&&(r[n]=o)})),r}function mj(t){var e=yj(t,"color");if(e){var n=yj(t,"colorAlpha"),r=yj(t,"colorSaturation");return r&&(e=Object(Ne["g"])(e,null,null,r)),n&&(e=Object(Ne["f"])(e,n)),e}}function gj(t,e){return null!=e?Object(Ne["g"])(e,null,null,t):null}function yj(t,e){var n=t[e];if(null!=n&&"none"!==n)return n}function vj(t,e,n,r,i,o){if(o&&o.length){var a=_j(e,"color")||null!=i.color&&"none"!==i.color&&(_j(e,"colorAlpha")||_j(e,"colorSaturation"));if(a){var s=e.get("visualMin"),l=e.get("visualMax"),u=n.dataExtent.slice();null!=s&&s<u[0]&&(u[0]=s),null!=l&&l>u[1]&&(u[1]=l);var c=e.get("colorMappingBy"),d={type:a.name,dataExtent:u,visual:a.range};"color"!==d.type||"index"!==c&&"id"!==c?d.mappingMethod="linear":(d.mappingMethod="category",d.loop=!0);var h=new uj(d);return dj(h).drColorMappingBy=c,h}}}function _j(t,e){var n=t.get(e);return Object(s["t"])(n)&&n.length?{name:e,range:n}:null}function bj(t,e,n,r,i,o){var a=Object(s["m"])({},e);if(i){var l=i.type,u="color"===l&&dj(i).drColorMappingBy,c="index"===u?r:"id"===u?o.mapIdToIndex(n.getId()):n.getValue(t.get("visualDimension"));a[l]=i.mapValueToVisual(c)}return a}var xj=Math.max,wj=Math.min,Mj=s["O"],Sj=s["k"],kj=["itemStyle","borderWidth"],Lj=["itemStyle","gapWidth"],Tj=["upperLabel","show"],Dj=["upperLabel","height"],Oj={seriesType:"treemap",reset:function(t,e,n,r){var i=n.getWidth(),o=n.getHeight(),a=t.option,l=Hc(t.getBoxLayoutParams(),{width:n.getWidth(),height:n.getHeight()}),u=a.size||[],c=Nr(Mj(l.width,u[0]),i),d=Nr(Mj(l.height,u[1]),o),h=r&&r.type,f=["treemapZoomToNode","treemapRootToNode"],p=eY(r,f,t),m="treemapRender"===h||"treemapMove"===h?r.rootRect:null,g=t.getViewRoot(),y=nY(g);if("treemapMove"!==h){var v="treemapZoomToNode"===h?Nj(t,p,g,c,d):m?[m.width,m.height]:[c,d],_=a.sort;_&&"asc"!==_&&"desc"!==_&&(_="desc");var b={squareRatio:a.squareRatio,sort:_,leafDepth:a.leafDepth};g.hostTree.clearLayouts();var x={x:0,y:0,width:v[0],height:v[1],area:v[0]*v[1]};g.setLayout(x),Ij(g,b,!1,0),x=g.getLayout(),Sj(y,(function(t,e){var n=(y[e+1]||g).getValue();t.setLayout(s["m"]({dataExtent:[n,n],borderWidth:0,upperHeight:0},x))}))}var w=t.getData().tree.root;w.setLayout(Rj(l,m,p),!0),t.setLayoutInfo(l),Hj(w,new Tt(-l.x,-l.y,i,o),y,g,0)}};function Ij(t,e,n,r){var i,o;if(!t.isRemoved()){var a=t.getLayout();i=a.width,o=a.height;var s=t.getModel(),l=s.get(kj),u=s.get(Lj)/2,c=zj(s),d=Math.max(l,c),h=l-u,f=d-u;t.setLayout({borderWidth:l,upperHeight:d,upperLabelHeight:c},!0),i=xj(i-2*h,0),o=xj(o-h-f,0);var p=i*o,m=Cj(t,s,p,e,n,r);if(m.length){var g={x:h,y:f,width:i,height:o},y=wj(i,o),v=1/0,_=[];_.area=0;for(var b=0,x=m.length;b<x;){var w=m[b];_.push(w),_.area+=w.getLayout().area;var M=Pj(_,y,e.squareRatio);M<=v?(b++,v=M):(_.area-=_.pop().getLayout().area,Ej(_,y,g,u,!1),y=wj(g.width,g.height),_.length=_.area=0,v=1/0)}if(_.length&&Ej(_,y,g,u,!0),!n){var S=s.get("childrenVisibleMin");null!=S&&p<S&&(n=!0)}for(b=0,x=m.length;b<x;b++)Ij(m[b],e,n,r+1)}}}function Cj(t,e,n,r,i,o){var a=t.children||[],l=r.sort;"asc"!==l&&"desc"!==l&&(l=null);var u=null!=r.leafDepth&&r.leafDepth<=o;if(i&&!u)return t.viewChildren=[];a=s["n"](a,(function(t){return!t.isRemoved()})),Yj(a,l);var c=jj(e,a,l);if(0===c.sum)return t.viewChildren=[];if(c.sum=Aj(e,n,c.sum,l,a),0===c.sum)return t.viewChildren=[];for(var d=0,h=a.length;d<h;d++){var f=a[d].getValue()/c.sum*n;a[d].setLayout({area:f})}return u&&(a.length&&t.setLayout({isLeafRoot:!0},!0),a.length=0),t.viewChildren=a,t.setLayout({dataExtent:c.dataExtent},!0),a}function Aj(t,e,n,r,i){if(!r)return n;for(var o=t.get("visibleMin"),a=i.length,s=a,l=a-1;l>=0;l--){var u=i["asc"===r?a-l-1:l].getValue();u/n*e<o&&(s=l,n-=u)}return"asc"===r?i.splice(0,a-s):i.splice(s,a-s),n}function Yj(t,e){return e&&t.sort((function(t,n){var r="asc"===e?t.getValue()-n.getValue():n.getValue()-t.getValue();return 0===r?"asc"===e?t.dataIndex-n.dataIndex:n.dataIndex-t.dataIndex:r})),t}function jj(t,e,n){for(var r=0,i=0,o=e.length;i<o;i++)r+=e[i].getValue();var a,s=t.get("visualDimension");return e&&e.length?"value"===s&&n?(a=[e[e.length-1].getValue(),e[0].getValue()],"asc"===n&&a.reverse()):(a=[1/0,-1/0],Sj(e,(function(t){var e=t.getValue(s);e<a[0]&&(a[0]=e),e>a[1]&&(a[1]=e)}))):a=[NaN,NaN],{sum:r,dataExtent:a}}function Pj(t,e,n){for(var r=0,i=1/0,o=0,a=void 0,s=t.length;o<s;o++)a=t[o].getLayout().area,a&&(a<i&&(i=a),a>r&&(r=a));var l=t.area*t.area,u=e*e*n;return l?xj(u*r/l,l/(u*i)):1/0}function Ej(t,e,n,r,i){var o=e===n.width?0:1,a=1-o,s=["x","y"],l=["width","height"],u=n[s[o]],c=e?t.area/e:0;(i||c>n[l[a]])&&(c=n[l[a]]);for(var d=0,h=t.length;d<h;d++){var f=t[d],p={},m=c?f.getLayout().area/c:0,g=p[l[a]]=xj(c-2*r,0),y=n[s[o]]+n[l[o]]-u,v=d===h-1||y<m?y:m,_=p[l[o]]=xj(v-2*r,0);p[s[a]]=n[s[a]]+wj(r,g/2),p[s[o]]=u+wj(r,_/2),u+=v,f.setLayout(p,!0)}n[s[a]]+=c,n[l[a]]-=c}function Nj(t,e,n,r,i){var o,a=(e||{}).node,s=[r,i];if(!a||a===n)return s;var l=r*i,u=l*t.option.zoomToNodeRatio;while(o=a.parentNode){for(var c=0,d=o.children,h=0,f=d.length;h<f;h++)c+=d[h].getValue();var p=a.getValue();if(0===p)return s;u*=c/p;var m=o.getModel(),g=m.get(kj),y=Math.max(g,zj(m));u+=4*g*g+(3*g+y)*Math.pow(u,.5),u>Gr&&(u=Gr),a=o}u<l&&(u=l);var v=Math.pow(u/l,.5);return[r*v,i*v]}function Rj(t,e,n){if(e)return{x:e.x,y:e.y};var r={x:0,y:0};if(!n)return r;var i=n.node,o=i.getLayout();if(!o)return r;var a=[o.width/2,o.height/2],s=i;while(s){var l=s.getLayout();a[0]+=l.x,a[1]+=l.y,s=s.parentNode}return{x:t.width/2-a[0],y:t.height/2-a[1]}}function Hj(t,e,n,r,i){var o=t.getLayout(),a=n[i],s=a&&a===t;if(!(a&&!s||i===n.length&&t!==r)){t.setLayout({isInView:!0,invisible:!s&&!e.intersect(o),isAboveViewRoot:s},!0);var l=new Tt(e.x-o.x,e.y-o.y,e.width,e.height);Sj(t.viewChildren||[],(function(t){Hj(t,l,n,r,i+1)}))}}function zj(t){return t.get(Tj)?t.get(Dj):0}function Fj(t){t.registerSeriesModel(bY),t.registerChartView(GY),t.registerVisual(hj),t.registerLayout(Oj),mY(t)}function Bj(t){var e=t.findComponents({mainType:"legend"});e&&e.length&&t.eachSeriesByType("graph",(function(t){var n=t.getCategoriesData(),r=t.getGraph(),i=r.data,o=n.mapArray(n.getName);i.filterSelf((function(t){var n=i.getItemModel(t),r=n.getShallow("category");if(null!=r){Object(s["z"])(r)&&(r=o[r]);for(var a=0;a<e.length;a++)if(!e[a].isSelected(r))return!1}return!0}))}))}function Wj(t){var e={};t.eachSeriesByType("graph",(function(t){var n=t.getCategoriesData(),r=t.getData(),i={};n.each((function(r){var o=n.getName(r);i["ec-"+o]=r;var a=n.getItemModel(r),s=a.getModel("itemStyle").getItemStyle();s.fill||(s.fill=t.getColorFromPalette(o,e)),n.setItemVisual(r,"style",s);for(var l=["symbol","symbolSize","symbolKeepAspect"],u=0;u<l.length;u++){var c=a.getShallow(l[u],!0);null!=c&&n.setItemVisual(r,l[u],c)}})),n.count()&&r.each((function(t){var e=r.getItemModel(t),o=e.getShallow("category");if(null!=o){Object(s["C"])(o)&&(o=i["ec-"+o]);var a=n.getItemVisual(o,"style"),l=r.ensureUniqueItemVisual(t,"style");Object(s["m"])(l,a);for(var u=["symbol","symbolSize","symbolKeepAspect"],c=0;c<u.length;c++)r.setItemVisual(t,u[c],n.getItemVisual(o,u[c]))}}))}))}function Vj(t){return t instanceof Array||(t=[t,t]),t}function Gj(t){t.eachSeriesByType("graph",(function(t){var e=t.getGraph(),n=t.getEdgeData(),r=Vj(t.get("edgeSymbol")),i=Vj(t.get("edgeSymbolSize"));n.setVisual("fromSymbol",r&&r[0]),n.setVisual("toSymbol",r&&r[1]),n.setVisual("fromSymbolSize",i&&i[0]),n.setVisual("toSymbolSize",i&&i[1]),n.setVisual("style",t.getModel("lineStyle").getLineStyle()),n.each((function(t){var r=n.getItemModel(t),i=e.getEdgeByIndex(t),o=Vj(r.getShallow("symbol",!0)),a=Vj(r.getShallow("symbolSize",!0)),l=r.getModel("lineStyle").getLineStyle(),u=n.ensureUniqueItemVisual(t,"style");switch(Object(s["m"])(u,l),u.stroke){case"source":var c=i.node1.getVisual("style");u.stroke=c&&c.fill;break;case"target":c=i.node2.getVisual("style");u.stroke=c&&c.fill;break}o[0]&&i.setVisual("fromSymbol",o[0]),o[1]&&i.setVisual("toSymbol",o[1]),a[0]&&i.setVisual("fromSymbolSize",a[0]),a[1]&&i.setVisual("toSymbolSize",a[1])}))}))}var $j="--\x3e",Uj=function(t){return t.get("autoCurveness")||null},qj=function(t,e){var n=Uj(t),r=20,i=[];if(s["z"](n))r=n;else if(s["t"](n))return void(t.__curvenessList=n);e>r&&(r=e);var o=r%2?r+2:r+3;i=[];for(var a=0;a<o;a++)i.push((a%2?a+1:a)/10*(a%2?-1:1));t.__curvenessList=i},Xj=function(t,e,n){var r=[t.id,t.dataIndex].join("."),i=[e.id,e.dataIndex].join(".");return[n.uid,r,i].join($j)},Zj=function(t){var e=t.split($j);return[e[0],e[2],e[1]].join($j)},Jj=function(t,e){var n=Xj(t.node1,t.node2,e);return e.__edgeMap[n]},Kj=function(t,e){var n=Qj(Xj(t.node1,t.node2,e),e),r=Qj(Xj(t.node2,t.node1,e),e);return n+r},Qj=function(t,e){var n=e.__edgeMap;return n[t]?n[t].length:0};function tP(t){Uj(t)&&(t.__curvenessList=[],t.__edgeMap={},qj(t))}function eP(t,e,n,r){if(Uj(n)){var i=Xj(t,e,n),o=n.__edgeMap,a=o[Zj(i)];o[i]&&!a?o[i].isForward=!0:a&&o[i]&&(a.isForward=!0,o[i].isForward=!1),o[i]=o[i]||[],o[i].push(r)}}function nP(t,e,n,r){var i=Uj(e),o=s["t"](i);if(!i)return null;var a=Jj(t,e);if(!a)return null;for(var l=-1,u=0;u<a.length;u++)if(a[u]===n){l=u;break}var c=Kj(t,e);qj(e,c),t.lineStyle=t.lineStyle||{};var d=Xj(t.node1,t.node2,e),h=e.__curvenessList,f=o||c%2?0:1;if(a.isForward)return h[f+l];var p=Zj(d),m=Qj(p,e),g=h[l+m+f];return r?o?i&&0===i[0]?(m+f)%2?g:-g:((m%2?0:1)+f)%2?g:-g:(m+f)%2?g:-g:h[l+m+f]}function rP(t){var e=t.coordinateSystem;if(!e||"view"===e.type){var n=t.getGraph();n.eachNode((function(t){var e=t.getModel();t.setLayout([+e.get("x"),+e.get("y")])})),iP(n,t)}}function iP(t,e){t.eachEdge((function(t,n){var r=s["Q"](t.getModel().get(["lineStyle","curveness"]),-nP(t,e,n,!0),0),i=c(t.node1.getLayout()),o=c(t.node2.getLayout()),a=[i,o];+r&&a.push([(i[0]+o[0])/2-(i[1]-o[1])*r,(i[1]+o[1])/2-(o[0]-i[0])*r]),t.setLayout(a)}))}function oP(t,e){t.eachSeriesByType("graph",(function(t){var e=t.get("layout"),n=t.coordinateSystem;if(n&&"view"!==n.type){var r=t.getData(),i=[];Object(s["k"])(n.dimensions,(function(t){i=i.concat(r.mapDimensionsAll(t))}));for(var o=0;o<r.count();o++){for(var a=[],l=!1,u=0;u<i.length;u++){var c=r.get(i[u],o);isNaN(c)||(l=!0),a.push(c)}l?r.setItemLayout(o,n.dataToPoint(a)):r.setItemLayout(o,[NaN,NaN])}iP(r.graph,t)}else e&&"none"!==e||rP(t)}))}function aP(t){var e=t.coordinateSystem;if("view"!==e.type)return 1;var n=t.option.nodeScaleRatio,r=e.scaleX,i=e.getZoom(),o=(i-1)*n+1;return o/r}function sP(t){var e=t.getVisual("symbolSize");return e instanceof Array&&(e=(e[0]+e[1])/2),+e}var lP=Math.PI,uP=[];function cP(t,e,n,r){var i=t.coordinateSystem;if(!i||"view"===i.type){var o=i.getBoundingRect(),a=t.getData(),l=a.graph,u=o.width/2+o.x,d=o.height/2+o.y,h=Math.min(o.width,o.height)/2,f=a.count();if(a.setLayout({cx:u,cy:d}),f){if(n){var p=i.pointToData(r),m=p[0],g=p[1],_=[m-u,g-d];v(_,_),y(_,_,h),n.setLayout([u+_[0],d+_[1]],!0);var b=t.get(["circular","rotateLabel"]);hP(n,b,u,d)}dP[e](t,l,a,h,u,d,f),l.eachEdge((function(e,n){var r,i=s["Q"](e.getModel().get(["lineStyle","curveness"]),nP(e,t,n),0),o=c(e.node1.getLayout()),a=c(e.node2.getLayout()),l=(o[0]+a[0])/2,h=(o[1]+a[1])/2;+i&&(i*=3,r=[u*i+l*(1-i),d*i+h*(1-i)]),e.setLayout([o,a,r])}))}}}var dP={value:function(t,e,n,r,i,o,a){var s=0,l=n.getSum("value"),u=2*Math.PI/(l||a);e.eachNode((function(t){var e=t.getValue("value"),n=u*(l?e:1)/2;s+=n,t.setLayout([r*Math.cos(s)+i,r*Math.sin(s)+o]),s+=n}))},symbolSize:function(t,e,n,r,i,o,a){var s=0;uP.length=a;var l=aP(t);e.eachNode((function(t){var e=sP(t);isNaN(e)&&(e=2),e<0&&(e=0),e*=l;var n=Math.asin(e/2/r);isNaN(n)&&(n=lP/2),uP[t.dataIndex]=n,s+=2*n}));var u=(2*lP-s)/a/2,c=0;e.eachNode((function(t){var e=u+uP[t.dataIndex];c+=e,(!t.getLayout()||!t.getLayout().fixed)&&t.setLayout([r*Math.cos(c)+i,r*Math.sin(c)+o]),c+=e}))}};function hP(t,e,n,r){var i=t.getGraphicEl();if(i){var o=t.getModel(),a=o.get(["label","rotate"])||0,l=i.getSymbolPath();if(e){var u=t.getLayout(),c=Math.atan2(u[1]-r,u[0]-n);c<0&&(c=2*Math.PI+c);var d=u[0]<n;d&&(c-=Math.PI);var h=d?"left":"right";l.setTextConfig({rotation:-c,position:h,origin:"center"});var f=l.ensureState("emphasis");s["m"](f.textConfig||(f.textConfig={}),{position:h})}else l.setTextConfig({rotation:a*=Math.PI/180})}}function fP(t){t.eachSeriesByType("graph",(function(t){"circular"===t.get("layout")&&cP(t,"symbolSize")}))}var pP=f;function mP(t,e,n){for(var r=t,i=e,o=n.rect,a=o.width,s=o.height,h=[o.x+a/2,o.y+s/2],f=null==n.gravity?.1:n.gravity,g=0;g<r.length;g++){var y=r[g];y.p||(y.p=l(a*(Math.random()-.5)+h[0],s*(Math.random()-.5)+h[1])),y.pp=c(y.p),y.edges=null}var _,b,x=null==n.friction?.6:n.friction,w=x;return{warmUp:function(){w=.8*x},setFixed:function(t){r[t].fixed=!0},setUnfixed:function(t){r[t].fixed=!1},beforeStep:function(t){_=t},afterStep:function(t){b=t},step:function(t){_&&_(r,i);for(var e=[],n=r.length,o=0;o<i.length;o++){var a=i[o];if(!a.ignoreForceLayout){var s=a.n1,l=a.n2;p(e,l.p,s.p);var c=m(e)-a.d,g=l.w/(s.w+l.w);isNaN(g)&&(g=0),v(e,e),!s.fixed&&pP(s.p,s.p,e,g*c*w),!l.fixed&&pP(l.p,l.p,e,-(1-g)*c*w)}}for(o=0;o<n;o++){var y=r[o];y.fixed||(p(e,h,y.p),pP(y.p,y.p,e,f*w))}for(o=0;o<n;o++){s=r[o];for(var x=o+1;x<n;x++){l=r[x];p(e,l.p,s.p);c=m(e);0===c&&(d(e,Math.random()-.5,Math.random()-.5),c=1);var M=(s.rep+l.rep)/c/c;!s.fixed&&pP(s.pp,s.pp,e,M),!l.fixed&&pP(l.pp,l.pp,e,-M)}}var S=[];for(o=0;o<n;o++){y=r[o];y.fixed||(p(S,y.p,y.pp),pP(y.p,y.p,S,w),u(y.pp,y.p))}w*=.992;var k=w<.01;b&&b(r,i,k),t&&t(k)}}}function gP(t){t.eachSeriesByType("graph",(function(t){var e=t.coordinateSystem;if(!e||"view"===e.type)if("force"===t.get("layout")){var n=t.preservedPoints||{},r=t.getGraph(),i=r.data,o=r.edgeData,a=t.getModel("force"),l=a.get("initLayout");t.preservedPoints?i.each((function(t){var e=i.getId(t);i.setItemLayout(t,n[e]||[NaN,NaN])})):l&&"none"!==l?"circular"===l&&cP(t,"value"):rP(t);var c=i.getDataExtent("value"),d=o.getDataExtent("value"),h=a.get("repulsion"),f=a.get("edgeLength"),p=s["t"](h)?h:[h,h],m=s["t"](f)?f:[f,f];m=[m[1],m[0]];var g=i.mapArray("value",(function(t,e){var n=i.getItemLayout(e),r=Er(t,c,p);return isNaN(r)&&(r=(p[0]+p[1])/2),{w:r,rep:r,fixed:i.getItemModel(e).get("fixed"),p:!n||isNaN(n[0])||isNaN(n[1])?null:n}})),y=o.mapArray("value",(function(e,n){var i=r.getEdgeByIndex(n),o=Er(e,d,m);isNaN(o)&&(o=(m[0]+m[1])/2);var a=i.getModel(),l=s["Q"](i.getModel().get(["lineStyle","curveness"]),-nP(i,t,n,!0),0);return{n1:g[i.node1.dataIndex],n2:g[i.node2.dataIndex],d:o,curveness:l,ignoreForceLayout:a.get("ignoreForceLayout")}})),v=e.getBoundingRect(),_=mP(g,y,{rect:v,gravity:a.get("gravity"),friction:a.get("friction")});_.beforeStep((function(t,e){for(var n=0,i=t.length;n<i;n++)t[n].fixed&&u(t[n].p,r.getNodeByIndex(n).getLayout())})),_.afterStep((function(t,e,o){for(var a=0,s=t.length;a<s;a++)t[a].fixed||r.getNodeByIndex(a).setLayout(t[a].p),n[i.getId(a)]=t[a].p;for(a=0,s=e.length;a<s;a++){var l=e[a],c=r.getEdgeByIndex(a),d=l.n1.p,h=l.n2.p,f=c.getLayout();f=f?f.slice():[],f[0]=f[0]||[],f[1]=f[1]||[],u(f[0],d),u(f[1],h),+l.curveness&&(f[2]=[(d[0]+h[0])/2-(d[1]-h[1])*l.curveness,(d[1]+h[1])/2-(h[0]-d[0])*l.curveness]),c.setLayout(f)}})),t.forceLayout=_,t.preservedPoints=n,_.step()}else t.forceLayout=null}))}function yP(t,e,n){var r=Object(s["m"])(t.getBoxLayoutParams(),{aspect:n});return Hc(r,{width:e.getWidth(),height:e.getHeight()})}function vP(t,e){var n=[];return t.eachSeriesByType("graph",(function(t){var r=t.get("coordinateSystem");if(!r||"view"===r){var i=t.getData(),o=i.mapArray((function(t){var e=i.getItemModel(t);return[+e.get("x"),+e.get("y")]})),a=[],s=[];$o(o,a,s),s[0]-a[0]===0&&(s[0]+=1,a[0]-=1),s[1]-a[1]===0&&(s[1]+=1,a[1]-=1);var l=(s[0]-a[0])/(s[1]-a[1]),u=yP(t,e,l);isNaN(l)&&(a=[u.x,u.y],s=[u.x+u.width,u.y+u.height]);var c=s[0]-a[0],d=s[1]-a[1],h=u.width,f=u.height,p=t.coordinateSystem=new ZC;p.zoomLimit=t.get("scaleLimit"),p.setBoundingRect(a[0],a[1],c,d),p.setViewRect(u.x,u.y,h,f),p.setCenter(t.get("center"),e),p.setZoom(t.get("zoom")),n.push(p)}})),n}var _P=jm.prototype,bP=Hm.prototype,xP=function(){function t(){this.x1=0,this.y1=0,this.x2=0,this.y2=0,this.percent=1}return t}();(function(t){function e(){return null!==t&&t.apply(this,arguments)||this}o(e,t)})(xP);function wP(t){return isNaN(+t.cpx1)||isNaN(+t.cpy1)}var MP=function(t){function e(e){var n=t.call(this,e)||this;return n.type="ec-line",n}return o(e,t),e.prototype.getDefaultStyle=function(){return{stroke:"#000",fill:null}},e.prototype.getDefaultShape=function(){return new xP},e.prototype.buildPath=function(t,e){wP(e)?_P.buildPath.call(this,t,e):bP.buildPath.call(this,t,e)},e.prototype.pointAt=function(t){return wP(this.shape)?_P.pointAt.call(this,t):bP.pointAt.call(this,t)},e.prototype.tangentAt=function(t){var e=this.shape,n=wP(e)?[e.x2-e.x1,e.y2-e.y1]:bP.tangentAt.call(this,t);return v(n,n)},e}(Va),SP=MP,kP=["fromSymbol","toSymbol"];function LP(t){return"_"+t+"Type"}function TP(t,e,n){var r=e.getItemVisual(n,t);if(!r||"none"===r)return r;var i=e.getItemVisual(n,t+"Size"),o=e.getItemVisual(n,t+"Rotate"),a=e.getItemVisual(n,t+"Offset"),s=e.getItemVisual(n,t+"KeepAspect"),l=av(i),u=sv(a||0,l);return r+l+u+(o||"")+(s||"")}function DP(t,e,n){var r=e.getItemVisual(n,t);if(r&&"none"!==r){var i=e.getItemVisual(n,t+"Size"),o=e.getItemVisual(n,t+"Rotate"),a=e.getItemVisual(n,t+"Offset"),s=e.getItemVisual(n,t+"KeepAspect"),l=av(i),u=sv(a||0,l),c=ov(r,-l[0]/2+u[0],-l[1]/2+u[1],l[0],l[1],null,s);return c.__specifiedRotation=null==o||isNaN(o)?void 0:+o*Math.PI/180||0,c.name=t,c}}function OP(t){var e=new SP({name:"line",subPixelOptimize:!0});return IP(e.shape,t),e}function IP(t,e){t.x1=e[0][0],t.y1=e[0][1],t.x2=e[1][0],t.y2=e[1][1],t.percent=1;var n=e[2];n?(t.cpx1=n[0],t.cpy1=n[1]):(t.cpx1=NaN,t.cpy1=NaN)}var CP=function(t){function e(e,n,r){var i=t.call(this)||this;return i._createLine(e,n,r),i}return o(e,t),e.prototype._createLine=function(t,e,n){var r=t.hostModel,i=t.getItemLayout(e),o=OP(i);o.shape.percent=0,Vl(o,{shape:{percent:1}},r,e),this.add(o),Object(s["k"])(kP,(function(n){var r=DP(n,t,e);this.add(r),this[LP(n)]=TP(n,t,e)}),this),this._updateCommonStl(t,e,n)},e.prototype.updateData=function(t,e,n){var r=t.hostModel,i=this.childOfName("line"),o=t.getItemLayout(e),a={shape:{}};IP(a.shape,o),Wl(i,a,r,e),Object(s["k"])(kP,(function(n){var r=TP(n,t,e),i=LP(n);if(this[i]!==r){this.remove(this.childOfName(n));var o=DP(n,t,e);this.add(o)}this[i]=r}),this),this._updateCommonStl(t,e,n)},e.prototype.getLinePath=function(){return this.childAt(0)},e.prototype._updateCommonStl=function(t,e,n){var r=t.hostModel,i=this.childOfName("line"),o=n&&n.emphasisLineStyle,a=n&&n.blurLineStyle,l=n&&n.selectLineStyle,u=n&&n.labelStatesModels,c=n&&n.emphasisDisabled,d=n&&n.focus,h=n&&n.blurScope;if(!n||t.hasItemOption){var f=t.getItemModel(e),p=f.getModel("emphasis");o=p.getModel("lineStyle").getLineStyle(),a=f.getModel(["blur","lineStyle"]).getLineStyle(),l=f.getModel(["select","lineStyle"]).getLineStyle(),c=p.get("disabled"),d=p.get("focus"),h=p.get("blurScope"),u=eu(f)}var m=t.getItemVisual(e,"style"),g=m.stroke;i.useStyle(m),i.style.fill=null,i.style.strokeNoScale=!0,i.ensureState("emphasis").style=o,i.ensureState("blur").style=a,i.ensureState("select").style=l,Object(s["k"])(kP,(function(t){var e=this.childOfName(t);if(e){e.setColor(g),e.style.opacity=m.opacity;for(var n=0;n<Ps.length;n++){var r=Ps[n],o=i.getState(r);if(o){var a=o.style||{},s=e.ensureState(r),l=s.style||(s.style={});null!=a.stroke&&(l[e.__isEmptyBrush?"stroke":"fill"]=a.stroke),null!=a.opacity&&(l.opacity=a.opacity)}}e.markRedraw()}}),this);var y=r.getRawValue(e);tu(this,u,{labelDataIndex:e,labelFetcher:{getFormattedLabel:function(e,n){return r.getFormattedLabel(e,n,t.dataType)}},inheritColor:g||"#000",defaultOpacity:m.opacity,defaultText:(null==y?t.getName(e):isFinite(y)?Rr(y):y)+""});var v=this.getTextContent();if(v){var _=u.normal;v.__align=v.style.align,v.__verticalAlign=v.style.verticalAlign,v.__position=_.get("position")||"middle";var b=_.get("distance");Object(s["t"])(b)||(b=[b,b]),v.__labelDistance=b}this.setTextConfig({position:null,local:!0,inside:!1}),Dl(this,d,h,c)},e.prototype.highlight=function(){ul(this)},e.prototype.downplay=function(){cl(this)},e.prototype.updateLayout=function(t,e){this.setLinePoints(t.getItemLayout(e))},e.prototype.setLinePoints=function(t){var e=this.childOfName("line");IP(e.shape,t),e.dirty()},e.prototype.beforeUpdate=function(){var t=this,e=t.childOfName("fromSymbol"),n=t.childOfName("toSymbol"),r=t.getTextContent();if(e||n||r&&!r.ignore){var i=1,o=this.parent;while(o)o.scaleX&&(i/=o.scaleX),o=o.parent;var a=t.childOfName("line");if(this.__dirty||a.__dirty){var s=a.shape.percent,l=a.pointAt(0),u=a.pointAt(s),c=p([],u,l);if(v(c,c),e&&(e.setPosition(l),k(e,0),e.scaleX=e.scaleY=i*s,e.markRedraw()),n&&(n.setPosition(u),k(n,1),n.scaleX=n.scaleY=i*s,n.markRedraw()),r&&!r.ignore){r.x=r.y=0,r.originX=r.originY=0;var d=void 0,h=void 0,f=r.__labelDistance,m=f[0]*i,g=f[1]*i,y=s/2,_=a.tangentAt(y),b=[_[1],-_[0]],x=a.pointAt(y);b[1]>0&&(b[0]=-b[0],b[1]=-b[1]);var w=_[0]<0?-1:1;if("start"!==r.__position&&"end"!==r.__position){var M=-Math.atan2(_[1],_[0]);u[0]<l[0]&&(M=Math.PI+M),r.rotation=M}var S=void 0;switch(r.__position){case"insideStartTop":case"insideMiddleTop":case"insideEndTop":case"middle":S=-g,h="bottom";break;case"insideStartBottom":case"insideMiddleBottom":case"insideEndBottom":S=g,h="top";break;default:S=0,h="middle"}switch(r.__position){case"end":r.x=c[0]*m+u[0],r.y=c[1]*g+u[1],d=c[0]>.8?"left":c[0]<-.8?"right":"center",h=c[1]>.8?"top":c[1]<-.8?"bottom":"middle";break;case"start":r.x=-c[0]*m+l[0],r.y=-c[1]*g+l[1],d=c[0]>.8?"right":c[0]<-.8?"left":"center",h=c[1]>.8?"bottom":c[1]<-.8?"top":"middle";break;case"insideStartTop":case"insideStart":case"insideStartBottom":r.x=m*w+l[0],r.y=l[1]+S,d=_[0]<0?"right":"left",r.originX=-m*w,r.originY=-S;break;case"insideMiddleTop":case"insideMiddle":case"insideMiddleBottom":case"middle":r.x=x[0],r.y=x[1]+S,d="center",r.originY=-S;break;case"insideEndTop":case"insideEnd":case"insideEndBottom":r.x=-m*w+u[0],r.y=u[1]+S,d=_[0]>=0?"right":"left",r.originX=m*w,r.originY=-S;break}r.scaleX=r.scaleY=i,r.setStyle({verticalAlign:r.__verticalAlign||h,align:r.__align||d})}}}function k(t,e){var n=t.__specifiedRotation;if(null==n){var r=a.tangentAt(e);t.attr("rotation",(1===e?-1:1)*Math.PI/2-Math.atan2(r[1],r[0]))}else t.attr("rotation",n)}},e}(wr),AP=CP,YP=function(){function t(t){this.group=new wr,this._LineCtor=t||AP}return t.prototype.updateData=function(t){var e=this;this._progressiveEls=null;var n=this,r=n.group,i=n._lineData;n._lineData=t,i||r.removeAll();var o=PP(t);t.diff(i).add((function(n){e._doAdd(t,n,o)})).update((function(n,r){e._doUpdate(i,t,r,n,o)})).remove((function(t){r.remove(i.getItemGraphicEl(t))})).execute()},t.prototype.updateLayout=function(){var t=this._lineData;t&&t.eachItemGraphicEl((function(e,n){e.updateLayout(t,n)}),this)},t.prototype.incrementalPrepareUpdate=function(t){this._seriesScope=PP(t),this._lineData=null,this.group.removeAll()},t.prototype.incrementalUpdate=function(t,e){function n(t){t.isGroup||jP(t)||(t.incremental=!0,t.ensureState("emphasis").hoverLayer=!0)}this._progressiveEls=[];for(var r=t.start;r<t.end;r++){var i=e.getItemLayout(r);if(NP(i)){var o=new this._LineCtor(e,r,this._seriesScope);o.traverse(n),this.group.add(o),e.setItemGraphicEl(r,o),this._progressiveEls.push(o)}}},t.prototype.remove=function(){this.group.removeAll()},t.prototype.eachRendered=function(t){Ng(this._progressiveEls||this.group,t)},t.prototype._doAdd=function(t,e,n){var r=t.getItemLayout(e);if(NP(r)){var i=new this._LineCtor(t,e,n);t.setItemGraphicEl(e,i),this.group.add(i)}},t.prototype._doUpdate=function(t,e,n,r,i){var o=t.getItemGraphicEl(n);NP(e.getItemLayout(r))?(o?o.updateData(e,r,i):o=new this._LineCtor(e,r,i),e.setItemGraphicEl(r,o),this.group.add(o)):this.group.remove(o)},t}();function jP(t){return t.animators&&t.animators.length>0}function PP(t){var e=t.hostModel,n=e.getModel("emphasis");return{lineStyle:e.getModel("lineStyle").getLineStyle(),emphasisLineStyle:n.getModel(["lineStyle"]).getLineStyle(),blurLineStyle:e.getModel(["blur","lineStyle"]).getLineStyle(),selectLineStyle:e.getModel(["select","lineStyle"]).getLineStyle(),emphasisDisabled:n.get("disabled"),blurScope:n.get("blurScope"),focus:n.get("focus"),labelStatesModels:eu(e)}}function EP(t){return isNaN(t[0])||isNaN(t[1])}function NP(t){return t&&!EP(t[0])&&!EP(t[1])}var RP=YP,HP=[],zP=[],FP=[],BP=Le,WP=w,VP=Math.abs;function GP(t,e,n){for(var r,i=t[0],o=t[1],a=t[2],s=1/0,l=n*n,u=.1,c=.1;c<=.9;c+=.1){HP[0]=BP(i[0],o[0],a[0],c),HP[1]=BP(i[1],o[1],a[1],c);var d=VP(WP(HP,e)-l);d<s&&(s=d,r=c)}for(var h=0;h<32;h++){var f=r+u;zP[0]=BP(i[0],o[0],a[0],r),zP[1]=BP(i[1],o[1],a[1],r),FP[0]=BP(i[0],o[0],a[0],f),FP[1]=BP(i[1],o[1],a[1],f);d=WP(zP,e)-l;if(VP(d)<.01)break;var p=WP(FP,e)-l;u/=2,d<0?p>=0?r+=u:r-=u:p>=0?r-=u:r+=u}return r}function $P(t,e){var n=[],r=Ie,i=[[],[],[]],o=[[],[]],a=[];e/=2,t.eachEdge((function(t,s){var l=t.getLayout(),d=t.getVisual("fromSymbol"),h=t.getVisual("toSymbol");l.__original||(l.__original=[c(l[0]),c(l[1])],l[2]&&l.__original.push(c(l[2])));var m=l.__original;if(null!=l[2]){if(u(i[0],m[0]),u(i[1],m[2]),u(i[2],m[1]),d&&"none"!==d){var g=sP(t.node1),y=GP(i,m[0],g*e);r(i[0][0],i[1][0],i[2][0],y,n),i[0][0]=n[3],i[1][0]=n[4],r(i[0][1],i[1][1],i[2][1],y,n),i[0][1]=n[3],i[1][1]=n[4]}if(h&&"none"!==h){g=sP(t.node2),y=GP(i,m[1],g*e);r(i[0][0],i[1][0],i[2][0],y,n),i[1][0]=n[1],i[2][0]=n[2],r(i[0][1],i[1][1],i[2][1],y,n),i[1][1]=n[1],i[2][1]=n[2]}u(l[0],i[0]),u(l[1],i[2]),u(l[2],i[1])}else{if(u(o[0],m[0]),u(o[1],m[1]),p(a,o[1],o[0]),v(a,a),d&&"none"!==d){g=sP(t.node1);f(o[0],o[0],a,g*e)}if(h&&"none"!==h){g=sP(t.node2);f(o[1],o[1],a,-g*e)}u(l[0],o[0]),u(l[1],o[1])}}))}function UP(t){return"view"===t.type}var qP=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(t,e){var n=new MS,r=new RP,i=this.group;this._controller=new dI(e.getZr()),this._controllerHost={target:i},i.add(n.group),i.add(r.group),this._symbolDraw=n,this._lineDraw=r,this._firstRender=!0},e.prototype.render=function(t,e,n){var r=this,i=t.coordinateSystem;this._model=t;var o=this._symbolDraw,a=this._lineDraw,s=this.group;if(UP(i)){var l={x:i.x,y:i.y,scaleX:i.scaleX,scaleY:i.scaleY};this._firstRender?s.attr(l):Wl(s,l,t)}$P(t.getGraph(),aP(t));var u=t.getData();o.updateData(u);var c=t.getEdgeData();a.updateData(c),this._updateNodeAndLinkScale(),this._updateController(t,e,n),clearTimeout(this._layoutTimeout);var d=t.forceLayout,h=t.get(["force","layoutAnimation"]);d&&this._startForceLayoutIteration(d,h);var f=t.get("layout");u.graph.eachNode((function(e){var n=e.dataIndex,i=e.getGraphicEl(),o=e.getModel();if(i){i.off("drag").off("dragend");var a=o.get("draggable");a&&i.on("drag",(function(o){switch(f){case"force":d.warmUp(),!r._layouting&&r._startForceLayoutIteration(d,h),d.setFixed(n),u.setItemLayout(n,[i.x,i.y]);break;case"circular":u.setItemLayout(n,[i.x,i.y]),e.setLayout({fixed:!0},!0),cP(t,"symbolSize",e,[o.offsetX,o.offsetY]),r.updateLayout(t);break;case"none":default:u.setItemLayout(n,[i.x,i.y]),iP(t.getGraph(),t),r.updateLayout(t);break}})).on("dragend",(function(){d&&d.setUnfixed(n)})),i.setDraggable(a,!!o.get("cursor"));var s=o.get(["emphasis","focus"]);"adjacency"===s&&(Ls(i).focus=e.getAdjacentDataIndices())}})),u.graph.eachEdge((function(t){var e=t.getGraphicEl(),n=t.getModel().get(["emphasis","focus"]);e&&"adjacency"===n&&(Ls(e).focus={edge:[t.dataIndex],node:[t.node1.dataIndex,t.node2.dataIndex]})}));var p="circular"===t.get("layout")&&t.get(["circular","rotateLabel"]),m=u.getLayout("cx"),g=u.getLayout("cy");u.graph.eachNode((function(t){hP(t,p,m,g)})),this._firstRender=!1},e.prototype.dispose=function(){this.remove(),this._controller&&this._controller.dispose(),this._controllerHost=null},e.prototype._startForceLayoutIteration=function(t,e){var n=this;(function r(){t.step((function(t){n.updateLayout(n._model),(n._layouting=!t)&&(e?n._layoutTimeout=setTimeout(r,16):r())}))})()},e.prototype._updateController=function(t,e,n){var r=this,i=this._controller,o=this._controllerHost,a=this.group;i.setPointerChecker((function(e,r,i){var o=a.getBoundingRect();return o.applyTransform(a.transform),o.contain(r,i)&&!gI(e,n,t)})),UP(t.coordinateSystem)?(i.enable(t.get("roam")),o.zoomLimit=t.get("scaleLimit"),o.zoom=t.coordinateSystem.getZoom(),i.off("pan").off("zoom").on("pan",(function(e){hI(o,e.dx,e.dy),n.dispatchAction({seriesId:t.id,type:"graphRoam",dx:e.dx,dy:e.dy})})).on("zoom",(function(e){fI(o,e.scale,e.originX,e.originY),n.dispatchAction({seriesId:t.id,type:"graphRoam",zoom:e.scale,originX:e.originX,originY:e.originY}),r._updateNodeAndLinkScale(),$P(t.getGraph(),aP(t)),r._lineDraw.updateLayout(),n.updateLabelLayout()}))):i.disable()},e.prototype._updateNodeAndLinkScale=function(){var t=this._model,e=t.getData(),n=aP(t);e.eachItemGraphicEl((function(t,e){t&&t.setSymbolScale(n)}))},e.prototype.updateLayout=function(t){$P(t.getGraph(),aP(t)),this._symbolDraw.updateLayout(),this._lineDraw.updateLayout()},e.prototype.remove=function(){clearTimeout(this._layoutTimeout),this._layouting=!1,this._layoutTimeout=null,this._symbolDraw&&this._symbolDraw.remove(),this._lineDraw&&this._lineDraw.remove()},e.type="graph",e}($g),XP=qP;function ZP(t){return"_EC_"+t}var JP=function(){function t(t){this.type="graph",this.nodes=[],this.edges=[],this._nodesMap={},this._edgesMap={},this._directed=t||!1}return t.prototype.isDirected=function(){return this._directed},t.prototype.addNode=function(t,e){t=null==t?""+e:""+t;var n=this._nodesMap;if(!n[ZP(t)]){var r=new KP(t,e);return r.hostGraph=this,this.nodes.push(r),n[ZP(t)]=r,r}},t.prototype.getNodeByIndex=function(t){var e=this.data.getRawIndex(t);return this.nodes[e]},t.prototype.getNodeById=function(t){return this._nodesMap[ZP(t)]},t.prototype.addEdge=function(t,e,n){var r=this._nodesMap,i=this._edgesMap;if(s["z"](t)&&(t=this.nodes[t]),s["z"](e)&&(e=this.nodes[e]),t instanceof KP||(t=r[ZP(t)]),e instanceof KP||(e=r[ZP(e)]),t&&e){var o=t.id+"-"+e.id,a=new QP(t,e,n);return a.hostGraph=this,this._directed&&(t.outEdges.push(a),e.inEdges.push(a)),t.edges.push(a),t!==e&&e.edges.push(a),this.edges.push(a),i[o]=a,a}},t.prototype.getEdgeByIndex=function(t){var e=this.edgeData.getRawIndex(t);return this.edges[e]},t.prototype.getEdge=function(t,e){t instanceof KP&&(t=t.id),e instanceof KP&&(e=e.id);var n=this._edgesMap;return this._directed?n[t+"-"+e]:n[t+"-"+e]||n[e+"-"+t]},t.prototype.eachNode=function(t,e){for(var n=this.nodes,r=n.length,i=0;i<r;i++)n[i].dataIndex>=0&&t.call(e,n[i],i)},t.prototype.eachEdge=function(t,e){for(var n=this.edges,r=n.length,i=0;i<r;i++)n[i].dataIndex>=0&&n[i].node1.dataIndex>=0&&n[i].node2.dataIndex>=0&&t.call(e,n[i],i)},t.prototype.breadthFirstTraverse=function(t,e,n,r){if(e instanceof KP||(e=this._nodesMap[ZP(e)]),e){for(var i="out"===n?"outEdges":"in"===n?"inEdges":"edges",o=0;o<this.nodes.length;o++)this.nodes[o].__visited=!1;if(!t.call(r,e,null)){var a=[e];while(a.length){var s=a.shift(),l=s[i];for(o=0;o<l.length;o++){var u=l[o],c=u.node1===s?u.node2:u.node1;if(!c.__visited){if(t.call(r,c,s))return;a.push(c),c.__visited=!0}}}}}},t.prototype.update=function(){for(var t=this.data,e=this.edgeData,n=this.nodes,r=this.edges,i=0,o=n.length;i<o;i++)n[i].dataIndex=-1;for(i=0,o=t.count();i<o;i++)n[t.getRawIndex(i)].dataIndex=i;e.filterSelf((function(t){var n=r[e.getRawIndex(t)];return n.node1.dataIndex>=0&&n.node2.dataIndex>=0}));for(i=0,o=r.length;i<o;i++)r[i].dataIndex=-1;for(i=0,o=e.count();i<o;i++)r[e.getRawIndex(i)].dataIndex=i},t.prototype.clone=function(){for(var e=new t(this._directed),n=this.nodes,r=this.edges,i=0;i<n.length;i++)e.addNode(n[i].id,n[i].dataIndex);for(i=0;i<r.length;i++){var o=r[i];e.addEdge(o.node1.id,o.node2.id,o.dataIndex)}return e},t}(),KP=function(){function t(t,e){this.inEdges=[],this.outEdges=[],this.edges=[],this.dataIndex=-1,this.id=null==t?"":t,this.dataIndex=null==e?-1:e}return t.prototype.degree=function(){return this.edges.length},t.prototype.inDegree=function(){return this.inEdges.length},t.prototype.outDegree=function(){return this.outEdges.length},t.prototype.getModel=function(t){if(!(this.dataIndex<0)){var e=this.hostGraph,n=e.data.getItemModel(this.dataIndex);return n.getModel(t)}},t.prototype.getAdjacentDataIndices=function(){for(var t={edge:[],node:[]},e=0;e<this.edges.length;e++){var n=this.edges[e];n.dataIndex<0||(t.edge.push(n.dataIndex),t.node.push(n.node1.dataIndex,n.node2.dataIndex))}return t},t.prototype.getTrajectoryDataIndices=function(){for(var t=s["f"](),e=s["f"](),n=0;n<this.edges.length;n++){var r=this.edges[n];if(!(r.dataIndex<0)){t.set(r.dataIndex,!0);var i=[r.node1],o=[r.node2],a=0;while(a<i.length){var l=i[a];a++,e.set(l.dataIndex,!0);for(var u=0;u<l.inEdges.length;u++)t.set(l.inEdges[u].dataIndex,!0),i.push(l.inEdges[u].node1)}a=0;while(a<o.length){var c=o[a];a++,e.set(c.dataIndex,!0);for(u=0;u<c.outEdges.length;u++)t.set(c.outEdges[u].dataIndex,!0),o.push(c.outEdges[u].node2)}}}return{edge:t.keys(),node:e.keys()}},t}(),QP=function(){function t(t,e,n){this.dataIndex=-1,this.node1=t,this.node2=e,this.dataIndex=null==n?-1:n}return t.prototype.getModel=function(t){if(!(this.dataIndex<0)){var e=this.hostGraph,n=e.edgeData.getItemModel(this.dataIndex);return n.getModel(t)}},t.prototype.getAdjacentDataIndices=function(){return{edge:[this.dataIndex],node:[this.node1.dataIndex,this.node2.dataIndex]}},t.prototype.getTrajectoryDataIndices=function(){var t=s["f"](),e=s["f"]();t.set(this.dataIndex,!0);var n=[this.node1],r=[this.node2],i=0;while(i<n.length){var o=n[i];i++,e.set(o.dataIndex,!0);for(var a=0;a<o.inEdges.length;a++)t.set(o.inEdges[a].dataIndex,!0),n.push(o.inEdges[a].node1)}i=0;while(i<r.length){var l=r[i];i++,e.set(l.dataIndex,!0);for(a=0;a<l.outEdges.length;a++)t.set(l.outEdges[a].dataIndex,!0),r.push(l.outEdges[a].node2)}return{edge:t.keys(),node:e.keys()}},t}();function tE(t,e){return{getValue:function(n){var r=this[t][e];return r.getStore().get(r.getDimensionIndex(n||"value"),this.dataIndex)},setVisual:function(n,r){this.dataIndex>=0&&this[t][e].setItemVisual(this.dataIndex,n,r)},getVisual:function(n){return this[t][e].getItemVisual(this.dataIndex,n)},setLayout:function(n,r){this.dataIndex>=0&&this[t][e].setItemLayout(this.dataIndex,n,r)},getLayout:function(){return this[t][e].getItemLayout(this.dataIndex)},getGraphicEl:function(){return this[t][e].getItemGraphicEl(this.dataIndex)},getRawIndex:function(){return this[t][e].getRawIndex(this.dataIndex)}}}s["K"](KP,tE("hostGraph","data")),s["K"](QP,tE("hostGraph","edgeData"));var eE=JP;function nE(t,e,n,r,i){for(var o=new eE(r),a=0;a<t.length;a++)o.addNode(s["O"](t[a].id,t[a].name,a),a);var l=[],u=[],c=0;for(a=0;a<e.length;a++){var d=e[a],h=d.source,f=d.target;o.addEdge(h,f,c)&&(u.push(d),l.push(s["O"](Mi(d.id,null),h+" > "+f)),c++)}var p,m=n.get("coordinateSystem");if("cartesian2d"===m||"polar"===m)p=dS(t,n);else{var g=$d.get(m),y=g&&g.dimensions||[];s["r"](y,"value")<0&&y.concat(["value"]);var v=qM(t,{coordDimensions:y,encodeDefine:n.getEncode()}).dimensions;p=new UM(v,n),p.initData(t)}var _=new UM(["value"],n);return _.initData(u,l),i&&i(p,_),ZA({mainData:p,struct:o,structAttr:"graph",datas:{node:p,edge:_},datasAttr:{node:"data",edge:"edgeData"}}),o.update(),o}var rE=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.hasSymbolVisual=!0,n}return o(e,t),e.prototype.init=function(e){t.prototype.init.apply(this,arguments);var n=this;function r(){return n._categoriesData}this.legendVisualProvider=new OL(r,r),this.fillDataTextStyle(e.edges||e.links),this._updateCategoriesData()},e.prototype.mergeOption=function(e){t.prototype.mergeOption.apply(this,arguments),this.fillDataTextStyle(e.edges||e.links),this._updateCategoriesData()},e.prototype.mergeDefaultAndTheme=function(e){t.prototype.mergeDefaultAndTheme.apply(this,arguments),ci(e,"edgeLabel",["show"])},e.prototype.getInitialData=function(t,e){var n=t.edges||t.links||[],r=t.data||t.nodes||[],i=this;if(r&&n){tP(this);var o=nE(r,n,this,!0,a);return s["k"](o.edges,(function(t){eP(t.node1,t.node2,this,t.dataIndex)}),this),o.data}function a(t,e){t.wrapMethod("getItemModel",(function(t){var e=i._categoriesModels,n=t.getShallow("category"),r=e[n];return r&&(r.parentModel=t.parentModel,t.parentModel=r),t}));var n=Lu.prototype.getModel;function r(t,e){var r=n.call(this,t,e);return r.resolveParentPath=o,r}function o(t){if(t&&("label"===t[0]||"label"===t[1])){var e=t.slice();return"label"===t[0]?e[0]="edgeLabel":"label"===t[1]&&(e[1]="edgeLabel"),e}return t}e.wrapMethod("getItemModel",(function(t){return t.resolveParentPath=o,t.getModel=r,t}))}},e.prototype.getGraph=function(){return this.getData().graph},e.prototype.getEdgeData=function(){return this.getGraph().edgeData},e.prototype.getCategoriesData=function(){return this._categoriesData},e.prototype.formatTooltip=function(t,e,n){if("edge"===n){var r=this.getData(),i=this.getDataParams(t,n),o=r.graph.getEdgeByIndex(t),a=r.getName(o.node1.dataIndex),s=r.getName(o.node2.dataIndex),l=[];return null!=a&&l.push(a),null!=s&&l.push(s),Uf("nameValue",{name:l.join(" > "),value:i.value,noValue:null==i.value})}var u=up({series:this,dataIndex:t,multipleSeries:e});return u},e.prototype._updateCategoriesData=function(){var t=s["H"](this.option.categories||[],(function(t){return null!=t.value?t:s["m"]({value:0},t)})),e=new UM(["value"],this);e.initData(t),this._categoriesData=e,this._categoriesModels=e.mapArray((function(t){return e.getItemModel(t)}))},e.prototype.setZoom=function(t){this.option.zoom=t},e.prototype.setCenter=function(t){this.option.center=t},e.prototype.isAnimationEnabled=function(){return t.prototype.isAnimationEnabled.call(this)&&!("force"===this.get("layout")&&this.get(["force","layoutAnimation"]))},e.type="series.graph",e.dependencies=["grid","polar","geo","singleAxis","calendar"],e.defaultOption={z:2,coordinateSystem:"view",legendHoverLink:!0,layout:null,circular:{rotateLabel:!1},force:{initLayout:null,repulsion:[0,50],gravity:.1,friction:.6,edgeLength:30,layoutAnimation:!0},left:"center",top:"center",symbol:"circle",symbolSize:10,edgeSymbol:["none","none"],edgeSymbolSize:10,edgeLabel:{position:"middle",distance:5},draggable:!1,roam:!1,center:null,zoom:1,nodeScaleRatio:.6,label:{show:!1,formatter:"{b}"},itemStyle:{},lineStyle:{color:"#aaa",width:1,opacity:.5},emphasis:{scale:!0,label:{show:!0}},select:{itemStyle:{borderColor:"#212121"}}},e}(Mp),iE=rE,oE={type:"graphRoam",event:"graphRoam",update:"none"};function aE(t){t.registerChartView(XP),t.registerSeriesModel(iE),t.registerProcessor(Bj),t.registerVisual(Wj),t.registerVisual(Gj),t.registerLayout(oP),t.registerLayout(t.PRIORITY.VISUAL.POST_CHART_LAYOUT,fP),t.registerLayout(gP),t.registerCoordinateSystem("graphView",{dimensions:ZC.dimensions,create:vP}),t.registerAction({type:"focusNodeAdjacency",event:"focusNodeAdjacency",update:"series:focusNodeAdjacency"},s["L"]),t.registerAction({type:"unfocusNodeAdjacency",event:"unfocusNodeAdjacency",update:"series:unfocusNodeAdjacency"},s["L"]),t.registerAction(oE,(function(t,e,n){e.eachComponent({mainType:"series",query:t},(function(e){var r=e.coordinateSystem,i=cA(r,t,void 0,n);e.setCenter&&e.setCenter(i.center),e.setZoom&&e.setZoom(i.zoom)}))}))}var sE=function(){function t(){this.angle=0,this.width=10,this.r=10,this.x=0,this.y=0}return t}(),lE=function(t){function e(e){var n=t.call(this,e)||this;return n.type="pointer",n}return o(e,t),e.prototype.getDefaultShape=function(){return new sE},e.prototype.buildPath=function(t,e){var n=Math.cos,r=Math.sin,i=e.r,o=e.width,a=e.angle,s=e.x-n(a)*o*(o>=i/3?1:2),l=e.y-r(a)*o*(o>=i/3?1:2);a=e.angle-Math.PI/2,t.moveTo(s,l),t.lineTo(e.x+n(a)*o,e.y+r(a)*o),t.lineTo(e.x+n(e.angle)*i,e.y+r(e.angle)*i),t.lineTo(e.x-n(a)*o,e.y-r(a)*o),t.lineTo(s,l)},e}(Va),uE=lE;function cE(t,e){var n=t.get("center"),r=e.getWidth(),i=e.getHeight(),o=Math.min(r,i),a=Nr(n[0],e.getWidth()),s=Nr(n[1],e.getHeight()),l=Nr(t.get("radius"),o/2);return{cx:a,cy:s,r:l}}function dE(t,e){var n=null==t?"":t+"";return e&&(Object(s["C"])(e)?n=e.replace("{value}",n):Object(s["w"])(e)&&(n=e(t))),n}var hE=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){this.group.removeAll();var r=t.get(["axisLine","lineStyle","color"]),i=cE(t,n);this._renderMain(t,e,n,r,i),this._data=t.getData()},e.prototype.dispose=function(){},e.prototype._renderMain=function(t,e,n,r,i){var o=this.group,a=t.get("clockwise"),l=-t.get("startAngle")/180*Math.PI,u=-t.get("endAngle")/180*Math.PI,c=t.getModel("axisLine"),d=c.get("roundCap"),h=d?jk:_m,f=c.get("show"),p=c.getModel("lineStyle"),m=p.get("width"),g=[l,u];ga(g,!a),l=g[0],u=g[1];for(var y=u-l,v=l,_=[],b=0;f&&b<r.length;b++){var x=Math.min(Math.max(r[b][0],0),1);u=l+y*x;var w=new h({shape:{startAngle:v,endAngle:u,cx:i.cx,cy:i.cy,clockwise:a,r0:i.r-m,r:i.r},silent:!0});w.setStyle({fill:r[b][1]}),w.setStyle(p.getLineStyle(["color","width"])),_.push(w),v=u}_.reverse(),Object(s["k"])(_,(function(t){return o.add(t)}));var M=function(t){if(t<=0)return r[0][1];var e;for(e=0;e<r.length;e++)if(r[e][0]>=t&&(0===e?0:r[e-1][0])<t)return r[e][1];return r[e-1][1]};this._renderTicks(t,e,n,M,i,l,u,a,m),this._renderTitleAndDetail(t,e,n,M,i),this._renderAnchor(t,i),this._renderPointer(t,e,n,M,i,l,u,a,m)},e.prototype._renderTicks=function(t,e,n,r,i,o,a,l,u){for(var c,d,h=this.group,f=i.cx,p=i.cy,m=i.r,g=+t.get("min"),y=+t.get("max"),v=t.getModel("splitLine"),_=t.getModel("axisTick"),b=t.getModel("axisLabel"),x=t.get("splitNumber"),w=_.get("splitNumber"),M=Nr(v.get("length"),m),S=Nr(_.get("length"),m),k=o,L=(a-o)/x,T=L/w,D=v.getModel("lineStyle").getLineStyle(),O=_.getModel("lineStyle").getLineStyle(),I=v.get("distance"),C=0;C<=x;C++){if(c=Math.cos(k),d=Math.sin(k),v.get("show")){var A=I?I+u:u,Y=new jm({shape:{x1:c*(m-A)+f,y1:d*(m-A)+p,x2:c*(m-M-A)+f,y2:d*(m-M-A)+p},style:D,silent:!0});"auto"===D.stroke&&Y.setStyle({stroke:r(C/x)}),h.add(Y)}if(b.get("show")){A=b.get("distance")+I;var j=dE(Rr(C/x*(y-g)+g),b.get("formatter")),P=r(C/x),E=c*(m-M-A)+f,N=d*(m-M-A)+p,R=b.get("rotate"),H=0;"radial"===R?(H=-k+2*Math.PI,H>Math.PI/2&&(H+=Math.PI)):"tangential"===R?H=-k-Math.PI/2:Object(s["z"])(R)&&(H=R*Math.PI/180),0===H?h.add(new ks({style:nu(b,{text:j,x:E,y:N,verticalAlign:d<-.8?"top":d>.8?"bottom":"middle",align:c<-.4?"left":c>.4?"right":"center"},{inheritColor:P}),silent:!0})):h.add(new ks({style:nu(b,{text:j,x:E,y:N,verticalAlign:"middle",align:"center"},{inheritColor:P}),silent:!0,originX:E,originY:N,rotation:H}))}if(_.get("show")&&C!==x){A=_.get("distance");A=A?A+u:u;for(var z=0;z<=w;z++){c=Math.cos(k),d=Math.sin(k);var F=new jm({shape:{x1:c*(m-A)+f,y1:d*(m-A)+p,x2:c*(m-S-A)+f,y2:d*(m-S-A)+p},silent:!0,style:O});"auto"===O.stroke&&F.setStyle({stroke:r((C+z/w)/x)}),h.add(F),k+=T}k-=T}else k+=L}},e.prototype._renderPointer=function(t,e,n,r,i,o,a,l,u){var c=this.group,d=this._data,h=this._progressEls,f=[],p=t.get(["pointer","show"]),m=t.getModel("progress"),g=m.get("show"),y=t.getData(),v=y.mapDimension("value"),_=+t.get("min"),b=+t.get("max"),x=[_,b],w=[o,a];function M(e,n){var r,o=y.getItemModel(e),a=o.getModel("pointer"),s=Nr(a.get("width"),i.r),l=Nr(a.get("length"),i.r),u=t.get(["pointer","icon"]),c=a.get("offsetCenter"),d=Nr(c[0],i.r),h=Nr(c[1],i.r),f=a.get("keepAspect");return r=u?ov(u,d-s/2,h-l,s,l,null,f):new uE({shape:{angle:-Math.PI/2,width:s,r:l,x:d,y:h}}),r.rotation=-(n+Math.PI/2),r.x=i.cx,r.y=i.cy,r}function S(t,e){var n=m.get("roundCap"),r=n?jk:_m,a=m.get("overlap"),s=a?m.get("width"):u/y.count(),c=a?i.r-s:i.r-(t+1)*s,d=a?i.r:i.r-t*s,h=new r({shape:{startAngle:o,endAngle:e,cx:i.cx,cy:i.cy,clockwise:l,r0:c,r:d}});return a&&(h.z2=b-y.get(v,t)%b),h}(g||p)&&(y.diff(d).add((function(e){var n=y.get(v,e);if(p){var r=M(e,o);Vl(r,{rotation:-((isNaN(+n)?w[0]:Er(n,x,w,!0))+Math.PI/2)},t),c.add(r),y.setItemGraphicEl(e,r)}if(g){var i=S(e,o),a=m.get("clip");Vl(i,{shape:{endAngle:Er(n,x,w,a)}},t),c.add(i),Ts(t.seriesIndex,y.dataType,e,i),f[e]=i}})).update((function(e,n){var r=y.get(v,e);if(p){var i=d.getItemGraphicEl(n),a=i?i.rotation:o,s=M(e,a);s.rotation=a,Wl(s,{rotation:-((isNaN(+r)?w[0]:Er(r,x,w,!0))+Math.PI/2)},t),c.add(s),y.setItemGraphicEl(e,s)}if(g){var l=h[n],u=l?l.shape.endAngle:o,_=S(e,u),b=m.get("clip");Wl(_,{shape:{endAngle:Er(r,x,w,b)}},t),c.add(_),Ts(t.seriesIndex,y.dataType,e,_),f[e]=_}})).execute(),y.each((function(t){var e=y.getItemModel(t),n=e.getModel("emphasis"),i=n.get("focus"),o=n.get("blurScope"),a=n.get("disabled");if(p){var l=y.getItemGraphicEl(t),u=y.getItemVisual(t,"style"),c=u.fill;if(l instanceof Ka){var d=l.style;l.useStyle(Object(s["m"])({image:d.image,x:d.x,y:d.y,width:d.width,height:d.height},u))}else l.useStyle(u),"pointer"!==l.type&&l.setColor(c);l.setStyle(e.getModel(["pointer","itemStyle"]).getItemStyle()),"auto"===l.style.fill&&l.setStyle("fill",r(Er(y.get(v,t),x,[0,1],!0))),l.z2EmphasisLift=0,Al(l,e),Dl(l,i,o,a)}if(g){var h=f[t];h.useStyle(y.getItemVisual(t,"style")),h.setStyle(e.getModel(["progress","itemStyle"]).getItemStyle()),h.z2EmphasisLift=0,Al(h,e),Dl(h,i,o,a)}})),this._progressEls=f)},e.prototype._renderAnchor=function(t,e){var n=t.getModel("anchor"),r=n.get("show");if(r){var i=n.get("size"),o=n.get("icon"),a=n.get("offsetCenter"),s=n.get("keepAspect"),l=ov(o,e.cx-i/2+Nr(a[0],e.r),e.cy-i/2+Nr(a[1],e.r),i,i,null,s);l.z2=n.get("showAbove")?1:0,l.setStyle(n.getModel("itemStyle").getItemStyle()),this.group.add(l)}},e.prototype._renderTitleAndDetail=function(t,e,n,r,i){var o=this,a=t.getData(),s=a.mapDimension("value"),l=+t.get("min"),u=+t.get("max"),c=new wr,d=[],h=[],f=t.isAnimationEnabled(),p=t.get(["pointer","showAbove"]);a.diff(this._data).add((function(t){d[t]=new ks({silent:!0}),h[t]=new ks({silent:!0})})).update((function(t,e){d[t]=o._titleEls[e],h[t]=o._detailEls[e]})).execute(),a.each((function(e){var n=a.getItemModel(e),o=a.get(s,e),m=new wr,g=r(Er(o,[l,u],[0,1],!0)),y=n.getModel("title");if(y.get("show")){var v=y.get("offsetCenter"),_=i.cx+Nr(v[0],i.r),b=i.cy+Nr(v[1],i.r),x=d[e];x.attr({z2:p?0:2,style:nu(y,{x:_,y:b,text:a.getName(e),align:"center",verticalAlign:"middle"},{inheritColor:g})}),m.add(x)}var w=n.getModel("detail");if(w.get("show")){var M=w.get("offsetCenter"),S=i.cx+Nr(M[0],i.r),k=i.cy+Nr(M[1],i.r),L=Nr(w.get("width"),i.r),T=Nr(w.get("height"),i.r),D=t.get(["progress","show"])?a.getItemVisual(e,"style").fill:g,O=(x=h[e],w.get("formatter"));x.attr({z2:p?0:2,style:nu(w,{x:S,y:k,text:dE(o,O),width:isNaN(L)?null:L,height:isNaN(T)?null:T,align:"center",verticalAlign:"middle"},{inheritColor:D})}),hu(x,{normal:w},o,(function(t){return dE(t,O)})),f&&fu(x,e,a,t,{getFormattedLabel:function(t,e,n,r,i,a){return dE(a?a.interpolatedValue:o,O)}}),m.add(x)}c.add(m)})),this.group.add(c),this._titleEls=d,this._detailEls=h},e.type="gauge",e}($g),fE=hE,pE=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.visualStyleAccessPath="itemStyle",n}return o(e,t),e.prototype.getInitialData=function(t,e){return TL(this,["value"])},e.type="series.gauge",e.defaultOption={z:2,colorBy:"data",center:["50%","50%"],legendHoverLink:!0,radius:"75%",startAngle:225,endAngle:-45,clockwise:!0,min:0,max:100,splitNumber:10,axisLine:{show:!0,roundCap:!1,lineStyle:{color:[[1,"#E6EBF8"]],width:10}},progress:{show:!1,overlap:!0,width:10,roundCap:!1,clip:!0},splitLine:{show:!0,length:10,distance:10,lineStyle:{color:"#63677A",width:3,type:"solid"}},axisTick:{show:!0,splitNumber:5,length:6,distance:10,lineStyle:{color:"#63677A",width:1,type:"solid"}},axisLabel:{show:!0,distance:15,color:"#464646",fontSize:12,rotate:0},pointer:{icon:null,offsetCenter:[0,0],show:!0,showAbove:!0,length:"60%",width:6,keepAspect:!1},anchor:{show:!1,showAbove:!1,size:6,icon:"circle",offsetCenter:[0,0],keepAspect:!1,itemStyle:{color:"#fff",borderWidth:0,borderColor:"#5470c6"}},title:{show:!0,offsetCenter:[0,"20%"],color:"#464646",fontSize:16,valueAnimation:!1},detail:{show:!0,backgroundColor:"rgba(0,0,0,0)",borderWidth:0,borderColor:"#ccc",width:100,height:null,padding:[5,10],offsetCenter:[0,"40%"],color:"#464646",fontSize:30,fontWeight:"bold",lineHeight:30,valueAnimation:!1}},e}(Mp),mE=pE;function gE(t){t.registerChartView(fE),t.registerSeriesModel(mE)}var yE=["itemStyle","opacity"],vE=function(t){function e(e,n){var r=t.call(this)||this,i=r,o=new Im,a=new ks;return i.setTextContent(a),r.setTextGuideLine(o),r.updateData(e,n,!0),r}return o(e,t),e.prototype.updateData=function(t,e,n){var r=this,i=t.hostModel,o=t.getItemModel(e),a=t.getItemLayout(e),s=o.getModel("emphasis"),l=o.get(yE);l=null==l?1:l,n||Xl(r),r.useStyle(t.getItemVisual(e,"style")),r.style.lineJoin="round",n?(r.setShape({points:a.points}),r.style.opacity=0,Vl(r,{style:{opacity:l}},i,e)):Wl(r,{style:{opacity:l},shape:{points:a.points}},i,e),Al(r,o),this._updateLabel(t,e),Dl(this,s.get("focus"),s.get("blurScope"),s.get("disabled"))},e.prototype._updateLabel=function(t,e){var n=this,r=this.getTextGuideLine(),i=n.getTextContent(),o=t.hostModel,a=t.getItemModel(e),s=t.getItemLayout(e),l=s.label,u=t.getItemVisual(e,"style"),c=u.fill;tu(i,eu(a),{labelFetcher:t.hostModel,labelDataIndex:e,defaultOpacity:u.opacity,defaultText:t.getName(e)},{normal:{align:l.textAlign,verticalAlign:l.verticalAlign}}),n.setTextConfig({local:!0,inside:!!l.inside,insideStroke:c,outsideFill:c});var d=l.linePoints;r.setShape({points:d}),n.textGuideLineConfig={anchor:d?new yt(d[0][0],d[0][1]):null},Wl(i,{style:{x:l.x,y:l.y}},o,e),i.attr({rotation:l.rotation,originX:l.x,originY:l.y,z2:10}),rx(n,ix(a),{stroke:c})},e}(Tm),_E=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.ignoreLabelLineUpdate=!0,n}return o(e,t),e.prototype.render=function(t,e,n){var r=t.getData(),i=this._data,o=this.group;r.diff(i).add((function(t){var e=new vE(r,t);r.setItemGraphicEl(t,e),o.add(e)})).update((function(t,e){var n=i.getItemGraphicEl(e);n.updateData(r,t),o.add(n),r.setItemGraphicEl(t,n)})).remove((function(e){var n=i.getItemGraphicEl(e);ql(n,t,e)})).execute(),this._data=r},e.prototype.remove=function(){this.group.removeAll(),this._data=null},e.prototype.dispose=function(){},e.type="funnel",e}($g),bE=_E,xE=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(e){t.prototype.init.apply(this,arguments),this.legendVisualProvider=new OL(s["c"](this.getData,this),s["c"](this.getRawData,this)),this._defaultLabelLine(e)},e.prototype.getInitialData=function(t,e){return TL(this,{coordDimensions:["value"],encodeDefaulter:s["h"](hd,this)})},e.prototype._defaultLabelLine=function(t){ci(t,"labelLine",["show"]);var e=t.labelLine,n=t.emphasis.labelLine;e.show=e.show&&t.label.show,n.show=n.show&&t.emphasis.label.show},e.prototype.getDataParams=function(e){var n=this.getData(),r=t.prototype.getDataParams.call(this,e),i=n.mapDimension("value"),o=n.getSum(i);return r.percent=o?+(n.get(i,e)/o*100).toFixed(2):0,r.$vars.push("percent"),r},e.type="series.funnel",e.defaultOption={z:2,legendHoverLink:!0,colorBy:"data",left:80,top:60,right:80,bottom:60,minSize:"0%",maxSize:"100%",sort:"descending",orient:"vertical",gap:0,funnelAlign:"center",label:{show:!0,position:"outer"},labelLine:{show:!0,length:20,lineStyle:{width:1}},itemStyle:{borderColor:"#fff",borderWidth:1},emphasis:{label:{show:!0}},select:{itemStyle:{borderColor:"#212121"}}},e}(Mp),wE=xE;function ME(t,e){return Hc(t.getBoxLayoutParams(),{width:e.getWidth(),height:e.getHeight()})}function SE(t,e){for(var n=t.mapDimension("value"),r=t.mapArray(n,(function(t){return t})),i=[],o="ascending"===e,a=0,l=t.count();a<l;a++)i[a]=a;return Object(s["w"])(e)?i.sort(e):"none"!==e&&i.sort((function(t,e){return o?r[t]-r[e]:r[e]-r[t]})),i}function kE(t){var e=t.hostModel,n=e.get("orient");t.each((function(e){var r,i,o,a,s=t.getItemModel(e),l=s.getModel("label"),u=l.get("position"),c=s.getModel("labelLine"),d=t.getItemLayout(e),h=d.points,f="inner"===u||"inside"===u||"center"===u||"insideLeft"===u||"insideRight"===u;if(f)"insideLeft"===u?(i=(h[0][0]+h[3][0])/2+5,o=(h[0][1]+h[3][1])/2,r="left"):"insideRight"===u?(i=(h[1][0]+h[2][0])/2-5,o=(h[1][1]+h[2][1])/2,r="right"):(i=(h[0][0]+h[1][0]+h[2][0]+h[3][0])/4,o=(h[0][1]+h[1][1]+h[2][1]+h[3][1])/4,r="center"),a=[[i,o],[i,o]];else{var p=void 0,m=void 0,g=void 0,y=void 0,v=c.get("length");0,"left"===u?(p=(h[3][0]+h[0][0])/2,m=(h[3][1]+h[0][1])/2,g=p-v,i=g-5,r="right"):"right"===u?(p=(h[1][0]+h[2][0])/2,m=(h[1][1]+h[2][1])/2,g=p+v,i=g+5,r="left"):"top"===u?(p=(h[3][0]+h[0][0])/2,m=(h[3][1]+h[0][1])/2,y=m-v,o=y-5,r="center"):"bottom"===u?(p=(h[1][0]+h[2][0])/2,m=(h[1][1]+h[2][1])/2,y=m+v,o=y+5,r="center"):"rightTop"===u?(p="horizontal"===n?h[3][0]:h[1][0],m="horizontal"===n?h[3][1]:h[1][1],"horizontal"===n?(y=m-v,o=y-5,r="center"):(g=p+v,i=g+5,r="top")):"rightBottom"===u?(p=h[2][0],m=h[2][1],"horizontal"===n?(y=m+v,o=y+5,r="center"):(g=p+v,i=g+5,r="bottom")):"leftTop"===u?(p=h[0][0],m="horizontal"===n?h[0][1]:h[1][1],"horizontal"===n?(y=m-v,o=y-5,r="center"):(g=p-v,i=g-5,r="right")):"leftBottom"===u?(p="horizontal"===n?h[1][0]:h[3][0],m="horizontal"===n?h[1][1]:h[2][1],"horizontal"===n?(y=m+v,o=y+5,r="center"):(g=p-v,i=g-5,r="right")):(p=(h[1][0]+h[2][0])/2,m=(h[1][1]+h[2][1])/2,"horizontal"===n?(y=m+v,o=y+5,r="center"):(g=p+v,i=g+5,r="left")),"horizontal"===n?(g=p,i=g):(y=m,o=y),a=[[p,m],[g,y]]}d.label={linePoints:a,x:i,y:o,verticalAlign:"middle",textAlign:r,inside:f}}))}function LE(t,e){t.eachSeriesByType("funnel",(function(t){var n=t.getData(),r=n.mapDimension("value"),i=t.get("sort"),o=ME(t,e),a=t.get("orient"),s=o.width,l=o.height,u=SE(n,i),c=o.x,d=o.y,h="horizontal"===a?[Nr(t.get("minSize"),l),Nr(t.get("maxSize"),l)]:[Nr(t.get("minSize"),s),Nr(t.get("maxSize"),s)],f=n.getDataExtent(r),p=t.get("min"),m=t.get("max");null==p&&(p=Math.min(f[0],0)),null==m&&(m=f[1]);var g=t.get("funnelAlign"),y=t.get("gap"),v="horizontal"===a?s:l,_=(v-y*(n.count()-1))/n.count(),b=function(t,e){if("horizontal"===a){var i=n.get(r,t)||0,o=Er(i,[p,m],h,!0),u=void 0;switch(g){case"top":u=d;break;case"center":u=d+(l-o)/2;break;case"bottom":u=d+(l-o);break}return[[e,u],[e,u+o]]}var f,y=n.get(r,t)||0,v=Er(y,[p,m],h,!0);switch(g){case"left":f=c;break;case"center":f=c+(s-v)/2;break;case"right":f=c+s-v;break}return[[f,e],[f+v,e]]};"ascending"===i&&(_=-_,y=-y,"horizontal"===a?c+=s:d+=l,u=u.reverse());for(var x=0;x<u.length;x++){var w=u[x],M=u[x+1],S=n.getItemModel(w);if("horizontal"===a){var k=S.get(["itemStyle","width"]);null==k?k=_:(k=Nr(k,s),"ascending"===i&&(k=-k));var L=b(w,c),T=b(M,c+k);c+=k+y,n.setItemLayout(w,{points:L.concat(T.slice().reverse())})}else{var D=S.get(["itemStyle","height"]);null==D?D=_:(D=Nr(D,l),"ascending"===i&&(D=-D));L=b(w,d),T=b(M,d+D);d+=D+y,n.setItemLayout(w,{points:L.concat(T.slice().reverse())})}}kE(n)}))}function TE(t){t.registerChartView(bE),t.registerSeriesModel(wE),t.registerLayout(LE),t.registerProcessor(yL("funnel"))}var DE=.3,OE=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n._dataGroup=new wr,n._initialized=!1,n}return o(e,t),e.prototype.init=function(){this.group.add(this._dataGroup)},e.prototype.render=function(t,e,n,r){this._progressiveEls=null;var i=this._dataGroup,o=t.getData(),a=this._data,s=t.coordinateSystem,l=s.dimensions,u=YE(t);function c(t){var e=AE(o,i,t,l,s);jE(e,o,t,u)}function d(e,n){var r=a.getItemGraphicEl(n),i=CE(o,e,l,s);o.setItemGraphicEl(e,r),Wl(r,{shape:{points:i}},t,e),Xl(r),jE(r,o,e,u)}function h(t){var e=a.getItemGraphicEl(t);i.remove(e)}if(o.diff(a).add(c).update(d).remove(h).execute(),!this._initialized){this._initialized=!0;var f=IE(s,t,(function(){setTimeout((function(){i.removeClipPath()}))}));i.setClipPath(f)}this._data=o},e.prototype.incrementalPrepareRender=function(t,e,n){this._initialized=!0,this._data=null,this._dataGroup.removeAll()},e.prototype.incrementalRender=function(t,e,n){for(var r=e.getData(),i=e.coordinateSystem,o=i.dimensions,a=YE(e),s=this._progressiveEls=[],l=t.start;l<t.end;l++){var u=AE(r,this._dataGroup,l,o,i);u.incremental=!0,jE(u,r,l,a),s.push(u)}},e.prototype.remove=function(){this._dataGroup&&this._dataGroup.removeAll(),this._data=null},e.type="parallel",e}($g);function IE(t,e,n){var r=t.model,i=t.getRect(),o=new ss({shape:{x:i.x,y:i.y,width:i.width,height:i.height}}),a="horizontal"===r.get("layout")?"width":"height";return o.setShape(a,0),Vl(o,{shape:{width:i.width,height:i.height}},e,n),o}function CE(t,e,n,r){for(var i=[],o=0;o<n.length;o++){var a=n[o],s=t.get(t.mapDimension(a),e);PE(s,r.getAxis(a).type)||i.push(r.dataToPoint(s,a))}return i}function AE(t,e,n,r,i){var o=CE(t,n,r,i),a=new Im({shape:{points:o},z2:10});return e.add(a),t.setItemGraphicEl(n,a),a}function YE(t){var e=t.get("smooth",!0);return!0===e&&(e=DE),e=ei(e),Object(s["l"])(e)&&(e=0),{smooth:e}}function jE(t,e,n,r){t.useStyle(e.getItemVisual(n,"style")),t.style.fill=null,t.setShape("smooth",r.smooth);var i=e.getItemModel(n),o=i.getModel("emphasis");Al(t,i,"lineStyle"),Dl(t,o.get("focus"),o.get("blurScope"),o.get("disabled"))}function PE(t,e){return"category"===e?null==t:null==t||isNaN(t)}var EE=OE,NE=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.visualStyleAccessPath="lineStyle",n.visualDrawType="stroke",n}return o(e,t),e.prototype.getInitialData=function(t,e){return dS(null,this,{useEncodeDefaulter:Object(s["c"])(RE,null,this)})},e.prototype.getRawIndicesByActiveState=function(t){var e=this.coordinateSystem,n=this.getData(),r=[];return e.eachActiveState(n,(function(e,i){t===e&&r.push(n.getRawIndex(i))})),r},e.type="series.parallel",e.dependencies=["parallel"],e.defaultOption={z:2,coordinateSystem:"parallel",parallelIndex:0,label:{show:!1},inactiveOpacity:.05,activeOpacity:1,lineStyle:{width:1,opacity:.45,type:"solid"},emphasis:{label:{show:!1}},progressive:500,smooth:!1,animationEasing:"linear"},e}(Mp);function RE(t){var e=t.ecModel.getComponent("parallel",t.get("parallelIndex"));if(e){var n={};return Object(s["k"])(e.dimensions,(function(t){var e=HE(t);n[t]=e})),n}}function HE(t){return+t.replace("dim","")}var zE=NE,FE=["lineStyle","opacity"],BE={seriesType:"parallel",reset:function(t,e){var n=t.coordinateSystem,r={normal:t.get(["lineStyle","opacity"]),active:t.get("activeOpacity"),inactive:t.get("inactiveOpacity")};return{progress:function(t,e){n.eachActiveState(e,(function(t,n){var i=r[t];if("normal"===t&&e.hasItemOption){var o=e.getItemModel(n).get(FE,!0);null!=o&&(i=o)}var a=e.ensureUniqueItemVisual(n,"style");a.opacity=i}),t.start,t.end)}}}},WE=BE;function VE(t){GE(t),$E(t)}function GE(t){if(!t.parallel){var e=!1;s["k"](t.series,(function(t){t&&"parallel"===t.type&&(e=!0)})),e&&(t.parallel=[{}])}}function $E(t){var e=ui(t.parallelAxis);s["k"](e,(function(e){if(s["A"](e)){var n=e.parallelIndex||0,r=ui(t.parallel)[n];r&&r.parallelAxisDefault&&s["I"](e,r.parallelAxisDefault,!1)}}))}var UE=5,qE=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){this._model=t,this._api=n,this._handlers||(this._handlers={},Object(s["k"])(XE,(function(t,e){n.getZr().on(e,this._handlers[e]=Object(s["c"])(t,this))}),this)),Jg(this,"_throttledDispatchExpand",t.get("axisExpandRate"),"fixRate")},e.prototype.dispose=function(t,e){Kg(this,"_throttledDispatchExpand"),Object(s["k"])(this._handlers,(function(t,n){e.getZr().off(n,t)})),this._handlers=null},e.prototype._throttledDispatchExpand=function(t){this._dispatchExpand(t)},e.prototype._dispatchExpand=function(t){t&&this._api.dispatchAction(Object(s["m"])({type:"parallelAxisExpand"},t))},e.type="parallel",e}(kp),XE={mousedown:function(t){ZE(this,"click")&&(this._mouseDownPoint=[t.offsetX,t.offsetY])},mouseup:function(t){var e=this._mouseDownPoint;if(ZE(this,"click")&&e){var n=[t.offsetX,t.offsetY],r=Math.pow(e[0]-n[0],2)+Math.pow(e[1]-n[1],2);if(r>UE)return;var i=this._model.coordinateSystem.getSlidedAxisExpandWindow([t.offsetX,t.offsetY]);"none"!==i.behavior&&this._dispatchExpand({axisExpandWindow:i.axisExpandWindow})}this._mouseDownPoint=null},mousemove:function(t){if(!this._mouseDownPoint&&ZE(this,"mousemove")){var e=this._model,n=e.coordinateSystem.getSlidedAxisExpandWindow([t.offsetX,t.offsetY]),r=n.behavior;"jump"===r&&this._throttledDispatchExpand.debounceNextCall(e.get("axisExpandDebounce")),this._throttledDispatchExpand("none"===r?null:{axisExpandWindow:n.axisExpandWindow,animation:"jump"===r?null:{duration:0}})}}};function ZE(t,e){var n=t._model;return n.get("axisExpandable")&&n.get("axisExpandTriggerOn")===e}var JE=qE,KE=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(){t.prototype.init.apply(this,arguments),this.mergeOption({})},e.prototype.mergeOption=function(t){var e=this.option;t&&s["I"](e,t,!0),this._initDimensions()},e.prototype.contains=function(t,e){var n=t.get("parallelIndex");return null!=n&&e.getComponent("parallel",n)===this},e.prototype.setAxisExpand=function(t){s["k"](["axisExpandable","axisExpandCenter","axisExpandCount","axisExpandWidth","axisExpandWindow"],(function(e){t.hasOwnProperty(e)&&(this.option[e]=t[e])}),this)},e.prototype._initDimensions=function(){var t=this.dimensions=[],e=this.parallelAxisIndex=[],n=s["n"](this.ecModel.queryComponents({mainType:"parallelAxis"}),(function(t){return(t.get("parallelIndex")||0)===this.componentIndex}),this);s["k"](n,(function(n){t.push("dim"+n.get("dim")),e.push(n.componentIndex)}))},e.type="parallel",e.dependencies=["parallelAxis"],e.layoutMode="box",e.defaultOption={z:0,left:80,top:60,right:80,bottom:60,layout:"horizontal",axisExpandable:!1,axisExpandCenter:null,axisExpandCount:0,axisExpandWidth:50,axisExpandRate:17,axisExpandDebounce:50,axisExpandSlideTriggerArea:[-.15,.05,.4],axisExpandTriggerOn:"click",parallelAxisDefault:null},e}(Xc),QE=KE,tN=function(t){function e(e,n,r,i,o){var a=t.call(this,e,n,r)||this;return a.type=i||"value",a.axisIndex=o,a}return o(e,t),e.prototype.isHorizontal=function(){return"horizontal"!==this.coordinateSystem.getModel().get("layout")},e}(PD),eN=tN;function nN(t,e,n,r,i,o){t=t||0;var a=n[1]-n[0];if(null!=i&&(i=iN(i,[0,a])),null!=o&&(o=Math.max(o,null!=i?i:0)),"all"===r){var s=Math.abs(e[1]-e[0]);s=iN(s,[0,a]),i=o=iN(s,[i,o]),r=0}e[0]=iN(e[0],n),e[1]=iN(e[1],n);var l=rN(e,r);e[r]+=t;var u,c=i||0,d=n.slice();return l.sign<0?d[0]+=c:d[1]-=c,e[r]=iN(e[r],d),u=rN(e,r),null!=i&&(u.sign!==l.sign||u.span<i)&&(e[1-r]=e[r]+l.sign*i),u=rN(e,r),null!=o&&u.span>o&&(e[1-r]=e[r]+u.sign*o),e}function rN(t,e){var n=t[e]-t[1-e];return{span:Math.abs(n),sign:n>0?-1:n<0?1:e?-1:1}}function iN(t,e){return Math.min(null!=e[1]?e[1]:1/0,Math.max(null!=e[0]?e[0]:-1/0,t))}var oN=s["k"],aN=Math.min,sN=Math.max,lN=Math.floor,uN=Math.ceil,cN=Rr,dN=Math.PI,hN=function(){function t(t,e,n){this.type="parallel",this._axesMap=s["f"](),this._axesLayout={},this.dimensions=t.dimensions,this._model=t,this._init(t,e,n)}return t.prototype._init=function(t,e,n){var r=t.dimensions,i=t.parallelAxisIndex;oN(r,(function(t,n){var r=i[n],o=e.getComponent("parallelAxis",r),a=this._axesMap.set(t,new eN(t,tD(o),[0,0],o.get("type"),r)),s="category"===a.type;a.onBand=s&&o.get("boundaryGap"),a.inverse=o.get("inverse"),o.axis=a,a.model=o,a.coordinateSystem=o.coordinateSystem=this}),this)},t.prototype.update=function(t,e){this._updateAxesFromSeries(this._model,t)},t.prototype.containPoint=function(t){var e=this._makeLayoutInfo(),n=e.axisBase,r=e.layoutBase,i=e.pixelDimIndex,o=t[1-i],a=t[i];return o>=n&&o<=n+e.axisLength&&a>=r&&a<=r+e.layoutLength},t.prototype.getModel=function(){return this._model},t.prototype._updateAxesFromSeries=function(t,e){e.eachSeries((function(n){if(t.contains(n,e)){var r=n.getData();oN(this.dimensions,(function(t){var e=this._axesMap.get(t);e.scale.unionExtentFromData(r,r.mapDimension(t)),QT(e.scale,e.model)}),this)}}),this)},t.prototype.resize=function(t,e){this._rect=Hc(t.getBoxLayoutParams(),{width:e.getWidth(),height:e.getHeight()}),this._layoutAxes()},t.prototype.getRect=function(){return this._rect},t.prototype._makeLayoutInfo=function(){var t,e=this._model,n=this._rect,r=["x","y"],i=["width","height"],o=e.get("layout"),a="horizontal"===o?0:1,s=n[i[a]],l=[0,s],u=this.dimensions.length,c=fN(e.get("axisExpandWidth"),l),d=fN(e.get("axisExpandCount")||0,[0,u]),h=e.get("axisExpandable")&&u>3&&u>d&&d>1&&c>0&&s>0,f=e.get("axisExpandWindow");if(f)t=fN(f[1]-f[0],l),f[1]=f[0]+t;else{t=fN(c*(d-1),l);var p=e.get("axisExpandCenter")||lN(u/2);f=[c*p-t/2],f[1]=f[0]+t}var m=(s-t)/(u-d);m<3&&(m=0);var g=[lN(cN(f[0]/c,1))+1,uN(cN(f[1]/c,1))-1],y=m/c*f[0];return{layout:o,pixelDimIndex:a,layoutBase:n[r[a]],layoutLength:s,axisBase:n[r[1-a]],axisLength:n[i[1-a]],axisExpandable:h,axisExpandWidth:c,axisCollapseWidth:m,axisExpandWindow:f,axisCount:u,winInnerIndices:g,axisExpandWindow0Pos:y}},t.prototype._layoutAxes=function(){var t=this._rect,e=this._axesMap,n=this.dimensions,r=this._makeLayoutInfo(),i=r.layout;e.each((function(t){var e=[0,r.axisLength],n=t.inverse?1:0;t.setExtent(e[n],e[1-n])})),oN(n,(function(e,n){var o=(r.axisExpandable?mN:pN)(n,r),a={horizontal:{x:o.position,y:r.axisLength},vertical:{x:0,y:o.position}},s={horizontal:dN/2,vertical:0},l=[a[i].x+t.x,a[i].y+t.y],u=s[i],c=st();ht(c,c,u),dt(c,c,l),this._axesLayout[e]={position:l,rotation:u,transform:c,axisNameAvailableWidth:o.axisNameAvailableWidth,axisLabelShow:o.axisLabelShow,nameTruncateMaxWidth:o.nameTruncateMaxWidth,tickDirection:1,labelDirection:1}}),this)},t.prototype.getAxis=function(t){return this._axesMap.get(t)},t.prototype.dataToPoint=function(t,e){return this.axisCoordToPoint(this._axesMap.get(e).dataToCoord(t),e)},t.prototype.eachActiveState=function(t,e,n,r){null==n&&(n=0),null==r&&(r=t.count());var i=this._axesMap,o=this.dimensions,a=[],l=[];s["k"](o,(function(e){a.push(t.mapDimension(e)),l.push(i.get(e).model)}));for(var u=this.hasAxisBrushed(),c=n;c<r;c++){var d=void 0;if(u){d="active";for(var h=t.getValues(a,c),f=0,p=o.length;f<p;f++){var m=l[f].getActiveState(h[f]);if("inactive"===m){d="inactive";break}}}else d="normal";e(d,c)}},t.prototype.hasAxisBrushed=function(){for(var t=this.dimensions,e=this._axesMap,n=!1,r=0,i=t.length;r<i;r++)"normal"!==e.get(t[r]).model.getActiveState()&&(n=!0);return n},t.prototype.axisCoordToPoint=function(t,e){var n=this._axesLayout[e];return Mg([t,0],n.transform)},t.prototype.getAxisLayout=function(t){return s["d"](this._axesLayout[t])},t.prototype.getSlidedAxisExpandWindow=function(t){var e=this._makeLayoutInfo(),n=e.pixelDimIndex,r=e.axisExpandWindow.slice(),i=r[1]-r[0],o=[0,e.axisExpandWidth*(e.axisCount-1)];if(!this.containPoint(t))return{behavior:"none",axisExpandWindow:r};var a,s=t[n]-e.layoutBase-e.axisExpandWindow0Pos,l="slide",u=e.axisCollapseWidth,c=this._model.get("axisExpandSlideTriggerArea"),d=null!=c[0];if(u)d&&u&&s<i*c[0]?(l="jump",a=s-i*c[2]):d&&u&&s>i*(1-c[0])?(l="jump",a=s-i*(1-c[2])):(a=s-i*c[1])>=0&&(a=s-i*(1-c[1]))<=0&&(a=0),a*=e.axisExpandWidth/u,a?nN(a,r,o,"all"):l="none";else{var h=r[1]-r[0],f=o[1]*s/h;r=[sN(0,f-h/2)],r[1]=aN(o[1],r[0]+h),r[0]=r[1]-h}return{axisExpandWindow:r,behavior:l}},t}();function fN(t,e){return aN(sN(t,e[0]),e[1])}function pN(t,e){var n=e.layoutLength/(e.axisCount-1);return{position:n*t,axisNameAvailableWidth:n,axisLabelShow:!0}}function mN(t,e){var n,r,i=e.layoutLength,o=e.axisExpandWidth,a=e.axisCount,s=e.axisCollapseWidth,l=e.winInnerIndices,u=s,c=!1;return t<l[0]?(n=t*s,r=s):t<=l[1]?(n=e.axisExpandWindow0Pos+t*o-e.axisExpandWindow[0],u=o,c=!0):(n=i-(a-1-t)*s,r=s),{position:n,axisNameAvailableWidth:u,axisLabelShow:c,nameTruncateMaxWidth:r}}var gN=hN;function yN(t,e){var n=[];return t.eachComponent("parallel",(function(r,i){var o=new gN(r,t,e);o.name="parallel_"+i,o.resize(r,e),r.coordinateSystem=o,o.model=r,n.push(o)})),t.eachSeries((function(t){if("parallel"===t.get("coordinateSystem")){var e=t.getReferringComponents("parallel",Pi).models[0];t.coordinateSystem=e.coordinateSystem}})),n}var vN={create:yN},_N=vN,bN=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.activeIntervals=[],n}return o(e,t),e.prototype.getAreaSelectStyle=function(){return ro([["fill","color"],["lineWidth","borderWidth"],["stroke","borderColor"],["width","width"],["opacity","opacity"]])(this.getModel("areaSelectStyle"))},e.prototype.setActiveIntervals=function(t){var e=this.activeIntervals=s["d"](t);if(e)for(var n=e.length-1;n>=0;n--)Hr(e[n])},e.prototype.getActiveState=function(t){var e=this.activeIntervals;if(!e.length)return"normal";if(null==t||isNaN(+t))return"inactive";if(1===e.length){var n=e[0];if(n[0]<=t&&t<=n[1])return"active"}else for(var r=0,i=e.length;r<i;r++)if(e[r][0]<=t&&t<=e[r][1])return"active";return"inactive"},e}(Xc);s["K"](bN,$L);var xN=bN,wN=!0,MN=Math.min,SN=Math.max,kN=Math.pow,LN=1e4,TN=6,DN=6,ON="globalPan",IN={w:[0,0],e:[0,1],n:[1,0],s:[1,1]},CN={w:"ew",e:"ew",n:"ns",s:"ns",ne:"nesw",sw:"nesw",nw:"nwse",se:"nwse"},AN={brushStyle:{lineWidth:2,stroke:"rgba(210,219,238,0.3)",fill:"#D2DBEE"},transformable:!0,brushMode:"single",removeOnClick:!1},YN=0,jN=function(t){function e(e){var n=t.call(this)||this;return n._track=[],n._covers=[],n._handlers={},n._zr=e,n.group=new wr,n._uid="brushController_"+YN++,Object(s["k"])(hR,(function(t,e){this._handlers[e]=Object(s["c"])(t,this)}),n),n}return o(e,t),e.prototype.enableBrush=function(t){return this._brushType&&this._doDisableBrush(),t.brushType&&this._doEnableBrush(t),this},e.prototype._doEnableBrush=function(t){var e=this._zr;this._enableGlobalPan||rI(e,ON,this._uid),Object(s["k"])(this._handlers,(function(t,n){e.on(n,t)})),this._brushType=t.brushType,this._brushOption=Object(s["I"])(Object(s["d"])(AN),t,!0)},e.prototype._doDisableBrush=function(){var t=this._zr;iI(t,ON,this._uid),Object(s["k"])(this._handlers,(function(e,n){t.off(n,e)})),this._brushType=this._brushOption=null},e.prototype.setPanels=function(t){if(t&&t.length){var e=this._panels={};Object(s["k"])(t,(function(t){e[t.panelId]=Object(s["d"])(t)}))}else this._panels=null;return this},e.prototype.mount=function(t){t=t||{},this._enableGlobalPan=t.enableGlobalPan;var e=this.group;return this._zr.add(e),e.attr({x:t.x||0,y:t.y||0,rotation:t.rotation||0,scaleX:t.scaleX||1,scaleY:t.scaleY||1}),this._transform=e.getLocalTransform(),this},e.prototype.updateCovers=function(t){t=Object(s["H"])(t,(function(t){return Object(s["I"])(Object(s["d"])(AN),t,!0)}));var e="\0-brush-index-",n=this._covers,r=this._covers=[],i=this,o=this._creatingCover;return new yM(n,t,l,a).add(u).update(u).remove(c).execute(),this;function a(t,n){return(null!=t.id?t.id:e+n)+"-"+t.brushType}function l(t,e){return a(t.__brushOption,e)}function u(e,a){var s=t[e];if(null!=a&&n[a]===o)r[e]=n[a];else{var l=r[e]=null!=a?(n[a].__brushOption=s,n[a]):EN(i,PN(i,s));HN(i,l)}}function c(t){n[t]!==o&&i.group.remove(n[t])}},e.prototype.unmount=function(){return this.enableBrush(!1),WN(this),this._zr.remove(this.group),this},e.prototype.dispose=function(){this.unmount(),this.off()},e}(C);function PN(t,e){var n=mR[e.brushType].createCover(t,e);return n.__brushOption=e,RN(n,e),t.group.add(n),n}function EN(t,e){var n=zN(e);return n.endCreating&&(n.endCreating(t,e),RN(e,e.__brushOption)),e}function NN(t,e){var n=e.__brushOption;zN(e).updateCoverShape(t,e,n.range,n)}function RN(t,e){var n=e.z;null==n&&(n=LN),t.traverse((function(t){t.z=n,t.z2=n}))}function HN(t,e){zN(e).updateCommon(t,e),NN(t,e)}function zN(t){return mR[t.__brushOption.brushType]}function FN(t,e,n){var r,i=t._panels;if(!i)return wN;var o=t._transform;return Object(s["k"])(i,(function(t){t.isTargetByCursor(e,n,o)&&(r=t)})),r}function BN(t,e){var n=t._panels;if(!n)return wN;var r=e.__brushOption.panelId;return null!=r?n[r]:wN}function WN(t){var e=t._covers,n=e.length;return Object(s["k"])(e,(function(e){t.group.remove(e)}),t),e.length=0,!!n}function VN(t,e){var n=Object(s["H"])(t._covers,(function(t){var e=t.__brushOption,n=Object(s["d"])(e.range);return{brushType:e.brushType,panelId:e.panelId,range:n}}));t.trigger("brush",{areas:n,isEnd:!!e.isEnd,removeOnClick:!!e.removeOnClick})}function GN(t){var e=t._track;if(!e.length)return!1;var n=e[e.length-1],r=e[0],i=n[0]-r[0],o=n[1]-r[1],a=kN(i*i+o*o,.5);return a>TN}function $N(t){var e=t.length-1;return e<0&&(e=0),[t[0],t[e]]}function UN(t,e,n,r){var i=new wr;return i.add(new ss({name:"main",style:JN(n),silent:!0,draggable:!0,cursor:"move",drift:Object(s["h"])(nR,t,e,i,["n","s","w","e"]),ondragend:Object(s["h"])(VN,e,{isEnd:!0})})),Object(s["k"])(r,(function(n){i.add(new ss({name:n.join(""),style:{opacity:0},draggable:!0,silent:!0,invisible:!0,drift:Object(s["h"])(nR,t,e,i,n),ondragend:Object(s["h"])(VN,e,{isEnd:!0})}))})),i}function qN(t,e,n,r){var i=r.brushStyle.lineWidth||0,o=SN(i,DN),a=n[0][0],s=n[1][0],l=a-i/2,u=s-i/2,c=n[0][1],d=n[1][1],h=c-o+i/2,f=d-o+i/2,p=c-a,m=d-s,g=p+i,y=m+i;ZN(t,e,"main",a,s,p,m),r.transformable&&(ZN(t,e,"w",l,u,o,y),ZN(t,e,"e",h,u,o,y),ZN(t,e,"n",l,u,g,o),ZN(t,e,"s",l,f,g,o),ZN(t,e,"nw",l,u,o,o),ZN(t,e,"ne",h,u,o,o),ZN(t,e,"sw",l,f,o,o),ZN(t,e,"se",h,f,o,o))}function XN(t,e){var n=e.__brushOption,r=n.transformable,i=e.childAt(0);i.useStyle(JN(n)),i.attr({silent:!r,cursor:r?"move":"default"}),Object(s["k"])([["w"],["e"],["n"],["s"],["s","e"],["s","w"],["n","e"],["n","w"]],(function(n){var i=e.childOfName(n.join("")),o=1===n.length?tR(t,n[0]):eR(t,n);i&&i.attr({silent:!r,invisible:!r,cursor:r?CN[o]+"-resize":null})}))}function ZN(t,e,n,r,i,o,a){var s=e.childOfName(n);s&&s.setShape(aR(oR(t,e,[[r,i],[r+o,i+a]])))}function JN(t){return Object(s["i"])({strokeNoScale:!0},t.brushStyle)}function KN(t,e,n,r){var i=[MN(t,n),MN(e,r)],o=[SN(t,n),SN(e,r)];return[[i[0],o[0]],[i[1],o[1]]]}function QN(t){return wg(t.group)}function tR(t,e){var n={w:"left",e:"right",n:"top",s:"bottom"},r={left:"w",right:"e",top:"n",bottom:"s"},i=Sg(n[e],QN(t));return r[i]}function eR(t,e){var n=[tR(t,e[0]),tR(t,e[1])];return("e"===n[0]||"w"===n[0])&&n.reverse(),n.join("")}function nR(t,e,n,r,i,o){var a=n.__brushOption,l=t.toRectRange(a.range),u=iR(e,i,o);Object(s["k"])(r,(function(t){var e=IN[t];l[e[0]][e[1]]+=u[e[0]]})),a.range=t.fromRectRange(KN(l[0][0],l[1][0],l[0][1],l[1][1])),HN(e,n),VN(e,{isEnd:!1})}function rR(t,e,n,r){var i=e.__brushOption.range,o=iR(t,n,r);Object(s["k"])(i,(function(t){t[0]+=o[0],t[1]+=o[1]})),HN(t,e),VN(t,{isEnd:!1})}function iR(t,e,n){var r=t.group,i=r.transformCoordToLocal(e,n),o=r.transformCoordToLocal(0,0);return[i[0]-o[0],i[1]-o[1]]}function oR(t,e,n){var r=BN(t,e);return r&&r!==wN?r.clipPath(n,t._transform):Object(s["d"])(n)}function aR(t){var e=MN(t[0][0],t[1][0]),n=MN(t[0][1],t[1][1]),r=SN(t[0][0],t[1][0]),i=SN(t[0][1],t[1][1]);return{x:e,y:n,width:r-e,height:i-n}}function sR(t,e,n){if(t._brushType&&!pR(t,e.offsetX,e.offsetY)){var r=t._zr,i=t._covers,o=FN(t,e,n);if(!t._dragging)for(var a=0;a<i.length;a++){var s=i[a].__brushOption;if(o&&(o===wN||s.panelId===o.panelId)&&mR[s.brushType].contain(i[a],n[0],n[1]))return}o&&r.setCursorStyle("crosshair")}}function lR(t){var e=t.event;e.preventDefault&&e.preventDefault()}function uR(t,e,n){return t.childOfName("main").contain(e,n)}function cR(t,e,n,r){var i,o=t._creatingCover,a=t._creatingPanel,l=t._brushOption;if(t._track.push(n.slice()),GN(t)||o){if(a&&!o){"single"===l.brushMode&&WN(t);var u=Object(s["d"])(l);u.brushType=dR(u.brushType,a),u.panelId=a===wN?null:a.panelId,o=t._creatingCover=PN(t,u),t._covers.push(o)}if(o){var c=mR[dR(t._brushType,a)],d=o.__brushOption;d.range=c.getCreatingRange(oR(t,o,t._track)),r&&(EN(t,o),c.updateCommon(t,o)),NN(t,o),i={isEnd:r}}}else r&&"single"===l.brushMode&&l.removeOnClick&&FN(t,e,n)&&WN(t)&&(i={isEnd:r,removeOnClick:!0});return i}function dR(t,e){return"auto"===t?e.defaultBrushType:t}var hR={mousedown:function(t){if(this._dragging)fR(this,t);else if(!t.target||!t.target.draggable){lR(t);var e=this.group.transformCoordToLocal(t.offsetX,t.offsetY);this._creatingCover=null;var n=this._creatingPanel=FN(this,t,e);n&&(this._dragging=!0,this._track=[e.slice()])}},mousemove:function(t){var e=t.offsetX,n=t.offsetY,r=this.group.transformCoordToLocal(e,n);if(sR(this,t,r),this._dragging){lR(t);var i=cR(this,t,r,!1);i&&VN(this,i)}},mouseup:function(t){fR(this,t)}};function fR(t,e){if(t._dragging){lR(e);var n=e.offsetX,r=e.offsetY,i=t.group.transformCoordToLocal(n,r),o=cR(t,e,i,!0);t._dragging=!1,t._track=[],t._creatingCover=null,o&&VN(t,o)}}function pR(t,e,n){var r=t._zr;return e<0||e>r.getWidth()||n<0||n>r.getHeight()}var mR={lineX:gR(0),lineY:gR(1),rect:{createCover:function(t,e){function n(t){return t}return UN({toRectRange:n,fromRectRange:n},t,e,[["w"],["e"],["n"],["s"],["s","e"],["s","w"],["n","e"],["n","w"]])},getCreatingRange:function(t){var e=$N(t);return KN(e[1][0],e[1][1],e[0][0],e[0][1])},updateCoverShape:function(t,e,n,r){qN(t,e,n,r)},updateCommon:XN,contain:uR},polygon:{createCover:function(t,e){var n=new wr;return n.add(new Im({name:"main",style:JN(e),silent:!0})),n},getCreatingRange:function(t){return t},endCreating:function(t,e){e.remove(e.childAt(0)),e.add(new Tm({name:"main",draggable:!0,drift:Object(s["h"])(rR,t,e),ondragend:Object(s["h"])(VN,t,{isEnd:!0})}))},updateCoverShape:function(t,e,n,r){e.childAt(0).setShape({points:oR(t,e,n)})},updateCommon:XN,contain:uR}};function gR(t){return{createCover:function(e,n){return UN({toRectRange:function(e){var n=[e,[0,100]];return t&&n.reverse(),n},fromRectRange:function(e){return e[t]}},e,n,[[["w"],["e"]],[["n"],["s"]]][t])},getCreatingRange:function(e){var n=$N(e),r=MN(n[0][t],n[1][t]),i=SN(n[0][t],n[1][t]);return[r,i]},updateCoverShape:function(e,n,r,i){var o,a=BN(e,n);if(a!==wN&&a.getLinearBrushOtherExtent)o=a.getLinearBrushOtherExtent(t);else{var s=e._zr;o=[0,[s.getWidth(),s.getHeight()][1-t]]}var l=[r,o];t&&l.reverse(),qN(e,n,l,i)},updateCommon:XN,contain:uR}}var yR=jN;function vR(t){return t=xR(t),function(e){return Dg(e,t)}}function _R(t,e){return t=xR(t),function(n){var r=null!=e?e:n,i=r?t.width:t.height,o=r?t.x:t.y;return[o,o+(i||0)]}}function bR(t,e,n){var r=xR(t);return function(t,i){return r.contain(i[0],i[1])&&!gI(t,e,n)}}function xR(t){return Tt.create(t)}var wR=["axisLine","axisTickLabel","axisName"],MR=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(e,n){t.prototype.init.apply(this,arguments),(this._brushController=new yR(n.getZr())).on("brush",s["c"](this._onBrush,this))},e.prototype.render=function(t,e,n,r){if(!SR(t,e,r)){this.axisModel=t,this.api=n,this.group.removeAll();var i=this._axisGroup;if(this._axisGroup=new wr,this.group.add(this._axisGroup),t.get("show")){var o=LR(t,e),a=o.coordinateSystem,l=t.getAreaSelectStyle(),u=l.width,c=t.axis.dim,d=a.getAxisLayout(c),h=s["m"]({strokeContainThreshold:u},d),f=new sO(t,h);s["k"](wR,f.add,f),this._axisGroup.add(f.getGroup()),this._refreshBrushController(h,l,t,o,u,n),Tg(i,this._axisGroup,t)}}},e.prototype._refreshBrushController=function(t,e,n,r,i,o){var a=n.axis.getExtent(),s=a[1]-a[0],l=Math.min(30,.1*Math.abs(s)),u=Tt.create({x:a[0],y:-i/2,width:s,height:i});u.x-=l,u.width+=2*l,this._brushController.mount({enableGlobalPan:!0,rotation:t.rotation,x:t.position[0],y:t.position[1]}).setPanels([{panelId:"pl",clipPath:vR(u),isTargetByCursor:bR(u,o,r),getLinearBrushOtherExtent:_R(u,0)}]).enableBrush({brushType:"lineX",brushStyle:e,removeOnClick:!0}).updateCovers(kR(n))},e.prototype._onBrush=function(t){var e=t.areas,n=this.axisModel,r=n.axis,i=s["H"](e,(function(t){return[r.coordToData(t.range[0],!0),r.coordToData(t.range[1],!0)]}));(!n.option.realtime===t.isEnd||t.removeOnClick)&&this.api.dispatchAction({type:"axisAreaSelect",parallelAxisId:n.id,intervals:i})},e.prototype.dispose=function(){this._brushController.dispose()},e.type="parallelAxis",e}(kp);function SR(t,e,n){return n&&"axisAreaSelect"===n.type&&e.findComponents({mainType:"parallelAxis",query:n})[0]===t}function kR(t){var e=t.axis;return s["H"](t.activeIntervals,(function(t){return{brushType:"lineX",panelId:"pl",range:[e.dataToCoord(t[0],!0),e.dataToCoord(t[1],!0)]}}))}function LR(t,e){return e.getComponent("parallel",t.get("parallelIndex"))}var TR=MR,DR={type:"axisAreaSelect",event:"axisAreaSelected"};function OR(t){t.registerAction(DR,(function(t,e){e.eachComponent({mainType:"parallelAxis",query:t},(function(e){e.axis.model.setActiveIntervals(t.intervals)}))})),t.registerAction("parallelAxisExpand",(function(t,e){e.eachComponent({mainType:"parallel",query:t},(function(e){e.setAxisExpand(t)}))}))}var IR={type:"value",areaSelectStyle:{width:20,borderWidth:1,borderColor:"rgba(160,197,232)",color:"rgba(160,197,232)",opacity:.3},realtime:!0,z:10};function CR(t){t.registerComponentView(JE),t.registerComponentModel(QE),t.registerCoordinateSystem("parallel",_N),t.registerPreprocessor(VE),t.registerComponentModel(xN),t.registerComponentView(TR),oT(t,"parallel",xN,IR),OR(t)}function AR(t){jb(CR),t.registerChartView(EE),t.registerSeriesModel(zE),t.registerVisual(t.PRIORITY.VISUAL.BRUSH,WE)}var YR=function(){function t(){this.x1=0,this.y1=0,this.x2=0,this.y2=0,this.cpx1=0,this.cpy1=0,this.cpx2=0,this.cpy2=0,this.extent=0}return t}(),jR=function(t){function e(e){return t.call(this,e)||this}return o(e,t),e.prototype.getDefaultShape=function(){return new YR},e.prototype.buildPath=function(t,e){var n=e.extent;t.moveTo(e.x1,e.y1),t.bezierCurveTo(e.cpx1,e.cpy1,e.cpx2,e.cpy2,e.x2,e.y2),"vertical"===e.orient?(t.lineTo(e.x2+n,e.y2),t.bezierCurveTo(e.cpx2+n,e.cpy2,e.cpx1+n,e.cpy1,e.x1+n,e.y1)):(t.lineTo(e.x2,e.y2+n),t.bezierCurveTo(e.cpx2,e.cpy2+n,e.cpx1,e.cpy1+n,e.x1,e.y1+n)),t.closePath()},e.prototype.highlight=function(){ul(this)},e.prototype.downplay=function(){cl(this)},e}(Va),PR=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n._focusAdjacencyDisabled=!1,n}return o(e,t),e.prototype.render=function(t,e,n){var r=this,i=t.getGraph(),o=this.group,a=t.layoutInfo,l=a.width,u=a.height,c=t.getData(),d=t.getData("edge"),h=t.get("orient");this._model=t,o.removeAll(),o.x=a.x,o.y=a.y,i.eachEdge((function(e){var n=new jR,r=Ls(n);r.dataIndex=e.dataIndex,r.seriesIndex=t.seriesIndex,r.dataType="edge";var i,a,c,f,p,m,g,y,v=e.getModel(),_=v.getModel("lineStyle"),b=_.get("curveness"),x=e.node1.getLayout(),w=e.node1.getModel(),M=w.get("localX"),S=w.get("localY"),k=e.node2.getLayout(),L=e.node2.getModel(),T=L.get("localX"),D=L.get("localY"),O=e.getLayout();n.shape.extent=Math.max(1,O.dy),n.shape.orient=h,"vertical"===h?(i=(null!=M?M*l:x.x)+O.sy,a=(null!=S?S*u:x.y)+x.dy,c=(null!=T?T*l:k.x)+O.ty,f=null!=D?D*u:k.y,p=i,m=a*(1-b)+f*b,g=c,y=a*b+f*(1-b)):(i=(null!=M?M*l:x.x)+x.dx,a=(null!=S?S*u:x.y)+O.sy,c=null!=T?T*l:k.x,f=(null!=D?D*u:k.y)+O.ty,p=i*(1-b)+c*b,m=a,g=i*b+c*(1-b),y=f),n.setShape({x1:i,y1:a,x2:c,y2:f,cpx1:p,cpy1:m,cpx2:g,cpy2:y}),n.useStyle(_.getItemStyle()),ER(n.style,h,e);var I=""+v.get("value"),C=eu(v,"edgeLabel");tu(n,C,{labelFetcher:{getFormattedLabel:function(e,n,r,i,o,a){return t.getFormattedLabel(e,n,"edge",i,Object(s["Q"])(o,C.normal&&C.normal.get("formatter"),I),a)}},labelDataIndex:e.dataIndex,defaultText:I}),n.setTextConfig({position:"inside"});var A=v.getModel("emphasis");Al(n,v,"lineStyle",(function(t){var n=t.getItemStyle();return ER(n,h,e),n})),o.add(n),d.setItemGraphicEl(e.dataIndex,n);var Y=A.get("focus");Dl(n,"adjacency"===Y?e.getAdjacentDataIndices():"trajectory"===Y?e.getTrajectoryDataIndices():Y,A.get("blurScope"),A.get("disabled"))})),i.eachNode((function(e){var n=e.getLayout(),r=e.getModel(),i=r.get("localX"),a=r.get("localY"),s=r.getModel("emphasis"),d=new ss({shape:{x:null!=i?i*l:n.x,y:null!=a?a*u:n.y,width:n.dx,height:n.dy},style:r.getModel("itemStyle").getItemStyle(),z2:10});tu(d,eu(r),{labelFetcher:{getFormattedLabel:function(e,n){return t.getFormattedLabel(e,n,"node")}},labelDataIndex:e.dataIndex,defaultText:e.id}),d.disableLabelAnimation=!0,d.setStyle("fill",e.getVisual("color")),d.setStyle("decal",e.getVisual("style").decal),Al(d,r),o.add(d),c.setItemGraphicEl(e.dataIndex,d),Ls(d).dataType="node";var h=s.get("focus");Dl(d,"adjacency"===h?e.getAdjacentDataIndices():"trajectory"===h?e.getTrajectoryDataIndices():h,s.get("blurScope"),s.get("disabled"))})),c.eachItemGraphicEl((function(e,i){var o=c.getItemModel(i);o.get("draggable")&&(e.drift=function(e,o){r._focusAdjacencyDisabled=!0,this.shape.x+=e,this.shape.y+=o,this.dirty(),n.dispatchAction({type:"dragNode",seriesId:t.id,dataIndex:c.getRawIndex(i),localX:this.shape.x/l,localY:this.shape.y/u})},e.ondragend=function(){r._focusAdjacencyDisabled=!1},e.draggable=!0,e.cursor="move")})),!this._data&&t.isAnimationEnabled()&&o.setClipPath(NR(o.getBoundingRect(),t,(function(){o.removeClipPath()}))),this._data=t.getData()},e.prototype.dispose=function(){},e.type="sankey",e}($g);function ER(t,e,n){switch(t.fill){case"source":t.fill=n.node1.getVisual("color"),t.decal=n.node1.getVisual("style").decal;break;case"target":t.fill=n.node2.getVisual("color"),t.decal=n.node2.getVisual("style").decal;break;case"gradient":var r=n.node1.getVisual("color"),i=n.node2.getVisual("color");Object(s["C"])(r)&&Object(s["C"])(i)&&(t.fill=new qm(0,0,+("horizontal"===e),+("vertical"===e),[{color:r,offset:0},{color:i,offset:1}]))}}function NR(t,e,n){var r=new ss({shape:{x:t.x-10,y:t.y-10,width:0,height:t.height+20}});return Vl(r,{shape:{width:t.width+20}},e,n),r}var RR=PR,HR=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.getInitialData=function(t,e){var n=t.edges||t.links,r=t.data||t.nodes,i=t.levels;this.levelModels=[];for(var o=this.levelModels,a=0;a<i.length;a++)null!=i[a].depth&&i[a].depth>=0&&(o[i[a].depth]=new Lu(i[a],this,e));if(r&&n){var s=nE(r,n,this,!0,l);return s.data}function l(t,e){t.wrapMethod("getItemModel",(function(t,e){var n=t.parentModel,r=n.getData().getItemLayout(e);if(r){var i=r.depth,o=n.levelModels[i];o&&(t.parentModel=o)}return t})),e.wrapMethod("getItemModel",(function(t,e){var n=t.parentModel,r=n.getGraph().getEdgeByIndex(e),i=r.node1.getLayout();if(i){var o=i.depth,a=n.levelModels[o];a&&(t.parentModel=a)}return t}))}},e.prototype.setNodePosition=function(t,e){var n=this.option.data||this.option.nodes,r=n[t];r.localX=e[0],r.localY=e[1]},e.prototype.getGraph=function(){return this.getData().graph},e.prototype.getEdgeData=function(){return this.getGraph().edgeData},e.prototype.formatTooltip=function(t,e,n){function r(t){return isNaN(t)||null==t}if("edge"===n){var i=this.getDataParams(t,n),o=i.data,a=i.value,s=o.source+" -- "+o.target;return Uf("nameValue",{name:s,value:a,noValue:r(a)})}var l=this.getGraph().getNodeByIndex(t),u=l.getLayout().value,c=this.getDataParams(t,n).data.name;return Uf("nameValue",{name:null!=c?c+"":null,value:u,noValue:r(u)})},e.prototype.optionUpdated=function(){},e.prototype.getDataParams=function(e,n){var r=t.prototype.getDataParams.call(this,e,n);if(null==r.value&&"node"===n){var i=this.getGraph().getNodeByIndex(e),o=i.getLayout().value;r.value=o}return r},e.type="series.sankey",e.defaultOption={z:2,coordinateSystem:"view",left:"5%",top:"5%",right:"20%",bottom:"5%",orient:"horizontal",nodeWidth:20,nodeGap:8,draggable:!0,layoutIterations:32,label:{show:!0,position:"right",fontSize:12},edgeLabel:{show:!1,fontSize:12},levels:[],nodeAlign:"justify",lineStyle:{color:"#314656",opacity:.2,curveness:.5},emphasis:{label:{show:!0},lineStyle:{opacity:.5}},select:{itemStyle:{borderColor:"#212121"}},animationEasing:"linear",animationDuration:1e3},e}(Mp),zR=HR;function FR(t,e){t.eachSeriesByType("sankey",(function(t){var n=t.get("nodeWidth"),r=t.get("nodeGap"),i=BR(t,e);t.layoutInfo=i;var o=i.width,a=i.height,l=t.getGraph(),u=l.nodes,c=l.edges;VR(u);var d=s["n"](u,(function(t){return 0===t.getLayout().value})),h=0!==d.length?0:t.get("layoutIterations"),f=t.get("orient"),p=t.get("nodeAlign");WR(u,c,n,r,o,a,h,f,p)}))}function BR(t,e){return Hc(t.getBoxLayoutParams(),{width:e.getWidth(),height:e.getHeight()})}function WR(t,e,n,r,i,o,a,s,l){GR(t,e,n,i,o,s,l),ZR(t,e,o,i,r,a,s),uH(t,s)}function VR(t){s["k"](t,(function(t){var e=sH(t.outEdges,aH),n=sH(t.inEdges,aH),r=t.getValue()||0,i=Math.max(e,n,r);t.setLayout({value:i},!0)}))}function GR(t,e,n,r,i,o,a){for(var s=[],l=[],u=[],c=[],d=0,h=0;h<e.length;h++)s[h]=1;for(h=0;h<t.length;h++)l[h]=t[h].inEdges.length,0===l[h]&&u.push(t[h]);var f=-1;while(u.length){for(var p=0;p<u.length;p++){var m=u[p],g=m.hostGraph.data.getRawDataItem(m.dataIndex),y=null!=g.depth&&g.depth>=0;y&&g.depth>f&&(f=g.depth),m.setLayout({depth:y?g.depth:d},!0),"vertical"===o?m.setLayout({dy:n},!0):m.setLayout({dx:n},!0);for(var v=0;v<m.outEdges.length;v++){var _=m.outEdges[v],b=e.indexOf(_);s[b]=0;var x=_.node2,w=t.indexOf(x);0===--l[w]&&c.indexOf(x)<0&&c.push(x)}}++d,u=c,c=[]}for(h=0;h<s.length;h++)if(1===s[h])throw new Error("Sankey is a DAG, the original data has cycle!");var M=f>d-1?f:d-1;a&&"left"!==a&&UR(t,a,o,M);var S="vertical"===o?(i-n)/M:(r-n)/M;XR(t,S,o)}function $R(t){var e=t.hostGraph.data.getRawDataItem(t.dataIndex);return null!=e.depth&&e.depth>=0}function UR(t,e,n,r){if("right"===e){var i=[],o=t,a=0;while(o.length){for(var l=0;l<o.length;l++){var u=o[l];u.setLayout({skNodeHeight:a},!0);for(var c=0;c<u.inEdges.length;c++){var d=u.inEdges[c];i.indexOf(d.node1)<0&&i.push(d.node1)}}o=i,i=[],++a}s["k"](t,(function(t){$R(t)||t.setLayout({depth:Math.max(0,r-t.getLayout().skNodeHeight)},!0)}))}else"justify"===e&&qR(t,r)}function qR(t,e){s["k"](t,(function(t){$R(t)||t.outEdges.length||t.setLayout({depth:e},!0)}))}function XR(t,e,n){s["k"](t,(function(t){var r=t.getLayout().depth*e;"vertical"===n?t.setLayout({y:r},!0):t.setLayout({x:r},!0)}))}function ZR(t,e,n,r,i,o,a){var s=JR(t,a);KR(s,e,n,r,i,a),QR(s,i,n,r,a);for(var l=1;o>0;o--)l*=.99,tH(s,l,a),QR(s,i,n,r,a),lH(s,l,a),QR(s,i,n,r,a)}function JR(t,e){var n=[],r="vertical"===e?"y":"x",i=Fi(t,(function(t){return t.getLayout()[r]}));return i.keys.sort((function(t,e){return t-e})),s["k"](i.keys,(function(t){n.push(i.buckets.get(t))})),n}function KR(t,e,n,r,i,o){var a=1/0;s["k"](t,(function(t){var e=t.length,l=0;s["k"](t,(function(t){l+=t.getLayout().value}));var u="vertical"===o?(r-(e-1)*i)/l:(n-(e-1)*i)/l;u<a&&(a=u)})),s["k"](t,(function(t){s["k"](t,(function(t,e){var n=t.getLayout().value*a;"vertical"===o?(t.setLayout({x:e},!0),t.setLayout({dx:n},!0)):(t.setLayout({y:e},!0),t.setLayout({dy:n},!0))}))})),s["k"](e,(function(t){var e=+t.getValue()*a;t.setLayout({dy:e},!0)}))}function QR(t,e,n,r,i){var o="vertical"===i?"x":"y";s["k"](t,(function(t){var a,s,l;t.sort((function(t,e){return t.getLayout()[o]-e.getLayout()[o]}));for(var u=0,c=t.length,d="vertical"===i?"dx":"dy",h=0;h<c;h++)s=t[h],l=u-s.getLayout()[o],l>0&&(a=s.getLayout()[o]+l,"vertical"===i?s.setLayout({x:a},!0):s.setLayout({y:a},!0)),u=s.getLayout()[o]+s.getLayout()[d]+e;var f="vertical"===i?r:n;if(l=u-e-f,l>0){a=s.getLayout()[o]-l,"vertical"===i?s.setLayout({x:a},!0):s.setLayout({y:a},!0),u=a;for(h=c-2;h>=0;--h)s=t[h],l=s.getLayout()[o]+s.getLayout()[d]+e-u,l>0&&(a=s.getLayout()[o]-l,"vertical"===i?s.setLayout({x:a},!0):s.setLayout({y:a},!0)),u=s.getLayout()[o]}}))}function tH(t,e,n){s["k"](t.slice().reverse(),(function(t){s["k"](t,(function(t){if(t.outEdges.length){var r=sH(t.outEdges,eH,n)/sH(t.outEdges,aH);if(isNaN(r)){var i=t.outEdges.length;r=i?sH(t.outEdges,nH,n)/i:0}if("vertical"===n){var o=t.getLayout().x+(r-oH(t,n))*e;t.setLayout({x:o},!0)}else{var a=t.getLayout().y+(r-oH(t,n))*e;t.setLayout({y:a},!0)}}}))}))}function eH(t,e){return oH(t.node2,e)*t.getValue()}function nH(t,e){return oH(t.node2,e)}function rH(t,e){return oH(t.node1,e)*t.getValue()}function iH(t,e){return oH(t.node1,e)}function oH(t,e){return"vertical"===e?t.getLayout().x+t.getLayout().dx/2:t.getLayout().y+t.getLayout().dy/2}function aH(t){return t.getValue()}function sH(t,e,n){var r=0,i=t.length,o=-1;while(++o<i){var a=+e(t[o],n);isNaN(a)||(r+=a)}return r}function lH(t,e,n){s["k"](t,(function(t){s["k"](t,(function(t){if(t.inEdges.length){var r=sH(t.inEdges,rH,n)/sH(t.inEdges,aH);if(isNaN(r)){var i=t.inEdges.length;r=i?sH(t.inEdges,iH,n)/i:0}if("vertical"===n){var o=t.getLayout().x+(r-oH(t,n))*e;t.setLayout({x:o},!0)}else{var a=t.getLayout().y+(r-oH(t,n))*e;t.setLayout({y:a},!0)}}}))}))}function uH(t,e){var n="vertical"===e?"x":"y";s["k"](t,(function(t){t.outEdges.sort((function(t,e){return t.node2.getLayout()[n]-e.node2.getLayout()[n]})),t.inEdges.sort((function(t,e){return t.node1.getLayout()[n]-e.node1.getLayout()[n]}))})),s["k"](t,(function(t){var e=0,n=0;s["k"](t.outEdges,(function(t){t.setLayout({sy:e},!0),e+=t.getLayout().dy})),s["k"](t.inEdges,(function(t){t.setLayout({ty:n},!0),n+=t.getLayout().dy}))}))}function cH(t){t.eachSeriesByType("sankey",(function(t){var e=t.getGraph(),n=e.nodes,r=e.edges;if(n.length){var i=1/0,o=-1/0;s["k"](n,(function(t){var e=t.getLayout().value;e<i&&(i=e),e>o&&(o=e)})),s["k"](n,(function(e){var n=new uj({type:"color",mappingMethod:"linear",dataExtent:[i,o],visual:t.get("color")}),r=n.mapValueToVisual(e.getLayout().value),a=e.getModel().get(["itemStyle","color"]);null!=a?(e.setVisual("color",a),e.setVisual("style",{fill:a})):(e.setVisual("color",r),e.setVisual("style",{fill:r}))}))}r.length&&s["k"](r,(function(t){var e=t.getModel().get("lineStyle");t.setVisual("style",e)}))}))}function dH(t){t.registerChartView(RR),t.registerSeriesModel(zR),t.registerLayout(FR),t.registerVisual(cH),t.registerAction({type:"dragNode",event:"dragnode",update:"update"},(function(t,e){e.eachComponent({mainType:"series",subType:"sankey",query:t},(function(e){e.setNodePosition(t.dataIndex,[t.localX,t.localY])}))}))}var hH=function(){function t(){}return t.prototype.getInitialData=function(t,e){var n,r,i=e.getComponent("xAxis",this.get("xAxisIndex")),o=e.getComponent("yAxis",this.get("yAxisIndex")),a=i.get("type"),l=o.get("type");"category"===a?(t.layout="horizontal",n=i.getOrdinalMeta(),r=!0):"category"===l?(t.layout="vertical",n=o.getOrdinalMeta(),r=!0):t.layout=t.layout||"horizontal";var u=["x","y"],c="horizontal"===t.layout?0:1,d=this._baseAxisDim=u[c],h=u[1-c],f=[i,o],p=f[c].get("type"),m=f[1-c].get("type"),g=t.data;if(g&&r){var y=[];s["k"](g,(function(t,e){var n;s["t"](t)?(n=t.slice(),t.unshift(e)):s["t"](t.value)?(n=s["m"]({},t),n.value=n.value.slice(),t.value.unshift(e)):n=t,y.push(n)})),t.data=y}var v=this.defaultValueDimensions,_=[{name:d,type:xM(p),ordinalMeta:n,otherDims:{tooltip:!1,itemName:0},dimsDef:["base"]},{name:h,type:xM(m),dimsDef:v.slice()}];return TL(this,{coordDimensions:_,dimensionsCount:v.length+1,encodeDefaulter:s["h"](dd,_,this)})},t.prototype.getBaseAxis=function(){var t=this._baseAxisDim;return this.ecModel.getComponent(t+"Axis",this.get(t+"AxisIndex")).axis},t}(),fH=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.defaultValueDimensions=[{name:"min",defaultTooltip:!0},{name:"Q1",defaultTooltip:!0},{name:"median",defaultTooltip:!0},{name:"Q3",defaultTooltip:!0},{name:"max",defaultTooltip:!0}],n.visualDrawType="stroke",n}return o(e,t),e.type="series.boxplot",e.dependencies=["xAxis","yAxis","grid"],e.defaultOption={z:2,coordinateSystem:"cartesian2d",legendHoverLink:!0,layout:null,boxWidth:[7,50],itemStyle:{color:"#fff",borderWidth:1},emphasis:{scale:!0,itemStyle:{borderWidth:2,shadowBlur:5,shadowOffsetX:1,shadowOffsetY:1,shadowColor:"rgba(0,0,0,0.2)"}},animationDuration:800},e}(Mp);Object(s["K"])(fH,hH,!0);var pH=fH,mH=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){var r=t.getData(),i=this.group,o=this._data;this._data||i.removeAll();var a="horizontal"===t.get("layout")?1:0;r.diff(o).add((function(t){if(r.hasValue(t)){var e=r.getItemLayout(t),n=vH(e,r,t,a,!0);r.setItemGraphicEl(t,n),i.add(n)}})).update((function(t,e){var n=o.getItemGraphicEl(e);if(r.hasValue(t)){var s=r.getItemLayout(t);n?(Xl(n),_H(s,n,r,t)):n=vH(s,r,t,a),i.add(n),r.setItemGraphicEl(t,n)}else i.remove(n)})).remove((function(t){var e=o.getItemGraphicEl(t);e&&i.remove(e)})).execute(),this._data=r},e.prototype.remove=function(t){var e=this.group,n=this._data;this._data=null,n&&n.eachItemGraphicEl((function(t){t&&e.remove(t)}))},e.type="boxplot",e}($g),gH=function(){function t(){}return t}(),yH=function(t){function e(e){var n=t.call(this,e)||this;return n.type="boxplotBoxPath",n}return o(e,t),e.prototype.getDefaultShape=function(){return new gH},e.prototype.buildPath=function(t,e){var n=e.points,r=0;for(t.moveTo(n[r][0],n[r][1]),r++;r<4;r++)t.lineTo(n[r][0],n[r][1]);for(t.closePath();r<n.length;r++)t.moveTo(n[r][0],n[r][1]),r++,t.lineTo(n[r][0],n[r][1])},e}(Va);function vH(t,e,n,r,i){var o=t.ends,a=new yH({shape:{points:i?bH(o,r,t):o}});return _H(t,a,e,n,i),a}function _H(t,e,n,i,o){var a=n.hostModel,s=r[o?"initProps":"updateProps"];s(e,{shape:{points:t.ends}},a,i),e.useStyle(n.getItemVisual(i,"style")),e.style.strokeNoScale=!0,e.z2=100;var l=n.getItemModel(i),u=l.getModel("emphasis");Al(e,l),Dl(e,u.get("focus"),u.get("blurScope"),u.get("disabled"))}function bH(t,e,n){return s["H"](t,(function(t){return t=t.slice(),t[e]=n.initBaseline,t}))}var xH=mH,wH=s["k"];function MH(t){var e=SH(t);wH(e,(function(t){var e=t.seriesModels;e.length&&(kH(t),wH(e,(function(e,n){LH(e,t.boxOffsetList[n],t.boxWidthList[n])})))}))}function SH(t){var e=[],n=[];return t.eachSeriesByType("boxplot",(function(t){var r=t.getBaseAxis(),i=s["r"](n,r);i<0&&(i=n.length,n[i]=r,e[i]={axis:r,seriesModels:[]}),e[i].seriesModels.push(t)})),e}function kH(t){var e,n=t.axis,r=t.seriesModels,i=r.length,o=t.boxWidthList=[],a=t.boxOffsetList=[],l=[];if("category"===n.type)e=n.getBandWidth();else{var u=0;wH(r,(function(t){u=Math.max(u,t.getData().count())}));var c=n.getExtent();e=Math.abs(c[1]-c[0])/u}wH(r,(function(t){var n=t.get("boxWidth");s["t"](n)||(n=[n,n]),l.push([Nr(n[0],e)||0,Nr(n[1],e)||0])}));var d=.8*e-2,h=d/i*.3,f=(d-h*(i-1))/i,p=f/2-d/2;wH(r,(function(t,e){a.push(p),p+=h+f,o.push(Math.min(Math.max(f,l[e][0]),l[e][1]))}))}function LH(t,e,n){var r=t.coordinateSystem,i=t.getData(),o=n/2,a="horizontal"===t.get("layout")?0:1,s=1-a,l=["x","y"],u=i.mapDimension(l[a]),c=i.mapDimensionsAll(l[s]);if(!(null==u||c.length<5))for(var d=0;d<i.count();d++){var h=i.get(u,d),f=_(h,c[2],d),p=_(h,c[0],d),m=_(h,c[1],d),g=_(h,c[3],d),y=_(h,c[4],d),v=[];b(v,m,!1),b(v,g,!0),v.push(p,m,y,g),x(v,p),x(v,y),x(v,f),i.setItemLayout(d,{initBaseline:f[s],ends:v})}function _(t,n,o){var l,u=i.get(n,o),c=[];return c[a]=t,c[s]=u,isNaN(t)||isNaN(u)?l=[NaN,NaN]:(l=r.dataToPoint(c),l[a]+=e),l}function b(t,e,n){var r=e.slice(),i=e.slice();r[a]+=o,i[a]-=o,n?t.push(r,i):t.push(i,r)}function x(t,e){var n=e.slice(),r=e.slice();n[a]-=o,r[a]+=o,t.push(n,r)}}function TH(t,e){e=e||{};for(var n=[],r=[],i=e.boundIQR,o="none"===i||0===i,a=0;a<t.length;a++){var l=Hr(t[a].slice()),u=Qr(l,.25),c=Qr(l,.5),d=Qr(l,.75),h=l[0],f=l[l.length-1],p=(null==i?1.5:i)*(d-u),m=o?h:Math.max(h,u-p),g=o?f:Math.min(f,d+p),y=e.itemNameFormatter,v=Object(s["w"])(y)?y({value:a}):Object(s["C"])(y)?y.replace("{value}",a+""):a+"";n.push([v,m,u,c,d,g]);for(var _=0;_<l.length;_++){var b=l[_];if(b<m||b>g){var x=[v,b];r.push(x)}}}return{boxData:n,outliers:r}}var DH={type:"echarts:boxplot",transform:function(t){var e=t.upstream;if(e.sourceFormat!==ed){var n="";0,Cd(n)}var r=TH(e.getRawData(),t.config);return[{dimensions:["ItemName","Low","Q1","Q2","Q3","High"],data:r.boxData},{data:r.outliers}]}};function OH(t){t.registerSeriesModel(pH),t.registerChartView(xH),t.registerLayout(MH),t.registerTransform(DH)}var IH=["color","borderColor"],CH=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){this.group.removeClipPath(),this._progressiveEls=null,this._updateDrawMode(t),this._isLargeDraw?this._renderLarge(t):this._renderNormal(t)},e.prototype.incrementalPrepareRender=function(t,e,n){this._clear(),this._updateDrawMode(t)},e.prototype.incrementalRender=function(t,e,n,r){this._progressiveEls=[],this._isLargeDraw?this._incrementalRenderLarge(t,e):this._incrementalRenderNormal(t,e)},e.prototype.eachRendered=function(t){Ng(this._progressiveEls||this.group,t)},e.prototype._updateDrawMode=function(t){var e=t.pipelineContext.large;null!=this._isLargeDraw&&e===this._isLargeDraw||(this._isLargeDraw=e,this._clear())},e.prototype._renderNormal=function(t){var e=t.getData(),n=this._data,r=this.group,i=e.getLayout("isSimpleBox"),o=t.get("clip",!0),a=t.coordinateSystem,s=a.getArea&&a.getArea();this._data||r.removeAll(),e.diff(n).add((function(n){if(e.hasValue(n)){var a=e.getItemLayout(n);if(o&&PH(s,a))return;var l=jH(a,n,!0);Vl(l,{shape:{points:a.ends}},t,n),EH(l,e,n,i),r.add(l),e.setItemGraphicEl(n,l)}})).update((function(a,l){var u=n.getItemGraphicEl(l);if(e.hasValue(a)){var c=e.getItemLayout(a);o&&PH(s,c)?r.remove(u):(u?(Wl(u,{shape:{points:c.ends}},t,a),Xl(u)):u=jH(c,a),EH(u,e,a,i),r.add(u),e.setItemGraphicEl(a,u))}else r.remove(u)})).remove((function(t){var e=n.getItemGraphicEl(t);e&&r.remove(e)})).execute(),this._data=e},e.prototype._renderLarge=function(t){this._clear(),zH(t,this.group);var e=t.get("clip",!0)?BS(t.coordinateSystem,!1,t):null;e?this.group.setClipPath(e):this.group.removeClipPath()},e.prototype._incrementalRenderNormal=function(t,e){var n,r=e.getData(),i=r.getLayout("isSimpleBox");while(null!=(n=t.next())){var o=r.getItemLayout(n),a=jH(o,n);EH(a,r,n,i),a.incremental=!0,this.group.add(a),this._progressiveEls.push(a)}},e.prototype._incrementalRenderLarge=function(t,e){zH(e,this.group,this._progressiveEls,!0)},e.prototype.remove=function(t){this._clear()},e.prototype._clear=function(){this.group.removeAll(),this._data=null},e.type="candlestick",e}($g),AH=function(){function t(){}return t}(),YH=function(t){function e(e){var n=t.call(this,e)||this;return n.type="normalCandlestickBox",n}return o(e,t),e.prototype.getDefaultShape=function(){return new AH},e.prototype.buildPath=function(t,e){var n=e.points;this.__simpleBox?(t.moveTo(n[4][0],n[4][1]),t.lineTo(n[6][0],n[6][1])):(t.moveTo(n[0][0],n[0][1]),t.lineTo(n[1][0],n[1][1]),t.lineTo(n[2][0],n[2][1]),t.lineTo(n[3][0],n[3][1]),t.closePath(),t.moveTo(n[4][0],n[4][1]),t.lineTo(n[5][0],n[5][1]),t.moveTo(n[6][0],n[6][1]),t.lineTo(n[7][0],n[7][1]))},e}(Va);function jH(t,e,n){var r=t.ends;return new YH({shape:{points:n?NH(r,t):r},z2:100})}function PH(t,e){for(var n=!0,r=0;r<e.ends.length;r++)if(t.contain(e.ends[r][0],e.ends[r][1])){n=!1;break}return n}function EH(t,e,n,r){var i=e.getItemModel(n);t.useStyle(e.getItemVisual(n,"style")),t.style.strokeNoScale=!0,t.__simpleBox=r,Al(t,i)}function NH(t,e){return s["H"](t,(function(t){return t=t.slice(),t[1]=e.initBaseline,t}))}var RH=function(){function t(){}return t}(),HH=function(t){function e(e){var n=t.call(this,e)||this;return n.type="largeCandlestickBox",n}return o(e,t),e.prototype.getDefaultShape=function(){return new RH},e.prototype.buildPath=function(t,e){for(var n=e.points,r=0;r<n.length;)if(this.__sign===n[r++]){var i=n[r++];t.moveTo(i,n[r++]),t.lineTo(i,n[r++])}else r+=3},e}(Va);function zH(t,e,n,r){var i=t.getData(),o=i.getLayout("largePoints"),a=new HH({shape:{points:o},__sign:1,ignoreCoarsePointer:!0});e.add(a);var s=new HH({shape:{points:o},__sign:-1,ignoreCoarsePointer:!0});e.add(s);var l=new HH({shape:{points:o},__sign:0,ignoreCoarsePointer:!0});e.add(l),FH(1,a,t,i),FH(-1,s,t,i),FH(0,l,t,i),r&&(a.incremental=!0,s.incremental=!0),n&&n.push(a,s)}function FH(t,e,n,r){var i=n.get(["itemStyle",t>0?"borderColor":"borderColor0"])||n.get(["itemStyle",t>0?"color":"color0"]);0===t&&(i=n.get(["itemStyle","borderColorDoji"]));var o=n.getModel("itemStyle").getItemStyle(IH);e.useStyle(o),e.style.fill=null,e.style.stroke=i}var BH=CH,WH=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.defaultValueDimensions=[{name:"open",defaultTooltip:!0},{name:"close",defaultTooltip:!0},{name:"lowest",defaultTooltip:!0},{name:"highest",defaultTooltip:!0}],n}return o(e,t),e.prototype.getShadowDim=function(){return"open"},e.prototype.brushSelector=function(t,e,n){var r=e.getItemLayout(t);return r&&n.rect(r.brushRect)},e.type="series.candlestick",e.dependencies=["xAxis","yAxis","grid"],e.defaultOption={z:2,coordinateSystem:"cartesian2d",legendHoverLink:!0,layout:null,clip:!0,itemStyle:{color:"#eb5454",color0:"#47b262",borderColor:"#eb5454",borderColor0:"#47b262",borderColorDoji:null,borderWidth:1},emphasis:{scale:!0,itemStyle:{borderWidth:2}},barMaxWidth:null,barMinWidth:null,barWidth:null,large:!0,largeThreshold:600,progressive:3e3,progressiveThreshold:1e4,progressiveChunkMode:"mod",animationEasing:"linear",animationDuration:300},e}(Mp);Object(s["K"])(WH,hH,!0);var VH=WH;function GH(t){t&&s["t"](t.series)&&s["k"](t.series,(function(t){s["A"](t)&&"k"===t.type&&(t.type="candlestick")}))}var $H=["itemStyle","borderColor"],UH=["itemStyle","borderColor0"],qH=["itemStyle","borderColorDoji"],XH=["itemStyle","color"],ZH=["itemStyle","color0"],JH={seriesType:"candlestick",plan:Lp(),performRawSeries:!0,reset:function(t,e){function n(t,e){return e.get(t>0?XH:ZH)}function r(t,e){return e.get(0===t?qH:t>0?$H:UH)}if(!e.isSeriesFiltered(t)){var i=t.pipelineContext.large;return!i&&{progress:function(t,e){var i;while(null!=(i=t.next())){var o=e.getItemModel(i),a=e.getItemLayout(i).sign,l=o.getItemStyle();l.fill=n(a,o),l.stroke=r(a,o)||l.fill;var u=e.ensureUniqueItemVisual(i,"style");Object(s["m"])(u,l)}}}}}},KH=JH,QH={seriesType:"candlestick",plan:Lp(),reset:function(t){var e=t.coordinateSystem,n=t.getData(),r=ez(t,n),i=0,o=1,a=["x","y"],l=n.getDimensionIndex(n.mapDimension(a[i])),u=Object(s["H"])(n.mapDimensionsAll(a[o]),n.getDimensionIndex,n),c=u[0],d=u[1],h=u[2],f=u[3];if(n.setLayout({candleWidth:r,isSimpleBox:r<=1.3}),!(l<0||u.length<4))return{progress:t.pipelineContext.large?m:p};function p(t,n){var a,s=n.getStore();while(null!=(a=t.next())){var u=s.get(l,a),p=s.get(c,a),m=s.get(d,a),g=s.get(h,a),y=s.get(f,a),v=Math.min(p,m),_=Math.max(p,m),b=T(v,u),x=T(_,u),w=T(g,u),M=T(y,u),S=[];D(S,x,0),D(S,b,1),S.push(I(M),I(x),I(w),I(b));var k=n.getItemModel(a),L=!!k.get(["itemStyle","borderColorDoji"]);n.setItemLayout(a,{sign:tz(s,a,p,m,d,L),initBaseline:p>m?x[o]:b[o],ends:S,brushRect:O(g,y,u)})}function T(t,n){var r=[];return r[i]=n,r[o]=t,isNaN(n)||isNaN(t)?[NaN,NaN]:e.dataToPoint(r)}function D(t,e,n){var o=e.slice(),a=e.slice();o[i]=xg(o[i]+r/2,1,!1),a[i]=xg(a[i]-r/2,1,!0),n?t.push(o,a):t.push(a,o)}function O(t,e,n){var a=T(t,n),s=T(e,n);return a[i]-=r/2,s[i]-=r/2,{x:a[0],y:a[1],width:o?r:s[0]-a[0],height:o?s[1]-a[1]:r}}function I(t){return t[i]=xg(t[i],1),t}}function m(n,r){var a,s,u=OS(4*n.count),p=0,m=[],g=[],y=r.getStore(),v=!!t.get(["itemStyle","borderColorDoji"]);while(null!=(s=n.next())){var _=y.get(l,s),b=y.get(c,s),x=y.get(d,s),w=y.get(h,s),M=y.get(f,s);isNaN(_)||isNaN(w)||isNaN(M)?(u[p++]=NaN,p+=3):(u[p++]=tz(y,s,b,x,d,v),m[i]=_,m[o]=w,a=e.dataToPoint(m,null,g),u[p++]=a?a[0]:NaN,u[p++]=a?a[1]:NaN,m[o]=M,a=e.dataToPoint(m,null,g),u[p++]=a?a[1]:NaN)}r.setLayout("largePoints",u)}}};function tz(t,e,n,r,i,o){var a;return a=n>r?-1:n<r?1:o?0:e>0?t.get(i,e-1)<=r?1:-1:1,a}function ez(t,e){var n,r=t.getBaseAxis(),i="category"===r.type?r.getBandWidth():(n=r.getExtent(),Math.abs(n[1]-n[0])/e.count()),o=Nr(Object(s["P"])(t.get("barMaxWidth"),i),i),a=Nr(Object(s["P"])(t.get("barMinWidth"),1),i),l=t.get("barWidth");return null!=l?Nr(l,i):Math.max(Math.min(i/2,o),a)}var nz=QH;function rz(t){t.registerChartView(BH),t.registerSeriesModel(VH),t.registerPreprocessor(GH),t.registerVisual(KH),t.registerLayout(nz)}function iz(t,e){var n=e.rippleEffectColor||e.color;t.eachChild((function(t){t.attr({z:e.z,zlevel:e.zlevel,style:{stroke:"stroke"===e.brushType?n:null,fill:"fill"===e.brushType?n:null}})}))}var oz=function(t){function e(e,n){var r=t.call(this)||this,i=new vS(e,n),o=new wr;return r.add(i),r.add(o),r.updateData(e,n),r}return o(e,t),e.prototype.stopEffectAnimation=function(){this.childAt(1).removeAll()},e.prototype.startEffectAnimation=function(t){for(var e=t.symbolType,n=t.color,r=t.rippleNumber,i=this.childAt(1),o=0;o<r;o++){var a=ov(e,-1,-1,2,2,n);a.attr({style:{strokeNoScale:!0},z2:99,silent:!0,scaleX:.5,scaleY:.5});var s=-o/r*t.period+t.effectOffset;a.animate("",!0).when(t.period,{scaleX:t.rippleScale/2,scaleY:t.rippleScale/2}).delay(s).start(),a.animateStyle(!0).when(t.period,{opacity:0}).delay(s).start(),i.add(a)}iz(i,t)},e.prototype.updateEffectAnimation=function(t){for(var e=this._effectCfg,n=this.childAt(1),r=["symbolType","period","rippleScale","rippleNumber"],i=0;i<r.length;i++){var o=r[i];if(e[o]!==t[o])return this.stopEffectAnimation(),void this.startEffectAnimation(t)}iz(n,t)},e.prototype.highlight=function(){ul(this)},e.prototype.downplay=function(){cl(this)},e.prototype.getSymbolType=function(){var t=this.childAt(0);return t&&t.getSymbolType()},e.prototype.updateData=function(t,e){var n=this,r=t.hostModel;this.childAt(0).updateData(t,e);var i=this.childAt(1),o=t.getItemModel(e),a=t.getItemVisual(e,"symbol"),s=av(t.getItemVisual(e,"symbolSize")),l=t.getItemVisual(e,"style"),u=l&&l.fill,c=o.getModel("emphasis");i.setScale(s),i.traverse((function(t){t.setStyle("fill",u)}));var d=sv(t.getItemVisual(e,"symbolOffset"),s);d&&(i.x=d[0],i.y=d[1]);var h=t.getItemVisual(e,"symbolRotate");i.rotation=(h||0)*Math.PI/180||0;var f={};f.showEffectOn=r.get("showEffectOn"),f.rippleScale=o.get(["rippleEffect","scale"]),f.brushType=o.get(["rippleEffect","brushType"]),f.period=1e3*o.get(["rippleEffect","period"]),f.effectOffset=e/t.count(),f.z=r.getShallow("z")||0,f.zlevel=r.getShallow("zlevel")||0,f.symbolType=a,f.color=u,f.rippleEffectColor=o.get(["rippleEffect","color"]),f.rippleNumber=o.get(["rippleEffect","number"]),"render"===f.showEffectOn?(this._effectCfg?this.updateEffectAnimation(f):this.startEffectAnimation(f),this._effectCfg=f):(this._effectCfg=null,this.stopEffectAnimation(),this.onHoverStateChange=function(t){"emphasis"===t?"render"!==f.showEffectOn&&n.startEffectAnimation(f):"normal"===t&&"render"!==f.showEffectOn&&n.stopEffectAnimation()}),this._effectCfg=f,Dl(this,c.get("focus"),c.get("blurScope"),c.get("disabled"))},e.prototype.fadeOut=function(t){t&&t()},e}(wr),az=oz,sz=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(){this._symbolDraw=new MS(az)},e.prototype.render=function(t,e,n){var r=t.getData(),i=this._symbolDraw;i.updateData(r,{clipShape:this._getClipShape(t)}),this.group.add(i.group)},e.prototype._getClipShape=function(t){var e=t.coordinateSystem,n=e&&e.getArea&&e.getArea();return t.get("clip",!0)?n:null},e.prototype.updateTransform=function(t,e,n){var r=t.getData();this.group.dirty();var i=uk("").reset(t,e,n);i.progress&&i.progress({start:0,end:r.count(),count:r.count()},r),this._symbolDraw.updateLayout()},e.prototype._updateGroupTransform=function(t){var e=t.coordinateSystem;e&&e.getRoamTransform&&(this.group.transform=mt(e.getRoamTransform()),this.group.decomposeTransform())},e.prototype.remove=function(t,e){this._symbolDraw&&this._symbolDraw.remove(!0)},e.type="effectScatter",e}($g),lz=sz,uz=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.hasSymbolVisual=!0,n}return o(e,t),e.prototype.getInitialData=function(t,e){return dS(null,this,{useEncodeDefaulter:!0})},e.prototype.brushSelector=function(t,e,n){return n.point(e.getItemLayout(t))},e.type="series.effectScatter",e.dependencies=["grid","polar"],e.defaultOption={coordinateSystem:"cartesian2d",z:2,legendHoverLink:!0,effectType:"ripple",progressive:0,showEffectOn:"render",clip:!0,rippleEffect:{period:4,scale:2.5,brushType:"fill",number:3},universalTransition:{divideShape:"clone"},symbolSize:10},e}(Mp),cz=uz;function dz(t){t.registerChartView(lz),t.registerSeriesModel(cz),t.registerLayout(uk("effectScatter"))}var hz=function(t){function e(e,n,r){var i=t.call(this)||this;return i.add(i.createLine(e,n,r)),i._updateEffectSymbol(e,n),i}return o(e,t),e.prototype.createLine=function(t,e,n){return new AP(t,e,n)},e.prototype._updateEffectSymbol=function(t,e){var n=t.getItemModel(e),r=n.getModel("effect"),i=r.get("symbolSize"),o=r.get("symbol");s["t"](i)||(i=[i,i]);var a=t.getItemVisual(e,"style"),l=r.get("color")||a&&a.stroke,u=this.childAt(1);this._symbolType!==o&&(this.remove(u),u=ov(o,-.5,-.5,1,1,l),u.z2=100,u.culling=!0,this.add(u)),u&&(u.setStyle("shadowColor",l),u.setStyle(r.getItemStyle(["color"])),u.scaleX=i[0],u.scaleY=i[1],u.setColor(l),this._symbolType=o,this._symbolScale=i,this._updateEffectAnimation(t,r,e))},e.prototype._updateEffectAnimation=function(t,e,n){var r=this.childAt(1);if(r){var i=t.getItemLayout(n),o=1e3*e.get("period"),a=e.get("loop"),l=e.get("roundTrip"),u=e.get("constantSpeed"),c=s["O"](e.get("delay"),(function(e){return e/t.count()*o/3}));if(r.ignore=!0,this._updateAnimationPoints(r,i),u>0&&(o=this._getLineLength(r)/u*1e3),o!==this._period||a!==this._loop||l!==this._roundTrip){r.stopAnimation();var d=void 0;d=s["w"](c)?c(n):c,r.__t>0&&(d=-o*r.__t),this._animateSymbol(r,o,d,a,l)}this._period=o,this._loop=a,this._roundTrip=l}},e.prototype._animateSymbol=function(t,e,n,r,i){if(e>0){t.__t=0;var o=this,a=t.animate("",r).when(i?2*e:e,{__t:i?2:1}).delay(n).during((function(){o._updateSymbolPosition(t)}));r||a.done((function(){o.remove(t)})),a.start()}},e.prototype._getLineLength=function(t){return b(t.__p1,t.__cp1)+b(t.__cp1,t.__p2)},e.prototype._updateAnimationPoints=function(t,e){t.__p1=e[0],t.__p2=e[1],t.__cp1=e[2]||[(e[0][0]+e[1][0])/2,(e[0][1]+e[1][1])/2]},e.prototype.updateData=function(t,e,n){this.childAt(0).updateData(t,e,n),this._updateEffectSymbol(t,e)},e.prototype._updateSymbolPosition=function(t){var e=t.__p1,n=t.__p2,r=t.__cp1,i=t.__t<1?t.__t:2-t.__t,o=[t.x,t.y],a=o.slice(),s=Le,l=Te;o[0]=s(e[0],r[0],n[0],i),o[1]=s(e[1],r[1],n[1],i);var u=t.__t<1?l(e[0],r[0],n[0],i):l(n[0],r[0],e[0],1-i),c=t.__t<1?l(e[1],r[1],n[1],i):l(n[1],r[1],e[1],1-i);t.rotation=-Math.atan2(c,u)-Math.PI/2,"line"!==this._symbolType&&"rect"!==this._symbolType&&"roundRect"!==this._symbolType||(void 0!==t.__lastT&&t.__lastT<t.__t?(t.scaleY=1.05*b(a,o),1===i&&(o[0]=a[0]+(o[0]-a[0])/2,o[1]=a[1]+(o[1]-a[1])/2)):1===t.__lastT?t.scaleY=2*b(e,o):t.scaleY=this._symbolScale[1]),t.__lastT=t.__t,t.ignore=!1,t.x=o[0],t.y=o[1]},e.prototype.updateLayout=function(t,e){this.childAt(0).updateLayout(t,e);var n=t.getItemModel(e).getModel("effect");this._updateEffectAnimation(t,n,e)},e}(wr),fz=hz,pz=function(t){function e(e,n,r){var i=t.call(this)||this;return i._createPolyline(e,n,r),i}return o(e,t),e.prototype._createPolyline=function(t,e,n){var r=t.getItemLayout(e),i=new Im({shape:{points:r}});this.add(i),this._updateCommonStl(t,e,n)},e.prototype.updateData=function(t,e,n){var r=t.hostModel,i=this.childAt(0),o={shape:{points:t.getItemLayout(e)}};Wl(i,o,r,e),this._updateCommonStl(t,e,n)},e.prototype._updateCommonStl=function(t,e,n){var r=this.childAt(0),i=t.getItemModel(e),o=n&&n.emphasisLineStyle,a=n&&n.focus,s=n&&n.blurScope,l=n&&n.emphasisDisabled;if(!n||t.hasItemOption){var u=i.getModel("emphasis");o=u.getModel("lineStyle").getLineStyle(),l=u.get("disabled"),a=u.get("focus"),s=u.get("blurScope")}r.useStyle(t.getItemVisual(e,"style")),r.style.fill=null,r.style.strokeNoScale=!0;var c=r.ensureState("emphasis");c.style=o,Dl(this,a,s,l)},e.prototype.updateLayout=function(t,e){var n=this.childAt(0);n.setShape("points",t.getItemLayout(e))},e}(wr),mz=pz,gz=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e._lastFrame=0,e._lastFramePercent=0,e}return o(e,t),e.prototype.createLine=function(t,e,n){return new mz(t,e,n)},e.prototype._updateAnimationPoints=function(t,e){this._points=e;for(var n=[0],r=0,i=1;i<e.length;i++){var o=e[i-1],a=e[i];r+=b(o,a),n.push(r)}if(0!==r){for(i=0;i<n.length;i++)n[i]/=r;this._offsets=n,this._length=r}else this._length=0},e.prototype._getLineLength=function(){return this._length},e.prototype._updateSymbolPosition=function(t){var e=t.__t<1?t.__t:2-t.__t,n=this._points,r=this._offsets,i=n.length;if(r){var o,a=this._lastFrame;if(e<this._lastFramePercent){var s=Math.min(a+1,i-1);for(o=s;o>=0;o--)if(r[o]<=e)break;o=Math.min(o,i-2)}else{for(o=a;o<i;o++)if(r[o]>e)break;o=Math.min(o-1,i-2)}var l=(e-r[o])/(r[o+1]-r[o]),u=n[o],c=n[o+1];t.x=u[0]*(1-l)+l*c[0],t.y=u[1]*(1-l)+l*c[1];var d=t.__t<1?c[0]-u[0]:u[0]-c[0],h=t.__t<1?c[1]-u[1]:u[1]-c[1];t.rotation=-Math.atan2(h,d)-Math.PI/2,this._lastFrame=o,this._lastFramePercent=e,t.ignore=!1}},e}(fz),yz=gz,vz=function(){function t(){this.polyline=!1,this.curveness=0,this.segs=[]}return t}(),_z=function(t){function e(e){var n=t.call(this,e)||this;return n._off=0,n.hoverDataIdx=-1,n}return o(e,t),e.prototype.reset=function(){this.notClear=!1,this._off=0},e.prototype.getDefaultStyle=function(){return{stroke:"#000",fill:null}},e.prototype.getDefaultShape=function(){return new vz},e.prototype.buildPath=function(t,e){var n,r=e.segs,i=e.curveness;if(e.polyline)for(n=this._off;n<r.length;){var o=r[n++];if(o>0){t.moveTo(r[n++],r[n++]);for(var a=1;a<o;a++)t.lineTo(r[n++],r[n++])}}else for(n=this._off;n<r.length;){var s=r[n++],l=r[n++],u=r[n++],c=r[n++];if(t.moveTo(s,l),i>0){var d=(s+u)/2-(l-c)*i,h=(l+c)/2-(u-s)*i;t.quadraticCurveTo(d,h,u,c)}else t.lineTo(u,c)}this.incremental&&(this._off=n,this.notClear=!0)},e.prototype.findDataIndex=function(t,e){var n=this.shape,r=n.segs,i=n.curveness,o=this.style.lineWidth;if(n.polyline)for(var a=0,s=0;s<r.length;){var l=r[s++];if(l>0)for(var u=r[s++],c=r[s++],d=1;d<l;d++){var h=r[s++],f=r[s++];if(_a(u,c,h,f,o,t,e))return a}a++}else for(a=0,s=0;s<r.length;){u=r[s++],c=r[s++],h=r[s++],f=r[s++];if(i>0){var p=(u+h)/2-(c-f)*i,m=(c+f)/2-(h-u)*i;if(xa(u,c,p,m,h,f,o,t,e))return a}else if(_a(u,c,h,f,o,t,e))return a;a++}return-1},e.prototype.contain=function(t,e){var n=this.transformCoordToLocal(t,e),r=this.getBoundingRect();if(t=n[0],e=n[1],r.contain(t,e)){var i=this.hoverDataIdx=this.findDataIndex(t,e);return i>=0}return this.hoverDataIdx=-1,!1},e.prototype.getBoundingRect=function(){var t=this._rect;if(!t){for(var e=this.shape,n=e.segs,r=1/0,i=1/0,o=-1/0,a=-1/0,s=0;s<n.length;){var l=n[s++],u=n[s++];r=Math.min(l,r),o=Math.max(l,o),i=Math.min(u,i),a=Math.max(u,a)}t=this._rect=new Tt(r,i,o,a)}return t},e}(Va),bz=function(){function t(){this.group=new wr}return t.prototype.updateData=function(t){this._clear();var e=this._create();e.setShape({segs:t.getLayout("linesPoints")}),this._setCommon(e,t)},t.prototype.incrementalPrepareUpdate=function(t){this.group.removeAll(),this._clear()},t.prototype.incrementalUpdate=function(t,e){var n=this._newAdded[0],r=e.getLayout("linesPoints"),i=n&&n.shape.segs;if(i&&i.length<2e4){var o=i.length,a=new Float32Array(o+r.length);a.set(i),a.set(r,o),n.setShape({segs:a})}else{this._newAdded=[];var s=this._create();s.incremental=!0,s.setShape({segs:r}),this._setCommon(s,e),s.__startIndex=t.start}},t.prototype.remove=function(){this._clear()},t.prototype.eachRendered=function(t){this._newAdded[0]&&t(this._newAdded[0])},t.prototype._create=function(){var t=new _z({cursor:"default",ignoreCoarsePointer:!0});return this._newAdded.push(t),this.group.add(t),t},t.prototype._setCommon=function(t,e,n){var r=e.hostModel;t.setShape({polyline:r.get("polyline"),curveness:r.get(["lineStyle","curveness"])}),t.useStyle(r.getModel("lineStyle").getLineStyle()),t.style.strokeNoScale=!0;var i=e.getVisual("style");i&&i.stroke&&t.setStyle("stroke",i.stroke),t.setStyle("fill",null);var o=Ls(t);o.seriesIndex=r.seriesIndex,t.on("mousemove",(function(e){o.dataIndex=null;var n=t.hoverDataIdx;n>0&&(o.dataIndex=n+t.__startIndex)}))},t.prototype._clear=function(){this._newAdded=[],this.group.removeAll()},t}(),xz=bz,wz={seriesType:"lines",plan:Lp(),reset:function(t){var e=t.coordinateSystem;if(e){var n=t.get("polyline"),r=t.pipelineContext.large;return{progress:function(i,o){var a=[];if(r){var s=void 0,l=i.end-i.start;if(n){for(var u=0,c=i.start;c<i.end;c++)u+=t.getLineCoordsCount(c);s=new Float32Array(l+2*u)}else s=new Float32Array(4*l);var d=0,h=[];for(c=i.start;c<i.end;c++){var f=t.getLineCoords(c,a);n&&(s[d++]=f);for(var p=0;p<f;p++)h=e.dataToPoint(a[p],!1,h),s[d++]=h[0],s[d++]=h[1]}o.setLayout("linesPoints",s)}else for(c=i.start;c<i.end;c++){var m=o.getItemModel(c),g=(f=t.getLineCoords(c,a),[]);if(n)for(var y=0;y<f;y++)g.push(e.dataToPoint(a[y]));else{g[0]=e.dataToPoint(a[0]),g[1]=e.dataToPoint(a[1]);var v=m.get(["lineStyle","curveness"]);+v&&(g[2]=[(g[0][0]+g[1][0])/2-(g[0][1]-g[1][1])*v,(g[0][1]+g[1][1])/2-(g[1][0]-g[0][0])*v])}o.setItemLayout(c,g)}}}}}},Mz=wz,Sz=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){var r=t.getData(),i=this._updateLineDraw(r,t),o=t.get("zlevel"),a=t.get(["effect","trailLength"]),s=n.getZr(),l="svg"===s.painter.getType();l||s.painter.getLayer(o).clear(!0),null==this._lastZlevel||l||s.configLayer(this._lastZlevel,{motionBlur:!1}),this._showEffect(t)&&a>0&&(l||s.configLayer(o,{motionBlur:!0,lastFrameAlpha:Math.max(Math.min(a/10+.9,1),0)})),i.updateData(r);var u=t.get("clip",!0)&&BS(t.coordinateSystem,!1,t);u?this.group.setClipPath(u):this.group.removeClipPath(),this._lastZlevel=o,this._finished=!0},e.prototype.incrementalPrepareRender=function(t,e,n){var r=t.getData(),i=this._updateLineDraw(r,t);i.incrementalPrepareUpdate(r),this._clearLayer(n),this._finished=!1},e.prototype.incrementalRender=function(t,e,n){this._lineDraw.incrementalUpdate(t,e.getData()),this._finished=t.end===e.getData().count()},e.prototype.eachRendered=function(t){this._lineDraw&&this._lineDraw.eachRendered(t)},e.prototype.updateTransform=function(t,e,n){var r=t.getData(),i=t.pipelineContext;if(!this._finished||i.large||i.progressiveRender)return{update:!0};var o=Mz.reset(t,e,n);o.progress&&o.progress({start:0,end:r.count(),count:r.count()},r),this._lineDraw.updateLayout(),this._clearLayer(n)},e.prototype._updateLineDraw=function(t,e){var n=this._lineDraw,r=this._showEffect(e),i=!!e.get("polyline"),o=e.pipelineContext,a=o.large;return n&&r===this._hasEffet&&i===this._isPolyline&&a===this._isLargeDraw||(n&&n.remove(),n=this._lineDraw=a?new xz:new RP(i?r?yz:mz:r?fz:AP),this._hasEffet=r,this._isPolyline=i,this._isLargeDraw=a),this.group.add(n.group),n},e.prototype._showEffect=function(t){return!!t.get(["effect","show"])},e.prototype._clearLayer=function(t){var e=t.getZr(),n="svg"===e.painter.getType();n||null==this._lastZlevel||e.painter.getLayer(this._lastZlevel).clear(!0)},e.prototype.remove=function(t,e){this._lineDraw&&this._lineDraw.remove(),this._lineDraw=null,this._clearLayer(e)},e.prototype.dispose=function(t,e){this.remove(t,e)},e.type="lines",e}($g),kz=Sz,Lz="undefined"===typeof Uint32Array?Array:Uint32Array,Tz="undefined"===typeof Float64Array?Array:Float64Array;function Dz(t){var e=t.data;e&&e[0]&&e[0][0]&&e[0][0].coord&&(t.data=Object(s["H"])(e,(function(t){var e=[t[0].coord,t[1].coord],n={coords:e};return t[0].name&&(n.fromName=t[0].name),t[1].name&&(n.toName=t[1].name),Object(s["J"])([n,t[0],t[1]])})))}var Oz=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.visualStyleAccessPath="lineStyle",n.visualDrawType="stroke",n}return o(e,t),e.prototype.init=function(e){e.data=e.data||[],Dz(e);var n=this._processFlatCoordsArray(e.data);this._flatCoords=n.flatCoords,this._flatCoordsOffset=n.flatCoordsOffset,n.flatCoords&&(e.data=new Float32Array(n.count)),t.prototype.init.apply(this,arguments)},e.prototype.mergeOption=function(e){if(Dz(e),e.data){var n=this._processFlatCoordsArray(e.data);this._flatCoords=n.flatCoords,this._flatCoordsOffset=n.flatCoordsOffset,n.flatCoords&&(e.data=new Float32Array(n.count))}t.prototype.mergeOption.apply(this,arguments)},e.prototype.appendData=function(t){var e=this._processFlatCoordsArray(t.data);e.flatCoords&&(this._flatCoords?(this._flatCoords=Object(s["e"])(this._flatCoords,e.flatCoords),this._flatCoordsOffset=Object(s["e"])(this._flatCoordsOffset,e.flatCoordsOffset)):(this._flatCoords=e.flatCoords,this._flatCoordsOffset=e.flatCoordsOffset),t.data=new Float32Array(e.count)),this.getRawData().appendData(t.data)},e.prototype._getCoordsFromItemModel=function(t){var e=this.getData().getItemModel(t),n=e.option instanceof Array?e.option:e.getShallow("coords");return n},e.prototype.getLineCoordsCount=function(t){return this._flatCoordsOffset?this._flatCoordsOffset[2*t+1]:this._getCoordsFromItemModel(t).length},e.prototype.getLineCoords=function(t,e){if(this._flatCoordsOffset){for(var n=this._flatCoordsOffset[2*t],r=this._flatCoordsOffset[2*t+1],i=0;i<r;i++)e[i]=e[i]||[],e[i][0]=this._flatCoords[n+2*i],e[i][1]=this._flatCoords[n+2*i+1];return r}var o=this._getCoordsFromItemModel(t);for(i=0;i<o.length;i++)e[i]=e[i]||[],e[i][0]=o[i][0],e[i][1]=o[i][1];return o.length},e.prototype._processFlatCoordsArray=function(t){var e=0;if(this._flatCoords&&(e=this._flatCoords.length),Object(s["z"])(t[0])){for(var n=t.length,r=new Lz(n),i=new Tz(n),o=0,a=0,l=0,u=0;u<n;){l++;var c=t[u++];r[a++]=o+e,r[a++]=c;for(var d=0;d<c;d++){var h=t[u++],f=t[u++];i[o++]=h,i[o++]=f}}return{flatCoordsOffset:new Uint32Array(r.buffer,0,a),flatCoords:i,count:l}}return{flatCoordsOffset:null,flatCoords:null,count:t.length}},e.prototype.getInitialData=function(t,e){var n=new UM(["value"],this);return n.hasItemOption=!1,n.initData(t.data,[],(function(t,e,r,i){if(t instanceof Array)return NaN;n.hasItemOption=!0;var o=t.value;return null!=o?o instanceof Array?o[i]:o:void 0})),n},e.prototype.formatTooltip=function(t,e,n){var r=this.getData(),i=r.getItemModel(t),o=i.get("name");if(o)return o;var a=i.get("fromName"),s=i.get("toName"),l=[];return null!=a&&l.push(a),null!=s&&l.push(s),Uf("nameValue",{name:l.join(" > ")})},e.prototype.preventIncremental=function(){return!!this.get(["effect","show"])},e.prototype.getProgressive=function(){var t=this.option.progressive;return null==t?this.option.large?1e4:this.get("progressive"):t},e.prototype.getProgressiveThreshold=function(){var t=this.option.progressiveThreshold;return null==t?this.option.large?2e4:this.get("progressiveThreshold"):t},e.prototype.getZLevelKey=function(){var t=this.getModel("effect"),e=t.get("trailLength");return this.getData().count()>this.getProgressiveThreshold()?this.id:t.get("show")&&e>0?e+"":""},e.type="series.lines",e.dependencies=["grid","polar","geo","calendar"],e.defaultOption={coordinateSystem:"geo",z:2,legendHoverLink:!0,xAxisIndex:0,yAxisIndex:0,symbol:["none","none"],symbolSize:[10,10],geoIndex:0,effect:{show:!1,period:4,constantSpeed:0,symbol:"circle",symbolSize:3,loop:!0,trailLength:.2},large:!1,largeThreshold:2e3,polyline:!1,clip:!0,label:{show:!1,position:"end"},lineStyle:{opacity:.5}},e}(Mp),Iz=Oz;function Cz(t){return t instanceof Array||(t=[t,t]),t}var Az={seriesType:"lines",reset:function(t){var e=Cz(t.get("symbol")),n=Cz(t.get("symbolSize")),r=t.getData();function i(t,e){var n=t.getItemModel(e),r=Cz(n.getShallow("symbol",!0)),i=Cz(n.getShallow("symbolSize",!0));r[0]&&t.setItemVisual(e,"fromSymbol",r[0]),r[1]&&t.setItemVisual(e,"toSymbol",r[1]),i[0]&&t.setItemVisual(e,"fromSymbolSize",i[0]),i[1]&&t.setItemVisual(e,"toSymbolSize",i[1])}return r.setVisual("fromSymbol",e&&e[0]),r.setVisual("toSymbol",e&&e[1]),r.setVisual("fromSymbolSize",n&&n[0]),r.setVisual("toSymbolSize",n&&n[1]),{dataEach:r.hasItemOption?i:null}}},Yz=Az;function jz(t){t.registerChartView(kz),t.registerSeriesModel(Iz),t.registerLayout(Mz),t.registerVisual(Yz)}var Pz=256,Ez=function(){function t(){this.blurSize=30,this.pointSize=20,this.maxOpacity=1,this.minOpacity=0,this._gradientPixels={inRange:null,outOfRange:null};var t=Jn["d"].createCanvas();this.canvas=t}return t.prototype.update=function(t,e,n,r,i,o){var a=this._getBrush(),s=this._getGradient(i,"inRange"),l=this._getGradient(i,"outOfRange"),u=this.pointSize+this.blurSize,c=this.canvas,d=c.getContext("2d"),h=t.length;c.width=e,c.height=n;for(var f=0;f<h;++f){var p=t[f],m=p[0],g=p[1],y=p[2],v=r(y);d.globalAlpha=v,d.drawImage(a,m-u,g-u)}if(!c.width||!c.height)return c;var _=d.getImageData(0,0,c.width,c.height),b=_.data,x=0,w=b.length,M=this.minOpacity,S=this.maxOpacity,k=S-M;while(x<w){v=b[x+3]/256;var L=4*Math.floor(v*(Pz-1));if(v>0){var T=o(v)?s:l;v>0&&(v=v*k+M),b[x++]=T[L],b[x++]=T[L+1],b[x++]=T[L+2],b[x++]=T[L+3]*v*256}else x+=4}return d.putImageData(_,0,0),c},t.prototype._getBrush=function(){var t=this._brushCanvas||(this._brushCanvas=Jn["d"].createCanvas()),e=this.pointSize+this.blurSize,n=2*e;t.width=n,t.height=n;var r=t.getContext("2d");return r.clearRect(0,0,n,n),r.shadowOffsetX=n,r.shadowBlur=this.blurSize,r.shadowColor="#000",r.beginPath(),r.arc(-e,e,this.pointSize,0,2*Math.PI,!0),r.closePath(),r.fill(),t},t.prototype._getGradient=function(t,e){for(var n=this._gradientPixels,r=n[e]||(n[e]=new Uint8ClampedArray(1024)),i=[0,0,0,0],o=0,a=0;a<256;a++)t[e](a/255,!0,i),r[o++]=i[0],r[o++]=i[1],r[o++]=i[2],r[o++]=i[3];return r},t}(),Nz=Ez;function Rz(t,e,n){var r=t[1]-t[0];e=s["H"](e,(function(e){return{interval:[(e.interval[0]-t[0])/r,(e.interval[1]-t[0])/r]}}));var i=e.length,o=0;return function(t){var r;for(r=o;r<i;r++){var a=e[r].interval;if(a[0]<=t&&t<=a[1]){o=r;break}}if(r===i)for(r=o-1;r>=0;r--){a=e[r].interval;if(a[0]<=t&&t<=a[1]){o=r;break}}return r>=0&&r<i&&n[r]}}function Hz(t,e){var n=t[1]-t[0];return e=[(e[0]-t[0])/n,(e[1]-t[0])/n],function(t){return t>=e[0]&&t<=e[1]}}function zz(t){var e=t.dimensions;return"lng"===e[0]&&"lat"===e[1]}var Fz=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){var r;e.eachComponent("visualMap",(function(e){e.eachTargetSeries((function(n){n===t&&(r=e)}))})),this._progressiveEls=null,this.group.removeAll();var i=t.coordinateSystem;"cartesian2d"===i.type||"calendar"===i.type?this._renderOnCartesianAndCalendar(t,n,0,t.getData().count()):zz(i)&&this._renderOnGeo(i,t,r,n)},e.prototype.incrementalPrepareRender=function(t,e,n){this.group.removeAll()},e.prototype.incrementalRender=function(t,e,n,r){var i=e.coordinateSystem;i&&(zz(i)?this.render(e,n,r):(this._progressiveEls=[],this._renderOnCartesianAndCalendar(e,r,t.start,t.end,!0)))},e.prototype.eachRendered=function(t){Ng(this._progressiveEls||this.group,t)},e.prototype._renderOnCartesianAndCalendar=function(t,e,n,r,i){var o,a,s,l,u=t.coordinateSystem,c=WS(u,"cartesian2d");if(c){var d=u.getAxis("x"),h=u.getAxis("y");0,o=d.getBandWidth()+.5,a=h.getBandWidth()+.5,s=d.scale.getExtent(),l=h.scale.getExtent()}for(var f=this.group,p=t.getData(),m=t.getModel(["emphasis","itemStyle"]).getItemStyle(),g=t.getModel(["blur","itemStyle"]).getItemStyle(),y=t.getModel(["select","itemStyle"]).getItemStyle(),v=t.get(["itemStyle","borderRadius"]),_=eu(t),b=t.getModel("emphasis"),x=b.get("focus"),w=b.get("blurScope"),M=b.get("disabled"),S=c?[p.mapDimension("x"),p.mapDimension("y"),p.mapDimension("value")]:[p.mapDimension("time"),p.mapDimension("value")],k=n;k<r;k++){var L=void 0,T=p.getItemVisual(k,"style");if(c){var D=p.get(S[0],k),O=p.get(S[1],k);if(isNaN(p.get(S[2],k))||isNaN(D)||isNaN(O)||D<s[0]||D>s[1]||O<l[0]||O>l[1])continue;var I=u.dataToPoint([D,O]);L=new ss({shape:{x:I[0]-o/2,y:I[1]-a/2,width:o,height:a},style:T})}else{if(isNaN(p.get(S[1],k)))continue;L=new ss({z2:1,shape:u.dataToRect([p.get(S[0],k)]).contentShape,style:T})}if(p.hasItemOption){var C=p.getItemModel(k),A=C.getModel("emphasis");m=A.getModel("itemStyle").getItemStyle(),g=C.getModel(["blur","itemStyle"]).getItemStyle(),y=C.getModel(["select","itemStyle"]).getItemStyle(),v=C.get(["itemStyle","borderRadius"]),x=A.get("focus"),w=A.get("blurScope"),M=A.get("disabled"),_=eu(C)}L.shape.r=v;var Y=t.getRawValue(k),j="-";Y&&null!=Y[2]&&(j=Y[2]+""),tu(L,_,{labelFetcher:t,labelDataIndex:k,defaultOpacity:T.opacity,defaultText:j}),L.ensureState("emphasis").style=m,L.ensureState("blur").style=g,L.ensureState("select").style=y,Dl(L,x,w,M),L.incremental=i,i&&(L.states.emphasis.hoverLayer=!0),f.add(L),p.setItemGraphicEl(k,L),this._progressiveEls&&this._progressiveEls.push(L)}},e.prototype._renderOnGeo=function(t,e,n,r){var i=n.targetVisuals.inRange,o=n.targetVisuals.outOfRange,a=e.getData(),s=this._hmLayer||this._hmLayer||new Nz;s.blurSize=e.get("blurSize"),s.pointSize=e.get("pointSize"),s.minOpacity=e.get("minOpacity"),s.maxOpacity=e.get("maxOpacity");var l=t.getViewRect().clone(),u=t.getRoamTransform();l.applyTransform(u);var c=Math.max(l.x,0),d=Math.max(l.y,0),h=Math.min(l.width+l.x,r.getWidth()),f=Math.min(l.height+l.y,r.getHeight()),p=h-c,m=f-d,g=[a.mapDimension("lng"),a.mapDimension("lat"),a.mapDimension("value")],y=a.mapArray(g,(function(e,n,r){var i=t.dataToPoint([e,n]);return i[0]-=c,i[1]-=d,i.push(r),i})),v=n.getExtent(),_="visualMap.continuous"===n.type?Hz(v,n.option.range):Rz(v,n.getPieceList(),n.option.selected);s.update(y,p,m,i.color.getNormalizer(),{inRange:i.color.getColorMapper(),outOfRange:o.color.getColorMapper()},_);var b=new Ka({style:{width:p,height:m,x:c,y:d,image:s.canvas},silent:!0});this.group.add(b)},e.type="heatmap",e}($g),Bz=Fz,Wz=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.getInitialData=function(t,e){return dS(null,this,{generateCoord:"value"})},e.prototype.preventIncremental=function(){var t=$d.get(this.get("coordinateSystem"));if(t&&t.dimensions)return"lng"===t.dimensions[0]&&"lat"===t.dimensions[1]},e.type="series.heatmap",e.dependencies=["grid","geo","calendar"],e.defaultOption={coordinateSystem:"cartesian2d",z:2,geoIndex:0,blurSize:30,pointSize:20,maxOpacity:1,minOpacity:0,select:{itemStyle:{borderColor:"#212121"}}},e}(Mp),Vz=Wz;function Gz(t){t.registerChartView(Bz),t.registerSeriesModel(Vz)}var $z=["itemStyle","borderWidth"],Uz=[{xy:"x",wh:"width",index:0,posDesc:["left","right"]},{xy:"y",wh:"height",index:1,posDesc:["top","bottom"]}],qz=new Kp,Xz=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){var r=this.group,i=t.getData(),o=this._data,a=t.coordinateSystem,s=a.getBaseAxis(),l=s.isHorizontal(),u=a.master.getRect(),c={ecSize:{width:n.getWidth(),height:n.getHeight()},seriesModel:t,coordSys:a,coordSysExtent:[[u.x,u.x+u.width],[u.y,u.y+u.height]],isHorizontal:l,valueDim:Uz[+l],categoryDim:Uz[1-+l]};i.diff(o).add((function(t){if(i.hasValue(t)){var e=sF(i,t),n=Zz(i,t,e,c),o=cF(i,c,n);i.setItemGraphicEl(t,o),r.add(o),gF(o,c,n)}})).update((function(t,e){var n=o.getItemGraphicEl(e);if(i.hasValue(t)){var a=sF(i,t),s=Zz(i,t,a,c),l=fF(i,s);n&&l!==n.__pictorialShapeStr&&(r.remove(n),i.setItemGraphicEl(t,null),n=null),n?dF(n,c,s):n=cF(i,c,s,!0),i.setItemGraphicEl(t,n),n.__pictorialSymbolMeta=s,r.add(n),gF(n,c,s)}else r.remove(n)})).remove((function(t){var e=o.getItemGraphicEl(t);e&&hF(o,t,e.__pictorialSymbolMeta.animationModel,e)})).execute();var d=t.get("clip",!0)?BS(t.coordinateSystem,!1,t):null;return d?r.setClipPath(d):r.removeClipPath(),this._data=i,this.group},e.prototype.remove=function(t,e){var n=this.group,r=this._data;t.get("animation")?r&&r.eachItemGraphicEl((function(e){hF(r,Ls(e).dataIndex,t,e)})):n.removeAll()},e.type="pictorialBar",e}($g);function Zz(t,e,n,r){var i=t.getItemLayout(e),o=n.get("symbolRepeat"),a=n.get("symbolClip"),s=n.get("symbolPosition")||"start",l=n.get("symbolRotate"),u=(l||0)*Math.PI/180||0,c=n.get("symbolPatternSize")||2,d=n.isAnimationEnabled(),h={dataIndex:e,layout:i,itemModel:n,symbolType:t.getItemVisual(e,"symbol")||"circle",style:t.getItemVisual(e,"style"),symbolClip:a,symbolRepeat:o,symbolRepeatDirection:n.get("symbolRepeatDirection"),symbolPatternSize:c,rotation:u,animationModel:d?n:null,hoverScale:d&&n.get(["emphasis","scale"]),z2:n.getShallow("z",!0)||0};Jz(n,o,i,r,h),Qz(t,e,i,o,a,h.boundingLength,h.pxSign,c,r,h),tF(n,h.symbolScale,u,r,h);var f=h.symbolSize,p=sv(n.get("symbolOffset"),f);return eF(n,f,i,o,a,p,s,h.valueLineWidth,h.boundingLength,h.repeatCutLength,r,h),h}function Jz(t,e,n,r,i){var o,a=r.valueDim,l=t.get("symbolBoundingData"),u=r.coordSys.getOtherAxis(r.coordSys.getBaseAxis()),c=u.toGlobalCoord(u.dataToCoord(0)),d=1-+(n[a.wh]<=0);if(s["t"](l)){var h=[Kz(u,l[0])-c,Kz(u,l[1])-c];h[1]<h[0]&&h.reverse(),o=h[d]}else o=null!=l?Kz(u,l)-c:e?r.coordSysExtent[a.index][d]-c:n[a.wh];i.boundingLength=o,e&&(i.repeatCutLength=n[a.wh]),i.pxSign=o>0?1:-1}function Kz(t,e){return t.toGlobalCoord(t.dataToCoord(t.scale.parse(e)))}function Qz(t,e,n,r,i,o,a,l,u,c){var d,h=u.valueDim,f=u.categoryDim,p=Math.abs(n[f.wh]),m=t.getItemVisual(e,"symbolSize");d=s["t"](m)?m.slice():null==m?["100%","100%"]:[m,m],d[f.index]=Nr(d[f.index],p),d[h.index]=Nr(d[h.index],r?p:Math.abs(o)),c.symbolSize=d;var g=c.symbolScale=[d[0]/l,d[1]/l];g[h.index]*=(u.isHorizontal?-1:1)*a}function tF(t,e,n,r,i){var o=t.get($z)||0;o&&(qz.attr({scaleX:e[0],scaleY:e[1],rotation:n}),qz.updateTransform(),o/=qz.getLineScale(),o*=e[r.valueDim.index]),i.valueLineWidth=o||0}function eF(t,e,n,r,i,o,a,l,u,c,d,h){var f=d.categoryDim,p=d.valueDim,m=h.pxSign,g=Math.max(e[p.index]+l,0),y=g;if(r){var v=Math.abs(u),_=s["O"](t.get("symbolMargin"),"15%")+"",b=!1;_.lastIndexOf("!")===_.length-1&&(b=!0,_=_.slice(0,_.length-1));var x=Nr(_,e[p.index]),w=Math.max(g+2*x,0),M=b?0:2*x,S=ni(r),k=S?r:yF((v+M)/w),L=v-k*g;x=L/2/(b?k:Math.max(k-1,1)),w=g+2*x,M=b?0:2*x,S||"fixed"===r||(k=c?yF((Math.abs(c)+M)/w):0),y=k*w-M,h.repeatTimes=k,h.symbolMargin=x}var T=m*(y/2),D=h.pathPosition=[];D[f.index]=n[f.wh]/2,D[p.index]="start"===a?T:"end"===a?u-T:u/2,o&&(D[0]+=o[0],D[1]+=o[1]);var O=h.bundlePosition=[];O[f.index]=n[f.xy],O[p.index]=n[p.xy];var I=h.barRectShape=s["m"]({},n);I[p.wh]=m*Math.max(Math.abs(n[p.wh]),Math.abs(D[p.index]+T)),I[f.wh]=n[f.wh];var C=h.clipShape={};C[f.xy]=-n[f.xy],C[f.wh]=d.ecSize[f.wh],C[p.xy]=0,C[p.wh]=n[p.wh]}function nF(t){var e=t.symbolPatternSize,n=ov(t.symbolType,-e/2,-e/2,e,e);return n.attr({culling:!0}),"image"!==n.type&&n.setStyle({strokeNoScale:!0}),n}function rF(t,e,n,r){var i=t.__pictorialBundle,o=n.symbolSize,a=n.valueLineWidth,s=n.pathPosition,l=e.valueDim,u=n.repeatTimes||0,c=0,d=o[e.valueDim.index]+a+2*n.symbolMargin;for(pF(t,(function(t){t.__pictorialAnimationIndex=c,t.__pictorialRepeatTimes=u,c<u?mF(t,null,p(c),n,r):mF(t,null,{scaleX:0,scaleY:0},n,r,(function(){i.remove(t)})),c++}));c<u;c++){var h=nF(n);h.__pictorialAnimationIndex=c,h.__pictorialRepeatTimes=u,i.add(h);var f=p(c);mF(h,{x:f.x,y:f.y,scaleX:0,scaleY:0},{scaleX:f.scaleX,scaleY:f.scaleY,rotation:f.rotation},n,r)}function p(t){var e=s.slice(),r=n.pxSign,i=t;return("start"===n.symbolRepeatDirection?r>0:r<0)&&(i=u-1-t),e[l.index]=d*(i-u/2+.5)+s[l.index],{x:e[0],y:e[1],scaleX:n.symbolScale[0],scaleY:n.symbolScale[1],rotation:n.rotation}}}function iF(t,e,n,r){var i=t.__pictorialBundle,o=t.__pictorialMainPath;o?mF(o,null,{x:n.pathPosition[0],y:n.pathPosition[1],scaleX:n.symbolScale[0],scaleY:n.symbolScale[1],rotation:n.rotation},n,r):(o=t.__pictorialMainPath=nF(n),i.add(o),mF(o,{x:n.pathPosition[0],y:n.pathPosition[1],scaleX:0,scaleY:0,rotation:n.rotation},{scaleX:n.symbolScale[0],scaleY:n.symbolScale[1]},n,r))}function oF(t,e,n){var r=s["m"]({},e.barRectShape),i=t.__pictorialBarRect;i?mF(i,null,{shape:r},e,n):(i=t.__pictorialBarRect=new ss({z2:2,shape:r,silent:!0,style:{stroke:"transparent",fill:"transparent",lineWidth:0}}),i.disableMorphing=!0,t.add(i))}function aF(t,e,n,i){if(n.symbolClip){var o=t.__pictorialClipPath,a=s["m"]({},n.clipShape),l=e.valueDim,u=n.animationModel,c=n.dataIndex;if(o)Wl(o,{shape:a},u,c);else{a[l.wh]=0,o=new ss({shape:a}),t.__pictorialBundle.setClipPath(o),t.__pictorialClipPath=o;var d={};d[l.wh]=n.clipShape[l.wh],r[i?"updateProps":"initProps"](o,{shape:d},u,c)}}}function sF(t,e){var n=t.getItemModel(e);return n.getAnimationDelayParams=lF,n.isAnimationEnabled=uF,n}function lF(t){return{index:t.__pictorialAnimationIndex,count:t.__pictorialRepeatTimes}}function uF(){return this.parentModel.isAnimationEnabled()&&!!this.getShallow("animation")}function cF(t,e,n,r){var i=new wr,o=new wr;return i.add(o),i.__pictorialBundle=o,o.x=n.bundlePosition[0],o.y=n.bundlePosition[1],n.symbolRepeat?rF(i,e,n):iF(i,e,n),oF(i,n,r),aF(i,e,n,r),i.__pictorialShapeStr=fF(t,n),i.__pictorialSymbolMeta=n,i}function dF(t,e,n){var r=n.animationModel,i=n.dataIndex,o=t.__pictorialBundle;Wl(o,{x:n.bundlePosition[0],y:n.bundlePosition[1]},r,i),n.symbolRepeat?rF(t,e,n,!0):iF(t,e,n,!0),oF(t,n,!0),aF(t,e,n,!0)}function hF(t,e,n,r){var i=r.__pictorialBarRect;i&&i.removeTextContent();var o=[];pF(r,(function(t){o.push(t)})),r.__pictorialMainPath&&o.push(r.__pictorialMainPath),r.__pictorialClipPath&&(n=null),s["k"](o,(function(t){$l(t,{scaleX:0,scaleY:0},n,e,(function(){r.parent&&r.parent.remove(r)}))})),t.setItemGraphicEl(e,null)}function fF(t,e){return[t.getItemVisual(e.dataIndex,"symbol")||"none",!!e.symbolRepeat,!!e.symbolClip].join(":")}function pF(t,e,n){s["k"](t.__pictorialBundle.children(),(function(r){r!==t.__pictorialBarRect&&e.call(n,r)}))}function mF(t,e,n,i,o,a){e&&t.attr(e),i.symbolClip&&!o?n&&t.attr(n):n&&r[o?"updateProps":"initProps"](t,n,i.animationModel,i.dataIndex,a)}function gF(t,e,n){var r=n.dataIndex,i=n.itemModel,o=i.getModel("emphasis"),a=o.getModel("itemStyle").getItemStyle(),l=i.getModel(["blur","itemStyle"]).getItemStyle(),u=i.getModel(["select","itemStyle"]).getItemStyle(),c=i.getShallow("cursor"),d=o.get("focus"),h=o.get("blurScope"),f=o.get("scale");pF(t,(function(t){if(t instanceof Ka){var e=t.style;t.useStyle(s["m"]({image:e.image,x:e.x,y:e.y,width:e.width,height:e.height},n.style))}else t.useStyle(n.style);var r=t.ensureState("emphasis");r.style=a,f&&(r.scaleX=1.1*t.scaleX,r.scaleY=1.1*t.scaleY),t.ensureState("blur").style=l,t.ensureState("select").style=u,c&&(t.cursor=c),t.z2=n.z2}));var p=e.valueDim.posDesc[+(n.boundingLength>0)],m=t.__pictorialBarRect;m.ignoreClip=!0,tu(m,eu(i),{labelFetcher:e.seriesModel,labelDataIndex:r,defaultText:pS(e.seriesModel.getData(),r),inheritColor:n.style.fill,defaultOpacity:n.style.opacity,defaultOutsidePosition:p}),Dl(t,d,h,o.get("disabled"))}function yF(t){var e=Math.round(t);return Math.abs(t-e)<1e-4?e:Math.ceil(t)}var vF=Xz,_F=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.hasSymbolVisual=!0,n.defaultSymbol="roundRect",n}return o(e,t),e.prototype.getInitialData=function(e){return e.stack=null,t.prototype.getInitialData.apply(this,arguments)},e.type="series.pictorialBar",e.dependencies=["grid"],e.defaultOption=Cu(Ok.defaultOption,{symbol:"circle",symbolSize:null,symbolRotate:null,symbolPosition:null,symbolOffset:null,symbolMargin:null,symbolRepeat:!1,symbolRepeatDirection:"end",symbolClip:!1,symbolBoundingData:null,symbolPatternSize:400,barGap:"-100%",clip:!1,progressive:0,emphasis:{scale:!1},select:{itemStyle:{borderColor:"#212121"}}}),e}(Ok),bF=_F;function xF(t){t.registerChartView(vF),t.registerSeriesModel(bF),t.registerLayout(t.PRIORITY.VISUAL.LAYOUT,Object(s["h"])(Mk,"pictorialBar")),t.registerLayout(t.PRIORITY.VISUAL.PROGRESSIVE_LAYOUT,Sk("pictorialBar"))}var wF=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n._layers=[],n}return o(e,t),e.prototype.render=function(t,e,n){var r=t.getData(),i=this,o=this.group,a=t.getLayerSeries(),l=r.getLayout("layoutInfo"),u=l.rect,c=l.boundaryGap;function d(t){return t.name}o.x=0,o.y=u.y+c[0];var h=new yM(this._layersSeries||[],a,d,d),f=[];function p(e,n,s){var l=i._layers;if("remove"!==e){for(var u,c,d=[],h=[],p=a[n].indices,m=0;m<p.length;m++){var g=r.getItemLayout(p[m]),y=g.x,v=g.y0,_=g.y;d.push(y,v),h.push(y,v+_),u=r.getItemVisual(p[m],"style")}var b=r.getItemLayout(p[0]),x=t.getModel("label"),w=x.get("margin"),M=t.getModel("emphasis");if("add"===e){var S=f[n]=new wr;c=new HS({shape:{points:d,stackedOnPoints:h,smooth:.4,stackedOnSmooth:.4,smoothConstraint:!1},z2:0}),S.add(c),o.add(S),t.isAnimationEnabled()&&c.setClipPath(MF(c.getBoundingRect(),t,(function(){c.removeClipPath()})))}else{S=l[s];c=S.childAt(0),o.add(S),f[n]=S,Wl(c,{shape:{points:d,stackedOnPoints:h}},t),Xl(c)}tu(c,eu(t),{labelDataIndex:p[m-1],defaultText:r.getName(p[m-1]),inheritColor:u.fill},{normal:{verticalAlign:"middle"}}),c.setTextConfig({position:null,local:!0});var k=c.getTextContent();k&&(k.x=b.x-w,k.y=b.y0+b.y/2),c.useStyle(u),r.setItemGraphicEl(n,c),Al(c,t),Dl(c,M.get("focus"),M.get("blurScope"),M.get("disabled"))}else o.remove(l[n])}h.add(Object(s["c"])(p,this,"add")).update(Object(s["c"])(p,this,"update")).remove(Object(s["c"])(p,this,"remove")).execute(),this._layersSeries=a,this._layers=f},e.type="themeRiver",e}($g);function MF(t,e,n){var r=new ss({shape:{x:t.x-10,y:t.y-10,width:0,height:t.height+20}});return Vl(r,{shape:{x:t.x-50,width:t.width+100,height:t.height+20}},e,n),r}var SF=wF,kF=2,LF=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(e){t.prototype.init.apply(this,arguments),this.legendVisualProvider=new OL(s["c"](this.getData,this),s["c"](this.getRawData,this))},e.prototype.fixData=function(t){var e=t.length,n={},r=Fi(t,(function(t){return n.hasOwnProperty(t[0]+"")||(n[t[0]+""]=-1),t[2]})),i=[];r.buckets.each((function(t,e){i.push({name:e,dataList:t})}));for(var o=i.length,a=0;a<o;++a){for(var s=i[a].name,l=0;l<i[a].dataList.length;++l){var u=i[a].dataList[l][0]+"";n[u]=a}for(var u in n)n.hasOwnProperty(u)&&n[u]!==a&&(n[u]=a,t[e]=[u,0,s],e++)}return t},e.prototype.getInitialData=function(t,e){for(var n=this.getReferringComponents("singleAxis",Pi).models[0],r=n.get("type"),i=s["n"](t.data,(function(t){return void 0!==t[2]})),o=this.fixData(i||[]),a=[],l=this.nameMap=s["f"](),u=0,c=0;c<o.length;++c)a.push(o[c][kF]),l.get(o[c][kF])||(l.set(o[c][kF],u),u++);var d=qM(o,{coordDimensions:["single"],dimensionsDefine:[{name:"time",type:xM(r)},{name:"value",type:"float"},{name:"name",type:"ordinal"}],encodeDefine:{single:0,value:1,itemName:2}}).dimensions,h=new UM(d,this);return h.initData(o),h},e.prototype.getLayerSeries=function(){for(var t=this.getData(),e=t.count(),n=[],r=0;r<e;++r)n[r]=r;var i=t.mapDimension("single"),o=Fi(n,(function(e){return t.get("name",e)})),a=[];return o.buckets.each((function(e,n){e.sort((function(e,n){return t.get(i,e)-t.get(i,n)})),a.push({name:n,indices:e})})),a},e.prototype.getAxisTooltipData=function(t,e,n){s["t"](t)||(t=t?[t]:[]);for(var r,i=this.getData(),o=this.getLayerSeries(),a=[],l=o.length,u=0;u<l;++u){for(var c=Number.MAX_VALUE,d=-1,h=o[u].indices.length,f=0;f<h;++f){var p=i.get(t[0],o[u].indices[f]),m=Math.abs(p-e);m<=c&&(r=p,c=m,d=o[u].indices[f])}a.push(d)}return{dataIndices:a,nestestValue:r}},e.prototype.formatTooltip=function(t,e,n){var r=this.getData(),i=r.getName(t),o=r.get(r.mapDimension("value"),t);return Uf("nameValue",{name:i,value:o})},e.type="series.themeRiver",e.dependencies=["singleAxis"],e.defaultOption={z:2,colorBy:"data",coordinateSystem:"singleAxis",boundaryGap:["10%","10%"],singleAxisIndex:0,animationEasing:"linear",label:{margin:4,show:!0,position:"left",fontSize:11},emphasis:{label:{show:!0}}},e}(Mp),TF=LF;function DF(t,e){t.eachSeriesByType("themeRiver",(function(t){var e=t.getData(),n=t.coordinateSystem,r={},i=n.getRect();r.rect=i;var o=t.get("boundaryGap"),a=n.getAxis();if(r.boundaryGap=o,"horizontal"===a.orient){o[0]=Nr(o[0],i.height),o[1]=Nr(o[1],i.height);var s=i.height-o[0]-o[1];OF(e,t,s)}else{o[0]=Nr(o[0],i.width),o[1]=Nr(o[1],i.width);var l=i.width-o[0]-o[1];OF(e,t,l)}e.setLayout("layoutInfo",r)}))}function OF(t,e,n){if(t.count())for(var r,i=e.coordinateSystem,o=e.getLayerSeries(),a=t.mapDimension("single"),l=t.mapDimension("value"),u=s["H"](o,(function(e){return s["H"](e.indices,(function(e){var n=i.dataToPoint(t.get(a,e));return n[1]=t.get(l,e),n}))})),c=IF(u),d=c.y0,h=n/c.max,f=o.length,p=o[0].indices.length,m=0;m<p;++m){r=d[m]*h,t.setItemLayout(o[0].indices[m],{layerIndex:0,x:u[0][m][0],y0:r,y:u[0][m][1]*h});for(var g=1;g<f;++g)r+=u[g-1][m][1]*h,t.setItemLayout(o[g].indices[m],{layerIndex:g,x:u[g][m][0],y0:r,y:u[g][m][1]*h})}}function IF(t){for(var e=t.length,n=t[0].length,r=[],i=[],o=0,a=0;a<n;++a){for(var s=0,l=0;l<e;++l)s+=t[l][a][1];s>o&&(o=s),r.push(s)}for(var u=0;u<n;++u)i[u]=(o-r[u])/2;o=0;for(var c=0;c<n;++c){var d=r[c]+i[c];d>o&&(o=d)}return{y0:i,max:o}}function CF(t){t.registerChartView(SF),t.registerSeriesModel(TF),t.registerLayout(DF),t.registerProcessor(yL("themeRiver"))}var AF=2,YF=4,jF=function(t){function e(e,n,r,i){var o=t.call(this)||this;o.z2=AF,o.textConfig={inside:!0},Ls(o).seriesIndex=n.seriesIndex;var a=new ks({z2:YF,silent:e.getModel().get(["label","silent"])});return o.setTextContent(a),o.updateData(!0,e,n,r,i),o}return o(e,t),e.prototype.updateData=function(t,e,n,r,i){this.node=e,e.piece=this,n=n||this._seriesModel,r=r||this._ecModel;var o=this;Ls(o).dataIndex=e.dataIndex;var a=e.getModel(),l=a.getModel("emphasis"),u=e.getLayout(),c=s["m"]({},u);c.label=null;var d=e.getVisual("style");d.lineJoin="bevel";var h=e.getVisual("decal");h&&(d.decal=qv(h,i));var f=Hk(a.getModel("itemStyle"),c,!0);s["m"](c,f),s["k"](Ps,(function(t){var e=o.ensureState(t),n=a.getModel([t,"itemStyle"]);e.style=n.getItemStyle();var r=Hk(n,c);r&&(e.shape=r)})),t?(o.setShape(c),o.shape.r=u.r0,Vl(o,{shape:{r:u.r}},n,e.dataIndex)):(Wl(o,{shape:c},n),Xl(o)),o.useStyle(d),this._updateLabel(n);var p=a.getShallow("cursor");p&&o.attr("cursor",p),this._seriesModel=n||this._seriesModel,this._ecModel=r||this._ecModel;var m=l.get("focus"),g="ancestor"===m?e.getAncestorsIndices():"descendant"===m?e.getDescendantIndices():m;Dl(this,g,l.get("blurScope"),l.get("disabled"))},e.prototype._updateLabel=function(t){var e=this,n=this.node.getModel(),r=n.getModel("label"),i=this.node.getLayout(),o=i.endAngle-i.startAngle,a=(i.startAngle+i.endAngle)/2,l=Math.cos(a),u=Math.sin(a),c=this,d=c.getTextContent(),h=this.node.dataIndex,f=r.get("minAngle")/180*Math.PI,p=r.get("show")&&!(null!=f&&Math.abs(o)<f);function m(t,e){var n=t.get(e);return null==n?r.get(e):n}d.ignore=!p,s["k"](Es,(function(r){var f="normal"===r?n.getModel("label"):n.getModel([r,"label"]),p="normal"===r,g=p?d:d.ensureState(r),y=t.getFormattedLabel(h,r);p&&(y=y||e.node.name),g.style=nu(f,{},null,"normal"!==r,!0),y&&(g.style.text=y);var v=f.get("show");null==v||p||(g.ignore=!v);var _,b=m(f,"position"),x=p?c:c.states[r],w=x.style.fill;x.textConfig={outsideFill:"inherit"===f.get("color")?w:null,inside:"outside"!==b};var M=m(f,"distance")||0,S=m(f,"align"),k=m(f,"rotate"),L=.5*Math.PI,T=1.5*Math.PI,D=Ma("tangential"===k?Math.PI/2-a:a),O=D>L&&!Ur(D-L)&&D<T;"outside"===b?(_=i.r+M,S=O?"right":"left"):S&&"center"!==S?"left"===S?(_=i.r0+M,S=O?"right":"left"):"right"===S&&(_=i.r-M,S=O?"left":"right"):(_=o===2*Math.PI&&0===i.r0?0:(i.r+i.r0)/2,S="center"),g.style.align=S,g.style.verticalAlign=m(f,"verticalAlign")||"middle",g.x=_*l+i.cx,g.y=_*u+i.cy;var I=0;"radial"===k?I=Ma(-a)+(O?Math.PI:0):"tangential"===k?I=Ma(Math.PI/2-a)+(O?Math.PI:0):s["z"](k)&&(I=k*Math.PI/180),g.rotation=Ma(I)})),d.dirtyStyle()},e}(_m),PF=jF,EF="sunburstRootToNode",NF="sunburstHighlight",RF="sunburstUnhighlight";function HF(t){t.registerAction({type:EF,update:"updateView"},(function(t,e){function n(e,n){var r=eY(t,[EF],e);if(r){var i=e.getViewRoot();i&&(t.direction=rY(i,r.node)?"rollUp":"drillDown"),e.resetViewRoot(r.node)}}e.eachComponent({mainType:"series",subType:"sunburst",query:t},n)})),t.registerAction({type:NF,update:"none"},(function(t,e,n){function r(e){var n=eY(t,[NF],e);n&&(t.dataIndex=n.node.dataIndex)}t=Object(s["m"])({},t),e.eachComponent({mainType:"series",subType:"sunburst",query:t},r),n.dispatchAction(Object(s["m"])(t,{type:"highlight"}))})),t.registerAction({type:RF,update:"updateView"},(function(t,e,n){t=Object(s["m"])({},t),n.dispatchAction(Object(s["m"])(t,{type:"downplay"}))}))}var zF=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n,r){var i=this;this.seriesModel=t,this.api=n,this.ecModel=e;var o=t.getData(),a=o.tree.root,l=t.getViewRoot(),u=this.group,c=t.get("renderLabelForZeroData"),d=[];l.eachNode((function(t){d.push(t)}));var h=this._oldChildren||[];function f(t,e){function n(t){return t.getId()}function r(n,r){var i=null==n?null:t[n],o=null==r?null:e[r];p(i,o)}0===t.length&&0===e.length||new yM(e,t,n,n).add(r).update(r).remove(s["h"](r,null)).execute()}function p(r,i){if(c||!r||r.getValue()||(r=null),r!==a&&i!==a)if(i&&i.piece)r?(i.piece.updateData(!1,r,t,e,n),o.setItemGraphicEl(r.dataIndex,i.piece)):m(i);else if(r){var s=new PF(r,t,e,n);u.add(s),o.setItemGraphicEl(r.dataIndex,s)}}function m(t){t&&t.piece&&(u.remove(t.piece),t.piece=null)}function g(r,o){o.depth>0?(i.virtualPiece?i.virtualPiece.updateData(!1,r,t,e,n):(i.virtualPiece=new PF(r,t,e,n),u.add(i.virtualPiece)),o.piece.off("click"),i.virtualPiece.on("click",(function(t){i._rootToNode(o.parentNode)}))):i.virtualPiece&&(u.remove(i.virtualPiece),i.virtualPiece=null)}f(d,h),g(a,l),this._initEvents(),this._oldChildren=d},e.prototype._initEvents=function(){var t=this;this.group.off("click"),this.group.on("click",(function(e){var n=!1,r=t.seriesModel.getViewRoot();r.eachNode((function(r){if(!n&&r.piece&&r.piece===e.target){var i=r.getModel().get("nodeClick");if("rootToNode"===i)t._rootToNode(r);else if("link"===i){var o=r.getModel(),a=o.get("link");if(a){var s=o.get("target",!0)||"_blank";Ac(a,s)}}n=!0}}))}))},e.prototype._rootToNode=function(t){t!==this.seriesModel.getViewRoot()&&this.api.dispatchAction({type:EF,from:this.uid,seriesId:this.seriesModel.id,targetNode:t})},e.prototype.containPoint=function(t,e){var n=e.getData(),r=n.getItemLayout(0);if(r){var i=t[0]-r.cx,o=t[1]-r.cy,a=Math.sqrt(i*i+o*o);return a<=r.r&&a>=r.r0}},e.type="sunburst",e}($g),FF=zF,BF=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.ignoreStyleOnData=!0,n}return o(e,t),e.prototype.getInitialData=function(t,e){var n={name:t.name,children:t.data};WF(n);var r=this._levelModels=s["H"](t.levels||[],(function(t){return new Lu(t,this,e)}),this),i=tY.createTree(n,this,o);function o(t){t.wrapMethod("getItemModel",(function(t,e){var n=i.getNodeByDataIndex(e),o=r[n.depth];return o&&(t.parentModel=o),t}))}return i.data},e.prototype.optionUpdated=function(){this.resetViewRoot()},e.prototype.getDataParams=function(e){var n=t.prototype.getDataParams.apply(this,arguments),r=this.getData().tree.getNodeByDataIndex(e);return n.treePathInfo=iY(r,this),n},e.prototype.getLevelModel=function(t){return this._levelModels&&this._levelModels[t.depth]},e.prototype.getViewRoot=function(){return this._viewRoot},e.prototype.resetViewRoot=function(t){t?this._viewRoot=t:t=this._viewRoot;var e=this.getRawData().tree.root;t&&(t===e||e.contains(t))||(this._viewRoot=e)},e.prototype.enableAriaDecal=function(){gY(this)},e.type="series.sunburst",e.defaultOption={z:2,center:["50%","50%"],radius:[0,"75%"],clockwise:!0,startAngle:90,minAngle:0,stillShowZeroSum:!0,nodeClick:"rootToNode",renderLabelForZeroData:!1,label:{rotate:"radial",show:!0,opacity:1,align:"center",position:"inside",distance:5,silent:!0},itemStyle:{borderWidth:1,borderColor:"white",borderType:"solid",shadowBlur:0,shadowColor:"rgba(0, 0, 0, 0.2)",shadowOffsetX:0,shadowOffsetY:0,opacity:1},emphasis:{focus:"descendant"},blur:{itemStyle:{opacity:.2},label:{opacity:.1}},animationType:"expansion",animationDuration:1e3,animationDurationUpdate:500,data:[],sort:"desc"},e}(Mp);function WF(t){var e=0;s["k"](t.children,(function(t){WF(t);var n=t.value;s["t"](n)&&(n=n[0]),e+=n}));var n=t.value;s["t"](n)&&(n=n[0]),(null==n||isNaN(n))&&(n=e),n<0&&(n=0),s["t"](t.value)?t.value[0]=n:t.value=n}var VF=BF,GF=Math.PI/180;function $F(t,e,n){e.eachSeriesByType(t,(function(t){var e=t.get("center"),r=t.get("radius");s["t"](r)||(r=[0,r]),s["t"](e)||(e=[e,e]);var i=n.getWidth(),o=n.getHeight(),a=Math.min(i,o),l=Nr(e[0],i),u=Nr(e[1],o),c=Nr(r[0],a/2),d=Nr(r[1],a/2),h=-t.get("startAngle")*GF,f=t.get("minAngle")*GF,p=t.getData().tree.root,m=t.getViewRoot(),g=m.depth,y=t.get("sort");null!=y&&UF(m,y);var v=0;s["k"](m.children,(function(t){!isNaN(t.getValue())&&v++}));var _=m.getValue(),b=Math.PI/(_||v)*2,x=m.depth>0,w=m.height-(x?-1:1),M=(d-c)/(w||1),S=t.get("clockwise"),k=t.get("stillShowZeroSum"),L=S?1:-1,T=function(e,n){if(e){var r=n;if(e!==p){var i=e.getValue(),o=0===_&&k?b:i*b;o<f&&(o=f),r=n+L*o;var d=e.depth-g-(x?-1:1),h=c+M*d,m=c+M*(d+1),y=t.getLevelModel(e);if(y){var v=y.get("r0",!0),w=y.get("r",!0),D=y.get("radius",!0);null!=D&&(v=D[0],w=D[1]),null!=v&&(h=Nr(v,a/2)),null!=w&&(m=Nr(w,a/2))}e.setLayout({angle:o,startAngle:n,endAngle:r,clockwise:S,cx:l,cy:u,r0:h,r:m})}if(e.children&&e.children.length){var O=0;s["k"](e.children,(function(t){O+=T(t,n+O)}))}return r-n}};if(x){var D=c,O=c+M,I=2*Math.PI;p.setLayout({angle:I,startAngle:h,endAngle:h+I,clockwise:S,cx:l,cy:u,r0:D,r:O})}T(m,h)}))}function UF(t,e){var n=t.children||[];t.children=qF(n,e),n.length&&s["k"](t.children,(function(t){UF(t,e)}))}function qF(t,e){if(s["w"](e)){var n=s["H"](t,(function(t,e){var n=t.getValue();return{params:{depth:t.depth,height:t.height,dataIndex:t.dataIndex,getValue:function(){return n}},index:e}}));return n.sort((function(t,n){return e(t.params,n.params)})),s["H"](n,(function(e){return t[e.index]}))}var r="asc"===e;return t.sort((function(t,e){var n=(t.getValue()-e.getValue())*(r?1:-1);return 0===n?(t.dataIndex-e.dataIndex)*(r?-1:1):n}))}function XF(t){var e={};function n(t,n,r){var i=t;while(i&&i.depth>1)i=i.parentNode;var o=n.getColorFromPalette(i.name||i.dataIndex+"",e);return t.depth>1&&Object(s["C"])(o)&&(o=Object(Ne["c"])(o,(t.depth-1)/(r-1)*.5)),o}t.eachSeriesByType("sunburst",(function(t){var e=t.getData(),r=e.tree;r.eachNode((function(i){var o=i.getModel(),a=o.getModel("itemStyle").getItemStyle();a.fill||(a.fill=n(i,t,r.root.height));var l=e.ensureUniqueItemVisual(i.dataIndex,"style");Object(s["m"])(l,a)}))}))}function ZF(t){t.registerChartView(FF),t.registerSeriesModel(VF),t.registerLayout(Object(s["h"])($F,"sunburst")),t.registerProcessor(Object(s["h"])(yL,"sunburst")),t.registerVisual(XF),HF(t)}var JF={color:"fill",borderColor:"stroke"},KF={symbol:1,symbolSize:1,symbolKeepAspect:1,legendIcon:1,visualMeta:1,liftZ:1,decal:1},QF=Ci(),tB=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.optionUpdated=function(){this.currentZLevel=this.get("zlevel",!0),this.currentZ=this.get("z",!0)},e.prototype.getInitialData=function(t,e){return dS(null,this)},e.prototype.getDataParams=function(e,n,r){var i=t.prototype.getDataParams.call(this,e,n);return r&&(i.info=QF(r).info),i},e.type="series.custom",e.dependencies=["grid","polar","geo","singleAxis","calendar"],e.defaultOption={coordinateSystem:"cartesian2d",z:2,legendHoverLink:!0,clip:!1},e}(Mp),eB=tB;function nB(t,e){return e=e||[0,0],s["H"](["x","y"],(function(n,r){var i=this.getAxis(n),o=e[r],a=t[r]/2;return"category"===i.type?i.getBandWidth():Math.abs(i.dataToCoord(o-a)-i.dataToCoord(o+a))}),this)}function rB(t){var e=t.master.getRect();return{coordSys:{type:"cartesian2d",x:e.x,y:e.y,width:e.width,height:e.height},api:{coord:function(e){return t.dataToPoint(e)},size:s["c"](nB,t)}}}function iB(t,e){return e=e||[0,0],s["H"]([0,1],(function(n){var r=e[n],i=t[n]/2,o=[],a=[];return o[n]=r-i,a[n]=r+i,o[1-n]=a[1-n]=e[1-n],Math.abs(this.dataToPoint(o)[n]-this.dataToPoint(a)[n])}),this)}function oB(t){var e=t.getBoundingRect();return{coordSys:{type:"geo",x:e.x,y:e.y,width:e.width,height:e.height,zoom:t.getZoom()},api:{coord:function(e){return t.dataToPoint(e)},size:s["c"](iB,t)}}}function aB(t,e){var n=this.getAxis(),r=e instanceof Array?e[0]:e,i=(t instanceof Array?t[0]:t)/2;return"category"===n.type?n.getBandWidth():Math.abs(n.dataToCoord(r-i)-n.dataToCoord(r+i))}function sB(t){var e=t.getRect();return{coordSys:{type:"singleAxis",x:e.x,y:e.y,width:e.width,height:e.height},api:{coord:function(e){return t.dataToPoint(e)},size:Object(s["c"])(aB,t)}}}function lB(t,e){return e=e||[0,0],s["H"](["Radius","Angle"],(function(n,r){var i="get"+n+"Axis",o=this[i](),a=e[r],s=t[r]/2,l="category"===o.type?o.getBandWidth():Math.abs(o.dataToCoord(a-s)-o.dataToCoord(a+s));return"Angle"===n&&(l=l*Math.PI/180),l}),this)}function uB(t){var e=t.getRadiusAxis(),n=t.getAngleAxis(),r=e.getExtent();return r[0]>r[1]&&r.reverse(),{coordSys:{type:"polar",cx:t.cx,cy:t.cy,r:r[1],r0:r[0]},api:{coord:function(r){var i=e.dataToRadius(r[0]),o=n.dataToAngle(r[1]),a=t.coordToPoint([i,o]);return a.push(i,o*Math.PI/180),a},size:s["c"](lB,t)}}}function cB(t){var e=t.getRect(),n=t.getRangeInfo();return{coordSys:{type:"calendar",x:e.x,y:e.y,width:e.width,height:e.height,cellWidth:t.getCellWidth(),cellHeight:t.getCellHeight(),rangeInfo:{start:n.start,end:n.end,weeks:n.weeks,dayCount:n.allDay}},api:{coord:function(e,n){return t.dataToPoint(e,n)}}}}function dB(t,e,n,r){return t&&(t.legacy||!1!==t.legacy&&!n&&!r&&"tspan"!==e&&("text"===e||Object(s["q"])(t,"text")))}function hB(t,e,n){var r,i,o,a=t;if("text"===e)o=a;else{o={},Object(s["q"])(a,"text")&&(o.text=a.text),Object(s["q"])(a,"rich")&&(o.rich=a.rich),Object(s["q"])(a,"textFill")&&(o.fill=a.textFill),Object(s["q"])(a,"textStroke")&&(o.stroke=a.textStroke),Object(s["q"])(a,"fontFamily")&&(o.fontFamily=a.fontFamily),Object(s["q"])(a,"fontSize")&&(o.fontSize=a.fontSize),Object(s["q"])(a,"fontStyle")&&(o.fontStyle=a.fontStyle),Object(s["q"])(a,"fontWeight")&&(o.fontWeight=a.fontWeight),i={type:"text",style:o,silent:!0},r={};var l=Object(s["q"])(a,"textPosition");n?r.position=l?a.textPosition:"inside":l&&(r.position=a.textPosition),Object(s["q"])(a,"textPosition")&&(r.position=a.textPosition),Object(s["q"])(a,"textOffset")&&(r.offset=a.textOffset),Object(s["q"])(a,"textRotation")&&(r.rotation=a.textRotation),Object(s["q"])(a,"textDistance")&&(r.distance=a.textDistance)}return fB(o,t),Object(s["k"])(o.rich,(function(t){fB(t,t)})),{textConfig:r,textContent:i}}function fB(t,e){e&&(e.font=e.textFont||e.font,Object(s["q"])(e,"textStrokeWidth")&&(t.lineWidth=e.textStrokeWidth),Object(s["q"])(e,"textAlign")&&(t.align=e.textAlign),Object(s["q"])(e,"textVerticalAlign")&&(t.verticalAlign=e.textVerticalAlign),Object(s["q"])(e,"textLineHeight")&&(t.lineHeight=e.textLineHeight),Object(s["q"])(e,"textWidth")&&(t.width=e.textWidth),Object(s["q"])(e,"textHeight")&&(t.height=e.textHeight),Object(s["q"])(e,"textBackgroundColor")&&(t.backgroundColor=e.textBackgroundColor),Object(s["q"])(e,"textPadding")&&(t.padding=e.textPadding),Object(s["q"])(e,"textBorderColor")&&(t.borderColor=e.textBorderColor),Object(s["q"])(e,"textBorderWidth")&&(t.borderWidth=e.textBorderWidth),Object(s["q"])(e,"textBorderRadius")&&(t.borderRadius=e.textBorderRadius),Object(s["q"])(e,"textBoxShadowColor")&&(t.shadowColor=e.textBoxShadowColor),Object(s["q"])(e,"textBoxShadowBlur")&&(t.shadowBlur=e.textBoxShadowBlur),Object(s["q"])(e,"textBoxShadowOffsetX")&&(t.shadowOffsetX=e.textBoxShadowOffsetX),Object(s["q"])(e,"textBoxShadowOffsetY")&&(t.shadowOffsetY=e.textBoxShadowOffsetY))}function pB(t,e,n){var r=t;r.textPosition=r.textPosition||n.position||"inside",null!=n.offset&&(r.textOffset=n.offset),null!=n.rotation&&(r.textRotation=n.rotation),null!=n.distance&&(r.textDistance=n.distance);var i=r.textPosition.indexOf("inside")>=0,o=t.fill||"#000";mB(r,e);var a=null==r.textFill;return i?a&&(r.textFill=n.insideFill||"#fff",!r.textStroke&&n.insideStroke&&(r.textStroke=n.insideStroke),!r.textStroke&&(r.textStroke=o),null==r.textStrokeWidth&&(r.textStrokeWidth=2)):(a&&(r.textFill=t.fill||n.outsideFill||"#000"),!r.textStroke&&n.outsideStroke&&(r.textStroke=n.outsideStroke)),r.text=e.text,r.rich=e.rich,Object(s["k"])(e.rich,(function(t){mB(t,t)})),r}function mB(t,e){e&&(Object(s["q"])(e,"fill")&&(t.textFill=e.fill),Object(s["q"])(e,"stroke")&&(t.textStroke=e.fill),Object(s["q"])(e,"lineWidth")&&(t.textStrokeWidth=e.lineWidth),Object(s["q"])(e,"font")&&(t.font=e.font),Object(s["q"])(e,"fontStyle")&&(t.fontStyle=e.fontStyle),Object(s["q"])(e,"fontWeight")&&(t.fontWeight=e.fontWeight),Object(s["q"])(e,"fontSize")&&(t.fontSize=e.fontSize),Object(s["q"])(e,"fontFamily")&&(t.fontFamily=e.fontFamily),Object(s["q"])(e,"align")&&(t.textAlign=e.align),Object(s["q"])(e,"verticalAlign")&&(t.textVerticalAlign=e.verticalAlign),Object(s["q"])(e,"lineHeight")&&(t.textLineHeight=e.lineHeight),Object(s["q"])(e,"width")&&(t.textWidth=e.width),Object(s["q"])(e,"height")&&(t.textHeight=e.height),Object(s["q"])(e,"backgroundColor")&&(t.textBackgroundColor=e.backgroundColor),Object(s["q"])(e,"padding")&&(t.textPadding=e.padding),Object(s["q"])(e,"borderColor")&&(t.textBorderColor=e.borderColor),Object(s["q"])(e,"borderWidth")&&(t.textBorderWidth=e.borderWidth),Object(s["q"])(e,"borderRadius")&&(t.textBorderRadius=e.borderRadius),Object(s["q"])(e,"shadowColor")&&(t.textBoxShadowColor=e.shadowColor),Object(s["q"])(e,"shadowBlur")&&(t.textBoxShadowBlur=e.shadowBlur),Object(s["q"])(e,"shadowOffsetX")&&(t.textBoxShadowOffsetX=e.shadowOffsetX),Object(s["q"])(e,"shadowOffsetY")&&(t.textBoxShadowOffsetY=e.shadowOffsetY),Object(s["q"])(e,"textShadowColor")&&(t.textShadowColor=e.textShadowColor),Object(s["q"])(e,"textShadowBlur")&&(t.textShadowBlur=e.textShadowBlur),Object(s["q"])(e,"textShadowOffsetX")&&(t.textShadowOffsetX=e.textShadowOffsetX),Object(s["q"])(e,"textShadowOffsetY")&&(t.textShadowOffsetY=e.textShadowOffsetY))}var gB={position:["x","y"],scale:["scaleX","scaleY"],origin:["originX","originY"]},yB=Object(s["F"])(gB),vB=(Object(s["N"])(Un,(function(t,e){return t[e]=1,t}),{}),Un.join(", "),["","style","shape","extra"]),_B=Ci();function bB(t,e,n,r,i){var o=t+"Animation",a=Fl(t,r,i)||{},l=_B(e).userDuring;return a.duration>0&&(a.during=l?Object(s["c"])(IB,{el:e,userDuring:l}):null,a.setToFinal=!0,a.scope=t),Object(s["m"])(a,n[o]),a}function xB(t,e,n,r){r=r||{};var i=r.dataIndex,o=r.isInit,a=r.clearStyle,l=n.isAnimationEnabled(),u=_B(t),c=e.style;u.userDuring=e.during;var d={},h={};if(jB(t,e,h),AB("shape",e,h),AB("extra",e,h),!o&&l&&(YB(t,e,d),CB("shape",t,e,d),CB("extra",t,e,d),PB(t,e,c,d)),h.style=c,kB(t,h,a),TB(t,e),l)if(o){var f={};Object(s["k"])(vB,(function(t){var n=t?e[t]:e;n&&n.enterFrom&&(t&&(f[t]=f[t]||{}),Object(s["m"])(t?f[t]:f,n.enterFrom))}));var p=bB("enter",t,e,n,i);p.duration>0&&t.animateFrom(f,p)}else LB(t,e,i||0,n,d);wB(t,e),c?t.dirty():t.markRedraw()}function wB(t,e){for(var n=_B(t).leaveToProps,r=0;r<vB.length;r++){var i=vB[r],o=i?e[i]:e;o&&o.leaveTo&&(n||(n=_B(t).leaveToProps={}),i&&(n[i]=n[i]||{}),Object(s["m"])(i?n[i]:n,o.leaveTo))}}function MB(t,e,n,r){if(t){var i=t.parent,o=_B(t).leaveToProps;if(o){var a=bB("update",t,e,n,0);a.done=function(){i.remove(t),r&&r()},t.animateTo(o,a)}else i.remove(t),r&&r()}}function SB(t){return"all"===t}function kB(t,e,n){var r=e.style;if(!t.isGroup&&r){if(n){t.useStyle({});for(var i=t.animators,o=0;o<i.length;o++){var a=i[o];"style"===a.targetName&&a.changeTarget(t.style)}}t.setStyle(r)}e&&(e.style=null,e&&t.attr(e),e.style=r)}function LB(t,e,n,r,i){if(i){var o=bB("update",t,e,r,n);o.duration>0&&t.animateFrom(i,o)}}function TB(t,e){Object(s["q"])(e,"silent")&&(t.silent=e.silent),Object(s["q"])(e,"ignore")&&(t.ignore=e.ignore),t instanceof No&&Object(s["q"])(e,"invisible")&&(t.invisible=e.invisible),t instanceof Va&&Object(s["q"])(e,"autoBatch")&&(t.autoBatch=e.autoBatch)}var DB={},OB={setTransform:function(t,e){return DB.el[t]=e,this},getTransform:function(t){return DB.el[t]},setShape:function(t,e){var n=DB.el,r=n.shape||(n.shape={});return r[t]=e,n.dirtyShape&&n.dirtyShape(),this},getShape:function(t){var e=DB.el.shape;if(e)return e[t]},setStyle:function(t,e){var n=DB.el,r=n.style;return r&&(r[t]=e,n.dirtyStyle&&n.dirtyStyle()),this},getStyle:function(t){var e=DB.el.style;if(e)return e[t]},setExtra:function(t,e){var n=DB.el.extra||(DB.el.extra={});return n[t]=e,this},getExtra:function(t){var e=DB.el.extra;if(e)return e[t]}};function IB(){var t=this,e=t.el;if(e){var n=_B(e).userDuring,r=t.userDuring;n===r?(DB.el=e,r(OB)):t.el=t.userDuring=null}}function CB(t,e,n,r){var i=n[t];if(i){var o,a=e[t];if(a){var l=n.transition,u=i.transition;if(u)if(!o&&(o=r[t]={}),SB(u))Object(s["m"])(o,a);else for(var c=ui(u),d=0;d<c.length;d++){var h=c[d],f=a[h];o[h]=f}else if(SB(l)||Object(s["r"])(l,t)>=0){!o&&(o=r[t]={});var p=Object(s["F"])(a);for(d=0;d<p.length;d++){h=p[d],f=a[h];EB(i[h],f)&&(o[h]=f)}}}}}function AB(t,e,n){var r=e[t];if(r)for(var i=n[t]={},o=Object(s["F"])(r),a=0;a<o.length;a++){var l=o[a];i[l]=Ue(r[l])}}function YB(t,e,n){for(var r=e.transition,i=SB(r)?Un:ui(r||[]),o=0;o<i.length;o++){var a=i[o];if("style"!==a&&"shape"!==a&&"extra"!==a){var s=t[a];0,n[a]=s}}}function jB(t,e,n){for(var r=0;r<yB.length;r++){var i=yB[r],o=gB[i],a=e[i];a&&(n[o[0]]=a[0],n[o[1]]=a[1])}for(r=0;r<Un.length;r++){var s=Un[r];null!=e[s]&&(n[s]=e[s])}}function PB(t,e,n,r){if(n){var i,o=t.style;if(o){var a=n.transition,l=e.transition;if(a&&!SB(a)){var u=ui(a);!i&&(i=r.style={});for(var c=0;c<u.length;c++){var d=u[c],h=o[d];i[d]=h}}else if(t.getAnimationStyleProps&&(SB(l)||SB(a)||Object(s["r"])(l,"style")>=0)){var f=t.getAnimationStyleProps(),p=f?f.style:null;if(p){!i&&(i=r.style={});var m=Object(s["F"])(n);for(c=0;c<m.length;c++){d=m[c];if(p[d]){h=o[d];i[d]=h}}}}}}}function EB(t,e){return Object(s["u"])(t)?t!==e:null!=t&&isFinite(t)}var NB=Ci(),RB=["percent","easing","shape","style","extra"];function HB(t){t.stopAnimation("keyframe"),t.attr(NB(t))}function zB(t,e,n){if(n.isAnimationEnabled()&&e)if(Object(s["t"])(e))Object(s["k"])(e,(function(e){zB(t,e,n)}));else{var r=e.keyframes,i=e.duration;if(n&&null==i){var o=Fl("enter",n,0);i=o&&o.duration}if(r&&i){var a=NB(t);Object(s["k"])(vB,(function(n){if(!n||t[n]){var o;r.sort((function(t,e){return t.percent-e.percent})),Object(s["k"])(r,(function(r){var l=t.animators,u=n?r[n]:r;if(u){var c=Object(s["F"])(u);if(n||(c=Object(s["n"])(c,(function(t){return Object(s["r"])(RB,t)<0}))),c.length){o||(o=t.animate(n,e.loop,!0),o.scope="keyframe");for(var d=0;d<l.length;d++)l[d]!==o&&l[d].targetName===o.targetName&&l[d].stopTracks(c);n&&(a[n]=a[n]||{});var h=n?a[n]:a;Object(s["k"])(c,(function(e){h[e]=((n?t[n]:t)||{})[e]})),o.whenWithKeys(i*r.percent,u,c,r.easing)}}})),o&&o.delay(e.delay||0).duration(i).start(e.easing)}}))}}}var FB="emphasis",BB="normal",WB="blur",VB="select",GB=[BB,FB,WB,VB],$B={normal:["itemStyle"],emphasis:[FB,"itemStyle"],blur:[WB,"itemStyle"],select:[VB,"itemStyle"]},UB={normal:["label"],emphasis:[FB,"label"],blur:[WB,"label"],select:[VB,"label"]},qB=["x","y"],XB="e\0\0",ZB={normal:{},emphasis:{},blur:{},select:{}},JB={cartesian2d:rB,geo:oB,single:sB,polar:uB,calendar:cB};function KB(t){return t instanceof Va}function QB(t){return t instanceof No}function tW(t,e){e.copyTransform(t),QB(e)&&QB(t)&&(e.setStyle(t.style),e.z=t.z,e.z2=t.z2,e.zlevel=t.zlevel,e.invisible=t.invisible,e.ignore=t.ignore,KB(e)&&KB(t)&&e.setShape(t.shape))}var eW=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n,r){this._progressiveEls=null;var i=this._data,o=t.getData(),a=this.group,s=lW(t,o,e,n);i||a.removeAll(),o.diff(i).add((function(e){cW(n,null,e,s(e,r),t,a,o)})).remove((function(e){var n=i.getItemGraphicEl(e);n&&MB(n,QF(n).option,t)})).update((function(e,l){var u=i.getItemGraphicEl(l);cW(n,u,e,s(e,r),t,a,o)})).execute();var l=t.get("clip",!0)?BS(t.coordinateSystem,!1,t):null;l?a.setClipPath(l):a.removeClipPath(),this._data=o},e.prototype.incrementalPrepareRender=function(t,e,n){this.group.removeAll(),this._data=null},e.prototype.incrementalRender=function(t,e,n,r,i){var o=e.getData(),a=lW(e,o,n,r),s=this._progressiveEls=[];function l(t){t.isGroup||(t.incremental=!0,t.ensureState("emphasis").hoverLayer=!0)}for(var u=t.start;u<t.end;u++){var c=cW(null,null,u,a(u,i),e,this.group,o);c&&(c.traverse(l),s.push(c))}},e.prototype.eachRendered=function(t){Ng(this._progressiveEls||this.group,t)},e.prototype.filterForExposedEvent=function(t,e,n,r){var i=e.element;if(null==i||n.name===i)return!0;while((n=n.__hostTarget||n.parent)&&n!==this.group)if(n.name===i)return!0;return!1},e.type="custom",e}($g),nW=eW;function rW(t){var e,n=t.type;if("path"===n){var r=t.shape,i=null!=r.width&&null!=r.height?{x:r.x||0,y:r.y||0,width:r.width,height:r.height}:null,o=SW(r);e=pg(o,null,i,r.layout||"center"),QF(e).customPathData=o}else if("image"===n)e=new Ka({}),QF(e).customImagePath=t.style.image;else if("text"===n)e=new ks({});else if("group"===n)e=new wr;else{if("compoundPath"===n)throw new Error('"compoundPath" is not supported yet.');var a=fg(n);if(!a){var s="";0,Cd(s)}e=new a}return QF(e).customGraphicType=n,e.name=t.name,e.z2EmphasisLift=1,e.z2SelectLift=1,e}function iW(t,e,n,r,i,o,a){HB(e);var l=i&&i.normal.cfg;l&&e.setTextConfig(l),r&&null==r.transition&&(r.transition=qB);var u=r&&r.style;if(u){if("text"===e.type){var c=u;Object(s["q"])(c,"textFill")&&(c.fill=c.textFill),Object(s["q"])(c,"textStroke")&&(c.stroke=c.textStroke)}var d=void 0,h=KB(e)?u.decal:null;t&&h&&(h.dirty=!0,d=qv(h,t)),u.__decalPattern=d}if(QB(e)&&u){d=u.__decalPattern;d&&(u.decal=d)}xB(e,r,o,{dataIndex:n,isInit:a,clearStyle:!0}),zB(e,r.keyframeAnimation,o)}function oW(t,e,n,r,i){var o=e.isGroup?null:e,a=i&&i[t].cfg;if(o){var s=o.ensureState(t);if(!1===r){var l=o.getState(t);l&&(l.style=null)}else s.style=r||null;a&&(s.textConfig=a),al(o)}}function aW(t,e,n){if(!t.isGroup){var r=t,i=n.currentZ,o=n.currentZLevel;r.z=i,r.zlevel=o;var a=e.z2;null!=a&&(r.z2=a||0);for(var s=0;s<GB.length;s++)sW(r,e,GB[s])}}function sW(t,e,n){var r,i=n===BB,o=i?e:gW(e,n),a=o?o.z2:null;null!=a&&(r=i?t:t.ensureState(n),r.z2=a||0)}function lW(t,e,n,r){var i=t.get("renderItem"),o=t.coordinateSystem,a={};o&&(a=o.prepareCustoms?o.prepareCustoms(o):JB[o.type](o));for(var l,u,c=Object(s["i"])({getWidth:r.getWidth,getHeight:r.getHeight,getZr:r.getZr,getDevicePixelRatio:r.getDevicePixelRatio,value:x,style:M,ordinalRawValue:w,styleEmphasis:S,visual:T,barLayout:D,currentSeriesIndices:O,font:I},a.api||{}),d={context:{},seriesId:t.id,seriesName:t.name,seriesIndex:t.seriesIndex,coordSys:a.coordSys,dataInsideLength:e.count(),encode:uW(t.getData())},h={},f={},p={},m={},g=0;g<GB.length;g++){var y=GB[g];p[y]=t.getModel($B[y]),m[y]=t.getModel(UB[y])}function v(t){return t===l?u||(u=e.getItemModel(t)):e.getItemModel(t)}function _(t,n){return e.hasItemOption?t===l?h[n]||(h[n]=v(t).getModel($B[n])):v(t).getModel($B[n]):p[n]}function b(t,n){return e.hasItemOption?t===l?f[n]||(f[n]=v(t).getModel(UB[n])):v(t).getModel(UB[n]):m[n]}return function(t,n){return l=t,u=null,h={},f={},i&&i(Object(s["i"])({dataIndexInside:t,dataIndex:e.getRawIndex(t),actionType:n?n.type:null},d),c)};function x(t,n){return null==n&&(n=l),e.getStore().get(e.getDimensionIndex(t||0),n)}function w(t,n){null==n&&(n=l),t=t||0;var r=e.getDimensionInfo(t);if(!r){var i=e.getDimensionIndex(t);return i>=0?e.getStore().get(i,n):void 0}var o=e.get(r.name,n),a=r&&r.ordinalMeta;return a?a.categories[o]:o}function M(n,r){null==r&&(r=l);var i=e.getItemVisual(r,"style"),o=i&&i.fill,a=i&&i.opacity,u=_(r,BB).getItemStyle();null!=o&&(u.fill=o),null!=a&&(u.opacity=a);var c={inheritColor:Object(s["C"])(o)?o:"#000"},d=b(r,BB),h=nu(d,null,c,!1,!0);h.text=d.getShallow("show")?Object(s["P"])(t.getFormattedLabel(r,BB),pS(e,r)):null;var f=ru(d,c,!1);return L(n,u),u=pB(u,h,f),n&&k(u,n),u.legacy=!0,u}function S(n,r){null==r&&(r=l);var i=_(r,FB).getItemStyle(),o=b(r,FB),a=nu(o,null,null,!0,!0);a.text=o.getShallow("show")?Object(s["Q"])(t.getFormattedLabel(r,FB),t.getFormattedLabel(r,BB),pS(e,r)):null;var u=ru(o,null,!0);return L(n,i),i=pB(i,a,u),n&&k(i,n),i.legacy=!0,i}function k(t,e){for(var n in e)Object(s["q"])(e,n)&&(t[n]=e[n])}function L(t,e){t&&(t.textFill&&(e.textFill=t.textFill),t.textPosition&&(e.textPosition=t.textPosition))}function T(t,n){if(null==n&&(n=l),Object(s["q"])(JF,t)){var r=e.getItemVisual(n,"style");return r?r[JF[t]]:null}if(Object(s["q"])(KF,t))return e.getItemVisual(n,t)}function D(t){if("cartesian2d"===o.type){var e=o.getBaseAxis();return yk(Object(s["i"])({axis:e},t))}}function O(){return n.getCurrentSeriesIndices()}function I(t){return cu(t,n)}}function uW(t){var e={};return Object(s["k"])(t.dimensions,(function(n){var r=t.getDimensionInfo(n);if(!r.isExtraCoord){var i=r.coordDim,o=e[i]=e[i]||[];o[r.coordDimIndex]=t.getDimensionIndex(n)}})),e}function cW(t,e,n,r,i,o,a){if(r){var s=dW(t,e,n,r,i,o);return s&&a.setItemGraphicEl(n,s),s&&Dl(s,r.focus,r.blurScope,r.emphasisDisabled),s}o.remove(e)}function dW(t,e,n,r,i,o){var a=-1,l=e;e&&hW(e,r,i)&&(a=Object(s["r"])(o.childrenRef(),e),e=null);var u=!e,c=e;c?c.clearStates():(c=rW(r),l&&tW(l,c)),!1===r.morph?c.disableMorphing=!0:c.disableMorphing&&(c.disableMorphing=!1),ZB.normal.cfg=ZB.normal.conOpt=ZB.emphasis.cfg=ZB.emphasis.conOpt=ZB.blur.cfg=ZB.blur.conOpt=ZB.select.cfg=ZB.select.conOpt=null,ZB.isLegacy=!1,pW(c,n,r,i,u,ZB),fW(c,n,r,i,u),iW(t,c,n,r,ZB,i,u),Object(s["q"])(r,"info")&&(QF(c).info=r.info);for(var d=0;d<GB.length;d++){var h=GB[d];if(h!==BB){var f=gW(r,h),p=yW(r,f,h);oW(h,c,f,p,ZB)}}return aW(c,r,i),"group"===r.type&&vW(t,c,n,r,i),a>=0?o.replaceAt(c,a):o.add(c),c}function hW(t,e,n){var r=QF(t),i=e.type,o=e.shape,a=e.style;return n.isUniversalTransitionEnabled()||null!=i&&i!==r.customGraphicType||"path"===i&&kW(o)&&SW(o)!==r.customPathData||"image"===i&&Object(s["q"])(a,"image")&&a.image!==r.customImagePath}function fW(t,e,n,r,i){var o=n.clipPath;if(!1===o)t&&t.getClipPath()&&t.removeClipPath();else if(o){var a=t.getClipPath();a&&hW(a,o,r)&&(a=null),a||(a=rW(o),t.setClipPath(a)),iW(null,a,e,o,null,r,i)}}function pW(t,e,n,r,i,o){if(!t.isGroup){mW(n,null,o),mW(n,FB,o);var a=o.normal.conOpt,s=o.emphasis.conOpt,l=o.blur.conOpt,u=o.select.conOpt;if(null!=a||null!=s||null!=u||null!=l){var c=t.getTextContent();if(!1===a)c&&t.removeTextContent();else{a=o.normal.conOpt=a||{type:"text"},c?c.clearStates():(c=rW(a),t.setTextContent(c)),iW(null,c,e,a,null,r,i);for(var d=a&&a.style,h=0;h<GB.length;h++){var f=GB[h];if(f!==BB){var p=o[f].conOpt;oW(f,c,p,yW(a,p,f),null)}}d?c.dirty():c.markRedraw()}}}}function mW(t,e,n){var r=e?gW(t,e):t,i=e?yW(t,r,FB):t.style,o=t.type,a=r?r.textConfig:null,s=t.textContent,l=s?e?gW(s,e):s:null;if(i&&(n.isLegacy||dB(i,o,!!a,!!l))){n.isLegacy=!0;var u=hB(i,o,!e);!a&&u.textConfig&&(a=u.textConfig),!l&&u.textContent&&(l=u.textContent)}if(!e&&l){var c=l;!c.type&&(c.type="text")}var d=e?n[e]:n.normal;d.cfg=a,d.conOpt=l}function gW(t,e){return e?t?t[e]:null:t}function yW(t,e,n){var r=e&&e.style;return null==r&&n===FB&&t&&(r=t.styleEmphasis),r}function vW(t,e,n,r,i){var o=r.children,a=o?o.length:0,s=r.$mergeChildren,l="byName"===s||r.diffChildrenByName,u=!1===s;if(a||l||u)if(l)bW({api:t,oldChildren:e.children()||[],newChildren:o||[],dataIndex:n,seriesModel:i,group:e});else{u&&e.removeAll();for(var c=0;c<a;c++){var d=o[c],h=e.childAt(c);d?(null==d.ignore&&(d.ignore=!1),dW(t,h,n,d,i,e)):h.ignore=!0}for(var f=e.childCount()-1;f>=c;f--){var p=e.childAt(f);_W(e,p,i)}}}function _W(t,e,n){e&&MB(e,QF(t).option,n)}function bW(t){new yM(t.oldChildren,t.newChildren,xW,xW,t).add(wW).update(wW).remove(MW).execute()}function xW(t,e){var n=t&&t.name;return null!=n?n:XB+e}function wW(t,e){var n=this.context,r=null!=t?n.newChildren[t]:null,i=null!=e?n.oldChildren[e]:null;dW(n.api,i,n.dataIndex,r,n.seriesModel,n.group)}function MW(t){var e=this.context,n=e.oldChildren[t];n&&MB(n,QF(n).option,e.seriesModel)}function SW(t){return t&&(t.pathData||t.d)}function kW(t){return t&&(Object(s["q"])(t,"pathData")||Object(s["q"])(t,"d"))}function LW(t){t.registerChartView(nW),t.registerSeriesModel(eB)}var TW=Ci(),DW=s["d"],OW=s["c"],IW=function(){function t(){this._dragging=!1,this.animationThreshold=15}return t.prototype.render=function(t,e,n,r){var i=e.get("value"),o=e.get("status");if(this._axisModel=t,this._axisPointerModel=e,this._api=n,r||this._lastValue!==i||this._lastStatus!==o){this._lastValue=i,this._lastStatus=o;var a=this._group,l=this._handle;if(!o||"hide"===o)return a&&a.hide(),void(l&&l.hide());a&&a.show(),l&&l.show();var u={};this.makeElOption(u,i,t,e,n);var c=u.graphicKey;c!==this._lastGraphicKey&&this.clear(n),this._lastGraphicKey=c;var d=this._moveAnimation=this.determineAnimation(t,e);if(a){var h=s["h"](CW,e,d);this.updatePointerEl(a,u,h),this.updateLabelEl(a,u,h,e)}else a=this._group=new wr,this.createPointerEl(a,u,t,e),this.createLabelEl(a,u,t,e),n.getZr().add(a);PW(a,e,!0),this._renderHandle(i)}},t.prototype.remove=function(t){this.clear(t)},t.prototype.dispose=function(t){this.clear(t)},t.prototype.determineAnimation=function(t,e){var n=e.get("animation"),r=t.axis,i="category"===r.type,o=e.get("snap");if(!o&&!i)return!1;if("auto"===n||null==n){var a=this.animationThreshold;if(i&&r.getBandWidth()>a)return!0;if(o){var s=mO(t).seriesDataCount,l=r.getExtent();return Math.abs(l[0]-l[1])/s>a}return!1}return!0===n},t.prototype.makeElOption=function(t,e,n,r,i){},t.prototype.createPointerEl=function(t,e,n,i){var o=e.pointer;if(o){var a=TW(t).pointerEl=new r[o.type](DW(e.pointer));t.add(a)}},t.prototype.createLabelEl=function(t,e,n,r){if(e.label){var i=TW(t).labelEl=new ks(DW(e.label));t.add(i),YW(i,r)}},t.prototype.updatePointerEl=function(t,e,n){var r=TW(t).pointerEl;r&&e.pointer&&(r.setStyle(e.pointer.style),n(r,{shape:e.pointer.shape}))},t.prototype.updateLabelEl=function(t,e,n,r){var i=TW(t).labelEl;i&&(i.setStyle(e.label.style),n(i,{x:e.label.x,y:e.label.y}),YW(i,r))},t.prototype._renderHandle=function(t){if(!this._dragging&&this.updateHandleTransform){var e,n=this._axisPointerModel,r=this._api.getZr(),i=this._handle,o=n.getModel("handle"),a=n.get("status");if(!o.get("show")||!a||"hide"===a)return i&&r.remove(i),void(this._handle=null);this._handle||(e=!0,i=this._handle=Ig(o.get("icon"),{cursor:"move",draggable:!0,onmousemove:function(t){et(t.event)},onmousedown:OW(this._onHandleDragMove,this,0,0),drift:OW(this._onHandleDragMove,this),ondragend:OW(this._onHandleDragEnd,this)}),r.add(i)),PW(i,n,!1),i.setStyle(o.getItemStyle(null,["color","borderColor","borderWidth","opacity","shadowColor","shadowBlur","shadowOffsetX","shadowOffsetY"]));var l=o.get("size");s["t"](l)||(l=[l,l]),i.scaleX=l[0]/2,i.scaleY=l[1]/2,Jg(this,"_doDispatchAxisPointer",o.get("throttle")||0,"fixRate"),this._moveHandleToValue(t,e)}},t.prototype._moveHandleToValue=function(t,e){CW(this._axisPointerModel,!e&&this._moveAnimation,this._handle,jW(this.getHandleTransform(t,this._axisModel,this._axisPointerModel)))},t.prototype._onHandleDragMove=function(t,e){var n=this._handle;if(n){this._dragging=!0;var r=this.updateHandleTransform(jW(n),[t,e],this._axisModel,this._axisPointerModel);this._payloadInfo=r,n.stopAnimation(),n.attr(jW(r)),TW(n).lastProp=null,this._doDispatchAxisPointer()}},t.prototype._doDispatchAxisPointer=function(){var t=this._handle;if(t){var e=this._payloadInfo,n=this._axisModel;this._api.dispatchAction({type:"updateAxisPointer",x:e.cursorPoint[0],y:e.cursorPoint[1],tooltipOption:e.tooltipOption,axesInfo:[{axisDim:n.axis.dim,axisIndex:n.componentIndex}]})}},t.prototype._onHandleDragEnd=function(){this._dragging=!1;var t=this._handle;if(t){var e=this._axisPointerModel.get("value");this._moveHandleToValue(e),this._api.dispatchAction({type:"hideTip"})}},t.prototype.clear=function(t){this._lastValue=null,this._lastStatus=null;var e=t.getZr(),n=this._group,r=this._handle;e&&n&&(this._lastGraphicKey=null,n&&e.remove(n),r&&e.remove(r),this._group=null,this._handle=null,this._payloadInfo=null),Kg(this,"_doDispatchAxisPointer")},t.prototype.doClear=function(){},t.prototype.buildLabel=function(t,e,n){return n=n||0,{x:t[n],y:t[1-n],width:e[n],height:e[1-n]}},t}();function CW(t,e,n,r){AW(TW(n).lastProp,r)||(TW(n).lastProp=r,e?Wl(n,r,t):(n.stopAnimation(),n.attr(r)))}function AW(t,e){if(s["A"](t)&&s["A"](e)){var n=!0;return s["k"](e,(function(e,r){n=n&&AW(t[r],e)})),!!n}return t===e}function YW(t,e){t[e.get(["label","show"])?"show":"hide"]()}function jW(t){return{x:t.x||0,y:t.y||0,rotation:t.rotation||0}}function PW(t,e,n){var r=e.get("z"),i=e.get("zlevel");t&&t.traverse((function(t){"group"!==t.type&&(null!=r&&(t.z=r),null!=i&&(t.zlevel=i),t.silent=n)}))}var EW=IW;function NW(t){var e,n=t.get("type"),r=t.getModel(n+"Style");return"line"===n?(e=r.getLineStyle(),e.fill=null):"shadow"===n&&(e=r.getAreaStyle(),e.stroke=null),e}function RW(t,e,n,r,i){var o=n.get("value"),a=zW(o,e.axis,e.ecModel,n.get("seriesDataIndices"),{precision:n.get(["label","precision"]),formatter:n.get(["label","formatter"])}),s=n.getModel("label"),l=Sc(s.get("padding")||0),u=s.getFont(),c=er(a,u),d=i.position,h=c.width+l[1]+l[3],f=c.height+l[0]+l[2],p=i.align;"right"===p&&(d[0]-=h),"center"===p&&(d[0]-=h/2);var m=i.verticalAlign;"bottom"===m&&(d[1]-=f),"middle"===m&&(d[1]-=f/2),HW(d,h,f,r);var g=s.get("backgroundColor");g&&"auto"!==g||(g=e.get(["axisLine","lineStyle","color"])),t.label={x:d[0],y:d[1],style:nu(s,{text:a,font:u,fill:s.getTextColor(),padding:l,backgroundColor:g}),z2:10}}function HW(t,e,n,r){var i=r.getWidth(),o=r.getHeight();t[0]=Math.min(t[0]+e,i)-e,t[1]=Math.min(t[1]+n,o)-n,t[0]=Math.max(t[0],0),t[1]=Math.max(t[1],0)}function zW(t,e,n,r,i){t=e.scale.parse(t);var o=e.scale.getLabel({value:t},{precision:i.precision}),a=i.formatter;if(a){var l={value:rD(e,{value:t}),axisDimension:e.dim,axisIndex:e.index,seriesData:[]};s["k"](r,(function(t){var e=n.getSeriesByIndex(t.seriesIndex),r=t.dataIndexInside,i=e&&e.getDataParams(r);i&&l.seriesData.push(i)})),s["C"](a)?o=a.replace("{value}",o):s["w"](a)&&(o=a(l))}return o}function FW(t,e,n){var r=st();return ht(r,r,n.rotation),dt(r,r,n.position),Mg([t.dataToCoord(e),(n.labelOffset||0)+(n.labelDirection||1)*(n.labelMargin||0)],r)}function BW(t,e,n,r,i,o){var a=sO.innerTextLayout(n.rotation,0,n.labelDirection);n.labelMargin=i.get(["label","margin"]),RW(e,r,i,o,{position:FW(r.axis,t,n),align:a.textAlign,verticalAlign:a.textVerticalAlign})}function WW(t,e,n){return n=n||0,{x1:t[n],y1:t[1-n],x2:e[n],y2:e[1-n]}}function VW(t,e,n){return n=n||0,{x:t[n],y:t[1-n],width:e[n],height:e[1-n]}}function GW(t,e,n,r,i,o){return{cx:t,cy:e,r0:n,r:r,startAngle:i,endAngle:o,clockwise:!0}}var $W=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.makeElOption=function(t,e,n,r,i){var o=n.axis,a=o.grid,s=r.get("type"),l=UW(a,o).getOtherAxis(o).getGlobalExtent(),u=o.toGlobalCoord(o.dataToCoord(e,!0));if(s&&"none"!==s){var c=NW(r),d=qW[s](o,u,l);d.style=c,t.graphicKey=d.type,t.pointer=d}var h=RD(a.model,n);BW(e,t,h,n,r,i)},e.prototype.getHandleTransform=function(t,e,n){var r=RD(e.axis.grid.model,e,{labelInside:!1});r.labelMargin=n.get(["handle","margin"]);var i=FW(e.axis,t,r);return{x:i[0],y:i[1],rotation:r.rotation+(r.labelDirection<0?Math.PI:0)}},e.prototype.updateHandleTransform=function(t,e,n,r){var i=n.axis,o=i.grid,a=i.getGlobalExtent(!0),s=UW(o,i).getOtherAxis(i).getGlobalExtent(),l="x"===i.dim?0:1,u=[t.x,t.y];u[l]+=e[l],u[l]=Math.min(a[1],u[l]),u[l]=Math.max(a[0],u[l]);var c=(s[1]+s[0])/2,d=[c,c];d[l]=u[l];var h=[{verticalAlign:"middle"},{align:"center"}];return{x:u[0],y:u[1],rotation:t.rotation,cursorPoint:d,tooltipOption:h[l]}},e}(EW);function UW(t,e){var n={};return n[e.dim+"AxisIndex"]=e.index,t.getCartesian(n)}var qW={line:function(t,e,n){var r=WW([e,n[0]],[e,n[1]],XW(t));return{type:"Line",subPixelOptimize:!0,shape:r}},shadow:function(t,e,n){var r=Math.max(1,t.getBandWidth()),i=n[1]-n[0];return{type:"Rect",shape:VW([e-r/2,n[0]],[r,i],XW(t))}}};function XW(t){return"x"===t.dim?0:1}var ZW=$W,JW=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="axisPointer",e.defaultOption={show:"auto",z:50,type:"line",snap:!1,triggerTooltip:!0,triggerEmphasis:!0,value:null,status:null,link:[],animation:null,animationDurationUpdate:200,lineStyle:{color:"#B9BEC9",width:1,type:"dashed"},shadowStyle:{color:"rgba(210,219,238,0.2)"},label:{show:!0,formatter:null,precision:"auto",margin:3,color:"#fff",padding:[5,7,5,7],backgroundColor:"auto",borderColor:null,borderWidth:0,borderRadius:3},handle:{show:!1,icon:"M10.7,11.9v-1.3H9.3v1.3c-4.9,0.3-8.8,4.4-8.8,9.4c0,5,3.9,9.1,8.8,9.4h1.3c4.9-0.3,8.8-4.4,8.8-9.4C19.5,16.3,15.6,12.2,10.7,11.9z M13.3,24.4H6.7v-1.2h6.6z M13.3,22H6.7v-1.2h6.6z M13.3,19.6H6.7v-1.2h6.6z",size:45,margin:50,color:"#333",shadowBlur:3,shadowColor:"#aaa",shadowOffsetX:0,shadowOffsetY:2,throttle:40}},e}(Xc),KW=JW,QW=Ci(),tV=s["k"];function eV(t,e,n){if(!a["a"].node){var r=e.getZr();QW(r).records||(QW(r).records={}),nV(r,e);var i=QW(r).records[t]||(QW(r).records[t]={});i.handler=n}}function nV(t,e){function n(n,r){t.on(n,(function(n){var i=aV(e);tV(QW(t).records,(function(t){t&&r(t,n,i.dispatchAction)})),rV(i.pendings,e)}))}QW(t).initialized||(QW(t).initialized=!0,n("click",s["h"](oV,"click")),n("mousemove",s["h"](oV,"mousemove")),n("globalout",iV))}function rV(t,e){var n,r=t.showTip.length,i=t.hideTip.length;r?n=t.showTip[r-1]:i&&(n=t.hideTip[i-1]),n&&(n.dispatchAction=null,e.dispatchAction(n))}function iV(t,e,n){t.handler("leave",null,n)}function oV(t,e,n,r){e.handler(t,n,r)}function aV(t){var e={showTip:[],hideTip:[]},n=function(r){var i=e[r.type];i?i.push(r):(r.dispatchAction=n,t.dispatchAction(r))};return{dispatchAction:n,pendings:e}}function sV(t,e){if(!a["a"].node){var n=e.getZr(),r=(QW(n).records||{})[t];r&&(QW(n).records[t]=null)}}var lV=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){var r=e.getComponent("tooltip"),i=t.get("triggerOn")||r&&r.get("triggerOn")||"mousemove|click";eV("axisPointer",n,(function(t,e,n){"none"!==i&&("leave"===t||i.indexOf(t)>=0)&&n({type:"updateAxisPointer",currTrigger:t,x:e&&e.offsetX,y:e&&e.offsetY})}))},e.prototype.remove=function(t,e){sV("axisPointer",e)},e.prototype.dispose=function(t,e){sV("axisPointer",e)},e.type="axisPointer",e}(kp),uV=lV;function cV(t,e){var n,r=[],i=t.seriesIndex;if(null==i||!(n=e.getSeriesByIndex(i)))return{point:[]};var o=n.getData(),a=Ii(o,t);if(null==a||a<0||s["t"](a))return{point:[]};var l=o.getItemGraphicEl(a),u=n.coordinateSystem;if(n.getTooltipPosition)r=n.getTooltipPosition(a)||[];else if(u&&u.dataToPoint)if(t.isStacked){var c=u.getBaseAxis(),d=u.getOtherAxis(c),h=d.dim,f=c.dim,p="x"===h||"radius"===h?1:0,m=o.mapDimension(f),g=[];g[p]=o.get(m,a),g[1-p]=o.get(o.getCalculationInfo("stackResultDimension"),a),r=u.dataToPoint(g)||[]}else r=u.dataToPoint(o.getValues(s["H"](u.dimensions,(function(t){return o.mapDimension(t)})),a))||[];else if(l){var y=l.getBoundingRect().clone();y.applyTransform(l.transform),r=[y.x+y.width/2,y.y+y.height/2]}return{point:r,el:l}}var dV=Ci();function hV(t,e,n){var r=t.currTrigger,i=[t.x,t.y],o=t,a=t.dispatchAction||Object(s["c"])(n.dispatchAction,n),l=e.getComponent("axisPointer").coordSysAxesInfo;if(l){wV(i)&&(i=cV({seriesIndex:o.seriesIndex,dataIndex:o.dataIndex},e).point);var u=wV(i),c=o.axesInfo,d=l.axesInfo,h="leave"===r||wV(i),f={},p={},m={list:[],map:{}},g={showPointer:Object(s["h"])(mV,p),showTooltip:Object(s["h"])(gV,m)};Object(s["k"])(l.coordSysMap,(function(t,e){var n=u||t.containPoint(i);Object(s["k"])(l.coordSysAxesInfo[e],(function(t,e){var r=t.axis,o=bV(c,t);if(!h&&n&&(!c||o)){var a=o&&o.value;null!=a||u||(a=r.pointToData(i)),null!=a&&fV(t,a,g,!1,f)}}))}));var y={};return Object(s["k"])(d,(function(t,e){var n=t.linkGroup;n&&!p[e]&&Object(s["k"])(n.axesInfo,(function(e,r){var i=p[r];if(e!==t&&i){var o=i.value;n.mapper&&(o=t.axis.scale.parse(n.mapper(o,xV(e),xV(t)))),y[t.key]=o}}))})),Object(s["k"])(y,(function(t,e){fV(d[e],t,g,!0,f)})),yV(p,d,f),vV(m,i,t,a),_V(d,a,n),f}}function fV(t,e,n,r,i){var o=t.axis;if(!o.scale.isBlank()&&o.containData(e))if(t.involveSeries){var a=pV(e,t),l=a.payloadBatch,u=a.snapToValue;l[0]&&null==i.seriesIndex&&Object(s["m"])(i,l[0]),!r&&t.snap&&o.containData(u)&&null!=u&&(e=u),n.showPointer(t,e,l),n.showTooltip(t,a,u)}else n.showPointer(t,e)}function pV(t,e){var n=e.axis,r=n.dim,i=t,o=[],a=Number.MAX_VALUE,l=-1;return Object(s["k"])(e.seriesModels,(function(e,u){var c,d,h=e.getData().mapDimensionsAll(r);if(e.getAxisTooltipData){var f=e.getAxisTooltipData(h,t,n);d=f.dataIndices,c=f.nestestValue}else{if(d=e.getData().indicesOfNearest(h[0],t,"category"===n.type?.5:null),!d.length)return;c=e.getData().get(h[0],d[0])}if(null!=c&&isFinite(c)){var p=t-c,m=Math.abs(p);m<=a&&((m<a||p>=0&&l<0)&&(a=m,l=p,i=c,o.length=0),Object(s["k"])(d,(function(t){o.push({seriesIndex:e.seriesIndex,dataIndexInside:t,dataIndex:e.getData().getRawIndex(t)})})))}})),{payloadBatch:o,snapToValue:i}}function mV(t,e,n,r){t[e.key]={value:n,payloadBatch:r}}function gV(t,e,n,r){var i=n.payloadBatch,o=e.axis,a=o.model,s=e.axisPointerModel;if(e.triggerTooltip&&i.length){var l=e.coordSys.model,u=vO(l),c=t.map[u];c||(c=t.map[u]={coordSysId:l.id,coordSysIndex:l.componentIndex,coordSysType:l.type,coordSysMainType:l.mainType,dataByAxis:[]},t.list.push(c)),c.dataByAxis.push({axisDim:o.dim,axisIndex:a.componentIndex,axisType:a.type,axisId:a.id,value:r,valueLabelOpt:{precision:s.get(["label","precision"]),formatter:s.get(["label","formatter"])},seriesDataIndices:i.slice()})}}function yV(t,e,n){var r=n.axesInfo=[];Object(s["k"])(e,(function(e,n){var i=e.axisPointerModel.option,o=t[n];o?(!e.useHandle&&(i.status="show"),i.value=o.value,i.seriesDataIndices=(o.payloadBatch||[]).slice()):!e.useHandle&&(i.status="hide"),"show"===i.status&&r.push({axisDim:e.axis.dim,axisIndex:e.axis.model.componentIndex,value:i.value})}))}function vV(t,e,n,r){if(!wV(e)&&t.list.length){var i=((t.list[0].dataByAxis[0]||{}).seriesDataIndices||[])[0]||{};r({type:"showTip",escapeConnect:!0,x:e[0],y:e[1],tooltipOption:n.tooltipOption,position:n.position,dataIndexInside:i.dataIndexInside,dataIndex:i.dataIndex,seriesIndex:i.seriesIndex,dataByCoordSys:t.list})}else r({type:"hideTip"})}function _V(t,e,n){var r=n.getZr(),i="axisPointerLastHighlights",o=dV(r)[i]||{},a=dV(r)[i]={};Object(s["k"])(t,(function(t,e){var n=t.axisPointerModel.option;"show"===n.status&&t.triggerEmphasis&&Object(s["k"])(n.seriesDataIndices,(function(t){var e=t.seriesIndex+" | "+t.dataIndex;a[e]=t}))}));var l=[],u=[];Object(s["k"])(o,(function(t,e){!a[e]&&u.push(t)})),Object(s["k"])(a,(function(t,e){!o[e]&&l.push(t)})),u.length&&n.dispatchAction({type:"downplay",escapeConnect:!0,notBlur:!0,batch:u}),l.length&&n.dispatchAction({type:"highlight",escapeConnect:!0,notBlur:!0,batch:l})}function bV(t,e){for(var n=0;n<(t||[]).length;n++){var r=t[n];if(e.axis.dim===r.axisDim&&e.axis.model.componentIndex===r.axisIndex)return r}}function xV(t){var e=t.axis.model,n={},r=n.axisDim=t.axis.dim;return n.axisIndex=n[r+"AxisIndex"]=e.componentIndex,n.axisName=n[r+"AxisName"]=e.name,n.axisId=n[r+"AxisId"]=e.id,n}function wV(t){return!t||null==t[0]||isNaN(t[0])||null==t[1]||isNaN(t[1])}function MV(t){xO.registerAxisPointerClass("CartesianAxisPointer",ZW),t.registerComponentModel(KW),t.registerComponentView(uV),t.registerPreprocessor((function(t){if(t){(!t.axisPointer||0===t.axisPointer.length)&&(t.axisPointer={});var e=t.axisPointer.link;e&&!Object(s["t"])(e)&&(t.axisPointer.link=[e])}})),t.registerProcessor(t.PRIORITY.PROCESSOR.STATISTIC,(function(t,e){t.getComponent("axisPointer").coordSysAxesInfo=lO(t,e)})),t.registerAction({type:"updateAxisPointer",event:"updateAxisPointer",update:":updateAxisPointer"},hV)}function SV(t){jb(YO),jb(MV)}var kV=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.makeElOption=function(t,e,n,r,i){var o=n.axis;"angle"===o.dim&&(this.animationThreshold=Math.PI/18);var a=o.polar,s=a.getOtherAxis(o),l=s.getExtent(),u=o.dataToCoord(e),c=r.get("type");if(c&&"none"!==c){var d=NW(r),h=TV[c](o,a,u,l);h.style=d,t.graphicKey=h.type,t.pointer=h}var f=r.get(["label","margin"]),p=LV(e,n,r,a,f);RW(t,n,r,i,p)},e}(EW);function LV(t,e,n,r,i){var o=e.axis,a=o.dataToCoord(t),s=r.getAngleAxis().getExtent()[0];s=s/180*Math.PI;var l,u,c,d=r.getRadiusAxis().getExtent();if("radius"===o.dim){var h=st();ht(h,h,s),dt(h,h,[r.cx,r.cy]),l=Mg([a,-i],h);var f=e.getModel("axisLabel").get("rotate")||0,p=sO.innerTextLayout(s,f*Math.PI/180,-1);u=p.textAlign,c=p.textVerticalAlign}else{var m=d[1];l=r.coordToPoint([m+i,a]);var g=r.cx,y=r.cy;u=Math.abs(l[0]-g)/m<.3?"center":l[0]>g?"left":"right",c=Math.abs(l[1]-y)/m<.3?"middle":l[1]>y?"top":"bottom"}return{position:l,align:u,verticalAlign:c}}var TV={line:function(t,e,n,r){return"angle"===t.dim?{type:"Line",shape:WW(e.coordToPoint([r[0],n]),e.coordToPoint([r[1],n]))}:{type:"Circle",shape:{cx:e.cx,cy:e.cy,r:n}}},shadow:function(t,e,n,r){var i=Math.max(1,t.getBandWidth()),o=Math.PI/180;return"angle"===t.dim?{type:"Sector",shape:GW(e.cx,e.cy,r[0],r[1],(-n-i/2)*o,(i/2-n)*o)}:{type:"Sector",shape:GW(e.cx,e.cy,n-i/2,n+i/2,0,2*Math.PI)}}},DV=kV,OV=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.findAxisModel=function(t){var e,n=this.ecModel;return n.eachComponent(t,(function(t){t.getCoordSysModel()===this&&(e=t)}),this),e},e.type="polar",e.dependencies=["radiusAxis","angleAxis"],e.defaultOption={z:0,center:["50%","50%"],radius:"80%"},e}(Xc),IV=OV,CV=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.getCoordSysModel=function(){return this.getReferringComponents("polar",Pi).models[0]},e.type="polarAxis",e}(Xc);s["K"](CV,$L);var AV=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="angleAxis",e}(CV),YV=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="radiusAxis",e}(CV),jV=function(t){function e(e,n){return t.call(this,"radius",e,n)||this}return o(e,t),e.prototype.pointToData=function(t,e){return this.polar.pointToData(t,e)["radius"===this.dim?0:1]},e}(PD);jV.prototype.dataToRadius=PD.prototype.dataToCoord,jV.prototype.radiusToData=PD.prototype.coordToData;var PV=jV,EV=Ci(),NV=function(t){function e(e,n){return t.call(this,"angle",e,n||[0,360])||this}return o(e,t),e.prototype.pointToData=function(t,e){return this.polar.pointToData(t,e)["radius"===this.dim?0:1]},e.prototype.calculateCategoryInterval=function(){var t=this,e=t.getLabelModel(),n=t.scale,r=n.getExtent(),i=n.count();if(r[1]-r[0]<1)return 0;var o=r[0],a=t.dataToCoord(o+1)-t.dataToCoord(o),s=Math.abs(a),l=er(null==o?"":o+"",e.getFont(),"center","top"),u=Math.max(l.height,7),c=u/s;isNaN(c)&&(c=1/0);var d=Math.max(0,Math.floor(c)),h=EV(t.model),f=h.lastAutoInterval,p=h.lastTickCount;return null!=f&&null!=p&&Math.abs(f-d)<=1&&Math.abs(p-i)<=1&&f>d?d=f:(h.lastTickCount=i,h.lastAutoInterval=d),d},e}(PD);NV.prototype.dataToAngle=PD.prototype.dataToCoord,NV.prototype.angleToData=PD.prototype.coordToData;var RV=NV,HV=["radius","angle"],zV=function(){function t(t){this.dimensions=HV,this.type="polar",this.cx=0,this.cy=0,this._radiusAxis=new PV,this._angleAxis=new RV,this.axisPointerEnabled=!0,this.name=t||"",this._radiusAxis.polar=this._angleAxis.polar=this}return t.prototype.containPoint=function(t){var e=this.pointToCoord(t);return this._radiusAxis.contain(e[0])&&this._angleAxis.contain(e[1])},t.prototype.containData=function(t){return this._radiusAxis.containData(t[0])&&this._angleAxis.containData(t[1])},t.prototype.getAxis=function(t){var e="_"+t+"Axis";return this[e]},t.prototype.getAxes=function(){return[this._radiusAxis,this._angleAxis]},t.prototype.getAxesByScale=function(t){var e=[],n=this._angleAxis,r=this._radiusAxis;return n.scale.type===t&&e.push(n),r.scale.type===t&&e.push(r),e},t.prototype.getAngleAxis=function(){return this._angleAxis},t.prototype.getRadiusAxis=function(){return this._radiusAxis},t.prototype.getOtherAxis=function(t){var e=this._angleAxis;return t===e?this._radiusAxis:e},t.prototype.getBaseAxis=function(){return this.getAxesByScale("ordinal")[0]||this.getAxesByScale("time")[0]||this.getAngleAxis()},t.prototype.getTooltipAxes=function(t){var e=null!=t&&"auto"!==t?this.getAxis(t):this.getBaseAxis();return{baseAxes:[e],otherAxes:[this.getOtherAxis(e)]}},t.prototype.dataToPoint=function(t,e){return this.coordToPoint([this._radiusAxis.dataToRadius(t[0],e),this._angleAxis.dataToAngle(t[1],e)])},t.prototype.pointToData=function(t,e){var n=this.pointToCoord(t);return[this._radiusAxis.radiusToData(n[0],e),this._angleAxis.angleToData(n[1],e)]},t.prototype.pointToCoord=function(t){var e=t[0]-this.cx,n=t[1]-this.cy,r=this.getAngleAxis(),i=r.getExtent(),o=Math.min(i[0],i[1]),a=Math.max(i[0],i[1]);r.inverse?o=a-360:a=o+360;var s=Math.sqrt(e*e+n*n);e/=s,n/=s;var l=Math.atan2(-n,e)/Math.PI*180,u=l<o?1:-1;while(l<o||l>a)l+=360*u;return[s,l]},t.prototype.coordToPoint=function(t){var e=t[0],n=t[1]/180*Math.PI,r=Math.cos(n)*e+this.cx,i=-Math.sin(n)*e+this.cy;return[r,i]},t.prototype.getArea=function(){var t=this.getAngleAxis(),e=this.getRadiusAxis(),n=e.getExtent().slice();n[0]>n[1]&&n.reverse();var r=t.getExtent(),i=Math.PI/180;return{cx:this.cx,cy:this.cy,r0:n[0],r:n[1],startAngle:-r[0]*i,endAngle:-r[1]*i,clockwise:t.inverse,contain:function(t,e){var n=t-this.cx,r=e-this.cy,i=n*n+r*r-1e-4,o=this.r,a=this.r0;return i<=o*o&&i>=a*a}}},t.prototype.convertToPixel=function(t,e,n){var r=FV(e);return r===this?this.dataToPoint(n):null},t.prototype.convertFromPixel=function(t,e,n){var r=FV(e);return r===this?this.pointToData(n):null},t}();function FV(t){var e=t.seriesModel,n=t.polarModel;return n&&n.coordinateSystem||e&&e.coordinateSystem}var BV=zV;function WV(t,e,n){var r=e.get("center"),i=n.getWidth(),o=n.getHeight();t.cx=Nr(r[0],i),t.cy=Nr(r[1],o);var a=t.getRadiusAxis(),l=Math.min(i,o)/2,u=e.get("radius");null==u?u=[0,"100%"]:s["t"](u)||(u=[0,u]);var c=[Nr(u[0],l),Nr(u[1],l)];a.inverse?a.setExtent(c[1],c[0]):a.setExtent(c[0],c[1])}function VV(t,e){var n=this,r=n.getAngleAxis(),i=n.getRadiusAxis();if(r.scale.setExtent(1/0,-1/0),i.scale.setExtent(1/0,-1/0),t.eachSeries((function(t){if(t.coordinateSystem===n){var e=t.getData();s["k"](lD(e,"radius"),(function(t){i.scale.unionExtentFromData(e,t)})),s["k"](lD(e,"angle"),(function(t){r.scale.unionExtentFromData(e,t)}))}})),QT(r.scale,r.model),QT(i.scale,i.model),"category"===r.type&&!r.onBand){var o=r.getExtent(),a=360/r.scale.count();r.inverse?o[1]+=a:o[1]-=a,r.setExtent(o[0],o[1])}}function GV(t){return"angleAxis"===t.mainType}function $V(t,e){var n;if(t.type=e.get("type"),t.scale=tD(e),t.onBand=e.get("boundaryGap")&&"category"===t.type,t.inverse=e.get("inverse"),GV(e)){t.inverse=t.inverse!==e.get("clockwise");var r=e.get("startAngle"),i=null!==(n=e.get("endAngle"))&&void 0!==n?n:r+(t.inverse?-360:360);t.setExtent(r,i)}e.axis=t,t.model=e}var UV={dimensions:HV,create:function(t,e){var n=[];return t.eachComponent("polar",(function(t,r){var i=new BV(r+"");i.update=VV;var o=i.getRadiusAxis(),a=i.getAngleAxis(),s=t.findAxisModel("radiusAxis"),l=t.findAxisModel("angleAxis");$V(o,s),$V(a,l),WV(i,t,e),n.push(i),t.coordinateSystem=i,i.model=t})),t.eachSeries((function(t){if("polar"===t.get("coordinateSystem")){var e=t.getReferringComponents("polar",Pi).models[0];0,t.coordinateSystem=e.coordinateSystem}})),n}},qV=UV,XV=["axisLine","axisLabel","axisTick","minorTick","splitLine","minorSplitLine","splitArea"];function ZV(t,e,n){e[1]>e[0]&&(e=e.slice().reverse());var r=t.coordToPoint([e[0],n]),i=t.coordToPoint([e[1],n]);return{x1:r[0],y1:r[1],x2:i[0],y2:i[1]}}function JV(t){var e=t.getRadiusAxis();return e.inverse?0:1}function KV(t){var e=t[0],n=t[t.length-1];e&&n&&Math.abs(Math.abs(e.coord-n.coord)-360)<1e-4&&t.pop()}var QV=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.axisPointerClass="PolarAxisPointer",n}return o(e,t),e.prototype.render=function(t,e){if(this.group.removeAll(),t.get("show")){var n=t.axis,r=n.polar,i=r.getRadiusAxis().getExtent(),o=n.getTicksCoords(),a=n.getMinorTicksCoords(),l=s["H"](n.getViewLabels(),(function(t){t=s["d"](t);var e=n.scale,r="ordinal"===e.type?e.getRawOrdinalNumber(t.tickValue):t.tickValue;return t.coord=n.dataToCoord(r),t}));KV(l),KV(o),s["k"](XV,(function(e){!t.get([e,"show"])||n.scale.isBlank()&&"axisLine"!==e||tG[e](this.group,t,r,o,a,i,l)}),this)}},e.type="angleAxis",e}(xO),tG={axisLine:function(t,e,n,i,o,a){var s,l=e.getModel(["axisLine","lineStyle"]),u=n.getAngleAxis(),c=Math.PI/180,d=u.getExtent(),h=JV(n),f=h?0:1,p=360===Math.abs(d[1]-d[0])?"Circle":"Arc";s=0===a[f]?new r[p]({shape:{cx:n.cx,cy:n.cy,r:a[h],startAngle:-d[0]*c,endAngle:-d[1]*c,clockwise:u.inverse},style:l.getLineStyle(),z2:1,silent:!0}):new wm({shape:{cx:n.cx,cy:n.cy,r:a[h],r0:a[f]},style:l.getLineStyle(),z2:1,silent:!0}),s.style.fill=null,t.add(s)},axisTick:function(t,e,n,r,i,o){var a=e.getModel("axisTick"),l=(a.get("inside")?-1:1)*a.get("length"),u=o[JV(n)],c=s["H"](r,(function(t){return new jm({shape:ZV(n,[u,u+l],t.coord)})}));t.add(yg(c,{style:s["i"](a.getModel("lineStyle").getLineStyle(),{stroke:e.get(["axisLine","lineStyle","color"])})}))},minorTick:function(t,e,n,r,i,o){if(i.length){for(var a=e.getModel("axisTick"),l=e.getModel("minorTick"),u=(a.get("inside")?-1:1)*l.get("length"),c=o[JV(n)],d=[],h=0;h<i.length;h++)for(var f=0;f<i[h].length;f++)d.push(new jm({shape:ZV(n,[c,c+u],i[h][f].coord)}));t.add(yg(d,{style:s["i"](l.getModel("lineStyle").getLineStyle(),s["i"](a.getLineStyle(),{stroke:e.get(["axisLine","lineStyle","color"])}))}))}},axisLabel:function(t,e,n,r,i,o,a){var l=e.getCategories(!0),u=e.getModel("axisLabel"),c=u.get("margin"),d=e.get("triggerEvent");s["k"](a,(function(r,i){var a=u,h=r.tickValue,f=o[JV(n)],p=n.coordToPoint([f+c,r.coord]),m=n.cx,g=n.cy,y=Math.abs(p[0]-m)/f<.3?"center":p[0]>m?"left":"right",v=Math.abs(p[1]-g)/f<.3?"middle":p[1]>g?"top":"bottom";if(l&&l[h]){var _=l[h];s["A"](_)&&_.textStyle&&(a=new Lu(_.textStyle,u,u.ecModel))}var b=new ks({silent:sO.isLabelSilent(e),style:nu(a,{x:p[0],y:p[1],fill:a.getTextColor()||e.get(["axisLine","lineStyle","color"]),text:r.formattedLabel,align:y,verticalAlign:v})});if(t.add(b),d){var x=sO.makeAxisEventDataBase(e);x.targetType="axisLabel",x.value=r.rawLabel,Ls(b).eventData=x}}),this)},splitLine:function(t,e,n,r,i,o){var a=e.getModel("splitLine"),l=a.getModel("lineStyle"),u=l.get("color"),c=0;u=u instanceof Array?u:[u];for(var d=[],h=0;h<r.length;h++){var f=c++%u.length;d[f]=d[f]||[],d[f].push(new jm({shape:ZV(n,o,r[h].coord)}))}for(h=0;h<d.length;h++)t.add(yg(d[h],{style:s["i"]({stroke:u[h%u.length]},l.getLineStyle()),silent:!0,z:e.get("z")}))},minorSplitLine:function(t,e,n,r,i,o){if(i.length){for(var a=e.getModel("minorSplitLine"),s=a.getModel("lineStyle"),l=[],u=0;u<i.length;u++)for(var c=0;c<i[u].length;c++)l.push(new jm({shape:ZV(n,o,i[u][c].coord)}));t.add(yg(l,{style:s.getLineStyle(),silent:!0,z:e.get("z")}))}},splitArea:function(t,e,n,r,i,o){if(r.length){var a=e.getModel("splitArea"),l=a.getModel("areaStyle"),u=l.get("color"),c=0;u=u instanceof Array?u:[u];for(var d=[],h=Math.PI/180,f=-r[0].coord*h,p=Math.min(o[0],o[1]),m=Math.max(o[0],o[1]),g=e.get("clockwise"),y=1,v=r.length;y<=v;y++){var _=y===v?r[0].coord:r[y].coord,b=c++%u.length;d[b]=d[b]||[],d[b].push(new _m({shape:{cx:n.cx,cy:n.cy,r0:p,r:m,startAngle:f,endAngle:-_*h,clockwise:g},silent:!0})),f=-_*h}for(y=0;y<d.length;y++)t.add(yg(d[y],{style:s["i"]({fill:u[y%u.length]},l.getAreaStyle()),silent:!0}))}}},eG=QV,nG=["axisLine","axisTickLabel","axisName"],rG=["splitLine","splitArea","minorSplitLine"],iG=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.axisPointerClass="PolarAxisPointer",n}return o(e,t),e.prototype.render=function(t,e){if(this.group.removeAll(),t.get("show")){var n=this._axisGroup,r=this._axisGroup=new wr;this.group.add(r);var i=t.axis,o=i.polar,a=o.getAngleAxis(),l=i.getTicksCoords(),u=i.getMinorTicksCoords(),c=a.getExtent()[0],d=i.getExtent(),h=aG(o,t,c),f=new sO(t,h);s["k"](nG,f.add,f),r.add(f.getGroup()),Tg(n,r,t),s["k"](rG,(function(e){t.get([e,"show"])&&!i.scale.isBlank()&&oG[e](this.group,t,o,c,d,l,u)}),this)}},e.type="radiusAxis",e}(xO),oG={splitLine:function(t,e,n,i,o,a){var l=e.getModel("splitLine"),u=l.getModel("lineStyle"),c=u.get("color"),d=0,h=n.getAngleAxis(),f=Math.PI/180,p=h.getExtent(),m=360===Math.abs(p[1]-p[0])?"Circle":"Arc";c=c instanceof Array?c:[c];for(var g=[],y=0;y<a.length;y++){var v=d++%c.length;g[v]=g[v]||[],g[v].push(new r[m]({shape:{cx:n.cx,cy:n.cy,r:Math.max(a[y].coord,0),startAngle:-p[0]*f,endAngle:-p[1]*f,clockwise:h.inverse}}))}for(y=0;y<g.length;y++)t.add(yg(g[y],{style:s["i"]({stroke:c[y%c.length],fill:null},u.getLineStyle()),silent:!0}))},minorSplitLine:function(t,e,n,r,i,o,a){if(a.length){for(var l=e.getModel("minorSplitLine"),u=l.getModel("lineStyle"),c=[],d=0;d<a.length;d++)for(var h=0;h<a[d].length;h++)c.push(new Kp({shape:{cx:n.cx,cy:n.cy,r:a[d][h].coord}}));t.add(yg(c,{style:s["i"]({fill:null},u.getLineStyle()),silent:!0}))}},splitArea:function(t,e,n,r,i,o){if(o.length){var a=e.getModel("splitArea"),l=a.getModel("areaStyle"),u=l.get("color"),c=0;u=u instanceof Array?u:[u];for(var d=[],h=o[0].coord,f=1;f<o.length;f++){var p=c++%u.length;d[p]=d[p]||[],d[p].push(new _m({shape:{cx:n.cx,cy:n.cy,r0:h,r:o[f].coord,startAngle:0,endAngle:2*Math.PI},silent:!0})),h=o[f].coord}for(f=0;f<d.length;f++)t.add(yg(d[f],{style:s["i"]({fill:u[f%u.length]},l.getAreaStyle()),silent:!0}))}}};function aG(t,e,n){return{position:[t.cx,t.cy],rotation:n/180*Math.PI,labelDirection:-1,tickDirection:-1,nameDirection:1,labelRotate:e.getModel("axisLabel").get("rotate"),z2:1}}var sG=iG;function lG(t){return t.get("stack")||"__ec_stack_"+t.seriesIndex}function uG(t,e){return e.dim+t.model.componentIndex}function cG(t,e,n){var r={},i=dG(s["n"](e.getSeriesByType(t),(function(t){return!e.isSeriesFiltered(t)&&t.coordinateSystem&&"polar"===t.coordinateSystem.type})));e.eachSeriesByType(t,(function(t){if("polar"===t.coordinateSystem.type){var e=t.getData(),n=t.coordinateSystem,o=n.getBaseAxis(),a=uG(n,o),s=lG(t),l=i[a][s],u=l.offset,c=l.width,d=n.getOtherAxis(o),h=t.coordinateSystem.cx,f=t.coordinateSystem.cy,p=t.get("barMinHeight")||0,m=t.get("barMinAngle")||0;r[s]=r[s]||[];for(var g=e.mapDimension(d.dim),y=e.mapDimension(o.dim),v=iS(e,g),_="radius"!==o.dim||!t.get("roundCap",!0),b=d.dataToCoord(0),x=0,w=e.count();x<w;x++){var M=e.get(g,x),S=e.get(y,x),k=M>=0?"p":"n",L=b;v&&(r[s][S]||(r[s][S]={p:b,n:b}),L=r[s][S][k]);var T=void 0,D=void 0,O=void 0,I=void 0;if("radius"===d.dim){var C=d.dataToCoord(M)-b,A=o.dataToCoord(S);Math.abs(C)<p&&(C=(C<0?-1:1)*p),T=L,D=L+C,O=A-u,I=O-c,v&&(r[s][S][k]=D)}else{var Y=d.dataToCoord(M,_)-b,j=o.dataToCoord(S);Math.abs(Y)<m&&(Y=(Y<0?-1:1)*m),T=j+u,D=T+c,O=L,I=L+Y,v&&(r[s][S][k]=I)}e.setItemLayout(x,{cx:h,cy:f,r0:T,r:D,startAngle:-O*Math.PI/180,endAngle:-I*Math.PI/180,clockwise:O>=I})}}}))}function dG(t){var e={};s["k"](t,(function(t,n){var r=t.getData(),i=t.coordinateSystem,o=i.getBaseAxis(),a=uG(i,o),s=o.getExtent(),l="category"===o.type?o.getBandWidth():Math.abs(s[1]-s[0])/r.count(),u=e[a]||{bandWidth:l,remainedWidth:l,autoWidthCount:0,categoryGap:"20%",gap:"30%",stacks:{}},c=u.stacks;e[a]=u;var d=lG(t);c[d]||u.autoWidthCount++,c[d]=c[d]||{width:0,maxWidth:0};var h=Nr(t.get("barWidth"),l),f=Nr(t.get("barMaxWidth"),l),p=t.get("barGap"),m=t.get("barCategoryGap");h&&!c[d].width&&(h=Math.min(u.remainedWidth,h),c[d].width=h,u.remainedWidth-=h),f&&(c[d].maxWidth=f),null!=p&&(u.gap=p),null!=m&&(u.categoryGap=m)}));var n={};return s["k"](e,(function(t,e){n[e]={};var r=t.stacks,i=t.bandWidth,o=Nr(t.categoryGap,i),a=Nr(t.gap,1),l=t.remainedWidth,u=t.autoWidthCount,c=(l-o)/(u+(u-1)*a);c=Math.max(c,0),s["k"](r,(function(t,e){var n=t.maxWidth;n&&n<c&&(n=Math.min(n,l),t.width&&(n=Math.min(n,t.width)),l-=n,t.width=n,u--)})),c=(l-o)/(u+(u-1)*a),c=Math.max(c,0);var d,h=0;s["k"](r,(function(t,e){t.width||(t.width=c),d=t,h+=t.width*(1+a)})),d&&(h-=d.width*a);var f=-h/2;s["k"](r,(function(t,r){n[e][r]=n[e][r]||{offset:f,width:t.width},f+=t.width*(1+a)}))})),n}var hG=cG,fG={startAngle:90,clockwise:!0,splitNumber:12,axisLabel:{rotate:0}},pG={splitNumber:5},mG=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="polar",e}(kp);function gG(t){jb(MV),xO.registerAxisPointerClass("PolarAxisPointer",DV),t.registerCoordinateSystem("polar",qV),t.registerComponentModel(IV),t.registerComponentView(mG),oT(t,"angle",AV,fG),oT(t,"radius",YV,pG),t.registerComponentView(eG),t.registerComponentView(sG),t.registerLayout(Object(s["h"])(hG,"bar"))}function yG(t,e){e=e||{};var n=t.coordinateSystem,r=t.axis,i={},o=r.position,a=r.orient,l=n.getRect(),u=[l.x,l.x+l.width,l.y,l.y+l.height],c={horizontal:{top:u[2],bottom:u[3]},vertical:{left:u[0],right:u[1]}};i.position=["vertical"===a?c.vertical[o]:u[0],"horizontal"===a?c.horizontal[o]:u[3]];var d={horizontal:0,vertical:1};i.rotation=Math.PI/2*d[a];var h={top:-1,bottom:1,right:1,left:-1};i.labelDirection=i.tickDirection=i.nameDirection=h[o],t.get(["axisTick","inside"])&&(i.tickDirection=-i.tickDirection),s["O"](e.labelInside,t.get(["axisLabel","inside"]))&&(i.labelDirection=-i.labelDirection);var f=e.rotate;return null==f&&(f=t.get(["axisLabel","rotate"])),i.labelRotation="top"===o?-f:f,i.z2=1,i}var vG=["axisLine","axisTickLabel","axisName"],_G=["splitArea","splitLine"],bG=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.axisPointerClass="SingleAxisPointer",n}return o(e,t),e.prototype.render=function(e,n,r,i){var o=this.group;o.removeAll();var a=this._axisGroup;this._axisGroup=new wr;var l=yG(e),u=new sO(e,l);s["k"](vG,u.add,u),o.add(this._axisGroup),o.add(u.getGroup()),s["k"](_G,(function(t){e.get([t,"show"])&&xG[t](this,this.group,this._axisGroup,e)}),this),Tg(a,this._axisGroup,e),t.prototype.render.call(this,e,n,r,i)},e.prototype.remove=function(){SO(this)},e.type="singleAxis",e}(xO),xG={splitLine:function(t,e,n,r){var i=r.axis;if(!i.scale.isBlank()){var o=r.getModel("splitLine"),a=o.getModel("lineStyle"),l=a.get("color");l=l instanceof Array?l:[l];for(var u=a.get("width"),c=r.coordinateSystem.getRect(),d=i.isHorizontal(),h=[],f=0,p=i.getTicksCoords({tickModel:o}),m=[],g=[],y=0;y<p.length;++y){var v=i.toGlobalCoord(p[y].coord);d?(m[0]=v,m[1]=c.y,g[0]=v,g[1]=c.y+c.height):(m[0]=c.x,m[1]=v,g[0]=c.x+c.width,g[1]=v);var _=new jm({shape:{x1:m[0],y1:m[1],x2:g[0],y2:g[1]},silent:!0});_g(_.shape,u);var b=f++%l.length;h[b]=h[b]||[],h[b].push(_)}var x=a.getLineStyle(["color"]);for(y=0;y<h.length;++y)e.add(yg(h[y],{style:s["i"]({stroke:l[y%l.length]},x),silent:!0}))}},splitArea:function(t,e,n,r){MO(t,n,r,r)}},wG=bG,MG=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.getCoordSysModel=function(){return this},e.type="singleAxis",e.layoutMode="box",e.defaultOption={left:"5%",top:"5%",right:"5%",bottom:"5%",type:"value",position:"bottom",orient:"horizontal",axisLine:{show:!0,lineStyle:{width:1,type:"solid"}},tooltip:{show:!0},axisTick:{show:!0,length:6,lineStyle:{width:1}},axisLabel:{show:!0,interval:"auto"},splitLine:{show:!0,lineStyle:{type:"dashed",opacity:.2}}},e}(Xc);Object(s["K"])(MG,$L.prototype);var SG=MG,kG=function(t){function e(e,n,r,i,o){var a=t.call(this,e,n,r)||this;return a.type=i||"value",a.position=o||"bottom",a}return o(e,t),e.prototype.isHorizontal=function(){var t=this.position;return"top"===t||"bottom"===t},e.prototype.pointToData=function(t,e){return this.coordinateSystem.pointToData(t)[0]},e}(PD),LG=kG,TG=["single"],DG=function(){function t(t,e,n){this.type="single",this.dimension="single",this.dimensions=TG,this.axisPointerEnabled=!0,this.model=t,this._init(t,e,n)}return t.prototype._init=function(t,e,n){var r=this.dimension,i=new LG(r,tD(t),[0,0],t.get("type"),t.get("position")),o="category"===i.type;i.onBand=o&&t.get("boundaryGap"),i.inverse=t.get("inverse"),i.orient=t.get("orient"),t.axis=i,i.model=t,i.coordinateSystem=this,this._axis=i},t.prototype.update=function(t,e){t.eachSeries((function(t){if(t.coordinateSystem===this){var e=t.getData();Object(s["k"])(e.mapDimensionsAll(this.dimension),(function(t){this._axis.scale.unionExtentFromData(e,t)}),this),QT(this._axis.scale,this._axis.model)}}),this)},t.prototype.resize=function(t,e){this._rect=Hc({left:t.get("left"),top:t.get("top"),right:t.get("right"),bottom:t.get("bottom"),width:t.get("width"),height:t.get("height")},{width:e.getWidth(),height:e.getHeight()}),this._adjustAxis()},t.prototype.getRect=function(){return this._rect},t.prototype._adjustAxis=function(){var t=this._rect,e=this._axis,n=e.isHorizontal(),r=n?[0,t.width]:[0,t.height],i=e.inverse?1:0;e.setExtent(r[i],r[1-i]),this._updateAxisTransform(e,n?t.x:t.y)},t.prototype._updateAxisTransform=function(t,e){var n=t.getExtent(),r=n[0]+n[1],i=t.isHorizontal();t.toGlobalCoord=i?function(t){return t+e}:function(t){return r-t+e},t.toLocalCoord=i?function(t){return t-e}:function(t){return r-t+e}},t.prototype.getAxis=function(){return this._axis},t.prototype.getBaseAxis=function(){return this._axis},t.prototype.getAxes=function(){return[this._axis]},t.prototype.getTooltipAxes=function(){return{baseAxes:[this.getAxis()],otherAxes:[]}},t.prototype.containPoint=function(t){var e=this.getRect(),n=this.getAxis(),r=n.orient;return"horizontal"===r?n.contain(n.toLocalCoord(t[0]))&&t[1]>=e.y&&t[1]<=e.y+e.height:n.contain(n.toLocalCoord(t[1]))&&t[0]>=e.y&&t[0]<=e.y+e.height},t.prototype.pointToData=function(t){var e=this.getAxis();return[e.coordToData(e.toLocalCoord(t["horizontal"===e.orient?0:1]))]},t.prototype.dataToPoint=function(t){var e=this.getAxis(),n=this.getRect(),r=[],i="horizontal"===e.orient?0:1;return t instanceof Array&&(t=t[0]),r[i]=e.toGlobalCoord(e.dataToCoord(+t)),r[1-i]=0===i?n.y+n.height/2:n.x+n.width/2,r},t.prototype.convertToPixel=function(t,e,n){var r=OG(e);return r===this?this.dataToPoint(n):null},t.prototype.convertFromPixel=function(t,e,n){var r=OG(e);return r===this?this.pointToData(n):null},t}();function OG(t){var e=t.seriesModel,n=t.singleAxisModel;return n&&n.coordinateSystem||e&&e.coordinateSystem}var IG=DG;function CG(t,e){var n=[];return t.eachComponent("singleAxis",(function(r,i){var o=new IG(r,t,e);o.name="single_"+i,o.resize(r,e),r.coordinateSystem=o,n.push(o)})),t.eachSeries((function(t){if("singleAxis"===t.get("coordinateSystem")){var e=t.getReferringComponents("singleAxis",Pi).models[0];t.coordinateSystem=e&&e.coordinateSystem}})),n}var AG={create:CG,dimensions:TG},YG=AG,jG=["x","y"],PG=["width","height"],EG=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.makeElOption=function(t,e,n,r,i){var o=n.axis,a=o.coordinateSystem,s=HG(a,1-RG(o)),l=a.dataToPoint(e)[0],u=r.get("type");if(u&&"none"!==u){var c=NW(r),d=NG[u](o,l,s);d.style=c,t.graphicKey=d.type,t.pointer=d}var h=yG(n);BW(e,t,h,n,r,i)},e.prototype.getHandleTransform=function(t,e,n){var r=yG(e,{labelInside:!1});r.labelMargin=n.get(["handle","margin"]);var i=FW(e.axis,t,r);return{x:i[0],y:i[1],rotation:r.rotation+(r.labelDirection<0?Math.PI:0)}},e.prototype.updateHandleTransform=function(t,e,n,r){var i=n.axis,o=i.coordinateSystem,a=RG(i),s=HG(o,a),l=[t.x,t.y];l[a]+=e[a],l[a]=Math.min(s[1],l[a]),l[a]=Math.max(s[0],l[a]);var u=HG(o,1-a),c=(u[1]+u[0])/2,d=[c,c];return d[a]=l[a],{x:l[0],y:l[1],rotation:t.rotation,cursorPoint:d,tooltipOption:{verticalAlign:"middle"}}},e}(EW),NG={line:function(t,e,n){var r=WW([e,n[0]],[e,n[1]],RG(t));return{type:"Line",subPixelOptimize:!0,shape:r}},shadow:function(t,e,n){var r=t.getBandWidth(),i=n[1]-n[0];return{type:"Rect",shape:VW([e-r/2,n[0]],[r,i],RG(t))}}};function RG(t){return t.isHorizontal()?0:1}function HG(t,e){var n=t.getRect();return[n[jG[e]],n[jG[e]]+n[PG[e]]]}var zG=EG,FG=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="single",e}(kp);function BG(t){jb(MV),xO.registerAxisPointerClass("SingleAxisPointer",zG),t.registerComponentView(FG),t.registerComponentView(wG),t.registerComponentModel(SG),oT(t,"single",SG,SG.defaultOption),t.registerCoordinateSystem("single",YG)}var WG=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(e,n,r){var i=Vc(e);t.prototype.init.apply(this,arguments),VG(e,i)},e.prototype.mergeOption=function(e){t.prototype.mergeOption.apply(this,arguments),VG(this.option,e)},e.prototype.getCellSize=function(){return this.option.cellSize},e.type="calendar",e.defaultOption={z:2,left:80,top:60,cellSize:20,orient:"horizontal",splitLine:{show:!0,lineStyle:{color:"#000",width:1,type:"solid"}},itemStyle:{color:"#fff",borderWidth:1,borderColor:"#ccc"},dayLabel:{show:!0,firstDay:0,position:"start",margin:"50%",color:"#000"},monthLabel:{show:!0,position:"start",margin:5,align:"center",formatter:null,color:"#000"},yearLabel:{show:!0,position:null,margin:30,formatter:null,color:"#ccc",fontFamily:"sans-serif",fontWeight:"bolder",fontSize:20}},e}(Xc);function VG(t,e){var n,r=t.cellSize;n=s["t"](r)?r:t.cellSize=[r,r],1===n.length&&(n[1]=n[0]);var i=s["H"]([0,1],(function(t){return Fc(e,t)&&(n[t]="auto"),null!=n[t]&&"auto"!==n[t]}));Wc(t,e,{type:"box",ignoreSize:i})}var GG=WG,$G=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){var r=this.group;r.removeAll();var i=t.coordinateSystem,o=i.getRangeInfo(),a=i.getOrient(),s=e.getLocaleModel();this._renderDayRect(t,o,r),this._renderLines(t,o,a,r),this._renderYearText(t,o,a,r),this._renderMonthText(t,s,a,r),this._renderWeekText(t,s,o,a,r)},e.prototype._renderDayRect=function(t,e,n){for(var r=t.coordinateSystem,i=t.getModel("itemStyle").getItemStyle(),o=r.getCellWidth(),a=r.getCellHeight(),s=e.start.time;s<=e.end.time;s=r.getNextNDay(s,1).time){var l=r.dataToRect([s],!1).tl,u=new ss({shape:{x:l[0],y:l[1],width:o,height:a},cursor:"default",style:i});n.add(u)}},e.prototype._renderLines=function(t,e,n,r){var i=this,o=t.coordinateSystem,a=t.getModel(["splitLine","lineStyle"]).getLineStyle(),s=t.get(["splitLine","show"]),l=a.lineWidth;this._tlpoints=[],this._blpoints=[],this._firstDayOfMonth=[],this._firstDayPoints=[];for(var u=e.start,c=0;u.time<=e.end.time;c++){h(u.formatedDate),0===c&&(u=o.getDateInfo(e.start.y+"-"+e.start.m));var d=u.date;d.setMonth(d.getMonth()+1),u=o.getDateInfo(d)}function h(e){i._firstDayOfMonth.push(o.getDateInfo(e)),i._firstDayPoints.push(o.dataToRect([e],!1).tl);var l=i._getLinePointsOfOneWeek(t,e,n);i._tlpoints.push(l[0]),i._blpoints.push(l[l.length-1]),s&&i._drawSplitline(l,a,r)}h(o.getNextNDay(e.end.time,1).formatedDate),s&&this._drawSplitline(i._getEdgesPoints(i._tlpoints,l,n),a,r),s&&this._drawSplitline(i._getEdgesPoints(i._blpoints,l,n),a,r)},e.prototype._getEdgesPoints=function(t,e,n){var r=[t[0].slice(),t[t.length-1].slice()],i="horizontal"===n?0:1;return r[0][i]=r[0][i]-e/2,r[1][i]=r[1][i]+e/2,r},e.prototype._drawSplitline=function(t,e,n){var r=new Im({z2:20,shape:{points:t},style:e});n.add(r)},e.prototype._getLinePointsOfOneWeek=function(t,e,n){for(var r=t.coordinateSystem,i=r.getDateInfo(e),o=[],a=0;a<7;a++){var s=r.getNextNDay(i.time,a),l=r.dataToRect([s.time],!1);o[2*s.day]=l.tl,o[2*s.day+1]=l["horizontal"===n?"bl":"tr"]}return o},e.prototype._formatterLabel=function(t,e){return Object(s["C"])(t)&&t?Oc(t,e):Object(s["w"])(t)?t(e):e.nameMap},e.prototype._yearTextPositionControl=function(t,e,n,r,i){var o=e[0],a=e[1],s=["center","bottom"];"bottom"===r?(a+=i,s=["center","top"]):"left"===r?o-=i:"right"===r?(o+=i,s=["center","top"]):a-=i;var l=0;return"left"!==r&&"right"!==r||(l=Math.PI/2),{rotation:l,x:o,y:a,style:{align:s[0],verticalAlign:s[1]}}},e.prototype._renderYearText=function(t,e,n,r){var i=t.getModel("yearLabel");if(i.get("show")){var o=i.get("margin"),a=i.get("position");a||(a="horizontal"!==n?"top":"left");var s=[this._tlpoints[this._tlpoints.length-1],this._blpoints[0]],l=(s[0][0]+s[1][0])/2,u=(s[0][1]+s[1][1])/2,c="horizontal"===n?0:1,d={top:[l,s[c][1]],bottom:[l,s[1-c][1]],left:[s[1-c][0],u],right:[s[c][0],u]},h=e.start.y;+e.end.y>+e.start.y&&(h=h+"-"+e.end.y);var f=i.get("formatter"),p={start:e.start.y,end:e.end.y,nameMap:h},m=this._formatterLabel(f,p),g=new ks({z2:30,style:nu(i,{text:m})});g.attr(this._yearTextPositionControl(g,d[a],n,a,o)),r.add(g)}},e.prototype._monthTextPositionControl=function(t,e,n,r,i){var o="left",a="top",s=t[0],l=t[1];return"horizontal"===n?(l+=i,e&&(o="center"),"start"===r&&(a="bottom")):(s+=i,e&&(a="middle"),"start"===r&&(o="right")),{x:s,y:l,align:o,verticalAlign:a}},e.prototype._renderMonthText=function(t,e,n,r){var i=t.getModel("monthLabel");if(i.get("show")){var o=i.get("nameMap"),a=i.get("margin"),l=i.get("position"),u=i.get("align"),c=[this._tlpoints,this._blpoints];o&&!Object(s["C"])(o)||(o&&(e=Bu(o)||e),o=e.get(["time","monthAbbr"])||[]);var d="start"===l?0:1,h="horizontal"===n?0:1;a="start"===l?-a:a;for(var f="center"===u,p=0;p<c[d].length-1;p++){var m=c[d][p].slice(),g=this._firstDayOfMonth[p];if(f){var y=this._firstDayPoints[p];m[h]=(y[h]+c[0][p+1][h])/2}var v=i.get("formatter"),_=o[+g.m-1],b={yyyy:g.y,yy:(g.y+"").slice(2),MM:g.m,M:+g.m,nameMap:_},x=this._formatterLabel(v,b),w=new ks({z2:30,style:Object(s["m"])(nu(i,{text:x}),this._monthTextPositionControl(m,f,n,l,a))});r.add(w)}}},e.prototype._weekTextPositionControl=function(t,e,n,r,i){var o="center",a="middle",s=t[0],l=t[1],u="start"===n;return"horizontal"===e?(s=s+r+(u?1:-1)*i[0]/2,o=u?"right":"left"):(l=l+r+(u?1:-1)*i[1]/2,a=u?"bottom":"top"),{x:s,y:l,align:o,verticalAlign:a}},e.prototype._renderWeekText=function(t,e,n,r,i){var o=t.getModel("dayLabel");if(o.get("show")){var a=t.coordinateSystem,l=o.get("position"),u=o.get("nameMap"),c=o.get("margin"),d=a.getFirstDayOfWeek();if(!u||Object(s["C"])(u)){u&&(e=Bu(u)||e);var h=e.get(["time","dayOfWeekShort"]);u=h||Object(s["H"])(e.get(["time","dayOfWeekAbbr"]),(function(t){return t[0]}))}var f=a.getNextNDay(n.end.time,7-n.lweek).time,p=[a.getCellWidth(),a.getCellHeight()];c=Nr(c,Math.min(p[1],p[0])),"start"===l&&(f=a.getNextNDay(n.start.time,-(7+n.fweek)).time,c=-c);for(var m=0;m<7;m++){var g=a.getNextNDay(f,m),y=a.dataToRect([g.time],!1).center,v=m;v=Math.abs((m+d)%7);var _=new ks({z2:30,style:Object(s["m"])(nu(o,{text:u[v]}),this._weekTextPositionControl(y,r,l,c,p))});i.add(_)}}},e.type="calendar",e}(kp),UG=$G,qG=864e5,XG=function(){function t(e,n,r){this.type="calendar",this.dimensions=t.dimensions,this.getDimensionsInfo=t.getDimensionsInfo,this._model=e}return t.getDimensionsInfo=function(){return[{name:"time",type:"time"},"value"]},t.prototype.getRangeInfo=function(){return this._rangeInfo},t.prototype.getModel=function(){return this._model},t.prototype.getRect=function(){return this._rect},t.prototype.getCellWidth=function(){return this._sw},t.prototype.getCellHeight=function(){return this._sh},t.prototype.getOrient=function(){return this._orient},t.prototype.getFirstDayOfWeek=function(){return this._firstDayOfWeek},t.prototype.getDateInfo=function(t){t=Xr(t);var e=t.getFullYear(),n=t.getMonth()+1,r=n<10?"0"+n:""+n,i=t.getDate(),o=i<10?"0"+i:""+i,a=t.getDay();return a=Math.abs((a+7-this.getFirstDayOfWeek())%7),{y:e+"",m:r,d:o,day:a,time:t.getTime(),formatedDate:e+"-"+r+"-"+o,date:t}},t.prototype.getNextNDay=function(t,e){return e=e||0,0===e||(t=new Date(this.getDateInfo(t).time),t.setDate(t.getDate()+e)),this.getDateInfo(t)},t.prototype.update=function(t,e){this._firstDayOfWeek=+this._model.getModel("dayLabel").get("firstDay"),this._orient=this._model.get("orient"),this._lineWidth=this._model.getModel("itemStyle").getItemStyle().lineWidth||0,this._rangeInfo=this._getRangeInfo(this._initRangeOption());var n=this._rangeInfo.weeks||1,r=["width","height"],i=this._model.getCellSize().slice(),o=this._model.getBoxLayoutParams(),a="horizontal"===this._orient?[n,7]:[7,n];s["k"]([0,1],(function(t){c(i,t)&&(o[r[t]]=i[t]*a[t])}));var l={width:e.getWidth(),height:e.getHeight()},u=this._rect=Hc(o,l);function c(t,e){return null!=t[e]&&"auto"!==t[e]}s["k"]([0,1],(function(t){c(i,t)||(i[t]=u[r[t]]/a[t])})),this._sw=i[0],this._sh=i[1]},t.prototype.dataToPoint=function(t,e){s["t"](t)&&(t=t[0]),null==e&&(e=!0);var n=this.getDateInfo(t),r=this._rangeInfo,i=n.formatedDate;if(e&&!(n.time>=r.start.time&&n.time<r.end.time+qG))return[NaN,NaN];var o=n.day,a=this._getRangeInfo([r.start.time,i]).nthWeek;return"vertical"===this._orient?[this._rect.x+o*this._sw+this._sw/2,this._rect.y+a*this._sh+this._sh/2]:[this._rect.x+a*this._sw+this._sw/2,this._rect.y+o*this._sh+this._sh/2]},t.prototype.pointToData=function(t){var e=this.pointToDate(t);return e&&e.time},t.prototype.dataToRect=function(t,e){var n=this.dataToPoint(t,e);return{contentShape:{x:n[0]-(this._sw-this._lineWidth)/2,y:n[1]-(this._sh-this._lineWidth)/2,width:this._sw-this._lineWidth,height:this._sh-this._lineWidth},center:n,tl:[n[0]-this._sw/2,n[1]-this._sh/2],tr:[n[0]+this._sw/2,n[1]-this._sh/2],br:[n[0]+this._sw/2,n[1]+this._sh/2],bl:[n[0]-this._sw/2,n[1]+this._sh/2]}},t.prototype.pointToDate=function(t){var e=Math.floor((t[0]-this._rect.x)/this._sw)+1,n=Math.floor((t[1]-this._rect.y)/this._sh)+1,r=this._rangeInfo.range;return"vertical"===this._orient?this._getDateByWeeksAndDay(n,e-1,r):this._getDateByWeeksAndDay(e,n-1,r)},t.prototype.convertToPixel=function(t,e,n){var r=ZG(e);return r===this?r.dataToPoint(n):null},t.prototype.convertFromPixel=function(t,e,n){var r=ZG(e);return r===this?r.pointToData(n):null},t.prototype.containPoint=function(t){return!1},t.prototype._initRangeOption=function(){var t,e=this._model.get("range");if(s["t"](e)&&1===e.length&&(e=e[0]),s["t"](e))t=e;else{var n=e.toString();if(/^\d{4}$/.test(n)&&(t=[n+"-01-01",n+"-12-31"]),/^\d{4}[\/|-]\d{1,2}$/.test(n)){var r=this.getDateInfo(n),i=r.date;i.setMonth(i.getMonth()+1);var o=this.getNextNDay(i,-1);t=[r.formatedDate,o.formatedDate]}/^\d{4}[\/|-]\d{1,2}[\/|-]\d{1,2}$/.test(n)&&(t=[n,n])}if(!t)return e;var a=this._getRangeInfo(t);return a.start.time>a.end.time&&t.reverse(),t},t.prototype._getRangeInfo=function(t){var e,n=[this.getDateInfo(t[0]),this.getDateInfo(t[1])];n[0].time>n[1].time&&(e=!0,n.reverse());var r=Math.floor(n[1].time/qG)-Math.floor(n[0].time/qG)+1,i=new Date(n[0].time),o=i.getDate(),a=n[1].date.getDate();i.setDate(o+r-1);var s=i.getDate();if(s!==a){var l=i.getTime()-n[1].time>0?1:-1;while((s=i.getDate())!==a&&(i.getTime()-n[1].time)*l>0)r-=l,i.setDate(s-l)}var u=Math.floor((r+n[0].day+6)/7),c=e?1-u:u-1;return e&&n.reverse(),{range:[n[0].formatedDate,n[1].formatedDate],start:n[0],end:n[1],allDay:r,weeks:u,nthWeek:c,fweek:n[0].day,lweek:n[1].day}},t.prototype._getDateByWeeksAndDay=function(t,e,n){var r=this._getRangeInfo(n);if(t>r.weeks||0===t&&e<r.fweek||t===r.weeks&&e>r.lweek)return null;var i=7*(t-1)-r.fweek+e,o=new Date(r.start.time);return o.setDate(+r.start.d+i),this.getDateInfo(o)},t.create=function(e,n){var r=[];return e.eachComponent("calendar",(function(i){var o=new t(i,e,n);r.push(o),i.coordinateSystem=o})),e.eachSeries((function(t){"calendar"===t.get("coordinateSystem")&&(t.coordinateSystem=r[t.get("calendarIndex")||0])})),r},t.dimensions=["time","value"],t}();function ZG(t){var e=t.calendarModel,n=t.seriesModel,r=e?e.coordinateSystem:n?n.coordinateSystem:null;return r}var JG=XG;function KG(t){t.registerComponentModel(GG),t.registerComponentView(UG),t.registerCoordinateSystem("calendar",JG)}function QG(t,e){var n=t.existing;if(e.id=t.keyInfo.id,!e.type&&n&&(e.type=n.type),null==e.parentId){var r=e.parentOption;r?e.parentId=r.id:n&&(e.parentId=n.parentId)}e.parentOption=null}function t$(t,e){var n;return s["k"](e,(function(e){null!=t[e]&&"auto"!==t[e]&&(n=!0)})),n}function e$(t,e,n){var r=s["m"]({},n),i=t[e],o=n.$action||"merge";"merge"===o?i?(s["I"](i,r,!0),Wc(i,r,{ignoreSize:!0}),Gc(n,i),i$(n,i),i$(n,i,"shape"),i$(n,i,"style"),i$(n,i,"extra"),n.clipPath=i.clipPath):t[e]=r:"replace"===o?t[e]=r:"remove"===o&&i&&(t[e]=null)}var n$=["transition","enterFrom","leaveTo"],r$=n$.concat(["enterAnimation","updateAnimation","leaveAnimation"]);function i$(t,e,n){if(n&&(!t[n]&&e[n]&&(t[n]={}),t=t[n],e=e[n]),t&&e)for(var r=n?n$:r$,i=0;i<r.length;i++){var o=r[i];null==t[o]&&null!=e[o]&&(t[o]=e[o])}}function o$(t,e){if(t&&(t.hv=e.hv=[t$(e,["left","right"]),t$(e,["top","bottom"])],"group"===t.type)){var n=t,r=e;null==n.width&&(n.width=r.width=0),null==n.height&&(n.height=r.height=0)}}var a$=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.preventAutoZ=!0,n}return o(e,t),e.prototype.mergeOption=function(e,n){var r=this.option.elements;this.option.elements=null,t.prototype.mergeOption.call(this,e,n),this.option.elements=r},e.prototype.optionUpdated=function(t,e){var n=this.option,r=(e?n:t).elements,i=n.elements=e?[]:n.elements,o=[];this._flatten(r,o,null);var a=pi(i,o,"normalMerge"),l=this._elOptionsToUpdate=[];s["k"](a,(function(t,e){var n=t.newOption;n&&(l.push(n),QG(t,n),e$(i,e,n),o$(i[e],n))}),this),n.elements=s["n"](i,(function(t){return t&&delete t.$action,null!=t}))},e.prototype._flatten=function(t,e,n){s["k"](t,(function(t){if(t){n&&(t.parentOption=n),e.push(t);var r=t.children;r&&r.length&&this._flatten(r,e,t),delete t.children}}),this)},e.prototype.useElOptionsToUpdate=function(){var t=this._elOptionsToUpdate;return this._elOptionsToUpdate=null,t},e.type="graphic",e.defaultOption={elements:[]},e}(Xc),s$={path:null,compoundPath:null,group:wr,image:Ka,text:ks},l$=Ci(),u$=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(){this._elMap=s["f"]()},e.prototype.render=function(t,e,n){t!==this._lastGraphicModel&&this._clear(),this._lastGraphicModel=t,this._updateElements(t),this._relocate(t,n)},e.prototype._updateElements=function(t){var e=t.useElOptionsToUpdate();if(e){var n=this._elMap,r=this.group,i=t.get("z"),o=t.get("zlevel");s["k"](e,(function(e){var a=Mi(e.id,null),s=null!=a?n.get(a):null,l=Mi(e.parentId,null),u=null!=l?n.get(l):r,c=e.type,d=e.style;"text"===c&&d&&e.hv&&e.hv[1]&&(d.textVerticalAlign=d.textBaseline=d.verticalAlign=d.align=null);var h=e.textContent,f=e.textConfig;if(d&&dB(d,c,!!f,!!h)){var p=hB(d,c,!0);!f&&p.textConfig&&(f=e.textConfig=p.textConfig),!h&&p.textContent&&(h=p.textContent)}var m=p$(e);var g=e.$action||"merge",y="merge"===g,v="replace"===g;if(y){var _=!s,b=s;_?b=d$(a,u,e.type,n):(b&&(l$(b).isNew=!1),HB(b)),b&&(xB(b,m,t,{isInit:_}),f$(b,e,i,o))}else if(v){h$(s,e,n,t);var x=d$(a,u,e.type,n);x&&(xB(x,m,t,{isInit:!0}),f$(x,e,i,o))}else"remove"===g&&(wB(s,e),h$(s,e,n,t));var w=n.get(a);if(w&&h)if(y){var M=w.getTextContent();M?M.attr(h):w.setTextContent(new ks(h))}else v&&w.setTextContent(new ks(h));if(w){var S=e.clipPath;if(S){var k=S.type,L=void 0;_=!1;if(y){var T=w.getClipPath();_=!T||l$(T).type!==k,L=_?c$(k):T}else v&&(_=!0,L=c$(k));w.setClipPath(L),xB(L,S,t,{isInit:_}),zB(L,S.keyframeAnimation,t)}var D=l$(w);w.setTextConfig(f),D.option=e,m$(w,t,e),Pg({el:w,componentModel:t,itemName:w.name,itemTooltipOption:e.tooltip}),zB(w,e.keyframeAnimation,t)}}))}},e.prototype._relocate=function(t,e){for(var n=t.option.elements,r=this.group,i=this._elMap,o=e.getWidth(),a=e.getHeight(),l=["x","y"],u=0;u<n.length;u++){var c=n[u],d=Mi(c.id,null),h=null!=d?i.get(d):null;if(h&&h.isGroup){var f=h.parent,p=f===r,m=l$(h),g=l$(f);m.width=Nr(m.option.width,p?o:g.width)||0,m.height=Nr(m.option.height,p?a:g.height)||0}}for(u=n.length-1;u>=0;u--){c=n[u],d=Mi(c.id,null),h=null!=d?i.get(d):null;if(h){f=h.parent,g=l$(f);var y=f===r?{width:o,height:a}:{width:g.width,height:g.height},v={},_=zc(h,c,y,null,{hv:c.hv,boundingMode:c.bounding},v);if(!l$(h).isNew&&_){for(var b=c.transition,x={},w=0;w<l.length;w++){var M=l[w],S=v[M];b&&(SB(b)||s["r"](b,M)>=0)?x[M]=S:h[M]=S}Wl(h,x,t,0)}else h.attr(v)}}},e.prototype._clear=function(){var t=this,e=this._elMap;e.each((function(n){h$(n,l$(n).option,e,t._lastGraphicModel)})),this._elMap=s["f"]()},e.prototype.dispose=function(){this._clear()},e.type="graphic",e}(kp);function c$(t){var e=s["q"](s$,t)?s$[t]:fg(t);var n=new e({});return l$(n).type=t,n}function d$(t,e,n,r){var i=c$(n);return e.add(i),r.set(t,i),l$(i).id=t,l$(i).isNew=!0,i}function h$(t,e,n,r){var i=t&&t.parent;i&&("group"===t.type&&t.traverse((function(t){h$(t,e,n,r)})),MB(t,e,r),n.removeKey(l$(t).id))}function f$(t,e,n,r){t.isGroup||s["k"]([["cursor",No.prototype.cursor],["zlevel",r||0],["z",n||0],["z2",0]],(function(n){var r=n[0];s["q"](e,r)?t[r]=s["P"](e[r],n[1]):null==t[r]&&(t[r]=n[1])})),s["k"](s["F"](e),(function(n){if(0===n.indexOf("on")){var r=e[n];t[n]=s["w"](r)?r:null}})),s["q"](e,"draggable")&&(t.draggable=e.draggable),null!=e.name&&(t.name=e.name),null!=e.id&&(t.id=e.id)}function p$(t){return t=s["m"]({},t),s["k"](["id","parentId","$action","hv","bounding","textContent","clipPath"].concat(jc),(function(e){delete t[e]})),t}function m$(t,e,n){var r=Ls(t).eventData;t.silent||t.ignore||r||(r=Ls(t).eventData={componentType:"graphic",componentIndex:e.componentIndex,name:t.name}),r&&(r.info=n.info)}function g$(t){t.registerComponentModel(a$),t.registerComponentView(u$),t.registerPreprocessor((function(t){var e=t.graphic;Object(s["t"])(e)?e[0]&&e[0].elements?t.graphic=[t.graphic[0]]:t.graphic=[{elements:e}]:e&&!e.elements&&(t.graphic=[{elements:[e]}])}))}var y$=["x","y","radius","angle","single"],v$=["cartesian2d","polar","singleAxis"];function _$(t){var e=t.get("coordinateSystem");return Object(s["r"])(v$,e)>=0}function b$(t){return t+"Axis"}function x$(t,e){var n,r=Object(s["f"])(),i=[],o=Object(s["f"])();t.eachComponent({mainType:"dataZoom",query:e},(function(t){o.get(t.uid)||l(t)}));do{n=!1,t.eachComponent("dataZoom",a)}while(n);function a(t){!o.get(t.uid)&&u(t)&&(l(t),n=!0)}function l(t){o.set(t.uid,!0),i.push(t),c(t)}function u(t){var e=!1;return t.eachTargetAxis((function(t,n){var i=r.get(t);i&&i[n]&&(e=!0)})),e}function c(t){t.eachTargetAxis((function(t,e){(r.get(t)||r.set(t,[]))[e]=!0}))}return i}function w$(t){var e=t.ecModel,n={infoList:[],infoMap:Object(s["f"])()};return t.eachTargetAxis((function(t,r){var i=e.getComponent(b$(t),r);if(i){var o=i.getCoordSysModel();if(o){var a=o.uid,s=n.infoMap.get(a);s||(s={model:o,axisModels:[]},n.infoList.push(s),n.infoMap.set(a,s)),s.axisModels.push(i)}}})),n}var M$=function(){function t(){this.indexList=[],this.indexMap=[]}return t.prototype.add=function(t){this.indexMap[t]||(this.indexList.push(t),this.indexMap[t]=!0)},t}(),S$=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n._autoThrottle=!0,n._noTarget=!0,n._rangePropMode=["percent","percent"],n}return o(e,t),e.prototype.init=function(t,e,n){var r=k$(t);this.settledOption=r,this.mergeDefaultAndTheme(t,n),this._doInit(r)},e.prototype.mergeOption=function(t){var e=k$(t);Object(s["I"])(this.option,t,!0),Object(s["I"])(this.settledOption,e,!0),this._doInit(e)},e.prototype._doInit=function(t){var e=this.option;this._setDefaultThrottle(t),this._updateRangeUse(t);var n=this.settledOption;Object(s["k"])([["start","startValue"],["end","endValue"]],(function(t,r){"value"===this._rangePropMode[r]&&(e[t[0]]=n[t[0]]=null)}),this),this._resetTarget()},e.prototype._resetTarget=function(){var t=this.get("orient",!0),e=this._targetAxisInfoMap=Object(s["f"])(),n=this._fillSpecifiedTargetAxis(e);n?this._orient=t||this._makeAutoOrientByTargetAxis():(this._orient=t||"horizontal",this._fillAutoTargetAxisByOrient(e,this._orient)),this._noTarget=!0,e.each((function(t){t.indexList.length&&(this._noTarget=!1)}),this)},e.prototype._fillSpecifiedTargetAxis=function(t){var e=!1;return Object(s["k"])(y$,(function(n){var r=this.getReferringComponents(b$(n),Ei);if(r.specified){e=!0;var i=new M$;Object(s["k"])(r.models,(function(t){i.add(t.componentIndex)})),t.set(n,i)}}),this),e},e.prototype._fillAutoTargetAxisByOrient=function(t,e){var n=this.ecModel,r=!0;if(r){var i="vertical"===e?"y":"x",o=n.findComponents({mainType:i+"Axis"});a(o,i)}if(r){o=n.findComponents({mainType:"singleAxis",filter:function(t){return t.get("orient",!0)===e}});a(o,"single")}function a(e,n){var i=e[0];if(i){var o=new M$;if(o.add(i.componentIndex),t.set(n,o),r=!1,"x"===n||"y"===n){var a=i.getReferringComponents("grid",Pi).models[0];a&&Object(s["k"])(e,(function(t){i.componentIndex!==t.componentIndex&&a===t.getReferringComponents("grid",Pi).models[0]&&o.add(t.componentIndex)}))}}}r&&Object(s["k"])(y$,(function(e){if(r){var i=n.findComponents({mainType:b$(e),filter:function(t){return"category"===t.get("type",!0)}});if(i[0]){var o=new M$;o.add(i[0].componentIndex),t.set(e,o),r=!1}}}),this)},e.prototype._makeAutoOrientByTargetAxis=function(){var t;return this.eachTargetAxis((function(e){!t&&(t=e)}),this),"y"===t?"vertical":"horizontal"},e.prototype._setDefaultThrottle=function(t){if(t.hasOwnProperty("throttle")&&(this._autoThrottle=!1),this._autoThrottle){var e=this.ecModel.option;this.option.throttle=e.animation&&e.animationDurationUpdate>0?100:20}},e.prototype._updateRangeUse=function(t){var e=this._rangePropMode,n=this.get("rangeMode");Object(s["k"])([["start","startValue"],["end","endValue"]],(function(r,i){var o=null!=t[r[0]],a=null!=t[r[1]];o&&!a?e[i]="percent":!o&&a?e[i]="value":n?e[i]=n[i]:o&&(e[i]="percent")}))},e.prototype.noTarget=function(){return this._noTarget},e.prototype.getFirstTargetAxisModel=function(){var t;return this.eachTargetAxis((function(e,n){null==t&&(t=this.ecModel.getComponent(b$(e),n))}),this),t},e.prototype.eachTargetAxis=function(t,e){this._targetAxisInfoMap.each((function(n,r){Object(s["k"])(n.indexList,(function(n){t.call(e,r,n)}))}))},e.prototype.getAxisProxy=function(t,e){var n=this.getAxisModel(t,e);if(n)return n.__dzAxisProxy},e.prototype.getAxisModel=function(t,e){var n=this._targetAxisInfoMap.get(t);if(n&&n.indexMap[e])return this.ecModel.getComponent(b$(t),e)},e.prototype.setRawRange=function(t){var e=this.option,n=this.settledOption;Object(s["k"])([["start","startValue"],["end","endValue"]],(function(r){null==t[r[0]]&&null==t[r[1]]||(e[r[0]]=n[r[0]]=t[r[0]],e[r[1]]=n[r[1]]=t[r[1]])}),this),this._updateRangeUse(t)},e.prototype.setCalculatedRange=function(t){var e=this.option;Object(s["k"])(["start","startValue","end","endValue"],(function(n){e[n]=t[n]}))},e.prototype.getPercentRange=function(){var t=this.findRepresentativeAxisProxy();if(t)return t.getDataPercentWindow()},e.prototype.getValueRange=function(t,e){if(null!=t||null!=e)return this.getAxisProxy(t,e).getDataValueWindow();var n=this.findRepresentativeAxisProxy();return n?n.getDataValueWindow():void 0},e.prototype.findRepresentativeAxisProxy=function(t){if(t)return t.__dzAxisProxy;for(var e,n=this._targetAxisInfoMap.keys(),r=0;r<n.length;r++)for(var i=n[r],o=this._targetAxisInfoMap.get(i),a=0;a<o.indexList.length;a++){var s=this.getAxisProxy(i,o.indexList[a]);if(s.hostedBy(this))return s;e||(e=s)}return e},e.prototype.getRangePropMode=function(){return this._rangePropMode.slice()},e.prototype.getOrient=function(){return this._orient},e.type="dataZoom",e.dependencies=["xAxis","yAxis","radiusAxis","angleAxis","singleAxis","series","toolbox"],e.defaultOption={z:4,filterMode:"filter",start:0,end:100},e}(Xc);function k$(t){var e={};return Object(s["k"])(["start","end","startValue","endValue","throttle"],(function(n){t.hasOwnProperty(n)&&(e[n]=t[n])})),e}var L$=S$,T$=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="dataZoom.select",e}(L$),D$=T$,O$=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n,r){this.dataZoomModel=t,this.ecModel=e,this.api=n},e.type="dataZoom",e}(kp),I$=O$,C$=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="dataZoom.select",e}(I$),A$=C$,Y$=s["k"],j$=Hr,P$=function(){function t(t,e,n,r){this._dimName=t,this._axisIndex=e,this.ecModel=r,this._dataZoomModel=n}return t.prototype.hostedBy=function(t){return this._dataZoomModel===t},t.prototype.getDataValueWindow=function(){return this._valueWindow.slice()},t.prototype.getDataPercentWindow=function(){return this._percentWindow.slice()},t.prototype.getTargetSeriesModels=function(){var t=[];return this.ecModel.eachSeries((function(e){if(_$(e)){var n=b$(this._dimName),r=e.getReferringComponents(n,Pi).models[0];r&&this._axisIndex===r.componentIndex&&t.push(e)}}),this),t},t.prototype.getAxisModel=function(){return this.ecModel.getComponent(this._dimName+"Axis",this._axisIndex)},t.prototype.getMinMaxSpan=function(){return s["d"](this._minMaxSpan)},t.prototype.calculateDataWindow=function(t){var e,n=this._dataExtent,r=this.getAxisModel(),i=r.axis.scale,o=this._dataZoomModel.getRangePropMode(),a=[0,100],s=[],l=[];Y$(["start","end"],(function(r,u){var c=t[r],d=t[r+"Value"];"percent"===o[u]?(null==c&&(c=a[u]),d=i.parse(Er(c,a,n))):(e=!0,d=null==d?n[u]:i.parse(d),c=Er(d,n,a)),l[u]=null==d||isNaN(d)?n[u]:d,s[u]=null==c||isNaN(c)?a[u]:c})),j$(l),j$(s);var u=this._minMaxSpan;function c(t,e,n,r,o){var a=o?"Span":"ValueSpan";nN(0,t,n,"all",u["min"+a],u["max"+a]);for(var s=0;s<2;s++)e[s]=Er(t[s],n,r,!0),o&&(e[s]=i.parse(e[s]))}return e?c(l,s,n,a,!1):c(s,l,a,n,!0),{valueWindow:l,percentWindow:s}},t.prototype.reset=function(t){if(t===this._dataZoomModel){var e=this.getTargetSeriesModels();this._dataExtent=E$(this,this._dimName,e),this._updateMinMaxSpan();var n=this.calculateDataWindow(t.settledOption);this._valueWindow=n.valueWindow,this._percentWindow=n.percentWindow,this._setAxisModel()}},t.prototype.filterData=function(t,e){if(t===this._dataZoomModel){var n=this._dimName,r=this.getTargetSeriesModels(),i=t.get("filterMode"),o=this._valueWindow;"none"!==i&&Y$(r,(function(t){var e=t.getData(),r=e.mapDimensionsAll(n);if(r.length){if("weakFilter"===i){var l=e.getStore(),u=s["H"](r,(function(t){return e.getDimensionIndex(t)}),e);e.filterSelf((function(t){for(var e,n,i,a=0;a<r.length;a++){var s=l.get(u[a],t),c=!isNaN(s),d=s<o[0],h=s>o[1];if(c&&!d&&!h)return!0;c&&(i=!0),d&&(e=!0),h&&(n=!0)}return i&&e&&n}))}else Y$(r,(function(n){if("empty"===i)t.setData(e=e.map(n,(function(t){return a(t)?t:NaN})));else{var r={};r[n]=o,e.selectRange(r)}}));Y$(r,(function(t){e.setApproximateExtent(o,t)}))}}))}function a(t){return t>=o[0]&&t<=o[1]}},t.prototype._updateMinMaxSpan=function(){var t=this._minMaxSpan={},e=this._dataZoomModel,n=this._dataExtent;Y$(["min","max"],(function(r){var i=e.get(r+"Span"),o=e.get(r+"ValueSpan");null!=o&&(o=this.getAxisModel().axis.scale.parse(o)),null!=o?i=Er(n[0]+o,n,[0,100],!0):null!=i&&(o=Er(i,[0,100],n,!0)-n[0]),t[r+"Span"]=i,t[r+"ValueSpan"]=o}),this)},t.prototype._setAxisModel=function(){var t=this.getAxisModel(),e=this._percentWindow,n=this._valueWindow;if(e){var r=Br(n,[0,500]);r=Math.min(r,20);var i=t.axis.scale.rawExtentInfo;0!==e[0]&&i.setDeterminedMinMax("min",+n[0].toFixed(r)),100!==e[1]&&i.setDeterminedMinMax("max",+n[1].toFixed(r)),i.freeze()}},t}();function E$(t,e,n){var r=[1/0,-1/0];Y$(n,(function(t){uD(r,t.getData(),e)}));var i=t.getAxisModel(),o=XT(i.axis.scale,i,r).calculate();return[o.min,o.max]}var N$=P$,R$={getTargetSeries:function(t){function e(e){t.eachComponent("dataZoom",(function(n){n.eachTargetAxis((function(r,i){var o=t.getComponent(b$(r),i);e(r,i,o,n)}))}))}e((function(t,e,n,r){n.__dzAxisProxy=null}));var n=[];e((function(e,r,i,o){i.__dzAxisProxy||(i.__dzAxisProxy=new N$(e,r,o,t),n.push(i.__dzAxisProxy))}));var r=Object(s["f"])();return Object(s["k"])(n,(function(t){Object(s["k"])(t.getTargetSeriesModels(),(function(t){r.set(t.uid,t)}))})),r},overallReset:function(t,e){t.eachComponent("dataZoom",(function(t){t.eachTargetAxis((function(e,n){t.getAxisProxy(e,n).reset(t)})),t.eachTargetAxis((function(n,r){t.getAxisProxy(n,r).filterData(t,e)}))})),t.eachComponent("dataZoom",(function(t){var e=t.findRepresentativeAxisProxy();if(e){var n=e.getDataPercentWindow(),r=e.getDataValueWindow();t.setCalculatedRange({start:n[0],end:n[1],startValue:r[0],endValue:r[1]})}}))}},H$=R$;function z$(t){t.registerAction("dataZoom",(function(t,e){var n=x$(e,t);Object(s["k"])(n,(function(e){e.setRawRange({start:t.start,end:t.end,startValue:t.startValue,endValue:t.endValue})}))}))}var F$=!1;function B$(t){F$||(F$=!0,t.registerProcessor(t.PRIORITY.PROCESSOR.FILTER,H$),z$(t),t.registerSubTypeDefaulter("dataZoom",(function(){return"slider"})))}function W$(t){t.registerComponentModel(D$),t.registerComponentView(A$),B$(t)}var V$=function(){function t(){}return t}(),G$={};function $$(t,e){G$[t]=e}function U$(t){return G$[t]}var q$=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.optionUpdated=function(){t.prototype.optionUpdated.apply(this,arguments);var e=this.ecModel;s["k"](this.option.feature,(function(t,n){var r=U$(n);r&&(r.getDefaultOption&&(r.defaultOption=r.getDefaultOption(e)),s["I"](t,r.defaultOption))}))},e.type="toolbox",e.layoutMode={type:"box",ignoreSize:!0},e.defaultOption={show:!0,z:6,orient:"horizontal",left:"right",top:"top",backgroundColor:"transparent",borderColor:"#ccc",borderRadius:0,borderWidth:0,padding:5,itemSize:15,itemGap:8,showTitle:!0,iconStyle:{borderColor:"#666",color:"none"},emphasis:{iconStyle:{borderColor:"#3E98C5"}},tooltip:{show:!1,position:"bottom"}},e}(Xc),X$=q$;function Z$(t,e,n){var r=e.getBoxLayoutParams(),i=e.get("padding"),o={width:n.getWidth(),height:n.getHeight()},a=Hc(r,o,i);Nc(e.get("orient"),t,e.get("itemGap"),a.width,a.height),zc(t,r,o,i)}function J$(t,e){var n=Sc(e.get("padding")),r=e.getItemStyle(["color","opacity"]);return r.fill=e.get("backgroundColor"),t=new ss({shape:{x:t.x-n[3],y:t.y-n[0],width:t.width+n[1]+n[3],height:t.height+n[0]+n[2],r:e.get("borderRadius")},style:r,silent:!0,z2:-1}),t}var K$=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.render=function(t,e,n,r){var i=this.group;if(i.removeAll(),t.get("show")){var o=+t.get("itemSize"),a="vertical"===t.get("orient"),l=t.get("feature")||{},u=this._features||(this._features={}),c=[];s["k"](l,(function(t,e){c.push(e)})),new yM(this._featureNames||[],c).add(d).update(d).remove(s["h"](d,null)).execute(),this._featureNames=c,Z$(i,t,n),i.add(J$(i.getBoundingRect(),t)),a||i.eachChild((function(t){var e=t.__title,r=t.ensureState("emphasis"),a=r.textConfig||(r.textConfig={}),l=t.getTextContent(),u=l&&l.ensureState("emphasis");if(u&&!s["w"](u)&&e){var c=u.style||(u.style={}),d=er(e,ks.makeFont(c)),h=t.x+i.x,f=t.y+i.y+o,p=!1;f+d.height>n.getHeight()&&(a.position="top",p=!0);var m=p?-5-d.height:o+10;h+d.width/2>n.getWidth()?(a.position=["100%",m],c.align="right"):h-d.width/2<0&&(a.position=[0,m],c.align="left")}}))}function d(i,o){var a,s=c[i],d=c[o],f=l[s],p=new Lu(f,t,t.ecModel);if(r&&null!=r.newTitle&&r.featureName===s&&(f.title=r.newTitle),s&&!d){if(Q$(s))a={onclick:p.option.onclick,featureName:s};else{var m=U$(s);if(!m)return;a=new m}u[s]=a}else if(a=u[d],!a)return;a.uid=Du("toolbox-feature"),a.model=p,a.ecModel=e,a.api=n;var g=a instanceof V$;s||!d?!p.get("show")||g&&a.unusable?g&&a.remove&&a.remove(e,n):(h(p,a,s),p.setIconStatus=function(t,e){var n=this.option,r=this.iconPaths;n.iconStatus=n.iconStatus||{},n.iconStatus[t]=e,r[t]&&("emphasis"===e?ul:cl)(r[t])},a instanceof V$&&a.render&&a.render(p,e,n,r)):g&&a.dispose&&a.dispose(e,n)}function h(r,l,u){var c,d,h=r.getModel("iconStyle"),f=r.getModel(["emphasis","iconStyle"]),p=l instanceof V$&&l.getIcons?l.getIcons():r.get("icon"),m=r.get("title")||{};s["C"](p)?(c={},c[u]=p):c=p,s["C"](m)?(d={},d[u]=m):d=m;var g=r.iconPaths={};s["k"](c,(function(u,c){var p=Ig(u,{},{x:-o/2,y:-o/2,width:o,height:o});p.setStyle(h.getItemStyle());var m=p.ensureState("emphasis");m.style=f.getItemStyle();var y=new ks({style:{text:d[c],align:f.get("textAlign"),borderRadius:f.get("textBorderRadius"),padding:f.get("textPadding"),fill:null,font:cu({fontStyle:f.get("textFontStyle"),fontFamily:f.get("textFontFamily"),fontSize:f.get("textFontSize"),fontWeight:f.get("textFontWeight")},e)},ignore:!0});p.setTextContent(y),Pg({el:p,componentModel:t,itemName:c,formatterParamsExtra:{title:d[c]}}),p.__title=d[c],p.on("mouseover",(function(){var e=f.getItemStyle(),r=a?null==t.get("right")&&"right"!==t.get("left")?"right":"left":null==t.get("bottom")&&"bottom"!==t.get("top")?"bottom":"top";y.setStyle({fill:f.get("textFill")||e.fill||e.stroke||"#000",backgroundColor:f.get("textBackgroundColor")}),p.setTextConfig({position:f.get("textPosition")||r}),y.ignore=!t.get("showTitle"),n.enterEmphasis(this)})).on("mouseout",(function(){"emphasis"!==r.get(["iconStatus",c])&&n.leaveEmphasis(this),y.hide()})),("emphasis"===r.get(["iconStatus",c])?ul:cl)(p),i.add(p),p.on("click",s["c"](l.onclick,l,e,n,c)),g[c]=p}))}},e.prototype.updateView=function(t,e,n,r){s["k"](this._features,(function(t){t instanceof V$&&t.updateView&&t.updateView(t.model,e,n,r)}))},e.prototype.remove=function(t,e){s["k"](this._features,(function(n){n instanceof V$&&n.remove&&n.remove(t,e)})),this.group.removeAll()},e.prototype.dispose=function(t,e){s["k"](this._features,(function(n){n instanceof V$&&n.dispose&&n.dispose(t,e)}))},e.type="toolbox",e}(kp);function Q$(t){return 0===t.indexOf("my")}var tU=K$,eU=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.onclick=function(t,e){var n=this.model,r=n.get("name")||t.get("title.0.text")||"echarts",i="svg"===e.getZr().painter.getType(),o=i?"svg":n.get("type",!0)||"png",l=e.getConnectedDataURL({type:o,backgroundColor:n.get("backgroundColor",!0)||t.get("backgroundColor")||"#fff",connectedBackgroundColor:n.get("connectedBackgroundColor"),excludeComponents:n.get("excludeComponents"),pixelRatio:n.get("pixelRatio")}),u=a["a"].browser;if(Object(s["w"])(MouseEvent)&&(u.newEdge||!u.ie&&!u.edge)){var c=document.createElement("a");c.download=r+"."+o,c.target="_blank",c.href=l;var d=new MouseEvent("click",{view:document.defaultView,bubbles:!0,cancelable:!1});c.dispatchEvent(d)}else if(window.navigator.msSaveOrOpenBlob||i){var h=l.split(","),f=h[0].indexOf("base64")>-1,p=i?decodeURIComponent(h[1]):h[1];f&&(p=window.atob(p));var m=r+"."+o;if(window.navigator.msSaveOrOpenBlob){var g=p.length,y=new Uint8Array(g);while(g--)y[g]=p.charCodeAt(g);var v=new Blob([y]);window.navigator.msSaveOrOpenBlob(v,m)}else{var _=document.createElement("iframe");document.body.appendChild(_);var b=_.contentWindow,x=b.document;x.open("image/svg+xml","replace"),x.write(p),x.close(),b.focus(),x.execCommand("SaveAs",!0,m),document.body.removeChild(_)}}else{var w=n.get("lang"),M='<body style="margin:0;"><img src="'+l+'" style="max-width:100%;" title="'+(w&&w[0]||"")+'" /></body>',S=window.open();S.document.write(M),S.document.title=r}},e.getDefaultOption=function(t){var e={show:!0,icon:"M4.7,22.9L29.3,45.5L54.7,23.4M4.6,43.6L4.6,58L53.8,58L53.8,43.6M29.2,45.1L29.2,0",title:t.getLocaleModel().get(["toolbox","saveAsImage","title"]),type:"png",connectedBackgroundColor:"#fff",name:"",excludeComponents:["toolbox"],lang:t.getLocaleModel().get(["toolbox","saveAsImage","lang"])};return e},e}(V$),nU=eU,rU="__ec_magicType_stack__",iU=[["line","bar"],["stack"]],oU=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.getIcons=function(){var t=this.model,e=t.get("icon"),n={};return s["k"](t.get("type"),(function(t){e[t]&&(n[t]=e[t])})),n},e.getDefaultOption=function(t){var e={show:!0,type:[],icon:{line:"M4.1,28.9h7.1l9.3-22l7.4,38l9.7-19.7l3,12.8h14.9M4.1,58h51.4",bar:"M6.7,22.9h10V48h-10V22.9zM24.9,13h10v35h-10V13zM43.2,2h10v46h-10V2zM3.1,58h53.7",stack:"M8.2,38.4l-8.4,4.1l30.6,15.3L60,42.5l-8.1-4.1l-21.5,11L8.2,38.4z M51.9,30l-8.1,4.2l-13.4,6.9l-13.9-6.9L8.2,30l-8.4,4.2l8.4,4.2l22.2,11l21.5-11l8.1-4.2L51.9,30z M51.9,21.7l-8.1,4.2L35.7,30l-5.3,2.8L24.9,30l-8.4-4.1l-8.3-4.2l-8.4,4.2L8.2,30l8.3,4.2l13.9,6.9l13.4-6.9l8.1-4.2l8.1-4.1L51.9,21.7zM30.4,2.2L-0.2,17.5l8.4,4.1l8.3,4.2l8.4,4.2l5.5,2.7l5.3-2.7l8.1-4.2l8.1-4.2l8.1-4.1L30.4,2.2z"},title:t.getLocaleModel().get(["toolbox","magicType","title"]),option:{},seriesIndex:{}};return e},e.prototype.onclick=function(t,e,n){var r=this.model,i=r.get(["seriesIndex",n]);if(aU[n]){var o,a={series:[]},l=function(t){var e=t.subType,i=t.id,o=aU[n](e,i,t,r);o&&(s["i"](o,t.option),a.series.push(o));var l=t.coordinateSystem;if(l&&"cartesian2d"===l.type&&("line"===n||"bar"===n)){var u=l.getAxesByScale("ordinal")[0];if(u){var c=u.dim,d=c+"Axis",h=t.getReferringComponents(d,Pi).models[0],f=h.componentIndex;a[d]=a[d]||[];for(var p=0;p<=f;p++)a[d][f]=a[d][f]||{};a[d][f].boundaryGap="bar"===n}}};s["k"](iU,(function(t){s["r"](t,n)>=0&&s["k"](t,(function(t){r.setIconStatus(t,"normal")}))})),r.setIconStatus(n,"emphasis"),t.eachComponent({mainType:"series",query:null==i?null:{seriesIndex:i}},l);var u=n;"stack"===n&&(o=s["I"]({stack:r.option.title.tiled,tiled:r.option.title.stack},r.option.title),"emphasis"!==r.get(["iconStatus",n])&&(u="tiled")),e.dispatchAction({type:"changeMagicType",currentType:u,newOption:a,newTitle:o,featureName:"magicType"})}},e}(V$),aU={line:function(t,e,n,r){if("bar"===t)return s["I"]({id:e,type:"line",data:n.get("data"),stack:n.get("stack"),markPoint:n.get("markPoint"),markLine:n.get("markLine")},r.get(["option","line"])||{},!0)},bar:function(t,e,n,r){if("line"===t)return s["I"]({id:e,type:"bar",data:n.get("data"),stack:n.get("stack"),markPoint:n.get("markPoint"),markLine:n.get("markLine")},r.get(["option","bar"])||{},!0)},stack:function(t,e,n,r){var i=n.get("stack")===rU;if("line"===t||"bar"===t)return r.setIconStatus("stack",i?"normal":"emphasis"),s["I"]({id:e,stack:i?"":rU},r.get(["option","stack"])||{},!0)}};Mb({type:"changeMagicType",event:"magicTypeChanged",update:"prepareAndUpdate"},(function(t,e){e.mergeOption(t.newOption)}));var sU=oU,lU=new Array(60).join("-"),uU="\t";function cU(t){var e={},n=[],r=[];return t.eachRawSeries((function(t){var i=t.coordinateSystem;if(!i||"cartesian2d"!==i.type&&"polar"!==i.type)n.push(t);else{var o=i.getBaseAxis();if("category"===o.type){var a=o.dim+"_"+o.index;e[a]||(e[a]={categoryAxis:o,valueAxis:i.getOtherAxis(o),series:[]},r.push({axisDim:o.dim,axisIndex:o.index})),e[a].series.push(t)}else n.push(t)}})),{seriesGroupByCategoryAxis:e,other:n,meta:r}}function dU(t){var e=[];return s["k"](t,(function(t,n){var r=t.categoryAxis,i=t.valueAxis,o=i.dim,a=[" "].concat(s["H"](t.series,(function(t){return t.name}))),l=[r.model.getCategories()];s["k"](t.series,(function(t){var e=t.getRawData();l.push(t.getRawData().mapArray(e.mapDimension(o),(function(t){return t})))}));for(var u=[a.join(uU)],c=0;c<l[0].length;c++){for(var d=[],h=0;h<l.length;h++)d.push(l[h][c]);u.push(d.join(uU))}e.push(u.join("\n"))})),e.join("\n\n"+lU+"\n\n")}function hU(t){return s["H"](t,(function(t){var e=t.getRawData(),n=[t.name],r=[];return e.each(e.dimensions,(function(){for(var t=arguments.length,i=arguments[t-1],o=e.getName(i),a=0;a<t-1;a++)r[a]=arguments[a];n.push((o?o+uU:"")+r.join(uU))})),n.join("\n")})).join("\n\n"+lU+"\n\n")}function fU(t){var e=cU(t);return{value:s["n"]([dU(e.seriesGroupByCategoryAxis),hU(e.other)],(function(t){return!!t.replace(/[\n\t\s]/g,"")})).join("\n\n"+lU+"\n\n"),meta:e.meta}}function pU(t){return t.replace(/^\s\s*/,"").replace(/\s\s*$/,"")}function mU(t){var e=t.slice(0,t.indexOf("\n"));if(e.indexOf(uU)>=0)return!0}var gU=new RegExp("["+uU+"]+","g");function yU(t){for(var e=t.split(/\n+/g),n=pU(e.shift()).split(gU),r=[],i=s["H"](n,(function(t){return{name:t,data:[]}})),o=0;o<e.length;o++){var a=pU(e[o]).split(gU);r.push(a.shift());for(var l=0;l<a.length;l++)i[l]&&(i[l].data[o]=a[l])}return{series:i,categories:r}}function vU(t){for(var e=t.split(/\n+/g),n=pU(e.shift()),r=[],i=0;i<e.length;i++){var o=pU(e[i]);if(o){var a=o.split(gU),s="",l=void 0,u=!1;isNaN(a[0])?(u=!0,s=a[0],a=a.slice(1),r[i]={name:s,value:[]},l=r[i].value):l=r[i]=[];for(var c=0;c<a.length;c++)l.push(+a[c]);1===l.length&&(u?r[i].value=l[0]:r[i]=l[0])}}return{name:n,data:r}}function _U(t,e){var n=t.split(new RegExp("\n*"+lU+"\n*","g")),r={series:[]};return s["k"](n,(function(t,n){if(mU(t)){var i=yU(t),o=e[n],a=o.axisDim+"Axis";o&&(r[a]=r[a]||[],r[a][o.axisIndex]={data:i.categories},r.series=r.series.concat(i.series))}else{i=vU(t);r.series.push(i)}})),r}var bU=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.onclick=function(t,e){setTimeout((function(){e.dispatchAction({type:"hideTip"})}));var n=e.getDom(),r=this.model;this._dom&&n.removeChild(this._dom);var i=document.createElement("div");i.style.cssText="position:absolute;top:0;bottom:0;left:0;right:0;padding:5px",i.style.backgroundColor=r.get("backgroundColor")||"#fff";var o=document.createElement("h4"),a=r.get("lang")||[];o.innerHTML=a[0]||r.get("title"),o.style.cssText="margin:10px 20px",o.style.color=r.get("textColor");var l=document.createElement("div"),u=document.createElement("textarea");l.style.cssText="overflow:auto";var c=r.get("optionToContent"),d=r.get("contentToOption"),h=fU(t);if(s["w"](c)){var f=c(e.getOption());s["C"](f)?l.innerHTML=f:s["v"](f)&&l.appendChild(f)}else{u.readOnly=r.get("readOnly");var p=u.style;p.cssText="display:block;width:100%;height:100%;font-family:monospace;font-size:14px;line-height:1.6rem;resize:none;box-sizing:border-box;outline:none",p.color=r.get("textColor"),p.borderColor=r.get("textareaBorderColor"),p.backgroundColor=r.get("textareaColor"),u.value=h.value,l.appendChild(u)}var m=h.meta,g=document.createElement("div");g.style.cssText="position:absolute;bottom:5px;left:0;right:0";var y="float:right;margin-right:20px;border:none;cursor:pointer;padding:2px 5px;font-size:12px;border-radius:3px",v=document.createElement("div"),_=document.createElement("div");y+=";background-color:"+r.get("buttonColor"),y+=";color:"+r.get("buttonTextColor");var b=this;function x(){n.removeChild(i),b._dom=null}Q(v,"click",x),Q(_,"click",(function(){if(null==d&&null!=c||null!=d&&null==c)x();else{var t;try{t=s["w"](d)?d(l,e.getOption()):_U(u.value,m)}catch(n){throw x(),new Error("Data view format error "+n)}t&&e.dispatchAction({type:"changeDataView",newOption:t}),x()}})),v.innerHTML=a[1],_.innerHTML=a[2],_.style.cssText=v.style.cssText=y,!r.get("readOnly")&&g.appendChild(_),g.appendChild(v),i.appendChild(o),i.appendChild(l),i.appendChild(g),l.style.height=n.clientHeight-80+"px",n.appendChild(i),this._dom=i},e.prototype.remove=function(t,e){this._dom&&e.getDom().removeChild(this._dom)},e.prototype.dispose=function(t,e){this.remove(t,e)},e.getDefaultOption=function(t){var e={show:!0,readOnly:!1,optionToContent:null,contentToOption:null,icon:"M17.5,17.3H33 M17.5,17.3H33 M45.4,29.5h-28 M11.5,2v56H51V14.8L38.4,2H11.5z M38.4,2.2v12.7H51 M45.4,41.7h-28",title:t.getLocaleModel().get(["toolbox","dataView","title"]),lang:t.getLocaleModel().get(["toolbox","dataView","lang"]),backgroundColor:"#fff",textColor:"#000",textareaColor:"#fff",textareaBorderColor:"#333",buttonColor:"#c23531",buttonTextColor:"#fff"};return e},e}(V$);function xU(t,e){return s["H"](t,(function(t,n){var r=e&&e[n];if(s["A"](r)&&!s["t"](r)){var i=s["A"](t)&&!s["t"](t);i||(t={value:t});var o=null!=r.name&&null==t.name;return t=s["i"](t,r),o&&delete t.name,t}return t}))}Mb({type:"changeDataView",event:"dataViewChanged",update:"prepareAndUpdate"},(function(t,e){var n=[];s["k"](t.newOption.series,(function(t){var r=e.getSeriesByName(t.name)[0];if(r){var i=r.get("data");n.push({name:t.name,data:xU(t.data,i)})}else n.push(s["m"]({type:"scatter"},t))})),e.mergeOption(s["i"]({series:n},t.newOption))}));var wU=bU,MU=s["k"],SU=Ci();function kU(t,e){var n=OU(t);MU(e,(function(e,r){for(var i=n.length-1;i>=0;i--){var o=n[i];if(o[r])break}if(i<0){var a=t.queryComponents({mainType:"dataZoom",subType:"select",id:r})[0];if(a){var s=a.getPercentRange();n[0][r]={dataZoomId:r,start:s[0],end:s[1]}}}})),n.push(e)}function LU(t){var e=OU(t),n=e[e.length-1];e.length>1&&e.pop();var r={};return MU(n,(function(t,n){for(var i=e.length-1;i>=0;i--)if(t=e[i][n],t){r[n]=t;break}})),r}function TU(t){SU(t).snapshots=null}function DU(t){return OU(t).length}function OU(t){var e=SU(t);return e.snapshots||(e.snapshots=[{}]),e.snapshots}var IU=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.onclick=function(t,e){TU(t),e.dispatchAction({type:"restore",from:this.uid})},e.getDefaultOption=function(t){var e={show:!0,icon:"M3.8,33.4 M47,18.9h9.8V8.7 M56.3,20.1 C52.1,9,40.5,0.6,26.8,2.1C12.6,3.7,1.6,16.2,2.1,30.6 M13,41.1H3.1v10.2 M3.7,39.9c4.2,11.1,15.8,19.5,29.5,18 c14.2-1.6,25.2-14.1,24.7-28.5",title:t.getLocaleModel().get(["toolbox","restore","title"])};return e},e}(V$);Mb({type:"restore",event:"restore",update:"prepareAndUpdate"},(function(t,e){e.resetOption("recreate")}));var CU=IU,AU=["grid","xAxis","yAxis","geo","graph","polar","radiusAxis","angleAxis","bmap"],YU=function(){function t(t,e,n){var r=this;this._targetInfoList=[];var i=PU(e,t);Object(s["k"])(EU,(function(t,e){(!n||!n.include||Object(s["r"])(n.include,e)>=0)&&t(i,r._targetInfoList)}))}return t.prototype.setOutputRanges=function(t,e){return this.matchOutputRanges(t,e,(function(t,e,n){if((t.coordRanges||(t.coordRanges=[])).push(e),!t.coordRange){t.coordRange=e;var r=HU[t.brushType](0,n,e);t.__rangeOffset={offset:FU[t.brushType](r.values,t.range,[1,1]),xyMinMax:r.xyMinMax}}})),t},t.prototype.matchOutputRanges=function(t,e,n){Object(s["k"])(t,(function(t){var r=this.findTargetInfo(t,e);r&&!0!==r&&Object(s["k"])(r.coordSyses,(function(r){var i=HU[t.brushType](1,r,t.range,!0);n(t,i.values,r,e)}))}),this)},t.prototype.setInputRanges=function(t,e){Object(s["k"])(t,(function(t){var n=this.findTargetInfo(t,e);if(t.range=t.range||[],n&&!0!==n){t.panelId=n.panelId;var r=HU[t.brushType](0,n.coordSys,t.coordRange),i=t.__rangeOffset;t.range=i?FU[t.brushType](r.values,i.offset,WU(r.xyMinMax,i.xyMinMax)):r.values}}),this)},t.prototype.makePanelOpts=function(t,e){return Object(s["H"])(this._targetInfoList,(function(n){var r=n.getPanelRect();return{panelId:n.panelId,defaultBrushType:e?e(n):null,clipPath:vR(r),isTargetByCursor:bR(r,t,n.coordSysModel),getLinearBrushOtherExtent:_R(r)}}))},t.prototype.controlSeries=function(t,e,n){var r=this.findTargetInfo(t,n);return!0===r||r&&Object(s["r"])(r.coordSyses,e.coordinateSystem)>=0},t.prototype.findTargetInfo=function(t,e){for(var n=this._targetInfoList,r=PU(e,t),i=0;i<n.length;i++){var o=n[i],a=t.panelId;if(a){if(o.panelId===a)return o}else for(var s=0;s<NU.length;s++)if(NU[s](r,o))return o}return!0},t}();function jU(t){return t[0]>t[1]&&t.reverse(),t}function PU(t,e){return Yi(t,e,{includeMainTypes:AU})}var EU={grid:function(t,e){var n=t.xAxisModels,r=t.yAxisModels,i=t.gridModels,o=Object(s["f"])(),a={},l={};(n||r||i)&&(Object(s["k"])(n,(function(t){var e=t.axis.grid.model;o.set(e.id,e),a[e.id]=!0})),Object(s["k"])(r,(function(t){var e=t.axis.grid.model;o.set(e.id,e),l[e.id]=!0})),Object(s["k"])(i,(function(t){o.set(t.id,t),a[t.id]=!0,l[t.id]=!0})),o.each((function(t){var i=t.coordinateSystem,o=[];Object(s["k"])(i.getCartesians(),(function(t,e){(Object(s["r"])(n,t.getAxis("x").model)>=0||Object(s["r"])(r,t.getAxis("y").model)>=0)&&o.push(t)})),e.push({panelId:"grid--"+t.id,gridModel:t,coordSysModel:t,coordSys:o[0],coordSyses:o,getPanelRect:RU.grid,xAxisDeclared:a[t.id],yAxisDeclared:l[t.id]})})))},geo:function(t,e){Object(s["k"])(t.geoModels,(function(t){var n=t.coordinateSystem;e.push({panelId:"geo--"+t.id,geoModel:t,coordSysModel:t,coordSys:n,coordSyses:[n],getPanelRect:RU.geo})}))}},NU=[function(t,e){var n=t.xAxisModel,r=t.yAxisModel,i=t.gridModel;return!i&&n&&(i=n.axis.grid.model),!i&&r&&(i=r.axis.grid.model),i&&i===e.gridModel},function(t,e){var n=t.geoModel;return n&&n===e.geoModel}],RU={grid:function(){return this.coordSys.master.getRect().clone()},geo:function(){var t=this.coordSys,e=t.getBoundingRect().clone();return e.applyTransform(wg(t)),e}},HU={lineX:Object(s["h"])(zU,0),lineY:Object(s["h"])(zU,1),rect:function(t,e,n,r){var i=t?e.pointToData([n[0][0],n[1][0]],r):e.dataToPoint([n[0][0],n[1][0]],r),o=t?e.pointToData([n[0][1],n[1][1]],r):e.dataToPoint([n[0][1],n[1][1]],r),a=[jU([i[0],o[0]]),jU([i[1],o[1]])];return{values:a,xyMinMax:a}},polygon:function(t,e,n,r){var i=[[1/0,-1/0],[1/0,-1/0]],o=Object(s["H"])(n,(function(n){var o=t?e.pointToData(n,r):e.dataToPoint(n,r);return i[0][0]=Math.min(i[0][0],o[0]),i[1][0]=Math.min(i[1][0],o[1]),i[0][1]=Math.max(i[0][1],o[0]),i[1][1]=Math.max(i[1][1],o[1]),o}));return{values:o,xyMinMax:i}}};function zU(t,e,n,r){var i=n.getAxis(["x","y"][t]),o=jU(Object(s["H"])([0,1],(function(t){return e?i.coordToData(i.toLocalCoord(r[t]),!0):i.toGlobalCoord(i.dataToCoord(r[t]))}))),a=[];return a[t]=o,a[1-t]=[NaN,NaN],{values:o,xyMinMax:a}}var FU={lineX:Object(s["h"])(BU,0),lineY:Object(s["h"])(BU,1),rect:function(t,e,n){return[[t[0][0]-n[0]*e[0][0],t[0][1]-n[0]*e[0][1]],[t[1][0]-n[1]*e[1][0],t[1][1]-n[1]*e[1][1]]]},polygon:function(t,e,n){return Object(s["H"])(t,(function(t,r){return[t[0]-n[0]*e[r][0],t[1]-n[1]*e[r][1]]}))}};function BU(t,e,n,r){return[e[0]-r[t]*n[0],e[1]-r[t]*n[1]]}function WU(t,e){var n=VU(t),r=VU(e),i=[n[0]/r[0],n[1]/r[1]];return isNaN(i[0])&&(i[0]=1),isNaN(i[1])&&(i[1]=1),i}function VU(t){return t?[t[0][1]-t[0][0],t[1][1]-t[1][0]]:[NaN,NaN]}var GU=YU,$U=s["k"],UU=Li("toolbox-dataZoom_"),qU=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.render=function(t,e,n,r){this._brushController||(this._brushController=new yR(n.getZr()),this._brushController.on("brush",s["c"](this._onBrush,this)).mount()),KU(t,e,this,r,n),JU(t,e)},e.prototype.onclick=function(t,e,n){XU[n].call(this)},e.prototype.remove=function(t,e){this._brushController&&this._brushController.unmount()},e.prototype.dispose=function(t,e){this._brushController&&this._brushController.dispose()},e.prototype._onBrush=function(t){var e=t.areas;if(t.isEnd&&e.length){var n={},r=this.ecModel;this._brushController.updateCovers([]);var i=new GU(ZU(this.model),r,{include:["grid"]});i.matchOutputRanges(e,r,(function(t,e,n){if("cartesian2d"===n.type){var r=t.brushType;"rect"===r?(o("x",n,e[0]),o("y",n,e[1])):o({lineX:"x",lineY:"y"}[r],n,e)}})),kU(r,n),this._dispatchZoomAction(n)}function o(t,e,i){var o=e.getAxis(t),s=o.model,l=a(t,s,r),u=l.findRepresentativeAxisProxy(s).getMinMaxSpan();null==u.minValueSpan&&null==u.maxValueSpan||(i=nN(0,i.slice(),o.scale.getExtent(),0,u.minValueSpan,u.maxValueSpan)),l&&(n[l.id]={dataZoomId:l.id,startValue:i[0],endValue:i[1]})}function a(t,e,n){var r;return n.eachComponent({mainType:"dataZoom",subType:"select"},(function(n){var i=n.getAxisModel(t,e.componentIndex);i&&(r=n)})),r}},e.prototype._dispatchZoomAction=function(t){var e=[];$U(t,(function(t,n){e.push(s["d"](t))})),e.length&&this.api.dispatchAction({type:"dataZoom",from:this.uid,batch:e})},e.getDefaultOption=function(t){var e={show:!0,filterMode:"filter",icon:{zoom:"M0,13.5h26.9 M13.5,26.9V0 M32.1,13.5H58V58H13.5 V32.1",back:"M22,1.4L9.9,13.5l12.3,12.3 M10.3,13.5H54.9v44.6 H10.3v-26"},title:t.getLocaleModel().get(["toolbox","dataZoom","title"]),brushStyle:{borderWidth:0,color:"rgba(210,219,238,0.2)"}};return e},e}(V$),XU={zoom:function(){var t=!this._isZoomActive;this.api.dispatchAction({type:"takeGlobalCursor",key:"dataZoomSelect",dataZoomSelectActive:t})},back:function(){this._dispatchZoomAction(LU(this.ecModel))}};function ZU(t){var e={xAxisIndex:t.get("xAxisIndex",!0),yAxisIndex:t.get("yAxisIndex",!0),xAxisId:t.get("xAxisId",!0),yAxisId:t.get("yAxisId",!0)};return null==e.xAxisIndex&&null==e.xAxisId&&(e.xAxisIndex="all"),null==e.yAxisIndex&&null==e.yAxisId&&(e.yAxisIndex="all"),e}function JU(t,e){t.setIconStatus("back",DU(e)>1?"emphasis":"normal")}function KU(t,e,n,r,i){var o=n._isZoomActive;r&&"takeGlobalCursor"===r.type&&(o="dataZoomSelect"===r.key&&r.dataZoomSelectActive),n._isZoomActive=o,t.setIconStatus("zoom",o?"emphasis":"normal");var a=new GU(ZU(t),e,{include:["grid"]}),s=a.makePanelOpts(i,(function(t){return t.xAxisDeclared&&!t.yAxisDeclared?"lineX":!t.xAxisDeclared&&t.yAxisDeclared?"lineY":"rect"}));n._brushController.setPanels(s).enableBrush(!(!o||!s.length)&&{brushType:"auto",brushStyle:t.getModel("brushStyle").getItemStyle()})}vd("dataZoom",(function(t){var e=t.getComponent("toolbox",0),n=["feature","dataZoom"];if(e&&null!=e.get(n)){var r=e.getModel(n),i=[],o=ZU(r),a=Yi(t,o);return $U(a.xAxisModels,(function(t){return s(t,"xAxis","xAxisIndex")})),$U(a.yAxisModels,(function(t){return s(t,"yAxis","yAxisIndex")})),i}function s(t,e,n){var o=t.componentIndex,a={type:"select",$fromToolbox:!0,filterMode:r.get("filterMode",!0)||"filter",id:UU+e+o};a[n]=o,i.push(a)}}));var QU=qU;function tq(t){t.registerComponentModel(X$),t.registerComponentView(tU),$$("saveAsImage",nU),$$("magicType",sU),$$("dataView",wU),$$("dataZoom",QU),$$("restore",CU),jb(W$)}var eq=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="tooltip",e.dependencies=["axisPointer"],e.defaultOption={z:60,show:!0,showContent:!0,trigger:"item",triggerOn:"mousemove|click",alwaysShowContent:!1,displayMode:"single",renderMode:"auto",confine:null,showDelay:0,hideDelay:100,transitionDuration:.4,enterable:!1,backgroundColor:"#fff",shadowBlur:10,shadowColor:"rgba(0, 0, 0, .2)",shadowOffsetX:1,shadowOffsetY:2,borderRadius:4,borderWidth:1,padding:null,extraCssText:"",axisPointer:{type:"line",axis:"auto",animation:"auto",animationDurationUpdate:200,animationEasingUpdate:"exponentialOut",crossStyle:{color:"#999",width:1,type:"dashed",textStyle:{}}},textStyle:{color:"#666",fontSize:14}},e}(Xc),nq=eq;function rq(t){var e=t.get("confine");return null!=e?!!e:"richText"===t.get("renderMode")}function iq(t){if(a["a"].domSupported)for(var e=document.documentElement.style,n=0,r=t.length;n<r;n++)if(t[n]in e)return t[n]}var oq=iq(["transform","webkitTransform","OTransform","MozTransform","msTransform"]),aq=iq(["webkitTransition","transition","OTransition","MozTransition","msTransition"]);function sq(t,e){if(!t)return e;e=Mc(e,!0);var n=t.indexOf(e);return t=-1===n?e:"-"+t.slice(0,n)+"-"+e,t.toLowerCase()}function lq(t,e){var n=t.currentStyle||document.defaultView&&document.defaultView.getComputedStyle(t);return n?e?n[e]:n:null}var uq=sq(aq,"transition"),cq=sq(oq,"transform"),dq="position:absolute;display:block;border-style:solid;white-space:nowrap;z-index:9999999;"+(a["a"].transform3dSupported?"will-change:transform;":"");function hq(t){return t="left"===t?"right":"right"===t?"left":"top"===t?"bottom":"top",t}function fq(t,e,n){if(!Object(s["C"])(n)||"inside"===n)return"";var r=t.get("backgroundColor"),i=t.get("borderWidth");e=Cc(e);var o,a=hq(n),l=Math.max(1.5*Math.round(i),6),u="",c=cq+":";Object(s["r"])(["left","right"],a)>-1?(u+="top:50%",c+="translateY(-50%) rotate("+(o="left"===a?-225:-45)+"deg)"):(u+="left:50%",c+="translateX(-50%) rotate("+(o="top"===a?225:45)+"deg)");var d=o*Math.PI/180,h=l+i,f=h*Math.abs(Math.cos(d))+h*Math.abs(Math.sin(d)),p=Math.round(100*((f-Math.SQRT2*i)/2+Math.SQRT2*i-(f-h)/2))/100;u+=";"+a+":-"+p+"px";var m=e+" solid "+i+"px;",g=["position:absolute;width:"+l+"px;height:"+l+"px;z-index:-1;",u+";"+c+";","border-bottom:"+m,"border-right:"+m,"background-color:"+r+";"];return'<div style="'+g.join("")+'"></div>'}function pq(t,e){var n="cubic-bezier(0.23,1,0.32,1)",r=" "+t/2+"s "+n,i="opacity"+r+",visibility"+r;return e||(r=" "+t+"s "+n,i+=a["a"].transformSupported?","+cq+r:",left"+r+",top"+r),uq+":"+i}function mq(t,e,n){var r=t.toFixed(0)+"px",i=e.toFixed(0)+"px";if(!a["a"].transformSupported)return n?"top:"+i+";left:"+r+";":[["top",i],["left",r]];var o=a["a"].transform3dSupported,s="translate"+(o?"3d":"")+"("+r+","+i+(o?",0":"")+")";return n?"top:0;left:0;"+cq+":"+s+";":[["top",0],["left",0],[oq,s]]}function gq(t){var e=[],n=t.get("fontSize"),r=t.getTextColor();r&&e.push("color:"+r),e.push("font:"+t.getFont()),n&&e.push("line-height:"+Math.round(3*n/2)+"px");var i=t.get("textShadowColor"),o=t.get("textShadowBlur")||0,a=t.get("textShadowOffsetX")||0,l=t.get("textShadowOffsetY")||0;return i&&o&&e.push("text-shadow:"+a+"px "+l+"px "+o+"px "+i),Object(s["k"])(["decoration","align"],(function(n){var r=t.get(n);r&&e.push("text-"+n+":"+r)})),e.join(";")}function yq(t,e,n){var r=[],i=t.get("transitionDuration"),o=t.get("backgroundColor"),a=t.get("shadowBlur"),l=t.get("shadowColor"),u=t.get("shadowOffsetX"),c=t.get("shadowOffsetY"),d=t.getModel("textStyle"),h=sp(t,"html"),f=u+"px "+c+"px "+a+"px "+l;return r.push("box-shadow:"+f),e&&i&&r.push(pq(i,n)),o&&r.push("background-color:"+o),Object(s["k"])(["width","color","radius"],(function(e){var n="border-"+e,i=Mc(n),o=t.get(i);null!=o&&r.push(n+":"+o+("color"===e?"":"px"))})),r.push(gq(d)),null!=h&&r.push("padding:"+Sc(h).join("px ")+"px"),r.join(";")+";"}function vq(t,e,n,r,i){var o=e&&e.painter;if(n){var a=o&&o.getViewportRoot();a&&N(t,a,n,r,i)}else{t[0]=r,t[1]=i;var s=o&&o.getViewportRootOffset();s&&(t[0]+=s.offsetLeft,t[1]+=s.offsetTop)}t[2]=t[0]/e.getWidth(),t[3]=t[1]/e.getHeight()}var _q=function(){function t(t,e){if(this._show=!1,this._styleCoord=[0,0,0,0],this._enterable=!0,this._alwaysShowContent=!1,this._firstShow=!0,this._longHide=!0,a["a"].wxa)return null;var n=document.createElement("div");n.domBelongToZr=!0,this.el=n;var r=this._zr=t.getZr(),i=e.appendTo,o=i&&(Object(s["C"])(i)?document.querySelector(i):Object(s["v"])(i)?i:Object(s["w"])(i)&&i(t.getDom()));vq(this._styleCoord,r,o,t.getWidth()/2,t.getHeight()/2),(o||t.getDom()).appendChild(n),this._api=t,this._container=o;var l=this;n.onmouseenter=function(){l._enterable&&(clearTimeout(l._hideTimeout),l._show=!0),l._inContent=!0},n.onmousemove=function(t){if(t=t||window.event,!l._enterable){var e=r.handler,n=r.painter.getViewportRoot();J(n,t,!0),e.dispatch("mousemove",t)}},n.onmouseleave=function(){l._inContent=!1,l._enterable&&l._show&&l.hideLater(l._hideDelay)}}return t.prototype.update=function(t){if(!this._container){var e=this._api.getDom(),n=lq(e,"position"),r=e.style;"absolute"!==r.position&&"absolute"!==n&&(r.position="relative")}var i=t.get("alwaysShowContent");i&&this._moveIfResized(),this._alwaysShowContent=i,this.el.className=t.get("className")||""},t.prototype.show=function(t,e){clearTimeout(this._hideTimeout),clearTimeout(this._longHideTimeout);var n=this.el,r=n.style,i=this._styleCoord;n.innerHTML?r.cssText=dq+yq(t,!this._firstShow,this._longHide)+mq(i[0],i[1],!0)+"border-color:"+Cc(e)+";"+(t.get("extraCssText")||"")+";pointer-events:"+(this._enterable?"auto":"none"):r.display="none",this._show=!0,this._firstShow=!1,this._longHide=!1},t.prototype.setContent=function(t,e,n,r,i){var o=this.el;if(null!=t){var a="";if(Object(s["C"])(i)&&"item"===n.get("trigger")&&!rq(n)&&(a=fq(n,r,i)),Object(s["C"])(t))o.innerHTML=t+a;else if(t){o.innerHTML="",Object(s["t"])(t)||(t=[t]);for(var l=0;l<t.length;l++)Object(s["v"])(t[l])&&t[l].parentNode!==o&&o.appendChild(t[l]);if(a&&o.childNodes.length){var u=document.createElement("div");u.innerHTML=a,o.appendChild(u)}}}else o.innerHTML=""},t.prototype.setEnterable=function(t){this._enterable=t},t.prototype.getSize=function(){var t=this.el;return[t.offsetWidth,t.offsetHeight]},t.prototype.moveTo=function(t,e){var n=this._styleCoord;if(vq(n,this._zr,this._container,t,e),null!=n[0]&&null!=n[1]){var r=this.el.style,i=mq(n[0],n[1]);Object(s["k"])(i,(function(t){r[t[0]]=t[1]}))}},t.prototype._moveIfResized=function(){var t=this._styleCoord[2],e=this._styleCoord[3];this.moveTo(t*this._zr.getWidth(),e*this._zr.getHeight())},t.prototype.hide=function(){var t=this,e=this.el.style;e.visibility="hidden",e.opacity="0",a["a"].transform3dSupported&&(e.willChange=""),this._show=!1,this._longHideTimeout=setTimeout((function(){return t._longHide=!0}),500)},t.prototype.hideLater=function(t){!this._show||this._inContent&&this._enterable||this._alwaysShowContent||(t?(this._hideDelay=t,this._show=!1,this._hideTimeout=setTimeout(Object(s["c"])(this.hide,this),t)):this.hide())},t.prototype.isShow=function(){return this._show},t.prototype.dispose=function(){clearTimeout(this._hideTimeout),clearTimeout(this._longHideTimeout);var t=this.el.parentNode;t&&t.removeChild(this.el),this.el=this._container=null},t}(),bq=_q,xq=function(){function t(t){this._show=!1,this._styleCoord=[0,0,0,0],this._alwaysShowContent=!1,this._enterable=!0,this._zr=t.getZr(),Sq(this._styleCoord,this._zr,t.getWidth()/2,t.getHeight()/2)}return t.prototype.update=function(t){var e=t.get("alwaysShowContent");e&&this._moveIfResized(),this._alwaysShowContent=e},t.prototype.show=function(){this._hideTimeout&&clearTimeout(this._hideTimeout),this.el.show(),this._show=!0},t.prototype.setContent=function(t,e,n,r,i){var o=this;s["A"](t)&&Cd(""),this.el&&this._zr.remove(this.el);var a=n.getModel("textStyle");this.el=new ks({style:{rich:e.richTextStyles,text:t,lineHeight:22,borderWidth:1,borderColor:r,textShadowColor:a.get("textShadowColor"),fill:n.get(["textStyle","color"]),padding:sp(n,"richText"),verticalAlign:"top",align:"left"},z:n.get("z")}),s["k"](["backgroundColor","borderRadius","shadowColor","shadowBlur","shadowOffsetX","shadowOffsetY"],(function(t){o.el.style[t]=n.get(t)})),s["k"](["textShadowBlur","textShadowOffsetX","textShadowOffsetY"],(function(t){o.el.style[t]=a.get(t)||0})),this._zr.add(this.el);var l=this;this.el.on("mouseover",(function(){l._enterable&&(clearTimeout(l._hideTimeout),l._show=!0),l._inContent=!0})),this.el.on("mouseout",(function(){l._enterable&&l._show&&l.hideLater(l._hideDelay),l._inContent=!1}))},t.prototype.setEnterable=function(t){this._enterable=t},t.prototype.getSize=function(){var t=this.el,e=this.el.getBoundingRect(),n=Mq(t.style);return[e.width+n.left+n.right,e.height+n.top+n.bottom]},t.prototype.moveTo=function(t,e){var n=this.el;if(n){var r=this._styleCoord;Sq(r,this._zr,t,e),t=r[0],e=r[1];var i=n.style,o=wq(i.borderWidth||0),a=Mq(i);n.x=t+o+a.left,n.y=e+o+a.top,n.markRedraw()}},t.prototype._moveIfResized=function(){var t=this._styleCoord[2],e=this._styleCoord[3];this.moveTo(t*this._zr.getWidth(),e*this._zr.getHeight())},t.prototype.hide=function(){this.el&&this.el.hide(),this._show=!1},t.prototype.hideLater=function(t){!this._show||this._inContent&&this._enterable||this._alwaysShowContent||(t?(this._hideDelay=t,this._show=!1,this._hideTimeout=setTimeout(s["c"](this.hide,this),t)):this.hide())},t.prototype.isShow=function(){return this._show},t.prototype.dispose=function(){this._zr.remove(this.el)},t}();function wq(t){return Math.max(0,t)}function Mq(t){var e=wq(t.shadowBlur||0),n=wq(t.shadowOffsetX||0),r=wq(t.shadowOffsetY||0);return{left:wq(e-n),right:wq(e+n),top:wq(e-r),bottom:wq(e+r)}}function Sq(t,e,n,r){t[0]=n,t[1]=r,t[2]=t[0]/e.getWidth(),t[3]=t[1]/e.getHeight()}var kq=xq,Lq=new ss({shape:{x:-1,y:-1,width:2,height:2}}),Tq=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(t,e){if(!a["a"].node&&e.getDom()){var n=t.getComponent("tooltip"),r=this._renderMode=zi(n.get("renderMode"));this._tooltipContent="richText"===r?new kq(e):new bq(e,{appendTo:n.get("appendToBody",!0)?"body":n.get("appendTo",!0)})}},e.prototype.render=function(t,e,n){if(!a["a"].node&&n.getDom()){this.group.removeAll(),this._tooltipModel=t,this._ecModel=e,this._api=n;var r=this._tooltipContent;r.update(t),r.setEnterable(t.get("enterable")),this._initGlobalListener(),this._keepShow(),"richText"!==this._renderMode&&t.get("transitionDuration")?Jg(this,"_updatePosition",50,"fixRate"):Kg(this,"_updatePosition")}},e.prototype._initGlobalListener=function(){var t=this._tooltipModel,e=t.get("triggerOn");eV("itemTooltip",this._api,Object(s["c"])((function(t,n,r){"none"!==e&&(e.indexOf(t)>=0?this._tryShow(n,r):"leave"===t&&this._hide(r))}),this))},e.prototype._keepShow=function(){var t=this._tooltipModel,e=this._ecModel,n=this._api,r=t.get("triggerOn");if(null!=this._lastX&&null!=this._lastY&&"none"!==r&&"click"!==r){var i=this;clearTimeout(this._refreshUpdateTimeout),this._refreshUpdateTimeout=setTimeout((function(){!n.isDisposed()&&i.manuallyShowTip(t,e,n,{x:i._lastX,y:i._lastY,dataByCoordSys:i._lastDataByCoordSys})}))}},e.prototype.manuallyShowTip=function(t,e,n,r){if(r.from!==this.uid&&!a["a"].node&&n.getDom()){var i=Oq(r,n);this._ticket="";var o=r.dataByCoordSys,s=jq(r,e,n);if(s){var l=s.el.getBoundingRect().clone();l.applyTransform(s.el.transform),this._tryShow({offsetX:l.x+l.width/2,offsetY:l.y+l.height/2,target:s.el,position:r.position,positionDefault:"bottom"},i)}else if(r.tooltip&&null!=r.x&&null!=r.y){var u=Lq;u.x=r.x,u.y=r.y,u.update(),Ls(u).tooltipConfig={name:null,option:r.tooltip},this._tryShow({offsetX:r.x,offsetY:r.y,target:u},i)}else if(o)this._tryShow({offsetX:r.x,offsetY:r.y,position:r.position,dataByCoordSys:o,tooltipOption:r.tooltipOption},i);else if(null!=r.seriesIndex){if(this._manuallyAxisShowTip(t,e,n,r))return;var c=cV(r,e),d=c.point[0],h=c.point[1];null!=d&&null!=h&&this._tryShow({offsetX:d,offsetY:h,target:c.el,position:r.position,positionDefault:"bottom"},i)}else null!=r.x&&null!=r.y&&(n.dispatchAction({type:"updateAxisPointer",x:r.x,y:r.y}),this._tryShow({offsetX:r.x,offsetY:r.y,position:r.position,target:n.getZr().findHover(r.x,r.y).target},i))}},e.prototype.manuallyHideTip=function(t,e,n,r){var i=this._tooltipContent;this._tooltipModel&&i.hideLater(this._tooltipModel.get("hideDelay")),this._lastX=this._lastY=this._lastDataByCoordSys=null,r.from!==this.uid&&this._hide(Oq(r,n))},e.prototype._manuallyAxisShowTip=function(t,e,n,r){var i=r.seriesIndex,o=r.dataIndex,a=e.getComponent("axisPointer").coordSysAxesInfo;if(null!=i&&null!=o&&null!=a){var s=e.getSeriesByIndex(i);if(s){var l=s.getData(),u=Dq([l.getItemModel(o),s,(s.coordinateSystem||{}).model],this._tooltipModel);if("axis"===u.get("trigger"))return n.dispatchAction({type:"updateAxisPointer",seriesIndex:i,dataIndex:o,position:r.position}),!0}}},e.prototype._tryShow=function(t,e){var n=t.target,r=this._tooltipModel;if(r){this._lastX=t.offsetX,this._lastY=t.offsetY;var i=t.dataByCoordSys;if(i&&i.length)this._showAxisTooltip(i,t);else if(n){var o,a,s=Ls(n);if("legend"===s.ssrType)return;this._lastDataByCoordSys=null,Gy(n,(function(t){return null!=Ls(t).dataIndex?(o=t,!0):null!=Ls(t).tooltipConfig?(a=t,!0):void 0}),!0),o?this._showSeriesItemTooltip(t,o,e):a?this._showComponentItemTooltip(t,a,e):this._hide(e)}else this._lastDataByCoordSys=null,this._hide(e)}},e.prototype._showOrMove=function(t,e){var n=t.get("showDelay");e=Object(s["c"])(e,this),clearTimeout(this._showTimout),n>0?this._showTimout=setTimeout(e,n):e()},e.prototype._showAxisTooltip=function(t,e){var n=this._ecModel,r=this._tooltipModel,i=[e.offsetX,e.offsetY],o=Dq([e.tooltipOption],r),a=this._renderMode,l=[],u=Uf("section",{blocks:[],noHeader:!0}),c=[],d=new lp;Object(s["k"])(t,(function(t){Object(s["k"])(t.dataByAxis,(function(t){var e=n.getComponent(t.axisDim+"Axis",t.axisIndex),i=t.value;if(e&&null!=i){var o=zW(i,e.axis,n,t.seriesDataIndices,t.valueLabelOpt),h=Uf("section",{header:o,noHeader:!Object(s["T"])(o),sortBlocks:!0,blocks:[]});u.blocks.push(h),Object(s["k"])(t.seriesDataIndices,(function(u){var f=n.getSeriesByIndex(u.seriesIndex),p=u.dataIndexInside,m=f.getDataParams(p);if(!(m.dataIndex<0)){m.axisDim=t.axisDim,m.axisIndex=t.axisIndex,m.axisType=t.axisType,m.axisId=t.axisId,m.axisValue=rD(e.axis,{value:i}),m.axisValueLabel=o,m.marker=d.makeTooltipMarker("item",Cc(m.color),a);var g=nf(f.formatTooltip(p,!0,null)),y=g.frag;if(y){var v=Dq([f],r).get("valueFormatter");h.blocks.push(v?Object(s["m"])({valueFormatter:v},y):y)}g.text&&c.push(g.text),l.push(m)}}))}}))})),u.blocks.reverse(),c.reverse();var h=e.position,f=o.get("order"),p=Qf(u,d,a,f,n.get("useUTC"),o.get("textStyle"));p&&c.unshift(p);var m="richText"===a?"\n\n":"<br/>",g=c.join(m);this._showOrMove(o,(function(){this._updateContentNotChangedOnAxis(t,l)?this._updatePosition(o,h,i[0],i[1],this._tooltipContent,l):this._showTooltipContent(o,g,l,Math.random()+"",i[0],i[1],h,null,d)}))},e.prototype._showSeriesItemTooltip=function(t,e,n){var r=this._ecModel,i=Ls(e),o=i.seriesIndex,a=r.getSeriesByIndex(o),l=i.dataModel||a,u=i.dataIndex,c=i.dataType,d=l.getData(c),h=this._renderMode,f=t.positionDefault,p=Dq([d.getItemModel(u),l,a&&(a.coordinateSystem||{}).model],this._tooltipModel,f?{position:f}:null),m=p.get("trigger");if(null==m||"item"===m){var g=l.getDataParams(u,c),y=new lp;g.marker=y.makeTooltipMarker("item",Cc(g.color),h);var v=nf(l.formatTooltip(u,!1,c)),_=p.get("order"),b=p.get("valueFormatter"),x=v.frag,w=x?Qf(b?Object(s["m"])({valueFormatter:b},x):x,y,h,_,r.get("useUTC"),p.get("textStyle")):v.text,M="item_"+l.name+"_"+u;this._showOrMove(p,(function(){this._showTooltipContent(p,w,g,M,t.offsetX,t.offsetY,t.position,t.target,y)})),n({type:"showTip",dataIndexInside:u,dataIndex:d.getRawIndex(u),seriesIndex:o,from:this.uid})}},e.prototype._showComponentItemTooltip=function(t,e,n){var r=Ls(e),i=r.tooltipConfig,o=i.option||{};if(Object(s["C"])(o)){var a=o;o={content:a,formatter:a}}var l=[o],u=this._ecModel.getComponent(r.componentMainType,r.componentIndex);u&&l.push(u),l.push({formatter:o.content});var c=t.positionDefault,d=Dq(l,this._tooltipModel,c?{position:c}:null),h=d.get("content"),f=Math.random()+"",p=new lp;this._showOrMove(d,(function(){var n=Object(s["d"])(d.get("formatterParams")||{});this._showTooltipContent(d,h,n,f,t.offsetX,t.offsetY,t.position,e,p)})),n({type:"showTip",from:this.uid})},e.prototype._showTooltipContent=function(t,e,n,r,i,o,a,l,u){if(this._ticket="",t.get("showContent")&&t.get("show")){var c=this._tooltipContent;c.setEnterable(t.get("enterable"));var d=t.get("formatter");a=a||t.get("position");var h=e,f=this._getNearestPoint([i,o],n,t.get("trigger"),t.get("borderColor")),p=f.color;if(d)if(Object(s["C"])(d)){var m=t.ecModel.get("useUTC"),g=Object(s["t"])(n)?n[0]:n,y=g&&g.axisType&&g.axisType.indexOf("time")>=0;h=d,y&&(h=ic(g.axisValue,h,m)),h=Dc(h,n,!0)}else if(Object(s["w"])(d)){var v=Object(s["c"])((function(e,r){e===this._ticket&&(c.setContent(r,u,t,p,a),this._updatePosition(t,a,i,o,c,n,l))}),this);this._ticket=r,h=d(n,r,v)}else h=d;c.setContent(h,u,t,p,a),c.show(t,p),this._updatePosition(t,a,i,o,c,n,l)}},e.prototype._getNearestPoint=function(t,e,n,r){return"axis"===n||Object(s["t"])(e)?{color:r||("html"===this._renderMode?"#fff":"none")}:Object(s["t"])(e)?void 0:{color:r||e.color||e.borderColor}},e.prototype._updatePosition=function(t,e,n,r,i,o,a){var l=this._api.getWidth(),u=this._api.getHeight();e=e||t.get("position");var c=i.getSize(),d=t.get("align"),h=t.get("verticalAlign"),f=a&&a.getBoundingRect().clone();if(a&&f.applyTransform(a.transform),Object(s["w"])(e)&&(e=e([n,r],o,i.el,f,{viewSize:[l,u],contentSize:c.slice()})),Object(s["t"])(e))n=Nr(e[0],l),r=Nr(e[1],u);else if(Object(s["A"])(e)){var p=e;p.width=c[0],p.height=c[1];var m=Hc(p,{width:l,height:u});n=m.x,r=m.y,d=null,h=null}else if(Object(s["C"])(e)&&a){var g=Aq(e,f,c,t.get("borderWidth"));n=g[0],r=g[1]}else{g=Iq(n,r,i,l,u,d?null:20,h?null:20);n=g[0],r=g[1]}if(d&&(n-=Yq(d)?c[0]/2:"right"===d?c[0]:0),h&&(r-=Yq(h)?c[1]/2:"bottom"===h?c[1]:0),rq(t)){g=Cq(n,r,i,l,u);n=g[0],r=g[1]}i.moveTo(n,r)},e.prototype._updateContentNotChangedOnAxis=function(t,e){var n=this._lastDataByCoordSys,r=this._cbParamsList,i=!!n&&n.length===t.length;return i&&Object(s["k"])(n,(function(n,o){var a=n.dataByAxis||[],l=t[o]||{},u=l.dataByAxis||[];i=i&&a.length===u.length,i&&Object(s["k"])(a,(function(t,n){var o=u[n]||{},a=t.seriesDataIndices||[],l=o.seriesDataIndices||[];i=i&&t.value===o.value&&t.axisType===o.axisType&&t.axisId===o.axisId&&a.length===l.length,i&&Object(s["k"])(a,(function(t,e){var n=l[e];i=i&&t.seriesIndex===n.seriesIndex&&t.dataIndex===n.dataIndex})),r&&Object(s["k"])(t.seriesDataIndices,(function(t){var n=t.seriesIndex,o=e[n],a=r[n];o&&a&&a.data!==o.data&&(i=!1)}))}))})),this._lastDataByCoordSys=t,this._cbParamsList=e,!!i},e.prototype._hide=function(t){this._lastDataByCoordSys=null,t({type:"hideTip",from:this.uid})},e.prototype.dispose=function(t,e){!a["a"].node&&e.getDom()&&(Kg(this,"_updatePosition"),this._tooltipContent.dispose(),sV("itemTooltip",e))},e.type="tooltip",e}(kp);function Dq(t,e,n){var r,i=e.ecModel;n?(r=new Lu(n,i,i),r=new Lu(e.option,r,i)):r=e;for(var o=t.length-1;o>=0;o--){var a=t[o];a&&(a instanceof Lu&&(a=a.get("tooltip",!0)),Object(s["C"])(a)&&(a={formatter:a}),a&&(r=new Lu(a,r,i)))}return r}function Oq(t,e){return t.dispatchAction||Object(s["c"])(e.dispatchAction,e)}function Iq(t,e,n,r,i,o,a){var s=n.getSize(),l=s[0],u=s[1];return null!=o&&(t+l+o+2>r?t-=l+o:t+=o),null!=a&&(e+u+a>i?e-=u+a:e+=a),[t,e]}function Cq(t,e,n,r,i){var o=n.getSize(),a=o[0],s=o[1];return t=Math.min(t+a,r)-a,e=Math.min(e+s,i)-s,t=Math.max(t,0),e=Math.max(e,0),[t,e]}function Aq(t,e,n,r){var i=n[0],o=n[1],a=Math.ceil(Math.SQRT2*r)+8,s=0,l=0,u=e.width,c=e.height;switch(t){case"inside":s=e.x+u/2-i/2,l=e.y+c/2-o/2;break;case"top":s=e.x+u/2-i/2,l=e.y-o-a;break;case"bottom":s=e.x+u/2-i/2,l=e.y+c+a;break;case"left":s=e.x-i-a,l=e.y+c/2-o/2;break;case"right":s=e.x+u+a,l=e.y+c/2-o/2}return[s,l]}function Yq(t){return"center"===t||"middle"===t}function jq(t,e,n){var r=ji(t).queryOptionMap,i=r.keys()[0];if(i&&"series"!==i){var o=Ni(e,i,r.get(i),{useDefault:!1,enableAll:!1,enableNone:!1}),a=o.models[0];if(a){var s,l=n.getViewOfComponentModel(a);return l.group.traverse((function(e){var n=Ls(e).tooltipConfig;if(n&&n.name===t.name)return s=e,!0})),s?{componentMainType:i,componentIndex:a.componentIndex,el:s}:void 0}}}var Pq=Tq;function Eq(t){jb(MV),t.registerComponentModel(nq),t.registerComponentView(Pq),t.registerAction({type:"showTip",event:"showTip",update:"tooltip:manuallyShowTip"},s["L"]),t.registerAction({type:"hideTip",event:"hideTip",update:"tooltip:manuallyHideTip"},s["L"])}var Nq=["rect","polygon","keep","clear"];function Rq(t,e){var n=ui(t?t.brush:[]);if(n.length){var r=[];s["k"](n,(function(t){var e=t.hasOwnProperty("toolbox")?t.toolbox:[];e instanceof Array&&(r=r.concat(e))}));var i=t&&t.toolbox;s["t"](i)&&(i=i[0]),i||(i={feature:{}},t.toolbox=[i]);var o=i.feature||(i.feature={}),a=o.brush||(o.brush={}),l=a.type||(a.type=[]);l.push.apply(l,r),Hq(l),e&&!l.length&&l.push.apply(l,Nq)}}function Hq(t){var e={};s["k"](t,(function(t){e[t]=1})),t.length=0,s["k"](e,(function(e,n){t.push(n)}))}var zq=s["k"];function Fq(t){if(t)for(var e in t)if(t.hasOwnProperty(e))return!0}function Bq(t,e,n){var r={};return zq(e,(function(e){var o=r[e]=i();zq(t[e],(function(t,r){if(uj.isValidType(r)){var i={type:r,visual:t};n&&n(i,e),o[r]=new uj(i),"opacity"===r&&(i=s["d"](i),i.type="colorAlpha",o.__hidden.__alphaForOpacity=new uj(i))}}))})),r;function i(){var t=function(){};t.prototype.__hidden=t.prototype;var e=new t;return e}}function Wq(t,e,n){var r;s["k"](n,(function(t){e.hasOwnProperty(t)&&Fq(e[t])&&(r=!0)})),r&&s["k"](n,(function(n){e.hasOwnProperty(n)&&Fq(e[n])?t[n]=s["d"](e[n]):delete t[n]}))}function Vq(t,e,n,r,i,o){var a,l={};function u(t){return Hy(n,a,t)}function c(t,e){Fy(n,a,t,e)}function d(t,s){a=null==o?t:s;var d=n.getRawDataItem(a);if(!d||!1!==d.visualMap)for(var h=r.call(i,t),f=e[h],p=l[h],m=0,g=p.length;m<g;m++){var y=p[m];f[y]&&f[y].applyVisual(t,u,c)}}s["k"](t,(function(t){var n=uj.prepareVisualTypes(e[t]);l[t]=n})),null==o?n.each(d):n.each([o],d)}function Gq(t,e,n,r){var i={};return s["k"](t,(function(t){var n=uj.prepareVisualTypes(e[t]);i[t]=n})),{progress:function(t,o){var a,s;function l(t){return Hy(o,s,t)}function u(t,e){Fy(o,s,t,e)}null!=r&&(a=o.getDimensionIndex(r));var c=o.getStore();while(null!=(s=t.next())){var d=o.getRawDataItem(s);if(!d||!1!==d.visualMap)for(var h=null!=r?c.get(a,s):s,f=n(h),p=e[f],m=i[f],g=0,y=m.length;g<y;g++){var v=m[g];p[v]&&p[v].applyVisual(h,l,u)}}}}}function $q(t){var e=t.brushType,n={point:function(r){return Uq[e].point(r,n,t)},rect:function(r){return Uq[e].rect(r,n,t)}};return n}var Uq={lineX:qq(0),lineY:qq(1),rect:{point:function(t,e,n){return t&&n.boundingRect.contain(t[0],t[1])},rect:function(t,e,n){return t&&n.boundingRect.intersect(t)}},polygon:{point:function(t,e,n){return t&&n.boundingRect.contain(t[0],t[1])&&GI(n.range,t[0],t[1])},rect:function(t,e,n){var r=n.range;if(!t||r.length<=1)return!1;var i=t.x,o=t.y,a=t.width,s=t.height,l=r[0];return!!(GI(r,i,o)||GI(r,i+a,o)||GI(r,i,o+s)||GI(r,i+a,o+s)||Tt.create(t).contain(l[0],l[1])||Cg(i,o,i+a,o,r)||Cg(i,o,i,o+s,r)||Cg(i+a,o,i+a,o+s,r)||Cg(i,o+s,i+a,o+s,r))||void 0}}};function qq(t){var e=["x","y"],n=["width","height"];return{point:function(e,n,r){if(e){var i=r.range,o=e[t];return Xq(o,i)}},rect:function(r,i,o){if(r){var a=o.range,s=[r[e[t]],r[e[t]]+r[n[t]]];return s[1]<s[0]&&s.reverse(),Xq(s[0],a)||Xq(s[1],a)||Xq(a[0],s)||Xq(a[1],s)}}}}function Xq(t,e){return e[0]<=t&&t<=e[1]}var Zq=["inBrush","outOfBrush"],Jq="__ecBrushSelect",Kq="__ecInBrushSelectEvent";function Qq(t){t.eachComponent({mainType:"brush"},(function(e){var n=e.brushTargetManager=new GU(e.option,t);n.setInputRanges(e.areas,t)}))}function tX(t,e,n){var r,i,o=[];t.eachComponent({mainType:"brush"},(function(t){n&&"takeGlobalCursor"===n.type&&t.setBrushOption("brush"===n.key?n.brushOption:{brushType:!1})})),Qq(t),t.eachComponent({mainType:"brush"},(function(e,n){var a={brushId:e.id,brushIndex:n,brushName:e.name,areas:s["d"](e.areas),selected:[]};o.push(a);var l=e.option,u=l.brushLink,c=[],d=[],h=[],f=!1;n||(r=l.throttleType,i=l.throttleDelay);var p=s["H"](e.areas,(function(t){var e=oX[t.brushType],n=s["i"]({boundingRect:e?e(t):void 0},t);return n.selectors=$q(n),n})),m=Bq(e.option,Zq,(function(t){t.mappingMethod="fixed"}));function g(t){return"all"===u||!!c[t]}function y(t){return!!t.length}function v(t,e){var n=t.coordinateSystem;f=f||n.hasAxisBrushed(),g(e)&&n.eachActiveState(t.getData(),(function(t,e){"active"===t&&(d[e]=1)}))}function _(n,r,i){if(n.brushSelector&&!iX(e,r)&&(s["k"](p,(function(r){e.brushTargetManager.controlSeries(r,n,t)&&i.push(r),f=f||y(i)})),g(r)&&y(i))){var o=n.getData();o.each((function(t){rX(n,i,o,t)&&(d[t]=1)}))}}s["t"](u)&&s["k"](u,(function(t){c[t]=1})),t.eachSeries((function(t,e){var n=h[e]=[];"parallel"===t.subType?v(t,e):_(t,e,n)})),t.eachSeries((function(t,e){var n={seriesId:t.id,seriesIndex:e,seriesName:t.name,dataIndex:[]};a.selected.push(n);var r=h[e],i=t.getData(),o=g(e)?function(t){return d[t]?(n.dataIndex.push(i.getRawIndex(t)),"inBrush"):"outOfBrush"}:function(e){return rX(t,r,i,e)?(n.dataIndex.push(i.getRawIndex(e)),"inBrush"):"outOfBrush"};(g(e)?f:y(r))&&Vq(Zq,m,i,o)}))})),eX(e,r,i,o,n)}function eX(t,e,n,r,i){if(i){var o=t.getZr();if(!o[Kq]){o[Jq]||(o[Jq]=nX);var a=Jg(o,Jq,n,e);a(t,r)}}}function nX(t,e){if(!t.isDisposed()){var n=t.getZr();n[Kq]=!0,t.dispatchAction({type:"brushSelect",batch:e}),n[Kq]=!1}}function rX(t,e,n,r){for(var i=0,o=e.length;i<o;i++){var a=e[i];if(t.brushSelector(r,n,a.selectors,a))return!0}}function iX(t,e){var n=t.option.seriesIndex;return null!=n&&"all"!==n&&(s["t"](n)?s["r"](n,e)<0:e!==n)}var oX={rect:function(t){return aX(t.range)},polygon:function(t){for(var e,n=t.range,r=0,i=n.length;r<i;r++){e=e||[[1/0,-1/0],[1/0,-1/0]];var o=n[r];o[0]<e[0][0]&&(e[0][0]=o[0]),o[0]>e[0][1]&&(e[0][1]=o[0]),o[1]<e[1][0]&&(e[1][0]=o[1]),o[1]>e[1][1]&&(e[1][1]=o[1])}return e&&aX(e)}};function aX(t){return new Tt(t[0][0],t[1][0],t[0][1]-t[0][0],t[1][1]-t[1][0])}var sX=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(t,e){this.ecModel=t,this.api=e,this.model,(this._brushController=new yR(e.getZr())).on("brush",s["c"](this._onBrush,this)).mount()},e.prototype.render=function(t,e,n,r){this.model=t,this._updateController(t,e,n,r)},e.prototype.updateTransform=function(t,e,n,r){Qq(e),this._updateController(t,e,n,r)},e.prototype.updateVisual=function(t,e,n,r){this.updateTransform(t,e,n,r)},e.prototype.updateView=function(t,e,n,r){this._updateController(t,e,n,r)},e.prototype._updateController=function(t,e,n,r){(!r||r.$from!==t.id)&&this._brushController.setPanels(t.brushTargetManager.makePanelOpts(n)).enableBrush(t.brushOption).updateCovers(t.areas.slice())},e.prototype.dispose=function(){this._brushController.dispose()},e.prototype._onBrush=function(t){var e=this.model.id,n=this.model.brushTargetManager.setOutputRanges(t.areas,this.ecModel);(!t.isEnd||t.removeOnClick)&&this.api.dispatchAction({type:"brush",brushId:e,areas:s["d"](n),$from:e}),t.isEnd&&this.api.dispatchAction({type:"brushEnd",brushId:e,areas:s["d"](n),$from:e})},e.type="brush",e}(kp),lX=sX,uX="#ddd",cX=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.areas=[],n.brushOption={},n}return o(e,t),e.prototype.optionUpdated=function(t,e){var n=this.option;!e&&Wq(n,t,["inBrush","outOfBrush"]);var r=n.inBrush=n.inBrush||{};n.outOfBrush=n.outOfBrush||{color:uX},r.hasOwnProperty("liftZ")||(r.liftZ=5)},e.prototype.setAreas=function(t){t&&(this.areas=s["H"](t,(function(t){return dX(this.option,t)}),this))},e.prototype.setBrushOption=function(t){this.brushOption=dX(this.option,t),this.brushType=this.brushOption.brushType},e.type="brush",e.dependencies=["geo","grid","xAxis","yAxis","parallel","series"],e.defaultOption={seriesIndex:"all",brushType:"rect",brushMode:"single",transformable:!0,brushStyle:{borderWidth:1,color:"rgba(210,219,238,0.3)",borderColor:"#D2DBEE"},throttleType:"fixRate",throttleDelay:0,removeOnClick:!0,z:1e4},e}(Xc);function dX(t,e){return s["I"]({brushType:t.brushType,brushMode:t.brushMode,transformable:t.transformable,brushStyle:new Lu(t.brushStyle).getItemStyle(),removeOnClick:t.removeOnClick,z:t.z},e,!0)}var hX=cX,fX=["rect","polygon","lineX","lineY","keep","clear"],pX=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.render=function(t,e,n){var r,i,o;e.eachComponent({mainType:"brush"},(function(t){r=t.brushType,i=t.brushOption.brushMode||"single",o=o||!!t.areas.length})),this._brushType=r,this._brushMode=i,s["k"](t.get("type",!0),(function(e){t.setIconStatus(e,("keep"===e?"multiple"===i:"clear"===e?o:e===r)?"emphasis":"normal")}))},e.prototype.updateView=function(t,e,n){this.render(t,e,n)},e.prototype.getIcons=function(){var t=this.model,e=t.get("icon",!0),n={};return s["k"](t.get("type",!0),(function(t){e[t]&&(n[t]=e[t])})),n},e.prototype.onclick=function(t,e,n){var r=this._brushType,i=this._brushMode;"clear"===n?(e.dispatchAction({type:"axisAreaSelect",intervals:[]}),e.dispatchAction({type:"brush",command:"clear",areas:[]})):e.dispatchAction({type:"takeGlobalCursor",key:"brush",brushOption:{brushType:"keep"===n?r:r!==n&&n,brushMode:"keep"===n?"multiple"===i?"single":"multiple":i}})},e.getDefaultOption=function(t){var e={show:!0,type:fX.slice(),icon:{rect:"M7.3,34.7 M0.4,10V-0.2h9.8 M89.6,10V-0.2h-9.8 M0.4,60v10.2h9.8 M89.6,60v10.2h-9.8 M12.3,22.4V10.5h13.1 M33.6,10.5h7.8 M49.1,10.5h7.8 M77.5,22.4V10.5h-13 M12.3,31.1v8.2 M77.7,31.1v8.2 M12.3,47.6v11.9h13.1 M33.6,59.5h7.6 M49.1,59.5 h7.7 M77.5,47.6v11.9h-13",polygon:"M55.2,34.9c1.7,0,3.1,1.4,3.1,3.1s-1.4,3.1-3.1,3.1 s-3.1-1.4-3.1-3.1S53.5,34.9,55.2,34.9z M50.4,51c1.7,0,3.1,1.4,3.1,3.1c0,1.7-1.4,3.1-3.1,3.1c-1.7,0-3.1-1.4-3.1-3.1 C47.3,52.4,48.7,51,50.4,51z M55.6,37.1l1.5-7.8 M60.1,13.5l1.6-8.7l-7.8,4 M59,19l-1,5.3 M24,16.1l6.4,4.9l6.4-3.3 M48.5,11.6 l-5.9,3.1 M19.1,12.8L9.7,5.1l1.1,7.7 M13.4,29.8l1,7.3l6.6,1.6 M11.6,18.4l1,6.1 M32.8,41.9 M26.6,40.4 M27.3,40.2l6.1,1.6 M49.9,52.1l-5.6-7.6l-4.9-1.2",lineX:"M15.2,30 M19.7,15.6V1.9H29 M34.8,1.9H40.4 M55.3,15.6V1.9H45.9 M19.7,44.4V58.1H29 M34.8,58.1H40.4 M55.3,44.4 V58.1H45.9 M12.5,20.3l-9.4,9.6l9.6,9.8 M3.1,29.9h16.5 M62.5,20.3l9.4,9.6L62.3,39.7 M71.9,29.9H55.4",lineY:"M38.8,7.7 M52.7,12h13.2v9 M65.9,26.6V32 M52.7,46.3h13.2v-9 M24.9,12H11.8v9 M11.8,26.6V32 M24.9,46.3H11.8v-9 M48.2,5.1l-9.3-9l-9.4,9.2 M38.9-3.9V12 M48.2,53.3l-9.3,9l-9.4-9.2 M38.9,62.3V46.4",keep:"M4,10.5V1h10.3 M20.7,1h6.1 M33,1h6.1 M55.4,10.5V1H45.2 M4,17.3v6.6 M55.6,17.3v6.6 M4,30.5V40h10.3 M20.7,40 h6.1 M33,40h6.1 M55.4,30.5V40H45.2 M21,18.9h62.9v48.6H21V18.9z",clear:"M22,14.7l30.9,31 M52.9,14.7L22,45.7 M4.7,16.8V4.2h13.1 M26,4.2h7.8 M41.6,4.2h7.8 M70.3,16.8V4.2H57.2 M4.7,25.9v8.6 M70.3,25.9v8.6 M4.7,43.2v12.6h13.1 M26,55.8h7.8 M41.6,55.8h7.8 M70.3,43.2v12.6H57.2"},title:t.getLocaleModel().get(["toolbox","brush","title"])};return e},e}(V$),mX=pX;function gX(t){t.registerComponentView(lX),t.registerComponentModel(hX),t.registerPreprocessor(Rq),t.registerVisual(t.PRIORITY.VISUAL.BRUSH,tX),t.registerAction({type:"brush",event:"brush",update:"updateVisual"},(function(t,e){e.eachComponent({mainType:"brush",query:t},(function(e){e.setAreas(t.areas)}))})),t.registerAction({type:"brushSelect",event:"brushSelected",update:"none"},s["L"]),t.registerAction({type:"brushEnd",event:"brushEnd",update:"none"},s["L"]),$$("brush",mX)}var yX=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.layoutMode={type:"box",ignoreSize:!0},n}return o(e,t),e.type="title",e.defaultOption={z:6,show:!0,text:"",target:"blank",subtext:"",subtarget:"blank",left:0,top:0,backgroundColor:"rgba(0,0,0,0)",borderColor:"#ccc",borderWidth:0,padding:5,itemGap:10,textStyle:{fontSize:18,fontWeight:"bold",color:"#464646"},subtextStyle:{fontSize:12,color:"#6E7079"}},e}(Xc),vX=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){if(this.group.removeAll(),t.get("show")){var r=this.group,i=t.getModel("textStyle"),o=t.getModel("subtextStyle"),a=t.get("textAlign"),l=s["P"](t.get("textBaseline"),t.get("textVerticalAlign")),u=new ks({style:nu(i,{text:t.get("text"),fill:i.getTextColor()},{disableBox:!0}),z2:10}),c=u.getBoundingRect(),d=t.get("subtext"),h=new ks({style:nu(o,{text:d,fill:o.getTextColor(),y:c.height+t.get("itemGap"),verticalAlign:"top"},{disableBox:!0}),z2:10}),f=t.get("link"),p=t.get("sublink"),m=t.get("triggerEvent",!0);u.silent=!f&&!m,h.silent=!p&&!m,f&&u.on("click",(function(){Ac(f,"_"+t.get("target"))})),p&&h.on("click",(function(){Ac(p,"_"+t.get("subtarget"))})),Ls(u).eventData=Ls(h).eventData=m?{componentType:"title",componentIndex:t.componentIndex}:null,r.add(u),d&&r.add(h);var g=r.getBoundingRect(),y=t.getBoxLayoutParams();y.width=g.width,y.height=g.height;var v=Hc(y,{width:n.getWidth(),height:n.getHeight()},t.get("padding"));a||(a=t.get("left")||t.get("right"),"middle"===a&&(a="center"),"right"===a?v.x+=v.width:"center"===a&&(v.x+=v.width/2)),l||(l=t.get("top")||t.get("bottom"),"center"===l&&(l="middle"),"bottom"===l?v.y+=v.height:"middle"===l&&(v.y+=v.height/2),l=l||"top"),r.x=v.x,r.y=v.y,r.markRedraw();var _={align:a,verticalAlign:l};u.setStyle(_),h.setStyle(_),g=r.getBoundingRect();var b=v.margin,x=t.getItemStyle(["color","opacity"]);x.fill=t.get("backgroundColor");var w=new ss({shape:{x:g.x-b[3],y:g.y-b[0],width:g.width+b[1]+b[3],height:g.height+b[0]+b[2],r:t.get("borderRadius")},style:x,subPixelOptimize:!0,silent:!0});r.add(w)}},e.type="title",e}(kp);function _X(t){t.registerComponentModel(yX),t.registerComponentView(vX)}var bX=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.layoutMode="box",n}return o(e,t),e.prototype.init=function(t,e,n){this.mergeDefaultAndTheme(t,n),this._initData()},e.prototype.mergeOption=function(e){t.prototype.mergeOption.apply(this,arguments),this._initData()},e.prototype.setCurrentIndex=function(t){null==t&&(t=this.option.currentIndex);var e=this._data.count();this.option.loop?t=(t%e+e)%e:(t>=e&&(t=e-1),t<0&&(t=0)),this.option.currentIndex=t},e.prototype.getCurrentIndex=function(){return this.option.currentIndex},e.prototype.isIndexMax=function(){return this.getCurrentIndex()>=this._data.count()-1},e.prototype.setPlayState=function(t){this.option.autoPlay=!!t},e.prototype.getPlayState=function(){return!!this.option.autoPlay},e.prototype._initData=function(){var t,e=this.option,n=e.data||[],r=e.axisType,i=this._names=[];"category"===r?(t=[],Object(s["k"])(n,(function(e,n){var r,o=Mi(hi(e),"");Object(s["A"])(e)?(r=Object(s["d"])(e),r.value=n):r=n,t.push(r),i.push(o)}))):t=n;var o={category:"ordinal",time:"time",value:"number"}[r]||"number",a=this._data=new UM([{name:"value",type:o}],this);a.initData(t,i)},e.prototype.getData=function(){return this._data},e.prototype.getCategories=function(){if("category"===this.get("axisType"))return this._names.slice()},e.type="timeline",e.defaultOption={z:4,show:!0,axisType:"time",realtime:!0,left:"20%",top:null,right:"20%",bottom:0,width:null,height:40,padding:5,controlPosition:"left",autoPlay:!1,rewind:!1,loop:!0,playInterval:2e3,currentIndex:0,itemStyle:{},label:{color:"#000"},data:[]},e}(Xc),xX=bX,wX=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="timeline.slider",e.defaultOption=Cu(xX.defaultOption,{backgroundColor:"rgba(0,0,0,0)",borderColor:"#ccc",borderWidth:0,orient:"horizontal",inverse:!1,tooltip:{trigger:"item"},symbol:"circle",symbolSize:12,lineStyle:{show:!0,width:2,color:"#DAE1F5"},label:{position:"auto",show:!0,interval:"auto",rotate:0,color:"#A4B1D7"},itemStyle:{color:"#A4B1D7",borderWidth:1},checkpointStyle:{symbol:"circle",symbolSize:15,color:"#316bf3",borderColor:"#fff",borderWidth:2,shadowBlur:2,shadowOffsetX:1,shadowOffsetY:1,shadowColor:"rgba(0, 0, 0, 0.3)",animation:!0,animationDuration:300,animationEasing:"quinticInOut"},controlStyle:{show:!0,showPlayBtn:!0,showPrevBtn:!0,showNextBtn:!0,itemSize:24,itemGap:12,position:"left",playIcon:"path://M31.6,53C17.5,53,6,41.5,6,27.4S17.5,1.8,31.6,1.8C45.7,1.8,57.2,13.3,57.2,27.4S45.7,53,31.6,53z M31.6,3.3 C18.4,3.3,7.5,14.1,7.5,27.4c0,13.3,10.8,24.1,24.1,24.1C44.9,51.5,55.7,40.7,55.7,27.4C55.7,14.1,44.9,3.3,31.6,3.3z M24.9,21.3 c0-2.2,1.6-3.1,3.5-2l10.5,6.1c1.899,1.1,1.899,2.9,0,4l-10.5,6.1c-1.9,1.1-3.5,0.2-3.5-2V21.3z",stopIcon:"path://M30.9,53.2C16.8,53.2,5.3,41.7,5.3,27.6S16.8,2,30.9,2C45,2,56.4,13.5,56.4,27.6S45,53.2,30.9,53.2z M30.9,3.5C17.6,3.5,6.8,14.4,6.8,27.6c0,13.3,10.8,24.1,24.101,24.1C44.2,51.7,55,40.9,55,27.6C54.9,14.4,44.1,3.5,30.9,3.5z M36.9,35.8c0,0.601-0.4,1-0.9,1h-1.3c-0.5,0-0.9-0.399-0.9-1V19.5c0-0.6,0.4-1,0.9-1H36c0.5,0,0.9,0.4,0.9,1V35.8z M27.8,35.8 c0,0.601-0.4,1-0.9,1h-1.3c-0.5,0-0.9-0.399-0.9-1V19.5c0-0.6,0.4-1,0.9-1H27c0.5,0,0.9,0.4,0.9,1L27.8,35.8L27.8,35.8z",nextIcon:"M2,18.5A1.52,1.52,0,0,1,.92,18a1.49,1.49,0,0,1,0-2.12L7.81,9.36,1,3.11A1.5,1.5,0,1,1,3,.89l8,7.34a1.48,1.48,0,0,1,.49,1.09,1.51,1.51,0,0,1-.46,1.1L3,18.08A1.5,1.5,0,0,1,2,18.5Z",prevIcon:"M10,.5A1.52,1.52,0,0,1,11.08,1a1.49,1.49,0,0,1,0,2.12L4.19,9.64,11,15.89a1.5,1.5,0,1,1-2,2.22L1,10.77A1.48,1.48,0,0,1,.5,9.68,1.51,1.51,0,0,1,1,8.58L9,.92A1.5,1.5,0,0,1,10,.5Z",prevBtnSize:18,nextBtnSize:18,color:"#A4B1D7",borderColor:"#A4B1D7",borderWidth:1},emphasis:{label:{show:!0,color:"#6f778d"},itemStyle:{color:"#316BF3"},controlStyle:{color:"#316BF3",borderColor:"#316BF3",borderWidth:2}},progress:{lineStyle:{color:"#316BF3"},itemStyle:{color:"#316BF3"},label:{color:"#6f778d"}},data:[]}),e}(xX);Object(s["K"])(wX,ef.prototype);var MX=wX,SX=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="timeline",e}(kp),kX=SX,LX=function(t){function e(e,n,r,i){var o=t.call(this,e,n,r)||this;return o.type=i||"value",o}return o(e,t),e.prototype.getLabelModel=function(){return this.model.getModel("label")},e.prototype.isHorizontal=function(){return"horizontal"===this.model.get("orient")},e}(PD),TX=LX,DX=Math.PI,OX=Ci(),IX=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(t,e){this.api=e},e.prototype.render=function(t,e,n){if(this.model=t,this.api=n,this.ecModel=e,this.group.removeAll(),t.get("show",!0)){var r=this._layout(t,n),i=this._createGroup("_mainGroup"),o=this._createGroup("_labelGroup"),a=this._axis=this._createAxis(r,t);t.formatTooltip=function(t){var e=a.scale.getLabel({value:t});return Uf("nameValue",{noName:!0,value:e})},Object(s["k"])(["AxisLine","AxisTick","Control","CurrentPointer"],(function(e){this["_render"+e](r,i,a,t)}),this),this._renderAxisLabel(r,o,a,t),this._position(r,t)}this._doPlayStop(),this._updateTicksStatus()},e.prototype.remove=function(){this._clearTimer(),this.group.removeAll()},e.prototype.dispose=function(){this._clearTimer()},e.prototype._layout=function(t,e){var n,r=t.get(["label","position"]),i=t.get("orient"),o=AX(t,e);n=null==r||"auto"===r?"horizontal"===i?o.y+o.height/2<e.getHeight()/2?"-":"+":o.x+o.width/2<e.getWidth()/2?"+":"-":Object(s["C"])(r)?{horizontal:{top:"-",bottom:"+"},vertical:{left:"-",right:"+"}}[i][r]:r;var a,l,u,c={horizontal:"center",vertical:n>=0||"+"===n?"left":"right"},d={horizontal:n>=0||"+"===n?"top":"bottom",vertical:"middle"},h={horizontal:0,vertical:DX/2},f="vertical"===i?o.height:o.width,p=t.getModel("controlStyle"),m=p.get("show",!0),g=m?p.get("itemSize"):0,y=m?p.get("itemGap"):0,v=g+y,_=t.get(["label","rotate"])||0;_=_*DX/180;var b=p.get("position",!0),x=m&&p.get("showPlayBtn",!0),w=m&&p.get("showPrevBtn",!0),M=m&&p.get("showNextBtn",!0),S=0,k=f;"left"===b||"bottom"===b?(x&&(a=[0,0],S+=v),w&&(l=[S,0],S+=v),M&&(u=[k-g,0],k-=v)):(x&&(a=[k-g,0],k-=v),w&&(l=[0,0],S+=v),M&&(u=[k-g,0],k-=v));var L=[S,k];return t.get("inverse")&&L.reverse(),{viewRect:o,mainLength:f,orient:i,rotation:h[i],labelRotation:_,labelPosOpt:n,labelAlign:t.get(["label","align"])||c[i],labelBaseline:t.get(["label","verticalAlign"])||t.get(["label","baseline"])||d[i],playPosition:a,prevBtnPosition:l,nextBtnPosition:u,axisExtent:L,controlSize:g,controlGap:y}},e.prototype._position=function(t,e){var n=this._mainGroup,r=this._labelGroup,i=t.viewRect;if("vertical"===t.orient){var o=st(),a=i.x,l=i.y+i.height;dt(o,o,[-a,-l]),ht(o,o,-DX/2),dt(o,o,[a,l]),i=i.clone(),i.applyTransform(o)}var u=y(i),c=y(n.getBoundingRect()),d=y(r.getBoundingRect()),h=[n.x,n.y],f=[r.x,r.y];f[0]=h[0]=u[0][0];var p=t.labelPosOpt;if(null==p||Object(s["C"])(p)){var m="+"===p?0:1;v(h,c,u,1,m),v(f,d,u,1,1-m)}else{m=p>=0?0:1;v(h,c,u,1,m),f[1]=h[1]+p}function g(t){t.originX=u[0][0]-t.x,t.originY=u[1][0]-t.y}function y(t){return[[t.x,t.x+t.width],[t.y,t.y+t.height]]}function v(t,e,n,r,i){t[r]+=n[r][i]-e[r][i]}n.setPosition(h),r.setPosition(f),n.rotation=r.rotation=t.rotation,g(n),g(r)},e.prototype._createAxis=function(t,e){var n=e.getData(),r=e.get("axisType"),i=CX(e,r);i.getTicks=function(){return n.mapArray(["value"],(function(t){return{value:t}}))};var o=n.getDataExtent("value");i.setExtent(o[0],o[1]),i.calcNiceTicks();var a=new TX("value",i,t.axisExtent,r);return a.model=e,a},e.prototype._createGroup=function(t){var e=this[t]=new wr;return this.group.add(e),e},e.prototype._renderAxisLine=function(t,e,n,r){var i=n.getExtent();if(r.get(["lineStyle","show"])){var o=new jm({shape:{x1:i[0],y1:0,x2:i[1],y2:0},style:Object(s["m"])({lineCap:"round"},r.getModel("lineStyle").getLineStyle()),silent:!0,z2:1});e.add(o);var a=this._progressLine=new jm({shape:{x1:i[0],x2:this._currentPointer?this._currentPointer.x:i[0],y1:0,y2:0},style:Object(s["i"])({lineCap:"round",lineWidth:o.style.lineWidth},r.getModel(["progress","lineStyle"]).getLineStyle()),silent:!0,z2:1});e.add(a)}},e.prototype._renderAxisTick=function(t,e,n,r){var i=this,o=r.getData(),a=n.scale.getTicks();this._tickSymbols=[],Object(s["k"])(a,(function(t){var a=n.dataToCoord(t.value),l=o.getItemModel(t.value),u=l.getModel("itemStyle"),c=l.getModel(["emphasis","itemStyle"]),d=l.getModel(["progress","itemStyle"]),h={x:a,y:0,onclick:Object(s["c"])(i._changeTimeline,i,t.value)},f=jX(l,u,e,h);f.ensureState("emphasis").style=c.getItemStyle(),f.ensureState("progress").style=d.getItemStyle(),Ll(f);var p=Ls(f);l.get("tooltip")?(p.dataIndex=t.value,p.dataModel=r):p.dataIndex=p.dataModel=null,i._tickSymbols.push(f)}))},e.prototype._renderAxisLabel=function(t,e,n,r){var i=this,o=n.getLabelModel();if(o.get("show")){var a=r.getData(),l=n.getViewLabels();this._tickLabels=[],Object(s["k"])(l,(function(r){var o=r.tickValue,l=a.getItemModel(o),u=l.getModel("label"),c=l.getModel(["emphasis","label"]),d=l.getModel(["progress","label"]),h=n.dataToCoord(r.tickValue),f=new ks({x:h,y:0,rotation:t.labelRotation-t.rotation,onclick:Object(s["c"])(i._changeTimeline,i,o),silent:!1,style:nu(u,{text:r.formattedLabel,align:t.labelAlign,verticalAlign:t.labelBaseline})});f.ensureState("emphasis").style=nu(c),f.ensureState("progress").style=nu(d),e.add(f),Ll(f),OX(f).dataIndex=o,i._tickLabels.push(f)}))}},e.prototype._renderControl=function(t,e,n,r){var i=t.controlSize,o=t.rotation,a=r.getModel("controlStyle").getItemStyle(),l=r.getModel(["emphasis","controlStyle"]).getItemStyle(),u=r.getPlayState(),c=r.get("inverse",!0);function d(t,n,u,c){if(t){var d=or(Object(s["P"])(r.get(["controlStyle",n+"BtnSize"]),i),i),h=[0,-d/2,d,d],f=YX(r,n+"Icon",h,{x:t[0],y:t[1],originX:i/2,originY:0,rotation:c?-o:0,rectHover:!0,style:a,onclick:u});f.ensureState("emphasis").style=l,e.add(f),Ll(f)}}d(t.nextBtnPosition,"next",Object(s["c"])(this._changeTimeline,this,c?"-":"+")),d(t.prevBtnPosition,"prev",Object(s["c"])(this._changeTimeline,this,c?"+":"-")),d(t.playPosition,u?"stop":"play",Object(s["c"])(this._handlePlayClick,this,!u),!0)},e.prototype._renderCurrentPointer=function(t,e,n,r){var i=r.getData(),o=r.getCurrentIndex(),a=i.getItemModel(o).getModel("checkpointStyle"),l=this,u={onCreate:function(t){t.draggable=!0,t.drift=Object(s["c"])(l._handlePointerDrag,l),t.ondragend=Object(s["c"])(l._handlePointerDragend,l),PX(t,l._progressLine,o,n,r,!0)},onUpdate:function(t){PX(t,l._progressLine,o,n,r)}};this._currentPointer=jX(a,a,this._mainGroup,{},this._currentPointer,u)},e.prototype._handlePlayClick=function(t){this._clearTimer(),this.api.dispatchAction({type:"timelinePlayChange",playState:t,from:this.uid})},e.prototype._handlePointerDrag=function(t,e,n){this._clearTimer(),this._pointerChangeTimeline([n.offsetX,n.offsetY])},e.prototype._handlePointerDragend=function(t){this._pointerChangeTimeline([t.offsetX,t.offsetY],!0)},e.prototype._pointerChangeTimeline=function(t,e){var n=this._toAxisCoord(t)[0],r=this._axis,i=Hr(r.getExtent().slice());n>i[1]&&(n=i[1]),n<i[0]&&(n=i[0]),this._currentPointer.x=n,this._currentPointer.markRedraw();var o=this._progressLine;o&&(o.shape.x2=n,o.dirty());var a=this._findNearestTick(n),s=this.model;(e||a!==s.getCurrentIndex()&&s.get("realtime"))&&this._changeTimeline(a)},e.prototype._doPlayStop=function(){var t=this;this._clearTimer(),this.model.getPlayState()&&(this._timer=setTimeout((function(){var e=t.model;t._changeTimeline(e.getCurrentIndex()+(e.get("rewind",!0)?-1:1))}),this.model.get("playInterval")))},e.prototype._toAxisCoord=function(t){var e=this._mainGroup.getLocalTransform();return Mg(t,e,!0)},e.prototype._findNearestTick=function(t){var e,n=this.model.getData(),r=1/0,i=this._axis;return n.each(["value"],(function(n,o){var a=i.dataToCoord(n),s=Math.abs(a-t);s<r&&(r=s,e=o)})),e},e.prototype._clearTimer=function(){this._timer&&(clearTimeout(this._timer),this._timer=null)},e.prototype._changeTimeline=function(t){var e=this.model.getCurrentIndex();"+"===t?t=e+1:"-"===t&&(t=e-1),this.api.dispatchAction({type:"timelineChange",currentIndex:t,from:this.uid})},e.prototype._updateTicksStatus=function(){var t=this.model.getCurrentIndex(),e=this._tickSymbols,n=this._tickLabels;if(e)for(var r=0;r<e.length;r++)e&&e[r]&&e[r].toggleState("progress",r<t);if(n)for(r=0;r<n.length;r++)n&&n[r]&&n[r].toggleState("progress",OX(n[r]).dataIndex<=t)},e.type="timeline.slider",e}(kX);function CX(t,e){if(e=e||t.get("type"),e)switch(e){case"category":return new _T({ordinalMeta:t.getCategories(),extent:[1/0,-1/0]});case"time":return new jT({locale:t.ecModel.getLocaleModel(),useUTC:t.ecModel.get("useUTC")});default:return new wT}}function AX(t,e){return Hc(t.getBoxLayoutParams(),{width:e.getWidth(),height:e.getHeight()},t.get("padding"))}function YX(t,e,n,r){var i=r.style,o=Ig(t.get(["controlStyle",e]),r||{},new Tt(n[0],n[1],n[2],n[3]));return i&&o.setStyle(i),o}function jX(t,e,n,r,i,o){var a=e.get("color");if(i)i.setColor(a),n.add(i),o&&o.onUpdate(i);else{var l=t.get("symbol");i=ov(l,-1,-1,2,2,a),i.setStyle("strokeNoScale",!0),n.add(i),o&&o.onCreate(i)}var u=e.getItemStyle(["color"]);i.setStyle(u),r=Object(s["I"])({rectHover:!0,z2:100},r,!0);var c=av(t.get("symbolSize"));r.scaleX=c[0]/2,r.scaleY=c[1]/2;var d=sv(t.get("symbolOffset"),c);d&&(r.x=(r.x||0)+d[0],r.y=(r.y||0)+d[1]);var h=t.get("symbolRotate");return r.rotation=(h||0)*Math.PI/180||0,i.attr(r),i.updateTransform(),i}function PX(t,e,n,r,i,o){if(!t.dragging){var a=i.getModel("checkpointStyle"),s=r.dataToCoord(i.getData().get("value",n));if(o||!a.get("animation",!0))t.attr({x:s,y:0}),e&&e.attr({shape:{x2:s}});else{var l={duration:a.get("animationDuration",!0),easing:a.get("animationEasing",!0)};t.stopAnimation(null,!0),t.animateTo({x:s,y:0},l),e&&e.animateTo({shape:{x2:s}},l)}}}var EX=IX;function NX(t){t.registerAction({type:"timelineChange",event:"timelineChanged",update:"prepareAndUpdate"},(function(t,e,n){var r=e.getComponent("timeline");return r&&null!=t.currentIndex&&(r.setCurrentIndex(t.currentIndex),!r.get("loop",!0)&&r.isIndexMax()&&r.getPlayState()&&(r.setPlayState(!1),n.dispatchAction({type:"timelinePlayChange",playState:!1,from:t.from}))),e.resetOption("timeline",{replaceMerge:r.get("replaceMerge",!0)}),Object(s["i"])({currentIndex:r.option.currentIndex},t)})),t.registerAction({type:"timelinePlayChange",event:"timelinePlayChanged",update:"update"},(function(t,e){var n=e.getComponent("timeline");n&&null!=t.playState&&n.setPlayState(t.playState)}))}function RX(t){var e=t&&t.timeline;s["t"](e)||(e=e?[e]:[]),s["k"](e,(function(t){t&&HX(t)}))}function HX(t){var e=t.type,n={number:"value",time:"time"};if(n[e]&&(t.axisType=n[e],delete t.type),zX(t),FX(t,"controlPosition")){var r=t.controlStyle||(t.controlStyle={});FX(r,"position")||(r.position=t.controlPosition),"none"!==r.position||FX(r,"show")||(r.show=!1,delete r.position),delete t.controlPosition}s["k"](t.data||[],(function(t){s["A"](t)&&!s["t"](t)&&(!FX(t,"value")&&FX(t,"name")&&(t.value=t.name),zX(t))}))}function zX(t){var e=t.itemStyle||(t.itemStyle={}),n=e.emphasis||(e.emphasis={}),r=t.label||t.label||{},i=r.normal||(r.normal={}),o={normal:1,emphasis:1};s["k"](r,(function(t,e){o[e]||FX(i,e)||(i[e]=t)})),n.label&&!FX(r,"emphasis")&&(r.emphasis=n.label,delete n.label)}function FX(t,e){return t.hasOwnProperty(e)}function BX(t){t.registerComponentModel(MX),t.registerComponentView(EX),t.registerSubTypeDefaulter("timeline",(function(){return"slider"})),NX(t),t.registerPreprocessor(RX)}function WX(t,e){if(!t)return!1;for(var n=Object(s["t"])(t)?t:[t],r=0;r<n.length;r++)if(n[r]&&n[r][e])return!0;return!1}function VX(t){ci(t,"label",["show"])}var GX=Ci(),$X=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.createdBySelf=!1,n}return o(e,t),e.prototype.init=function(t,e,n){this.mergeDefaultAndTheme(t,n),this._mergeOption(t,n,!1,!0)},e.prototype.isAnimationEnabled=function(){if(a["a"].node)return!1;var t=this.__hostSeries;return this.getShallow("animation")&&t&&t.isAnimationEnabled()},e.prototype.mergeOption=function(t,e){this._mergeOption(t,e,!1,!1)},e.prototype._mergeOption=function(t,e,n,r){var i=this.mainType;n||e.eachSeries((function(t){var n=t.get(this.mainType,!0),o=GX(t)[i];n&&n.data?(o?o._mergeOption(n,e,!0):(r&&VX(n),s["k"](n.data,(function(t){t instanceof Array?(VX(t[0]),VX(t[1])):VX(t)})),o=this.createMarkerModelFromSeries(n,this,e),s["m"](o,{mainType:this.mainType,seriesIndex:t.seriesIndex,name:t.name,createdBySelf:!0}),o.__hostSeries=t),GX(t)[i]=o):GX(t)[i]=null}),this)},e.prototype.formatTooltip=function(t,e,n){var r=this.getData(),i=this.getRawValue(t),o=r.getName(t);return Uf("section",{header:this.name,blocks:[Uf("nameValue",{name:o,value:i,noName:!o,noValue:null==i})]})},e.prototype.getData=function(){return this._data},e.prototype.setData=function(t){this._data=t},e.getMarkerModelFromSeries=function(t,e){return GX(t)[e]},e.type="marker",e.dependencies=["series","grid","polar","geo"],e}(Xc);s["K"]($X,ef.prototype);var UX=$X,qX=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.createMarkerModelFromSeries=function(t,n,r){return new e(t,n,r)},e.type="markPoint",e.defaultOption={z:5,symbol:"pin",symbolSize:50,tooltip:{trigger:"item"},label:{show:!0,position:"inside"},itemStyle:{borderWidth:2},emphasis:{label:{show:!0}}},e}(UX),XX=qX;function ZX(t){return!(isNaN(parseFloat(t.x))&&isNaN(parseFloat(t.y)))}function JX(t){return!isNaN(parseFloat(t.x))&&!isNaN(parseFloat(t.y))}function KX(t,e,n,r,i,o){var a=[],s=iS(e,r),l=s?e.getCalculationInfo("stackResultDimension"):r,u=aZ(e,l,t),c=e.indicesOfNearest(l,u)[0];a[i]=e.get(n,c),a[o]=e.get(l,c);var d=e.get(r,c),h=zr(e.get(r,c));return h=Math.min(h,20),h>=0&&(a[o]=+a[o].toFixed(h)),[a,d]}var QX={min:Object(s["h"])(KX,"min"),max:Object(s["h"])(KX,"max"),average:Object(s["h"])(KX,"average"),median:Object(s["h"])(KX,"median")};function tZ(t,e){if(e){var n=t.getData(),r=t.coordinateSystem,i=r&&r.dimensions;if(!JX(e)&&!Object(s["t"])(e.coord)&&Object(s["t"])(i)){var o=eZ(e,n,r,t);if(e=Object(s["d"])(e),e.type&&QX[e.type]&&o.baseAxis&&o.valueAxis){var a=Object(s["r"])(i,o.baseAxis.dim),l=Object(s["r"])(i,o.valueAxis.dim),u=QX[e.type](n,o.baseDataDim,o.valueDataDim,a,l);e.coord=u[0],e.value=u[1]}else e.coord=[null!=e.xAxis?e.xAxis:e.radiusAxis,null!=e.yAxis?e.yAxis:e.angleAxis]}if(null!=e.coord&&Object(s["t"])(i))for(var c=e.coord,d=0;d<2;d++)QX[c[d]]&&(c[d]=aZ(n,n.mapDimension(i[d]),c[d]));else e.coord=[];return e}}function eZ(t,e,n,r){var i={};return null!=t.valueIndex||null!=t.valueDim?(i.valueDataDim=null!=t.valueIndex?e.getDimension(t.valueIndex):t.valueDim,i.valueAxis=n.getAxis(nZ(r,i.valueDataDim)),i.baseAxis=n.getOtherAxis(i.valueAxis),i.baseDataDim=e.mapDimension(i.baseAxis.dim)):(i.baseAxis=r.getBaseAxis(),i.valueAxis=n.getOtherAxis(i.baseAxis),i.baseDataDim=e.mapDimension(i.baseAxis.dim),i.valueDataDim=e.mapDimension(i.valueAxis.dim)),i}function nZ(t,e){var n=t.getData().getDimensionInfo(e);return n&&n.coordDim}function rZ(t,e){return!(t&&t.containData&&e.coord&&!ZX(e))||t.containData(e.coord)}function iZ(t,e,n){return!(t&&t.containZone&&e.coord&&n.coord&&!ZX(e)&&!ZX(n))||t.containZone(e.coord,n.coord)}function oZ(t,e){return t?function(t,n,r,i){var o=i<2?t.coord&&t.coord[i]:t.value;return sf(o,e[i])}:function(t,n,r,i){return sf(t.value,e[i])}}function aZ(t,e,n){if("average"===n){var r=0,i=0;return t.each(e,(function(t,e){isNaN(t)||(r+=t,i++)})),r/i}return"median"===n?t.getMedian(e):t.getDataExtent(e)["max"===n?1:0]}var sZ=Ci(),lZ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(){this.markerGroupMap=Object(s["f"])()},e.prototype.render=function(t,e,n){var r=this,i=this.markerGroupMap;i.each((function(t){sZ(t).keep=!1})),e.eachSeries((function(t){var i=UX.getMarkerModelFromSeries(t,r.type);i&&r.renderSeries(t,i,e,n)})),i.each((function(t){!sZ(t).keep&&r.group.remove(t.group)}))},e.prototype.markKeep=function(t){sZ(t).keep=!0},e.prototype.toggleBlurSeries=function(t,e){var n=this;Object(s["k"])(t,(function(t){var r=UX.getMarkerModelFromSeries(t,n.type);if(r){var i=r.getData();i.eachItemGraphicEl((function(t){t&&(e?dl(t):hl(t))}))}}))},e.type="marker",e}(kp),uZ=lZ;function cZ(t,e,n){var r=e.coordinateSystem;t.each((function(i){var o,a=t.getItemModel(i),s=Nr(a.get("x"),n.getWidth()),l=Nr(a.get("y"),n.getHeight());if(isNaN(s)||isNaN(l)){if(e.getMarkerPosition)o=e.getMarkerPosition(t.getValues(t.dimensions,i));else if(r){var u=t.get(r.dimensions[0],i),c=t.get(r.dimensions[1],i);o=r.dataToPoint([u,c])}}else o=[s,l];isNaN(s)||(o[0]=s),isNaN(l)||(o[1]=l),t.setItemLayout(i,o)}))}var dZ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.updateTransform=function(t,e,n){e.eachSeries((function(t){var e=UX.getMarkerModelFromSeries(t,"markPoint");e&&(cZ(e.getData(),t,n),this.markerGroupMap.get(t.id).updateLayout())}),this)},e.prototype.renderSeries=function(t,e,n,r){var i=t.coordinateSystem,o=t.id,a=t.getData(),l=this.markerGroupMap,u=l.get(o)||l.set(o,new MS),c=hZ(i,t,e);e.setData(c),cZ(e.getData(),t,r),c.each((function(t){var n=c.getItemModel(t),r=n.getShallow("symbol"),i=n.getShallow("symbolSize"),o=n.getShallow("symbolRotate"),l=n.getShallow("symbolOffset"),u=n.getShallow("symbolKeepAspect");if(Object(s["w"])(r)||Object(s["w"])(i)||Object(s["w"])(o)||Object(s["w"])(l)){var d=e.getRawValue(t),h=e.getDataParams(t);Object(s["w"])(r)&&(r=r(d,h)),Object(s["w"])(i)&&(i=i(d,h)),Object(s["w"])(o)&&(o=o(d,h)),Object(s["w"])(l)&&(l=l(d,h))}var f=n.getModel("itemStyle").getItemStyle(),p=zy(a,"color");f.fill||(f.fill=p),c.setItemVisual(t,{symbol:r,symbolSize:i,symbolRotate:o,symbolOffset:l,symbolKeepAspect:u,style:f})})),u.updateData(c),this.group.add(u.group),c.eachItemGraphicEl((function(t){t.traverse((function(t){Ls(t).dataModel=e}))})),this.markKeep(u),u.group.silent=e.get("silent")||t.get("silent")},e.type="markPoint",e}(uZ);function hZ(t,e,n){var r;r=t?Object(s["H"])(t&&t.dimensions,(function(t){var n=e.getData().getDimensionInfo(e.getData().mapDimension(t))||{};return Object(s["m"])(Object(s["m"])({},n),{name:t,ordinalMeta:null})})):[{name:"value",type:"float"}];var i=new UM(r,n),o=Object(s["H"])(n.get("data"),Object(s["h"])(tZ,e));t&&(o=Object(s["n"])(o,Object(s["h"])(rZ,t)));var a=oZ(!!t,r);return i.initData(o,null,a),i}var fZ=dZ;function pZ(t){t.registerComponentModel(XX),t.registerComponentView(fZ),t.registerPreprocessor((function(t){WX(t.series,"markPoint")&&(t.markPoint=t.markPoint||{})}))}var mZ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.createMarkerModelFromSeries=function(t,n,r){return new e(t,n,r)},e.type="markLine",e.defaultOption={z:5,symbol:["circle","arrow"],symbolSize:[8,16],symbolOffset:0,precision:2,tooltip:{trigger:"item"},label:{show:!0,position:"end",distance:5},lineStyle:{type:"dashed"},emphasis:{label:{show:!0},lineStyle:{width:3}},animationEasing:"linear"},e}(UX),gZ=mZ,yZ=Ci(),vZ=function(t,e,n,r){var i,o=t.getData();if(Object(s["t"])(r))i=r;else{var a=r.type;if("min"===a||"max"===a||"average"===a||"median"===a||null!=r.xAxis||null!=r.yAxis){var l=void 0,u=void 0;if(null!=r.yAxis||null!=r.xAxis)l=e.getAxis(null!=r.yAxis?"y":"x"),u=Object(s["O"])(r.yAxis,r.xAxis);else{var c=eZ(r,o,e,t);l=c.valueAxis;var d=oS(o,c.valueDataDim);u=aZ(o,d,a)}var h="x"===l.dim?0:1,f=1-h,p=Object(s["d"])(r),m={coord:[]};p.type=null,p.coord=[],p.coord[f]=-1/0,m.coord[f]=1/0;var g=n.get("precision");g>=0&&Object(s["z"])(u)&&(u=+u.toFixed(Math.min(g,20))),p.coord[h]=m.coord[h]=u,i=[p,m,{type:a,valueIndex:r.valueIndex,value:u}]}else i=[]}var y=[tZ(t,i[0]),tZ(t,i[1]),Object(s["m"])({},i[2])];return y[2].type=y[2].type||null,Object(s["I"])(y[2],y[0]),Object(s["I"])(y[2],y[1]),y};function _Z(t){return!isNaN(t)&&!isFinite(t)}function bZ(t,e,n,r){var i=1-t,o=r.dimensions[t];return _Z(e[i])&&_Z(n[i])&&e[t]===n[t]&&r.getAxis(o).containData(e[t])}function xZ(t,e){if("cartesian2d"===t.type){var n=e[0].coord,r=e[1].coord;if(n&&r&&(bZ(1,n,r,t)||bZ(0,n,r,t)))return!0}return rZ(t,e[0])&&rZ(t,e[1])}function wZ(t,e,n,r,i){var o,a=r.coordinateSystem,s=t.getItemModel(e),l=Nr(s.get("x"),i.getWidth()),u=Nr(s.get("y"),i.getHeight());if(isNaN(l)||isNaN(u)){if(r.getMarkerPosition)o=r.getMarkerPosition(t.getValues(t.dimensions,e));else{var c=a.dimensions,d=t.get(c[0],e),h=t.get(c[1],e);o=a.dataToPoint([d,h])}if(WS(a,"cartesian2d")){var f=a.getAxis("x"),p=a.getAxis("y");c=a.dimensions;_Z(t.get(c[0],e))?o[0]=f.toGlobalCoord(f.getExtent()[n?0:1]):_Z(t.get(c[1],e))&&(o[1]=p.toGlobalCoord(p.getExtent()[n?0:1]))}isNaN(l)||(o[0]=l),isNaN(u)||(o[1]=u)}else o=[l,u];t.setItemLayout(e,o)}var MZ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.updateTransform=function(t,e,n){e.eachSeries((function(t){var e=UX.getMarkerModelFromSeries(t,"markLine");if(e){var r=e.getData(),i=yZ(e).from,o=yZ(e).to;i.each((function(e){wZ(i,e,!0,t,n),wZ(o,e,!1,t,n)})),r.each((function(t){r.setItemLayout(t,[i.getItemLayout(t),o.getItemLayout(t)])})),this.markerGroupMap.get(t.id).updateLayout()}}),this)},e.prototype.renderSeries=function(t,e,n,r){var i=t.coordinateSystem,o=t.id,a=t.getData(),l=this.markerGroupMap,u=l.get(o)||l.set(o,new RP);this.group.add(u.group);var c=SZ(i,t,e),d=c.from,h=c.to,f=c.line;yZ(e).from=d,yZ(e).to=h,e.setData(f);var p=e.get("symbol"),m=e.get("symbolSize"),g=e.get("symbolRotate"),y=e.get("symbolOffset");function v(e,n,i){var o=e.getItemModel(n);wZ(e,n,i,t,r);var l=o.getModel("itemStyle").getItemStyle();null==l.fill&&(l.fill=zy(a,"color")),e.setItemVisual(n,{symbolKeepAspect:o.get("symbolKeepAspect"),symbolOffset:Object(s["P"])(o.get("symbolOffset",!0),y[i?0:1]),symbolRotate:Object(s["P"])(o.get("symbolRotate",!0),g[i?0:1]),symbolSize:Object(s["P"])(o.get("symbolSize"),m[i?0:1]),symbol:Object(s["P"])(o.get("symbol",!0),p[i?0:1]),style:l})}Object(s["t"])(p)||(p=[p,p]),Object(s["t"])(m)||(m=[m,m]),Object(s["t"])(g)||(g=[g,g]),Object(s["t"])(y)||(y=[y,y]),c.from.each((function(t){v(d,t,!0),v(h,t,!1)})),f.each((function(t){var e=f.getItemModel(t).getModel("lineStyle").getLineStyle();f.setItemLayout(t,[d.getItemLayout(t),h.getItemLayout(t)]),null==e.stroke&&(e.stroke=d.getItemVisual(t,"style").fill),f.setItemVisual(t,{fromSymbolKeepAspect:d.getItemVisual(t,"symbolKeepAspect"),fromSymbolOffset:d.getItemVisual(t,"symbolOffset"),fromSymbolRotate:d.getItemVisual(t,"symbolRotate"),fromSymbolSize:d.getItemVisual(t,"symbolSize"),fromSymbol:d.getItemVisual(t,"symbol"),toSymbolKeepAspect:h.getItemVisual(t,"symbolKeepAspect"),toSymbolOffset:h.getItemVisual(t,"symbolOffset"),toSymbolRotate:h.getItemVisual(t,"symbolRotate"),toSymbolSize:h.getItemVisual(t,"symbolSize"),toSymbol:h.getItemVisual(t,"symbol"),style:e})})),u.updateData(f),c.line.eachItemGraphicEl((function(t){Ls(t).dataModel=e,t.traverse((function(t){Ls(t).dataModel=e}))})),this.markKeep(u),u.group.silent=e.get("silent")||t.get("silent")},e.type="markLine",e}(uZ);function SZ(t,e,n){var r;r=t?Object(s["H"])(t&&t.dimensions,(function(t){var n=e.getData().getDimensionInfo(e.getData().mapDimension(t))||{};return Object(s["m"])(Object(s["m"])({},n),{name:t,ordinalMeta:null})})):[{name:"value",type:"float"}];var i=new UM(r,n),o=new UM(r,n),a=new UM([],n),l=Object(s["H"])(n.get("data"),Object(s["h"])(vZ,e,t,n));t&&(l=Object(s["n"])(l,Object(s["h"])(xZ,t)));var u=oZ(!!t,r);return i.initData(Object(s["H"])(l,(function(t){return t[0]})),null,u),o.initData(Object(s["H"])(l,(function(t){return t[1]})),null,u),a.initData(Object(s["H"])(l,(function(t){return t[2]}))),a.hasItemOption=!0,{from:i,to:o,line:a}}var kZ=MZ;function LZ(t){t.registerComponentModel(gZ),t.registerComponentView(kZ),t.registerPreprocessor((function(t){WX(t.series,"markLine")&&(t.markLine=t.markLine||{})}))}var TZ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.createMarkerModelFromSeries=function(t,n,r){return new e(t,n,r)},e.type="markArea",e.defaultOption={z:1,tooltip:{trigger:"item"},animation:!1,label:{show:!0,position:"top"},itemStyle:{borderWidth:0},emphasis:{label:{show:!0,position:"top"}}},e}(UX),DZ=TZ,OZ=Ci(),IZ=function(t,e,n,r){var i=r[0],o=r[1];if(i&&o){var a=tZ(t,i),l=tZ(t,o),u=a.coord,c=l.coord;u[0]=Object(s["O"])(u[0],-1/0),u[1]=Object(s["O"])(u[1],-1/0),c[0]=Object(s["O"])(c[0],1/0),c[1]=Object(s["O"])(c[1],1/0);var d=Object(s["J"])([{},a,l]);return d.coord=[a.coord,l.coord],d.x0=a.x,d.y0=a.y,d.x1=l.x,d.y1=l.y,d}};function CZ(t){return!isNaN(t)&&!isFinite(t)}function AZ(t,e,n,r){var i=1-t;return CZ(e[i])&&CZ(n[i])}function YZ(t,e){var n=e.coord[0],r=e.coord[1],i={coord:n,x:e.x0,y:e.y0},o={coord:r,x:e.x1,y:e.y1};return WS(t,"cartesian2d")?!(!n||!r||!AZ(1,n,r,t)&&!AZ(0,n,r,t))||iZ(t,i,o):rZ(t,i)||rZ(t,o)}function jZ(t,e,n,r,i){var o,a=r.coordinateSystem,s=t.getItemModel(e),l=Nr(s.get(n[0]),i.getWidth()),u=Nr(s.get(n[1]),i.getHeight());if(isNaN(l)||isNaN(u)){if(r.getMarkerPosition){var c=t.getValues(["x0","y0"],e),d=t.getValues(["x1","y1"],e),h=a.clampData(c),f=a.clampData(d),p=[];"x0"===n[0]?p[0]=h[0]>f[0]?d[0]:c[0]:p[0]=h[0]>f[0]?c[0]:d[0],"y0"===n[1]?p[1]=h[1]>f[1]?d[1]:c[1]:p[1]=h[1]>f[1]?c[1]:d[1],o=r.getMarkerPosition(p,n,!0)}else{var m=t.get(n[0],e),g=t.get(n[1],e),y=[m,g];a.clampData&&a.clampData(y,y),o=a.dataToPoint(y,!0)}if(WS(a,"cartesian2d")){var v=a.getAxis("x"),_=a.getAxis("y");m=t.get(n[0],e),g=t.get(n[1],e);CZ(m)?o[0]=v.toGlobalCoord(v.getExtent()["x0"===n[0]?0:1]):CZ(g)&&(o[1]=_.toGlobalCoord(_.getExtent()["y0"===n[1]?0:1]))}isNaN(l)||(o[0]=l),isNaN(u)||(o[1]=u)}else o=[l,u];return o}var PZ=[["x0","y0"],["x1","y0"],["x1","y1"],["x0","y1"]],EZ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.updateTransform=function(t,e,n){e.eachSeries((function(t){var e=UX.getMarkerModelFromSeries(t,"markArea");if(e){var r=e.getData();r.each((function(e){var i=Object(s["H"])(PZ,(function(i){return jZ(r,e,i,t,n)}));r.setItemLayout(e,i);var o=r.getItemGraphicEl(e);o.setShape("points",i)}))}}),this)},e.prototype.renderSeries=function(t,e,n,r){var i=t.coordinateSystem,o=t.id,a=t.getData(),l=this.markerGroupMap,u=l.get(o)||l.set(o,{group:new wr});this.group.add(u.group),this.markKeep(u);var c=NZ(i,t,e);e.setData(c),c.each((function(e){var n=Object(s["H"])(PZ,(function(n){return jZ(c,e,n,t,r)})),o=i.getAxis("x").scale,l=i.getAxis("y").scale,u=o.getExtent(),d=l.getExtent(),h=[o.parse(c.get("x0",e)),o.parse(c.get("x1",e))],f=[l.parse(c.get("y0",e)),l.parse(c.get("y1",e))];Hr(h),Hr(f);var p=!(u[0]>h[1]||u[1]<h[0]||d[0]>f[1]||d[1]<f[0]),m=!p;c.setItemLayout(e,{points:n,allClipped:m});var g=c.getItemModel(e).getModel("itemStyle").getItemStyle(),y=zy(a,"color");g.fill||(g.fill=y,Object(s["C"])(g.fill)&&(g.fill=Ne["f"](g.fill,.4))),g.stroke||(g.stroke=y),c.setItemVisual(e,"style",g)})),c.diff(OZ(u).data).add((function(t){var e=c.getItemLayout(t);if(!e.allClipped){var n=new Tm({shape:{points:e.points}});c.setItemGraphicEl(t,n),u.group.add(n)}})).update((function(t,n){var r=OZ(u).data.getItemGraphicEl(n),i=c.getItemLayout(t);i.allClipped?r&&u.group.remove(r):(r?Wl(r,{shape:{points:i.points}},e,t):r=new Tm({shape:{points:i.points}}),c.setItemGraphicEl(t,r),u.group.add(r))})).remove((function(t){var e=OZ(u).data.getItemGraphicEl(t);u.group.remove(e)})).execute(),c.eachItemGraphicEl((function(t,n){var r=c.getItemModel(n),i=c.getItemVisual(n,"style");t.useStyle(c.getItemVisual(n,"style")),tu(t,eu(r),{labelFetcher:e,labelDataIndex:n,defaultText:c.getName(n)||"",inheritColor:Object(s["C"])(i.fill)?Ne["f"](i.fill,1):"#000"}),Al(t,r),Dl(t,null,null,r.get(["emphasis","disabled"])),Ls(t).dataModel=e})),OZ(u).data=c,u.group.silent=e.get("silent")||t.get("silent")},e.type="markArea",e}(uZ);function NZ(t,e,n){var r,i,o=["x0","y0","x1","y1"];if(t){var a=Object(s["H"])(t&&t.dimensions,(function(t){var n=e.getData(),r=n.getDimensionInfo(n.mapDimension(t))||{};return Object(s["m"])(Object(s["m"])({},r),{name:t,ordinalMeta:null})}));i=Object(s["H"])(o,(function(t,e){return{name:t,type:a[e%2].type}})),r=new UM(i,n)}else i=[{name:"value",type:"float"}],r=new UM(i,n);var l=Object(s["H"])(n.get("data"),Object(s["h"])(IZ,e,t,n));t&&(l=Object(s["n"])(l,Object(s["h"])(YZ,t)));var u=t?function(t,e,n,r){var o=t.coord[Math.floor(r/2)][r%2];return sf(o,i[r])}:function(t,e,n,r){return sf(t.value,i[r])};return r.initData(l,null,u),r.hasItemOption=!0,r}var RZ=EZ;function HZ(t){t.registerComponentModel(DZ),t.registerComponentView(RZ),t.registerPreprocessor((function(t){WX(t.series,"markArea")&&(t.markArea=t.markArea||{})}))}var zZ=function(t,e){return"all"===e?{type:"all",title:t.getLocaleModel().get(["legend","selector","all"])}:"inverse"===e?{type:"inverse",title:t.getLocaleModel().get(["legend","selector","inverse"])}:void 0},FZ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.layoutMode={type:"box",ignoreSize:!0},n}return o(e,t),e.prototype.init=function(t,e,n){this.mergeDefaultAndTheme(t,n),t.selected=t.selected||{},this._updateSelector(t)},e.prototype.mergeOption=function(e,n){t.prototype.mergeOption.call(this,e,n),this._updateSelector(e)},e.prototype._updateSelector=function(t){var e=t.selector,n=this.ecModel;!0===e&&(e=t.selector=["all","inverse"]),s["t"](e)&&s["k"](e,(function(t,r){s["C"](t)&&(t={type:t}),e[r]=s["I"](t,zZ(n,t.type))}))},e.prototype.optionUpdated=function(){this._updateData(this.ecModel);var t=this._data;if(t[0]&&"single"===this.get("selectedMode")){for(var e=!1,n=0;n<t.length;n++){var r=t[n].get("name");if(this.isSelected(r)){this.select(r),e=!0;break}}!e&&this.select(t[0].get("name"))}},e.prototype._updateData=function(t){var e=[],n=[];t.eachRawSeries((function(r){var i,o=r.name;if(n.push(o),r.legendVisualProvider){var a=r.legendVisualProvider,s=a.getAllNames();t.isSeriesFiltered(r)||(n=n.concat(s)),s.length?e=e.concat(s):i=!0}else i=!0;i&&Si(r)&&e.push(r.name)})),this._availableNames=n;var r=this.get("data")||e,i=s["f"](),o=s["H"](r,(function(t){return(s["C"](t)||s["z"](t))&&(t={name:t}),i.get(t.name)?null:(i.set(t.name,!0),new Lu(t,this,this.ecModel))}),this);this._data=s["n"](o,(function(t){return!!t}))},e.prototype.getData=function(){return this._data},e.prototype.select=function(t){var e=this.option.selected,n=this.get("selectedMode");if("single"===n){var r=this._data;s["k"](r,(function(t){e[t.get("name")]=!1}))}e[t]=!0},e.prototype.unSelect=function(t){"single"!==this.get("selectedMode")&&(this.option.selected[t]=!1)},e.prototype.toggleSelected=function(t){var e=this.option.selected;e.hasOwnProperty(t)||(e[t]=!0),this[e[t]?"unSelect":"select"](t)},e.prototype.allSelect=function(){var t=this._data,e=this.option.selected;s["k"](t,(function(t){e[t.get("name",!0)]=!0}))},e.prototype.inverseSelect=function(){var t=this._data,e=this.option.selected;s["k"](t,(function(t){var n=t.get("name",!0);e.hasOwnProperty(n)||(e[n]=!0),e[n]=!e[n]}))},e.prototype.isSelected=function(t){var e=this.option.selected;return!(e.hasOwnProperty(t)&&!e[t])&&s["r"](this._availableNames,t)>=0},e.prototype.getOrient=function(){return"vertical"===this.get("orient")?{index:1,name:"vertical"}:{index:0,name:"horizontal"}},e.type="legend.plain",e.dependencies=["series"],e.defaultOption={z:4,show:!0,orient:"horizontal",left:"center",top:0,align:"auto",backgroundColor:"rgba(0,0,0,0)",borderColor:"#ccc",borderRadius:0,borderWidth:0,padding:5,itemGap:10,itemWidth:25,itemHeight:14,symbolRotate:"inherit",symbolKeepAspect:!0,inactiveColor:"#ccc",inactiveBorderColor:"#ccc",inactiveBorderWidth:"auto",itemStyle:{color:"inherit",opacity:"inherit",borderColor:"inherit",borderWidth:"auto",borderCap:"inherit",borderJoin:"inherit",borderDashOffset:"inherit",borderMiterLimit:"inherit"},lineStyle:{width:"auto",color:"inherit",inactiveColor:"#ccc",inactiveWidth:2,opacity:"inherit",type:"inherit",cap:"inherit",join:"inherit",dashOffset:"inherit",miterLimit:"inherit"},textStyle:{color:"#333"},selectedMode:!0,selector:!1,selectorLabel:{show:!0,borderRadius:10,padding:[3,5,3,5],fontSize:12,fontFamily:"sans-serif",color:"#666",borderWidth:1,borderColor:"#666"},emphasis:{selectorLabel:{show:!0,color:"#eee",backgroundColor:"#666"}},selectorPosition:"auto",selectorItemGap:7,selectorButtonGap:10,tooltip:{show:!1}},e}(Xc),BZ=FZ,WZ=s["h"],VZ=s["k"],GZ=wr,$Z=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.newlineDisabled=!1,n}return o(e,t),e.prototype.init=function(){this.group.add(this._contentGroup=new GZ),this.group.add(this._selectorGroup=new GZ),this._isFirstRender=!0},e.prototype.getContentGroup=function(){return this._contentGroup},e.prototype.getSelectorGroup=function(){return this._selectorGroup},e.prototype.render=function(t,e,n){var r=this._isFirstRender;if(this._isFirstRender=!1,this.resetInner(),t.get("show",!0)){var i=t.get("align"),o=t.get("orient");i&&"auto"!==i||(i="right"===t.get("left")&&"vertical"===o?"right":"left");var a=t.get("selector",!0),l=t.get("selectorPosition",!0);!a||l&&"auto"!==l||(l="horizontal"===o?"end":"start"),this.renderInner(i,t,e,n,a,o,l);var u=t.getBoxLayoutParams(),c={width:n.getWidth(),height:n.getHeight()},d=t.get("padding"),h=Hc(u,c,d),f=this.layoutInner(t,i,h,r,a,l),p=Hc(s["i"]({width:f.width,height:f.height},u),c,d);this.group.x=p.x-f.x,this.group.y=p.y-f.y,this.group.markRedraw(),this.group.add(this._backgroundEl=J$(f,t))}},e.prototype.resetInner=function(){this.getContentGroup().removeAll(),this._backgroundEl&&this.group.remove(this._backgroundEl),this.getSelectorGroup().removeAll()},e.prototype.renderInner=function(t,e,n,r,i,o,a){var l=this.getContentGroup(),u=s["f"](),c=e.get("selectedMode"),d=[];n.eachRawSeries((function(t){!t.get("legendHoverLink")&&d.push(t.id)})),VZ(e.getData(),(function(i,o){var a=i.get("name");if(!this.newlineDisabled&&(""===a||"\n"===a)){var h=new GZ;return h.newline=!0,void l.add(h)}var f=n.getSeriesByName(a)[0];if(!u.get(a)){if(f){var p=f.getData(),m=p.getVisual("legendLineStyle")||{},g=p.getVisual("legendIcon"),y=p.getVisual("style"),v=this._createItem(f,a,o,i,e,t,m,y,g,c,r);v.on("click",WZ(XZ,a,null,r,d)).on("mouseover",WZ(JZ,f.name,null,r,d)).on("mouseout",WZ(KZ,f.name,null,r,d)),n.ssr&&v.eachChild((function(t){var e=Ls(t);e.seriesIndex=f.seriesIndex,e.dataIndex=o,e.ssrType="legend"})),u.set(a,!0)}else n.eachRawSeries((function(l){if(!u.get(a)&&l.legendVisualProvider){var h=l.legendVisualProvider;if(!h.containName(a))return;var f=h.indexOfName(a),p=h.getItemVisual(f,"style"),m=h.getItemVisual(f,"legendIcon"),g=Object(Ne["h"])(p.fill);g&&0===g[3]&&(g[3]=.2,p=s["m"](s["m"]({},p),{fill:Object(Ne["i"])(g,"rgba")}));var y=this._createItem(l,a,o,i,e,t,{},p,m,c,r);y.on("click",WZ(XZ,null,a,r,d)).on("mouseover",WZ(JZ,null,a,r,d)).on("mouseout",WZ(KZ,null,a,r,d)),n.ssr&&y.eachChild((function(t){var e=Ls(t);e.seriesIndex=l.seriesIndex,e.dataIndex=o,e.ssrType="legend"})),u.set(a,!0)}}),this);0}}),this),i&&this._createSelector(i,e,r,o,a)},e.prototype._createSelector=function(t,e,n,r,i){var o=this.getSelectorGroup();VZ(t,(function(t){var r=t.type,i=new ks({style:{x:0,y:0,align:"center",verticalAlign:"middle"},onclick:function(){n.dispatchAction({type:"all"===r?"legendAllSelect":"legendInverseSelect"})}});o.add(i);var a=e.getModel("selectorLabel"),s=e.getModel(["emphasis","selectorLabel"]);tu(i,{normal:a,emphasis:s},{defaultText:t.title}),Ll(i)}))},e.prototype._createItem=function(t,e,n,r,i,o,a,l,u,c,d){var h=t.visualDrawType,f=i.get("itemWidth"),p=i.get("itemHeight"),m=i.isSelected(e),g=r.get("symbolRotate"),y=r.get("symbolKeepAspect"),v=r.get("icon");u=v||u||"roundRect";var _=UZ(u,r,a,l,h,m,d),b=new GZ,x=r.getModel("textStyle");if(!s["w"](t.getLegendIcon)||v&&"inherit"!==v){var w="inherit"===v&&t.getData().getVisual("symbol")?"inherit"===g?t.getData().getVisual("symbolRotate"):g:0;b.add(qZ({itemWidth:f,itemHeight:p,icon:u,iconRotate:w,itemStyle:_.itemStyle,lineStyle:_.lineStyle,symbolKeepAspect:y}))}else b.add(t.getLegendIcon({itemWidth:f,itemHeight:p,icon:u,iconRotate:g,itemStyle:_.itemStyle,lineStyle:_.lineStyle,symbolKeepAspect:y}));var M="left"===o?f+5:-5,S=o,k=i.get("formatter"),L=e;s["C"](k)&&k?L=k.replace("{name}",null!=e?e:""):s["w"](k)&&(L=k(e));var T=m?x.getTextColor():r.get("inactiveColor");b.add(new ks({style:nu(x,{text:L,x:M,y:p/2,fill:T,align:S,verticalAlign:"middle"},{inheritColor:T})}));var D=new ss({shape:b.getBoundingRect(),style:{fill:"transparent"}}),O=r.getModel("tooltip");return O.get("show")&&Pg({el:D,componentModel:i,itemName:e,itemTooltipOption:O.option}),b.add(D),b.eachChild((function(t){t.silent=!0})),D.silent=!c,this.getContentGroup().add(b),Ll(b),b.__legendDataIndex=n,b},e.prototype.layoutInner=function(t,e,n,r,i,o){var a=this.getContentGroup(),s=this.getSelectorGroup();Nc(t.get("orient"),a,t.get("itemGap"),n.width,n.height);var l=a.getBoundingRect(),u=[-l.x,-l.y];if(s.markRedraw(),a.markRedraw(),i){Nc("horizontal",s,t.get("selectorItemGap",!0));var c=s.getBoundingRect(),d=[-c.x,-c.y],h=t.get("selectorButtonGap",!0),f=t.getOrient().index,p=0===f?"width":"height",m=0===f?"height":"width",g=0===f?"y":"x";"end"===o?d[f]+=l[p]+h:u[f]+=c[p]+h,d[1-f]+=l[m]/2-c[m]/2,s.x=d[0],s.y=d[1],a.x=u[0],a.y=u[1];var y={x:0,y:0};return y[p]=l[p]+h+c[p],y[m]=Math.max(l[m],c[m]),y[g]=Math.min(0,c[g]+d[1-f]),y}return a.x=u[0],a.y=u[1],this.group.getBoundingRect()},e.prototype.remove=function(){this.getContentGroup().removeAll(),this._isFirstRender=!0},e.type="legend.plain",e}(kp);function UZ(t,e,n,r,i,o,a){function s(t,e){"auto"===t.lineWidth&&(t.lineWidth=e.lineWidth>0?2:0),VZ(t,(function(n,r){"inherit"===t[r]&&(t[r]=e[r])}))}var l=e.getModel("itemStyle"),u=l.getItemStyle(),c=0===t.lastIndexOf("empty",0)?"fill":"stroke",d=l.getShallow("decal");u.decal=d&&"inherit"!==d?qv(d,a):r.decal,"inherit"===u.fill&&(u.fill=r[i]),"inherit"===u.stroke&&(u.stroke=r[c]),"inherit"===u.opacity&&(u.opacity=("fill"===i?r:n).opacity),s(u,r);var h=e.getModel("lineStyle"),f=h.getLineStyle();if(s(f,n),"auto"===u.fill&&(u.fill=r.fill),"auto"===u.stroke&&(u.stroke=r.fill),"auto"===f.stroke&&(f.stroke=r.fill),!o){var p=e.get("inactiveBorderWidth"),m=u[c];u.lineWidth="auto"===p?r.lineWidth>0&&m?2:0:u.lineWidth,u.fill=e.get("inactiveColor"),u.stroke=e.get("inactiveBorderColor"),f.stroke=h.get("inactiveColor"),f.lineWidth=h.get("inactiveWidth")}return{itemStyle:u,lineStyle:f}}function qZ(t){var e=t.icon||"roundRect",n=ov(e,0,0,t.itemWidth,t.itemHeight,t.itemStyle.fill,t.symbolKeepAspect);return n.setStyle(t.itemStyle),n.rotation=(t.iconRotate||0)*Math.PI/180,n.setOrigin([t.itemWidth/2,t.itemHeight/2]),e.indexOf("empty")>-1&&(n.style.stroke=n.style.fill,n.style.fill="#fff",n.style.lineWidth=2),n}function XZ(t,e,n,r){KZ(t,e,n,r),n.dispatchAction({type:"legendToggleSelect",name:null!=t?t:e}),JZ(t,e,n,r)}function ZZ(t){var e,n=t.getZr().storage.getDisplayList(),r=0,i=n.length;while(r<i&&!(e=n[r].states.emphasis))r++;return e&&e.hoverLayer}function JZ(t,e,n,r){ZZ(n)||n.dispatchAction({type:"highlight",seriesName:t,name:e,excludeSeriesId:r})}function KZ(t,e,n,r){ZZ(n)||n.dispatchAction({type:"downplay",seriesName:t,name:e,excludeSeriesId:r})}var QZ=$Z;function tJ(t){var e=t.findComponents({mainType:"legend"});e&&e.length&&t.filterSeries((function(t){for(var n=0;n<e.length;n++)if(!e[n].isSelected(t.name))return!1;return!0}))}function eJ(t,e,n){var r,i={},o="toggleSelected"===t;return n.eachComponent("legend",(function(n){o&&null!=r?n[r?"select":"unSelect"](e.name):"allSelect"===t||"inverseSelect"===t?n[t]():(n[t](e.name),r=n.isSelected(e.name));var a=n.getData();Object(s["k"])(a,(function(t){var e=t.get("name");if("\n"!==e&&""!==e){var r=n.isSelected(e);i.hasOwnProperty(e)?i[e]=i[e]&&r:i[e]=r}}))})),"allSelect"===t||"inverseSelect"===t?{selected:i}:{name:e.name,selected:i}}function nJ(t){t.registerAction("legendToggleSelect","legendselectchanged",Object(s["h"])(eJ,"toggleSelected")),t.registerAction("legendAllSelect","legendselectall",Object(s["h"])(eJ,"allSelect")),t.registerAction("legendInverseSelect","legendinverseselect",Object(s["h"])(eJ,"inverseSelect")),t.registerAction("legendSelect","legendselected",Object(s["h"])(eJ,"select")),t.registerAction("legendUnSelect","legendunselected",Object(s["h"])(eJ,"unSelect"))}function rJ(t){t.registerComponentModel(BZ),t.registerComponentView(QZ),t.registerProcessor(t.PRIORITY.PROCESSOR.SERIES_FILTER,tJ),t.registerSubTypeDefaulter("legend",(function(){return"plain"})),nJ(t)}var iJ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.setScrollDataIndex=function(t){this.option.scrollDataIndex=t},e.prototype.init=function(e,n,r){var i=Vc(e);t.prototype.init.call(this,e,n,r),oJ(this,e,i)},e.prototype.mergeOption=function(e,n){t.prototype.mergeOption.call(this,e,n),oJ(this,this.option,e)},e.type="legend.scroll",e.defaultOption=Cu(BZ.defaultOption,{scrollDataIndex:0,pageButtonItemGap:5,pageButtonGap:null,pageButtonPosition:"end",pageFormatter:"{current}/{total}",pageIcons:{horizontal:["M0,0L12,-10L12,10z","M0,0L-12,-10L-12,10z"],vertical:["M0,0L20,0L10,-20z","M0,0L20,0L10,20z"]},pageIconColor:"#2f4554",pageIconInactiveColor:"#aaa",pageIconSize:15,pageTextStyle:{color:"#333"},animationDurationUpdate:800}),e}(BZ);function oJ(t,e,n){var r=t.getOrient(),i=[1,1];i[r.index]=0,Wc(e,n,{type:"box",ignoreSize:!!i})}var aJ=iJ,sJ=wr,lJ=["width","height"],uJ=["x","y"],cJ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.newlineDisabled=!0,n._currentIndex=0,n}return o(e,t),e.prototype.init=function(){t.prototype.init.call(this),this.group.add(this._containerGroup=new sJ),this._containerGroup.add(this.getContentGroup()),this.group.add(this._controllerGroup=new sJ)},e.prototype.resetInner=function(){t.prototype.resetInner.call(this),this._controllerGroup.removeAll(),this._containerGroup.removeClipPath(),this._containerGroup.__rectSize=null},e.prototype.renderInner=function(e,n,r,i,o,a,l){var u=this;t.prototype.renderInner.call(this,e,n,r,i,o,a,l);var c=this._controllerGroup,d=n.get("pageIconSize",!0),h=s["t"](d)?d:[d,d];p("pagePrev",0);var f=n.getModel("pageTextStyle");function p(t,e){var r=t+"DataIndex",o=Ig(n.get("pageIcons",!0)[n.getOrient().name][e],{onclick:s["c"](u._pageGo,u,r,n,i)},{x:-h[0]/2,y:-h[1]/2,width:h[0],height:h[1]});o.name=t,c.add(o)}c.add(new ks({name:"pageText",style:{text:"xx/xx",fill:f.getTextColor(),font:f.getFont(),verticalAlign:"middle",align:"center"},silent:!0})),p("pageNext",1)},e.prototype.layoutInner=function(t,e,n,r,i,o){var a=this.getSelectorGroup(),l=t.getOrient().index,u=lJ[l],c=uJ[l],d=lJ[1-l],h=uJ[1-l];i&&Nc("horizontal",a,t.get("selectorItemGap",!0));var f=t.get("selectorButtonGap",!0),p=a.getBoundingRect(),m=[-p.x,-p.y],g=s["d"](n);i&&(g[u]=n[u]-p[u]-f);var y=this._layoutContentAndController(t,r,g,l,u,d,h,c);if(i){if("end"===o)m[l]+=y[u]+f;else{var v=p[u]+f;m[l]-=v,y[c]-=v}y[u]+=p[u]+f,m[1-l]+=y[h]+y[d]/2-p[d]/2,y[d]=Math.max(y[d],p[d]),y[h]=Math.min(y[h],p[h]+m[1-l]),a.x=m[0],a.y=m[1],a.markRedraw()}return y},e.prototype._layoutContentAndController=function(t,e,n,r,i,o,a,l){var u=this.getContentGroup(),c=this._containerGroup,d=this._controllerGroup;Nc(t.get("orient"),u,t.get("itemGap"),r?n.width:null,r?null:n.height),Nc("horizontal",d,t.get("pageButtonItemGap",!0));var h=u.getBoundingRect(),f=d.getBoundingRect(),p=this._showController=h[i]>n[i],m=[-h.x,-h.y];e||(m[r]=u[l]);var g=[0,0],y=[-f.x,-f.y],v=s["P"](t.get("pageButtonGap",!0),t.get("itemGap",!0));if(p){var _=t.get("pageButtonPosition",!0);"end"===_?y[r]+=n[i]-f[i]:g[r]+=f[i]+v}y[1-r]+=h[o]/2-f[o]/2,u.setPosition(m),c.setPosition(g),d.setPosition(y);var b={x:0,y:0};if(b[i]=p?n[i]:h[i],b[o]=Math.max(h[o],f[o]),b[a]=Math.min(0,f[a]+y[1-r]),c.__rectSize=n[i],p){var x={x:0,y:0};x[i]=Math.max(n[i]-f[i]-v,0),x[o]=b[o],c.setClipPath(new ss({shape:x})),c.__rectSize=x[i]}else d.eachChild((function(t){t.attr({invisible:!0,silent:!0})}));var w=this._getPageInfo(t);return null!=w.pageIndex&&Wl(u,{x:w.contentPosition[0],y:w.contentPosition[1]},p?t:null),this._updatePageInfoView(t,w),b},e.prototype._pageGo=function(t,e,n){var r=this._getPageInfo(e)[t];null!=r&&n.dispatchAction({type:"legendScroll",scrollDataIndex:r,legendId:e.id})},e.prototype._updatePageInfoView=function(t,e){var n=this._controllerGroup;s["k"](["pagePrev","pageNext"],(function(r){var i=r+"DataIndex",o=null!=e[i],a=n.childOfName(r);a&&(a.setStyle("fill",o?t.get("pageIconColor",!0):t.get("pageIconInactiveColor",!0)),a.cursor=o?"pointer":"default")}));var r=n.childOfName("pageText"),i=t.get("pageFormatter"),o=e.pageIndex,a=null!=o?o+1:0,l=e.pageCount;r&&i&&r.setStyle("text",s["C"](i)?i.replace("{current}",null==a?"":a+"").replace("{total}",null==l?"":l+""):i({current:a,total:l}))},e.prototype._getPageInfo=function(t){var e=t.get("scrollDataIndex",!0),n=this.getContentGroup(),r=this._containerGroup.__rectSize,i=t.getOrient().index,o=lJ[i],a=uJ[i],s=this._findTargetItemIndex(e),l=n.children(),u=l[s],c=l.length,d=c?1:0,h={contentPosition:[n.x,n.y],pageCount:d,pageIndex:d-1,pagePrevDataIndex:null,pageNextDataIndex:null};if(!u)return h;var f=v(u);h.contentPosition[i]=-f.s;for(var p=s+1,m=f,g=f,y=null;p<=c;++p)y=v(l[p]),(!y&&g.e>m.s+r||y&&!_(y,m.s))&&(m=g.i>m.i?g:y,m&&(null==h.pageNextDataIndex&&(h.pageNextDataIndex=m.i),++h.pageCount)),g=y;for(p=s-1,m=f,g=f,y=null;p>=-1;--p)y=v(l[p]),y&&_(g,y.s)||!(m.i<g.i)||(g=m,null==h.pagePrevDataIndex&&(h.pagePrevDataIndex=m.i),++h.pageCount,++h.pageIndex),m=y;return h;function v(t){if(t){var e=t.getBoundingRect(),n=e[a]+t[a];return{s:n,e:n+e[o],i:t.__legendDataIndex}}}function _(t,e){return t.e>=e&&t.s<=e+r}},e.prototype._findTargetItemIndex=function(t){if(!this._showController)return 0;var e,n,r=this.getContentGroup();return r.eachChild((function(r,i){var o=r.__legendDataIndex;null==n&&null!=o&&(n=i),o===t&&(e=i)})),null!=e?e:n},e.type="legend.scroll",e}(QZ),dJ=cJ;function hJ(t){t.registerAction("legendScroll","legendscroll",(function(t,e){var n=t.scrollDataIndex;null!=n&&e.eachComponent({mainType:"legend",subType:"scroll",query:t},(function(t){t.setScrollDataIndex(n)}))}))}function fJ(t){jb(rJ),t.registerComponentModel(aJ),t.registerComponentView(dJ),hJ(t)}function pJ(t){jb(rJ),jb(fJ)}var mJ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="dataZoom.inside",e.defaultOption=Cu(L$.defaultOption,{disabled:!1,zoomLock:!1,zoomOnMouseWheel:!0,moveOnMouseMove:!0,moveOnMouseWheel:!1,preventDefaultMouseMove:!0}),e}(L$),gJ=mJ,yJ=Ci();function vJ(t,e,n){yJ(t).coordSysRecordMap.each((function(t){var r=t.dataZoomInfoMap.get(e.uid);r&&(r.getRange=n)}))}function _J(t,e){for(var n=yJ(t).coordSysRecordMap,r=n.keys(),i=0;i<r.length;i++){var o=r[i],a=n.get(o),s=a.dataZoomInfoMap;if(s){var l=e.uid,u=s.get(l);u&&(s.removeKey(l),s.keys().length||bJ(n,a))}}}function bJ(t,e){if(e){t.removeKey(e.model.uid);var n=e.controller;n&&n.dispose()}}function xJ(t,e){var n={model:e,containsPoint:Object(s["h"])(MJ,e),dispatchAction:Object(s["h"])(wJ,t),dataZoomInfoMap:null,controller:null},r=n.controller=new dI(t.getZr());return Object(s["k"])(["pan","zoom","scrollMove"],(function(t){r.on(t,(function(e){var r=[];n.dataZoomInfoMap.each((function(i){if(e.isAvailableBehavior(i.model.option)){var o=(i.getRange||{})[t],a=o&&o(i.dzReferCoordSysInfo,n.model.mainType,n.controller,e);!i.model.get("disabled",!0)&&a&&r.push({dataZoomId:i.model.id,start:a[0],end:a[1]})}})),r.length&&n.dispatchAction(r)}))})),n}function wJ(t,e){t.isDisposed()||t.dispatchAction({type:"dataZoom",animation:{easing:"cubicOut",duration:100},batch:e})}function MJ(t,e,n,r){return t.coordinateSystem.containPoint([n,r])}function SJ(t){var e,n="type_",r={type_true:2,type_move:1,type_false:0,type_undefined:-1},i=!0;return t.each((function(t){var o=t.model,a=!o.get("disabled",!0)&&(!o.get("zoomLock",!0)||"move");r[n+a]>r[n+e]&&(e=a),i=i&&o.get("preventDefaultMouseMove",!0)})),{controlType:e,opt:{zoomOnMouseWheel:!0,moveOnMouseMove:!0,moveOnMouseWheel:!0,preventDefaultMouseMove:!!i}}}function kJ(t){t.registerProcessor(t.PRIORITY.PROCESSOR.FILTER,(function(t,e){var n=yJ(e),r=n.coordSysRecordMap||(n.coordSysRecordMap=Object(s["f"])());r.each((function(t){t.dataZoomInfoMap=null})),t.eachComponent({mainType:"dataZoom",subType:"inside"},(function(t){var n=w$(t);Object(s["k"])(n.infoList,(function(n){var i=n.model.uid,o=r.get(i)||r.set(i,xJ(e,n.model)),a=o.dataZoomInfoMap||(o.dataZoomInfoMap=Object(s["f"])());a.set(t.uid,{dzReferCoordSysInfo:n,model:t,getRange:null})}))})),r.each((function(t){var e,n=t.controller,i=t.dataZoomInfoMap;if(i){var o=i.keys()[0];null!=o&&(e=i.get(o))}if(e){var a=SJ(i);n.enable(a.controlType,a.opt),n.setPointerChecker(t.containsPoint),Jg(t,"dispatchAction",e.model.get("throttle",!0),"fixRate")}else bJ(r,t)}))}))}var LJ=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.type="dataZoom.inside",e}return o(e,t),e.prototype.render=function(e,n,r){t.prototype.render.apply(this,arguments),e.noTarget()?this._clear():(this.range=e.getPercentRange(),vJ(r,e,{pan:Object(s["c"])(TJ.pan,this),zoom:Object(s["c"])(TJ.zoom,this),scrollMove:Object(s["c"])(TJ.scrollMove,this)}))},e.prototype.dispose=function(){this._clear(),t.prototype.dispose.apply(this,arguments)},e.prototype._clear=function(){_J(this.api,this.dataZoomModel),this.range=null},e.type="dataZoom.inside",e}(I$),TJ={zoom:function(t,e,n,r){var i=this.range,o=i.slice(),a=t.axisModels[0];if(a){var s=OJ[e](null,[r.originX,r.originY],a,n,t),l=(s.signal>0?s.pixelStart+s.pixelLength-s.pixel:s.pixel-s.pixelStart)/s.pixelLength*(o[1]-o[0])+o[0],u=Math.max(1/r.scale,0);o[0]=(o[0]-l)*u+l,o[1]=(o[1]-l)*u+l;var c=this.dataZoomModel.findRepresentativeAxisProxy().getMinMaxSpan();return nN(0,o,[0,100],0,c.minSpan,c.maxSpan),this.range=o,i[0]!==o[0]||i[1]!==o[1]?o:void 0}},pan:DJ((function(t,e,n,r,i,o){var a=OJ[r]([o.oldX,o.oldY],[o.newX,o.newY],e,i,n);return a.signal*(t[1]-t[0])*a.pixel/a.pixelLength})),scrollMove:DJ((function(t,e,n,r,i,o){var a=OJ[r]([0,0],[o.scrollDelta,o.scrollDelta],e,i,n);return a.signal*(t[1]-t[0])*o.scrollDelta}))};function DJ(t){return function(e,n,r,i){var o=this.range,a=o.slice(),s=e.axisModels[0];if(s){var l=t(a,s,e,n,r,i);return nN(l,a,[0,100],"all"),this.range=a,o[0]!==a[0]||o[1]!==a[1]?a:void 0}}}var OJ={grid:function(t,e,n,r,i){var o=n.axis,a={},s=i.model.coordinateSystem.getRect();return t=t||[0,0],"x"===o.dim?(a.pixel=e[0]-t[0],a.pixelLength=s.width,a.pixelStart=s.x,a.signal=o.inverse?1:-1):(a.pixel=e[1]-t[1],a.pixelLength=s.height,a.pixelStart=s.y,a.signal=o.inverse?-1:1),a},polar:function(t,e,n,r,i){var o=n.axis,a={},s=i.model.coordinateSystem,l=s.getRadiusAxis().getExtent(),u=s.getAngleAxis().getExtent();return t=t?s.pointToCoord(t):[0,0],e=s.pointToCoord(e),"radiusAxis"===n.mainType?(a.pixel=e[0]-t[0],a.pixelLength=l[1]-l[0],a.pixelStart=l[0],a.signal=o.inverse?1:-1):(a.pixel=e[1]-t[1],a.pixelLength=u[1]-u[0],a.pixelStart=u[0],a.signal=o.inverse?-1:1),a},singleAxis:function(t,e,n,r,i){var o=n.axis,a=i.model.coordinateSystem.getRect(),s={};return t=t||[0,0],"horizontal"===o.orient?(s.pixel=e[0]-t[0],s.pixelLength=a.width,s.pixelStart=a.x,s.signal=o.inverse?1:-1):(s.pixel=e[1]-t[1],s.pixelLength=a.height,s.pixelStart=a.y,s.signal=o.inverse?-1:1),s}},IJ=LJ;function CJ(t){B$(t),t.registerComponentModel(gJ),t.registerComponentView(IJ),kJ(t)}var AJ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="dataZoom.slider",e.layoutMode="box",e.defaultOption=Cu(L$.defaultOption,{show:!0,right:"ph",top:"ph",width:"ph",height:"ph",left:null,bottom:null,borderColor:"#d2dbee",borderRadius:3,backgroundColor:"rgba(47,69,84,0)",dataBackground:{lineStyle:{color:"#d2dbee",width:.5},areaStyle:{color:"#d2dbee",opacity:.2}},selectedDataBackground:{lineStyle:{color:"#8fb0f7",width:.5},areaStyle:{color:"#8fb0f7",opacity:.2}},fillerColor:"rgba(135,175,274,0.2)",handleIcon:"path://M-9.35,34.56V42m0-40V9.5m-2,0h4a2,2,0,0,1,2,2v21a2,2,0,0,1-2,2h-4a2,2,0,0,1-2-2v-21A2,2,0,0,1-11.35,9.5Z",handleSize:"100%",handleStyle:{color:"#fff",borderColor:"#ACB8D1"},moveHandleSize:7,moveHandleIcon:"path://M-320.9-50L-320.9-50c18.1,0,27.1,9,27.1,27.1V85.7c0,18.1-9,27.1-27.1,27.1l0,0c-18.1,0-27.1-9-27.1-27.1V-22.9C-348-41-339-50-320.9-50z M-212.3-50L-212.3-50c18.1,0,27.1,9,27.1,27.1V85.7c0,18.1-9,27.1-27.1,27.1l0,0c-18.1,0-27.1-9-27.1-27.1V-22.9C-239.4-41-230.4-50-212.3-50z M-103.7-50L-103.7-50c18.1,0,27.1,9,27.1,27.1V85.7c0,18.1-9,27.1-27.1,27.1l0,0c-18.1,0-27.1-9-27.1-27.1V-22.9C-130.9-41-121.8-50-103.7-50z",moveHandleStyle:{color:"#D2DBEE",opacity:.7},showDetail:!0,showDataShadow:"auto",realtime:!0,zoomLock:!1,textStyle:{color:"#6E7079"},brushSelect:!0,brushStyle:{color:"rgba(135,175,274,0.15)"},emphasis:{handleStyle:{borderColor:"#8FB0F7"},moveHandleStyle:{color:"#8FB0F7"}}}),e}(L$),YJ=AJ,jJ=ss,PJ=7,EJ=1,NJ=30,RJ=7,HJ="horizontal",zJ="vertical",FJ=5,BJ=["line","bar","candlestick","scatter"],WJ={easing:"cubicOut",duration:100,delay:0},VJ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n._displayables={},n}return o(e,t),e.prototype.init=function(t,e){this.api=e,this._onBrush=Object(s["c"])(this._onBrush,this),this._onBrushEnd=Object(s["c"])(this._onBrushEnd,this)},e.prototype.render=function(e,n,r,i){if(t.prototype.render.apply(this,arguments),Jg(this,"_dispatchZoomAction",e.get("throttle"),"fixRate"),this._orient=e.getOrient(),!1!==e.get("show")){if(e.noTarget())return this._clear(),void this.group.removeAll();i&&"dataZoom"===i.type&&i.from===this.uid||this._buildView(),this._updateView()}else this.group.removeAll()},e.prototype.dispose=function(){this._clear(),t.prototype.dispose.apply(this,arguments)},e.prototype._clear=function(){Kg(this,"_dispatchZoomAction");var t=this.api.getZr();t.off("mousemove",this._onBrush),t.off("mouseup",this._onBrushEnd)},e.prototype._buildView=function(){var t=this.group;t.removeAll(),this._brushing=!1,this._displayables.brushRect=null,this._resetLocation(),this._resetInterval();var e=this._displayables.sliderGroup=new wr;this._renderBackground(),this._renderHandle(),this._renderDataShadow(),t.add(e),this._positionGroup()},e.prototype._resetLocation=function(){var t=this.dataZoomModel,e=this.api,n=t.get("brushSelect"),r=n?RJ:0,i=this._findCoordRect(),o={width:e.getWidth(),height:e.getHeight()},a=this._orient===HJ?{right:o.width-i.x-i.width,top:o.height-NJ-PJ-r,width:i.width,height:NJ}:{right:PJ,top:i.y,width:NJ,height:i.height},l=Vc(t.option);Object(s["k"])(["right","top","width","height"],(function(t){"ph"===l[t]&&(l[t]=a[t])}));var u=Hc(l,o);this._location={x:u.x,y:u.y},this._size=[u.width,u.height],this._orient===zJ&&this._size.reverse()},e.prototype._positionGroup=function(){var t=this.group,e=this._location,n=this._orient,r=this.dataZoomModel.getFirstTargetAxisModel(),i=r&&r.get("inverse"),o=this._displayables.sliderGroup,a=(this._dataShadowInfo||{}).otherAxisInverse;o.attr(n!==HJ||i?n===HJ&&i?{scaleY:a?1:-1,scaleX:-1}:n!==zJ||i?{scaleY:a?-1:1,scaleX:-1,rotation:Math.PI/2}:{scaleY:a?-1:1,scaleX:1,rotation:Math.PI/2}:{scaleY:a?1:-1,scaleX:1});var s=t.getBoundingRect([o]);t.x=e.x-s.x,t.y=e.y-s.y,t.markRedraw()},e.prototype._getViewExtent=function(){return[0,this._size[0]]},e.prototype._renderBackground=function(){var t=this.dataZoomModel,e=this._size,n=this._displayables.sliderGroup,r=t.get("brushSelect");n.add(new jJ({silent:!0,shape:{x:0,y:0,width:e[0],height:e[1]},style:{fill:t.get("backgroundColor")},z2:-40}));var i=new jJ({shape:{x:0,y:0,width:e[0],height:e[1]},style:{fill:"transparent"},z2:0,onclick:Object(s["c"])(this._onClickPanel,this)}),o=this.api.getZr();r?(i.on("mousedown",this._onBrushStart,this),i.cursor="crosshair",o.on("mousemove",this._onBrush),o.on("mouseup",this._onBrushEnd)):(o.off("mousemove",this._onBrush),o.off("mouseup",this._onBrushEnd)),n.add(i)},e.prototype._renderDataShadow=function(){var t=this._dataShadowInfo=this._prepareDataShadowInfo();if(this._displayables.dataShadowSegs=[],t){var e=this._size,n=this._shadowSize||[],r=t.series,i=r.getRawData(),o=r.getShadowDim&&r.getShadowDim(),a=o&&i.getDimensionInfo(o)?r.getShadowDim():t.otherDim;if(null!=a){var s=this._shadowPolygonPts,l=this._shadowPolylinePts;if(i!==this._shadowData||a!==this._shadowDim||e[0]!==n[0]||e[1]!==n[1]){var u=i.getDataExtent(a),c=.3*(u[1]-u[0]);u=[u[0]-c,u[1]+c];var d,h=[0,e[1]],f=[0,e[0]],p=[[e[0],0],[0,0]],m=[],g=f[1]/(i.count()-1),y=0,v=Math.round(i.count()/e[0]);i.each([a],(function(t,e){if(v>0&&e%v)y+=g;else{var n=null==t||isNaN(t)||""===t,r=n?0:Er(t,u,h,!0);n&&!d&&e?(p.push([p[p.length-1][0],0]),m.push([m[m.length-1][0],0])):!n&&d&&(p.push([y,0]),m.push([y,0])),p.push([y,r]),m.push([y,r]),y+=g,d=n}})),s=this._shadowPolygonPts=p,l=this._shadowPolylinePts=m}this._shadowData=i,this._shadowDim=a,this._shadowSize=[e[0],e[1]];for(var _=this.dataZoomModel,b=0;b<3;b++){var x=w(1===b);this._displayables.sliderGroup.add(x),this._displayables.dataShadowSegs.push(x)}}}function w(t){var e=_.getModel(t?"selectedDataBackground":"dataBackground"),n=new wr,r=new Tm({shape:{points:s},segmentIgnoreThreshold:1,style:e.getModel("areaStyle").getAreaStyle(),silent:!0,z2:-20}),i=new Im({shape:{points:l},segmentIgnoreThreshold:1,style:e.getModel("lineStyle").getLineStyle(),silent:!0,z2:-19});return n.add(r),n.add(i),n}},e.prototype._prepareDataShadowInfo=function(){var t=this.dataZoomModel,e=t.get("showDataShadow");if(!1!==e){var n,r=this.ecModel;return t.eachTargetAxis((function(i,o){var a=t.getAxisProxy(i,o).getTargetSeriesModels();Object(s["k"])(a,(function(t){if(!n&&!(!0!==e&&Object(s["r"])(BJ,t.get("type"))<0)){var a,l=r.getComponent(b$(i),o).axis,u=GJ(i),c=t.coordinateSystem;null!=u&&c.getOtherAxis&&(a=c.getOtherAxis(l).inverse),u=t.getData().mapDimension(u),n={thisAxis:l,series:t,thisDim:i,otherDim:u,otherAxisInverse:a}}}),this)}),this),n}},e.prototype._renderHandle=function(){var t=this.group,e=this._displayables,n=e.handles=[null,null],r=e.handleLabels=[null,null],i=this._displayables.sliderGroup,o=this._size,a=this.dataZoomModel,l=this.api,u=a.get("borderRadius")||0,c=a.get("brushSelect"),d=e.filler=new jJ({silent:c,style:{fill:a.get("fillerColor")},textConfig:{position:"inside"}});i.add(d),i.add(new jJ({silent:!0,subPixelOptimize:!0,shape:{x:0,y:0,width:o[0],height:o[1],r:u},style:{stroke:a.get("dataBackgroundColor")||a.get("borderColor"),lineWidth:EJ,fill:"rgba(0,0,0,0)"}})),Object(s["k"])([0,1],(function(e){var o=a.get("handleIcon");!nv[o]&&o.indexOf("path://")<0&&o.indexOf("image://")<0&&(o="path://"+o);var l=ov(o,-1,0,2,2,null,!0);l.attr({cursor:$J(this._orient),draggable:!0,drift:Object(s["c"])(this._onDragMove,this,e),ondragend:Object(s["c"])(this._onDragEnd,this),onmouseover:Object(s["c"])(this._showDataInfo,this,!0),onmouseout:Object(s["c"])(this._showDataInfo,this,!1),z2:5});var u=l.getBoundingRect(),c=a.get("handleSize");this._handleHeight=Nr(c,this._size[1]),this._handleWidth=u.width/u.height*this._handleHeight,l.setStyle(a.getModel("handleStyle").getItemStyle()),l.style.strokeNoScale=!0,l.rectHover=!0,l.ensureState("emphasis").style=a.getModel(["emphasis","handleStyle"]).getItemStyle(),Ll(l);var d=a.get("handleColor");null!=d&&(l.style.fill=d),i.add(n[e]=l);var h=a.getModel("textStyle");t.add(r[e]=new ks({silent:!0,invisible:!0,style:nu(h,{x:0,y:0,text:"",verticalAlign:"middle",align:"center",fill:h.getTextColor(),font:h.getFont()}),z2:10}))}),this);var h=d;if(c){var f=Nr(a.get("moveHandleSize"),o[1]),p=e.moveHandle=new ss({style:a.getModel("moveHandleStyle").getItemStyle(),silent:!0,shape:{r:[0,0,2,2],y:o[1]-.5,height:f}}),m=.8*f,g=e.moveHandleIcon=ov(a.get("moveHandleIcon"),-m/2,-m/2,m,m,"#fff",!0);g.silent=!0,g.y=o[1]+f/2-.5,p.ensureState("emphasis").style=a.getModel(["emphasis","moveHandleStyle"]).getItemStyle();var y=Math.min(o[1]/2,Math.max(f,10));h=e.moveZone=new ss({invisible:!0,shape:{y:o[1]-y,height:f+y}}),h.on("mouseover",(function(){l.enterEmphasis(p)})).on("mouseout",(function(){l.leaveEmphasis(p)})),i.add(p),i.add(g),i.add(h)}h.attr({draggable:!0,cursor:$J(this._orient),drift:Object(s["c"])(this._onDragMove,this,"all"),ondragstart:Object(s["c"])(this._showDataInfo,this,!0),ondragend:Object(s["c"])(this._onDragEnd,this),onmouseover:Object(s["c"])(this._showDataInfo,this,!0),onmouseout:Object(s["c"])(this._showDataInfo,this,!1)})},e.prototype._resetInterval=function(){var t=this._range=this.dataZoomModel.getPercentRange(),e=this._getViewExtent();this._handleEnds=[Er(t[0],[0,100],e,!0),Er(t[1],[0,100],e,!0)]},e.prototype._updateInterval=function(t,e){var n=this.dataZoomModel,r=this._handleEnds,i=this._getViewExtent(),o=n.findRepresentativeAxisProxy().getMinMaxSpan(),a=[0,100];nN(e,r,i,n.get("zoomLock")?"all":t,null!=o.minSpan?Er(o.minSpan,a,i,!0):null,null!=o.maxSpan?Er(o.maxSpan,a,i,!0):null);var s=this._range,l=this._range=Hr([Er(r[0],i,a,!0),Er(r[1],i,a,!0)]);return!s||s[0]!==l[0]||s[1]!==l[1]},e.prototype._updateView=function(t){var e=this._displayables,n=this._handleEnds,r=Hr(n.slice()),i=this._size;Object(s["k"])([0,1],(function(t){var r=e.handles[t],o=this._handleHeight;r.attr({scaleX:o/2,scaleY:o/2,x:n[t]+(t?-1:1),y:i[1]/2-o/2})}),this),e.filler.setShape({x:r[0],y:0,width:r[1]-r[0],height:i[1]});var o={x:r[0],width:r[1]-r[0]};e.moveHandle&&(e.moveHandle.setShape(o),e.moveZone.setShape(o),e.moveZone.getBoundingRect(),e.moveHandleIcon&&e.moveHandleIcon.attr("x",o.x+o.width/2));for(var a=e.dataShadowSegs,l=[0,r[0],r[1],i[0]],u=0;u<a.length;u++){var c=a[u],d=c.getClipPath();d||(d=new ss,c.setClipPath(d)),d.setShape({x:l[u],y:0,width:l[u+1]-l[u],height:i[1]})}this._updateDataInfo(t)},e.prototype._updateDataInfo=function(t){var e=this.dataZoomModel,n=this._displayables,r=n.handleLabels,i=this._orient,o=["",""];if(e.get("showDetail")){var a=e.findRepresentativeAxisProxy();if(a){var s=a.getAxisModel().axis,l=this._range,u=t?a.calculateDataWindow({start:l[0],end:l[1]}).valueWindow:a.getDataValueWindow();o=[this._formatLabel(u[0],s),this._formatLabel(u[1],s)]}}var c=Hr(this._handleEnds.slice());function d(t){var e=wg(n.handles[t].parent,this.group),a=Sg(0===t?"right":"left",e),s=this._handleWidth/2+FJ,l=Mg([c[t]+(0===t?-s:s),this._size[1]/2],e);r[t].setStyle({x:l[0],y:l[1],verticalAlign:i===HJ?"middle":a,align:i===HJ?a:"center",text:o[t]})}d.call(this,0),d.call(this,1)},e.prototype._formatLabel=function(t,e){var n=this.dataZoomModel,r=n.get("labelFormatter"),i=n.get("labelPrecision");null!=i&&"auto"!==i||(i=e.getPixelPrecision());var o=null==t||isNaN(t)?"":"category"===e.type||"time"===e.type?e.scale.getLabel({value:Math.round(t)}):t.toFixed(Math.min(i,20));return Object(s["w"])(r)?r(t,o):Object(s["C"])(r)?r.replace("{value}",o):o},e.prototype._showDataInfo=function(t){t=this._dragging||t;var e=this._displayables,n=e.handleLabels;n[0].attr("invisible",!t),n[1].attr("invisible",!t),e.moveHandle&&this.api[t?"enterEmphasis":"leaveEmphasis"](e.moveHandle,1)},e.prototype._onDragMove=function(t,e,n,r){this._dragging=!0,et(r.event);var i=this._displayables.sliderGroup.getLocalTransform(),o=Mg([e,n],i,!0),a=this._updateInterval(t,o[0]),s=this.dataZoomModel.get("realtime");this._updateView(!s),a&&s&&this._dispatchZoomAction(!0)},e.prototype._onDragEnd=function(){this._dragging=!1,this._showDataInfo(!1);var t=this.dataZoomModel.get("realtime");!t&&this._dispatchZoomAction(!1)},e.prototype._onClickPanel=function(t){var e=this._size,n=this._displayables.sliderGroup.transformCoordToLocal(t.offsetX,t.offsetY);if(!(n[0]<0||n[0]>e[0]||n[1]<0||n[1]>e[1])){var r=this._handleEnds,i=(r[0]+r[1])/2,o=this._updateInterval("all",n[0]-i);this._updateView(),o&&this._dispatchZoomAction(!1)}},e.prototype._onBrushStart=function(t){var e=t.offsetX,n=t.offsetY;this._brushStart=new yt(e,n),this._brushing=!0,this._brushStartTime=+new Date},e.prototype._onBrushEnd=function(t){if(this._brushing){var e=this._displayables.brushRect;if(this._brushing=!1,e){e.attr("ignore",!0);var n=e.shape,r=+new Date;if(!(r-this._brushStartTime<200&&Math.abs(n.width)<5)){var i=this._getViewExtent(),o=[0,100];this._range=Hr([Er(n.x,i,o,!0),Er(n.x+n.width,i,o,!0)]),this._handleEnds=[n.x,n.x+n.width],this._updateView(),this._dispatchZoomAction(!1)}}}},e.prototype._onBrush=function(t){this._brushing&&(et(t.event),this._updateBrushRect(t.offsetX,t.offsetY))},e.prototype._updateBrushRect=function(t,e){var n=this._displayables,r=this.dataZoomModel,i=n.brushRect;i||(i=n.brushRect=new jJ({silent:!0,style:r.getModel("brushStyle").getItemStyle()}),n.sliderGroup.add(i)),i.attr("ignore",!1);var o=this._brushStart,a=this._displayables.sliderGroup,s=a.transformCoordToLocal(t,e),l=a.transformCoordToLocal(o.x,o.y),u=this._size;s[0]=Math.max(Math.min(u[0],s[0]),0),i.setShape({x:l[0],y:0,width:s[0]-l[0],height:u[1]})},e.prototype._dispatchZoomAction=function(t){var e=this._range;this.api.dispatchAction({type:"dataZoom",from:this.uid,dataZoomId:this.dataZoomModel.id,animation:t?WJ:null,start:e[0],end:e[1]})},e.prototype._findCoordRect=function(){var t,e=w$(this.dataZoomModel).infoList;if(!t&&e.length){var n=e[0].model.coordinateSystem;t=n.getRect&&n.getRect()}if(!t){var r=this.api.getWidth(),i=this.api.getHeight();t={x:.2*r,y:.2*i,width:.6*r,height:.6*i}}return t},e.type="dataZoom.slider",e}(I$);function GJ(t){var e={x:"y",y:"x",radius:"angle",angle:"radius"};return e[t]}function $J(t){return"vertical"===t?"ns-resize":"ew-resize"}var UJ=VJ;function qJ(t){t.registerComponentModel(YJ),t.registerComponentView(UJ),B$(t)}function XJ(t){jb(CJ),jb(qJ)}var ZJ={get:function(t,e,n){var r=s["d"]((JJ[t]||{})[e]);return n&&s["t"](r)?r[r.length-1]:r}},JJ={color:{active:["#006edd","#e0ffff"],inactive:["rgba(0,0,0,0)"]},colorHue:{active:[0,360],inactive:[0,0]},colorSaturation:{active:[.3,1],inactive:[0,0]},colorLightness:{active:[.9,.5],inactive:[0,0]},colorAlpha:{active:[.3,1],inactive:[0,0]},opacity:{active:[.3,1],inactive:[0,0]},symbol:{active:["circle","roundRect","diamond"],inactive:["none"]},symbolSize:{active:[10,50],inactive:[0,0]}},KJ=ZJ,QJ=uj.mapVisual,tK=uj.eachVisual,eK=s["t"],nK=s["k"],rK=Hr,iK=Er,oK=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.stateList=["inRange","outOfRange"],n.replacableOptionKeys=["inRange","outOfRange","target","controller","color"],n.layoutMode={type:"box",ignoreSize:!0},n.dataBound=[-1/0,1/0],n.targetVisuals={},n.controllerVisuals={},n}return o(e,t),e.prototype.init=function(t,e,n){this.mergeDefaultAndTheme(t,n)},e.prototype.optionUpdated=function(t,e){var n=this.option;!e&&Wq(n,t,this.replacableOptionKeys),this.textStyleModel=this.getModel("textStyle"),this.resetItemSize(),this.completeVisualOption()},e.prototype.resetVisual=function(t){var e=this.stateList;t=s["c"](t,this),this.controllerVisuals=Bq(this.option.controller,e,t),this.targetVisuals=Bq(this.option.target,e,t)},e.prototype.getItemSymbol=function(){return null},e.prototype.getTargetSeriesIndices=function(){var t=this.option.seriesIndex,e=[];return null==t||"all"===t?this.ecModel.eachSeries((function(t,n){e.push(n)})):e=ui(t),e},e.prototype.eachTargetSeries=function(t,e){s["k"](this.getTargetSeriesIndices(),(function(n){var r=this.ecModel.getSeriesByIndex(n);r&&t.call(e,r)}),this)},e.prototype.isTargetSeries=function(t){var e=!1;return this.eachTargetSeries((function(n){n===t&&(e=!0)})),e},e.prototype.formatValueText=function(t,e,n){var r,i=this.option,o=i.precision,a=this.dataBound,l=i.formatter;n=n||["<",">"],s["t"](t)&&(t=t.slice(),r=!0);var u=e?t:r?[c(t[0]),c(t[1])]:c(t);return s["C"](l)?l.replace("{value}",r?u[0]:u).replace("{value2}",r?u[1]:u):s["w"](l)?r?l(t[0],t[1]):l(t):r?t[0]===a[0]?n[0]+" "+u[1]:t[1]===a[1]?n[1]+" "+u[0]:u[0]+" - "+u[1]:u;function c(t){return t===a[0]?"min":t===a[1]?"max":(+t).toFixed(Math.min(o,20))}},e.prototype.resetExtent=function(){var t=this.option,e=rK([t.min,t.max]);this._dataExtent=e},e.prototype.getDataDimensionIndex=function(t){var e=this.option.dimension;if(null!=e)return t.getDimensionIndex(e);for(var n=t.dimensions,r=n.length-1;r>=0;r--){var i=n[r],o=t.getDimensionInfo(i);if(!o.isCalculationCoord)return o.storeDimIndex}},e.prototype.getExtent=function(){return this._dataExtent.slice()},e.prototype.completeVisualOption=function(){var t=this.ecModel,e=this.option,n={inRange:e.inRange,outOfRange:e.outOfRange},r=e.target||(e.target={}),i=e.controller||(e.controller={});s["I"](r,n),s["I"](i,n);var o=this.isCategory();function a(n){eK(e.color)&&!n.inRange&&(n.inRange={color:e.color.slice().reverse()}),n.inRange=n.inRange||{color:t.get("gradientColor")}}function l(t,e,n){var r=t[e],i=t[n];r&&!i&&(i=t[n]={},nK(r,(function(t,e){if(uj.isValidType(e)){var n=KJ.get(e,"inactive",o);null!=n&&(i[e]=n,"color"!==e||i.hasOwnProperty("opacity")||i.hasOwnProperty("colorAlpha")||(i.opacity=[0,0]))}})))}function u(t){var e=(t.inRange||{}).symbol||(t.outOfRange||{}).symbol,n=(t.inRange||{}).symbolSize||(t.outOfRange||{}).symbolSize,r=this.get("inactiveColor"),i=this.getItemSymbol(),a=i||"roundRect";nK(this.stateList,(function(i){var l=this.itemSize,u=t[i];u||(u=t[i]={color:o?r:[r]}),null==u.symbol&&(u.symbol=e&&s["d"](e)||(o?a:[a])),null==u.symbolSize&&(u.symbolSize=n&&s["d"](n)||(o?l[0]:[l[0],l[0]])),u.symbol=QJ(u.symbol,(function(t){return"none"===t?a:t}));var c=u.symbolSize;if(null!=c){var d=-1/0;tK(c,(function(t){t>d&&(d=t)})),u.symbolSize=QJ(c,(function(t){return iK(t,[0,d],[0,l[0]],!0)}))}}),this)}a.call(this,r),a.call(this,i),l.call(this,r,"inRange","outOfRange"),u.call(this,i)},e.prototype.resetItemSize=function(){this.itemSize=[parseFloat(this.get("itemWidth")),parseFloat(this.get("itemHeight"))]},e.prototype.isCategory=function(){return!!this.option.categories},e.prototype.setSelected=function(t){},e.prototype.getSelected=function(){return null},e.prototype.getValueState=function(t){return null},e.prototype.getVisualMeta=function(t){return null},e.type="visualMap",e.dependencies=["series"],e.defaultOption={show:!0,z:4,seriesIndex:"all",min:0,max:200,left:0,right:null,top:null,bottom:0,itemWidth:null,itemHeight:null,inverse:!1,orient:"vertical",backgroundColor:"rgba(0,0,0,0)",borderColor:"#ccc",contentColor:"#5793f3",inactiveColor:"#aaa",borderWidth:0,padding:5,textGap:10,precision:0,textStyle:{color:"#333"}},e}(Xc),aK=oK,sK=[20,140],lK=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.optionUpdated=function(e,n){t.prototype.optionUpdated.apply(this,arguments),this.resetExtent(),this.resetVisual((function(t){t.mappingMethod="linear",t.dataExtent=this.getExtent()})),this._resetRange()},e.prototype.resetItemSize=function(){t.prototype.resetItemSize.apply(this,arguments);var e=this.itemSize;(null==e[0]||isNaN(e[0]))&&(e[0]=sK[0]),(null==e[1]||isNaN(e[1]))&&(e[1]=sK[1])},e.prototype._resetRange=function(){var t=this.getExtent(),e=this.option.range;!e||e.auto?(t.auto=1,this.option.range=t):s["t"](e)&&(e[0]>e[1]&&e.reverse(),e[0]=Math.max(e[0],t[0]),e[1]=Math.min(e[1],t[1]))},e.prototype.completeVisualOption=function(){t.prototype.completeVisualOption.apply(this,arguments),s["k"](this.stateList,(function(t){var e=this.option.controller[t].symbolSize;e&&e[0]!==e[1]&&(e[0]=e[1]/3)}),this)},e.prototype.setSelected=function(t){this.option.range=t.slice(),this._resetRange()},e.prototype.getSelected=function(){var t=this.getExtent(),e=Hr((this.get("range")||[]).slice());return e[0]>t[1]&&(e[0]=t[1]),e[1]>t[1]&&(e[1]=t[1]),e[0]<t[0]&&(e[0]=t[0]),e[1]<t[0]&&(e[1]=t[0]),e},e.prototype.getValueState=function(t){var e=this.option.range,n=this.getExtent();return(e[0]<=n[0]||e[0]<=t)&&(e[1]>=n[1]||t<=e[1])?"inRange":"outOfRange"},e.prototype.findTargetDataIndices=function(t){var e=[];return this.eachTargetSeries((function(n){var r=[],i=n.getData();i.each(this.getDataDimensionIndex(i),(function(e,n){t[0]<=e&&e<=t[1]&&r.push(n)}),this),e.push({seriesId:n.id,dataIndex:r})}),this),e},e.prototype.getVisualMeta=function(t){var e=uK(this,"outOfRange",this.getExtent()),n=uK(this,"inRange",this.option.range.slice()),r=[];function i(e,n){r.push({value:e,color:t(e,n)})}for(var o=0,a=0,s=n.length,l=e.length;a<l&&(!n.length||e[a]<=n[0]);a++)e[a]<n[o]&&i(e[a],"outOfRange");for(var u=1;o<s;o++,u=0)u&&r.length&&i(n[o],"outOfRange"),i(n[o],"inRange");for(u=1;a<l;a++)(!n.length||n[n.length-1]<e[a])&&(u&&(r.length&&i(r[r.length-1].value,"outOfRange"),u=0),i(e[a],"outOfRange"));var c=r.length;return{stops:r,outerColors:[c?r[0].color:"transparent",c?r[c-1].color:"transparent"]}},e.type="visualMap.continuous",e.defaultOption=Cu(aK.defaultOption,{align:"auto",calculable:!1,hoverLink:!0,realtime:!0,handleIcon:"path://M-11.39,9.77h0a3.5,3.5,0,0,1-3.5,3.5h-22a3.5,3.5,0,0,1-3.5-3.5h0a3.5,3.5,0,0,1,3.5-3.5h22A3.5,3.5,0,0,1-11.39,9.77Z",handleSize:"120%",handleStyle:{borderColor:"#fff",borderWidth:1},indicatorIcon:"circle",indicatorSize:"50%",indicatorStyle:{borderColor:"#fff",borderWidth:2,shadowBlur:2,shadowOffsetX:1,shadowOffsetY:1,shadowColor:"rgba(0,0,0,0.2)"}}),e}(aK);function uK(t,e,n){if(n[0]===n[1])return n.slice();for(var r=200,i=(n[1]-n[0])/r,o=n[0],a=[],s=0;s<=r&&o<n[1];s++)a.push(o),o+=i;return a.push(n[1]),a}var cK=lK,dK=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.autoPositionValues={left:1,right:1,top:1,bottom:1},n}return o(e,t),e.prototype.init=function(t,e){this.ecModel=t,this.api=e},e.prototype.render=function(t,e,n,r){this.visualMapModel=t,!1!==t.get("show")?this.doRender(t,e,n,r):this.group.removeAll()},e.prototype.renderBackground=function(t){var e=this.visualMapModel,n=Sc(e.get("padding")||0),r=t.getBoundingRect();t.add(new ss({z2:-1,silent:!0,shape:{x:r.x-n[3],y:r.y-n[0],width:r.width+n[3]+n[1],height:r.height+n[0]+n[2]},style:{fill:e.get("backgroundColor"),stroke:e.get("borderColor"),lineWidth:e.get("borderWidth")}}))},e.prototype.getControllerVisual=function(t,e,n){n=n||{};var r=n.forceState,i=this.visualMapModel,o={};if("color"===e){var a=i.get("contentColor");o.color=a}function l(t){return o[t]}function u(t,e){o[t]=e}var c=i.controllerVisuals[r||i.getValueState(t)],d=uj.prepareVisualTypes(c);return s["k"](d,(function(r){var i=c[r];n.convertOpacityToAlpha&&"opacity"===r&&(r="colorAlpha",i=c.__alphaForOpacity),uj.dependsOn(r,e)&&i&&i.applyVisual(t,l,u)})),o[e]},e.prototype.positionGroup=function(t){var e=this.visualMapModel,n=this.api;zc(t,e.getBoxLayoutParams(),{width:n.getWidth(),height:n.getHeight()})},e.prototype.doRender=function(t,e,n,r){},e.type="visualMap",e}(kp),hK=dK,fK=[["left","right","width"],["top","bottom","height"]];function pK(t,e,n){var r=t.option,i=r.align;if(null!=i&&"auto"!==i)return i;for(var o={width:e.getWidth(),height:e.getHeight()},a="horizontal"===r.orient?1:0,s=fK[a],l=[0,null,10],u={},c=0;c<3;c++)u[fK[1-a][c]]=l[c],u[s[c]]=2===c?n[0]:r[s[c]];var d=[["x","width",3],["y","height",0]][a],h=Hc(u,o,r.padding);return s[(h.margin[d[2]]||0)+h[d[0]]+.5*h[d[1]]<.5*o[d[1]]?0:1]}function mK(t,e){return s["k"](t||[],(function(t){null!=t.dataIndex&&(t.dataIndexInside=t.dataIndex,t.dataIndex=null),t.highlightKey="visualMap"+(e?e.componentIndex:"")})),t}var gK=Er,yK=s["k"],vK=Math.min,_K=Math.max,bK=12,xK=6,wK=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n._shapes={},n._dataInterval=[],n._handleEnds=[],n._hoverLinkDataIndices=[],n}return o(e,t),e.prototype.init=function(e,n){t.prototype.init.call(this,e,n),this._hoverLinkFromSeriesMouseOver=s["c"](this._hoverLinkFromSeriesMouseOver,this),this._hideIndicator=s["c"](this._hideIndicator,this)},e.prototype.doRender=function(t,e,n,r){r&&"selectDataRange"===r.type&&r.from===this.uid||this._buildView()},e.prototype._buildView=function(){this.group.removeAll();var t=this.visualMapModel,e=this.group;this._orient=t.get("orient"),this._useHandle=t.get("calculable"),this._resetInterval(),this._renderBar(e);var n=t.get("text");this._renderEndsText(e,n,0),this._renderEndsText(e,n,1),this._updateView(!0),this.renderBackground(e),this._updateView(),this._enableHoverLinkToSeries(),this._enableHoverLinkFromSeries(),this.positionGroup(e)},e.prototype._renderEndsText=function(t,e,n){if(e){var r=e[1-n];r=null!=r?r+"":"";var i=this.visualMapModel,o=i.get("textGap"),a=i.itemSize,s=this._shapes.mainGroup,l=this._applyTransform([a[0]/2,0===n?-o:a[1]+o],s),u=this._applyTransform(0===n?"bottom":"top",s),c=this._orient,d=this.visualMapModel.textStyleModel;this.group.add(new ks({style:nu(d,{x:l[0],y:l[1],verticalAlign:"horizontal"===c?"middle":u,align:"horizontal"===c?u:"center",text:r})}))}},e.prototype._renderBar=function(t){var e=this.visualMapModel,n=this._shapes,r=e.itemSize,i=this._orient,o=this._useHandle,a=pK(e,this.api,r),l=n.mainGroup=this._createBarGroup(a),u=new wr;l.add(u),u.add(n.outOfRange=MK()),u.add(n.inRange=MK(null,o?LK(this._orient):null,s["c"](this._dragHandle,this,"all",!1),s["c"](this._dragHandle,this,"all",!0))),u.setClipPath(new ss({shape:{x:0,y:0,width:r[0],height:r[1],r:3}}));var c=e.textStyleModel.getTextRect("国"),d=_K(c.width,c.height);o&&(n.handleThumbs=[],n.handleLabels=[],n.handleLabelPoints=[],this._createHandle(e,l,0,r,d,i),this._createHandle(e,l,1,r,d,i)),this._createIndicator(e,l,r,d,i),t.add(l)},e.prototype._createHandle=function(t,e,n,r,i,o){var a=s["c"](this._dragHandle,this,n,!1),l=s["c"](this._dragHandle,this,n,!0),u=or(t.get("handleSize"),r[0]),c=ov(t.get("handleIcon"),-u/2,-u/2,u,u,null,!0),d=LK(this._orient);c.attr({cursor:d,draggable:!0,drift:a,ondragend:l,onmousemove:function(t){et(t.event)}}),c.x=r[0]/2,c.useStyle(t.getModel("handleStyle").getItemStyle()),c.setStyle({strokeNoScale:!0,strokeFirst:!0}),c.style.lineWidth*=2,c.ensureState("emphasis").style=t.getModel(["emphasis","handleStyle"]).getItemStyle(),Yl(c,!0),e.add(c);var h=this.visualMapModel.textStyleModel,f=new ks({cursor:d,draggable:!0,drift:a,onmousemove:function(t){et(t.event)},ondragend:l,style:nu(h,{x:0,y:0,text:""})});f.ensureState("blur").style={opacity:.1},f.stateTransition={duration:200},this.group.add(f);var p=[u,0],m=this._shapes;m.handleThumbs[n]=c,m.handleLabelPoints[n]=p,m.handleLabels[n]=f},e.prototype._createIndicator=function(t,e,n,r,i){var o=or(t.get("indicatorSize"),n[0]),a=ov(t.get("indicatorIcon"),-o/2,-o/2,o,o,null,!0);a.attr({cursor:"move",invisible:!0,silent:!0,x:n[0]/2});var l=t.getModel("indicatorStyle").getItemStyle();if(a instanceof Ka){var u=a.style;a.useStyle(s["m"]({image:u.image,x:u.x,y:u.y,width:u.width,height:u.height},l))}else a.useStyle(l);e.add(a);var c=this.visualMapModel.textStyleModel,d=new ks({silent:!0,invisible:!0,style:nu(c,{x:0,y:0,text:""})});this.group.add(d);var h=[("horizontal"===i?r/2:xK)+n[0]/2,0],f=this._shapes;f.indicator=a,f.indicatorLabel=d,f.indicatorLabelPoint=h,this._firstShowIndicator=!0},e.prototype._dragHandle=function(t,e,n,r){if(this._useHandle){if(this._dragging=!e,!e){var i=this._applyTransform([n,r],this._shapes.mainGroup,!0);this._updateInterval(t,i[1]),this._hideIndicator(),this._updateView()}e===!this.visualMapModel.get("realtime")&&this.api.dispatchAction({type:"selectDataRange",from:this.uid,visualMapId:this.visualMapModel.id,selected:this._dataInterval.slice()}),e?!this._hovering&&this._clearHoverLinkToSeries():kK(this.visualMapModel)&&this._doHoverLinkToSeries(this._handleEnds[t],!1)}},e.prototype._resetInterval=function(){var t=this.visualMapModel,e=this._dataInterval=t.getSelected(),n=t.getExtent(),r=[0,t.itemSize[1]];this._handleEnds=[gK(e[0],n,r,!0),gK(e[1],n,r,!0)]},e.prototype._updateInterval=function(t,e){e=e||0;var n=this.visualMapModel,r=this._handleEnds,i=[0,n.itemSize[1]];nN(e,r,i,t,0);var o=n.getExtent();this._dataInterval=[gK(r[0],i,o,!0),gK(r[1],i,o,!0)]},e.prototype._updateView=function(t){var e=this.visualMapModel,n=e.getExtent(),r=this._shapes,i=[0,e.itemSize[1]],o=t?i:this._handleEnds,a=this._createBarVisual(this._dataInterval,n,o,"inRange"),s=this._createBarVisual(n,n,i,"outOfRange");r.inRange.setStyle({fill:a.barColor}).setShape("points",a.barPoints),r.outOfRange.setStyle({fill:s.barColor}).setShape("points",s.barPoints),this._updateHandle(o,a)},e.prototype._createBarVisual=function(t,e,n,r){var i={forceState:r,convertOpacityToAlpha:!0},o=this._makeColorGradient(t,i),a=[this.getControllerVisual(t[0],"symbolSize",i),this.getControllerVisual(t[1],"symbolSize",i)],s=this._createBarPoints(n,a);return{barColor:new qm(0,0,0,1,o),barPoints:s,handlesColor:[o[0].color,o[o.length-1].color]}},e.prototype._makeColorGradient=function(t,e){var n=100,r=[],i=(t[1]-t[0])/n;r.push({color:this.getControllerVisual(t[0],"color",e),offset:0});for(var o=1;o<n;o++){var a=t[0]+i*o;if(a>t[1])break;r.push({color:this.getControllerVisual(a,"color",e),offset:o/n})}return r.push({color:this.getControllerVisual(t[1],"color",e),offset:1}),r},e.prototype._createBarPoints=function(t,e){var n=this.visualMapModel.itemSize;return[[n[0]-e[0],t[0]],[n[0],t[0]],[n[0],t[1]],[n[0]-e[1],t[1]]]},e.prototype._createBarGroup=function(t){var e=this._orient,n=this.visualMapModel.get("inverse");return new wr("horizontal"!==e||n?"horizontal"===e&&n?{scaleX:"bottom"===t?-1:1,rotation:-Math.PI/2}:"vertical"!==e||n?{scaleX:"left"===t?1:-1}:{scaleX:"left"===t?1:-1,scaleY:-1}:{scaleX:"bottom"===t?1:-1,rotation:Math.PI/2})},e.prototype._updateHandle=function(t,e){if(this._useHandle){var n=this._shapes,r=this.visualMapModel,i=n.handleThumbs,o=n.handleLabels,a=r.itemSize,s=r.getExtent();yK([0,1],(function(l){var u=i[l];u.setStyle("fill",e.handlesColor[l]),u.y=t[l];var c=gK(t[l],[0,a[1]],s,!0),d=this.getControllerVisual(c,"symbolSize");u.scaleX=u.scaleY=d/a[0],u.x=a[0]-d/2;var h=Mg(n.handleLabelPoints[l],wg(u,this.group));o[l].setStyle({x:h[0],y:h[1],text:r.formatValueText(this._dataInterval[l]),verticalAlign:"middle",align:"vertical"===this._orient?this._applyTransform("left",n.mainGroup):"center"})}),this)}},e.prototype._showIndicator=function(t,e,n,r){var i=this.visualMapModel,o=i.getExtent(),a=i.itemSize,s=[0,a[1]],l=this._shapes,u=l.indicator;if(u){u.attr("invisible",!1);var c={convertOpacityToAlpha:!0},d=this.getControllerVisual(t,"color",c),h=this.getControllerVisual(t,"symbolSize"),f=gK(t,o,s,!0),p=a[0]-h/2,m={x:u.x,y:u.y};u.y=f,u.x=p;var g=Mg(l.indicatorLabelPoint,wg(u,this.group)),y=l.indicatorLabel;y.attr("invisible",!1);var v=this._applyTransform("left",l.mainGroup),_=this._orient,b="horizontal"===_;y.setStyle({text:(n||"")+i.formatValueText(e),verticalAlign:b?v:"middle",align:b?"center":v});var x={x:p,y:f,style:{fill:d}},w={style:{x:g[0],y:g[1]}};if(i.ecModel.isAnimationEnabled()&&!this._firstShowIndicator){var M={duration:100,easing:"cubicInOut",additive:!0};u.x=m.x,u.y=m.y,u.animateTo(x,M),y.animateTo(w,M)}else u.attr(x),y.attr(w);this._firstShowIndicator=!1;var S=this._shapes.handleLabels;if(S)for(var k=0;k<S.length;k++)this.api.enterBlur(S[k])}},e.prototype._enableHoverLinkToSeries=function(){var t=this;this._shapes.mainGroup.on("mousemove",(function(e){if(t._hovering=!0,!t._dragging){var n=t.visualMapModel.itemSize,r=t._applyTransform([e.offsetX,e.offsetY],t._shapes.mainGroup,!0,!0);r[1]=vK(_K(0,r[1]),n[1]),t._doHoverLinkToSeries(r[1],0<=r[0]&&r[0]<=n[0])}})).on("mouseout",(function(){t._hovering=!1,!t._dragging&&t._clearHoverLinkToSeries()}))},e.prototype._enableHoverLinkFromSeries=function(){var t=this.api.getZr();this.visualMapModel.option.hoverLink?(t.on("mouseover",this._hoverLinkFromSeriesMouseOver,this),t.on("mouseout",this._hideIndicator,this)):this._clearHoverLinkFromSeries()},e.prototype._doHoverLinkToSeries=function(t,e){var n=this.visualMapModel,r=n.itemSize;if(n.option.hoverLink){var i=[0,r[1]],o=n.getExtent();t=vK(_K(i[0],t),i[1]);var a=SK(n,o,i),s=[t-a,t+a],l=gK(t,i,o,!0),u=[gK(s[0],i,o,!0),gK(s[1],i,o,!0)];s[0]<i[0]&&(u[0]=-1/0),s[1]>i[1]&&(u[1]=1/0),e&&(u[0]===-1/0?this._showIndicator(l,u[1],"< ",a):u[1]===1/0?this._showIndicator(l,u[0],"> ",a):this._showIndicator(l,l,"≈ ",a));var c=this._hoverLinkDataIndices,d=[];(e||kK(n))&&(d=this._hoverLinkDataIndices=n.findTargetDataIndices(u));var h=Oi(c,d);this._dispatchHighDown("downplay",mK(h[0],n)),this._dispatchHighDown("highlight",mK(h[1],n))}},e.prototype._hoverLinkFromSeriesMouseOver=function(t){var e;if(Gy(t.target,(function(t){var n=Ls(t);if(null!=n.dataIndex)return e=n,!0}),!0),e){var n=this.ecModel.getSeriesByIndex(e.seriesIndex),r=this.visualMapModel;if(r.isTargetSeries(n)){var i=n.getData(e.dataType),o=i.getStore().get(r.getDataDimensionIndex(i),e.dataIndex);isNaN(o)||this._showIndicator(o,o)}}},e.prototype._hideIndicator=function(){var t=this._shapes;t.indicator&&t.indicator.attr("invisible",!0),t.indicatorLabel&&t.indicatorLabel.attr("invisible",!0);var e=this._shapes.handleLabels;if(e)for(var n=0;n<e.length;n++)this.api.leaveBlur(e[n])},e.prototype._clearHoverLinkToSeries=function(){this._hideIndicator();var t=this._hoverLinkDataIndices;this._dispatchHighDown("downplay",mK(t,this.visualMapModel)),t.length=0},e.prototype._clearHoverLinkFromSeries=function(){this._hideIndicator();var t=this.api.getZr();t.off("mouseover",this._hoverLinkFromSeriesMouseOver),t.off("mouseout",this._hideIndicator)},e.prototype._applyTransform=function(t,e,n,r){var i=wg(e,r?null:this.group);return s["t"](t)?Mg(t,i,n):Sg(t,i,n)},e.prototype._dispatchHighDown=function(t,e){e&&e.length&&this.api.dispatchAction({type:t,batch:e})},e.prototype.dispose=function(){this._clearHoverLinkFromSeries(),this._clearHoverLinkToSeries()},e.type="visualMap.continuous",e}(hK);function MK(t,e,n,r){return new Tm({shape:{points:t},draggable:!!n,cursor:e,drift:n,onmousemove:function(t){et(t.event)},ondragend:r})}function SK(t,e,n){var r=bK/2,i=t.get("hoverLinkDataSize");return i&&(r=gK(i,e,n,!0)/2),r}function kK(t){var e=t.get("hoverLinkOnHandle");return!!(null==e?t.get("realtime"):e)}function LK(t){return"vertical"===t?"ns-resize":"ew-resize"}var TK=wK,DK={type:"selectDataRange",event:"dataRangeSelected",update:"update"},OK=function(t,e){e.eachComponent({mainType:"visualMap",query:t},(function(e){e.setSelected(t.selected)}))},IK=[{createOnAllSeries:!0,reset:function(t,e){var n=[];return e.eachComponent("visualMap",(function(e){var r=t.pipelineContext;!e.isTargetSeries(t)||r&&r.large||n.push(Gq(e.stateList,e.targetVisuals,s["c"](e.getValueState,e),e.getDataDimensionIndex(t.getData())))})),n}},{createOnAllSeries:!0,reset:function(t,e){var n=t.getData(),r=[];e.eachComponent("visualMap",(function(e){if(e.isTargetSeries(t)){var i=e.getVisualMeta(s["c"](CK,null,t,e))||{stops:[],outerColors:[]},o=e.getDataDimensionIndex(n);o>=0&&(i.dimension=o,r.push(i))}})),t.getData().setVisual("visualMeta",r)}}];function CK(t,e,n,r){for(var i=e.targetVisuals[r],o=uj.prepareVisualTypes(i),a={color:zy(t.getData(),"color")},s=0,l=o.length;s<l;s++){var u=o[s],c=i["opacity"===u?"__alphaForOpacity":u];c&&c.applyVisual(n,d,h)}return a.color;function d(t){return a[t]}function h(t,e){a[t]=e}}var AK=s["k"];function YK(t){var e=t&&t.visualMap;s["t"](e)||(e=e?[e]:[]),AK(e,(function(t){if(t){jK(t,"splitList")&&!jK(t,"pieces")&&(t.pieces=t.splitList,delete t.splitList);var e=t.pieces;e&&s["t"](e)&&AK(e,(function(t){s["A"](t)&&(jK(t,"start")&&!jK(t,"min")&&(t.min=t.start),jK(t,"end")&&!jK(t,"max")&&(t.max=t.end))}))}}))}function jK(t,e){return t&&t.hasOwnProperty&&t.hasOwnProperty(e)}var PK=!1;function EK(t){PK||(PK=!0,t.registerSubTypeDefaulter("visualMap",(function(t){return t.categories||(t.pieces?t.pieces.length>0:t.splitNumber>0)&&!t.calculable?"piecewise":"continuous"})),t.registerAction(DK,OK),Object(s["k"])(IK,(function(e){t.registerVisual(t.PRIORITY.VISUAL.COMPONENT,e)})),t.registerPreprocessor(YK))}function NK(t){t.registerComponentModel(cK),t.registerComponentView(TK),EK(t)}var RK=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n._pieceList=[],n}return o(e,t),e.prototype.optionUpdated=function(e,n){t.prototype.optionUpdated.apply(this,arguments),this.resetExtent();var r=this._mode=this._determineMode();this._pieceList=[],HK[this._mode].call(this,this._pieceList),this._resetSelected(e,n);var i=this.option.categories;this.resetVisual((function(t,e){"categories"===r?(t.mappingMethod="category",t.categories=s["d"](i)):(t.dataExtent=this.getExtent(),t.mappingMethod="piecewise",t.pieceList=s["H"](this._pieceList,(function(t){return t=s["d"](t),"inRange"!==e&&(t.visual=null),t})))}))},e.prototype.completeVisualOption=function(){var e=this.option,n={},r=uj.listVisualTypes(),i=this.isCategory();function o(t,e,n){return t&&t[e]&&t[e].hasOwnProperty(n)}s["k"](e.pieces,(function(t){s["k"](r,(function(e){t.hasOwnProperty(e)&&(n[e]=1)}))})),s["k"](n,(function(t,n){var r=!1;s["k"](this.stateList,(function(t){r=r||o(e,t,n)||o(e.target,t,n)}),this),!r&&s["k"](this.stateList,(function(t){(e[t]||(e[t]={}))[n]=KJ.get(n,"inRange"===t?"active":"inactive",i)}))}),this),t.prototype.completeVisualOption.apply(this,arguments)},e.prototype._resetSelected=function(t,e){var n=this.option,r=this._pieceList,i=(e?n:t).selected||{};if(n.selected=i,s["k"](r,(function(t,e){var n=this.getSelectedMapKey(t);i.hasOwnProperty(n)||(i[n]=!0)}),this),"single"===n.selectedMode){var o=!1;s["k"](r,(function(t,e){var n=this.getSelectedMapKey(t);i[n]&&(o?i[n]=!1:o=!0)}),this)}},e.prototype.getItemSymbol=function(){return this.get("itemSymbol")},e.prototype.getSelectedMapKey=function(t){return"categories"===this._mode?t.value+"":t.index+""},e.prototype.getPieceList=function(){return this._pieceList},e.prototype._determineMode=function(){var t=this.option;return t.pieces&&t.pieces.length>0?"pieces":this.option.categories?"categories":"splitNumber"},e.prototype.setSelected=function(t){this.option.selected=s["d"](t)},e.prototype.getValueState=function(t){var e=uj.findPieceIndex(t,this._pieceList);return null!=e&&this.option.selected[this.getSelectedMapKey(this._pieceList[e])]?"inRange":"outOfRange"},e.prototype.findTargetDataIndices=function(t){var e=[],n=this._pieceList;return this.eachTargetSeries((function(r){var i=[],o=r.getData();o.each(this.getDataDimensionIndex(o),(function(e,r){var o=uj.findPieceIndex(e,n);o===t&&i.push(r)}),this),e.push({seriesId:r.id,dataIndex:i})}),this),e},e.prototype.getRepresentValue=function(t){var e;if(this.isCategory())e=t.value;else if(null!=t.value)e=t.value;else{var n=t.interval||[];e=n[0]===-1/0&&n[1]===1/0?0:(n[0]+n[1])/2}return e},e.prototype.getVisualMeta=function(t){if(!this.isCategory()){var e=[],n=["",""],r=this,i=this._pieceList.slice();if(i.length){var o=i[0].interval[0];o!==-1/0&&i.unshift({interval:[-1/0,o]}),o=i[i.length-1].interval[1],o!==1/0&&i.push({interval:[o,1/0]})}else i.push({interval:[-1/0,1/0]});var a=-1/0;return s["k"](i,(function(t){var e=t.interval;e&&(e[0]>a&&l([a,e[0]],"outOfRange"),l(e.slice()),a=e[1])}),this),{stops:e,outerColors:n}}function l(i,o){var a=r.getRepresentValue({interval:i});o||(o=r.getValueState(a));var s=t(a,o);i[0]===-1/0?n[0]=s:i[1]===1/0?n[1]=s:e.push({value:i[0],color:s},{value:i[1],color:s})}},e.type="visualMap.piecewise",e.defaultOption=Cu(aK.defaultOption,{selected:null,minOpen:!1,maxOpen:!1,align:"auto",itemWidth:20,itemHeight:14,itemSymbol:"roundRect",pieces:null,categories:null,splitNumber:5,selectedMode:"multiple",itemGap:10,hoverLink:!0}),e}(aK),HK={splitNumber:function(t){var e=this.option,n=Math.min(e.precision,20),r=this.getExtent(),i=e.splitNumber;i=Math.max(parseInt(i,10),1),e.splitNumber=i;var o=(r[1]-r[0])/i;while(+o.toFixed(n)!==o&&n<5)n++;e.precision=n,o=+o.toFixed(n),e.minOpen&&t.push({interval:[-1/0,r[0]],close:[0,0]});for(var a=0,l=r[0];a<i;l+=o,a++){var u=a===i-1?r[1]:l+o;t.push({interval:[l,u],close:[1,1]})}e.maxOpen&&t.push({interval:[r[1],1/0],close:[0,0]}),ti(t),s["k"](t,(function(t,e){t.index=e,t.text=this.formatValueText(t.interval)}),this)},categories:function(t){var e=this.option;s["k"](e.categories,(function(e){t.push({text:this.formatValueText(e,!0),value:e})}),this),zK(e,t)},pieces:function(t){var e=this.option;s["k"](e.pieces,(function(e,n){s["A"](e)||(e={value:e});var r={text:"",index:n};if(null!=e.label&&(r.text=e.label),e.hasOwnProperty("value")){var i=r.value=e.value;r.interval=[i,i],r.close=[1,1]}else{for(var o=r.interval=[],a=r.close=[0,0],l=[1,0,1],u=[-1/0,1/0],c=[],d=0;d<2;d++){for(var h=[["gte","gt","min"],["lte","lt","max"]][d],f=0;f<3&&null==o[d];f++)o[d]=e[h[f]],a[d]=l[f],c[d]=2===f;null==o[d]&&(o[d]=u[d])}c[0]&&o[1]===1/0&&(a[0]=0),c[1]&&o[0]===-1/0&&(a[1]=0),o[0]===o[1]&&a[0]&&a[1]&&(r.value=o[0])}r.visual=uj.retrieveVisuals(e),t.push(r)}),this),zK(e,t),ti(t),s["k"](t,(function(t){var e=t.close,n=[["<","≤"][e[1]],[">","≥"][e[0]]];t.text=t.text||this.formatValueText(null!=t.value?t.value:t.interval,!1,n)}),this)}};function zK(t,e){var n=t.inverse;("vertical"===t.orient?!n:n)&&e.reverse()}var FK=RK,BK=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.doRender=function(){var t=this.group;t.removeAll();var e=this.visualMapModel,n=e.get("textGap"),r=e.textStyleModel,i=r.getFont(),o=r.getTextColor(),a=this._getItemAlign(),l=e.itemSize,u=this._getViewData(),c=u.endsText,d=s["O"](e.get("showLabel",!0),!c);c&&this._renderEndsText(t,c[0],l,d,a),s["k"](u.viewPieceList,(function(r){var u=r.piece,c=new wr;c.onclick=s["c"](this._onItemClick,this,u),this._enableHoverLink(c,r.indexInModelPieceList);var h=e.getRepresentValue(u);if(this._createItemSymbol(c,h,[0,0,l[0],l[1]]),d){var f=this.visualMapModel.getValueState(h);c.add(new ks({style:{x:"right"===a?-n:l[0]+n,y:l[1]/2,text:u.text,verticalAlign:"middle",align:a,font:i,fill:o,opacity:"outOfRange"===f?.5:1}}))}t.add(c)}),this),c&&this._renderEndsText(t,c[1],l,d,a),Nc(e.get("orient"),t,e.get("itemGap")),this.renderBackground(t),this.positionGroup(t)},e.prototype._enableHoverLink=function(t,e){var n=this;t.on("mouseover",(function(){return r("highlight")})).on("mouseout",(function(){return r("downplay")}));var r=function(t){var r=n.visualMapModel;r.option.hoverLink&&n.api.dispatchAction({type:t,batch:mK(r.findTargetDataIndices(e),r)})}},e.prototype._getItemAlign=function(){var t=this.visualMapModel,e=t.option;if("vertical"===e.orient)return pK(t,this.api,t.itemSize);var n=e.align;return n&&"auto"!==n||(n="left"),n},e.prototype._renderEndsText=function(t,e,n,r,i){if(e){var o=new wr,a=this.visualMapModel.textStyleModel;o.add(new ks({style:nu(a,{x:r?"right"===i?n[0]:0:n[0]/2,y:n[1]/2,verticalAlign:"middle",align:r?i:"center",text:e})})),t.add(o)}},e.prototype._getViewData=function(){var t=this.visualMapModel,e=s["H"](t.getPieceList(),(function(t,e){return{piece:t,indexInModelPieceList:e}})),n=t.get("text"),r=t.get("orient"),i=t.get("inverse");return("horizontal"===r?i:!i)?e.reverse():n&&(n=n.slice().reverse()),{viewPieceList:e,endsText:n}},e.prototype._createItemSymbol=function(t,e,n){t.add(ov(this.getControllerVisual(e,"symbol"),n[0],n[1],n[2],n[3],this.getControllerVisual(e,"color")))},e.prototype._onItemClick=function(t){var e=this.visualMapModel,n=e.option,r=n.selectedMode;if(r){var i=s["d"](n.selected),o=e.getSelectedMapKey(t);"single"===r||!0===r?(i[o]=!0,s["k"](i,(function(t,e){i[e]=e===o}))):i[o]=!i[o],this.api.dispatchAction({type:"selectDataRange",from:this.uid,visualMapId:this.visualMapModel.id,selected:i})}},e.type="visualMap.piecewise",e}(hK),WK=BK;function VK(t){t.registerComponentModel(FK),t.registerComponentView(WK),EK(t)}function GK(t){jb(NK),jb(VK)}var $K={label:{enabled:!0},decal:{show:!1}},UK=Ci(),qK={};function XK(t,e){var n=t.getModel("aria");if(n.get("enabled")){var r=s["d"]($K);s["I"](r.label,t.getLocaleModel().get("aria"),!1),s["I"](n.option,r,!1),i(),o()}function i(){var e=n.getModel("decal"),r=e.get("show");if(r){var i=s["f"]();t.eachSeries((function(t){if(!t.isColorBySeries()){var e=i.get(t.type);e||(e={},i.set(t.type,e)),UK(t).scope=e}})),t.eachRawSeries((function(e){if(!t.isSeriesFiltered(e))if(s["w"](e.enableAriaDecal))e.enableAriaDecal();else{var n=e.getData();if(e.isColorBySeries()){var r=Md(e.ecModel,e.name,qK,t.getSeriesCount()),i=n.getVisual("decal");n.setVisual("decal",c(i,r))}else{var o=e.getRawData(),a={},l=UK(e).scope;n.each((function(t){var e=n.getRawIndex(t);a[e]=t}));var u=o.count();o.each((function(t){var r=a[t],i=o.getName(t)||t+"",s=Md(e.ecModel,i,l,u),d=n.getItemVisual(r,"decal");n.setItemVisual(r,"decal",c(d,s))}))}}function c(t,e){var n=t?s["m"](s["m"]({},e),t):e;return n.dirty=!0,n}}))}}function o(){var r=t.getLocaleModel().get("aria"),i=n.getModel("label");if(i.option=s["i"](i.option,r),i.get("enabled")){var o=e.getZr().dom;if(i.get("description"))o.setAttribute("aria-label",i.get("description"));else{var c,d=t.getSeriesCount(),h=i.get(["data","maxCount"])||10,f=i.get(["series","maxCount"])||10,p=Math.min(d,f);if(!(d<1)){var m=l();if(m){var g=i.get(["general","withTitle"]);c=a(g,{title:m})}else c=i.get(["general","withoutTitle"]);var y=[],v=d>1?i.get(["series","multiple","prefix"]):i.get(["series","single","prefix"]);c+=a(v,{seriesCount:d}),t.eachSeries((function(t,e){if(e<p){var n=void 0,r=t.get("name"),o=r?"withName":"withoutName";n=d>1?i.get(["series","multiple",o]):i.get(["series","single",o]),n=a(n,{seriesId:t.seriesIndex,seriesName:t.get("name"),seriesType:u(t.subType)});var s=t.getData();if(s.count()>h){var l=i.get(["data","partialData"]);n+=a(l,{displayCnt:h})}else n+=i.get(["data","allData"]);for(var c=i.get(["data","separator","middle"]),f=i.get(["data","separator","end"]),m=[],g=0;g<s.count();g++)if(g<h){var v=s.getName(g),_=s.getValues(g),b=i.get(["data",v?"withName":"withoutName"]);m.push(a(b,{name:v,value:_.join(c)}))}n+=m.join(c)+f,y.push(n)}}));var _=i.getModel(["series","multiple","separator"]),b=_.get("middle"),x=_.get("end");c+=y.join(b)+x,o.setAttribute("aria-label",c)}}}}function a(t,e){if(!s["C"](t))return t;var n=t;return s["k"](e,(function(t,e){n=n.replace(new RegExp("\\{\\s*"+e+"\\s*\\}","g"),t)})),n}function l(){var e=t.get("title");return e&&e.length&&(e=e[0]),e&&e.text}function u(e){var n=t.getLocaleModel().get(["series","typeNames"]);return n[e]||n.chart}}function ZK(t){if(t&&t.aria){var e=t.aria;null!=e.show&&(e.enabled=e.show),e.label=e.label||{},s["k"](["description","general","series","data"],(function(t){null!=e[t]&&(e.label[t]=e[t])}))}}function JK(t){t.registerPreprocessor(ZK),t.registerVisual(t.PRIORITY.VISUAL.ARIA,XK)}var KK={value:"eq","<":"lt","<=":"lte",">":"gt",">=":"gte","=":"eq","!=":"ne","<>":"ne"},QK=function(){function t(t){var e=this._condVal=Object(s["C"])(t)?new RegExp(t):Object(s["B"])(t)?t:null;if(null==e){var n="";0,Cd(n)}}return t.prototype.evaluate=function(t){var e=typeof t;return Object(s["C"])(e)?this._condVal.test(t):!!Object(s["z"])(e)&&this._condVal.test(t+"")},t}(),tQ=function(){function t(){}return t.prototype.evaluate=function(){return this.value},t}(),eQ=function(){function t(){}return t.prototype.evaluate=function(){for(var t=this.children,e=0;e<t.length;e++)if(!t[e].evaluate())return!1;return!0},t}(),nQ=function(){function t(){}return t.prototype.evaluate=function(){for(var t=this.children,e=0;e<t.length;e++)if(t[e].evaluate())return!0;return!1},t}(),rQ=function(){function t(){}return t.prototype.evaluate=function(){return!this.child.evaluate()},t}(),iQ=function(){function t(){}return t.prototype.evaluate=function(){for(var t=!!this.valueParser,e=this.getValue,n=e(this.valueGetterParam),r=t?this.valueParser(n):null,i=0;i<this.subCondList.length;i++)if(!this.subCondList[i].evaluate(t?r:n))return!1;return!0},t}();function oQ(t,e){if(!0===t||!1===t){var n=new tQ;return n.value=t,n}var r="";return uQ(t)||Cd(r),t.and?aQ("and",t,e):t.or?aQ("or",t,e):t.not?sQ(t,e):lQ(t,e)}function aQ(t,e,n){var r=e[t],i="";Object(s["t"])(r)||Cd(i),r.length||Cd(i);var o="and"===t?new eQ:new nQ;return o.children=Object(s["H"])(r,(function(t){return oQ(t,n)})),o.children.length||Cd(i),o}function sQ(t,e){var n=t.not,r="";uQ(n)||Cd(r);var i=new rQ;return i.child=oQ(n,e),i.child||Cd(r),i}function lQ(t,e){for(var n="",r=e.prepareGetValue(t),i=[],o=Object(s["F"])(t),a=t.parser,l=a?uf(a):null,u=0;u<o.length;u++){var c=o[u];if("parser"!==c&&!e.valueGetterAttrMap.get(c)){var d=Object(s["q"])(KK,c)?KK[c]:c,h=t[c],f=l?l(h):h,p=pf(d,f)||"reg"===d&&new QK(f);p||Cd(n),i.push(p)}}i.length||Cd(n);var m=new iQ;return m.valueGetterParam=r,m.valueParser=l,m.getValue=e.getValue,m.subCondList=i,m}function uQ(t){return Object(s["A"])(t)&&!Object(s["u"])(t)}var cQ=function(){function t(t,e){this._cond=oQ(t,e)}return t.prototype.evaluate=function(){return this._cond.evaluate()},t}();function dQ(t,e){return new cQ(t,e)}var hQ={type:"echarts:filter",transform:function(t){for(var e,n=t.upstream,r=dQ(t.config,{valueGetterAttrMap:Object(s["f"])({dimension:!0}),prepareGetValue:function(t){var e="",r=t.dimension;Object(s["q"])(t,"dimension")||Cd(e);var i=n.getDimensionInfo(r);return i||Cd(e),{dimIdx:i.index}},getValue:function(t){return n.retrieveValueFromItem(e,t.dimIdx)}}),i=[],o=0,a=n.count();o<a;o++)e=n.getRawDataItem(o),r.evaluate()&&i.push(e);return{data:i}}};var fQ={type:"echarts:sort",transform:function(t){var e=t.upstream,n=t.config,r="",i=ui(n);i.length||Cd(r);var o=[];Object(s["k"])(i,(function(t){var n=t.dimension,i=t.order,a=t.parser,s=t.incomparable;if(null==n&&Cd(r),"asc"!==i&&"desc"!==i&&Cd(r),s&&"min"!==s&&"max"!==s){var l="";0,Cd(l)}if("asc"!==i&&"desc"!==i){var u="";0,Cd(u)}var c=e.getDimensionInfo(n);c||Cd(r);var d=a?uf(a):null;a&&!d&&Cd(r),o.push({dimIdx:c.index,parser:d,comparator:new hf(i,s)})}));var a=e.sourceFormat;a!==ed&&a!==nd&&Cd(r);for(var l=[],u=0,c=e.count();u<c;u++)l.push(e.getRawDataItem(u));return l.sort((function(t,n){for(var r=0;r<o.length;r++){var i=o[r],a=e.retrieveValueFromItem(t,i.dimIdx),s=e.retrieveValueFromItem(n,i.dimIdx);i.parser&&(a=i.parser(a),s=i.parser(s));var l=i.comparator.evaluate(a,s);if(0!==l)return l}return 0})),{data:l}}};function pQ(t){t.registerTransform(hQ),t.registerTransform(fQ)}var mQ=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.type="dataset",e}return o(e,t),e.prototype.init=function(e,n,r){t.prototype.init.call(this,e,n,r),this._sourceManager=new Hf(this),zf(this)},e.prototype.mergeOption=function(e,n){t.prototype.mergeOption.call(this,e,n),zf(this)},e.prototype.optionUpdated=function(){this._sourceManager.dirty()},e.prototype.getSourceManager=function(){return this._sourceManager},e.type="dataset",e.defaultOption={seriesLayoutBy:ad},e}(Xc),gQ=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.type="dataset",e}return o(e,t),e.type="dataset",e}(kp);function yQ(t){t.registerComponentModel(mQ),t.registerComponentView(gQ)}var vQ=va.CMD;function _Q(t,e){return Math.abs(t-e)<1e-5}function bQ(t){var e,n,r,i,o,a=t.data,s=t.len(),l=[],u=0,c=0,d=0,h=0;function f(t,n){e&&e.length>2&&l.push(e),e=[t,n]}function p(t,n,r,i){_Q(t,r)&&_Q(n,i)||e.push(t,n,r,i,r,i)}function m(t,n,r,i,o,a){var s=Math.abs(n-t),l=4*Math.tan(s/4)/3,u=n<t?-1:1,c=Math.cos(t),d=Math.sin(t),h=Math.cos(n),f=Math.sin(n),p=c*o+r,m=d*a+i,g=h*o+r,y=f*a+i,v=o*l*u,_=a*l*u;e.push(p-v*d,m+_*c,g+v*f,y-_*h,g,y)}for(var g=0;g<s;){var y=a[g++],v=1===g;switch(v&&(u=a[g],c=a[g+1],d=u,h=c,y!==vQ.L&&y!==vQ.C&&y!==vQ.Q||(e=[d,h])),y){case vQ.M:u=d=a[g++],c=h=a[g++],f(d,h);break;case vQ.L:n=a[g++],r=a[g++],p(u,c,n,r),u=n,c=r;break;case vQ.C:e.push(a[g++],a[g++],a[g++],a[g++],u=a[g++],c=a[g++]);break;case vQ.Q:n=a[g++],r=a[g++],i=a[g++],o=a[g++],e.push(u+2/3*(n-u),c+2/3*(r-c),i+2/3*(n-i),o+2/3*(r-o),i,o),u=i,c=o;break;case vQ.A:var _=a[g++],b=a[g++],x=a[g++],w=a[g++],M=a[g++],S=a[g++]+M;g+=1;var k=!a[g++];n=Math.cos(M)*x+_,r=Math.sin(M)*w+b,v?(d=n,h=r,f(d,h)):p(u,c,n,r),u=Math.cos(S)*x+_,c=Math.sin(S)*w+b;for(var L=(k?-1:1)*Math.PI/2,T=M;k?T>S:T<S;T+=L){var D=k?Math.max(T+L,S):Math.min(T+L,S);m(T,D,_,b,x,w)}break;case vQ.R:d=u=a[g++],h=c=a[g++],n=d+a[g++],r=h+a[g++],f(n,h),p(n,h,n,r),p(n,r,d,r),p(d,r,d,h),p(d,h,n,h);break;case vQ.Z:e&&p(u,c,d,h),u=d,c=h;break}}return e&&e.length>2&&l.push(e),l}function xQ(t,e,n,r,i,o,a,s,l,u){if(_Q(t,n)&&_Q(e,r)&&_Q(i,a)&&_Q(o,s))l.push(a,s);else{var c=2/u,d=c*c,h=a-t,f=s-e,p=Math.sqrt(h*h+f*f);h/=p,f/=p;var m=n-t,g=r-e,y=i-a,v=o-s,_=m*m+g*g,b=y*y+v*v;if(_<d&&b<d)l.push(a,s);else{var x=h*m+f*g,w=-h*y-f*v,M=_-x*x,S=b-w*w;if(M<d&&x>=0&&S<d&&w>=0)l.push(a,s);else{var k=[],L=[];Me(t,n,i,a,.5,k),Me(e,r,o,s,.5,L),xQ(k[0],L[0],k[1],L[1],k[2],L[2],k[3],L[3],l,u),xQ(k[4],L[4],k[5],L[5],k[6],L[6],k[7],L[7],l,u)}}}}function wQ(t,e){var n=bQ(t),r=[];e=e||1;for(var i=0;i<n.length;i++){var o=n[i],a=[],s=o[0],l=o[1];a.push(s,l);for(var u=2;u<o.length;){var c=o[u++],d=o[u++],h=o[u++],f=o[u++],p=o[u++],m=o[u++];xQ(s,l,c,d,h,f,p,m,a,e),s=p,l=m}r.push(a)}return r}function MQ(t,e,n){var r=t[e],i=t[1-e],o=Math.abs(r/i),a=Math.ceil(Math.sqrt(o*n)),s=Math.floor(n/a);0===s&&(s=1,a=n);for(var l=[],u=0;u<a;u++)l.push(s);var c=a*s,d=n-c;if(d>0)for(u=0;u<d;u++)l[u%a]+=1;return l}function SQ(t,e,n){for(var r=t.r0,i=t.r,o=t.startAngle,a=t.endAngle,s=Math.abs(a-o),l=s*i,u=i-r,c=l>Math.abs(u),d=MQ([l,u],c?0:1,e),h=(c?s:u)/d.length,f=0;f<d.length;f++)for(var p=(c?u:s)/d[f],m=0;m<d[f];m++){var g={};c?(g.startAngle=o+h*f,g.endAngle=o+h*(f+1),g.r0=r+p*m,g.r=r+p*(m+1)):(g.startAngle=o+p*m,g.endAngle=o+p*(m+1),g.r0=r+h*f,g.r=r+h*(f+1)),g.clockwise=t.clockwise,g.cx=t.cx,g.cy=t.cy,n.push(g)}}function kQ(t,e,n){for(var r=t.width,i=t.height,o=r>i,a=MQ([r,i],o?0:1,e),s=o?"width":"height",l=o?"height":"width",u=o?"x":"y",c=o?"y":"x",d=t[s]/a.length,h=0;h<a.length;h++)for(var f=t[l]/a[h],p=0;p<a[h];p++){var m={};m[u]=h*d,m[c]=p*f,m[s]=d,m[l]=f,m.x+=t.x,m.y+=t.y,n.push(m)}}function LQ(t,e,n,r){return t*r-n*e}function TQ(t,e,n,r,i,o,a,s){var l=n-t,u=r-e,c=a-i,d=s-o,h=LQ(c,d,l,u);if(Math.abs(h)<1e-6)return null;var f=t-i,p=e-o,m=LQ(f,p,c,d)/h;return m<0||m>1?null:new yt(m*l+t,m*u+e)}function DQ(t,e,n){var r=new yt;yt.sub(r,n,e),r.normalize();var i=new yt;yt.sub(i,t,e);var o=i.dot(r);return o}function OQ(t,e){var n=t[t.length-1];n&&n[0]===e[0]&&n[1]===e[1]||t.push(e)}function IQ(t,e,n){for(var r=t.length,i=[],o=0;o<r;o++){var a=t[o],s=t[(o+1)%r],l=TQ(a[0],a[1],s[0],s[1],e.x,e.y,n.x,n.y);l&&i.push({projPt:DQ(l,e,n),pt:l,idx:o})}if(i.length<2)return[{points:t},{points:t}];i.sort((function(t,e){return t.projPt-e.projPt}));var u=i[0],c=i[i.length-1];if(c.idx<u.idx){var d=u;u=c,c=d}var h=[u.pt.x,u.pt.y],f=[c.pt.x,c.pt.y],p=[h],m=[f];for(o=u.idx+1;o<=c.idx;o++)OQ(p,t[o].slice());OQ(p,f),OQ(p,h);for(o=c.idx+1;o<=u.idx+r;o++)OQ(m,t[o%r].slice());return OQ(m,h),OQ(m,f),[{points:p},{points:m}]}function CQ(t){var e=t.points,n=[],r=[];$o(e,n,r);var i=new Tt(n[0],n[1],r[0]-n[0],r[1]-n[1]),o=i.width,a=i.height,s=i.x,l=i.y,u=new yt,c=new yt;return o>a?(u.x=c.x=s+o/2,u.y=l,c.y=l+a):(u.y=c.y=l+a/2,u.x=s,c.x=s+o),IQ(e,u,c)}function AQ(t,e,n,r){if(1===n)r.push(e);else{var i=Math.floor(n/2),o=t(e);AQ(t,o[0],i,r),AQ(t,o[1],n-i,r)}return r}function YQ(t,e){for(var n=[],r=0;r<e;r++)n.push(Xp(t));return n}function jQ(t,e){e.setStyle(t.style),e.z=t.z,e.z2=t.z2,e.zlevel=t.zlevel}function PQ(t){for(var e=[],n=0;n<t.length;)e.push([t[n++],t[n++]]);return e}function EQ(t,e){var n,r=[],i=t.shape;switch(t.type){case"rect":kQ(i,e,r),n=ss;break;case"sector":SQ(i,e,r),n=_m;break;case"circle":SQ({r0:0,r:i.r,startAngle:0,endAngle:2*Math.PI,cx:i.cx,cy:i.cy},e,r),n=_m;break;default:var o=t.getComputedTransform(),a=o?Math.sqrt(Math.max(o[0]*o[0]+o[1]*o[1],o[2]*o[2]+o[3]*o[3])):1,l=Object(s["H"])(wQ(t.getUpdatedPathProxy(),a),(function(t){return PQ(t)})),u=l.length;if(0===u)AQ(CQ,{points:l[0]},e,r);else if(u===e)for(var c=0;c<u;c++)r.push({points:l[c]});else{var d=0,h=Object(s["H"])(l,(function(t){var e=[],n=[];$o(t,e,n);var r=(n[1]-e[1])*(n[0]-e[0]);return d+=r,{poly:t,area:r}}));h.sort((function(t,e){return e.area-t.area}));var f=e;for(c=0;c<u;c++){var p=h[c];if(f<=0)break;var m=c===u-1?f:Math.ceil(p.area/d*e);m<0||(AQ(CQ,{points:p.poly},m,r),f-=m)}}n=Tm;break}if(!n)return YQ(t,e);var g=[];for(c=0;c<r.length;c++){var y=new n;y.setShape(r[c]),jQ(t,y),g.push(y)}return g}function NQ(t,e){var n=t.length,r=e.length;if(n===r)return[t,e];for(var i=[],o=[],a=n<r?t:e,s=Math.min(n,r),l=Math.abs(r-n)/6,u=(s-2)/6,c=Math.ceil(l/u)+1,d=[a[0],a[1]],h=l,f=2;f<s;){var p=a[f-2],m=a[f-1],g=a[f++],y=a[f++],v=a[f++],_=a[f++],b=a[f++],x=a[f++];if(h<=0)d.push(g,y,v,_,b,x);else{for(var w=Math.min(h,c-1)+1,M=1;M<=w;M++){var S=M/w;Me(p,g,v,b,S,i),Me(m,y,_,x,S,o),p=i[3],m=o[3],d.push(i[1],o[1],i[2],o[2],p,m),g=i[5],y=o[5],v=i[6],_=o[6]}h-=w-1}}return a===t?[d,e]:[t,d]}function RQ(t,e){for(var n=t.length,r=t[n-2],i=t[n-1],o=[],a=0;a<e.length;)o[a++]=r,o[a++]=i;return o}function HQ(t,e){for(var n,r,i,o=[],a=[],s=0;s<Math.max(t.length,e.length);s++){var l=t[s],u=e[s],c=void 0,d=void 0;l?u?(n=NQ(l,u),c=n[0],d=n[1],r=c,i=d):(d=RQ(i||l,l),c=l):(c=RQ(r||u,u),d=u),o.push(c),a.push(d)}return[o,a]}function zQ(t){for(var e=0,n=0,r=0,i=t.length,o=0,a=i-2;o<i;a=o,o+=2){var s=t[a],l=t[a+1],u=t[o],c=t[o+1],d=s*c-u*l;e+=d,n+=(s+u)*d,r+=(l+c)*d}return 0===e?[t[0]||0,t[1]||0]:[n/e/3,r/e/3,e]}function FQ(t,e,n,r){for(var i=(t.length-2)/6,o=1/0,a=0,s=t.length,l=s-2,u=0;u<i;u++){for(var c=6*u,d=0,h=0;h<s;h+=2){var f=0===h?c:(c+h-2)%l+2,p=t[f]-n[0],m=t[f+1]-n[1],g=e[h]-r[0],y=e[h+1]-r[1],v=g-p,_=y-m;d+=v*v+_*_}d<o&&(o=d,a=u)}return a}function BQ(t){for(var e=[],n=t.length,r=0;r<n;r+=2)e[r]=t[n-r-2],e[r+1]=t[n-r-1];return e}function WQ(t,e,n,r){for(var i,o=[],a=0;a<t.length;a++){var s=t[a],l=e[a],u=zQ(s),c=zQ(l);null==i&&(i=u[2]<0!==c[2]<0);var d=[],h=[],f=0,p=1/0,m=[],g=s.length;i&&(s=BQ(s));for(var y=6*FQ(s,l,u,c),v=g-2,_=0;_<v;_+=2){var b=(y+_)%v+2;d[_+2]=s[b]-u[0],d[_+3]=s[b+1]-u[1]}if(d[0]=s[y]-u[0],d[1]=s[y+1]-u[1],n>0)for(var x=r/n,w=-r/2;w<=r/2;w+=x){var M=Math.sin(w),S=Math.cos(w),k=0;for(_=0;_<s.length;_+=2){var L=d[_],T=d[_+1],D=l[_]-c[0],O=l[_+1]-c[1],I=D*S-O*M,C=D*M+O*S;m[_]=I,m[_+1]=C;var A=I-L,Y=C-T;k+=A*A+Y*Y}if(k<p){p=k,f=w;for(var j=0;j<m.length;j++)h[j]=m[j]}}else for(var P=0;P<g;P+=2)h[P]=l[P]-c[0],h[P+1]=l[P+1]-c[1];o.push({from:d,to:h,fromCp:u,toCp:c,rotation:-f})}return o}function VQ(t){return t.__isCombineMorphing}var GQ="__mOriginal_";function $Q(t,e,n){var r=GQ+e,i=t[r]||t[e];t[r]||(t[r]=t[e]);var o=n.replace,a=n.after,s=n.before;t[e]=function(){var t,e=arguments;return s&&s.apply(this,e),t=o?o.apply(this,e):i.apply(this,e),a&&a.apply(this,e),t}}function UQ(t,e){var n=GQ+e;t[n]&&(t[e]=t[n],t[n]=null)}function qQ(t,e){for(var n=0;n<t.length;n++)for(var r=t[n],i=0;i<r.length;){var o=r[i],a=r[i+1];r[i++]=e[0]*o+e[2]*a+e[4],r[i++]=e[1]*o+e[3]*a+e[5]}}function XQ(t,e){var n=t.getUpdatedPathProxy(),r=e.getUpdatedPathProxy(),i=HQ(bQ(n),bQ(r)),o=i[0],a=i[1],s=t.getComputedTransform(),l=e.getComputedTransform();function u(){this.transform=null}s&&qQ(o,s),l&&qQ(a,l),$Q(e,"updateTransform",{replace:u}),e.transform=null;var c=WQ(o,a,10,Math.PI),d=[];$Q(e,"buildPath",{replace:function(t){for(var n=e.__morphT,r=1-n,i=[],o=0;o<c.length;o++){var a=c[o],s=a.from,l=a.to,u=a.rotation*n,h=a.fromCp,f=a.toCp,p=Math.sin(u),m=Math.cos(u);M(i,h,f,n);for(var g=0;g<s.length;g+=2){var y=s[g],v=s[g+1],_=l[g],b=l[g+1],x=y*r+_*n,w=v*r+b*n;d[g]=x*m-w*p+i[0],d[g+1]=x*p+w*m+i[1]}var S=d[0],k=d[1];t.moveTo(S,k);for(g=2;g<s.length;){_=d[g++],b=d[g++];var L=d[g++],T=d[g++],D=d[g++],O=d[g++];S===_&&k===b&&L===D&&T===O?t.lineTo(D,O):t.bezierCurveTo(_,b,L,T,D,O),S=D,k=O}}}})}function ZQ(t,e,n){if(!t||!e)return e;var r=n.done,i=n.during;function o(){UQ(e,"buildPath"),UQ(e,"updateTransform"),e.__morphT=-1,e.createPathProxy(),e.dirtyShape()}return XQ(t,e),e.__morphT=0,e.animateTo({__morphT:1},Object(s["i"])({during:function(t){e.dirtyShape(),i&&i(t)},done:function(){o(),r&&r()}},n)),e}function JQ(t,e,n,r,i,o){var a=16;t=i===n?0:Math.round(32767*(t-n)/(i-n)),e=o===r?0:Math.round(32767*(e-r)/(o-r));for(var s,l=0,u=(1<<a)/2;u>0;u/=2){var c=0,d=0;(t&u)>0&&(c=1),(e&u)>0&&(d=1),l+=u*u*(3*c^d),0===d&&(1===c&&(t=u-1-t,e=u-1-e),s=t,t=e,e=s)}return l}function KQ(t){var e=1/0,n=1/0,r=-1/0,i=-1/0,o=Object(s["H"])(t,(function(t){var o=t.getBoundingRect(),a=t.getComputedTransform(),s=o.x+o.width/2+(a?a[4]:0),l=o.y+o.height/2+(a?a[5]:0);return e=Math.min(s,e),n=Math.min(l,n),r=Math.max(s,r),i=Math.max(l,i),[s,l]})),a=Object(s["H"])(o,(function(o,a){return{cp:o,z:JQ(o[0],o[1],e,n,r,i),path:t[a]}}));return a.sort((function(t,e){return t.z-e.z})).map((function(t){return t.path}))}function QQ(t){return EQ(t.path,t.count)}function t0(){return{fromIndividuals:[],toIndividuals:[],count:0}}function e0(t,e,n){var r=[];function i(t){for(var e=0;e<t.length;e++){var n=t[e];VQ(n)?i(n.childrenRef()):n instanceof Va&&r.push(n)}}i(t);var o=r.length;if(!o)return t0();var a=n.dividePath||QQ,l=a({path:e,count:o});if(l.length!==o)return t0();r=KQ(r),l=KQ(l);for(var u=n.done,c=n.during,d=n.individualDelay,h=new Xn,f=0;f<o;f++){var p=r[f],m=l[f];m.parent=e,m.copyTransform(h),d||XQ(p,m)}function g(t){for(var e=0;e<l.length;e++)l[e].addSelfToZr(t)}function y(){e.__isCombineMorphing=!1,e.__morphT=-1,e.childrenRef=null,UQ(e,"addSelfToZr"),UQ(e,"removeSelfFromZr")}e.__isCombineMorphing=!0,e.childrenRef=function(){return l},$Q(e,"addSelfToZr",{after:function(t){g(t)}}),$Q(e,"removeSelfFromZr",{after:function(t){for(var e=0;e<l.length;e++)l[e].removeSelfFromZr(t)}});var v=l.length;if(d){var _=v,b=function(){_--,0===_&&(y(),u&&u())};for(f=0;f<v;f++){var x=d?Object(s["i"])({delay:(n.delay||0)+d(f,v,r[f],l[f]),done:b},n):n;ZQ(r[f],l[f],x)}}else e.__morphT=0,e.animateTo({__morphT:1},Object(s["i"])({during:function(t){for(var n=0;n<v;n++){var r=l[n];r.__morphT=e.__morphT,r.dirtyShape()}c&&c(t)},done:function(){y();for(var e=0;e<t.length;e++)UQ(t[e],"updateTransform");u&&u()}},n));return e.__zr&&g(e.__zr),{fromIndividuals:r,toIndividuals:l,count:v}}function n0(t,e,n){var r=e.length,i=[],o=n.dividePath||QQ;function a(t){for(var e=0;e<t.length;e++){var n=t[e];VQ(n)?a(n.childrenRef()):n instanceof Va&&i.push(n)}}if(VQ(t)){a(t.childrenRef());var l=i.length;if(l<r)for(var u=0,c=l;c<r;c++)i.push(Xp(i[u++%l]));i.length=r}else{i=o({path:t,count:r});var d=t.getComputedTransform();for(c=0;c<i.length;c++)i[c].setLocalTransform(d);if(i.length!==r)return t0()}i=KQ(i),e=KQ(e);var h=n.individualDelay;for(c=0;c<r;c++){var f=h?Object(s["i"])({delay:(n.delay||0)+h(c,r,i[c],e[c])},n):n;ZQ(i[c],e[c],f)}return{fromIndividuals:i,toIndividuals:e,count:e.length}}function r0(t){return Object(s["t"])(t[0])}function i0(t,e){for(var n=[],r=t.length,i=0;i<r;i++)n.push({one:t[i],many:[]});for(i=0;i<e.length;i++){var o=e[i].length,a=void 0;for(a=0;a<o;a++)n[a%r].many.push(e[i][a])}var s=0;for(i=r-1;i>=0;i--)if(!n[i].many.length){var l=n[s].many;if(l.length<=1){if(!s)return n;s=0}o=l.length;var u=Math.ceil(o/2);n[i].many=l.slice(u,o),n[s].many=l.slice(0,u),s++}return n}var o0={clone:function(t){for(var e=[],n=1-Math.pow(1-t.path.style.opacity,1/t.count),r=0;r<t.count;r++){var i=Xp(t.path);i.setStyle("opacity",n),e.push(i)}return e},split:null};function a0(t,e,n,r,i,o){if(t.length&&e.length){var a=Fl("update",r,i);if(a&&a.duration>0){var l,u,c=r.getModel("universalTransition").get("delay"),d=Object.assign({setToFinal:!0},a);r0(t)&&(l=t,u=e),r0(e)&&(l=e,u=t);for(var h=l?l===t:t.length>e.length,f=l?i0(u,l):i0(h?e:t,[h?t:e]),p=0,m=0;m<f.length;m++)p+=f[m].many.length;var g=0;for(m=0;m<f.length;m++)y(f[m],h,g,p),g+=f[m].many.length}}function y(t,e,r,i,a){var l=t.many,u=t.one;if(1!==l.length||a)for(var h=Object(s["i"])({dividePath:o0[n],individualDelay:c&&function(t,e,n,o){return c(t+r,i)}},d),f=e?e0(l,u,h):n0(u,l,h),p=f.fromIndividuals,m=f.toIndividuals,g=p.length,v=0;v<g;v++){x=c?Object(s["i"])({delay:c(v,g)},d):d;o(p[v],m[v],e?l[v]:t.one,e?t.one:l[v],x)}else{var _=e?l[0]:u,b=e?u:l[0];if(VQ(_))y({many:[_],one:b},!0,r,i,!0);else{var x=c?Object(s["i"])({delay:c(r,i)},d):d;ZQ(_,b,x),o(_,b,_,b,x)}}}}function s0(t){if(!t)return[];if(Object(s["t"])(t)){for(var e=[],n=0;n<t.length;n++)e.push(s0(t[n]));return e}var r=[];return t.traverse((function(t){t instanceof Va&&!t.disableMorphing&&!t.invisible&&!t.ignore&&r.push(t)})),r}var l0=1e4,u0=0,c0=1,d0=2,h0=Ci();function f0(t,e){for(var n=t.dimensions,r=0;r<n.length;r++){var i=t.getDimensionInfo(n[r]);if(i&&0===i.otherDims[e])return n[r]}}function p0(t,e,n){var r=t.getDimensionInfo(n),i=r&&r.ordinalMeta;if(r){var o=t.get(r.name,e);return i&&i.categories[o]||o+""}}function m0(t,e,n,r){var i=r?"itemChildGroupId":"itemGroupId",o=f0(t,i);if(o){var a=p0(t,e,o);return a}var s=t.getRawDataItem(e),l=r?"childGroupId":"groupId";return s&&s[l]?s[l]+"":r?void 0:n||t.getId(e)}function g0(t){var e=[];return Object(s["k"])(t,(function(t){var n=t.data,r=t.dataGroupId;if(!(n.count()>l0))for(var i=n.getIndices(),o=0;o<i.length;o++)e.push({data:n,groupId:m0(n,o,r,!1),childGroupId:m0(n,o,r,!0),divide:t.divide,dataIndex:o})})),e}function y0(t,e,n){t.traverse((function(t){t instanceof Va&&Vl(t,{style:{opacity:0}},e,{dataIndex:n,isFrom:!0})}))}function v0(t){if(t.parent){var e=t.getComputedTransform();t.setLocalTransform(e),t.parent.remove(t)}}function _0(t){t.stopAnimation(),t.isGroup&&t.traverse((function(t){t.stopAnimation()}))}function b0(t,e,n){var r=Fl("update",n,e);r&&t.traverse((function(t){if(t instanceof No){var e=Zl(t);e&&t.animateFrom({style:e},r)}}))}function x0(t,e){var n=t.length;if(n!==e.length)return!1;for(var r=0;r<n;r++){var i=t[r],o=e[r];if(i.data.getId(i.dataIndex)!==o.data.getId(o.dataIndex))return!1}return!0}function w0(t,e,n){var r=g0(t),i=g0(e);function o(t,e,n,r,i){(n||t)&&e.animateFrom({style:n&&n!==t?Object(s["m"])(Object(s["m"])({},n.style),t.style):t.style},i)}var a=!1,l=u0,u=Object(s["f"])(),c=Object(s["f"])();r.forEach((function(t){t.groupId&&u.set(t.groupId,!0),t.childGroupId&&c.set(t.childGroupId,!0)}));for(var d=0;d<i.length;d++){var h=i[d].groupId;if(c.get(h)){l=c0;break}var f=i[d].childGroupId;if(f&&u.get(f)){l=d0;break}}function p(t,e){return function(n){var r=n.data,i=n.dataIndex;return e?r.getId(i):t?l===c0?n.childGroupId:n.groupId:l===d0?n.childGroupId:n.groupId}}var m=x0(r,i),g={};if(!m)for(d=0;d<i.length;d++){var y=i[d],v=y.data.getItemGraphicEl(y.dataIndex);v&&(g[v.id]=!0)}function _(t,e){var n=r[e],s=i[t],l=s.data.hostModel,u=n.data.getItemGraphicEl(n.dataIndex),c=s.data.getItemGraphicEl(s.dataIndex);u!==c?u&&g[u.id]||c&&(_0(c),u?(_0(u),v0(u),a=!0,a0(s0(u),s0(c),s.divide,l,t,o)):y0(c,l,t)):c&&b0(c,s.dataIndex,l)}new yM(r,i,p(!0,m),p(!1,m),null,"multiple").update(_).updateManyToOne((function(t,e){var n=i[t],l=n.data,u=l.hostModel,c=l.getItemGraphicEl(n.dataIndex),d=Object(s["n"])(Object(s["H"])(e,(function(t){return r[t].data.getItemGraphicEl(r[t].dataIndex)})),(function(t){return t&&t!==c&&!g[t.id]}));c&&(_0(c),d.length?(Object(s["k"])(d,(function(t){_0(t),v0(t)})),a=!0,a0(s0(d),s0(c),n.divide,u,t,o)):y0(c,u,n.dataIndex))})).updateOneToMany((function(t,e){var n=r[e],l=n.data.getItemGraphicEl(n.dataIndex);if(!l||!g[l.id]){var u=Object(s["n"])(Object(s["H"])(t,(function(t){return i[t].data.getItemGraphicEl(i[t].dataIndex)})),(function(t){return t&&t!==l})),c=i[t[0]].data.hostModel;u.length&&(Object(s["k"])(u,(function(t){return _0(t)})),l?(_0(l),v0(l),a=!0,a0(s0(l),s0(u),n.divide,c,t[0],o)):Object(s["k"])(u,(function(e){return y0(e,c,t[0])})))}})).updateManyToMany((function(t,e){new yM(e,t,(function(t){return r[t].data.getId(r[t].dataIndex)}),(function(t){return i[t].data.getId(i[t].dataIndex)})).update((function(n,r){_(t[n],e[r])})).execute()})).execute(),a&&Object(s["k"])(e,(function(t){var e=t.data,r=e.hostModel,i=r&&n.getViewOfSeriesModel(r),o=Fl("update",r,0);i&&r.isAnimationEnabled()&&o&&o.duration>0&&i.group.traverse((function(t){t instanceof Va&&!t.animators.length&&t.animateFrom({style:{opacity:0}},o)}))}))}function M0(t){var e=t.getModel("universalTransition").get("seriesKey");return e||t.id}function S0(t){return Object(s["t"])(t)?t.sort().join(","):t}function k0(t){if(t.hostModel)return t.hostModel.getModel("universalTransition").get("divideShape")}function L0(t,e){var n=Object(s["f"])(),r=Object(s["f"])(),i=Object(s["f"])();return Object(s["k"])(t.oldSeries,(function(e,n){var o=t.oldDataGroupIds[n],a=t.oldData[n],l=M0(e),u=S0(l);r.set(u,{dataGroupId:o,data:a}),Object(s["t"])(l)&&Object(s["k"])(l,(function(t){i.set(t,{key:u,dataGroupId:o,data:a})}))})),Object(s["k"])(e.updatedSeries,(function(t){if(t.isUniversalTransitionEnabled()&&t.isAnimationEnabled()){var e=t.get("dataGroupId"),o=t.getData(),a=M0(t),l=S0(a),u=r.get(l);if(u)n.set(l,{oldSeries:[{dataGroupId:u.dataGroupId,divide:k0(u.data),data:u.data}],newSeries:[{dataGroupId:e,divide:k0(o),data:o}]});else if(Object(s["t"])(a)){0;var c=[];Object(s["k"])(a,(function(t){var e=r.get(t);e.data&&c.push({dataGroupId:e.dataGroupId,divide:k0(e.data),data:e.data})})),c.length&&n.set(l,{oldSeries:c,newSeries:[{dataGroupId:e,data:o,divide:k0(o)}]})}else{var d=i.get(a);if(d){var h=n.get(d.key);h||(h={oldSeries:[{dataGroupId:d.dataGroupId,data:d.data,divide:k0(d.data)}],newSeries:[]},n.set(d.key,h)),h.newSeries.push({dataGroupId:e,data:o,divide:k0(o)})}}}})),n}function T0(t,e){for(var n=0;n<t.length;n++){var r=null!=e.seriesIndex&&e.seriesIndex===t[n].seriesIndex||null!=e.seriesId&&e.seriesId===t[n].id;if(r)return n}}function D0(t,e,n,r){var i=[],o=[];Object(s["k"])(ui(t.from),(function(t){var n=T0(e.oldSeries,t);n>=0&&i.push({dataGroupId:e.oldDataGroupIds[n],data:e.oldData[n],divide:k0(e.oldData[n]),groupIdDim:t.dimension})})),Object(s["k"])(ui(t.to),(function(t){var r=T0(n.updatedSeries,t);if(r>=0){var i=n.updatedSeries[r].getData();o.push({dataGroupId:e.oldDataGroupIds[r],data:i,divide:k0(i),groupIdDim:t.dimension})}})),i.length>0&&o.length>0&&w0(i,o,r)}function O0(t){t.registerUpdateLifecycle("series:beforeupdate",(function(t,e,n){Object(s["k"])(ui(n.seriesTransition),(function(t){Object(s["k"])(ui(t.to),(function(t){for(var e=n.updatedSeries,r=0;r<e.length;r++)(null!=t.seriesIndex&&t.seriesIndex===e[r].seriesIndex||null!=t.seriesId&&t.seriesId===e[r].id)&&(e[r][fp]=!0)}))}))})),t.registerUpdateLifecycle("series:transition",(function(t,e,n){var r=h0(e);if(r.oldSeries&&n.updatedSeries&&n.optionChanged){var i=n.seriesTransition;if(i)Object(s["k"])(ui(i),(function(t){D0(t,r,n,e)}));else{var o=L0(r,n);Object(s["k"])(o.keys(),(function(t){var n=o.get(t);w0(n.oldSeries,n.newSeries,e)}))}Object(s["k"])(n.updatedSeries,(function(t){t[fp]&&(t[fp]=!1)}))}for(var a=t.getSeries(),l=r.oldSeries=[],u=r.oldDataGroupIds=[],c=r.oldData=[],d=0;d<a.length;d++){var h=a[d].getData();h.count()<l0&&(l.push(a[d]),u.push(a[d].get("dataGroupId")),c.push(h))}}))}jb([Yx]),jb([fM]),jb([fk,dL,jL,jO,eI,mA,fY,Fj,aE,gE,TE,AR,dH,OH,rz,dz,jz,Gz,xF,CF,ZF,LW]),jb(SV),jb(gG),jb(pA),jb(BG),jb(CR),jb(KG),jb(g$),jb(tq),jb(Eq),jb(MV),jb(gX),jb(_X),jb(BX),jb(pZ),jb(LZ),jb(HZ),jb(pJ),jb(XJ),jb(CJ),jb(qJ),jb(GK),jb(NK),jb(VK),jb(JK),jb(pQ),jb(yQ),jb(O0),jb(xx)},"31de":function(t,e,n){"use strict";var r=function(){var t=this,e=t._self._c;return e("ul",{staticClass:"ivu-dropdown-menu"},[t._t("default")],2)},i=[],o={name:"DropdownMenu"},a=o,s=n("2877"),l=Object(s["a"])(a,r,i,!1,null,null,null);e["a"]=l.exports},"32b3":function(t,e,n){var r=n("872a"),i=n("9638"),o=Object.prototype,a=o.hasOwnProperty;function s(t,e,n){var o=t[e];a.call(t,e)&&i(o,n)&&(void 0!==n||e in t)||r(t,e,n)}t.exports=s},"32f4":function(t,e,n){var r=n("2d7c"),i=n("d327"),o=Object.prototype,a=o.propertyIsEnumerable,s=Object.getOwnPropertySymbols,l=s?function(t){return null==t?[]:(t=Object(t),r(s(t),(function(e){return a.call(t,e)})))}:i;t.exports=l},"342f":function(t,e,n){"use strict";t.exports="undefined"!=typeof navigator&&String(navigator.userAgent)||""},"34ac":function(t,e,n){var r=n("9520"),i=n("1368"),o=n("1a8c"),a=n("dc57"),s=/[\\^$.*+?()[\]{}|]/g,l=/^\[object .+?Constructor\]$/,u=Function.prototype,c=Object.prototype,d=u.toString,h=c.hasOwnProperty,f=RegExp("^"+d.call(h).replace(s,"\\$&").replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g,"$1.*?")+"$");function p(t){if(!o(t)||i(t))return!1;var e=r(t)?f:l;return e.test(a(t))}t.exports=p},3511:function(t,e,n){"use strict";var r=TypeError,i=9007199254740991;t.exports=function(t){if(t>i)throw r("Maximum allowed index exceeded");return t}},3529:function(t,e,n){"use strict";var r=n("23e7"),i=n("c65b"),o=n("59ed"),a=n("f069"),s=n("e667"),l=n("2266"),u=n("5eed");r({target:"Promise",stat:!0,forced:u},{race:function(t){var e=this,n=a.f(e),r=n.reject,u=s((function(){var a=o(e.resolve);l(t,(function(t){i(a,e,t).then(n.resolve,r)}))}));return u.error&&r(u.value),n.promise}})},"35a1":function(t,e,n){"use strict";var r=n("f5df"),i=n("dc4a"),o=n("7234"),a=n("3f8c"),s=n("b622"),l=s("iterator");t.exports=function(t){if(!o(t))return i(t,l)||i(t,"@@iterator")||a[r(t)]}},3698:function(t,e){function n(t,e){return null==t?void 0:t[e]}t.exports=n},3729:function(t,e,n){var r=n("9e69"),i=n("00fd"),o=n("29f3"),a="[object Null]",s="[object Undefined]",l=r?r.toStringTag:void 0;function u(t){return null==t?void 0===t?s:a:l&&l in Object(t)?i(t):o(t)}t.exports=u},"37e8":function(t,e,n){"use strict";var r=n("83ab"),i=n("aed9"),o=n("9bf2"),a=n("825a"),s=n("fc6a"),l=n("df75");e.f=r&&!i?Object.defineProperties:function(t,e){a(t);var n,r=s(e),i=l(e),u=i.length,c=0;while(u>c)o.f(t,n=i[c++],r[n]);return t}},3818:function(t,e,n){var r=n("7e64"),i=n("8057"),o=n("32b3"),a=n("5b01"),s=n("0f0f"),l=n("e5383"),u=n("4359"),c=n("54eb"),d=n("1041"),h=n("a994"),f=n("1bac"),p=n("42a2"),m=n("c87c"),g=n("c2b6"),y=n("fa21"),v=n("6747"),_=n("0d24"),b=n("cc45"),x=n("1a8c"),w=n("d7ee"),M=n("ec69"),S=n("9934"),k=1,L=2,T=4,D="[object Arguments]",O="[object Array]",I="[object Boolean]",C="[object Date]",A="[object Error]",Y="[object Function]",j="[object GeneratorFunction]",P="[object Map]",E="[object Number]",N="[object Object]",R="[object RegExp]",H="[object Set]",z="[object String]",F="[object Symbol]",B="[object WeakMap]",W="[object ArrayBuffer]",V="[object DataView]",G="[object Float32Array]",$="[object Float64Array]",U="[object Int8Array]",q="[object Int16Array]",X="[object Int32Array]",Z="[object Uint8Array]",J="[object Uint8ClampedArray]",K="[object Uint16Array]",Q="[object Uint32Array]",tt={};function et(t,e,n,O,I,C){var A,P=e&k,E=e&L,R=e&T;if(n&&(A=I?n(t,O,I,C):n(t)),void 0!==A)return A;if(!x(t))return t;var H=v(t);if(H){if(A=m(t),!P)return u(t,A)}else{var z=p(t),F=z==Y||z==j;if(_(t))return l(t,P);if(z==N||z==D||F&&!I){if(A=E||F?{}:y(t),!P)return E?d(t,s(A,t)):c(t,a(A,t))}else{if(!tt[z])return I?t:{};A=g(t,z,P)}}C||(C=new r);var B=C.get(t);if(B)return B;C.set(t,A),w(t)?t.forEach((function(r){A.add(et(r,e,n,r,t,C))})):b(t)&&t.forEach((function(r,i){A.set(i,et(r,e,n,i,t,C))}));var W=R?E?f:h:E?S:M,V=H?void 0:W(t);return i(V||t,(function(r,i){V&&(i=r,r=t[i]),o(A,i,et(r,e,n,i,t,C))})),A}tt[D]=tt[O]=tt[W]=tt[V]=tt[I]=tt[C]=tt[G]=tt[$]=tt[U]=tt[q]=tt[X]=tt[P]=tt[E]=tt[N]=tt[R]=tt[H]=tt[z]=tt[F]=tt[Z]=tt[J]=tt[K]=tt[Q]=!0,tt[A]=tt[Y]=tt[B]=!1,t.exports=et},3852:function(t,e,n){var r=n("96f3"),i=n("e2c0");function o(t,e){return null!=t&&i(t,e,r)}t.exports=o},3886:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("en-ca",{months:"January_February_March_April_May_June_July_August_September_October_November_December".split("_"),monthsShort:"Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec".split("_"),weekdays:"Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday".split("_"),weekdaysShort:"Sun_Mon_Tue_Wed_Thu_Fri_Sat".split("_"),weekdaysMin:"Su_Mo_Tu_We_Th_Fr_Sa".split("_"),longDateFormat:{LT:"h:mm A",LTS:"h:mm:ss A",L:"YYYY-MM-DD",LL:"MMMM D, YYYY",LLL:"MMMM D, YYYY h:mm A",LLLL:"dddd, MMMM D, YYYY h:mm A"},calendar:{sameDay:"[Today at] LT",nextDay:"[Tomorrow at] LT",nextWeek:"dddd [at] LT",lastDay:"[Yesterday at] LT",lastWeek:"[Last] dddd [at] LT",sameElse:"L"},relativeTime:{future:"in %s",past:"%s ago",s:"a few seconds",ss:"%d seconds",m:"a minute",mm:"%d minutes",h:"an hour",hh:"%d hours",d:"a day",dd:"%d days",M:"a month",MM:"%d months",y:"a year",yy:"%d years"},dayOfMonthOrdinalParse:/\d{1,2}(st|nd|rd|th)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n}});return e}))},"39a6":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("en-gb",{months:"January_February_March_April_May_June_July_August_September_October_November_December".split("_"),monthsShort:"Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec".split("_"),weekdays:"Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday".split("_"),weekdaysShort:"Sun_Mon_Tue_Wed_Thu_Fri_Sat".split("_"),weekdaysMin:"Su_Mo_Tu_We_Th_Fr_Sa".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Today at] LT",nextDay:"[Tomorrow at] LT",nextWeek:"dddd [at] LT",lastDay:"[Yesterday at] LT",lastWeek:"[Last] dddd [at] LT",sameElse:"L"},relativeTime:{future:"in %s",past:"%s ago",s:"a few seconds",ss:"%d seconds",m:"a minute",mm:"%d minutes",h:"an hour",hh:"%d hours",d:"a day",dd:"%d days",M:"a month",MM:"%d months",y:"a year",yy:"%d years"},dayOfMonthOrdinalParse:/\d{1,2}(st|nd|rd|th)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n},week:{dow:1,doy:4}});return e}))},"39bd":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"१",2:"२",3:"३",4:"४",5:"५",6:"६",7:"७",8:"८",9:"९",0:"०"},n={"१":"1","२":"2","३":"3","४":"4","५":"5","६":"6","७":"7","८":"8","९":"9","०":"0"};function r(t,e,n,r){var i="";if(e)switch(n){case"s":i="काही सेकंद";break;case"ss":i="%d सेकंद";break;case"m":i="एक मिनिट";break;case"mm":i="%d मिनिटे";break;case"h":i="एक तास";break;case"hh":i="%d तास";break;case"d":i="एक दिवस";break;case"dd":i="%d दिवस";break;case"M":i="एक महिना";break;case"MM":i="%d महिने";break;case"y":i="एक वर्ष";break;case"yy":i="%d वर्षे";break}else switch(n){case"s":i="काही सेकंदां";break;case"ss":i="%d सेकंदां";break;case"m":i="एका मिनिटा";break;case"mm":i="%d मिनिटां";break;case"h":i="एका तासा";break;case"hh":i="%d तासां";break;case"d":i="एका दिवसा";break;case"dd":i="%d दिवसां";break;case"M":i="एका महिन्या";break;case"MM":i="%d महिन्यां";break;case"y":i="एका वर्षा";break;case"yy":i="%d वर्षां";break}return i.replace(/%d/i,t)}var i=t.defineLocale("mr",{months:"जानेवारी_फेब्रुवारी_मार्च_एप्रिल_मे_जून_जुलै_ऑगस्ट_सप्टेंबर_ऑक्टोबर_नोव्हेंबर_डिसेंबर".split("_"),monthsShort:"जाने._फेब्रु._मार्च._एप्रि._मे._जून._जुलै._ऑग._सप्टें._ऑक्टो._नोव्हें._डिसें.".split("_"),monthsParseExact:!0,weekdays:"रविवार_सोमवार_मंगळवार_बुधवार_गुरूवार_शुक्रवार_शनिवार".split("_"),weekdaysShort:"रवि_सोम_मंगळ_बुध_गुरू_शुक्र_शनि".split("_"),weekdaysMin:"र_सो_मं_बु_गु_शु_श".split("_"),longDateFormat:{LT:"A h:mm वाजता",LTS:"A h:mm:ss वाजता",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm वाजता",LLLL:"dddd, D MMMM YYYY, A h:mm वाजता"},calendar:{sameDay:"[आज] LT",nextDay:"[उद्या] LT",nextWeek:"dddd, LT",lastDay:"[काल] LT",lastWeek:"[मागील] dddd, LT",sameElse:"L"},relativeTime:{future:"%sमध्ये",past:"%sपूर्वी",s:r,ss:r,m:r,mm:r,h:r,hh:r,d:r,dd:r,M:r,MM:r,y:r,yy:r},preparse:function(t){return t.replace(/[१२३४५६७८९०]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/पहाटे|सकाळी|दुपारी|सायंकाळी|रात्री/,meridiemHour:function(t,e){return 12===t&&(t=0),"पहाटे"===e||"सकाळी"===e?t:"दुपारी"===e||"सायंकाळी"===e||"रात्री"===e?t>=12?t:t+12:void 0},meridiem:function(t,e,n){return t>=0&&t<6?"पहाटे":t<12?"सकाळी":t<17?"दुपारी":t<20?"सायंकाळी":"रात्री"},week:{dow:0,doy:6}});return i}))},"39ff":function(t,e,n){var r=n("0b07"),i=n("2b3e"),o=r(i,"WeakMap");t.exports=o},"3a0e":function(t,e,n){var r=n("86e1"),i=n("100e"),o=i(r);t.exports=o},"3a27":function(t,e,n){"use strict";n.d(e,"j",(function(){return o})),n.d(e,"h",(function(){return s})),n.d(e,"a",(function(){return l})),n.d(e,"i",(function(){return h})),n.d(e,"b",(function(){return p})),n.d(e,"f",(function(){return m})),n.d(e,"e",(function(){return g})),n.d(e,"g",(function(){return y})),n.d(e,"d",(function(){return v})),n.d(e,"c",(function(){return _})),n.d(e,"k",(function(){return b}));var r=n("2b0e");const i=r["a"].prototype.$isServer;function o(t,e){for(let n=0;n<e.length;n++)if(t===e[n])return!0;return!1}let a;function s(t){if(i)return 0;if(t||void 0===a){const t=document.createElement("div");t.style.width="100%",t.style.height="200px";const e=document.createElement("div"),n=e.style;n.position="absolute",n.top=0,n.left=0,n.pointerEvents="none",n.visibility="hidden",n.width="200px",n.height="150px",n.overflow="hidden",e.appendChild(t),document.body.appendChild(e);const r=t.offsetWidth;e.style.overflow="scroll";let i=t.offsetWidth;r===i&&(i=e.clientWidth),document.body.removeChild(e),a=r-i}return a}const l=!i&&(window.MutationObserver||window.WebKitMutationObserver||window.MozMutationObserver||!1),u=/([\:\-\_]+(.))/g,c=/^moz([A-Z])/;function d(t){return t.replace(u,(function(t,e,n,r){return r?n.toUpperCase():n})).replace(c,"Moz$1")}function h(t,e){if(!t||!e)return null;e=d(e),"float"===e&&(e="cssFloat");try{const n=document.defaultView.getComputedStyle(t,"");return t.style[e]||n?n[e]:null}catch(n){return t.style[e]}}function f(t){const e=Object.prototype.toString,n={"[object Boolean]":"boolean","[object Number]":"number","[object String]":"string","[object Function]":"function","[object Array]":"array","[object Date]":"date","[object RegExp]":"regExp","[object Undefined]":"undefined","[object Null]":"null","[object Object]":"object"};return n[e.call(t)]}function p(t){const e=f(t);let n;if("array"===e)n=[];else{if("object"!==e)return t;n={}}if("array"===e)for(let r=0;r<t.length;r++)n.push(p(t[r]));else if("object"===e)for(let r in t)n[r]=p(t[r]);return n}function m(t,e,n){n="string"===typeof e?[e]:e;let r=t.$parent,i=r.$options.name;while(r&&(!i||n.indexOf(i)<0))r=r.$parent,r&&(i=r.$options.name);return r}function g(t,e){const n=t.$children;let r=null;if(n.length)for(const i of n){const t=i.$options.name;if(t===e){r=i;break}if(r=g(i,e),r)break}return r}function y(t,e){return t.$children.reduce((t,n)=>{n.$options.name===e&&t.push(n);const r=y(n,e);return t.concat(r)},[])}function v(t,e,n=!0){let r=t.$parent.$children.filter(t=>t.$options.name===e),i=r.findIndex(e=>e._uid===t._uid);return n&&r.splice(i,1),r}const _={xs:"480px",sm:"576px",md:"768px",lg:"992px",xl:"1200px",xxl:"1600px"};function b(){if("undefined"!==typeof window){const t=t=>({media:t,matches:!1,on(){},off(){}});window.matchMedia=window.matchMedia||t}}},"3a34":function(t,e,n){"use strict";var r=n("83ab"),i=n("e8b5"),o=TypeError,a=Object.getOwnPropertyDescriptor,s=r&&!function(){if(void 0!==this)return!0;try{Object.defineProperty([],"length",{writable:!1}).length=1}catch(t){return t instanceof TypeError}}();t.exports=s?function(t,e){if(i(t)&&!a(t,"length").writable)throw new o("Cannot set read only .length");return t.length=e}:function(t,e){return t.length=e}},"3a39":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"१",2:"२",3:"३",4:"४",5:"५",6:"६",7:"७",8:"८",9:"९",0:"०"},n={"१":"1","२":"2","३":"3","४":"4","५":"5","६":"6","७":"7","८":"8","९":"9","०":"0"},r=t.defineLocale("ne",{months:"जनवरी_फेब्रुवरी_मार्च_अप्रिल_मई_जुन_जुलाई_अगष्ट_सेप्टेम्बर_अक्टोबर_नोभेम्बर_डिसेम्बर".split("_"),monthsShort:"जन._फेब्रु._मार्च_अप्रि._मई_जुन_जुलाई._अग._सेप्ट._अक्टो._नोभे._डिसे.".split("_"),monthsParseExact:!0,weekdays:"आइतबार_सोमबार_मङ्गलबार_बुधबार_बिहिबार_शुक्रबार_शनिबार".split("_"),weekdaysShort:"आइत._सोम._मङ्गल._बुध._बिहि._शुक्र._शनि.".split("_"),weekdaysMin:"आ._सो._मं._बु._बि._शु._श.".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"Aको h:mm बजे",LTS:"Aको h:mm:ss बजे",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, Aको h:mm बजे",LLLL:"dddd, D MMMM YYYY, Aको h:mm बजे"},preparse:function(t){return t.replace(/[१२३४५६७८९०]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/राति|बिहान|दिउँसो|साँझ/,meridiemHour:function(t,e){return 12===t&&(t=0),"राति"===e?t<4?t:t+12:"बिहान"===e?t:"दिउँसो"===e?t>=10?t:t+12:"साँझ"===e?t+12:void 0},meridiem:function(t,e,n){return t<3?"राति":t<12?"बिहान":t<16?"दिउँसो":t<20?"साँझ":"राति"},calendar:{sameDay:"[आज] LT",nextDay:"[भोलि] LT",nextWeek:"[आउँदो] dddd[,] LT",lastDay:"[हिजो] LT",lastWeek:"[गएको] dddd[,] LT",sameElse:"L"},relativeTime:{future:"%sमा",past:"%s अगाडि",s:"केही क्षण",ss:"%d सेकेण्ड",m:"एक मिनेट",mm:"%d मिनेट",h:"एक घण्टा",hh:"%d घण्टा",d:"एक दिन",dd:"%d दिन",M:"एक महिना",MM:"%d महिना",y:"एक बर्ष",yy:"%d बर्ष"},week:{dow:0,doy:6}});return r}))},"3a6c":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("zh-mo",{months:"一月_二月_三月_四月_五月_六月_七月_八月_九月_十月_十一月_十二月".split("_"),monthsShort:"1月_2月_3月_4月_5月_6月_7月_8月_9月_10月_11月_12月".split("_"),weekdays:"星期日_星期一_星期二_星期三_星期四_星期五_星期六".split("_"),weekdaysShort:"週日_週一_週二_週三_週四_週五_週六".split("_"),weekdaysMin:"日_一_二_三_四_五_六".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"YYYY年M月D日",LLL:"YYYY年M月D日 HH:mm",LLLL:"YYYY年M月D日dddd HH:mm",l:"D/M/YYYY",ll:"YYYY年M月D日",lll:"YYYY年M月D日 HH:mm",llll:"YYYY年M月D日dddd HH:mm"},meridiemParse:/凌晨|早上|上午|中午|下午|晚上/,meridiemHour:function(t,e){return 12===t&&(t=0),"凌晨"===e||"早上"===e||"上午"===e?t:"中午"===e?t>=11?t:t+12:"下午"===e||"晚上"===e?t+12:void 0},meridiem:function(t,e,n){var r=100*t+e;return r<600?"凌晨":r<900?"早上":r<1130?"上午":r<1230?"中午":r<1800?"下午":"晚上"},calendar:{sameDay:"[今天] LT",nextDay:"[明天] LT",nextWeek:"[下]dddd LT",lastDay:"[昨天] LT",lastWeek:"[上]dddd LT",sameElse:"L"},dayOfMonthOrdinalParse:/\d{1,2}(日|月|週)/,ordinal:function(t,e){switch(e){case"d":case"D":case"DDD":return t+"日";case"M":return t+"月";case"w":case"W":return t+"週";default:return t}},relativeTime:{future:"%s內",past:"%s前",s:"幾秒",ss:"%d 秒",m:"1 分鐘",mm:"%d 分鐘",h:"1 小時",hh:"%d 小時",d:"1 天",dd:"%d 天",M:"1 個月",MM:"%d 個月",y:"1 年",yy:"%d 年"}});return e}))},"3a9b":function(t,e,n){"use strict";var r=n("e330");t.exports=r({}.isPrototypeOf)},"3b1b":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={0:"-ум",1:"-ум",2:"-юм",3:"-юм",4:"-ум",5:"-ум",6:"-ум",7:"-ум",8:"-ум",9:"-ум",10:"-ум",12:"-ум",13:"-ум",20:"-ум",30:"-юм",40:"-ум",50:"-ум",60:"-ум",70:"-ум",80:"-ум",90:"-ум",100:"-ум"},n=t.defineLocale("tg",{months:{format:"январи_феврали_марти_апрели_майи_июни_июли_августи_сентябри_октябри_ноябри_декабри".split("_"),standalone:"январ_феврал_март_апрел_май_июн_июл_август_сентябр_октябр_ноябр_декабр".split("_")},monthsShort:"янв_фев_мар_апр_май_июн_июл_авг_сен_окт_ноя_дек".split("_"),weekdays:"якшанбе_душанбе_сешанбе_чоршанбе_панҷшанбе_ҷумъа_шанбе".split("_"),weekdaysShort:"яшб_дшб_сшб_чшб_пшб_ҷум_шнб".split("_"),weekdaysMin:"яш_дш_сш_чш_пш_ҷм_шб".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Имрӯз соати] LT",nextDay:"[Фардо соати] LT",lastDay:"[Дирӯз соати] LT",nextWeek:"dddd[и] [ҳафтаи оянда соати] LT",lastWeek:"dddd[и] [ҳафтаи гузашта соати] LT",sameElse:"L"},relativeTime:{future:"баъди %s",past:"%s пеш",s:"якчанд сония",m:"як дақиқа",mm:"%d дақиқа",h:"як соат",hh:"%d соат",d:"як рӯз",dd:"%d рӯз",M:"як моҳ",MM:"%d моҳ",y:"як сол",yy:"%d сол"},meridiemParse:/шаб|субҳ|рӯз|бегоҳ/,meridiemHour:function(t,e){return 12===t&&(t=0),"шаб"===e?t<4?t:t+12:"субҳ"===e?t:"рӯз"===e?t>=11?t:t+12:"бегоҳ"===e?t+12:void 0},meridiem:function(t,e,n){return t<4?"шаб":t<11?"субҳ":t<16?"рӯз":t<19?"бегоҳ":"шаб"},dayOfMonthOrdinalParse:/\d{1,2}-(ум|юм)/,ordinal:function(t){var n=t%10,r=t>=100?100:null;return t+(e[t]||e[n]||e[r])},week:{dow:1,doy:7}});return n}))},"3b4a":function(t,e,n){var r=n("0b07"),i=function(){try{var t=r(Object,"defineProperty");return t({},"",{}),t}catch(e){}}();t.exports=i},"3bb4":function(t,e,n){var r=n("08cc"),i=n("ec69");function o(t){var e=i(t),n=e.length;while(n--){var o=e[n],a=t[o];e[n]=[o,a,r(a)]}return e}t.exports=o},"3bbe":function(t,e,n){"use strict";var r=n("1787"),i=String,o=TypeError;t.exports=function(t){if(r(t))return t;throw new o("Can't set "+i(t)+" as a prototype")}},"3c0d":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={standalone:"leden_únor_březen_duben_květen_červen_červenec_srpen_září_říjen_listopad_prosinec".split("_"),format:"ledna_února_března_dubna_května_června_července_srpna_září_října_listopadu_prosince".split("_"),isFormat:/DD?[o.]?(\[[^\[\]]*\]|\s)+MMMM/},n="led_úno_bře_dub_kvě_čvn_čvc_srp_zář_říj_lis_pro".split("_"),r=[/^led/i,/^úno/i,/^bře/i,/^dub/i,/^kvě/i,/^(čvn|červen$|června)/i,/^(čvc|červenec|července)/i,/^srp/i,/^zář/i,/^říj/i,/^lis/i,/^pro/i],i=/^(leden|únor|březen|duben|květen|červenec|července|červen|června|srpen|září|říjen|listopad|prosinec|led|úno|bře|dub|kvě|čvn|čvc|srp|zář|říj|lis|pro)/i;function o(t){return t>1&&t<5&&1!==~~(t/10)}function a(t,e,n,r){var i=t+" ";switch(n){case"s":return e||r?"pár sekund":"pár sekundami";case"ss":return e||r?i+(o(t)?"sekundy":"sekund"):i+"sekundami";case"m":return e?"minuta":r?"minutu":"minutou";case"mm":return e||r?i+(o(t)?"minuty":"minut"):i+"minutami";case"h":return e?"hodina":r?"hodinu":"hodinou";case"hh":return e||r?i+(o(t)?"hodiny":"hodin"):i+"hodinami";case"d":return e||r?"den":"dnem";case"dd":return e||r?i+(o(t)?"dny":"dní"):i+"dny";case"M":return e||r?"měsíc":"měsícem";case"MM":return e||r?i+(o(t)?"měsíce":"měsíců"):i+"měsíci";case"y":return e||r?"rok":"rokem";case"yy":return e||r?i+(o(t)?"roky":"let"):i+"lety"}}var s=t.defineLocale("cs",{months:e,monthsShort:n,monthsRegex:i,monthsShortRegex:i,monthsStrictRegex:/^(leden|ledna|února|únor|březen|března|duben|dubna|květen|května|červenec|července|červen|června|srpen|srpna|září|říjen|října|listopadu|listopad|prosinec|prosince)/i,monthsShortStrictRegex:/^(led|úno|bře|dub|kvě|čvn|čvc|srp|zář|říj|lis|pro)/i,monthsParse:r,longMonthsParse:r,shortMonthsParse:r,weekdays:"neděle_pondělí_úterý_středa_čtvrtek_pátek_sobota".split("_"),weekdaysShort:"ne_po_út_st_čt_pá_so".split("_"),weekdaysMin:"ne_po_út_st_čt_pá_so".split("_"),longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY H:mm",LLLL:"dddd D. MMMM YYYY H:mm",l:"D. M. YYYY"},calendar:{sameDay:"[dnes v] LT",nextDay:"[zítra v] LT",nextWeek:function(){switch(this.day()){case 0:return"[v neděli v] LT";case 1:case 2:return"[v] dddd [v] LT";case 3:return"[ve středu v] LT";case 4:return"[ve čtvrtek v] LT";case 5:return"[v pátek v] LT";case 6:return"[v sobotu v] LT"}},lastDay:"[včera v] LT",lastWeek:function(){switch(this.day()){case 0:return"[minulou neděli v] LT";case 1:case 2:return"[minulé] dddd [v] LT";case 3:return"[minulou středu v] LT";case 4:case 5:return"[minulý] dddd [v] LT";case 6:return"[minulou sobotu v] LT"}},sameElse:"L"},relativeTime:{future:"za %s",past:"před %s",s:a,ss:a,m:a,mm:a,h:a,hh:a,d:a,dd:a,M:a,MM:a,y:a,yy:a},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return s}))},"3ca3":function(t,e,n){"use strict";var r=n("6547").charAt,i=n("577e"),o=n("69f3"),a=n("c6d2"),s=n("4754"),l="String Iterator",u=o.set,c=o.getterFor(l);a(String,"String",(function(t){u(this,{type:l,string:i(t),index:0})}),(function(){var t,e=c(this),n=e.string,i=e.index;return i>=n.length?s(void 0,!0):(t=r(n,i),e.index+=t.length,s(t,!1))}))},"3cfe":function(t,e,n){var r=n("4bb5");function i(t,e){return null==t||r(t,e)}t.exports=i},"3de5":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"௧",2:"௨",3:"௩",4:"௪",5:"௫",6:"௬",7:"௭",8:"௮",9:"௯",0:"௦"},n={"௧":"1","௨":"2","௩":"3","௪":"4","௫":"5","௬":"6","௭":"7","௮":"8","௯":"9","௦":"0"},r=t.defineLocale("ta",{months:"ஜனவரி_பிப்ரவரி_மார்ச்_ஏப்ரல்_மே_ஜூன்_ஜூலை_ஆகஸ்ட்_செப்டெம்பர்_அக்டோபர்_நவம்பர்_டிசம்பர்".split("_"),monthsShort:"ஜனவரி_பிப்ரவரி_மார்ச்_ஏப்ரல்_மே_ஜூன்_ஜூலை_ஆகஸ்ட்_செப்டெம்பர்_அக்டோபர்_நவம்பர்_டிசம்பர்".split("_"),weekdays:"ஞாயிற்றுக்கிழமை_திங்கட்கிழமை_செவ்வாய்கிழமை_புதன்கிழமை_வியாழக்கிழமை_வெள்ளிக்கிழமை_சனிக்கிழமை".split("_"),weekdaysShort:"ஞாயிறு_திங்கள்_செவ்வாய்_புதன்_வியாழன்_வெள்ளி_சனி".split("_"),weekdaysMin:"ஞா_தி_செ_பு_வி_வெ_ச".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, HH:mm",LLLL:"dddd, D MMMM YYYY, HH:mm"},calendar:{sameDay:"[இன்று] LT",nextDay:"[நாளை] LT",nextWeek:"dddd, LT",lastDay:"[நேற்று] LT",lastWeek:"[கடந்த வாரம்] dddd, LT",sameElse:"L"},relativeTime:{future:"%s இல்",past:"%s முன்",s:"ஒரு சில விநாடிகள்",ss:"%d விநாடிகள்",m:"ஒரு நிமிடம்",mm:"%d நிமிடங்கள்",h:"ஒரு மணி நேரம்",hh:"%d மணி நேரம்",d:"ஒரு நாள்",dd:"%d நாட்கள்",M:"ஒரு மாதம்",MM:"%d மாதங்கள்",y:"ஒரு வருடம்",yy:"%d ஆண்டுகள்"},dayOfMonthOrdinalParse:/\d{1,2}வது/,ordinal:function(t){return t+"வது"},preparse:function(t){return t.replace(/[௧௨௩௪௫௬௭௮௯௦]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/யாமம்|வைகறை|காலை|நண்பகல்|எற்பாடு|மாலை/,meridiem:function(t,e,n){return t<2?" யாமம்":t<6?" வைகறை":t<10?" காலை":t<14?" நண்பகல்":t<18?" எற்பாடு":t<22?" மாலை":" யாமம்"},meridiemHour:function(t,e){return 12===t&&(t=0),"யாமம்"===e?t<2?t:t+12:"வைகறை"===e||"காலை"===e||"நண்பகல்"===e&&t>=10?t:t+12},week:{dow:0,doy:6}});return r}))},"3e92":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"೧",2:"೨",3:"೩",4:"೪",5:"೫",6:"೬",7:"೭",8:"೮",9:"೯",0:"೦"},n={"೧":"1","೨":"2","೩":"3","೪":"4","೫":"5","೬":"6","೭":"7","೮":"8","೯":"9","೦":"0"},r=t.defineLocale("kn",{months:"ಜನವರಿ_ಫೆಬ್ರವರಿ_ಮಾರ್ಚ್_ಏಪ್ರಿಲ್_ಮೇ_ಜೂನ್_ಜುಲೈ_ಆಗಸ್ಟ್_ಸೆಪ್ಟೆಂಬರ್_ಅಕ್ಟೋಬರ್_ನವೆಂಬರ್_ಡಿಸೆಂಬರ್".split("_"),monthsShort:"ಜನ_ಫೆಬ್ರ_ಮಾರ್ಚ್_ಏಪ್ರಿಲ್_ಮೇ_ಜೂನ್_ಜುಲೈ_ಆಗಸ್ಟ್_ಸೆಪ್ಟೆಂ_ಅಕ್ಟೋ_ನವೆಂ_ಡಿಸೆಂ".split("_"),monthsParseExact:!0,weekdays:"ಭಾನುವಾರ_ಸೋಮವಾರ_ಮಂಗಳವಾರ_ಬುಧವಾರ_ಗುರುವಾರ_ಶುಕ್ರವಾರ_ಶನಿವಾರ".split("_"),weekdaysShort:"ಭಾನು_ಸೋಮ_ಮಂಗಳ_ಬುಧ_ಗುರು_ಶುಕ್ರ_ಶನಿ".split("_"),weekdaysMin:"ಭಾ_ಸೋ_ಮಂ_ಬು_ಗು_ಶು_ಶ".split("_"),longDateFormat:{LT:"A h:mm",LTS:"A h:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm",LLLL:"dddd, D MMMM YYYY, A h:mm"},calendar:{sameDay:"[ಇಂದು] LT",nextDay:"[ನಾಳೆ] LT",nextWeek:"dddd, LT",lastDay:"[ನಿನ್ನೆ] LT",lastWeek:"[ಕೊನೆಯ] dddd, LT",sameElse:"L"},relativeTime:{future:"%s ನಂತರ",past:"%s ಹಿಂದೆ",s:"ಕೆಲವು ಕ್ಷಣಗಳು",ss:"%d ಸೆಕೆಂಡುಗಳು",m:"ಒಂದು ನಿಮಿಷ",mm:"%d ನಿಮಿಷ",h:"ಒಂದು ಗಂಟೆ",hh:"%d ಗಂಟೆ",d:"ಒಂದು ದಿನ",dd:"%d ದಿನ",M:"ಒಂದು ತಿಂಗಳು",MM:"%d ತಿಂಗಳು",y:"ಒಂದು ವರ್ಷ",yy:"%d ವರ್ಷ"},preparse:function(t){return t.replace(/[೧೨೩೪೫೬೭೮೯೦]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/ರಾತ್ರಿ|ಬೆಳಿಗ್ಗೆ|ಮಧ್ಯಾಹ್ನ|ಸಂಜೆ/,meridiemHour:function(t,e){return 12===t&&(t=0),"ರಾತ್ರಿ"===e?t<4?t:t+12:"ಬೆಳಿಗ್ಗೆ"===e?t:"ಮಧ್ಯಾಹ್ನ"===e?t>=10?t:t+12:"ಸಂಜೆ"===e?t+12:void 0},meridiem:function(t,e,n){return t<4?"ರಾತ್ರಿ":t<10?"ಬೆಳಿಗ್ಗೆ":t<17?"ಮಧ್ಯಾಹ್ನ":t<20?"ಸಂಜೆ":"ರಾತ್ರಿ"},dayOfMonthOrdinalParse:/\d{1,2}(ನೇ)/,ordinal:function(t){return t+"ನೇ"},week:{dow:0,doy:6}});return r}))},"3eea":function(t,e,n){var r=n("7948"),i=n("3818"),o=n("4bb5"),a=n("e2e4"),s=n("8eeb"),l=n("e0e7"),u=n("c6cf"),c=n("1bac"),d=1,h=2,f=4,p=u((function(t,e){var n={};if(null==t)return n;var u=!1;e=r(e,(function(e){return e=a(e,t),u||(u=e.length>1),e})),s(t,c(t),n),u&&(n=i(n,d|h|f,l));var p=e.length;while(p--)o(n,e[p]);return n}));t.exports=p},"3f84":function(t,e,n){var r=n("85e3"),i=n("100e"),o=n("e031"),a=n("2411"),s=i((function(t){return t.push(void 0,o),r(a,void 0,t)}));t.exports=s},"3f8c":function(t,e,n){"use strict";t.exports={}},"3ff1":function(t,e,n){var r=n("266a"),i=n("ec69");function o(t){return null==t?[]:r(t,i(t))}t.exports=o},"408a":function(t,e,n){"use strict";var r=n("e330");t.exports=r(1..valueOf)},"40d5":function(t,e,n){"use strict";var r=n("d039");t.exports=!r((function(){var t=function(){}.bind();return"function"!=typeof t||t.hasOwnProperty("prototype")}))},"41c3":function(t,e,n){var r=n("1a8c"),i=n("eac5"),o=n("ec8c"),a=Object.prototype,s=a.hasOwnProperty;function l(t){if(!r(t))return o(t);var e=i(t),n=[];for(var a in t)("constructor"!=a||!e&&s.call(t,a))&&n.push(a);return n}t.exports=l},"41ef":function(t,e,n){"use strict";n.d(e,"h",(function(){return v})),n.d(e,"c",(function(){return x})),n.d(e,"a",(function(){return w})),n.d(e,"b",(function(){return M})),n.d(e,"g",(function(){return S})),n.d(e,"f",(function(){return k})),n.d(e,"i",(function(){return L})),n.d(e,"e",(function(){return T})),n.d(e,"d",(function(){return O}));var r=n("d51b"),i=n("6d8b"),o={transparent:[0,0,0,0],aliceblue:[240,248,255,1],antiquewhite:[250,235,215,1],aqua:[0,255,255,1],aquamarine:[127,255,212,1],azure:[240,255,255,1],beige:[245,245,220,1],bisque:[255,228,196,1],black:[0,0,0,1],blanchedalmond:[255,235,205,1],blue:[0,0,255,1],blueviolet:[138,43,226,1],brown:[165,42,42,1],burlywood:[222,184,135,1],cadetblue:[95,158,160,1],chartreuse:[127,255,0,1],chocolate:[210,105,30,1],coral:[255,127,80,1],cornflowerblue:[100,149,237,1],cornsilk:[255,248,220,1],crimson:[220,20,60,1],cyan:[0,255,255,1],darkblue:[0,0,139,1],darkcyan:[0,139,139,1],darkgoldenrod:[184,134,11,1],darkgray:[169,169,169,1],darkgreen:[0,100,0,1],darkgrey:[169,169,169,1],darkkhaki:[189,183,107,1],darkmagenta:[139,0,139,1],darkolivegreen:[85,107,47,1],darkorange:[255,140,0,1],darkorchid:[153,50,204,1],darkred:[139,0,0,1],darksalmon:[233,150,122,1],darkseagreen:[143,188,143,1],darkslateblue:[72,61,139,1],darkslategray:[47,79,79,1],darkslategrey:[47,79,79,1],darkturquoise:[0,206,209,1],darkviolet:[148,0,211,1],deeppink:[255,20,147,1],deepskyblue:[0,191,255,1],dimgray:[105,105,105,1],dimgrey:[105,105,105,1],dodgerblue:[30,144,255,1],firebrick:[178,34,34,1],floralwhite:[255,250,240,1],forestgreen:[34,139,34,1],fuchsia:[255,0,255,1],gainsboro:[220,220,220,1],ghostwhite:[248,248,255,1],gold:[255,215,0,1],goldenrod:[218,165,32,1],gray:[128,128,128,1],green:[0,128,0,1],greenyellow:[173,255,47,1],grey:[128,128,128,1],honeydew:[240,255,240,1],hotpink:[255,105,180,1],indianred:[205,92,92,1],indigo:[75,0,130,1],ivory:[255,255,240,1],khaki:[240,230,140,1],lavender:[230,230,250,1],lavenderblush:[255,240,245,1],lawngreen:[124,252,0,1],lemonchiffon:[255,250,205,1],lightblue:[173,216,230,1],lightcoral:[240,128,128,1],lightcyan:[224,255,255,1],lightgoldenrodyellow:[250,250,210,1],lightgray:[211,211,211,1],lightgreen:[144,238,144,1],lightgrey:[211,211,211,1],lightpink:[255,182,193,1],lightsalmon:[255,160,122,1],lightseagreen:[32,178,170,1],lightskyblue:[135,206,250,1],lightslategray:[119,136,153,1],lightslategrey:[119,136,153,1],lightsteelblue:[176,196,222,1],lightyellow:[255,255,224,1],lime:[0,255,0,1],limegreen:[50,205,50,1],linen:[250,240,230,1],magenta:[255,0,255,1],maroon:[128,0,0,1],mediumaquamarine:[102,205,170,1],mediumblue:[0,0,205,1],mediumorchid:[186,85,211,1],mediumpurple:[147,112,219,1],mediumseagreen:[60,179,113,1],mediumslateblue:[123,104,238,1],mediumspringgreen:[0,250,154,1],mediumturquoise:[72,209,204,1],mediumvioletred:[199,21,133,1],midnightblue:[25,25,112,1],mintcream:[245,255,250,1],mistyrose:[255,228,225,1],moccasin:[255,228,181,1],navajowhite:[255,222,173,1],navy:[0,0,128,1],oldlace:[253,245,230,1],olive:[128,128,0,1],olivedrab:[107,142,35,1],orange:[255,165,0,1],orangered:[255,69,0,1],orchid:[218,112,214,1],palegoldenrod:[238,232,170,1],palegreen:[152,251,152,1],paleturquoise:[175,238,238,1],palevioletred:[219,112,147,1],papayawhip:[255,239,213,1],peachpuff:[255,218,185,1],peru:[205,133,63,1],pink:[255,192,203,1],plum:[221,160,221,1],powderblue:[176,224,230,1],purple:[128,0,128,1],red:[255,0,0,1],rosybrown:[188,143,143,1],royalblue:[65,105,225,1],saddlebrown:[139,69,19,1],salmon:[250,128,114,1],sandybrown:[244,164,96,1],seagreen:[46,139,87,1],seashell:[255,245,238,1],sienna:[160,82,45,1],silver:[192,192,192,1],skyblue:[135,206,235,1],slateblue:[106,90,205,1],slategray:[112,128,144,1],slategrey:[112,128,144,1],snow:[255,250,250,1],springgreen:[0,255,127,1],steelblue:[70,130,180,1],tan:[210,180,140,1],teal:[0,128,128,1],thistle:[216,191,216,1],tomato:[255,99,71,1],turquoise:[64,224,208,1],violet:[238,130,238,1],wheat:[245,222,179,1],white:[255,255,255,1],whitesmoke:[245,245,245,1],yellow:[255,255,0,1],yellowgreen:[154,205,50,1]};function a(t){return t=Math.round(t),t<0?0:t>255?255:t}function s(t){return t=Math.round(t),t<0?0:t>360?360:t}function l(t){return t<0?0:t>1?1:t}function u(t){var e=t;return e.length&&"%"===e.charAt(e.length-1)?a(parseFloat(e)/100*255):a(parseInt(e,10))}function c(t){var e=t;return e.length&&"%"===e.charAt(e.length-1)?l(parseFloat(e)/100):l(parseFloat(e))}function d(t,e,n){return n<0?n+=1:n>1&&(n-=1),6*n<1?t+(e-t)*n*6:2*n<1?e:3*n<2?t+(e-t)*(2/3-n)*6:t}function h(t,e,n){return t+(e-t)*n}function f(t,e,n,r,i){return t[0]=e,t[1]=n,t[2]=r,t[3]=i,t}function p(t,e){return t[0]=e[0],t[1]=e[1],t[2]=e[2],t[3]=e[3],t}var m=new r["a"](20),g=null;function y(t,e){g&&p(g,e),g=m.put(t,g||e.slice())}function v(t,e){if(t){e=e||[];var n=m.get(t);if(n)return p(e,n);t+="";var r=t.replace(/ /g,"").toLowerCase();if(r in o)return p(e,o[r]),y(t,e),e;var i=r.length;if("#"!==r.charAt(0)){var a=r.indexOf("("),s=r.indexOf(")");if(-1!==a&&s+1===i){var l=r.substr(0,a),d=r.substr(a+1,s-(a+1)).split(","),h=1;switch(l){case"rgba":if(4!==d.length)return 3===d.length?f(e,+d[0],+d[1],+d[2],1):f(e,0,0,0,1);h=c(d.pop());case"rgb":return d.length>=3?(f(e,u(d[0]),u(d[1]),u(d[2]),3===d.length?h:c(d[3])),y(t,e),e):void f(e,0,0,0,1);case"hsla":return 4!==d.length?void f(e,0,0,0,1):(d[3]=c(d[3]),_(d,e),y(t,e),e);case"hsl":return 3!==d.length?void f(e,0,0,0,1):(_(d,e),y(t,e),e);default:return}}f(e,0,0,0,1)}else{if(4===i||5===i){var g=parseInt(r.slice(1,4),16);return g>=0&&g<=4095?(f(e,(3840&g)>>4|(3840&g)>>8,240&g|(240&g)>>4,15&g|(15&g)<<4,5===i?parseInt(r.slice(4),16)/15:1),y(t,e),e):void f(e,0,0,0,1)}if(7===i||9===i){g=parseInt(r.slice(1,7),16);return g>=0&&g<=16777215?(f(e,(16711680&g)>>16,(65280&g)>>8,255&g,9===i?parseInt(r.slice(7),16)/255:1),y(t,e),e):void f(e,0,0,0,1)}}}}function _(t,e){var n=(parseFloat(t[0])%360+360)%360/360,r=c(t[1]),i=c(t[2]),o=i<=.5?i*(r+1):i+r-i*r,s=2*i-o;return e=e||[],f(e,a(255*d(s,o,n+1/3)),a(255*d(s,o,n)),a(255*d(s,o,n-1/3)),1),4===t.length&&(e[3]=t[3]),e}function b(t){if(t){var e,n,r=t[0]/255,i=t[1]/255,o=t[2]/255,a=Math.min(r,i,o),s=Math.max(r,i,o),l=s-a,u=(s+a)/2;if(0===l)e=0,n=0;else{n=u<.5?l/(s+a):l/(2-s-a);var c=((s-r)/6+l/2)/l,d=((s-i)/6+l/2)/l,h=((s-o)/6+l/2)/l;r===s?e=h-d:i===s?e=1/3+c-h:o===s&&(e=2/3+d-c),e<0&&(e+=1),e>1&&(e-=1)}var f=[360*e,n,u];return null!=t[3]&&f.push(t[3]),f}}function x(t,e){var n=v(t);if(n){for(var r=0;r<3;r++)n[r]=e<0?n[r]*(1-e)|0:(255-n[r])*e+n[r]|0,n[r]>255?n[r]=255:n[r]<0&&(n[r]=0);return L(n,4===n.length?"rgba":"rgb")}}function w(t,e,n){if(e&&e.length&&t>=0&&t<=1){n=n||[];var r=t*(e.length-1),i=Math.floor(r),o=Math.ceil(r),s=e[i],u=e[o],c=r-i;return n[0]=a(h(s[0],u[0],c)),n[1]=a(h(s[1],u[1],c)),n[2]=a(h(s[2],u[2],c)),n[3]=l(h(s[3],u[3],c)),n}}function M(t,e,n){if(e&&e.length&&t>=0&&t<=1){var r=t*(e.length-1),i=Math.floor(r),o=Math.ceil(r),s=v(e[i]),u=v(e[o]),c=r-i,d=L([a(h(s[0],u[0],c)),a(h(s[1],u[1],c)),a(h(s[2],u[2],c)),l(h(s[3],u[3],c))],"rgba");return n?{color:d,leftIndex:i,rightIndex:o,value:r}:d}}function S(t,e,n,r){var i=v(t);if(t)return i=b(i),null!=e&&(i[0]=s(e)),null!=n&&(i[1]=c(n)),null!=r&&(i[2]=c(r)),L(_(i),"rgba")}function k(t,e){var n=v(t);if(n&&null!=e)return n[3]=l(e),L(n,"rgba")}function L(t,e){if(t&&t.length){var n=t[0]+","+t[1]+","+t[2];return"rgba"!==e&&"hsva"!==e&&"hsla"!==e||(n+=","+t[3]),e+"("+n+")"}}function T(t,e){var n=v(t);return n?(.299*n[0]+.587*n[1]+.114*n[2])*n[3]/255+(1-n[3])*e:0}var D=new r["a"](100);function O(t){if(Object(i["C"])(t)){var e=D.get(t);return e||(e=x(t,-.1),D.put(t,e)),e}if(Object(i["x"])(t)){var n=Object(i["m"])({},t);return n.colorStops=Object(i["H"])(t.colorStops,(function(t){return{offset:t.offset,color:x(t.color,-.1)}})),n}return t}},"423e":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ar-kw",{months:"يناير_فبراير_مارس_أبريل_ماي_يونيو_يوليوز_غشت_شتنبر_أكتوبر_نونبر_دجنبر".split("_"),monthsShort:"يناير_فبراير_مارس_أبريل_ماي_يونيو_يوليوز_غشت_شتنبر_أكتوبر_نونبر_دجنبر".split("_"),weekdays:"الأحد_الإتنين_الثلاثاء_الأربعاء_الخميس_الجمعة_السبت".split("_"),weekdaysShort:"احد_اتنين_ثلاثاء_اربعاء_خميس_جمعة_سبت".split("_"),weekdaysMin:"ح_ن_ث_ر_خ_ج_س".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[اليوم على الساعة] LT",nextDay:"[غدا على الساعة] LT",nextWeek:"dddd [على الساعة] LT",lastDay:"[أمس على الساعة] LT",lastWeek:"dddd [على الساعة] LT",sameElse:"L"},relativeTime:{future:"في %s",past:"منذ %s",s:"ثوان",ss:"%d ثانية",m:"دقيقة",mm:"%d دقائق",h:"ساعة",hh:"%d ساعات",d:"يوم",dd:"%d أيام",M:"شهر",MM:"%d أشهر",y:"سنة",yy:"%d سنوات"},week:{dow:0,doy:12}});return e}))},4245:function(t,e,n){var r=n("1290");function i(t,e){var n=t.__data__;return r(e)?n["string"==typeof e?"string":"hash"]:n.map}t.exports=i},42454:function(t,e,n){var r=n("f909"),i=n("2ec1"),o=i((function(t,e,n){r(t,e,n)}));t.exports=o},4284:function(t,e){function n(t,e){var n=-1,r=null==t?0:t.length;while(++n<r)if(e(t[n],n,t))return!0;return!1}t.exports=n},"428f":function(t,e,n){"use strict";var r=n("da84");t.exports=r},"42a2":function(t,e,n){var r=n("b5a7"),i=n("79bc"),o=n("1cec"),a=n("c869"),s=n("39ff"),l=n("3729"),u=n("dc57"),c="[object Map]",d="[object Object]",h="[object Promise]",f="[object Set]",p="[object WeakMap]",m="[object DataView]",g=u(r),y=u(i),v=u(o),_=u(a),b=u(s),x=l;(r&&x(new r(new ArrayBuffer(1)))!=m||i&&x(new i)!=c||o&&x(o.resolve())!=h||a&&x(new a)!=f||s&&x(new s)!=p)&&(x=function(t){var e=l(t),n=e==d?t.constructor:void 0,r=n?u(n):"";if(r)switch(r){case g:return m;case y:return c;case v:return h;case _:return f;case b:return p}return e}),t.exports=x},"432f":function(t,e,n){"use strict";
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */const r="undefined"!==typeof window&&null!=window.customElements&&void 0!==window.customElements.polyfillWrapFlushCallback,i=(t,e,n=null)=>{while(e!==n){const n=e.nextSibling;t.removeChild(e),e=n}},o=`{{lit-${String(Math.random()).slice(2)}}}`,a=`\x3c!--${o}--\x3e`,s=new RegExp(`${o}|${a}`),l="$lit$";class u{constructor(t,e){this.parts=[],this.element=e;const n=[],r=[],i=document.createTreeWalker(e.content,133,null,!1);let a=0,u=-1,d=0;const{strings:p,values:{length:m}}=t;while(d<m){const t=i.nextNode();if(null!==t){if(u++,1===t.nodeType){if(t.hasAttributes()){const e=t.attributes,{length:n}=e;let r=0;for(let t=0;t<n;t++)c(e[t].name,l)&&r++;while(r-- >0){const e=p[d],n=f.exec(e)[2],r=n.toLowerCase()+l,i=t.getAttribute(r);t.removeAttribute(r);const o=i.split(s);this.parts.push({type:"attribute",index:u,name:n,strings:o}),d+=o.length-1}}"TEMPLATE"===t.tagName&&(r.push(t),i.currentNode=t.content)}else if(3===t.nodeType){const e=t.data;if(e.indexOf(o)>=0){const r=t.parentNode,i=e.split(s),o=i.length-1;for(let e=0;e<o;e++){let n,o=i[e];if(""===o)n=h();else{const t=f.exec(o);null!==t&&c(t[2],l)&&(o=o.slice(0,t.index)+t[1]+t[2].slice(0,-l.length)+t[3]),n=document.createTextNode(o)}r.insertBefore(n,t),this.parts.push({type:"node",index:++u})}""===i[o]?(r.insertBefore(h(),t),n.push(t)):t.data=i[o],d+=o}}else if(8===t.nodeType)if(t.data===o){const e=t.parentNode;null!==t.previousSibling&&u!==a||(u++,e.insertBefore(h(),t)),a=u,this.parts.push({type:"node",index:u}),null===t.nextSibling?t.data="":(n.push(t),u--),d++}else{let e=-1;while(-1!==(e=t.data.indexOf(o,e+1)))this.parts.push({type:"node",index:-1}),d++}}else i.currentNode=r.pop()}for(const o of n)o.parentNode.removeChild(o)}}const c=(t,e)=>{const n=t.length-e.length;return n>=0&&t.slice(n)===e},d=t=>-1!==t.index,h=()=>document.createComment(""),f=/([ \x09\x0a\x0c\x0d])([^\0-\x1F\x7F-\x9F "'>=/]+)([ \x09\x0a\x0c\x0d]*=[ \x09\x0a\x0c\x0d]*(?:[^ \x09\x0a\x0c\x0d"'`<>=]*|"[^"]*|'[^']*))$/,p=133;function m(t,e){const{element:{content:n},parts:r}=t,i=document.createTreeWalker(n,p,null,!1);let o=y(r),a=r[o],s=-1,l=0;const u=[];let c=null;while(i.nextNode()){s++;const t=i.currentNode;t.previousSibling===c&&(c=null),e.has(t)&&(u.push(t),null===c&&(c=t)),null!==c&&l++;while(void 0!==a&&a.index===s)a.index=null!==c?-1:a.index-l,o=y(r,o),a=r[o]}u.forEach(t=>t.parentNode.removeChild(t))}const g=t=>{let e=11===t.nodeType?0:1;const n=document.createTreeWalker(t,p,null,!1);while(n.nextNode())e++;return e},y=(t,e=-1)=>{for(let n=e+1;n<t.length;n++){const e=t[n];if(d(e))return n}return-1};function v(t,e,n=null){const{element:{content:r},parts:i}=t;if(null===n||void 0===n)return void r.appendChild(e);const o=document.createTreeWalker(r,p,null,!1);let a=y(i),s=0,l=-1;while(o.nextNode()){l++;const t=o.currentNode;t===n&&(s=g(e),n.parentNode.insertBefore(e,n));while(-1!==a&&i[a].index===l){if(s>0){while(-1!==a)i[a].index+=s,a=y(i,a);return}a=y(i,a)}}}
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */const _=new WeakMap,b=t=>"function"===typeof t&&_.has(t),x={},w={};
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
class M{constructor(t,e,n){this.__parts=[],this.template=t,this.processor=e,this.options=n}update(t){let e=0;for(const n of this.__parts)void 0!==n&&n.setValue(t[e]),e++;for(const n of this.__parts)void 0!==n&&n.commit()}_clone(){const t=r?this.template.element.content.cloneNode(!0):document.importNode(this.template.element.content,!0),e=[],n=this.template.parts,i=document.createTreeWalker(t,133,null,!1);let o,a=0,s=0,l=i.nextNode();while(a<n.length)if(o=n[a],d(o)){while(s<o.index)s++,"TEMPLATE"===l.nodeName&&(e.push(l),i.currentNode=l.content),null===(l=i.nextNode())&&(i.currentNode=e.pop(),l=i.nextNode());if("node"===o.type){const t=this.processor.handleTextExpression(this.options);t.insertAfterNode(l.previousSibling),this.__parts.push(t)}else this.__parts.push(...this.processor.handleAttributeExpressions(l,o.name,o.strings,this.options));a++}else this.__parts.push(void 0),a++;return r&&(document.adoptNode(t),customElements.upgrade(t)),t}}
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */const S=window.trustedTypes&&trustedTypes.createPolicy("lit-html",{createHTML:t=>t}),k=` ${o} `;class L{constructor(t,e,n,r){this.strings=t,this.values=e,this.type=n,this.processor=r}getHTML(){const t=this.strings.length-1;let e="",n=!1;for(let r=0;r<t;r++){const t=this.strings[r],i=t.lastIndexOf("\x3c!--");n=(i>-1||n)&&-1===t.indexOf("--\x3e",i+1);const s=f.exec(t);e+=null===s?t+(n?k:a):t.substr(0,s.index)+s[1]+s[2]+l+s[3]+o}return e+=this.strings[t],e}getTemplateElement(){const t=document.createElement("template");let e=this.getHTML();return void 0!==S&&(e=S.createHTML(e)),t.innerHTML=e,t}}
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
const T=t=>null===t||!("object"===typeof t||"function"===typeof t),D=t=>Array.isArray(t)||!(!t||!t[Symbol.iterator]);class O{constructor(t,e,n){this.dirty=!0,this.element=t,this.name=e,this.strings=n,this.parts=[];for(let r=0;r<n.length-1;r++)this.parts[r]=this._createPart()}_createPart(){return new I(this)}_getValue(){const t=this.strings,e=t.length-1,n=this.parts;if(1===e&&""===t[0]&&""===t[1]){const t=n[0].value;if("symbol"===typeof t)return String(t);if("string"===typeof t||!D(t))return t}let r="";for(let i=0;i<e;i++){r+=t[i];const e=n[i];if(void 0!==e){const t=e.value;if(T(t)||!D(t))r+="string"===typeof t?t:String(t);else for(const e of t)r+="string"===typeof e?e:String(e)}}return r+=t[e],r}commit(){this.dirty&&(this.dirty=!1,this.element.setAttribute(this.name,this._getValue()))}}class I{constructor(t){this.value=void 0,this.committer=t}setValue(t){t===x||T(t)&&t===this.value||(this.value=t,b(t)||(this.committer.dirty=!0))}commit(){while(b(this.value)){const t=this.value;this.value=x,t(this)}this.value!==x&&this.committer.commit()}}class C{constructor(t){this.value=void 0,this.__pendingValue=void 0,this.options=t}appendInto(t){this.startNode=t.appendChild(h()),this.endNode=t.appendChild(h())}insertAfterNode(t){this.startNode=t,this.endNode=t.nextSibling}appendIntoPart(t){t.__insert(this.startNode=h()),t.__insert(this.endNode=h())}insertAfterPart(t){t.__insert(this.startNode=h()),this.endNode=t.endNode,t.endNode=this.startNode}setValue(t){this.__pendingValue=t}commit(){if(null===this.startNode.parentNode)return;while(b(this.__pendingValue)){const t=this.__pendingValue;this.__pendingValue=x,t(this)}const t=this.__pendingValue;t!==x&&(T(t)?t!==this.value&&this.__commitText(t):t instanceof L?this.__commitTemplateResult(t):t instanceof Node?this.__commitNode(t):D(t)?this.__commitIterable(t):t===w?(this.value=w,this.clear()):this.__commitText(t))}__insert(t){this.endNode.parentNode.insertBefore(t,this.endNode)}__commitNode(t){this.value!==t&&(this.clear(),this.__insert(t),this.value=t)}__commitText(t){const e=this.startNode.nextSibling;t=null==t?"":t;const n="string"===typeof t?t:String(t);e===this.endNode.previousSibling&&3===e.nodeType?e.data=n:this.__commitNode(document.createTextNode(n)),this.value=t}__commitTemplateResult(t){const e=this.options.templateFactory(t);if(this.value instanceof M&&this.value.template===e)this.value.update(t.values);else{const n=new M(e,t.processor,this.options),r=n._clone();n.update(t.values),this.__commitNode(r),this.value=n}}__commitIterable(t){Array.isArray(this.value)||(this.value=[],this.clear());const e=this.value;let n,r=0;for(const i of t)n=e[r],void 0===n&&(n=new C(this.options),e.push(n),0===r?n.appendIntoPart(this):n.insertAfterPart(e[r-1])),n.setValue(i),n.commit(),r++;r<e.length&&(e.length=r,this.clear(n&&n.endNode))}clear(t=this.startNode){i(this.startNode.parentNode,t.nextSibling,this.endNode)}}class A{constructor(t,e,n){if(this.value=void 0,this.__pendingValue=void 0,2!==n.length||""!==n[0]||""!==n[1])throw new Error("Boolean attributes can only contain a single expression");this.element=t,this.name=e,this.strings=n}setValue(t){this.__pendingValue=t}commit(){while(b(this.__pendingValue)){const t=this.__pendingValue;this.__pendingValue=x,t(this)}if(this.__pendingValue===x)return;const t=!!this.__pendingValue;this.value!==t&&(t?this.element.setAttribute(this.name,""):this.element.removeAttribute(this.name),this.value=t),this.__pendingValue=x}}class Y extends O{constructor(t,e,n){super(t,e,n),this.single=2===n.length&&""===n[0]&&""===n[1]}_createPart(){return new j(this)}_getValue(){return this.single?this.parts[0].value:super._getValue()}commit(){this.dirty&&(this.dirty=!1,this.element[this.name]=this._getValue())}}class j extends I{}let P=!1;(()=>{try{const t={get capture(){return P=!0,!1}};window.addEventListener("test",t,t),window.removeEventListener("test",t,t)}catch(t){}})();class E{constructor(t,e,n){this.value=void 0,this.__pendingValue=void 0,this.element=t,this.eventName=e,this.eventContext=n,this.__boundHandleEvent=t=>this.handleEvent(t)}setValue(t){this.__pendingValue=t}commit(){while(b(this.__pendingValue)){const t=this.__pendingValue;this.__pendingValue=x,t(this)}if(this.__pendingValue===x)return;const t=this.__pendingValue,e=this.value,n=null==t||null!=e&&(t.capture!==e.capture||t.once!==e.once||t.passive!==e.passive),r=null!=t&&(null==e||n);n&&this.element.removeEventListener(this.eventName,this.__boundHandleEvent,this.__options),r&&(this.__options=N(t),this.element.addEventListener(this.eventName,this.__boundHandleEvent,this.__options)),this.value=t,this.__pendingValue=x}handleEvent(t){"function"===typeof this.value?this.value.call(this.eventContext||this.element,t):this.value.handleEvent(t)}}const N=t=>t&&(P?{capture:t.capture,passive:t.passive,once:t.once}:t.capture)
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */;function R(t){let e=H.get(t.type);void 0===e&&(e={stringsArray:new WeakMap,keyString:new Map},H.set(t.type,e));let n=e.stringsArray.get(t.strings);if(void 0!==n)return n;const r=t.strings.join(o);return n=e.keyString.get(r),void 0===n&&(n=new u(t,t.getTemplateElement()),e.keyString.set(r,n)),e.stringsArray.set(t.strings,n),n}const H=new Map,z=new WeakMap,F=(t,e,n)=>{let r=z.get(e);void 0===r&&(i(e,e.firstChild),z.set(e,r=new C(Object.assign({templateFactory:R},n))),r.appendInto(e)),r.setValue(t),r.commit()};
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
class B{handleAttributeExpressions(t,e,n,r){const i=e[0];if("."===i){const r=new Y(t,e.slice(1),n);return r.parts}if("@"===i)return[new E(t,e.slice(1),r.eventContext)];if("?"===i)return[new A(t,e.slice(1),n)];const o=new O(t,e,n);return o.parts}handleTextExpression(t){return new C(t)}}const W=new B;
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */"undefined"!==typeof window&&(window["litHtmlVersions"]||(window["litHtmlVersions"]=[])).push("1.4.1");const V=(t,...e)=>new L(t,e,"html",W),G=(t,e)=>`${t}--${e}`;let $=!0;("undefined"===typeof window.ShadyCSS||"undefined"===typeof window.ShadyCSS.prepareTemplateDom)&&($=!1);const U=t=>e=>{const n=G(e.type,t);let r=H.get(n);void 0===r&&(r={stringsArray:new WeakMap,keyString:new Map},H.set(n,r));let i=r.stringsArray.get(e.strings);if(void 0!==i)return i;const a=e.strings.join(o);if(i=r.keyString.get(a),void 0===i){const n=e.getTemplateElement();$&&window.ShadyCSS.prepareTemplateDom(n,t),i=new u(e,n),r.keyString.set(a,i)}return r.stringsArray.set(e.strings,i),i},q=["html","svg"],X=t=>{q.forEach(e=>{const n=H.get(G(e,t));void 0!==n&&n.keyString.forEach(t=>{const{element:{content:e}}=t,n=new Set;Array.from(e.querySelectorAll("style")).forEach(t=>{n.add(t)}),m(t,n)})})},Z=new Set,J=(t,e,n)=>{Z.add(t);const r=n?n.element:document.createElement("template"),i=e.querySelectorAll("style"),{length:o}=i;if(0===o)return void window.ShadyCSS.prepareTemplateStyles(r,t);const a=document.createElement("style");for(let u=0;u<o;u++){const t=i[u];t.parentNode.removeChild(t),a.textContent+=t.textContent}X(t);const s=r.content;n?v(n,a,s.firstChild):s.insertBefore(a,s.firstChild),window.ShadyCSS.prepareTemplateStyles(r,t);const l=s.querySelector("style");if(window.ShadyCSS.nativeShadow&&null!==l)e.insertBefore(l.cloneNode(!0),e.firstChild);else if(n){s.insertBefore(a,s.firstChild);const t=new Set;t.add(a),m(n,t)}},K=(t,e,n)=>{if(!n||"object"!==typeof n||!n.scopeName)throw new Error("The `scopeName` option is required.");const r=n.scopeName,o=z.has(e),a=$&&11===e.nodeType&&!!e.host,s=a&&!Z.has(r),l=s?document.createDocumentFragment():e;if(F(t,l,Object.assign({templateFactory:U(r)},n)),s){const t=z.get(l);z.delete(l);const n=t.value instanceof M?t.value.template:void 0;J(r,l,n),i(e,e.firstChild),e.appendChild(l),z.set(e,t)}!o&&a&&window.ShadyCSS.styleElement(e.host)};
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
var Q;window.JSCompiler_renameProperty=(t,e)=>t;const tt={toAttribute(t,e){switch(e){case Boolean:return t?"":null;case Object:case Array:return null==t?t:JSON.stringify(t)}return t},fromAttribute(t,e){switch(e){case Boolean:return null!==t;case Number:return null===t?null:Number(t);case Object:case Array:return JSON.parse(t)}return t}},et=(t,e)=>e!==t&&(e===e||t===t),nt={attribute:!0,type:String,converter:tt,reflect:!1,hasChanged:et},rt=1,it=4,ot=8,at=16,st="finalized";class lt extends HTMLElement{constructor(){super(),this.initialize()}static get observedAttributes(){this.finalize();const t=[];return this._classProperties.forEach((e,n)=>{const r=this._attributeNameForProperty(n,e);void 0!==r&&(this._attributeToPropertyMap.set(r,n),t.push(r))}),t}static _ensureClassProperties(){if(!this.hasOwnProperty(JSCompiler_renameProperty("_classProperties",this))){this._classProperties=new Map;const t=Object.getPrototypeOf(this)._classProperties;void 0!==t&&t.forEach((t,e)=>this._classProperties.set(e,t))}}static createProperty(t,e=nt){if(this._ensureClassProperties(),this._classProperties.set(t,e),e.noAccessor||this.prototype.hasOwnProperty(t))return;const n="symbol"===typeof t?Symbol():"__"+t,r=this.getPropertyDescriptor(t,n,e);void 0!==r&&Object.defineProperty(this.prototype,t,r)}static getPropertyDescriptor(t,e,n){return{get(){return this[e]},set(r){const i=this[t];this[e]=r,this.requestUpdateInternal(t,i,n)},configurable:!0,enumerable:!0}}static getPropertyOptions(t){return this._classProperties&&this._classProperties.get(t)||nt}static finalize(){const t=Object.getPrototypeOf(this);if(t.hasOwnProperty(st)||t.finalize(),this[st]=!0,this._ensureClassProperties(),this._attributeToPropertyMap=new Map,this.hasOwnProperty(JSCompiler_renameProperty("properties",this))){const t=this.properties,e=[...Object.getOwnPropertyNames(t),..."function"===typeof Object.getOwnPropertySymbols?Object.getOwnPropertySymbols(t):[]];for(const n of e)this.createProperty(n,t[n])}}static _attributeNameForProperty(t,e){const n=e.attribute;return!1===n?void 0:"string"===typeof n?n:"string"===typeof t?t.toLowerCase():void 0}static _valueHasChanged(t,e,n=et){return n(t,e)}static _propertyValueFromAttribute(t,e){const n=e.type,r=e.converter||tt,i="function"===typeof r?r:r.fromAttribute;return i?i(t,n):t}static _propertyValueToAttribute(t,e){if(void 0===e.reflect)return;const n=e.type,r=e.converter,i=r&&r.toAttribute||tt.toAttribute;return i(t,n)}initialize(){this._updateState=0,this._updatePromise=new Promise(t=>this._enableUpdatingResolver=t),this._changedProperties=new Map,this._saveInstanceProperties(),this.requestUpdateInternal()}_saveInstanceProperties(){this.constructor._classProperties.forEach((t,e)=>{if(this.hasOwnProperty(e)){const t=this[e];delete this[e],this._instanceProperties||(this._instanceProperties=new Map),this._instanceProperties.set(e,t)}})}_applyInstanceProperties(){this._instanceProperties.forEach((t,e)=>this[e]=t),this._instanceProperties=void 0}connectedCallback(){this.enableUpdating()}enableUpdating(){void 0!==this._enableUpdatingResolver&&(this._enableUpdatingResolver(),this._enableUpdatingResolver=void 0)}disconnectedCallback(){}attributeChangedCallback(t,e,n){e!==n&&this._attributeToProperty(t,n)}_propertyToAttribute(t,e,n=nt){const r=this.constructor,i=r._attributeNameForProperty(t,n);if(void 0!==i){const t=r._propertyValueToAttribute(e,n);if(void 0===t)return;this._updateState=this._updateState|ot,null==t?this.removeAttribute(i):this.setAttribute(i,t),this._updateState=this._updateState&~ot}}_attributeToProperty(t,e){if(this._updateState&ot)return;const n=this.constructor,r=n._attributeToPropertyMap.get(t);if(void 0!==r){const t=n.getPropertyOptions(r);this._updateState=this._updateState|at,this[r]=n._propertyValueFromAttribute(e,t),this._updateState=this._updateState&~at}}requestUpdateInternal(t,e,n){let r=!0;if(void 0!==t){const i=this.constructor;n=n||i.getPropertyOptions(t),i._valueHasChanged(this[t],e,n.hasChanged)?(this._changedProperties.has(t)||this._changedProperties.set(t,e),!0!==n.reflect||this._updateState&at||(void 0===this._reflectingProperties&&(this._reflectingProperties=new Map),this._reflectingProperties.set(t,n))):r=!1}!this._hasRequestedUpdate&&r&&(this._updatePromise=this._enqueueUpdate())}requestUpdate(t,e){return this.requestUpdateInternal(t,e),this.updateComplete}async _enqueueUpdate(){this._updateState=this._updateState|it;try{await this._updatePromise}catch(e){}const t=this.performUpdate();return null!=t&&await t,!this._hasRequestedUpdate}get _hasRequestedUpdate(){return this._updateState&it}get hasUpdated(){return this._updateState&rt}performUpdate(){if(!this._hasRequestedUpdate)return;this._instanceProperties&&this._applyInstanceProperties();let t=!1;const e=this._changedProperties;try{t=this.shouldUpdate(e),t?this.update(e):this._markUpdated()}catch(n){throw t=!1,this._markUpdated(),n}t&&(this._updateState&rt||(this._updateState=this._updateState|rt,this.firstUpdated(e)),this.updated(e))}_markUpdated(){this._changedProperties=new Map,this._updateState=this._updateState&~it}get updateComplete(){return this._getUpdateComplete()}_getUpdateComplete(){return this.getUpdateComplete()}getUpdateComplete(){return this._updatePromise}shouldUpdate(t){return!0}update(t){void 0!==this._reflectingProperties&&this._reflectingProperties.size>0&&(this._reflectingProperties.forEach((t,e)=>this._propertyToAttribute(e,this[e],t)),this._reflectingProperties=void 0),this._markUpdated()}updated(t){}firstUpdated(t){}}Q=st,lt[Q]=!0;const ut=Element.prototype;ut.msMatchesSelector||ut.webkitMatchesSelector;
/**
@license
Copyright (c) 2019 The Polymer Project Authors. All rights reserved.
This code may only be used under the BSD style license found at
http://polymer.github.io/LICENSE.txt The complete set of authors may be found at
http://polymer.github.io/AUTHORS.txt The complete set of contributors may be
found at http://polymer.github.io/CONTRIBUTORS.txt Code distributed by Google as
part of the polymer project is also subject to an additional IP rights grant
found at http://polymer.github.io/PATENTS.txt
*/
const ct=window.ShadowRoot&&(void 0===window.ShadyCSS||window.ShadyCSS.nativeShadow)&&"adoptedStyleSheets"in Document.prototype&&"replace"in CSSStyleSheet.prototype,dt=Symbol();class ht{constructor(t,e){if(e!==dt)throw new Error("CSSResult is not constructable. Use `unsafeCSS` or `css` instead.");this.cssText=t}get styleSheet(){return void 0===this._styleSheet&&(ct?(this._styleSheet=new CSSStyleSheet,this._styleSheet.replaceSync(this.cssText)):this._styleSheet=null),this._styleSheet}toString(){return this.cssText}}const ft=t=>new ht(String(t),dt),pt=t=>{if(t instanceof ht)return t.cssText;if("number"===typeof t)return t;throw new Error(`Value passed to 'css' function must be a 'css' function result: ${t}. Use 'unsafeCSS' to pass non-literal values, but\n            take care to ensure page security.`)},mt=(t,...e)=>{const n=e.reduce((e,n,r)=>e+pt(n)+t[r+1],t[0]);return new ht(n,dt)};
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
(window["litElementVersions"]||(window["litElementVersions"]=[])).push("2.5.1");const gt={};class yt extends lt{static getStyles(){return this.styles}static _getUniqueStyles(){if(this.hasOwnProperty(JSCompiler_renameProperty("_styles",this)))return;const t=this.getStyles();if(Array.isArray(t)){const e=(t,n)=>t.reduceRight((t,n)=>Array.isArray(n)?e(n,t):(t.add(n),t),n),n=e(t,new Set),r=[];n.forEach(t=>r.unshift(t)),this._styles=r}else this._styles=void 0===t?[]:[t];this._styles=this._styles.map(t=>{if(t instanceof CSSStyleSheet&&!ct){const e=Array.prototype.slice.call(t.cssRules).reduce((t,e)=>t+e.cssText,"");return ft(e)}return t})}initialize(){super.initialize(),this.constructor._getUniqueStyles(),this.renderRoot=this.createRenderRoot(),window.ShadowRoot&&this.renderRoot instanceof window.ShadowRoot&&this.adoptStyles()}createRenderRoot(){return this.attachShadow(this.constructor.shadowRootOptions)}adoptStyles(){const t=this.constructor._styles;0!==t.length&&(void 0===window.ShadyCSS||window.ShadyCSS.nativeShadow?ct?this.renderRoot.adoptedStyleSheets=t.map(t=>t instanceof CSSStyleSheet?t:t.styleSheet):this._needsShimAdoptedStyleSheets=!0:window.ShadyCSS.ScopingShim.prepareAdoptedCssText(t.map(t=>t.cssText),this.localName))}connectedCallback(){super.connectedCallback(),this.hasUpdated&&void 0!==window.ShadyCSS&&window.ShadyCSS.styleElement(this)}update(t){const e=this.render();super.update(t),e!==gt&&this.constructor.render(e,this.renderRoot,{scopeName:this.localName,eventContext:this}),this._needsShimAdoptedStyleSheets&&(this._needsShimAdoptedStyleSheets=!1,this.constructor._styles.forEach(t=>{const e=document.createElement("style");e.textContent=t.cssText,this.renderRoot.appendChild(e)}))}render(){return gt}}yt["finalized"]=!0,yt.render=K,yt.shadowRootOptions={mode:"open"};const vt=[{name:"初始资金",key:"start_balance",fix_decs:2},{name:"结束资金",key:"end_balance",fix_decs:2},{name:"开始交易日",key:"start_date"},{name:"结束交易日",key:"end_date"},{name:"累计盈利天数",key:"cum_profit_days",fix_decs:0},{name:"累计亏损天数",key:"cum_loss_days",fix_decs:0},{name:"年化收益率",key:"annual_yield",is_percent:!0,fix_decs:2},{name:"最大回撤",key:"max_drawdown",is_percent:!0,fix_decs:2},{name:"总手续费",key:"commission",fix_decs:2},{name:"开仓次数",key:"open_times"},{name:"平仓次数",key:"close_times"},{name:"日均风险度",key:"daily_risk_ratio",is_percent:!0,fix_decs:2},{name:"最大连续盈利天数",key:"max_cont_profit_days"},{name:"最大连续亏损天数",key:"max_cont_loss_days"},{name:"年化夏普率",key:"sharpe_ratio",fix_decs:2},{name:"年化卡玛比率",key:"calmar_ratio",fix_decs:2},{name:"年化索提诺比率",key:"sortino_ratio",fix_decs:2}];class _t extends yt{static get properties(){return{metries:{type:Object}}}static get styles(){return mt`
    table.metries {
      text-indent: 0px;
      border: 2px solid black;
      border-collapse: collapse;
    }
    table.metries>tr {
        border: 1px solid rgba(0, 0, 0, .2);
    }
    table.metries>tr>td {
      padding: 0.5rem;
    }
    table.metries>tr>td.text-right {
      text-align: right;
    }
  `}constructor(){super(),this.metries={},this.columns=vt}formatter(t){return t.fix_decs?t.is_percent?Number(100*this.metries[t.key]).toFixed(t.fix_decs)+"%":Number(this.metries[t.key]).toFixed(t.fix_decs):this.metries[t.key]}render(){return V`
      <table class="metries">
        ${this.columns.map((t,e,n)=>V`
            <tr>
              <td>${t.name}</td>
              <td class="text-right">${this.formatter(t)}</td>
            </tr>`)}
      </table>
      `}}customElements.define("tq-report-metrics",_t)},4359:function(t,e){function n(t,e){var n=-1,r=t.length;e||(e=Array(r));while(++n<r)e[n]=t[n];return e}t.exports=n},"43ad":function(t,e,n){var r=n("23e2");function i(t,e){return function(n,i){return r(n,t,e(i),{})}}t.exports=i},"440c":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,r){var i={m:["eng Minutt","enger Minutt"],h:["eng Stonn","enger Stonn"],d:["een Dag","engem Dag"],M:["ee Mount","engem Mount"],y:["ee Joer","engem Joer"]};return e?i[n][0]:i[n][1]}function n(t){var e=t.substr(0,t.indexOf(" "));return i(e)?"a "+t:"an "+t}function r(t){var e=t.substr(0,t.indexOf(" "));return i(e)?"viru "+t:"virun "+t}function i(t){if(t=parseInt(t,10),isNaN(t))return!1;if(t<0)return!0;if(t<10)return 4<=t&&t<=7;if(t<100){var e=t%10,n=t/10;return i(0===e?n:e)}if(t<1e4){while(t>=10)t/=10;return i(t)}return t/=1e3,i(t)}var o=t.defineLocale("lb",{months:"Januar_Februar_Mäerz_Abrëll_Mee_Juni_Juli_August_September_Oktober_November_Dezember".split("_"),monthsShort:"Jan._Febr._Mrz._Abr._Mee_Jun._Jul._Aug._Sept._Okt._Nov._Dez.".split("_"),monthsParseExact:!0,weekdays:"Sonndeg_Méindeg_Dënschdeg_Mëttwoch_Donneschdeg_Freideg_Samschdeg".split("_"),weekdaysShort:"So._Mé._Dë._Më._Do._Fr._Sa.".split("_"),weekdaysMin:"So_Mé_Dë_Më_Do_Fr_Sa".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm [Auer]",LTS:"H:mm:ss [Auer]",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY H:mm [Auer]",LLLL:"dddd, D. MMMM YYYY H:mm [Auer]"},calendar:{sameDay:"[Haut um] LT",sameElse:"L",nextDay:"[Muer um] LT",nextWeek:"dddd [um] LT",lastDay:"[Gëschter um] LT",lastWeek:function(){switch(this.day()){case 2:case 4:return"[Leschten] dddd [um] LT";default:return"[Leschte] dddd [um] LT"}}},relativeTime:{future:n,past:r,s:"e puer Sekonnen",ss:"%d Sekonnen",m:e,mm:"%d Minutten",h:e,hh:"%d Stonnen",d:e,dd:"%d Deeg",M:e,MM:"%d Méint",y:e,yy:"%d Joer"},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return o}))},4416:function(t,e){function n(t){var e=null==t?0:t.length;return e?t[e-1]:void 0}t.exports=n},4467:function(t,e,n){var r=n("e2e4"),i=n("9520"),o=n("f4d6");function a(t,e,n){e=r(e,t);var a=-1,s=e.length;s||(s=1,t=void 0);while(++a<s){var l=null==t?void 0:t[o(e[a])];void 0===l&&(a=s,l=n),t=i(l)?l.call(t):l}return t}t.exports=a},4472:function(t,e,n){var r=n("872a"),i=n("242e"),o=n("badf");function a(t,e){var n={};return e=o(e,3),i(t,(function(t,i,o){r(n,e(t,i,o),t)})),n}t.exports=a},"44ad":function(t,e,n){"use strict";var r=n("e330"),i=n("d039"),o=n("c6b6"),a=Object,s=r("".split);t.exports=i((function(){return!a("z").propertyIsEnumerable(0)}))?function(t){return"String"===o(t)?s(t,""):a(t)}:a},"44d2":function(t,e,n){"use strict";var r=n("b622"),i=n("7c73"),o=n("9bf2").f,a=r("unscopables"),s=Array.prototype;void 0===s[a]&&o(s,a,{configurable:!0,value:i(null)}),t.exports=function(t){s[a][t]=!0}},"44de":function(t,e,n){"use strict";t.exports=function(t,e){}},"44e7":function(t,e,n){"use strict";var r=n("861d"),i=n("c6b6"),o=n("b622"),a=o("match");t.exports=function(t){var e;return r(t)&&(void 0!==(e=t[a])?!!e:"RegExp"===i(t))}},4625:function(t,e,n){"use strict";var r=n("c6b6"),i=n("e330");t.exports=function(t){if("Function"===r(t))return i(t)}},"466d":function(t,e,n){"use strict";var r=n("c65b"),i=n("d784"),o=n("825a"),a=n("7234"),s=n("50c4"),l=n("577e"),u=n("1d80"),c=n("dc4a"),d=n("8aa5"),h=n("14c3");i("match",(function(t,e,n){return[function(e){var n=u(this),i=a(e)?void 0:c(e,t);return i?r(i,e,n):new RegExp(e)[t](l(n))},function(t){var r=o(this),i=l(t),a=n(e,r,i);if(a.done)return a.value;if(!r.global)return h(r,i);var u=r.unicode;r.lastIndex=0;var c,f=[],p=0;while(null!==(c=h(r,i))){var m=l(c[0]);f[p]=m,""===m&&(r.lastIndex=d(i,s(r.lastIndex),u)),p++}return 0===p?null:f}]}))},4738:function(t,e,n){"use strict";var r=n("da84"),i=n("d256"),o=n("1626"),a=n("94ca"),s=n("8925"),l=n("b622"),u=n("6069"),c=n("6c59"),d=n("c430"),h=n("2d00"),f=i&&i.prototype,p=l("species"),m=!1,g=o(r.PromiseRejectionEvent),y=a("Promise",(function(){var t=s(i),e=t!==String(i);if(!e&&66===h)return!0;if(d&&(!f["catch"]||!f["finally"]))return!0;if(!h||h<51||!/native code/.test(t)){var n=new i((function(t){t(1)})),r=function(t){t((function(){}),(function(){}))},o=n.constructor={};if(o[p]=r,m=n.then((function(){}))instanceof r,!m)return!0}return!e&&(u||c)&&!g}));t.exports={CONSTRUCTOR:y,REJECTION_EVENT:g,SUBCLASSING:m}},4754:function(t,e,n){"use strict";t.exports=function(t,e){return{value:t,done:e}}},4840:function(t,e,n){"use strict";var r=n("825a"),i=n("5087"),o=n("7234"),a=n("b622"),s=a("species");t.exports=function(t,e){var n,a=r(t).constructor;return void 0===a||o(n=r(a)[s])?e:i(n)}},"485a":function(t,e,n){"use strict";var r=n("c65b"),i=n("1626"),o=n("861d"),a=TypeError;t.exports=function(t,e){var n,s;if("string"===e&&i(n=t.toString)&&!o(s=r(n,t)))return s;if(i(n=t.valueOf)&&!o(s=r(n,t)))return s;if("string"!==e&&i(n=t.toString)&&!o(s=r(n,t)))return s;throw new a("Can't convert object to primitive value")}},"485c":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"-inci",5:"-inci",8:"-inci",70:"-inci",80:"-inci",2:"-nci",7:"-nci",20:"-nci",50:"-nci",3:"-üncü",4:"-üncü",100:"-üncü",6:"-ncı",9:"-uncu",10:"-uncu",30:"-uncu",60:"-ıncı",90:"-ıncı"},n=t.defineLocale("az",{months:"yanvar_fevral_mart_aprel_may_iyun_iyul_avqust_sentyabr_oktyabr_noyabr_dekabr".split("_"),monthsShort:"yan_fev_mar_apr_may_iyn_iyl_avq_sen_okt_noy_dek".split("_"),weekdays:"Bazar_Bazar ertəsi_Çərşənbə axşamı_Çərşənbə_Cümə axşamı_Cümə_Şənbə".split("_"),weekdaysShort:"Baz_BzE_ÇAx_Çər_CAx_Cüm_Şən".split("_"),weekdaysMin:"Bz_BE_ÇA_Çə_CA_Cü_Şə".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[bugün saat] LT",nextDay:"[sabah saat] LT",nextWeek:"[gələn həftə] dddd [saat] LT",lastDay:"[dünən] LT",lastWeek:"[keçən həftə] dddd [saat] LT",sameElse:"L"},relativeTime:{future:"%s sonra",past:"%s əvvəl",s:"bir neçə saniyə",ss:"%d saniyə",m:"bir dəqiqə",mm:"%d dəqiqə",h:"bir saat",hh:"%d saat",d:"bir gün",dd:"%d gün",M:"bir ay",MM:"%d ay",y:"bir il",yy:"%d il"},meridiemParse:/gecə|səhər|gündüz|axşam/,isPM:function(t){return/^(gündüz|axşam)$/.test(t)},meridiem:function(t,e,n){return t<4?"gecə":t<12?"səhər":t<17?"gündüz":"axşam"},dayOfMonthOrdinalParse:/\d{1,2}-(ıncı|inci|nci|üncü|ncı|uncu)/,ordinal:function(t){if(0===t)return t+"-ıncı";var n=t%10,r=t%100-n,i=t>=100?100:null;return t+(e[n]||e[r]||e[i])},week:{dow:1,doy:7}});return n}))},"49ab":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("zh-hk",{months:"一月_二月_三月_四月_五月_六月_七月_八月_九月_十月_十一月_十二月".split("_"),monthsShort:"1月_2月_3月_4月_5月_6月_7月_8月_9月_10月_11月_12月".split("_"),weekdays:"星期日_星期一_星期二_星期三_星期四_星期五_星期六".split("_"),weekdaysShort:"週日_週一_週二_週三_週四_週五_週六".split("_"),weekdaysMin:"日_一_二_三_四_五_六".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY/MM/DD",LL:"YYYY年M月D日",LLL:"YYYY年M月D日 HH:mm",LLLL:"YYYY年M月D日dddd HH:mm",l:"YYYY/M/D",ll:"YYYY年M月D日",lll:"YYYY年M月D日 HH:mm",llll:"YYYY年M月D日dddd HH:mm"},meridiemParse:/凌晨|早上|上午|中午|下午|晚上/,meridiemHour:function(t,e){return 12===t&&(t=0),"凌晨"===e||"早上"===e||"上午"===e?t:"中午"===e?t>=11?t:t+12:"下午"===e||"晚上"===e?t+12:void 0},meridiem:function(t,e,n){var r=100*t+e;return r<600?"凌晨":r<900?"早上":r<1200?"上午":1200===r?"中午":r<1800?"下午":"晚上"},calendar:{sameDay:"[今天]LT",nextDay:"[明天]LT",nextWeek:"[下]ddddLT",lastDay:"[昨天]LT",lastWeek:"[上]ddddLT",sameElse:"L"},dayOfMonthOrdinalParse:/\d{1,2}(日|月|週)/,ordinal:function(t,e){switch(e){case"d":case"D":case"DDD":return t+"日";case"M":return t+"月";case"w":case"W":return t+"週";default:return t}},relativeTime:{future:"%s後",past:"%s前",s:"幾秒",ss:"%d 秒",m:"1 分鐘",mm:"%d 分鐘",h:"1 小時",hh:"%d 小時",d:"1 天",dd:"%d 天",M:"1 個月",MM:"%d 個月",y:"1 年",yy:"%d 年"}});return e}))},"49ad":function(t,e,n){"use strict";t.exports=function(t){var e={};function n(n){var r=t.get(n);return void 0===r?[]:e[r]||[]}function r(n,r){var i=t.get(n);e[i]||(e[i]=[]),e[i].push(r)}function i(t,e){for(var r=n(t),i=0,o=r.length;i<o;++i)if(r[i]===e){r.splice(i,1);break}}function o(t){var e=n(t);e&&(e.length=0)}return{get:n,add:r,removeListener:i,removeAllListeners:o}}},"49f4":function(t,e,n){var r=n("6044");function i(){this.__data__=r?r(null):{},this.size=0}t.exports=i},"4ba9":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n){var r=t+" ";switch(n){case"ss":return r+=1===t?"sekunda":2===t||3===t||4===t?"sekunde":"sekundi",r;case"m":return e?"jedna minuta":"jedne minute";case"mm":return r+=1===t?"minuta":2===t||3===t||4===t?"minute":"minuta",r;case"h":return e?"jedan sat":"jednog sata";case"hh":return r+=1===t?"sat":2===t||3===t||4===t?"sata":"sati",r;case"dd":return r+=1===t?"dan":"dana",r;case"MM":return r+=1===t?"mjesec":2===t||3===t||4===t?"mjeseca":"mjeseci",r;case"yy":return r+=1===t?"godina":2===t||3===t||4===t?"godine":"godina",r}}var n=t.defineLocale("hr",{months:{format:"siječnja_veljače_ožujka_travnja_svibnja_lipnja_srpnja_kolovoza_rujna_listopada_studenoga_prosinca".split("_"),standalone:"siječanj_veljača_ožujak_travanj_svibanj_lipanj_srpanj_kolovoz_rujan_listopad_studeni_prosinac".split("_")},monthsShort:"sij._velj._ožu._tra._svi._lip._srp._kol._ruj._lis._stu._pro.".split("_"),monthsParseExact:!0,weekdays:"nedjelja_ponedjeljak_utorak_srijeda_četvrtak_petak_subota".split("_"),weekdaysShort:"ned._pon._uto._sri._čet._pet._sub.".split("_"),weekdaysMin:"ne_po_ut_sr_če_pe_su".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD.MM.YYYY",LL:"Do MMMM YYYY",LLL:"Do MMMM YYYY H:mm",LLLL:"dddd, Do MMMM YYYY H:mm"},calendar:{sameDay:"[danas u] LT",nextDay:"[sutra u] LT",nextWeek:function(){switch(this.day()){case 0:return"[u] [nedjelju] [u] LT";case 3:return"[u] [srijedu] [u] LT";case 6:return"[u] [subotu] [u] LT";case 1:case 2:case 4:case 5:return"[u] dddd [u] LT"}},lastDay:"[jučer u] LT",lastWeek:function(){switch(this.day()){case 0:return"[prošlu] [nedjelju] [u] LT";case 3:return"[prošlu] [srijedu] [u] LT";case 6:return"[prošle] [subote] [u] LT";case 1:case 2:case 4:case 5:return"[prošli] dddd [u] LT"}},sameElse:"L"},relativeTime:{future:"za %s",past:"prije %s",s:"par sekundi",ss:e,m:e,mm:e,h:e,hh:e,d:"dan",dd:e,M:"mjesec",MM:e,y:"godinu",yy:e},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:7}});return n}))},"4bb5":function(t,e,n){var r=n("e2e4"),i=n("4416"),o=n("8296"),a=n("f4d6");function s(t,e){return e=r(e,t),t=o(t,e),null==t||delete t[a(i(e))]}t.exports=s},"4be8":function(t,e,n){var r=n("30e0"),i=n("ec69");function o(t,e){return t&&r(t,e,i)}t.exports=o},"4c98":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"١",2:"٢",3:"٣",4:"٤",5:"٥",6:"٦",7:"٧",8:"٨",9:"٩",0:"٠"},n={"١":"1","٢":"2","٣":"3","٤":"4","٥":"5","٦":"6","٧":"7","٨":"8","٩":"9","٠":"0"},r=t.defineLocale("ar-ps",{months:"كانون الثاني_شباط_آذار_نيسان_أيّار_حزيران_تمّوز_آب_أيلول_تشري الأوّل_تشرين الثاني_كانون الأوّل".split("_"),monthsShort:"ك٢_شباط_آذار_نيسان_أيّار_حزيران_تمّوز_آب_أيلول_ت١_ت٢_ك١".split("_"),weekdays:"الأحد_الإثنين_الثلاثاء_الأربعاء_الخميس_الجمعة_السبت".split("_"),weekdaysShort:"أحد_إثنين_ثلاثاء_أربعاء_خميس_جمعة_سبت".split("_"),weekdaysMin:"ح_ن_ث_ر_خ_ج_س".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},meridiemParse:/ص|م/,isPM:function(t){return"م"===t},meridiem:function(t,e,n){return t<12?"ص":"م"},calendar:{sameDay:"[اليوم على الساعة] LT",nextDay:"[غدا على الساعة] LT",nextWeek:"dddd [على الساعة] LT",lastDay:"[أمس على الساعة] LT",lastWeek:"dddd [على الساعة] LT",sameElse:"L"},relativeTime:{future:"في %s",past:"منذ %s",s:"ثوان",ss:"%d ثانية",m:"دقيقة",mm:"%d دقائق",h:"ساعة",hh:"%d ساعات",d:"يوم",dd:"%d أيام",M:"شهر",MM:"%d أشهر",y:"سنة",yy:"%d سنوات"},preparse:function(t){return t.replace(/[٣٤٥٦٧٨٩٠]/g,(function(t){return n[t]})).split("").reverse().join("").replace(/[١٢](?![\u062a\u0643])/g,(function(t){return n[t]})).split("").reverse().join("").replace(/،/g,",")},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]})).replace(/,/g,"،")},week:{dow:0,doy:6}});return r}))},"4d64":function(t,e,n){"use strict";var r=n("fc6a"),i=n("23cb"),o=n("07fa"),a=function(t){return function(e,n,a){var s=r(e),l=o(s);if(0===l)return!t&&-1;var u,c=i(a,l);if(t&&n!==n){while(l>c)if(u=s[c++],u!==u)return!0}else for(;l>c;c++)if((t||c in s)&&s[c]===n)return t||c||0;return!t&&-1}};t.exports={includes:a(!0),indexOf:a(!1)}},"4d8b":function(t,e,n){var r=n("9b02");function i(t,e){var n=-1,i=e.length,o=Array(i),a=null==t;while(++n<i)o[n]=a?void 0:r(t,e[n]);return o}t.exports=i},"4d8c":function(t,e,n){var r=n("5c69");function i(t){var e=null==t?0:t.length;return e?r(t,1):[]}t.exports=i},"4de4":function(t,e,n){"use strict";var r=n("23e7"),i=n("b727").filter,o=n("1dde"),a=o("filter");r({target:"Array",proto:!0,forced:!a},{filter:function(t){return i(this,t,arguments.length>1?arguments[1]:void 0)}})},"4df4":function(t,e,n){"use strict";var r=n("0366"),i=n("c65b"),o=n("7b0b"),a=n("9bdd"),s=n("e95a"),l=n("68ee"),u=n("07fa"),c=n("8418"),d=n("9a1f"),h=n("35a1"),f=Array;t.exports=function(t){var e=o(t),n=l(this),p=arguments.length,m=p>1?arguments[1]:void 0,g=void 0!==m;g&&(m=r(m,p>2?arguments[2]:void 0));var y,v,_,b,x,w,M=h(e),S=0;if(!M||this===f&&s(M))for(y=u(e),v=n?new this(y):f(y);y>S;S++)w=g?m(e[S],S):e[S],c(v,S,w);else for(v=n?new this:[],b=d(e,M),x=b.next;!(_=i(x,b)).done;S++)w=g?a(b,m,[_.value,S],!0):_.value,c(v,S,w);return v.length=S,v}},"4e82":function(t,e,n){"use strict";var r=n("23e7"),i=n("e330"),o=n("59ed"),a=n("7b0b"),s=n("07fa"),l=n("083a"),u=n("577e"),c=n("d039"),d=n("addb"),h=n("a640"),f=n("04d1"),p=n("d998"),m=n("2d00"),g=n("512c"),y=[],v=i(y.sort),_=i(y.push),b=c((function(){y.sort(void 0)})),x=c((function(){y.sort(null)})),w=h("sort"),M=!c((function(){if(m)return m<70;if(!(f&&f>3)){if(p)return!0;if(g)return g<603;var t,e,n,r,i="";for(t=65;t<76;t++){switch(e=String.fromCharCode(t),t){case 66:case 69:case 70:case 72:n=3;break;case 68:case 71:n=4;break;default:n=2}for(r=0;r<47;r++)y.push({k:e+r,v:n})}for(y.sort((function(t,e){return e.v-t.v})),r=0;r<y.length;r++)e=y[r].k.charAt(0),i.charAt(i.length-1)!==e&&(i+=e);return"DGBEFHACIJK"!==i}})),S=b||!x||!w||!M,k=function(t){return function(e,n){return void 0===n?-1:void 0===e?1:void 0!==t?+t(e,n)||0:u(e)>u(n)?1:-1}};r({target:"Array",proto:!0,forced:S},{sort:function(t){void 0!==t&&o(t);var e=a(this);if(M)return void 0===t?v(e):v(e,t);var n,r,i=[],u=s(e);for(r=0;r<u;r++)r in e&&_(i,e[r]);d(i,k(t)),n=s(i),r=0;while(r<n)e[r]=i[r++];while(r<u)l(e,r++);return e}})},"4f50":function(t,e,n){var r=n("b760"),i=n("e5383"),o=n("c8fe"),a=n("4359"),s=n("fa21"),l=n("d370"),u=n("6747"),c=n("dcbe"),d=n("0d24"),h=n("9520"),f=n("1a8c"),p=n("60ed"),m=n("73ac"),g=n("8adb"),y=n("8de2");function v(t,e,n,v,_,b,x){var w=g(t,n),M=g(e,n),S=x.get(M);if(S)r(t,n,S);else{var k=b?b(w,M,n+"",t,e,x):void 0,L=void 0===k;if(L){var T=u(M),D=!T&&d(M),O=!T&&!D&&m(M);k=M,T||D||O?u(w)?k=w:c(w)?k=a(w):D?(L=!1,k=i(M,!0)):O?(L=!1,k=o(M,!0)):k=[]:p(M)||l(M)?(k=w,l(w)?k=y(w):f(w)&&!h(w)||(k=s(M))):L=!1}L&&(x.set(M,k),_(k,M,v,b,x),x["delete"](M)),r(t,n,k)}}t.exports=v},5038:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("id",{months:"Januari_Februari_Maret_April_Mei_Juni_Juli_Agustus_September_Oktober_November_Desember".split("_"),monthsShort:"Jan_Feb_Mar_Apr_Mei_Jun_Jul_Agt_Sep_Okt_Nov_Des".split("_"),weekdays:"Minggu_Senin_Selasa_Rabu_Kamis_Jumat_Sabtu".split("_"),weekdaysShort:"Min_Sen_Sel_Rab_Kam_Jum_Sab".split("_"),weekdaysMin:"Mg_Sn_Sl_Rb_Km_Jm_Sb".split("_"),longDateFormat:{LT:"HH.mm",LTS:"HH.mm.ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY [pukul] HH.mm",LLLL:"dddd, D MMMM YYYY [pukul] HH.mm"},meridiemParse:/pagi|siang|sore|malam/,meridiemHour:function(t,e){return 12===t&&(t=0),"pagi"===e?t:"siang"===e?t>=11?t:t+12:"sore"===e||"malam"===e?t+12:void 0},meridiem:function(t,e,n){return t<11?"pagi":t<15?"siang":t<19?"sore":"malam"},calendar:{sameDay:"[Hari ini pukul] LT",nextDay:"[Besok pukul] LT",nextWeek:"dddd [pukul] LT",lastDay:"[Kemarin pukul] LT",lastWeek:"dddd [lalu pukul] LT",sameElse:"L"},relativeTime:{future:"dalam %s",past:"%s yang lalu",s:"beberapa detik",ss:"%d detik",m:"semenit",mm:"%d menit",h:"sejam",hh:"%d jam",d:"sehari",dd:"%d hari",M:"sebulan",MM:"%d bulan",y:"setahun",yy:"%d tahun"},week:{dow:0,doy:6}});return e}))},5058:function(t,e,n){"use strict";t.exports=function(t){var e=t.idGenerator,n=t.stateHandler.getState;function r(t){var e=n(t);return e&&void 0!==e.id?e.id:null}function i(t){var r=n(t);if(!r)throw new Error("setId required the element to have a resize detection state.");var i=e.generate();return r.id=i,i}return{get:r,set:i}}},5087:function(t,e,n){"use strict";var r=n("68ee"),i=n("0d51"),o=TypeError;t.exports=function(t){if(r(t))return t;throw new o(i(t)+" is not a constructor")}},"50bf":function(t,e,n){"use strict";var r=t.exports={};function i(t,e,n){var r=t[e];return void 0!==r&&null!==r||void 0===n?r:n}r.getOption=i},"50c4":function(t,e,n){"use strict";var r=n("5926"),i=Math.min;t.exports=function(t){var e=r(t);return e>0?i(e,9007199254740991):0}},"50ca":function(t,e,n){var r=n("8057"),i=n("7530"),o=n("242e"),a=n("badf"),s=n("2dcb"),l=n("6747"),u=n("0d24"),c=n("9520"),d=n("1a8c"),h=n("73ac");function f(t,e,n){var f=l(t),p=f||u(t)||h(t);if(e=a(e,4),null==n){var m=t&&t.constructor;n=p?f?new m:[]:d(t)&&c(m)?i(s(t)):{}}return(p?r:o)(t,(function(t,r,i){return e(n,t,r,i)})),n}t.exports=f},"50d8":function(t,e){function n(t,e){var n=-1,r=Array(t);while(++n<t)r[n]=e(n);return r}t.exports=n},5120:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=["Eanáir","Feabhra","Márta","Aibreán","Bealtaine","Meitheamh","Iúil","Lúnasa","Meán Fómhair","Deireadh Fómhair","Samhain","Nollaig"],n=["Ean","Feabh","Márt","Aib","Beal","Meith","Iúil","Lún","M.F.","D.F.","Samh","Noll"],r=["Dé Domhnaigh","Dé Luain","Dé Máirt","Dé Céadaoin","Déardaoin","Dé hAoine","Dé Sathairn"],i=["Domh","Luan","Máirt","Céad","Déar","Aoine","Sath"],o=["Do","Lu","Má","Cé","Dé","A","Sa"],a=t.defineLocale("ga",{months:e,monthsShort:n,monthsParseExact:!0,weekdays:r,weekdaysShort:i,weekdaysMin:o,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Inniu ag] LT",nextDay:"[Amárach ag] LT",nextWeek:"dddd [ag] LT",lastDay:"[Inné ag] LT",lastWeek:"dddd [seo caite] [ag] LT",sameElse:"L"},relativeTime:{future:"i %s",past:"%s ó shin",s:"cúpla soicind",ss:"%d soicind",m:"nóiméad",mm:"%d nóiméad",h:"uair an chloig",hh:"%d uair an chloig",d:"lá",dd:"%d lá",M:"mí",MM:"%d míonna",y:"bliain",yy:"%d bliain"},dayOfMonthOrdinalParse:/\d{1,2}(d|na|mh)/,ordinal:function(t){var e=1===t?"d":t%10===2?"na":"mh";return t+e},week:{dow:1,doy:4}});return a}))},"512c":function(t,e,n){"use strict";var r=n("342f"),i=r.match(/AppleWebKit\/(\d+)\./);t.exports=!!i&&+i[1]},"51eb":function(t,e,n){"use strict";var r=n("825a"),i=n("485a"),o=TypeError;t.exports=function(t){if(r(this),"string"===t||"default"===t)t="string";else if("number"!==t)throw new o("Incorrect hint");return i(this,t)}},5250:function(t,e,n){"use strict";var r=n("099a");e["a"]=r["a"]},5294:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=["جنوری","فروری","مارچ","اپریل","مئی","جون","جولائی","اگست","ستمبر","اکتوبر","نومبر","دسمبر"],n=["اتوار","پیر","منگل","بدھ","جمعرات","جمعہ","ہفتہ"],r=t.defineLocale("ur",{months:e,monthsShort:e,weekdays:n,weekdaysShort:n,weekdaysMin:n,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd، D MMMM YYYY HH:mm"},meridiemParse:/صبح|شام/,isPM:function(t){return"شام"===t},meridiem:function(t,e,n){return t<12?"صبح":"شام"},calendar:{sameDay:"[آج بوقت] LT",nextDay:"[کل بوقت] LT",nextWeek:"dddd [بوقت] LT",lastDay:"[گذشتہ روز بوقت] LT",lastWeek:"[گذشتہ] dddd [بوقت] LT",sameElse:"L"},relativeTime:{future:"%s بعد",past:"%s قبل",s:"چند سیکنڈ",ss:"%d سیکنڈ",m:"ایک منٹ",mm:"%d منٹ",h:"ایک گھنٹہ",hh:"%d گھنٹے",d:"ایک دن",dd:"%d دن",M:"ایک ماہ",MM:"%d ماہ",y:"ایک سال",yy:"%d سال"},preparse:function(t){return t.replace(/،/g,",")},postformat:function(t){return t.replace(/,/g,"،")},week:{dow:1,doy:4}});return r}))},"52bd":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ss",{months:"Bhimbidvwane_Indlovana_Indlov'lenkhulu_Mabasa_Inkhwekhweti_Inhlaba_Kholwane_Ingci_Inyoni_Imphala_Lweti_Ingongoni".split("_"),monthsShort:"Bhi_Ina_Inu_Mab_Ink_Inh_Kho_Igc_Iny_Imp_Lwe_Igo".split("_"),weekdays:"Lisontfo_Umsombuluko_Lesibili_Lesitsatfu_Lesine_Lesihlanu_Umgcibelo".split("_"),weekdaysShort:"Lis_Umb_Lsb_Les_Lsi_Lsh_Umg".split("_"),weekdaysMin:"Li_Us_Lb_Lt_Ls_Lh_Ug".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"h:mm A",LTS:"h:mm:ss A",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY h:mm A",LLLL:"dddd, D MMMM YYYY h:mm A"},calendar:{sameDay:"[Namuhla nga] LT",nextDay:"[Kusasa nga] LT",nextWeek:"dddd [nga] LT",lastDay:"[Itolo nga] LT",lastWeek:"dddd [leliphelile] [nga] LT",sameElse:"L"},relativeTime:{future:"nga %s",past:"wenteka nga %s",s:"emizuzwana lomcane",ss:"%d mzuzwana",m:"umzuzu",mm:"%d emizuzu",h:"lihora",hh:"%d emahora",d:"lilanga",dd:"%d emalanga",M:"inyanga",MM:"%d tinyanga",y:"umnyaka",yy:"%d iminyaka"},meridiemParse:/ekuseni|emini|entsambama|ebusuku/,meridiem:function(t,e,n){return t<11?"ekuseni":t<15?"emini":t<19?"entsambama":"ebusuku"},meridiemHour:function(t,e){return 12===t&&(t=0),"ekuseni"===e?t:"emini"===e?t>=11?t:t+12:"entsambama"===e||"ebusuku"===e?0===t?0:t+12:void 0},dayOfMonthOrdinalParse:/\d{1,2}/,ordinal:"%d",week:{dow:1,doy:4}});return e}))},5334:function(t,e,n){"use strict";var r=function(){var t=this,e=t._self._c;return e("div",{directives:[{name:"show",rawName:"v-show",value:t.show,expression:"show"}],class:t.prefixCls,style:t.contentStyle},[t._t("default")],2)},i=[],o=(n("b0c0"),n("a9e3"),"ivu-tabs-tabpane"),a={name:"TabPane",inject:["TabsInstance"],props:{name:{type:String},label:{type:[String,Function],default:""},icon:{type:String},disabled:{type:Boolean,default:!1},closable:{type:Boolean,default:null},tab:{type:String},index:{type:Number},contextMenu:{type:Boolean,default:!1}},data:function(){return{prefixCls:o,show:!0,currentName:this.name}},computed:{contentStyle:function(){return{visibility:this.TabsInstance.activeKey!==this.currentName?"hidden":"visible"}}},methods:{updateNav:function(){this.TabsInstance.updateNav()}},watch:{name:function(t){this.currentName=t,this.updateNav()},label:function(){this.updateNav()},icon:function(){this.updateNav()},disabled:function(){this.updateNav()}},mounted:function(){this.updateNav()},destroyed:function(){this.updateNav()}},s=a,l=n("2877"),u=Object(l["a"])(s,r,i,!1,null,null,null);e["a"]=u.exports},"53ca":function(t,e,n){"use strict";n.d(e,"a",(function(){return r}));n("a4d3"),n("e01a"),n("d28b"),n("d3b7"),n("3ca3"),n("ddb0");function r(t){return r="function"==typeof Symbol&&"symbol"==typeof Symbol.iterator?function(t){return typeof t}:function(t){return t&&"function"==typeof Symbol&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},r(t)}},"54eb":function(t,e,n){var r=n("8eeb"),i=n("32f4");function o(t,e){return r(t,i(t),e)}t.exports=o},"55a3":function(t,e){function n(t){return this.__data__.has(t)}t.exports=n},"55c9":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="ene._feb._mar._abr._may._jun._jul._ago._sep._oct._nov._dic.".split("_"),n="ene_feb_mar_abr_may_jun_jul_ago_sep_oct_nov_dic".split("_"),r=[/^ene/i,/^feb/i,/^mar/i,/^abr/i,/^may/i,/^jun/i,/^jul/i,/^ago/i,/^sep/i,/^oct/i,/^nov/i,/^dic/i],i=/^(enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre|ene\.?|feb\.?|mar\.?|abr\.?|may\.?|jun\.?|jul\.?|ago\.?|sep\.?|oct\.?|nov\.?|dic\.?)/i,o=t.defineLocale("es-us",{months:"enero_febrero_marzo_abril_mayo_junio_julio_agosto_septiembre_octubre_noviembre_diciembre".split("_"),monthsShort:function(t,r){return t?/-MMM-/.test(r)?n[t.month()]:e[t.month()]:e},monthsRegex:i,monthsShortRegex:i,monthsStrictRegex:/^(enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre)/i,monthsShortStrictRegex:/^(ene\.?|feb\.?|mar\.?|abr\.?|may\.?|jun\.?|jul\.?|ago\.?|sep\.?|oct\.?|nov\.?|dic\.?)/i,monthsParse:r,longMonthsParse:r,shortMonthsParse:r,weekdays:"domingo_lunes_martes_miércoles_jueves_viernes_sábado".split("_"),weekdaysShort:"dom._lun._mar._mié._jue._vie._sáb.".split("_"),weekdaysMin:"do_lu_ma_mi_ju_vi_sá".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"h:mm A",LTS:"h:mm:ss A",L:"MM/DD/YYYY",LL:"D [de] MMMM [de] YYYY",LLL:"D [de] MMMM [de] YYYY h:mm A",LLLL:"dddd, D [de] MMMM [de] YYYY h:mm A"},calendar:{sameDay:function(){return"[hoy a la"+(1!==this.hours()?"s":"")+"] LT"},nextDay:function(){return"[mañana a la"+(1!==this.hours()?"s":"")+"] LT"},nextWeek:function(){return"dddd [a la"+(1!==this.hours()?"s":"")+"] LT"},lastDay:function(){return"[ayer a la"+(1!==this.hours()?"s":"")+"] LT"},lastWeek:function(){return"[el] dddd [pasado a la"+(1!==this.hours()?"s":"")+"] LT"},sameElse:"L"},relativeTime:{future:"en %s",past:"hace %s",s:"unos segundos",ss:"%d segundos",m:"un minuto",mm:"%d minutos",h:"una hora",hh:"%d horas",d:"un día",dd:"%d días",w:"una semana",ww:"%d semanas",M:"un mes",MM:"%d meses",y:"un año",yy:"%d años"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:0,doy:6}});return o}))},5692:function(t,e,n){"use strict";var r=n("c6cd");t.exports=function(t,e){return r[t]||(r[t]=e||{})}},"56ef":function(t,e,n){"use strict";var r=n("d066"),i=n("e330"),o=n("241c"),a=n("7418"),s=n("825a"),l=i([].concat);t.exports=r("Reflect","ownKeys")||function(t){var e=o.f(s(t)),n=a.f;return n?l(e,n(t)):e}},"576c":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("tet",{months:"Janeiru_Fevereiru_Marsu_Abril_Maiu_Juñu_Jullu_Agustu_Setembru_Outubru_Novembru_Dezembru".split("_"),monthsShort:"Jan_Fev_Mar_Abr_Mai_Jun_Jul_Ago_Set_Out_Nov_Dez".split("_"),weekdays:"Domingu_Segunda_Tersa_Kuarta_Kinta_Sesta_Sabadu".split("_"),weekdaysShort:"Dom_Seg_Ters_Kua_Kint_Sest_Sab".split("_"),weekdaysMin:"Do_Seg_Te_Ku_Ki_Ses_Sa".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Ohin iha] LT",nextDay:"[Aban iha] LT",nextWeek:"dddd [iha] LT",lastDay:"[Horiseik iha] LT",lastWeek:"dddd [semana kotuk] [iha] LT",sameElse:"L"},relativeTime:{future:"iha %s",past:"%s liuba",s:"segundu balun",ss:"segundu %d",m:"minutu ida",mm:"minutu %d",h:"oras ida",hh:"oras %d",d:"loron ida",dd:"loron %d",M:"fulan ida",MM:"fulan %d",y:"tinan ida",yy:"tinan %d"},dayOfMonthOrdinalParse:/\d{1,2}(st|nd|rd|th)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n},week:{dow:1,doy:4}});return e}))},"577e":function(t,e,n){"use strict";var r=n("f5df"),i=String;t.exports=function(t){if("Symbol"===r(t))throw new TypeError("Cannot convert a Symbol value to a string");return i(t)}},"57a5":function(t,e,n){var r=n("91e9"),i=r(Object.keys,Object);t.exports=i},"57b9":function(t,e,n){"use strict";var r=n("c65b"),i=n("d066"),o=n("b622"),a=n("cb2d");t.exports=function(){var t=i("Symbol"),e=t&&t.prototype,n=e&&e.valueOf,s=o("toPrimitive");e&&!e[s]&&a(e,s,(function(t){return r(n,this)}),{arity:1})}},"585a":function(t,e,n){(function(e){var n="object"==typeof e&&e&&e.Object===Object&&e;t.exports=n}).call(this,n("c8ba"))},5899:function(t,e,n){"use strict";t.exports="\t\n\v\f\r                　\u2028\u2029\ufeff"},"58a8":function(t,e,n){"use strict";var r=n("e330"),i=n("1d80"),o=n("577e"),a=n("5899"),s=r("".replace),l=RegExp("^["+a+"]+"),u=RegExp("(^|[^"+a+"])["+a+"]+$"),c=function(t){return function(e){var n=o(i(e));return 1&t&&(n=s(n,l,"")),2&t&&(n=s(n,u,"$1")),n}};t.exports={start:c(1),end:c(2),trim:c(3)}},5926:function(t,e,n){"use strict";var r=n("b42e");t.exports=function(t){var e=+t;return e!==e||0===e?0:r(e)}},"598a":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=["ޖެނުއަރީ","ފެބްރުއަރީ","މާރިޗު","އޭޕްރީލު","މޭ","ޖޫން","ޖުލައި","އޯގަސްޓު","ސެޕްޓެމްބަރު","އޮކްޓޯބަރު","ނޮވެމްބަރު","ޑިސެމްބަރު"],n=["އާދިއްތަ","ހޯމަ","އަންގާރަ","ބުދަ","ބުރާސްފަތި","ހުކުރު","ހޮނިހިރު"],r=t.defineLocale("dv",{months:e,monthsShort:e,weekdays:n,weekdaysShort:n,weekdaysMin:"އާދި_ހޯމަ_އަން_ބުދަ_ބުރާ_ހުކު_ހޮނި".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"D/M/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},meridiemParse:/މކ|މފ/,isPM:function(t){return"މފ"===t},meridiem:function(t,e,n){return t<12?"މކ":"މފ"},calendar:{sameDay:"[މިއަދު] LT",nextDay:"[މާދަމާ] LT",nextWeek:"dddd LT",lastDay:"[އިއްޔެ] LT",lastWeek:"[ފާއިތުވި] dddd LT",sameElse:"L"},relativeTime:{future:"ތެރޭގައި %s",past:"ކުރިން %s",s:"ސިކުންތުކޮޅެއް",ss:"d% ސިކުންތު",m:"މިނިޓެއް",mm:"މިނިޓު %d",h:"ގަޑިއިރެއް",hh:"ގަޑިއިރު %d",d:"ދުވަހެއް",dd:"ދުވަސް %d",M:"މަހެއް",MM:"މަސް %d",y:"އަހަރެއް",yy:"އަހަރު %d"},preparse:function(t){return t.replace(/،/g,",")},postformat:function(t){return t.replace(/,/g,"،")},week:{dow:7,doy:12}});return r}))},"59ed":function(t,e,n){"use strict";var r=n("1626"),i=n("0d51"),o=TypeError;t.exports=function(t){if(r(t))return t;throw new o(i(t)+" is not a function")}},"5a34":function(t,e,n){"use strict";var r=n("44e7"),i=TypeError;t.exports=function(t){if(r(t))throw new i("The method doesn't accept regular expressions");return t}},"5a47":function(t,e,n){"use strict";var r=n("23e7"),i=n("04f8"),o=n("d039"),a=n("7418"),s=n("7b0b"),l=!i||o((function(){a.f(1)}));r({target:"Object",stat:!0,forced:l},{getOwnPropertySymbols:function(t){var e=a.f;return e?e(s(t)):[]}})},"5aff":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"'inji",5:"'inji",8:"'inji",70:"'inji",80:"'inji",2:"'nji",7:"'nji",20:"'nji",50:"'nji",3:"'ünji",4:"'ünji",100:"'ünji",6:"'njy",9:"'unjy",10:"'unjy",30:"'unjy",60:"'ynjy",90:"'ynjy"},n=t.defineLocale("tk",{months:"Ýanwar_Fewral_Mart_Aprel_Maý_Iýun_Iýul_Awgust_Sentýabr_Oktýabr_Noýabr_Dekabr".split("_"),monthsShort:"Ýan_Few_Mar_Apr_Maý_Iýn_Iýl_Awg_Sen_Okt_Noý_Dek".split("_"),weekdays:"Ýekşenbe_Duşenbe_Sişenbe_Çarşenbe_Penşenbe_Anna_Şenbe".split("_"),weekdaysShort:"Ýek_Duş_Siş_Çar_Pen_Ann_Şen".split("_"),weekdaysMin:"Ýk_Dş_Sş_Çr_Pn_An_Şn".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[bugün sagat] LT",nextDay:"[ertir sagat] LT",nextWeek:"[indiki] dddd [sagat] LT",lastDay:"[düýn] LT",lastWeek:"[geçen] dddd [sagat] LT",sameElse:"L"},relativeTime:{future:"%s soň",past:"%s öň",s:"birnäçe sekunt",m:"bir minut",mm:"%d minut",h:"bir sagat",hh:"%d sagat",d:"bir gün",dd:"%d gün",M:"bir aý",MM:"%d aý",y:"bir ýyl",yy:"%d ýyl"},ordinal:function(t,n){switch(n){case"d":case"D":case"Do":case"DD":return t;default:if(0===t)return t+"'unjy";var r=t%10,i=t%100-r,o=t>=100?100:null;return t+(e[r]||e[i]||e[o])}},week:{dow:1,doy:7}});return n}))},"5b01":function(t,e,n){var r=n("8eeb"),i=n("ec69");function o(t,e){return t&&r(e,i(e),t)}t.exports=o},"5b14":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="vasárnap hétfőn kedden szerdán csütörtökön pénteken szombaton".split(" ");function n(t,e,n,r){var i=t;switch(n){case"s":return r||e?"néhány másodperc":"néhány másodperce";case"ss":return i+(r||e)?" másodperc":" másodperce";case"m":return"egy"+(r||e?" perc":" perce");case"mm":return i+(r||e?" perc":" perce");case"h":return"egy"+(r||e?" óra":" órája");case"hh":return i+(r||e?" óra":" órája");case"d":return"egy"+(r||e?" nap":" napja");case"dd":return i+(r||e?" nap":" napja");case"M":return"egy"+(r||e?" hónap":" hónapja");case"MM":return i+(r||e?" hónap":" hónapja");case"y":return"egy"+(r||e?" év":" éve");case"yy":return i+(r||e?" év":" éve")}return""}function r(t){return(t?"":"[múlt] ")+"["+e[this.day()]+"] LT[-kor]"}var i=t.defineLocale("hu",{months:"január_február_március_április_május_június_július_augusztus_szeptember_október_november_december".split("_"),monthsShort:"jan._feb._márc._ápr._máj._jún._júl._aug._szept._okt._nov._dec.".split("_"),monthsParseExact:!0,weekdays:"vasárnap_hétfő_kedd_szerda_csütörtök_péntek_szombat".split("_"),weekdaysShort:"vas_hét_kedd_sze_csüt_pén_szo".split("_"),weekdaysMin:"v_h_k_sze_cs_p_szo".split("_"),longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"YYYY.MM.DD.",LL:"YYYY. MMMM D.",LLL:"YYYY. MMMM D. H:mm",LLLL:"YYYY. MMMM D., dddd H:mm"},meridiemParse:/de|du/i,isPM:function(t){return"u"===t.charAt(1).toLowerCase()},meridiem:function(t,e,n){return t<12?!0===n?"de":"DE":!0===n?"du":"DU"},calendar:{sameDay:"[ma] LT[-kor]",nextDay:"[holnap] LT[-kor]",nextWeek:function(){return r.call(this,!0)},lastDay:"[tegnap] LT[-kor]",lastWeek:function(){return r.call(this,!1)},sameElse:"L"},relativeTime:{future:"%s múlva",past:"%s",s:n,ss:n,m:n,mm:n,h:n,hh:n,d:n,dd:n,M:n,MM:n,y:n,yy:n},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return i}))},"5be5":function(t,e,n){"use strict";t.exports=function(t){var e=t.stateHandler.getState;function n(t){var n=e(t);return n&&!!n.isDetectable}function r(t){e(t).isDetectable=!0}function i(t){return!!e(t).busy}function o(t,n){e(t).busy=!!n}return{isDetectable:n,markAsDetectable:r,isBusy:i,markBusy:o}}},"5c3a":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("zh-cn",{months:"一月_二月_三月_四月_五月_六月_七月_八月_九月_十月_十一月_十二月".split("_"),monthsShort:"1月_2月_3月_4月_5月_6月_7月_8月_9月_10月_11月_12月".split("_"),weekdays:"星期日_星期一_星期二_星期三_星期四_星期五_星期六".split("_"),weekdaysShort:"周日_周一_周二_周三_周四_周五_周六".split("_"),weekdaysMin:"日_一_二_三_四_五_六".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY/MM/DD",LL:"YYYY年M月D日",LLL:"YYYY年M月D日Ah点mm分",LLLL:"YYYY年M月D日ddddAh点mm分",l:"YYYY/M/D",ll:"YYYY年M月D日",lll:"YYYY年M月D日 HH:mm",llll:"YYYY年M月D日dddd HH:mm"},meridiemParse:/凌晨|早上|上午|中午|下午|晚上/,meridiemHour:function(t,e){return 12===t&&(t=0),"凌晨"===e||"早上"===e||"上午"===e?t:"下午"===e||"晚上"===e?t+12:t>=11?t:t+12},meridiem:function(t,e,n){var r=100*t+e;return r<600?"凌晨":r<900?"早上":r<1130?"上午":r<1230?"中午":r<1800?"下午":"晚上"},calendar:{sameDay:"[今天]LT",nextDay:"[明天]LT",nextWeek:function(t){return t.week()!==this.week()?"[下]dddLT":"[本]dddLT"},lastDay:"[昨天]LT",lastWeek:function(t){return this.week()!==t.week()?"[上]dddLT":"[本]dddLT"},sameElse:"L"},dayOfMonthOrdinalParse:/\d{1,2}(日|月|周)/,ordinal:function(t,e){switch(e){case"d":case"D":case"DDD":return t+"日";case"M":return t+"月";case"w":case"W":return t+"周";default:return t}},relativeTime:{future:"%s后",past:"%s前",s:"几秒",ss:"%d 秒",m:"1 分钟",mm:"%d 分钟",h:"1 小时",hh:"%d 小时",d:"1 天",dd:"%d 天",w:"1 周",ww:"%d 周",M:"1 个月",MM:"%d 个月",y:"1 年",yy:"%d 年"},week:{dow:1,doy:4}});return e}))},"5c69":function(t,e,n){var r=n("087d"),i=n("0621");function o(t,e,n,a,s){var l=-1,u=t.length;n||(n=i),s||(s=[]);while(++l<u){var c=t[l];e>0&&n(c)?e>1?o(c,e-1,n,a,s):r(s,c):a||(s[s.length]=c)}return s}t.exports=o},"5c6c":function(t,e,n){"use strict";t.exports=function(t,e){return{enumerable:!(1&t),configurable:!(2&t),writable:!(4&t),value:e}}},"5cbb":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("te",{months:"జనవరి_ఫిబ్రవరి_మార్చి_ఏప్రిల్_మే_జూన్_జులై_ఆగస్టు_సెప్టెంబర్_అక్టోబర్_నవంబర్_డిసెంబర్".split("_"),monthsShort:"జన._ఫిబ్ర._మార్చి_ఏప్రి._మే_జూన్_జులై_ఆగ._సెప్._అక్టో._నవ._డిసె.".split("_"),monthsParseExact:!0,weekdays:"ఆదివారం_సోమవారం_మంగళవారం_బుధవారం_గురువారం_శుక్రవారం_శనివారం".split("_"),weekdaysShort:"ఆది_సోమ_మంగళ_బుధ_గురు_శుక్ర_శని".split("_"),weekdaysMin:"ఆ_సో_మం_బు_గు_శు_శ".split("_"),longDateFormat:{LT:"A h:mm",LTS:"A h:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm",LLLL:"dddd, D MMMM YYYY, A h:mm"},calendar:{sameDay:"[నేడు] LT",nextDay:"[రేపు] LT",nextWeek:"dddd, LT",lastDay:"[నిన్న] LT",lastWeek:"[గత] dddd, LT",sameElse:"L"},relativeTime:{future:"%s లో",past:"%s క్రితం",s:"కొన్ని క్షణాలు",ss:"%d సెకన్లు",m:"ఒక నిమిషం",mm:"%d నిమిషాలు",h:"ఒక గంట",hh:"%d గంటలు",d:"ఒక రోజు",dd:"%d రోజులు",M:"ఒక నెల",MM:"%d నెలలు",y:"ఒక సంవత్సరం",yy:"%d సంవత్సరాలు"},dayOfMonthOrdinalParse:/\d{1,2}వ/,ordinal:"%dవ",meridiemParse:/రాత్రి|ఉదయం|మధ్యాహ్నం|సాయంత్రం/,meridiemHour:function(t,e){return 12===t&&(t=0),"రాత్రి"===e?t<4?t:t+12:"ఉదయం"===e?t:"మధ్యాహ్నం"===e?t>=10?t:t+12:"సాయంత్రం"===e?t+12:void 0},meridiem:function(t,e,n){return t<4?"రాత్రి":t<10?"ఉదయం":t<17?"మధ్యాహ్నం":t<20?"సాయంత్రం":"రాత్రి"},week:{dow:0,doy:6}});return e}))},"5d89":function(t,e,n){var r=n("f8af");function i(t,e){var n=e?r(t.buffer):t.buffer;return new t.constructor(n,t.byteOffset,t.byteLength)}t.exports=i},"5e2e":function(t,e,n){var r=n("28c9"),i=n("69d5"),o=n("b4c0"),a=n("fba5"),s=n("67ca");function l(t){var e=-1,n=null==t?0:t.length;this.clear();while(++e<n){var r=t[e];this.set(r[0],r[1])}}l.prototype.clear=r,l.prototype["delete"]=i,l.prototype.get=o,l.prototype.has=a,l.prototype.set=s,t.exports=l},"5e77":function(t,e,n){"use strict";var r=n("83ab"),i=n("1a2d"),o=Function.prototype,a=r&&Object.getOwnPropertyDescriptor,s=i(o,"name"),l=s&&"something"===function(){}.name,u=s&&(!r||r&&a(o,"name").configurable);t.exports={EXISTS:s,PROPER:l,CONFIGURABLE:u}},"5e7e":function(t,e,n){"use strict";var r,i,o,a,s=n("23e7"),l=n("c430"),u=n("605d"),c=n("da84"),d=n("c65b"),h=n("cb2d"),f=n("d2bb"),p=n("d44e"),m=n("2626"),g=n("59ed"),y=n("1626"),v=n("861d"),_=n("19aa"),b=n("4840"),x=n("2cf4").set,w=n("b575"),M=n("44de"),S=n("e667"),k=n("01b4"),L=n("69f3"),T=n("d256"),D=n("4738"),O=n("f069"),I="Promise",C=D.CONSTRUCTOR,A=D.REJECTION_EVENT,Y=D.SUBCLASSING,j=L.getterFor(I),P=L.set,E=T&&T.prototype,N=T,R=E,H=c.TypeError,z=c.document,F=c.process,B=O.f,W=B,V=!!(z&&z.createEvent&&c.dispatchEvent),G="unhandledrejection",$="rejectionhandled",U=0,q=1,X=2,Z=1,J=2,K=function(t){var e;return!(!v(t)||!y(e=t.then))&&e},Q=function(t,e){var n,r,i,o=e.value,a=e.state===q,s=a?t.ok:t.fail,l=t.resolve,u=t.reject,c=t.domain;try{s?(a||(e.rejection===J&&it(e),e.rejection=Z),!0===s?n=o:(c&&c.enter(),n=s(o),c&&(c.exit(),i=!0)),n===t.promise?u(new H("Promise-chain cycle")):(r=K(n))?d(r,n,l,u):l(n)):u(o)}catch(h){c&&!i&&c.exit(),u(h)}},tt=function(t,e){t.notified||(t.notified=!0,w((function(){var n,r=t.reactions;while(n=r.get())Q(n,t);t.notified=!1,e&&!t.rejection&&nt(t)})))},et=function(t,e,n){var r,i;V?(r=z.createEvent("Event"),r.promise=e,r.reason=n,r.initEvent(t,!1,!0),c.dispatchEvent(r)):r={promise:e,reason:n},!A&&(i=c["on"+t])?i(r):t===G&&M("Unhandled promise rejection",n)},nt=function(t){d(x,c,(function(){var e,n=t.facade,r=t.value,i=rt(t);if(i&&(e=S((function(){u?F.emit("unhandledRejection",r,n):et(G,n,r)})),t.rejection=u||rt(t)?J:Z,e.error))throw e.value}))},rt=function(t){return t.rejection!==Z&&!t.parent},it=function(t){d(x,c,(function(){var e=t.facade;u?F.emit("rejectionHandled",e):et($,e,t.value)}))},ot=function(t,e,n){return function(r){t(e,r,n)}},at=function(t,e,n){t.done||(t.done=!0,n&&(t=n),t.value=e,t.state=X,tt(t,!0))},st=function(t,e,n){if(!t.done){t.done=!0,n&&(t=n);try{if(t.facade===e)throw new H("Promise can't be resolved itself");var r=K(e);r?w((function(){var n={done:!1};try{d(r,e,ot(st,n,t),ot(at,n,t))}catch(i){at(n,i,t)}})):(t.value=e,t.state=q,tt(t,!1))}catch(i){at({done:!1},i,t)}}};if(C&&(N=function(t){_(this,R),g(t),d(r,this);var e=j(this);try{t(ot(st,e),ot(at,e))}catch(n){at(e,n)}},R=N.prototype,r=function(t){P(this,{type:I,done:!1,notified:!1,parent:!1,reactions:new k,rejection:!1,state:U,value:void 0})},r.prototype=h(R,"then",(function(t,e){var n=j(this),r=B(b(this,N));return n.parent=!0,r.ok=!y(t)||t,r.fail=y(e)&&e,r.domain=u?F.domain:void 0,n.state===U?n.reactions.add(r):w((function(){Q(r,n)})),r.promise})),i=function(){var t=new r,e=j(t);this.promise=t,this.resolve=ot(st,e),this.reject=ot(at,e)},O.f=B=function(t){return t===N||t===o?new i(t):W(t)},!l&&y(T)&&E!==Object.prototype)){a=E.then,Y||h(E,"then",(function(t,e){var n=this;return new N((function(t,e){d(a,n,t,e)})).then(t,e)}),{unsafe:!0});try{delete E.constructor}catch(lt){}f&&f(E,R)}s({global:!0,constructor:!0,wrap:!0,forced:C},{Promise:N}),p(N,I,!1,!0),m(I)},"5eed":function(t,e,n){"use strict";var r=n("d256"),i=n("1c7e"),o=n("4738").CONSTRUCTOR;t.exports=o||!i((function(t){r.all(t).then(void 0,(function(){}))}))},"5fbd":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("sv",{months:"januari_februari_mars_april_maj_juni_juli_augusti_september_oktober_november_december".split("_"),monthsShort:"jan_feb_mar_apr_maj_jun_jul_aug_sep_okt_nov_dec".split("_"),weekdays:"söndag_måndag_tisdag_onsdag_torsdag_fredag_lördag".split("_"),weekdaysShort:"sön_mån_tis_ons_tor_fre_lör".split("_"),weekdaysMin:"sö_må_ti_on_to_fr_lö".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY-MM-DD",LL:"D MMMM YYYY",LLL:"D MMMM YYYY [kl.] HH:mm",LLLL:"dddd D MMMM YYYY [kl.] HH:mm",lll:"D MMM YYYY HH:mm",llll:"ddd D MMM YYYY HH:mm"},calendar:{sameDay:"[Idag] LT",nextDay:"[Imorgon] LT",lastDay:"[Igår] LT",nextWeek:"[På] dddd LT",lastWeek:"[I] dddd[s] LT",sameElse:"L"},relativeTime:{future:"om %s",past:"för %s sedan",s:"några sekunder",ss:"%d sekunder",m:"en minut",mm:"%d minuter",h:"en timme",hh:"%d timmar",d:"en dag",dd:"%d dagar",M:"en månad",MM:"%d månader",y:"ett år",yy:"%d år"},dayOfMonthOrdinalParse:/\d{1,2}(\:e|\:a)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?":e":1===e||2===e?":a":":e";return t+n},week:{dow:1,doy:4}});return e}))},6029:function(t,e,n){"use strict";var r=Array.prototype.slice,i=function(t){return t},o=1,a=2,s=3,l=4,u=1e-6;function c(t){return"translate("+t+",0)"}function d(t){return"translate(0,"+t+")"}function h(t){return e=>+t(e)}function f(t,e){return e=Math.max(0,t.bandwidth()-2*e)/2,t.round()&&(e=Math.round(e)),n=>+t(n)+e}function p(){return!this.__axis}function m(t,e){var n=[],m=null,g=null,y=6,v=6,_=3,b="undefined"!==typeof window&&window.devicePixelRatio>1?0:.5,x=t===o||t===l?-1:1,w=t===l||t===a?"x":"y",M=t===o||t===s?c:d;function S(r){var c=null==m?e.ticks?e.ticks.apply(e,n):e.domain():m,d=null==g?e.tickFormat?e.tickFormat.apply(e,n):i:g,S=Math.max(y,0)+_,k=e.range(),L=+k[0]+b,T=+k[k.length-1]+b,D=(e.bandwidth?f:h)(e.copy(),b),O=r.selection?r.selection():r,I=O.selectAll(".domain").data([null]),C=O.selectAll(".tick").data(c,e).order(),A=C.exit(),Y=C.enter().append("g").attr("class","tick"),j=C.select("line"),P=C.select("text");I=I.merge(I.enter().insert("path",".tick").attr("class","domain").attr("stroke","currentColor")),C=C.merge(Y),j=j.merge(Y.append("line").attr("stroke","currentColor").attr(w+"2",x*y)),P=P.merge(Y.append("text").attr("fill","currentColor").attr(w,x*S).attr("dy",t===o?"0em":t===s?"0.71em":"0.32em")),r!==O&&(I=I.transition(r),C=C.transition(r),j=j.transition(r),P=P.transition(r),A=A.transition(r).attr("opacity",u).attr("transform",(function(t){return isFinite(t=D(t))?M(t+b):this.getAttribute("transform")})),Y.attr("opacity",u).attr("transform",(function(t){var e=this.parentNode.__axis;return M((e&&isFinite(e=e(t))?e:D(t))+b)}))),A.remove(),I.attr("d",t===l||t===a?v?"M"+x*v+","+L+"H"+b+"V"+T+"H"+x*v:"M"+b+","+L+"V"+T:v?"M"+L+","+x*v+"V"+b+"H"+T+"V"+x*v:"M"+L+","+b+"H"+T),C.attr("opacity",1).attr("transform",(function(t){return M(D(t)+b)})),j.attr(w+"2",x*y),P.attr(w,x*S).text(d),O.filter(p).attr("fill","none").attr("font-size",10).attr("font-family","sans-serif").attr("text-anchor",t===a?"start":t===l?"end":"middle"),O.each((function(){this.__axis=D}))}return S.scale=function(t){return arguments.length?(e=t,S):e},S.ticks=function(){return n=r.call(arguments),S},S.tickArguments=function(t){return arguments.length?(n=null==t?[]:r.call(t),S):n.slice()},S.tickValues=function(t){return arguments.length?(m=null==t?null:r.call(t),S):m&&m.slice()},S.tickFormat=function(t){return arguments.length?(g=t,S):g},S.tickSize=function(t){return arguments.length?(y=v=+t,S):y},S.tickSizeInner=function(t){return arguments.length?(y=+t,S):y},S.tickSizeOuter=function(t){return arguments.length?(v=+t,S):v},S.tickPadding=function(t){return arguments.length?(_=+t,S):_},S.offset=function(t){return arguments.length?(b=+t,S):b},S}function g(t){return m(a,t)}function y(t){return m(s,t)}function v(t){return m(l,t)}var _=Math.sqrt(50),b=Math.sqrt(10),x=Math.sqrt(2),w=function(t,e,n){var r,i,o,a,s=-1;if(e=+e,t=+t,n=+n,t===e&&n>0)return[t];if((r=e<t)&&(i=t,t=e,e=i),0===(a=M(t,e,n))||!isFinite(a))return[];if(a>0){let n=Math.round(t/a),r=Math.round(e/a);n*a<t&&++n,r*a>e&&--r,o=new Array(i=r-n+1);while(++s<i)o[s]=(n+s)*a}else{a=-a;let n=Math.round(t*a),r=Math.round(e*a);n/a<t&&++n,r/a>e&&--r,o=new Array(i=r-n+1);while(++s<i)o[s]=(n+s)/a}return r&&o.reverse(),o};function M(t,e,n){var r=(e-t)/Math.max(0,n),i=Math.floor(Math.log(r)/Math.LN10),o=r/Math.pow(10,i);return i>=0?(o>=_?10:o>=b?5:o>=x?2:1)*Math.pow(10,i):-Math.pow(10,-i)/(o>=_?10:o>=b?5:o>=x?2:1)}function S(t,e,n){var r=Math.abs(e-t)/Math.max(0,n),i=Math.pow(10,Math.floor(Math.log(r)/Math.LN10)),o=r/i;return o>=_?i*=10:o>=b?i*=5:o>=x&&(i*=2),e<t?-i:i}var k=function(t,e){return t<e?-1:t>e?1:t>=e?0:NaN},L=function(t){let e=t,n=t;function r(t,e,r,i){null==r&&(r=0),null==i&&(i=t.length);while(r<i){const o=r+i>>>1;n(t[o],e)<0?r=o+1:i=o}return r}function i(t,e,r,i){null==r&&(r=0),null==i&&(i=t.length);while(r<i){const o=r+i>>>1;n(t[o],e)>0?i=o:r=o+1}return r}function o(t,n,i,o){null==i&&(i=0),null==o&&(o=t.length);const a=r(t,n,i,o-1);return a>i&&e(t[a-1],n)>-e(t[a],n)?a-1:a}return 1===t.length&&(e=(e,n)=>t(e)-n,n=T(t)),{left:r,center:o,right:i}};function T(t){return(e,n)=>k(t(e),n)}var D=function(t){return null===t?NaN:+t};const O=L(k),I=O.right;O.left,L(D).center;var C=I,A=function(t,e,n){t.prototype=e.prototype=n,n.constructor=t};function Y(t,e){var n=Object.create(t.prototype);for(var r in e)n[r]=e[r];return n}function j(){}var P=.7,E=1/P,N="\\s*([+-]?\\d+)\\s*",R="\\s*([+-]?\\d*\\.?\\d+(?:[eE][+-]?\\d+)?)\\s*",H="\\s*([+-]?\\d*\\.?\\d+(?:[eE][+-]?\\d+)?)%\\s*",z=/^#([0-9a-f]{3,8})$/,F=new RegExp("^rgb\\("+[N,N,N]+"\\)$"),B=new RegExp("^rgb\\("+[H,H,H]+"\\)$"),W=new RegExp("^rgba\\("+[N,N,N,R]+"\\)$"),V=new RegExp("^rgba\\("+[H,H,H,R]+"\\)$"),G=new RegExp("^hsl\\("+[R,H,H]+"\\)$"),$=new RegExp("^hsla\\("+[R,H,H,R]+"\\)$"),U={aliceblue:15792383,antiquewhite:16444375,aqua:65535,aquamarine:8388564,azure:15794175,beige:16119260,bisque:16770244,black:0,blanchedalmond:16772045,blue:255,blueviolet:9055202,brown:10824234,burlywood:14596231,cadetblue:6266528,chartreuse:8388352,chocolate:13789470,coral:16744272,cornflowerblue:6591981,cornsilk:16775388,crimson:14423100,cyan:65535,darkblue:139,darkcyan:35723,darkgoldenrod:12092939,darkgray:11119017,darkgreen:25600,darkgrey:11119017,darkkhaki:12433259,darkmagenta:9109643,darkolivegreen:5597999,darkorange:16747520,darkorchid:10040012,darkred:9109504,darksalmon:15308410,darkseagreen:9419919,darkslateblue:4734347,darkslategray:3100495,darkslategrey:3100495,darkturquoise:52945,darkviolet:9699539,deeppink:16716947,deepskyblue:49151,dimgray:6908265,dimgrey:6908265,dodgerblue:2003199,firebrick:11674146,floralwhite:16775920,forestgreen:2263842,fuchsia:16711935,gainsboro:14474460,ghostwhite:16316671,gold:16766720,goldenrod:14329120,gray:8421504,green:32768,greenyellow:11403055,grey:8421504,honeydew:15794160,hotpink:16738740,indianred:13458524,indigo:4915330,ivory:16777200,khaki:15787660,lavender:15132410,lavenderblush:16773365,lawngreen:8190976,lemonchiffon:16775885,lightblue:11393254,lightcoral:15761536,lightcyan:14745599,lightgoldenrodyellow:16448210,lightgray:13882323,lightgreen:9498256,lightgrey:13882323,lightpink:16758465,lightsalmon:16752762,lightseagreen:2142890,lightskyblue:8900346,lightslategray:7833753,lightslategrey:7833753,lightsteelblue:11584734,lightyellow:16777184,lime:65280,limegreen:3329330,linen:16445670,magenta:16711935,maroon:8388608,mediumaquamarine:6737322,mediumblue:205,mediumorchid:12211667,mediumpurple:9662683,mediumseagreen:3978097,mediumslateblue:8087790,mediumspringgreen:64154,mediumturquoise:4772300,mediumvioletred:13047173,midnightblue:1644912,mintcream:16121850,mistyrose:16770273,moccasin:16770229,navajowhite:16768685,navy:128,oldlace:16643558,olive:8421376,olivedrab:7048739,orange:16753920,orangered:16729344,orchid:14315734,palegoldenrod:15657130,palegreen:10025880,paleturquoise:11529966,palevioletred:14381203,papayawhip:16773077,peachpuff:16767673,peru:13468991,pink:16761035,plum:14524637,powderblue:11591910,purple:8388736,rebeccapurple:6697881,red:16711680,rosybrown:12357519,royalblue:4286945,saddlebrown:9127187,salmon:16416882,sandybrown:16032864,seagreen:3050327,seashell:16774638,sienna:10506797,silver:12632256,skyblue:8900331,slateblue:6970061,slategray:7372944,slategrey:7372944,snow:16775930,springgreen:65407,steelblue:4620980,tan:13808780,teal:32896,thistle:14204888,tomato:16737095,turquoise:4251856,violet:15631086,wheat:16113331,white:16777215,whitesmoke:16119285,yellow:16776960,yellowgreen:10145074};function q(){return this.rgb().formatHex()}function X(){return st(this).formatHsl()}function Z(){return this.rgb().formatRgb()}function J(t){var e,n;return t=(t+"").trim().toLowerCase(),(e=z.exec(t))?(n=e[1].length,e=parseInt(e[1],16),6===n?K(e):3===n?new nt(e>>8&15|e>>4&240,e>>4&15|240&e,(15&e)<<4|15&e,1):8===n?Q(e>>24&255,e>>16&255,e>>8&255,(255&e)/255):4===n?Q(e>>12&15|e>>8&240,e>>8&15|e>>4&240,e>>4&15|240&e,((15&e)<<4|15&e)/255):null):(e=F.exec(t))?new nt(e[1],e[2],e[3],1):(e=B.exec(t))?new nt(255*e[1]/100,255*e[2]/100,255*e[3]/100,1):(e=W.exec(t))?Q(e[1],e[2],e[3],e[4]):(e=V.exec(t))?Q(255*e[1]/100,255*e[2]/100,255*e[3]/100,e[4]):(e=G.exec(t))?at(e[1],e[2]/100,e[3]/100,1):(e=$.exec(t))?at(e[1],e[2]/100,e[3]/100,e[4]):U.hasOwnProperty(t)?K(U[t]):"transparent"===t?new nt(NaN,NaN,NaN,0):null}function K(t){return new nt(t>>16&255,t>>8&255,255&t,1)}function Q(t,e,n,r){return r<=0&&(t=e=n=NaN),new nt(t,e,n,r)}function tt(t){return t instanceof j||(t=J(t)),t?(t=t.rgb(),new nt(t.r,t.g,t.b,t.opacity)):new nt}function et(t,e,n,r){return 1===arguments.length?tt(t):new nt(t,e,n,null==r?1:r)}function nt(t,e,n,r){this.r=+t,this.g=+e,this.b=+n,this.opacity=+r}function rt(){return"#"+ot(this.r)+ot(this.g)+ot(this.b)}function it(){var t=this.opacity;return t=isNaN(t)?1:Math.max(0,Math.min(1,t)),(1===t?"rgb(":"rgba(")+Math.max(0,Math.min(255,Math.round(this.r)||0))+", "+Math.max(0,Math.min(255,Math.round(this.g)||0))+", "+Math.max(0,Math.min(255,Math.round(this.b)||0))+(1===t?")":", "+t+")")}function ot(t){return t=Math.max(0,Math.min(255,Math.round(t)||0)),(t<16?"0":"")+t.toString(16)}function at(t,e,n,r){return r<=0?t=e=n=NaN:n<=0||n>=1?t=e=NaN:e<=0&&(t=NaN),new ut(t,e,n,r)}function st(t){if(t instanceof ut)return new ut(t.h,t.s,t.l,t.opacity);if(t instanceof j||(t=J(t)),!t)return new ut;if(t instanceof ut)return t;t=t.rgb();var e=t.r/255,n=t.g/255,r=t.b/255,i=Math.min(e,n,r),o=Math.max(e,n,r),a=NaN,s=o-i,l=(o+i)/2;return s?(a=e===o?(n-r)/s+6*(n<r):n===o?(r-e)/s+2:(e-n)/s+4,s/=l<.5?o+i:2-o-i,a*=60):s=l>0&&l<1?0:a,new ut(a,s,l,t.opacity)}function lt(t,e,n,r){return 1===arguments.length?st(t):new ut(t,e,n,null==r?1:r)}function ut(t,e,n,r){this.h=+t,this.s=+e,this.l=+n,this.opacity=+r}function ct(t,e,n){return 255*(t<60?e+(n-e)*t/60:t<180?n:t<240?e+(n-e)*(240-t)/60:e)}function dt(t,e,n,r,i){var o=t*t,a=o*t;return((1-3*t+3*o-a)*e+(4-6*o+3*a)*n+(1+3*t+3*o-3*a)*r+a*i)/6}A(j,J,{copy:function(t){return Object.assign(new this.constructor,this,t)},displayable:function(){return this.rgb().displayable()},hex:q,formatHex:q,formatHsl:X,formatRgb:Z,toString:Z}),A(nt,et,Y(j,{brighter:function(t){return t=null==t?E:Math.pow(E,t),new nt(this.r*t,this.g*t,this.b*t,this.opacity)},darker:function(t){return t=null==t?P:Math.pow(P,t),new nt(this.r*t,this.g*t,this.b*t,this.opacity)},rgb:function(){return this},displayable:function(){return-.5<=this.r&&this.r<255.5&&-.5<=this.g&&this.g<255.5&&-.5<=this.b&&this.b<255.5&&0<=this.opacity&&this.opacity<=1},hex:rt,formatHex:rt,formatRgb:it,toString:it})),A(ut,lt,Y(j,{brighter:function(t){return t=null==t?E:Math.pow(E,t),new ut(this.h,this.s,this.l*t,this.opacity)},darker:function(t){return t=null==t?P:Math.pow(P,t),new ut(this.h,this.s,this.l*t,this.opacity)},rgb:function(){var t=this.h%360+360*(this.h<0),e=isNaN(t)||isNaN(this.s)?0:this.s,n=this.l,r=n+(n<.5?n:1-n)*e,i=2*n-r;return new nt(ct(t>=240?t-240:t+120,i,r),ct(t,i,r),ct(t<120?t+240:t-120,i,r),this.opacity)},displayable:function(){return(0<=this.s&&this.s<=1||isNaN(this.s))&&0<=this.l&&this.l<=1&&0<=this.opacity&&this.opacity<=1},formatHsl:function(){var t=this.opacity;return t=isNaN(t)?1:Math.max(0,Math.min(1,t)),(1===t?"hsl(":"hsla(")+(this.h||0)+", "+100*(this.s||0)+"%, "+100*(this.l||0)+"%"+(1===t?")":", "+t+")")}}));var ht=function(t){var e=t.length-1;return function(n){var r=n<=0?n=0:n>=1?(n=1,e-1):Math.floor(n*e),i=t[r],o=t[r+1],a=r>0?t[r-1]:2*i-o,s=r<e-1?t[r+2]:2*o-i;return dt((n-r/e)*e,a,i,o,s)}},ft=function(t){var e=t.length;return function(n){var r=Math.floor(((n%=1)<0?++n:n)*e),i=t[(r+e-1)%e],o=t[r%e],a=t[(r+1)%e],s=t[(r+2)%e];return dt((n-r/e)*e,i,o,a,s)}},pt=t=>()=>t;function mt(t,e){return function(n){return t+n*e}}function gt(t,e,n){return t=Math.pow(t,n),e=Math.pow(e,n)-t,n=1/n,function(r){return Math.pow(t+r*e,n)}}function yt(t){return 1===(t=+t)?vt:function(e,n){return n-e?gt(e,n,t):pt(isNaN(e)?n:e)}}function vt(t,e){var n=e-t;return n?mt(t,n):pt(isNaN(t)?e:t)}var _t=function t(e){var n=yt(e);function r(t,e){var r=n((t=et(t)).r,(e=et(e)).r),i=n(t.g,e.g),o=n(t.b,e.b),a=vt(t.opacity,e.opacity);return function(e){return t.r=r(e),t.g=i(e),t.b=o(e),t.opacity=a(e),t+""}}return r.gamma=t,r}(1);function bt(t){return function(e){var n,r,i=e.length,o=new Array(i),a=new Array(i),s=new Array(i);for(n=0;n<i;++n)r=et(e[n]),o[n]=r.r||0,a[n]=r.g||0,s[n]=r.b||0;return o=t(o),a=t(a),s=t(s),r.opacity=1,function(t){return r.r=o(t),r.g=a(t),r.b=s(t),r+""}}}bt(ht),bt(ft);var xt=function(t,e){e||(e=[]);var n,r=t?Math.min(e.length,t.length):0,i=e.slice();return function(o){for(n=0;n<r;++n)i[n]=t[n]*(1-o)+e[n]*o;return i}};function wt(t){return ArrayBuffer.isView(t)&&!(t instanceof DataView)}function Mt(t,e){var n,r=e?e.length:0,i=t?Math.min(r,t.length):0,o=new Array(i),a=new Array(r);for(n=0;n<i;++n)o[n]=At(t[n],e[n]);for(;n<r;++n)a[n]=e[n];return function(t){for(n=0;n<i;++n)a[n]=o[n](t);return a}}var St=function(t,e){var n=new Date;return t=+t,e=+e,function(r){return n.setTime(t*(1-r)+e*r),n}},kt=function(t,e){return t=+t,e=+e,function(n){return t*(1-n)+e*n}},Lt=function(t,e){var n,r={},i={};for(n in null!==t&&"object"===typeof t||(t={}),null!==e&&"object"===typeof e||(e={}),e)n in t?r[n]=At(t[n],e[n]):i[n]=e[n];return function(t){for(n in r)i[n]=r[n](t);return i}},Tt=/[-+]?(?:\d+\.?\d*|\.?\d+)(?:[eE][-+]?\d+)?/g,Dt=new RegExp(Tt.source,"g");function Ot(t){return function(){return t}}function It(t){return function(e){return t(e)+""}}var Ct=function(t,e){var n,r,i,o=Tt.lastIndex=Dt.lastIndex=0,a=-1,s=[],l=[];t+="",e+="";while((n=Tt.exec(t))&&(r=Dt.exec(e)))(i=r.index)>o&&(i=e.slice(o,i),s[a]?s[a]+=i:s[++a]=i),(n=n[0])===(r=r[0])?s[a]?s[a]+=r:s[++a]=r:(s[++a]=null,l.push({i:a,x:kt(n,r)})),o=Dt.lastIndex;return o<e.length&&(i=e.slice(o),s[a]?s[a]+=i:s[++a]=i),s.length<2?l[0]?It(l[0].x):Ot(e):(e=l.length,function(t){for(var n,r=0;r<e;++r)s[(n=l[r]).i]=n.x(t);return s.join("")})},At=function(t,e){var n,r=typeof e;return null==e||"boolean"===r?pt(e):("number"===r?kt:"string"===r?(n=J(e))?(e=n,_t):Ct:e instanceof J?_t:e instanceof Date?St:wt(e)?xt:Array.isArray(e)?Mt:"function"!==typeof e.valueOf&&"function"!==typeof e.toString||isNaN(e)?Lt:kt)(t,e)},Yt=function(t,e){return t=+t,e=+e,function(n){return Math.round(t*(1-n)+e*n)}};function jt(t){return function(){return t}}function Pt(t){return+t}var Et=[0,1];function Nt(t){return t}function Rt(t,e){return(e-=t=+t)?function(n){return(n-t)/e}:jt(isNaN(e)?NaN:.5)}function Ht(t,e){var n;return t>e&&(n=t,t=e,e=n),function(n){return Math.max(t,Math.min(e,n))}}function zt(t,e,n){var r=t[0],i=t[1],o=e[0],a=e[1];return i<r?(r=Rt(i,r),o=n(a,o)):(r=Rt(r,i),o=n(o,a)),function(t){return o(r(t))}}function Ft(t,e,n){var r=Math.min(t.length,e.length)-1,i=new Array(r),o=new Array(r),a=-1;t[r]<t[0]&&(t=t.slice().reverse(),e=e.slice().reverse());while(++a<r)i[a]=Rt(t[a],t[a+1]),o[a]=n(e[a],e[a+1]);return function(e){var n=C(t,e,1,r)-1;return o[n](i[n](e))}}function Bt(t,e){return e.domain(t.domain()).range(t.range()).interpolate(t.interpolate()).clamp(t.clamp()).unknown(t.unknown())}function Wt(){var t,e,n,r,i,o,a=Et,s=Et,l=At,u=Nt;function c(){var t=Math.min(a.length,s.length);return u!==Nt&&(u=Ht(a[0],a[t-1])),r=t>2?Ft:zt,i=o=null,d}function d(e){return null==e||isNaN(e=+e)?n:(i||(i=r(a.map(t),s,l)))(t(u(e)))}return d.invert=function(n){return u(e((o||(o=r(s,a.map(t),kt)))(n)))},d.domain=function(t){return arguments.length?(a=Array.from(t,Pt),c()):a.slice()},d.range=function(t){return arguments.length?(s=Array.from(t),c()):s.slice()},d.rangeRound=function(t){return s=Array.from(t),l=Yt,c()},d.clamp=function(t){return arguments.length?(u=!!t||Nt,c()):u!==Nt},d.interpolate=function(t){return arguments.length?(l=t,c()):l},d.unknown=function(t){return arguments.length?(n=t,d):n},function(n,r){return t=n,e=r,c()}}function Vt(){return Wt()(Nt,Nt)}function Gt(t,e){switch(arguments.length){case 0:break;case 1:this.range(t);break;default:this.range(e).domain(t);break}return this}var $t=/^(?:(.)?([<>=^]))?([+\-( ])?([$#])?(0)?(\d+)?(,)?(\.\d+)?(~)?([a-z%])?$/i;function Ut(t){if(!(e=$t.exec(t)))throw new Error("invalid format: "+t);var e;return new qt({fill:e[1],align:e[2],sign:e[3],symbol:e[4],zero:e[5],width:e[6],comma:e[7],precision:e[8]&&e[8].slice(1),trim:e[9],type:e[10]})}function qt(t){this.fill=void 0===t.fill?" ":t.fill+"",this.align=void 0===t.align?">":t.align+"",this.sign=void 0===t.sign?"-":t.sign+"",this.symbol=void 0===t.symbol?"":t.symbol+"",this.zero=!!t.zero,this.width=void 0===t.width?void 0:+t.width,this.comma=!!t.comma,this.precision=void 0===t.precision?void 0:+t.precision,this.trim=!!t.trim,this.type=void 0===t.type?"":t.type+""}Ut.prototype=qt.prototype,qt.prototype.toString=function(){return this.fill+this.align+this.sign+this.symbol+(this.zero?"0":"")+(void 0===this.width?"":Math.max(1,0|this.width))+(this.comma?",":"")+(void 0===this.precision?"":"."+Math.max(0,0|this.precision))+(this.trim?"~":"")+this.type};var Xt=function(t){return Math.abs(t=Math.round(t))>=1e21?t.toLocaleString("en").replace(/,/g,""):t.toString(10)};function Zt(t,e){if((n=(t=e?t.toExponential(e-1):t.toExponential()).indexOf("e"))<0)return null;var n,r=t.slice(0,n);return[r.length>1?r[0]+r.slice(2):r,+t.slice(n+1)]}var Jt,Kt,Qt,te,ee=function(t){return t=Zt(Math.abs(t)),t?t[1]:NaN},ne=function(t,e){return Math.max(0,3*Math.max(-8,Math.min(8,Math.floor(ee(e)/3)))-ee(Math.abs(t)))},re=function(t,e){return function(n,r){var i=n.length,o=[],a=0,s=t[0],l=0;while(i>0&&s>0){if(l+s+1>r&&(s=Math.max(1,r-l)),o.push(n.substring(i-=s,i+s)),(l+=s+1)>r)break;s=t[a=(a+1)%t.length]}return o.reverse().join(e)}},ie=function(t){return function(e){return e.replace(/[0-9]/g,(function(e){return t[+e]}))}},oe=function(t){t:for(var e,n=t.length,r=1,i=-1;r<n;++r)switch(t[r]){case".":i=e=r;break;case"0":0===i&&(i=r),e=r;break;default:if(!+t[r])break t;i>0&&(i=0);break}return i>0?t.slice(0,i)+t.slice(e+1):t},ae=function(t,e){var n=Zt(t,e);if(!n)return t+"";var r=n[0],i=n[1],o=i-(Jt=3*Math.max(-8,Math.min(8,Math.floor(i/3))))+1,a=r.length;return o===a?r:o>a?r+new Array(o-a+1).join("0"):o>0?r.slice(0,o)+"."+r.slice(o):"0."+new Array(1-o).join("0")+Zt(t,Math.max(0,e+o-1))[0]},se=function(t,e){var n=Zt(t,e);if(!n)return t+"";var r=n[0],i=n[1];return i<0?"0."+new Array(-i).join("0")+r:r.length>i+1?r.slice(0,i+1)+"."+r.slice(i+1):r+new Array(i-r.length+2).join("0")},le={"%":(t,e)=>(100*t).toFixed(e),b:t=>Math.round(t).toString(2),c:t=>t+"",d:Xt,e:(t,e)=>t.toExponential(e),f:(t,e)=>t.toFixed(e),g:(t,e)=>t.toPrecision(e),o:t=>Math.round(t).toString(8),p:(t,e)=>se(100*t,e),r:se,s:ae,X:t=>Math.round(t).toString(16).toUpperCase(),x:t=>Math.round(t).toString(16)},ue=function(t){return t},ce=Array.prototype.map,de=["y","z","a","f","p","n","µ","m","","k","M","G","T","P","E","Z","Y"],he=function(t){var e=void 0===t.grouping||void 0===t.thousands?ue:re(ce.call(t.grouping,Number),t.thousands+""),n=void 0===t.currency?"":t.currency[0]+"",r=void 0===t.currency?"":t.currency[1]+"",i=void 0===t.decimal?".":t.decimal+"",o=void 0===t.numerals?ue:ie(ce.call(t.numerals,String)),a=void 0===t.percent?"%":t.percent+"",s=void 0===t.minus?"−":t.minus+"",l=void 0===t.nan?"NaN":t.nan+"";function u(t){t=Ut(t);var u=t.fill,c=t.align,d=t.sign,h=t.symbol,f=t.zero,p=t.width,m=t.comma,g=t.precision,y=t.trim,v=t.type;"n"===v?(m=!0,v="g"):le[v]||(void 0===g&&(g=12),y=!0,v="g"),(f||"0"===u&&"="===c)&&(f=!0,u="0",c="=");var _="$"===h?n:"#"===h&&/[boxX]/.test(v)?"0"+v.toLowerCase():"",b="$"===h?r:/[%p]/.test(v)?a:"",x=le[v],w=/[defgprs%]/.test(v);function M(t){var n,r,a,h=_,M=b;if("c"===v)M=x(t)+M,t="";else{t=+t;var S=t<0||1/t<0;if(t=isNaN(t)?l:x(Math.abs(t),g),y&&(t=oe(t)),S&&0===+t&&"+"!==d&&(S=!1),h=(S?"("===d?d:s:"-"===d||"("===d?"":d)+h,M=("s"===v?de[8+Jt/3]:"")+M+(S&&"("===d?")":""),w){n=-1,r=t.length;while(++n<r)if(a=t.charCodeAt(n),48>a||a>57){M=(46===a?i+t.slice(n+1):t.slice(n))+M,t=t.slice(0,n);break}}}m&&!f&&(t=e(t,1/0));var k=h.length+t.length+M.length,L=k<p?new Array(p-k+1).join(u):"";switch(m&&f&&(t=e(L+t,L.length?p-M.length:1/0),L=""),c){case"<":t=h+t+M+L;break;case"=":t=h+L+t+M;break;case"^":t=L.slice(0,k=L.length>>1)+h+t+M+L.slice(k);break;default:t=L+h+t+M;break}return o(t)}return g=void 0===g?6:/[gprs]/.test(v)?Math.max(1,Math.min(21,g)):Math.max(0,Math.min(20,g)),M.toString=function(){return t+""},M}function c(t,e){var n=u((t=Ut(t),t.type="f",t)),r=3*Math.max(-8,Math.min(8,Math.floor(ee(e)/3))),i=Math.pow(10,-r),o=de[8+r/3];return function(t){return n(i*t)+o}}return{format:u,formatPrefix:c}};function fe(t){return Kt=he(t),Qt=Kt.format,te=Kt.formatPrefix,Kt}fe({thousands:",",grouping:[3],currency:["$",""]});var pe=function(t,e){return t=Math.abs(t),e=Math.abs(e)-t,Math.max(0,ee(e)-ee(t))+1},me=function(t){return Math.max(0,-ee(Math.abs(t)))};function ge(t,e,n,r){var i,o=S(t,e,n);switch(r=Ut(null==r?",f":r),r.type){case"s":var a=Math.max(Math.abs(t),Math.abs(e));return null!=r.precision||isNaN(i=ne(o,a))||(r.precision=i),te(r,a);case"":case"e":case"g":case"p":case"r":null!=r.precision||isNaN(i=pe(o,Math.max(Math.abs(t),Math.abs(e))))||(r.precision=i-("e"===r.type));break;case"f":case"%":null!=r.precision||isNaN(i=me(o))||(r.precision=i-2*("%"===r.type));break}return Qt(r)}function ye(t){var e=t.domain;return t.ticks=function(t){var n=e();return w(n[0],n[n.length-1],null==t?10:t)},t.tickFormat=function(t,n){var r=e();return ge(r[0],r[r.length-1],null==t?10:t,n)},t.nice=function(n){null==n&&(n=10);var r,i,o=e(),a=0,s=o.length-1,l=o[a],u=o[s],c=10;u<l&&(i=l,l=u,u=i,i=a,a=s,s=i);while(c-- >0){if(i=M(l,u,n),i===r)return o[a]=l,o[s]=u,e(o);if(i>0)l=Math.floor(l/i)*i,u=Math.ceil(u/i)*i;else{if(!(i<0))break;l=Math.ceil(l*i)/i,u=Math.floor(u*i)/i}r=i}return t},t}function ve(){var t=Vt();return t.copy=function(){return Bt(t,ve())},Gt.apply(t,arguments),ye(t)}function _e(){}var be=function(t){return null==t?_e:function(){return this.querySelector(t)}},xe=function(t){"function"!==typeof t&&(t=be(t));for(var e=this._groups,n=e.length,r=new Array(n),i=0;i<n;++i)for(var o,a,s=e[i],l=s.length,u=r[i]=new Array(l),c=0;c<l;++c)(o=s[c])&&(a=t.call(o,o.__data__,c,s))&&("__data__"in o&&(a.__data__=o.__data__),u[c]=a);return new mr(r,this._parents)},we=function(t){return"object"===typeof t&&"length"in t?t:Array.from(t)};function Me(){return[]}var Se=function(t){return null==t?Me:function(){return this.querySelectorAll(t)}};function ke(t){return function(){var e=t.apply(this,arguments);return null==e?[]:we(e)}}var Le=function(t){t="function"===typeof t?ke(t):Se(t);for(var e=this._groups,n=e.length,r=[],i=[],o=0;o<n;++o)for(var a,s=e[o],l=s.length,u=0;u<l;++u)(a=s[u])&&(r.push(t.call(a,a.__data__,u,s)),i.push(a));return new mr(r,i)},Te=function(t){return function(){return this.matches(t)}};function De(t){return function(e){return e.matches(t)}}var Oe=Array.prototype.find;function Ie(t){return function(){return Oe.call(this.children,t)}}function Ce(){return this.firstElementChild}var Ae=function(t){return this.select(null==t?Ce:Ie("function"===typeof t?t:De(t)))},Ye=Array.prototype.filter;function je(){return this.children}function Pe(t){return function(){return Ye.call(this.children,t)}}var Ee=function(t){return this.selectAll(null==t?je:Pe("function"===typeof t?t:De(t)))},Ne=function(t){"function"!==typeof t&&(t=Te(t));for(var e=this._groups,n=e.length,r=new Array(n),i=0;i<n;++i)for(var o,a=e[i],s=a.length,l=r[i]=[],u=0;u<s;++u)(o=a[u])&&t.call(o,o.__data__,u,a)&&l.push(o);return new mr(r,this._parents)},Re=function(t){return new Array(t.length)},He=function(){return new mr(this._enter||this._groups.map(Re),this._parents)};function ze(t,e){this.ownerDocument=t.ownerDocument,this.namespaceURI=t.namespaceURI,this._next=null,this._parent=t,this.__data__=e}ze.prototype={constructor:ze,appendChild:function(t){return this._parent.insertBefore(t,this._next)},insertBefore:function(t,e){return this._parent.insertBefore(t,e)},querySelector:function(t){return this._parent.querySelector(t)},querySelectorAll:function(t){return this._parent.querySelectorAll(t)}};var Fe=function(t){return function(){return t}};function Be(t,e,n,r,i,o){for(var a,s=0,l=e.length,u=o.length;s<u;++s)(a=e[s])?(a.__data__=o[s],r[s]=a):n[s]=new ze(t,o[s]);for(;s<l;++s)(a=e[s])&&(i[s]=a)}function We(t,e,n,r,i,o,a){var s,l,u,c=new Map,d=e.length,h=o.length,f=new Array(d);for(s=0;s<d;++s)(l=e[s])&&(f[s]=u=a.call(l,l.__data__,s,e)+"",c.has(u)?i[s]=l:c.set(u,l));for(s=0;s<h;++s)u=a.call(t,o[s],s,o)+"",(l=c.get(u))?(r[s]=l,l.__data__=o[s],c.delete(u)):n[s]=new ze(t,o[s]);for(s=0;s<d;++s)(l=e[s])&&c.get(f[s])===l&&(i[s]=l)}function Ve(t){return t.__data__}var Ge=function(t,e){if(!arguments.length)return Array.from(this,Ve);var n=e?We:Be,r=this._parents,i=this._groups;"function"!==typeof t&&(t=Fe(t));for(var o=i.length,a=new Array(o),s=new Array(o),l=new Array(o),u=0;u<o;++u){var c=r[u],d=i[u],h=d.length,f=we(t.call(c,c&&c.__data__,u,r)),p=f.length,m=s[u]=new Array(p),g=a[u]=new Array(p),y=l[u]=new Array(h);n(c,d,m,g,y,f,e);for(var v,_,b=0,x=0;b<p;++b)if(v=m[b]){b>=x&&(x=b+1);while(!(_=g[x])&&++x<p);v._next=_||null}}return a=new mr(a,r),a._enter=s,a._exit=l,a},$e=function(){return new mr(this._exit||this._groups.map(Re),this._parents)},Ue=function(t,e,n){var r=this.enter(),i=this,o=this.exit();return r="function"===typeof t?t(r):r.append(t+""),null!=e&&(i=e(i)),null==n?o.remove():n(o),r&&i?r.merge(i).order():i},qe=function(t){if(!(t instanceof mr))throw new Error("invalid merge");for(var e=this._groups,n=t._groups,r=e.length,i=n.length,o=Math.min(r,i),a=new Array(r),s=0;s<o;++s)for(var l,u=e[s],c=n[s],d=u.length,h=a[s]=new Array(d),f=0;f<d;++f)(l=u[f]||c[f])&&(h[f]=l);for(;s<r;++s)a[s]=e[s];return new mr(a,this._parents)},Xe=function(){for(var t=this._groups,e=-1,n=t.length;++e<n;)for(var r,i=t[e],o=i.length-1,a=i[o];--o>=0;)(r=i[o])&&(a&&4^r.compareDocumentPosition(a)&&a.parentNode.insertBefore(r,a),a=r);return this},Ze=function(t){function e(e,n){return e&&n?t(e.__data__,n.__data__):!e-!n}t||(t=Je);for(var n=this._groups,r=n.length,i=new Array(r),o=0;o<r;++o){for(var a,s=n[o],l=s.length,u=i[o]=new Array(l),c=0;c<l;++c)(a=s[c])&&(u[c]=a);u.sort(e)}return new mr(i,this._parents).order()};function Je(t,e){return t<e?-1:t>e?1:t>=e?0:NaN}var Ke=function(){var t=arguments[0];return arguments[0]=this,t.apply(null,arguments),this},Qe=function(){return Array.from(this)},tn=function(){for(var t=this._groups,e=0,n=t.length;e<n;++e)for(var r=t[e],i=0,o=r.length;i<o;++i){var a=r[i];if(a)return a}return null},en=function(){let t=0;for(const e of this)++t;return t},nn=function(){return!this.node()},rn=function(t){for(var e=this._groups,n=0,r=e.length;n<r;++n)for(var i,o=e[n],a=0,s=o.length;a<s;++a)(i=o[a])&&t.call(i,i.__data__,a,o);return this},on="http://www.w3.org/1999/xhtml",an={svg:"http://www.w3.org/2000/svg",xhtml:on,xlink:"http://www.w3.org/1999/xlink",xml:"http://www.w3.org/XML/1998/namespace",xmlns:"http://www.w3.org/2000/xmlns/"},sn=function(t){var e=t+="",n=e.indexOf(":");return n>=0&&"xmlns"!==(e=t.slice(0,n))&&(t=t.slice(n+1)),an.hasOwnProperty(e)?{space:an[e],local:t}:t};function ln(t){return function(){this.removeAttribute(t)}}function un(t){return function(){this.removeAttributeNS(t.space,t.local)}}function cn(t,e){return function(){this.setAttribute(t,e)}}function dn(t,e){return function(){this.setAttributeNS(t.space,t.local,e)}}function hn(t,e){return function(){var n=e.apply(this,arguments);null==n?this.removeAttribute(t):this.setAttribute(t,n)}}function fn(t,e){return function(){var n=e.apply(this,arguments);null==n?this.removeAttributeNS(t.space,t.local):this.setAttributeNS(t.space,t.local,n)}}var pn=function(t,e){var n=sn(t);if(arguments.length<2){var r=this.node();return n.local?r.getAttributeNS(n.space,n.local):r.getAttribute(n)}return this.each((null==e?n.local?un:ln:"function"===typeof e?n.local?fn:hn:n.local?dn:cn)(n,e))},mn=function(t){return t.ownerDocument&&t.ownerDocument.defaultView||t.document&&t||t.defaultView};function gn(t){return function(){this.style.removeProperty(t)}}function yn(t,e,n){return function(){this.style.setProperty(t,e,n)}}function vn(t,e,n){return function(){var r=e.apply(this,arguments);null==r?this.style.removeProperty(t):this.style.setProperty(t,r,n)}}var _n=function(t,e,n){return arguments.length>1?this.each((null==e?gn:"function"===typeof e?vn:yn)(t,e,null==n?"":n)):bn(this.node(),t)};function bn(t,e){return t.style.getPropertyValue(e)||mn(t).getComputedStyle(t,null).getPropertyValue(e)}function xn(t){return function(){delete this[t]}}function wn(t,e){return function(){this[t]=e}}function Mn(t,e){return function(){var n=e.apply(this,arguments);null==n?delete this[t]:this[t]=n}}var Sn=function(t,e){return arguments.length>1?this.each((null==e?xn:"function"===typeof e?Mn:wn)(t,e)):this.node()[t]};function kn(t){return t.trim().split(/^|\s+/)}function Ln(t){return t.classList||new Tn(t)}function Tn(t){this._node=t,this._names=kn(t.getAttribute("class")||"")}function Dn(t,e){var n=Ln(t),r=-1,i=e.length;while(++r<i)n.add(e[r])}function On(t,e){var n=Ln(t),r=-1,i=e.length;while(++r<i)n.remove(e[r])}function In(t){return function(){Dn(this,t)}}function Cn(t){return function(){On(this,t)}}function An(t,e){return function(){(e.apply(this,arguments)?Dn:On)(this,t)}}Tn.prototype={add:function(t){var e=this._names.indexOf(t);e<0&&(this._names.push(t),this._node.setAttribute("class",this._names.join(" ")))},remove:function(t){var e=this._names.indexOf(t);e>=0&&(this._names.splice(e,1),this._node.setAttribute("class",this._names.join(" ")))},contains:function(t){return this._names.indexOf(t)>=0}};var Yn=function(t,e){var n=kn(t+"");if(arguments.length<2){var r=Ln(this.node()),i=-1,o=n.length;while(++i<o)if(!r.contains(n[i]))return!1;return!0}return this.each(("function"===typeof e?An:e?In:Cn)(n,e))};function jn(){this.textContent=""}function Pn(t){return function(){this.textContent=t}}function En(t){return function(){var e=t.apply(this,arguments);this.textContent=null==e?"":e}}var Nn=function(t){return arguments.length?this.each(null==t?jn:("function"===typeof t?En:Pn)(t)):this.node().textContent};function Rn(){this.innerHTML=""}function Hn(t){return function(){this.innerHTML=t}}function zn(t){return function(){var e=t.apply(this,arguments);this.innerHTML=null==e?"":e}}var Fn=function(t){return arguments.length?this.each(null==t?Rn:("function"===typeof t?zn:Hn)(t)):this.node().innerHTML};function Bn(){this.nextSibling&&this.parentNode.appendChild(this)}var Wn=function(){return this.each(Bn)};function Vn(){this.previousSibling&&this.parentNode.insertBefore(this,this.parentNode.firstChild)}var Gn=function(){return this.each(Vn)};function $n(t){return function(){var e=this.ownerDocument,n=this.namespaceURI;return n===on&&e.documentElement.namespaceURI===on?e.createElement(t):e.createElementNS(n,t)}}function Un(t){return function(){return this.ownerDocument.createElementNS(t.space,t.local)}}var qn=function(t){var e=sn(t);return(e.local?Un:$n)(e)},Xn=function(t){var e="function"===typeof t?t:qn(t);return this.select((function(){return this.appendChild(e.apply(this,arguments))}))};function Zn(){return null}var Jn=function(t,e){var n="function"===typeof t?t:qn(t),r=null==e?Zn:"function"===typeof e?e:be(e);return this.select((function(){return this.insertBefore(n.apply(this,arguments),r.apply(this,arguments)||null)}))};function Kn(){var t=this.parentNode;t&&t.removeChild(this)}var Qn=function(){return this.each(Kn)};function tr(){var t=this.cloneNode(!1),e=this.parentNode;return e?e.insertBefore(t,this.nextSibling):t}function er(){var t=this.cloneNode(!0),e=this.parentNode;return e?e.insertBefore(t,this.nextSibling):t}var nr=function(t){return this.select(t?er:tr)},rr=function(t){return arguments.length?this.property("__data__",t):this.node().__data__};function ir(t){return function(e){t.call(this,e,this.__data__)}}function or(t){return t.trim().split(/^|\s+/).map((function(t){var e="",n=t.indexOf(".");return n>=0&&(e=t.slice(n+1),t=t.slice(0,n)),{type:t,name:e}}))}function ar(t){return function(){var e=this.__on;if(e){for(var n,r=0,i=-1,o=e.length;r<o;++r)n=e[r],t.type&&n.type!==t.type||n.name!==t.name?e[++i]=n:this.removeEventListener(n.type,n.listener,n.options);++i?e.length=i:delete this.__on}}}function sr(t,e,n){return function(){var r,i=this.__on,o=ir(e);if(i)for(var a=0,s=i.length;a<s;++a)if((r=i[a]).type===t.type&&r.name===t.name)return this.removeEventListener(r.type,r.listener,r.options),this.addEventListener(r.type,r.listener=o,r.options=n),void(r.value=e);this.addEventListener(t.type,o,n),r={type:t.type,name:t.name,value:e,listener:o,options:n},i?i.push(r):this.__on=[r]}}var lr=function(t,e,n){var r,i,o=or(t+""),a=o.length;if(!(arguments.length<2)){for(s=e?sr:ar,r=0;r<a;++r)this.each(s(o[r],e,n));return this}var s=this.node().__on;if(s)for(var l,u=0,c=s.length;u<c;++u)for(r=0,l=s[u];r<a;++r)if((i=o[r]).type===l.type&&i.name===l.name)return l.value};function ur(t,e,n){var r=mn(t),i=r.CustomEvent;"function"===typeof i?i=new i(e,n):(i=r.document.createEvent("Event"),n?(i.initEvent(e,n.bubbles,n.cancelable),i.detail=n.detail):i.initEvent(e,!1,!1)),t.dispatchEvent(i)}function cr(t,e){return function(){return ur(this,t,e)}}function dr(t,e){return function(){return ur(this,t,e.apply(this,arguments))}}var hr=function(t,e){return this.each(("function"===typeof e?dr:cr)(t,e))},fr=function*(){for(var t=this._groups,e=0,n=t.length;e<n;++e)for(var r,i=t[e],o=0,a=i.length;o<a;++o)(r=i[o])&&(yield r)},pr=[null];function mr(t,e){this._groups=t,this._parents=e}function gr(){return new mr([[document.documentElement]],pr)}function yr(){return this}mr.prototype=gr.prototype={constructor:mr,select:xe,selectAll:Le,selectChild:Ae,selectChildren:Ee,filter:Ne,data:Ge,enter:He,exit:$e,join:Ue,merge:qe,selection:yr,order:Xe,sort:Ze,call:Ke,nodes:Qe,node:tn,size:en,empty:nn,each:rn,attr:pn,style:_n,property:Sn,classed:Yn,text:Nn,html:Fn,raise:Wn,lower:Gn,append:Xn,insert:Jn,remove:Qn,clone:nr,datum:rr,on:lr,dispatch:hr,[Symbol.iterator]:fr};var vr=function(t){return"string"===typeof t?new mr([[document.querySelector(t)]],[document.documentElement]):new mr([[t]],pr)};function _r(){return window.devicePixelRatio||1}function br(t=12,e="normal",n="Helvetica Neue"){return`${e} ${t*_r()}px ${n}`}function xr(t,e,n,{color:r,lineWidth:i}){const o=_r();wr(t,`M ${e.x*o} ${e.y*o} L ${n.x*o} ${n.y*o}`,{lineWidth:i||1,strokeStyle:r||"#FF0000"})}function wr(t,e,{lineWidth:n,lineDash:r,strokeStyle:i,fillStyle:o}){t.beginPath(),r&&t.setLineDash(r),t.lineWidth=n||_r(),t.strokeStyle=i||"none",t.fillStyle=o||"none",t[i?"stroke":"fill"](new Path2D(e))}function Mr(t){let e="string"==typeof t?t:"";if(n=t,Number.isInteger(Number(n))){const n=t.toString(16).padStart(8,"F"),r=parseInt(n.slice(0,2),16)/255;e=`rgba(${parseInt(n.slice(2,4),16)}, ${parseInt(n.slice(4,6),16)}, ${parseInt(n.slice(6),16)}, ${r})`}var n;return e}class Sr{constructor({chart:t,boardId:e,width:n,height:r,align:i,container:o,priceDecs:a}){this.chart=t,this.width=n,this.height=r,this.boardId=e,this.container=o.dom,this.align=i||"right",this.priceDecs=a||0,this.scale=ve().rangeRound([this.height,0]),this.axis="left"===this.align?v(this.scale):g(this.scale),this.rootElement=vr(this.container).append("svg").attr("class","axis").attr("width",this.width).attr("height",this.height),this.rootG=this.rootElement.append("g").attr("class","y axis "+this.align),this.annotationG=this.rootElement.append("g"),this.annotationRect=this.annotationG.append("rect").attr("fill","#fbd3a0"),this.annotationText=this.annotationG.append("text"),"left"===this.align&&this.rootG.attr("transform","translate("+this.width+",0)")}update({width:t,height:e}){t&&t!=this.width&&(this.width=t),e&&e!=this.height&&(this.height=e),this.scale.rangeRound([this.height,0]),this.rootElement.attr("width",this.width).attr("height",this.height)}getTicks(){const t=[],e=this;return this.rootG.selectAll(".tick").each((function(n){t.push(Math.round(e.scale(n)))})),t}draw(t){this.scale.domain(t),this.axis.ticks("main"===this.boardId?6:3,`.${this.priceDecs}f`),this.rootG.call(this.axis)}showYAnnotation(t){const e=this.scale.invert(t).toFixed(this.priceDecs);isNaN(e)?this.annotationG.attr("visibility","hidden"):(this.annotationG.attr("visibility","visible"),this.annotationRect.attr("width",this.width-2).attr("height",12).attr("transform","translate(1,"+(t-6)+")"),this.annotationText.text(e).attr("transform","translate(3,"+(t+5)+")"))}hideYAnnotation(){this.annotationG.attr("visibility","hidden")}}class kr{constructor(t,e,n){this.fn=t,this.context=e,this.once=n||!1}}class Lr{constructor(){this._events={}}addListener(t,e,n,r){if("function"!=typeof e)throw new TypeError("The listener must be a function");const i=new kr(e,n||this,r||!1);this._events[t]?this._events[t].push(i):this._events[t]=[i]}once(t,e,n){this.addListener(t,e,n||this,!0)}removeListener(t,e){if("function"!=typeof e)throw new TypeError("The listener must be a function");if(!this._events[t])return;let n=!1;for(const r in this._events[t])e==this._events[t][r].fn&&(n=!0,this._events[t][r]=void 0);n&&(this._events[t]=this._events[t].filter(t=>t)),0==this._events[t].length&&delete this._events[t]}removeAllListeners(t){this._events[t]&&delete this._events[t]}emit(t,...e){if(!this._events[t])return;let n=!1;for(const r in this._events[t]){const i=this._events[t][r];i.once&&(n=!0,this._events[t][r]=void 0),i.fn.call(i.context,...e)}n&&(this._events[t]=this._events[t].filter(t=>t),0==this._events[t].length&&delete this._events[t])}getEventNames(){return Object.keys(this._events)}getListeners(t){if(!this._events[t])return[];const e=[];for(const n of this._events[t])e.push(n);return e}}Lr.prototype.off=Lr.prototype.removeListener,Lr.prototype.on=Lr.prototype.addListener;class Tr{constructor({id:t,chart:e,board:n,yAlign:r}){this.id=t||(new Date).getTime(),this.chart=e,this.board=n,this.boardId=this.board.boardId,this.yAlign=r||"right",this._show=!0}update(t){this.xPos1=void 0===t.xPos1?this.xPos1:t.xPos1,this.x1=void 0===t.x1?this.x1:t.x1,this.dt1=void 0===t.dt1?this.dt1:t.dt1,this.yPos1=void 0===t.yPos1?this.yPos1:t.yPos1,this.y1=void 0===t.y1?this.y1:t.y1,this.xPos2=void 0===t.xPos2?this.xPos2:t.xPos2,this.x2=void 0===t.x2?this.x2:t.x2,this.dt2=void 0===t.dt2?this.dt2:t.dt2,this.yPos2=void 0===t.yPos2?this.yPos2:t.yPos2,this.y2=void 0===t.y2?this.y2:t.y2}show(){this._show||(this._show=!0)}hide(){this._show&&(this._show=!1)}revertSomeToPos(t){const e={x:null,y:null};return Number.isFinite(this["xPos"+t])?e.x=this["xPos"+t]:Number.isFinite(this["x"+t])?e.x=this.revertIdToPos(this["x"+t]):Number.isFinite(this["dt"+t])&&(e.x=this.revertIdToPos(this.revertDtToId(this["dt"+t]))),Number.isFinite(this["yPos"+t])?e.y=this["yPos"+t]:Number.isFinite(this["y"+t])&&(e.y=this.revertValueToYPos(this["y"+t])),e}revertDtToId(t){const e=this.chart.mainSeries,[n,r]=[this.chart.range.leftId,this.chart.range.rightId];if(!e||!e.data||!e.data[n]||t<e.data[n].datetime)return null;const i=e.data[r]?r:e.last_id;if(t<=e.data[i].datetime)for(let o=n;o<i+1;o++)if(t-e.data[o].datetime<=0)return o;return null}revertIdToPos(t){return Math.round(this.chart.timeAxis.scale(t))+Math.floor(this.chart.bar.barWidth/2)}revertValueToYPos(t){return t&&this.board[this.yAlign+"YAxis"]?Math.round(this.board[this.yAlign+"YAxis"].scale(t)):null}isPointInChart(t){const{x:e,y:n}=t;return e>=0&&e<=this.board.width&&n>=0&&n<=this.board.height}}class Dr extends Tr{constructor(t){super(t),this.type="text",this.color=t.color||"#000000",this.xPos1=t.xPos1,this.x1=t.x1,this.dt1=t.dt1,this.yPos1=t.yPos1,this.y1=t.y1,this.text=t.text,this.size=t.size||12}update(t){super.update(t),this.text=void 0===t.text?this.text:t.text}draw(t){if(!this._show)return;const e=this.chart.bar,n=this.x1*e.step,r=this.revertValueToYPos(this.y1),i=_r();t.fillStyle=this.color,t.font=br(this.size),t.fillText(this.text,n*i,(r+this.size)*i)}}class Or extends Tr{constructor(t){super(t),this.type="line",this.color=t.color,this.lineWidth=t.lineWidth,this.xPos1=t.xPos1,this.x1=t.x1,this.dt1=t.dt1,this.yPos1=t.yPos1,this.y1=t.y1,this.xPos2=t.xPos2,this.x2=t.x2,this.dt2=t.dt2,this.yPos2=t.yPos2,this.y2=t.y2,this.text=t.text,this.clamp=t.clamp}update(t){super.update(t),this.x1=t.x1||this.x1,this.y1=t.y1||this.y1,this.text=void 0===t.text?this.text:t.text,this.clamp=void 0===t.clamp?this.clamp:t.clamp}draw(t){if(!this._show)return;const e=this.chart.bar;xr(t,{x:this.x1*e.step,y:this.revertValueToYPos(this.y1)},{x:this.x2*e.step,y:this.revertValueToYPos(this.y2)},{color:this.color,lineWidth:this.lineWidth})}}class Ir extends Tr{constructor(t){super(t),this.type="h-line",this.color=t.stroke,this.lineWidth=t.lineWidth,this.y1=t.y1,this.text=t.text||""}update(t){super.update(t),this.y1=t.y1||this.y1,this.text=t.text||this.text}draw(t){if(!this._show)return;const e=this.revertValueToYPos(this.y1),n=this.chart.chartCtrl.pixelLeft,r=this.chart.chartCtrl.pixelLeft+this.board.width,i=_r();t.font=br(12),xr(t,{x:n,y:e},{x:r,y:e},{color:this.color,lineWidth:this.lineWidth}),t.fillStyle=this.color,t.fillText(this.text,n*i,(e-1)*i)}}class Cr extends Tr{constructor(t){super(t),this.type="rect",this.fill=t.fill||"none",this.color=t.color||"#bbbbbb",this.lineWidth=t.lineWidth||1,this.xPos1=t.xPos1,this.x1=t.x1,this.dt1=t.dt1,this.yPos1=t.yPos1,this.y1=t.y1,this.xPos2=t.xPos2,this.x2=t.x2,this.dt2=t.dt2,this.yPos2=t.yPos2,this.y2=t.y2,this.width=t.width,this.height=t.height}update(t){super.update(t),this.width=void 0===t.width?this.width:t.width,this.height=void 0===t.height?this.height:t.height}draw(t){if(!this._show)return;const e=this.chart.bar,n=this.x1*e.step,r=this.revertValueToYPos(this.y1),i=this.x2*e.step+this.chart.bar.barWidth-n,o=this.revertValueToYPos(this.y2)-r;t.lineWidth=this.lineWidth,t.fillStyle=this.fill||"blue",t.strokeStyle=this.color||"red";const a=_r();t.rect(n*a,r*a,i*a,o*a),t.fill(),t.stroke()}}class Ar extends Tr{constructor(t){super(t),this.type="hightlight-bar",this.fillStyle=t.fill||"#d0edfd",this.x1=t.x1}update(t){super.update(t),this.x1=t.x1||this.x1}draw(t){if(!this._show)return;const e=this.chart.bar,n=this.x1*e.step,r=e.barWidth,i=this.board.height;t.beginPath(),t.fillStyle=this.fillStyle;const o=_r();t.rect(n*o,0,r*o,i*o),t.fill()}}class Yr extends Tr{constructor(t){super(t),this.type="arrow",this.x1=t.x1,this.width=t.width,this.height=t.height,this.direction=t.direction}update(t){super.update(t),this.width=void 0===t.width?this.width:t.width,this.height=void 0===t.height?this.height:t.height,this.direction=void 0===t.direction?this.direction:t.direction}draw(){const t=this.chart.bar,e={x:this.x1*t.step,y:this.revertValueToYPos(this.y1)},n=this.width||t.barWidth,r=this.height||t.barWidth,i="up"===this.direction?1:-1,o="none"===this.stroke?0:.5;if(Number.isFinite(e.x)&&Number.isFinite(e.y)&&this.isPointInChart(e)){const t=Math.round(n/3),a=`M ${e.x+n/2} ${e.y} l ${-n/2} ${i*r/2} l ${t-o} 0 l 0 ${i*r/2} l ${t} 0 l 0 ${-1*i*r/2} l ${t} 0 z`;this.arrowG.attr("d",a),this.show()}else this.hide()}}class jr{constructor(t,e,n){this.chart=t,this.board=e,this.canvas=n,this._showAll=!0,this.marks={}}addMark(t){return t.id=t.id||(new Date).getTime(),this.marks[t.id]?this.marks[t.id].update(t):(t.chart=t.chart||this.chart,t.boardId=t.boardId||"main",t.board=t.board||this.chart.boards[t.boardId],t.yAlign=t.yAlign||"right","line"===t.type?this.marks[t.id]=new Or(t):"h-line"===t.type?this.marks[t.id]=new Ir(t):"rect"===t.type?this.marks[t.id]=new Cr(t):"arrow"===t.type?this.marks[t.id]=new Yr(t):"text"===t.type?this.marks[t.id]=new Dr(t):"hightlight-bar"===t.type&&(this.marks[t.id]=new Ar(t))),this.marks[t.id]}removeMarkAll(){this.marks={}}removeMark(t){this.marks[t]&&delete this.marks[t]}showMark(t){t&&this.marks[t]&&this.marks[t].show()}hideMark(t){t&&this.marks[t]&&this.marks[t].hide()}showAllMark(){this._showAll||(this._showAll=!0)}hideAllMark(){this._showAll&&(this._showAll=!1)}draw(){if(this._showAll)for(const t in this.marks)this.marks[t].draw(this.canvas.ctx)}}class Pr{constructor({width:t,height:e}){this.dom=document.createElement("table"),this.dom.setAttribute("cellpading","0"),this.dom.setAttribute("cellspacing","0"),this.dom.setAttribute("class","tqchart-table"),this.width=t||500,this.height=e||300}resize({width:t,height:e}){this.width=t||this.width,this.height=e||this.height,this.dom.style.width=this.width+"px",this.dom.style.height=this.height+"px"}}class Er{constructor(t){this.dom=document.createElement("tr");for(const e in t)this.dom.setAttribute(e,t[e])}}class Nr{constructor({width:t,height:e,minWidth:n},r){this.dom=document.createElement("td"),this.dom.style.width=t+"px",this.dom.style.height=e+"px",this.width=t||400,this.height=e||150,n&&(this.dom.style.minWidth=n+"px");for(const i in r)this.dom.setAttribute(i,r[i])}resize({width:t,height:e}){this.width=t||this.width,this.height=e||this.height,this.dom.style.width=this.width+"px",this.dom.style.height=this.height+"px",this.dom.style.minWidth=this.width+"px"}}class Rr{constructor({width:t,height:e},n){this.dom=document.createElement("canvas"),this.ctx=this.dom.getContext("2d"),this.width=t||300,this.height=e||150,this.resize({width:t,height:e});for(const r in n)this.dom.setAttribute(r,n[r])}resize({width:t,height:e}){this.width=t||this.width,this.height=e||this.height;const n=_r();this.dom.width=this.width*n,this.dom.height=this.height*n,this.dom.style.width=this.width+"px",this.dom.style.height=this.height+"px"}clear(){this.clearRect({width:this.width,height:this.height})}clearRect({width:t,height:e}){const n=_r();t=t||this.width,e=e||this.height,this.ctx.clearRect(0,0,t*n,e*n)}}class Hr extends Lr{constructor({boardId:t,chart:e,container:n,width:r,height:i,leftWidth:o,rightWidth:a,rightPriceTick:s,rightPriceDecs:l,leftPriceTick:u,leftPriceDecs:c}){super(),this.boardId=t,this.chart=e,this.container=n,this.width=r||0,this.height=i||0,this.leftWidth=o||0,this.rightWidth=a||0,this._initElement(),this._initYAxis(l,c),this._initManagers(),this.show=!0}_initElement(){this._rootElement=new Er({class:"board-container"}),this._leftTd=new Nr({width:this.leftWidth,height:this.height,minWidth:this.leftWidth},{class:"left-y-container"}),this._centerTd=new Nr({width:this.width-this.leftWidth-this.rightWidth,height:this.height},{class:"canvas-container"}),this._centerCanvas=new Rr({width:this.width-this.leftWidth-this.rightWidth,height:this.height},{class:"plot-canvas"}),this._centerTopCanvas=new Rr({width:this.width-this.leftWidth-this.rightWidth,height:this.height},{class:"plot-top-canvas","data-board-id":this.boardId}),this._centerTd.dom.appendChild(this._centerCanvas.dom),this._centerTd.dom.appendChild(this._centerTopCanvas.dom),this._rightTd=new Nr({width:this.rightWidth,height:this.height,minWidth:this.rightWidth},{class:"right-y-container"}),this._rootElement.dom.appendChild(this._leftTd.dom),this._rootElement.dom.appendChild(this._centerTd.dom),this._rootElement.dom.appendChild(this._rightTd.dom)}_initYAxis(t,e){this._rightTd&&(this.rightYAxis=new Sr({chart:this.chart,width:this.rightWidth,height:this.height,boardId:this.boardId,container:this._rightTd,align:"right",priceDecs:t})),this._leftTd&&(this.leftYAxis=new Sr({chart:this.chart,width:this.leftWidth,height:this.height,boardId:this.boardId,container:this._leftTd,align:"left",priceDecs:e}))}_initManagers(){this.marksManager=new jr(this.chart,this,this._centerCanvas),this.hightlightManager=new jr(this.chart,this,this._centerCanvas)}setDomain(t){this.domain=t}setPriceDecs(t){this.rightYAxis.priceDecs=t}drawYAnnotation(t,e){e?(this.leftYAxis.showYAnnotation(t),this.rightYAxis.showYAnnotation(t)):(this.leftYAxis.hideYAnnotation(),this.rightYAxis.hideYAnnotation())}addMark(t){"back"===t.layer?this.hightlightManager.addMark(t):this.marksManager.addMark(t)}removeMark(t){this.hightlightManager.removeMark(t),this.marksManager.removeMark(t)}removeMarkAll(){this.hightlightManager.removeMarkAll(),this.marksManager.removeMarkAll()}update({width:t,height:e}){if((t<=-1||e<=-1)&&this.show?(this.show=!1,this._rootElement.dom.previousElementSibling.getAttribute("class").indexOf("separator-container")>-1&&this._rootElement.dom.previousElementSibling.setAttribute("class","separator-container hide"),this._rootElement.dom.setAttribute("class","board-container hide"),this._centerCanvas.clear(),this._centerTopCanvas.clear()):!this.show&&t>0&&e>0&&(this.show=!0,this._rootElement.dom.previousElementSibling.getAttribute("class").indexOf("separator-container")>-1&&this._rootElement.dom.previousElementSibling.setAttribute("class","separator-container"),this._rootElement.dom.setAttribute("class","board-container")),t>0||e>0){t>0&&this.width!=t&&(this.width=t),e>0&&this.height!=e&&(this.height=e,this._leftTd.resize({width:this.leftWidth,height:this.height}),this._rightTd.resize({width:this.rightWidth,height:this.height}),this.leftYAxis&&this.leftYAxis.update({width:this.leftWidth,height:this.height}),this.rightYAxis&&this.rightYAxis.update({width:this.rightWidth,height:this.height}));const n=this.width-this.leftWidth-this.rightWidth;this._centerTd.resize({width:n,height:this.height}),this._centerCanvas.resize({width:n,height:this.height}),this._centerTopCanvas.resize({width:n,height:this.height})}}drawClearRect(){this.show&&this._centerCanvas.clear()}drawYAxis(){this.show}drawBackground(){this.show&&this.hightlightManager.draw()}drawFrontMark(){this.show&&this.marksManager.draw()}hideCrosshair(){this._centerTopCanvas.clear()}drawTitleText(t,e){const n=this._centerTopCanvas.ctx,r=_r();n.font=br(14),n.fillStyle="#000000";let i="";const o=this.chart.bar.getKlineId(t);if(this.chart.mainSeries&&this.chart.mainSeries.data[o])for(const a in this.chart.plots){const t=this.chart.plots[a];t.boardId===this.boardId&&(i+=t.getText(o))}n.fillText(i,1*r,15*r)}drawCrosshair(t,e){if(!this.show)return;this._centerTopCanvas.clear();const n=this._centerTopCanvas.ctx,r=_r();n.save(),n.lineWidth=1*r,n.setLineDash([5,5]),n.strokeStyle="#666666",n.beginPath(),n.moveTo(t*r,0*r),n.lineTo(t*r,this.height*r),n.moveTo(0*r,e*r),n.lineTo(this.width*r,e*r),n.stroke(),n.restore()}drawCrosshairVertical(t){this._centerTopCanvas.clear();const e=this._centerTopCanvas.ctx,n=_r();e.save(),e.lineWidth=1*n,e.setLineDash([5,5]),e.strokeStyle="#666666",e.beginPath(),e.moveTo(t*n,0*n),e.lineTo(t*n,this.height*n),e.stroke(),e.restore()}}class zr extends Hr{_initElement(){this._rootElement=new Er({class:"board-container"}),this._leftTd=new Nr({width:this.leftWidth,height:this.height}),this._centerTd=new Nr({width:this.width-this.leftWidth-this.rightWidth,height:this.height}),this._rightTd=new Nr({width:this.rightWidth,height:this.height}),this._rootElement.dom.appendChild(this._leftTd.dom),this._rootElement.dom.appendChild(this._centerTd.dom),this._rootElement.dom.appendChild(this._rightTd.dom)}_initYAxis(){}_initManagers(){}}class Fr{constructor(){this._rootElement=document.createElement("tr"),this._rootElement.setAttribute("class","separator-container"),this._td=document.createElement("td"),this._td.setAttribute("class","separator"),this._td.setAttribute("colspan","3"),this._separatorDiv=document.createElement("div"),this._separatorDiv.setAttribute("class","handle"),this._td.appendChild(this._separatorDiv),this._rootElement.appendChild(this._td)}}class Br extends Lr{init({chart:t,width:e,height:n,margin:r}){this.chart=t,this.width=e,this.height=n,this.margin=r,this.table=new Pr({width:this.width,height:this.height}),this.root=this.table.dom,this.boards=[],this.initX=null,this.isClick=!0,this.isDraging=!1,this.dragTimeout=null,this._initEventListeners(),this.timeAxisBoard=new zr({chart:this.chart,width:this.width,height:this.margin.bottom,leftWidth:this.margin.left,rightWidth:this.margin.right}),this.table.dom.appendChild(this.timeAxisBoard._rootElement.dom)}getBoardContainerByBoardId(t){for(const e of this.boards)if(t===e.boardId)return e;return null}update(){const t=function(t,e,n){const r=e-1*(t-1),i=[];if(n){i[0]={top:0,height:r};for(let e=1;e<t;e++)i[e]={top:-1,height:-1}}else{const e=1/t,n=Math.ceil(r*e*.7),o=r-n*(t-1);i[0]={top:0,height:o};for(let r=1;r<t;r++)i[r]={height:n,top:i[r-1].top+i[r-1].height+1}}return i}(this.boards.length,this.height-this.margin.bottom,this.chart.onlyMainBoard||!1);for(let e=0;e<t.length;e++)this.boards[e].update({width:this.width,height:t[e].height})}resize({width:t,height:e}){this.width=t,this.height=e,this.table.resize({width:this.width,height:this.height}),this.update()}addBoard(t){for(const e of this.boards)if(e.boardId===t.boardId)return;if(this.boards.push(t),this.boards.length>1){const t=new Fr;this.table.dom.insertBefore(t._rootElement,this.timeAxisBoard._rootElement.dom)}return this.table.dom.insertBefore(t._rootElement.dom,this.timeAxisBoard._rootElement.dom),this.update(),t}_getBoardIdByEvent(t){return"CANVAS"==t.target.nodeName&&"plot-top-canvas"==t.target.getAttribute("class")&&t.target.dataset.boardId}_initEventListeners(){const t=this;this.root.addEventListener("mouseup",(function(e){e.preventDefault(),t.dragTimeout&&(clearTimeout(t.dragTimeout),t.dragTimeout=null)}),!1),this.root.addEventListener("mousemove",(function(e){e.preventDefault();const n=t._getBoardIdByEvent(e);n&&(t.isDraging?t.emit("dragmove",n,e):t.emit("move",n,e))}),!1);const e=function(n){n.preventDefault(),n.stopPropagation(),t.isDraging&&(t.isClick=!1,t.isDraging=!1,t.emit("dragend",t.boardId,n),t.root.removeEventListener("mouseup",e))};this.root.addEventListener("mousedown",(function(n){n.preventDefault(),n.stopPropagation();const r=t._getBoardIdByEvent(n);r&&(t.dragTimeout=setTimeout((function(){t.isClick=!0,t.isDraging=!0,t.emit("dragstart",r,n),t.root.addEventListener("mouseup",e,!1)}),100))}),!1),this.root.addEventListener("mouseenter",(function(t){t.preventDefault(),t.stopPropagation()}),!1),this.root.addEventListener("mouseleave",(function(e){e.preventDefault(),e.stopPropagation(),t.isDraging&&(t.isDraging=!1,t.emit("dragend",t.boardId,e))}),!1),this.root.addEventListener("wheel",(function(e){e.preventDefault(),e.stopPropagation(),t._getBoardIdByEvent(e)&&t.emit("zoom",t.boardId,e)}))}boardId(t,e,n){throw new Error("Method not implemented.")}deleteBoard(t){if("main"===t)return;let e=-1;for(let i=0;i<this.boards.length;i++)this.boards[i].boardId===t&&(e=i);if(-1===e)return;const n=this.boards[e];this.boards.splice(e,1);const r=n._rootElement.dom.previousSibling;return this.table.dom.removeChild(r),this.table.dom.removeChild(n._rootElement.dom),this.update(),n}hideRow(){}showRow(){}}const Wr=/^Y{1,4}|M{1,4}|D{1,2}|d{1,4}|H{1,2}|m{1,2}|s{1,2}|SSS/g;function Vr(t,e="YYYY-MM-DD HH:mm:ss"){const n=(t=t instanceof Date?t:new Date(t)).getFullYear(),r=t.getMonth(),i=t.getDate(),o=t.getHours(),a=t.getMinutes(),s=t.getSeconds(),l=t.getMilliseconds(),u={YY:String(n).slice(-2),YYYY:String(n),M:String(r+1),MM:String(r+1).padStart(2,"0"),D:String(i),DD:String(i).padStart(2,"0"),H:String(o),HH:String(o).padStart(2,"0"),m:String(a),mm:String(a).padStart(2,"0"),s:String(s),ss:String(s).padStart(2,"0"),SSS:String(l).padStart(3,"0")};return e.replace(Wr,t=>u[t])}class Gr{constructor({chart:t,width:e,height:n,container:r}){this.chart=t,this.width=e,this.height=n,this.container=r.dom,this.scale=ve().rangeRound([0,this.width]),this.axis=y(this.scale),this.rootElement=vr(this.container).append("svg").attr("class","axis").attr("width",this.width).attr("height",this.height),this.rootG=this.rootElement.append("g").attr("class","y axis "+this.align),this.annotationG=this.rootElement.append("g"),this.annotationRect=this.annotationG.append("rect").attr("fill","#fbd3a0"),this.annotationText=this.annotationG.append("text")}update({width:t,height:e}){t&&t!=this.width&&(this.width=t),e&&e!=this.height&&(this.height=e),this.rootElement.attr("width",this.width).attr("height",this.height)}getTicks(){return this.axis.getTicks()}draw(t,e,n){const r=t+Math.ceil(this.width/this.chart.bar.step);this.scale.rangeRound([0,(r-t)*this.chart.bar.step]),this.scale.domain([t,r]);const i=t*this.chart.bar.step-this.chart.chartCtrl.pixelLeft+Math.round(this.chart.bar.barWidth/2);this.rootG.attr("transform","translate("+i+",0)"),this.axis.tickFormat((t,e)=>{return n[t]&&n[t].datetime?(r=n[t].datetime/1e6,i=this.chart.duration,Vr(r,i/1e9/86400==0?"YYYY-MM-DD":i/1e3%3600==0?"MM-DD HH:mm":i/1e3%60==0?"HH:mm":"HH:mm:ss")):"";var r,i}),this.rootG.call(this.axis)}drawAnnotation(){}showXAnnotation(t){const e=Math.floor(t/this.chart.bar.step),n=this.scale.domain()[0]+e;if(isNaN(n)||!this.chart.mainSeries||!this.chart.mainSeries.data[n])return void this.annotationG.attr("visibility","hidden");const r=Vr(this.chart.mainSeries.data[n].datetime/1e6,"YYYY-MM-DD HH:mm:ss");this.annotationG.attr("visibility","visible"),this.annotationRect.attr("width",118).attr("height",16).attr("transform","translate("+(t-42)+",1)"),this.annotationText.text(r).attr("transform","translate("+(t-40)+",14)")}hideXAnnotation(){this.annotationG.attr("visibility","hidden")}}var $r={min:function(t,e){let n;if(void 0===e)for(const r of t)null!=r&&(n>r||void 0===n&&r>=r)&&(n=r);else{let r=-1;for(let i of t)null!=(i=e(i,++r,t))&&(n>i||void 0===n&&i>=i)&&(n=i)}return n},max:function(t,e){let n;if(void 0===e)for(const r of t)null!=r&&(n<r||void 0===n&&r>=r)&&(n=r);else{let r=-1;for(let i of t)null!=(i=e(i,++r,t))&&(n<i||void 0===n&&i>=i)&&(n=i)}return n}};const Ur={candle:{getYDomain(t,e,n){const r=n.slice(t,e+1),i=$r.min(r,t=>t&&t.low),o=$r.max(r,t=>t&&t.high),a=.5*(o-i)||4*this.chart.price_ticks;return[i-a,o+a]},draw(t,e,n,r){const i=_r(),o={upBody:"",upLine:"",downBody:"",downLine:"",equalLine:""};for(let a=e;a<=n;a++){const t=r[a];if(!t)continue;if(t.datetime>this.chart.maxDatetime)break;const e=this.yScale(t.open),n=this.yScale(t.close),s=this.yScale(t.high),l=this.yScale(t.low),u=a*this.xScale.step,c=Math.round(u+this.xScale.barWidth/2)-1;if(t.close===t.open)o.equalLine+=`M${u*i} ${(e+.5)*i} l${this.xScale.barWidth*i} 0 M${(c+.5)*i} ${s*i} l0 ${(l-s)*i}`;else{const r=t.close>t.open?"up":"down",a=n>e?Math.max(n-e,1):Math.min(n-e,-1);o[r+"Body"]+=`M${u*i} ${e*i} l${this.xScale.barWidth*i} 0 l0 ${a*i} l${-this.xScale.barWidth*i} 0 l0 ${-a*i}`,o[r+"Line"]+=`M${(c+.5)*i} ${s*i} l0 ${(l-s)*i}`}}wr(t,o.upBody,{fillStyle:"red"}),wr(t,o.upLine,{lineWidth:i,strokeStyle:"red"}),wr(t,o.downBody,{fillStyle:"green"}),wr(t,o.downLine,{lineWidth:i,strokeStyle:"green"}),wr(t,o.equalLine,{lineWidth:i,strokeStyle:"black"})}},close:{getYDomain(t,e,n){const r=n.slice(t,e+1);let[i,o]=[$r.min(r,t=>t&&t.close),$r.max(r,t=>t&&t.close)];return i-=.2*(o-i),o+=.2*(o-i),[i,o]},draw(t,e,n,r){const i=_r();let o="";for(let a=e;a<=n;a++){const t=r[a];if(t){if(t.datetime>this.chart.maxDatetime)break;o+=`${a==e?"M":"L"} ${(a*this.xScale.step+Math.round(this.xScale.barWidth/2)-1)*i} ${this.yScale(t.close)*i}`}}wr(t,o,{lineWidth:i,strokeStyle:"#C62828"})}},volume:{getYDomain(t,e,n){const r=n.slice(t,e+1);return[0,1.2*$r.max(r,t=>t&&t.volume)]},draw(t,e,n,r){const i=_r();let o={up:"",down:""};for(let a=e;a<=n;a++){const t=r[a];if(!t)continue;if(!t.volume)continue;if(t.datetime>this.chart.maxDatetime)break;const e=a*this.xScale.step*i,n=this.xScale.barWidth*i,s=this.yScale.range()[0]*i,l=this.yScale(t.volume)*i-s;o[t.close>=t.open?"up":"down"]+=`M${e} ${s} l${n} 0 l0 ${l} l${-n} 0 l0 ${-l}`}wr(t,o.up,{fillStyle:"red"}),wr(t,o.down,{fillStyle:"green"})}},oi:{getYDomain(t,e,n){const r=n.slice(t,e+1);let[i,o]=[$r.min(r,t=>t&&t.close_oi),$r.max(r,t=>t&&t.close_oi)];return i-=.2*(o-i),o+=.2*(o-i),[i,o]},draw(t,e,n,r){const i=_r();let o=null,a="";for(let s=e;s<=n;s++){const t=r[s];if(t&&t.close_oi){if(t.datetime>this.chart.maxDatetime)break;a+=`${o==s-1?"L":"M"} ${(s*this.xScale.step+Math.round(this.xScale.barWidth/2)-1)*i} ${this.yScale(t.close_oi)*i}`,o=s}}wr(t,a,{lineWidth:2*i,strokeStyle:"#C62828"})}},line:{getYDomain(t,e,n){if(Number.isFinite(this.min)&&Number.isFinite(this.max))return[this.min,this.max];{const r=n.slice(t,e+1),[i,o]=[$r.min(r,t=>t&&t[this.propName]),$r.max(r,t=>t&&t[this.propName])],a=.2*(o-i);return[i-a,o+a]}},draw(t,e,n,r){if(!this.yScale||!this.xScale)return;const i=_r();let o=null,a="";for(let l=e;l<=n;l++){const t=r[l],e=l*this.xScale.step+Math.round(this.xScale.barWidth/2)-1;t&&t[this.propName]&&(a+=`${o&&o==l-1?"L":"M"} ${e*i} ${this.yScale(t[this.propName])*i}`,o=l)}const s={lineWidth:this.width*i,strokeStyle:this.color};this.dash&&(s.lineDash=[5,5]),wr(t,a,s)}},bar:{getYDomain(t,e,n){if(Number.isFinite(this.min)&&Number.isFinite(this.max))return[this.min,this.max];{const r=n.slice(t,e+1);return[0,$r.max(r,t=>t&&t[this.propName])]}},draw(t,e,n,r){if(!this.yScale||!this.xScale)return;const i=_r();let o="";for(let a=e;a<=n;a++){const t=r[a];if(!t)continue;if(!t[this.propName])continue;const e=a*this.xScale.step*i,n=this.yScale(t[this.propName])*i,s=this.xScale.barWidth*i,l=(n-this.yScale.range[0])*i;o+=`M${e} ${n} l${s} 0 l0 ${l} l${-s} 0 l0 ${-l}`}wr(t,o,{fillStyle:this.color||"red"})}},dot:{getYDomain(t,e,n){if(Number.isFinite(this.min)&&Number.isFinite(this.max))return[this.min,this.max];{const r=n.slice(t,e+1);return[$r.min(r,t=>t&&t[this.propName]),$r.max(r,t=>t&&t[this.propName])]}},draw(t,e,n,r){if(!this.yScale||!this.xScale)return;const i=_r();let o="";for(let a=e;a<=n;a++){const t=r[a];if(!t)continue;if(!t[this.propName])continue;let e=a*this.xScale.step+this.bar.step/2,n=this.yScale(t[this.propName]);if(Number.isNaN(n)||Number.isNaN(e))continue;const s=3*i,l=3*i;e*=i,n*=i,o+="M"+(e-s)+" "+n+"a"+s+" "+l+" 0 1 0 "+2*s+" 0a"+s+" "+l+" 0 1 0 "+-2*s+" 0z"}wr(t,o,{fillStyle:this.color||"red"})}}};class qr{constructor(t,e,n,r,i,o={}){this.chart=t,this.boardId=e,this.board=this.chart.boards[this.boardId],this.plotId=n,this.type=r,this.yAlign=i||"right",this.xScale=this.chart.bar,this.yScale=this.board[this.yAlign+"YAxis"].scale,this.data=o.data,this.color=o.color||"red",this.width=Number.isFinite(o.width)?Math.round(o.width):1,this.dash=!!o.dash,this.min=o.min,this.max=o.max,this.isHollow=o.isHollow,this._initType(o)}_initType(t){switch(this.propName=t.propName||"value",this.type){case"candle":break;case"volume":this.propName="volume";break;case"oi":this.propName="close_oi";break;case"close":this.propName="close"}}getYDomain(t,e,n){let r=this.type;return"dash"==this.type&&(r="line"),Ur[r].getYDomain.bind(this)(t,e,n)}getText(t){if(this.data&&this.data[t]){const e=this.data[t];if("candle"===this.type&&e)return` O(${e.open}) H(${e.high}) L(${e.low}) C(${e.close})`;if(e&&e[this.propName]){let t=String(e[this.propName]);return t.length>12?(t=function(t,e=2){const n=Number(t),r=Number.isInteger(e)?e:2;return Number.isFinite(n)?n.toFixed(r):String(t)}(e[this.propName],"main"===this.boardId&&"right"===this.yAlign&&this.chart.price_decs?this.chart.price_decs:2),` ${this.plotId} (${t})`):` ${this.plotId} (${e[this.propName]})`}}return""}draw(t,e,n,r){this.data=r||this.data;let i=this.type;"dash"==this.type&&(i="line"),Ur[i].draw.bind(this)(t,e,n,this.data)}}class Xr{constructor(t,e){this._reset(t,e)}_reset(t,e){this.leftId=Number.isInteger(t)?t:-1,this.rightId=Number.isInteger(e)?e:-1}reset(t,e){return!(!this.isInvalid()&&t===this.leftId&&e===this.rightId)&&(this._reset(t,e),!0)}isInvalid(){return-1===this.leftId||-1===this.rightId}isEqual(t){return t.leftId===this.leftId&&t.rightId===this.rightId}copy(t){t instanceof Xr&&(this.leftId=t.leftId,this.rightId=t.rightId)}static copy(t){return new Xr(t.leftId,t.rightId)}}class Zr{constructor(t){this.chart=t,this.range=new Xr(-1,-1),this.pixelLeft=-1}get fullPixel(){return this.lastId&&this.lastId>-1?Math.max(this.lastId*this.chart.bar.step,this.canvasWidth):this.canvasWidth}get lastId(){return this.chart.mainSeries&&this.chart.mainSeries.last_id}get canvasWidth(){return this.chart.width-this.chart.margin.left-this.chart.margin.right}init(){this._initRange()}_initRange(){let[t,e]=[-1,-1];this.chart.mainSeries&&-1!==this.chart.mainSeries.last_id&&(e=this.chart.mainSeries.last_id,t=e-this.chart.bar.barNumbers+1,t+=5,t<0&&(t=0),this.pixelLeft=t*this.chart.bar.step),this.range.reset(t,e)}getRange(){return this.range.isInvalid()&&this._initRange(),this.range}setRange(t,e){t>=0&&e>=t?this.range.reset(t,e)&&(this.pixelLeft=this.range.leftId*this.chart.bar.step):this.range.reset(-1,-1)}move(t){if(this.getRange().isInvalid())return!1;let e=this.pixelLeft-t;if(e<0?e=0:e>=this.fullPixel-this.canvasWidth/2&&(e=Math.round(this.fullPixel-this.canvasWidth/2)),this.pixelLeft===e)return!1;this.pixelLeft=e;const n=Math.floor(e/this.chart.bar.step);let r=n+Math.ceil(this.canvasWidth/this.chart.bar.step);return r>this.lastId&&(r=this.lastId),this.range.reset(n,r),!0}zoom(t){const e=this.chart.bar.step-t,n=this.range.rightId*e-this.pixelLeft;this.pixelLeft=this.range.rightId*this.chart.bar.step-n;const r=Math.ceil(n/this.chart.bar.step);let i=this.range.rightId-r+1;i<0&&(i=0);let o=i+Math.ceil(this.canvasWidth/this.chart.bar.step);return o>this.lastId&&(o=this.lastId),this.range.reset(i,o),!0}resetByKlineId(t){const e=(t=t||this.range.rightId)*this.chart.bar.step-this.pixelLeft;this.pixelLeft=this.range.rightId*this.chart.bar.step-e;const n=Math.ceil(e/this.chart.bar.step);let r=this.range.rightId-n+1;r<0&&(r=0);let i=r+Math.ceil(this.canvasWidth/this.chart.bar.step);return i>this.lastId&&(i=this.lastId),this.range.reset(r,i),!0}revertIdToPx(t){return t*this.chart.bar.step}revertDtToId(t,e,n){const r=this.chart.mainSeries;if(!r||!r.data||this.range.isInvalid())return null;let i=!1;e>=0&&n>e?i=!0:(e=this.range.leftId,n=this.range.rightId);let[o,a]=[e,n];if(a=a<=r.last_id?a:r.last_id,!r.data[o]||!r.data[a])return null;if(t>=r.data[a].datetime&&t<r.data[a].datetime+this.chart.duration)return a;if(t>=r.data[o].datetime&&t<r.data[a].datetime+this.chart.duration){for(let e=o;e<a+1&&r.data[e];e++)if(t>=r.data[e].datetime&&t<r.data[e].datetime+this.chart.duration)return e;return null}if(i)return null;if(t>r.data[a].datetime){if(a===r.last_id)return null;for(let e=a;e<r.last_id&&r.data[e]&&r.data[e];e++)if(t-r.data[e].datetime<=0)return e>=1?e-1:e;return null}if(t<r.data[o].datetime){if(0===o)return null;for(let e=o;e>=0&&r.data[e];e--)if(t-r.data[e].datetime>=0)return e;return null}}}class Jr{constructor(t){this.chart=t,this.board=this.chart.boards.main,this.trades={}}add(t,e){this.trades[t]||(this.trades[t]=e)}removeAll(){this.trades={}}revertValueToYPos(t){return t&&this.board.rightYAxis?Math.round(this.board.rightYAxis.scale(t)):null}draw(){const t=[],e={},n={},r=this.chart.mainSeries;let[i,o]=[this.chart.range.leftId,this.chart.range.rightId];if(o=r.data[o]?o:r.last_id,!r.data[i]||!r.data[o])return;const[a,s]=[r.data[i].datetime,r.data[o].datetime];for(const l in this.trades){const u=this.trades[l];if(u.trade_date_time>=a&&u.trade_date_time<=s+this.chart.duration){const a=this.chart.chartCtrl.revertDtToId(u.trade_date_time,i,o);if(!a)continue;if("BUY"===u.direction)if(e[a])e[a].volume+=u.volume;else{const n={x:this.chart.chartCtrl.revertIdToPx(a),y:this.revertValueToYPos(r.data[a].low),dir:"up",diff:"OPEN"===u.offset?0:.5,width:this.chart.bar.barWidth,volume:u.volume};e[a]=n,t.push(n)}else if("SELL"===u.direction)if(n[a])n[a].volume+=u.volume;else{const e={x:this.chart.chartCtrl.revertIdToPx(a),y:this.revertValueToYPos(r.data[a].high),dir:"down",diff:"OPEN"===u.offset?0:.5,width:this.chart.bar.barWidth,volume:u.volume};n[a]=e,t.push(e)}}}this.drawArrows(t)}drawArrows(t){const e=this.board._centerCanvas.ctx,n=_r(),r={up:"",down:""};for(let i=0;i<t.length;i++){const e=t[i],o="up"===e.dir?1:-1,a=e.width,s=e.width,l=Math.round(s/3);r[e.dir]+=`M ${(e.x+s/2)*n} ${e.y*n} l ${-s/2*n} ${o*a/2*n} l ${(l-e.diff)*n} 0 l 0 ${o*a/2*n} l ${l*n} 0 l 0 ${-1*o*a/2*n} l ${l*n} 0 z`}e.beginPath(),e.fillStyle="red",e.fill(new Path2D(r.up)),e.fillStyle="green",e.fill(new Path2D(r.down))}}class Kr{constructor(t){this.fullWidth=t,this.barWidth=7,this.barNumbers=this._getBarNumbers(),this._domain=null,this.rangeToDomain={},this.domainToRange={},this.domainRange={}}get padding(){return this.barWidth<=3?0:this.barWidth<=17?1:2}get step(){return this.barWidth+(this.barWidth<=3?0:this.barWidth<=17?1:2)}_getBarNumbers(){return Math.round(this.fullWidth/this.step)}set domain(t){this._domain=t,this._updateDomainRange()}get domain(){return this._domain}_updateDomainRange(){this.domainRange={};for(let t=0;t<this.barNumbers;t++)this.domainToRange[this.domain[0]+t]=t*this.step}zoom(t){if(0===(t=Math.round(t)))return;const e=this.barWidth+t+t%2;e<1||e>35||(this.barWidth=e,this.barNumbers=this._getBarNumbers(),this._updateDomainRange())}setFullWidth(t){this.fullWidth=t,this.barNumbers=this._getBarNumbers()}setBarWidth(t=7){this.barWidth=Math.max(Math.min(Math.round(t),35),1),this.barNumbers=this._getBarNumbers()}setBarNumbers(t){this.barNumbers=Math.round(t);let e=Math.floor(this.fullWidth/this.barNumbers);e=Math.max(Math.min(e,35),1),this.barWidth!==e&&this.setBarWidth(e),this._updateDomainRange()}setDomain(t){this.domain=t,this._updateDomainRange()}getPosX(t){return this.domainToRange[t]}getKlineId(t){if(this.domain&&this.domain[0])return this.domain[0]+Math.floor(t/this.step)}getKlineIdTicks(){const t=Math.round((this.domain[1]-this.domain[0]-4)/7),e=[this.domain[0]+2];for(let n=1;n<7;n++)e.push(e[n-1]+t);return e}}e["a"]=class extends Lr{constructor(t){if(super(),"string"==typeof t)this.rootContainer=document.querySelector("#"+t);else{if(!(t instanceof HTMLElement&&t.tagName))return;this.rootContainer=t}}init({width:t,height:e,margin:n,symbol:r,duration:i,price_decs:o,price_ticks:a,mainSeries:s}){this.width=t,this.height=e,this.rootContainer.style.width=this.width+"px",this.rootContainer.style.height=this.height+"px",this.margin=n?Object.assign(n,{top:0,right:0,bottom:0,left:0}):{top:1,right:50,bottom:20,left:50},this.symbol=r||"",this.duration=i||6e10,this.price_decs=o||2,this.price_ticks=a||1,this.bar=new Kr(t-this.margin.left-this.margin.right),this.range=new Xr,this.mainSeries=s,this.tableContainer=new Br,this.chartCtrl=new Zr(this),this.tableContainer.init({chart:this,width:this.width,height:this.height,margin:this.margin}),this._initEventListenners(),this.timeAxis=new Gr({chart:this,width:this.width-this.margin.left-this.margin.right,height:this.margin.bottom,container:this.tableContainer.timeAxisBoard._centerTd}),this.mainBoard=new Hr({boardId:"main",chart:this,width:this.width,leftWidth:this.margin.left,rightWidth:this.margin.right,rightPriceDecs:this.price_decs}),this.boards={main:this.mainBoard,sub:new Hr({boardId:"sub",chart:this,width:this.width,leftWidth:this.margin.left,rightWidth:this.margin.right,rightPriceDecs:0,leftPriceDecs:0})},this.tableContainer.addBoard(this.mainBoard),this.tableContainer.addBoard(this.boards.sub),this.tableContainer.update(),this.plots={mainkline:new qr(this,"main","candle","candle","right"),volume:new qr(this,"sub","volume","volume","right"),oi:new qr(this,"sub","oi","oi","left")},this.chartCtrl.init(),this.rootContainer.appendChild(this.tableContainer.root),this.tradeArrowManager=new Jr(this)}setRange(t,e){"close"===this.mainType&&this.bar.setBarNumbers(e-t+1),this.chartCtrl.setRange(t,e),this.draw()}setPriceDecs(t){this.price_decs=t||this.price_decs,this.boards.main.setPriceDecs(this.price_decs)}setMainSeries(t,e,n,r="candle",{price_decs:i,price_ticks:o}){t&&e&&n&&(this.symbol=t,this.duration=e,this.mainSeries=n,this.mainType=r,this.chartCtrl.init(),this.price_decs=i||this.price_decs,this.price_ticks=o||this.price_ticks,this.boards.main.setPriceDecs(this.price_decs),this._resetBoards(),this.draw())}_resetBoards(){let t=Object.keys(this.plots);for(let n in t)-1===["mainkline","volume","oi"].indexOf(t[n])&&delete this.plots[t[n]];let e=Object.keys(this.boards);for(let n in e)-1===["main","sub"].indexOf(e[n])&&(delete this.boards[e[n]],this.tableContainer.deleteBoard(e[n]))}draw(t){if("candle"===this.mainType&&"close"===this.plots.mainkline.type&&(this.bar.setBarWidth(),this.plots.mainkline.type="candle",this.chartCtrl.setRange()),"close"===this.mainType&&"candle"===this.plots.mainkline.type&&(this.bar.setBarWidth(),this.plots.mainkline.type="close"),!this.mainSeries)return;if(!this.mainSeries.last_id||-1===this.mainSeries.last_id)return;if(this.mainSeries.last_id-this.range.rightId==1){const{leftId:t,rightId:e}=this.chartCtrl.getRange();this.bar.barNumbers-(e-t)>2?this.chartCtrl.setRange(t,this.mainSeries.last_id):this.chartCtrl.setRange(this.mainSeries.last_id-(e-t),this.mainSeries.last_id)}const{leftId:e,rightId:n}=this.chartCtrl.getRange();this.bar.setDomain([e,n]),this.timeAxis.draw(e,n,this.mainSeries.data);for(const r in this.boards){const t=this.boards[r],i={left:[1/0,-1/0],right:[1/0,-1/0]};for(const s in this.plots){const t=this.plots[s];if(r!=t.boardId)continue;let o=null;o="mainkline"===s||"volume"===s||"oi"===s?t.getYDomain(e,n,this.mainSeries.data):t.getYDomain(e,n,t.data||this.mainSeries.data),o[0]=Math.min(i[t.yAlign][0],o[0]),o[1]=Math.max(i[t.yAlign][1],o[1]),Number.isFinite(o[0])&&Number.isFinite(o[1])&&(i[t.yAlign]=o)}t.leftYAxis.draw(i.left),t.rightYAxis.draw(i.right),t.drawClearRect(e,n,this.mainSeries.data);const o=t._centerCanvas.ctx,a=_r();o.save(),o.translate(-this.chartCtrl.pixelLeft*a,0),t.drawBackground(e,n,this.mainSeries.data);for(const s in this.plots){const t=this.plots[s];r==t.boardId&&("mainkline"===s||"volume"===s||"oi"===s?t.draw(o,e,n,this.mainSeries.data):t.draw(o,e,n,this.plots[s].data||this.mainSeries.data))}t.drawFrontMark(e,n,this.mainSeries.data),"main"==r&&this.tradeArrowManager.draw(),o.restore()}if(!this.chartCtrl.range.isEqual(this.range)){this.range.copy(this.chartCtrl.range);const t=this.range.leftId-100,e=this.range.rightId+100;this.mainSeries.data[t]&&this.mainSeries.data[e]||this.emit("showRangeChanged",this.range)}}resize({width:t,height:e}){t===this.width&&e===this.height||(this.width=t,this.height=e,this.bar.setFullWidth(this.width-this.margin.left-this.margin.right),this.timeAxis.update({width:this.width-this.margin.left-this.margin.right,height:this.margin.bottom}),this.tableContainer.resize({width:this.width,height:this.height}),this.chartCtrl.resetByKlineId(),this.draw())}removeMark(t){for(const e in this.boards)this.boards[e].removeMark(t)}removeMarkAll(){for(const t in this.boards)this.boards[t].removeMarkAll()}addHighlightBar(t,e){for(const n in this.boards)this.boards[n].addMark({id:t,boardId:"main",type:"hightlight-bar",x1:e,layer:"back"});this.draw()}_getKlineIdByDatetime(t){let e=null,n=1/0;for(let r=this.range.leftId;r<=this.range.rightId;r++)if(e||(e=this.mainSeries[r]),e&&e<=t){const i=t-e;if(!(i>0))return 0==i?r:n+i>0?r-1:r;n=i}else if(e&&e>t)return null;return null}addTradeArrow(t,e){this.tradeArrowManager.add(t,e)}addPositionRecord(t,e){}_debounce(t,e){let n=null;return function(){const r=this,i=arguments;clearTimeout(n),n=setTimeout((function(){t.apply(r,i)}),e)}}_initEventListenners(){const t=this;this.tableContainer.addListener("dragstart",(function(e,n){if("close"!==t.mainType)for(const r in t.boards)t.boards[r].hideCrosshair()})),this.tableContainer.addListener("dragmove",(function(e,n){"close"!==t.mainType&&0!=n.movementX&&t.chartCtrl.move(n.movementX)&&t.draw("dragmove")})),this.tableContainer.addListener("dragend",(function(e,n){t.mainType})),this.tableContainer.addListener("zoom",(function(e,n){if("close"===t.mainType)return;const r=n.deltaY;Math.abs(r)>0&&t._debounce(t._zoom,60).bind(t)(r>0?-2:2)})),this.tableContainer.addListener("move",(function(e,n){for(const r in t.boards){const i=t.boards[r];e===r?(i.drawCrosshair(n.offsetX,n.offsetY),i.drawYAnnotation(n.offsetY,!0)):(i.drawCrosshairVertical(n.offsetX),i.drawYAnnotation(n.offsetY,!1)),i.drawTitleText(n.offsetX,n.offsetY)}t.timeAxis.showXAnnotation(n.offsetX)}))}_zoom(t){if("fixed"===this.mode)return;const e=this.bar.step;this.bar.zoom(t),e!==this.bar.step&&(this.chartCtrl.zoom(this.bar.step-e),this.draw("zoom"))}addSeries(t,e){const n=e.board?e.board.toLowerCase():"main";if(!this.boards[n]){const t=new Hr({boardId:n,chart:this,width:this.width,leftWidth:this.margin.left,rightWidth:this.margin.right,rightPriceTick:1,rightPriceDecs:0,leftPriceTick:1,leftPriceDecs:0});this.boards[n]=t,this.tableContainer.addBoard(t)}"KSERIAL"!==e.type&&"SERIAL"!==e.type||this.plots[t]?this.plots[t]&&e.color&&(this.plots[t].color=Mr(e.color)):"KSERIAL"===e.type?this.plots[t]=new qr(this,n,t,"candle",e.yAlign,{data:e.data}):this.plots[t]=new qr(this,n,t,e.style.toLowerCase(),e.yAlign,{data:e.data,color:Mr(e.color),width:e.width,dash:"dash"===e.style.toLowerCase()}),"TEXT"===e.type?this.boards[n].addMark({id:t,boardId:n,type:"text",x1:e.x1,y1:e.y1,color:Mr(e.color),size:e.size,text:e.text}):["LINE","SEG","RAY"].includes(e.type)?this.boards[n].addMark({id:t,boardId:n,type:"line",x1:e.x1,x2:e.x2,y1:e.y1,y2:e.y2,color:Mr(e.color),lineWidth:e.width,text:"",clamp:!1}):"BOX"===e.type&&this.boards[n].addMark({id:t,boardId:n,type:"rect",x1:e.x1,x2:e.x2,y1:e.y1,y2:e.y2,fill:Mr(e.bg_color),color:Mr(e.color),lineWidth:e.width||1}),this.draw()}addMark(t){this.boards[t.boardId]&&this.boards[t.boardId].addMark(t)}setOnlyMainBoard(t){this.onlyMainBoard!==t&&(this.onlyMainBoard=t,this.tableContainer.update(),this.draw())}setMaxDatetime(t){this.maxDatetime=t}}},6044:function(t,e,n){var r=n("0b07"),i=r(Object,"create");t.exports=i},"605d":function(t,e,n){"use strict";var r=n("da84"),i=n("c6b6");t.exports="process"===i(r.process)},6069:function(t,e,n){"use strict";var r=n("6c59"),i=n("605d");t.exports=!r&&!i&&"object"==typeof window&&"object"==typeof document},"60da":function(t,e,n){"use strict";var r=n("83ab"),i=n("e330"),o=n("c65b"),a=n("d039"),s=n("df75"),l=n("7418"),u=n("d1e7"),c=n("7b0b"),d=n("44ad"),h=Object.assign,f=Object.defineProperty,p=i([].concat);t.exports=!h||a((function(){if(r&&1!==h({b:1},h(f({},"a",{enumerable:!0,get:function(){f(this,"b",{value:3,enumerable:!1})}}),{b:2})).b)return!0;var t={},e={},n=Symbol("assign detection"),i="abcdefghijklmnopqrst";return t[n]=7,i.split("").forEach((function(t){e[t]=t})),7!==h({},t)[n]||s(h({},e)).join("")!==i}))?function(t,e){var n=c(t),i=arguments.length,a=1,h=l.f,f=u.f;while(i>a){var m,g=d(arguments[a++]),y=h?p(s(g),h(g)):s(g),v=y.length,_=0;while(v>_)m=y[_++],r&&!o(f,g,m)||(n[m]=g[m])}return n}:h},"60ed":function(t,e,n){var r=n("3729"),i=n("2dcb"),o=n("1310"),a="[object Object]",s=Function.prototype,l=Object.prototype,u=s.toString,c=l.hasOwnProperty,d=u.call(Object);function h(t){if(!o(t)||r(t)!=a)return!1;var e=i(t);if(null===e)return!0;var n=c.call(e,"constructor")&&e.constructor;return"function"==typeof n&&n instanceof n&&u.call(n)==d}t.exports=h},6117:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ug-cn",{months:"يانۋار_فېۋرال_مارت_ئاپرېل_ماي_ئىيۇن_ئىيۇل_ئاۋغۇست_سېنتەبىر_ئۆكتەبىر_نويابىر_دېكابىر".split("_"),monthsShort:"يانۋار_فېۋرال_مارت_ئاپرېل_ماي_ئىيۇن_ئىيۇل_ئاۋغۇست_سېنتەبىر_ئۆكتەبىر_نويابىر_دېكابىر".split("_"),weekdays:"يەكشەنبە_دۈشەنبە_سەيشەنبە_چارشەنبە_پەيشەنبە_جۈمە_شەنبە".split("_"),weekdaysShort:"يە_دۈ_سە_چا_پە_جۈ_شە".split("_"),weekdaysMin:"يە_دۈ_سە_چا_پە_جۈ_شە".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY-MM-DD",LL:"YYYY-يىلىM-ئاينىڭD-كۈنى",LLL:"YYYY-يىلىM-ئاينىڭD-كۈنى، HH:mm",LLLL:"dddd، YYYY-يىلىM-ئاينىڭD-كۈنى، HH:mm"},meridiemParse:/يېرىم كېچە|سەھەر|چۈشتىن بۇرۇن|چۈش|چۈشتىن كېيىن|كەچ/,meridiemHour:function(t,e){return 12===t&&(t=0),"يېرىم كېچە"===e||"سەھەر"===e||"چۈشتىن بۇرۇن"===e?t:"چۈشتىن كېيىن"===e||"كەچ"===e?t+12:t>=11?t:t+12},meridiem:function(t,e,n){var r=100*t+e;return r<600?"يېرىم كېچە":r<900?"سەھەر":r<1130?"چۈشتىن بۇرۇن":r<1230?"چۈش":r<1800?"چۈشتىن كېيىن":"كەچ"},calendar:{sameDay:"[بۈگۈن سائەت] LT",nextDay:"[ئەتە سائەت] LT",nextWeek:"[كېلەركى] dddd [سائەت] LT",lastDay:"[تۆنۈگۈن] LT",lastWeek:"[ئالدىنقى] dddd [سائەت] LT",sameElse:"L"},relativeTime:{future:"%s كېيىن",past:"%s بۇرۇن",s:"نەچچە سېكونت",ss:"%d سېكونت",m:"بىر مىنۇت",mm:"%d مىنۇت",h:"بىر سائەت",hh:"%d سائەت",d:"بىر كۈن",dd:"%d كۈن",M:"بىر ئاي",MM:"%d ئاي",y:"بىر يىل",yy:"%d يىل"},dayOfMonthOrdinalParse:/\d{1,2}(-كۈنى|-ئاي|-ھەپتە)/,ordinal:function(t,e){switch(e){case"d":case"D":case"DDD":return t+"-كۈنى";case"w":case"W":return t+"-ھەپتە";default:return t}},preparse:function(t){return t.replace(/،/g,",")},postformat:function(t){return t.replace(/,/g,"،")},week:{dow:1,doy:7}});return e}))},6252:function(t,e,n){"use strict";var r=n("a77e");const i="ivu-notice",o="ivu-icon",a="ivu_notice_key_";let s,l=24,u=4.5,c=1;const d={info:"ios-information-circle",success:"ios-checkmark-circle",warning:"ios-alert",error:"ios-close-circle"};function h(){return s=s||r["a"].newInstance({prefixCls:i,styles:{top:l+"px",right:0}}),s}function f(t,e){const n=e.title||"",r=e.desc||"",s=e.name||`${a}${c}`,l=e.onClose||function(){},f=e.render,p=0===e.duration?0:e.duration||u;c++;let m,g,y=h();const v=e.render&&!n?"":r||e.render?` ${i}-with-desc`:"";if("normal"==t)g=!1,m=`\n            <div class="${i}-custom-content ${i}-with-normal ${v}">\n                <div class="${i}-title">${n}</div>\n                <div class="${i}-desc">${r}</div>\n            </div>\n        `;else{const e=d[t],a=""===v?"":"-outline";g=!0,m=`\n            <div class="${i}-custom-content ${i}-with-icon ${i}-with-${t} ${v}">\n                <span class="${i}-icon ${i}-icon-${t}">\n                    <i class="${o} ${o}-${e}${a}"></i>\n                </span>\n                <div class="${i}-title">${n}</div>\n                <div class="${i}-desc">${r}</div>\n            </div>\n        `}y.notice({name:s.toString(),duration:p,styles:{},transitionName:"move-notice",content:m,withIcon:g,render:f,hasTitle:!!n,onClose:l,closable:!0,type:"notice"})}e["a"]={open(t){return f("normal",t)},info(t){return f("info",t)},success(t){return f("success",t)},warning(t){return f("warning",t)},error(t){return f("error",t)},config(t){t.top&&(l=t.top),(t.duration||0===t.duration)&&(u=t.duration)},close(t){if(!t)return!1;t=t.toString(),s&&s.remove(t)},destroy(){let t=h();s=null,t.destroy("ivu-notice")}}},"62e4":function(t,e){t.exports=function(t){return t.webpackPolyfill||(t.deprecate=function(){},t.paths=[],t.children||(t.children=[]),Object.defineProperty(t,"loaded",{enumerable:!0,get:function(){return t.l}}),Object.defineProperty(t,"id",{enumerable:!0,get:function(){return t.i}}),t.webpackPolyfill=1),t}},6374:function(t,e,n){"use strict";var r=n("da84"),i=Object.defineProperty;t.exports=function(t,e){try{i(r,t,{value:e,configurable:!0,writable:!0})}catch(n){r[t]=e}return e}},6403:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ms-my",{months:"Januari_Februari_Mac_April_Mei_Jun_Julai_Ogos_September_Oktober_November_Disember".split("_"),monthsShort:"Jan_Feb_Mac_Apr_Mei_Jun_Jul_Ogs_Sep_Okt_Nov_Dis".split("_"),weekdays:"Ahad_Isnin_Selasa_Rabu_Khamis_Jumaat_Sabtu".split("_"),weekdaysShort:"Ahd_Isn_Sel_Rab_Kha_Jum_Sab".split("_"),weekdaysMin:"Ah_Is_Sl_Rb_Km_Jm_Sb".split("_"),longDateFormat:{LT:"HH.mm",LTS:"HH.mm.ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY [pukul] HH.mm",LLLL:"dddd, D MMMM YYYY [pukul] HH.mm"},meridiemParse:/pagi|tengahari|petang|malam/,meridiemHour:function(t,e){return 12===t&&(t=0),"pagi"===e?t:"tengahari"===e?t>=11?t:t+12:"petang"===e||"malam"===e?t+12:void 0},meridiem:function(t,e,n){return t<11?"pagi":t<15?"tengahari":t<19?"petang":"malam"},calendar:{sameDay:"[Hari ini pukul] LT",nextDay:"[Esok pukul] LT",nextWeek:"dddd [pukul] LT",lastDay:"[Kelmarin pukul] LT",lastWeek:"dddd [lepas pukul] LT",sameElse:"L"},relativeTime:{future:"dalam %s",past:"%s yang lepas",s:"beberapa saat",ss:"%d saat",m:"seminit",mm:"%d minit",h:"sejam",hh:"%d jam",d:"sehari",dd:"%d hari",M:"sebulan",MM:"%d bulan",y:"setahun",yy:"%d tahun"},week:{dow:1,doy:7}});return e}))},"642a":function(t,e,n){var r=n("966f"),i=n("3bb4"),o=n("20ec");function a(t){var e=i(t);return 1==e.length&&e[0][2]?o(e[0][0],e[0][1]):function(n){return n===t||r(n,t,e)}}t.exports=a},6547:function(t,e,n){"use strict";var r=n("e330"),i=n("5926"),o=n("577e"),a=n("1d80"),s=r("".charAt),l=r("".charCodeAt),u=r("".slice),c=function(t){return function(e,n){var r,c,d=o(a(e)),h=i(n),f=d.length;return h<0||h>=f?t?"":void 0:(r=l(d,h),r<55296||r>56319||h+1===f||(c=l(d,h+1))<56320||c>57343?t?s(d,h):r:t?u(d,h,h+2):c-56320+(r-55296<<10)+65536)}};t.exports={codeAt:c(!1),charAt:c(!0)}},"656b":function(t,e,n){var r=n("e2e4"),i=n("f4d6");function o(t,e){e=r(e,t);var n=0,o=e.length;while(null!=t&&n<o)t=t[i(e[n++])];return n&&n==o?t:void 0}t.exports=o},"65db":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("eo",{months:"januaro_februaro_marto_aprilo_majo_junio_julio_aŭgusto_septembro_oktobro_novembro_decembro".split("_"),monthsShort:"jan_feb_mart_apr_maj_jun_jul_aŭg_sept_okt_nov_dec".split("_"),weekdays:"dimanĉo_lundo_mardo_merkredo_ĵaŭdo_vendredo_sabato".split("_"),weekdaysShort:"dim_lun_mard_merk_ĵaŭ_ven_sab".split("_"),weekdaysMin:"di_lu_ma_me_ĵa_ve_sa".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY-MM-DD",LL:"[la] D[-an de] MMMM, YYYY",LLL:"[la] D[-an de] MMMM, YYYY HH:mm",LLLL:"dddd[n], [la] D[-an de] MMMM, YYYY HH:mm",llll:"ddd, [la] D[-an de] MMM, YYYY HH:mm"},meridiemParse:/[ap]\.t\.m/i,isPM:function(t){return"p"===t.charAt(0).toLowerCase()},meridiem:function(t,e,n){return t>11?n?"p.t.m.":"P.T.M.":n?"a.t.m.":"A.T.M."},calendar:{sameDay:"[Hodiaŭ je] LT",nextDay:"[Morgaŭ je] LT",nextWeek:"dddd[n je] LT",lastDay:"[Hieraŭ je] LT",lastWeek:"[pasintan] dddd[n je] LT",sameElse:"L"},relativeTime:{future:"post %s",past:"antaŭ %s",s:"kelkaj sekundoj",ss:"%d sekundoj",m:"unu minuto",mm:"%d minutoj",h:"unu horo",hh:"%d horoj",d:"unu tago",dd:"%d tagoj",M:"unu monato",MM:"%d monatoj",y:"unu jaro",yy:"%d jaroj"},dayOfMonthOrdinalParse:/\d{1,2}a/,ordinal:"%da",week:{dow:1,doy:7}});return e}))},"65f0":function(t,e,n){"use strict";var r=n("0b42");t.exports=function(t,e){return new(r(t))(0===e?0:e)}},6747:function(t,e){var n=Array.isArray;t.exports=n},6784:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=["جنوري","فيبروري","مارچ","اپريل","مئي","جون","جولاءِ","آگسٽ","سيپٽمبر","آڪٽوبر","نومبر","ڊسمبر"],n=["آچر","سومر","اڱارو","اربع","خميس","جمع","ڇنڇر"],r=t.defineLocale("sd",{months:e,monthsShort:e,weekdays:n,weekdaysShort:n,weekdaysMin:n,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd، D MMMM YYYY HH:mm"},meridiemParse:/صبح|شام/,isPM:function(t){return"شام"===t},meridiem:function(t,e,n){return t<12?"صبح":"شام"},calendar:{sameDay:"[اڄ] LT",nextDay:"[سڀاڻي] LT",nextWeek:"dddd [اڳين هفتي تي] LT",lastDay:"[ڪالهه] LT",lastWeek:"[گزريل هفتي] dddd [تي] LT",sameElse:"L"},relativeTime:{future:"%s پوء",past:"%s اڳ",s:"چند سيڪنڊ",ss:"%d سيڪنڊ",m:"هڪ منٽ",mm:"%d منٽ",h:"هڪ ڪلاڪ",hh:"%d ڪلاڪ",d:"هڪ ڏينهن",dd:"%d ڏينهن",M:"هڪ مهينو",MM:"%d مهينا",y:"هڪ سال",yy:"%d سال"},preparse:function(t){return t.replace(/،/g,",")},postformat:function(t){return t.replace(/,/g,"،")},week:{dow:1,doy:4}});return r}))},"67ca":function(t,e,n){var r=n("cb5a");function i(t,e){var n=this.__data__,i=r(n,t);return i<0?(++this.size,n.push([t,e])):n[i][1]=e,this}t.exports=i},6887:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n){var r={mm:"munutenn",MM:"miz",dd:"devezh"};return t+" "+i(r[n],t)}function n(t){switch(r(t)){case 1:case 3:case 4:case 5:case 9:return t+" bloaz";default:return t+" vloaz"}}function r(t){return t>9?r(t%10):t}function i(t,e){return 2===e?o(t):t}function o(t){var e={m:"v",b:"v",d:"z"};return void 0===e[t.charAt(0)]?t:e[t.charAt(0)]+t.substring(1)}var a=[/^gen/i,/^c[ʼ\']hwe/i,/^meu/i,/^ebr/i,/^mae/i,/^(mez|eve)/i,/^gou/i,/^eos/i,/^gwe/i,/^her/i,/^du/i,/^ker/i],s=/^(genver|c[ʼ\']hwevrer|meurzh|ebrel|mae|mezheven|gouere|eost|gwengolo|here|du|kerzu|gen|c[ʼ\']hwe|meu|ebr|mae|eve|gou|eos|gwe|her|du|ker)/i,l=/^(genver|c[ʼ\']hwevrer|meurzh|ebrel|mae|mezheven|gouere|eost|gwengolo|here|du|kerzu)/i,u=/^(gen|c[ʼ\']hwe|meu|ebr|mae|eve|gou|eos|gwe|her|du|ker)/i,c=[/^sul/i,/^lun/i,/^meurzh/i,/^merc[ʼ\']her/i,/^yaou/i,/^gwener/i,/^sadorn/i],d=[/^Sul/i,/^Lun/i,/^Meu/i,/^Mer/i,/^Yao/i,/^Gwe/i,/^Sad/i],h=[/^Su/i,/^Lu/i,/^Me([^r]|$)/i,/^Mer/i,/^Ya/i,/^Gw/i,/^Sa/i],f=t.defineLocale("br",{months:"Genver_Cʼhwevrer_Meurzh_Ebrel_Mae_Mezheven_Gouere_Eost_Gwengolo_Here_Du_Kerzu".split("_"),monthsShort:"Gen_Cʼhwe_Meu_Ebr_Mae_Eve_Gou_Eos_Gwe_Her_Du_Ker".split("_"),weekdays:"Sul_Lun_Meurzh_Mercʼher_Yaou_Gwener_Sadorn".split("_"),weekdaysShort:"Sul_Lun_Meu_Mer_Yao_Gwe_Sad".split("_"),weekdaysMin:"Su_Lu_Me_Mer_Ya_Gw_Sa".split("_"),weekdaysParse:h,fullWeekdaysParse:c,shortWeekdaysParse:d,minWeekdaysParse:h,monthsRegex:s,monthsShortRegex:s,monthsStrictRegex:l,monthsShortStrictRegex:u,monthsParse:a,longMonthsParse:a,shortMonthsParse:a,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D [a viz] MMMM YYYY",LLL:"D [a viz] MMMM YYYY HH:mm",LLLL:"dddd, D [a viz] MMMM YYYY HH:mm"},calendar:{sameDay:"[Hiziv da] LT",nextDay:"[Warcʼhoazh da] LT",nextWeek:"dddd [da] LT",lastDay:"[Decʼh da] LT",lastWeek:"dddd [paset da] LT",sameElse:"L"},relativeTime:{future:"a-benn %s",past:"%s ʼzo",s:"un nebeud segondennoù",ss:"%d eilenn",m:"ur vunutenn",mm:e,h:"un eur",hh:"%d eur",d:"un devezh",dd:e,M:"ur miz",MM:e,y:"ur bloaz",yy:n},dayOfMonthOrdinalParse:/\d{1,2}(añ|vet)/,ordinal:function(t){var e=1===t?"añ":"vet";return t+e},week:{dow:1,doy:4},meridiemParse:/a.m.|g.m./,isPM:function(t){return"g.m."===t},meridiem:function(t,e,n){return t<12?"a.m.":"g.m."}});return f}))},"688b":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("mi",{months:"Kohi-tāte_Hui-tanguru_Poutū-te-rangi_Paenga-whāwhā_Haratua_Pipiri_Hōngoingoi_Here-turi-kōkā_Mahuru_Whiringa-ā-nuku_Whiringa-ā-rangi_Hakihea".split("_"),monthsShort:"Kohi_Hui_Pou_Pae_Hara_Pipi_Hōngoi_Here_Mahu_Whi-nu_Whi-ra_Haki".split("_"),monthsRegex:/(?:['a-z\u0101\u014D\u016B]+\-?){1,3}/i,monthsStrictRegex:/(?:['a-z\u0101\u014D\u016B]+\-?){1,3}/i,monthsShortRegex:/(?:['a-z\u0101\u014D\u016B]+\-?){1,3}/i,monthsShortStrictRegex:/(?:['a-z\u0101\u014D\u016B]+\-?){1,2}/i,weekdays:"Rātapu_Mane_Tūrei_Wenerei_Tāite_Paraire_Hātarei".split("_"),weekdaysShort:"Ta_Ma_Tū_We_Tāi_Pa_Hā".split("_"),weekdaysMin:"Ta_Ma_Tū_We_Tāi_Pa_Hā".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY [i] HH:mm",LLLL:"dddd, D MMMM YYYY [i] HH:mm"},calendar:{sameDay:"[i teie mahana, i] LT",nextDay:"[apopo i] LT",nextWeek:"dddd [i] LT",lastDay:"[inanahi i] LT",lastWeek:"dddd [whakamutunga i] LT",sameElse:"L"},relativeTime:{future:"i roto i %s",past:"%s i mua",s:"te hēkona ruarua",ss:"%d hēkona",m:"he meneti",mm:"%d meneti",h:"te haora",hh:"%d haora",d:"he ra",dd:"%d ra",M:"he marama",MM:"%d marama",y:"he tau",yy:"%d tau"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:1,doy:4}});return e}))},"68ee":function(t,e,n){"use strict";var r=n("e330"),i=n("d039"),o=n("1626"),a=n("f5df"),s=n("d066"),l=n("8925"),u=function(){},c=s("Reflect","construct"),d=/^\s*(?:class|function)\b/,h=r(d.exec),f=!d.test(u),p=function(t){if(!o(t))return!1;try{return c(u,[],t),!0}catch(e){return!1}},m=function(t){if(!o(t))return!1;switch(a(t)){case"AsyncFunction":case"GeneratorFunction":case"AsyncGeneratorFunction":return!1}try{return f||!!h(d,l(t))}catch(e){return!0}};m.sham=!0,t.exports=!c||i((function(){var t;return p(p.call)||!p(Object)||!p((function(){t=!0}))||t}))?m:p},6909:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("mk",{months:"јануари_февруари_март_април_мај_јуни_јули_август_септември_октомври_ноември_декември".split("_"),monthsShort:"јан_фев_мар_апр_мај_јун_јул_авг_сеп_окт_ное_дек".split("_"),weekdays:"недела_понеделник_вторник_среда_четврток_петок_сабота".split("_"),weekdaysShort:"нед_пон_вто_сре_чет_пет_саб".split("_"),weekdaysMin:"нe_пo_вт_ср_че_пе_сa".split("_"),longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"D.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY H:mm",LLLL:"dddd, D MMMM YYYY H:mm"},calendar:{sameDay:"[Денес во] LT",nextDay:"[Утре во] LT",nextWeek:"[Во] dddd [во] LT",lastDay:"[Вчера во] LT",lastWeek:function(){switch(this.day()){case 0:case 3:case 6:return"[Изминатата] dddd [во] LT";case 1:case 2:case 4:case 5:return"[Изминатиот] dddd [во] LT"}},sameElse:"L"},relativeTime:{future:"за %s",past:"пред %s",s:"неколку секунди",ss:"%d секунди",m:"една минута",mm:"%d минути",h:"еден час",hh:"%d часа",d:"еден ден",dd:"%d дена",M:"еден месец",MM:"%d месеци",y:"една година",yy:"%d години"},dayOfMonthOrdinalParse:/\d{1,2}-(ев|ен|ти|ви|ри|ми)/,ordinal:function(t){var e=t%10,n=t%100;return 0===t?t+"-ев":0===n?t+"-ен":n>10&&n<20?t+"-ти":1===e?t+"-ви":2===e?t+"-ри":7===e||8===e?t+"-ми":t+"-ти"},week:{dow:1,doy:7}});return e}))},"69d5":function(t,e,n){var r=n("cb5a"),i=Array.prototype,o=i.splice;function a(t){var e=this.__data__,n=r(e,t);if(n<0)return!1;var i=e.length-1;return n==i?e.pop():o.call(e,n,1),--this.size,!0}t.exports=a},"69f3":function(t,e,n){"use strict";var r,i,o,a=n("cdce"),s=n("da84"),l=n("861d"),u=n("9112"),c=n("1a2d"),d=n("c6cd"),h=n("f772"),f=n("d012"),p="Object already initialized",m=s.TypeError,g=s.WeakMap,y=function(t){return o(t)?i(t):r(t,{})},v=function(t){return function(e){var n;if(!l(e)||(n=i(e)).type!==t)throw new m("Incompatible receiver, "+t+" required");return n}};if(a||d.state){var _=d.state||(d.state=new g);_.get=_.get,_.has=_.has,_.set=_.set,r=function(t,e){if(_.has(t))throw new m(p);return e.facade=t,_.set(t,e),e},i=function(t){return _.get(t)||{}},o=function(t){return _.has(t)}}else{var b=h("state");f[b]=!0,r=function(t,e){if(c(t,b))throw new m(p);return e.facade=t,u(t,b,e),e},i=function(t){return c(t,b)?t[b]:{}},o=function(t){return c(t,b)}}t.exports={set:r,get:i,has:o,enforce:y,getterFor:v}},"6b75":function(t,e,n){"use strict";function r(t,e){(null==e||e>t.length)&&(e=t.length);for(var n=0,r=new Array(e);n<e;n++)r[n]=t[n];return r}n.d(e,"a",(function(){return r}))},"6c59":function(t,e,n){"use strict";t.exports="object"==typeof Deno&&Deno&&"object"==typeof Deno.version},"6ce3":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("nb",{months:"januar_februar_mars_april_mai_juni_juli_august_september_oktober_november_desember".split("_"),monthsShort:"jan._feb._mars_apr._mai_juni_juli_aug._sep._okt._nov._des.".split("_"),monthsParseExact:!0,weekdays:"søndag_mandag_tirsdag_onsdag_torsdag_fredag_lørdag".split("_"),weekdaysShort:"sø._ma._ti._on._to._fr._lø.".split("_"),weekdaysMin:"sø_ma_ti_on_to_fr_lø".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY [kl.] HH:mm",LLLL:"dddd D. MMMM YYYY [kl.] HH:mm"},calendar:{sameDay:"[i dag kl.] LT",nextDay:"[i morgen kl.] LT",nextWeek:"dddd [kl.] LT",lastDay:"[i går kl.] LT",lastWeek:"[forrige] dddd [kl.] LT",sameElse:"L"},relativeTime:{future:"om %s",past:"%s siden",s:"noen sekunder",ss:"%d sekunder",m:"ett minutt",mm:"%d minutter",h:"én time",hh:"%d timer",d:"én dag",dd:"%d dager",w:"én uke",ww:"%d uker",M:"én måned",MM:"%d måneder",y:"ett år",yy:"%d år"},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return e}))},"6d79":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={0:"-ші",1:"-ші",2:"-ші",3:"-ші",4:"-ші",5:"-ші",6:"-шы",7:"-ші",8:"-ші",9:"-шы",10:"-шы",20:"-шы",30:"-шы",40:"-шы",50:"-ші",60:"-шы",70:"-ші",80:"-ші",90:"-шы",100:"-ші"},n=t.defineLocale("kk",{months:"қаңтар_ақпан_наурыз_сәуір_мамыр_маусым_шілде_тамыз_қыркүйек_қазан_қараша_желтоқсан".split("_"),monthsShort:"қаң_ақп_нау_сәу_мам_мау_шіл_там_қыр_қаз_қар_жел".split("_"),weekdays:"жексенбі_дүйсенбі_сейсенбі_сәрсенбі_бейсенбі_жұма_сенбі".split("_"),weekdaysShort:"жек_дүй_сей_сәр_бей_жұм_сен".split("_"),weekdaysMin:"жк_дй_сй_ср_бй_жм_сн".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Бүгін сағат] LT",nextDay:"[Ертең сағат] LT",nextWeek:"dddd [сағат] LT",lastDay:"[Кеше сағат] LT",lastWeek:"[Өткен аптаның] dddd [сағат] LT",sameElse:"L"},relativeTime:{future:"%s ішінде",past:"%s бұрын",s:"бірнеше секунд",ss:"%d секунд",m:"бір минут",mm:"%d минут",h:"бір сағат",hh:"%d сағат",d:"бір күн",dd:"%d күн",M:"бір ай",MM:"%d ай",y:"бір жыл",yy:"%d жыл"},dayOfMonthOrdinalParse:/\d{1,2}-(ші|шы)/,ordinal:function(t){var n=t%10,r=t>=100?100:null;return t+(e[t]||e[n]||e[r])},week:{dow:1,doy:7}});return n}))},"6d83":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ar-tn",{months:"جانفي_فيفري_مارس_أفريل_ماي_جوان_جويلية_أوت_سبتمبر_أكتوبر_نوفمبر_ديسمبر".split("_"),monthsShort:"جانفي_فيفري_مارس_أفريل_ماي_جوان_جويلية_أوت_سبتمبر_أكتوبر_نوفمبر_ديسمبر".split("_"),weekdays:"الأحد_الإثنين_الثلاثاء_الأربعاء_الخميس_الجمعة_السبت".split("_"),weekdaysShort:"أحد_إثنين_ثلاثاء_أربعاء_خميس_جمعة_سبت".split("_"),weekdaysMin:"ح_ن_ث_ر_خ_ج_س".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[اليوم على الساعة] LT",nextDay:"[غدا على الساعة] LT",nextWeek:"dddd [على الساعة] LT",lastDay:"[أمس على الساعة] LT",lastWeek:"dddd [على الساعة] LT",sameElse:"L"},relativeTime:{future:"في %s",past:"منذ %s",s:"ثوان",ss:"%d ثانية",m:"دقيقة",mm:"%d دقائق",h:"ساعة",hh:"%d ساعات",d:"يوم",dd:"%d أيام",M:"شهر",MM:"%d أشهر",y:"سنة",yy:"%d سنوات"},week:{dow:1,doy:4}});return e}))},"6d8b":function(t,e,n){"use strict";n.d(e,"p",(function(){return g})),n.d(e,"G",(function(){return y})),n.d(e,"d",(function(){return v})),n.d(e,"I",(function(){return _})),n.d(e,"J",(function(){return b})),n.d(e,"m",(function(){return x})),n.d(e,"i",(function(){return w})),n.d(e,"r",(function(){return M})),n.d(e,"s",(function(){return S})),n.d(e,"K",(function(){return k})),n.d(e,"u",(function(){return L})),n.d(e,"k",(function(){return T})),n.d(e,"H",(function(){return D})),n.d(e,"N",(function(){return O})),n.d(e,"n",(function(){return I})),n.d(e,"o",(function(){return C})),n.d(e,"F",(function(){return A})),n.d(e,"c",(function(){return j})),n.d(e,"h",(function(){return P})),n.d(e,"t",(function(){return E})),n.d(e,"w",(function(){return N})),n.d(e,"C",(function(){return R})),n.d(e,"D",(function(){return H})),n.d(e,"z",(function(){return z})),n.d(e,"A",(function(){return F})),n.d(e,"E",(function(){return W})),n.d(e,"v",(function(){return V})),n.d(e,"x",(function(){return G})),n.d(e,"y",(function(){return $})),n.d(e,"B",(function(){return U})),n.d(e,"l",(function(){return q})),n.d(e,"O",(function(){return X})),n.d(e,"P",(function(){return Z})),n.d(e,"Q",(function(){return J})),n.d(e,"S",(function(){return K})),n.d(e,"M",(function(){return Q})),n.d(e,"b",(function(){return tt})),n.d(e,"T",(function(){return et})),n.d(e,"R",(function(){return rt})),n.d(e,"f",(function(){return ut})),n.d(e,"e",(function(){return ct})),n.d(e,"g",(function(){return dt})),n.d(e,"j",(function(){return ht})),n.d(e,"q",(function(){return ft})),n.d(e,"L",(function(){return pt})),n.d(e,"a",(function(){return mt}));var r=n("726e"),i=O(["Function","RegExp","Date","Error","CanvasGradient","CanvasPattern","Image","Canvas"],(function(t,e){return t["[object "+e+"]"]=!0,t}),{}),o=O(["Int8","Uint8","Uint8Clamped","Int16","Uint16","Int32","Uint32","Float32","Float64"],(function(t,e){return t["[object "+e+"Array]"]=!0,t}),{}),a=Object.prototype.toString,s=Array.prototype,l=s.forEach,u=s.filter,c=s.slice,d=s.map,h=function(){}.constructor,f=h?h.prototype:null,p="__proto__",m=2311;function g(){return m++}function y(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e]}function v(t){if(null==t||"object"!==typeof t)return t;var e=t,n=a.call(t);if("[object Array]"===n){if(!it(t)){e=[];for(var r=0,s=t.length;r<s;r++)e[r]=v(t[r])}}else if(o[n]){if(!it(t)){var l=t.constructor;if(l.from)e=l.from(t);else{e=new l(t.length);for(r=0,s=t.length;r<s;r++)e[r]=t[r]}}}else if(!i[n]&&!it(t)&&!V(t))for(var u in e={},t)t.hasOwnProperty(u)&&u!==p&&(e[u]=v(t[u]));return e}function _(t,e,n){if(!F(e)||!F(t))return n?v(e):t;for(var r in e)if(e.hasOwnProperty(r)&&r!==p){var i=t[r],o=e[r];!F(o)||!F(i)||E(o)||E(i)||V(o)||V(i)||B(o)||B(i)||it(o)||it(i)?!n&&r in t||(t[r]=v(e[r])):_(i,o,n)}return t}function b(t,e){for(var n=t[0],r=1,i=t.length;r<i;r++)n=_(n,t[r],e);return n}function x(t,e){if(Object.assign)Object.assign(t,e);else for(var n in e)e.hasOwnProperty(n)&&n!==p&&(t[n]=e[n]);return t}function w(t,e,n){for(var r=A(e),i=0;i<r.length;i++){var o=r[i];(n?null!=e[o]:null==t[o])&&(t[o]=e[o])}return t}r["d"].createCanvas;function M(t,e){if(t){if(t.indexOf)return t.indexOf(e);for(var n=0,r=t.length;n<r;n++)if(t[n]===e)return n}return-1}function S(t,e){var n=t.prototype;function r(){}for(var i in r.prototype=e.prototype,t.prototype=new r,n)n.hasOwnProperty(i)&&(t.prototype[i]=n[i]);t.prototype.constructor=t,t.superClass=e}function k(t,e,n){if(t="prototype"in t?t.prototype:t,e="prototype"in e?e.prototype:e,Object.getOwnPropertyNames)for(var r=Object.getOwnPropertyNames(e),i=0;i<r.length;i++){var o=r[i];"constructor"!==o&&(n?null!=e[o]:null==t[o])&&(t[o]=e[o])}else w(t,e,n)}function L(t){return!!t&&("string"!==typeof t&&"number"===typeof t.length)}function T(t,e,n){if(t&&e)if(t.forEach&&t.forEach===l)t.forEach(e,n);else if(t.length===+t.length)for(var r=0,i=t.length;r<i;r++)e.call(n,t[r],r,t);else for(var o in t)t.hasOwnProperty(o)&&e.call(n,t[o],o,t)}function D(t,e,n){if(!t)return[];if(!e)return K(t);if(t.map&&t.map===d)return t.map(e,n);for(var r=[],i=0,o=t.length;i<o;i++)r.push(e.call(n,t[i],i,t));return r}function O(t,e,n,r){if(t&&e){for(var i=0,o=t.length;i<o;i++)n=e.call(r,n,t[i],i,t);return n}}function I(t,e,n){if(!t)return[];if(!e)return K(t);if(t.filter&&t.filter===u)return t.filter(e,n);for(var r=[],i=0,o=t.length;i<o;i++)e.call(n,t[i],i,t)&&r.push(t[i]);return r}function C(t,e,n){if(t&&e)for(var r=0,i=t.length;r<i;r++)if(e.call(n,t[r],r,t))return t[r]}function A(t){if(!t)return[];if(Object.keys)return Object.keys(t);var e=[];for(var n in t)t.hasOwnProperty(n)&&e.push(n);return e}function Y(t,e){for(var n=[],r=2;r<arguments.length;r++)n[r-2]=arguments[r];return function(){return t.apply(e,n.concat(c.call(arguments)))}}var j=f&&N(f.bind)?f.call.bind(f.bind):Y;function P(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];return function(){return t.apply(this,e.concat(c.call(arguments)))}}function E(t){return Array.isArray?Array.isArray(t):"[object Array]"===a.call(t)}function N(t){return"function"===typeof t}function R(t){return"string"===typeof t}function H(t){return"[object String]"===a.call(t)}function z(t){return"number"===typeof t}function F(t){var e=typeof t;return"function"===e||!!t&&"object"===e}function B(t){return!!i[a.call(t)]}function W(t){return!!o[a.call(t)]}function V(t){return"object"===typeof t&&"number"===typeof t.nodeType&&"object"===typeof t.ownerDocument}function G(t){return null!=t.colorStops}function $(t){return null!=t.image}function U(t){return"[object RegExp]"===a.call(t)}function q(t){return t!==t}function X(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];for(var n=0,r=t.length;n<r;n++)if(null!=t[n])return t[n]}function Z(t,e){return null!=t?t:e}function J(t,e,n){return null!=t?t:null!=e?e:n}function K(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];return c.apply(t,e)}function Q(t){if("number"===typeof t)return[t,t,t,t];var e=t.length;return 2===e?[t[0],t[1],t[0],t[1]]:3===e?[t[0],t[1],t[2],t[1]]:t}function tt(t,e){if(!t)throw new Error(e)}function et(t){return null==t?null:"function"===typeof t.trim?t.trim():t.replace(/^[\s\uFEFF\xA0]+|[\s\uFEFF\xA0]+$/g,"")}var nt="__ec_primitive__";function rt(t){t[nt]=!0}function it(t){return t[nt]}var ot=function(){function t(){this.data={}}return t.prototype["delete"]=function(t){var e=this.has(t);return e&&delete this.data[t],e},t.prototype.has=function(t){return this.data.hasOwnProperty(t)},t.prototype.get=function(t){return this.data[t]},t.prototype.set=function(t,e){return this.data[t]=e,this},t.prototype.keys=function(){return A(this.data)},t.prototype.forEach=function(t){var e=this.data;for(var n in e)e.hasOwnProperty(n)&&t(e[n],n)},t}(),at="function"===typeof Map;function st(){return at?new Map:new ot}var lt=function(){function t(e){var n=E(e);this.data=st();var r=this;function i(t,e){n?r.set(t,e):r.set(e,t)}e instanceof t?e.each(i):e&&T(e,i)}return t.prototype.hasKey=function(t){return this.data.has(t)},t.prototype.get=function(t){return this.data.get(t)},t.prototype.set=function(t,e){return this.data.set(t,e),e},t.prototype.each=function(t,e){this.data.forEach((function(n,r){t.call(e,n,r)}))},t.prototype.keys=function(){var t=this.data.keys();return at?Array.from(t):t},t.prototype.removeKey=function(t){this.data["delete"](t)},t}();function ut(t){return new lt(t)}function ct(t,e){for(var n=new t.constructor(t.length+e.length),r=0;r<t.length;r++)n[r]=t[r];var i=t.length;for(r=0;r<e.length;r++)n[r+i]=e[r];return n}function dt(t,e){var n;if(Object.create)n=Object.create(t);else{var r=function(){};r.prototype=t,n=new r}return e&&x(n,e),n}function ht(t){var e=t.style;e.webkitUserSelect="none",e.userSelect="none",e.webkitTapHighlightColor="rgba(0,0,0,0)",e["-webkit-touch-callout"]="none"}function ft(t,e){return t.hasOwnProperty(e)}function pt(){}var mt=180/Math.PI},"6def":function(t,e,n){"use strict";var r=function(){var t=this,e=t._self._c;return t.dot?e("span",{ref:"badge",class:t.classes},[t._t("default"),e("sup",{directives:[{name:"show",rawName:"v-show",value:t.badge,expression:"badge"}],class:t.dotClasses,style:t.styles})],2):t.status||t.color?e("span",{ref:"badge",staticClass:"ivu-badge-status",class:t.classes},[e("span",{class:t.statusClasses,style:t.statusStyles}),e("span",{staticClass:"ivu-badge-status-text"},[t._t("text",(function(){return[t._v(t._s(t.text))]}))],2)]):e("span",{ref:"badge",class:t.classes},[t._t("default"),t.$slots.count?e("sup",{class:t.customCountClasses,style:t.styles},[t._t("count")],2):t.hasCount?e("sup",{directives:[{name:"show",rawName:"v-show",value:t.badge,expression:"badge"}],class:t.countClasses,style:t.styles},[t._t("text",(function(){return[t._v(t._s(t.finalCount))]}))],2):t._e()],2)},i=[],o=n("ade3"),a=(n("99af"),n("a9e3"),n("3a27")),s=["blue","green","red","yellow","pink","magenta","volcano","orange","gold","lime","cyan","geekblue","purple"],l="ivu-badge",u={name:"Badge",props:{count:Number,dot:{type:Boolean,default:!1},overflowCount:{type:[Number,String],default:99},className:String,showZero:{type:Boolean,default:!1},text:{type:String,default:""},status:{validator:function(t){return Object(a["j"])(t,["success","processing","default","error","warning"])}},type:{validator:function(t){return Object(a["j"])(t,["success","primary","normal","error","warning","info"])}},offset:{type:Array},color:{type:String}},computed:{classes:function(){return"".concat(l)},dotClasses:function(){return"".concat(l,"-dot")},countClasses:function(){return["".concat(l,"-count"),Object(o["a"])(Object(o["a"])(Object(o["a"])({},"".concat(this.className),!!this.className),"".concat(l,"-count-alone"),this.alone),"".concat(l,"-count-").concat(this.type),!!this.type)]},customCountClasses:function(){return["".concat(l,"-count"),"".concat(l,"-count-custom"),Object(o["a"])({},"".concat(this.className),!!this.className)]},statusClasses:function(){return["".concat(l,"-status-dot"),Object(o["a"])(Object(o["a"])({},"".concat(l,"-status-").concat(this.status),!!this.status),"".concat(l,"-status-").concat(this.color),!!this.color&&Object(a["j"])(this.color,s))]},statusStyles:function(){return Object(a["j"])(this.color,s)?{}:{backgroundColor:this.color}},styles:function(){var t={};return this.offset&&2===this.offset.length&&(t["margin-top"]="".concat(this.offset[0],"px"),t["margin-right"]="".concat(this.offset[1],"px")),t},finalCount:function(){return""!==this.text?this.text:parseInt(this.count)>=parseInt(this.overflowCount)?"".concat(this.overflowCount,"+"):this.count},badge:function(){var t=!1;return this.count&&(t=!(0===parseInt(this.count))),this.dot&&(t=!0,null!==this.count&&0===parseInt(this.count)&&(t=!1)),""!==this.text&&(t=!0),t||this.showZero},hasCount:function(){return!(!this.count&&""===this.text)||!(!this.showZero||0!==parseInt(this.count))},alone:function(){return void 0===this.$slots.default}}},c=u,d=n("2877"),h=Object(d["a"])(c,r,i,!1,null,null,null),f=h.exports;e["a"]=f},"6e29":function(t,e,n){"use strict";var r=function(){var t=this,e=t._self._c;return e("div",{directives:[{name:"click-outside",rawName:"v-click-outside",value:t.onClickoutside,expression:"onClickoutside"}],class:[t.prefixCls],on:{mouseenter:t.handleMouseenter,mouseleave:t.handleMouseleave}},[e("div",{ref:"reference",class:t.relClasses,on:{click:t.handleClick,contextmenu:function(e){return e.preventDefault(),t.handleRightClick.apply(null,arguments)}}},[t._t("default")],2),e("transition",{attrs:{name:"transition-drop"}},[e("Drop",{directives:[{name:"show",rawName:"v-show",value:t.currentVisible,expression:"currentVisible"},{name:"transfer-dom",rawName:"v-transfer-dom"}],ref:"drop",class:t.dropdownCls,attrs:{placement:t.placement,eventsEnabled:t.eventsEnabled,"data-transfer":t.transfer,transfer:t.transfer},nativeOn:{mouseenter:function(e){return t.handleMouseenter.apply(null,arguments)},mouseleave:function(e){return t.handleMouseleave.apply(null,arguments)}}},[t._t("list")],2)],1)],1)},i=[],o=n("ade3"),a=(n("d3b7"),function(){var t=this,e=t._self._c;return e("div",{staticClass:"ivu-select-dropdown",class:t.className,style:t.styles},[t._t("default")],2)}),s=[],l=(n("b0c0"),n("2b0e")),u=n("3a27"),c=n("c177"),d=l["a"].prototype.$isServer,h=d?function(){}:n("d2e4"),f={name:"Drop",props:{placement:{type:String,default:"bottom-start"},className:{type:String},transfer:{type:Boolean},eventsEnabled:{type:Boolean,default:!1}},data:function(){return{popper:null,width:"",popperStatus:!1,tIndex:this.handleGetIndex()}},computed:{styles:function(){var t={};return this.width&&(t.minWidth="".concat(this.width,"px")),this.transfer&&(t["z-index"]=1060+this.tIndex),t}},methods:{update:function(){var t=this;d||this.$nextTick((function(){t.popper?(t.popper.update(),t.popperStatus=!0):t.popper=new h(t.$parent.$refs.reference,t.$el,{eventsEnabled:t.eventsEnabled,placement:t.placement,modifiers:{computeStyle:{gpuAcceleration:!1},preventOverflow:{boundariesElement:"window"}},onCreate:function(){t.resetTransformOrigin(),t.$nextTick(t.popper.update())},onUpdate:function(){t.resetTransformOrigin()}}),"iSelect"===t.$parent.$options.name&&(t.width=parseInt(Object(u["i"])(t.$parent.$el,"width"))),t.tIndex=t.handleGetIndex()}))},destroy:function(){var t=this;this.popper&&setTimeout((function(){t.popper&&!t.popperStatus&&(t.popper.popper.style.display="none",t.popper.destroy(),t.popper=null),t.popperStatus=!1}),300)},resetTransformOrigin:function(){if(this.popper){var t=this.popper.popper.getAttribute("x-placement"),e=t.split("-")[0],n=t.split("-")[1],r="left"===t||"right"===t;r||(this.popper.popper.style.transformOrigin="bottom"===e||"top"!==e&&"start"===n?"center top":"center bottom")}},handleGetIndex:function(){return Object(c["c"])(),c["d"]}},created:function(){this.$on("on-update-popper",this.update),this.$on("on-destroy-popper",this.destroy)},beforeDestroy:function(){this.$off("on-update-popper",this.update),this.$off("on-destroy-popper",this.destroy),this.popper&&(this.popper.destroy(),this.popper=null)}},p=f,m=n("2877"),g=Object(m["a"])(p,a,s,!1,null,null,null),y=g.exports,v={bind(t,e,n){function r(n){if(t.contains(n.target))return!1;e.expression&&e.value(n)}t.__vueClickOutside__=r,document.addEventListener("click",r)},update(){},unbind(t,e){document.removeEventListener("click",t.__vueClickOutside__),delete t.__vueClickOutside__}},_=n("c780"),b="ivu-dropdown",x={name:"Dropdown",directives:{clickOutside:v,TransferDom:_["a"]},components:{Drop:y},props:{trigger:{validator:function(t){return Object(u["j"])(t,["click","hover","custom","contextMenu"])},default:"hover"},placement:{validator:function(t){return Object(u["j"])(t,["top","top-start","top-end","bottom","bottom-start","bottom-end","left","left-start","left-end","right","right-start","right-end"])},default:"bottom"},visible:{type:Boolean,default:!1},transfer:{type:Boolean,default:function(){return!(!this.$IVIEW||""===this.$IVIEW.transfer)&&this.$IVIEW.transfer}},transferClassName:{type:String},stopPropagation:{type:Boolean,default:!1},capture:{type:Boolean,default:function(){return!this.$IVIEW||this.$IVIEW.capture}},eventsEnabled:{type:Boolean,default:!1}},computed:{transition:function(){return["bottom-start","bottom","bottom-end"].indexOf(this.placement)>-1?"slide-up":"fade"},dropdownCls:function(){return Object(o["a"])(Object(o["a"])({},b+"-transfer",this.transfer),this.transferClassName,this.transferClassName)},relClasses:function(){return["".concat(b,"-rel"),Object(o["a"])({},"".concat(b,"-rel-user-select-none"),"contextMenu"===this.trigger)]}},data:function(){return{prefixCls:b,currentVisible:this.visible}},watch:{visible:function(t){this.currentVisible=t},currentVisible:function(t){t?this.$refs.drop.update():this.$refs.drop.destroy(),this.$emit("on-visible-change",t)}},methods:{handleClick:function(){if("custom"===this.trigger)return!1;if("click"!==this.trigger)return!1;var t=this.hasParent();t||(this.currentVisible=!this.currentVisible)},handleRightClick:function(){return"custom"!==this.trigger&&("contextMenu"===this.trigger&&void(this.currentVisible=!this.currentVisible))},handleMouseenter:function(){var t=this;return"custom"!==this.trigger&&("hover"===this.trigger&&(this.timeout&&clearTimeout(this.timeout),void(this.timeout=setTimeout((function(){t.currentVisible=!0}),250))))},handleMouseleave:function(){var t=this;return"custom"!==this.trigger&&("hover"===this.trigger&&void(this.timeout&&(clearTimeout(this.timeout),this.timeout=setTimeout((function(){t.currentVisible=!1}),150))))},onClickoutside:function(t){this.handleClose(),this.handleRightClose(),this.currentVisible&&this.$emit("on-clickoutside",t)},handleClose:function(){return"custom"!==this.trigger&&("click"===this.trigger&&void(this.currentVisible=!1))},handleRightClose:function(){return"custom"!==this.trigger&&("contextMenu"===this.trigger&&void(this.currentVisible=!1))},hasParent:function(){var t=Object(u["f"])(this,"Dropdown");return t||!1}},mounted:function(){var t=this;this.$on("on-click",(function(e){if(!t.stopPropagation){var n=t.hasParent();n&&n.$emit("on-click",e)}})),this.$on("on-hover-click",(function(){var e=t.hasParent();e?(t.$nextTick((function(){if("custom"===t.trigger)return!1;t.currentVisible=!1})),e.$emit("on-hover-click")):t.$nextTick((function(){if("custom"===t.trigger)return!1;t.currentVisible=!1}))})),this.$on("on-haschild-click",(function(){t.$nextTick((function(){if("custom"===t.trigger)return!1;t.currentVisible=!0}));var e=t.hasParent();e&&e.$emit("on-haschild-click")}))}},w=x,M=Object(m["a"])(w,r,i,!1,null,null,null);e["a"]=M.exports},"6e98":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("it",{months:"gennaio_febbraio_marzo_aprile_maggio_giugno_luglio_agosto_settembre_ottobre_novembre_dicembre".split("_"),monthsShort:"gen_feb_mar_apr_mag_giu_lug_ago_set_ott_nov_dic".split("_"),weekdays:"domenica_lunedì_martedì_mercoledì_giovedì_venerdì_sabato".split("_"),weekdaysShort:"dom_lun_mar_mer_gio_ven_sab".split("_"),weekdaysMin:"do_lu_ma_me_gi_ve_sa".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:function(){return"[Oggi a"+(this.hours()>1?"lle ":0===this.hours()?" ":"ll'")+"]LT"},nextDay:function(){return"[Domani a"+(this.hours()>1?"lle ":0===this.hours()?" ":"ll'")+"]LT"},nextWeek:function(){return"dddd [a"+(this.hours()>1?"lle ":0===this.hours()?" ":"ll'")+"]LT"},lastDay:function(){return"[Ieri a"+(this.hours()>1?"lle ":0===this.hours()?" ":"ll'")+"]LT"},lastWeek:function(){switch(this.day()){case 0:return"[La scorsa] dddd [a"+(this.hours()>1?"lle ":0===this.hours()?" ":"ll'")+"]LT";default:return"[Lo scorso] dddd [a"+(this.hours()>1?"lle ":0===this.hours()?" ":"ll'")+"]LT"}},sameElse:"L"},relativeTime:{future:"tra %s",past:"%s fa",s:"alcuni secondi",ss:"%d secondi",m:"un minuto",mm:"%d minuti",h:"un'ora",hh:"%d ore",d:"un giorno",dd:"%d giorni",w:"una settimana",ww:"%d settimane",M:"un mese",MM:"%d mesi",y:"un anno",yy:"%d anni"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:1,doy:4}});return e}))},"6f12":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("it-ch",{months:"gennaio_febbraio_marzo_aprile_maggio_giugno_luglio_agosto_settembre_ottobre_novembre_dicembre".split("_"),monthsShort:"gen_feb_mar_apr_mag_giu_lug_ago_set_ott_nov_dic".split("_"),weekdays:"domenica_lunedì_martedì_mercoledì_giovedì_venerdì_sabato".split("_"),weekdaysShort:"dom_lun_mar_mer_gio_ven_sab".split("_"),weekdaysMin:"do_lu_ma_me_gi_ve_sa".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[Oggi alle] LT",nextDay:"[Domani alle] LT",nextWeek:"dddd [alle] LT",lastDay:"[Ieri alle] LT",lastWeek:function(){switch(this.day()){case 0:return"[la scorsa] dddd [alle] LT";default:return"[lo scorso] dddd [alle] LT"}},sameElse:"L"},relativeTime:{future:function(t){return(/^[0-9].+$/.test(t)?"tra":"in")+" "+t},past:"%s fa",s:"alcuni secondi",ss:"%d secondi",m:"un minuto",mm:"%d minuti",h:"un'ora",hh:"%d ore",d:"un giorno",dd:"%d giorni",M:"un mese",MM:"%d mesi",y:"un anno",yy:"%d anni"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:1,doy:4}});return e}))},"6f19":function(t,e,n){"use strict";var r=n("9112"),i=n("0d26"),o=n("b980"),a=Error.captureStackTrace;t.exports=function(t,e,n,s){o&&(a?a(t,e):r(t,"stack",i(n,s)))}},"6f50":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("en-nz",{months:"January_February_March_April_May_June_July_August_September_October_November_December".split("_"),monthsShort:"Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec".split("_"),weekdays:"Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday".split("_"),weekdaysShort:"Sun_Mon_Tue_Wed_Thu_Fri_Sat".split("_"),weekdaysMin:"Su_Mo_Tu_We_Th_Fr_Sa".split("_"),longDateFormat:{LT:"h:mm A",LTS:"h:mm:ss A",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY h:mm A",LLLL:"dddd, D MMMM YYYY h:mm A"},calendar:{sameDay:"[Today at] LT",nextDay:"[Tomorrow at] LT",nextWeek:"dddd [at] LT",lastDay:"[Yesterday at] LT",lastWeek:"[Last] dddd [at] LT",sameElse:"L"},relativeTime:{future:"in %s",past:"%s ago",s:"a few seconds",ss:"%d seconds",m:"a minute",mm:"%d minutes",h:"an hour",hh:"%d hours",d:"a day",dd:"%d days",M:"a month",MM:"%d months",y:"a year",yy:"%d years"},dayOfMonthOrdinalParse:/\d{1,2}(st|nd|rd|th)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n},week:{dow:1,doy:4}});return e}))},"6f53":function(t,e,n){"use strict";var r=n("83ab"),i=n("d039"),o=n("e330"),a=n("e163"),s=n("df75"),l=n("fc6a"),u=n("d1e7").f,c=o(u),d=o([].push),h=r&&i((function(){var t=Object.create(null);return t[2]=2,!c(t,2)})),f=function(t){return function(e){var n,i=l(e),o=s(i),u=h&&null===a(i),f=o.length,p=0,m=[];while(f>p)n=o[p++],r&&!(u?n in i:c(i,n))||d(m,t?[n,i[n]]:i[n]);return m}};t.exports={entries:f(!0),values:f(!1)}},"6f6c":function(t,e){var n=/\w*$/;function r(t){var e=new t.constructor(t.source,n.exec(t));return e.lastIndex=t.lastIndex,e}t.exports=r},"6fcd":function(t,e,n){var r=n("50d8"),i=n("d370"),o=n("6747"),a=n("0d24"),s=n("c098"),l=n("73ac"),u=Object.prototype,c=u.hasOwnProperty;function d(t,e){var n=o(t),u=!n&&i(t),d=!n&&!u&&a(t),h=!n&&!u&&!d&&l(t),f=n||u||d||h,p=f?r(t.length,String):[],m=p.length;for(var g in t)!e&&!c.call(t,g)||f&&("length"==g||d&&("offset"==g||"parent"==g)||h&&("buffer"==g||"byteLength"==g||"byteOffset"==g)||s(g,m))||p.push(g);return p}t.exports=d},7118:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="jan._feb._mrt._apr._mai_jun._jul._aug._sep._okt._nov._des.".split("_"),n="jan_feb_mrt_apr_mai_jun_jul_aug_sep_okt_nov_des".split("_"),r=t.defineLocale("fy",{months:"jannewaris_febrewaris_maart_april_maaie_juny_july_augustus_septimber_oktober_novimber_desimber".split("_"),monthsShort:function(t,r){return t?/-MMM-/.test(r)?n[t.month()]:e[t.month()]:e},monthsParseExact:!0,weekdays:"snein_moandei_tiisdei_woansdei_tongersdei_freed_sneon".split("_"),weekdaysShort:"si._mo._ti._wo._to._fr._so.".split("_"),weekdaysMin:"Si_Mo_Ti_Wo_To_Fr_So".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD-MM-YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[hjoed om] LT",nextDay:"[moarn om] LT",nextWeek:"dddd [om] LT",lastDay:"[juster om] LT",lastWeek:"[ôfrûne] dddd [om] LT",sameElse:"L"},relativeTime:{future:"oer %s",past:"%s lyn",s:"in pear sekonden",ss:"%d sekonden",m:"ien minút",mm:"%d minuten",h:"ien oere",hh:"%d oeren",d:"ien dei",dd:"%d dagen",M:"ien moanne",MM:"%d moannen",y:"ien jier",yy:"%d jierren"},dayOfMonthOrdinalParse:/\d{1,2}(ste|de)/,ordinal:function(t){return t+(1===t||8===t||t>=20?"ste":"de")},week:{dow:1,doy:4}});return r}))},7149:function(t,e,n){"use strict";var r=n("23e7"),i=n("d066"),o=n("c430"),a=n("d256"),s=n("4738").CONSTRUCTOR,l=n("cdf9"),u=i("Promise"),c=o&&!s;r({target:"Promise",stat:!0,forced:o||s},{resolve:function(t){return l(c&&this===u?a:this,t)}})},7156:function(t,e,n){"use strict";var r=n("1626"),i=n("861d"),o=n("d2bb");t.exports=function(t,e,n){var a,s;return o&&r(a=e.constructor)&&a!==n&&i(s=a.prototype)&&s!==n.prototype&&o(t,s),t}},7159:function(t,e,n){var r=n("266a"),i=n("9934");function o(t){return null==t?[]:r(t,i(t))}t.exports=o},7234:function(t,e,n){"use strict";t.exports=function(t){return null===t||void 0===t}},"726e":function(t,e,n){"use strict";n.d(e,"c",(function(){return r})),n.d(e,"b",(function(){return i})),n.d(e,"a",(function(){return o})),n.d(e,"d",(function(){return d})),n.d(e,"e",(function(){return h}));var r=12,i="sans-serif",o=r+"px "+i,a=20,s=100,l="007LLmW'55;N0500LLLLLLLLLL00NNNLzWW\\\\WQb\\0FWLg\\bWb\\WQ\\WrWWQ000CL5LLFLL0LL**F*gLLLL5F0LF\\FFF5.5N";function u(t){var e={};if("undefined"===typeof JSON)return e;for(var n=0;n<t.length;n++){var r=String.fromCharCode(n+32),i=(t.charCodeAt(n)-a)/s;e[r]=i}return e}var c=u(l),d={createCanvas:function(){return"undefined"!==typeof document&&document.createElement("canvas")},measureText:function(){var t,e;return function(n,i){if(!t){var a=d.createCanvas();t=a&&a.getContext("2d")}if(t)return e!==i&&(e=t.font=i||o),t.measureText(n);n=n||"",i=i||o;var s=/(\d+)px/.exec(i),l=s&&+s[1]||r,u=0;if(i.indexOf("mono")>=0)u=l*n.length;else for(var h=0;h<n.length;h++){var f=c[n[h]];u+=null==f?l:f*l}return{width:u}}}(),loadImage:function(t,e,n){var r=new Image;return r.onload=e,r.onerror=n,r.src=t,r}};function h(t){for(var e in d)t[e]&&(d[e]=t[e])}},7282:function(t,e,n){"use strict";var r=n("e330"),i=n("59ed");t.exports=function(t,e,n){try{return r(i(Object.getOwnPropertyDescriptor(t,e)[n]))}catch(o){}}},"72af":function(t,e,n){var r=n("99cd"),i=r();t.exports=i},"72f0":function(t,e){function n(t){return function(){return t}}t.exports=n},7333:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("en-il",{months:"January_February_March_April_May_June_July_August_September_October_November_December".split("_"),monthsShort:"Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec".split("_"),weekdays:"Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday".split("_"),weekdaysShort:"Sun_Mon_Tue_Wed_Thu_Fri_Sat".split("_"),weekdaysMin:"Su_Mo_Tu_We_Th_Fr_Sa".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Today at] LT",nextDay:"[Tomorrow at] LT",nextWeek:"dddd [at] LT",lastDay:"[Yesterday at] LT",lastWeek:"[Last] dddd [at] LT",sameElse:"L"},relativeTime:{future:"in %s",past:"%s ago",s:"a few seconds",ss:"%d seconds",m:"a minute",mm:"%d minutes",h:"an hour",hh:"%d hours",d:"a day",dd:"%d days",M:"a month",MM:"%d months",y:"a year",yy:"%d years"},dayOfMonthOrdinalParse:/\d{1,2}(st|nd|rd|th)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n}});return e}))},"73ac":function(t,e,n){var r=n("743f"),i=n("b047"),o=n("99d3"),a=o&&o.isTypedArray,s=a?i(a):r;t.exports=s},7418:function(t,e,n){"use strict";e.f=Object.getOwnPropertySymbols},"743f":function(t,e,n){var r=n("3729"),i=n("b218"),o=n("1310"),a="[object Arguments]",s="[object Array]",l="[object Boolean]",u="[object Date]",c="[object Error]",d="[object Function]",h="[object Map]",f="[object Number]",p="[object Object]",m="[object RegExp]",g="[object Set]",y="[object String]",v="[object WeakMap]",_="[object ArrayBuffer]",b="[object DataView]",x="[object Float32Array]",w="[object Float64Array]",M="[object Int8Array]",S="[object Int16Array]",k="[object Int32Array]",L="[object Uint8Array]",T="[object Uint8ClampedArray]",D="[object Uint16Array]",O="[object Uint32Array]",I={};function C(t){return o(t)&&i(t.length)&&!!I[r(t)]}I[x]=I[w]=I[M]=I[S]=I[k]=I[L]=I[T]=I[D]=I[O]=!0,I[a]=I[s]=I[_]=I[l]=I[b]=I[u]=I[c]=I[d]=I[h]=I[f]=I[p]=I[m]=I[g]=I[y]=I[v]=!1,t.exports=C},"74c8":function(t,e,n){var r=n("972ca"),i=n("242e"),o=n("badf");function a(t,e){return r(t,o(e,3),i)}t.exports=a},"74dc":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("sw",{months:"Januari_Februari_Machi_Aprili_Mei_Juni_Julai_Agosti_Septemba_Oktoba_Novemba_Desemba".split("_"),monthsShort:"Jan_Feb_Mac_Apr_Mei_Jun_Jul_Ago_Sep_Okt_Nov_Des".split("_"),weekdays:"Jumapili_Jumatatu_Jumanne_Jumatano_Alhamisi_Ijumaa_Jumamosi".split("_"),weekdaysShort:"Jpl_Jtat_Jnne_Jtan_Alh_Ijm_Jmos".split("_"),weekdaysMin:"J2_J3_J4_J5_Al_Ij_J1".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"hh:mm A",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[leo saa] LT",nextDay:"[kesho saa] LT",nextWeek:"[wiki ijayo] dddd [saat] LT",lastDay:"[jana] LT",lastWeek:"[wiki iliyopita] dddd [saat] LT",sameElse:"L"},relativeTime:{future:"%s baadaye",past:"tokea %s",s:"hivi punde",ss:"sekunde %d",m:"dakika moja",mm:"dakika %d",h:"saa limoja",hh:"masaa %d",d:"siku moja",dd:"siku %d",M:"mwezi mmoja",MM:"miezi %d",y:"mwaka mmoja",yy:"miaka %d"},week:{dow:1,doy:7}});return e}))},7530:function(t,e,n){var r=n("1a8c"),i=Object.create,o=function(){function t(){}return function(e){if(!r(e))return{};if(i)return i(e);t.prototype=e;var n=new t;return t.prototype=void 0,n}}();t.exports=o},7558:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,r){var i={s:["çend sanîye","çend sanîyeyan"],ss:[t+" sanîye",t+" sanîyeyan"],m:["deqîqeyek","deqîqeyekê"],mm:[t+" deqîqe",t+" deqîqeyan"],h:["saetek","saetekê"],hh:[t+" saet",t+" saetan"],d:["rojek","rojekê"],dd:[t+" roj",t+" rojan"],w:["hefteyek","hefteyekê"],ww:[t+" hefte",t+" hefteyan"],M:["mehek","mehekê"],MM:[t+" meh",t+" mehan"],y:["salek","salekê"],yy:[t+" sal",t+" salan"]};return e?i[n][0]:i[n][1]}function n(t){t=""+t;var e=t.substring(t.length-1),n=t.length>1?t.substring(t.length-2):"";return 12==n||13==n||"2"!=e&&"3"!=e&&"50"!=n&&"70"!=e&&"80"!=e?"ê":"yê"}var r=t.defineLocale("ku-kmr",{months:"Rêbendan_Sibat_Adar_Nîsan_Gulan_Hezîran_Tîrmeh_Tebax_Îlon_Cotmeh_Mijdar_Berfanbar".split("_"),monthsShort:"Rêb_Sib_Ada_Nîs_Gul_Hez_Tîr_Teb_Îlo_Cot_Mij_Ber".split("_"),monthsParseExact:!0,weekdays:"Yekşem_Duşem_Sêşem_Çarşem_Pêncşem_În_Şemî".split("_"),weekdaysShort:"Yek_Du_Sê_Çar_Pên_În_Şem".split("_"),weekdaysMin:"Ye_Du_Sê_Ça_Pê_În_Şe".split("_"),meridiem:function(t,e,n){return t<12?n?"bn":"BN":n?"pn":"PN"},meridiemParse:/bn|BN|pn|PN/,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"Do MMMM[a] YYYY[an]",LLL:"Do MMMM[a] YYYY[an] HH:mm",LLLL:"dddd, Do MMMM[a] YYYY[an] HH:mm",ll:"Do MMM[.] YYYY[an]",lll:"Do MMM[.] YYYY[an] HH:mm",llll:"ddd[.], Do MMM[.] YYYY[an] HH:mm"},calendar:{sameDay:"[Îro di saet] LT [de]",nextDay:"[Sibê di saet] LT [de]",nextWeek:"dddd [di saet] LT [de]",lastDay:"[Duh di saet] LT [de]",lastWeek:"dddd[a borî di saet] LT [de]",sameElse:"L"},relativeTime:{future:"di %s de",past:"berî %s",s:e,ss:e,m:e,mm:e,h:e,hh:e,d:e,dd:e,w:e,ww:e,M:e,MM:e,y:e,yy:e},dayOfMonthOrdinalParse:/\d{1,2}(?:yê|ê|\.)/,ordinal:function(t,e){var r=e.toLowerCase();return r.includes("w")||r.includes("m")?t+".":t+n(t)},week:{dow:1,doy:4}});return r}))},"76dd":function(t,e,n){var r=n("ce86");function i(t){return null==t?"":r(t)}t.exports=i},"76ea":function(t,e,n){"use strict";var r=function(){var t=this,e=t._self._c;return e("div",{class:t.classes,style:t.styles},[t._t("default")],2)},i=[],o=n("53ca"),a=n("ade3"),s=(n("99af"),n("14d9"),n("a9e3"),n("b64b"),n("d3b7"),n("ac1f"),n("00b4"),n("159b"),n("3a27")),l="ivu-col";function u(t){return"number"===typeof t?"".concat(t," ").concat(t," auto"):/^\d+(\.\d+)?(px|em|rem|%)$/.test(t)?"0 0 ".concat(t):t}var c={name:"iCol",props:{span:[Number,String],order:[Number,String],offset:[Number,String],push:[Number,String],pull:[Number,String],className:String,xs:[Number,Object],sm:[Number,Object],md:[Number,Object],lg:[Number,Object],xl:[Number,Object],xxl:[Number,Object],flex:{type:[Number,String],default:""}},data:function(){return{gutter:0}},computed:{classes:function(){var t=this,e=["".concat(l),Object(a["a"])(Object(a["a"])(Object(a["a"])(Object(a["a"])(Object(a["a"])(Object(a["a"])({},"".concat(l,"-span-").concat(this.span),this.span),"".concat(l,"-order-").concat(this.order),this.order),"".concat(l,"-offset-").concat(this.offset),this.offset),"".concat(l,"-push-").concat(this.push),this.push),"".concat(l,"-pull-").concat(this.pull),this.pull),"".concat(this.className),!!this.className)];return["xs","sm","md","lg","xl","xxl"].forEach((function(n){if("number"===typeof t[n])e.push("".concat(l,"-span-").concat(n,"-").concat(t[n]));else if("object"===Object(o["a"])(t[n])){var r=t[n];Object.keys(r).forEach((function(t){e.push("span"!==t?"".concat(l,"-").concat(n,"-").concat(t,"-").concat(r[t]):"".concat(l,"-span-").concat(n,"-").concat(r[t]))}))}})),e},styles:function(){var t={};return 0!==this.gutter&&(t={paddingLeft:this.gutter/2+"px",paddingRight:this.gutter/2+"px"}),this.flex&&(t.flex=u(this.flex)),t}},methods:{updateGutter:function(){var t=Object(s["f"])(this,"Row");t&&t.updateGutter(t.gutter)}},mounted:function(){this.updateGutter()},beforeDestroy:function(){this.updateGutter()}},d=c,h=n("2877"),f=Object(h["a"])(d,r,i,!1,null,null,null),p=f.exports;e["a"]=p},"77c1":function(t,e,n){var r=n("7948"),i=n("badf"),o=n("89d9"),a=n("1bac");function s(t,e){if(null==t)return{};var n=r(a(t),(function(t){return[t]}));return e=i(e),o(t,n,(function(t,n){return e(t,n[0])}))}t.exports=s},7839:function(t,e,n){"use strict";t.exports=["constructor","hasOwnProperty","isPrototypeOf","propertyIsEnumerable","toLocaleString","toString","valueOf"]},"785a":function(t,e,n){"use strict";var r=n("cc12"),i=r("span").classList,o=i&&i.constructor&&i.constructor.prototype;t.exports=o===Object.prototype?void 0:o},7948:function(t,e){function n(t,e){var n=-1,r=null==t?0:t.length,i=Array(r);while(++n<r)i[n]=e(t[n],n,t);return i}t.exports=n},"79bc":function(t,e,n){var r=n("0b07"),i=n("2b3e"),o=r(i,"Map");t.exports=o},"7a12":function(t,e,n){"use strict";var r=function(){var t=this,e=t._self._c;return e(t.tagName,t._b({tag:"component",class:t.classes,attrs:{disabled:t.itemDisabled},on:{click:t.handleClickLink}},"component",t.tagProps,!1),[t.loading?e("Icon",{staticClass:"ivu-load-loop",attrs:{type:"ios-loading"}}):t._e(),!t.icon&&!t.customIcon||t.loading?t._e():e("Icon",{attrs:{type:t.icon,custom:t.customIcon}}),t.showSlot?e("span",{ref:"slot"},[t._t("default")],2):t._e()],1)},i=[],o=n("ade3"),a=(n("99af"),n("5250")),s=n("3a27"),l={props:{to:{type:[Object,String]},replace:{type:Boolean,default:!1},target:{type:String,validator(t){return Object(s["j"])(t,["_blank","_self","_parent","_top"])},default:"_self"},append:{type:Boolean,required:!1,default:!1}},computed:{linkUrl(){const t=typeof this.to;if("string"!==t)return null;if(this.to.includes("//"))return this.to;const e=this.$router;if(e){const t=this.$route,n=e.resolve(this.to,t,this.append);return n?n.href:this.to}return this.to}},methods:{handleOpenTo(){const t=this.$router;let e=this.to;if(t){const n=this.$route,r=t.resolve(this.to,n,this.append);e=r?r.href:this.to}"string"!==typeof this.to&&window.open(e)},handleClick(t=!1){const e=this.$router;t?this.handleOpenTo():e?"string"===typeof this.to&&this.to.includes("//")?window.location.href=this.to:this.replace?this.$router.replace(this.to,()=>{}):this.$router.push(this.to,()=>{}):window.location.href=this.to},handleCheckClick(t,e=!1){if(this.to){if("_blank"===this.target)return this.handleOpenTo(),!1;t.preventDefault(),this.handleClick(e)}}}},u=n("9f81"),c="ivu-btn",d={name:"Button",mixins:[l,u["a"]],components:{Icon:a["a"]},props:{type:{validator:function(t){return Object(s["j"])(t,["default","primary","dashed","text","info","success","warning","error"])},default:"default"},shape:{validator:function(t){return Object(s["j"])(t,["circle","circle-outline"])}},size:{validator:function(t){return Object(s["j"])(t,["small","large","default"])},default:function(){return this.$IVIEW&&""!==this.$IVIEW.size?this.$IVIEW.size:"default"}},loading:Boolean,disabled:Boolean,htmlType:{default:"button",validator:function(t){return Object(s["j"])(t,["button","submit","reset"])}},icon:{type:String,default:""},customIcon:{type:String,default:""},long:{type:Boolean,default:!1},ghost:{type:Boolean,default:!1}},computed:{showSlot:function(){return!!this.$slots.default},classes:function(){return["".concat(c),"".concat(c,"-").concat(this.type),Object(o["a"])(Object(o["a"])(Object(o["a"])(Object(o["a"])(Object(o["a"])(Object(o["a"])({},"".concat(c,"-long"),this.long),"".concat(c,"-").concat(this.shape),!!this.shape),"".concat(c,"-").concat(this.size),"default"!==this.size),"".concat(c,"-loading"),null!=this.loading&&this.loading),"".concat(c,"-icon-only"),!this.showSlot&&(!!this.icon||!!this.customIcon||this.loading)),"".concat(c,"-ghost"),this.ghost)]},isHrefPattern:function(){var t=this.to;return!!t},tagName:function(){var t=this.isHrefPattern;return t?"a":"button"},tagProps:function(){var t=this.isHrefPattern;if(t){var e=this.linkUrl,n=this.target;return{href:e,target:n}}var r=this.htmlType;return{type:r}}},methods:{handleClickLink:function(t){this.$emit("click",t);var e=t.ctrlKey||t.metaKey;this.handleCheckClick(t,e)}}},h=d,f=n("2877"),p=Object(f["a"])(h,r,i,!1,null,null,null);e["a"]=p.exports},"7a29":function(t,e,n){"use strict";(function(t){n.d(e,"p",(function(){return s})),n.d(e,"j",(function(){return u})),n.d(e,"q",(function(){return d})),n.d(e,"e",(function(){return h})),n.d(e,"a",(function(){return f})),n.d(e,"b",(function(){return p})),n.d(e,"i",(function(){return m})),n.d(e,"h",(function(){return g})),n.d(e,"l",(function(){return y})),n.d(e,"n",(function(){return _})),n.d(e,"m",(function(){return b})),n.d(e,"o",(function(){return x})),n.d(e,"k",(function(){return w})),n.d(e,"d",(function(){return M})),n.d(e,"f",(function(){return S})),n.d(e,"g",(function(){return k})),n.d(e,"c",(function(){return L}));var r=n("6d8b"),i=n("41ef"),o=n("22d1"),a=Math.round;function s(t){var e;if(t&&"transparent"!==t){if("string"===typeof t&&t.indexOf("rgba")>-1){var n=Object(i["h"])(t);n&&(t="rgb("+n[0]+","+n[1]+","+n[2]+")",e=n[3])}}else t="none";return{color:t,opacity:null==e?1:e}}var l=1e-4;function u(t){return t<l&&t>-l}function c(t){return a(1e3*t)/1e3}function d(t){return a(1e4*t)/1e4}function h(t){return"matrix("+c(t[0])+","+c(t[1])+","+c(t[2])+","+c(t[3])+","+d(t[4])+","+d(t[5])+")"}var f={left:"start",right:"end",center:"middle",middle:"middle"};function p(t,e,n){return"top"===n?t+=e/2:"bottom"===n&&(t-=e/2),t}function m(t){return t&&(t.shadowBlur||t.shadowOffsetX||t.shadowOffsetY)}function g(t){var e=t.style,n=t.getGlobalScale();return[e.shadowColor,(e.shadowBlur||0).toFixed(2),(e.shadowOffsetX||0).toFixed(2),(e.shadowOffsetY||0).toFixed(2),n[0],n[1]].join(",")}function y(t){return t&&!!t.image}function v(t){return t&&!!t.svgElement}function _(t){return y(t)||v(t)}function b(t){return"linear"===t.type}function x(t){return"radial"===t.type}function w(t){return t&&("linear"===t.type||"radial"===t.type)}function M(t){return"url(#"+t+")"}function S(t){var e=t.getGlobalScale(),n=Math.max(e[0],e[1]);return Math.max(Math.ceil(Math.log(n)/Math.log(10)),1)}function k(t){var e=t.x||0,n=t.y||0,i=(t.rotation||0)*r["a"],o=Object(r["P"])(t.scaleX,1),s=Object(r["P"])(t.scaleY,1),l=t.skewX||0,u=t.skewY||0,c=[];return(e||n)&&c.push("translate("+e+"px,"+n+"px)"),i&&c.push("rotate("+i+")"),1===o&&1===s||c.push("scale("+o+","+s+")"),(l||u)&&c.push("skew("+a(l*r["a"])+"deg, "+a(u*r["a"])+"deg)"),c.join(" ")}var L=function(){return o["a"].hasGlobalWindow&&Object(r["w"])(window.btoa)?function(t){return window.btoa(unescape(encodeURIComponent(t)))}:"undefined"!==typeof t?function(e){return t.from(e).toString("base64")}:function(t){return null}}()}).call(this,n("b639").Buffer)},"7a48":function(t,e,n){var r=n("6044"),i=Object.prototype,o=i.hasOwnProperty;function a(t){var e=this.__data__;return r?void 0!==e[t]:o.call(e,t)}t.exports=a},"7aa2":function(t,e,n){var r=n("ec47"),i=n("9934"),o=r(i);t.exports=o},"7b0b":function(t,e,n){"use strict";var r=n("1d80"),i=Object;t.exports=function(t){return i(r(t))}},"7b83":function(t,e,n){var r=n("7c64"),i=n("93ed"),o=n("2478"),a=n("a524"),s=n("1fc8");function l(t){var e=-1,n=null==t?0:t.length;this.clear();while(++e<n){var r=t[e];this.set(r[0],r[1])}}l.prototype.clear=r,l.prototype["delete"]=i,l.prototype.get=o,l.prototype.has=a,l.prototype.set=s,t.exports=l},"7b97":function(t,e,n){var r=n("7e64"),i=n("a2be"),o=n("1c3c"),a=n("b1e5"),s=n("42a2"),l=n("6747"),u=n("0d24"),c=n("73ac"),d=1,h="[object Arguments]",f="[object Array]",p="[object Object]",m=Object.prototype,g=m.hasOwnProperty;function y(t,e,n,m,y,v){var _=l(t),b=l(e),x=_?f:s(t),w=b?f:s(e);x=x==h?p:x,w=w==h?p:w;var M=x==p,S=w==p,k=x==w;if(k&&u(t)){if(!u(e))return!1;_=!0,M=!1}if(k&&!M)return v||(v=new r),_||c(t)?i(t,e,n,m,y,v):o(t,e,x,n,m,y,v);if(!(n&d)){var L=M&&g.call(t,"__wrapped__"),T=S&&g.call(e,"__wrapped__");if(L||T){var D=L?t.value():t,O=T?e.value():e;return v||(v=new r),y(D,O,n,m,v)}}return!!k&&(v||(v=new r),a(t,e,n,m,y,v))}t.exports=y},"7be6":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="január_február_marec_apríl_máj_jún_júl_august_september_október_november_december".split("_"),n="jan_feb_mar_apr_máj_jún_júl_aug_sep_okt_nov_dec".split("_");function r(t){return t>1&&t<5}function i(t,e,n,i){var o=t+" ";switch(n){case"s":return e||i?"pár sekúnd":"pár sekundami";case"ss":return e||i?o+(r(t)?"sekundy":"sekúnd"):o+"sekundami";case"m":return e?"minúta":i?"minútu":"minútou";case"mm":return e||i?o+(r(t)?"minúty":"minút"):o+"minútami";case"h":return e?"hodina":i?"hodinu":"hodinou";case"hh":return e||i?o+(r(t)?"hodiny":"hodín"):o+"hodinami";case"d":return e||i?"deň":"dňom";case"dd":return e||i?o+(r(t)?"dni":"dní"):o+"dňami";case"M":return e||i?"mesiac":"mesiacom";case"MM":return e||i?o+(r(t)?"mesiace":"mesiacov"):o+"mesiacmi";case"y":return e||i?"rok":"rokom";case"yy":return e||i?o+(r(t)?"roky":"rokov"):o+"rokmi"}}var o=t.defineLocale("sk",{months:e,monthsShort:n,weekdays:"nedeľa_pondelok_utorok_streda_štvrtok_piatok_sobota".split("_"),weekdaysShort:"ne_po_ut_st_št_pi_so".split("_"),weekdaysMin:"ne_po_ut_st_št_pi_so".split("_"),longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY H:mm",LLLL:"dddd D. MMMM YYYY H:mm"},calendar:{sameDay:"[dnes o] LT",nextDay:"[zajtra o] LT",nextWeek:function(){switch(this.day()){case 0:return"[v nedeľu o] LT";case 1:case 2:return"[v] dddd [o] LT";case 3:return"[v stredu o] LT";case 4:return"[vo štvrtok o] LT";case 5:return"[v piatok o] LT";case 6:return"[v sobotu o] LT"}},lastDay:"[včera o] LT",lastWeek:function(){switch(this.day()){case 0:return"[minulú nedeľu o] LT";case 1:case 2:return"[minulý] dddd [o] LT";case 3:return"[minulú stredu o] LT";case 4:case 5:return"[minulý] dddd [o] LT";case 6:return"[minulú sobotu o] LT"}},sameElse:"L"},relativeTime:{future:"za %s",past:"pred %s",s:i,ss:i,m:i,mm:i,h:i,hh:i,d:i,dd:i,M:i,MM:i,y:i,yy:i},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return o}))},"7c64":function(t,e,n){var r=n("e24b"),i=n("5e2e"),o=n("79bc");function a(){this.size=0,this.__data__={hash:new r,map:new(o||i),string:new r}}t.exports=a},"7c73":function(t,e,n){"use strict";var r,i=n("825a"),o=n("37e8"),a=n("7839"),s=n("d012"),l=n("1be4"),u=n("cc12"),c=n("f772"),d=">",h="<",f="prototype",p="script",m=c("IE_PROTO"),g=function(){},y=function(t){return h+p+d+t+h+"/"+p+d},v=function(t){t.write(y("")),t.close();var e=t.parentWindow.Object;return t=null,e},_=function(){var t,e=u("iframe"),n="java"+p+":";return e.style.display="none",l.appendChild(e),e.src=String(n),t=e.contentWindow.document,t.open(),t.write(y("document.F=Object")),t.close(),t.F},b=function(){try{r=new ActiveXObject("htmlfile")}catch(e){}b="undefined"!=typeof document?document.domain&&r?v(r):_():v(r);var t=a.length;while(t--)delete b[f][a[t]];return b()};s[m]=!0,t.exports=Object.create||function(t,e){var n;return null!==t?(g[f]=i(t),n=new g,g[f]=null,n[m]=t):n=b(),void 0===e?n:o.f(n,e)}},"7d1f":function(t,e,n){var r=n("087d"),i=n("6747");function o(t,e,n){var o=e(t);return i(t)?o:r(o,n(t))}t.exports=o},"7e64":function(t,e,n){var r=n("5e2e"),i=n("efb6"),o=n("2fcc"),a=n("802a"),s=n("55a3"),l=n("d02c");function u(t){var e=this.__data__=new r(t);this.size=e.size}u.prototype.clear=i,u.prototype["delete"]=o,u.prototype.get=a,u.prototype.has=s,u.prototype.set=l,t.exports=u},"7ed2":function(t,e){var n="__lodash_hash_undefined__";function r(t){return this.__data__.set(t,n),this}t.exports=r},"7f33":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("yo",{months:"Sẹ́rẹ́_Èrèlè_Ẹrẹ̀nà_Ìgbé_Èbibi_Òkùdu_Agẹmo_Ògún_Owewe_Ọ̀wàrà_Bélú_Ọ̀pẹ̀̀".split("_"),monthsShort:"Sẹ́r_Èrl_Ẹrn_Ìgb_Èbi_Òkù_Agẹ_Ògú_Owe_Ọ̀wà_Bél_Ọ̀pẹ̀̀".split("_"),weekdays:"Àìkú_Ajé_Ìsẹ́gun_Ọjọ́rú_Ọjọ́bọ_Ẹtì_Àbámẹ́ta".split("_"),weekdaysShort:"Àìk_Ajé_Ìsẹ́_Ọjr_Ọjb_Ẹtì_Àbá".split("_"),weekdaysMin:"Àì_Aj_Ìs_Ọr_Ọb_Ẹt_Àb".split("_"),longDateFormat:{LT:"h:mm A",LTS:"h:mm:ss A",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY h:mm A",LLLL:"dddd, D MMMM YYYY h:mm A"},calendar:{sameDay:"[Ònì ni] LT",nextDay:"[Ọ̀la ni] LT",nextWeek:"dddd [Ọsẹ̀ tón'bọ] [ni] LT",lastDay:"[Àna ni] LT",lastWeek:"dddd [Ọsẹ̀ tólọ́] [ni] LT",sameElse:"L"},relativeTime:{future:"ní %s",past:"%s kọjá",s:"ìsẹjú aayá die",ss:"aayá %d",m:"ìsẹjú kan",mm:"ìsẹjú %d",h:"wákati kan",hh:"wákati %d",d:"ọjọ́ kan",dd:"ọjọ́ %d",M:"osù kan",MM:"osù %d",y:"ọdún kan",yy:"ọdún %d"},dayOfMonthOrdinalParse:/ọjọ́\s\d{1,2}/,ordinal:"ọjọ́ %d",week:{dow:1,doy:4}});return e}))},"802a":function(t,e){function n(t){return this.__data__.get(t)}t.exports=n},8057:function(t,e){function n(t,e){var n=-1,r=null==t?0:t.length;while(++n<r)if(!1===e(t[n],n,t))break;return t}t.exports=n},8144:function(t,e,n){"use strict";var r=n("2430");e["a"]=r["a"]},8155:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,r){var i=t+" ";switch(n){case"s":return e||r?"nekaj sekund":"nekaj sekundami";case"ss":return i+=1===t?e?"sekundo":"sekundi":2===t?e||r?"sekundi":"sekundah":t<5?e||r?"sekunde":"sekundah":"sekund",i;case"m":return e?"ena minuta":"eno minuto";case"mm":return i+=1===t?e?"minuta":"minuto":2===t?e||r?"minuti":"minutama":t<5?e||r?"minute":"minutami":e||r?"minut":"minutami",i;case"h":return e?"ena ura":"eno uro";case"hh":return i+=1===t?e?"ura":"uro":2===t?e||r?"uri":"urama":t<5?e||r?"ure":"urami":e||r?"ur":"urami",i;case"d":return e||r?"en dan":"enim dnem";case"dd":return i+=1===t?e||r?"dan":"dnem":2===t?e||r?"dni":"dnevoma":e||r?"dni":"dnevi",i;case"M":return e||r?"en mesec":"enim mesecem";case"MM":return i+=1===t?e||r?"mesec":"mesecem":2===t?e||r?"meseca":"mesecema":t<5?e||r?"mesece":"meseci":e||r?"mesecev":"meseci",i;case"y":return e||r?"eno leto":"enim letom";case"yy":return i+=1===t?e||r?"leto":"letom":2===t?e||r?"leti":"letoma":t<5?e||r?"leta":"leti":e||r?"let":"leti",i}}var n=t.defineLocale("sl",{months:"januar_februar_marec_april_maj_junij_julij_avgust_september_oktober_november_december".split("_"),monthsShort:"jan._feb._mar._apr._maj._jun._jul._avg._sep._okt._nov._dec.".split("_"),monthsParseExact:!0,weekdays:"nedelja_ponedeljek_torek_sreda_četrtek_petek_sobota".split("_"),weekdaysShort:"ned._pon._tor._sre._čet._pet._sob.".split("_"),weekdaysMin:"ne_po_to_sr_če_pe_so".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD. MM. YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY H:mm",LLLL:"dddd, D. MMMM YYYY H:mm"},calendar:{sameDay:"[danes ob] LT",nextDay:"[jutri ob] LT",nextWeek:function(){switch(this.day()){case 0:return"[v] [nedeljo] [ob] LT";case 3:return"[v] [sredo] [ob] LT";case 6:return"[v] [soboto] [ob] LT";case 1:case 2:case 4:case 5:return"[v] dddd [ob] LT"}},lastDay:"[včeraj ob] LT",lastWeek:function(){switch(this.day()){case 0:return"[prejšnjo] [nedeljo] [ob] LT";case 3:return"[prejšnjo] [sredo] [ob] LT";case 6:return"[prejšnjo] [soboto] [ob] LT";case 1:case 2:case 4:case 5:return"[prejšnji] dddd [ob] LT"}},sameElse:"L"},relativeTime:{future:"čez %s",past:"pred %s",s:e,ss:e,m:e,mm:e,h:e,hh:e,d:e,dd:e,M:e,MM:e,y:e,yy:e},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:7}});return n}))},8172:function(t,e,n){"use strict";var r=n("e065"),i=n("57b9");r("toPrimitive"),i()},"81e9":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="nolla yksi kaksi kolme neljä viisi kuusi seitsemän kahdeksan yhdeksän".split(" "),n=["nolla","yhden","kahden","kolmen","neljän","viiden","kuuden",e[7],e[8],e[9]];function r(t,e,n,r){var o="";switch(n){case"s":return r?"muutaman sekunnin":"muutama sekunti";case"ss":o=r?"sekunnin":"sekuntia";break;case"m":return r?"minuutin":"minuutti";case"mm":o=r?"minuutin":"minuuttia";break;case"h":return r?"tunnin":"tunti";case"hh":o=r?"tunnin":"tuntia";break;case"d":return r?"päivän":"päivä";case"dd":o=r?"päivän":"päivää";break;case"M":return r?"kuukauden":"kuukausi";case"MM":o=r?"kuukauden":"kuukautta";break;case"y":return r?"vuoden":"vuosi";case"yy":o=r?"vuoden":"vuotta";break}return o=i(t,r)+" "+o,o}function i(t,r){return t<10?r?n[t]:e[t]:t}var o=t.defineLocale("fi",{months:"tammikuu_helmikuu_maaliskuu_huhtikuu_toukokuu_kesäkuu_heinäkuu_elokuu_syyskuu_lokakuu_marraskuu_joulukuu".split("_"),monthsShort:"tammi_helmi_maalis_huhti_touko_kesä_heinä_elo_syys_loka_marras_joulu".split("_"),weekdays:"sunnuntai_maanantai_tiistai_keskiviikko_torstai_perjantai_lauantai".split("_"),weekdaysShort:"su_ma_ti_ke_to_pe_la".split("_"),weekdaysMin:"su_ma_ti_ke_to_pe_la".split("_"),longDateFormat:{LT:"HH.mm",LTS:"HH.mm.ss",L:"DD.MM.YYYY",LL:"Do MMMM[ta] YYYY",LLL:"Do MMMM[ta] YYYY, [klo] HH.mm",LLLL:"dddd, Do MMMM[ta] YYYY, [klo] HH.mm",l:"D.M.YYYY",ll:"Do MMM YYYY",lll:"Do MMM YYYY, [klo] HH.mm",llll:"ddd, Do MMM YYYY, [klo] HH.mm"},calendar:{sameDay:"[tänään] [klo] LT",nextDay:"[huomenna] [klo] LT",nextWeek:"dddd [klo] LT",lastDay:"[eilen] [klo] LT",lastWeek:"[viime] dddd[na] [klo] LT",sameElse:"L"},relativeTime:{future:"%s päästä",past:"%s sitten",s:r,ss:r,m:r,mm:r,h:r,hh:r,d:r,dd:r,M:r,MM:r,y:r,yy:r},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return o}))},8230:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"١",2:"٢",3:"٣",4:"٤",5:"٥",6:"٦",7:"٧",8:"٨",9:"٩",0:"٠"},n={"١":"1","٢":"2","٣":"3","٤":"4","٥":"5","٦":"6","٧":"7","٨":"8","٩":"9","٠":"0"},r=t.defineLocale("ar-sa",{months:"يناير_فبراير_مارس_أبريل_مايو_يونيو_يوليو_أغسطس_سبتمبر_أكتوبر_نوفمبر_ديسمبر".split("_"),monthsShort:"يناير_فبراير_مارس_أبريل_مايو_يونيو_يوليو_أغسطس_سبتمبر_أكتوبر_نوفمبر_ديسمبر".split("_"),weekdays:"الأحد_الإثنين_الثلاثاء_الأربعاء_الخميس_الجمعة_السبت".split("_"),weekdaysShort:"أحد_إثنين_ثلاثاء_أربعاء_خميس_جمعة_سبت".split("_"),weekdaysMin:"ح_ن_ث_ر_خ_ج_س".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},meridiemParse:/ص|م/,isPM:function(t){return"م"===t},meridiem:function(t,e,n){return t<12?"ص":"م"},calendar:{sameDay:"[اليوم على الساعة] LT",nextDay:"[غدا على الساعة] LT",nextWeek:"dddd [على الساعة] LT",lastDay:"[أمس على الساعة] LT",lastWeek:"dddd [على الساعة] LT",sameElse:"L"},relativeTime:{future:"في %s",past:"منذ %s",s:"ثوان",ss:"%d ثانية",m:"دقيقة",mm:"%d دقائق",h:"ساعة",hh:"%d ساعات",d:"يوم",dd:"%d أيام",M:"شهر",MM:"%d أشهر",y:"سنة",yy:"%d سنوات"},preparse:function(t){return t.replace(/[١٢٣٤٥٦٧٨٩٠]/g,(function(t){return n[t]})).replace(/،/g,",")},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]})).replace(/,/g,"،")},week:{dow:0,doy:6}});return r}))},"825a":function(t,e,n){"use strict";var r=n("861d"),i=String,o=TypeError;t.exports=function(t){if(r(t))return t;throw new o(i(t)+" is not an object")}},8296:function(t,e,n){var r=n("656b"),i=n("2b10");function o(t,e){return e.length<2?t:r(t,i(e,0,-1))}t.exports=o},8332:function(t,e,n){t.exports=n("7aa2")},"83ab":function(t,e,n){"use strict";var r=n("d039");t.exports=!r((function(){return 7!==Object.defineProperty({},1,{get:function(){return 7}})[1]}))},8418:function(t,e,n){"use strict";var r=n("83ab"),i=n("9bf2"),o=n("5c6c");t.exports=function(t,e,n){r?i.f(t,e,o(0,n)):t[e]=n}},"84aa":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("bg",{months:"януари_февруари_март_април_май_юни_юли_август_септември_октомври_ноември_декември".split("_"),monthsShort:"яну_фев_мар_апр_май_юни_юли_авг_сеп_окт_ное_дек".split("_"),weekdays:"неделя_понеделник_вторник_сряда_четвъртък_петък_събота".split("_"),weekdaysShort:"нед_пон_вто_сря_чет_пет_съб".split("_"),weekdaysMin:"нд_пн_вт_ср_чт_пт_сб".split("_"),longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"D.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY H:mm",LLLL:"dddd, D MMMM YYYY H:mm"},calendar:{sameDay:"[Днес в] LT",nextDay:"[Утре в] LT",nextWeek:"dddd [в] LT",lastDay:"[Вчера в] LT",lastWeek:function(){switch(this.day()){case 0:case 3:case 6:return"[Миналата] dddd [в] LT";case 1:case 2:case 4:case 5:return"[Миналия] dddd [в] LT"}},sameElse:"L"},relativeTime:{future:"след %s",past:"преди %s",s:"няколко секунди",ss:"%d секунди",m:"минута",mm:"%d минути",h:"час",hh:"%d часа",d:"ден",dd:"%d дена",w:"седмица",ww:"%d седмици",M:"месец",MM:"%d месеца",y:"година",yy:"%d години"},dayOfMonthOrdinalParse:/\d{1,2}-(ев|ен|ти|ви|ри|ми)/,ordinal:function(t){var e=t%10,n=t%100;return 0===t?t+"-ев":0===n?t+"-ен":n>10&&n<20?t+"-ти":1===e?t+"-ви":2===e?t+"-ри":7===e||8===e?t+"-ми":t+"-ти"},week:{dow:1,doy:7}});return e}))},"85e3":function(t,e){function n(t,e,n){switch(n.length){case 0:return t.call(e);case 1:return t.call(e,n[0]);case 2:return t.call(e,n[0],n[1]);case 3:return t.call(e,n[0],n[1],n[2])}return t.apply(e,n)}t.exports=n},8604:function(t,e,n){var r=n("26e8"),i=n("e2c0");function o(t,e){return null!=t&&i(t,e,r)}t.exports=o},"861d":function(t,e,n){"use strict";var r=n("1626");t.exports=function(t){return"object"==typeof t?null!==t:r(t)}},8629:function(t,e,n){"use strict";var r=n("5334");e["a"]=r["a"]},8679:function(t,e,n){"use strict";var r=function(){var t=this,e=t._self._c;return e("div",{class:t.wrapClasses},[t._t("default")],2)},i=[],o=n("ade3"),a=(n("b0c0"),n("d3b7"),"ivu-layout"),s={name:"Layout",data:function(){return{hasSider:!1}},computed:{wrapClasses:function(){return["".concat(a),Object(o["a"])({},"".concat(a,"-has-sider"),this.hasSider)]}},methods:{findSider:function(){return this.$children.some((function(t){return"Sider"===t.$options.name}))}},mounted:function(){this.hasSider=this.findSider()}},l=s,u=n("2877"),c=Object(u["a"])(l,r,i,!1,null,null,null),d=c.exports,h=function(){var t=this,e=t._self._c;return e("div",{class:t.wrapClasses},[t._t("default")],2)},f=[],p="ivu-layout",m={name:"Header",computed:{wrapClasses:function(){return"".concat(p,"-header")}}},g=m,y=Object(u["a"])(g,h,f,!1,null,null,null),v=y.exports,_=function(){var t=this,e=t._self._c;return e("div",{class:t.wrapClasses,style:t.wrapStyles},[e("span",{directives:[{name:"show",rawName:"v-show",value:t.showZeroTrigger,expression:"showZeroTrigger"}],class:t.zeroWidthTriggerClasses,on:{click:t.toggleCollapse}},[e("i",{staticClass:"ivu-icon ivu-icon-ios-menu"})]),e("div",{class:t.childClasses},[t._t("default")],2),t._t("trigger",(function(){return[e("div",{directives:[{name:"show",rawName:"v-show",value:t.showBottomTrigger,expression:"showBottomTrigger"}],class:t.triggerClasses,style:{width:t.siderWidth+"px"},on:{click:t.toggleCollapse}},[e("i",{class:t.triggerIconClasses})])]}))],2)},b=[],x=(n("a9e3"),n("de48")),w=n("3a27"),M="ivu-layout-sider";Object(w["k"])();var S={name:"Sider",props:{value:{type:Boolean,default:!1},width:{type:[Number,String],default:200},collapsedWidth:{type:[Number,String],default:64},hideTrigger:{type:Boolean,default:!1},breakpoint:{type:String,validator:function(t){return Object(w["j"])(t,["xs","sm","md","lg","xl","xxl"])}},collapsible:{type:Boolean,default:!1},defaultCollapsed:{type:Boolean,default:!1},reverseArrow:{type:Boolean,default:!1}},data:function(){return{prefixCls:M,mediaMatched:!1}},computed:{wrapClasses:function(){return["".concat(M),this.siderWidth?"":"".concat(M,"-zero-width"),this.value?"".concat(M,"-collapsed"):""]},wrapStyles:function(){return{width:"".concat(this.siderWidth,"px"),minWidth:"".concat(this.siderWidth,"px"),maxWidth:"".concat(this.siderWidth,"px"),flex:"0 0 ".concat(this.siderWidth,"px")}},triggerClasses:function(){return["".concat(M,"-trigger"),this.value?"".concat(M,"-trigger-collapsed"):""]},childClasses:function(){return"".concat(this.prefixCls,"-children")},zeroWidthTriggerClasses:function(){return["".concat(M,"-zero-width-trigger"),this.reverseArrow?"".concat(M,"-zero-width-trigger-left"):""]},triggerIconClasses:function(){return["ivu-icon","ivu-icon-ios-arrow-".concat(this.reverseArrow?"forward":"back"),"".concat(M,"-trigger-icon")]},siderWidth:function(){return this.collapsible?this.value?this.mediaMatched?0:parseInt(this.collapsedWidth):parseInt(this.width):this.width},showZeroTrigger:function(){return!!this.collapsible&&(this.mediaMatched&&!this.hideTrigger||0===parseInt(this.collapsedWidth)&&this.value&&!this.hideTrigger)},showBottomTrigger:function(){return!!this.collapsible&&(!this.mediaMatched&&!this.hideTrigger)}},methods:{toggleCollapse:function(){var t=!!this.collapsible&&!this.value;this.$emit("input",t)},matchMedia:function(){var t;window.matchMedia&&(t=window.matchMedia);var e=this.mediaMatched;this.mediaMatched=t("(max-width: ".concat(w["c"][this.breakpoint],")")).matches,this.mediaMatched!==e&&this.$emit("input",this.mediaMatched)},onWindowResize:function(){this.matchMedia()}},watch:{value:function(t){this.$emit("on-collapse",t)}},mounted:function(){this.defaultCollapsed&&this.$emit("input",this.defaultCollapsed),void 0!==this.breakpoint&&(Object(x["b"])(window,"resize",this.onWindowResize),this.matchMedia())},beforeDestroy:function(){void 0!==this.breakpoint&&Object(x["a"])(window,"resize",this.onWindowResize)}},k=S,L=Object(u["a"])(k,_,b,!1,null,null,null),T=L.exports,D=function(){var t=this,e=t._self._c;return e("div",{class:t.wrapClasses},[t._t("default")],2)},O=[],I="ivu-layout",C={name:"Content",computed:{wrapClasses:function(){return"".concat(I,"-content")}}},A=C,Y=Object(u["a"])(A,D,O,!1,null,null,null),j=Y.exports,P=function(){var t=this,e=t._self._c;return e("div",{class:t.wrapClasses},[t._t("default")],2)},E=[],N="ivu-layout",R={name:"Footer",computed:{wrapClasses:function(){return"".concat(N,"-footer")}}},H=R,z=Object(u["a"])(H,P,E,!1,null,null,null),F=z.exports;d.Header=v,d.Sider=T,d.Content=j,d.Footer=F;e["a"]=d},8689:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"၁",2:"၂",3:"၃",4:"၄",5:"၅",6:"၆",7:"၇",8:"၈",9:"၉",0:"၀"},n={"၁":"1","၂":"2","၃":"3","၄":"4","၅":"5","၆":"6","၇":"7","၈":"8","၉":"9","၀":"0"},r=t.defineLocale("my",{months:"ဇန်နဝါရီ_ဖေဖော်ဝါရီ_မတ်_ဧပြီ_မေ_ဇွန်_ဇူလိုင်_သြဂုတ်_စက်တင်ဘာ_အောက်တိုဘာ_နိုဝင်ဘာ_ဒီဇင်ဘာ".split("_"),monthsShort:"ဇန်_ဖေ_မတ်_ပြီ_မေ_ဇွန်_လိုင်_သြ_စက်_အောက်_နို_ဒီ".split("_"),weekdays:"တနင်္ဂနွေ_တနင်္လာ_အင်္ဂါ_ဗုဒ္ဓဟူး_ကြာသပတေး_သောကြာ_စနေ".split("_"),weekdaysShort:"နွေ_လာ_ဂါ_ဟူး_ကြာ_သော_နေ".split("_"),weekdaysMin:"နွေ_လာ_ဂါ_ဟူး_ကြာ_သော_နေ".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[ယနေ.] LT [မှာ]",nextDay:"[မနက်ဖြန်] LT [မှာ]",nextWeek:"dddd LT [မှာ]",lastDay:"[မနေ.က] LT [မှာ]",lastWeek:"[ပြီးခဲ့သော] dddd LT [မှာ]",sameElse:"L"},relativeTime:{future:"လာမည့် %s မှာ",past:"လွန်ခဲ့သော %s က",s:"စက္ကန်.အနည်းငယ်",ss:"%d စက္ကန့်",m:"တစ်မိနစ်",mm:"%d မိနစ်",h:"တစ်နာရီ",hh:"%d နာရီ",d:"တစ်ရက်",dd:"%d ရက်",M:"တစ်လ",MM:"%d လ",y:"တစ်နှစ်",yy:"%d နှစ်"},preparse:function(t){return t.replace(/[၁၂၃၄၅၆၇၈၉၀]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},week:{dow:1,doy:4}});return r}))},"86e1":function(t,e,n){var r=n("85e3"),i=n("e2e4"),o=n("4416"),a=n("8296"),s=n("f4d6");function l(t,e,n){e=i(e,t),t=a(t,e);var l=null==t?t:t[s(o(e))];return null==l?void 0:r(l,t,n)}t.exports=l},"872a":function(t,e,n){var r=n("3b4a");function i(t,e,n){"__proto__"==e&&r?r(t,e,{configurable:!0,enumerable:!0,value:n,writable:!0}):t[e]=n}t.exports=i},8840:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("gl",{months:"xaneiro_febreiro_marzo_abril_maio_xuño_xullo_agosto_setembro_outubro_novembro_decembro".split("_"),monthsShort:"xan._feb._mar._abr._mai._xuñ._xul._ago._set._out._nov._dec.".split("_"),monthsParseExact:!0,weekdays:"domingo_luns_martes_mércores_xoves_venres_sábado".split("_"),weekdaysShort:"dom._lun._mar._mér._xov._ven._sáb.".split("_"),weekdaysMin:"do_lu_ma_mé_xo_ve_sá".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD/MM/YYYY",LL:"D [de] MMMM [de] YYYY",LLL:"D [de] MMMM [de] YYYY H:mm",LLLL:"dddd, D [de] MMMM [de] YYYY H:mm"},calendar:{sameDay:function(){return"[hoxe "+(1!==this.hours()?"ás":"á")+"] LT"},nextDay:function(){return"[mañá "+(1!==this.hours()?"ás":"á")+"] LT"},nextWeek:function(){return"dddd ["+(1!==this.hours()?"ás":"a")+"] LT"},lastDay:function(){return"[onte "+(1!==this.hours()?"á":"a")+"] LT"},lastWeek:function(){return"[o] dddd [pasado "+(1!==this.hours()?"ás":"a")+"] LT"},sameElse:"L"},relativeTime:{future:function(t){return 0===t.indexOf("un")?"n"+t:"en "+t},past:"hai %s",s:"uns segundos",ss:"%d segundos",m:"un minuto",mm:"%d minutos",h:"unha hora",hh:"%d horas",d:"un día",dd:"%d días",M:"un mes",MM:"%d meses",y:"un ano",yy:"%d anos"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:1,doy:4}});return e}))},8925:function(t,e,n){"use strict";var r=n("e330"),i=n("1626"),o=n("c6cd"),a=r(Function.toString);i(o.inspectSource)||(o.inspectSource=function(t){return a(t)}),t.exports=o.inspectSource},"898b":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="ene._feb._mar._abr._may._jun._jul._ago._sep._oct._nov._dic.".split("_"),n="ene_feb_mar_abr_may_jun_jul_ago_sep_oct_nov_dic".split("_"),r=[/^ene/i,/^feb/i,/^mar/i,/^abr/i,/^may/i,/^jun/i,/^jul/i,/^ago/i,/^sep/i,/^oct/i,/^nov/i,/^dic/i],i=/^(enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre|ene\.?|feb\.?|mar\.?|abr\.?|may\.?|jun\.?|jul\.?|ago\.?|sep\.?|oct\.?|nov\.?|dic\.?)/i,o=t.defineLocale("es",{months:"enero_febrero_marzo_abril_mayo_junio_julio_agosto_septiembre_octubre_noviembre_diciembre".split("_"),monthsShort:function(t,r){return t?/-MMM-/.test(r)?n[t.month()]:e[t.month()]:e},monthsRegex:i,monthsShortRegex:i,monthsStrictRegex:/^(enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre)/i,monthsShortStrictRegex:/^(ene\.?|feb\.?|mar\.?|abr\.?|may\.?|jun\.?|jul\.?|ago\.?|sep\.?|oct\.?|nov\.?|dic\.?)/i,monthsParse:r,longMonthsParse:r,shortMonthsParse:r,weekdays:"domingo_lunes_martes_miércoles_jueves_viernes_sábado".split("_"),weekdaysShort:"dom._lun._mar._mié._jue._vie._sáb.".split("_"),weekdaysMin:"do_lu_ma_mi_ju_vi_sá".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD/MM/YYYY",LL:"D [de] MMMM [de] YYYY",LLL:"D [de] MMMM [de] YYYY H:mm",LLLL:"dddd, D [de] MMMM [de] YYYY H:mm"},calendar:{sameDay:function(){return"[hoy a la"+(1!==this.hours()?"s":"")+"] LT"},nextDay:function(){return"[mañana a la"+(1!==this.hours()?"s":"")+"] LT"},nextWeek:function(){return"dddd [a la"+(1!==this.hours()?"s":"")+"] LT"},lastDay:function(){return"[ayer a la"+(1!==this.hours()?"s":"")+"] LT"},lastWeek:function(){return"[el] dddd [pasado a la"+(1!==this.hours()?"s":"")+"] LT"},sameElse:"L"},relativeTime:{future:"en %s",past:"hace %s",s:"unos segundos",ss:"%d segundos",m:"un minuto",mm:"%d minutos",h:"una hora",hh:"%d horas",d:"un día",dd:"%d días",w:"una semana",ww:"%d semanas",M:"un mes",MM:"%d meses",y:"un año",yy:"%d años"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:1,doy:4},invalidDate:"Fecha inválida"});return o}))},"89d9":function(t,e,n){var r=n("656b"),i=n("159a"),o=n("e2e4");function a(t,e,n){var a=-1,s=e.length,l={};while(++a<s){var u=e[a],c=r(t,u);n(c,u)&&i(l,o(u,t),c)}return l}t.exports=a},"8aa5":function(t,e,n){"use strict";var r=n("6547").charAt;t.exports=function(t,e,n){return e+(n?r(t,e).length:1)}},"8adb":function(t,e){function n(t,e){if(("constructor"!==e||"function"!==typeof t[e])&&"__proto__"!=e)return t[e]}t.exports=n},"8ba4":function(t,e,n){"use strict";var r=n("23e7"),i=n("eac50");r({target:"Number",stat:!0},{isInteger:i})},"8c4f":function(t,e,n){"use strict";function r(t,e){for(var n in e)t[n]=e[n];return t}n.d(e,"a",(function(){return we}));var i=/[!'()*]/g,o=function(t){return"%"+t.charCodeAt(0).toString(16)},a=/%2C/g,s=function(t){return encodeURIComponent(t).replace(i,o).replace(a,",")};function l(t){try{return decodeURIComponent(t)}catch(e){0}return t}function u(t,e,n){void 0===e&&(e={});var r,i=n||d;try{r=i(t||"")}catch(s){r={}}for(var o in e){var a=e[o];r[o]=Array.isArray(a)?a.map(c):c(a)}return r}var c=function(t){return null==t||"object"===typeof t?t:String(t)};function d(t){var e={};return t=t.trim().replace(/^(\?|#|&)/,""),t?(t.split("&").forEach((function(t){var n=t.replace(/\+/g," ").split("="),r=l(n.shift()),i=n.length>0?l(n.join("=")):null;void 0===e[r]?e[r]=i:Array.isArray(e[r])?e[r].push(i):e[r]=[e[r],i]})),e):e}function h(t){var e=t?Object.keys(t).map((function(e){var n=t[e];if(void 0===n)return"";if(null===n)return s(e);if(Array.isArray(n)){var r=[];return n.forEach((function(t){void 0!==t&&(null===t?r.push(s(e)):r.push(s(e)+"="+s(t)))})),r.join("&")}return s(e)+"="+s(n)})).filter((function(t){return t.length>0})).join("&"):null;return e?"?"+e:""}var f=/\/?$/;function p(t,e,n,r){var i=r&&r.options.stringifyQuery,o=e.query||{};try{o=m(o)}catch(s){}var a={name:e.name||t&&t.name,meta:t&&t.meta||{},path:e.path||"/",hash:e.hash||"",query:o,params:e.params||{},fullPath:v(e,i),matched:t?y(t):[]};return n&&(a.redirectedFrom=v(n,i)),Object.freeze(a)}function m(t){if(Array.isArray(t))return t.map(m);if(t&&"object"===typeof t){var e={};for(var n in t)e[n]=m(t[n]);return e}return t}var g=p(null,{path:"/"});function y(t){var e=[];while(t)e.unshift(t),t=t.parent;return e}function v(t,e){var n=t.path,r=t.query;void 0===r&&(r={});var i=t.hash;void 0===i&&(i="");var o=e||h;return(n||"/")+o(r)+i}function _(t,e,n){return e===g?t===e:!!e&&(t.path&&e.path?t.path.replace(f,"")===e.path.replace(f,"")&&(n||t.hash===e.hash&&b(t.query,e.query)):!(!t.name||!e.name)&&(t.name===e.name&&(n||t.hash===e.hash&&b(t.query,e.query)&&b(t.params,e.params))))}function b(t,e){if(void 0===t&&(t={}),void 0===e&&(e={}),!t||!e)return t===e;var n=Object.keys(t).sort(),r=Object.keys(e).sort();return n.length===r.length&&n.every((function(n,i){var o=t[n],a=r[i];if(a!==n)return!1;var s=e[n];return null==o||null==s?o===s:"object"===typeof o&&"object"===typeof s?b(o,s):String(o)===String(s)}))}function x(t,e){return 0===t.path.replace(f,"/").indexOf(e.path.replace(f,"/"))&&(!e.hash||t.hash===e.hash)&&w(t.query,e.query)}function w(t,e){for(var n in e)if(!(n in t))return!1;return!0}function M(t){for(var e=0;e<t.matched.length;e++){var n=t.matched[e];for(var r in n.instances){var i=n.instances[r],o=n.enteredCbs[r];if(i&&o){delete n.enteredCbs[r];for(var a=0;a<o.length;a++)i._isBeingDestroyed||o[a](i)}}}}var S={name:"RouterView",functional:!0,props:{name:{type:String,default:"default"}},render:function(t,e){var n=e.props,i=e.children,o=e.parent,a=e.data;a.routerView=!0;var s=o.$createElement,l=n.name,u=o.$route,c=o._routerViewCache||(o._routerViewCache={}),d=0,h=!1;while(o&&o._routerRoot!==o){var f=o.$vnode?o.$vnode.data:{};f.routerView&&d++,f.keepAlive&&o._directInactive&&o._inactive&&(h=!0),o=o.$parent}if(a.routerViewDepth=d,h){var p=c[l],m=p&&p.component;return m?(p.configProps&&k(m,a,p.route,p.configProps),s(m,a,i)):s()}var g=u.matched[d],y=g&&g.components[l];if(!g||!y)return c[l]=null,s();c[l]={component:y},a.registerRouteInstance=function(t,e){var n=g.instances[l];(e&&n!==t||!e&&n===t)&&(g.instances[l]=e)},(a.hook||(a.hook={})).prepatch=function(t,e){g.instances[l]=e.componentInstance},a.hook.init=function(t){t.data.keepAlive&&t.componentInstance&&t.componentInstance!==g.instances[l]&&(g.instances[l]=t.componentInstance),M(u)};var v=g.props&&g.props[l];return v&&(r(c[l],{route:u,configProps:v}),k(y,a,u,v)),s(y,a,i)}};function k(t,e,n,i){var o=e.props=L(n,i);if(o){o=e.props=r({},o);var a=e.attrs=e.attrs||{};for(var s in o)t.props&&s in t.props||(a[s]=o[s],delete o[s])}}function L(t,e){switch(typeof e){case"undefined":return;case"object":return e;case"function":return e(t);case"boolean":return e?t.params:void 0;default:0}}function T(t,e,n){var r=t.charAt(0);if("/"===r)return t;if("?"===r||"#"===r)return e+t;var i=e.split("/");n&&i[i.length-1]||i.pop();for(var o=t.replace(/^\//,"").split("/"),a=0;a<o.length;a++){var s=o[a];".."===s?i.pop():"."!==s&&i.push(s)}return""!==i[0]&&i.unshift(""),i.join("/")}function D(t){var e="",n="",r=t.indexOf("#");r>=0&&(e=t.slice(r),t=t.slice(0,r));var i=t.indexOf("?");return i>=0&&(n=t.slice(i+1),t=t.slice(0,i)),{path:t,query:n,hash:e}}function O(t){return t.replace(/\/(?:\s*\/)+/g,"/")}var I=Array.isArray||function(t){return"[object Array]"==Object.prototype.toString.call(t)},C=Z,A=N,Y=R,j=F,P=X,E=new RegExp(["(\\\\.)","([\\/.])?(?:(?:\\:(\\w+)(?:\\(((?:\\\\.|[^\\\\()])+)\\))?|\\(((?:\\\\.|[^\\\\()])+)\\))([+*?])?|(\\*))"].join("|"),"g");function N(t,e){var n,r=[],i=0,o=0,a="",s=e&&e.delimiter||"/";while(null!=(n=E.exec(t))){var l=n[0],u=n[1],c=n.index;if(a+=t.slice(o,c),o=c+l.length,u)a+=u[1];else{var d=t[o],h=n[2],f=n[3],p=n[4],m=n[5],g=n[6],y=n[7];a&&(r.push(a),a="");var v=null!=h&&null!=d&&d!==h,_="+"===g||"*"===g,b="?"===g||"*"===g,x=n[2]||s,w=p||m;r.push({name:f||i++,prefix:h||"",delimiter:x,optional:b,repeat:_,partial:v,asterisk:!!y,pattern:w?W(w):y?".*":"[^"+B(x)+"]+?"})}}return o<t.length&&(a+=t.substr(o)),a&&r.push(a),r}function R(t,e){return F(N(t,e),e)}function H(t){return encodeURI(t).replace(/[\/?#]/g,(function(t){return"%"+t.charCodeAt(0).toString(16).toUpperCase()}))}function z(t){return encodeURI(t).replace(/[?#]/g,(function(t){return"%"+t.charCodeAt(0).toString(16).toUpperCase()}))}function F(t,e){for(var n=new Array(t.length),r=0;r<t.length;r++)"object"===typeof t[r]&&(n[r]=new RegExp("^(?:"+t[r].pattern+")$",G(e)));return function(e,r){for(var i="",o=e||{},a=r||{},s=a.pretty?H:encodeURIComponent,l=0;l<t.length;l++){var u=t[l];if("string"!==typeof u){var c,d=o[u.name];if(null==d){if(u.optional){u.partial&&(i+=u.prefix);continue}throw new TypeError('Expected "'+u.name+'" to be defined')}if(I(d)){if(!u.repeat)throw new TypeError('Expected "'+u.name+'" to not repeat, but received `'+JSON.stringify(d)+"`");if(0===d.length){if(u.optional)continue;throw new TypeError('Expected "'+u.name+'" to not be empty')}for(var h=0;h<d.length;h++){if(c=s(d[h]),!n[l].test(c))throw new TypeError('Expected all "'+u.name+'" to match "'+u.pattern+'", but received `'+JSON.stringify(c)+"`");i+=(0===h?u.prefix:u.delimiter)+c}}else{if(c=u.asterisk?z(d):s(d),!n[l].test(c))throw new TypeError('Expected "'+u.name+'" to match "'+u.pattern+'", but received "'+c+'"');i+=u.prefix+c}}else i+=u}return i}}function B(t){return t.replace(/([.+*?=^!:${}()[\]|\/\\])/g,"\\$1")}function W(t){return t.replace(/([=!:$\/()])/g,"\\$1")}function V(t,e){return t.keys=e,t}function G(t){return t&&t.sensitive?"":"i"}function $(t,e){var n=t.source.match(/\((?!\?)/g);if(n)for(var r=0;r<n.length;r++)e.push({name:r,prefix:null,delimiter:null,optional:!1,repeat:!1,partial:!1,asterisk:!1,pattern:null});return V(t,e)}function U(t,e,n){for(var r=[],i=0;i<t.length;i++)r.push(Z(t[i],e,n).source);var o=new RegExp("(?:"+r.join("|")+")",G(n));return V(o,e)}function q(t,e,n){return X(N(t,n),e,n)}function X(t,e,n){I(e)||(n=e||n,e=[]),n=n||{};for(var r=n.strict,i=!1!==n.end,o="",a=0;a<t.length;a++){var s=t[a];if("string"===typeof s)o+=B(s);else{var l=B(s.prefix),u="(?:"+s.pattern+")";e.push(s),s.repeat&&(u+="(?:"+l+u+")*"),u=s.optional?s.partial?l+"("+u+")?":"(?:"+l+"("+u+"))?":l+"("+u+")",o+=u}}var c=B(n.delimiter||"/"),d=o.slice(-c.length)===c;return r||(o=(d?o.slice(0,-c.length):o)+"(?:"+c+"(?=$))?"),o+=i?"$":r&&d?"":"(?="+c+"|$)",V(new RegExp("^"+o,G(n)),e)}function Z(t,e,n){return I(e)||(n=e||n,e=[]),n=n||{},t instanceof RegExp?$(t,e):I(t)?U(t,e,n):q(t,e,n)}C.parse=A,C.compile=Y,C.tokensToFunction=j,C.tokensToRegExp=P;var J=Object.create(null);function K(t,e,n){e=e||{};try{var r=J[t]||(J[t]=C.compile(t));return"string"===typeof e.pathMatch&&(e[0]=e.pathMatch),r(e,{pretty:!0})}catch(i){return""}finally{delete e[0]}}function Q(t,e,n,i){var o="string"===typeof t?{path:t}:t;if(o._normalized)return o;if(o.name){o=r({},t);var a=o.params;return a&&"object"===typeof a&&(o.params=r({},a)),o}if(!o.path&&o.params&&e){o=r({},o),o._normalized=!0;var s=r(r({},e.params),o.params);if(e.name)o.name=e.name,o.params=s;else if(e.matched.length){var l=e.matched[e.matched.length-1].path;o.path=K(l,s,"path "+e.path)}else 0;return o}var c=D(o.path||""),d=e&&e.path||"/",h=c.path?T(c.path,d,n||o.append):d,f=u(c.query,o.query,i&&i.options.parseQuery),p=o.hash||c.hash;return p&&"#"!==p.charAt(0)&&(p="#"+p),{_normalized:!0,path:h,query:f,hash:p}}var tt,et=[String,Object],nt=[String,Array],rt=function(){},it={name:"RouterLink",props:{to:{type:et,required:!0},tag:{type:String,default:"a"},custom:Boolean,exact:Boolean,exactPath:Boolean,append:Boolean,replace:Boolean,activeClass:String,exactActiveClass:String,ariaCurrentValue:{type:String,default:"page"},event:{type:nt,default:"click"}},render:function(t){var e=this,n=this.$router,i=this.$route,o=n.resolve(this.to,i,this.append),a=o.location,s=o.route,l=o.href,u={},c=n.options.linkActiveClass,d=n.options.linkExactActiveClass,h=null==c?"router-link-active":c,f=null==d?"router-link-exact-active":d,m=null==this.activeClass?h:this.activeClass,g=null==this.exactActiveClass?f:this.exactActiveClass,y=s.redirectedFrom?p(null,Q(s.redirectedFrom),null,n):s;u[g]=_(i,y,this.exactPath),u[m]=this.exact||this.exactPath?u[g]:x(i,y);var v=u[g]?this.ariaCurrentValue:null,b=function(t){ot(t)&&(e.replace?n.replace(a,rt):n.push(a,rt))},w={click:ot};Array.isArray(this.event)?this.event.forEach((function(t){w[t]=b})):w[this.event]=b;var M={class:u},S=!this.$scopedSlots.$hasNormal&&this.$scopedSlots.default&&this.$scopedSlots.default({href:l,route:s,navigate:b,isActive:u[m],isExactActive:u[g]});if(S){if(1===S.length)return S[0];if(S.length>1||!S.length)return 0===S.length?t():t("span",{},S)}if("a"===this.tag)M.on=w,M.attrs={href:l,"aria-current":v};else{var k=at(this.$slots.default);if(k){k.isStatic=!1;var L=k.data=r({},k.data);for(var T in L.on=L.on||{},L.on){var D=L.on[T];T in w&&(L.on[T]=Array.isArray(D)?D:[D])}for(var O in w)O in L.on?L.on[O].push(w[O]):L.on[O]=b;var I=k.data.attrs=r({},k.data.attrs);I.href=l,I["aria-current"]=v}else M.on=w}return t(this.tag,M,this.$slots.default)}};function ot(t){if(!(t.metaKey||t.altKey||t.ctrlKey||t.shiftKey)&&!t.defaultPrevented&&(void 0===t.button||0===t.button)){if(t.currentTarget&&t.currentTarget.getAttribute){var e=t.currentTarget.getAttribute("target");if(/\b_blank\b/i.test(e))return}return t.preventDefault&&t.preventDefault(),!0}}function at(t){if(t)for(var e,n=0;n<t.length;n++){if(e=t[n],"a"===e.tag)return e;if(e.children&&(e=at(e.children)))return e}}function st(t){if(!st.installed||tt!==t){st.installed=!0,tt=t;var e=function(t){return void 0!==t},n=function(t,n){var r=t.$options._parentVnode;e(r)&&e(r=r.data)&&e(r=r.registerRouteInstance)&&r(t,n)};t.mixin({beforeCreate:function(){e(this.$options.router)?(this._routerRoot=this,this._router=this.$options.router,this._router.init(this),t.util.defineReactive(this,"_route",this._router.history.current)):this._routerRoot=this.$parent&&this.$parent._routerRoot||this,n(this,this)},destroyed:function(){n(this)}}),Object.defineProperty(t.prototype,"$router",{get:function(){return this._routerRoot._router}}),Object.defineProperty(t.prototype,"$route",{get:function(){return this._routerRoot._route}}),t.component("RouterView",S),t.component("RouterLink",it);var r=t.config.optionMergeStrategies;r.beforeRouteEnter=r.beforeRouteLeave=r.beforeRouteUpdate=r.created}}var lt="undefined"!==typeof window;function ut(t,e,n,r,i){var o=e||[],a=n||Object.create(null),s=r||Object.create(null);t.forEach((function(t){ct(o,a,s,t,i)}));for(var l=0,u=o.length;l<u;l++)"*"===o[l]&&(o.push(o.splice(l,1)[0]),u--,l--);return{pathList:o,pathMap:a,nameMap:s}}function ct(t,e,n,r,i,o){var a=r.path,s=r.name;var l=r.pathToRegexpOptions||{},u=ht(a,i,l.strict);"boolean"===typeof r.caseSensitive&&(l.sensitive=r.caseSensitive);var c={path:u,regex:dt(u,l),components:r.components||{default:r.component},alias:r.alias?"string"===typeof r.alias?[r.alias]:r.alias:[],instances:{},enteredCbs:{},name:s,parent:i,matchAs:o,redirect:r.redirect,beforeEnter:r.beforeEnter,meta:r.meta||{},props:null==r.props?{}:r.components?r.props:{default:r.props}};if(r.children&&r.children.forEach((function(r){var i=o?O(o+"/"+r.path):void 0;ct(t,e,n,r,c,i)})),e[c.path]||(t.push(c.path),e[c.path]=c),void 0!==r.alias)for(var d=Array.isArray(r.alias)?r.alias:[r.alias],h=0;h<d.length;++h){var f=d[h];0;var p={path:f,children:r.children};ct(t,e,n,p,i,c.path||"/")}s&&(n[s]||(n[s]=c))}function dt(t,e){var n=C(t,[],e);return n}function ht(t,e,n){return n||(t=t.replace(/\/$/,"")),"/"===t[0]||null==e?t:O(e.path+"/"+t)}function ft(t,e){var n=ut(t),r=n.pathList,i=n.pathMap,o=n.nameMap;function a(t){ut(t,r,i,o)}function s(t,e){var n="object"!==typeof t?o[t]:void 0;ut([e||t],r,i,o,n),n&&n.alias.length&&ut(n.alias.map((function(t){return{path:t,children:[e]}})),r,i,o,n)}function l(){return r.map((function(t){return i[t]}))}function u(t,n,a){var s=Q(t,n,!1,e),l=s.name;if(l){var u=o[l];if(!u)return h(null,s);var c=u.regex.keys.filter((function(t){return!t.optional})).map((function(t){return t.name}));if("object"!==typeof s.params&&(s.params={}),n&&"object"===typeof n.params)for(var d in n.params)!(d in s.params)&&c.indexOf(d)>-1&&(s.params[d]=n.params[d]);return s.path=K(u.path,s.params,'named route "'+l+'"'),h(u,s,a)}if(s.path){s.params={};for(var f=0;f<r.length;f++){var p=r[f],m=i[p];if(pt(m.regex,s.path,s.params))return h(m,s,a)}}return h(null,s)}function c(t,n){var r=t.redirect,i="function"===typeof r?r(p(t,n,null,e)):r;if("string"===typeof i&&(i={path:i}),!i||"object"!==typeof i)return h(null,n);var a=i,s=a.name,l=a.path,c=n.query,d=n.hash,f=n.params;if(c=a.hasOwnProperty("query")?a.query:c,d=a.hasOwnProperty("hash")?a.hash:d,f=a.hasOwnProperty("params")?a.params:f,s){o[s];return u({_normalized:!0,name:s,query:c,hash:d,params:f},void 0,n)}if(l){var m=mt(l,t),g=K(m,f,'redirect route with path "'+m+'"');return u({_normalized:!0,path:g,query:c,hash:d},void 0,n)}return h(null,n)}function d(t,e,n){var r=K(n,e.params,'aliased route with path "'+n+'"'),i=u({_normalized:!0,path:r});if(i){var o=i.matched,a=o[o.length-1];return e.params=i.params,h(a,e)}return h(null,e)}function h(t,n,r){return t&&t.redirect?c(t,r||n):t&&t.matchAs?d(t,n,t.matchAs):p(t,n,r,e)}return{match:u,addRoute:s,getRoutes:l,addRoutes:a}}function pt(t,e,n){var r=e.match(t);if(!r)return!1;if(!n)return!0;for(var i=1,o=r.length;i<o;++i){var a=t.keys[i-1];a&&(n[a.name||"pathMatch"]="string"===typeof r[i]?l(r[i]):r[i])}return!0}function mt(t,e){return T(t,e.parent?e.parent.path:"/",!0)}var gt=lt&&window.performance&&window.performance.now?window.performance:Date;function yt(){return gt.now().toFixed(3)}var vt=yt();function _t(){return vt}function bt(t){return vt=t}var xt=Object.create(null);function wt(){"scrollRestoration"in window.history&&(window.history.scrollRestoration="manual");var t=window.location.protocol+"//"+window.location.host,e=window.location.href.replace(t,""),n=r({},window.history.state);return n.key=_t(),window.history.replaceState(n,"",e),window.addEventListener("popstate",kt),function(){window.removeEventListener("popstate",kt)}}function Mt(t,e,n,r){if(t.app){var i=t.options.scrollBehavior;i&&t.app.$nextTick((function(){var o=Lt(),a=i.call(t,e,n,r?o:null);a&&("function"===typeof a.then?a.then((function(t){Yt(t,o)})).catch((function(t){0})):Yt(a,o))}))}}function St(){var t=_t();t&&(xt[t]={x:window.pageXOffset,y:window.pageYOffset})}function kt(t){St(),t.state&&t.state.key&&bt(t.state.key)}function Lt(){var t=_t();if(t)return xt[t]}function Tt(t,e){var n=document.documentElement,r=n.getBoundingClientRect(),i=t.getBoundingClientRect();return{x:i.left-r.left-e.x,y:i.top-r.top-e.y}}function Dt(t){return Ct(t.x)||Ct(t.y)}function Ot(t){return{x:Ct(t.x)?t.x:window.pageXOffset,y:Ct(t.y)?t.y:window.pageYOffset}}function It(t){return{x:Ct(t.x)?t.x:0,y:Ct(t.y)?t.y:0}}function Ct(t){return"number"===typeof t}var At=/^#\d/;function Yt(t,e){var n="object"===typeof t;if(n&&"string"===typeof t.selector){var r=At.test(t.selector)?document.getElementById(t.selector.slice(1)):document.querySelector(t.selector);if(r){var i=t.offset&&"object"===typeof t.offset?t.offset:{};i=It(i),e=Tt(r,i)}else Dt(t)&&(e=Ot(t))}else n&&Dt(t)&&(e=Ot(t));e&&("scrollBehavior"in document.documentElement.style?window.scrollTo({left:e.x,top:e.y,behavior:t.behavior}):window.scrollTo(e.x,e.y))}var jt=lt&&function(){var t=window.navigator.userAgent;return(-1===t.indexOf("Android 2.")&&-1===t.indexOf("Android 4.0")||-1===t.indexOf("Mobile Safari")||-1!==t.indexOf("Chrome")||-1!==t.indexOf("Windows Phone"))&&(window.history&&"function"===typeof window.history.pushState)}();function Pt(t,e){St();var n=window.history;try{if(e){var i=r({},n.state);i.key=_t(),n.replaceState(i,"",t)}else n.pushState({key:bt(yt())},"",t)}catch(o){window.location[e?"replace":"assign"](t)}}function Et(t){Pt(t,!0)}var Nt={redirected:2,aborted:4,cancelled:8,duplicated:16};function Rt(t,e){return Bt(t,e,Nt.redirected,'Redirected when going from "'+t.fullPath+'" to "'+Vt(e)+'" via a navigation guard.')}function Ht(t,e){var n=Bt(t,e,Nt.duplicated,'Avoided redundant navigation to current location: "'+t.fullPath+'".');return n.name="NavigationDuplicated",n}function zt(t,e){return Bt(t,e,Nt.cancelled,'Navigation cancelled from "'+t.fullPath+'" to "'+e.fullPath+'" with a new navigation.')}function Ft(t,e){return Bt(t,e,Nt.aborted,'Navigation aborted from "'+t.fullPath+'" to "'+e.fullPath+'" via a navigation guard.')}function Bt(t,e,n,r){var i=new Error(r);return i._isRouter=!0,i.from=t,i.to=e,i.type=n,i}var Wt=["params","query","hash"];function Vt(t){if("string"===typeof t)return t;if("path"in t)return t.path;var e={};return Wt.forEach((function(n){n in t&&(e[n]=t[n])})),JSON.stringify(e,null,2)}function Gt(t){return Object.prototype.toString.call(t).indexOf("Error")>-1}function $t(t,e){return Gt(t)&&t._isRouter&&(null==e||t.type===e)}function Ut(t,e,n){var r=function(i){i>=t.length?n():t[i]?e(t[i],(function(){r(i+1)})):r(i+1)};r(0)}function qt(t){return function(e,n,r){var i=!1,o=0,a=null;Xt(t,(function(t,e,n,s){if("function"===typeof t&&void 0===t.cid){i=!0,o++;var l,u=Qt((function(e){Kt(e)&&(e=e.default),t.resolved="function"===typeof e?e:tt.extend(e),n.components[s]=e,o--,o<=0&&r()})),c=Qt((function(t){var e="Failed to resolve async component "+s+": "+t;a||(a=Gt(t)?t:new Error(e),r(a))}));try{l=t(u,c)}catch(h){c(h)}if(l)if("function"===typeof l.then)l.then(u,c);else{var d=l.component;d&&"function"===typeof d.then&&d.then(u,c)}}})),i||r()}}function Xt(t,e){return Zt(t.map((function(t){return Object.keys(t.components).map((function(n){return e(t.components[n],t.instances[n],t,n)}))})))}function Zt(t){return Array.prototype.concat.apply([],t)}var Jt="function"===typeof Symbol&&"symbol"===typeof Symbol.toStringTag;function Kt(t){return t.__esModule||Jt&&"Module"===t[Symbol.toStringTag]}function Qt(t){var e=!1;return function(){var n=[],r=arguments.length;while(r--)n[r]=arguments[r];if(!e)return e=!0,t.apply(this,n)}}var te=function(t,e){this.router=t,this.base=ee(e),this.current=g,this.pending=null,this.ready=!1,this.readyCbs=[],this.readyErrorCbs=[],this.errorCbs=[],this.listeners=[]};function ee(t){if(!t)if(lt){var e=document.querySelector("base");t=e&&e.getAttribute("href")||"/",t=t.replace(/^https?:\/\/[^\/]+/,"")}else t="/";return"/"!==t.charAt(0)&&(t="/"+t),t.replace(/\/$/,"")}function ne(t,e){var n,r=Math.max(t.length,e.length);for(n=0;n<r;n++)if(t[n]!==e[n])break;return{updated:e.slice(0,n),activated:e.slice(n),deactivated:t.slice(n)}}function re(t,e,n,r){var i=Xt(t,(function(t,r,i,o){var a=ie(t,e);if(a)return Array.isArray(a)?a.map((function(t){return n(t,r,i,o)})):n(a,r,i,o)}));return Zt(r?i.reverse():i)}function ie(t,e){return"function"!==typeof t&&(t=tt.extend(t)),t.options[e]}function oe(t){return re(t,"beforeRouteLeave",se,!0)}function ae(t){return re(t,"beforeRouteUpdate",se)}function se(t,e){if(e)return function(){return t.apply(e,arguments)}}function le(t){return re(t,"beforeRouteEnter",(function(t,e,n,r){return ue(t,n,r)}))}function ue(t,e,n){return function(r,i,o){return t(r,i,(function(t){"function"===typeof t&&(e.enteredCbs[n]||(e.enteredCbs[n]=[]),e.enteredCbs[n].push(t)),o(t)}))}}te.prototype.listen=function(t){this.cb=t},te.prototype.onReady=function(t,e){this.ready?t():(this.readyCbs.push(t),e&&this.readyErrorCbs.push(e))},te.prototype.onError=function(t){this.errorCbs.push(t)},te.prototype.transitionTo=function(t,e,n){var r,i=this;try{r=this.router.match(t,this.current)}catch(a){throw this.errorCbs.forEach((function(t){t(a)})),a}var o=this.current;this.confirmTransition(r,(function(){i.updateRoute(r),e&&e(r),i.ensureURL(),i.router.afterHooks.forEach((function(t){t&&t(r,o)})),i.ready||(i.ready=!0,i.readyCbs.forEach((function(t){t(r)})))}),(function(t){n&&n(t),t&&!i.ready&&($t(t,Nt.redirected)&&o===g||(i.ready=!0,i.readyErrorCbs.forEach((function(e){e(t)}))))}))},te.prototype.confirmTransition=function(t,e,n){var r=this,i=this.current;this.pending=t;var o=function(t){!$t(t)&&Gt(t)&&r.errorCbs.length&&r.errorCbs.forEach((function(e){e(t)})),n&&n(t)},a=t.matched.length-1,s=i.matched.length-1;if(_(t,i)&&a===s&&t.matched[a]===i.matched[s])return this.ensureURL(),t.hash&&Mt(this.router,i,t,!1),o(Ht(i,t));var l=ne(this.current.matched,t.matched),u=l.updated,c=l.deactivated,d=l.activated,h=[].concat(oe(c),this.router.beforeHooks,ae(u),d.map((function(t){return t.beforeEnter})),qt(d)),f=function(e,n){if(r.pending!==t)return o(zt(i,t));try{e(t,i,(function(e){!1===e?(r.ensureURL(!0),o(Ft(i,t))):Gt(e)?(r.ensureURL(!0),o(e)):"string"===typeof e||"object"===typeof e&&("string"===typeof e.path||"string"===typeof e.name)?(o(Rt(i,t)),"object"===typeof e&&e.replace?r.replace(e):r.push(e)):n(e)}))}catch(a){o(a)}};Ut(h,f,(function(){var n=le(d),a=n.concat(r.router.resolveHooks);Ut(a,f,(function(){if(r.pending!==t)return o(zt(i,t));r.pending=null,e(t),r.router.app&&r.router.app.$nextTick((function(){M(t)}))}))}))},te.prototype.updateRoute=function(t){this.current=t,this.cb&&this.cb(t)},te.prototype.setupListeners=function(){},te.prototype.teardown=function(){this.listeners.forEach((function(t){t()})),this.listeners=[],this.current=g,this.pending=null};var ce=function(t){function e(e,n){t.call(this,e,n),this._startLocation=de(this.base)}return t&&(e.__proto__=t),e.prototype=Object.create(t&&t.prototype),e.prototype.constructor=e,e.prototype.setupListeners=function(){var t=this;if(!(this.listeners.length>0)){var e=this.router,n=e.options.scrollBehavior,r=jt&&n;r&&this.listeners.push(wt());var i=function(){var n=t.current,i=de(t.base);t.current===g&&i===t._startLocation||t.transitionTo(i,(function(t){r&&Mt(e,t,n,!0)}))};window.addEventListener("popstate",i),this.listeners.push((function(){window.removeEventListener("popstate",i)}))}},e.prototype.go=function(t){window.history.go(t)},e.prototype.push=function(t,e,n){var r=this,i=this,o=i.current;this.transitionTo(t,(function(t){Pt(O(r.base+t.fullPath)),Mt(r.router,t,o,!1),e&&e(t)}),n)},e.prototype.replace=function(t,e,n){var r=this,i=this,o=i.current;this.transitionTo(t,(function(t){Et(O(r.base+t.fullPath)),Mt(r.router,t,o,!1),e&&e(t)}),n)},e.prototype.ensureURL=function(t){if(de(this.base)!==this.current.fullPath){var e=O(this.base+this.current.fullPath);t?Pt(e):Et(e)}},e.prototype.getCurrentLocation=function(){return de(this.base)},e}(te);function de(t){var e=window.location.pathname,n=e.toLowerCase(),r=t.toLowerCase();return!t||n!==r&&0!==n.indexOf(O(r+"/"))||(e=e.slice(t.length)),(e||"/")+window.location.search+window.location.hash}var he=function(t){function e(e,n,r){t.call(this,e,n),r&&fe(this.base)||pe()}return t&&(e.__proto__=t),e.prototype=Object.create(t&&t.prototype),e.prototype.constructor=e,e.prototype.setupListeners=function(){var t=this;if(!(this.listeners.length>0)){var e=this.router,n=e.options.scrollBehavior,r=jt&&n;r&&this.listeners.push(wt());var i=function(){var e=t.current;pe()&&t.transitionTo(me(),(function(n){r&&Mt(t.router,n,e,!0),jt||ve(n.fullPath)}))},o=jt?"popstate":"hashchange";window.addEventListener(o,i),this.listeners.push((function(){window.removeEventListener(o,i)}))}},e.prototype.push=function(t,e,n){var r=this,i=this,o=i.current;this.transitionTo(t,(function(t){ye(t.fullPath),Mt(r.router,t,o,!1),e&&e(t)}),n)},e.prototype.replace=function(t,e,n){var r=this,i=this,o=i.current;this.transitionTo(t,(function(t){ve(t.fullPath),Mt(r.router,t,o,!1),e&&e(t)}),n)},e.prototype.go=function(t){window.history.go(t)},e.prototype.ensureURL=function(t){var e=this.current.fullPath;me()!==e&&(t?ye(e):ve(e))},e.prototype.getCurrentLocation=function(){return me()},e}(te);function fe(t){var e=de(t);if(!/^\/#/.test(e))return window.location.replace(O(t+"/#"+e)),!0}function pe(){var t=me();return"/"===t.charAt(0)||(ve("/"+t),!1)}function me(){var t=window.location.href,e=t.indexOf("#");return e<0?"":(t=t.slice(e+1),t)}function ge(t){var e=window.location.href,n=e.indexOf("#"),r=n>=0?e.slice(0,n):e;return r+"#"+t}function ye(t){jt?Pt(ge(t)):window.location.hash=t}function ve(t){jt?Et(ge(t)):window.location.replace(ge(t))}var _e=function(t){function e(e,n){t.call(this,e,n),this.stack=[],this.index=-1}return t&&(e.__proto__=t),e.prototype=Object.create(t&&t.prototype),e.prototype.constructor=e,e.prototype.push=function(t,e,n){var r=this;this.transitionTo(t,(function(t){r.stack=r.stack.slice(0,r.index+1).concat(t),r.index++,e&&e(t)}),n)},e.prototype.replace=function(t,e,n){var r=this;this.transitionTo(t,(function(t){r.stack=r.stack.slice(0,r.index).concat(t),e&&e(t)}),n)},e.prototype.go=function(t){var e=this,n=this.index+t;if(!(n<0||n>=this.stack.length)){var r=this.stack[n];this.confirmTransition(r,(function(){var t=e.current;e.index=n,e.updateRoute(r),e.router.afterHooks.forEach((function(e){e&&e(r,t)}))}),(function(t){$t(t,Nt.duplicated)&&(e.index=n)}))}},e.prototype.getCurrentLocation=function(){var t=this.stack[this.stack.length-1];return t?t.fullPath:"/"},e.prototype.ensureURL=function(){},e}(te),be=function(t){void 0===t&&(t={}),this.app=null,this.apps=[],this.options=t,this.beforeHooks=[],this.resolveHooks=[],this.afterHooks=[],this.matcher=ft(t.routes||[],this);var e=t.mode||"hash";switch(this.fallback="history"===e&&!jt&&!1!==t.fallback,this.fallback&&(e="hash"),lt||(e="abstract"),this.mode=e,e){case"history":this.history=new ce(this,t.base);break;case"hash":this.history=new he(this,t.base,this.fallback);break;case"abstract":this.history=new _e(this,t.base);break;default:0}},xe={currentRoute:{configurable:!0}};be.prototype.match=function(t,e,n){return this.matcher.match(t,e,n)},xe.currentRoute.get=function(){return this.history&&this.history.current},be.prototype.init=function(t){var e=this;if(this.apps.push(t),t.$once("hook:destroyed",(function(){var n=e.apps.indexOf(t);n>-1&&e.apps.splice(n,1),e.app===t&&(e.app=e.apps[0]||null),e.app||e.history.teardown()})),!this.app){this.app=t;var n=this.history;if(n instanceof ce||n instanceof he){var r=function(t){var r=n.current,i=e.options.scrollBehavior,o=jt&&i;o&&"fullPath"in t&&Mt(e,t,r,!1)},i=function(t){n.setupListeners(),r(t)};n.transitionTo(n.getCurrentLocation(),i,i)}n.listen((function(t){e.apps.forEach((function(e){e._route=t}))}))}},be.prototype.beforeEach=function(t){return Me(this.beforeHooks,t)},be.prototype.beforeResolve=function(t){return Me(this.resolveHooks,t)},be.prototype.afterEach=function(t){return Me(this.afterHooks,t)},be.prototype.onReady=function(t,e){this.history.onReady(t,e)},be.prototype.onError=function(t){this.history.onError(t)},be.prototype.push=function(t,e,n){var r=this;if(!e&&!n&&"undefined"!==typeof Promise)return new Promise((function(e,n){r.history.push(t,e,n)}));this.history.push(t,e,n)},be.prototype.replace=function(t,e,n){var r=this;if(!e&&!n&&"undefined"!==typeof Promise)return new Promise((function(e,n){r.history.replace(t,e,n)}));this.history.replace(t,e,n)},be.prototype.go=function(t){this.history.go(t)},be.prototype.back=function(){this.go(-1)},be.prototype.forward=function(){this.go(1)},be.prototype.getMatchedComponents=function(t){var e=t?t.matched?t:this.resolve(t).route:this.currentRoute;return e?[].concat.apply([],e.matched.map((function(t){return Object.keys(t.components).map((function(e){return t.components[e]}))}))):[]},be.prototype.resolve=function(t,e,n){e=e||this.history.current;var r=Q(t,e,n,this),i=this.match(r,e),o=i.redirectedFrom||i.fullPath,a=this.history.base,s=Se(a,o,this.mode);return{location:r,route:i,href:s,normalizedTo:r,resolved:i}},be.prototype.getRoutes=function(){return this.matcher.getRoutes()},be.prototype.addRoute=function(t,e){this.matcher.addRoute(t,e),this.history.current!==g&&this.history.transitionTo(this.history.getCurrentLocation())},be.prototype.addRoutes=function(t){this.matcher.addRoutes(t),this.history.current!==g&&this.history.transitionTo(this.history.getCurrentLocation())},Object.defineProperties(be.prototype,xe);var we=be;function Me(t,e){return t.push(e),function(){var n=t.indexOf(e);n>-1&&t.splice(n,1)}}function Se(t,e,n){var r="hash"===n?"#"+e:e;return t?O(t+"/"+r):r}be.install=st,be.version="3.6.5",be.isNavigationFailure=$t,be.NavigationFailureType=Nt,be.START_LOCATION=g,lt&&window.Vue&&window.Vue.use(be)},"8d47":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t){return"undefined"!==typeof Function&&t instanceof Function||"[object Function]"===Object.prototype.toString.call(t)}var n=t.defineLocale("el",{monthsNominativeEl:"Ιανουάριος_Φεβρουάριος_Μάρτιος_Απρίλιος_Μάιος_Ιούνιος_Ιούλιος_Αύγουστος_Σεπτέμβριος_Οκτώβριος_Νοέμβριος_Δεκέμβριος".split("_"),monthsGenitiveEl:"Ιανουαρίου_Φεβρουαρίου_Μαρτίου_Απριλίου_Μαΐου_Ιουνίου_Ιουλίου_Αυγούστου_Σεπτεμβρίου_Οκτωβρίου_Νοεμβρίου_Δεκεμβρίου".split("_"),months:function(t,e){return t?"string"===typeof e&&/D/.test(e.substring(0,e.indexOf("MMMM")))?this._monthsGenitiveEl[t.month()]:this._monthsNominativeEl[t.month()]:this._monthsNominativeEl},monthsShort:"Ιαν_Φεβ_Μαρ_Απρ_Μαϊ_Ιουν_Ιουλ_Αυγ_Σεπ_Οκτ_Νοε_Δεκ".split("_"),weekdays:"Κυριακή_Δευτέρα_Τρίτη_Τετάρτη_Πέμπτη_Παρασκευή_Σάββατο".split("_"),weekdaysShort:"Κυρ_Δευ_Τρι_Τετ_Πεμ_Παρ_Σαβ".split("_"),weekdaysMin:"Κυ_Δε_Τρ_Τε_Πε_Πα_Σα".split("_"),meridiem:function(t,e,n){return t>11?n?"μμ":"ΜΜ":n?"πμ":"ΠΜ"},isPM:function(t){return"μ"===(t+"").toLowerCase()[0]},meridiemParse:/[ΠΜ]\.?Μ?\.?/i,longDateFormat:{LT:"h:mm A",LTS:"h:mm:ss A",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY h:mm A",LLLL:"dddd, D MMMM YYYY h:mm A"},calendarEl:{sameDay:"[Σήμερα {}] LT",nextDay:"[Αύριο {}] LT",nextWeek:"dddd [{}] LT",lastDay:"[Χθες {}] LT",lastWeek:function(){switch(this.day()){case 6:return"[το προηγούμενο] dddd [{}] LT";default:return"[την προηγούμενη] dddd [{}] LT"}},sameElse:"L"},calendar:function(t,n){var r=this._calendarEl[t],i=n&&n.hours();return e(r)&&(r=r.apply(n)),r.replace("{}",i%12===1?"στη":"στις")},relativeTime:{future:"σε %s",past:"%s πριν",s:"λίγα δευτερόλεπτα",ss:"%d δευτερόλεπτα",m:"ένα λεπτό",mm:"%d λεπτά",h:"μία ώρα",hh:"%d ώρες",d:"μία μέρα",dd:"%d μέρες",M:"ένας μήνας",MM:"%d μήνες",y:"ένας χρόνος",yy:"%d χρόνια"},dayOfMonthOrdinalParse:/\d{1,2}η/,ordinal:"%dη",week:{dow:1,doy:4}});return n}))},"8d57":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="styczeń_luty_marzec_kwiecień_maj_czerwiec_lipiec_sierpień_wrzesień_październik_listopad_grudzień".split("_"),n="stycznia_lutego_marca_kwietnia_maja_czerwca_lipca_sierpnia_września_października_listopada_grudnia".split("_"),r=[/^sty/i,/^lut/i,/^mar/i,/^kwi/i,/^maj/i,/^cze/i,/^lip/i,/^sie/i,/^wrz/i,/^paź/i,/^lis/i,/^gru/i];function i(t){return t%10<5&&t%10>1&&~~(t/10)%10!==1}function o(t,e,n){var r=t+" ";switch(n){case"ss":return r+(i(t)?"sekundy":"sekund");case"m":return e?"minuta":"minutę";case"mm":return r+(i(t)?"minuty":"minut");case"h":return e?"godzina":"godzinę";case"hh":return r+(i(t)?"godziny":"godzin");case"ww":return r+(i(t)?"tygodnie":"tygodni");case"MM":return r+(i(t)?"miesiące":"miesięcy");case"yy":return r+(i(t)?"lata":"lat")}}var a=t.defineLocale("pl",{months:function(t,r){return t?/D MMMM/.test(r)?n[t.month()]:e[t.month()]:e},monthsShort:"sty_lut_mar_kwi_maj_cze_lip_sie_wrz_paź_lis_gru".split("_"),monthsParse:r,longMonthsParse:r,shortMonthsParse:r,weekdays:"niedziela_poniedziałek_wtorek_środa_czwartek_piątek_sobota".split("_"),weekdaysShort:"ndz_pon_wt_śr_czw_pt_sob".split("_"),weekdaysMin:"Nd_Pn_Wt_Śr_Cz_Pt_So".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Dziś o] LT",nextDay:"[Jutro o] LT",nextWeek:function(){switch(this.day()){case 0:return"[W niedzielę o] LT";case 2:return"[We wtorek o] LT";case 3:return"[W środę o] LT";case 6:return"[W sobotę o] LT";default:return"[W] dddd [o] LT"}},lastDay:"[Wczoraj o] LT",lastWeek:function(){switch(this.day()){case 0:return"[W zeszłą niedzielę o] LT";case 3:return"[W zeszłą środę o] LT";case 6:return"[W zeszłą sobotę o] LT";default:return"[W zeszły] dddd [o] LT"}},sameElse:"L"},relativeTime:{future:"za %s",past:"%s temu",s:"kilka sekund",ss:o,m:o,mm:o,h:o,hh:o,d:"1 dzień",dd:"%d dni",w:"tydzień",ww:o,M:"miesiąc",MM:o,y:"rok",yy:o},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return a}))},"8de2":function(t,e,n){var r=n("8eeb"),i=n("9934");function o(t){return r(t,i(t))}t.exports=o},"8df4":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"۱",2:"۲",3:"۳",4:"۴",5:"۵",6:"۶",7:"۷",8:"۸",9:"۹",0:"۰"},n={"۱":"1","۲":"2","۳":"3","۴":"4","۵":"5","۶":"6","۷":"7","۸":"8","۹":"9","۰":"0"},r=t.defineLocale("fa",{months:"ژانویه_فوریه_مارس_آوریل_مه_ژوئن_ژوئیه_اوت_سپتامبر_اکتبر_نوامبر_دسامبر".split("_"),monthsShort:"ژانویه_فوریه_مارس_آوریل_مه_ژوئن_ژوئیه_اوت_سپتامبر_اکتبر_نوامبر_دسامبر".split("_"),weekdays:"یک‌شنبه_دوشنبه_سه‌شنبه_چهارشنبه_پنج‌شنبه_جمعه_شنبه".split("_"),weekdaysShort:"یک‌شنبه_دوشنبه_سه‌شنبه_چهارشنبه_پنج‌شنبه_جمعه_شنبه".split("_"),weekdaysMin:"ی_د_س_چ_پ_ج_ش".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},meridiemParse:/قبل از ظهر|بعد از ظهر/,isPM:function(t){return/بعد از ظهر/.test(t)},meridiem:function(t,e,n){return t<12?"قبل از ظهر":"بعد از ظهر"},calendar:{sameDay:"[امروز ساعت] LT",nextDay:"[فردا ساعت] LT",nextWeek:"dddd [ساعت] LT",lastDay:"[دیروز ساعت] LT",lastWeek:"dddd [پیش] [ساعت] LT",sameElse:"L"},relativeTime:{future:"در %s",past:"%s پیش",s:"چند ثانیه",ss:"%d ثانیه",m:"یک دقیقه",mm:"%d دقیقه",h:"یک ساعت",hh:"%d ساعت",d:"یک روز",dd:"%d روز",M:"یک ماه",MM:"%d ماه",y:"یک سال",yy:"%d سال"},preparse:function(t){return t.replace(/[۰-۹]/g,(function(t){return n[t]})).replace(/،/g,",")},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]})).replace(/,/g,"،")},dayOfMonthOrdinalParse:/\d{1,2}م/,ordinal:"%dم",week:{dow:6,doy:12}});return r}))},"8e73":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"١",2:"٢",3:"٣",4:"٤",5:"٥",6:"٦",7:"٧",8:"٨",9:"٩",0:"٠"},n={"١":"1","٢":"2","٣":"3","٤":"4","٥":"5","٦":"6","٧":"7","٨":"8","٩":"9","٠":"0"},r=function(t){return 0===t?0:1===t?1:2===t?2:t%100>=3&&t%100<=10?3:t%100>=11?4:5},i={s:["أقل من ثانية","ثانية واحدة",["ثانيتان","ثانيتين"],"%d ثوان","%d ثانية","%d ثانية"],m:["أقل من دقيقة","دقيقة واحدة",["دقيقتان","دقيقتين"],"%d دقائق","%d دقيقة","%d دقيقة"],h:["أقل من ساعة","ساعة واحدة",["ساعتان","ساعتين"],"%d ساعات","%d ساعة","%d ساعة"],d:["أقل من يوم","يوم واحد",["يومان","يومين"],"%d أيام","%d يومًا","%d يوم"],M:["أقل من شهر","شهر واحد",["شهران","شهرين"],"%d أشهر","%d شهرا","%d شهر"],y:["أقل من عام","عام واحد",["عامان","عامين"],"%d أعوام","%d عامًا","%d عام"]},o=function(t){return function(e,n,o,a){var s=r(e),l=i[t][r(e)];return 2===s&&(l=l[n?0:1]),l.replace(/%d/i,e)}},a=["يناير","فبراير","مارس","أبريل","مايو","يونيو","يوليو","أغسطس","سبتمبر","أكتوبر","نوفمبر","ديسمبر"],s=t.defineLocale("ar",{months:a,monthsShort:a,weekdays:"الأحد_الإثنين_الثلاثاء_الأربعاء_الخميس_الجمعة_السبت".split("_"),weekdaysShort:"أحد_إثنين_ثلاثاء_أربعاء_خميس_جمعة_سبت".split("_"),weekdaysMin:"ح_ن_ث_ر_خ_ج_س".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"D/‏M/‏YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},meridiemParse:/ص|م/,isPM:function(t){return"م"===t},meridiem:function(t,e,n){return t<12?"ص":"م"},calendar:{sameDay:"[اليوم عند الساعة] LT",nextDay:"[غدًا عند الساعة] LT",nextWeek:"dddd [عند الساعة] LT",lastDay:"[أمس عند الساعة] LT",lastWeek:"dddd [عند الساعة] LT",sameElse:"L"},relativeTime:{future:"بعد %s",past:"منذ %s",s:o("s"),ss:o("s"),m:o("m"),mm:o("m"),h:o("h"),hh:o("h"),d:o("d"),dd:o("d"),M:o("M"),MM:o("M"),y:o("y"),yy:o("y")},preparse:function(t){return t.replace(/[١٢٣٤٥٦٧٨٩٠]/g,(function(t){return n[t]})).replace(/،/g,",")},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]})).replace(/,/g,"،")},week:{dow:6,doy:12}});return s}))},"8eeb":function(t,e,n){var r=n("32b3"),i=n("872a");function o(t,e,n,o){var a=!n;n||(n={});var s=-1,l=e.length;while(++s<l){var u=e[s],c=o?o(n[u],t[u],u,n,t):void 0;void 0===c&&(c=t[u]),a?i(n,u,c):r(n,u,c)}return n}t.exports=o},"900d":function(t,e,n){var r=n("8eeb"),i=n("2ec1"),o=n("ec69"),a=i((function(t,e,n,i){r(e,o(e),t,i)}));t.exports=a},9043:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"১",2:"২",3:"৩",4:"৪",5:"৫",6:"৬",7:"৭",8:"৮",9:"৯",0:"০"},n={"১":"1","২":"2","৩":"3","৪":"4","৫":"5","৬":"6","৭":"7","৮":"8","৯":"9","০":"0"},r=t.defineLocale("bn",{months:"জানুয়ারি_ফেব্রুয়ারি_মার্চ_এপ্রিল_মে_জুন_জুলাই_আগস্ট_সেপ্টেম্বর_অক্টোবর_নভেম্বর_ডিসেম্বর".split("_"),monthsShort:"জানু_ফেব্রু_মার্চ_এপ্রিল_মে_জুন_জুলাই_আগস্ট_সেপ্ট_অক্টো_নভে_ডিসে".split("_"),weekdays:"রবিবার_সোমবার_মঙ্গলবার_বুধবার_বৃহস্পতিবার_শুক্রবার_শনিবার".split("_"),weekdaysShort:"রবি_সোম_মঙ্গল_বুধ_বৃহস্পতি_শুক্র_শনি".split("_"),weekdaysMin:"রবি_সোম_মঙ্গল_বুধ_বৃহ_শুক্র_শনি".split("_"),longDateFormat:{LT:"A h:mm সময়",LTS:"A h:mm:ss সময়",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm সময়",LLLL:"dddd, D MMMM YYYY, A h:mm সময়"},calendar:{sameDay:"[আজ] LT",nextDay:"[আগামীকাল] LT",nextWeek:"dddd, LT",lastDay:"[গতকাল] LT",lastWeek:"[গত] dddd, LT",sameElse:"L"},relativeTime:{future:"%s পরে",past:"%s আগে",s:"কয়েক সেকেন্ড",ss:"%d সেকেন্ড",m:"এক মিনিট",mm:"%d মিনিট",h:"এক ঘন্টা",hh:"%d ঘন্টা",d:"এক দিন",dd:"%d দিন",M:"এক মাস",MM:"%d মাস",y:"এক বছর",yy:"%d বছর"},preparse:function(t){return t.replace(/[১২৩৪৫৬৭৮৯০]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/রাত|সকাল|দুপুর|বিকাল|রাত/,meridiemHour:function(t,e){return 12===t&&(t=0),"রাত"===e&&t>=4||"দুপুর"===e&&t<5||"বিকাল"===e?t+12:t},meridiem:function(t,e,n){return t<4?"রাত":t<10?"সকাল":t<17?"দুপুর":t<20?"বিকাল":"রাত"},week:{dow:0,doy:6}});return r}))},"90b2":function(t,e,n){"use strict";var r=n("31de");e["a"]=r["a"]},"90d8":function(t,e,n){"use strict";var r=n("c65b"),i=n("1a2d"),o=n("3a9b"),a=n("ad6d"),s=RegExp.prototype;t.exports=function(t){var e=t.flags;return void 0!==e||"flags"in s||i(t,"flags")||!o(s,t)?e:r(a,t)}},"90e3":function(t,e,n){"use strict";var r=n("e330"),i=0,o=Math.random(),a=r(1..toString);t.exports=function(t){return"Symbol("+(void 0===t?"":t)+")_"+a(++i+o,36)}},"90ea":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("zh-tw",{months:"一月_二月_三月_四月_五月_六月_七月_八月_九月_十月_十一月_十二月".split("_"),monthsShort:"1月_2月_3月_4月_5月_6月_7月_8月_9月_10月_11月_12月".split("_"),weekdays:"星期日_星期一_星期二_星期三_星期四_星期五_星期六".split("_"),weekdaysShort:"週日_週一_週二_週三_週四_週五_週六".split("_"),weekdaysMin:"日_一_二_三_四_五_六".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY/MM/DD",LL:"YYYY年M月D日",LLL:"YYYY年M月D日 HH:mm",LLLL:"YYYY年M月D日dddd HH:mm",l:"YYYY/M/D",ll:"YYYY年M月D日",lll:"YYYY年M月D日 HH:mm",llll:"YYYY年M月D日dddd HH:mm"},meridiemParse:/凌晨|早上|上午|中午|下午|晚上/,meridiemHour:function(t,e){return 12===t&&(t=0),"凌晨"===e||"早上"===e||"上午"===e?t:"中午"===e?t>=11?t:t+12:"下午"===e||"晚上"===e?t+12:void 0},meridiem:function(t,e,n){var r=100*t+e;return r<600?"凌晨":r<900?"早上":r<1130?"上午":r<1230?"中午":r<1800?"下午":"晚上"},calendar:{sameDay:"[今天] LT",nextDay:"[明天] LT",nextWeek:"[下]dddd LT",lastDay:"[昨天] LT",lastWeek:"[上]dddd LT",sameElse:"L"},dayOfMonthOrdinalParse:/\d{1,2}(日|月|週)/,ordinal:function(t,e){switch(e){case"d":case"D":case"DDD":return t+"日";case"M":return t+"月";case"w":case"W":return t+"週";default:return t}},relativeTime:{future:"%s後",past:"%s前",s:"幾秒",ss:"%d 秒",m:"1 分鐘",mm:"%d 分鐘",h:"1 小時",hh:"%d 小時",d:"1 天",dd:"%d 天",M:"1 個月",MM:"%d 個月",y:"1 年",yy:"%d 年"}});return e}))},9112:function(t,e,n){"use strict";var r=n("83ab"),i=n("9bf2"),o=n("5c6c");t.exports=r?function(t,e,n){return i.f(t,e,o(1,n))}:function(t,e,n){return t[e]=n,t}},9129:function(t,e,n){"use strict";var r=n("23e7");r({target:"Number",stat:!0},{isNaN:function(t){return t!==t}})},9152:function(t,e){
/*! ieee754. BSD-3-Clause License. Feross Aboukhadijeh <https://feross.org/opensource> */
e.read=function(t,e,n,r,i){var o,a,s=8*i-r-1,l=(1<<s)-1,u=l>>1,c=-7,d=n?i-1:0,h=n?-1:1,f=t[e+d];for(d+=h,o=f&(1<<-c)-1,f>>=-c,c+=s;c>0;o=256*o+t[e+d],d+=h,c-=8);for(a=o&(1<<-c)-1,o>>=-c,c+=r;c>0;a=256*a+t[e+d],d+=h,c-=8);if(0===o)o=1-u;else{if(o===l)return a?NaN:1/0*(f?-1:1);a+=Math.pow(2,r),o-=u}return(f?-1:1)*a*Math.pow(2,o-r)},e.write=function(t,e,n,r,i,o){var a,s,l,u=8*o-i-1,c=(1<<u)-1,d=c>>1,h=23===i?Math.pow(2,-24)-Math.pow(2,-77):0,f=r?0:o-1,p=r?1:-1,m=e<0||0===e&&1/e<0?1:0;for(e=Math.abs(e),isNaN(e)||e===1/0?(s=isNaN(e)?1:0,a=c):(a=Math.floor(Math.log(e)/Math.LN2),e*(l=Math.pow(2,-a))<1&&(a--,l*=2),e+=a+d>=1?h/l:h*Math.pow(2,1-d),e*l>=2&&(a++,l/=2),a+d>=c?(s=0,a=c):a+d>=1?(s=(e*l-1)*Math.pow(2,i),a+=d):(s=e*Math.pow(2,d-1)*Math.pow(2,i),a=0));i>=8;t[n+f]=255&s,f+=p,s/=256,i-=8);for(a=a<<i|s,u+=i;u>0;t[n+f]=255&a,f+=p,a/=256,u-=8);t[n+f-p]|=128*m}},"91e9":function(t,e){function n(t,e){return function(n){return t(e(n))}}t.exports=n},9259:function(t,e,n){"use strict";var r=n("2b0e"),i=function(){var t=this,e=t._self._c;return e("div",{directives:[{name:"transfer-dom",rawName:"v-transfer-dom"}],attrs:{"data-transfer":t.transfer}},[e("transition",{attrs:{name:t.transitionNames[1]}},[t.showMask?e("div",{directives:[{name:"show",rawName:"v-show",value:t.visible,expression:"visible"}],class:t.maskClasses,style:t.wrapStyles,on:{click:t.handleMask}}):t._e()]),e("div",{class:t.wrapClasses,style:t.wrapStyles,on:{click:t.handleWrapClick}},[e("transition",{attrs:{name:t.transitionNames[0]},on:{"after-leave":t.animationFinish}},[e("div",{directives:[{name:"show",rawName:"v-show",value:t.visible,expression:"visible"}],class:t.classes,style:t.mainStyles,on:{mousedown:t.handleMousedown}},[e("div",{ref:"content",class:t.contentClasses,style:t.contentStyles,on:{click:t.handleClickModal}},[t.closable?e("a",{class:[t.prefixCls+"-close"],on:{click:t.close}},[t._t("close",(function(){return[e("Icon",{attrs:{type:"ios-close"}})]}))],2):t._e(),t.showHead?e("div",{class:[t.prefixCls+"-header"],on:{mousedown:t.handleMoveStart}},[t._t("header",(function(){return[e("div",{class:[t.prefixCls+"-header-inner"]},[t._v(t._s(t.title))])]}))],2):t._e(),e("div",{class:[t.prefixCls+"-body"]},[t._t("default")],2),t.footerHide?t._e():e("div",{class:[t.prefixCls+"-footer"]},[t._t("footer",(function(){return[e("i-button",{attrs:{type:"text"},nativeOn:{click:function(e){return t.cancel.apply(null,arguments)}}},[t._v(t._s(t.localeCancelText))]),e("i-button",{attrs:{type:"primary",loading:t.buttonLoading},nativeOn:{click:function(e){return t.ok.apply(null,arguments)}}},[t._v(t._s(t.localeOkText))])]}))],2)])])])],1)],1)},o=[],a=n("ade3"),s=(n("4de4"),n("4e82"),n("a9e3"),n("d3b7"),n("2c3e"),n("5250")),l=n("7a12"),u=n("c780");const c=r["a"].prototype.$isServer;var d=function(t){c||"undefined"!==typeof window.iview&&("langs"in iview||(iview.langs={}),iview.langs[t.i.locale]=t)};const h={i:{locale:"zh-CN",select:{placeholder:"请选择",noMatch:"无匹配数据",loading:"加载中"},table:{noDataText:"暂无数据",noFilteredDataText:"暂无筛选结果",confirmFilter:"筛选",resetFilter:"重置",clearFilter:"全部",sumText:"合计"},datepicker:{selectDate:"选择日期",selectTime:"选择时间",startTime:"开始时间",endTime:"结束时间",clear:"清空",ok:"确定",datePanelLabel:"[yyyy年] [m月]",month:"月",month1:"1 月",month2:"2 月",month3:"3 月",month4:"4 月",month5:"5 月",month6:"6 月",month7:"7 月",month8:"8 月",month9:"9 月",month10:"10 月",month11:"11 月",month12:"12 月",year:"年",weekStartDay:"0",weeks:{sun:"日",mon:"一",tue:"二",wed:"三",thu:"四",fri:"五",sat:"六"},months:{m1:"1月",m2:"2月",m3:"3月",m4:"4月",m5:"5月",m6:"6月",m7:"7月",m8:"8月",m9:"9月",m10:"10月",m11:"11月",m12:"12月"}},transfer:{titles:{source:"源列表",target:"目的列表"},filterPlaceholder:"请输入搜索内容",notFoundText:"列表为空"},modal:{okText:"确定",cancelText:"取消"},poptip:{okText:"确定",cancelText:"取消"},page:{prev:"上一页",next:"下一页",total:"共",item:"条",items:"条",prev5:"向前 5 页",next5:"向后 5 页",page:"条/页",goto:"跳至",p:"页"},rate:{star:"星",stars:"星"},time:{before:"前",after:"后",just:"刚刚",seconds:"秒",minutes:"分钟",hours:"小时",days:"天"},tree:{emptyText:"暂无数据"}}};d(h);var f=h,p=function(t){return m(t)&&!g(t)};function m(t){return!!t&&"object"===typeof t}function g(t){var e=Object.prototype.toString.call(t);return"[object RegExp]"===e||"[object Date]"===e||_(t)}var y="function"===typeof Symbol&&Symbol.for,v=y?Symbol.for("react.element"):60103;function _(t){return t.$$typeof===v}function b(t){return Array.isArray(t)?[]:{}}function x(t,e){return!1!==e.clone&&e.isMergeableObject(t)?S(b(t),t,e):t}function w(t,e,n){return t.concat(e).map((function(t){return x(t,n)}))}function M(t,e,n){var r={};return n.isMergeableObject(t)&&Object.keys(t).forEach((function(e){r[e]=x(t[e],n)})),Object.keys(e).forEach((function(i){n.isMergeableObject(e[i])&&t[i]?r[i]=S(t[i],e[i],n):r[i]=x(e[i],n)})),r}function S(t,e,n){n=n||{},n.arrayMerge=n.arrayMerge||w,n.isMergeableObject=n.isMergeableObject||p;var r=Array.isArray(e),i=Array.isArray(t),o=r===i;return o?r?n.arrayMerge(t,e,n):M(t,e,n):x(e,n)}S.all=function(t,e){if(!Array.isArray(t))throw new Error("first argument should be an array");return t.reduce((function(t,n){return S(t,n,e)}),{})};var k=S,L=k;const T=/(%|)\{([0-9a-zA-Z_]+)\}/g;var D=function(){function t(t,e){return Object.prototype.hasOwnProperty.call(t,e)}function e(e,...n){return 1===n.length&&"object"===typeof n[0]&&(n=n[0]),n&&n.hasOwnProperty||(n={}),e.replace(T,(r,i,o,a)=>{let s;return"{"===e[a-1]&&"}"===e[a+r.length]?o:(s=t(n,o)?n[o]:null,null===s||void 0===s?"":s)})}return e};const O=D(r["a"]);let I=f,C=!1,A=function(){const t=Object.getPrototypeOf(this||r["a"]).$t;if("function"===typeof t&&r["a"].locale)return C||(C=!0,r["a"].locale(r["a"].config.lang,L(I,r["a"].locale(r["a"].config.lang)||{},{clone:!0}))),t.apply(this,arguments)};const Y=function(t,e){let n=A.apply(this,arguments);if(null!==n&&void 0!==n)return n;const r=t.split(".");let i=I;for(let o=0,a=r.length;o<a;o++){const t=r[o];if(n=i[t],o===a-1)return O(n,e);if(!n)return"";i=n}return""};var j={methods:{t(...t){return Y.apply(this,t)}}},P=n("2507"),E=n("b2e6"),N=n("de48"),R=n("3a27"),H=n("c177"),z="ivu-modal",F={x:null,y:null,dragX:null,dragY:null,dragging:!1,rect:null},B={name:"Modal",mixins:[j,P["a"],E["a"]],components:{Icon:s["a"],iButton:l["a"]},directives:{TransferDom:u["a"]},props:{value:{type:Boolean,default:!1},closable:{type:Boolean,default:!0},maskClosable:{type:Boolean,default:function(){return!this.$IVIEW||""===this.$IVIEW.modal.maskClosable||this.$IVIEW.modal.maskClosable}},title:{type:String},width:{type:[Number,String],default:520},okText:{type:String},cancelText:{type:String},loading:{type:Boolean,default:!1},styles:{type:Object,default:function(){return{}}},className:{type:String},footerHide:{type:Boolean,default:!1},scrollable:{type:Boolean,default:!1},transitionNames:{type:Array,default:function(){return["ease","fade"]}},transfer:{type:Boolean,default:function(){return!this.$IVIEW||""===this.$IVIEW.transfer||this.$IVIEW.transfer}},fullscreen:{type:Boolean,default:!1},mask:{type:Boolean,default:!0},draggable:{type:Boolean,default:!1},sticky:{type:Boolean,default:!1},stickyDistance:{type:Number,default:10},resetDragPosition:{type:Boolean,default:!1},zIndex:{type:Number,default:1e3},beforeClose:Function},data:function(){return{prefixCls:z,wrapShow:!1,showHead:!0,buttonLoading:!1,visible:this.value,dragData:Object(R["b"])(F),modalIndex:this.handleGetModalIndex(),isMouseTriggerIn:!1}},computed:{wrapClasses:function(){return["".concat(z,"-wrap"),Object(a["a"])(Object(a["a"])(Object(a["a"])({},"".concat(z,"-hidden"),!this.wrapShow),"".concat(this.className),!!this.className),"".concat(z,"-no-mask"),!this.showMask)]},wrapStyles:function(){return{zIndex:this.modalIndex+this.zIndex}},maskClasses:function(){return"".concat(z,"-mask")},classes:function(){return["".concat(z),Object(a["a"])(Object(a["a"])(Object(a["a"])({},"".concat(z,"-fullscreen"),this.fullscreen),"".concat(z,"-fullscreen-no-header"),this.fullscreen&&!this.showHead),"".concat(z,"-fullscreen-no-footer"),this.fullscreen&&this.footerHide)]},contentClasses:function(){return["".concat(z,"-content"),Object(a["a"])(Object(a["a"])(Object(a["a"])({},"".concat(z,"-content-no-mask"),!this.showMask),"".concat(z,"-content-drag"),this.draggable&&!this.fullscreen),"".concat(z,"-content-dragging"),this.draggable&&this.dragData.dragging)]},mainStyles:function(){var t={},e=parseInt(this.width),n=null!==this.dragData.x?{top:0}:{width:"".concat(e,e<=100?"%":"px")},r=this.styles?this.styles:{};return Object.assign(t,n,r),t},contentStyles:function(){var t={};if(this.draggable&&!this.fullscreen){var e=this.styles.top?parseFloat(this.styles.top):0,n=this.styles.left?parseFloat(this.styles.left):0;null!==this.dragData.x&&(t.left="".concat(this.dragData.x-n,"px")),null!==this.dragData.y&&(t.top="".concat(this.dragData.y,"px")),null!==this.dragData.y&&(t.top="".concat(this.dragData.y-e,"px"));var r=parseInt(this.width),i={width:"".concat(r,r<=100?"%":"px")};Object.assign(t,i)}return t},localeOkText:function(){return void 0===this.okText?this.t("i.modal.okText"):this.okText},localeCancelText:function(){return void 0===this.cancelText?this.t("i.modal.cancelText"):this.cancelText},showMask:function(){return this.mask}},methods:{close:function(){var t=this;if(!this.beforeClose)return this.handleClose();var e=this.beforeClose();e&&e.then?e.then((function(){t.handleClose()})):this.handleClose()},handleClose:function(){this.visible=!1,this.$emit("input",!1),this.$emit("on-cancel")},handleMask:function(){this.maskClosable&&this.showMask&&this.close()},handleWrapClick:function(t){if(this.isMouseTriggerIn)this.isMouseTriggerIn=!1;else{var e=t.target.getAttribute("class");e&&e.indexOf("".concat(z,"-wrap"))>-1&&this.handleMask()}},handleMousedown:function(){this.isMouseTriggerIn=!0},cancel:function(){this.close()},ok:function(){this.loading?this.buttonLoading=!0:(this.visible=!1,this.$emit("input",!1)),this.$emit("on-ok")},EscClose:function(t){if(this.visible&&this.closable&&27===t.keyCode){var e=Object(R["g"])(this.$root,"Modal").filter((function(t){return t.$data.visible&&t.$props.closable})),n=e.sort((function(t,e){return t.$data.modalIndex<e.$data.modalIndex?1:-1}))[0];setTimeout((function(){n.close()}),0)}},animationFinish:function(){this.$emit("on-hidden")},handleMoveStart:function(t){if(!this.draggable||this.fullscreen)return!1;var e=this.$refs.content,n=e.getBoundingClientRect();this.dragData.rect=n,this.dragData.x=n.x||n.left,this.dragData.y=n.y||n.top;var r={x:t.clientX,y:t.clientY};this.dragData.dragX=r.x,this.dragData.dragY=r.y,this.dragData.dragging=!0,Object(N["b"])(window,"mousemove",this.handleMoveMove),Object(N["b"])(window,"mouseup",this.handleMoveEnd)},handleMoveMove:function(t){if(!this.dragData.dragging||this.fullscreen)return!1;var e={x:t.clientX,y:t.clientY},n={x:e.x-this.dragData.dragX,y:e.y-this.dragData.dragY};if(this.sticky){var r=document.documentElement.clientWidth,i=document.documentElement.clientHeight;this.dragData.x+n.x<=this.stickyDistance&&n.x<0?this.dragData.x=0:this.dragData.x+this.dragData.rect.width-r>-this.stickyDistance&&n.x>0?this.dragData.x=r-this.dragData.rect.width:this.dragData.x+=n.x,this.dragData.y+n.y<=this.stickyDistance&&n.y<0?this.dragData.y=0:this.dragData.y+this.dragData.rect.height-i>-this.stickyDistance&&n.y>0?this.dragData.y=i-this.dragData.rect.height:this.dragData.y+=n.y}else this.dragData.x+=n.x,this.dragData.y+=n.y;this.dragData.dragX=e.x,this.dragData.dragY=e.y},handleMoveEnd:function(){this.dragData.dragging=!1,Object(N["a"])(window,"mousemove",this.handleMoveMove),Object(N["a"])(window,"mouseup",this.handleMoveEnd)},handleGetModalIndex:function(){return Object(H["c"])(),H["d"]},handleClickModal:function(){if(this.draggable){if(H["b"]!==this.lastVisibleIndex)return void(this.lastVisibleIndex=H["b"]);this.modalIndex=this.handleGetModalIndex()}}},mounted:function(){this.visible&&(this.wrapShow=!0);var t=!0;void 0!==this.$slots.header||this.title||(t=!1),this.showHead=t,document.addEventListener("keydown",this.EscClose)},beforeDestroy:function(){document.removeEventListener("keydown",this.EscClose),this.removeScrollEffect()},watch:{value:function(t){this.visible=t},visible:function(t){var e=this;!1===t?(this.buttonLoading=!1,this.timer=setTimeout((function(){e.wrapShow=!1,e.removeScrollEffect()}),300)):(this.lastVisible!==t&&(this.modalIndex=this.handleGetModalIndex(),Object(H["a"])()),this.timer&&clearTimeout(this.timer),this.wrapShow=!0,this.scrollable||this.addScrollEffect()),this.broadcast("Table","on-visible-change",t),this.broadcast("Slider","on-visible-change",t),this.$emit("on-visible-change",t),this.lastVisible=t,this.lastVisibleIndex=H["b"],t&&this.resetDragPosition&&(this.dragData=Object(R["b"])(F))},loading:function(t){t||(this.buttonLoading=!1)},scrollable:function(t){t?this.removeScrollEffect():this.addScrollEffect()},title:function(t){void 0===this.$slots.header&&(this.showHead=!!t)}}},W=B,V=n("2877"),G=Object(V["a"])(W,i,o,!1,null,null,null),$=G.exports;const U="ivu-modal-confirm";$.newInstance=t=>{const e=t||{},n=new r["a"]({mixins:[j],data:Object.assign({},e,{visible:!1,width:416,title:"",body:"",iconType:"",iconName:"",okText:void 0,cancelText:void 0,showCancel:!1,loading:!1,buttonLoading:!1,scrollable:!1,closable:!1,closing:!1}),render(t){let n,r,i=[];return this.showCancel&&i.push(t(l["a"],{props:{type:"text"},on:{click:this.cancel}},this.localeCancelText)),i.push(t(l["a"],{props:{type:"primary",loading:this.buttonLoading},on:{click:this.ok}},this.localeOkText)),n=this.render?t("div",{attrs:{class:`${U}-body ${U}-body-render`}},[this.render(t)]):t("div",{attrs:{class:U+"-body"}},[t("div",{domProps:{innerHTML:this.body}})]),this.title&&(r=t("div",{attrs:{class:U+"-head"}},[t("div",{class:this.iconTypeCls},[t("i",{class:this.iconNameCls})]),t("div",{attrs:{class:U+"-head-title"},domProps:{innerHTML:this.title}})])),t($,{props:Object.assign({},e,{width:this.width,scrollable:this.scrollable,closable:this.closable}),domProps:{value:this.visible},on:{input:t=>{this.visible=t},"on-cancel":this.cancel}},[t("div",{attrs:{class:U}},[r,n,t("div",{attrs:{class:U+"-footer"}},i)])])},computed:{iconTypeCls(){return[U+"-head-icon",`${U}-head-icon-${this.iconType}`]},iconNameCls(){return["ivu-icon","ivu-icon-"+this.iconName]},localeOkText(){return this.okText?this.okText:this.t("i.modal.okText")},localeCancelText(){return this.cancelText?this.cancelText:this.t("i.modal.cancelText")}},methods:{cancel(){this.closing||(this.$children[0].visible=!1,this.buttonLoading=!1,this.onCancel(),this.remove())},ok(){this.closing||(this.loading?this.buttonLoading=!0:(this.$children[0].visible=!1,this.remove()),this.onOk())},remove(){this.closing=!0,setTimeout(()=>{this.closing=!1,this.destroy()},300)},destroy(){this.$destroy(),this.$el&&document.body.removeChild(this.$el),this.onRemove()},onOk(){},onCancel(){},onRemove(){}}}),i=n.$mount();document.body.appendChild(i.$el);const o=n.$children[0];return{show(t){switch(o.$parent.showCancel=t.showCancel,o.$parent.iconType=t.icon,t.icon){case"info":o.$parent.iconName="ios-information-circle";break;case"success":o.$parent.iconName="ios-checkmark-circle";break;case"warning":o.$parent.iconName="ios-alert";break;case"error":o.$parent.iconName="ios-close-circle";break;case"confirm":o.$parent.iconName="ios-help-circle";break}"width"in t&&(o.$parent.width=t.width),"closable"in t&&(o.$parent.closable=t.closable),"title"in t&&(o.$parent.title=t.title),"content"in t&&(o.$parent.body=t.content),"okText"in t&&(o.$parent.okText=t.okText),"cancelText"in t&&(o.$parent.cancelText=t.cancelText),"onCancel"in t&&(o.$parent.onCancel=t.onCancel),"onOk"in t&&(o.$parent.onOk=t.onOk),"loading"in t&&(o.$parent.loading=t.loading),"scrollable"in t&&(o.$parent.scrollable=t.scrollable),o.$parent.onRemove=t.onRemove,o.visible=!0},remove(){o.visible=!1,o.$parent.buttonLoading=!1,o.$parent.remove()},component:o}};var q=$;let X;function Z(t,e=!0){return X=X||q.newInstance({closable:!1,maskClosable:!1,footerHide:!0,render:t,lockScroll:e}),X}function J(t){const e="render"in t?t.render:void 0,n=!("lockScroll"in t)||t.lockScroll;let r=Z(e,n);t.onRemove=function(){X=null},r.show(t)}q.info=function(t={}){return t.icon="info",t.showCancel=!1,J(t)},q.success=function(t={}){return t.icon="success",t.showCancel=!1,J(t)},q.warning=function(t={}){return t.icon="warning",t.showCancel=!1,J(t)},q.error=function(t={}){return t.icon="error",t.showCancel=!1,J(t)},q.confirm=function(t={}){return t.icon="confirm",t.showCancel=!0,J(t)},q.remove=function(){if(!X)return!1;const t=Z();t.remove()};e["a"]=q},9263:function(t,e,n){"use strict";var r=n("c65b"),i=n("e330"),o=n("577e"),a=n("ad6d"),s=n("9f7f"),l=n("5692"),u=n("7c73"),c=n("69f3").get,d=n("fce3"),h=n("107c"),f=l("native-string-replace",String.prototype.replace),p=RegExp.prototype.exec,m=p,g=i("".charAt),y=i("".indexOf),v=i("".replace),_=i("".slice),b=function(){var t=/a/,e=/b*/g;return r(p,t,"a"),r(p,e,"a"),0!==t.lastIndex||0!==e.lastIndex}(),x=s.BROKEN_CARET,w=void 0!==/()??/.exec("")[1],M=b||w||x||d||h;M&&(m=function(t){var e,n,i,s,l,d,h,M=this,S=c(M),k=o(t),L=S.raw;if(L)return L.lastIndex=M.lastIndex,e=r(m,L,k),M.lastIndex=L.lastIndex,e;var T=S.groups,D=x&&M.sticky,O=r(a,M),I=M.source,C=0,A=k;if(D&&(O=v(O,"y",""),-1===y(O,"g")&&(O+="g"),A=_(k,M.lastIndex),M.lastIndex>0&&(!M.multiline||M.multiline&&"\n"!==g(k,M.lastIndex-1))&&(I="(?: "+I+")",A=" "+A,C++),n=new RegExp("^(?:"+I+")",O)),w&&(n=new RegExp("^"+I+"$(?!\\s)",O)),b&&(i=M.lastIndex),s=r(p,D?n:M,A),D?s?(s.input=_(s.input,C),s[0]=_(s[0],C),s.index=M.lastIndex,M.lastIndex+=s[0].length):M.lastIndex=0:b&&s&&(M.lastIndex=M.global?s.index+s[0].length:i),w&&s&&s.length>1&&r(f,s[0],n,(function(){for(l=1;l<arguments.length-2;l++)void 0===arguments[l]&&(s[l]=void 0)})),s&&T)for(s.groups=d=u(null),l=0;l<T.length;l++)h=T[l],d[h[0]]=s[h[1]];return s}),t.exports=m},"93ed":function(t,e,n){var r=n("4245");function i(t){var e=r(this,t)["delete"](t);return this.size-=e?1:0,e}t.exports=i},"94ca":function(t,e,n){"use strict";var r=n("d039"),i=n("1626"),o=/#|\.prototype\./,a=function(t,e){var n=l[s(t)];return n===c||n!==u&&(i(e)?r(e):!!e)},s=a.normalize=function(t){return String(t).replace(o,".").toLowerCase()},l=a.data={},u=a.NATIVE="N",c=a.POLYFILL="P";t.exports=a},9520:function(t,e,n){var r=n("3729"),i=n("1a8c"),o="[object AsyncFunction]",a="[object Function]",s="[object GeneratorFunction]",l="[object Proxy]";function u(t){if(!i(t))return!1;var e=r(t);return e==a||e==s||e==o||e==l}t.exports=u},"957c":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e){var n=t.split("_");return e%10===1&&e%100!==11?n[0]:e%10>=2&&e%10<=4&&(e%100<10||e%100>=20)?n[1]:n[2]}function n(t,n,r){var i={ss:n?"секунда_секунды_секунд":"секунду_секунды_секунд",mm:n?"минута_минуты_минут":"минуту_минуты_минут",hh:"час_часа_часов",dd:"день_дня_дней",ww:"неделя_недели_недель",MM:"месяц_месяца_месяцев",yy:"год_года_лет"};return"m"===r?n?"минута":"минуту":t+" "+e(i[r],+t)}var r=[/^янв/i,/^фев/i,/^мар/i,/^апр/i,/^ма[йя]/i,/^июн/i,/^июл/i,/^авг/i,/^сен/i,/^окт/i,/^ноя/i,/^дек/i],i=t.defineLocale("ru",{months:{format:"января_февраля_марта_апреля_мая_июня_июля_августа_сентября_октября_ноября_декабря".split("_"),standalone:"январь_февраль_март_апрель_май_июнь_июль_август_сентябрь_октябрь_ноябрь_декабрь".split("_")},monthsShort:{format:"янв._февр._мар._апр._мая_июня_июля_авг._сент._окт._нояб._дек.".split("_"),standalone:"янв._февр._март_апр._май_июнь_июль_авг._сент._окт._нояб._дек.".split("_")},weekdays:{standalone:"воскресенье_понедельник_вторник_среда_четверг_пятница_суббота".split("_"),format:"воскресенье_понедельник_вторник_среду_четверг_пятницу_субботу".split("_"),isFormat:/\[ ?[Вв] ?(?:прошлую|следующую|эту)? ?] ?dddd/},weekdaysShort:"вс_пн_вт_ср_чт_пт_сб".split("_"),weekdaysMin:"вс_пн_вт_ср_чт_пт_сб".split("_"),monthsParse:r,longMonthsParse:r,shortMonthsParse:r,monthsRegex:/^(январ[ья]|янв\.?|феврал[ья]|февр?\.?|марта?|мар\.?|апрел[ья]|апр\.?|ма[йя]|июн[ья]|июн\.?|июл[ья]|июл\.?|августа?|авг\.?|сентябр[ья]|сент?\.?|октябр[ья]|окт\.?|ноябр[ья]|нояб?\.?|декабр[ья]|дек\.?)/i,monthsShortRegex:/^(январ[ья]|янв\.?|феврал[ья]|февр?\.?|марта?|мар\.?|апрел[ья]|апр\.?|ма[йя]|июн[ья]|июн\.?|июл[ья]|июл\.?|августа?|авг\.?|сентябр[ья]|сент?\.?|октябр[ья]|окт\.?|ноябр[ья]|нояб?\.?|декабр[ья]|дек\.?)/i,monthsStrictRegex:/^(январ[яь]|феврал[яь]|марта?|апрел[яь]|ма[яй]|июн[яь]|июл[яь]|августа?|сентябр[яь]|октябр[яь]|ноябр[яь]|декабр[яь])/i,monthsShortStrictRegex:/^(янв\.|февр?\.|мар[т.]|апр\.|ма[яй]|июн[ья.]|июл[ья.]|авг\.|сент?\.|окт\.|нояб?\.|дек\.)/i,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY г.",LLL:"D MMMM YYYY г., H:mm",LLLL:"dddd, D MMMM YYYY г., H:mm"},calendar:{sameDay:"[Сегодня, в] LT",nextDay:"[Завтра, в] LT",lastDay:"[Вчера, в] LT",nextWeek:function(t){if(t.week()===this.week())return 2===this.day()?"[Во] dddd, [в] LT":"[В] dddd, [в] LT";switch(this.day()){case 0:return"[В следующее] dddd, [в] LT";case 1:case 2:case 4:return"[В следующий] dddd, [в] LT";case 3:case 5:case 6:return"[В следующую] dddd, [в] LT"}},lastWeek:function(t){if(t.week()===this.week())return 2===this.day()?"[Во] dddd, [в] LT":"[В] dddd, [в] LT";switch(this.day()){case 0:return"[В прошлое] dddd, [в] LT";case 1:case 2:case 4:return"[В прошлый] dddd, [в] LT";case 3:case 5:case 6:return"[В прошлую] dddd, [в] LT"}},sameElse:"L"},relativeTime:{future:"через %s",past:"%s назад",s:"несколько секунд",ss:n,m:n,mm:n,h:"час",hh:n,d:"день",dd:n,w:"неделя",ww:n,M:"месяц",MM:n,y:"год",yy:n},meridiemParse:/ночи|утра|дня|вечера/i,isPM:function(t){return/^(дня|вечера)$/.test(t)},meridiem:function(t,e,n){return t<4?"ночи":t<12?"утра":t<17?"дня":"вечера"},dayOfMonthOrdinalParse:/\d{1,2}-(й|го|я)/,ordinal:function(t,e){switch(e){case"M":case"d":case"DDD":return t+"-й";case"D":return t+"-го";case"w":case"W":return t+"-я";default:return t}},week:{dow:1,doy:4}});return i}))},"958b":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,r){switch(n){case"s":return e?"хэдхэн секунд":"хэдхэн секундын";case"ss":return t+(e?" секунд":" секундын");case"m":case"mm":return t+(e?" минут":" минутын");case"h":case"hh":return t+(e?" цаг":" цагийн");case"d":case"dd":return t+(e?" өдөр":" өдрийн");case"M":case"MM":return t+(e?" сар":" сарын");case"y":case"yy":return t+(e?" жил":" жилийн");default:return t}}var n=t.defineLocale("mn",{months:"Нэгдүгээр сар_Хоёрдугаар сар_Гуравдугаар сар_Дөрөвдүгээр сар_Тавдугаар сар_Зургадугаар сар_Долдугаар сар_Наймдугаар сар_Есдүгээр сар_Аравдугаар сар_Арван нэгдүгээр сар_Арван хоёрдугаар сар".split("_"),monthsShort:"1 сар_2 сар_3 сар_4 сар_5 сар_6 сар_7 сар_8 сар_9 сар_10 сар_11 сар_12 сар".split("_"),monthsParseExact:!0,weekdays:"Ням_Даваа_Мягмар_Лхагва_Пүрэв_Баасан_Бямба".split("_"),weekdaysShort:"Ням_Дав_Мяг_Лха_Пүр_Баа_Бям".split("_"),weekdaysMin:"Ня_Да_Мя_Лх_Пү_Ба_Бя".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY-MM-DD",LL:"YYYY оны MMMMын D",LLL:"YYYY оны MMMMын D HH:mm",LLLL:"dddd, YYYY оны MMMMын D HH:mm"},meridiemParse:/ҮӨ|ҮХ/i,isPM:function(t){return"ҮХ"===t},meridiem:function(t,e,n){return t<12?"ҮӨ":"ҮХ"},calendar:{sameDay:"[Өнөөдөр] LT",nextDay:"[Маргааш] LT",nextWeek:"[Ирэх] dddd LT",lastDay:"[Өчигдөр] LT",lastWeek:"[Өнгөрсөн] dddd LT",sameElse:"L"},relativeTime:{future:"%s дараа",past:"%s өмнө",s:e,ss:e,m:e,mm:e,h:e,hh:e,d:e,dd:e,M:e,MM:e,y:e,yy:e},dayOfMonthOrdinalParse:/\d{1,2} өдөр/,ordinal:function(t,e){switch(e){case"d":case"D":case"DDD":return t+" өдөр";default:return t}}});return n}))},"95ae":function(t,e,n){var r=n("100e"),i=n("9638"),o=n("9aff"),a=n("9934"),s=Object.prototype,l=s.hasOwnProperty,u=r((function(t,e){t=Object(t);var n=-1,r=e.length,u=r>2?e[2]:void 0;u&&o(e[0],e[1],u)&&(r=1);while(++n<r){var c=e[n],d=a(c),h=-1,f=d.length;while(++h<f){var p=d[h],m=t[p];(void 0===m||i(m,s[p])&&!l.call(t,p))&&(t[p]=c[p])}}return t}));t.exports=u},9609:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={0:"-чү",1:"-чи",2:"-чи",3:"-чү",4:"-чү",5:"-чи",6:"-чы",7:"-чи",8:"-чи",9:"-чу",10:"-чу",20:"-чы",30:"-чу",40:"-чы",50:"-чү",60:"-чы",70:"-чи",80:"-чи",90:"-чу",100:"-чү"},n=t.defineLocale("ky",{months:"январь_февраль_март_апрель_май_июнь_июль_август_сентябрь_октябрь_ноябрь_декабрь".split("_"),monthsShort:"янв_фев_март_апр_май_июнь_июль_авг_сен_окт_ноя_дек".split("_"),weekdays:"Жекшемби_Дүйшөмбү_Шейшемби_Шаршемби_Бейшемби_Жума_Ишемби".split("_"),weekdaysShort:"Жек_Дүй_Шей_Шар_Бей_Жум_Ише".split("_"),weekdaysMin:"Жк_Дй_Шй_Шр_Бй_Жм_Иш".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Бүгүн саат] LT",nextDay:"[Эртең саат] LT",nextWeek:"dddd [саат] LT",lastDay:"[Кечээ саат] LT",lastWeek:"[Өткөн аптанын] dddd [күнү] [саат] LT",sameElse:"L"},relativeTime:{future:"%s ичинде",past:"%s мурун",s:"бирнече секунд",ss:"%d секунд",m:"бир мүнөт",mm:"%d мүнөт",h:"бир саат",hh:"%d саат",d:"бир күн",dd:"%d күн",M:"бир ай",MM:"%d ай",y:"бир жыл",yy:"%d жыл"},dayOfMonthOrdinalParse:/\d{1,2}-(чи|чы|чү|чу)/,ordinal:function(t){var n=t%10,r=t>=100?100:null;return t+(e[t]||e[n]||e[r])},week:{dow:1,doy:7}});return n}))},9638:function(t,e){function n(t,e){return t===e||t!==t&&e!==e}t.exports=n},"966f":function(t,e,n){var r=n("7e64"),i=n("c05f"),o=1,a=2;function s(t,e,n,s){var l=n.length,u=l,c=!s;if(null==t)return!u;t=Object(t);while(l--){var d=n[l];if(c&&d[2]?d[1]!==t[d[0]]:!(d[0]in t))return!1}while(++l<u){d=n[l];var h=d[0],f=t[h],p=d[1];if(c&&d[2]){if(void 0===f&&!(h in t))return!1}else{var m=new r;if(s)var g=s(f,p,h,t,e,m);if(!(void 0===g?i(p,f,o|a,s,m):g))return!1}}return!0}t.exports=s},9686:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"১",2:"২",3:"৩",4:"৪",5:"৫",6:"৬",7:"৭",8:"৮",9:"৯",0:"০"},n={"১":"1","২":"2","৩":"3","৪":"4","৫":"5","৬":"6","৭":"7","৮":"8","৯":"9","০":"0"},r=t.defineLocale("bn-bd",{months:"জানুয়ারি_ফেব্রুয়ারি_মার্চ_এপ্রিল_মে_জুন_জুলাই_আগস্ট_সেপ্টেম্বর_অক্টোবর_নভেম্বর_ডিসেম্বর".split("_"),monthsShort:"জানু_ফেব্রু_মার্চ_এপ্রিল_মে_জুন_জুলাই_আগস্ট_সেপ্ট_অক্টো_নভে_ডিসে".split("_"),weekdays:"রবিবার_সোমবার_মঙ্গলবার_বুধবার_বৃহস্পতিবার_শুক্রবার_শনিবার".split("_"),weekdaysShort:"রবি_সোম_মঙ্গল_বুধ_বৃহস্পতি_শুক্র_শনি".split("_"),weekdaysMin:"রবি_সোম_মঙ্গল_বুধ_বৃহ_শুক্র_শনি".split("_"),longDateFormat:{LT:"A h:mm সময়",LTS:"A h:mm:ss সময়",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm সময়",LLLL:"dddd, D MMMM YYYY, A h:mm সময়"},calendar:{sameDay:"[আজ] LT",nextDay:"[আগামীকাল] LT",nextWeek:"dddd, LT",lastDay:"[গতকাল] LT",lastWeek:"[গত] dddd, LT",sameElse:"L"},relativeTime:{future:"%s পরে",past:"%s আগে",s:"কয়েক সেকেন্ড",ss:"%d সেকেন্ড",m:"এক মিনিট",mm:"%d মিনিট",h:"এক ঘন্টা",hh:"%d ঘন্টা",d:"এক দিন",dd:"%d দিন",M:"এক মাস",MM:"%d মাস",y:"এক বছর",yy:"%d বছর"},preparse:function(t){return t.replace(/[১২৩৪৫৬৭৮৯০]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/রাত|ভোর|সকাল|দুপুর|বিকাল|সন্ধ্যা|রাত/,meridiemHour:function(t,e){return 12===t&&(t=0),"রাত"===e?t<4?t:t+12:"ভোর"===e||"সকাল"===e?t:"দুপুর"===e?t>=3?t:t+12:"বিকাল"===e||"সন্ধ্যা"===e?t+12:void 0},meridiem:function(t,e,n){return t<4?"রাত":t<6?"ভোর":t<12?"সকাল":t<15?"দুপুর":t<18?"বিকাল":t<20?"সন্ধ্যা":"রাত"},week:{dow:0,doy:6}});return r}))},"96f3":function(t,e){var n=Object.prototype,r=n.hasOwnProperty;function i(t,e){return null!=t&&r.call(t,e)}t.exports=i},"972c":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n){var r={ss:"secunde",mm:"minute",hh:"ore",dd:"zile",ww:"săptămâni",MM:"luni",yy:"ani"},i=" ";return(t%100>=20||t>=100&&t%100===0)&&(i=" de "),t+i+r[n]}var n=t.defineLocale("ro",{months:"ianuarie_februarie_martie_aprilie_mai_iunie_iulie_august_septembrie_octombrie_noiembrie_decembrie".split("_"),monthsShort:"ian._feb._mart._apr._mai_iun._iul._aug._sept._oct._nov._dec.".split("_"),monthsParseExact:!0,weekdays:"duminică_luni_marți_miercuri_joi_vineri_sâmbătă".split("_"),weekdaysShort:"Dum_Lun_Mar_Mie_Joi_Vin_Sâm".split("_"),weekdaysMin:"Du_Lu_Ma_Mi_Jo_Vi_Sâ".split("_"),longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY H:mm",LLLL:"dddd, D MMMM YYYY H:mm"},calendar:{sameDay:"[azi la] LT",nextDay:"[mâine la] LT",nextWeek:"dddd [la] LT",lastDay:"[ieri la] LT",lastWeek:"[fosta] dddd [la] LT",sameElse:"L"},relativeTime:{future:"peste %s",past:"%s în urmă",s:"câteva secunde",ss:e,m:"un minut",mm:e,h:"o oră",hh:e,d:"o zi",dd:e,w:"o săptămână",ww:e,M:"o lună",MM:e,y:"un an",yy:e},week:{dow:1,doy:7}});return n}))},"972ca":function(t,e){function n(t,e,n){var r;return n(t,(function(t,n,i){if(e(t,n,i))return r=n,!1})),r}t.exports=n},9797:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("cy",{months:"Ionawr_Chwefror_Mawrth_Ebrill_Mai_Mehefin_Gorffennaf_Awst_Medi_Hydref_Tachwedd_Rhagfyr".split("_"),monthsShort:"Ion_Chwe_Maw_Ebr_Mai_Meh_Gor_Aws_Med_Hyd_Tach_Rhag".split("_"),weekdays:"Dydd Sul_Dydd Llun_Dydd Mawrth_Dydd Mercher_Dydd Iau_Dydd Gwener_Dydd Sadwrn".split("_"),weekdaysShort:"Sul_Llun_Maw_Mer_Iau_Gwe_Sad".split("_"),weekdaysMin:"Su_Ll_Ma_Me_Ia_Gw_Sa".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Heddiw am] LT",nextDay:"[Yfory am] LT",nextWeek:"dddd [am] LT",lastDay:"[Ddoe am] LT",lastWeek:"dddd [diwethaf am] LT",sameElse:"L"},relativeTime:{future:"mewn %s",past:"%s yn ôl",s:"ychydig eiliadau",ss:"%d eiliad",m:"munud",mm:"%d munud",h:"awr",hh:"%d awr",d:"diwrnod",dd:"%d diwrnod",M:"mis",MM:"%d mis",y:"blwyddyn",yy:"%d flynedd"},dayOfMonthOrdinalParse:/\d{1,2}(fed|ain|af|il|ydd|ed|eg)/,ordinal:function(t){var e=t,n="",r=["","af","il","ydd","ydd","ed","ed","ed","fed","fed","fed","eg","fed","eg","eg","fed","eg","eg","fed","eg","fed"];return e>20?n=40===e||50===e||60===e||80===e||100===e?"fed":"ain":e>0&&(n=r[e]),t+n},week:{dow:1,doy:4}});return e}))},9934:function(t,e,n){var r=n("6fcd"),i=n("41c3"),o=n("30c9");function a(t){return o(t)?r(t,!0):i(t)}t.exports=a},9948:function(t,e,n){var r=n("72af"),i=n("1304"),o=n("9934");function a(t,e){return null==t?t:r(t,i(e),o)}t.exports=a},"99af":function(t,e,n){"use strict";var r=n("23e7"),i=n("d039"),o=n("e8b5"),a=n("861d"),s=n("7b0b"),l=n("07fa"),u=n("3511"),c=n("8418"),d=n("65f0"),h=n("1dde"),f=n("b622"),p=n("2d00"),m=f("isConcatSpreadable"),g=p>=51||!i((function(){var t=[];return t[m]=!1,t.concat()[0]!==t})),y=function(t){if(!a(t))return!1;var e=t[m];return void 0!==e?!!e:o(t)},v=!g||!h("concat");r({target:"Array",proto:!0,arity:1,forced:v},{concat:function(t){var e,n,r,i,o,a=s(this),h=d(a,0),f=0;for(e=-1,r=arguments.length;e<r;e++)if(o=-1===e?a:arguments[e],y(o))for(i=l(o),u(f+i),n=0;n<i;n++,f++)n in o&&c(h,f,o[n]);else u(f+1),c(h,f++,o);return h.length=f,h}})},"99cd":function(t,e){function n(t){return function(e,n,r){var i=-1,o=Object(e),a=r(e),s=a.length;while(s--){var l=a[t?s:++i];if(!1===n(o[l],l,o))break}return e}}t.exports=n},"99d3":function(t,e,n){(function(t){var r=n("585a"),i=e&&!e.nodeType&&e,o=i&&"object"==typeof t&&t&&!t.nodeType&&t,a=o&&o.exports===i,s=a&&r.process,l=function(){try{var t=o&&o.require&&o.require("util").types;return t||s&&s.binding&&s.binding("util")}catch(e){}}();t.exports=l}).call(this,n("62e4")(t))},"9a1f":function(t,e,n){"use strict";var r=n("c65b"),i=n("59ed"),o=n("825a"),a=n("0d51"),s=n("35a1"),l=TypeError;t.exports=function(t,e){var n=arguments.length<2?s(t):e;if(i(n))return o(r(n,t));throw new l(a(t)+" is not iterable")}},"9aff":function(t,e,n){var r=n("9638"),i=n("30c9"),o=n("c098"),a=n("1a8c");function s(t,e,n){if(!a(n))return!1;var s=typeof e;return!!("number"==s?i(n)&&o(e,n.length):"string"==s&&e in n)&&r(n[e],t)}t.exports=s},"9b02":function(t,e,n){var r=n("656b");function i(t,e,n){var i=null==t?void 0:r(t,e);return void 0===i?n:i}t.exports=i},"9bdd":function(t,e,n){"use strict";var r=n("825a"),i=n("2a62");t.exports=function(t,e,n,o){try{return o?e(r(n)[0],n[1]):e(n)}catch(a){i(t,"throw",a)}}},"9bf2":function(t,e,n){"use strict";var r=n("83ab"),i=n("0cfb"),o=n("aed9"),a=n("825a"),s=n("a04b"),l=TypeError,u=Object.defineProperty,c=Object.getOwnPropertyDescriptor,d="enumerable",h="configurable",f="writable";e.f=r?o?function(t,e,n){if(a(t),e=s(e),a(n),"function"===typeof t&&"prototype"===e&&"value"in n&&f in n&&!n[f]){var r=c(t,e);r&&r[f]&&(t[e]=n.value,n={configurable:h in n?n[h]:r[h],enumerable:d in n?n[d]:r[d],writable:!1})}return u(t,e,n)}:u:function(t,e,n){if(a(t),e=s(e),a(n),i)try{return u(t,e,n)}catch(r){}if("get"in n||"set"in n)throw new l("Accessors not supported");return"value"in n&&(t[e]=n.value),t}},"9e69":function(t,e,n){var r=n("2b3e"),i=r.Symbol;t.exports=i},"9e86":function(t,e,n){var r=n("872a"),i=n("242e"),o=n("badf");function a(t,e){var n={};return e=o(e,3),i(t,(function(t,i,o){r(n,i,e(t,i,o))})),n}t.exports=a},"9f26":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=/^(janvier|février|mars|avril|mai|juin|juillet|août|septembre|octobre|novembre|décembre)/i,n=/(janv\.?|févr\.?|mars|avr\.?|mai|juin|juil\.?|août|sept\.?|oct\.?|nov\.?|déc\.?)/i,r=/(janv\.?|févr\.?|mars|avr\.?|mai|juin|juil\.?|août|sept\.?|oct\.?|nov\.?|déc\.?|janvier|février|mars|avril|mai|juin|juillet|août|septembre|octobre|novembre|décembre)/i,i=[/^janv/i,/^févr/i,/^mars/i,/^avr/i,/^mai/i,/^juin/i,/^juil/i,/^août/i,/^sept/i,/^oct/i,/^nov/i,/^déc/i],o=t.defineLocale("fr",{months:"janvier_février_mars_avril_mai_juin_juillet_août_septembre_octobre_novembre_décembre".split("_"),monthsShort:"janv._févr._mars_avr._mai_juin_juil._août_sept._oct._nov._déc.".split("_"),monthsRegex:r,monthsShortRegex:r,monthsStrictRegex:e,monthsShortStrictRegex:n,monthsParse:i,longMonthsParse:i,shortMonthsParse:i,weekdays:"dimanche_lundi_mardi_mercredi_jeudi_vendredi_samedi".split("_"),weekdaysShort:"dim._lun._mar._mer._jeu._ven._sam.".split("_"),weekdaysMin:"di_lu_ma_me_je_ve_sa".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[Aujourd’hui à] LT",nextDay:"[Demain à] LT",nextWeek:"dddd [à] LT",lastDay:"[Hier à] LT",lastWeek:"dddd [dernier à] LT",sameElse:"L"},relativeTime:{future:"dans %s",past:"il y a %s",s:"quelques secondes",ss:"%d secondes",m:"une minute",mm:"%d minutes",h:"une heure",hh:"%d heures",d:"un jour",dd:"%d jours",w:"une semaine",ww:"%d semaines",M:"un mois",MM:"%d mois",y:"un an",yy:"%d ans"},dayOfMonthOrdinalParse:/\d{1,2}(er|)/,ordinal:function(t,e){switch(e){case"D":return t+(1===t?"er":"");default:case"M":case"Q":case"DDD":case"d":return t+(1===t?"er":"e");case"w":case"W":return t+(1===t?"re":"e")}},week:{dow:1,doy:4}});return o}))},"9f7f":function(t,e,n){"use strict";var r=n("d039"),i=n("da84"),o=i.RegExp,a=r((function(){var t=o("a","y");return t.lastIndex=2,null!==t.exec("abcd")})),s=a||r((function(){return!o("a","y").sticky})),l=a||r((function(){var t=o("^r","gy");return t.lastIndex=2,null!==t.exec("str")}));t.exports={BROKEN_CARET:l,MISSED_STICKY:s,UNSUPPORTED_Y:a}},"9f81":function(t,e,n){"use strict";e["a"]={inject:{FormInstance:{default:""}},computed:{itemDisabled(){let t=this.disabled;return!t&&this.FormInstance&&(t=this.FormInstance.disabled),t}}}},a029:function(t,e,n){var r=n("087d"),i=n("2dcb"),o=n("32f4"),a=n("d327"),s=Object.getOwnPropertySymbols,l=s?function(t){var e=[];while(t)r(e,o(t)),t=i(t);return e}:a;t.exports=l},a04b:function(t,e,n){"use strict";var r=n("c04e"),i=n("d9b5");t.exports=function(t){var e=r(t,"string");return i(e)?e:e+""}},a0ac:function(t,e){var n="Expected a function";function r(t){if("function"!=typeof t)throw new TypeError(n);return function(){var e=arguments;switch(e.length){case 0:return!t.call(this);case 1:return!t.call(this,e[0]);case 2:return!t.call(this,e[0],e[1]);case 3:return!t.call(this,e[0],e[1],e[2])}return!t.apply(this,e)}}t.exports=r},a2be:function(t,e,n){var r=n("d612"),i=n("4284"),o=n("c584"),a=1,s=2;function l(t,e,n,l,u,c){var d=n&a,h=t.length,f=e.length;if(h!=f&&!(d&&f>h))return!1;var p=c.get(t),m=c.get(e);if(p&&m)return p==e&&m==t;var g=-1,y=!0,v=n&s?new r:void 0;c.set(t,e),c.set(e,t);while(++g<h){var _=t[g],b=e[g];if(l)var x=d?l(b,_,g,e,t,c):l(_,b,g,t,e,c);if(void 0!==x){if(x)continue;y=!1;break}if(v){if(!i(e,(function(t,e){if(!o(v,e)&&(_===t||u(_,t,n,l,c)))return v.push(e)}))){y=!1;break}}else if(_!==b&&!u(_,b,n,l,c)){y=!1;break}}return c["delete"](t),c["delete"](e),y}t.exports=l},a2db:function(t,e,n){var r=n("9e69"),i=r?r.prototype:void 0,o=i?i.valueOf:void 0;function a(t){return o?Object(o.call(t)):{}}t.exports=a},a356:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=function(t){return 0===t?0:1===t?1:2===t?2:t%100>=3&&t%100<=10?3:t%100>=11?4:5},n={s:["أقل من ثانية","ثانية واحدة",["ثانيتان","ثانيتين"],"%d ثوان","%d ثانية","%d ثانية"],m:["أقل من دقيقة","دقيقة واحدة",["دقيقتان","دقيقتين"],"%d دقائق","%d دقيقة","%d دقيقة"],h:["أقل من ساعة","ساعة واحدة",["ساعتان","ساعتين"],"%d ساعات","%d ساعة","%d ساعة"],d:["أقل من يوم","يوم واحد",["يومان","يومين"],"%d أيام","%d يومًا","%d يوم"],M:["أقل من شهر","شهر واحد",["شهران","شهرين"],"%d أشهر","%d شهرا","%d شهر"],y:["أقل من عام","عام واحد",["عامان","عامين"],"%d أعوام","%d عامًا","%d عام"]},r=function(t){return function(r,i,o,a){var s=e(r),l=n[t][e(r)];return 2===s&&(l=l[i?0:1]),l.replace(/%d/i,r)}},i=["جانفي","فيفري","مارس","أفريل","ماي","جوان","جويلية","أوت","سبتمبر","أكتوبر","نوفمبر","ديسمبر"],o=t.defineLocale("ar-dz",{months:i,monthsShort:i,weekdays:"الأحد_الإثنين_الثلاثاء_الأربعاء_الخميس_الجمعة_السبت".split("_"),weekdaysShort:"أحد_إثنين_ثلاثاء_أربعاء_خميس_جمعة_سبت".split("_"),weekdaysMin:"ح_ن_ث_ر_خ_ج_س".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"D/‏M/‏YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},meridiemParse:/ص|م/,isPM:function(t){return"م"===t},meridiem:function(t,e,n){return t<12?"ص":"م"},calendar:{sameDay:"[اليوم عند الساعة] LT",nextDay:"[غدًا عند الساعة] LT",nextWeek:"dddd [عند الساعة] LT",lastDay:"[أمس عند الساعة] LT",lastWeek:"dddd [عند الساعة] LT",sameElse:"L"},relativeTime:{future:"بعد %s",past:"منذ %s",s:r("s"),ss:r("s"),m:r("m"),mm:r("m"),h:r("h"),hh:r("h"),d:r("d"),dd:r("d"),M:r("M"),MM:r("M"),y:r("y"),yy:r("y")},postformat:function(t){return t.replace(/,/g,"،")},week:{dow:0,doy:4}});return o}))},a3fd:function(t,e,n){var r=n("7948");function i(t,e){return r(e,(function(e){return[e,t[e]]}))}t.exports=i},a434:function(t,e,n){"use strict";var r=n("23e7"),i=n("7b0b"),o=n("23cb"),a=n("5926"),s=n("07fa"),l=n("3a34"),u=n("3511"),c=n("65f0"),d=n("8418"),h=n("083a"),f=n("1dde"),p=f("splice"),m=Math.max,g=Math.min;r({target:"Array",proto:!0,forced:!p},{splice:function(t,e){var n,r,f,p,y,v,_=i(this),b=s(_),x=o(t,b),w=arguments.length;for(0===w?n=r=0:1===w?(n=0,r=b-x):(n=w-2,r=g(m(a(e),0),b-x)),u(b+n-r),f=c(_,r),p=0;p<r;p++)y=x+p,y in _&&d(f,p,_[y]);if(f.length=r,n<r){for(p=x;p<b-r;p++)y=p+r,v=p+n,y in _?_[v]=_[y]:h(_,v);for(p=b;p>b-r+n;p--)h(_,p-1)}else if(n>r)for(p=b-r;p>x;p--)y=p+r-1,v=p+n-1,y in _?_[v]=_[y]:h(_,v);for(p=0;p<n;p++)_[p+x]=arguments[p+2];return l(_,b-r+n),f}})},a454:function(t,e,n){var r=n("72f0"),i=n("3b4a"),o=n("cd9d"),a=i?function(t,e){return i(t,"toString",{configurable:!0,enumerable:!1,value:r(e),writable:!0})}:o;t.exports=a},a4b4:function(t,e,n){"use strict";var r=n("342f");t.exports=/web0s(?!.*chrome)/i.test(r)},a4d3:function(t,e,n){"use strict";n("d9f5"),n("b4f8"),n("c513"),n("e9c4"),n("5a47")},a524:function(t,e,n){var r=n("4245");function i(t){return r(this,t).has(t)}t.exports=i},a630:function(t,e,n){"use strict";var r=n("23e7"),i=n("4df4"),o=n("1c7e"),a=!o((function(t){Array.from(t)}));r({target:"Array",stat:!0,forced:a},{from:i})},a640:function(t,e,n){"use strict";var r=n("d039");t.exports=function(t,e){var n=[][t];return!!n&&r((function(){n.call(null,e||function(){return 1},1)}))}},a6fb:function(t,e,n){t.exports={assign:n("dce5"),assignIn:n("a9b9"),assignInWith:n("c30c"),assignWith:n("900d"),at:n("b8bb"),create:n("c3f4"),defaults:n("95ae"),defaultsDeep:n("3f84"),entries:n("0b49"),entriesIn:n("8332"),extend:n("cdd8"),extendWith:n("d6dd"),findKey:n("74c8"),findLastKey:n("ab3e"),forIn:n("9948"),forInRight:n("18f6"),forOwn:n("020f"),forOwnRight:n("ca7c"),functions:n("fb25"),functionsIn:n("1693"),get:n("9b02"),has:n("3852"),hasIn:n("8604"),invert:n("27f3"),invertBy:n("e759"),invoke:n("3a0e"),keys:n("ec69"),keysIn:n("9934"),mapKeys:n("4472"),mapValues:n("9e86"),merge:n("42454"),mergeWith:n("2411"),omit:n("3eea"),omitBy:n("dd65"),pick:n("2593"),pickBy:n("77c1"),result:n("4467"),set:n("0f5c"),setWith:n("193b"),toPairs:n("f542"),toPairsIn:n("7aa2"),transform:n("50ca"),unset:n("3cfe"),update:n("ee45"),updateWith:n("e9d1"),values:n("3ff1"),valuesIn:n("7159")}},a77e:function(t,e,n){"use strict";n("b0c0");var r=function(){var t=this,e=t._self._c;return e("div",{class:t.classes,style:t.wrapStyles},t._l(t.notices,(function(n){return e("Notice",{key:n.name,attrs:{"prefix-cls":t.prefixCls,styles:n.styles,type:n.type,content:n.content,duration:n.duration,render:n.render,"has-title":n.hasTitle,withIcon:n.withIcon,closable:n.closable,name:n.name,"transition-name":n.transitionName,background:n.background,"msg-type":n.msgType,"on-close":n.onClose}})})),1)},i=[],o=n("ade3"),a=(n("14d9"),n("a434"),function(){var t=this,e=t._self._c;return e("transition",{attrs:{name:t.transitionName,appear:""},on:{enter:t.handleEnter,leave:t.handleLeave}},[e("div",{class:t.classes,style:t.styles},["notice"===t.type?[e("div",{ref:"content",class:t.contentClasses,domProps:{innerHTML:t._s(t.content)}}),e("div",{class:t.contentWithIcon},[e("render-cell",{attrs:{render:t.renderFunc}})],1),t.closable?e("a",{class:[t.baseClass+"-close"],on:{click:t.close}},[e("i",{staticClass:"ivu-icon ivu-icon-ios-close"})]):t._e()]:t._e(),"message"===t.type?[e("div",{ref:"content",class:t.messageContentClasses},[e("div",{class:[t.baseClass+"-content-text"],domProps:{innerHTML:t._s(t.content)}}),e("div",{class:[t.baseClass+"-content-text"]},[e("render-cell",{attrs:{render:t.renderFunc}})],1),t.closable?e("a",{class:[t.baseClass+"-close"],on:{click:t.close}},[e("i",{staticClass:"ivu-icon ivu-icon-ios-close"})]):t._e()])]:t._e()],2)])}),s=[],l=(n("99af"),n("a9e3"),n("1766")),u={components:{RenderCell:l["a"]},props:{prefixCls:{type:String,default:""},duration:{type:Number,default:1.5},type:{type:String},content:{type:String,default:""},withIcon:Boolean,render:{type:Function},hasTitle:Boolean,styles:{type:Object,default:function(){return{right:"50%"}}},closable:{type:Boolean,default:!1},className:{type:String},name:{type:String,required:!0},onClose:{type:Function},transitionName:{type:String},background:{type:Boolean,default:!1},msgType:{type:String}},data:function(){return{withDesc:!1}},computed:{baseClass:function(){return"".concat(this.prefixCls,"-notice")},renderFunc:function(){return this.render||function(){}},classes:function(){return[this.baseClass,Object(o["a"])(Object(o["a"])(Object(o["a"])(Object(o["a"])({},"".concat(this.className),!!this.className),"".concat(this.baseClass,"-closable"),this.closable),"".concat(this.baseClass,"-with-desc"),this.withDesc),"".concat(this.baseClass,"-with-background"),this.background)]},contentClasses:function(){return["".concat(this.baseClass,"-content"),void 0!==this.render?"".concat(this.baseClass,"-content-with-render"):""]},messageContentClasses:function(){return["".concat(this.baseClass,"-content"),Object(o["a"])(Object(o["a"])({},"".concat(this.baseClass,"-content-").concat(this.msgType),this.msgType),"".concat(this.baseClass,"-content-background"),this.background)]},contentWithIcon:function(){return[this.withIcon?"".concat(this.prefixCls,"-content-with-icon"):"",!this.hasTitle&&this.withIcon?"".concat(this.prefixCls,"-content-with-render-notitle"):""]},messageClasses:function(){return["".concat(this.baseClass,"-content"),void 0!==this.render?"".concat(this.baseClass,"-content-with-render"):""]}},methods:{clearCloseTimer:function(){this.closeTimer&&(clearTimeout(this.closeTimer),this.closeTimer=null)},close:function(){this.clearCloseTimer(),this.onClose(),this.$parent.close(this.name)},handleEnter:function(t){"message"===this.type&&(t.style.height=t.scrollHeight+"px")},handleLeave:function(t){"message"===this.type&&1!==document.getElementsByClassName("ivu-message-notice").length&&(t.style.height=0,t.style.paddingTop=0,t.style.paddingBottom=0)}},mounted:function(){var t=this;if(this.clearCloseTimer(),0!==this.duration&&(this.closeTimer=setTimeout((function(){t.close()}),1e3*this.duration)),"ivu-notice"===this.prefixCls){var e=this.$refs.content.querySelectorAll(".".concat(this.prefixCls,"-desc"))[0];this.withDesc=!!this.render||!!e&&""!==e.innerHTML}},beforeDestroy:function(){this.clearCloseTimer()}},c=u,d=n("2877"),h=Object(d["a"])(c,a,s,!1,null,null,null),f=h.exports,p=n("c177"),m="ivu-notification",g=0,y=Date.now();function v(){return"ivuNotification_"+y+"_"+g++}var _={components:{Notice:f},props:{prefixCls:{type:String,default:m},styles:{type:Object,default:function(){return{top:"65px",left:"50%"}}},content:{type:String},className:{type:String}},data:function(){return{notices:[],tIndex:this.handleGetIndex()}},computed:{classes:function(){return["".concat(this.prefixCls),Object(o["a"])({},"".concat(this.className),!!this.className)]},wrapStyles:function(){var t=Object.assign({},this.styles);return t["z-index"]=1010+this.tIndex,t}},methods:{add:function(t){var e=t.name||v(),n=Object.assign({styles:{right:"50%"},content:"",duration:1.5,closable:!1,name:e},t);this.notices.push(n),this.tIndex=this.handleGetIndex()},close:function(t){for(var e=this.notices,n=0;n<e.length;n++)if(e[n].name===t){this.notices.splice(n,1);break}},closeAll:function(){this.notices=[]},handleGetIndex:function(){return Object(p["c"])(),p["d"]}}},b=_,x=Object(d["a"])(b,r,i,!1,null,null,null),w=x.exports,M=n("2b0e");w.newInstance=t=>{const e=t||{},n=new M["a"]({render(t){return t(w,{props:e})}}),r=n.$mount();document.body.appendChild(r.$el);const i=n.$children[0];return{notice(t){i.add(t)},remove(t){i.close(t)},component:i,destroy(t){i.closeAll(),setTimeout((function(){document.body.removeChild(document.getElementsByClassName(t)[0])}),500)}}};e["a"]=w},a78b:function(t,e,n){var r=n("656b"),i=n("159a");function o(t,e,n,o){return i(t,e,n(r(t,e)),o)}t.exports=o},a79d:function(t,e,n){"use strict";var r=n("23e7"),i=n("c430"),o=n("d256"),a=n("d039"),s=n("d066"),l=n("1626"),u=n("4840"),c=n("cdf9"),d=n("cb2d"),h=o&&o.prototype,f=!!o&&a((function(){h["finally"].call({then:function(){}},(function(){}))}));if(r({target:"Promise",proto:!0,real:!0,forced:f},{finally:function(t){var e=u(this,s("Promise")),n=l(t);return this.then(n?function(n){return c(e,t()).then((function(){return n}))}:t,n?function(n){return c(e,t()).then((function(){throw n}))}:t)}}),!i&&l(o)){var p=s("Promise").prototype["finally"];h["finally"]!==p&&d(h,"finally",p,{unsafe:!0})}},a7fa:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("bm",{months:"Zanwuyekalo_Fewuruyekalo_Marisikalo_Awirilikalo_Mɛkalo_Zuwɛnkalo_Zuluyekalo_Utikalo_Sɛtanburukalo_ɔkutɔburukalo_Nowanburukalo_Desanburukalo".split("_"),monthsShort:"Zan_Few_Mar_Awi_Mɛ_Zuw_Zul_Uti_Sɛt_ɔku_Now_Des".split("_"),weekdays:"Kari_Ntɛnɛn_Tarata_Araba_Alamisa_Juma_Sibiri".split("_"),weekdaysShort:"Kar_Ntɛ_Tar_Ara_Ala_Jum_Sib".split("_"),weekdaysMin:"Ka_Nt_Ta_Ar_Al_Ju_Si".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"MMMM [tile] D [san] YYYY",LLL:"MMMM [tile] D [san] YYYY [lɛrɛ] HH:mm",LLLL:"dddd MMMM [tile] D [san] YYYY [lɛrɛ] HH:mm"},calendar:{sameDay:"[Bi lɛrɛ] LT",nextDay:"[Sini lɛrɛ] LT",nextWeek:"dddd [don lɛrɛ] LT",lastDay:"[Kunu lɛrɛ] LT",lastWeek:"dddd [tɛmɛnen lɛrɛ] LT",sameElse:"L"},relativeTime:{future:"%s kɔnɔ",past:"a bɛ %s bɔ",s:"sanga dama dama",ss:"sekondi %d",m:"miniti kelen",mm:"miniti %d",h:"lɛrɛ kelen",hh:"lɛrɛ %d",d:"tile kelen",dd:"tile %d",M:"kalo kelen",MM:"kalo %d",y:"san kelen",yy:"san %d"},week:{dow:1,doy:4}});return e}))},a994:function(t,e,n){var r=n("7d1f"),i=n("32f4"),o=n("ec69");function a(t){return r(t,o,i)}t.exports=a},a9b9:function(t,e,n){var r=n("8eeb"),i=n("2ec1"),o=n("9934"),a=i((function(t,e){r(e,o(e),t)}));t.exports=a},a9e3:function(t,e,n){"use strict";var r=n("23e7"),i=n("c430"),o=n("83ab"),a=n("da84"),s=n("428f"),l=n("e330"),u=n("94ca"),c=n("1a2d"),d=n("7156"),h=n("3a9b"),f=n("d9b5"),p=n("c04e"),m=n("d039"),g=n("241c").f,y=n("06cf").f,v=n("9bf2").f,_=n("408a"),b=n("58a8").trim,x="Number",w=a[x],M=s[x],S=w.prototype,k=a.TypeError,L=l("".slice),T=l("".charCodeAt),D=function(t){var e=p(t,"number");return"bigint"==typeof e?e:O(e)},O=function(t){var e,n,r,i,o,a,s,l,u=p(t,"number");if(f(u))throw new k("Cannot convert a Symbol value to a number");if("string"==typeof u&&u.length>2)if(u=b(u),e=T(u,0),43===e||45===e){if(n=T(u,2),88===n||120===n)return NaN}else if(48===e){switch(T(u,1)){case 66:case 98:r=2,i=49;break;case 79:case 111:r=8,i=55;break;default:return+u}for(o=L(u,2),a=o.length,s=0;s<a;s++)if(l=T(o,s),l<48||l>i)return NaN;return parseInt(o,r)}return+u},I=u(x,!w(" 0o1")||!w("0b1")||w("+0x1")),C=function(t){return h(S,t)&&m((function(){_(t)}))},A=function(t){var e=arguments.length<1?0:w(D(t));return C(this)?d(Object(e),this,A):e};A.prototype=S,I&&!i&&(S.constructor=A),r({global:!0,constructor:!0,wrap:!0,forced:I},{Number:A});var Y=function(t,e){for(var n,r=o?g(e):"MAX_VALUE,MIN_VALUE,NaN,NEGATIVE_INFINITY,POSITIVE_INFINITY,EPSILON,MAX_SAFE_INTEGER,MIN_SAFE_INTEGER,isFinite,isInteger,isNaN,isSafeInteger,parseFloat,parseInt,fromString,range".split(","),i=0;r.length>i;i++)c(e,n=r[i])&&!c(t,n)&&v(t,n,y(e,n))};i&&M&&Y(s[x],M),(I||i)&&Y(s[x],w)},aa30:function(t,e,n){"use strict";var r=n("a77e");const i="ivu-message",o="ivu-icon",a="ivu_message_key_",s={top:24,duration:1.5};let l,u=1;const c={info:"ios-information-circle",success:"ios-checkmark-circle",warning:"ios-alert",error:"ios-close-circle",loading:"ios-loading"};function d(){return l=l||r["a"].newInstance({prefixCls:i,styles:{top:s.top+"px"}}),l}function h(t="",e=s.duration,n,r=function(){},l=!1,h=function(){},f=!1){const p=c[n],m="loading"===n?" ivu-load-loop":"";let g=d();return g.notice({name:`${a}${u}`,duration:e,styles:{},transitionName:"move-up",content:`\n            <div class="${i}-custom-content ${i}-${n}">\n                <i class="${o} ${o}-${p} ${m}"></i>\n                <span>${t}</span>\n            </div>\n        `,render:h,onClose:r,closable:l,type:"message",msgType:n,background:f}),function(){let t=u++;return function(){g.remove(`${a}${t}`)}}()}e["a"]={name:"Message",info(t){return this.message("info",t)},success(t){return this.message("success",t)},warning(t){return this.message("warning",t)},error(t){return this.message("error",t)},loading(t){return this.message("loading",t)},message(t,e){return"string"===typeof e&&(e={content:e}),h(e.content,e.duration,t,e.onClose,e.closable,e.render,e.background)},config(t){(t.top||0===t.top)&&(s.top=t.top),(t.duration||0===t.duration)&&(s.duration=t.duration)},destroy(){let t=d();l=null,t.destroy("ivu-message")}}},aaf2:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,r){var i={s:["थोडया सॅकंडांनी","थोडे सॅकंड"],ss:[t+" सॅकंडांनी",t+" सॅकंड"],m:["एका मिणटान","एक मिनूट"],mm:[t+" मिणटांनी",t+" मिणटां"],h:["एका वरान","एक वर"],hh:[t+" वरांनी",t+" वरां"],d:["एका दिसान","एक दीस"],dd:[t+" दिसांनी",t+" दीस"],M:["एका म्हयन्यान","एक म्हयनो"],MM:[t+" म्हयन्यानी",t+" म्हयने"],y:["एका वर्सान","एक वर्स"],yy:[t+" वर्सांनी",t+" वर्सां"]};return r?i[n][0]:i[n][1]}var n=t.defineLocale("gom-deva",{months:{standalone:"जानेवारी_फेब्रुवारी_मार्च_एप्रील_मे_जून_जुलय_ऑगस्ट_सप्टेंबर_ऑक्टोबर_नोव्हेंबर_डिसेंबर".split("_"),format:"जानेवारीच्या_फेब्रुवारीच्या_मार्चाच्या_एप्रीलाच्या_मेयाच्या_जूनाच्या_जुलयाच्या_ऑगस्टाच्या_सप्टेंबराच्या_ऑक्टोबराच्या_नोव्हेंबराच्या_डिसेंबराच्या".split("_"),isFormat:/MMMM(\s)+D[oD]?/},monthsShort:"जाने._फेब्रु._मार्च_एप्री._मे_जून_जुल._ऑग._सप्टें._ऑक्टो._नोव्हें._डिसें.".split("_"),monthsParseExact:!0,weekdays:"आयतार_सोमार_मंगळार_बुधवार_बिरेस्तार_सुक्रार_शेनवार".split("_"),weekdaysShort:"आयत._सोम._मंगळ._बुध._ब्रेस्त._सुक्र._शेन.".split("_"),weekdaysMin:"आ_सो_मं_बु_ब्रे_सु_शे".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"A h:mm [वाजतां]",LTS:"A h:mm:ss [वाजतां]",L:"DD-MM-YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY A h:mm [वाजतां]",LLLL:"dddd, MMMM Do, YYYY, A h:mm [वाजतां]",llll:"ddd, D MMM YYYY, A h:mm [वाजतां]"},calendar:{sameDay:"[आयज] LT",nextDay:"[फाल्यां] LT",nextWeek:"[फुडलो] dddd[,] LT",lastDay:"[काल] LT",lastWeek:"[फाटलो] dddd[,] LT",sameElse:"L"},relativeTime:{future:"%s",past:"%s आदीं",s:e,ss:e,m:e,mm:e,h:e,hh:e,d:e,dd:e,M:e,MM:e,y:e,yy:e},dayOfMonthOrdinalParse:/\d{1,2}(वेर)/,ordinal:function(t,e){switch(e){case"D":return t+"वेर";default:case"M":case"Q":case"DDD":case"d":case"w":case"W":return t}},week:{dow:0,doy:3},meridiemParse:/राती|सकाळीं|दनपारां|सांजे/,meridiemHour:function(t,e){return 12===t&&(t=0),"राती"===e?t<4?t:t+12:"सकाळीं"===e?t:"दनपारां"===e?t>12?t:t+12:"सांजे"===e?t+12:void 0},meridiem:function(t,e,n){return t<4?"राती":t<12?"सकाळीं":t<16?"दनपारां":t<20?"सांजे":"राती"}});return n}))},ab13:function(t,e,n){"use strict";var r=n("b622"),i=r("match");t.exports=function(t){var e=/./;try{"/./"[t](e)}catch(n){try{return e[i]=!1,"/./"[t](e)}catch(r){}}return!1}},ab36:function(t,e,n){"use strict";var r=n("861d"),i=n("9112");t.exports=function(t,e){r(e)&&"cause"in e&&i(t,"cause",e.cause)}},ab3e:function(t,e,n){var r=n("972ca"),i=n("4be8"),o=n("badf");function a(t,e){return r(t,o(e,3),i)}t.exports=a},abb4:function(t,e,n){"use strict";t.exports=function(t){function e(){}var n={log:e,warn:e,error:e};if(!t&&window.console){var r=function(t,e){t[e]=function(){var t=console[e];if(t.apply)t.apply(console,arguments);else for(var n=0;n<arguments.length;n++)t(arguments[n])}};r(n,"log"),r(n,"warn"),r(n,"error")}return n}},ac1f:function(t,e,n){"use strict";var r=n("23e7"),i=n("9263");r({target:"RegExp",proto:!0,forced:/./.exec!==i},{exec:i})},ac41:function(t,e){function n(t){var e=-1,n=Array(t.size);return t.forEach((function(t){n[++e]=t})),n}t.exports=n},ad6d:function(t,e,n){"use strict";var r=n("825a");t.exports=function(){var t=r(this),e="";return t.hasIndices&&(e+="d"),t.global&&(e+="g"),t.ignoreCase&&(e+="i"),t.multiline&&(e+="m"),t.dotAll&&(e+="s"),t.unicode&&(e+="u"),t.unicodeSets&&(e+="v"),t.sticky&&(e+="y"),e}},ada2:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e){var n=t.split("_");return e%10===1&&e%100!==11?n[0]:e%10>=2&&e%10<=4&&(e%100<10||e%100>=20)?n[1]:n[2]}function n(t,n,r){var i={ss:n?"секунда_секунди_секунд":"секунду_секунди_секунд",mm:n?"хвилина_хвилини_хвилин":"хвилину_хвилини_хвилин",hh:n?"година_години_годин":"годину_години_годин",dd:"день_дні_днів",MM:"місяць_місяці_місяців",yy:"рік_роки_років"};return"m"===r?n?"хвилина":"хвилину":"h"===r?n?"година":"годину":t+" "+e(i[r],+t)}function r(t,e){var n,r={nominative:"неділя_понеділок_вівторок_середа_четвер_п’ятниця_субота".split("_"),accusative:"неділю_понеділок_вівторок_середу_четвер_п’ятницю_суботу".split("_"),genitive:"неділі_понеділка_вівторка_середи_четверга_п’ятниці_суботи".split("_")};return!0===t?r["nominative"].slice(1,7).concat(r["nominative"].slice(0,1)):t?(n=/(\[[ВвУу]\]) ?dddd/.test(e)?"accusative":/\[?(?:минулої|наступної)? ?\] ?dddd/.test(e)?"genitive":"nominative",r[n][t.day()]):r["nominative"]}function i(t){return function(){return t+"о"+(11===this.hours()?"б":"")+"] LT"}}var o=t.defineLocale("uk",{months:{format:"січня_лютого_березня_квітня_травня_червня_липня_серпня_вересня_жовтня_листопада_грудня".split("_"),standalone:"січень_лютий_березень_квітень_травень_червень_липень_серпень_вересень_жовтень_листопад_грудень".split("_")},monthsShort:"січ_лют_бер_квіт_трав_черв_лип_серп_вер_жовт_лист_груд".split("_"),weekdays:r,weekdaysShort:"нд_пн_вт_ср_чт_пт_сб".split("_"),weekdaysMin:"нд_пн_вт_ср_чт_пт_сб".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY р.",LLL:"D MMMM YYYY р., HH:mm",LLLL:"dddd, D MMMM YYYY р., HH:mm"},calendar:{sameDay:i("[Сьогодні "),nextDay:i("[Завтра "),lastDay:i("[Вчора "),nextWeek:i("[У] dddd ["),lastWeek:function(){switch(this.day()){case 0:case 3:case 5:case 6:return i("[Минулої] dddd [").call(this);case 1:case 2:case 4:return i("[Минулого] dddd [").call(this)}},sameElse:"L"},relativeTime:{future:"за %s",past:"%s тому",s:"декілька секунд",ss:n,m:n,mm:n,h:"годину",hh:n,d:"день",dd:n,M:"місяць",MM:n,y:"рік",yy:n},meridiemParse:/ночі|ранку|дня|вечора/,isPM:function(t){return/^(дня|вечора)$/.test(t)},meridiem:function(t,e,n){return t<4?"ночі":t<12?"ранку":t<17?"дня":"вечора"},dayOfMonthOrdinalParse:/\d{1,2}-(й|го)/,ordinal:function(t,e){switch(e){case"M":case"d":case"DDD":case"w":case"W":return t+"-й";case"D":return t+"-го";default:return t}},week:{dow:1,doy:7}});return o}))},addb:function(t,e,n){"use strict";var r=n("f36a"),i=Math.floor,o=function(t,e){var n=t.length;if(n<8){var a,s,l=1;while(l<n){s=l,a=t[l];while(s&&e(t[s-1],a)>0)t[s]=t[--s];s!==l++&&(t[s]=a)}}else{var u=i(n/2),c=o(r(t,0,u),e),d=o(r(t,u),e),h=c.length,f=d.length,p=0,m=0;while(p<h||m<f)t[p+m]=p<h&&m<f?e(c[p],d[m])<=0?c[p++]:d[m++]:p<h?c[p++]:d[m++]}return t};t.exports=o},ade3:function(t,e,n){"use strict";n.d(e,"a",(function(){return a}));var r=n("53ca");n("8172"),n("d9e2"),n("efec"),n("a9e3");function i(t,e){if("object"!=Object(r["a"])(t)||!t)return t;var n=t[Symbol.toPrimitive];if(void 0!==n){var i=n.call(t,e||"default");if("object"!=Object(r["a"])(i))return i;throw new TypeError("@@toPrimitive must return a primitive value.")}return("string"===e?String:Number)(t)}function o(t){var e=i(t,"string");return"symbol"==Object(r["a"])(e)?e:e+""}function a(t,e,n){return e=o(e),e in t?Object.defineProperty(t,e,{value:n,enumerable:!0,configurable:!0,writable:!0}):t[e]=n,t}},ae93:function(t,e,n){"use strict";var r,i,o,a=n("d039"),s=n("1626"),l=n("861d"),u=n("7c73"),c=n("e163"),d=n("cb2d"),h=n("b622"),f=n("c430"),p=h("iterator"),m=!1;[].keys&&(o=[].keys(),"next"in o?(i=c(c(o)),i!==Object.prototype&&(r=i)):m=!0);var g=!l(r)||a((function(){var t={};return r[p].call(t)!==t}));g?r={}:f&&(r=u(r)),s(r[p])||d(r,p,(function(){return this})),t.exports={IteratorPrototype:r,BUGGY_SAFARI_ITERATORS:m}},aeb0:function(t,e,n){"use strict";var r=n("9bf2").f;t.exports=function(t,e,n){n in t||r(t,n,{configurable:!0,get:function(){return e[n]},set:function(t){e[n]=t}})}},aed9:function(t,e,n){"use strict";var r=n("83ab"),i=n("d039");t.exports=r&&i((function(){return 42!==Object.defineProperty((function(){}),"prototype",{value:42,writable:!1}).prototype}))},afb9:function(t,e,n){var r=n("2d7c"),i=n("9520");function o(t,e){return r(e,(function(e){return i(t[e])}))}t.exports=o},b041:function(t,e,n){"use strict";var r=n("00ee"),i=n("f5df");t.exports=r?{}.toString:function(){return"[object "+i(this)+"]"}},b047:function(t,e){function n(t){return function(e){return t(e)}}t.exports=n},b0c0:function(t,e,n){"use strict";var r=n("83ab"),i=n("5e77").EXISTS,o=n("e330"),a=n("edd0"),s=Function.prototype,l=o(s.toString),u=/function\b(?:\s|\/\*[\S\s]*?\*\/|\/\/[^\n\r]*[\n\r]+)*([^\s(/]*)/,c=o(u.exec),d="name";r&&!i&&a(s,d,{configurable:!0,get:function(){try{return c(u,l(this))[1]}catch(t){return""}}})},b1e5:function(t,e,n){var r=n("a994"),i=1,o=Object.prototype,a=o.hasOwnProperty;function s(t,e,n,o,s,l){var u=n&i,c=r(t),d=c.length,h=r(e),f=h.length;if(d!=f&&!u)return!1;var p=d;while(p--){var m=c[p];if(!(u?m in e:a.call(e,m)))return!1}var g=l.get(t),y=l.get(e);if(g&&y)return g==e&&y==t;var v=!0;l.set(t,e),l.set(e,t);var _=u;while(++p<d){m=c[p];var b=t[m],x=e[m];if(o)var w=u?o(x,b,m,e,t,l):o(b,x,m,t,e,l);if(!(void 0===w?b===x||s(b,x,n,o,l):w)){v=!1;break}_||(_="constructor"==m)}if(v&&!_){var M=t.constructor,S=e.constructor;M==S||!("constructor"in t)||!("constructor"in e)||"function"==typeof M&&M instanceof M&&"function"==typeof S&&S instanceof S||(v=!1)}return l["delete"](t),l["delete"](e),v}t.exports=s},b218:function(t,e){var n=9007199254740991;function r(t){return"number"==typeof t&&t>-1&&t%1==0&&t<=n}t.exports=r},b29d:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("lo",{months:"ມັງກອນ_ກຸມພາ_ມີນາ_ເມສາ_ພຶດສະພາ_ມິຖຸນາ_ກໍລະກົດ_ສິງຫາ_ກັນຍາ_ຕຸລາ_ພະຈິກ_ທັນວາ".split("_"),monthsShort:"ມັງກອນ_ກຸມພາ_ມີນາ_ເມສາ_ພຶດສະພາ_ມິຖຸນາ_ກໍລະກົດ_ສິງຫາ_ກັນຍາ_ຕຸລາ_ພະຈິກ_ທັນວາ".split("_"),weekdays:"ອາທິດ_ຈັນ_ອັງຄານ_ພຸດ_ພະຫັດ_ສຸກ_ເສົາ".split("_"),weekdaysShort:"ທິດ_ຈັນ_ອັງຄານ_ພຸດ_ພະຫັດ_ສຸກ_ເສົາ".split("_"),weekdaysMin:"ທ_ຈ_ອຄ_ພ_ພຫ_ສກ_ສ".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"ວັນdddd D MMMM YYYY HH:mm"},meridiemParse:/ຕອນເຊົ້າ|ຕອນແລງ/,isPM:function(t){return"ຕອນແລງ"===t},meridiem:function(t,e,n){return t<12?"ຕອນເຊົ້າ":"ຕອນແລງ"},calendar:{sameDay:"[ມື້ນີ້ເວລາ] LT",nextDay:"[ມື້ອື່ນເວລາ] LT",nextWeek:"[ວັນ]dddd[ໜ້າເວລາ] LT",lastDay:"[ມື້ວານນີ້ເວລາ] LT",lastWeek:"[ວັນ]dddd[ແລ້ວນີ້ເວລາ] LT",sameElse:"L"},relativeTime:{future:"ອີກ %s",past:"%sຜ່ານມາ",s:"ບໍ່ເທົ່າໃດວິນາທີ",ss:"%d ວິນາທີ",m:"1 ນາທີ",mm:"%d ນາທີ",h:"1 ຊົ່ວໂມງ",hh:"%d ຊົ່ວໂມງ",d:"1 ມື້",dd:"%d ມື້",M:"1 ເດືອນ",MM:"%d ເດືອນ",y:"1 ປີ",yy:"%d ປີ"},dayOfMonthOrdinalParse:/(ທີ່)\d{1,2}/,ordinal:function(t){return"ທີ່"+t}});return e}))},b2e6:function(t,e,n){"use strict";var r=n("3a27");e["a"]={props:{lockScroll:{type:Boolean,default:!0}},methods:{checkScrollBar(){let t=window.innerWidth;if(!t){const e=document.documentElement.getBoundingClientRect();t=e.right-Math.abs(e.left)}this.bodyIsOverflowing=document.body.clientWidth<t,this.bodyIsOverflowing&&(this.scrollBarWidth=Object(r["h"])())},checkMaskInVisible(){let t=document.getElementsByClassName("ivu-modal-mask")||[];return Array.from(t).every(t=>"none"===t.style.display||t.classList.contains("fade-leave-to"))},setScrollBar(){this.bodyIsOverflowing&&void 0!==this.scrollBarWidth&&(document.body.style.paddingRight=this.scrollBarWidth+"px")},resetScrollBar(){document.body.style.paddingRight=""},addScrollEffect(){this.lockScroll&&(this.checkScrollBar(),this.setScrollBar(),document.body.style.overflow="hidden")},removeScrollEffect(){this.lockScroll&&this.checkMaskInVisible()&&(document.body.style.overflow="",this.resetScrollBar())}}}},b3e6:function(t,e,n){"use strict";var r=function(){var t=this,e=t._self._c;return e("div",{class:t.classes,style:t.styles},[t._t("default")],2)},i=[],o=n("ade3"),a=(n("99af"),n("a9e3"),n("d3b7"),n("159b"),n("3a27")),s="ivu-row",l={name:"Row",props:{type:{validator:function(t){return Object(a["j"])(t,["flex"])}},align:{validator:function(t){return Object(a["j"])(t,["top","middle","bottom"])}},justify:{validator:function(t){return Object(a["j"])(t,["start","end","center","space-around","space-between"])}},gutter:{type:Number,default:0},className:String,wrap:{type:Boolean,default:!0}},computed:{classes:function(){return["".concat(s),Object(o["a"])(Object(o["a"])(Object(o["a"])(Object(o["a"])(Object(o["a"])(Object(o["a"])(Object(o["a"])({},"".concat(s,"-").concat(this.type),!!this.type),"".concat(s,"-").concat(this.type,"-").concat(this.align),!!this.align&&this.type),"".concat(s,"-").concat(this.type,"-").concat(this.justify),!!this.justify&&this.type),"".concat(s,"-").concat(this.align),!!this.align),"".concat(s,"-").concat(this.justify),!!this.justify),"".concat(this.className),!!this.className),"".concat(s,"-no-wrap"),!this.wrap)]},styles:function(){var t={};return 0!==this.gutter&&(t={marginLeft:this.gutter/-2+"px",marginRight:this.gutter/-2+"px"}),t}},methods:{updateGutter:function(t){var e=Object(a["e"])(this,"iCol"),n=Object(a["d"])(e,"iCol",!1);n.length&&n.forEach((function(e){0!==t&&(e.gutter=t)}))}},watch:{gutter:function(t){this.updateGutter(t)}}},u=l,c=n("2877"),d=Object(c["a"])(u,r,i,!1,null,null,null),h=d.exports;e["a"]=h},b3eb:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,r){var i={m:["eine Minute","einer Minute"],h:["eine Stunde","einer Stunde"],d:["ein Tag","einem Tag"],dd:[t+" Tage",t+" Tagen"],w:["eine Woche","einer Woche"],M:["ein Monat","einem Monat"],MM:[t+" Monate",t+" Monaten"],y:["ein Jahr","einem Jahr"],yy:[t+" Jahre",t+" Jahren"]};return e?i[n][0]:i[n][1]}var n=t.defineLocale("de-at",{months:"Jänner_Februar_März_April_Mai_Juni_Juli_August_September_Oktober_November_Dezember".split("_"),monthsShort:"Jän._Feb._März_Apr._Mai_Juni_Juli_Aug._Sep._Okt._Nov._Dez.".split("_"),monthsParseExact:!0,weekdays:"Sonntag_Montag_Dienstag_Mittwoch_Donnerstag_Freitag_Samstag".split("_"),weekdaysShort:"So._Mo._Di._Mi._Do._Fr._Sa.".split("_"),weekdaysMin:"So_Mo_Di_Mi_Do_Fr_Sa".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY HH:mm",LLLL:"dddd, D. MMMM YYYY HH:mm"},calendar:{sameDay:"[heute um] LT [Uhr]",sameElse:"L",nextDay:"[morgen um] LT [Uhr]",nextWeek:"dddd [um] LT [Uhr]",lastDay:"[gestern um] LT [Uhr]",lastWeek:"[letzten] dddd [um] LT [Uhr]"},relativeTime:{future:"in %s",past:"vor %s",s:"ein paar Sekunden",ss:"%d Sekunden",m:e,mm:"%d Minuten",h:e,hh:"%d Stunden",d:e,dd:e,w:e,ww:"%d Wochen",M:e,MM:e,y:e,yy:e},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return n}))},b42e:function(t,e,n){"use strict";var r=Math.ceil,i=Math.floor;t.exports=Math.trunc||function(t){var e=+t;return(e>0?i:r)(e)}},b469:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,r){var i={m:["eine Minute","einer Minute"],h:["eine Stunde","einer Stunde"],d:["ein Tag","einem Tag"],dd:[t+" Tage",t+" Tagen"],w:["eine Woche","einer Woche"],M:["ein Monat","einem Monat"],MM:[t+" Monate",t+" Monaten"],y:["ein Jahr","einem Jahr"],yy:[t+" Jahre",t+" Jahren"]};return e?i[n][0]:i[n][1]}var n=t.defineLocale("de",{months:"Januar_Februar_März_April_Mai_Juni_Juli_August_September_Oktober_November_Dezember".split("_"),monthsShort:"Jan._Feb._März_Apr._Mai_Juni_Juli_Aug._Sep._Okt._Nov._Dez.".split("_"),monthsParseExact:!0,weekdays:"Sonntag_Montag_Dienstag_Mittwoch_Donnerstag_Freitag_Samstag".split("_"),weekdaysShort:"So._Mo._Di._Mi._Do._Fr._Sa.".split("_"),weekdaysMin:"So_Mo_Di_Mi_Do_Fr_Sa".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY HH:mm",LLLL:"dddd, D. MMMM YYYY HH:mm"},calendar:{sameDay:"[heute um] LT [Uhr]",sameElse:"L",nextDay:"[morgen um] LT [Uhr]",nextWeek:"dddd [um] LT [Uhr]",lastDay:"[gestern um] LT [Uhr]",lastWeek:"[letzten] dddd [um] LT [Uhr]"},relativeTime:{future:"in %s",past:"vor %s",s:"ein paar Sekunden",ss:"%d Sekunden",m:e,mm:"%d Minuten",h:e,hh:"%d Stunden",d:e,dd:e,w:e,ww:"%d Wochen",M:e,MM:e,y:e,yy:e},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return n}))},b4c0:function(t,e,n){var r=n("cb5a");function i(t){var e=this.__data__,n=r(e,t);return n<0?void 0:e[n][1]}t.exports=i},b4f8:function(t,e,n){"use strict";var r=n("23e7"),i=n("d066"),o=n("1a2d"),a=n("577e"),s=n("5692"),l=n("0b43"),u=s("string-to-symbol-registry"),c=s("symbol-to-string-registry");r({target:"Symbol",stat:!0,forced:!l},{for:function(t){var e=a(t);if(o(u,e))return u[e];var n=i("Symbol")(e);return u[e]=n,c[n]=e,n}})},b53d:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("tzm-latn",{months:"innayr_brˤayrˤ_marˤsˤ_ibrir_mayyw_ywnyw_ywlywz_ɣwšt_šwtanbir_ktˤwbrˤ_nwwanbir_dwjnbir".split("_"),monthsShort:"innayr_brˤayrˤ_marˤsˤ_ibrir_mayyw_ywnyw_ywlywz_ɣwšt_šwtanbir_ktˤwbrˤ_nwwanbir_dwjnbir".split("_"),weekdays:"asamas_aynas_asinas_akras_akwas_asimwas_asiḍyas".split("_"),weekdaysShort:"asamas_aynas_asinas_akras_akwas_asimwas_asiḍyas".split("_"),weekdaysMin:"asamas_aynas_asinas_akras_akwas_asimwas_asiḍyas".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[asdkh g] LT",nextDay:"[aska g] LT",nextWeek:"dddd [g] LT",lastDay:"[assant g] LT",lastWeek:"dddd [g] LT",sameElse:"L"},relativeTime:{future:"dadkh s yan %s",past:"yan %s",s:"imik",ss:"%d imik",m:"minuḍ",mm:"%d minuḍ",h:"saɛa",hh:"%d tassaɛin",d:"ass",dd:"%d ossan",M:"ayowr",MM:"%d iyyirn",y:"asgas",yy:"%d isgasn"},week:{dow:6,doy:12}});return e}))},b540:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("jv",{months:"Januari_Februari_Maret_April_Mei_Juni_Juli_Agustus_September_Oktober_Nopember_Desember".split("_"),monthsShort:"Jan_Feb_Mar_Apr_Mei_Jun_Jul_Ags_Sep_Okt_Nop_Des".split("_"),weekdays:"Minggu_Senen_Seloso_Rebu_Kemis_Jemuwah_Septu".split("_"),weekdaysShort:"Min_Sen_Sel_Reb_Kem_Jem_Sep".split("_"),weekdaysMin:"Mg_Sn_Sl_Rb_Km_Jm_Sp".split("_"),longDateFormat:{LT:"HH.mm",LTS:"HH.mm.ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY [pukul] HH.mm",LLLL:"dddd, D MMMM YYYY [pukul] HH.mm"},meridiemParse:/enjing|siyang|sonten|ndalu/,meridiemHour:function(t,e){return 12===t&&(t=0),"enjing"===e?t:"siyang"===e?t>=11?t:t+12:"sonten"===e||"ndalu"===e?t+12:void 0},meridiem:function(t,e,n){return t<11?"enjing":t<15?"siyang":t<19?"sonten":"ndalu"},calendar:{sameDay:"[Dinten puniko pukul] LT",nextDay:"[Mbenjang pukul] LT",nextWeek:"dddd [pukul] LT",lastDay:"[Kala wingi pukul] LT",lastWeek:"dddd [kepengker pukul] LT",sameElse:"L"},relativeTime:{future:"wonten ing %s",past:"%s ingkang kepengker",s:"sawetawis detik",ss:"%d detik",m:"setunggal menit",mm:"%d menit",h:"setunggal jam",hh:"%d jam",d:"sedinten",dd:"%d dinten",M:"sewulan",MM:"%d wulan",y:"setaun",yy:"%d taun"},week:{dow:1,doy:7}});return e}))},b575:function(t,e,n){"use strict";var r,i,o,a,s,l=n("da84"),u=n("157a"),c=n("0366"),d=n("2cf4").set,h=n("01b4"),f=n("1cdc"),p=n("d4c3"),m=n("a4b4"),g=n("605d"),y=l.MutationObserver||l.WebKitMutationObserver,v=l.document,_=l.process,b=l.Promise,x=u("queueMicrotask");if(!x){var w=new h,M=function(){var t,e;g&&(t=_.domain)&&t.exit();while(e=w.get())try{e()}catch(n){throw w.head&&r(),n}t&&t.enter()};f||g||m||!y||!v?!p&&b&&b.resolve?(a=b.resolve(void 0),a.constructor=b,s=c(a.then,a),r=function(){s(M)}):g?r=function(){_.nextTick(M)}:(d=c(d,l),r=function(){d(M)}):(i=!0,o=v.createTextNode(""),new y(M).observe(o,{characterData:!0}),r=function(){o.data=i=!i}),x=function(t){w.head||r(),w.add(t)}}t.exports=x},b5a7:function(t,e,n){var r=n("0b07"),i=n("2b3e"),o=r(i,"DataView");t.exports=o},b5b7:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="ene._feb._mar._abr._may._jun._jul._ago._sep._oct._nov._dic.".split("_"),n="ene_feb_mar_abr_may_jun_jul_ago_sep_oct_nov_dic".split("_"),r=[/^ene/i,/^feb/i,/^mar/i,/^abr/i,/^may/i,/^jun/i,/^jul/i,/^ago/i,/^sep/i,/^oct/i,/^nov/i,/^dic/i],i=/^(enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre|ene\.?|feb\.?|mar\.?|abr\.?|may\.?|jun\.?|jul\.?|ago\.?|sep\.?|oct\.?|nov\.?|dic\.?)/i,o=t.defineLocale("es-mx",{months:"enero_febrero_marzo_abril_mayo_junio_julio_agosto_septiembre_octubre_noviembre_diciembre".split("_"),monthsShort:function(t,r){return t?/-MMM-/.test(r)?n[t.month()]:e[t.month()]:e},monthsRegex:i,monthsShortRegex:i,monthsStrictRegex:/^(enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre)/i,monthsShortStrictRegex:/^(ene\.?|feb\.?|mar\.?|abr\.?|may\.?|jun\.?|jul\.?|ago\.?|sep\.?|oct\.?|nov\.?|dic\.?)/i,monthsParse:r,longMonthsParse:r,shortMonthsParse:r,weekdays:"domingo_lunes_martes_miércoles_jueves_viernes_sábado".split("_"),weekdaysShort:"dom._lun._mar._mié._jue._vie._sáb.".split("_"),weekdaysMin:"do_lu_ma_mi_ju_vi_sá".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD/MM/YYYY",LL:"D [de] MMMM [de] YYYY",LLL:"D [de] MMMM [de] YYYY H:mm",LLLL:"dddd, D [de] MMMM [de] YYYY H:mm"},calendar:{sameDay:function(){return"[hoy a la"+(1!==this.hours()?"s":"")+"] LT"},nextDay:function(){return"[mañana a la"+(1!==this.hours()?"s":"")+"] LT"},nextWeek:function(){return"dddd [a la"+(1!==this.hours()?"s":"")+"] LT"},lastDay:function(){return"[ayer a la"+(1!==this.hours()?"s":"")+"] LT"},lastWeek:function(){return"[el] dddd [pasado a la"+(1!==this.hours()?"s":"")+"] LT"},sameElse:"L"},relativeTime:{future:"en %s",past:"hace %s",s:"unos segundos",ss:"%d segundos",m:"un minuto",mm:"%d minutos",h:"una hora",hh:"%d horas",d:"un día",dd:"%d días",w:"una semana",ww:"%d semanas",M:"un mes",MM:"%d meses",y:"un año",yy:"%d años"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:0,doy:4},invalidDate:"Fecha inválida"});return o}))},b622:function(t,e,n){"use strict";var r=n("da84"),i=n("5692"),o=n("1a2d"),a=n("90e3"),s=n("04f8"),l=n("fdbf"),u=r.Symbol,c=i("wks"),d=l?u["for"]||u:u&&u.withoutSetter||a;t.exports=function(t){return o(c,t)||(c[t]=s&&o(u,t)?u[t]:d("Symbol."+t)),c[t]}},b639:function(t,e,n){"use strict";(function(t){
/*!
 * The buffer module from node.js, for the browser.
 *
 * @author   Feross Aboukhadijeh <http://feross.org>
 * @license  MIT
 */
var r=n("1fb5"),i=n("9152"),o=n("e3db");function a(){try{var t=new Uint8Array(1);return t.__proto__={__proto__:Uint8Array.prototype,foo:function(){return 42}},42===t.foo()&&"function"===typeof t.subarray&&0===t.subarray(1,1).byteLength}catch(e){return!1}}function s(){return u.TYPED_ARRAY_SUPPORT?2147483647:1073741823}function l(t,e){if(s()<e)throw new RangeError("Invalid typed array length");return u.TYPED_ARRAY_SUPPORT?(t=new Uint8Array(e),t.__proto__=u.prototype):(null===t&&(t=new u(e)),t.length=e),t}function u(t,e,n){if(!u.TYPED_ARRAY_SUPPORT&&!(this instanceof u))return new u(t,e,n);if("number"===typeof t){if("string"===typeof e)throw new Error("If encoding is specified then the first argument must be a string");return f(this,t)}return c(this,t,e,n)}function c(t,e,n,r){if("number"===typeof e)throw new TypeError('"value" argument must not be a number');return"undefined"!==typeof ArrayBuffer&&e instanceof ArrayBuffer?g(t,e,n,r):"string"===typeof e?p(t,e,n):y(t,e)}function d(t){if("number"!==typeof t)throw new TypeError('"size" argument must be a number');if(t<0)throw new RangeError('"size" argument must not be negative')}function h(t,e,n,r){return d(e),e<=0?l(t,e):void 0!==n?"string"===typeof r?l(t,e).fill(n,r):l(t,e).fill(n):l(t,e)}function f(t,e){if(d(e),t=l(t,e<0?0:0|v(e)),!u.TYPED_ARRAY_SUPPORT)for(var n=0;n<e;++n)t[n]=0;return t}function p(t,e,n){if("string"===typeof n&&""!==n||(n="utf8"),!u.isEncoding(n))throw new TypeError('"encoding" must be a valid string encoding');var r=0|b(e,n);t=l(t,r);var i=t.write(e,n);return i!==r&&(t=t.slice(0,i)),t}function m(t,e){var n=e.length<0?0:0|v(e.length);t=l(t,n);for(var r=0;r<n;r+=1)t[r]=255&e[r];return t}function g(t,e,n,r){if(e.byteLength,n<0||e.byteLength<n)throw new RangeError("'offset' is out of bounds");if(e.byteLength<n+(r||0))throw new RangeError("'length' is out of bounds");return e=void 0===n&&void 0===r?new Uint8Array(e):void 0===r?new Uint8Array(e,n):new Uint8Array(e,n,r),u.TYPED_ARRAY_SUPPORT?(t=e,t.__proto__=u.prototype):t=m(t,e),t}function y(t,e){if(u.isBuffer(e)){var n=0|v(e.length);return t=l(t,n),0===t.length?t:(e.copy(t,0,0,n),t)}if(e){if("undefined"!==typeof ArrayBuffer&&e.buffer instanceof ArrayBuffer||"length"in e)return"number"!==typeof e.length||et(e.length)?l(t,0):m(t,e);if("Buffer"===e.type&&o(e.data))return m(t,e.data)}throw new TypeError("First argument must be a string, Buffer, ArrayBuffer, Array, or array-like object.")}function v(t){if(t>=s())throw new RangeError("Attempt to allocate Buffer larger than maximum size: 0x"+s().toString(16)+" bytes");return 0|t}function _(t){return+t!=t&&(t=0),u.alloc(+t)}function b(t,e){if(u.isBuffer(t))return t.length;if("undefined"!==typeof ArrayBuffer&&"function"===typeof ArrayBuffer.isView&&(ArrayBuffer.isView(t)||t instanceof ArrayBuffer))return t.byteLength;"string"!==typeof t&&(t=""+t);var n=t.length;if(0===n)return 0;for(var r=!1;;)switch(e){case"ascii":case"latin1":case"binary":return n;case"utf8":case"utf-8":case void 0:return Z(t).length;case"ucs2":case"ucs-2":case"utf16le":case"utf-16le":return 2*n;case"hex":return n>>>1;case"base64":return Q(t).length;default:if(r)return Z(t).length;e=(""+e).toLowerCase(),r=!0}}function x(t,e,n){var r=!1;if((void 0===e||e<0)&&(e=0),e>this.length)return"";if((void 0===n||n>this.length)&&(n=this.length),n<=0)return"";if(n>>>=0,e>>>=0,n<=e)return"";t||(t="utf8");while(1)switch(t){case"hex":return N(this,e,n);case"utf8":case"utf-8":return A(this,e,n);case"ascii":return P(this,e,n);case"latin1":case"binary":return E(this,e,n);case"base64":return C(this,e,n);case"ucs2":case"ucs-2":case"utf16le":case"utf-16le":return R(this,e,n);default:if(r)throw new TypeError("Unknown encoding: "+t);t=(t+"").toLowerCase(),r=!0}}function w(t,e,n){var r=t[e];t[e]=t[n],t[n]=r}function M(t,e,n,r,i){if(0===t.length)return-1;if("string"===typeof n?(r=n,n=0):n>2147483647?n=2147483647:n<-2147483648&&(n=-2147483648),n=+n,isNaN(n)&&(n=i?0:t.length-1),n<0&&(n=t.length+n),n>=t.length){if(i)return-1;n=t.length-1}else if(n<0){if(!i)return-1;n=0}if("string"===typeof e&&(e=u.from(e,r)),u.isBuffer(e))return 0===e.length?-1:S(t,e,n,r,i);if("number"===typeof e)return e&=255,u.TYPED_ARRAY_SUPPORT&&"function"===typeof Uint8Array.prototype.indexOf?i?Uint8Array.prototype.indexOf.call(t,e,n):Uint8Array.prototype.lastIndexOf.call(t,e,n):S(t,[e],n,r,i);throw new TypeError("val must be string, number or Buffer")}function S(t,e,n,r,i){var o,a=1,s=t.length,l=e.length;if(void 0!==r&&(r=String(r).toLowerCase(),"ucs2"===r||"ucs-2"===r||"utf16le"===r||"utf-16le"===r)){if(t.length<2||e.length<2)return-1;a=2,s/=2,l/=2,n/=2}function u(t,e){return 1===a?t[e]:t.readUInt16BE(e*a)}if(i){var c=-1;for(o=n;o<s;o++)if(u(t,o)===u(e,-1===c?0:o-c)){if(-1===c&&(c=o),o-c+1===l)return c*a}else-1!==c&&(o-=o-c),c=-1}else for(n+l>s&&(n=s-l),o=n;o>=0;o--){for(var d=!0,h=0;h<l;h++)if(u(t,o+h)!==u(e,h)){d=!1;break}if(d)return o}return-1}function k(t,e,n,r){n=Number(n)||0;var i=t.length-n;r?(r=Number(r),r>i&&(r=i)):r=i;var o=e.length;if(o%2!==0)throw new TypeError("Invalid hex string");r>o/2&&(r=o/2);for(var a=0;a<r;++a){var s=parseInt(e.substr(2*a,2),16);if(isNaN(s))return a;t[n+a]=s}return a}function L(t,e,n,r){return tt(Z(e,t.length-n),t,n,r)}function T(t,e,n,r){return tt(J(e),t,n,r)}function D(t,e,n,r){return T(t,e,n,r)}function O(t,e,n,r){return tt(Q(e),t,n,r)}function I(t,e,n,r){return tt(K(e,t.length-n),t,n,r)}function C(t,e,n){return 0===e&&n===t.length?r.fromByteArray(t):r.fromByteArray(t.slice(e,n))}function A(t,e,n){n=Math.min(t.length,n);var r=[],i=e;while(i<n){var o,a,s,l,u=t[i],c=null,d=u>239?4:u>223?3:u>191?2:1;if(i+d<=n)switch(d){case 1:u<128&&(c=u);break;case 2:o=t[i+1],128===(192&o)&&(l=(31&u)<<6|63&o,l>127&&(c=l));break;case 3:o=t[i+1],a=t[i+2],128===(192&o)&&128===(192&a)&&(l=(15&u)<<12|(63&o)<<6|63&a,l>2047&&(l<55296||l>57343)&&(c=l));break;case 4:o=t[i+1],a=t[i+2],s=t[i+3],128===(192&o)&&128===(192&a)&&128===(192&s)&&(l=(15&u)<<18|(63&o)<<12|(63&a)<<6|63&s,l>65535&&l<1114112&&(c=l))}null===c?(c=65533,d=1):c>65535&&(c-=65536,r.push(c>>>10&1023|55296),c=56320|1023&c),r.push(c),i+=d}return j(r)}e.Buffer=u,e.SlowBuffer=_,e.INSPECT_MAX_BYTES=50,u.TYPED_ARRAY_SUPPORT=void 0!==t.TYPED_ARRAY_SUPPORT?t.TYPED_ARRAY_SUPPORT:a(),e.kMaxLength=s(),u.poolSize=8192,u._augment=function(t){return t.__proto__=u.prototype,t},u.from=function(t,e,n){return c(null,t,e,n)},u.TYPED_ARRAY_SUPPORT&&(u.prototype.__proto__=Uint8Array.prototype,u.__proto__=Uint8Array,"undefined"!==typeof Symbol&&Symbol.species&&u[Symbol.species]===u&&Object.defineProperty(u,Symbol.species,{value:null,configurable:!0})),u.alloc=function(t,e,n){return h(null,t,e,n)},u.allocUnsafe=function(t){return f(null,t)},u.allocUnsafeSlow=function(t){return f(null,t)},u.isBuffer=function(t){return!(null==t||!t._isBuffer)},u.compare=function(t,e){if(!u.isBuffer(t)||!u.isBuffer(e))throw new TypeError("Arguments must be Buffers");if(t===e)return 0;for(var n=t.length,r=e.length,i=0,o=Math.min(n,r);i<o;++i)if(t[i]!==e[i]){n=t[i],r=e[i];break}return n<r?-1:r<n?1:0},u.isEncoding=function(t){switch(String(t).toLowerCase()){case"hex":case"utf8":case"utf-8":case"ascii":case"latin1":case"binary":case"base64":case"ucs2":case"ucs-2":case"utf16le":case"utf-16le":return!0;default:return!1}},u.concat=function(t,e){if(!o(t))throw new TypeError('"list" argument must be an Array of Buffers');if(0===t.length)return u.alloc(0);var n;if(void 0===e)for(e=0,n=0;n<t.length;++n)e+=t[n].length;var r=u.allocUnsafe(e),i=0;for(n=0;n<t.length;++n){var a=t[n];if(!u.isBuffer(a))throw new TypeError('"list" argument must be an Array of Buffers');a.copy(r,i),i+=a.length}return r},u.byteLength=b,u.prototype._isBuffer=!0,u.prototype.swap16=function(){var t=this.length;if(t%2!==0)throw new RangeError("Buffer size must be a multiple of 16-bits");for(var e=0;e<t;e+=2)w(this,e,e+1);return this},u.prototype.swap32=function(){var t=this.length;if(t%4!==0)throw new RangeError("Buffer size must be a multiple of 32-bits");for(var e=0;e<t;e+=4)w(this,e,e+3),w(this,e+1,e+2);return this},u.prototype.swap64=function(){var t=this.length;if(t%8!==0)throw new RangeError("Buffer size must be a multiple of 64-bits");for(var e=0;e<t;e+=8)w(this,e,e+7),w(this,e+1,e+6),w(this,e+2,e+5),w(this,e+3,e+4);return this},u.prototype.toString=function(){var t=0|this.length;return 0===t?"":0===arguments.length?A(this,0,t):x.apply(this,arguments)},u.prototype.equals=function(t){if(!u.isBuffer(t))throw new TypeError("Argument must be a Buffer");return this===t||0===u.compare(this,t)},u.prototype.inspect=function(){var t="",n=e.INSPECT_MAX_BYTES;return this.length>0&&(t=this.toString("hex",0,n).match(/.{2}/g).join(" "),this.length>n&&(t+=" ... ")),"<Buffer "+t+">"},u.prototype.compare=function(t,e,n,r,i){if(!u.isBuffer(t))throw new TypeError("Argument must be a Buffer");if(void 0===e&&(e=0),void 0===n&&(n=t?t.length:0),void 0===r&&(r=0),void 0===i&&(i=this.length),e<0||n>t.length||r<0||i>this.length)throw new RangeError("out of range index");if(r>=i&&e>=n)return 0;if(r>=i)return-1;if(e>=n)return 1;if(e>>>=0,n>>>=0,r>>>=0,i>>>=0,this===t)return 0;for(var o=i-r,a=n-e,s=Math.min(o,a),l=this.slice(r,i),c=t.slice(e,n),d=0;d<s;++d)if(l[d]!==c[d]){o=l[d],a=c[d];break}return o<a?-1:a<o?1:0},u.prototype.includes=function(t,e,n){return-1!==this.indexOf(t,e,n)},u.prototype.indexOf=function(t,e,n){return M(this,t,e,n,!0)},u.prototype.lastIndexOf=function(t,e,n){return M(this,t,e,n,!1)},u.prototype.write=function(t,e,n,r){if(void 0===e)r="utf8",n=this.length,e=0;else if(void 0===n&&"string"===typeof e)r=e,n=this.length,e=0;else{if(!isFinite(e))throw new Error("Buffer.write(string, encoding, offset[, length]) is no longer supported");e|=0,isFinite(n)?(n|=0,void 0===r&&(r="utf8")):(r=n,n=void 0)}var i=this.length-e;if((void 0===n||n>i)&&(n=i),t.length>0&&(n<0||e<0)||e>this.length)throw new RangeError("Attempt to write outside buffer bounds");r||(r="utf8");for(var o=!1;;)switch(r){case"hex":return k(this,t,e,n);case"utf8":case"utf-8":return L(this,t,e,n);case"ascii":return T(this,t,e,n);case"latin1":case"binary":return D(this,t,e,n);case"base64":return O(this,t,e,n);case"ucs2":case"ucs-2":case"utf16le":case"utf-16le":return I(this,t,e,n);default:if(o)throw new TypeError("Unknown encoding: "+r);r=(""+r).toLowerCase(),o=!0}},u.prototype.toJSON=function(){return{type:"Buffer",data:Array.prototype.slice.call(this._arr||this,0)}};var Y=4096;function j(t){var e=t.length;if(e<=Y)return String.fromCharCode.apply(String,t);var n="",r=0;while(r<e)n+=String.fromCharCode.apply(String,t.slice(r,r+=Y));return n}function P(t,e,n){var r="";n=Math.min(t.length,n);for(var i=e;i<n;++i)r+=String.fromCharCode(127&t[i]);return r}function E(t,e,n){var r="";n=Math.min(t.length,n);for(var i=e;i<n;++i)r+=String.fromCharCode(t[i]);return r}function N(t,e,n){var r=t.length;(!e||e<0)&&(e=0),(!n||n<0||n>r)&&(n=r);for(var i="",o=e;o<n;++o)i+=X(t[o]);return i}function R(t,e,n){for(var r=t.slice(e,n),i="",o=0;o<r.length;o+=2)i+=String.fromCharCode(r[o]+256*r[o+1]);return i}function H(t,e,n){if(t%1!==0||t<0)throw new RangeError("offset is not uint");if(t+e>n)throw new RangeError("Trying to access beyond buffer length")}function z(t,e,n,r,i,o){if(!u.isBuffer(t))throw new TypeError('"buffer" argument must be a Buffer instance');if(e>i||e<o)throw new RangeError('"value" argument is out of bounds');if(n+r>t.length)throw new RangeError("Index out of range")}function F(t,e,n,r){e<0&&(e=65535+e+1);for(var i=0,o=Math.min(t.length-n,2);i<o;++i)t[n+i]=(e&255<<8*(r?i:1-i))>>>8*(r?i:1-i)}function B(t,e,n,r){e<0&&(e=4294967295+e+1);for(var i=0,o=Math.min(t.length-n,4);i<o;++i)t[n+i]=e>>>8*(r?i:3-i)&255}function W(t,e,n,r,i,o){if(n+r>t.length)throw new RangeError("Index out of range");if(n<0)throw new RangeError("Index out of range")}function V(t,e,n,r,o){return o||W(t,e,n,4,34028234663852886e22,-34028234663852886e22),i.write(t,e,n,r,23,4),n+4}function G(t,e,n,r,o){return o||W(t,e,n,8,17976931348623157e292,-17976931348623157e292),i.write(t,e,n,r,52,8),n+8}u.prototype.slice=function(t,e){var n,r=this.length;if(t=~~t,e=void 0===e?r:~~e,t<0?(t+=r,t<0&&(t=0)):t>r&&(t=r),e<0?(e+=r,e<0&&(e=0)):e>r&&(e=r),e<t&&(e=t),u.TYPED_ARRAY_SUPPORT)n=this.subarray(t,e),n.__proto__=u.prototype;else{var i=e-t;n=new u(i,void 0);for(var o=0;o<i;++o)n[o]=this[o+t]}return n},u.prototype.readUIntLE=function(t,e,n){t|=0,e|=0,n||H(t,e,this.length);var r=this[t],i=1,o=0;while(++o<e&&(i*=256))r+=this[t+o]*i;return r},u.prototype.readUIntBE=function(t,e,n){t|=0,e|=0,n||H(t,e,this.length);var r=this[t+--e],i=1;while(e>0&&(i*=256))r+=this[t+--e]*i;return r},u.prototype.readUInt8=function(t,e){return e||H(t,1,this.length),this[t]},u.prototype.readUInt16LE=function(t,e){return e||H(t,2,this.length),this[t]|this[t+1]<<8},u.prototype.readUInt16BE=function(t,e){return e||H(t,2,this.length),this[t]<<8|this[t+1]},u.prototype.readUInt32LE=function(t,e){return e||H(t,4,this.length),(this[t]|this[t+1]<<8|this[t+2]<<16)+16777216*this[t+3]},u.prototype.readUInt32BE=function(t,e){return e||H(t,4,this.length),16777216*this[t]+(this[t+1]<<16|this[t+2]<<8|this[t+3])},u.prototype.readIntLE=function(t,e,n){t|=0,e|=0,n||H(t,e,this.length);var r=this[t],i=1,o=0;while(++o<e&&(i*=256))r+=this[t+o]*i;return i*=128,r>=i&&(r-=Math.pow(2,8*e)),r},u.prototype.readIntBE=function(t,e,n){t|=0,e|=0,n||H(t,e,this.length);var r=e,i=1,o=this[t+--r];while(r>0&&(i*=256))o+=this[t+--r]*i;return i*=128,o>=i&&(o-=Math.pow(2,8*e)),o},u.prototype.readInt8=function(t,e){return e||H(t,1,this.length),128&this[t]?-1*(255-this[t]+1):this[t]},u.prototype.readInt16LE=function(t,e){e||H(t,2,this.length);var n=this[t]|this[t+1]<<8;return 32768&n?4294901760|n:n},u.prototype.readInt16BE=function(t,e){e||H(t,2,this.length);var n=this[t+1]|this[t]<<8;return 32768&n?4294901760|n:n},u.prototype.readInt32LE=function(t,e){return e||H(t,4,this.length),this[t]|this[t+1]<<8|this[t+2]<<16|this[t+3]<<24},u.prototype.readInt32BE=function(t,e){return e||H(t,4,this.length),this[t]<<24|this[t+1]<<16|this[t+2]<<8|this[t+3]},u.prototype.readFloatLE=function(t,e){return e||H(t,4,this.length),i.read(this,t,!0,23,4)},u.prototype.readFloatBE=function(t,e){return e||H(t,4,this.length),i.read(this,t,!1,23,4)},u.prototype.readDoubleLE=function(t,e){return e||H(t,8,this.length),i.read(this,t,!0,52,8)},u.prototype.readDoubleBE=function(t,e){return e||H(t,8,this.length),i.read(this,t,!1,52,8)},u.prototype.writeUIntLE=function(t,e,n,r){if(t=+t,e|=0,n|=0,!r){var i=Math.pow(2,8*n)-1;z(this,t,e,n,i,0)}var o=1,a=0;this[e]=255&t;while(++a<n&&(o*=256))this[e+a]=t/o&255;return e+n},u.prototype.writeUIntBE=function(t,e,n,r){if(t=+t,e|=0,n|=0,!r){var i=Math.pow(2,8*n)-1;z(this,t,e,n,i,0)}var o=n-1,a=1;this[e+o]=255&t;while(--o>=0&&(a*=256))this[e+o]=t/a&255;return e+n},u.prototype.writeUInt8=function(t,e,n){return t=+t,e|=0,n||z(this,t,e,1,255,0),u.TYPED_ARRAY_SUPPORT||(t=Math.floor(t)),this[e]=255&t,e+1},u.prototype.writeUInt16LE=function(t,e,n){return t=+t,e|=0,n||z(this,t,e,2,65535,0),u.TYPED_ARRAY_SUPPORT?(this[e]=255&t,this[e+1]=t>>>8):F(this,t,e,!0),e+2},u.prototype.writeUInt16BE=function(t,e,n){return t=+t,e|=0,n||z(this,t,e,2,65535,0),u.TYPED_ARRAY_SUPPORT?(this[e]=t>>>8,this[e+1]=255&t):F(this,t,e,!1),e+2},u.prototype.writeUInt32LE=function(t,e,n){return t=+t,e|=0,n||z(this,t,e,4,4294967295,0),u.TYPED_ARRAY_SUPPORT?(this[e+3]=t>>>24,this[e+2]=t>>>16,this[e+1]=t>>>8,this[e]=255&t):B(this,t,e,!0),e+4},u.prototype.writeUInt32BE=function(t,e,n){return t=+t,e|=0,n||z(this,t,e,4,4294967295,0),u.TYPED_ARRAY_SUPPORT?(this[e]=t>>>24,this[e+1]=t>>>16,this[e+2]=t>>>8,this[e+3]=255&t):B(this,t,e,!1),e+4},u.prototype.writeIntLE=function(t,e,n,r){if(t=+t,e|=0,!r){var i=Math.pow(2,8*n-1);z(this,t,e,n,i-1,-i)}var o=0,a=1,s=0;this[e]=255&t;while(++o<n&&(a*=256))t<0&&0===s&&0!==this[e+o-1]&&(s=1),this[e+o]=(t/a>>0)-s&255;return e+n},u.prototype.writeIntBE=function(t,e,n,r){if(t=+t,e|=0,!r){var i=Math.pow(2,8*n-1);z(this,t,e,n,i-1,-i)}var o=n-1,a=1,s=0;this[e+o]=255&t;while(--o>=0&&(a*=256))t<0&&0===s&&0!==this[e+o+1]&&(s=1),this[e+o]=(t/a>>0)-s&255;return e+n},u.prototype.writeInt8=function(t,e,n){return t=+t,e|=0,n||z(this,t,e,1,127,-128),u.TYPED_ARRAY_SUPPORT||(t=Math.floor(t)),t<0&&(t=255+t+1),this[e]=255&t,e+1},u.prototype.writeInt16LE=function(t,e,n){return t=+t,e|=0,n||z(this,t,e,2,32767,-32768),u.TYPED_ARRAY_SUPPORT?(this[e]=255&t,this[e+1]=t>>>8):F(this,t,e,!0),e+2},u.prototype.writeInt16BE=function(t,e,n){return t=+t,e|=0,n||z(this,t,e,2,32767,-32768),u.TYPED_ARRAY_SUPPORT?(this[e]=t>>>8,this[e+1]=255&t):F(this,t,e,!1),e+2},u.prototype.writeInt32LE=function(t,e,n){return t=+t,e|=0,n||z(this,t,e,4,2147483647,-2147483648),u.TYPED_ARRAY_SUPPORT?(this[e]=255&t,this[e+1]=t>>>8,this[e+2]=t>>>16,this[e+3]=t>>>24):B(this,t,e,!0),e+4},u.prototype.writeInt32BE=function(t,e,n){return t=+t,e|=0,n||z(this,t,e,4,2147483647,-2147483648),t<0&&(t=4294967295+t+1),u.TYPED_ARRAY_SUPPORT?(this[e]=t>>>24,this[e+1]=t>>>16,this[e+2]=t>>>8,this[e+3]=255&t):B(this,t,e,!1),e+4},u.prototype.writeFloatLE=function(t,e,n){return V(this,t,e,!0,n)},u.prototype.writeFloatBE=function(t,e,n){return V(this,t,e,!1,n)},u.prototype.writeDoubleLE=function(t,e,n){return G(this,t,e,!0,n)},u.prototype.writeDoubleBE=function(t,e,n){return G(this,t,e,!1,n)},u.prototype.copy=function(t,e,n,r){if(n||(n=0),r||0===r||(r=this.length),e>=t.length&&(e=t.length),e||(e=0),r>0&&r<n&&(r=n),r===n)return 0;if(0===t.length||0===this.length)return 0;if(e<0)throw new RangeError("targetStart out of bounds");if(n<0||n>=this.length)throw new RangeError("sourceStart out of bounds");if(r<0)throw new RangeError("sourceEnd out of bounds");r>this.length&&(r=this.length),t.length-e<r-n&&(r=t.length-e+n);var i,o=r-n;if(this===t&&n<e&&e<r)for(i=o-1;i>=0;--i)t[i+e]=this[i+n];else if(o<1e3||!u.TYPED_ARRAY_SUPPORT)for(i=0;i<o;++i)t[i+e]=this[i+n];else Uint8Array.prototype.set.call(t,this.subarray(n,n+o),e);return o},u.prototype.fill=function(t,e,n,r){if("string"===typeof t){if("string"===typeof e?(r=e,e=0,n=this.length):"string"===typeof n&&(r=n,n=this.length),1===t.length){var i=t.charCodeAt(0);i<256&&(t=i)}if(void 0!==r&&"string"!==typeof r)throw new TypeError("encoding must be a string");if("string"===typeof r&&!u.isEncoding(r))throw new TypeError("Unknown encoding: "+r)}else"number"===typeof t&&(t&=255);if(e<0||this.length<e||this.length<n)throw new RangeError("Out of range index");if(n<=e)return this;var o;if(e>>>=0,n=void 0===n?this.length:n>>>0,t||(t=0),"number"===typeof t)for(o=e;o<n;++o)this[o]=t;else{var a=u.isBuffer(t)?t:Z(new u(t,r).toString()),s=a.length;for(o=0;o<n-e;++o)this[o+e]=a[o%s]}return this};var $=/[^+\/0-9A-Za-z-_]/g;function U(t){if(t=q(t).replace($,""),t.length<2)return"";while(t.length%4!==0)t+="=";return t}function q(t){return t.trim?t.trim():t.replace(/^\s+|\s+$/g,"")}function X(t){return t<16?"0"+t.toString(16):t.toString(16)}function Z(t,e){var n;e=e||1/0;for(var r=t.length,i=null,o=[],a=0;a<r;++a){if(n=t.charCodeAt(a),n>55295&&n<57344){if(!i){if(n>56319){(e-=3)>-1&&o.push(239,191,189);continue}if(a+1===r){(e-=3)>-1&&o.push(239,191,189);continue}i=n;continue}if(n<56320){(e-=3)>-1&&o.push(239,191,189),i=n;continue}n=65536+(i-55296<<10|n-56320)}else i&&(e-=3)>-1&&o.push(239,191,189);if(i=null,n<128){if((e-=1)<0)break;o.push(n)}else if(n<2048){if((e-=2)<0)break;o.push(n>>6|192,63&n|128)}else if(n<65536){if((e-=3)<0)break;o.push(n>>12|224,n>>6&63|128,63&n|128)}else{if(!(n<1114112))throw new Error("Invalid code point");if((e-=4)<0)break;o.push(n>>18|240,n>>12&63|128,n>>6&63|128,63&n|128)}}return o}function J(t){for(var e=[],n=0;n<t.length;++n)e.push(255&t.charCodeAt(n));return e}function K(t,e){for(var n,r,i,o=[],a=0;a<t.length;++a){if((e-=2)<0)break;n=t.charCodeAt(a),r=n>>8,i=n%256,o.push(i),o.push(r)}return o}function Q(t){return r.toByteArray(U(t))}function tt(t,e,n,r){for(var i=0;i<r;++i){if(i+n>=e.length||i>=t.length)break;e[i+n]=t[i]}return i}function et(t){return t!==t}}).call(this,n("c8ba"))},b64b:function(t,e,n){"use strict";var r=n("23e7"),i=n("7b0b"),o=n("df75"),a=n("d039"),s=a((function(){o(1)}));r({target:"Object",stat:!0,forced:s},{keys:function(t){return o(i(t))}})},b680:function(t,e,n){"use strict";var r=n("23e7"),i=n("e330"),o=n("5926"),a=n("408a"),s=n("1148"),l=n("d039"),u=RangeError,c=String,d=Math.floor,h=i(s),f=i("".slice),p=i(1..toFixed),m=function(t,e,n){return 0===e?n:e%2===1?m(t,e-1,n*t):m(t*t,e/2,n)},g=function(t){var e=0,n=t;while(n>=4096)e+=12,n/=4096;while(n>=2)e+=1,n/=2;return e},y=function(t,e,n){var r=-1,i=n;while(++r<6)i+=e*t[r],t[r]=i%1e7,i=d(i/1e7)},v=function(t,e){var n=6,r=0;while(--n>=0)r+=t[n],t[n]=d(r/e),r=r%e*1e7},_=function(t){var e=6,n="";while(--e>=0)if(""!==n||0===e||0!==t[e]){var r=c(t[e]);n=""===n?r:n+h("0",7-r.length)+r}return n},b=l((function(){return"0.000"!==p(8e-5,3)||"1"!==p(.9,0)||"1.25"!==p(1.255,2)||"1000000000000000128"!==p(0xde0b6b3a7640080,0)}))||!l((function(){p({})}));r({target:"Number",proto:!0,forced:b},{toFixed:function(t){var e,n,r,i,s=a(this),l=o(t),d=[0,0,0,0,0,0],p="",b="0";if(l<0||l>20)throw new u("Incorrect fraction digits");if(s!==s)return"NaN";if(s<=-1e21||s>=1e21)return c(s);if(s<0&&(p="-",s=-s),s>1e-21)if(e=g(s*m(2,69,1))-69,n=e<0?s*m(2,-e,1):s/m(2,e,1),n*=4503599627370496,e=52-e,e>0){y(d,0,n),r=l;while(r>=7)y(d,1e7,0),r-=7;y(d,m(10,r,1),0),r=e-1;while(r>=23)v(d,1<<23),r-=23;v(d,1<<r),y(d,1,1),v(d,2),b=_(d)}else y(d,0,n),y(d,1<<-e,0),b=_(d)+h("0",l);return l>0?(i=b.length,b=p+(i<=l?"0."+h("0",l-i)+b:f(b,0,i-l)+"."+f(b,i-l))):b=p+b,b}})},b727:function(t,e,n){"use strict";var r=n("0366"),i=n("e330"),o=n("44ad"),a=n("7b0b"),s=n("07fa"),l=n("65f0"),u=i([].push),c=function(t){var e=1===t,n=2===t,i=3===t,c=4===t,d=6===t,h=7===t,f=5===t||d;return function(p,m,g,y){for(var v,_,b=a(p),x=o(b),w=s(x),M=r(m,g),S=0,k=y||l,L=e?k(p,w):n||h?k(p,0):void 0;w>S;S++)if((f||S in x)&&(v=x[S],_=M(v,S,b),t))if(e)L[S]=_;else if(_)switch(t){case 3:return!0;case 5:return v;case 6:return S;case 2:u(L,v)}else switch(t){case 4:return!1;case 7:u(L,v)}return d?-1:i||c?c:L}};t.exports={forEach:c(0),map:c(1),filter:c(2),some:c(3),every:c(4),find:c(5),findIndex:c(6),filterReject:c(7)}},b760:function(t,e,n){var r=n("872a"),i=n("9638");function o(t,e,n){(void 0!==n&&!i(t[e],n)||void 0===n&&!(e in t))&&r(t,e,n)}t.exports=o},b770:function(t,e,n){"use strict";var r=t.exports={};r.forEach=function(t,e){for(var n=0;n<t.length;n++){var r=e(t[n]);if(r)return r}}},b7e9:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("en-sg",{months:"January_February_March_April_May_June_July_August_September_October_November_December".split("_"),monthsShort:"Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec".split("_"),weekdays:"Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday".split("_"),weekdaysShort:"Sun_Mon_Tue_Wed_Thu_Fri_Sat".split("_"),weekdaysMin:"Su_Mo_Tu_We_Th_Fr_Sa".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Today at] LT",nextDay:"[Tomorrow at] LT",nextWeek:"dddd [at] LT",lastDay:"[Yesterday at] LT",lastWeek:"[Last] dddd [at] LT",sameElse:"L"},relativeTime:{future:"in %s",past:"%s ago",s:"a few seconds",ss:"%d seconds",m:"a minute",mm:"%d minutes",h:"an hour",hh:"%d hours",d:"a day",dd:"%d days",M:"a month",MM:"%d months",y:"a year",yy:"%d years"},dayOfMonthOrdinalParse:/\d{1,2}(st|nd|rd|th)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n},week:{dow:1,doy:4}});return e}))},b84c:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("nn",{months:"januar_februar_mars_april_mai_juni_juli_august_september_oktober_november_desember".split("_"),monthsShort:"jan._feb._mars_apr._mai_juni_juli_aug._sep._okt._nov._des.".split("_"),monthsParseExact:!0,weekdays:"sundag_måndag_tysdag_onsdag_torsdag_fredag_laurdag".split("_"),weekdaysShort:"su._må._ty._on._to._fr._lau.".split("_"),weekdaysMin:"su_må_ty_on_to_fr_la".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY [kl.] H:mm",LLLL:"dddd D. MMMM YYYY [kl.] HH:mm"},calendar:{sameDay:"[I dag klokka] LT",nextDay:"[I morgon klokka] LT",nextWeek:"dddd [klokka] LT",lastDay:"[I går klokka] LT",lastWeek:"[Føregåande] dddd [klokka] LT",sameElse:"L"},relativeTime:{future:"om %s",past:"%s sidan",s:"nokre sekund",ss:"%d sekund",m:"eit minutt",mm:"%d minutt",h:"ein time",hh:"%d timar",d:"ein dag",dd:"%d dagar",w:"ei veke",ww:"%d veker",M:"ein månad",MM:"%d månader",y:"eit år",yy:"%d år"},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return e}))},b85c:function(t,e,n){"use strict";n.d(e,"a",(function(){return i}));n("a4d3"),n("e01a"),n("d28b"),n("d9e2"),n("d3b7"),n("3ca3"),n("ddb0");var r=n("06c5");function i(t,e){var n="undefined"!==typeof Symbol&&t[Symbol.iterator]||t["@@iterator"];if(!n){if(Array.isArray(t)||(n=Object(r["a"])(t))||e&&t&&"number"===typeof t.length){n&&(t=n);var i=0,o=function(){};return{s:o,n:function(){return i>=t.length?{done:!0}:{done:!1,value:t[i++]}},e:function(t){throw t},f:o}}throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.")}var a,s=!0,l=!1;return{s:function(){n=n.call(t)},n:function(){var t=n.next();return s=t.done,t},e:function(t){l=!0,a=t},f:function(){try{s||null==n["return"]||n["return"]()}finally{if(l)throw a}}}}},b8bb:function(t,e,n){var r=n("4d8b"),i=n("c6cf"),o=i(r);t.exports=o},b97c:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={ss:"sekundes_sekundēm_sekunde_sekundes".split("_"),m:"minūtes_minūtēm_minūte_minūtes".split("_"),mm:"minūtes_minūtēm_minūte_minūtes".split("_"),h:"stundas_stundām_stunda_stundas".split("_"),hh:"stundas_stundām_stunda_stundas".split("_"),d:"dienas_dienām_diena_dienas".split("_"),dd:"dienas_dienām_diena_dienas".split("_"),M:"mēneša_mēnešiem_mēnesis_mēneši".split("_"),MM:"mēneša_mēnešiem_mēnesis_mēneši".split("_"),y:"gada_gadiem_gads_gadi".split("_"),yy:"gada_gadiem_gads_gadi".split("_")};function n(t,e,n){return n?e%10===1&&e%100!==11?t[2]:t[3]:e%10===1&&e%100!==11?t[0]:t[1]}function r(t,r,i){return t+" "+n(e[i],t,r)}function i(t,r,i){return n(e[i],t,r)}function o(t,e){return e?"dažas sekundes":"dažām sekundēm"}var a=t.defineLocale("lv",{months:"janvāris_februāris_marts_aprīlis_maijs_jūnijs_jūlijs_augusts_septembris_oktobris_novembris_decembris".split("_"),monthsShort:"jan_feb_mar_apr_mai_jūn_jūl_aug_sep_okt_nov_dec".split("_"),weekdays:"svētdiena_pirmdiena_otrdiena_trešdiena_ceturtdiena_piektdiena_sestdiena".split("_"),weekdaysShort:"Sv_P_O_T_C_Pk_S".split("_"),weekdaysMin:"Sv_P_O_T_C_Pk_S".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY.",LL:"YYYY. [gada] D. MMMM",LLL:"YYYY. [gada] D. MMMM, HH:mm",LLLL:"YYYY. [gada] D. MMMM, dddd, HH:mm"},calendar:{sameDay:"[Šodien pulksten] LT",nextDay:"[Rīt pulksten] LT",nextWeek:"dddd [pulksten] LT",lastDay:"[Vakar pulksten] LT",lastWeek:"[Pagājušā] dddd [pulksten] LT",sameElse:"L"},relativeTime:{future:"pēc %s",past:"pirms %s",s:o,ss:r,m:i,mm:r,h:i,hh:r,d:i,dd:r,M:i,MM:r,y:i,yy:r},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return a}))},b980:function(t,e,n){"use strict";var r=n("d039"),i=n("5c6c");t.exports=!r((function(){var t=new Error("a");return!("stack"in t)||(Object.defineProperty(t,"stack",i(1,7)),7!==t.stack)}))},badf:function(t,e,n){var r=n("642a"),i=n("1838"),o=n("cd9d"),a=n("6747"),s=n("f9ce");function l(t){return"function"==typeof t?t:null==t?o:"object"==typeof t?a(t)?i(t[0],t[1]):r(t):s(t)}t.exports=l},bb71:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,r){var i={m:["eine Minute","einer Minute"],h:["eine Stunde","einer Stunde"],d:["ein Tag","einem Tag"],dd:[t+" Tage",t+" Tagen"],w:["eine Woche","einer Woche"],M:["ein Monat","einem Monat"],MM:[t+" Monate",t+" Monaten"],y:["ein Jahr","einem Jahr"],yy:[t+" Jahre",t+" Jahren"]};return e?i[n][0]:i[n][1]}var n=t.defineLocale("de-ch",{months:"Januar_Februar_März_April_Mai_Juni_Juli_August_September_Oktober_November_Dezember".split("_"),monthsShort:"Jan._Feb._März_Apr._Mai_Juni_Juli_Aug._Sep._Okt._Nov._Dez.".split("_"),monthsParseExact:!0,weekdays:"Sonntag_Montag_Dienstag_Mittwoch_Donnerstag_Freitag_Samstag".split("_"),weekdaysShort:"So_Mo_Di_Mi_Do_Fr_Sa".split("_"),weekdaysMin:"So_Mo_Di_Mi_Do_Fr_Sa".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY HH:mm",LLLL:"dddd, D. MMMM YYYY HH:mm"},calendar:{sameDay:"[heute um] LT [Uhr]",sameElse:"L",nextDay:"[morgen um] LT [Uhr]",nextWeek:"dddd [um] LT [Uhr]",lastDay:"[gestern um] LT [Uhr]",lastWeek:"[letzten] dddd [um] LT [Uhr]"},relativeTime:{future:"in %s",past:"vor %s",s:"ein paar Sekunden",ss:"%d Sekunden",m:e,mm:"%d Minuten",h:e,hh:"%d Stunden",d:e,dd:e,w:e,ww:"%d Wochen",M:e,MM:e,y:e,yy:e},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return n}))},bbc0:function(t,e,n){var r=n("6044"),i="__lodash_hash_undefined__",o=Object.prototype,a=o.hasOwnProperty;function s(t){var e=this.__data__;if(r){var n=e[t];return n===i?void 0:n}return a.call(e,t)?e[t]:void 0}t.exports=s},c04e:function(t,e,n){"use strict";var r=n("c65b"),i=n("861d"),o=n("d9b5"),a=n("dc4a"),s=n("485a"),l=n("b622"),u=TypeError,c=l("toPrimitive");t.exports=function(t,e){if(!i(t)||o(t))return t;var n,l=a(t,c);if(l){if(void 0===e&&(e="default"),n=r(l,t,e),!i(n)||o(n))return n;throw new u("Can't convert object to primitive value")}return void 0===e&&(e="number"),s(t,e)}},c05f:function(t,e,n){var r=n("7b97"),i=n("1310");function o(t,e,n,a,s){return t===e||(null==t||null==e||!i(t)&&!i(e)?t!==t&&e!==e:r(t,e,n,a,o,s))}t.exports=o},c098:function(t,e){var n=9007199254740991,r=/^(?:0|[1-9]\d*)$/;function i(t,e){var i=typeof t;return e=null==e?n:e,!!e&&("number"==i||"symbol"!=i&&r.test(t))&&t>-1&&t%1==0&&t<e}t.exports=i},c109:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("tzm",{months:"ⵉⵏⵏⴰⵢⵔ_ⴱⵕⴰⵢⵕ_ⵎⴰⵕⵚ_ⵉⴱⵔⵉⵔ_ⵎⴰⵢⵢⵓ_ⵢⵓⵏⵢⵓ_ⵢⵓⵍⵢⵓⵣ_ⵖⵓⵛⵜ_ⵛⵓⵜⴰⵏⴱⵉⵔ_ⴽⵟⵓⴱⵕ_ⵏⵓⵡⴰⵏⴱⵉⵔ_ⴷⵓⵊⵏⴱⵉⵔ".split("_"),monthsShort:"ⵉⵏⵏⴰⵢⵔ_ⴱⵕⴰⵢⵕ_ⵎⴰⵕⵚ_ⵉⴱⵔⵉⵔ_ⵎⴰⵢⵢⵓ_ⵢⵓⵏⵢⵓ_ⵢⵓⵍⵢⵓⵣ_ⵖⵓⵛⵜ_ⵛⵓⵜⴰⵏⴱⵉⵔ_ⴽⵟⵓⴱⵕ_ⵏⵓⵡⴰⵏⴱⵉⵔ_ⴷⵓⵊⵏⴱⵉⵔ".split("_"),weekdays:"ⴰⵙⴰⵎⴰⵙ_ⴰⵢⵏⴰⵙ_ⴰⵙⵉⵏⴰⵙ_ⴰⴽⵔⴰⵙ_ⴰⴽⵡⴰⵙ_ⴰⵙⵉⵎⵡⴰⵙ_ⴰⵙⵉⴹⵢⴰⵙ".split("_"),weekdaysShort:"ⴰⵙⴰⵎⴰⵙ_ⴰⵢⵏⴰⵙ_ⴰⵙⵉⵏⴰⵙ_ⴰⴽⵔⴰⵙ_ⴰⴽⵡⴰⵙ_ⴰⵙⵉⵎⵡⴰⵙ_ⴰⵙⵉⴹⵢⴰⵙ".split("_"),weekdaysMin:"ⴰⵙⴰⵎⴰⵙ_ⴰⵢⵏⴰⵙ_ⴰⵙⵉⵏⴰⵙ_ⴰⴽⵔⴰⵙ_ⴰⴽⵡⴰⵙ_ⴰⵙⵉⵎⵡⴰⵙ_ⴰⵙⵉⴹⵢⴰⵙ".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[ⴰⵙⴷⵅ ⴴ] LT",nextDay:"[ⴰⵙⴽⴰ ⴴ] LT",nextWeek:"dddd [ⴴ] LT",lastDay:"[ⴰⵚⴰⵏⵜ ⴴ] LT",lastWeek:"dddd [ⴴ] LT",sameElse:"L"},relativeTime:{future:"ⴷⴰⴷⵅ ⵙ ⵢⴰⵏ %s",past:"ⵢⴰⵏ %s",s:"ⵉⵎⵉⴽ",ss:"%d ⵉⵎⵉⴽ",m:"ⵎⵉⵏⵓⴺ",mm:"%d ⵎⵉⵏⵓⴺ",h:"ⵙⴰⵄⴰ",hh:"%d ⵜⴰⵙⵙⴰⵄⵉⵏ",d:"ⴰⵙⵙ",dd:"%d oⵙⵙⴰⵏ",M:"ⴰⵢoⵓⵔ",MM:"%d ⵉⵢⵢⵉⵔⵏ",y:"ⴰⵙⴳⴰⵙ",yy:"%d ⵉⵙⴳⴰⵙⵏ"},week:{dow:6,doy:12}});return e}))},c177:function(t,e,n){"use strict";n.d(e,"d",(function(){return r})),n.d(e,"c",(function(){return o})),n.d(e,"b",(function(){return i})),n.d(e,"a",(function(){return a}));let r=0,i=0;function o(){r++}function a(){i++}},c1c9:function(t,e,n){var r=n("a454"),i=n("f3c1"),o=i(r);t.exports=o},c1df:function(t,e,n){(function(t){var e;//! moment.js
//! version : 2.30.1
//! authors : Tim Wood, Iskren Chernev, Moment.js contributors
//! license : MIT
//! momentjs.com
(function(e,n){t.exports=n()})(0,(function(){"use strict";var r,i;function o(){return r.apply(null,arguments)}function a(t){r=t}function s(t){return t instanceof Array||"[object Array]"===Object.prototype.toString.call(t)}function l(t){return null!=t&&"[object Object]"===Object.prototype.toString.call(t)}function u(t,e){return Object.prototype.hasOwnProperty.call(t,e)}function c(t){if(Object.getOwnPropertyNames)return 0===Object.getOwnPropertyNames(t).length;var e;for(e in t)if(u(t,e))return!1;return!0}function d(t){return void 0===t}function h(t){return"number"===typeof t||"[object Number]"===Object.prototype.toString.call(t)}function f(t){return t instanceof Date||"[object Date]"===Object.prototype.toString.call(t)}function p(t,e){var n,r=[],i=t.length;for(n=0;n<i;++n)r.push(e(t[n],n));return r}function m(t,e){for(var n in e)u(e,n)&&(t[n]=e[n]);return u(e,"toString")&&(t.toString=e.toString),u(e,"valueOf")&&(t.valueOf=e.valueOf),t}function g(t,e,n,r){return Zn(t,e,n,r,!0).utc()}function y(){return{empty:!1,unusedTokens:[],unusedInput:[],overflow:-2,charsLeftOver:0,nullInput:!1,invalidEra:null,invalidMonth:null,invalidFormat:!1,userInvalidated:!1,iso:!1,parsedDateParts:[],era:null,meridiem:null,rfc2822:!1,weekdayMismatch:!1}}function v(t){return null==t._pf&&(t._pf=y()),t._pf}function _(t){var e=null,n=!1,r=t._d&&!isNaN(t._d.getTime());return r&&(e=v(t),n=i.call(e.parsedDateParts,(function(t){return null!=t})),r=e.overflow<0&&!e.empty&&!e.invalidEra&&!e.invalidMonth&&!e.invalidWeekday&&!e.weekdayMismatch&&!e.nullInput&&!e.invalidFormat&&!e.userInvalidated&&(!e.meridiem||e.meridiem&&n),t._strict&&(r=r&&0===e.charsLeftOver&&0===e.unusedTokens.length&&void 0===e.bigHour)),null!=Object.isFrozen&&Object.isFrozen(t)?r:(t._isValid=r,t._isValid)}function b(t){var e=g(NaN);return null!=t?m(v(e),t):v(e).userInvalidated=!0,e}i=Array.prototype.some?Array.prototype.some:function(t){var e,n=Object(this),r=n.length>>>0;for(e=0;e<r;e++)if(e in n&&t.call(this,n[e],e,n))return!0;return!1};var x=o.momentProperties=[],w=!1;function M(t,e){var n,r,i,o=x.length;if(d(e._isAMomentObject)||(t._isAMomentObject=e._isAMomentObject),d(e._i)||(t._i=e._i),d(e._f)||(t._f=e._f),d(e._l)||(t._l=e._l),d(e._strict)||(t._strict=e._strict),d(e._tzm)||(t._tzm=e._tzm),d(e._isUTC)||(t._isUTC=e._isUTC),d(e._offset)||(t._offset=e._offset),d(e._pf)||(t._pf=v(e)),d(e._locale)||(t._locale=e._locale),o>0)for(n=0;n<o;n++)r=x[n],i=e[r],d(i)||(t[r]=i);return t}function S(t){M(this,t),this._d=new Date(null!=t._d?t._d.getTime():NaN),this.isValid()||(this._d=new Date(NaN)),!1===w&&(w=!0,o.updateOffset(this),w=!1)}function k(t){return t instanceof S||null!=t&&null!=t._isAMomentObject}function L(t){!1===o.suppressDeprecationWarnings&&"undefined"!==typeof console&&console.warn}function T(t,e){var n=!0;return m((function(){if(null!=o.deprecationHandler&&o.deprecationHandler(null,t),n){var r,i,a,s=[],l=arguments.length;for(i=0;i<l;i++){if(r="","object"===typeof arguments[i]){for(a in r+="\n["+i+"] ",arguments[0])u(arguments[0],a)&&(r+=a+": "+arguments[0][a]+", ");r=r.slice(0,-2)}else r=arguments[i];s.push(r)}L(t+"\nArguments: "+Array.prototype.slice.call(s).join("")+"\n"+(new Error).stack),n=!1}return e.apply(this,arguments)}),e)}var D,O={};function I(t,e){null!=o.deprecationHandler&&o.deprecationHandler(t,e),O[t]||(L(e),O[t]=!0)}function C(t){return"undefined"!==typeof Function&&t instanceof Function||"[object Function]"===Object.prototype.toString.call(t)}function A(t){var e,n;for(n in t)u(t,n)&&(e=t[n],C(e)?this[n]=e:this["_"+n]=e);this._config=t,this._dayOfMonthOrdinalParseLenient=new RegExp((this._dayOfMonthOrdinalParse.source||this._ordinalParse.source)+"|"+/\d{1,2}/.source)}function Y(t,e){var n,r=m({},t);for(n in e)u(e,n)&&(l(t[n])&&l(e[n])?(r[n]={},m(r[n],t[n]),m(r[n],e[n])):null!=e[n]?r[n]=e[n]:delete r[n]);for(n in t)u(t,n)&&!u(e,n)&&l(t[n])&&(r[n]=m({},r[n]));return r}function j(t){null!=t&&this.set(t)}o.suppressDeprecationWarnings=!1,o.deprecationHandler=null,D=Object.keys?Object.keys:function(t){var e,n=[];for(e in t)u(t,e)&&n.push(e);return n};var P={sameDay:"[Today at] LT",nextDay:"[Tomorrow at] LT",nextWeek:"dddd [at] LT",lastDay:"[Yesterday at] LT",lastWeek:"[Last] dddd [at] LT",sameElse:"L"};function E(t,e,n){var r=this._calendar[t]||this._calendar["sameElse"];return C(r)?r.call(e,n):r}function N(t,e,n){var r=""+Math.abs(t),i=e-r.length,o=t>=0;return(o?n?"+":"":"-")+Math.pow(10,Math.max(0,i)).toString().substr(1)+r}var R=/(\[[^\[]*\])|(\\)?([Hh]mm(ss)?|Mo|MM?M?M?|Do|DDDo|DD?D?D?|ddd?d?|do?|w[o|w]?|W[o|W]?|Qo?|N{1,5}|YYYYYY|YYYYY|YYYY|YY|y{2,4}|yo?|gg(ggg?)?|GG(GGG?)?|e|E|a|A|hh?|HH?|kk?|mm?|ss?|S{1,9}|x|X|zz?|ZZ?|.)/g,H=/(\[[^\[]*\])|(\\)?(LTS|LT|LL?L?L?|l{1,4})/g,z={},F={};function B(t,e,n,r){var i=r;"string"===typeof r&&(i=function(){return this[r]()}),t&&(F[t]=i),e&&(F[e[0]]=function(){return N(i.apply(this,arguments),e[1],e[2])}),n&&(F[n]=function(){return this.localeData().ordinal(i.apply(this,arguments),t)})}function W(t){return t.match(/\[[\s\S]/)?t.replace(/^\[|\]$/g,""):t.replace(/\\/g,"")}function V(t){var e,n,r=t.match(R);for(e=0,n=r.length;e<n;e++)F[r[e]]?r[e]=F[r[e]]:r[e]=W(r[e]);return function(e){var i,o="";for(i=0;i<n;i++)o+=C(r[i])?r[i].call(e,t):r[i];return o}}function G(t,e){return t.isValid()?(e=$(e,t.localeData()),z[e]=z[e]||V(e),z[e](t)):t.localeData().invalidDate()}function $(t,e){var n=5;function r(t){return e.longDateFormat(t)||t}H.lastIndex=0;while(n>=0&&H.test(t))t=t.replace(H,r),H.lastIndex=0,n-=1;return t}var U={LTS:"h:mm:ss A",LT:"h:mm A",L:"MM/DD/YYYY",LL:"MMMM D, YYYY",LLL:"MMMM D, YYYY h:mm A",LLLL:"dddd, MMMM D, YYYY h:mm A"};function q(t){var e=this._longDateFormat[t],n=this._longDateFormat[t.toUpperCase()];return e||!n?e:(this._longDateFormat[t]=n.match(R).map((function(t){return"MMMM"===t||"MM"===t||"DD"===t||"dddd"===t?t.slice(1):t})).join(""),this._longDateFormat[t])}var X="Invalid date";function Z(){return this._invalidDate}var J="%d",K=/\d{1,2}/;function Q(t){return this._ordinal.replace("%d",t)}var tt={future:"in %s",past:"%s ago",s:"a few seconds",ss:"%d seconds",m:"a minute",mm:"%d minutes",h:"an hour",hh:"%d hours",d:"a day",dd:"%d days",w:"a week",ww:"%d weeks",M:"a month",MM:"%d months",y:"a year",yy:"%d years"};function et(t,e,n,r){var i=this._relativeTime[n];return C(i)?i(t,e,n,r):i.replace(/%d/i,t)}function nt(t,e){var n=this._relativeTime[t>0?"future":"past"];return C(n)?n(e):n.replace(/%s/i,e)}var rt={D:"date",dates:"date",date:"date",d:"day",days:"day",day:"day",e:"weekday",weekdays:"weekday",weekday:"weekday",E:"isoWeekday",isoweekdays:"isoWeekday",isoweekday:"isoWeekday",DDD:"dayOfYear",dayofyears:"dayOfYear",dayofyear:"dayOfYear",h:"hour",hours:"hour",hour:"hour",ms:"millisecond",milliseconds:"millisecond",millisecond:"millisecond",m:"minute",minutes:"minute",minute:"minute",M:"month",months:"month",month:"month",Q:"quarter",quarters:"quarter",quarter:"quarter",s:"second",seconds:"second",second:"second",gg:"weekYear",weekyears:"weekYear",weekyear:"weekYear",GG:"isoWeekYear",isoweekyears:"isoWeekYear",isoweekyear:"isoWeekYear",w:"week",weeks:"week",week:"week",W:"isoWeek",isoweeks:"isoWeek",isoweek:"isoWeek",y:"year",years:"year",year:"year"};function it(t){return"string"===typeof t?rt[t]||rt[t.toLowerCase()]:void 0}function ot(t){var e,n,r={};for(n in t)u(t,n)&&(e=it(n),e&&(r[e]=t[n]));return r}var at={date:9,day:11,weekday:11,isoWeekday:11,dayOfYear:4,hour:13,millisecond:16,minute:14,month:8,quarter:7,second:15,weekYear:1,isoWeekYear:1,week:5,isoWeek:5,year:1};function st(t){var e,n=[];for(e in t)u(t,e)&&n.push({unit:e,priority:at[e]});return n.sort((function(t,e){return t.priority-e.priority})),n}var lt,ut=/\d/,ct=/\d\d/,dt=/\d{3}/,ht=/\d{4}/,ft=/[+-]?\d{6}/,pt=/\d\d?/,mt=/\d\d\d\d?/,gt=/\d\d\d\d\d\d?/,yt=/\d{1,3}/,vt=/\d{1,4}/,_t=/[+-]?\d{1,6}/,bt=/\d+/,xt=/[+-]?\d+/,wt=/Z|[+-]\d\d:?\d\d/gi,Mt=/Z|[+-]\d\d(?::?\d\d)?/gi,St=/[+-]?\d+(\.\d{1,3})?/,kt=/[0-9]{0,256}['a-z\u00A0-\u05FF\u0700-\uD7FF\uF900-\uFDCF\uFDF0-\uFF07\uFF10-\uFFEF]{1,256}|[\u0600-\u06FF\/]{1,256}(\s*?[\u0600-\u06FF]{1,256}){1,2}/i,Lt=/^[1-9]\d?/,Tt=/^([1-9]\d|\d)/;function Dt(t,e,n){lt[t]=C(e)?e:function(t,r){return t&&n?n:e}}function Ot(t,e){return u(lt,t)?lt[t](e._strict,e._locale):new RegExp(It(t))}function It(t){return Ct(t.replace("\\","").replace(/\\(\[)|\\(\])|\[([^\]\[]*)\]|\\(.)/g,(function(t,e,n,r,i){return e||n||r||i})))}function Ct(t){return t.replace(/[-\/\\^$*+?.()|[\]{}]/g,"\\$&")}function At(t){return t<0?Math.ceil(t)||0:Math.floor(t)}function Yt(t){var e=+t,n=0;return 0!==e&&isFinite(e)&&(n=At(e)),n}lt={};var jt={};function Pt(t,e){var n,r,i=e;for("string"===typeof t&&(t=[t]),h(e)&&(i=function(t,n){n[e]=Yt(t)}),r=t.length,n=0;n<r;n++)jt[t[n]]=i}function Et(t,e){Pt(t,(function(t,n,r,i){r._w=r._w||{},e(t,r._w,r,i)}))}function Nt(t,e,n){null!=e&&u(jt,t)&&jt[t](e,n._a,n,t)}function Rt(t){return t%4===0&&t%100!==0||t%400===0}var Ht=0,zt=1,Ft=2,Bt=3,Wt=4,Vt=5,Gt=6,$t=7,Ut=8;function qt(t){return Rt(t)?366:365}B("Y",0,0,(function(){var t=this.year();return t<=9999?N(t,4):"+"+t})),B(0,["YY",2],0,(function(){return this.year()%100})),B(0,["YYYY",4],0,"year"),B(0,["YYYYY",5],0,"year"),B(0,["YYYYYY",6,!0],0,"year"),Dt("Y",xt),Dt("YY",pt,ct),Dt("YYYY",vt,ht),Dt("YYYYY",_t,ft),Dt("YYYYYY",_t,ft),Pt(["YYYYY","YYYYYY"],Ht),Pt("YYYY",(function(t,e){e[Ht]=2===t.length?o.parseTwoDigitYear(t):Yt(t)})),Pt("YY",(function(t,e){e[Ht]=o.parseTwoDigitYear(t)})),Pt("Y",(function(t,e){e[Ht]=parseInt(t,10)})),o.parseTwoDigitYear=function(t){return Yt(t)+(Yt(t)>68?1900:2e3)};var Xt,Zt=Kt("FullYear",!0);function Jt(){return Rt(this.year())}function Kt(t,e){return function(n){return null!=n?(te(this,t,n),o.updateOffset(this,e),this):Qt(this,t)}}function Qt(t,e){if(!t.isValid())return NaN;var n=t._d,r=t._isUTC;switch(e){case"Milliseconds":return r?n.getUTCMilliseconds():n.getMilliseconds();case"Seconds":return r?n.getUTCSeconds():n.getSeconds();case"Minutes":return r?n.getUTCMinutes():n.getMinutes();case"Hours":return r?n.getUTCHours():n.getHours();case"Date":return r?n.getUTCDate():n.getDate();case"Day":return r?n.getUTCDay():n.getDay();case"Month":return r?n.getUTCMonth():n.getMonth();case"FullYear":return r?n.getUTCFullYear():n.getFullYear();default:return NaN}}function te(t,e,n){var r,i,o,a,s;if(t.isValid()&&!isNaN(n)){switch(r=t._d,i=t._isUTC,e){case"Milliseconds":return void(i?r.setUTCMilliseconds(n):r.setMilliseconds(n));case"Seconds":return void(i?r.setUTCSeconds(n):r.setSeconds(n));case"Minutes":return void(i?r.setUTCMinutes(n):r.setMinutes(n));case"Hours":return void(i?r.setUTCHours(n):r.setHours(n));case"Date":return void(i?r.setUTCDate(n):r.setDate(n));case"FullYear":break;default:return}o=n,a=t.month(),s=t.date(),s=29!==s||1!==a||Rt(o)?s:28,i?r.setUTCFullYear(o,a,s):r.setFullYear(o,a,s)}}function ee(t){return t=it(t),C(this[t])?this[t]():this}function ne(t,e){if("object"===typeof t){t=ot(t);var n,r=st(t),i=r.length;for(n=0;n<i;n++)this[r[n].unit](t[r[n].unit])}else if(t=it(t),C(this[t]))return this[t](e);return this}function re(t,e){return(t%e+e)%e}function ie(t,e){if(isNaN(t)||isNaN(e))return NaN;var n=re(e,12);return t+=(e-n)/12,1===n?Rt(t)?29:28:31-n%7%2}Xt=Array.prototype.indexOf?Array.prototype.indexOf:function(t){var e;for(e=0;e<this.length;++e)if(this[e]===t)return e;return-1},B("M",["MM",2],"Mo",(function(){return this.month()+1})),B("MMM",0,0,(function(t){return this.localeData().monthsShort(this,t)})),B("MMMM",0,0,(function(t){return this.localeData().months(this,t)})),Dt("M",pt,Lt),Dt("MM",pt,ct),Dt("MMM",(function(t,e){return e.monthsShortRegex(t)})),Dt("MMMM",(function(t,e){return e.monthsRegex(t)})),Pt(["M","MM"],(function(t,e){e[zt]=Yt(t)-1})),Pt(["MMM","MMMM"],(function(t,e,n,r){var i=n._locale.monthsParse(t,r,n._strict);null!=i?e[zt]=i:v(n).invalidMonth=t}));var oe="January_February_March_April_May_June_July_August_September_October_November_December".split("_"),ae="Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec".split("_"),se=/D[oD]?(\[[^\[\]]*\]|\s)+MMMM?/,le=kt,ue=kt;function ce(t,e){return t?s(this._months)?this._months[t.month()]:this._months[(this._months.isFormat||se).test(e)?"format":"standalone"][t.month()]:s(this._months)?this._months:this._months["standalone"]}function de(t,e){return t?s(this._monthsShort)?this._monthsShort[t.month()]:this._monthsShort[se.test(e)?"format":"standalone"][t.month()]:s(this._monthsShort)?this._monthsShort:this._monthsShort["standalone"]}function he(t,e,n){var r,i,o,a=t.toLocaleLowerCase();if(!this._monthsParse)for(this._monthsParse=[],this._longMonthsParse=[],this._shortMonthsParse=[],r=0;r<12;++r)o=g([2e3,r]),this._shortMonthsParse[r]=this.monthsShort(o,"").toLocaleLowerCase(),this._longMonthsParse[r]=this.months(o,"").toLocaleLowerCase();return n?"MMM"===e?(i=Xt.call(this._shortMonthsParse,a),-1!==i?i:null):(i=Xt.call(this._longMonthsParse,a),-1!==i?i:null):"MMM"===e?(i=Xt.call(this._shortMonthsParse,a),-1!==i?i:(i=Xt.call(this._longMonthsParse,a),-1!==i?i:null)):(i=Xt.call(this._longMonthsParse,a),-1!==i?i:(i=Xt.call(this._shortMonthsParse,a),-1!==i?i:null))}function fe(t,e,n){var r,i,o;if(this._monthsParseExact)return he.call(this,t,e,n);for(this._monthsParse||(this._monthsParse=[],this._longMonthsParse=[],this._shortMonthsParse=[]),r=0;r<12;r++){if(i=g([2e3,r]),n&&!this._longMonthsParse[r]&&(this._longMonthsParse[r]=new RegExp("^"+this.months(i,"").replace(".","")+"$","i"),this._shortMonthsParse[r]=new RegExp("^"+this.monthsShort(i,"").replace(".","")+"$","i")),n||this._monthsParse[r]||(o="^"+this.months(i,"")+"|^"+this.monthsShort(i,""),this._monthsParse[r]=new RegExp(o.replace(".",""),"i")),n&&"MMMM"===e&&this._longMonthsParse[r].test(t))return r;if(n&&"MMM"===e&&this._shortMonthsParse[r].test(t))return r;if(!n&&this._monthsParse[r].test(t))return r}}function pe(t,e){if(!t.isValid())return t;if("string"===typeof e)if(/^\d+$/.test(e))e=Yt(e);else if(e=t.localeData().monthsParse(e),!h(e))return t;var n=e,r=t.date();return r=r<29?r:Math.min(r,ie(t.year(),n)),t._isUTC?t._d.setUTCMonth(n,r):t._d.setMonth(n,r),t}function me(t){return null!=t?(pe(this,t),o.updateOffset(this,!0),this):Qt(this,"Month")}function ge(){return ie(this.year(),this.month())}function ye(t){return this._monthsParseExact?(u(this,"_monthsRegex")||_e.call(this),t?this._monthsShortStrictRegex:this._monthsShortRegex):(u(this,"_monthsShortRegex")||(this._monthsShortRegex=le),this._monthsShortStrictRegex&&t?this._monthsShortStrictRegex:this._monthsShortRegex)}function ve(t){return this._monthsParseExact?(u(this,"_monthsRegex")||_e.call(this),t?this._monthsStrictRegex:this._monthsRegex):(u(this,"_monthsRegex")||(this._monthsRegex=ue),this._monthsStrictRegex&&t?this._monthsStrictRegex:this._monthsRegex)}function _e(){function t(t,e){return e.length-t.length}var e,n,r,i,o=[],a=[],s=[];for(e=0;e<12;e++)n=g([2e3,e]),r=Ct(this.monthsShort(n,"")),i=Ct(this.months(n,"")),o.push(r),a.push(i),s.push(i),s.push(r);o.sort(t),a.sort(t),s.sort(t),this._monthsRegex=new RegExp("^("+s.join("|")+")","i"),this._monthsShortRegex=this._monthsRegex,this._monthsStrictRegex=new RegExp("^("+a.join("|")+")","i"),this._monthsShortStrictRegex=new RegExp("^("+o.join("|")+")","i")}function be(t,e,n,r,i,o,a){var s;return t<100&&t>=0?(s=new Date(t+400,e,n,r,i,o,a),isFinite(s.getFullYear())&&s.setFullYear(t)):s=new Date(t,e,n,r,i,o,a),s}function xe(t){var e,n;return t<100&&t>=0?(n=Array.prototype.slice.call(arguments),n[0]=t+400,e=new Date(Date.UTC.apply(null,n)),isFinite(e.getUTCFullYear())&&e.setUTCFullYear(t)):e=new Date(Date.UTC.apply(null,arguments)),e}function we(t,e,n){var r=7+e-n,i=(7+xe(t,0,r).getUTCDay()-e)%7;return-i+r-1}function Me(t,e,n,r,i){var o,a,s=(7+n-r)%7,l=we(t,r,i),u=1+7*(e-1)+s+l;return u<=0?(o=t-1,a=qt(o)+u):u>qt(t)?(o=t+1,a=u-qt(t)):(o=t,a=u),{year:o,dayOfYear:a}}function Se(t,e,n){var r,i,o=we(t.year(),e,n),a=Math.floor((t.dayOfYear()-o-1)/7)+1;return a<1?(i=t.year()-1,r=a+ke(i,e,n)):a>ke(t.year(),e,n)?(r=a-ke(t.year(),e,n),i=t.year()+1):(i=t.year(),r=a),{week:r,year:i}}function ke(t,e,n){var r=we(t,e,n),i=we(t+1,e,n);return(qt(t)-r+i)/7}function Le(t){return Se(t,this._week.dow,this._week.doy).week}B("w",["ww",2],"wo","week"),B("W",["WW",2],"Wo","isoWeek"),Dt("w",pt,Lt),Dt("ww",pt,ct),Dt("W",pt,Lt),Dt("WW",pt,ct),Et(["w","ww","W","WW"],(function(t,e,n,r){e[r.substr(0,1)]=Yt(t)}));var Te={dow:0,doy:6};function De(){return this._week.dow}function Oe(){return this._week.doy}function Ie(t){var e=this.localeData().week(this);return null==t?e:this.add(7*(t-e),"d")}function Ce(t){var e=Se(this,1,4).week;return null==t?e:this.add(7*(t-e),"d")}function Ae(t,e){return"string"!==typeof t?t:isNaN(t)?(t=e.weekdaysParse(t),"number"===typeof t?t:null):parseInt(t,10)}function Ye(t,e){return"string"===typeof t?e.weekdaysParse(t)%7||7:isNaN(t)?null:t}function je(t,e){return t.slice(e,7).concat(t.slice(0,e))}B("d",0,"do","day"),B("dd",0,0,(function(t){return this.localeData().weekdaysMin(this,t)})),B("ddd",0,0,(function(t){return this.localeData().weekdaysShort(this,t)})),B("dddd",0,0,(function(t){return this.localeData().weekdays(this,t)})),B("e",0,0,"weekday"),B("E",0,0,"isoWeekday"),Dt("d",pt),Dt("e",pt),Dt("E",pt),Dt("dd",(function(t,e){return e.weekdaysMinRegex(t)})),Dt("ddd",(function(t,e){return e.weekdaysShortRegex(t)})),Dt("dddd",(function(t,e){return e.weekdaysRegex(t)})),Et(["dd","ddd","dddd"],(function(t,e,n,r){var i=n._locale.weekdaysParse(t,r,n._strict);null!=i?e.d=i:v(n).invalidWeekday=t})),Et(["d","e","E"],(function(t,e,n,r){e[r]=Yt(t)}));var Pe="Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday".split("_"),Ee="Sun_Mon_Tue_Wed_Thu_Fri_Sat".split("_"),Ne="Su_Mo_Tu_We_Th_Fr_Sa".split("_"),Re=kt,He=kt,ze=kt;function Fe(t,e){var n=s(this._weekdays)?this._weekdays:this._weekdays[t&&!0!==t&&this._weekdays.isFormat.test(e)?"format":"standalone"];return!0===t?je(n,this._week.dow):t?n[t.day()]:n}function Be(t){return!0===t?je(this._weekdaysShort,this._week.dow):t?this._weekdaysShort[t.day()]:this._weekdaysShort}function We(t){return!0===t?je(this._weekdaysMin,this._week.dow):t?this._weekdaysMin[t.day()]:this._weekdaysMin}function Ve(t,e,n){var r,i,o,a=t.toLocaleLowerCase();if(!this._weekdaysParse)for(this._weekdaysParse=[],this._shortWeekdaysParse=[],this._minWeekdaysParse=[],r=0;r<7;++r)o=g([2e3,1]).day(r),this._minWeekdaysParse[r]=this.weekdaysMin(o,"").toLocaleLowerCase(),this._shortWeekdaysParse[r]=this.weekdaysShort(o,"").toLocaleLowerCase(),this._weekdaysParse[r]=this.weekdays(o,"").toLocaleLowerCase();return n?"dddd"===e?(i=Xt.call(this._weekdaysParse,a),-1!==i?i:null):"ddd"===e?(i=Xt.call(this._shortWeekdaysParse,a),-1!==i?i:null):(i=Xt.call(this._minWeekdaysParse,a),-1!==i?i:null):"dddd"===e?(i=Xt.call(this._weekdaysParse,a),-1!==i?i:(i=Xt.call(this._shortWeekdaysParse,a),-1!==i?i:(i=Xt.call(this._minWeekdaysParse,a),-1!==i?i:null))):"ddd"===e?(i=Xt.call(this._shortWeekdaysParse,a),-1!==i?i:(i=Xt.call(this._weekdaysParse,a),-1!==i?i:(i=Xt.call(this._minWeekdaysParse,a),-1!==i?i:null))):(i=Xt.call(this._minWeekdaysParse,a),-1!==i?i:(i=Xt.call(this._weekdaysParse,a),-1!==i?i:(i=Xt.call(this._shortWeekdaysParse,a),-1!==i?i:null)))}function Ge(t,e,n){var r,i,o;if(this._weekdaysParseExact)return Ve.call(this,t,e,n);for(this._weekdaysParse||(this._weekdaysParse=[],this._minWeekdaysParse=[],this._shortWeekdaysParse=[],this._fullWeekdaysParse=[]),r=0;r<7;r++){if(i=g([2e3,1]).day(r),n&&!this._fullWeekdaysParse[r]&&(this._fullWeekdaysParse[r]=new RegExp("^"+this.weekdays(i,"").replace(".","\\.?")+"$","i"),this._shortWeekdaysParse[r]=new RegExp("^"+this.weekdaysShort(i,"").replace(".","\\.?")+"$","i"),this._minWeekdaysParse[r]=new RegExp("^"+this.weekdaysMin(i,"").replace(".","\\.?")+"$","i")),this._weekdaysParse[r]||(o="^"+this.weekdays(i,"")+"|^"+this.weekdaysShort(i,"")+"|^"+this.weekdaysMin(i,""),this._weekdaysParse[r]=new RegExp(o.replace(".",""),"i")),n&&"dddd"===e&&this._fullWeekdaysParse[r].test(t))return r;if(n&&"ddd"===e&&this._shortWeekdaysParse[r].test(t))return r;if(n&&"dd"===e&&this._minWeekdaysParse[r].test(t))return r;if(!n&&this._weekdaysParse[r].test(t))return r}}function $e(t){if(!this.isValid())return null!=t?this:NaN;var e=Qt(this,"Day");return null!=t?(t=Ae(t,this.localeData()),this.add(t-e,"d")):e}function Ue(t){if(!this.isValid())return null!=t?this:NaN;var e=(this.day()+7-this.localeData()._week.dow)%7;return null==t?e:this.add(t-e,"d")}function qe(t){if(!this.isValid())return null!=t?this:NaN;if(null!=t){var e=Ye(t,this.localeData());return this.day(this.day()%7?e:e-7)}return this.day()||7}function Xe(t){return this._weekdaysParseExact?(u(this,"_weekdaysRegex")||Ke.call(this),t?this._weekdaysStrictRegex:this._weekdaysRegex):(u(this,"_weekdaysRegex")||(this._weekdaysRegex=Re),this._weekdaysStrictRegex&&t?this._weekdaysStrictRegex:this._weekdaysRegex)}function Ze(t){return this._weekdaysParseExact?(u(this,"_weekdaysRegex")||Ke.call(this),t?this._weekdaysShortStrictRegex:this._weekdaysShortRegex):(u(this,"_weekdaysShortRegex")||(this._weekdaysShortRegex=He),this._weekdaysShortStrictRegex&&t?this._weekdaysShortStrictRegex:this._weekdaysShortRegex)}function Je(t){return this._weekdaysParseExact?(u(this,"_weekdaysRegex")||Ke.call(this),t?this._weekdaysMinStrictRegex:this._weekdaysMinRegex):(u(this,"_weekdaysMinRegex")||(this._weekdaysMinRegex=ze),this._weekdaysMinStrictRegex&&t?this._weekdaysMinStrictRegex:this._weekdaysMinRegex)}function Ke(){function t(t,e){return e.length-t.length}var e,n,r,i,o,a=[],s=[],l=[],u=[];for(e=0;e<7;e++)n=g([2e3,1]).day(e),r=Ct(this.weekdaysMin(n,"")),i=Ct(this.weekdaysShort(n,"")),o=Ct(this.weekdays(n,"")),a.push(r),s.push(i),l.push(o),u.push(r),u.push(i),u.push(o);a.sort(t),s.sort(t),l.sort(t),u.sort(t),this._weekdaysRegex=new RegExp("^("+u.join("|")+")","i"),this._weekdaysShortRegex=this._weekdaysRegex,this._weekdaysMinRegex=this._weekdaysRegex,this._weekdaysStrictRegex=new RegExp("^("+l.join("|")+")","i"),this._weekdaysShortStrictRegex=new RegExp("^("+s.join("|")+")","i"),this._weekdaysMinStrictRegex=new RegExp("^("+a.join("|")+")","i")}function Qe(){return this.hours()%12||12}function tn(){return this.hours()||24}function en(t,e){B(t,0,0,(function(){return this.localeData().meridiem(this.hours(),this.minutes(),e)}))}function nn(t,e){return e._meridiemParse}function rn(t){return"p"===(t+"").toLowerCase().charAt(0)}B("H",["HH",2],0,"hour"),B("h",["hh",2],0,Qe),B("k",["kk",2],0,tn),B("hmm",0,0,(function(){return""+Qe.apply(this)+N(this.minutes(),2)})),B("hmmss",0,0,(function(){return""+Qe.apply(this)+N(this.minutes(),2)+N(this.seconds(),2)})),B("Hmm",0,0,(function(){return""+this.hours()+N(this.minutes(),2)})),B("Hmmss",0,0,(function(){return""+this.hours()+N(this.minutes(),2)+N(this.seconds(),2)})),en("a",!0),en("A",!1),Dt("a",nn),Dt("A",nn),Dt("H",pt,Tt),Dt("h",pt,Lt),Dt("k",pt,Lt),Dt("HH",pt,ct),Dt("hh",pt,ct),Dt("kk",pt,ct),Dt("hmm",mt),Dt("hmmss",gt),Dt("Hmm",mt),Dt("Hmmss",gt),Pt(["H","HH"],Bt),Pt(["k","kk"],(function(t,e,n){var r=Yt(t);e[Bt]=24===r?0:r})),Pt(["a","A"],(function(t,e,n){n._isPm=n._locale.isPM(t),n._meridiem=t})),Pt(["h","hh"],(function(t,e,n){e[Bt]=Yt(t),v(n).bigHour=!0})),Pt("hmm",(function(t,e,n){var r=t.length-2;e[Bt]=Yt(t.substr(0,r)),e[Wt]=Yt(t.substr(r)),v(n).bigHour=!0})),Pt("hmmss",(function(t,e,n){var r=t.length-4,i=t.length-2;e[Bt]=Yt(t.substr(0,r)),e[Wt]=Yt(t.substr(r,2)),e[Vt]=Yt(t.substr(i)),v(n).bigHour=!0})),Pt("Hmm",(function(t,e,n){var r=t.length-2;e[Bt]=Yt(t.substr(0,r)),e[Wt]=Yt(t.substr(r))})),Pt("Hmmss",(function(t,e,n){var r=t.length-4,i=t.length-2;e[Bt]=Yt(t.substr(0,r)),e[Wt]=Yt(t.substr(r,2)),e[Vt]=Yt(t.substr(i))}));var on=/[ap]\.?m?\.?/i,an=Kt("Hours",!0);function sn(t,e,n){return t>11?n?"pm":"PM":n?"am":"AM"}var ln,un={calendar:P,longDateFormat:U,invalidDate:X,ordinal:J,dayOfMonthOrdinalParse:K,relativeTime:tt,months:oe,monthsShort:ae,week:Te,weekdays:Pe,weekdaysMin:Ne,weekdaysShort:Ee,meridiemParse:on},cn={},dn={};function hn(t,e){var n,r=Math.min(t.length,e.length);for(n=0;n<r;n+=1)if(t[n]!==e[n])return n;return r}function fn(t){return t?t.toLowerCase().replace("_","-"):t}function pn(t){var e,n,r,i,o=0;while(o<t.length){i=fn(t[o]).split("-"),e=i.length,n=fn(t[o+1]),n=n?n.split("-"):null;while(e>0){if(r=gn(i.slice(0,e).join("-")),r)return r;if(n&&n.length>=e&&hn(i,n)>=e-1)break;e--}o++}return ln}function mn(t){return!(!t||!t.match("^[^/\\\\]*$"))}function gn(r){var i=null;if(void 0===cn[r]&&"undefined"!==typeof t&&t&&t.exports&&mn(r))try{i=ln._abbr,e,n("4678")("./"+r),yn(i)}catch(o){cn[r]=null}return cn[r]}function yn(t,e){var n;return t&&(n=d(e)?bn(t):vn(t,e),n?ln=n:"undefined"!==typeof console&&console.warn),ln._abbr}function vn(t,e){if(null!==e){var n,r=un;if(e.abbr=t,null!=cn[t])I("defineLocaleOverride","use moment.updateLocale(localeName, config) to change an existing locale. moment.defineLocale(localeName, config) should only be used for creating a new locale See http://momentjs.com/guides/#/warnings/define-locale/ for more info."),r=cn[t]._config;else if(null!=e.parentLocale)if(null!=cn[e.parentLocale])r=cn[e.parentLocale]._config;else{if(n=gn(e.parentLocale),null==n)return dn[e.parentLocale]||(dn[e.parentLocale]=[]),dn[e.parentLocale].push({name:t,config:e}),null;r=n._config}return cn[t]=new j(Y(r,e)),dn[t]&&dn[t].forEach((function(t){vn(t.name,t.config)})),yn(t),cn[t]}return delete cn[t],null}function _n(t,e){if(null!=e){var n,r,i=un;null!=cn[t]&&null!=cn[t].parentLocale?cn[t].set(Y(cn[t]._config,e)):(r=gn(t),null!=r&&(i=r._config),e=Y(i,e),null==r&&(e.abbr=t),n=new j(e),n.parentLocale=cn[t],cn[t]=n),yn(t)}else null!=cn[t]&&(null!=cn[t].parentLocale?(cn[t]=cn[t].parentLocale,t===yn()&&yn(t)):null!=cn[t]&&delete cn[t]);return cn[t]}function bn(t){var e;if(t&&t._locale&&t._locale._abbr&&(t=t._locale._abbr),!t)return ln;if(!s(t)){if(e=gn(t),e)return e;t=[t]}return pn(t)}function xn(){return D(cn)}function wn(t){var e,n=t._a;return n&&-2===v(t).overflow&&(e=n[zt]<0||n[zt]>11?zt:n[Ft]<1||n[Ft]>ie(n[Ht],n[zt])?Ft:n[Bt]<0||n[Bt]>24||24===n[Bt]&&(0!==n[Wt]||0!==n[Vt]||0!==n[Gt])?Bt:n[Wt]<0||n[Wt]>59?Wt:n[Vt]<0||n[Vt]>59?Vt:n[Gt]<0||n[Gt]>999?Gt:-1,v(t)._overflowDayOfYear&&(e<Ht||e>Ft)&&(e=Ft),v(t)._overflowWeeks&&-1===e&&(e=$t),v(t)._overflowWeekday&&-1===e&&(e=Ut),v(t).overflow=e),t}var Mn=/^\s*((?:[+-]\d{6}|\d{4})-(?:\d\d-\d\d|W\d\d-\d|W\d\d|\d\d\d|\d\d))(?:(T| )(\d\d(?::\d\d(?::\d\d(?:[.,]\d+)?)?)?)([+-]\d\d(?::?\d\d)?|\s*Z)?)?$/,Sn=/^\s*((?:[+-]\d{6}|\d{4})(?:\d\d\d\d|W\d\d\d|W\d\d|\d\d\d|\d\d|))(?:(T| )(\d\d(?:\d\d(?:\d\d(?:[.,]\d+)?)?)?)([+-]\d\d(?::?\d\d)?|\s*Z)?)?$/,kn=/Z|[+-]\d\d(?::?\d\d)?/,Ln=[["YYYYYY-MM-DD",/[+-]\d{6}-\d\d-\d\d/],["YYYY-MM-DD",/\d{4}-\d\d-\d\d/],["GGGG-[W]WW-E",/\d{4}-W\d\d-\d/],["GGGG-[W]WW",/\d{4}-W\d\d/,!1],["YYYY-DDD",/\d{4}-\d{3}/],["YYYY-MM",/\d{4}-\d\d/,!1],["YYYYYYMMDD",/[+-]\d{10}/],["YYYYMMDD",/\d{8}/],["GGGG[W]WWE",/\d{4}W\d{3}/],["GGGG[W]WW",/\d{4}W\d{2}/,!1],["YYYYDDD",/\d{7}/],["YYYYMM",/\d{6}/,!1],["YYYY",/\d{4}/,!1]],Tn=[["HH:mm:ss.SSSS",/\d\d:\d\d:\d\d\.\d+/],["HH:mm:ss,SSSS",/\d\d:\d\d:\d\d,\d+/],["HH:mm:ss",/\d\d:\d\d:\d\d/],["HH:mm",/\d\d:\d\d/],["HHmmss.SSSS",/\d\d\d\d\d\d\.\d+/],["HHmmss,SSSS",/\d\d\d\d\d\d,\d+/],["HHmmss",/\d\d\d\d\d\d/],["HHmm",/\d\d\d\d/],["HH",/\d\d/]],Dn=/^\/?Date\((-?\d+)/i,On=/^(?:(Mon|Tue|Wed|Thu|Fri|Sat|Sun),?\s)?(\d{1,2})\s(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\s(\d{2,4})\s(\d\d):(\d\d)(?::(\d\d))?\s(?:(UT|GMT|[ECMP][SD]T)|([Zz])|([+-]\d{4}))$/,In={UT:0,GMT:0,EDT:-240,EST:-300,CDT:-300,CST:-360,MDT:-360,MST:-420,PDT:-420,PST:-480};function Cn(t){var e,n,r,i,o,a,s=t._i,l=Mn.exec(s)||Sn.exec(s),u=Ln.length,c=Tn.length;if(l){for(v(t).iso=!0,e=0,n=u;e<n;e++)if(Ln[e][1].exec(l[1])){i=Ln[e][0],r=!1!==Ln[e][2];break}if(null==i)return void(t._isValid=!1);if(l[3]){for(e=0,n=c;e<n;e++)if(Tn[e][1].exec(l[3])){o=(l[2]||" ")+Tn[e][0];break}if(null==o)return void(t._isValid=!1)}if(!r&&null!=o)return void(t._isValid=!1);if(l[4]){if(!kn.exec(l[4]))return void(t._isValid=!1);a="Z"}t._f=i+(o||"")+(a||""),Wn(t)}else t._isValid=!1}function An(t,e,n,r,i,o){var a=[Yn(t),ae.indexOf(e),parseInt(n,10),parseInt(r,10),parseInt(i,10)];return o&&a.push(parseInt(o,10)),a}function Yn(t){var e=parseInt(t,10);return e<=49?2e3+e:e<=999?1900+e:e}function jn(t){return t.replace(/\([^()]*\)|[\n\t]/g," ").replace(/(\s\s+)/g," ").replace(/^\s\s*/,"").replace(/\s\s*$/,"")}function Pn(t,e,n){if(t){var r=Ee.indexOf(t),i=new Date(e[0],e[1],e[2]).getDay();if(r!==i)return v(n).weekdayMismatch=!0,n._isValid=!1,!1}return!0}function En(t,e,n){if(t)return In[t];if(e)return 0;var r=parseInt(n,10),i=r%100,o=(r-i)/100;return 60*o+i}function Nn(t){var e,n=On.exec(jn(t._i));if(n){if(e=An(n[4],n[3],n[2],n[5],n[6],n[7]),!Pn(n[1],e,t))return;t._a=e,t._tzm=En(n[8],n[9],n[10]),t._d=xe.apply(null,t._a),t._d.setUTCMinutes(t._d.getUTCMinutes()-t._tzm),v(t).rfc2822=!0}else t._isValid=!1}function Rn(t){var e=Dn.exec(t._i);null===e?(Cn(t),!1===t._isValid&&(delete t._isValid,Nn(t),!1===t._isValid&&(delete t._isValid,t._strict?t._isValid=!1:o.createFromInputFallback(t)))):t._d=new Date(+e[1])}function Hn(t,e,n){return null!=t?t:null!=e?e:n}function zn(t){var e=new Date(o.now());return t._useUTC?[e.getUTCFullYear(),e.getUTCMonth(),e.getUTCDate()]:[e.getFullYear(),e.getMonth(),e.getDate()]}function Fn(t){var e,n,r,i,o,a=[];if(!t._d){for(r=zn(t),t._w&&null==t._a[Ft]&&null==t._a[zt]&&Bn(t),null!=t._dayOfYear&&(o=Hn(t._a[Ht],r[Ht]),(t._dayOfYear>qt(o)||0===t._dayOfYear)&&(v(t)._overflowDayOfYear=!0),n=xe(o,0,t._dayOfYear),t._a[zt]=n.getUTCMonth(),t._a[Ft]=n.getUTCDate()),e=0;e<3&&null==t._a[e];++e)t._a[e]=a[e]=r[e];for(;e<7;e++)t._a[e]=a[e]=null==t._a[e]?2===e?1:0:t._a[e];24===t._a[Bt]&&0===t._a[Wt]&&0===t._a[Vt]&&0===t._a[Gt]&&(t._nextDay=!0,t._a[Bt]=0),t._d=(t._useUTC?xe:be).apply(null,a),i=t._useUTC?t._d.getUTCDay():t._d.getDay(),null!=t._tzm&&t._d.setUTCMinutes(t._d.getUTCMinutes()-t._tzm),t._nextDay&&(t._a[Bt]=24),t._w&&"undefined"!==typeof t._w.d&&t._w.d!==i&&(v(t).weekdayMismatch=!0)}}function Bn(t){var e,n,r,i,o,a,s,l,u;e=t._w,null!=e.GG||null!=e.W||null!=e.E?(o=1,a=4,n=Hn(e.GG,t._a[Ht],Se(Jn(),1,4).year),r=Hn(e.W,1),i=Hn(e.E,1),(i<1||i>7)&&(l=!0)):(o=t._locale._week.dow,a=t._locale._week.doy,u=Se(Jn(),o,a),n=Hn(e.gg,t._a[Ht],u.year),r=Hn(e.w,u.week),null!=e.d?(i=e.d,(i<0||i>6)&&(l=!0)):null!=e.e?(i=e.e+o,(e.e<0||e.e>6)&&(l=!0)):i=o),r<1||r>ke(n,o,a)?v(t)._overflowWeeks=!0:null!=l?v(t)._overflowWeekday=!0:(s=Me(n,r,i,o,a),t._a[Ht]=s.year,t._dayOfYear=s.dayOfYear)}function Wn(t){if(t._f!==o.ISO_8601)if(t._f!==o.RFC_2822){t._a=[],v(t).empty=!0;var e,n,r,i,a,s,l,u=""+t._i,c=u.length,d=0;for(r=$(t._f,t._locale).match(R)||[],l=r.length,e=0;e<l;e++)i=r[e],n=(u.match(Ot(i,t))||[])[0],n&&(a=u.substr(0,u.indexOf(n)),a.length>0&&v(t).unusedInput.push(a),u=u.slice(u.indexOf(n)+n.length),d+=n.length),F[i]?(n?v(t).empty=!1:v(t).unusedTokens.push(i),Nt(i,n,t)):t._strict&&!n&&v(t).unusedTokens.push(i);v(t).charsLeftOver=c-d,u.length>0&&v(t).unusedInput.push(u),t._a[Bt]<=12&&!0===v(t).bigHour&&t._a[Bt]>0&&(v(t).bigHour=void 0),v(t).parsedDateParts=t._a.slice(0),v(t).meridiem=t._meridiem,t._a[Bt]=Vn(t._locale,t._a[Bt],t._meridiem),s=v(t).era,null!==s&&(t._a[Ht]=t._locale.erasConvertYear(s,t._a[Ht])),Fn(t),wn(t)}else Nn(t);else Cn(t)}function Vn(t,e,n){var r;return null==n?e:null!=t.meridiemHour?t.meridiemHour(e,n):null!=t.isPM?(r=t.isPM(n),r&&e<12&&(e+=12),r||12!==e||(e=0),e):e}function Gn(t){var e,n,r,i,o,a,s=!1,l=t._f.length;if(0===l)return v(t).invalidFormat=!0,void(t._d=new Date(NaN));for(i=0;i<l;i++)o=0,a=!1,e=M({},t),null!=t._useUTC&&(e._useUTC=t._useUTC),e._f=t._f[i],Wn(e),_(e)&&(a=!0),o+=v(e).charsLeftOver,o+=10*v(e).unusedTokens.length,v(e).score=o,s?o<r&&(r=o,n=e):(null==r||o<r||a)&&(r=o,n=e,a&&(s=!0));m(t,n||e)}function $n(t){if(!t._d){var e=ot(t._i),n=void 0===e.day?e.date:e.day;t._a=p([e.year,e.month,n,e.hour,e.minute,e.second,e.millisecond],(function(t){return t&&parseInt(t,10)})),Fn(t)}}function Un(t){var e=new S(wn(qn(t)));return e._nextDay&&(e.add(1,"d"),e._nextDay=void 0),e}function qn(t){var e=t._i,n=t._f;return t._locale=t._locale||bn(t._l),null===e||void 0===n&&""===e?b({nullInput:!0}):("string"===typeof e&&(t._i=e=t._locale.preparse(e)),k(e)?new S(wn(e)):(f(e)?t._d=e:s(n)?Gn(t):n?Wn(t):Xn(t),_(t)||(t._d=null),t))}function Xn(t){var e=t._i;d(e)?t._d=new Date(o.now()):f(e)?t._d=new Date(e.valueOf()):"string"===typeof e?Rn(t):s(e)?(t._a=p(e.slice(0),(function(t){return parseInt(t,10)})),Fn(t)):l(e)?$n(t):h(e)?t._d=new Date(e):o.createFromInputFallback(t)}function Zn(t,e,n,r,i){var o={};return!0!==e&&!1!==e||(r=e,e=void 0),!0!==n&&!1!==n||(r=n,n=void 0),(l(t)&&c(t)||s(t)&&0===t.length)&&(t=void 0),o._isAMomentObject=!0,o._useUTC=o._isUTC=i,o._l=n,o._i=t,o._f=e,o._strict=r,Un(o)}function Jn(t,e,n,r){return Zn(t,e,n,r,!1)}o.createFromInputFallback=T("value provided is not in a recognized RFC2822 or ISO format. moment construction falls back to js Date(), which is not reliable across all browsers and versions. Non RFC2822/ISO date formats are discouraged. Please refer to http://momentjs.com/guides/#/warnings/js-date/ for more info.",(function(t){t._d=new Date(t._i+(t._useUTC?" UTC":""))})),o.ISO_8601=function(){},o.RFC_2822=function(){};var Kn=T("moment().min is deprecated, use moment.max instead. http://momentjs.com/guides/#/warnings/min-max/",(function(){var t=Jn.apply(null,arguments);return this.isValid()&&t.isValid()?t<this?this:t:b()})),Qn=T("moment().max is deprecated, use moment.min instead. http://momentjs.com/guides/#/warnings/min-max/",(function(){var t=Jn.apply(null,arguments);return this.isValid()&&t.isValid()?t>this?this:t:b()}));function tr(t,e){var n,r;if(1===e.length&&s(e[0])&&(e=e[0]),!e.length)return Jn();for(n=e[0],r=1;r<e.length;++r)e[r].isValid()&&!e[r][t](n)||(n=e[r]);return n}function er(){var t=[].slice.call(arguments,0);return tr("isBefore",t)}function nr(){var t=[].slice.call(arguments,0);return tr("isAfter",t)}var rr=function(){return Date.now?Date.now():+new Date},ir=["year","quarter","month","week","day","hour","minute","second","millisecond"];function or(t){var e,n,r=!1,i=ir.length;for(e in t)if(u(t,e)&&(-1===Xt.call(ir,e)||null!=t[e]&&isNaN(t[e])))return!1;for(n=0;n<i;++n)if(t[ir[n]]){if(r)return!1;parseFloat(t[ir[n]])!==Yt(t[ir[n]])&&(r=!0)}return!0}function ar(){return this._isValid}function sr(){return Ir(NaN)}function lr(t){var e=ot(t),n=e.year||0,r=e.quarter||0,i=e.month||0,o=e.week||e.isoWeek||0,a=e.day||0,s=e.hour||0,l=e.minute||0,u=e.second||0,c=e.millisecond||0;this._isValid=or(e),this._milliseconds=+c+1e3*u+6e4*l+1e3*s*60*60,this._days=+a+7*o,this._months=+i+3*r+12*n,this._data={},this._locale=bn(),this._bubble()}function ur(t){return t instanceof lr}function cr(t){return t<0?-1*Math.round(-1*t):Math.round(t)}function dr(t,e,n){var r,i=Math.min(t.length,e.length),o=Math.abs(t.length-e.length),a=0;for(r=0;r<i;r++)(n&&t[r]!==e[r]||!n&&Yt(t[r])!==Yt(e[r]))&&a++;return a+o}function hr(t,e){B(t,0,0,(function(){var t=this.utcOffset(),n="+";return t<0&&(t=-t,n="-"),n+N(~~(t/60),2)+e+N(~~t%60,2)}))}hr("Z",":"),hr("ZZ",""),Dt("Z",Mt),Dt("ZZ",Mt),Pt(["Z","ZZ"],(function(t,e,n){n._useUTC=!0,n._tzm=pr(Mt,t)}));var fr=/([\+\-]|\d\d)/gi;function pr(t,e){var n,r,i,o=(e||"").match(t);return null===o?null:(n=o[o.length-1]||[],r=(n+"").match(fr)||["-",0,0],i=60*r[1]+Yt(r[2]),0===i?0:"+"===r[0]?i:-i)}function mr(t,e){var n,r;return e._isUTC?(n=e.clone(),r=(k(t)||f(t)?t.valueOf():Jn(t).valueOf())-n.valueOf(),n._d.setTime(n._d.valueOf()+r),o.updateOffset(n,!1),n):Jn(t).local()}function gr(t){return-Math.round(t._d.getTimezoneOffset())}function yr(t,e,n){var r,i=this._offset||0;if(!this.isValid())return null!=t?this:NaN;if(null!=t){if("string"===typeof t){if(t=pr(Mt,t),null===t)return this}else Math.abs(t)<16&&!n&&(t*=60);return!this._isUTC&&e&&(r=gr(this)),this._offset=t,this._isUTC=!0,null!=r&&this.add(r,"m"),i!==t&&(!e||this._changeInProgress?Pr(this,Ir(t-i,"m"),1,!1):this._changeInProgress||(this._changeInProgress=!0,o.updateOffset(this,!0),this._changeInProgress=null)),this}return this._isUTC?i:gr(this)}function vr(t,e){return null!=t?("string"!==typeof t&&(t=-t),this.utcOffset(t,e),this):-this.utcOffset()}function _r(t){return this.utcOffset(0,t)}function br(t){return this._isUTC&&(this.utcOffset(0,t),this._isUTC=!1,t&&this.subtract(gr(this),"m")),this}function xr(){if(null!=this._tzm)this.utcOffset(this._tzm,!1,!0);else if("string"===typeof this._i){var t=pr(wt,this._i);null!=t?this.utcOffset(t):this.utcOffset(0,!0)}return this}function wr(t){return!!this.isValid()&&(t=t?Jn(t).utcOffset():0,(this.utcOffset()-t)%60===0)}function Mr(){return this.utcOffset()>this.clone().month(0).utcOffset()||this.utcOffset()>this.clone().month(5).utcOffset()}function Sr(){if(!d(this._isDSTShifted))return this._isDSTShifted;var t,e={};return M(e,this),e=qn(e),e._a?(t=e._isUTC?g(e._a):Jn(e._a),this._isDSTShifted=this.isValid()&&dr(e._a,t.toArray())>0):this._isDSTShifted=!1,this._isDSTShifted}function kr(){return!!this.isValid()&&!this._isUTC}function Lr(){return!!this.isValid()&&this._isUTC}function Tr(){return!!this.isValid()&&(this._isUTC&&0===this._offset)}o.updateOffset=function(){};var Dr=/^(-|\+)?(?:(\d*)[. ])?(\d+):(\d+)(?::(\d+)(\.\d*)?)?$/,Or=/^(-|\+)?P(?:([-+]?[0-9,.]*)Y)?(?:([-+]?[0-9,.]*)M)?(?:([-+]?[0-9,.]*)W)?(?:([-+]?[0-9,.]*)D)?(?:T(?:([-+]?[0-9,.]*)H)?(?:([-+]?[0-9,.]*)M)?(?:([-+]?[0-9,.]*)S)?)?$/;function Ir(t,e){var n,r,i,o=t,a=null;return ur(t)?o={ms:t._milliseconds,d:t._days,M:t._months}:h(t)||!isNaN(+t)?(o={},e?o[e]=+t:o.milliseconds=+t):(a=Dr.exec(t))?(n="-"===a[1]?-1:1,o={y:0,d:Yt(a[Ft])*n,h:Yt(a[Bt])*n,m:Yt(a[Wt])*n,s:Yt(a[Vt])*n,ms:Yt(cr(1e3*a[Gt]))*n}):(a=Or.exec(t))?(n="-"===a[1]?-1:1,o={y:Cr(a[2],n),M:Cr(a[3],n),w:Cr(a[4],n),d:Cr(a[5],n),h:Cr(a[6],n),m:Cr(a[7],n),s:Cr(a[8],n)}):null==o?o={}:"object"===typeof o&&("from"in o||"to"in o)&&(i=Yr(Jn(o.from),Jn(o.to)),o={},o.ms=i.milliseconds,o.M=i.months),r=new lr(o),ur(t)&&u(t,"_locale")&&(r._locale=t._locale),ur(t)&&u(t,"_isValid")&&(r._isValid=t._isValid),r}function Cr(t,e){var n=t&&parseFloat(t.replace(",","."));return(isNaN(n)?0:n)*e}function Ar(t,e){var n={};return n.months=e.month()-t.month()+12*(e.year()-t.year()),t.clone().add(n.months,"M").isAfter(e)&&--n.months,n.milliseconds=+e-+t.clone().add(n.months,"M"),n}function Yr(t,e){var n;return t.isValid()&&e.isValid()?(e=mr(e,t),t.isBefore(e)?n=Ar(t,e):(n=Ar(e,t),n.milliseconds=-n.milliseconds,n.months=-n.months),n):{milliseconds:0,months:0}}function jr(t,e){return function(n,r){var i,o;return null===r||isNaN(+r)||(I(e,"moment()."+e+"(period, number) is deprecated. Please use moment()."+e+"(number, period). See http://momentjs.com/guides/#/warnings/add-inverted-param/ for more info."),o=n,n=r,r=o),i=Ir(n,r),Pr(this,i,t),this}}function Pr(t,e,n,r){var i=e._milliseconds,a=cr(e._days),s=cr(e._months);t.isValid()&&(r=null==r||r,s&&pe(t,Qt(t,"Month")+s*n),a&&te(t,"Date",Qt(t,"Date")+a*n),i&&t._d.setTime(t._d.valueOf()+i*n),r&&o.updateOffset(t,a||s))}Ir.fn=lr.prototype,Ir.invalid=sr;var Er=jr(1,"add"),Nr=jr(-1,"subtract");function Rr(t){return"string"===typeof t||t instanceof String}function Hr(t){return k(t)||f(t)||Rr(t)||h(t)||Fr(t)||zr(t)||null===t||void 0===t}function zr(t){var e,n,r=l(t)&&!c(t),i=!1,o=["years","year","y","months","month","M","days","day","d","dates","date","D","hours","hour","h","minutes","minute","m","seconds","second","s","milliseconds","millisecond","ms"],a=o.length;for(e=0;e<a;e+=1)n=o[e],i=i||u(t,n);return r&&i}function Fr(t){var e=s(t),n=!1;return e&&(n=0===t.filter((function(e){return!h(e)&&Rr(t)})).length),e&&n}function Br(t){var e,n,r=l(t)&&!c(t),i=!1,o=["sameDay","nextDay","lastDay","nextWeek","lastWeek","sameElse"];for(e=0;e<o.length;e+=1)n=o[e],i=i||u(t,n);return r&&i}function Wr(t,e){var n=t.diff(e,"days",!0);return n<-6?"sameElse":n<-1?"lastWeek":n<0?"lastDay":n<1?"sameDay":n<2?"nextDay":n<7?"nextWeek":"sameElse"}function Vr(t,e){1===arguments.length&&(arguments[0]?Hr(arguments[0])?(t=arguments[0],e=void 0):Br(arguments[0])&&(e=arguments[0],t=void 0):(t=void 0,e=void 0));var n=t||Jn(),r=mr(n,this).startOf("day"),i=o.calendarFormat(this,r)||"sameElse",a=e&&(C(e[i])?e[i].call(this,n):e[i]);return this.format(a||this.localeData().calendar(i,this,Jn(n)))}function Gr(){return new S(this)}function $r(t,e){var n=k(t)?t:Jn(t);return!(!this.isValid()||!n.isValid())&&(e=it(e)||"millisecond","millisecond"===e?this.valueOf()>n.valueOf():n.valueOf()<this.clone().startOf(e).valueOf())}function Ur(t,e){var n=k(t)?t:Jn(t);return!(!this.isValid()||!n.isValid())&&(e=it(e)||"millisecond","millisecond"===e?this.valueOf()<n.valueOf():this.clone().endOf(e).valueOf()<n.valueOf())}function qr(t,e,n,r){var i=k(t)?t:Jn(t),o=k(e)?e:Jn(e);return!!(this.isValid()&&i.isValid()&&o.isValid())&&(r=r||"()",("("===r[0]?this.isAfter(i,n):!this.isBefore(i,n))&&(")"===r[1]?this.isBefore(o,n):!this.isAfter(o,n)))}function Xr(t,e){var n,r=k(t)?t:Jn(t);return!(!this.isValid()||!r.isValid())&&(e=it(e)||"millisecond","millisecond"===e?this.valueOf()===r.valueOf():(n=r.valueOf(),this.clone().startOf(e).valueOf()<=n&&n<=this.clone().endOf(e).valueOf()))}function Zr(t,e){return this.isSame(t,e)||this.isAfter(t,e)}function Jr(t,e){return this.isSame(t,e)||this.isBefore(t,e)}function Kr(t,e,n){var r,i,o;if(!this.isValid())return NaN;if(r=mr(t,this),!r.isValid())return NaN;switch(i=6e4*(r.utcOffset()-this.utcOffset()),e=it(e),e){case"year":o=Qr(this,r)/12;break;case"month":o=Qr(this,r);break;case"quarter":o=Qr(this,r)/3;break;case"second":o=(this-r)/1e3;break;case"minute":o=(this-r)/6e4;break;case"hour":o=(this-r)/36e5;break;case"day":o=(this-r-i)/864e5;break;case"week":o=(this-r-i)/6048e5;break;default:o=this-r}return n?o:At(o)}function Qr(t,e){if(t.date()<e.date())return-Qr(e,t);var n,r,i=12*(e.year()-t.year())+(e.month()-t.month()),o=t.clone().add(i,"months");return e-o<0?(n=t.clone().add(i-1,"months"),r=(e-o)/(o-n)):(n=t.clone().add(i+1,"months"),r=(e-o)/(n-o)),-(i+r)||0}function ti(){return this.clone().locale("en").format("ddd MMM DD YYYY HH:mm:ss [GMT]ZZ")}function ei(t){if(!this.isValid())return null;var e=!0!==t,n=e?this.clone().utc():this;return n.year()<0||n.year()>9999?G(n,e?"YYYYYY-MM-DD[T]HH:mm:ss.SSS[Z]":"YYYYYY-MM-DD[T]HH:mm:ss.SSSZ"):C(Date.prototype.toISOString)?e?this.toDate().toISOString():new Date(this.valueOf()+60*this.utcOffset()*1e3).toISOString().replace("Z",G(n,"Z")):G(n,e?"YYYY-MM-DD[T]HH:mm:ss.SSS[Z]":"YYYY-MM-DD[T]HH:mm:ss.SSSZ")}function ni(){if(!this.isValid())return"moment.invalid(/* "+this._i+" */)";var t,e,n,r,i="moment",o="";return this.isLocal()||(i=0===this.utcOffset()?"moment.utc":"moment.parseZone",o="Z"),t="["+i+'("]',e=0<=this.year()&&this.year()<=9999?"YYYY":"YYYYYY",n="-MM-DD[T]HH:mm:ss.SSS",r=o+'[")]',this.format(t+e+n+r)}function ri(t){t||(t=this.isUtc()?o.defaultFormatUtc:o.defaultFormat);var e=G(this,t);return this.localeData().postformat(e)}function ii(t,e){return this.isValid()&&(k(t)&&t.isValid()||Jn(t).isValid())?Ir({to:this,from:t}).locale(this.locale()).humanize(!e):this.localeData().invalidDate()}function oi(t){return this.from(Jn(),t)}function ai(t,e){return this.isValid()&&(k(t)&&t.isValid()||Jn(t).isValid())?Ir({from:this,to:t}).locale(this.locale()).humanize(!e):this.localeData().invalidDate()}function si(t){return this.to(Jn(),t)}function li(t){var e;return void 0===t?this._locale._abbr:(e=bn(t),null!=e&&(this._locale=e),this)}o.defaultFormat="YYYY-MM-DDTHH:mm:ssZ",o.defaultFormatUtc="YYYY-MM-DDTHH:mm:ss[Z]";var ui=T("moment().lang() is deprecated. Instead, use moment().localeData() to get the language configuration. Use moment().locale() to change languages.",(function(t){return void 0===t?this.localeData():this.locale(t)}));function ci(){return this._locale}var di=1e3,hi=60*di,fi=60*hi,pi=3506328*fi;function mi(t,e){return(t%e+e)%e}function gi(t,e,n){return t<100&&t>=0?new Date(t+400,e,n)-pi:new Date(t,e,n).valueOf()}function yi(t,e,n){return t<100&&t>=0?Date.UTC(t+400,e,n)-pi:Date.UTC(t,e,n)}function vi(t){var e,n;if(t=it(t),void 0===t||"millisecond"===t||!this.isValid())return this;switch(n=this._isUTC?yi:gi,t){case"year":e=n(this.year(),0,1);break;case"quarter":e=n(this.year(),this.month()-this.month()%3,1);break;case"month":e=n(this.year(),this.month(),1);break;case"week":e=n(this.year(),this.month(),this.date()-this.weekday());break;case"isoWeek":e=n(this.year(),this.month(),this.date()-(this.isoWeekday()-1));break;case"day":case"date":e=n(this.year(),this.month(),this.date());break;case"hour":e=this._d.valueOf(),e-=mi(e+(this._isUTC?0:this.utcOffset()*hi),fi);break;case"minute":e=this._d.valueOf(),e-=mi(e,hi);break;case"second":e=this._d.valueOf(),e-=mi(e,di);break}return this._d.setTime(e),o.updateOffset(this,!0),this}function _i(t){var e,n;if(t=it(t),void 0===t||"millisecond"===t||!this.isValid())return this;switch(n=this._isUTC?yi:gi,t){case"year":e=n(this.year()+1,0,1)-1;break;case"quarter":e=n(this.year(),this.month()-this.month()%3+3,1)-1;break;case"month":e=n(this.year(),this.month()+1,1)-1;break;case"week":e=n(this.year(),this.month(),this.date()-this.weekday()+7)-1;break;case"isoWeek":e=n(this.year(),this.month(),this.date()-(this.isoWeekday()-1)+7)-1;break;case"day":case"date":e=n(this.year(),this.month(),this.date()+1)-1;break;case"hour":e=this._d.valueOf(),e+=fi-mi(e+(this._isUTC?0:this.utcOffset()*hi),fi)-1;break;case"minute":e=this._d.valueOf(),e+=hi-mi(e,hi)-1;break;case"second":e=this._d.valueOf(),e+=di-mi(e,di)-1;break}return this._d.setTime(e),o.updateOffset(this,!0),this}function bi(){return this._d.valueOf()-6e4*(this._offset||0)}function xi(){return Math.floor(this.valueOf()/1e3)}function wi(){return new Date(this.valueOf())}function Mi(){var t=this;return[t.year(),t.month(),t.date(),t.hour(),t.minute(),t.second(),t.millisecond()]}function Si(){var t=this;return{years:t.year(),months:t.month(),date:t.date(),hours:t.hours(),minutes:t.minutes(),seconds:t.seconds(),milliseconds:t.milliseconds()}}function ki(){return this.isValid()?this.toISOString():null}function Li(){return _(this)}function Ti(){return m({},v(this))}function Di(){return v(this).overflow}function Oi(){return{input:this._i,format:this._f,locale:this._locale,isUTC:this._isUTC,strict:this._strict}}function Ii(t,e){var n,r,i,a=this._eras||bn("en")._eras;for(n=0,r=a.length;n<r;++n){switch(typeof a[n].since){case"string":i=o(a[n].since).startOf("day"),a[n].since=i.valueOf();break}switch(typeof a[n].until){case"undefined":a[n].until=1/0;break;case"string":i=o(a[n].until).startOf("day").valueOf(),a[n].until=i.valueOf();break}}return a}function Ci(t,e,n){var r,i,o,a,s,l=this.eras();for(t=t.toUpperCase(),r=0,i=l.length;r<i;++r)if(o=l[r].name.toUpperCase(),a=l[r].abbr.toUpperCase(),s=l[r].narrow.toUpperCase(),n)switch(e){case"N":case"NN":case"NNN":if(a===t)return l[r];break;case"NNNN":if(o===t)return l[r];break;case"NNNNN":if(s===t)return l[r];break}else if([o,a,s].indexOf(t)>=0)return l[r]}function Ai(t,e){var n=t.since<=t.until?1:-1;return void 0===e?o(t.since).year():o(t.since).year()+(e-t.offset)*n}function Yi(){var t,e,n,r=this.localeData().eras();for(t=0,e=r.length;t<e;++t){if(n=this.clone().startOf("day").valueOf(),r[t].since<=n&&n<=r[t].until)return r[t].name;if(r[t].until<=n&&n<=r[t].since)return r[t].name}return""}function ji(){var t,e,n,r=this.localeData().eras();for(t=0,e=r.length;t<e;++t){if(n=this.clone().startOf("day").valueOf(),r[t].since<=n&&n<=r[t].until)return r[t].narrow;if(r[t].until<=n&&n<=r[t].since)return r[t].narrow}return""}function Pi(){var t,e,n,r=this.localeData().eras();for(t=0,e=r.length;t<e;++t){if(n=this.clone().startOf("day").valueOf(),r[t].since<=n&&n<=r[t].until)return r[t].abbr;if(r[t].until<=n&&n<=r[t].since)return r[t].abbr}return""}function Ei(){var t,e,n,r,i=this.localeData().eras();for(t=0,e=i.length;t<e;++t)if(n=i[t].since<=i[t].until?1:-1,r=this.clone().startOf("day").valueOf(),i[t].since<=r&&r<=i[t].until||i[t].until<=r&&r<=i[t].since)return(this.year()-o(i[t].since).year())*n+i[t].offset;return this.year()}function Ni(t){return u(this,"_erasNameRegex")||Vi.call(this),t?this._erasNameRegex:this._erasRegex}function Ri(t){return u(this,"_erasAbbrRegex")||Vi.call(this),t?this._erasAbbrRegex:this._erasRegex}function Hi(t){return u(this,"_erasNarrowRegex")||Vi.call(this),t?this._erasNarrowRegex:this._erasRegex}function zi(t,e){return e.erasAbbrRegex(t)}function Fi(t,e){return e.erasNameRegex(t)}function Bi(t,e){return e.erasNarrowRegex(t)}function Wi(t,e){return e._eraYearOrdinalRegex||bt}function Vi(){var t,e,n,r,i,o=[],a=[],s=[],l=[],u=this.eras();for(t=0,e=u.length;t<e;++t)n=Ct(u[t].name),r=Ct(u[t].abbr),i=Ct(u[t].narrow),a.push(n),o.push(r),s.push(i),l.push(n),l.push(r),l.push(i);this._erasRegex=new RegExp("^("+l.join("|")+")","i"),this._erasNameRegex=new RegExp("^("+a.join("|")+")","i"),this._erasAbbrRegex=new RegExp("^("+o.join("|")+")","i"),this._erasNarrowRegex=new RegExp("^("+s.join("|")+")","i")}function Gi(t,e){B(0,[t,t.length],0,e)}function $i(t){return Ki.call(this,t,this.week(),this.weekday()+this.localeData()._week.dow,this.localeData()._week.dow,this.localeData()._week.doy)}function Ui(t){return Ki.call(this,t,this.isoWeek(),this.isoWeekday(),1,4)}function qi(){return ke(this.year(),1,4)}function Xi(){return ke(this.isoWeekYear(),1,4)}function Zi(){var t=this.localeData()._week;return ke(this.year(),t.dow,t.doy)}function Ji(){var t=this.localeData()._week;return ke(this.weekYear(),t.dow,t.doy)}function Ki(t,e,n,r,i){var o;return null==t?Se(this,r,i).year:(o=ke(t,r,i),e>o&&(e=o),Qi.call(this,t,e,n,r,i))}function Qi(t,e,n,r,i){var o=Me(t,e,n,r,i),a=xe(o.year,0,o.dayOfYear);return this.year(a.getUTCFullYear()),this.month(a.getUTCMonth()),this.date(a.getUTCDate()),this}function to(t){return null==t?Math.ceil((this.month()+1)/3):this.month(3*(t-1)+this.month()%3)}B("N",0,0,"eraAbbr"),B("NN",0,0,"eraAbbr"),B("NNN",0,0,"eraAbbr"),B("NNNN",0,0,"eraName"),B("NNNNN",0,0,"eraNarrow"),B("y",["y",1],"yo","eraYear"),B("y",["yy",2],0,"eraYear"),B("y",["yyy",3],0,"eraYear"),B("y",["yyyy",4],0,"eraYear"),Dt("N",zi),Dt("NN",zi),Dt("NNN",zi),Dt("NNNN",Fi),Dt("NNNNN",Bi),Pt(["N","NN","NNN","NNNN","NNNNN"],(function(t,e,n,r){var i=n._locale.erasParse(t,r,n._strict);i?v(n).era=i:v(n).invalidEra=t})),Dt("y",bt),Dt("yy",bt),Dt("yyy",bt),Dt("yyyy",bt),Dt("yo",Wi),Pt(["y","yy","yyy","yyyy"],Ht),Pt(["yo"],(function(t,e,n,r){var i;n._locale._eraYearOrdinalRegex&&(i=t.match(n._locale._eraYearOrdinalRegex)),n._locale.eraYearOrdinalParse?e[Ht]=n._locale.eraYearOrdinalParse(t,i):e[Ht]=parseInt(t,10)})),B(0,["gg",2],0,(function(){return this.weekYear()%100})),B(0,["GG",2],0,(function(){return this.isoWeekYear()%100})),Gi("gggg","weekYear"),Gi("ggggg","weekYear"),Gi("GGGG","isoWeekYear"),Gi("GGGGG","isoWeekYear"),Dt("G",xt),Dt("g",xt),Dt("GG",pt,ct),Dt("gg",pt,ct),Dt("GGGG",vt,ht),Dt("gggg",vt,ht),Dt("GGGGG",_t,ft),Dt("ggggg",_t,ft),Et(["gggg","ggggg","GGGG","GGGGG"],(function(t,e,n,r){e[r.substr(0,2)]=Yt(t)})),Et(["gg","GG"],(function(t,e,n,r){e[r]=o.parseTwoDigitYear(t)})),B("Q",0,"Qo","quarter"),Dt("Q",ut),Pt("Q",(function(t,e){e[zt]=3*(Yt(t)-1)})),B("D",["DD",2],"Do","date"),Dt("D",pt,Lt),Dt("DD",pt,ct),Dt("Do",(function(t,e){return t?e._dayOfMonthOrdinalParse||e._ordinalParse:e._dayOfMonthOrdinalParseLenient})),Pt(["D","DD"],Ft),Pt("Do",(function(t,e){e[Ft]=Yt(t.match(pt)[0])}));var eo=Kt("Date",!0);function no(t){var e=Math.round((this.clone().startOf("day")-this.clone().startOf("year"))/864e5)+1;return null==t?e:this.add(t-e,"d")}B("DDD",["DDDD",3],"DDDo","dayOfYear"),Dt("DDD",yt),Dt("DDDD",dt),Pt(["DDD","DDDD"],(function(t,e,n){n._dayOfYear=Yt(t)})),B("m",["mm",2],0,"minute"),Dt("m",pt,Tt),Dt("mm",pt,ct),Pt(["m","mm"],Wt);var ro=Kt("Minutes",!1);B("s",["ss",2],0,"second"),Dt("s",pt,Tt),Dt("ss",pt,ct),Pt(["s","ss"],Vt);var io,oo,ao=Kt("Seconds",!1);for(B("S",0,0,(function(){return~~(this.millisecond()/100)})),B(0,["SS",2],0,(function(){return~~(this.millisecond()/10)})),B(0,["SSS",3],0,"millisecond"),B(0,["SSSS",4],0,(function(){return 10*this.millisecond()})),B(0,["SSSSS",5],0,(function(){return 100*this.millisecond()})),B(0,["SSSSSS",6],0,(function(){return 1e3*this.millisecond()})),B(0,["SSSSSSS",7],0,(function(){return 1e4*this.millisecond()})),B(0,["SSSSSSSS",8],0,(function(){return 1e5*this.millisecond()})),B(0,["SSSSSSSSS",9],0,(function(){return 1e6*this.millisecond()})),Dt("S",yt,ut),Dt("SS",yt,ct),Dt("SSS",yt,dt),io="SSSS";io.length<=9;io+="S")Dt(io,bt);function so(t,e){e[Gt]=Yt(1e3*("0."+t))}for(io="S";io.length<=9;io+="S")Pt(io,so);function lo(){return this._isUTC?"UTC":""}function uo(){return this._isUTC?"Coordinated Universal Time":""}oo=Kt("Milliseconds",!1),B("z",0,0,"zoneAbbr"),B("zz",0,0,"zoneName");var co=S.prototype;function ho(t){return Jn(1e3*t)}function fo(){return Jn.apply(null,arguments).parseZone()}function po(t){return t}co.add=Er,co.calendar=Vr,co.clone=Gr,co.diff=Kr,co.endOf=_i,co.format=ri,co.from=ii,co.fromNow=oi,co.to=ai,co.toNow=si,co.get=ee,co.invalidAt=Di,co.isAfter=$r,co.isBefore=Ur,co.isBetween=qr,co.isSame=Xr,co.isSameOrAfter=Zr,co.isSameOrBefore=Jr,co.isValid=Li,co.lang=ui,co.locale=li,co.localeData=ci,co.max=Qn,co.min=Kn,co.parsingFlags=Ti,co.set=ne,co.startOf=vi,co.subtract=Nr,co.toArray=Mi,co.toObject=Si,co.toDate=wi,co.toISOString=ei,co.inspect=ni,"undefined"!==typeof Symbol&&null!=Symbol.for&&(co[Symbol.for("nodejs.util.inspect.custom")]=function(){return"Moment<"+this.format()+">"}),co.toJSON=ki,co.toString=ti,co.unix=xi,co.valueOf=bi,co.creationData=Oi,co.eraName=Yi,co.eraNarrow=ji,co.eraAbbr=Pi,co.eraYear=Ei,co.year=Zt,co.isLeapYear=Jt,co.weekYear=$i,co.isoWeekYear=Ui,co.quarter=co.quarters=to,co.month=me,co.daysInMonth=ge,co.week=co.weeks=Ie,co.isoWeek=co.isoWeeks=Ce,co.weeksInYear=Zi,co.weeksInWeekYear=Ji,co.isoWeeksInYear=qi,co.isoWeeksInISOWeekYear=Xi,co.date=eo,co.day=co.days=$e,co.weekday=Ue,co.isoWeekday=qe,co.dayOfYear=no,co.hour=co.hours=an,co.minute=co.minutes=ro,co.second=co.seconds=ao,co.millisecond=co.milliseconds=oo,co.utcOffset=yr,co.utc=_r,co.local=br,co.parseZone=xr,co.hasAlignedHourOffset=wr,co.isDST=Mr,co.isLocal=kr,co.isUtcOffset=Lr,co.isUtc=Tr,co.isUTC=Tr,co.zoneAbbr=lo,co.zoneName=uo,co.dates=T("dates accessor is deprecated. Use date instead.",eo),co.months=T("months accessor is deprecated. Use month instead",me),co.years=T("years accessor is deprecated. Use year instead",Zt),co.zone=T("moment().zone is deprecated, use moment().utcOffset instead. http://momentjs.com/guides/#/warnings/zone/",vr),co.isDSTShifted=T("isDSTShifted is deprecated. See http://momentjs.com/guides/#/warnings/dst-shifted/ for more information",Sr);var mo=j.prototype;function go(t,e,n,r){var i=bn(),o=g().set(r,e);return i[n](o,t)}function yo(t,e,n){if(h(t)&&(e=t,t=void 0),t=t||"",null!=e)return go(t,e,n,"month");var r,i=[];for(r=0;r<12;r++)i[r]=go(t,r,n,"month");return i}function vo(t,e,n,r){"boolean"===typeof t?(h(e)&&(n=e,e=void 0),e=e||""):(e=t,n=e,t=!1,h(e)&&(n=e,e=void 0),e=e||"");var i,o=bn(),a=t?o._week.dow:0,s=[];if(null!=n)return go(e,(n+a)%7,r,"day");for(i=0;i<7;i++)s[i]=go(e,(i+a)%7,r,"day");return s}function _o(t,e){return yo(t,e,"months")}function bo(t,e){return yo(t,e,"monthsShort")}function xo(t,e,n){return vo(t,e,n,"weekdays")}function wo(t,e,n){return vo(t,e,n,"weekdaysShort")}function Mo(t,e,n){return vo(t,e,n,"weekdaysMin")}mo.calendar=E,mo.longDateFormat=q,mo.invalidDate=Z,mo.ordinal=Q,mo.preparse=po,mo.postformat=po,mo.relativeTime=et,mo.pastFuture=nt,mo.set=A,mo.eras=Ii,mo.erasParse=Ci,mo.erasConvertYear=Ai,mo.erasAbbrRegex=Ri,mo.erasNameRegex=Ni,mo.erasNarrowRegex=Hi,mo.months=ce,mo.monthsShort=de,mo.monthsParse=fe,mo.monthsRegex=ve,mo.monthsShortRegex=ye,mo.week=Le,mo.firstDayOfYear=Oe,mo.firstDayOfWeek=De,mo.weekdays=Fe,mo.weekdaysMin=We,mo.weekdaysShort=Be,mo.weekdaysParse=Ge,mo.weekdaysRegex=Xe,mo.weekdaysShortRegex=Ze,mo.weekdaysMinRegex=Je,mo.isPM=rn,mo.meridiem=sn,yn("en",{eras:[{since:"0001-01-01",until:1/0,offset:1,name:"Anno Domini",narrow:"AD",abbr:"AD"},{since:"0000-12-31",until:-1/0,offset:1,name:"Before Christ",narrow:"BC",abbr:"BC"}],dayOfMonthOrdinalParse:/\d{1,2}(th|st|nd|rd)/,ordinal:function(t){var e=t%10,n=1===Yt(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n}}),o.lang=T("moment.lang is deprecated. Use moment.locale instead.",yn),o.langData=T("moment.langData is deprecated. Use moment.localeData instead.",bn);var So=Math.abs;function ko(){var t=this._data;return this._milliseconds=So(this._milliseconds),this._days=So(this._days),this._months=So(this._months),t.milliseconds=So(t.milliseconds),t.seconds=So(t.seconds),t.minutes=So(t.minutes),t.hours=So(t.hours),t.months=So(t.months),t.years=So(t.years),this}function Lo(t,e,n,r){var i=Ir(e,n);return t._milliseconds+=r*i._milliseconds,t._days+=r*i._days,t._months+=r*i._months,t._bubble()}function To(t,e){return Lo(this,t,e,1)}function Do(t,e){return Lo(this,t,e,-1)}function Oo(t){return t<0?Math.floor(t):Math.ceil(t)}function Io(){var t,e,n,r,i,o=this._milliseconds,a=this._days,s=this._months,l=this._data;return o>=0&&a>=0&&s>=0||o<=0&&a<=0&&s<=0||(o+=864e5*Oo(Ao(s)+a),a=0,s=0),l.milliseconds=o%1e3,t=At(o/1e3),l.seconds=t%60,e=At(t/60),l.minutes=e%60,n=At(e/60),l.hours=n%24,a+=At(n/24),i=At(Co(a)),s+=i,a-=Oo(Ao(i)),r=At(s/12),s%=12,l.days=a,l.months=s,l.years=r,this}function Co(t){return 4800*t/146097}function Ao(t){return 146097*t/4800}function Yo(t){if(!this.isValid())return NaN;var e,n,r=this._milliseconds;if(t=it(t),"month"===t||"quarter"===t||"year"===t)switch(e=this._days+r/864e5,n=this._months+Co(e),t){case"month":return n;case"quarter":return n/3;case"year":return n/12}else switch(e=this._days+Math.round(Ao(this._months)),t){case"week":return e/7+r/6048e5;case"day":return e+r/864e5;case"hour":return 24*e+r/36e5;case"minute":return 1440*e+r/6e4;case"second":return 86400*e+r/1e3;case"millisecond":return Math.floor(864e5*e)+r;default:throw new Error("Unknown unit "+t)}}function jo(t){return function(){return this.as(t)}}var Po=jo("ms"),Eo=jo("s"),No=jo("m"),Ro=jo("h"),Ho=jo("d"),zo=jo("w"),Fo=jo("M"),Bo=jo("Q"),Wo=jo("y"),Vo=Po;function Go(){return Ir(this)}function $o(t){return t=it(t),this.isValid()?this[t+"s"]():NaN}function Uo(t){return function(){return this.isValid()?this._data[t]:NaN}}var qo=Uo("milliseconds"),Xo=Uo("seconds"),Zo=Uo("minutes"),Jo=Uo("hours"),Ko=Uo("days"),Qo=Uo("months"),ta=Uo("years");function ea(){return At(this.days()/7)}var na=Math.round,ra={ss:44,s:45,m:45,h:22,d:26,w:null,M:11};function ia(t,e,n,r,i){return i.relativeTime(e||1,!!n,t,r)}function oa(t,e,n,r){var i=Ir(t).abs(),o=na(i.as("s")),a=na(i.as("m")),s=na(i.as("h")),l=na(i.as("d")),u=na(i.as("M")),c=na(i.as("w")),d=na(i.as("y")),h=o<=n.ss&&["s",o]||o<n.s&&["ss",o]||a<=1&&["m"]||a<n.m&&["mm",a]||s<=1&&["h"]||s<n.h&&["hh",s]||l<=1&&["d"]||l<n.d&&["dd",l];return null!=n.w&&(h=h||c<=1&&["w"]||c<n.w&&["ww",c]),h=h||u<=1&&["M"]||u<n.M&&["MM",u]||d<=1&&["y"]||["yy",d],h[2]=e,h[3]=+t>0,h[4]=r,ia.apply(null,h)}function aa(t){return void 0===t?na:"function"===typeof t&&(na=t,!0)}function sa(t,e){return void 0!==ra[t]&&(void 0===e?ra[t]:(ra[t]=e,"s"===t&&(ra.ss=e-1),!0))}function la(t,e){if(!this.isValid())return this.localeData().invalidDate();var n,r,i=!1,o=ra;return"object"===typeof t&&(e=t,t=!1),"boolean"===typeof t&&(i=t),"object"===typeof e&&(o=Object.assign({},ra,e),null!=e.s&&null==e.ss&&(o.ss=e.s-1)),n=this.localeData(),r=oa(this,!i,o,n),i&&(r=n.pastFuture(+this,r)),n.postformat(r)}var ua=Math.abs;function ca(t){return(t>0)-(t<0)||+t}function da(){if(!this.isValid())return this.localeData().invalidDate();var t,e,n,r,i,o,a,s,l=ua(this._milliseconds)/1e3,u=ua(this._days),c=ua(this._months),d=this.asSeconds();return d?(t=At(l/60),e=At(t/60),l%=60,t%=60,n=At(c/12),c%=12,r=l?l.toFixed(3).replace(/\.?0+$/,""):"",i=d<0?"-":"",o=ca(this._months)!==ca(d)?"-":"",a=ca(this._days)!==ca(d)?"-":"",s=ca(this._milliseconds)!==ca(d)?"-":"",i+"P"+(n?o+n+"Y":"")+(c?o+c+"M":"")+(u?a+u+"D":"")+(e||t||l?"T":"")+(e?s+e+"H":"")+(t?s+t+"M":"")+(l?s+r+"S":"")):"P0D"}var ha=lr.prototype;return ha.isValid=ar,ha.abs=ko,ha.add=To,ha.subtract=Do,ha.as=Yo,ha.asMilliseconds=Po,ha.asSeconds=Eo,ha.asMinutes=No,ha.asHours=Ro,ha.asDays=Ho,ha.asWeeks=zo,ha.asMonths=Fo,ha.asQuarters=Bo,ha.asYears=Wo,ha.valueOf=Vo,ha._bubble=Io,ha.clone=Go,ha.get=$o,ha.milliseconds=qo,ha.seconds=Xo,ha.minutes=Zo,ha.hours=Jo,ha.days=Ko,ha.weeks=ea,ha.months=Qo,ha.years=ta,ha.humanize=la,ha.toISOString=da,ha.toString=da,ha.toJSON=da,ha.locale=li,ha.localeData=ci,ha.toIsoString=T("toIsoString() is deprecated. Please use toISOString() instead (notice the capitals)",da),ha.lang=ui,B("X",0,0,"unix"),B("x",0,0,"valueOf"),Dt("x",xt),Dt("X",St),Pt("X",(function(t,e,n){n._d=new Date(1e3*parseFloat(t))})),Pt("x",(function(t,e,n){n._d=new Date(Yt(t))})),
//! moment.js
o.version="2.30.1",a(Jn),o.fn=co,o.min=er,o.max=nr,o.now=rr,o.utc=g,o.unix=ho,o.months=_o,o.isDate=f,o.locale=yn,o.invalid=b,o.duration=Ir,o.isMoment=k,o.weekdays=xo,o.parseZone=fo,o.localeData=bn,o.isDuration=ur,o.monthsShort=bo,o.weekdaysMin=Mo,o.defineLocale=vn,o.updateLocale=_n,o.locales=xn,o.weekdaysShort=wo,o.normalizeUnits=it,o.relativeTimeRounding=aa,o.relativeTimeThreshold=sa,o.calendarFormat=Wr,o.prototype=co,o.HTML5_FMT={DATETIME_LOCAL:"YYYY-MM-DDTHH:mm",DATETIME_LOCAL_SECONDS:"YYYY-MM-DDTHH:mm:ss",DATETIME_LOCAL_MS:"YYYY-MM-DDTHH:mm:ss.SSS",DATE:"YYYY-MM-DD",TIME:"HH:mm",TIME_SECONDS:"HH:mm:ss",TIME_MS:"HH:mm:ss.SSS",WEEK:"GGGG-[W]WW",MONTH:"YYYY-MM"},o}))}).call(this,n("62e4")(t))},c20d:function(t,e,n){"use strict";var r=n("da84"),i=n("d039"),o=n("e330"),a=n("577e"),s=n("58a8").trim,l=n("5899"),u=r.parseInt,c=r.Symbol,d=c&&c.iterator,h=/^[+-]?0x/i,f=o(h.exec),p=8!==u(l+"08")||22!==u(l+"0x16")||d&&!i((function(){u(Object(d))}));t.exports=p?function(t,e){var n=s(a(t));return u(n,e>>>0||(f(h,n)?16:10))}:u},c274:function(t,e,n){"use strict";var r=n("50bf");function i(){var t={},e=0,n=0,r=0;function i(i,o){o||(o=i,i=0),i>n?n=i:i<r&&(r=i),t[i]||(t[i]=[]),t[i].push(o),e++}function o(){for(var e=r;e<=n;e++)for(var i=t[e],o=0;o<i.length;o++){var a=i[o];a()}}function a(){return e}return{add:i,process:o,size:a}}t.exports=function(t){t=t||{};var e=t.reporter,n=r.getOption(t,"async",!0),o=r.getOption(t,"auto",!0);o&&!n&&(e&&e.warn("Invalid options combination. auto=true and async=false is invalid. Setting async=true."),n=!0);var a,s=i(),l=!1;function u(t,e){!l&&o&&n&&0===s.size()&&h(),s.add(t,e)}function c(){l=!0;while(s.size()){var t=s;s=i(),t.process()}l=!1}function d(t){l||(void 0===t&&(t=n),a&&(f(a),a=null),t?h():c())}function h(){a=p(c)}function f(t){var e=clearTimeout;return e(t)}function p(t){var e=function(t){return setTimeout(t,0)};return e(t)}return{add:u,force:d}}},c2b6:function(t,e,n){var r=n("f8af"),i=n("5d89"),o=n("6f6c"),a=n("a2db"),s=n("c8fe"),l="[object Boolean]",u="[object Date]",c="[object Map]",d="[object Number]",h="[object RegExp]",f="[object Set]",p="[object String]",m="[object Symbol]",g="[object ArrayBuffer]",y="[object DataView]",v="[object Float32Array]",_="[object Float64Array]",b="[object Int8Array]",x="[object Int16Array]",w="[object Int32Array]",M="[object Uint8Array]",S="[object Uint8ClampedArray]",k="[object Uint16Array]",L="[object Uint32Array]";function T(t,e,n){var T=t.constructor;switch(e){case g:return r(t);case l:case u:return new T(+t);case y:return i(t,n);case v:case _:case b:case x:case w:case M:case S:case k:case L:return s(t,n);case c:return new T;case d:case p:return new T(t);case h:return o(t);case f:return new T;case m:return a(t)}}t.exports=T},c30c:function(t,e,n){var r=n("8eeb"),i=n("2ec1"),o=n("9934"),a=i((function(t,e,n,i){r(e,o(e),t,i)}));t.exports=a},c3f4:function(t,e,n){var r=n("5b01"),i=n("7530");function o(t,e){var n=i(t);return null==e?n:r(n,e)}t.exports=o},c3fc:function(t,e,n){var r=n("42a2"),i=n("1310"),o="[object Set]";function a(t){return i(t)&&r(t)==o}t.exports=a},c430:function(t,e,n){"use strict";t.exports=!1},c513:function(t,e,n){"use strict";var r=n("23e7"),i=n("1a2d"),o=n("d9b5"),a=n("0d51"),s=n("5692"),l=n("0b43"),u=s("symbol-to-string-registry");r({target:"Symbol",stat:!0,forced:!l},{keyFor:function(t){if(!o(t))throw new TypeError(a(t)+" is not a symbol");if(i(u,t))return u[t]}})},c584:function(t,e){function n(t,e){return t.has(e)}t.exports=n},c5c1:function(t,e,n){"use strict";var r=n("6e29"),i=n("31de"),o=n("245a");r["a"].Menu=i["a"],r["a"].Item=o["a"],e["a"]=r["a"]},c65b:function(t,e,n){"use strict";var r=n("40d5"),i=Function.prototype.call;t.exports=r?i.bind(i):function(){return i.apply(i,arguments)}},c6a7:function(t,e,n){"use strict";t.exports="function"==typeof Bun&&Bun&&"string"==typeof Bun.version},c6b6:function(t,e,n){"use strict";var r=n("e330"),i=r({}.toString),o=r("".slice);t.exports=function(t){return o(i(t),8,-1)}},c6cd:function(t,e,n){"use strict";var r=n("c430"),i=n("da84"),o=n("6374"),a="__core-js_shared__",s=t.exports=i[a]||o(a,{});(s.versions||(s.versions=[])).push({version:"3.37.1",mode:r?"pure":"global",copyright:"© 2014-2024 Denis Pushkarev (zloirock.ru)",license:"https://github.com/zloirock/core-js/blob/v3.37.1/LICENSE",source:"https://github.com/zloirock/core-js"})},c6cf:function(t,e,n){var r=n("4d8c"),i=n("2286"),o=n("c1c9");function a(t){return o(i(t,void 0,r),t+"")}t.exports=a},c6d2:function(t,e,n){"use strict";var r=n("23e7"),i=n("c65b"),o=n("c430"),a=n("5e77"),s=n("1626"),l=n("dcc3"),u=n("e163"),c=n("d2bb"),d=n("d44e"),h=n("9112"),f=n("cb2d"),p=n("b622"),m=n("3f8c"),g=n("ae93"),y=a.PROPER,v=a.CONFIGURABLE,_=g.IteratorPrototype,b=g.BUGGY_SAFARI_ITERATORS,x=p("iterator"),w="keys",M="values",S="entries",k=function(){return this};t.exports=function(t,e,n,a,p,g,L){l(n,e,a);var T,D,O,I=function(t){if(t===p&&P)return P;if(!b&&t&&t in Y)return Y[t];switch(t){case w:return function(){return new n(this,t)};case M:return function(){return new n(this,t)};case S:return function(){return new n(this,t)}}return function(){return new n(this)}},C=e+" Iterator",A=!1,Y=t.prototype,j=Y[x]||Y["@@iterator"]||p&&Y[p],P=!b&&j||I(p),E="Array"===e&&Y.entries||j;if(E&&(T=u(E.call(new t)),T!==Object.prototype&&T.next&&(o||u(T)===_||(c?c(T,_):s(T[x])||f(T,x,k)),d(T,C,!0,!0),o&&(m[C]=k))),y&&p===M&&j&&j.name!==M&&(!o&&v?h(Y,"name",M):(A=!0,P=function(){return i(j,this)})),p)if(D={values:I(M),keys:g?P:I(w),entries:I(S)},L)for(O in D)(b||A||!(O in Y))&&f(Y,O,D[O]);else r({target:e,proto:!0,forced:b||A},D);return o&&!L||Y[x]===P||f(Y,x,P,{name:p}),m[e]=P,D}},c740:function(t,e,n){"use strict";var r=n("23e7"),i=n("b727").findIndex,o=n("44d2"),a="findIndex",s=!0;a in[]&&Array(1)[a]((function(){s=!1})),r({target:"Array",proto:!0,forced:s},{findIndex:function(t){return i(this,t,arguments.length>1?arguments[1]:void 0)}}),o(a)},c780:function(t,e,n){"use strict";function r(t){return void 0===t&&(t=document.body),!0===t?document.body:t instanceof window.Node?t:document.querySelector(t)}const i={inserted(t,{value:e},n){if(t.dataset&&"true"!==t.dataset.transfer)return!1;t.className=t.className?t.className+" v-transfer-dom":"v-transfer-dom";const i=t.parentNode;if(!i)return;const o=document.createComment("");let a=!1;!1!==e&&(i.replaceChild(o,t),r(e).appendChild(t),a=!0),t.__transferDomData||(t.__transferDomData={parentNode:i,home:o,target:r(e),hasMovedOut:a})},componentUpdated(t,{value:e}){if(t.dataset&&"true"!==t.dataset.transfer)return!1;const n=t.__transferDomData;if(!n)return;const i=n.parentNode,o=n.home,a=n.hasMovedOut;!a&&e?(i.replaceChild(o,t),r(e).appendChild(t),t.__transferDomData=Object.assign({},t.__transferDomData,{hasMovedOut:!0,target:r(e)})):a&&!1===e?(i.replaceChild(t,o),t.__transferDomData=Object.assign({},t.__transferDomData,{hasMovedOut:!1,target:r(e)})):e&&r(e).appendChild(t)},unbind(t){if(t.dataset&&"true"!==t.dataset.transfer)return!1;t.className=t.className.replace("v-transfer-dom","");const e=t.__transferDomData;e&&(!0===t.__transferDomData.hasMovedOut&&t.__transferDomData.parentNode&&t.__transferDomData.parentNode.appendChild(t),t.__transferDomData=null)}};e["a"]=i},c7aa:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("he",{months:"ינואר_פברואר_מרץ_אפריל_מאי_יוני_יולי_אוגוסט_ספטמבר_אוקטובר_נובמבר_דצמבר".split("_"),monthsShort:"ינו׳_פבר׳_מרץ_אפר׳_מאי_יוני_יולי_אוג׳_ספט׳_אוק׳_נוב׳_דצמ׳".split("_"),weekdays:"ראשון_שני_שלישי_רביעי_חמישי_שישי_שבת".split("_"),weekdaysShort:"א׳_ב׳_ג׳_ד׳_ה׳_ו׳_ש׳".split("_"),weekdaysMin:"א_ב_ג_ד_ה_ו_ש".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D [ב]MMMM YYYY",LLL:"D [ב]MMMM YYYY HH:mm",LLLL:"dddd, D [ב]MMMM YYYY HH:mm",l:"D/M/YYYY",ll:"D MMM YYYY",lll:"D MMM YYYY HH:mm",llll:"ddd, D MMM YYYY HH:mm"},calendar:{sameDay:"[היום ב־]LT",nextDay:"[מחר ב־]LT",nextWeek:"dddd [בשעה] LT",lastDay:"[אתמול ב־]LT",lastWeek:"[ביום] dddd [האחרון בשעה] LT",sameElse:"L"},relativeTime:{future:"בעוד %s",past:"לפני %s",s:"מספר שניות",ss:"%d שניות",m:"דקה",mm:"%d דקות",h:"שעה",hh:function(t){return 2===t?"שעתיים":t+" שעות"},d:"יום",dd:function(t){return 2===t?"יומיים":t+" ימים"},M:"חודש",MM:function(t){return 2===t?"חודשיים":t+" חודשים"},y:"שנה",yy:function(t){return 2===t?"שנתיים":t%10===0&&10!==t?t+" שנה":t+" שנים"}},meridiemParse:/אחה"צ|לפנה"צ|אחרי הצהריים|לפני הצהריים|לפנות בוקר|בבוקר|בערב/i,isPM:function(t){return/^(אחה"צ|אחרי הצהריים|בערב)$/.test(t)},meridiem:function(t,e,n){return t<5?"לפנות בוקר":t<10?"בבוקר":t<12?n?'לפנה"צ':"לפני הצהריים":t<18?n?'אחה"צ':"אחרי הצהריים":"בערב"}});return e}))},c869:function(t,e,n){var r=n("0b07"),i=n("2b3e"),o=r(i,"Set");t.exports=o},c87c:function(t,e){var n=Object.prototype,r=n.hasOwnProperty;function i(t){var e=t.length,n=new t.constructor(e);return e&&"string"==typeof t[0]&&r.call(t,"index")&&(n.index=t.index,n.input=t.input),n}t.exports=i},c8ba:function(t,e){var n;n=function(){return this}();try{n=n||new Function("return this")()}catch(r){"object"===typeof window&&(n=window)}t.exports=n},c8f3:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("sq",{months:"Janar_Shkurt_Mars_Prill_Maj_Qershor_Korrik_Gusht_Shtator_Tetor_Nëntor_Dhjetor".split("_"),monthsShort:"Jan_Shk_Mar_Pri_Maj_Qer_Kor_Gus_Sht_Tet_Nën_Dhj".split("_"),weekdays:"E Diel_E Hënë_E Martë_E Mërkurë_E Enjte_E Premte_E Shtunë".split("_"),weekdaysShort:"Die_Hën_Mar_Mër_Enj_Pre_Sht".split("_"),weekdaysMin:"D_H_Ma_Më_E_P_Sh".split("_"),weekdaysParseExact:!0,meridiemParse:/PD|MD/,isPM:function(t){return"M"===t.charAt(0)},meridiem:function(t,e,n){return t<12?"PD":"MD"},longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Sot në] LT",nextDay:"[Nesër në] LT",nextWeek:"dddd [në] LT",lastDay:"[Dje në] LT",lastWeek:"dddd [e kaluar në] LT",sameElse:"L"},relativeTime:{future:"në %s",past:"%s më parë",s:"disa sekonda",ss:"%d sekonda",m:"një minutë",mm:"%d minuta",h:"një orë",hh:"%d orë",d:"një ditë",dd:"%d ditë",M:"një muaj",MM:"%d muaj",y:"një vit",yy:"%d vite"},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return e}))},c8fe:function(t,e,n){var r=n("f8af");function i(t,e){var n=e?r(t.buffer):t.buffer;return new t.constructor(n,t.byteOffset,t.length)}t.exports=i},c946:function(t,e,n){"use strict";var r=n("b770").forEach;t.exports=function(t){t=t||{};var e=t.reporter,n=t.batchProcessor,i=t.stateHandler.getState,o=(t.stateHandler.hasState,t.idHandler);if(!n)throw new Error("Missing required dependency: batchProcessor");if(!e)throw new Error("Missing required dependency: reporter.");var a=d(),s="erd_scroll_detection_scrollbar_style",l="erd_scroll_detection_container";function u(t){h(t,s,l)}function c(e){var n=t.important?" !important; ":"; ";return(e.join(n)+n).trim()}function d(){var t=500,e=500,n=document.createElement("div");n.style.cssText=c(["position: absolute","width: "+2*t+"px","height: "+2*e+"px","visibility: hidden","margin: 0","padding: 0"]);var r=document.createElement("div");r.style.cssText=c(["position: absolute","width: "+t+"px","height: "+e+"px","overflow: scroll","visibility: none","top: "+3*-t+"px","left: "+3*-e+"px","visibility: hidden","margin: 0","padding: 0"]),r.appendChild(n),document.body.insertBefore(r,document.body.firstChild);var i=t-r.clientWidth,o=e-r.clientHeight;return document.body.removeChild(r),{width:i,height:o}}function h(t,e,n){function r(n,r){r=r||function(e){t.head.appendChild(e)};var i=t.createElement("style");return i.innerHTML=n,i.id=e,r(i),i}if(!t.getElementById(e)){var i=n+"_animation",o=n+"_animation_active",a="/* Created by the element-resize-detector library. */\n";a+="."+n+" > div::-webkit-scrollbar { "+c(["display: none"])+" }\n\n",a+="."+o+" { "+c(["-webkit-animation-duration: 0.1s","animation-duration: 0.1s","-webkit-animation-name: "+i,"animation-name: "+i])+" }\n",a+="@-webkit-keyframes "+i+" { 0% { opacity: 1; } 50% { opacity: 0; } 100% { opacity: 1; } }\n",a+="@keyframes "+i+" { 0% { opacity: 1; } 50% { opacity: 0; } 100% { opacity: 1; } }",r(a)}}function f(t){t.className+=" "+l+"_animation_active"}function p(t,n,r){if(t.addEventListener)t.addEventListener(n,r);else{if(!t.attachEvent)return e.error("[scroll] Don't know how to add event listeners.");t.attachEvent("on"+n,r)}}function m(t,n,r){if(t.removeEventListener)t.removeEventListener(n,r);else{if(!t.detachEvent)return e.error("[scroll] Don't know how to remove event listeners.");t.detachEvent("on"+n,r)}}function g(t){return i(t).container.childNodes[0].childNodes[0].childNodes[0]}function y(t){return i(t).container.childNodes[0].childNodes[0].childNodes[1]}function v(t,e){var n=i(t).listeners;if(!n.push)throw new Error("Cannot add listener to an element that is not detectable.");i(t).listeners.push(e)}function _(t,s,u){function d(){if(t.debug){var n=Array.prototype.slice.call(arguments);if(n.unshift(o.get(s),"Scroll: "),e.log.apply)e.log.apply(null,n);else for(var r=0;r<n.length;r++)e.log(n[r])}}function h(t){function e(t){var e=t.getRootNode&&t.getRootNode().contains(t);return t===t.ownerDocument.body||t.ownerDocument.body.contains(t)||e}return!e(t)||null===window.getComputedStyle(t)}function m(t){var e=i(t).container.childNodes[0],n=window.getComputedStyle(e);return!n.width||-1===n.width.indexOf("px")}function v(){var t=window.getComputedStyle(s),e={};return e.position=t.position,e.width=s.offsetWidth,e.height=s.offsetHeight,e.top=t.top,e.right=t.right,e.bottom=t.bottom,e.left=t.left,e.widthCSS=t.width,e.heightCSS=t.height,e}function _(){var t=v();i(s).startSize={width:t.width,height:t.height},d("Element start size",i(s).startSize)}function b(){i(s).listeners=[]}function x(){if(d("storeStyle invoked."),i(s)){var t=v();i(s).style=t}else d("Aborting because element has been uninstalled")}function w(t,e,n){i(t).lastWidth=e,i(t).lastHeight=n}function M(t){return g(t).childNodes[0]}function S(){return 2*a.width+1}function k(){return 2*a.height+1}function L(t){return t+10+S()}function T(t){return t+10+k()}function D(t){return 2*t+S()}function O(t){return 2*t+k()}function I(t,e,n){var r=g(t),i=y(t),o=L(e),a=T(n),s=D(e),l=O(n);r.scrollLeft=o,r.scrollTop=a,i.scrollLeft=s,i.scrollTop=l}function C(){var t=i(s).container;if(!t){t=document.createElement("div"),t.className=l,t.style.cssText=c(["visibility: hidden","display: inline","width: 0px","height: 0px","z-index: -1","overflow: hidden","margin: 0","padding: 0"]),i(s).container=t,f(t),s.appendChild(t);var e=function(){i(s).onRendered&&i(s).onRendered()};p(t,"animationstart",e),i(s).onAnimationStart=e}return t}function A(){function n(){var n=i(s).style;if("static"===n.position){s.style.setProperty("position","relative",t.important?"important":"");var r=function(t,e,n,r){function i(t){return t.replace(/[^-\d\.]/g,"")}var o=n[r];"auto"!==o&&"0"!==i(o)&&(t.warn("An element that is positioned static has style."+r+"="+o+" which is ignored due to the static positioning. The element will need to be positioned relative, so the style."+r+" will be set to 0. Element: ",e),e.style[r]=0)};r(e,s,n,"top"),r(e,s,n,"right"),r(e,s,n,"bottom"),r(e,s,n,"left")}}function r(t,e,n,r){return t=t?t+"px":"0",e=e?e+"px":"0",n=n?n+"px":"0",r=r?r+"px":"0",["left: "+t,"top: "+e,"right: "+r,"bottom: "+n]}if(d("Injecting elements"),i(s)){n();var o=i(s).container;o||(o=C());var u=a.width,h=a.height,f=c(["position: absolute","flex: none","overflow: hidden","z-index: -1","visibility: hidden","width: 100%","height: 100%","left: 0px","top: 0px"]),m=c(["position: absolute","flex: none","overflow: hidden","z-index: -1","visibility: hidden"].concat(r(-(1+u),-(1+h),-h,-u))),g=c(["position: absolute","flex: none","overflow: scroll","z-index: -1","visibility: hidden","width: 100%","height: 100%"]),y=c(["position: absolute","flex: none","overflow: scroll","z-index: -1","visibility: hidden","width: 100%","height: 100%"]),v=c(["position: absolute","left: 0","top: 0"]),_=c(["position: absolute","width: 200%","height: 200%"]),b=document.createElement("div"),x=document.createElement("div"),w=document.createElement("div"),M=document.createElement("div"),S=document.createElement("div"),k=document.createElement("div");b.dir="ltr",b.style.cssText=f,b.className=l,x.className=l,x.style.cssText=m,w.style.cssText=g,M.style.cssText=v,S.style.cssText=y,k.style.cssText=_,w.appendChild(M),S.appendChild(k),x.appendChild(w),x.appendChild(S),b.appendChild(x),o.appendChild(b),p(w,"scroll",L),p(S,"scroll",T),i(s).onExpandScroll=L,i(s).onShrinkScroll=T}else d("Aborting because element has been uninstalled");function L(){var t=i(s);t&&t.onExpand?t.onExpand():d("Aborting expand scroll handler: element has been uninstalled")}function T(){var t=i(s);t&&t.onShrink?t.onShrink():d("Aborting shrink scroll handler: element has been uninstalled")}}function Y(){function a(e,n,r){var i=M(e),o=L(n),a=T(r);i.style.setProperty("width",o+"px",t.important?"important":""),i.style.setProperty("height",a+"px",t.important?"important":"")}function l(r){var l=s.offsetWidth,c=s.offsetHeight,h=l!==i(s).lastWidth||c!==i(s).lastHeight;d("Storing current size",l,c),w(s,l,c),n.add(0,(function(){if(h)if(i(s))if(u()){if(t.debug){var n=s.offsetWidth,r=s.offsetHeight;n===l&&r===c||e.warn(o.get(s),"Scroll: Size changed before updating detector elements.")}a(s,l,c)}else d("Aborting because element container has not been initialized");else d("Aborting because element has been uninstalled")})),n.add(1,(function(){i(s)?u()?I(s,l,c):d("Aborting because element container has not been initialized"):d("Aborting because element has been uninstalled")})),h&&r&&n.add(2,(function(){i(s)?u()?r():d("Aborting because element container has not been initialized"):d("Aborting because element has been uninstalled")}))}function u(){return!!i(s).container}function c(){function t(){return void 0===i(s).lastNotifiedWidth}d("notifyListenersIfNeeded invoked");var e=i(s);return t()&&e.lastWidth===e.startSize.width&&e.lastHeight===e.startSize.height?d("Not notifying: Size is the same as the start size, and there has been no notification yet."):e.lastWidth===e.lastNotifiedWidth&&e.lastHeight===e.lastNotifiedHeight?d("Not notifying: Size already notified"):(d("Current size not notified, notifying..."),e.lastNotifiedWidth=e.lastWidth,e.lastNotifiedHeight=e.lastHeight,void r(i(s).listeners,(function(t){t(s)})))}function h(){if(d("startanimation triggered."),m(s))d("Ignoring since element is still unrendered...");else{d("Element rendered.");var t=g(s),e=y(s);0!==t.scrollLeft&&0!==t.scrollTop&&0!==e.scrollLeft&&0!==e.scrollTop||(d("Scrollbars out of sync. Updating detector elements..."),l(c))}}function f(){d("Scroll detected."),m(s)?d("Scroll event fired while unrendered. Ignoring..."):l(c)}if(d("registerListenersAndPositionElements invoked."),i(s)){i(s).onRendered=h,i(s).onExpand=f,i(s).onShrink=f;var p=i(s).style;a(s,p.width,p.height)}else d("Aborting because element has been uninstalled")}function j(){if(d("finalizeDomMutation invoked."),i(s)){var t=i(s).style;w(s,t.width,t.height),I(s,t.width,t.height)}else d("Aborting because element has been uninstalled")}function P(){u(s)}function E(){d("Installing..."),b(),_(),n.add(0,x),n.add(1,A),n.add(2,Y),n.add(3,j),n.add(4,P)}u||(u=s,s=t,t=null),t=t||{},d("Making detectable..."),h(s)?(d("Element is detached"),C(),d("Waiting until element is attached..."),i(s).onRendered=function(){d("Element is now attached"),E()}):E()}function b(t){var e=i(t);e&&(e.onExpandScroll&&m(g(t),"scroll",e.onExpandScroll),e.onShrinkScroll&&m(y(t),"scroll",e.onShrinkScroll),e.onAnimationStart&&m(e.container,"animationstart",e.onAnimationStart),e.container&&t.removeChild(e.container))}return u(window.document),{makeDetectable:_,addListener:v,uninstall:b,initDocument:u}}},ca7c:function(t,e,n){var r=n("4be8"),i=n("1304");function o(t,e){return t&&r(t,i(e))}t.exports=o},ca84:function(t,e,n){"use strict";var r=n("e330"),i=n("1a2d"),o=n("fc6a"),a=n("4d64").indexOf,s=n("d012"),l=r([].push);t.exports=function(t,e){var n,r=o(t),u=0,c=[];for(n in r)!i(s,n)&&i(r,n)&&l(c,n);while(e.length>u)i(r,n=e[u++])&&(~a(c,n)||l(c,n));return c}},caad:function(t,e,n){"use strict";var r=n("23e7"),i=n("4d64").includes,o=n("d039"),a=n("44d2"),s=o((function(){return!Array(1).includes()}));r({target:"Array",proto:!0,forced:s},{includes:function(t){return i(this,t,arguments.length>1?arguments[1]:void 0)}}),a("includes")},cb2d:function(t,e,n){"use strict";var r=n("1626"),i=n("9bf2"),o=n("13d2"),a=n("6374");t.exports=function(t,e,n,s){s||(s={});var l=s.enumerable,u=void 0!==s.name?s.name:e;if(r(n)&&o(n,u,s),s.global)l?t[e]=n:a(e,n);else{try{s.unsafe?t[e]&&(l=!0):delete t[e]}catch(c){}l?t[e]=n:i.f(t,e,{value:n,enumerable:!1,configurable:!s.nonConfigurable,writable:!s.nonWritable})}return t}},cb5a:function(t,e,n){var r=n("9638");function i(t,e){var n=t.length;while(n--)if(r(t[n][0],e))return n;return-1}t.exports=i},cc12:function(t,e,n){"use strict";var r=n("da84"),i=n("861d"),o=r.document,a=i(o)&&i(o.createElement);t.exports=function(t){return a?o.createElement(t):{}}},cc45:function(t,e,n){var r=n("1a2d0"),i=n("b047"),o=n("99d3"),a=o&&o.isMap,s=a?i(a):r;t.exports=s},cc98:function(t,e,n){"use strict";var r=n("23e7"),i=n("c430"),o=n("4738").CONSTRUCTOR,a=n("d256"),s=n("d066"),l=n("1626"),u=n("cb2d"),c=a&&a.prototype;if(r({target:"Promise",proto:!0,forced:o,real:!0},{catch:function(t){return this.then(void 0,t)}}),!i&&l(a)){var d=s("Promise").prototype["catch"];c["catch"]!==d&&u(c,"catch",d,{unsafe:!0})}},cca6:function(t,e,n){"use strict";var r=n("23e7"),i=n("60da");r({target:"Object",stat:!0,arity:2,forced:Object.assign!==i},{assign:i})},cd9d:function(t,e){function n(t){return t}t.exports=n},cdce:function(t,e,n){"use strict";var r=n("da84"),i=n("1626"),o=r.WeakMap;t.exports=i(o)&&/native code/.test(String(o))},cdd8:function(t,e,n){t.exports=n("a9b9")},cdf9:function(t,e,n){"use strict";var r=n("825a"),i=n("861d"),o=n("f069");t.exports=function(t,e){if(r(t),i(e)&&e.constructor===t)return e;var n=o.f(t),a=n.resolve;return a(e),n.promise}},ce86:function(t,e,n){var r=n("9e69"),i=n("7948"),o=n("6747"),a=n("ffd6"),s=1/0,l=r?r.prototype:void 0,u=l?l.toString:void 0;function c(t){if("string"==typeof t)return t;if(o(t))return i(t,c)+"";if(a(t))return u?u.call(t):"";var e=t+"";return"0"==e&&1/t==-s?"-0":e}t.exports=c},cebd:function(t,e){function n(t){var e=-1,n=Array(t.size);return t.forEach((function(t){n[++e]=[t,t]})),n}t.exports=n},cf1e:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={words:{ss:["sekunda","sekunde","sekundi"],m:["jedan minut","jednog minuta"],mm:["minut","minuta","minuta"],h:["jedan sat","jednog sata"],hh:["sat","sata","sati"],d:["jedan dan","jednog dana"],dd:["dan","dana","dana"],M:["jedan mesec","jednog meseca"],MM:["mesec","meseca","meseci"],y:["jednu godinu","jedne godine"],yy:["godinu","godine","godina"]},correctGrammaticalCase:function(t,e){return t%10>=1&&t%10<=4&&(t%100<10||t%100>=20)?t%10===1?e[0]:e[1]:e[2]},translate:function(t,n,r,i){var o,a=e.words[r];return 1===r.length?"y"===r&&n?"jedna godina":i||n?a[0]:a[1]:(o=e.correctGrammaticalCase(t,a),"yy"===r&&n&&"godinu"===o?t+" godina":t+" "+o)}},n=t.defineLocale("sr",{months:"januar_februar_mart_april_maj_jun_jul_avgust_septembar_oktobar_novembar_decembar".split("_"),monthsShort:"jan._feb._mar._apr._maj_jun_jul_avg._sep._okt._nov._dec.".split("_"),monthsParseExact:!0,weekdays:"nedelja_ponedeljak_utorak_sreda_četvrtak_petak_subota".split("_"),weekdaysShort:"ned._pon._uto._sre._čet._pet._sub.".split("_"),weekdaysMin:"ne_po_ut_sr_če_pe_su".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"D. M. YYYY.",LL:"D. MMMM YYYY.",LLL:"D. MMMM YYYY. H:mm",LLLL:"dddd, D. MMMM YYYY. H:mm"},calendar:{sameDay:"[danas u] LT",nextDay:"[sutra u] LT",nextWeek:function(){switch(this.day()){case 0:return"[u] [nedelju] [u] LT";case 3:return"[u] [sredu] [u] LT";case 6:return"[u] [subotu] [u] LT";case 1:case 2:case 4:case 5:return"[u] dddd [u] LT"}},lastDay:"[juče u] LT",lastWeek:function(){var t=["[prošle] [nedelje] [u] LT","[prošlog] [ponedeljka] [u] LT","[prošlog] [utorka] [u] LT","[prošle] [srede] [u] LT","[prošlog] [četvrtka] [u] LT","[prošlog] [petka] [u] LT","[prošle] [subote] [u] LT"];return t[this.day()]},sameElse:"L"},relativeTime:{future:"za %s",past:"pre %s",s:"nekoliko sekundi",ss:e.translate,m:e.translate,mm:e.translate,h:e.translate,hh:e.translate,d:e.translate,dd:e.translate,M:e.translate,MM:e.translate,y:e.translate,yy:e.translate},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:7}});return n}))},cf51:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("tzl",{months:"Januar_Fevraglh_Març_Avrïu_Mai_Gün_Julia_Guscht_Setemvar_Listopäts_Noemvar_Zecemvar".split("_"),monthsShort:"Jan_Fev_Mar_Avr_Mai_Gün_Jul_Gus_Set_Lis_Noe_Zec".split("_"),weekdays:"Súladi_Lúneçi_Maitzi_Márcuri_Xhúadi_Viénerçi_Sáturi".split("_"),weekdaysShort:"Súl_Lún_Mai_Már_Xhú_Vié_Sát".split("_"),weekdaysMin:"Sú_Lú_Ma_Má_Xh_Vi_Sá".split("_"),longDateFormat:{LT:"HH.mm",LTS:"HH.mm.ss",L:"DD.MM.YYYY",LL:"D. MMMM [dallas] YYYY",LLL:"D. MMMM [dallas] YYYY HH.mm",LLLL:"dddd, [li] D. MMMM [dallas] YYYY HH.mm"},meridiemParse:/d\'o|d\'a/i,isPM:function(t){return"d'o"===t.toLowerCase()},meridiem:function(t,e,n){return t>11?n?"d'o":"D'O":n?"d'a":"D'A"},calendar:{sameDay:"[oxhi à] LT",nextDay:"[demà à] LT",nextWeek:"dddd [à] LT",lastDay:"[ieiri à] LT",lastWeek:"[sür el] dddd [lasteu à] LT",sameElse:"L"},relativeTime:{future:"osprei %s",past:"ja%s",s:n,ss:n,m:n,mm:n,h:n,hh:n,d:n,dd:n,M:n,MM:n,y:n,yy:n},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});function n(t,e,n,r){var i={s:["viensas secunds","'iensas secunds"],ss:[t+" secunds",t+" secunds"],m:["'n míut","'iens míut"],mm:[t+" míuts",t+" míuts"],h:["'n þora","'iensa þora"],hh:[t+" þoras",t+" þoras"],d:["'n ziua","'iensa ziua"],dd:[t+" ziuas",t+" ziuas"],M:["'n mes","'iens mes"],MM:[t+" mesen",t+" mesen"],y:["'n ar","'iens ar"],yy:[t+" ars",t+" ars"]};return r||e?i[n][0]:i[n][1]}return e}))},cf75:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="pagh_wa’_cha’_wej_loS_vagh_jav_Soch_chorgh_Hut".split("_");function n(t){var e=t;return e=-1!==t.indexOf("jaj")?e.slice(0,-3)+"leS":-1!==t.indexOf("jar")?e.slice(0,-3)+"waQ":-1!==t.indexOf("DIS")?e.slice(0,-3)+"nem":e+" pIq",e}function r(t){var e=t;return e=-1!==t.indexOf("jaj")?e.slice(0,-3)+"Hu’":-1!==t.indexOf("jar")?e.slice(0,-3)+"wen":-1!==t.indexOf("DIS")?e.slice(0,-3)+"ben":e+" ret",e}function i(t,e,n,r){var i=o(t);switch(n){case"ss":return i+" lup";case"mm":return i+" tup";case"hh":return i+" rep";case"dd":return i+" jaj";case"MM":return i+" jar";case"yy":return i+" DIS"}}function o(t){var n=Math.floor(t%1e3/100),r=Math.floor(t%100/10),i=t%10,o="";return n>0&&(o+=e[n]+"vatlh"),r>0&&(o+=(""!==o?" ":"")+e[r]+"maH"),i>0&&(o+=(""!==o?" ":"")+e[i]),""===o?"pagh":o}var a=t.defineLocale("tlh",{months:"tera’ jar wa’_tera’ jar cha’_tera’ jar wej_tera’ jar loS_tera’ jar vagh_tera’ jar jav_tera’ jar Soch_tera’ jar chorgh_tera’ jar Hut_tera’ jar wa’maH_tera’ jar wa’maH wa’_tera’ jar wa’maH cha’".split("_"),monthsShort:"jar wa’_jar cha’_jar wej_jar loS_jar vagh_jar jav_jar Soch_jar chorgh_jar Hut_jar wa’maH_jar wa’maH wa’_jar wa’maH cha’".split("_"),monthsParseExact:!0,weekdays:"lojmItjaj_DaSjaj_povjaj_ghItlhjaj_loghjaj_buqjaj_ghInjaj".split("_"),weekdaysShort:"lojmItjaj_DaSjaj_povjaj_ghItlhjaj_loghjaj_buqjaj_ghInjaj".split("_"),weekdaysMin:"lojmItjaj_DaSjaj_povjaj_ghItlhjaj_loghjaj_buqjaj_ghInjaj".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[DaHjaj] LT",nextDay:"[wa’leS] LT",nextWeek:"LLL",lastDay:"[wa’Hu’] LT",lastWeek:"LLL",sameElse:"L"},relativeTime:{future:n,past:r,s:"puS lup",ss:i,m:"wa’ tup",mm:i,h:"wa’ rep",hh:i,d:"wa’ jaj",dd:i,M:"wa’ jar",MM:i,y:"wa’ DIS",yy:i},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return a}))},d012:function(t,e,n){"use strict";t.exports={}},d02c:function(t,e,n){var r=n("5e2e"),i=n("79bc"),o=n("7b83"),a=200;function s(t,e){var n=this.__data__;if(n instanceof r){var s=n.__data__;if(!i||s.length<a-1)return s.push([t,e]),this.size=++n.size,this;n=this.__data__=new o(s)}return n.set(t,e),this.size=n.size,this}t.exports=s},d039:function(t,e,n){"use strict";t.exports=function(t){try{return!!t()}catch(e){return!0}}},d066:function(t,e,n){"use strict";var r=n("da84"),i=n("1626"),o=function(t){return i(t)?t:void 0};t.exports=function(t,e){return arguments.length<2?o(r[t]):r[t]&&r[t][e]}},d1e7:function(t,e,n){"use strict";var r={}.propertyIsEnumerable,i=Object.getOwnPropertyDescriptor,o=i&&!r.call({1:2},1);e.f=o?function(t){var e=i(this,t);return!!e&&e.enumerable}:r},d256:function(t,e,n){"use strict";var r=n("da84");t.exports=r.Promise},d26a:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"༡",2:"༢",3:"༣",4:"༤",5:"༥",6:"༦",7:"༧",8:"༨",9:"༩",0:"༠"},n={"༡":"1","༢":"2","༣":"3","༤":"4","༥":"5","༦":"6","༧":"7","༨":"8","༩":"9","༠":"0"},r=t.defineLocale("bo",{months:"ཟླ་བ་དང་པོ_ཟླ་བ་གཉིས་པ_ཟླ་བ་གསུམ་པ_ཟླ་བ་བཞི་པ_ཟླ་བ་ལྔ་པ_ཟླ་བ་དྲུག་པ_ཟླ་བ་བདུན་པ_ཟླ་བ་བརྒྱད་པ_ཟླ་བ་དགུ་པ_ཟླ་བ་བཅུ་པ_ཟླ་བ་བཅུ་གཅིག་པ_ཟླ་བ་བཅུ་གཉིས་པ".split("_"),monthsShort:"ཟླ་1_ཟླ་2_ཟླ་3_ཟླ་4_ཟླ་5_ཟླ་6_ཟླ་7_ཟླ་8_ཟླ་9_ཟླ་10_ཟླ་11_ཟླ་12".split("_"),monthsShortRegex:/^(ཟླ་\d{1,2})/,monthsParseExact:!0,weekdays:"གཟའ་ཉི་མ་_གཟའ་ཟླ་བ་_གཟའ་མིག་དམར་_གཟའ་ལྷག་པ་_གཟའ་ཕུར་བུ_གཟའ་པ་སངས་_གཟའ་སྤེན་པ་".split("_"),weekdaysShort:"ཉི་མ་_ཟླ་བ་_མིག་དམར་_ལྷག་པ་_ཕུར་བུ_པ་སངས་_སྤེན་པ་".split("_"),weekdaysMin:"ཉི_ཟླ_མིག_ལྷག_ཕུར_སངས_སྤེན".split("_"),longDateFormat:{LT:"A h:mm",LTS:"A h:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm",LLLL:"dddd, D MMMM YYYY, A h:mm"},calendar:{sameDay:"[དི་རིང] LT",nextDay:"[སང་ཉིན] LT",nextWeek:"[བདུན་ཕྲག་རྗེས་མ], LT",lastDay:"[ཁ་སང] LT",lastWeek:"[བདུན་ཕྲག་མཐའ་མ] dddd, LT",sameElse:"L"},relativeTime:{future:"%s ལ་",past:"%s སྔན་ལ",s:"ལམ་སང",ss:"%d སྐར་ཆ།",m:"སྐར་མ་གཅིག",mm:"%d སྐར་མ",h:"ཆུ་ཚོད་གཅིག",hh:"%d ཆུ་ཚོད",d:"ཉིན་གཅིག",dd:"%d ཉིན་",M:"ཟླ་བ་གཅིག",MM:"%d ཟླ་བ",y:"ལོ་གཅིག",yy:"%d ལོ"},preparse:function(t){return t.replace(/[༡༢༣༤༥༦༧༨༩༠]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/མཚན་མོ|ཞོགས་ཀས|ཉིན་གུང|དགོང་དག|མཚན་མོ/,meridiemHour:function(t,e){return 12===t&&(t=0),"མཚན་མོ"===e&&t>=4||"ཉིན་གུང"===e&&t<5||"དགོང་དག"===e?t+12:t},meridiem:function(t,e,n){return t<4?"མཚན་མོ":t<10?"ཞོགས་ཀས":t<17?"ཉིན་གུང":t<20?"དགོང་དག":"མཚན་མོ"},week:{dow:0,doy:6}});return r}))},d28b:function(t,e,n){"use strict";var r=n("e065");r("iterator")},d2bb:function(t,e,n){"use strict";var r=n("7282"),i=n("861d"),o=n("1d80"),a=n("3bbe");t.exports=Object.setPrototypeOf||("__proto__"in{}?function(){var t,e=!1,n={};try{t=r(Object.prototype,"__proto__","set"),t(n,[]),e=n instanceof Array}catch(s){}return function(n,r){return o(n),a(r),i(n)?(e?t(n,r):n.__proto__=r,n):n}}():void 0)},d2d4:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("pt-br",{months:"janeiro_fevereiro_março_abril_maio_junho_julho_agosto_setembro_outubro_novembro_dezembro".split("_"),monthsShort:"jan_fev_mar_abr_mai_jun_jul_ago_set_out_nov_dez".split("_"),weekdays:"domingo_segunda-feira_terça-feira_quarta-feira_quinta-feira_sexta-feira_sábado".split("_"),weekdaysShort:"dom_seg_ter_qua_qui_sex_sáb".split("_"),weekdaysMin:"do_2ª_3ª_4ª_5ª_6ª_sá".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D [de] MMMM [de] YYYY",LLL:"D [de] MMMM [de] YYYY [às] HH:mm",LLLL:"dddd, D [de] MMMM [de] YYYY [às] HH:mm"},calendar:{sameDay:"[Hoje às] LT",nextDay:"[Amanhã às] LT",nextWeek:"dddd [às] LT",lastDay:"[Ontem às] LT",lastWeek:function(){return 0===this.day()||6===this.day()?"[Último] dddd [às] LT":"[Última] dddd [às] LT"},sameElse:"L"},relativeTime:{future:"em %s",past:"há %s",s:"poucos segundos",ss:"%d segundos",m:"um minuto",mm:"%d minutos",h:"uma hora",hh:"%d horas",d:"um dia",dd:"%d dias",M:"um mês",MM:"%d meses",y:"um ano",yy:"%d anos"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",invalidDate:"Data inválida"});return e}))},d2e4:function(t,e,n){(function(e){
/**!
 * @fileOverview Kickass library to create and place poppers near their reference elements.
 * @version 1.16.1
 * @license
 * Copyright (c) 2016 Federico Zivolo and contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
(function(e,n){t.exports=n()})(0,(function(){"use strict";var t="undefined"!==typeof window&&"undefined"!==typeof document&&"undefined"!==typeof navigator,n=function(){for(var e=["Edge","Trident","Firefox"],n=0;n<e.length;n+=1)if(t&&navigator.userAgent.indexOf(e[n])>=0)return 1;return 0}();function r(t){var e=!1;return function(){e||(e=!0,window.Promise.resolve().then((function(){e=!1,t()})))}}function i(t){var e=!1;return function(){e||(e=!0,setTimeout((function(){e=!1,t()}),n))}}var o=t&&window.Promise,a=o?r:i;function s(t){var e={};return t&&"[object Function]"===e.toString.call(t)}function l(t,e){if(1!==t.nodeType)return[];var n=t.ownerDocument.defaultView,r=n.getComputedStyle(t,null);return e?r[e]:r}function u(t){return"HTML"===t.nodeName?t:t.parentNode||t.host}function c(t){if(!t)return document.body;switch(t.nodeName){case"HTML":case"BODY":return t.ownerDocument.body;case"#document":return t.body}var e=l(t),n=e.overflow,r=e.overflowX,i=e.overflowY;return/(auto|scroll|overlay)/.test(n+i+r)?t:c(u(t))}function d(t){return t&&t.referenceNode?t.referenceNode:t}var h=t&&!(!window.MSInputMethodContext||!document.documentMode),f=t&&/MSIE 10/.test(navigator.userAgent);function p(t){return 11===t?h:10===t?f:h||f}function m(t){if(!t)return document.documentElement;var e=p(10)?document.body:null,n=t.offsetParent||null;while(n===e&&t.nextElementSibling)n=(t=t.nextElementSibling).offsetParent;var r=n&&n.nodeName;return r&&"BODY"!==r&&"HTML"!==r?-1!==["TH","TD","TABLE"].indexOf(n.nodeName)&&"static"===l(n,"position")?m(n):n:t?t.ownerDocument.documentElement:document.documentElement}function g(t){var e=t.nodeName;return"BODY"!==e&&("HTML"===e||m(t.firstElementChild)===t)}function y(t){return null!==t.parentNode?y(t.parentNode):t}function v(t,e){if(!t||!t.nodeType||!e||!e.nodeType)return document.documentElement;var n=t.compareDocumentPosition(e)&Node.DOCUMENT_POSITION_FOLLOWING,r=n?t:e,i=n?e:t,o=document.createRange();o.setStart(r,0),o.setEnd(i,0);var a=o.commonAncestorContainer;if(t!==a&&e!==a||r.contains(i))return g(a)?a:m(a);var s=y(t);return s.host?v(s.host,e):v(t,y(e).host)}function _(t){var e=arguments.length>1&&void 0!==arguments[1]?arguments[1]:"top",n="top"===e?"scrollTop":"scrollLeft",r=t.nodeName;if("BODY"===r||"HTML"===r){var i=t.ownerDocument.documentElement,o=t.ownerDocument.scrollingElement||i;return o[n]}return t[n]}function b(t,e){var n=arguments.length>2&&void 0!==arguments[2]&&arguments[2],r=_(e,"top"),i=_(e,"left"),o=n?-1:1;return t.top+=r*o,t.bottom+=r*o,t.left+=i*o,t.right+=i*o,t}function x(t,e){var n="x"===e?"Left":"Top",r="Left"===n?"Right":"Bottom";return parseFloat(t["border"+n+"Width"])+parseFloat(t["border"+r+"Width"])}function w(t,e,n,r){return Math.max(e["offset"+t],e["scroll"+t],n["client"+t],n["offset"+t],n["scroll"+t],p(10)?parseInt(n["offset"+t])+parseInt(r["margin"+("Height"===t?"Top":"Left")])+parseInt(r["margin"+("Height"===t?"Bottom":"Right")]):0)}function M(t){var e=t.body,n=t.documentElement,r=p(10)&&getComputedStyle(n);return{height:w("Height",e,n,r),width:w("Width",e,n,r)}}var S=function(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")},k=function(){function t(t,e){for(var n=0;n<e.length;n++){var r=e[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(t,r.key,r)}}return function(e,n,r){return n&&t(e.prototype,n),r&&t(e,r),e}}(),L=function(t,e,n){return e in t?Object.defineProperty(t,e,{value:n,enumerable:!0,configurable:!0,writable:!0}):t[e]=n,t},T=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var n=arguments[e];for(var r in n)Object.prototype.hasOwnProperty.call(n,r)&&(t[r]=n[r])}return t};function D(t){return T({},t,{right:t.left+t.width,bottom:t.top+t.height})}function O(t){var e={};try{if(p(10)){e=t.getBoundingClientRect();var n=_(t,"top"),r=_(t,"left");e.top+=n,e.left+=r,e.bottom+=n,e.right+=r}else e=t.getBoundingClientRect()}catch(h){}var i={left:e.left,top:e.top,width:e.right-e.left,height:e.bottom-e.top},o="HTML"===t.nodeName?M(t.ownerDocument):{},a=o.width||t.clientWidth||i.width,s=o.height||t.clientHeight||i.height,u=t.offsetWidth-a,c=t.offsetHeight-s;if(u||c){var d=l(t);u-=x(d,"x"),c-=x(d,"y"),i.width-=u,i.height-=c}return D(i)}function I(t,e){var n=arguments.length>2&&void 0!==arguments[2]&&arguments[2],r=p(10),i="HTML"===e.nodeName,o=O(t),a=O(e),s=c(t),u=l(e),d=parseFloat(u.borderTopWidth),h=parseFloat(u.borderLeftWidth);n&&i&&(a.top=Math.max(a.top,0),a.left=Math.max(a.left,0));var f=D({top:o.top-a.top-d,left:o.left-a.left-h,width:o.width,height:o.height});if(f.marginTop=0,f.marginLeft=0,!r&&i){var m=parseFloat(u.marginTop),g=parseFloat(u.marginLeft);f.top-=d-m,f.bottom-=d-m,f.left-=h-g,f.right-=h-g,f.marginTop=m,f.marginLeft=g}return(r&&!n?e.contains(s):e===s&&"BODY"!==s.nodeName)&&(f=b(f,e)),f}function C(t){var e=arguments.length>1&&void 0!==arguments[1]&&arguments[1],n=t.ownerDocument.documentElement,r=I(t,n),i=Math.max(n.clientWidth,window.innerWidth||0),o=Math.max(n.clientHeight,window.innerHeight||0),a=e?0:_(n),s=e?0:_(n,"left"),l={top:a-r.top+r.marginTop,left:s-r.left+r.marginLeft,width:i,height:o};return D(l)}function A(t){var e=t.nodeName;if("BODY"===e||"HTML"===e)return!1;if("fixed"===l(t,"position"))return!0;var n=u(t);return!!n&&A(n)}function Y(t){if(!t||!t.parentElement||p())return document.documentElement;var e=t.parentElement;while(e&&"none"===l(e,"transform"))e=e.parentElement;return e||document.documentElement}function j(t,e,n,r){var i=arguments.length>4&&void 0!==arguments[4]&&arguments[4],o={top:0,left:0},a=i?Y(t):v(t,d(e));if("viewport"===r)o=C(a,i);else{var s=void 0;"scrollParent"===r?(s=c(u(e)),"BODY"===s.nodeName&&(s=t.ownerDocument.documentElement)):s="window"===r?t.ownerDocument.documentElement:r;var l=I(s,a,i);if("HTML"!==s.nodeName||A(a))o=l;else{var h=M(t.ownerDocument),f=h.height,p=h.width;o.top+=l.top-l.marginTop,o.bottom=f+l.top,o.left+=l.left-l.marginLeft,o.right=p+l.left}}n=n||0;var m="number"===typeof n;return o.left+=m?n:n.left||0,o.top+=m?n:n.top||0,o.right-=m?n:n.right||0,o.bottom-=m?n:n.bottom||0,o}function P(t){var e=t.width,n=t.height;return e*n}function E(t,e,n,r,i){var o=arguments.length>5&&void 0!==arguments[5]?arguments[5]:0;if(-1===t.indexOf("auto"))return t;var a=j(n,r,o,i),s={top:{width:a.width,height:e.top-a.top},right:{width:a.right-e.right,height:a.height},bottom:{width:a.width,height:a.bottom-e.bottom},left:{width:e.left-a.left,height:a.height}},l=Object.keys(s).map((function(t){return T({key:t},s[t],{area:P(s[t])})})).sort((function(t,e){return e.area-t.area})),u=l.filter((function(t){var e=t.width,r=t.height;return e>=n.clientWidth&&r>=n.clientHeight})),c=u.length>0?u[0].key:l[0].key,d=t.split("-")[1];return c+(d?"-"+d:"")}function N(t,e,n){var r=arguments.length>3&&void 0!==arguments[3]?arguments[3]:null,i=r?Y(e):v(e,d(n));return I(n,i,r)}function R(t){var e=t.ownerDocument.defaultView,n=e.getComputedStyle(t),r=parseFloat(n.marginTop||0)+parseFloat(n.marginBottom||0),i=parseFloat(n.marginLeft||0)+parseFloat(n.marginRight||0),o={width:t.offsetWidth+i,height:t.offsetHeight+r};return o}function H(t){var e={left:"right",right:"left",bottom:"top",top:"bottom"};return t.replace(/left|right|bottom|top/g,(function(t){return e[t]}))}function z(t,e,n){n=n.split("-")[0];var r=R(t),i={width:r.width,height:r.height},o=-1!==["right","left"].indexOf(n),a=o?"top":"left",s=o?"left":"top",l=o?"height":"width",u=o?"width":"height";return i[a]=e[a]+e[l]/2-r[l]/2,i[s]=n===s?e[s]-r[u]:e[H(s)],i}function F(t,e){return Array.prototype.find?t.find(e):t.filter(e)[0]}function B(t,e,n){if(Array.prototype.findIndex)return t.findIndex((function(t){return t[e]===n}));var r=F(t,(function(t){return t[e]===n}));return t.indexOf(r)}function W(t,e,n){var r=void 0===n?t:t.slice(0,B(t,"name",n));return r.forEach((function(t){t["function"];var n=t["function"]||t.fn;t.enabled&&s(n)&&(e.offsets.popper=D(e.offsets.popper),e.offsets.reference=D(e.offsets.reference),e=n(e,t))})),e}function V(){if(!this.state.isDestroyed){var t={instance:this,styles:{},arrowStyles:{},attributes:{},flipped:!1,offsets:{}};t.offsets.reference=N(this.state,this.popper,this.reference,this.options.positionFixed),t.placement=E(this.options.placement,t.offsets.reference,this.popper,this.reference,this.options.modifiers.flip.boundariesElement,this.options.modifiers.flip.padding),t.originalPlacement=t.placement,t.positionFixed=this.options.positionFixed,t.offsets.popper=z(this.popper,t.offsets.reference,t.placement),t.offsets.popper.position=this.options.positionFixed?"fixed":"absolute",t=W(this.modifiers,t),this.state.isCreated?this.options.onUpdate(t):(this.state.isCreated=!0,this.options.onCreate(t))}}function G(t,e){return t.some((function(t){var n=t.name,r=t.enabled;return r&&n===e}))}function $(t){for(var e=[!1,"ms","Webkit","Moz","O"],n=t.charAt(0).toUpperCase()+t.slice(1),r=0;r<e.length;r++){var i=e[r],o=i?""+i+n:t;if("undefined"!==typeof document.body.style[o])return o}return null}function U(){return this.state.isDestroyed=!0,G(this.modifiers,"applyStyle")&&(this.popper.removeAttribute("x-placement"),this.popper.style.position="",this.popper.style.top="",this.popper.style.left="",this.popper.style.right="",this.popper.style.bottom="",this.popper.style.willChange="",this.popper.style[$("transform")]=""),this.disableEventListeners(),this.options.removeOnDestroy&&this.popper.parentNode.removeChild(this.popper),this}function q(t){var e=t.ownerDocument;return e?e.defaultView:window}function X(t,e,n,r){var i="BODY"===t.nodeName,o=i?t.ownerDocument.defaultView:t;o.addEventListener(e,n,{passive:!0}),i||X(c(o.parentNode),e,n,r),r.push(o)}function Z(t,e,n,r){n.updateBound=r,q(t).addEventListener("resize",n.updateBound,{passive:!0});var i=c(t);return X(i,"scroll",n.updateBound,n.scrollParents),n.scrollElement=i,n.eventsEnabled=!0,n}function J(){this.state.eventsEnabled||(this.state=Z(this.reference,this.options,this.state,this.scheduleUpdate))}function K(t,e){return q(t).removeEventListener("resize",e.updateBound),e.scrollParents.forEach((function(t){t.removeEventListener("scroll",e.updateBound)})),e.updateBound=null,e.scrollParents=[],e.scrollElement=null,e.eventsEnabled=!1,e}function Q(){this.state.eventsEnabled&&(cancelAnimationFrame(this.scheduleUpdate),this.state=K(this.reference,this.state))}function tt(t){return""!==t&&!isNaN(parseFloat(t))&&isFinite(t)}function et(t,e){Object.keys(e).forEach((function(n){var r="";-1!==["width","height","top","right","bottom","left"].indexOf(n)&&tt(e[n])&&(r="px"),t.style[n]=e[n]+r}))}function nt(t,e){Object.keys(e).forEach((function(n){var r=e[n];!1!==r?t.setAttribute(n,e[n]):t.removeAttribute(n)}))}function rt(t){return et(t.instance.popper,t.styles),nt(t.instance.popper,t.attributes),t.arrowElement&&Object.keys(t.arrowStyles).length&&et(t.arrowElement,t.arrowStyles),t}function it(t,e,n,r,i){var o=N(i,e,t,n.positionFixed),a=E(n.placement,o,e,t,n.modifiers.flip.boundariesElement,n.modifiers.flip.padding);return e.setAttribute("x-placement",a),et(e,{position:n.positionFixed?"fixed":"absolute"}),n}function ot(t,e){var n=t.offsets,r=n.popper,i=n.reference,o=Math.round,a=Math.floor,s=function(t){return t},l=o(i.width),u=o(r.width),c=-1!==["left","right"].indexOf(t.placement),d=-1!==t.placement.indexOf("-"),h=l%2===u%2,f=l%2===1&&u%2===1,p=e?c||d||h?o:a:s,m=e?o:s;return{left:p(f&&!d&&e?r.left-1:r.left),top:m(r.top),bottom:m(r.bottom),right:p(r.right)}}var at=t&&/Firefox/i.test(navigator.userAgent);function st(t,e){var n=e.x,r=e.y,i=t.offsets.popper,o=F(t.instance.modifiers,(function(t){return"applyStyle"===t.name})).gpuAcceleration,a=void 0!==o?o:e.gpuAcceleration,s=m(t.instance.popper),l=O(s),u={position:i.position},c=ot(t,window.devicePixelRatio<2||!at),d="bottom"===n?"top":"bottom",h="right"===r?"left":"right",f=$("transform"),p=void 0,g=void 0;if(g="bottom"===d?"HTML"===s.nodeName?-s.clientHeight+c.bottom:-l.height+c.bottom:c.top,p="right"===h?"HTML"===s.nodeName?-s.clientWidth+c.right:-l.width+c.right:c.left,a&&f)u[f]="translate3d("+p+"px, "+g+"px, 0)",u[d]=0,u[h]=0,u.willChange="transform";else{var y="bottom"===d?-1:1,v="right"===h?-1:1;u[d]=g*y,u[h]=p*v,u.willChange=d+", "+h}var _={"x-placement":t.placement};return t.attributes=T({},_,t.attributes),t.styles=T({},u,t.styles),t.arrowStyles=T({},t.offsets.arrow,t.arrowStyles),t}function lt(t,e,n){var r=F(t,(function(t){var n=t.name;return n===e})),i=!!r&&t.some((function(t){return t.name===n&&t.enabled&&t.order<r.order}));if(!i);return i}function ut(t,e){var n;if(!lt(t.instance.modifiers,"arrow","keepTogether"))return t;var r=e.element;if("string"===typeof r){if(r=t.instance.popper.querySelector(r),!r)return t}else if(!t.instance.popper.contains(r))return t;var i=t.placement.split("-")[0],o=t.offsets,a=o.popper,s=o.reference,u=-1!==["left","right"].indexOf(i),c=u?"height":"width",d=u?"Top":"Left",h=d.toLowerCase(),f=u?"left":"top",p=u?"bottom":"right",m=R(r)[c];s[p]-m<a[h]&&(t.offsets.popper[h]-=a[h]-(s[p]-m)),s[h]+m>a[p]&&(t.offsets.popper[h]+=s[h]+m-a[p]),t.offsets.popper=D(t.offsets.popper);var g=s[h]+s[c]/2-m/2,y=l(t.instance.popper),v=parseFloat(y["margin"+d]),_=parseFloat(y["border"+d+"Width"]),b=g-t.offsets.popper[h]-v-_;return b=Math.max(Math.min(a[c]-m,b),0),t.arrowElement=r,t.offsets.arrow=(n={},L(n,h,Math.round(b)),L(n,f,""),n),t}function ct(t){return"end"===t?"start":"start"===t?"end":t}var dt=["auto-start","auto","auto-end","top-start","top","top-end","right-start","right","right-end","bottom-end","bottom","bottom-start","left-end","left","left-start"],ht=dt.slice(3);function ft(t){var e=arguments.length>1&&void 0!==arguments[1]&&arguments[1],n=ht.indexOf(t),r=ht.slice(n+1).concat(ht.slice(0,n));return e?r.reverse():r}var pt={FLIP:"flip",CLOCKWISE:"clockwise",COUNTERCLOCKWISE:"counterclockwise"};function mt(t,e){if(G(t.instance.modifiers,"inner"))return t;if(t.flipped&&t.placement===t.originalPlacement)return t;var n=j(t.instance.popper,t.instance.reference,e.padding,e.boundariesElement,t.positionFixed),r=t.placement.split("-")[0],i=H(r),o=t.placement.split("-")[1]||"",a=[];switch(e.behavior){case pt.FLIP:a=[r,i];break;case pt.CLOCKWISE:a=ft(r);break;case pt.COUNTERCLOCKWISE:a=ft(r,!0);break;default:a=e.behavior}return a.forEach((function(s,l){if(r!==s||a.length===l+1)return t;r=t.placement.split("-")[0],i=H(r);var u=t.offsets.popper,c=t.offsets.reference,d=Math.floor,h="left"===r&&d(u.right)>d(c.left)||"right"===r&&d(u.left)<d(c.right)||"top"===r&&d(u.bottom)>d(c.top)||"bottom"===r&&d(u.top)<d(c.bottom),f=d(u.left)<d(n.left),p=d(u.right)>d(n.right),m=d(u.top)<d(n.top),g=d(u.bottom)>d(n.bottom),y="left"===r&&f||"right"===r&&p||"top"===r&&m||"bottom"===r&&g,v=-1!==["top","bottom"].indexOf(r),_=!!e.flipVariations&&(v&&"start"===o&&f||v&&"end"===o&&p||!v&&"start"===o&&m||!v&&"end"===o&&g),b=!!e.flipVariationsByContent&&(v&&"start"===o&&p||v&&"end"===o&&f||!v&&"start"===o&&g||!v&&"end"===o&&m),x=_||b;(h||y||x)&&(t.flipped=!0,(h||y)&&(r=a[l+1]),x&&(o=ct(o)),t.placement=r+(o?"-"+o:""),t.offsets.popper=T({},t.offsets.popper,z(t.instance.popper,t.offsets.reference,t.placement)),t=W(t.instance.modifiers,t,"flip"))})),t}function gt(t){var e=t.offsets,n=e.popper,r=e.reference,i=t.placement.split("-")[0],o=Math.floor,a=-1!==["top","bottom"].indexOf(i),s=a?"right":"bottom",l=a?"left":"top",u=a?"width":"height";return n[s]<o(r[l])&&(t.offsets.popper[l]=o(r[l])-n[u]),n[l]>o(r[s])&&(t.offsets.popper[l]=o(r[s])),t}function yt(t,e,n,r){var i=t.match(/((?:\-|\+)?\d*\.?\d*)(.*)/),o=+i[1],a=i[2];if(!o)return t;if(0===a.indexOf("%")){var s=void 0;switch(a){case"%p":s=n;break;case"%":case"%r":default:s=r}var l=D(s);return l[e]/100*o}if("vh"===a||"vw"===a){var u=void 0;return u="vh"===a?Math.max(document.documentElement.clientHeight,window.innerHeight||0):Math.max(document.documentElement.clientWidth,window.innerWidth||0),u/100*o}return o}function vt(t,e,n,r){var i=[0,0],o=-1!==["right","left"].indexOf(r),a=t.split(/(\+|\-)/).map((function(t){return t.trim()})),s=a.indexOf(F(a,(function(t){return-1!==t.search(/,|\s/)})));a[s]&&a[s].indexOf(",");var l=/\s*,\s*|\s+/,u=-1!==s?[a.slice(0,s).concat([a[s].split(l)[0]]),[a[s].split(l)[1]].concat(a.slice(s+1))]:[a];return u=u.map((function(t,r){var i=(1===r?!o:o)?"height":"width",a=!1;return t.reduce((function(t,e){return""===t[t.length-1]&&-1!==["+","-"].indexOf(e)?(t[t.length-1]=e,a=!0,t):a?(t[t.length-1]+=e,a=!1,t):t.concat(e)}),[]).map((function(t){return yt(t,i,e,n)}))})),u.forEach((function(t,e){t.forEach((function(n,r){tt(n)&&(i[e]+=n*("-"===t[r-1]?-1:1))}))})),i}function _t(t,e){var n=e.offset,r=t.placement,i=t.offsets,o=i.popper,a=i.reference,s=r.split("-")[0],l=void 0;return l=tt(+n)?[+n,0]:vt(n,o,a,s),"left"===s?(o.top+=l[0],o.left-=l[1]):"right"===s?(o.top+=l[0],o.left+=l[1]):"top"===s?(o.left+=l[0],o.top-=l[1]):"bottom"===s&&(o.left+=l[0],o.top+=l[1]),t.popper=o,t}function bt(t,e){var n=e.boundariesElement||m(t.instance.popper);t.instance.reference===n&&(n=m(n));var r=$("transform"),i=t.instance.popper.style,o=i.top,a=i.left,s=i[r];i.top="",i.left="",i[r]="";var l=j(t.instance.popper,t.instance.reference,e.padding,n,t.positionFixed);i.top=o,i.left=a,i[r]=s,e.boundaries=l;var u=e.priority,c=t.offsets.popper,d={primary:function(t){var n=c[t];return c[t]<l[t]&&!e.escapeWithReference&&(n=Math.max(c[t],l[t])),L({},t,n)},secondary:function(t){var n="right"===t?"left":"top",r=c[n];return c[t]>l[t]&&!e.escapeWithReference&&(r=Math.min(c[n],l[t]-("right"===t?c.width:c.height))),L({},n,r)}};return u.forEach((function(t){var e=-1!==["left","top"].indexOf(t)?"primary":"secondary";c=T({},c,d[e](t))})),t.offsets.popper=c,t}function xt(t){var e=t.placement,n=e.split("-")[0],r=e.split("-")[1];if(r){var i=t.offsets,o=i.reference,a=i.popper,s=-1!==["bottom","top"].indexOf(n),l=s?"left":"top",u=s?"width":"height",c={start:L({},l,o[l]),end:L({},l,o[l]+o[u]-a[u])};t.offsets.popper=T({},a,c[r])}return t}function wt(t){if(!lt(t.instance.modifiers,"hide","preventOverflow"))return t;var e=t.offsets.reference,n=F(t.instance.modifiers,(function(t){return"preventOverflow"===t.name})).boundaries;if(e.bottom<n.top||e.left>n.right||e.top>n.bottom||e.right<n.left){if(!0===t.hide)return t;t.hide=!0,t.attributes["x-out-of-boundaries"]=""}else{if(!1===t.hide)return t;t.hide=!1,t.attributes["x-out-of-boundaries"]=!1}return t}function Mt(t){var e=t.placement,n=e.split("-")[0],r=t.offsets,i=r.popper,o=r.reference,a=-1!==["left","right"].indexOf(n),s=-1===["top","left"].indexOf(n);return i[a?"left":"top"]=o[n]-(s?i[a?"width":"height"]:0),t.placement=H(e),t.offsets.popper=D(i),t}var St={shift:{order:100,enabled:!0,fn:xt},offset:{order:200,enabled:!0,fn:_t,offset:0},preventOverflow:{order:300,enabled:!0,fn:bt,priority:["left","right","top","bottom"],padding:5,boundariesElement:"scrollParent"},keepTogether:{order:400,enabled:!0,fn:gt},arrow:{order:500,enabled:!0,fn:ut,element:"[x-arrow]"},flip:{order:600,enabled:!0,fn:mt,behavior:"flip",padding:5,boundariesElement:"viewport",flipVariations:!1,flipVariationsByContent:!1},inner:{order:700,enabled:!1,fn:Mt},hide:{order:800,enabled:!0,fn:wt},computeStyle:{order:850,enabled:!0,fn:st,gpuAcceleration:!0,x:"bottom",y:"right"},applyStyle:{order:900,enabled:!0,fn:rt,onLoad:it,gpuAcceleration:void 0}},kt={placement:"bottom",positionFixed:!1,eventsEnabled:!0,removeOnDestroy:!1,onCreate:function(){},onUpdate:function(){},modifiers:St},Lt=function(){function t(e,n){var r=this,i=arguments.length>2&&void 0!==arguments[2]?arguments[2]:{};S(this,t),this.scheduleUpdate=function(){return requestAnimationFrame(r.update)},this.update=a(this.update.bind(this)),this.options=T({},t.Defaults,i),this.state={isDestroyed:!1,isCreated:!1,scrollParents:[]},this.reference=e&&e.jquery?e[0]:e,this.popper=n&&n.jquery?n[0]:n,this.options.modifiers={},Object.keys(T({},t.Defaults.modifiers,i.modifiers)).forEach((function(e){r.options.modifiers[e]=T({},t.Defaults.modifiers[e]||{},i.modifiers?i.modifiers[e]:{})})),this.modifiers=Object.keys(this.options.modifiers).map((function(t){return T({name:t},r.options.modifiers[t])})).sort((function(t,e){return t.order-e.order})),this.modifiers.forEach((function(t){t.enabled&&s(t.onLoad)&&t.onLoad(r.reference,r.popper,r.options,t,r.state)})),this.update();var o=this.options.eventsEnabled;o&&this.enableEventListeners(),this.state.eventsEnabled=o}return k(t,[{key:"update",value:function(){return V.call(this)}},{key:"destroy",value:function(){return U.call(this)}},{key:"enableEventListeners",value:function(){return J.call(this)}},{key:"disableEventListeners",value:function(){return Q.call(this)}}]),t}();return Lt.Utils=("undefined"!==typeof window?window:e).PopperUtils,Lt.placements=dt,Lt.Defaults=kt,Lt}))}).call(this,n("c8ba"))},d327:function(t,e){function n(){return[]}t.exports=n},d370:function(t,e,n){var r=n("253c"),i=n("1310"),o=Object.prototype,a=o.hasOwnProperty,s=o.propertyIsEnumerable,l=r(function(){return arguments}())?r:function(t){return i(t)&&a.call(t,"callee")&&!s.call(t,"callee")};t.exports=l},d3b7:function(t,e,n){"use strict";var r=n("00ee"),i=n("cb2d"),o=n("b041");r||i(Object.prototype,"toString",o,{unsafe:!0})},d44e:function(t,e,n){"use strict";var r=n("9bf2").f,i=n("1a2d"),o=n("b622"),a=o("toStringTag");t.exports=function(t,e,n){t&&!n&&(t=t.prototype),t&&!i(t,a)&&r(t,a,{configurable:!0,value:e})}},d4c3:function(t,e,n){"use strict";var r=n("342f");t.exports=/ipad|iphone|ipod/i.test(r)&&"undefined"!=typeof Pebble},d51b:function(t,e,n){"use strict";var r=function(){function t(t){this.value=t}return t}(),i=function(){function t(){this._len=0}return t.prototype.insert=function(t){var e=new r(t);return this.insertEntry(e),e},t.prototype.insertEntry=function(t){this.head?(this.tail.next=t,t.prev=this.tail,t.next=null,this.tail=t):this.head=this.tail=t,this._len++},t.prototype.remove=function(t){var e=t.prev,n=t.next;e?e.next=n:this.head=n,n?n.prev=e:this.tail=e,t.next=t.prev=null,this._len--},t.prototype.len=function(){return this._len},t.prototype.clear=function(){this.head=this.tail=null,this._len=0},t}(),o=function(){function t(t){this._list=new i,this._maxSize=10,this._map={},this._maxSize=t}return t.prototype.put=function(t,e){var n=this._list,i=this._map,o=null;if(null==i[t]){var a=n.len(),s=this._lastRemovedEntry;if(a>=this._maxSize&&a>0){var l=n.head;n.remove(l),delete i[l.key],o=l.value,this._lastRemovedEntry=l}s?s.value=e:s=new r(e),s.key=t,n.insertEntry(s),i[t]=s}return o},t.prototype.get=function(t){var e=this._map[t],n=this._list;if(null!=e)return e!==n.tail&&(n.remove(e),n.insertEntry(e)),e.value},t.prototype.clear=function(){this._list.clear(),this._map={}},t.prototype.len=function(){return this._list.len()},t}();e["a"]=o},d612:function(t,e,n){var r=n("7b83"),i=n("7ed2"),o=n("dc0f");function a(t){var e=-1,n=null==t?0:t.length;this.__data__=new r;while(++e<n)this.add(t[e])}a.prototype.add=a.prototype.push=i,a.prototype.has=o,t.exports=a},d69a:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("fil",{months:"Enero_Pebrero_Marso_Abril_Mayo_Hunyo_Hulyo_Agosto_Setyembre_Oktubre_Nobyembre_Disyembre".split("_"),monthsShort:"Ene_Peb_Mar_Abr_May_Hun_Hul_Ago_Set_Okt_Nob_Dis".split("_"),weekdays:"Linggo_Lunes_Martes_Miyerkules_Huwebes_Biyernes_Sabado".split("_"),weekdaysShort:"Lin_Lun_Mar_Miy_Huw_Biy_Sab".split("_"),weekdaysMin:"Li_Lu_Ma_Mi_Hu_Bi_Sab".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"MM/D/YYYY",LL:"MMMM D, YYYY",LLL:"MMMM D, YYYY HH:mm",LLLL:"dddd, MMMM DD, YYYY HH:mm"},calendar:{sameDay:"LT [ngayong araw]",nextDay:"[Bukas ng] LT",nextWeek:"LT [sa susunod na] dddd",lastDay:"LT [kahapon]",lastWeek:"LT [noong nakaraang] dddd",sameElse:"L"},relativeTime:{future:"sa loob ng %s",past:"%s ang nakalipas",s:"ilang segundo",ss:"%d segundo",m:"isang minuto",mm:"%d minuto",h:"isang oras",hh:"%d oras",d:"isang araw",dd:"%d araw",M:"isang buwan",MM:"%d buwan",y:"isang taon",yy:"%d taon"},dayOfMonthOrdinalParse:/\d{1,2}/,ordinal:function(t){return t},week:{dow:1,doy:4}});return e}))},d6b6:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("hy-am",{months:{format:"հունվարի_փետրվարի_մարտի_ապրիլի_մայիսի_հունիսի_հուլիսի_օգոստոսի_սեպտեմբերի_հոկտեմբերի_նոյեմբերի_դեկտեմբերի".split("_"),standalone:"հունվար_փետրվար_մարտ_ապրիլ_մայիս_հունիս_հուլիս_օգոստոս_սեպտեմբեր_հոկտեմբեր_նոյեմբեր_դեկտեմբեր".split("_")},monthsShort:"հնվ_փտր_մրտ_ապր_մյս_հնս_հլս_օգս_սպտ_հկտ_նմբ_դկտ".split("_"),weekdays:"կիրակի_երկուշաբթի_երեքշաբթի_չորեքշաբթի_հինգշաբթի_ուրբաթ_շաբաթ".split("_"),weekdaysShort:"կրկ_երկ_երք_չրք_հնգ_ուրբ_շբթ".split("_"),weekdaysMin:"կրկ_երկ_երք_չրք_հնգ_ուրբ_շբթ".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY թ.",LLL:"D MMMM YYYY թ., HH:mm",LLLL:"dddd, D MMMM YYYY թ., HH:mm"},calendar:{sameDay:"[այսօր] LT",nextDay:"[վաղը] LT",lastDay:"[երեկ] LT",nextWeek:function(){return"dddd [օրը ժամը] LT"},lastWeek:function(){return"[անցած] dddd [օրը ժամը] LT"},sameElse:"L"},relativeTime:{future:"%s հետո",past:"%s առաջ",s:"մի քանի վայրկյան",ss:"%d վայրկյան",m:"րոպե",mm:"%d րոպե",h:"ժամ",hh:"%d ժամ",d:"օր",dd:"%d օր",M:"ամիս",MM:"%d ամիս",y:"տարի",yy:"%d տարի"},meridiemParse:/գիշերվա|առավոտվա|ցերեկվա|երեկոյան/,isPM:function(t){return/^(ցերեկվա|երեկոյան)$/.test(t)},meridiem:function(t){return t<4?"գիշերվա":t<12?"առավոտվա":t<17?"ցերեկվա":"երեկոյան"},dayOfMonthOrdinalParse:/\d{1,2}|\d{1,2}-(ին|րդ)/,ordinal:function(t,e){switch(e){case"DDD":case"w":case"W":case"DDDo":return 1===t?t+"-ին":t+"-րդ";default:return t}},week:{dow:1,doy:7}});return e}))},d6d6:function(t,e,n){"use strict";var r=TypeError;t.exports=function(t,e){if(t<e)throw new r("Not enough arguments");return t}},d6dd:function(t,e,n){t.exports=n("c30c")},d6eb:function(t,e,n){"use strict";var r="_erd";function i(t){return t[r]={},o(t)}function o(t){return t[r]}function a(t){delete t[r]}t.exports={initState:i,getState:o,cleanState:a}},d716:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ca",{months:{standalone:"gener_febrer_març_abril_maig_juny_juliol_agost_setembre_octubre_novembre_desembre".split("_"),format:"de gener_de febrer_de març_d'abril_de maig_de juny_de juliol_d'agost_de setembre_d'octubre_de novembre_de desembre".split("_"),isFormat:/D[oD]?(\s)+MMMM/},monthsShort:"gen._febr._març_abr._maig_juny_jul._ag._set._oct._nov._des.".split("_"),monthsParseExact:!0,weekdays:"diumenge_dilluns_dimarts_dimecres_dijous_divendres_dissabte".split("_"),weekdaysShort:"dg._dl._dt._dc._dj._dv._ds.".split("_"),weekdaysMin:"dg_dl_dt_dc_dj_dv_ds".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM [de] YYYY",ll:"D MMM YYYY",LLL:"D MMMM [de] YYYY [a les] H:mm",lll:"D MMM YYYY, H:mm",LLLL:"dddd D MMMM [de] YYYY [a les] H:mm",llll:"ddd D MMM YYYY, H:mm"},calendar:{sameDay:function(){return"[avui a "+(1!==this.hours()?"les":"la")+"] LT"},nextDay:function(){return"[demà a "+(1!==this.hours()?"les":"la")+"] LT"},nextWeek:function(){return"dddd [a "+(1!==this.hours()?"les":"la")+"] LT"},lastDay:function(){return"[ahir a "+(1!==this.hours()?"les":"la")+"] LT"},lastWeek:function(){return"[el] dddd [passat a "+(1!==this.hours()?"les":"la")+"] LT"},sameElse:"L"},relativeTime:{future:"d'aquí %s",past:"fa %s",s:"uns segons",ss:"%d segons",m:"un minut",mm:"%d minuts",h:"una hora",hh:"%d hores",d:"un dia",dd:"%d dies",M:"un mes",MM:"%d mesos",y:"un any",yy:"%d anys"},dayOfMonthOrdinalParse:/\d{1,2}(r|n|t|è|a)/,ordinal:function(t,e){var n=1===t?"r":2===t?"n":3===t?"r":4===t?"t":"è";return"w"!==e&&"W"!==e||(n="a"),t+n},week:{dow:1,doy:4}});return e}))},d784:function(t,e,n){"use strict";n("ac1f");var r=n("c65b"),i=n("cb2d"),o=n("9263"),a=n("d039"),s=n("b622"),l=n("9112"),u=s("species"),c=RegExp.prototype;t.exports=function(t,e,n,d){var h=s(t),f=!a((function(){var e={};return e[h]=function(){return 7},7!==""[t](e)})),p=f&&!a((function(){var e=!1,n=/a/;return"split"===t&&(n={},n.constructor={},n.constructor[u]=function(){return n},n.flags="",n[h]=/./[h]),n.exec=function(){return e=!0,null},n[h](""),!e}));if(!f||!p||n){var m=/./[h],g=e(h,""[t],(function(t,e,n,i,a){var s=e.exec;return s===o||s===c.exec?f&&!a?{done:!0,value:r(m,e,n,i)}:{done:!0,value:r(t,n,e,i)}:{done:!1}}));i(String.prototype,t,g[0]),i(c,h,g[1])}d&&l(c[h],"sham",!0)}},d7ee:function(t,e,n){var r=n("c3fc"),i=n("b047"),o=n("99d3"),a=o&&o.isSet,s=a?i(a):r;t.exports=s},d81d:function(t,e,n){"use strict";var r=n("23e7"),i=n("b727").map,o=n("1dde"),a=o("map");r({target:"Array",proto:!0,forced:!a},{map:function(t){return i(this,t,arguments.length>1?arguments[1]:void 0)}})},d998:function(t,e,n){"use strict";var r=n("342f");t.exports=/MSIE|Trident/.test(r)},d9b5:function(t,e,n){"use strict";var r=n("d066"),i=n("1626"),o=n("3a9b"),a=n("fdbf"),s=Object;t.exports=a?function(t){return"symbol"==typeof t}:function(t){var e=r("Symbol");return i(e)&&o(e.prototype,s(t))}},d9e2:function(t,e,n){"use strict";var r=n("23e7"),i=n("da84"),o=n("2ba4"),a=n("e5cb"),s="WebAssembly",l=i[s],u=7!==new Error("e",{cause:7}).cause,c=function(t,e){var n={};n[t]=a(t,e,u),r({global:!0,constructor:!0,arity:1,forced:u},n)},d=function(t,e){if(l&&l[t]){var n={};n[t]=a(s+"."+t,e,u),r({target:s,stat:!0,constructor:!0,arity:1,forced:u},n)}};c("Error",(function(t){return function(e){return o(t,this,arguments)}})),c("EvalError",(function(t){return function(e){return o(t,this,arguments)}})),c("RangeError",(function(t){return function(e){return o(t,this,arguments)}})),c("ReferenceError",(function(t){return function(e){return o(t,this,arguments)}})),c("SyntaxError",(function(t){return function(e){return o(t,this,arguments)}})),c("TypeError",(function(t){return function(e){return o(t,this,arguments)}})),c("URIError",(function(t){return function(e){return o(t,this,arguments)}})),d("CompileError",(function(t){return function(e){return o(t,this,arguments)}})),d("LinkError",(function(t){return function(e){return o(t,this,arguments)}})),d("RuntimeError",(function(t){return function(e){return o(t,this,arguments)}}))},d9f5:function(t,e,n){"use strict";var r=n("23e7"),i=n("da84"),o=n("c65b"),a=n("e330"),s=n("c430"),l=n("83ab"),u=n("04f8"),c=n("d039"),d=n("1a2d"),h=n("3a9b"),f=n("825a"),p=n("fc6a"),m=n("a04b"),g=n("577e"),y=n("5c6c"),v=n("7c73"),_=n("df75"),b=n("241c"),x=n("057f"),w=n("7418"),M=n("06cf"),S=n("9bf2"),k=n("37e8"),L=n("d1e7"),T=n("cb2d"),D=n("edd0"),O=n("5692"),I=n("f772"),C=n("d012"),A=n("90e3"),Y=n("b622"),j=n("e538"),P=n("e065"),E=n("57b9"),N=n("d44e"),R=n("69f3"),H=n("b727").forEach,z=I("hidden"),F="Symbol",B="prototype",W=R.set,V=R.getterFor(F),G=Object[B],$=i.Symbol,U=$&&$[B],q=i.RangeError,X=i.TypeError,Z=i.QObject,J=M.f,K=S.f,Q=x.f,tt=L.f,et=a([].push),nt=O("symbols"),rt=O("op-symbols"),it=O("wks"),ot=!Z||!Z[B]||!Z[B].findChild,at=function(t,e,n){var r=J(G,e);r&&delete G[e],K(t,e,n),r&&t!==G&&K(G,e,r)},st=l&&c((function(){return 7!==v(K({},"a",{get:function(){return K(this,"a",{value:7}).a}})).a}))?at:K,lt=function(t,e){var n=nt[t]=v(U);return W(n,{type:F,tag:t,description:e}),l||(n.description=e),n},ut=function(t,e,n){t===G&&ut(rt,e,n),f(t);var r=m(e);return f(n),d(nt,r)?(n.enumerable?(d(t,z)&&t[z][r]&&(t[z][r]=!1),n=v(n,{enumerable:y(0,!1)})):(d(t,z)||K(t,z,y(1,v(null))),t[z][r]=!0),st(t,r,n)):K(t,r,n)},ct=function(t,e){f(t);var n=p(e),r=_(n).concat(mt(n));return H(r,(function(e){l&&!o(ht,n,e)||ut(t,e,n[e])})),t},dt=function(t,e){return void 0===e?v(t):ct(v(t),e)},ht=function(t){var e=m(t),n=o(tt,this,e);return!(this===G&&d(nt,e)&&!d(rt,e))&&(!(n||!d(this,e)||!d(nt,e)||d(this,z)&&this[z][e])||n)},ft=function(t,e){var n=p(t),r=m(e);if(n!==G||!d(nt,r)||d(rt,r)){var i=J(n,r);return!i||!d(nt,r)||d(n,z)&&n[z][r]||(i.enumerable=!0),i}},pt=function(t){var e=Q(p(t)),n=[];return H(e,(function(t){d(nt,t)||d(C,t)||et(n,t)})),n},mt=function(t){var e=t===G,n=Q(e?rt:p(t)),r=[];return H(n,(function(t){!d(nt,t)||e&&!d(G,t)||et(r,nt[t])})),r};u||($=function(){if(h(U,this))throw new X("Symbol is not a constructor");var t=arguments.length&&void 0!==arguments[0]?g(arguments[0]):void 0,e=A(t),n=function(t){var r=void 0===this?i:this;r===G&&o(n,rt,t),d(r,z)&&d(r[z],e)&&(r[z][e]=!1);var a=y(1,t);try{st(r,e,a)}catch(s){if(!(s instanceof q))throw s;at(r,e,a)}};return l&&ot&&st(G,e,{configurable:!0,set:n}),lt(e,t)},U=$[B],T(U,"toString",(function(){return V(this).tag})),T($,"withoutSetter",(function(t){return lt(A(t),t)})),L.f=ht,S.f=ut,k.f=ct,M.f=ft,b.f=x.f=pt,w.f=mt,j.f=function(t){return lt(Y(t),t)},l&&(D(U,"description",{configurable:!0,get:function(){return V(this).description}}),s||T(G,"propertyIsEnumerable",ht,{unsafe:!0}))),r({global:!0,constructor:!0,wrap:!0,forced:!u,sham:!u},{Symbol:$}),H(_(it),(function(t){P(t)})),r({target:F,stat:!0,forced:!u},{useSetter:function(){ot=!0},useSimple:function(){ot=!1}}),r({target:"Object",stat:!0,forced:!u,sham:!l},{create:dt,defineProperty:ut,defineProperties:ct,getOwnPropertyDescriptor:ft}),r({target:"Object",stat:!0,forced:!u},{getOwnPropertyNames:pt}),E(),N($,F),C[z]=!0},d9f8:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("fr-ca",{months:"janvier_février_mars_avril_mai_juin_juillet_août_septembre_octobre_novembre_décembre".split("_"),monthsShort:"janv._févr._mars_avr._mai_juin_juil._août_sept._oct._nov._déc.".split("_"),monthsParseExact:!0,weekdays:"dimanche_lundi_mardi_mercredi_jeudi_vendredi_samedi".split("_"),weekdaysShort:"dim._lun._mar._mer._jeu._ven._sam.".split("_"),weekdaysMin:"di_lu_ma_me_je_ve_sa".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY-MM-DD",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[Aujourd’hui à] LT",nextDay:"[Demain à] LT",nextWeek:"dddd [à] LT",lastDay:"[Hier à] LT",lastWeek:"dddd [dernier à] LT",sameElse:"L"},relativeTime:{future:"dans %s",past:"il y a %s",s:"quelques secondes",ss:"%d secondes",m:"une minute",mm:"%d minutes",h:"une heure",hh:"%d heures",d:"un jour",dd:"%d jours",M:"un mois",MM:"%d mois",y:"un an",yy:"%d ans"},dayOfMonthOrdinalParse:/\d{1,2}(er|e)/,ordinal:function(t,e){switch(e){default:case"M":case"Q":case"D":case"DDD":case"d":return t+(1===t?"er":"e");case"w":case"W":return t+(1===t?"re":"e")}}});return e}))},da03:function(t,e,n){var r=n("2b3e"),i=r["__core-js_shared__"];t.exports=i},da84:function(t,e,n){"use strict";(function(e){var n=function(t){return t&&t.Math===Math&&t};t.exports=n("object"==typeof globalThis&&globalThis)||n("object"==typeof window&&window)||n("object"==typeof self&&self)||n("object"==typeof e&&e)||n("object"==typeof this&&this)||function(){return this}()||Function("return this")()}).call(this,n("c8ba"))},db29:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="jan._feb._mrt._apr._mei_jun._jul._aug._sep._okt._nov._dec.".split("_"),n="jan_feb_mrt_apr_mei_jun_jul_aug_sep_okt_nov_dec".split("_"),r=[/^jan/i,/^feb/i,/^(maart|mrt\.?)$/i,/^apr/i,/^mei$/i,/^jun[i.]?$/i,/^jul[i.]?$/i,/^aug/i,/^sep/i,/^okt/i,/^nov/i,/^dec/i],i=/^(januari|februari|maart|april|mei|ju[nl]i|augustus|september|oktober|november|december|jan\.?|feb\.?|mrt\.?|apr\.?|ju[nl]\.?|aug\.?|sep\.?|okt\.?|nov\.?|dec\.?)/i,o=t.defineLocale("nl-be",{months:"januari_februari_maart_april_mei_juni_juli_augustus_september_oktober_november_december".split("_"),monthsShort:function(t,r){return t?/-MMM-/.test(r)?n[t.month()]:e[t.month()]:e},monthsRegex:i,monthsShortRegex:i,monthsStrictRegex:/^(januari|februari|maart|april|mei|ju[nl]i|augustus|september|oktober|november|december)/i,monthsShortStrictRegex:/^(jan\.?|feb\.?|mrt\.?|apr\.?|mei|ju[nl]\.?|aug\.?|sep\.?|okt\.?|nov\.?|dec\.?)/i,monthsParse:r,longMonthsParse:r,shortMonthsParse:r,weekdays:"zondag_maandag_dinsdag_woensdag_donderdag_vrijdag_zaterdag".split("_"),weekdaysShort:"zo._ma._di._wo._do._vr._za.".split("_"),weekdaysMin:"zo_ma_di_wo_do_vr_za".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[vandaag om] LT",nextDay:"[morgen om] LT",nextWeek:"dddd [om] LT",lastDay:"[gisteren om] LT",lastWeek:"[afgelopen] dddd [om] LT",sameElse:"L"},relativeTime:{future:"over %s",past:"%s geleden",s:"een paar seconden",ss:"%d seconden",m:"één minuut",mm:"%d minuten",h:"één uur",hh:"%d uur",d:"één dag",dd:"%d dagen",M:"één maand",MM:"%d maanden",y:"één jaar",yy:"%d jaar"},dayOfMonthOrdinalParse:/\d{1,2}(ste|de)/,ordinal:function(t){return t+(1===t||8===t||t>=20?"ste":"de")},week:{dow:1,doy:4}});return o}))},dc0f:function(t,e){function n(t){return this.__data__.has(t)}t.exports=n},dc4a:function(t,e,n){"use strict";var r=n("59ed"),i=n("7234");t.exports=function(t,e){var n=t[e];return i(n)?void 0:r(n)}},dc4d:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"१",2:"२",3:"३",4:"४",5:"५",6:"६",7:"७",8:"८",9:"९",0:"०"},n={"१":"1","२":"2","३":"3","४":"4","५":"5","६":"6","७":"7","८":"8","९":"9","०":"0"},r=[/^जन/i,/^फ़र|फर/i,/^मार्च/i,/^अप्रै/i,/^मई/i,/^जून/i,/^जुल/i,/^अग/i,/^सितं|सित/i,/^अक्टू/i,/^नव|नवं/i,/^दिसं|दिस/i],i=[/^जन/i,/^फ़र/i,/^मार्च/i,/^अप्रै/i,/^मई/i,/^जून/i,/^जुल/i,/^अग/i,/^सित/i,/^अक्टू/i,/^नव/i,/^दिस/i],o=t.defineLocale("hi",{months:{format:"जनवरी_फ़रवरी_मार्च_अप्रैल_मई_जून_जुलाई_अगस्त_सितम्बर_अक्टूबर_नवम्बर_दिसम्बर".split("_"),standalone:"जनवरी_फरवरी_मार्च_अप्रैल_मई_जून_जुलाई_अगस्त_सितंबर_अक्टूबर_नवंबर_दिसंबर".split("_")},monthsShort:"जन._फ़र._मार्च_अप्रै._मई_जून_जुल._अग._सित._अक्टू._नव._दिस.".split("_"),weekdays:"रविवार_सोमवार_मंगलवार_बुधवार_गुरूवार_शुक्रवार_शनिवार".split("_"),weekdaysShort:"रवि_सोम_मंगल_बुध_गुरू_शुक्र_शनि".split("_"),weekdaysMin:"र_सो_मं_बु_गु_शु_श".split("_"),longDateFormat:{LT:"A h:mm बजे",LTS:"A h:mm:ss बजे",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm बजे",LLLL:"dddd, D MMMM YYYY, A h:mm बजे"},monthsParse:r,longMonthsParse:r,shortMonthsParse:i,monthsRegex:/^(जनवरी|जन\.?|फ़रवरी|फरवरी|फ़र\.?|मार्च?|अप्रैल|अप्रै\.?|मई?|जून?|जुलाई|जुल\.?|अगस्त|अग\.?|सितम्बर|सितंबर|सित\.?|अक्टूबर|अक्टू\.?|नवम्बर|नवंबर|नव\.?|दिसम्बर|दिसंबर|दिस\.?)/i,monthsShortRegex:/^(जनवरी|जन\.?|फ़रवरी|फरवरी|फ़र\.?|मार्च?|अप्रैल|अप्रै\.?|मई?|जून?|जुलाई|जुल\.?|अगस्त|अग\.?|सितम्बर|सितंबर|सित\.?|अक्टूबर|अक्टू\.?|नवम्बर|नवंबर|नव\.?|दिसम्बर|दिसंबर|दिस\.?)/i,monthsStrictRegex:/^(जनवरी?|फ़रवरी|फरवरी?|मार्च?|अप्रैल?|मई?|जून?|जुलाई?|अगस्त?|सितम्बर|सितंबर|सित?\.?|अक्टूबर|अक्टू\.?|नवम्बर|नवंबर?|दिसम्बर|दिसंबर?)/i,monthsShortStrictRegex:/^(जन\.?|फ़र\.?|मार्च?|अप्रै\.?|मई?|जून?|जुल\.?|अग\.?|सित\.?|अक्टू\.?|नव\.?|दिस\.?)/i,calendar:{sameDay:"[आज] LT",nextDay:"[कल] LT",nextWeek:"dddd, LT",lastDay:"[कल] LT",lastWeek:"[पिछले] dddd, LT",sameElse:"L"},relativeTime:{future:"%s में",past:"%s पहले",s:"कुछ ही क्षण",ss:"%d सेकंड",m:"एक मिनट",mm:"%d मिनट",h:"एक घंटा",hh:"%d घंटे",d:"एक दिन",dd:"%d दिन",M:"एक महीने",MM:"%d महीने",y:"एक वर्ष",yy:"%d वर्ष"},preparse:function(t){return t.replace(/[१२३४५६७८९०]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/रात|सुबह|दोपहर|शाम/,meridiemHour:function(t,e){return 12===t&&(t=0),"रात"===e?t<4?t:t+12:"सुबह"===e?t:"दोपहर"===e?t>=10?t:t+12:"शाम"===e?t+12:void 0},meridiem:function(t,e,n){return t<4?"रात":t<10?"सुबह":t<17?"दोपहर":t<20?"शाम":"रात"},week:{dow:0,doy:6}});return o}))},dc57:function(t,e){var n=Function.prototype,r=n.toString;function i(t){if(null!=t){try{return r.call(t)}catch(e){}try{return t+""}catch(e){}}return""}t.exports=i},dcbe:function(t,e,n){var r=n("30c9"),i=n("1310");function o(t){return i(t)&&r(t)}t.exports=o},dcbf:function(t,e,n){"use strict";var r=n("f500");e["a"]=r["a"]},dcc3:function(t,e,n){"use strict";var r=n("ae93").IteratorPrototype,i=n("7c73"),o=n("5c6c"),a=n("d44e"),s=n("3f8c"),l=function(){return this};t.exports=function(t,e,n,u){var c=e+" Iterator";return t.prototype=i(r,{next:o(+!u,n)}),a(t,c,!1,!0),s[c]=l,t}},dce5:function(t,e,n){var r=n("32b3"),i=n("8eeb"),o=n("2ec1"),a=n("30c9"),s=n("eac5"),l=n("ec69"),u=Object.prototype,c=u.hasOwnProperty,d=o((function(t,e){if(s(e)||a(e))i(e,l(e),t);else for(var n in e)c.call(e,n)&&r(t,n,e[n])}));t.exports=d},dd65:function(t,e,n){var r=n("badf"),i=n("a0ac"),o=n("77c1");function a(t,e){return o(t,i(r(e)))}t.exports=a},ddb0:function(t,e,n){"use strict";var r=n("da84"),i=n("fdbc"),o=n("785a"),a=n("e260"),s=n("9112"),l=n("d44e"),u=n("b622"),c=u("iterator"),d=a.values,h=function(t,e){if(t){if(t[c]!==d)try{s(t,c,d)}catch(r){t[c]=d}if(l(t,e,!0),i[e])for(var n in a)if(t[n]!==a[n])try{s(t,n,a[n])}catch(r){t[n]=a[n]}}};for(var f in i)h(r[f]&&r[f].prototype,f);h(o,"DOMTokenList")},de48:function(t,e,n){"use strict";n.d(e,"b",(function(){return o})),n.d(e,"a",(function(){return a}));var r=n("2b0e");const i=r["a"].prototype.$isServer,o=function(){return!i&&document.addEventListener?function(t,e,n,r=!1){t&&e&&n&&t.addEventListener(e,n,r)}:function(t,e,n){t&&e&&n&&t.attachEvent("on"+e,n)}}(),a=function(){return!i&&document.removeEventListener?function(t,e,n,r=!1){t&&e&&t.removeEventListener(e,n,r)}:function(t,e,n){t&&e&&t.detachEvent("on"+e,n)}}()},df75:function(t,e,n){"use strict";var r=n("ca84"),i=n("7839");t.exports=Object.keys||function(t){return r(t,i)}},e01a:function(t,e,n){"use strict";var r=n("23e7"),i=n("83ab"),o=n("da84"),a=n("e330"),s=n("1a2d"),l=n("1626"),u=n("3a9b"),c=n("577e"),d=n("edd0"),h=n("e893"),f=o.Symbol,p=f&&f.prototype;if(i&&l(f)&&(!("description"in p)||void 0!==f().description)){var m={},g=function(){var t=arguments.length<1||void 0===arguments[0]?void 0:c(arguments[0]),e=u(p,this)?new f(t):void 0===t?f():f(t);return""===t&&(m[e]=!0),e};h(g,f),g.prototype=p,p.constructor=g;var y="Symbol(description detection)"===String(f("description detection")),v=a(p.valueOf),_=a(p.toString),b=/^Symbol\((.*)\)[^)]+$/,x=a("".replace),w=a("".slice);d(p,"description",{configurable:!0,get:function(){var t=v(this);if(s(m,t))return"";var e=_(t),n=y?w(e,7,-1):x(e,b,"$1");return""===n?void 0:n}}),r({global:!0,constructor:!0,forced:!0},{Symbol:g})}},e031:function(t,e,n){var r=n("f909"),i=n("1a8c");function o(t,e,n,a,s,l){return i(t)&&i(e)&&(l.set(e,t),r(t,e,void 0,o,l),l["delete"](e)),t}t.exports=o},e065:function(t,e,n){"use strict";var r=n("428f"),i=n("1a2d"),o=n("e538"),a=n("9bf2").f;t.exports=function(t){var e=r.Symbol||(r.Symbol={});i(e,t)||a(e,t,{value:o.f(t)})}},e0b9:function(t,e,n){"use strict";var r=function(){var t=this,e=t._self._c;return e("transition",{attrs:{name:"fade"}},[e("div",{directives:[{name:"show",rawName:"v-show",value:t.show,expression:"show"}],class:t.classes,style:t.outerStyles},[e("div",{class:t.innerClasses,style:t.styles})])])},i=[],o=n("ade3"),a=(n("a9e3"),"ivu-loading-bar"),s={name:"LoadingBar",props:{color:{type:String,default:"primary"},failedColor:{type:String,default:"error"},height:{type:Number,default:2}},data:function(){return{percent:0,status:"success",show:!1}},computed:{classes:function(){return"".concat(a)},innerClasses:function(){return["".concat(a,"-inner"),Object(o["a"])(Object(o["a"])({},"".concat(a,"-inner-color-primary"),"primary"===this.color&&"success"===this.status),"".concat(a,"-inner-failed-color-error"),"error"===this.failedColor&&"error"===this.status)]},outerStyles:function(){return{height:"".concat(this.height,"px")}},styles:function(){var t={width:"".concat(this.percent,"%"),height:"".concat(this.height,"px")};return"primary"!==this.color&&"success"===this.status&&(t.backgroundColor=this.color),"error"!==this.failedColor&&"error"===this.status&&(t.backgroundColor=this.failedColor),t}}},l=s,u=n("2877"),c=Object(u["a"])(l,r,i,!1,null,null,null),d=c.exports,h=n("2b0e");d.newInstance=t=>{const e=t||{},n=new h["a"]({data:e,render(t){return t(d,{props:e})}}),r=n.$mount();document.body.appendChild(r.$el);const i=n.$children[0];return{update(t){"percent"in t&&(i.percent=t.percent),t.status&&(i.status=t.status),"show"in t&&(i.show=t.show)},component:i,destroy(){document.body.removeChild(document.getElementsByClassName("ivu-loading-bar")[0])}}};var f=d;let p,m,g="primary",y=800,v="error",_=2;function b(){return p=p||f.newInstance({color:g,failedColor:v,height:_}),p}function x(t){let e=b();e.update(t)}function w(){setTimeout(()=>{x({show:!1}),setTimeout(()=>{x({percent:0})},200)},y)}function M(){m&&(clearInterval(m),m=null)}e["a"]={start(){if(m)return;let t=0;x({percent:t,status:"success",show:!0}),m=setInterval(()=>{t+=Math.floor(3*Math.random()+1),t>95&&M(),x({percent:t,status:"success",show:!0})},200)},update(t){M(),x({percent:t,status:"success",show:!0})},finish(){M(),x({percent:100,status:"success",show:!0}),w()},error(){M(),x({percent:100,status:"error",show:!0}),w()},config(t){t.color&&(g=t.color),t.duration&&(y=t.duration),t.failedColor&&(v=t.failedColor),t.height&&(_=t.height)},destroy(){M();let t=b();p=null,t.destroy()}}},e0c5:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"૧",2:"૨",3:"૩",4:"૪",5:"૫",6:"૬",7:"૭",8:"૮",9:"૯",0:"૦"},n={"૧":"1","૨":"2","૩":"3","૪":"4","૫":"5","૬":"6","૭":"7","૮":"8","૯":"9","૦":"0"},r=t.defineLocale("gu",{months:"જાન્યુઆરી_ફેબ્રુઆરી_માર્ચ_એપ્રિલ_મે_જૂન_જુલાઈ_ઑગસ્ટ_સપ્ટેમ્બર_ઑક્ટ્બર_નવેમ્બર_ડિસેમ્બર".split("_"),monthsShort:"જાન્યુ._ફેબ્રુ._માર્ચ_એપ્રિ._મે_જૂન_જુલા._ઑગ._સપ્ટે._ઑક્ટ્._નવે._ડિસે.".split("_"),monthsParseExact:!0,weekdays:"રવિવાર_સોમવાર_મંગળવાર_બુધ્વાર_ગુરુવાર_શુક્રવાર_શનિવાર".split("_"),weekdaysShort:"રવિ_સોમ_મંગળ_બુધ્_ગુરુ_શુક્ર_શનિ".split("_"),weekdaysMin:"ર_સો_મં_બુ_ગુ_શુ_શ".split("_"),longDateFormat:{LT:"A h:mm વાગ્યે",LTS:"A h:mm:ss વાગ્યે",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm વાગ્યે",LLLL:"dddd, D MMMM YYYY, A h:mm વાગ્યે"},calendar:{sameDay:"[આજ] LT",nextDay:"[કાલે] LT",nextWeek:"dddd, LT",lastDay:"[ગઇકાલે] LT",lastWeek:"[પાછલા] dddd, LT",sameElse:"L"},relativeTime:{future:"%s મા",past:"%s પહેલા",s:"અમુક પળો",ss:"%d સેકંડ",m:"એક મિનિટ",mm:"%d મિનિટ",h:"એક કલાક",hh:"%d કલાક",d:"એક દિવસ",dd:"%d દિવસ",M:"એક મહિનો",MM:"%d મહિનો",y:"એક વર્ષ",yy:"%d વર્ષ"},preparse:function(t){return t.replace(/[૧૨૩૪૫૬૭૮૯૦]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/રાત|બપોર|સવાર|સાંજ/,meridiemHour:function(t,e){return 12===t&&(t=0),"રાત"===e?t<4?t:t+12:"સવાર"===e?t:"બપોર"===e?t>=10?t:t+12:"સાંજ"===e?t+12:void 0},meridiem:function(t,e,n){return t<4?"રાત":t<10?"સવાર":t<17?"બપોર":t<20?"સાંજ":"રાત"},week:{dow:0,doy:6}});return r}))},e0e7:function(t,e,n){var r=n("60ed");function i(t){return r(t)?void 0:t}t.exports=i},e163:function(t,e,n){"use strict";var r=n("1a2d"),i=n("1626"),o=n("7b0b"),a=n("f772"),s=n("e177"),l=a("IE_PROTO"),u=Object,c=u.prototype;t.exports=s?u.getPrototypeOf:function(t){var e=o(t);if(r(e,l))return e[l];var n=e.constructor;return i(n)&&e instanceof n?n.prototype:e instanceof u?c:null}},e177:function(t,e,n){"use strict";var r=n("d039");t.exports=!r((function(){function t(){}return t.prototype.constructor=null,Object.getPrototypeOf(new t)!==t.prototype}))},e1d3:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("en-ie",{months:"January_February_March_April_May_June_July_August_September_October_November_December".split("_"),monthsShort:"Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec".split("_"),weekdays:"Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday".split("_"),weekdaysShort:"Sun_Mon_Tue_Wed_Thu_Fri_Sat".split("_"),weekdaysMin:"Su_Mo_Tu_We_Th_Fr_Sa".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[Today at] LT",nextDay:"[Tomorrow at] LT",nextWeek:"dddd [at] LT",lastDay:"[Yesterday at] LT",lastWeek:"[Last] dddd [at] LT",sameElse:"L"},relativeTime:{future:"in %s",past:"%s ago",s:"a few seconds",ss:"%d seconds",m:"a minute",mm:"%d minutes",h:"an hour",hh:"%d hours",d:"a day",dd:"%d days",M:"a month",MM:"%d months",y:"a year",yy:"%d years"},dayOfMonthOrdinalParse:/\d{1,2}(st|nd|rd|th)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n},week:{dow:1,doy:4}});return e}))},e20c:function(t,e,n){"use strict";var r=n("23e7"),i=n("da84"),o=n("2cf4").clear;r({global:!0,bind:!0,enumerable:!0,forced:i.clearImmediate!==o},{clearImmediate:o})},e24b:function(t,e,n){var r=n("49f4"),i=n("1efc"),o=n("bbc0"),a=n("7a48"),s=n("2524");function l(t){var e=-1,n=null==t?0:t.length;this.clear();while(++e<n){var r=t[e];this.set(r[0],r[1])}}l.prototype.clear=r,l.prototype["delete"]=i,l.prototype.get=o,l.prototype.has=a,l.prototype.set=s,t.exports=l},e260:function(t,e,n){"use strict";var r=n("fc6a"),i=n("44d2"),o=n("3f8c"),a=n("69f3"),s=n("9bf2").f,l=n("c6d2"),u=n("4754"),c=n("c430"),d=n("83ab"),h="Array Iterator",f=a.set,p=a.getterFor(h);t.exports=l(Array,"Array",(function(t,e){f(this,{type:h,target:r(t),index:0,kind:e})}),(function(){var t=p(this),e=t.target,n=t.index++;if(!e||n>=e.length)return t.target=void 0,u(void 0,!0);switch(t.kind){case"keys":return u(n,!1);case"values":return u(e[n],!1)}return u([n,e[n]],!1)}),"values");var m=o.Arguments=o.Array;if(i("keys"),i("values"),i("entries"),!c&&d&&"values"!==m.name)try{s(m,"name",{value:"values"})}catch(g){}},e267:function(t,e,n){"use strict";var r=n("e330"),i=n("e8b5"),o=n("1626"),a=n("c6b6"),s=n("577e"),l=r([].push);t.exports=function(t){if(o(t))return t;if(i(t)){for(var e=t.length,n=[],r=0;r<e;r++){var u=t[r];"string"==typeof u?l(n,u):"number"!=typeof u&&"Number"!==a(u)&&"String"!==a(u)||l(n,s(u))}var c=n.length,d=!0;return function(t,e){if(d)return d=!1,e;if(i(this))return e;for(var r=0;r<c;r++)if(n[r]===t)return e}}}},e285:function(t,e,n){"use strict";var r=n("da84"),i=r.isFinite;t.exports=Number.isFinite||function(t){return"number"==typeof t&&i(t)}},e2c0:function(t,e,n){var r=n("e2e4"),i=n("d370"),o=n("6747"),a=n("c098"),s=n("b218"),l=n("f4d6");function u(t,e,n){e=r(e,t);var u=-1,c=e.length,d=!1;while(++u<c){var h=l(e[u]);if(!(d=null!=t&&n(t,h)))break;t=t[h]}return d||++u!=c?d:(c=null==t?0:t.length,!!c&&s(c)&&a(h,c)&&(o(t)||i(t)))}t.exports=u},e2e4:function(t,e,n){var r=n("6747"),i=n("f608"),o=n("18d8"),a=n("76dd");function s(t,e){return r(t)?t:i(t,e)?[t]:o(a(t))}t.exports=s},e330:function(t,e,n){"use strict";var r=n("40d5"),i=Function.prototype,o=i.call,a=r&&i.bind.bind(o,o);t.exports=r?a:function(t){return function(){return o.apply(t,arguments)}}},e380:function(t,e,n){var r=n("7b83"),i="Expected a function";function o(t,e){if("function"!=typeof t||null!=e&&"function"!=typeof e)throw new TypeError(i);var n=function(){var r=arguments,i=e?e.apply(this,r):r[0],o=n.cache;if(o.has(i))return o.get(i);var a=t.apply(this,r);return n.cache=o.set(i,a)||o,a};return n.cache=new(o.Cache||r),n}o.Cache=r,t.exports=o},e391:function(t,e,n){"use strict";var r=n("577e");t.exports=function(t,e){return void 0===t?arguments.length<2?"":e:r(t)}},e3db:function(t,e){var n={}.toString;t.exports=Array.isArray||function(t){return"[object Array]"==n.call(t)}},e3f5:function(t,e,n){"use strict";var r=n("7a12"),i=n("f500");r["a"].Group=i["a"],e["a"]=r["a"]},e3f8:function(t,e,n){var r=n("656b");function i(t){return function(e){return r(e,t)}}t.exports=i},e538:function(t,e,n){"use strict";var r=n("b622");e.f=r},e5383:function(t,e,n){(function(t){var r=n("2b3e"),i=e&&!e.nodeType&&e,o=i&&"object"==typeof t&&t&&!t.nodeType&&t,a=o&&o.exports===i,s=a?r.Buffer:void 0,l=s?s.allocUnsafe:void 0;function u(t,e){if(e)return t.slice();var n=t.length,r=l?l(n):new t.constructor(n);return t.copy(r),r}t.exports=u}).call(this,n("62e4")(t))},e5cb:function(t,e,n){"use strict";var r=n("d066"),i=n("1a2d"),o=n("9112"),a=n("3a9b"),s=n("d2bb"),l=n("e893"),u=n("aeb0"),c=n("7156"),d=n("e391"),h=n("ab36"),f=n("6f19"),p=n("83ab"),m=n("c430");t.exports=function(t,e,n,g){var y="stackTraceLimit",v=g?2:1,_=t.split("."),b=_[_.length-1],x=r.apply(null,_);if(x){var w=x.prototype;if(!m&&i(w,"cause")&&delete w.cause,!n)return x;var M=r("Error"),S=e((function(t,e){var n=d(g?e:t,void 0),r=g?new x(t):new x;return void 0!==n&&o(r,"message",n),f(r,S,r.stack,2),this&&a(w,this)&&c(r,this,S),arguments.length>v&&h(r,arguments[v]),r}));if(S.prototype=w,"Error"!==b?s?s(S,M):l(S,M,{name:!0}):p&&y in x&&(u(S,x,y),u(S,x,"prepareStackTrace")),l(S,x),!m)try{w.name!==b&&o(w,"name",b),w.constructor=S}catch(k){}return S}}},e667:function(t,e,n){"use strict";t.exports=function(t){try{return{error:!1,value:t()}}catch(e){return{error:!0,value:e}}}},e6cf:function(t,e,n){"use strict";n("5e7e"),n("14e5"),n("cc98"),n("3529"),n("f22b"),n("7149")},e759:function(t,e,n){var r=n("badf"),i=n("43ad"),o=Object.prototype,a=o.hasOwnProperty,s=o.toString,l=i((function(t,e,n){null!=e&&"function"!=typeof e.toString&&(e=s.call(e)),a.call(t,e)?t[e].push(n):t[e]=[n]}),r);t.exports=l},e79d:function(t,e,n){"use strict";var r=n("245a");e["a"]=r["a"]},e81d:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"១",2:"២",3:"៣",4:"៤",5:"៥",6:"៦",7:"៧",8:"៨",9:"៩",0:"០"},n={"១":"1","២":"2","៣":"3","៤":"4","៥":"5","៦":"6","៧":"7","៨":"8","៩":"9","០":"0"},r=t.defineLocale("km",{months:"មករា_កុម្ភៈ_មីនា_មេសា_ឧសភា_មិថុនា_កក្កដា_សីហា_កញ្ញា_តុលា_វិច្ឆិកា_ធ្នូ".split("_"),monthsShort:"មករា_កុម្ភៈ_មីនា_មេសា_ឧសភា_មិថុនា_កក្កដា_សីហា_កញ្ញា_តុលា_វិច្ឆិកា_ធ្នូ".split("_"),weekdays:"អាទិត្យ_ច័ន្ទ_អង្គារ_ពុធ_ព្រហស្បតិ៍_សុក្រ_សៅរ៍".split("_"),weekdaysShort:"អា_ច_អ_ព_ព្រ_សុ_ស".split("_"),weekdaysMin:"អា_ច_អ_ព_ព្រ_សុ_ស".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},meridiemParse:/ព្រឹក|ល្ងាច/,isPM:function(t){return"ល្ងាច"===t},meridiem:function(t,e,n){return t<12?"ព្រឹក":"ល្ងាច"},calendar:{sameDay:"[ថ្ងៃនេះ ម៉ោង] LT",nextDay:"[ស្អែក ម៉ោង] LT",nextWeek:"dddd [ម៉ោង] LT",lastDay:"[ម្សិលមិញ ម៉ោង] LT",lastWeek:"dddd [សប្តាហ៍មុន] [ម៉ោង] LT",sameElse:"L"},relativeTime:{future:"%sទៀត",past:"%sមុន",s:"ប៉ុន្មានវិនាទី",ss:"%d វិនាទី",m:"មួយនាទី",mm:"%d នាទី",h:"មួយម៉ោង",hh:"%d ម៉ោង",d:"មួយថ្ងៃ",dd:"%d ថ្ងៃ",M:"មួយខែ",MM:"%d ខែ",y:"មួយឆ្នាំ",yy:"%d ឆ្នាំ"},dayOfMonthOrdinalParse:/ទី\d{1,2}/,ordinal:"ទី%d",preparse:function(t){return t.replace(/[១២៣៤៥៦៧៨៩០]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},week:{dow:1,doy:4}});return r}))},e893:function(t,e,n){"use strict";var r=n("1a2d"),i=n("56ef"),o=n("06cf"),a=n("9bf2");t.exports=function(t,e,n){for(var s=i(e),l=a.f,u=o.f,c=0;c<s.length;c++){var d=s[c];r(t,d)||n&&r(n,d)||l(t,d,u(e,d))}}},e8b5:function(t,e,n){"use strict";var r=n("c6b6");t.exports=Array.isArray||function(t){return"Array"===r(t)}},e95a:function(t,e,n){"use strict";var r=n("b622"),i=n("3f8c"),o=r("iterator"),a=Array.prototype;t.exports=function(t){return void 0!==t&&(i.Array===t||a[o]===t)}},e9c4:function(t,e,n){"use strict";var r=n("23e7"),i=n("d066"),o=n("2ba4"),a=n("c65b"),s=n("e330"),l=n("d039"),u=n("1626"),c=n("d9b5"),d=n("f36a"),h=n("e267"),f=n("04f8"),p=String,m=i("JSON","stringify"),g=s(/./.exec),y=s("".charAt),v=s("".charCodeAt),_=s("".replace),b=s(1..toString),x=/[\uD800-\uDFFF]/g,w=/^[\uD800-\uDBFF]$/,M=/^[\uDC00-\uDFFF]$/,S=!f||l((function(){var t=i("Symbol")("stringify detection");return"[null]"!==m([t])||"{}"!==m({a:t})||"{}"!==m(Object(t))})),k=l((function(){return'"\\udf06\\ud834"'!==m("\udf06\ud834")||'"\\udead"'!==m("\udead")})),L=function(t,e){var n=d(arguments),r=h(e);if(u(r)||void 0!==t&&!c(t))return n[1]=function(t,e){if(u(r)&&(e=a(r,this,p(t),e)),!c(e))return e},o(m,null,n)},T=function(t,e,n){var r=y(n,e-1),i=y(n,e+1);return g(w,t)&&!g(M,i)||g(M,t)&&!g(w,r)?"\\u"+b(v(t,0),16):t};m&&r({target:"JSON",stat:!0,arity:3,forced:S||k},{stringify:function(t,e,n){var r=d(arguments),i=o(S?L:m,null,r);return k&&"string"==typeof i?_(i,x,T):i}})},e9d1:function(t,e,n){var r=n("a78b"),i=n("1304");function o(t,e,n,o){return o="function"==typeof o?o:void 0,null==t?t:r(t,e,i(n),o)}t.exports=o},eac5:function(t,e){var n=Object.prototype;function r(t){var e=t&&t.constructor,r="function"==typeof e&&e.prototype||n;return t===r}t.exports=r},eac50:function(t,e,n){"use strict";var r=n("861d"),i=Math.floor;t.exports=Number.isInteger||function(t){return!r(t)&&isFinite(t)&&i(t)===t}},ebe4:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ms",{months:"Januari_Februari_Mac_April_Mei_Jun_Julai_Ogos_September_Oktober_November_Disember".split("_"),monthsShort:"Jan_Feb_Mac_Apr_Mei_Jun_Jul_Ogs_Sep_Okt_Nov_Dis".split("_"),weekdays:"Ahad_Isnin_Selasa_Rabu_Khamis_Jumaat_Sabtu".split("_"),weekdaysShort:"Ahd_Isn_Sel_Rab_Kha_Jum_Sab".split("_"),weekdaysMin:"Ah_Is_Sl_Rb_Km_Jm_Sb".split("_"),longDateFormat:{LT:"HH.mm",LTS:"HH.mm.ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY [pukul] HH.mm",LLLL:"dddd, D MMMM YYYY [pukul] HH.mm"},meridiemParse:/pagi|tengahari|petang|malam/,meridiemHour:function(t,e){return 12===t&&(t=0),"pagi"===e?t:"tengahari"===e?t>=11?t:t+12:"petang"===e||"malam"===e?t+12:void 0},meridiem:function(t,e,n){return t<11?"pagi":t<15?"tengahari":t<19?"petang":"malam"},calendar:{sameDay:"[Hari ini pukul] LT",nextDay:"[Esok pukul] LT",nextWeek:"dddd [pukul] LT",lastDay:"[Kelmarin pukul] LT",lastWeek:"dddd [lepas pukul] LT",sameElse:"L"},relativeTime:{future:"dalam %s",past:"%s yang lepas",s:"beberapa saat",ss:"%d saat",m:"seminit",mm:"%d minit",h:"sejam",hh:"%d jam",d:"sehari",dd:"%d hari",M:"sebulan",MM:"%d bulan",y:"setahun",yy:"%d tahun"},week:{dow:1,doy:7}});return e}))},ec18:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,r){var i={s:["mõne sekundi","mõni sekund","paar sekundit"],ss:[t+"sekundi",t+"sekundit"],m:["ühe minuti","üks minut"],mm:[t+" minuti",t+" minutit"],h:["ühe tunni","tund aega","üks tund"],hh:[t+" tunni",t+" tundi"],d:["ühe päeva","üks päev"],M:["kuu aja","kuu aega","üks kuu"],MM:[t+" kuu",t+" kuud"],y:["ühe aasta","aasta","üks aasta"],yy:[t+" aasta",t+" aastat"]};return e?i[n][2]?i[n][2]:i[n][1]:r?i[n][0]:i[n][1]}var n=t.defineLocale("et",{months:"jaanuar_veebruar_märts_aprill_mai_juuni_juuli_august_september_oktoober_november_detsember".split("_"),monthsShort:"jaan_veebr_märts_apr_mai_juuni_juuli_aug_sept_okt_nov_dets".split("_"),weekdays:"pühapäev_esmaspäev_teisipäev_kolmapäev_neljapäev_reede_laupäev".split("_"),weekdaysShort:"P_E_T_K_N_R_L".split("_"),weekdaysMin:"P_E_T_K_N_R_L".split("_"),longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY H:mm",LLLL:"dddd, D. MMMM YYYY H:mm"},calendar:{sameDay:"[Täna,] LT",nextDay:"[Homme,] LT",nextWeek:"[Järgmine] dddd LT",lastDay:"[Eile,] LT",lastWeek:"[Eelmine] dddd LT",sameElse:"L"},relativeTime:{future:"%s pärast",past:"%s tagasi",s:e,ss:e,m:e,mm:e,h:e,hh:e,d:e,dd:"%d päeva",M:e,MM:e,y:e,yy:e},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return n}))},ec2e:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("en-in",{months:"January_February_March_April_May_June_July_August_September_October_November_December".split("_"),monthsShort:"Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec".split("_"),weekdays:"Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday".split("_"),weekdaysShort:"Sun_Mon_Tue_Wed_Thu_Fri_Sat".split("_"),weekdaysMin:"Su_Mo_Tu_We_Th_Fr_Sa".split("_"),longDateFormat:{LT:"h:mm A",LTS:"h:mm:ss A",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY h:mm A",LLLL:"dddd, D MMMM YYYY h:mm A"},calendar:{sameDay:"[Today at] LT",nextDay:"[Tomorrow at] LT",nextWeek:"dddd [at] LT",lastDay:"[Yesterday at] LT",lastWeek:"[Last] dddd [at] LT",sameElse:"L"},relativeTime:{future:"in %s",past:"%s ago",s:"a few seconds",ss:"%d seconds",m:"a minute",mm:"%d minutes",h:"an hour",hh:"%d hours",d:"a day",dd:"%d days",M:"a month",MM:"%d months",y:"a year",yy:"%d years"},dayOfMonthOrdinalParse:/\d{1,2}(st|nd|rd|th)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n},week:{dow:0,doy:6}});return e}))},ec47:function(t,e,n){var r=n("a3fd"),i=n("42a2"),o=n("edfa"),a=n("cebd"),s="[object Map]",l="[object Set]";function u(t){return function(e){var n=i(e);return n==s?o(e):n==l?a(e):r(e,t(e))}}t.exports=u},ec69:function(t,e,n){var r=n("6fcd"),i=n("03dd"),o=n("30c9");function a(t){return o(t)?r(t):i(t)}t.exports=a},ec8c:function(t,e){function n(t){var e=[];if(null!=t)for(var n in Object(t))e.push(n);return e}t.exports=n},eda5:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("si",{months:"ජනවාරි_පෙබරවාරි_මාර්තු_අප්‍රේල්_මැයි_ජූනි_ජූලි_අගෝස්තු_සැප්තැම්බර්_ඔක්තෝබර්_නොවැම්බර්_දෙසැම්බර්".split("_"),monthsShort:"ජන_පෙබ_මාර්_අප්_මැයි_ජූනි_ජූලි_අගෝ_සැප්_ඔක්_නොවැ_දෙසැ".split("_"),weekdays:"ඉරිදා_සඳුදා_අඟහරුවාදා_බදාදා_බ්‍රහස්පතින්දා_සිකුරාදා_සෙනසුරාදා".split("_"),weekdaysShort:"ඉරි_සඳු_අඟ_බදා_බ්‍රහ_සිකු_සෙන".split("_"),weekdaysMin:"ඉ_ස_අ_බ_බ්‍ර_සි_සෙ".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"a h:mm",LTS:"a h:mm:ss",L:"YYYY/MM/DD",LL:"YYYY MMMM D",LLL:"YYYY MMMM D, a h:mm",LLLL:"YYYY MMMM D [වැනි] dddd, a h:mm:ss"},calendar:{sameDay:"[අද] LT[ට]",nextDay:"[හෙට] LT[ට]",nextWeek:"dddd LT[ට]",lastDay:"[ඊයේ] LT[ට]",lastWeek:"[පසුගිය] dddd LT[ට]",sameElse:"L"},relativeTime:{future:"%sකින්",past:"%sකට පෙර",s:"තත්පර කිහිපය",ss:"තත්පර %d",m:"මිනිත්තුව",mm:"මිනිත්තු %d",h:"පැය",hh:"පැය %d",d:"දිනය",dd:"දින %d",M:"මාසය",MM:"මාස %d",y:"වසර",yy:"වසර %d"},dayOfMonthOrdinalParse:/\d{1,2} වැනි/,ordinal:function(t){return t+" වැනි"},meridiemParse:/පෙර වරු|පස් වරු|පෙ.ව|ප.ව./,isPM:function(t){return"ප.ව."===t||"පස් වරු"===t},meridiem:function(t,e,n){return t>11?n?"ප.ව.":"පස් වරු":n?"පෙ.ව.":"පෙර වරු"}});return e}))},edd0:function(t,e,n){"use strict";var r=n("13d2"),i=n("9bf2");t.exports=function(t,e,n){return n.get&&r(n.get,e,{getter:!0}),n.set&&r(n.set,e,{setter:!0}),i.f(t,e,n)}},edfa:function(t,e){function n(t){var e=-1,n=Array(t.size);return t.forEach((function(t,r){n[++e]=[r,t]})),n}t.exports=n},ee45:function(t,e,n){var r=n("a78b"),i=n("1304");function o(t,e,n){return null==t?t:r(t,e,i(n))}t.exports=o},eec4:function(t,e,n){"use strict";var r=n("b770").forEach,i=n("5be5"),o=n("49ad"),a=n("2cef"),s=n("5058"),l=n("abb4"),u=n("18e9"),c=n("c274"),d=n("d6eb"),h=n("18d2"),f=n("c946");function p(t){return Array.isArray(t)||void 0!==t.length}function m(t){if(Array.isArray(t))return t;var e=[];return r(t,(function(t){e.push(t)})),e}function g(t){return t&&1===t.nodeType}function y(t,e,n){var r=t[e];return void 0!==r&&null!==r||void 0===n?r:n}t.exports=function(t){var e;if(t=t||{},t.idHandler)e={get:function(e){return t.idHandler.get(e,!0)},set:t.idHandler.set};else{var n=a(),v=s({idGenerator:n,stateHandler:d});e=v}var _=t.reporter;if(!_){var b=!1===_;_=l(b)}var x=y(t,"batchProcessor",c({reporter:_})),w={};w.callOnAdd=!!y(t,"callOnAdd",!0),w.debug=!!y(t,"debug",!1);var M,S=o(e),k=i({stateHandler:d}),L=y(t,"strategy","object"),T=y(t,"important",!1),D={reporter:_,batchProcessor:x,stateHandler:d,idHandler:e,important:T};if("scroll"===L&&(u.isLegacyOpera()?(_.warn("Scroll strategy is not supported on legacy Opera. Changing to object strategy."),L="object"):u.isIE(9)&&(_.warn("Scroll strategy is not supported on IE9. Changing to object strategy."),L="object")),"scroll"===L)M=f(D);else{if("object"!==L)throw new Error("Invalid strategy name: "+L);M=h(D)}var O={};function I(t,n,i){function o(t){var e=S.get(t);r(e,(function(e){e(t)}))}function a(t,e,n){S.add(e,n),t&&n(e)}if(i||(i=n,n=t,t={}),!n)throw new Error("At least one element required.");if(!i)throw new Error("Listener required.");if(g(n))n=[n];else{if(!p(n))return _.error("Invalid arguments. Must be a DOM element or a collection of DOM elements.");n=m(n)}var s=0,l=y(t,"callOnAdd",w.callOnAdd),u=y(t,"onReady",(function(){})),c=y(t,"debug",w.debug);r(n,(function(t){d.getState(t)||(d.initState(t),e.set(t));var h=e.get(t);if(c&&_.log("Attaching listener to element",h,t),!k.isDetectable(t))return c&&_.log(h,"Not detectable."),k.isBusy(t)?(c&&_.log(h,"System busy making it detectable"),a(l,t,i),O[h]=O[h]||[],void O[h].push((function(){s++,s===n.length&&u()}))):(c&&_.log(h,"Making detectable..."),k.markBusy(t,!0),M.makeDetectable({debug:c,important:T},t,(function(t){if(c&&_.log(h,"onElementDetectable"),d.getState(t)){k.markAsDetectable(t),k.markBusy(t,!1),M.addListener(t,o),a(l,t,i);var e=d.getState(t);if(e&&e.startSize){var f=t.offsetWidth,p=t.offsetHeight;e.startSize.width===f&&e.startSize.height===p||o(t)}O[h]&&r(O[h],(function(t){t()}))}else c&&_.log(h,"Element uninstalled before being detectable.");delete O[h],s++,s===n.length&&u()})));c&&_.log(h,"Already detecable, adding listener."),a(l,t,i),s++})),s===n.length&&u()}function C(t){if(!t)return _.error("At least one element is required.");if(g(t))t=[t];else{if(!p(t))return _.error("Invalid arguments. Must be a DOM element or a collection of DOM elements.");t=m(t)}r(t,(function(t){S.removeAllListeners(t),M.uninstall(t),d.cleanState(t)}))}function A(t){M.initDocument&&M.initDocument(t)}return{listenTo:I,removeListener:S.removeListener,removeAllListeners:S.removeAllListeners,uninstall:C,initDocument:A}}},ef5d:function(t,e){function n(t){return function(e){return null==e?void 0:e[t]}}t.exports=n},efb6:function(t,e,n){var r=n("5e2e");function i(){this.__data__=new r,this.size=0}t.exports=i},efc9:function(t,e,n){"use strict";n("99af"),n("fb6a"),n("b0c0");var r=function(){var t=this,e=t._self._c;return e("label",{class:t.wrapClasses},[e("span",{class:t.checkboxClasses},[e("span",{class:t.innerClasses}),t.group?e("input",{directives:[{name:"model",rawName:"v-model",value:t.model,expression:"model"}],class:t.inputClasses,attrs:{type:"checkbox",disabled:t.itemDisabled,name:t.name},domProps:{value:t.label,checked:Array.isArray(t.model)?t._i(t.model,t.label)>-1:t.model},on:{change:[function(e){var n=t.model,r=e.target,i=!!r.checked;if(Array.isArray(n)){var o=t.label,a=t._i(n,o);r.checked?a<0&&(t.model=n.concat([o])):a>-1&&(t.model=n.slice(0,a).concat(n.slice(a+1)))}else t.model=i},t.change],focus:t.onFocus,blur:t.onBlur}}):e("input",{class:t.inputClasses,attrs:{type:"checkbox",disabled:t.itemDisabled,name:t.name},domProps:{checked:t.currentValue},on:{change:t.change,focus:t.onFocus,blur:t.onBlur}})]),t._t("default",(function(){return[t.showSlot?e("span",[t._v(t._s(t.label))]):t._e()]}))],2)},i=[],o=n("ade3"),a=(n("a9e3"),n("3a27")),s=n("2507"),l=n("9f81"),u="ivu-checkbox",c={name:"Checkbox",mixins:[s["a"],l["a"]],props:{disabled:{type:Boolean,default:!1},value:{type:[String,Number,Boolean],default:!1},trueValue:{type:[String,Number,Boolean],default:!0},falseValue:{type:[String,Number,Boolean],default:!1},label:{type:[String,Number,Boolean]},indeterminate:{type:Boolean,default:!1},size:{validator:function(t){return Object(a["j"])(t,["small","large","default"])},default:function(){return this.$IVIEW&&""!==this.$IVIEW.size?this.$IVIEW.size:"default"}},name:{type:String},border:{type:Boolean,default:!1}},data:function(){return{model:[],currentValue:this.value,group:!1,showSlot:!0,parent:Object(a["f"])(this,"CheckboxGroup"),focusInner:!1}},computed:{wrapClasses:function(){return["".concat(u,"-wrapper"),Object(o["a"])(Object(o["a"])(Object(o["a"])(Object(o["a"])(Object(o["a"])({},"".concat(u,"-group-item"),this.group),"".concat(u,"-wrapper-checked"),this.currentValue),"".concat(u,"-wrapper-disabled"),this.itemDisabled),"".concat(u,"-").concat(this.size),!!this.size),"".concat(u,"-border"),this.border)]},checkboxClasses:function(){return["".concat(u),Object(o["a"])(Object(o["a"])(Object(o["a"])({},"".concat(u,"-checked"),this.currentValue),"".concat(u,"-disabled"),this.itemDisabled),"".concat(u,"-indeterminate"),this.indeterminate)]},innerClasses:function(){return["".concat(u,"-inner"),Object(o["a"])({},"".concat(u,"-focus"),this.focusInner)]},inputClasses:function(){return"".concat(u,"-input")}},mounted:function(){this.parent=Object(a["f"])(this,"CheckboxGroup"),this.parent&&(this.group=!0),this.group?this.parent.updateModel(!0):(this.updateModel(),this.showSlot=void 0!==this.$slots.default)},methods:{change:function(t){if(this.itemDisabled)return!1;var e=t.target.checked;this.currentValue=e;var n=e?this.trueValue:this.falseValue;this.$emit("input",n),this.group?this.parent.change(this.model):(this.$emit("on-change",n),this.dispatch("FormItem","on-form-change",n))},updateModel:function(){this.currentValue=this.value===this.trueValue},onBlur:function(){this.focusInner=!1},onFocus:function(){this.focusInner=!0}},watch:{value:function(t){if(t!==this.trueValue&&t!==this.falseValue)throw"Value should be trueValue or falseValue.";this.updateModel()}}},d=c,h=n("2877"),f=Object(h["a"])(d,r,i,!1,null,null,null),p=f.exports,m=function(){var t=this,e=t._self._c;return e("div",{class:t.classes},[t._t("default")],2)},g=[],y=(n("d3b7"),n("159b"),"ivu-checkbox-group"),v={name:"CheckboxGroup",mixins:[s["a"]],props:{value:{type:Array,default:function(){return[]}},size:{validator:function(t){return Object(a["j"])(t,["small","large","default"])},default:function(){return this.$IVIEW&&""!==this.$IVIEW.size?this.$IVIEW.size:"default"}}},data:function(){return{currentValue:this.value||[],children:[]}},computed:{classes:function(){return["".concat(y),Object(o["a"])({},"ivu-checkbox-".concat(this.size),!!this.size)]}},mounted:function(){this.updateModel(!0)},methods:{updateModel:function(t){if(this.children=Object(a["g"])(this,"Checkbox"),this.children){var e=this.value||[];this.children.forEach((function(n){n.model=e,t&&(n.currentValue=e.indexOf(n.label)>=0,n.group=!0)}))}},change:function(t){this.currentValue=t,this.$emit("input",t),this.$emit("on-change",t),this.dispatch("FormItem","on-form-change",t)}},watch:{value:function(){this.updateModel(!0)}}},_=v,b=Object(h["a"])(_,m,g,!1,null,null,null),x=b.exports;p.Group=x;e["a"]=p},efec:function(t,e,n){"use strict";var r=n("1a2d"),i=n("cb2d"),o=n("51eb"),a=n("b622"),s=a("toPrimitive"),l=Date.prototype;r(l,s)||i(l,s,o)},f00c:function(t,e,n){"use strict";var r=n("23e7"),i=n("e285");r({target:"Number",stat:!0},{isFinite:i})},f069:function(t,e,n){"use strict";var r=n("59ed"),i=TypeError,o=function(t){var e,n;this.promise=new t((function(t,r){if(void 0!==e||void 0!==n)throw new i("Bad Promise constructor");e=t,n=r})),this.resolve=r(e),this.reject=r(n)};t.exports.f=function(t){return new o(t)}},f22b:function(t,e,n){"use strict";var r=n("23e7"),i=n("f069"),o=n("4738").CONSTRUCTOR;r({target:"Promise",stat:!0,forced:o},{reject:function(t){var e=i.f(this),n=e.reject;return n(t),e.promise}})},f260:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("pt",{months:"janeiro_fevereiro_março_abril_maio_junho_julho_agosto_setembro_outubro_novembro_dezembro".split("_"),monthsShort:"jan_fev_mar_abr_mai_jun_jul_ago_set_out_nov_dez".split("_"),weekdays:"Domingo_Segunda-feira_Terça-feira_Quarta-feira_Quinta-feira_Sexta-feira_Sábado".split("_"),weekdaysShort:"Dom_Seg_Ter_Qua_Qui_Sex_Sáb".split("_"),weekdaysMin:"Do_2ª_3ª_4ª_5ª_6ª_Sá".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D [de] MMMM [de] YYYY",LLL:"D [de] MMMM [de] YYYY HH:mm",LLLL:"dddd, D [de] MMMM [de] YYYY HH:mm"},calendar:{sameDay:"[Hoje às] LT",nextDay:"[Amanhã às] LT",nextWeek:"dddd [às] LT",lastDay:"[Ontem às] LT",lastWeek:function(){return 0===this.day()||6===this.day()?"[Último] dddd [às] LT":"[Última] dddd [às] LT"},sameElse:"L"},relativeTime:{future:"em %s",past:"há %s",s:"segundos",ss:"%d segundos",m:"um minuto",mm:"%d minutos",h:"uma hora",hh:"%d horas",d:"um dia",dd:"%d dias",w:"uma semana",ww:"%d semanas",M:"um mês",MM:"%d meses",y:"um ano",yy:"%d anos"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:1,doy:4}});return e}))},f36a:function(t,e,n){"use strict";var r=n("e330");t.exports=r([].slice)},f3c1:function(t,e){var n=800,r=16,i=Date.now;function o(t){var e=0,o=0;return function(){var a=i(),s=r-(a-o);if(o=a,s>0){if(++e>=n)return arguments[0]}else e=0;return t.apply(void 0,arguments)}}t.exports=o},f3ff:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"੧",2:"੨",3:"੩",4:"੪",5:"੫",6:"੬",7:"੭",8:"੮",9:"੯",0:"੦"},n={"੧":"1","੨":"2","੩":"3","੪":"4","੫":"5","੬":"6","੭":"7","੮":"8","੯":"9","੦":"0"},r=t.defineLocale("pa-in",{months:"ਜਨਵਰੀ_ਫ਼ਰਵਰੀ_ਮਾਰਚ_ਅਪ੍ਰੈਲ_ਮਈ_ਜੂਨ_ਜੁਲਾਈ_ਅਗਸਤ_ਸਤੰਬਰ_ਅਕਤੂਬਰ_ਨਵੰਬਰ_ਦਸੰਬਰ".split("_"),monthsShort:"ਜਨਵਰੀ_ਫ਼ਰਵਰੀ_ਮਾਰਚ_ਅਪ੍ਰੈਲ_ਮਈ_ਜੂਨ_ਜੁਲਾਈ_ਅਗਸਤ_ਸਤੰਬਰ_ਅਕਤੂਬਰ_ਨਵੰਬਰ_ਦਸੰਬਰ".split("_"),weekdays:"ਐਤਵਾਰ_ਸੋਮਵਾਰ_ਮੰਗਲਵਾਰ_ਬੁਧਵਾਰ_ਵੀਰਵਾਰ_ਸ਼ੁੱਕਰਵਾਰ_ਸ਼ਨੀਚਰਵਾਰ".split("_"),weekdaysShort:"ਐਤ_ਸੋਮ_ਮੰਗਲ_ਬੁਧ_ਵੀਰ_ਸ਼ੁਕਰ_ਸ਼ਨੀ".split("_"),weekdaysMin:"ਐਤ_ਸੋਮ_ਮੰਗਲ_ਬੁਧ_ਵੀਰ_ਸ਼ੁਕਰ_ਸ਼ਨੀ".split("_"),longDateFormat:{LT:"A h:mm ਵਜੇ",LTS:"A h:mm:ss ਵਜੇ",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm ਵਜੇ",LLLL:"dddd, D MMMM YYYY, A h:mm ਵਜੇ"},calendar:{sameDay:"[ਅਜ] LT",nextDay:"[ਕਲ] LT",nextWeek:"[ਅਗਲਾ] dddd, LT",lastDay:"[ਕਲ] LT",lastWeek:"[ਪਿਛਲੇ] dddd, LT",sameElse:"L"},relativeTime:{future:"%s ਵਿੱਚ",past:"%s ਪਿਛਲੇ",s:"ਕੁਝ ਸਕਿੰਟ",ss:"%d ਸਕਿੰਟ",m:"ਇਕ ਮਿੰਟ",mm:"%d ਮਿੰਟ",h:"ਇੱਕ ਘੰਟਾ",hh:"%d ਘੰਟੇ",d:"ਇੱਕ ਦਿਨ",dd:"%d ਦਿਨ",M:"ਇੱਕ ਮਹੀਨਾ",MM:"%d ਮਹੀਨੇ",y:"ਇੱਕ ਸਾਲ",yy:"%d ਸਾਲ"},preparse:function(t){return t.replace(/[੧੨੩੪੫੬੭੮੯੦]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/ਰਾਤ|ਸਵੇਰ|ਦੁਪਹਿਰ|ਸ਼ਾਮ/,meridiemHour:function(t,e){return 12===t&&(t=0),"ਰਾਤ"===e?t<4?t:t+12:"ਸਵੇਰ"===e?t:"ਦੁਪਹਿਰ"===e?t>=10?t:t+12:"ਸ਼ਾਮ"===e?t+12:void 0},meridiem:function(t,e,n){return t<4?"ਰਾਤ":t<10?"ਸਵੇਰ":t<17?"ਦੁਪਹਿਰ":t<20?"ਸ਼ਾਮ":"ਰਾਤ"},week:{dow:0,doy:6}});return r}))},f4d6:function(t,e,n){var r=n("ffd6"),i=1/0;function o(t){if("string"==typeof t||r(t))return t;var e=t+"";return"0"==e&&1/t==-i?"-0":e}t.exports=o},f500:function(t,e,n){"use strict";var r=function(){var t=this,e=t._self._c;return e("div",{class:t.classes},[t._t("default")],2)},i=[],o=n("ade3"),a=(n("99af"),n("3a27")),s="ivu-btn-group",l={name:"ButtonGroup",props:{size:{validator:function(t){return Object(a["j"])(t,["small","large","default"])},default:function(){return this.$IVIEW&&""!==this.$IVIEW.size?this.$IVIEW.size:"default"}},shape:{validator:function(t){return Object(a["j"])(t,["circle","circle-outline"])}},vertical:{type:Boolean,default:!1}},computed:{classes:function(){return["".concat(s),Object(o["a"])(Object(o["a"])(Object(o["a"])({},"".concat(s,"-").concat(this.size),!!this.size),"".concat(s,"-").concat(this.shape),!!this.shape),"".concat(s,"-vertical"),this.vertical)]}}},u=l,c=n("2877"),d=Object(c["a"])(u,r,i,!1,null,null,null);e["a"]=d.exports},f542:function(t,e,n){var r=n("ec47"),i=n("ec69"),o=r(i);t.exports=o},f5df:function(t,e,n){"use strict";var r=n("00ee"),i=n("1626"),o=n("c6b6"),a=n("b622"),s=a("toStringTag"),l=Object,u="Arguments"===o(function(){return arguments}()),c=function(t,e){try{return t[e]}catch(n){}};t.exports=r?o:function(t){var e,n,r;return void 0===t?"Undefined":null===t?"Null":"string"==typeof(n=c(e=l(t),s))?n:u?o(e):"Object"===(r=o(e))&&i(e.callee)?"Arguments":r}},f608:function(t,e,n){var r=n("6747"),i=n("ffd6"),o=/\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/,a=/^\w*$/;function s(t,e){if(r(t))return!1;var n=typeof t;return!("number"!=n&&"symbol"!=n&&"boolean"!=n&&null!=t&&!i(t))||(a.test(t)||!o.test(t)||null!=e&&t in Object(e))}t.exports=s},f6b4:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=["Am Faoilleach","An Gearran","Am Màrt","An Giblean","An Cèitean","An t-Ògmhios","An t-Iuchar","An Lùnastal","An t-Sultain","An Dàmhair","An t-Samhain","An Dùbhlachd"],n=["Faoi","Gear","Màrt","Gibl","Cèit","Ògmh","Iuch","Lùn","Sult","Dàmh","Samh","Dùbh"],r=["Didòmhnaich","Diluain","Dimàirt","Diciadain","Diardaoin","Dihaoine","Disathairne"],i=["Did","Dil","Dim","Dic","Dia","Dih","Dis"],o=["Dò","Lu","Mà","Ci","Ar","Ha","Sa"],a=t.defineLocale("gd",{months:e,monthsShort:n,monthsParseExact:!0,weekdays:r,weekdaysShort:i,weekdaysMin:o,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[An-diugh aig] LT",nextDay:"[A-màireach aig] LT",nextWeek:"dddd [aig] LT",lastDay:"[An-dè aig] LT",lastWeek:"dddd [seo chaidh] [aig] LT",sameElse:"L"},relativeTime:{future:"ann an %s",past:"bho chionn %s",s:"beagan diogan",ss:"%d diogan",m:"mionaid",mm:"%d mionaidean",h:"uair",hh:"%d uairean",d:"latha",dd:"%d latha",M:"mìos",MM:"%d mìosan",y:"bliadhna",yy:"%d bliadhna"},dayOfMonthOrdinalParse:/\d{1,2}(d|na|mh)/,ordinal:function(t){var e=1===t?"d":t%10===2?"na":"mh";return t+e},week:{dow:1,doy:4}});return a}))},f772:function(t,e,n){"use strict";var r=n("5692"),i=n("90e3"),o=r("keys");t.exports=function(t){return o[t]||(o[t]=i(t))}},f843:function(t,e,n){"use strict";var r=function(){var t=this,e=t._self._c;return e("div",{ref:"tabsWrap",class:t.classes},[e("div",{class:[t.prefixCls+"-bar"]},[t.showSlot?e("div",{class:[t.prefixCls+"-nav-right"]},[t._t("extra")],2):t._e(),e("div",{ref:"navContainer",class:[t.prefixCls+"-nav-container"],attrs:{tabindex:"0"},on:{keydown:[t.handleTabKeyNavigation,function(e){return!e.type.indexOf("key")&&t._k(e.keyCode,"space",32,e.key,[" ","Spacebar"])?null:(e.preventDefault(),t.handleTabKeyboardSelect(!1))}]}},[e("div",{ref:"navWrap",class:[t.prefixCls+"-nav-wrap",t.scrollable?t.prefixCls+"-nav-scrollable":""]},[e("span",{class:[t.prefixCls+"-nav-prev",t.scrollable?"":t.prefixCls+"-nav-scroll-disabled"],on:{click:t.scrollPrev}},[e("Icon",{attrs:{type:"ios-arrow-back"}})],1),e("span",{class:[t.prefixCls+"-nav-next",t.scrollable?"":t.prefixCls+"-nav-scroll-disabled"],on:{click:t.scrollNext}},[e("Icon",{attrs:{type:"ios-arrow-forward"}})],1),e("div",{ref:"navScroll",class:[t.prefixCls+"-nav-scroll"],on:{DOMMouseScroll:t.handleScroll,mousewheel:t.handleScroll}},[e("div",{ref:"nav",class:[t.prefixCls+"-nav"],style:t.navStyle},[e("div",{class:t.barClasses,style:t.barStyle}),t._l(t.navList,(function(n,r){return e("div",{class:t.tabCls(n),attrs:{draggable:t.draggable},on:{click:function(e){return t.handleChange(r)},dblclick:function(e){return t.handleDblclick(r)},contextmenu:function(e){return e.stopPropagation(),t.handleContextmenu(r,e)},selectstart:function(e){return e.stopPropagation(),t.handlePreventSelect(r,e)},dragstart:function(e){return t.handleDrag(r,e)},drop:function(e){return t.handleDrop(r,e)},dragover:function(t){t.preventDefault()}}},[""!==n.icon?e("Icon",{attrs:{type:n.icon}}):t._e(),"function"===n.labelType?e("Render",{attrs:{render:n.label}}):[t._v(t._s(n.label))],t.showClose(n)?e("Icon",{class:[t.prefixCls+"-close"],attrs:{type:t.arrowType,custom:t.customArrowType,size:t.arrowSize},nativeOn:{click:function(e){return e.stopPropagation(),t.handleRemove(r)}}}):t._e()],2)}))],2)])])])]),e("div",{ref:"panes",class:t.contentClasses,style:t.contentStyle},[t._t("default")],2),e("div",{staticClass:"ivu-tabs-context-menu",style:t.contextMenuStyles},[e("Dropdown",{attrs:{trigger:"custom",visible:t.contextMenuVisible,transfer:""},on:{"on-clickoutside":t.handleClickContextMenuOutside}},[e("DropdownMenu",{attrs:{slot:"list"},slot:"list"},[t._t("contextMenu")],2)],1)],1)])},i=[],o=n("6b75");function a(t){if(Array.isArray(t))return Object(o["a"])(t)}n("a4d3"),n("e01a"),n("d28b"),n("a630"),n("d3b7"),n("3ca3"),n("ddb0");function s(t){if("undefined"!==typeof Symbol&&null!=t[Symbol.iterator]||null!=t["@@iterator"])return Array.from(t)}var l=n("06c5");n("d9e2");function u(){throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.")}function c(t){return a(t)||s(t)||Object(l["a"])(t)||u()}var d=n("53ca"),h=n("ade3"),f=n("b85c"),p=(n("99af"),n("4de4"),n("c740"),n("d81d"),n("14d9"),n("4e82"),n("a434"),n("b0c0"),n("a9e3"),n("ac1f"),n("466d"),n("159b"),n("099a")),m=n("1766"),g=n("6e29"),y=n("31de"),v=n("3a27"),_=n("2507"),b=n("eec4"),x=n.n(b),w="ivu-tabs",M=300,S=function t(e,n,r,i){var o=e.findIndex((function(t){return t.name===n})),a=(o+r+e.length)%e.length,s=e[a];return s.disabled?t(e,s.name,r,i):s},k=function t(e,n){try{e.focus()}catch(s){}if(document.activeElement==e&&e!==n)return!0;var r,i=e.children,o=Object(f["a"])(i);try{for(o.s();!(r=o.n()).done;){var a=r.value;if(t(a,n))return!0}}catch(s){o.e(s)}finally{o.f()}return!1},L={name:"Tabs",mixins:[_["a"]],components:{Icon:p["a"],Render:m["a"],Dropdown:g["a"],DropdownMenu:y["a"]},provide:function(){return{TabsInstance:this}},props:{value:{type:[String,Number]},type:{validator:function(t){return Object(v["j"])(t,["line","card"])},default:"line"},size:{validator:function(t){return Object(v["j"])(t,["small","default"])},default:"default"},animated:{type:Boolean,default:!0},captureFocus:{type:Boolean,default:!1},closable:{type:Boolean,default:!1},beforeRemove:Function,name:{type:String},draggable:{type:Boolean,default:!1}},data:function(){return{prefixCls:w,navList:[],barWidth:0,barOffset:0,activeKey:this.value,focusedKey:this.value,showSlot:!1,navStyle:{transform:""},scrollable:!1,transitioning:!1,contextMenuVisible:!1,contextMenuStyles:{top:0,left:0}}},computed:{classes:function(){return["".concat(w),Object(h["a"])(Object(h["a"])(Object(h["a"])({},"".concat(w,"-card"),"card"===this.type),"".concat(w,"-mini"),"small"===this.size&&"line"===this.type),"".concat(w,"-no-animation"),!this.animated)]},contentClasses:function(){return["".concat(w,"-content"),Object(h["a"])({},"".concat(w,"-content-animated"),this.animated)]},barClasses:function(){return["".concat(w,"-ink-bar"),Object(h["a"])({},"".concat(w,"-ink-bar-animated"),this.animated)]},contentStyle:function(){var t=this.getTabIndex(this.activeKey),e=0===t?"0%":"-".concat(t,"00%"),n={};return t>-1&&(n={transform:"translateX(".concat(e,") translateZ(0px)")}),n},barStyle:function(){var t={visibility:"hidden",width:"".concat(this.barWidth,"px")};return"line"===this.type&&(t.visibility="visible"),this.animated?t.transform="translate3d(".concat(this.barOffset,"px, 0px, 0px)"):t.left="".concat(this.barOffset,"px"),t},arrowType:function(){var t="ios-close";return this.$IVIEW&&(this.$IVIEW.tabs.customCloseIcon?t="":this.$IVIEW.tabs.closeIcon&&(t=this.$IVIEW.tabs.closeIcon)),t},customArrowType:function(){var t="";return this.$IVIEW&&this.$IVIEW.tabs.customCloseIcon&&(t=this.$IVIEW.tabs.customCloseIcon),t},arrowSize:function(){var t="";return this.$IVIEW&&this.$IVIEW.tabs.closeIconSize&&(t=this.$IVIEW.tabs.closeIconSize),t}},methods:{getTabs:function(){var t=this,e=Object(v["g"])(this,"TabPane"),n=[];return e.forEach((function(e){e.tab&&t.name?e.tab===t.name&&n.push(e):n.push(e)})),n.sort((function(t,e){if(t.index&&e.index)return t.index>e.index?1:-1})),n},updateNav:function(){var t=this;this.navList=[],this.getTabs().forEach((function(e,n){t.navList.push({labelType:Object(d["a"])(e.label),label:e.label,icon:e.icon||"",name:e.currentName||n,disabled:e.disabled,closable:e.closable,contextMenu:e.contextMenu}),e.currentName||(e.currentName=n),0===n&&(t.activeKey||(t.activeKey=e.currentName||n))})),this.updateStatus(),this.updateBar()},updateBar:function(){var t=this;this.$nextTick((function(){var e=t.getTabIndex(t.activeKey);if(t.$refs.nav){var n=t.$refs.nav.querySelectorAll(".".concat(w,"-tab")),r=n[e];if(t.barWidth=r?parseFloat(r.offsetWidth):0,e>0){for(var i=0,o="small"===t.size?0:16,a=0;a<e;a++)i+=parseFloat(n[a].offsetWidth)+o;t.barOffset=i}else t.barOffset=0;t.updateNavScroll()}}))},updateStatus:function(){var t=this,e=this.getTabs();e.forEach((function(e){return e.show=e.currentName===t.activeKey||t.animated}))},tabCls:function(t){return["".concat(w,"-tab"),Object(h["a"])(Object(h["a"])(Object(h["a"])({},"".concat(w,"-tab-disabled"),t.disabled),"".concat(w,"-tab-active"),t.name===this.activeKey),"".concat(w,"-tab-focused"),t.name===this.focusedKey)]},handleChange:function(t){var e=this;if(!this.transitioning){this.transitioning=!0,setTimeout((function(){return e.transitioning=!1}),M);var n=this.navList[t];n&&!n.disabled&&(this.activeKey=n.name,this.$emit("input",n.name),this.$emit("on-click",n.name))}},handleDblclick:function(t){var e=this.navList[t];e&&!e.disabled&&this.$emit("on-dblclick",e.name)},handleContextmenu:function(t,e){var n=this;this.contextMenuVisible&&this.handleClickContextMenuOutside(),this.$nextTick((function(){var r=n.navList[t];if(r&&!r.disabled&&r.contextMenu){e.preventDefault();var i=n.$refs.tabsWrap,o=i.getBoundingClientRect(),a={left:"".concat(e.clientX-o.left,"px"),top:"".concat(e.clientY-o.top,"px")};n.contextMenuStyles=a,n.contextMenuVisible=!0,n.$emit("on-contextmenu",r,e,a)}}))},handleClickContextMenuOutside:function(){this.contextMenuVisible=!1},handlePreventSelect:function(t,e){var n=this.navList[t];n&&!n.disabled&&n.contextMenu&&e.preventDefault()},handleTabKeyNavigation:function(t){if(37===t.keyCode||39===t.keyCode){var e=39===t.keyCode?1:-1,n=S(this.navList,this.focusedKey,e);this.focusedKey=n.name}},handleTabKeyboardSelect:function(){var t=arguments.length>0&&void 0!==arguments[0]&&arguments[0];if(!t){var e=this.focusedKey||0,n=this.getTabIndex(e);this.handleChange(n)}},handleRemove:function(t){var e=this;if(!this.beforeRemove)return this.handleRemoveTab(t);var n=this.beforeRemove(t);n&&n.then?n.then((function(){e.handleRemoveTab(t)})):this.handleRemoveTab(t)},handleRemoveTab:function(t){var e=this.getTabs(),n=e[t];if(n.$destroy(),n.currentName===this.activeKey){var r=this.getTabs(),i=-1;if(r.length){var o=e.filter((function(e,n){return!e.disabled&&n<t})),a=e.filter((function(e,n){return!e.disabled&&n>t}));i=a.length?a[0].currentName:o.length?o[o.length-1].currentName:r[0].currentName}this.activeKey=i,this.$emit("input",i)}this.$emit("on-tab-remove",n.currentName),this.updateNav()},showClose:function(t){return"card"===this.type&&(null!==t.closable?t.closable:this.closable)},scrollPrev:function(){var t=this.$refs.navScroll.offsetWidth,e=this.getCurrentScrollOffset();if(e){var n=e>t?e-t:0;this.setOffset(n)}},scrollNext:function(){var t=this.$refs.nav.offsetWidth,e=this.$refs.navScroll.offsetWidth,n=this.getCurrentScrollOffset();if(!(t-n<=e)){var r=t-n>2*e?n+e:t-e;this.setOffset(r)}},getCurrentScrollOffset:function(){var t=this.navStyle;return t.transform?Number(t.transform.match(/translateX\(-(\d+(\.\d+)*)px\)/)[1]):0},getTabIndex:function(t){return this.navList.findIndex((function(e){return e.name===t}))},setOffset:function(t){this.navStyle.transform="translateX(-".concat(t,"px)")},scrollToActiveTab:function(){if(this.scrollable){var t=this.$refs.nav,e=this.$el.querySelector(".".concat(w,"-tab-active"));if(e){var n=this.$refs.navScroll,r=e.getBoundingClientRect(),i=n.getBoundingClientRect(),o=t.getBoundingClientRect(),a=this.getCurrentScrollOffset(),s=a;o.right<i.right&&(s=t.offsetWidth-i.width),r.left<i.left?s=a-(i.left-r.left):r.right>i.right&&(s=a+r.right-i.right),a!==s&&this.setOffset(Math.max(s,0))}}},updateNavScroll:function(){var t=this.$refs.nav.offsetWidth,e=this.$refs.navScroll.offsetWidth,n=this.getCurrentScrollOffset();e<t?(this.scrollable=!0,t-n<e&&this.setOffset(t-e)):(this.scrollable=!1,n>0&&this.setOffset(0))},handleScroll:function(t){t.preventDefault(),t.stopPropagation();var e=t.type,n=0;"DOMMouseScroll"!==e&&"mousewheel"!==e||(n=t.wheelDelta?t.wheelDelta:40*-(t.detail||0)),n>0?this.scrollPrev():this.scrollNext()},handleResize:function(){this.updateNavScroll()},isInsideHiddenElement:function(){var t=this.$el.parentNode;while(t&&t!==document.body){if(t.style&&"none"===t.style.display)return t;t=t.parentNode}return!1},updateVisibility:function(t){var e=this;c(this.$refs.panes.querySelectorAll(".".concat(w,"-tabpane"))).forEach((function(n,r){t===r?(c(n.children).filter((function(t){return t.classList.contains("".concat(w,"-tabpane"))})).forEach((function(t){return t.style.visibility="visible"})),e.captureFocus&&setTimeout((function(){return k(n,n)}),M)):setTimeout((function(){c(n.children).filter((function(t){return t.classList.contains("".concat(w,"-tabpane"))})).forEach((function(t){return t.style.visibility="hidden"}))}),M)}))},handleDrag:function(t,e){var n=this.navList[t];n&&e.dataTransfer.setData("tab-name",n.name)},handleDrop:function(t,e){var n=this.navList[t];if(n){var r=e.dataTransfer.getData("tab-name");e.preventDefault();var i=this.navList.map((function(t){return t.name})),o=parseInt(i.findIndex((function(t){return t===r}))),a=parseInt(i.findIndex((function(t){return t===n.name})));i.splice.apply(i,[a,1].concat(c(i.splice(o,1,i[a])))),this.$emit("on-drag-drop",r,n.name,o,a,i)}}},watch:{value:function(t){this.activeKey=t,this.focusedKey=t},activeKey:function(t){var e=this;this.focusedKey=t,this.updateBar(),this.updateStatus(),this.broadcast("Table","on-visible-change",!0),this.$nextTick((function(){e.scrollToActiveTab()}));var n=Math.max(this.getTabIndex(this.focusedKey),0);this.updateVisibility(n)}},mounted:function(){var t=this;this.showSlot=void 0!==this.$slots.extra,this.observer=x()(),this.observer.listenTo(this.$refs.navWrap,this.handleResize);var e=this.isInsideHiddenElement();e&&(this.mutationObserver=new v["a"]((function(){"none"!==e.style.display&&(t.updateBar(),t.mutationObserver.disconnect())})),this.mutationObserver.observe(e,{attributes:!0,childList:!0,characterData:!0,attributeFilter:["style"]})),this.handleTabKeyboardSelect(!0),this.updateVisibility(this.getTabIndex(this.activeKey))},beforeDestroy:function(){this.observer.removeListener(this.$refs.navWrap,this.handleResize),this.mutationObserver&&this.mutationObserver.disconnect()}},T=L,D=n("2877"),O=Object(D["a"])(T,r,i,!1,null,null,null),I=O.exports,C=n("5334");I.Pane=C["a"];e["a"]=I},f8af:function(t,e,n){var r=n("2474");function i(t){var e=new t.constructor(t.byteLength);return new r(e).set(new r(t)),e}t.exports=i},f909:function(t,e,n){var r=n("7e64"),i=n("b760"),o=n("72af"),a=n("4f50"),s=n("1a8c"),l=n("9934"),u=n("8adb");function c(t,e,n,d,h){t!==e&&o(e,(function(o,l){if(h||(h=new r),s(o))a(t,e,l,n,c,d,h);else{var f=d?d(u(t,l),o,l+"",t,e,h):void 0;void 0===f&&(f=o),i(t,l,f)}}),l)}t.exports=c},f9ce:function(t,e,n){var r=n("ef5d"),i=n("e3f8"),o=n("f608"),a=n("f4d6");function s(t){return o(t)?r(a(t)):i(t)}t.exports=s},fa21:function(t,e,n){var r=n("7530"),i=n("2dcb"),o=n("eac5");function a(t){return"function"!=typeof t.constructor||o(t)?{}:r(i(t))}t.exports=a},facd:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="jan._feb._mrt._apr._mei_jun._jul._aug._sep._okt._nov._dec.".split("_"),n="jan_feb_mrt_apr_mei_jun_jul_aug_sep_okt_nov_dec".split("_"),r=[/^jan/i,/^feb/i,/^(maart|mrt\.?)$/i,/^apr/i,/^mei$/i,/^jun[i.]?$/i,/^jul[i.]?$/i,/^aug/i,/^sep/i,/^okt/i,/^nov/i,/^dec/i],i=/^(januari|februari|maart|april|mei|ju[nl]i|augustus|september|oktober|november|december|jan\.?|feb\.?|mrt\.?|apr\.?|ju[nl]\.?|aug\.?|sep\.?|okt\.?|nov\.?|dec\.?)/i,o=t.defineLocale("nl",{months:"januari_februari_maart_april_mei_juni_juli_augustus_september_oktober_november_december".split("_"),monthsShort:function(t,r){return t?/-MMM-/.test(r)?n[t.month()]:e[t.month()]:e},monthsRegex:i,monthsShortRegex:i,monthsStrictRegex:/^(januari|februari|maart|april|mei|ju[nl]i|augustus|september|oktober|november|december)/i,monthsShortStrictRegex:/^(jan\.?|feb\.?|mrt\.?|apr\.?|mei|ju[nl]\.?|aug\.?|sep\.?|okt\.?|nov\.?|dec\.?)/i,monthsParse:r,longMonthsParse:r,shortMonthsParse:r,weekdays:"zondag_maandag_dinsdag_woensdag_donderdag_vrijdag_zaterdag".split("_"),weekdaysShort:"zo._ma._di._wo._do._vr._za.".split("_"),weekdaysMin:"zo_ma_di_wo_do_vr_za".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD-MM-YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[vandaag om] LT",nextDay:"[morgen om] LT",nextWeek:"dddd [om] LT",lastDay:"[gisteren om] LT",lastWeek:"[afgelopen] dddd [om] LT",sameElse:"L"},relativeTime:{future:"over %s",past:"%s geleden",s:"een paar seconden",ss:"%d seconden",m:"één minuut",mm:"%d minuten",h:"één uur",hh:"%d uur",d:"één dag",dd:"%d dagen",w:"één week",ww:"%d weken",M:"één maand",MM:"%d maanden",y:"één jaar",yy:"%d jaar"},dayOfMonthOrdinalParse:/\d{1,2}(ste|de)/,ordinal:function(t){return t+(1===t||8===t||t>=20?"ste":"de")},week:{dow:1,doy:4}});return o}))},fb25:function(t,e,n){var r=n("afb9"),i=n("ec69");function o(t){return null==t?[]:r(t,i(t))}t.exports=o},fb6a:function(t,e,n){"use strict";var r=n("23e7"),i=n("e8b5"),o=n("68ee"),a=n("861d"),s=n("23cb"),l=n("07fa"),u=n("fc6a"),c=n("8418"),d=n("b622"),h=n("1dde"),f=n("f36a"),p=h("slice"),m=d("species"),g=Array,y=Math.max;r({target:"Array",proto:!0,forced:!p},{slice:function(t,e){var n,r,d,h=u(this),p=l(h),v=s(t,p),_=s(void 0===e?p:e,p);if(i(h)&&(n=h.constructor,o(n)&&(n===g||i(n.prototype))?n=void 0:a(n)&&(n=n[m],null===n&&(n=void 0)),n===g||void 0===n))return f(h,v,_);for(r=new(void 0===n?g:n)(y(_-v,0)),d=0;v<_;v++,d++)v in h&&c(r,d,h[v]);return r.length=d,r}})},fba5:function(t,e,n){var r=n("cb5a");function i(t){return r(this.__data__,t)>-1}t.exports=i},fc6a:function(t,e,n){"use strict";var r=n("44ad"),i=n("1d80");t.exports=function(t){return r(i(t))}},fce3:function(t,e,n){"use strict";var r=n("d039"),i=n("da84"),o=i.RegExp;t.exports=r((function(){var t=o(".","s");return!(t.dotAll&&t.test("\n")&&"s"===t.flags)}))},fd7e:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("x-pseudo",{months:"J~áñúá~rý_F~ébrú~árý_~Márc~h_Áp~ríl_~Máý_~Júñé~_Júl~ý_Áú~gúst~_Sép~témb~ér_Ó~ctób~ér_Ñ~óvém~bér_~Décé~mbér".split("_"),monthsShort:"J~áñ_~Féb_~Már_~Ápr_~Máý_~Júñ_~Júl_~Áúg_~Sép_~Óct_~Ñóv_~Déc".split("_"),monthsParseExact:!0,weekdays:"S~úñdá~ý_Mó~ñdáý~_Túé~sdáý~_Wéd~ñésd~áý_T~húrs~dáý_~Fríd~áý_S~átúr~dáý".split("_"),weekdaysShort:"S~úñ_~Móñ_~Túé_~Wéd_~Thú_~Frí_~Sát".split("_"),weekdaysMin:"S~ú_Mó~_Tú_~Wé_T~h_Fr~_Sá".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[T~ódá~ý át] LT",nextDay:"[T~ómó~rró~w át] LT",nextWeek:"dddd [át] LT",lastDay:"[Ý~ést~érdá~ý át] LT",lastWeek:"[L~ást] dddd [át] LT",sameElse:"L"},relativeTime:{future:"í~ñ %s",past:"%s á~gó",s:"á ~féw ~sécó~ñds",ss:"%d s~écóñ~ds",m:"á ~míñ~úté",mm:"%d m~íñú~tés",h:"á~ñ hó~úr",hh:"%d h~óúrs",d:"á ~dáý",dd:"%d d~áýs",M:"á ~móñ~th",MM:"%d m~óñt~hs",y:"á ~ýéár",yy:"%d ý~éárs"},dayOfMonthOrdinalParse:/\d{1,2}(th|st|nd|rd)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n},week:{dow:1,doy:4}});return e}))},fdbc:function(t,e,n){"use strict";t.exports={CSSRuleList:0,CSSStyleDeclaration:0,CSSValueList:0,ClientRectList:0,DOMRectList:0,DOMStringList:0,DOMTokenList:1,DataTransferItemList:0,FileList:0,HTMLAllCollection:0,HTMLCollection:0,HTMLFormElement:0,HTMLSelectElement:0,MediaList:0,MimeTypeArray:0,NamedNodeMap:0,NodeList:1,PaintRequestList:0,Plugin:0,PluginArray:0,SVGLengthList:0,SVGNumberList:0,SVGPathSegList:0,SVGPointList:0,SVGStringList:0,SVGTransformList:0,SourceBufferList:0,StyleSheetList:0,TextTrackCueList:0,TextTrackList:0,TouchList:0}},fdbf:function(t,e,n){"use strict";var r=n("04f8");t.exports=r&&!Symbol.sham&&"symbol"==typeof Symbol.iterator},fe66:function(t,e,n){"use strict";var r=n("2b0e"),i=function(){var t=this,e=t._self._c;return e("transition",{attrs:{name:"fade"}},[t.fullscreenVisible?e("div",{class:t.classes},[e("div",{class:t.mainClasses},[e("span",{class:t.dotClasses}),e("div",{class:t.textClasses},[t._t("default")],2)])]):t._e()])},o=[],a=n("ade3"),s=(n("99af"),n("3a27")),l=n("b2e6"),u="ivu-spin",c={name:"Spin",mixins:[l["a"]],props:{size:{validator:function(t){return Object(s["j"])(t,["small","large","default"])},default:function(){return this.$IVIEW&&""!==this.$IVIEW.size?this.$IVIEW.size:"default"}},fix:{type:Boolean,default:!1},fullscreen:{type:Boolean,default:!1}},data:function(){return{showText:!1,visible:!1}},computed:{classes:function(){return["".concat(u),Object(a["a"])(Object(a["a"])(Object(a["a"])(Object(a["a"])({},"".concat(u,"-").concat(this.size),!!this.size),"".concat(u,"-fix"),this.fix),"".concat(u,"-show-text"),this.showText),"".concat(u,"-fullscreen"),this.fullscreen)]},mainClasses:function(){return"".concat(u,"-main")},dotClasses:function(){return"".concat(u,"-dot")},textClasses:function(){return"".concat(u,"-text")},fullscreenVisible:function(){return!this.fullscreen||this.visible}},watch:{visible:function(t){t?this.addScrollEffect():this.removeScrollEffect()}},mounted:function(){this.showText=void 0!==this.$slots.default}},d=c,h=n("2877"),f=Object(h["a"])(d,i,o,!1,null,null,null),p=f.exports,m=n("c177");function g(){return Object(m["c"])(),m["d"]}let y=g();p.newInstance=t=>{const e=t||{},n=new r["a"]({data:Object.assign({},e,{}),render(t){let e="";return e=this.render?t(p,{props:{fix:!0,fullscreen:!0}},[this.render(t)]):t(p,{props:{size:"large",fix:!0,fullscreen:!0}}),t("div",{class:"ivu-spin-fullscreen ivu-spin-fullscreen-wrapper",style:{"z-index":2010+y}},[e])}}),i=n.$mount();document.body.appendChild(i.$el);const o=n.$children[0];return{show(){o.visible=!0,y=g()},remove(t){o.visible=!1,setTimeout((function(){o.$parent.$destroy(),void 0!==document.getElementsByClassName("ivu-spin-fullscreen")[0]&&document.body.removeChild(document.getElementsByClassName("ivu-spin-fullscreen")[0]),t()}),500)},component:o}};var v=p;let _;function b(t){return _=_||v.newInstance({render:t}),_}function x(t){const e="render"in t?t.render:void 0;let n=b(e);n.show(t)}v.show=function(t={}){return x(t)},v.hide=function(){if(!_)return!1;const t=b();t.remove(()=>{_=null})};e["a"]=v},ffd6:function(t,e,n){var r=n("3729"),i=n("1310"),o="[object Symbol]";function a(t){return"symbol"==typeof t||i(t)&&r(t)==o}t.exports=a},ffff:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("se",{months:"ođđajagemánnu_guovvamánnu_njukčamánnu_cuoŋománnu_miessemánnu_geassemánnu_suoidnemánnu_borgemánnu_čakčamánnu_golggotmánnu_skábmamánnu_juovlamánnu".split("_"),monthsShort:"ođđj_guov_njuk_cuo_mies_geas_suoi_borg_čakč_golg_skáb_juov".split("_"),weekdays:"sotnabeaivi_vuossárga_maŋŋebárga_gaskavahkku_duorastat_bearjadat_lávvardat".split("_"),weekdaysShort:"sotn_vuos_maŋ_gask_duor_bear_láv".split("_"),weekdaysMin:"s_v_m_g_d_b_L".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"MMMM D. [b.] YYYY",LLL:"MMMM D. [b.] YYYY [ti.] HH:mm",LLLL:"dddd, MMMM D. [b.] YYYY [ti.] HH:mm"},calendar:{sameDay:"[otne ti] LT",nextDay:"[ihttin ti] LT",nextWeek:"dddd [ti] LT",lastDay:"[ikte ti] LT",lastWeek:"[ovddit] dddd [ti] LT",sameElse:"L"},relativeTime:{future:"%s geažes",past:"maŋit %s",s:"moadde sekunddat",ss:"%d sekunddat",m:"okta minuhta",mm:"%d minuhtat",h:"okta diimmu",hh:"%d diimmut",d:"okta beaivi",dd:"%d beaivvit",M:"okta mánnu",MM:"%d mánut",y:"okta jahki",yy:"%d jagit"},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return e}))}}]);